{-# OPTIONS -fglasgow-exts #-}
------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Main where

{-

This module defines the main function of the Paradox program.

-}

import Flags
import ParseProblem
import SolveProblem
import Output
import Sat ( identify )

import Utils
  ( thd3
  )

import IO
  ( try
  )

-------------------------------------------------------------------------
-- main

main :: IO ()
main =
  do flags <- getFlags
     sat   <- identify
     putStrLn ("This is Paradox v. 1.0-casc (with " ++ sat ++ "), July 2003.")
     if null (files flags)
       then putStrLn "Warning: No file(s) given! Try --help."
       else sequence_ [ problem flags (sizes flags) file
                      | file <- files flags
                      ]

-------------------------------------------------------------------------
-- problem

problem :: Flags -> [Int] -> FilePath -> IO ()
problem flags ns file =    
  do putHeader "Reading"
     mecs <- try $ readProblemWithRoots ("" : map (++"/") (roots flags)) file
     case mecs of
       Left err ->
         do putStrLn (show err)
            return ()
       
       Right cs ->
         do solveProblem flags' ns (map thd3 cs)
        where
         flags' = flags{ dimacs = patch `fmap` dimacs flags }

         patch ('%':'f':s) = name file ++ patch s
         patch (c:s)       = c : patch s
         patch ""          = ""

         name = reverse
              . dropExtension
              . takeWhile (not . isSlash)
              . reverse

         dropExtension s = case drop 1 (dropWhile (/= '.') s) of
                             "" -> s
                             s' -> s'

         isSlash '/'  = True
         isSlash '\\' = True
         isSlash _    = False

-------------------------------------------------------------------------
-- the end.
