(module connection
	(include "dctp.sch")
	(export (clause-connections clauses))
	(import (globals "globals.scm") (unification "unification.scm")))

;;;*************************************************************************
;;--------------------------------
;; The connection graph functions
;;--------------------------------

(define (clause-connections clauses)
  (if (null? clauses)
      #t
      (begin
	(subgoals-connections (clause-subgoals (car clauses)) (cdr clauses))
	(if (and (unit-clause (car clauses)) (contains-unit-connection (subgoal-links (car (clause-subgoals (car clauses))))))
	    #f
	    (clause-connections (cdr clauses))))))

(define (subgoals-connections subgoals clauses)
  (if (null? subgoals)
      #t
      (begin
	(subgoal-connections (car subgoals) clauses)
	(subgoals-connections (cdr subgoals) clauses))))

(define (subgoal-connections subgoal clauses)
  (if (null? clauses)
      #t
      (begin
	(subgoal-clause-connections subgoal (clause-subgoals (car clauses)))
	(subgoal-connections subgoal (cdr clauses)))))

(define (subgoal-clause-connections subgoal subgoals)
  (if (null? subgoals)
      #t
      (let ((other-subgoal (car subgoals)))
	(if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
		 (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
		 (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms other-subgoal)))
	    (begin
	      (unbind-vars)
	      (set-subgoal-links subgoal (cons other-subgoal (subgoal-links subgoal)))
	      (set-subgoal-links other-subgoal (cons subgoal (subgoal-links other-subgoal)))))
	(subgoal-clause-connections subgoal (cdr subgoals)))))

(define (subgoal-clause-connections-internal subgoal renamed subgoals)
  (if (null? subgoals)
      #t
      (let ((other-subgoal (car subgoals)))
	(if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
		 (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
		 (unify-lists-stop (subgoal-terms renamed) (subgoal-terms other-subgoal)))
	    (begin
	      (unbind-vars)
	      ;;;(print "Found internal link in clause " (subgoal-depth other-subgoal) " (" (subgoal-depth subgoal)  ")")
	      (set-subgoal-links subgoal (cons other-subgoal (subgoal-links subgoal)))
	      (set-subgoal-links other-subgoal (cons subgoal (subgoal-links other-subgoal)))))
	(subgoal-clause-connections subgoal (cdr subgoals)))))

(define (contains-unit-connection connections)
  (if (pair? connections)
      (if (unit-clause (subgoal-clause (car connections)))
	  #t
	  (contains-unit-connection (cdr connections)))
      #f))
