;;;*************************************************************************
;; ---------------------------------------------------------------
;; FORMULA, CLAUSE, SUBGOAL, AND SIGNATURE ACCESS AND MODIFICATION
;; ---------------------------------------------------------------

(define-macro (formula-clauses f)       `(vector-ref ,f 0))      
(define-macro (formula-statistics f)    `(vector-ref ,f 1))      
(define-macro (formula-lemmas f)        `(vector-ref ,f 2))      

(define-macro (set-formula-clauses f arg)     `(vector-set! ,f 0 ,arg))      
(define-macro (set-formula-statistics f arg)  `(vector-set! ,f 1 ,arg))      
(define-macro (set-formula-lemmas f arg)      `(vector-set! ,f 2 ,arg))      

;; ---------------------------------------------------------------
;; Tableau clause data structure
;; ---------------------------------------------------------------

(define-macro (new-clause) `(make-vector 11 #f))

(define-macro (clause-subgoals cl)        `(vector-ref ,cl 0))      
(define-macro (clause-nr-variables cl)    `(vector-ref ,cl 1))      
(define-macro (clause-variables cl)       `(vector-ref ,cl 2))      
(define-macro (clause-path cl)            `(vector-ref ,cl 3))      
(define-macro (clause-predecessor cl)     `(vector-ref ,cl 4))      
(define-macro (clause-input cl)           `(vector-ref ,cl 5))      
(define-macro (clause-depth cl)           `(vector-ref ,cl 6))      
(define-macro (clause-initial cl)         `(vector-ref ,cl 7))      
(define-macro (clause-relevant cl)        `(vector-ref ,cl 8))      
(define-macro (clause-unsimplified cl)    `(vector-ref ,cl 9))      
(define-macro (clause-foldup-lits cl)     `(vector-ref ,cl 10))      

(define-macro (set-clause-subgoals cl arg)        `(vector-set! ,cl 0 ,arg)) 
(define-macro (set-clause-nr-variables cl arg)    `(vector-set! ,cl 1 ,arg))
(define-macro (set-clause-variables cl arg)       `(vector-set! ,cl 2 ,arg))
(define-macro (set-clause-path cl arg)            `(vector-set! ,cl 3 ,arg))
(define-macro (set-clause-predecessor cl arg)     `(vector-set! ,cl 4 ,arg))
(define-macro (set-clause-input cl arg)           `(vector-set! ,cl 5 ,arg))
(define-macro (set-clause-depth cl arg)           `(vector-set! ,cl 6 ,arg))
(define-macro (set-clause-initial cl arg)         `(vector-set! ,cl 7 ,arg))
(define-macro (set-clause-relevant cl arg)        `(vector-set! ,cl 8 ,arg))
(define-macro (set-clause-unsimplified cl arg)    `(vector-set! ,cl 9 ,arg))
(define-macro (set-clause-foldup-lits cl arg)     `(vector-set! ,cl 10 ,arg))

;; ---------------------------------------------------------------
;; Input clause data structure
;; ---------------------------------------------------------------

(define-macro (new-input) `(make-vector 9 #f))

(define-macro (input-copies in)        `(vector-ref ,in 0))      
(define-macro (input-length in)        `(vector-ref ,in 1))      
(define-macro (input-type in)          `(vector-ref ,in 2))      
(define-macro (input-number in)        `(vector-ref ,in 3))      
(define-macro (input-name in)          `(vector-ref ,in 4))      
(define-macro (input-all-copies in)    `(vector-ref ,in 5))      
(define-macro (input-complexity in)    `(vector-ref ,in 6))      
(define-macro (input-signature in)     `(vector-ref ,in 7))      
(define-macro (input-variant-index in) `(vector-ref ,in 8))      

(define-macro (set-input-copies in arg)        `(vector-set! ,in 0 ,arg))
(define-macro (set-input-length in arg)        `(vector-set! ,in 1 ,arg))
(define-macro (set-input-type in arg)          `(vector-set! ,in 2 ,arg))
(define-macro (set-input-number in arg)        `(vector-set! ,in 3 ,arg))
(define-macro (set-input-name in arg)          `(vector-set! ,in 4 ,arg))
(define-macro (set-input-all-copies in arg)    `(vector-set! ,in 5 ,arg))
(define-macro (set-input-complexity in arg)    `(vector-set! ,in 6 ,arg))
(define-macro (set-input-signature in arg)     `(vector-set! ,in 7 ,arg))
(define-macro (set-input-variant-index in arg) `(vector-set! ,in 8 ,arg))

;; ---------------------------------------------------------------
;; Subgoal data structure
;; ---------------------------------------------------------------

(define-macro (subgoal-size) `12)
(define-macro (new-subgoal) `(make-vector (subgoal-size) #f))

(define-macro (subgoal-sign sg)                 `(vector-ref ,sg 0))
(define-macro (subgoal-pred sg)                 `(vector-ref ,sg 1))
(define-macro (subgoal-terms sg)                `(vector-ref ,sg 2))
(define-macro (subgoal-links sg)                `(vector-ref ,sg 3))
(define-macro (subgoal-clause sg)               `(vector-ref ,sg 4))
(define-macro (subgoal-selected sg)             `(vector-ref ,sg 5))
(define-macro (subgoal-next sg)                 `(vector-ref ,sg 6))
(define-macro (subgoal-number sg)               `(vector-ref ,sg 7))
(define-macro (subgoal-weight sg)               `(vector-ref ,sg 8))
(define-macro (subgoal-simplified sg)           `(vector-ref ,sg 9))
(define-macro (subgoal-unlinkable sg)           `(vector-ref ,sg 10))
(define-macro (subgoal-misc sg)                 `(vector-ref ,sg 11))

(define-macro (subgoal-depth sg)                `(clause-depth (subgoal-clause ,sg)))
(define-macro (subgoal-predecessor sg)          `(clause-path (subgoal-clause ,sg)))
(define-macro (subgoal-clause-number sg)        `(input-number (clause-input (subgoal-clause ,sg))))

(define-macro (set-subgoal-sign sg arg)                `(vector-set! ,sg 0 ,arg))
(define-macro (set-subgoal-pred sg arg)                `(vector-set! ,sg 1 ,arg))
(define-macro (set-subgoal-terms sg arg)               `(vector-set! ,sg 2 ,arg))
(define-macro (set-subgoal-links sg arg)               `(vector-set! ,sg 3 ,arg))
(define-macro (set-subgoal-clause sg arg)              `(vector-set! ,sg 4 ,arg))
(define-macro (set-subgoal-selected sg arg)            `(vector-set! ,sg 5 ,arg))
(define-macro (set-subgoal-next sg arg)                `(vector-set! ,sg 6 ,arg))
(define-macro (set-subgoal-number sg arg)              `(vector-set! ,sg 7 ,arg))
(define-macro (set-subgoal-weight sg arg)              `(vector-set! ,sg 8 ,arg))
(define-macro (set-subgoal-simplified sg arg)          `(vector-set! ,sg 9 ,arg))
(define-macro (set-subgoal-unlinkable sg arg)          `(vector-set! ,sg 10 ,arg))
(define-macro (set-subgoal-misc sg arg)                `(vector-set! ,sg 11 ,arg))

;; ---------------------------------------------------------------
;; Term data structures
;; ---------------------------------------------------------------

(define-macro (new-predsym) `(make-vector 10 #f))

(define-macro (predsym-symbol p)          `(vector-ref ,p 0))
(define-macro (predsym-arity p)           `(vector-ref ,p 1))
(define-macro (predsym-number p)          `(vector-ref ,p 2))
(define-macro (predsym-weight p)          `(vector-ref ,p 3))
(define-macro (predsym-pos-path-index p)  `(vector-ref ,p 4))
(define-macro (predsym-neg-path-index p)  `(vector-ref ,p 5))
(define-macro (predsym-pos-gr-index p)    `(vector-ref ,p 6))
(define-macro (predsym-neg-gr-index p)    `(vector-ref ,p 7))
(define-macro (predsym-pos-unit-index p)  `(vector-ref ,p 8))
(define-macro (predsym-neg-unit-index p)  `(vector-ref ,p 9))

(define-macro (set-predsym-symbol p arg)          `(vector-set! ,p 0 ,arg))
(define-macro (set-predsym-arity p arg)           `(vector-set! ,p 1 ,arg))
(define-macro (set-predsym-number p arg)          `(vector-set! ,p 2 ,arg))
(define-macro (set-predsym-weight p arg)          `(vector-set! ,p 3 ,arg))
(define-macro (set-predsym-pos-path-index p arg)  `(vector-set! ,p 4 ,arg))
(define-macro (set-predsym-neg-path-index p arg)  `(vector-set! ,p 5 ,arg))
(define-macro (set-predsym-pos-gr-index p arg)    `(vector-set! ,p 6 ,arg))
(define-macro (set-predsym-neg-gr-index p arg)    `(vector-set! ,p 7 ,arg))
(define-macro (set-predsym-pos-unit-index p arg)  `(vector-set! ,p 8 ,arg))
(define-macro (set-predsym-neg-unit-index p arg)  `(vector-set! ,p 9 ,arg))

(define-macro (new-functor) `(make-vector 4 #f))

(define-macro (functor-symbol sg)     `(vector-ref ,sg 0))
(define-macro (functor-arity sg)      `(vector-ref ,sg 1))
(define-macro (functor-number sg)     `(vector-ref ,sg 2))
(define-macro (functor-weight sg)     `(vector-ref ,sg 3))

(define-macro (set-functor-symbol sg arg)     `(vector-set! ,sg 0 ,arg))
(define-macro (set-functor-arity sg arg)      `(vector-set! ,sg 1 ,arg))
(define-macro (set-functor-number sg arg)     `(vector-set! ,sg 2 ,arg))
(define-macro (set-functor-weight sg arg)     `(vector-set! ,sg 3 ,arg))

;; ---------------------------------------------------------------
;; Signature data structure
;; ---------------------------------------------------------------

(define-macro (new-signature) `(make-vector 4 #f))

(define-macro (signature-symbol sg)     `(vector-ref ,sg 0))
(define-macro (signature-arity sg)      `(vector-ref ,sg 1))
(define-macro (signature-number sg)     `(vector-ref ,sg 2))
(define-macro (signature-weight sg)     `(vector-ref ,sg 3))

(define-macro (set-signature-symbol sg arg)     `(vector-set! ,sg 0 ,arg))
(define-macro (set-signature-arity sg arg)      `(vector-set! ,sg 1 ,arg))
(define-macro (set-signature-number sg arg)     `(vector-set! ,sg 2 ,arg))
(define-macro (set-signature-weight sg arg)     `(vector-set! ,sg 3 ,arg))

;; ---------------------------------------------------------------
;; Variable data structure
;; ---------------------------------------------------------------

(define-macro (new-variable) `(cons '() (cons #f #f)))

(define-macro (variable-name var)             `(cddr ,var))
(define-macro (variable-renaming var)         `(cadr ,var))
(define-macro (variable-binding var)          `(car ,var))

(define-macro (set-variable-name var arg)     `(set-cdr! (cdr ,var) ,arg))
(define-macro (set-variable-renaming var arg) `(set-car! (cdr ,var) ,arg))
(define-macro (set-variable-binding var arg)  `(set-car! ,var ,arg))

;; ---------------------------------------------------------------
;; Link Data Structure
;; ---------------------------------------------------------------

(define-macro (new-link) `(make-vector 5 #f))

(define-macro (link-lita link)              `(vector-ref ,link 0))
(define-macro (link-litb link)              `(vector-ref ,link 1))
(define-macro (link-active link)            `(vector-ref ,link 2))
(define-macro (link-weight link)            `(vector-ref ,link 3))
(define-macro (link-depth link)             `(vector-ref ,link 4))

(define-macro (set-link-lita link arg)      `(vector-set! ,link 0 ,arg))
(define-macro (set-link-litb link arg)      `(vector-set! ,link 1 ,arg))
(define-macro (set-link-active link arg)    `(vector-set! ,link 2 ,arg))
(define-macro (set-link-weight link arg)    `(vector-set! ,link 3 ,arg))
(define-macro (set-link-depth link arg)     `(vector-set! ,link 4 ,arg))


(define-macro (unit-subgoal sg) `(= 1 (input-length (clause-input (subgoal-clause ,sg)))))
(define-macro (unit-clause cl)  `(= 1 (input-length (clause-input ,cl))))
(define-macro (non-unit-subgoal sg) `(< 1 (input-length (clause-input (subgoal-clause ,sg)))))
(define-macro (non-unit-clause cl)  `(< 1 (input-length (clause-input ,cl))))

(define-macro (clause-generic-subgoal clause offset) 
; returns generic subgoal with offset in clause; 
; note: first subgoal has offset 0
`(vector-ref (clause-generic ,clause) ,offset))

(define-macro (subgoals-offset-clause generic offset) 
; given a <generic> and an offset, the respective subgoal in the first
; clause copy is returned; note: first subgoal has offset 0
`(vector-ref                                      
    (copy-subgoals                                       
        (car (clause-copies (subgoal-copy (vector-ref ,generic ,offset)))))
    ,offset))

(define-macro (clause-copy-subgoal clause offset)
; returns entry subgoal with offset in clause; note: first subgoal has offset 0
`(vector-ref (copy-subgoals (car (clause-copies ,clause))) ,offset))

;;;*************************************************************************
;; ---------------
;; TERM TYPE TESTS
;; ---------------

(define-macro (constant term) `(and (pair? ,term) (null? (cdr ,term))))

(define-macro (complex term) 
`(and (pair? ,term) (not (variable ,term)) (pair? (cdr ,term))))

(define-macro (variable term) `(and (pair? ,term) (not (vector? (car ,term))) (pair? (cdr ,term))))

(define-macro (domain-element term) `(and (pair? ,term) (not (cdr ,term))))

(define-macro (non-variable term) `(and (pair? ,term) (vector? (car ,term))))

(define-macro (free-variable term) `(and (pair? ,term) (eq? (car ,term) '())))

(define-macro (bound-variable term) `(and (pair? ,term) (or (integer? ,term) (pair? (car ,term)))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; VARIABLE BINDING, EQUATION EXTRACTION AND TRAIL MODIFICATION
;; ------------------------------------------------------------

(define-macro (push-trail data)
`(begin (trail-next) (set-car! (car trail) ,data)))

(define-macro (trail-next)
`(if (null? (cdr trail))
     (begin (set-cdr! trail (list (cons #f trail)))
            (set! trail (cdr trail)))
     (set! trail (cdr trail))))

(define-macro (trail-stop)
`(begin (trail-next) (set-car! (car trail) #f)))

(define-macro (trail-previous) `(set! trail (cdar trail)))

(define-macro (trail-content) `(caar trail))

(define-macro (bind var term)
;; binds var to term destructively, next trail position is used
`(begin (set-car! ,var ,term)
        (trail-next)
        (set-car! (car trail) ,var)
	#t))

(define-macro (unbind s t)
`(unbind-vars))

(define-macro (unify-lists-stop k l)
`(begin (trail-stop) (unify-lists ,k ,l)))

(define-macro (unify-terms-stop s t)
`(begin (trail-stop) (unify-terms ,s ,t)))

(define-macro (extension-unify-lists-stop k l)
`(begin (trail-stop) 
        (set! optimized-unification 0) 
        (extension-unify-lists ,k ,l)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index macros
;; ------------------------------------------------------------

;; Index macros need to be partially adapted to support a count of valid sub-and-cells

(define-macro (new-or-and-cell)             `(cons (cons #f '()) '()))

(define-macro (and-content i)               `(caar ,i))      
(define-macro (and-next i)                  `(cdar ,i))      
(define-macro (or-current i)                `(car ,i))      
(define-macro (or-next i)                   `(cdr ,i))      

(define-macro (set-and-content i arg)       `(set-car! (car ,i) ,arg))      
(define-macro (set-and-next i arg)          `(set-cdr! (car ,i) ,arg))      
(define-macro (set-or-current i arg)        `(set-car! ,i ,arg))      
(define-macro (set-or-next i arg)           `(set-cdr! ,i ,arg))      

(define-macro (index-variable t) `(integer? ,t))
(define-macro (index-constant t) `(and (vector? ,t) (= (vector-ref ,t 1) 0)))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Term trail modification
;; ------------------------------------------------------------

;; New macros for the term trail are needed

(define-macro (push-term-trail data)
`(begin (term-trail-next) (set-car! (car term-trail) ,data)))

(define-macro (pop-term-trail)
`(if (term-trail-content)
     (begin
       (set! term-trail (cdar term-trail))
       (caadr term-trail))
     #f))

(define-macro (term-trail-next)
`(if (null? (cdr term-trail))
     (begin (set-cdr! term-trail (list (cons #f term-trail)))
            (set! term-trail (cdr term-trail)))
     (set! term-trail (cdr term-trail))))

(define-macro (term-trail-stop)
`(begin (term-trail-next) (set-car! (car term-trail) #f)))

(define-macro (term-trail-previous) `(set! term-trail (cdar term-trail)))

(define-macro (term-trail-content) `(caar term-trail))

