(module display
	(include "dctp.sch")
	(export (display-link-vector) (display-current-link-list l) (display-subgoal subgoal)
		(display-branch-subgoals subgoal) (display-dewey-number subgoal) (display-clause cl) 
		(display-subgoal-list subgoal-list) (display-clauses clauses) (display-link link) (display-ref term) 
		(display-subgoal-terms terms) (display-ref-list terms) (print-statistics) (print-tptp-lemmas subgoal))
	(import (globals "globals.scm") (unification "unification.scm")))

;;--------------------------------
;; Display functions
;;--------------------------------
(define (display-ref term)
  (if (index-variable term)
      (begin
	(display "X")
	(display term))
      (if (variable term)
	  (display (variable-name term))
	  (begin (display (functor-symbol (car term))) 
		 (if (pair? (cdr term))
		     (begin (display "(") 
			    (display-ref-list (cdr term))
			    (display ")")))))))

(define (display-ref-list terms)
 (if (pair? terms)
     (begin (display-ref (ref (car terms)))
            (if (pair? (cdr terms))
	        (begin (display ",") (display-ref-list (cdr terms)))))
     '#\ ))

(define (display-subgoal-list subgoal-list)
  (trail-stop)
  (bind-termlist-to-index-variables (variables-in-subgoals subgoal-list '()) 0)
  (display-subgoal-list1 subgoal-list)
  (unbind-vars))

(define (display-subgoal-list1 subgoal-list)
 (if (pair? subgoal-list)
     (begin (display-subgoal1 (car subgoal-list))
	    (if (pair? (cdr subgoal-list))
		(display ","))
            (display-subgoal-list1 (cdr subgoal-list)))))

(define (display-subgoal subgoal)
  (trail-stop)
  (bind-termlist-to-index-variables (free-vars-in-termlist (subgoal-terms subgoal) '()) 0)
  (display-subgoal1 subgoal)
  (unbind-vars))

(define (display-subgoal-terms terms)
  (trail-stop)
  (bind-termlist-to-index-variables (free-vars-in-termlist terms '()) 0)
  (display-ref-list terms)
  (unbind-vars))

(define (display-subgoal1 subgoal)
  (if (subgoal-selected subgoal) (display "["))
  (display (subgoal-sign subgoal)) 
  (display (predsym-symbol (subgoal-pred subgoal))) 
  (if (> (predsym-arity (subgoal-pred subgoal)) 0) 
      (begin (display "(")
	     (display-ref-list (subgoal-terms subgoal)) 
	     (display ")")))
  (if (subgoal-simplified subgoal) (display "*"))
  (if (subgoal-selected subgoal) (display "]")))

(define (display-clauses clauses)
  (trail-stop)
  (bind-termlist-to-index-variables (clauselist-variables clauses) 0)
  (display-clauses1 clauses)
  (unbind-vars))

(define (display-clauses1 clauses)
 (if (pair? clauses)
     (begin (display-clause1 (car clauses))
	    (if (pair? (cdr clauses)) (newline))
            (display-clauses1 (cdr clauses)))))

(define (display-clause cl)
  (trail-stop)
  (bind-termlist-to-index-variables (clause-variables cl) 0)
  (display-clause1 cl)
  (unbind-vars))

(define (display-clause1 cl)
 (display "clause ") (display (input-number (clause-input cl)))
 (display "(") (display (clause-depth cl)) (display ")")
 (display "[") (display (clause-unsimplified cl)) (display "]")
 (display ": ") 
 (display-subgoal-list1 (clause-subgoals cl)))

(define (display-link-vector)
  (do ((i 0 (+ i 1)))
      ((= i link-vector-size) (newline))
    (if (pair? (vector-ref link-vector i))
	(display-current-link-list (vector-ref link-vector i)))))

(define (display-link-list l)
  (if (pair? l)
      (begin
	(display-link (car l))
	(if (pair? (cdr l)) (display " , "))
	(display-link-list (cdr l)))
      (newline)))

(define (display-current-link-list l)
  (if (pair? l)
      (begin
	(if (not (eq? 'delete (link-active (car l))))
	    (begin
	      (display-link (car l))
	      (if (pair? (cdr l)) (display " , "))))
	(display-current-link-list (cdr l)))
      (newline)))

(define (display-active-link-list l)
  (if (pair? l)
      (begin
	(if (eq? #t (link-active (car l)))
	    (begin
	      (display-link (car l))
	      (if (pair? (cdr l)) (display " , "))))
	(display-active-link-list (cdr l)))
      (newline)))

(define (display-link link)
  (display "<")
  (display-subgoal (link-lita link))
  (display "(")
  (display (subgoal-depth (link-lita link)))
  (display "),")
  (display-subgoal (link-litb link))
  (display "(")
  (display (subgoal-depth (link-litb link)))
  (display "),")
  (display (link-weight link))
  (display ",")
  (display (link-active link))
  (display ",")
  (display (link-depth link))
  (display ">"))

(define (display-branch-subgoals subgoal)
  (if subgoal
      (begin
	(display-branch-subgoals (subgoal-predecessor subgoal))
	(if (subgoal-predecessor subgoal)
	    (display "-"))
	(display-subgoal subgoal))))

(define (display-dewey-number subgoal)
  (display (reverse (dewey-list subgoal))))

(define (dewey-list subgoal)
  (if subgoal
      (cons (subgoal-number subgoal) (dewey-list (subgoal-predecessor subgoal)))
      '()))

(define (bind-termlist-to-index-variables terms value)
  (if (pair? terms)
      (bind-termlist-to-index-variables (cdr terms) (bind-term-to-index-variables (ref (car terms)) value))
      value))

(define (bind-term-to-index-variables term value)
  ;;(print "binding " term " to " value)
  (if (index-variable term)
      value
      (if (variable term)
	  (begin
	    (bind term value)
	    (+ value 1))
	  (if (constant term)
	      value
	      (bind-termlist-to-index-variables (cdr term) value)))))

(define (clauselist-variables clauses)
  (if (pair? clauses)
      (append (clause-variables (car clauses)) (clauselist-variables (cdr clauses)))
      '()))

(define (variables-in-subgoals subgoals vars)
  (if (null? subgoals)
      vars
      (variables-in-subgoals (cdr subgoals) (free-vars-in-termlist (subgoal-terms (car subgoals)) vars))))

(define (free-vars-in-termlist terms vars)
  ;; finds all free variables in terms and inserts them into vars exactly once
  ;;(print vars)
  (if (not (null? terms))
      (begin
	(free-vars-in-termlist (cdr terms) (free-vars-in-term (ref (car terms)) vars)))
      vars))

(define (free-vars-in-term term vars)          
  ;; checks if a term is a variable.
  ;; if yes, it is added to the variable list vars if it is not already in there
  ;; otherwise if it is a proper functor, the argument list of term is searched for
  ;; free variables
  (if (free-variable term)
      (begin
	(find-or-insert-var term vars))
      (if (not (constant term))
	  (free-vars-in-termlist (cdr term) vars)
	  vars)))

(define (find-or-insert-var term vars)
  (if (pair? vars)
      (begin
	(if (not (eq? term (car vars)))
	    (if (pair? (cdr vars))
		(find-or-insert-var term (cdr vars))
		(set-cdr! vars (list term))))
	vars)
      (list term)))

;;;*************************************************************************
;; ----------
;; Statistics
;; ----------

(define (print-statistics)	
  (if (not silent)
      (begin
	(print "Proof search statistics:")
	(print "========================")
	(print "Length initial path:        " initial-length)
	(print "Linking steps (succ/fail):  " successful-inferences " / " (- total-inferences successful-inferences))
	(if context-lemmas 
	    (print "GR-closures total/c-foldup: " successful-reduction-steps " / " context-reduction-steps)
	    (print "GR-closures:                " successful-reduction-steps))
	(print "Total links generated:      " total-links)
	;;(print "Links tagged/untagged:      " links-tagged " / " links-untagged)
	;;(print "Links pushed/popped:        " links-pushed " / " links-popped)
	(if pruning 
	    (print "Clauses in tableau/pruned:  " tableau-clauses " / " nr-prunings)
	    (print "Clauses in tableau:         " tableau-clauses))
	(if regularity 
	    (if tautologies
		(print "Ground regularity fails:    " regularity-fails)
		(print "Regularity/Tautology fails: " regularity-fails " / " tautology-fails))
	    (if (not tautologies) (print "Tautology fails:            " tautology-fails)))
	(print "Variants deleted/subsumed:  " nr-variants " / " variant-subsumptions) 
	;;(if (> variant-subsumption-tests 0)
	;;(print "Average subgoals/inputs:    " (quotient potential-subgoals variant-subsumption-tests) " / " 
	;;(quotient potential-inputs variant-subsumption-tests)))
	(print "Variant subsumption tests:  " variant-subsumption-tests)
	(print "Fails pot/ful/sig/mat/len:  " potential-fails " / " full-fails " / " signature-fails " / " single-match-fails " / " length-fails)
	(if subsumption 
	    (if back-subsume
		(print "Subsumptions forward/back:  " nr-subsumption-deletions " / " backward-subsumptions)
		(print "Subsumption deletions:      " nr-subsumption-deletions))
	    (if back-subsume (print "Back subsumptions:          " backward-subsumptions)))
	(if couples (print "Coupled eliminations:       " coupled-eliminations))
	(if simplify 
	    (if top-down-lemmas
		(print "Simplifications/lemmas:     " nr-simplifications " / " nr-top-down-lemmas)
		(print "Simplifications:            " nr-simplifications))
	    (if top-down-lemmas (print "Top-down-lemmas:            " nr-top-down-lemmas)))
	(if foldup 
	    (if context-lemmas
		(print "Foldups root/context:       " nr-foldups " / " nr-context-foldups)
		(print "Root foldups:               " nr-foldups))
	    (if context-lemmas (print "Context foldups:            " nr-context-foldups)))
	(print "Maximum term depth/compl:   " max-term-depth " / " max-complexity)
	(print "Maximum tableau depth:      " max-depth))))


;;;*************************************************************************
;; ------------
;; Lemma output
;; ------------

(define (print-tptp-lemmas subgoal)
  (if subgoal
      (begin
	(if (and (eq? 'lemma (input-type (clause-input (subgoal-clause subgoal))))
		 (not (subgoal-unlinkable subgoal)))
	    (begin
	      (display "input_clause(")
	      (display (input-name (clause-input (subgoal-clause subgoal))))
	      (display ",lemma,[")
	      (if (eq? (subgoal-sign subgoal) '+)
		  (display "++ ")
		  (display "-- "))
	      (if (eq? (predsym-symbol (subgoal-pred subgoal)) 'equal)
		  (display "equal")
		  (display (predsym-symbol (subgoal-pred subgoal))))
	      (display "(")
	      (display-subgoal-terms (subgoal-terms subgoal))
	      (display ")]).")
	      (newline)))
	(print-tptp-lemmas (subgoal-predecessor subgoal)))))
