(module link
	(include "dctp.sch")
	(export (make-link subgoal predsubgoal) 
		(insert-link link weight) (mark-links-deletable links subgoal) (print-link-statistics) 
		(insert-links subgoal predsubgoals) (restore-link-cells link-cells)  (select-link-cell index depth) 
		(reset-link-statistics) (insert-lemma-links lemma-sg predsubgoal) (init-link-bias))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm")	(index "index.scm") 
		(select "select.scm") (subgoal-simplifiable pruning "pruning.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define total-weight 0)
(define min-weight 100000)
(define max-weight 0)
(define total-select 0)
(define min-select 100000)
(define max-select 0)
(define deleted-select 0)
(define new-links 0)
(define new-selects 0)

(define max-sg-weight 0)
(define open1 0)
(define delete1 #f)
(define open2 0)
(define delete2 #f)
(define index-end1 #f)
(define index-end2 #f)

(define unit-offset 10)
(define active-links-exist #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; The link functions
;; ------------------------------------------------------------

(define (reset-link-statistics)
  (set! total-select 0)
  (set! min-select 100000)
  (set! max-select 0)
  (set! deleted-select 0)
  (set! new-links 0)
  (set! new-selects 0)
  (set! total-weight 0)
  (set! min-weight 100000)
  (set! max-weight 0))

(define (insert-links subgoal predsubgoals)
  (if (pair? predsubgoals)
      (begin
	(if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign (car predsubgoals))))
		 (eq? (subgoal-pred subgoal) (subgoal-pred (car predsubgoals)))
		 (not (subgoal-unlinkable (car predsubgoals)))
		 (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms (car predsubgoals))))
	    (begin
	      (set! linked-terms-ground #t)
	      (set! max-sg-weight 0)
	      (set! current-depth 0)
	      (set! current-complexity 0)
	      (set! index-end1 (clause-in-index? (subgoal-clause subgoal)))
	      (if (and (vector? index-end1) (> (vector-ref index-end1 0) 0)) ;; if the first clause would become a variant
		  (begin
		    (set! delete1 #t)
		    (set! open1 0))
		  (if (and subsumption
			   (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause subgoal)))
			   (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause subgoal))))
		      (begin ;; or if it were subsumed
			(set! delete1 #t)
			(set! open1 0))
		      (if (and regularity (index-irregular-subgoals (clause-subgoals (subgoal-clause subgoal))))
			  (begin
			    (set! delete1 #t)
			    (set! open1 0))
			  (begin
			    (set! delete1 #f)
			    (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause subgoal)) subgoal))))))
	      (set! index-end2 (clause-in-index? (subgoal-clause (car predsubgoals))))
	      (if (and (vector? index-end2) (> (vector-ref index-end2 0) 0)) ;; if the second clause would become a variant
		  (begin
		    (set! delete2 #t)
		    (set! open2 0))
		  (if (and subsumption
			   (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (car predsubgoals))))
			   (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (car predsubgoals)))))
		      (begin ;; or if it were subsumed
			(set! delete2 #t)
			(set! open2 0))
		      (if (and regularity (index-irregular-subgoals (clause-subgoals (subgoal-clause (car predsubgoals)))))
			  (begin
			    (set! delete2 #t)
			    (set! open2 0))
			  (begin
			    (set! delete2 #f)
			    (set! open2 (get-open-subgoals (clause-subgoals (subgoal-clause (car predsubgoals))) 
							   (car predsubgoals)))))))
	      (unbind-vars)
	      (if (not (or (and delete1 delete2)
			   (and couples (non-unit-clause (subgoal-clause subgoal)) 
				(non-unit-clause (subgoal-clause (car predsubgoals)))
				(or (not (vector? index-end1))
				    (= (vector-ref index-end1 0) 0)
				    (not (eq? (car (vector-ref index-end1 1)) (subgoal-clause subgoal))))
				(or (not (vector? index-end2))
				    (= (vector-ref index-end2 0) 0)
				    (not (eq? (car (vector-ref index-end2 1)) (subgoal-clause (car predsubgoals)))))
				(or delete1 delete2))))
		  (let ((link (make-link subgoal (car predsubgoals))))
		    (set-subgoal-links subgoal (cons link (subgoal-links subgoal)))
		    (set-subgoal-links (car predsubgoals) (cons link (subgoal-links (car predsubgoals))))
		    ;;(set! links-pushed (+ links-pushed 1))
		    (set-link-weight link (compute-link-weight subgoal (car predsubgoals) open1 open2 max-sg-weight))
		    (set-link-depth link current-complexity)
; 		    (if (and (or (clause-relevant (subgoal-clause subgoal))
; 				 (clause-relevant (subgoal-clause (car predsubgoals))))
; 			     (> current-complexity min-relevant-depth))
; 			(set! min-relevant-depth current-complexity))
		    (if (< current-complexity min-relevant-depth)
			(set! min-relevant-depth current-complexity))
		    (if (> debug-level 4)
			(begin
			  (display "inserting link ")
			  (newline)))
		    (insert-link link (link-weight link))
		    (if (and link-info (= 0 (remainder total-links link-info)))
			(begin
			  (display "link statistics after ")
			  (display total-links)
			  (display " links (average min max):")
			  (newline)
			  (print-link-statistics)))))))
	(insert-links subgoal (cdr predsubgoals)))))

(define (print-link-statistics)
  (display "weight: ")
  (display (quotient total-weight new-links))
  (display " ")
  (display min-weight)
  (display " ")
  (display max-weight)
  (display "    select: ")
  (if (> new-selects 0)
      (display (quotient total-select new-selects))
      (display "0"))
  (display " ")
  (display min-select)
  (display " ")
  (display max-select)
  (display " deleted ")
  (display deleted-select)
  (newline)
  (set! total-weight 0)
  (set! min-weight 1000000)
  (set! max-weight 0)
  (set! total-select 0)
  (set! min-select 1000000)
  (set! max-select 0)
  (set! deleted-select 0)
  (set! new-selects 0)
  (set! new-links 0))

(define (insert-link link weight)
  (let (index)
    (if (or (>= weight link-vector-size) (< weight 0))
	(set! index (- link-vector-size 1))
	(set! index weight))
    (if (and (< index search-start-index) (<= (link-depth link) allowed-term-depth))
	(set! search-start-index index)
	(set! active-links-exist #t))
    (if (= index (- link-vector-size 1))
	(if (null? (vector-ref link-vector index))
	    (vector-set! link-vector index (list link))
	    (linear-insert-link link weight (vector-ref link-vector index)))
	(vector-set! link-vector index (cons link (vector-ref link-vector index))))))

(define (linear-insert-link link weight link-list)
  (if (<= weight (link-weight (car link-list)))
      (begin
	(set-cdr! link-list (cons (car link-list) (cdr link-list)))
	(set-car! link-list link))
      (if (pair? (cdr link-list))
	  (linear-insert-link link weight (cdr link-list))
	  (set-cdr! link-list (list link)))))

(define (restore-link-cells link-cells)
  (if (pair? link-cells)
      (let ((rest-cells (cdr link-cells)))
	(if (not (link-active (car link-cells))) ;; do not restore deletable links
	    (begin
	      ;;(set! links-untagged (+ links-untagged 1))
	      (restore-link-cell link-cells)))
	(restore-link-cells rest-cells))))

(define (restore-link-cell link-cell)
  (let (index)
    (if (or (>= (link-weight (car link-cell)) link-vector-size) (< (link-weight (car link-cell)) 0))
	(set! index (- link-vector-size 1))
	(set! index (link-weight (car link-cell))))
    (if (and (< index search-start-index) (<= (link-depth (car link-cell)) allowed-term-depth))
	(set! search-start-index index)
	(set! active-links-exist #t))
    (set-link-active (car link-cell) #t)
    (if (= index (- link-vector-size 1))
	(if (null? (vector-ref link-vector index))
	    (begin
	      (vector-set! link-vector index link-cell)
	      (set-cdr! link-cell '()))
	    (linear-insert-link (car link-cell) (link-weight (car link-cell)) (vector-ref link-vector index)))
	(begin
	  (set-cdr! link-cell (vector-ref link-vector index))
	  (vector-set! link-vector index link-cell)))))

(define (make-link subgoal predsubgoal)
  (let ((link (new-link)))
    (set! total-links (+ total-links 1))
    (set! new-links (+ new-links 1))
    (if unit-on-top
	(if (and (unit-subgoal predsubgoal) (non-unit-subgoal subgoal))
	    (begin
	      (set-link-lita link predsubgoal)
	      (set-link-litb link subgoal))
	    (begin
	      (set-link-lita link subgoal)
	      (set-link-litb link predsubgoal)))
	(if (and (unit-subgoal subgoal) (non-unit-subgoal predsubgoal))
	    (begin
	      (set-link-lita link predsubgoal)
	      (set-link-litb link subgoal))
	    (begin
	      (set-link-lita link subgoal)
	      (set-link-litb link predsubgoal))))
    (set-link-active link #t)
    link))
  
(define (link-relevant link)
  (or (clause-relevant (subgoal-clause (link-lita link)))
      (clause-relevant (subgoal-clause (link-litb link)))))

(define (compute-link-weight subgoal predsubgoal open1 open2 complexity)
  (let (weight)
    (set! weight (+ (* sg-bias (+ open1 open2)) (* (+ open1 open2) term-bias complexity)))
    (if (> open1 0)
 	(set! open1 1))
    (if (> open2 0)
 	(set! open2 1))
    (if (not (or norel
		 (clause-relevant (subgoal-clause subgoal))
		 (clause-relevant (subgoal-clause predsubgoal))))
 	(set! weight (bit-lsh weight 1)))
    (if (and (non-unit-subgoal subgoal) (non-unit-subgoal predsubgoal))
	(set! weight (+ unit-offset weight)))
    (if (and (> open1 0) (> open2 0))
 	(set! weight (bit-lsh weight 1)))
;     (if (not linked-terms-ground)
;  	(set! weight (+ weight 10)))
    (if (not limit-term-depth)
	(set! weight (+ weight current-depth)))
;     (if (and (clause-initial (subgoal-clause predsubgoal)) (clause-initial (subgoal-clause subgoal)))
; 	(set! weight (quotient weight 4)))
    (if (and (= open1 0) (= open2 0))
	(set! weight 0))
    (if (< weight 0) (set! weight maxint))
    (if link-info
	(begin
	  (if (< weight min-weight) (set! min-weight weight))
	  (if (> weight max-weight) (set! max-weight weight))
	  (set! total-weight (+ total-weight weight))))
    weight))

(define (select-link-cell index depth)
  (if (>= index link-vector-size)
      (if (and limit-term-depth active-links-exist)
	  (begin
	    (set! allowed-term-depth (+ allowed-term-depth 1))
	    (set! active-links-exist #f)
	    (set! search-start-index 0)         ;; next searches might must be starting from scratch
	    (if dewey-numbers
		(begin
		  ;;(newline)
		  (display "allowed link term complexity increased to ")
		  (display allowed-term-depth)
		  (newline)))
	    (select-link-cell 0 (+ depth 1)))
	  #f)      ;; once we get past the last link-vector entry, there are no more links: affirmation state
      (let ((link-cell (extract-active-link-from-list (vector-ref link-vector index) (vector-ref link-vector index) index)))
	(if (not link-cell)
	    (begin       ;; if the current link-vector entry contains no active link-cells
	      (set! depth (+ depth 1))
	      (set! index (+ index 1))
	      (do ()     ;; look for the next non-empty link-vector entry
		  ((or (>= index link-vector-size) (pair? (vector-ref link-vector index))) #t)
		(set! depth (+ depth 1))
		(set! index (+ index 1)))
	      (select-link-cell index (+ depth 1))) ;; then search through that entry
	    (begin
	      (set! total-select (+ total-select depth))
	      (set! new-selects (+ new-selects 1))
	      (if (< depth min-select) (set! min-select depth))
	      (if (> depth max-select) (set! max-select depth))
	      ;;(set! link-cell (vector-ref link-vector index)) ;; select the first active link-cell in the entry
	      ;;(vector-set! link-vector index (cdr (vector-ref link-vector index))) ;; delete it from that entry
	      ;; if we selected the only link-cell in the entry
	      (if (null? (vector-ref link-vector index))
		  (set! search-start-index (+ index 1))    ;; next search begins one entry down the vector
		  (set! search-start-index index))         ;; next search might be successful in this entry
	      ;;(set! active-links-exist #f)
	      (set-cdr! link-cell '())
	      link-cell)))))

(define (extract-active-link-from-list list pred-list index)
  (if (pair? list)
      (if (eq? (link-active (car list)) #t)
	  (if (<= (link-depth (car list)) allowed-term-depth)
	      (if (eq? list (vector-ref link-vector index))
		  (begin
		    (vector-set! link-vector index (cdr list))
		    list)
		  (begin
		    (set-cdr! pred-list (cdr list))
		    list))
	      (begin
		(set! active-links-exist #t)
		(extract-active-link-from-list (cdr list) list index)))
	  (if (eq? list (vector-ref link-vector index))
	      (begin
		(vector-set! link-vector index (cdr list))
		(extract-active-link-from-list (vector-ref link-vector index) (vector-ref link-vector index) index))
	      (begin
		(set-cdr! pred-list (cdr list))
		(extract-active-link-from-list (cdr pred-list) pred-list index))))
      #f))

(define (mark-links-deletable links subgoal)
  (if (pair? links)
      (begin 
	(if (not (eq? (link-active (car links)) 'delete))
	    (set-link-active (car links) 'delete))
	    ;;(set! links-popped (+ links-popped 1))))
	(if (and (link-lita (car links)) (not (eq? (link-lita (car links)) subgoal)))
	    (delete-link-from-list (car links) (link-lita (car links)))
	    (if (and (link-litb (car links)) (not (eq? (link-litb (car links)) subgoal)))
		(delete-link-from-list (car links) (link-litb (car links)))))
	(set-link-lita (car links) #f)
	(set-link-litb (car links) #f)
	(set-link-depth (car links) #f)
	(mark-links-deletable (cdr links) subgoal)
	(set-cdr! links '()))))

(define (delete-link-from-list link subgoal)
  (set-subgoal-links subgoal (delete-element-from-list link (subgoal-links subgoal) (subgoal-links subgoal))))

(define (delete-element-from-list element list pred-list)
  (if (pair? list)
      (if (eq? (car list) element)
          (if (eq? list pred-list)
              (cdr list)
              (begin
                (set-cdr! pred-list (cdr list))
                (cdr list)))
          (if (delete-element-from-list element (cdr list) list)
              list
              #f))  
      #f))

(define (get-open-subgoals subgoals link-sg)
  (if (pair? subgoals)
      (if (or (eq? (car subgoals) link-sg)
	      (ground-closed-with-index (car subgoals))
	      (subgoal-simplified (car subgoals))
	      (match-subgoal-with-index (car subgoals)))
	  (get-open-subgoals (cdr subgoals) link-sg)
	  (let (sg-weight)
	    (set! complexity 0)
	    (set! sg-weight (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals)) 1) 
				     (predsym-weight (subgoal-pred (car subgoals)))))
	    (if (> complexity current-complexity)
		(set! current-complexity complexity))
	    (if (> complexity max-complexity)
		(set! max-complexity complexity))
	    (if (> sg-weight max-sg-weight)
		(set! max-sg-weight sg-weight))
	    (+ 1 (get-open-subgoals (cdr subgoals) link-sg))))
      0))

(define (insert-lemma-links lemma-sg predsubgoal)
  (if predsubgoal
      (begin
	(if (and (not (eq? (subgoal-sign lemma-sg) (subgoal-sign predsubgoal)))
		 (eq? (subgoal-pred lemma-sg) (subgoal-pred predsubgoal))
		 (not (subgoal-unlinkable predsubgoal))
		 (unify-lists-stop (subgoal-terms lemma-sg) (subgoal-terms predsubgoal)))
	    (begin
	      (set! linked-terms-ground #t)
	      (set! max-sg-weight 0)
	      (set! current-depth 0)
	      (set! current-complexity 0)
	      ;; we have to omit these redundancy checks for the time being. These checks all refer to the current 
	      ;; state of the tableau. we must, however, refer to the state of the tableau at the lower point of the
	      ;; new link. the new link might be rejected due to a regularity failure with a subgoal lower on the tableau.
	      ;; on another branch this subgoal might not exist, so we must have the link in that case.
	      ;; subsumption is an exception because it only operates with input subgoals which are valid in every state of
	      ;; the tableau.
	      ;;(set! index-end1 (clause-in-index? (subgoal-clause lemma-sg)))
	      ;;(if (and (vector? index-end1) (> (vector-ref index-end1 0) 0)) ;; if the second clause would become a variant
		  ;;(begin
		    ;;(set! delete1 #t)
		    ;;(set! open1 0))
	      (if (and subsumption
		       (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause lemma-sg)))
		       (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause lemma-sg))))
		  (begin ;; or if it were subsumed
		    (set! delete1 #t)
		    (set! open1 0))
		  ;;(if (and regularity (index-irregular-subgoals (clause-subgoals (subgoal-clause lemma-sg))))
		      ;;(begin
			;;(set! delete1 #t)
			;(set! open1 0))
		  (begin
		    (set! delete1 #f)
		    (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause lemma-sg)) lemma-sg))))
	      ;;(set! index-end2 (clause-in-index? (subgoal-clause predsubgoal)))
	      ;;(if (and (vector? index-end2) (> (vector-ref index-end2 0) 0)) ;; if the second clause would become a variant
		  ;;(begin
		    ;;(set! delete2 #t)
		    ;;(set! open2 0))
	      (if (and subsumption
		       (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause predsubgoal)))
		       (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause predsubgoal))))
		  (begin ;; or if it were subsumed
		    (set! delete2 #t)
		    (set! open2 0))
		  ;;(if (and regularity (index-irregular-subgoals (clause-subgoals (subgoal-clause predsubgoal))))
		      ;;(begin
			;;(set! delete2 #t)
			;;(set! open2 0))
		  (begin
		    (set! delete2 #f)
		    (set! open2 (get-open-subgoals (clause-subgoals (subgoal-clause predsubgoal)) predsubgoal))))
	      (unbind-vars)
	      (if (> debug-level 8)
		  (begin
		    (display "Found unifying subgoal ")
		    (display-subgoal predsubgoal)
		    (display " of depth ")
		    (display (subgoal-depth predsubgoal))
		    (newline)))
	      (if (unit-clause (subgoal-clause predsubgoal)) ;; if we find a unifiable unit we are done anyway
		  (begin
		    (if dewey-numbers
			(begin
			  (display "Found complementary unit for new lemma: ")
			  (display-clause (subgoal-clause predsubgoal))
			  (newline)))
		    (subgoal-depth predsubgoal))
		  (if (and (subgoal-simplifiable predsubgoal) ;; if we can simplify a path subgoal with the new lemma...
			   ;; we must not allow pruning up to non-unit input clauses!
			   (not (clause-initial (subgoal-clause predsubgoal))))
		      (let (tmp-result)
			(set! tmp-result (insert-lemma-links lemma-sg (subgoal-predecessor predsubgoal)))
			;; we must not allow pruning up to non-unit input clauses!
			(if tmp-result
			    tmp-result
			    (subgoal-depth predsubgoal)))
		      (begin
			(if (not (and delete1 delete2))
			    (let ((link (make-link lemma-sg predsubgoal)))
			      (set-subgoal-links predsubgoal (cons link (subgoal-links predsubgoal)))
			      (set-subgoal-links lemma-sg (cons link (subgoal-links lemma-sg)))
			      ;;(set! links-pushed (+ links-pushed 1))
			      (set-link-weight link (compute-link-weight lemma-sg predsubgoal 0 open2 max-sg-weight))
			      (set-link-depth link current-complexity)
			      (if (> debug-level 4)
				  (begin
				    (display "inserting link ")
				    (newline)))
			      (insert-link link (link-weight link))
			      (if (and link-info (= 0 (remainder total-links link-info)))
				  (begin
				    (print "link statistics after " total-links " links ")
				    (print-link-statistics)))))
			(insert-lemma-links lemma-sg (subgoal-predecessor predsubgoal))))))
	    (begin
	      (if (and back-subsume (not (subgoal-unlinkable predsubgoal))
		       (or unit-subsume (non-unit-clause (subgoal-clause predsubgoal))))
		  (let ((tmp (subgoal-subsumed-by-index predsubgoal)))
		    (if (and tmp (not (eq? tmp predsubgoal)))
			(begin ;; back subsumption
			  (if dewey-numbers
			      (begin
				(display "back subsuming with new lemma: ")
				(display-clause (subgoal-clause predsubgoal))
				(newline)))
			  (set! backward-subsumptions (+ 1 backward-subsumptions))
			  (delete-subgoals-links (clause-subgoals (subgoal-clause predsubgoal)) predsubgoal)))))
	      (insert-lemma-links lemma-sg (subgoal-predecessor predsubgoal)))))
      #f))

(define (delete-subgoals-links subgoals path-sg)
  (if (pair? subgoals)
      (begin
	(mark-links-deletable (subgoal-links (car subgoals)) (car subgoals))
	(set-subgoal-selected (car subgoals) #t)
	(set-subgoal-unlinkable (car subgoals) #t)
	(delete-subgoals-links (cdr subgoals) path-sg))))

(define (index-irregular-subgoals subgoals)
  (if (pair? subgoals)
      (let ((result (ground-irregular-subgoal? (car subgoals))))
	(if result
	    result
	    (index-irregular-subgoals (cdr subgoals))))
      #f))

(define (init-link-bias)
  (if (and groundable (or (not sg-bias) (not term-bias)))
      (begin
	(set! sg-bias 4)
	(set! term-bias 1))
      (begin
	(if (not sg-bias)
	    (set! sg-bias 4))
	(if (not term-bias)
	    (set! term-bias 4))))
  (if horn
      (set! unit-offset 100)))
