(module path
	(include "dctp.sch")
	(export (initialize-path-literal-vectors) (add-path-literal subgoal) (add-unit-lemma subgoal) 
		(delete-path-literal subgoal) (potential-linked-lits subgoal) (potential-linked-lemmas subgoal))
	(import (globals "globals.scm") (parser "parser.scm") (display "display.scm")))
;;(index "index.scm") (dctp "dctp.scm") (select "select.scm"))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define pos-path-literals #f)
(define neg-path-literals #f)
(define pos-path-lemmas #f)
(define neg-path-lemmas #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Initialization
;; ------------------------------------------------------------

(define (initialize-path-literal-vectors)
  (set! pos-path-literals (make-vector (+ 1 nr-predicates) '()))
  (set! neg-path-literals (make-vector (+ 1 nr-predicates) '()))
  (set! pos-path-lemmas (make-vector (+ 1 nr-predicates) '()))
  (set! neg-path-lemmas (make-vector (+ 1 nr-predicates) '())))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Adding literals
;; ------------------------------------------------------------

(define (add-path-literal subgoal)
  (let (lists pred-num)
    (set! pred-num (predsym-number (subgoal-pred subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! lists pos-path-literals)
	(set! lists neg-path-literals))
    (if (null? (vector-ref lists pred-num))
	(let ((new-sg (cons subgoal '())))
	  (vector-set! lists pred-num new-sg))
	(vector-set! lists pred-num (cons subgoal (vector-ref lists pred-num))))))

(define (add-unit-lemma subgoal)
  (let (lists pred-num)
    (set! pred-num (predsym-number (subgoal-pred subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! lists pos-path-lemmas)
	(set! lists neg-path-lemmas))
    (if (null? (vector-ref lists pred-num))
	(let ((new-sg (cons subgoal '())))
	  (vector-set! lists pred-num new-sg))
	(vector-set! lists pred-num (cons subgoal (vector-ref lists pred-num))))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Removing literals
;; ------------------------------------------------------------

(define (delete-path-literal subgoal)  
  (let (lists list-ends pred-num)
    (set! pred-num (predsym-number (subgoal-pred subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! lists pos-path-literals)
	(set! lists neg-path-literals))
    (if (or (null? (vector-ref lists pred-num)) (not (eq? subgoal (car (vector-ref lists pred-num)))))
	(begin
	  (display "path literal deletion error for subgoal ")
	  (display-subgoal subgoal)
	  (display " on depth ")
	  (display (subgoal-depth  subgoal))
	  (newline)
	  (if (not (null? (vector-ref lists pred-num)))
	      (begin
		(display "found subgoal ")
		(display-subgoal (car (vector-ref lists pred-num)))
		(display " of depth ")
		(display (subgoal-depth (car (vector-ref lists pred-num))))
		(newline))))
	(begin
	  ;;(set-car! (vector-ref lists pred-num) #f)
	  (vector-set! lists pred-num (cdr (vector-ref lists pred-num)))))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Obtaining the list of possible linked path literals
;; ------------------------------------------------------------

(define (potential-linked-lits subgoal)
  (if (eq? '+ (subgoal-sign subgoal))
      (vector-ref neg-path-literals (predsym-number (subgoal-pred subgoal)))
      (vector-ref pos-path-literals (predsym-number (subgoal-pred subgoal)))))

(define (potential-linked-lemmas subgoal)
  (if (eq? '+ (subgoal-sign subgoal))
      (vector-ref neg-path-lemmas (predsym-number (subgoal-pred subgoal)))
      (vector-ref pos-path-lemmas (predsym-number (subgoal-pred subgoal)))))
