(module pruning
	(include "dctp.sch")
	(export (merge-rel-vectors rel-vector-a rel-vector-b depth) (mark-rel-vector rel-vector depth) 
		(subgoal-relevant subgoal rel-vector) (new-rel-vector depth) (initialize-table table offset) 
		(non-tautological clause) (ref-equal s t) (initialize-bit-mask-table) 
		(is-context-free rel-vector depth) (display-rel-vector r) (make-top-down-lemma subgoal) 
		(subgoal-simplifiable subgoal) (foldup-subgoal subgoal rel-vector) (delete-clause-foldup-lits clause)
		(path-vector-set i clause) (simplified-unit subgoal))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm")
		(index "index.scm") (link "link.scm")  (dctp "dctp.scm") (select "select.scm") (path "path.scm")
		(renamed-termlist normal-form "normal-form.scm")))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define path-vector (make-vector 5000 #f))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Path vector access
;; ------------------------------------------------------------

(define (path-vector-ref i)
  (if (< i (vector-length path-vector))
      (vector-ref path-vector i)
      (let ((new-path-vector (make-vector (* 2 (vector-length path-vector)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length path-vector)) #f)
	  (vector-set! new-path-vector j (vector-ref path-vector j)))
	(set! path-vector new-path-vector)
	#f)))
	
(define (path-vector-set i clause)
  (if (< i (vector-length path-vector))
      (vector-set! path-vector i clause)
      (let ((new-path-vector (make-vector (* 2 (vector-length path-vector)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length path-vector)) #f)
	  (vector-set! new-path-vector j (vector-ref path-vector j)))
	(set! path-vector new-path-vector)
	(vector-set! path-vector i clause)
	#t)))
	
;;;*************************************************************************
;; ------------------------------------------------------------
;; Term equality
;; ------------------------------------------------------------

(define (ref-equals k l)
;; checks whether termlists k and l are equal 
 (if (null? k)
     #t
     (if (ref-equal (ref (car k)) (ref (car l)))
	 (ref-equals (cdr k) (cdr l))
	 #f)))

(define (ref-equal s t)          
;; checks whether terms s and t are equal 
 (if (eq? s t)
     #t
     (if (or (variable s) (variable t))
         #f
         (and (eq? (car s) (car t)) 
              (ref-equals (cdr s) (cdr t))))))

;;;*************************************************************************
;; -------
;; Pruning
;; -------

(define (merge-rel-vectors rel-vector-a rel-vector-b depth)
  (if (and rel-vector-a rel-vector-b)
      (if (eq? #t rel-vector-a)
	  rel-vector-b
	  (if (eq? #t rel-vector-b)
	      rel-vector-a
	      (begin
		(merge-boolean-vectors rel-vector-a rel-vector-b (rel-vec-index depth))
		rel-vector-a)))
      #f))

(define (merge-boolean-vectors vec1 vec2 index)
 (if (>= index 0)
     (begin (vector-set! vec1 index (bit-or (vector-ref vec1 index)
					    (vector-ref vec2 index)))
	    (merge-boolean-vectors vec1 vec2 (- index 1)))))

(define (mark-rel-vector rel-vector depth)
  (if (>= depth 0)
      (vector-set! rel-vector (rel-vec-index depth)
		   (bit-or (rel-vec-position depth) (vector-ref rel-vector (rel-vec-index depth))))))

(define (subgoal-relevant subgoal rel-vector)
  (let (depth result)
    (set! depth (subgoal-depth subgoal))
    (if (>= depth 0)
	(begin
	  (set! result (bit-and (rel-vec-position depth) (vector-ref rel-vector (rel-vec-index depth))))
	  (if (= 0 result)
	      #f
	      #t))
	#t)))

(define (new-rel-vector depth)
  (make-vector (max min-rel-vec-length (+ 2 (rel-vec-index depth))) 0))

(define (copy-rel-vector rel-vector)
  (let ((new-vector (make-vector (vector-length rel-vector))))
    (do ((i 0 (+ i 1)))
	((>= i (vector-length new-vector)) new-vector)
      (vector-set! new-vector i (vector-ref rel-vector i)))))

(define (initialize-table table offset)
  (do ((i offset (+ i 1)))
      ((>= i (vector-length table)) #t)
    (vector-set! table i 
		 (cons (quotient i bit-vector-length) (bit-lsh 1 (- bit-vector-length 1 (remainder i bit-vector-length)))))))

(define (rel-vec-index depth)
  (if (>= depth (vector-length rel-vec-indexing-table))
      (let ((new-table (make-vector (* 2 (vector-length rel-vec-indexing-table)) #f)))
	(do ((i 0 (+ i 1)))
	    ((>= i (vector-length rel-vec-indexing-table)) #t)
	  (vector-set! new-table i (vector-ref rel-vec-indexing-table i)))
	(initialize-table new-table (vector-length rel-vec-indexing-table))
	(set! rel-vec-indexing-table new-table)))
  (car (vector-ref rel-vec-indexing-table depth)))

(define (rel-vec-position depth)
  (if (>= depth (vector-length rel-vec-indexing-table))
      (let ((new-table (make-vector (* 2 (vector-length rel-vec-indexing-table)) #f)))
	(do ((i 0 (+ i 1)))
	    ((>= i (vector-length rel-vec-indexing-table)) #t)
	  (vector-set! new-table i (vector-ref rel-vec-indexing-table i)))
	(initialize-table new-table (vector-length rel-vec-indexing-table))
	(set! rel-vec-indexing-table new-table)))
  (cdr (vector-ref rel-vec-indexing-table depth)))

(define (initialize-bit-mask-table)
  (let ((mask (bit-lsh 1 (- bit-vector-length 1))))
    (set! bit-mask-table (make-vector bit-vector-length 0))
    (do ((i 1 (+ i 1)))
	((= i bit-vector-length) #t)
      (vector-set! bit-mask-table i mask)
      (set! mask (bit-or mask (bit-rsh mask 1))))))

(define (is-context-free rel-vector depth)
  (let ((rel-vector-index 0))
    (if (do ((i 0 (+ i 1)))
	    ((or (not (= (vector-ref rel-vector i) 0)) (< depth bit-vector-length)) (< depth bit-vector-length))
	  (set! rel-vector-index (+ rel-vector-index 1))
	  (set! depth (- depth bit-vector-length)))
	(begin
	  ;;(display depth)
	  ;;(display " ")
	  ;;(display (vector-ref bit-mask-table depth))
	  ;;(display " ")
	  (= 0 (bit-and (vector-ref rel-vector rel-vector-index) (vector-ref bit-mask-table depth))))
	#f)))

(define (foldup-depth r depth)
  (let ((l 0) 
	(depth-offset (quotient depth bit-vector-length)) 
	(depth-index (remainder depth bit-vector-length)) 
	tmp)
    (set! tmp (vector-ref r depth-offset)) ;; save the rel-vector entry of depth and blank out levels depth and below
    (vector-set! r depth-offset (bit-and (vector-ref r depth-offset) (vector-ref bit-mask-table depth-index)))
    (set! l (do ((i depth-offset (- i 1))) ;; search for the first non-zero rel-vector element
		    ((or (< i 0) (not (= (vector-ref r i) 0))) i)))
    (if (< l 0) 
	(begin
	  (vector-set! r depth-offset tmp) ;; restore levels depth and below
	  l)
	(let (index)
	  (set! index (do ((j 0 (+ j 1))) ;; from the right end of the bit-vector, search the first non-zero bit
			  ((or (>= j bit-vector-length) (not (= 0 (bit-and (vector-ref r l) (bit-lsh 1 j))))) j)))
	  (vector-set! r depth-offset tmp)
	  (+ (- bit-vector-length index 1) (* bit-vector-length l))))))
    
(define (copy-up-to rel-vector depth) ;; depth is the first level not copied!
  (let ((new-vector (make-vector (vector-length rel-vector) 0))
	(depth-offset (quotient depth bit-vector-length)) 
	(depth-index (remainder depth bit-vector-length)))
    (do ((i 0 (+ i 1)))
	((> i depth-offset) new-vector)
      (vector-set! new-vector i (vector-ref rel-vector i)))
    (vector-set! new-vector depth-offset (bit-and (vector-ref rel-vector depth-offset) 
						  (vector-ref bit-mask-table depth-index)))
    new-vector))

(define (clear-depth-and-below r depth)
  (let ((depth-offset (quotient depth bit-vector-length)) 
	(depth-index (remainder depth bit-vector-length)))
    (vector-set! r depth-offset (bit-and (vector-ref r depth-offset) (vector-ref bit-mask-table depth-index)))
    (do ((i (+ 1 depth-offset) (+ i 1)))
	((>= i (vector-length r)) #t)
      (vector-set! r i 0))))

(define (display-rel-vector r)
  (do ((i 0 (+ i 1)))
      ((>= i (vector-length r)) #t)
    (if (not (= (vector-ref r i) 0))
	(do ((j 1 (+ j 1)))
	    ((> j bit-vector-length) #t)
	  (if (not (= 0 (bit-and (vector-ref r i) (bit-lsh 1 (- bit-vector-length j)))))
	      (begin
		(display " ")
		(display (+ (- j 1) (* bit-vector-length i)))))))))



;;;*************************************************************************
;; ---------------------
;; Tautology elimination
;; ---------------------

(define (non-tautological clause)
  (if (non-tautological-subgoals (clause-subgoals clause))
      #t
      (begin
	(set! tautology-fails (+ tautology-fails 1))
	(if (or dewey-numbers (> debug-level 4))
	    (begin
	      (display "rejected tautological ")
	      (display-clause clause)
	      (newline)))
	#f)))

(define (non-tautological-subgoals subgoals)
  (if (pair? subgoals)
      (and (not (tautology-check (car subgoals) (cdr subgoals)))
	   (non-tautological-subgoals (cdr subgoals)))
      #t))

(define (tautology-check subgoal subgoals)
  (if (pair? subgoals)
      (if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign (car subgoals))))
	       (eq? (subgoal-pred subgoal) (subgoal-pred (car subgoals)))
	       (ref-equals (subgoal-terms subgoal) (subgoal-terms (car subgoals))))
	  #t
	  (tautology-check subgoal (cdr subgoals)))
      (if (eq? #t (subgoal-pred subgoal))
	  #t
	  #f)))

;;;*************************************************************************
;; -------------------------
;; Foldup and simplification
;; -------------------------

(define (subgoal-simplifiable subgoal)
  (if simplify
      (if (subgoal-simplified subgoal)
	  #t
	  (let (index-end)
	    (set! index-end (match-subgoal-with-index subgoal))
	    (if index-end
		(begin
		  (if (or dewey-numbers (> debug-level 4))
		      (begin
			(display "simplifying subgoal ")
			(display-subgoal subgoal)
			(display " of depth ")
			(display (subgoal-depth subgoal))
			(display " matching unit ")
			(display-clause (subgoal-clause index-end))
			(newline)))
		  (set-subgoal-simplified subgoal #t)
		  (set-clause-unsimplified (subgoal-clause subgoal) (- (clause-unsimplified (subgoal-clause subgoal)) 1))
		  (set! nr-simplifications (+ nr-simplifications 1))
		  #t)
		#f)))
      #f))

(define (root-foldup subgoal)
  (let ((foldup-clause (new-clause)) (single-var (new-variable)) (foldup-input (new-input)))
    (set-variable-name single-var "X")
    (set-clause-subgoals foldup-clause (list (foldup-subgoal-copy subgoal single-var)))
    (set-clause-variables foldup-clause (variables-in-subgoals (clause-subgoals foldup-clause) '()))
    (set-clause-nr-variables foldup-clause (length (clause-variables foldup-clause)))
    (clause-pointers-in-subgoals (clause-subgoals foldup-clause) foldup-clause)
    (set-input-copies foldup-input (list foldup-clause))       
    (set-input-length foldup-input 1)       
    (set-clause-unsimplified foldup-clause 1)
    (set-input-type foldup-input 'lemma)         
    (set-input-number foldup-input -1)       
    (set-input-name foldup-input "foldup_lemma")         
    (set-input-all-copies foldup-input (input-copies foldup-input))   
    (set-input-complexity foldup-input 1)   
    (set-input-signature foldup-input '())    
    (set-clause-input foldup-clause foldup-input)
    (set-clause-initial foldup-clause #t)
    (set-clause-relevant foldup-clause (clause-relevant (subgoal-clause subgoal)))
    (set-clause-predecessor foldup-clause #f)
    (set-subgoal-next (car (clause-subgoals foldup-clause)) (subgoal-clause input-root))
    (set-clause-path (subgoal-clause input-root) (car (clause-subgoals foldup-clause)))
    (set-clause-depth foldup-clause (- (subgoal-depth input-root) 1))
    (set! input-root (car (clause-subgoals foldup-clause)))
    (insert-subgoal-in-index (car (clause-subgoals foldup-clause)) 'unit)
    (insert-subgoal-in-gr-index (car (clause-subgoals foldup-clause)))
    (set! nr-foldups (+ nr-foldups 1))
    (set-input-variant-index foldup-input (cons 0 '()))
    (insert-clause-in-index foldup-clause)
    (if (or dewey-numbers (> debug-level 4))
	(begin
	  (display "folding up subgoal ")
	  (display-subgoal subgoal)
	  (display "(")
	  (display (clause-depth foldup-clause))
	  (display ")")
	  (display " of depth ")
	  (display (subgoal-depth subgoal))
	  (display " to root as ")
	  (display-subgoal (car (clause-subgoals foldup-clause)))
	  (newline)))
    (add-unit-lemma (car (clause-subgoals foldup-clause)))
    (insert-lemma-links (car (clause-subgoals foldup-clause)) (subgoal-predecessor subgoal))))

(define (foldup-subgoal-copy subgoal single-var)
  (let ((sg (new-subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set-subgoal-sign sg '-)
	(set-subgoal-sign sg '+))
    (set-subgoal-pred sg (subgoal-pred subgoal))
    (set-subgoal-terms sg (foldup-termlist-copy (subgoal-terms subgoal) single-var))
    (set-subgoal-selected sg #f)
    (set-subgoal-number sg 1)
    sg))

(define (foldup-termlist-copy terms single-var)
  (if (null? terms)
      '()
      (cons (foldup-term-copy (ref (car terms)) single-var) (foldup-termlist-copy (cdr terms) single-var))))

(define (foldup-term-copy term single-var)
  (if (variable term)
      single-var
      (if (constant term)
	  term
	  (cons (car term) (foldup-termlist-copy (cdr term) single-var)))))

(define (make-top-down-lemma subgoal)
  (let ((lemma-clause (new-clause)) (lemma-input (new-input)))
    (set-clause-subgoals lemma-clause (list (lemma-subgoal-copy subgoal)))
    (set-clause-variables lemma-clause (variables-in-subgoals (clause-subgoals lemma-clause) '()))
    (set-clause-nr-variables lemma-clause (length (clause-variables lemma-clause)))
    (clause-pointers-in-subgoals (clause-subgoals lemma-clause) lemma-clause)
    (set-input-copies lemma-input (list lemma-clause))       
    (set-input-length lemma-input 1)       
    (set-clause-unsimplified lemma-clause 1)
    (set-input-type lemma-input 'lemma)         
    (set-input-number lemma-input -1)       
    (set-input-name lemma-input "simplification_lemma")         
    (set-input-all-copies lemma-input (input-copies lemma-input))   
    (set-input-complexity lemma-input 1)   
    (set-input-signature lemma-input '())    
    (set-clause-input lemma-clause lemma-input)
    (set-clause-initial lemma-clause #t)
    (set-clause-relevant lemma-clause (clause-relevant (subgoal-clause subgoal)))
    (set-clause-predecessor lemma-clause #f)
    (set-subgoal-next (car (clause-subgoals lemma-clause)) (subgoal-clause input-root))
    (set-clause-path (subgoal-clause input-root) (car (clause-subgoals lemma-clause)))
    (set-clause-depth lemma-clause (- (subgoal-depth input-root) 1))
    (set! input-root (car (clause-subgoals lemma-clause)))
    (insert-subgoal-in-index (car (clause-subgoals lemma-clause)) 'unit)
    (insert-subgoal-in-gr-index (car (clause-subgoals lemma-clause)))
    (set! nr-top-down-lemmas (+ nr-top-down-lemmas 1))
    (set-input-variant-index lemma-input (cons 0 '()))
    (insert-clause-in-index lemma-clause)
    (if (or dewey-numbers (> debug-level 4))
	(begin
	  (display "making simplification lemma ")
	  (display-subgoal subgoal)
	  (display "(")
	  (display (clause-depth lemma-clause))
	  (display ")")
	  (display " on depth ")
	  (display (subgoal-depth subgoal))
	  (newline)))
    (add-unit-lemma (car (clause-subgoals lemma-clause)))
    (insert-lemma-links (car (clause-subgoals lemma-clause)) (subgoal-predecessor subgoal))))

(define (lemma-subgoal-copy subgoal)
  (let ((sg (new-subgoal)))
    (set-subgoal-sign sg (subgoal-sign subgoal))
    (set-subgoal-pred sg (subgoal-pred subgoal))
    (set-subgoal-terms sg (renamed-termlist (subgoal-terms subgoal)))
    (set-subgoal-selected sg #f)
    (set-subgoal-simplified sg #f)
    (set-subgoal-number sg 1)
    sg))

(define (simplified-unit subgoal)
  (if (subgoal-simplified subgoal)
      #f
      (if (and (= (clause-unsimplified (subgoal-clause subgoal)) 1) (not (unit-clause (subgoal-clause subgoal))))
	  #t
	  #f)))

(define (foldup-subgoal subgoal rel-vector)
  (let (first-level)
    (set! first-level (foldup-depth rel-vector (subgoal-depth subgoal)))
    (if (< first-level 0)
	  (root-foldup subgoal)
	(begin
	  (if context-lemmas (context-foldup subgoal (+ first-level 1) rel-vector))
	  #f))))

(define (context-foldup subgoal depth rel-vector)
  (let (sg path-clause new-vector (single-var (new-variable)))
    (set! sg (foldup-subgoal-copy subgoal single-var))
    (set-subgoal-number sg depth) ;; misuse of subgoal-number to store the depth of the clauseless foldup lemma
    (set! new-vector (copy-up-to rel-vector depth))
    (set-subgoal-links sg new-vector)  ;; misuse of subgoal-links to store the lemma's rel-vector
    (insert-subgoal-in-index sg 'path) ;; use of path index as gr index is ok, as sg contains at most one variable
    (set! path-clause (path-vector-ref depth))
    (set-clause-foldup-lits path-clause (cons sg (clause-foldup-lits path-clause)))
    (set! nr-context-foldups (+ nr-context-foldups 1))
    (if dewey-numbers
	(begin
	  (display "folding up context subgoal ")
	  (display-subgoal subgoal)
	  (display " of depth ")
	  (display (subgoal-depth subgoal))
	  (display " to depth ")
	  (display depth)
	  (display " as ")
	  (display-subgoal sg)
	  (newline)))))

(define (delete-clause-foldup-lits clause)
  (if (clause-foldup-lits clause)
      (begin
	(delete-foldup-lits (clause-foldup-lits clause))
	(set-clause-foldup-lits clause #f))))

(define (delete-foldup-lits subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-links (car subgoals) #f)
	(delete-subgoal-from-index (car subgoals) 'path)
	(set-car! subgoals #f)
	(delete-foldup-lits (cdr subgoals)))))
