;;;*************************************************************************
;; ---------------------------------------------------------------
;; FORMULA, CLAUSE, SUBGOAL, AND SIGNATURE ACCESS AND MODIFICATION
;; ---------------------------------------------------------------

;; ---------------------------------------------------------------
;; Bit vector access masks
;; ---------------------------------------------------------------

(define (bit_0) `1)
(define (bit_1) `2)
(define (bit_2) `4)
(define (bit_3) `8)
(define (bit_4) `16)
(define (bit_5) `32)
(define (bit_6) `64)
(define (bit_7) `128)
(define (bit_8) `256)
(define (bit_9) `512)

(define (not_bit_0) `-2)
(define (not_bit_1) `-3)
(define (not_bit_2) `-5)
(define (not_bit_3) `-9)
(define (not_bit_4) `-17)
(define (not_bit_5) `-33)
(define (not_bit_6) `-65)
(define (not_bit_7) `-129)
(define (not_bit_8) `-257)
(define (not_bit_9) `-513)

;; ---------------------------------------------------------------
;; CNF formula data structure
;; ---------------------------------------------------------------

(define-macro (formula-clauses f)       `(vector-ref ,f 0))      
(define-macro (formula-statistics f)    `(vector-ref ,f 1))      
(define-macro (formula-lemmas f)        `(vector-ref ,f 2))      

(define-macro (set-formula-clauses f arg)     `(vector-set! ,f 0 ,arg))      
(define-macro (set-formula-statistics f arg)  `(vector-set! ,f 1 ,arg))      
(define-macro (set-formula-lemmas f arg)      `(vector-set! ,f 2 ,arg))      

;; ---------------------------------------------------------------
;; Tableau clause data structure
;; ---------------------------------------------------------------

(define-macro (new-clause) `(let ((v (make-vector 8 #f))) (vector-set! v 7 0) v))

(define-macro (clause-subgoals cl)        `(vector-ref ,cl 0))      
(define-macro (clause-foldup-lits cl)     `(vector-ref ,cl 1))      
(define-macro (clause-variables cl)       `(vector-ref ,cl 2))      
(define-macro (clause-path cl)            `(vector-ref ,cl 3))      
(define-macro (clause-input cl)           `(vector-ref ,cl 4))      
(define-macro (clause-depth cl)           `(vector-ref ,cl 5))      
(define-macro (clause-unsimplified cl)    `(vector-ref ,cl 6))      
(define-macro (clause-relevant cl)        `(not (= (bit-and (bit_0) (vector-ref ,cl 7)) 0)))
(define-macro (clause-initial cl)         `(not (= (bit-and (bit_1) (vector-ref ,cl 7)) 0)))

(define-macro (set-clause-subgoals cl arg)        `(vector-set! ,cl 0 ,arg)) 
(define-macro (set-clause-foldup-lits cl arg)     `(vector-set! ,cl 1 ,arg))
(define-macro (set-clause-variables cl arg)       `(vector-set! ,cl 2 ,arg))
(define-macro (set-clause-path cl arg)            `(vector-set! ,cl 3 ,arg))
(define-macro (set-clause-input cl arg)           `(vector-set! ,cl 4 ,arg))
(define-macro (set-clause-depth cl arg)           `(vector-set! ,cl 5 ,arg))
(define-macro (set-clause-unsimplified cl arg)    `(vector-set! ,cl 6 ,arg))
(define-macro (set-clause-relevant cl arg)        `(if ,arg 
						       (vector-set! ,cl 7 (bit-or (bit_0) (vector-ref ,cl 7)))
						       (vector-set! ,cl 7 (bit-and (not_bit_0) (vector-ref ,cl 7)))))
(define-macro (set-clause-initial cl arg)         `(if ,arg 
						       (vector-set! ,cl 7 (bit-or (bit_1) (vector-ref ,cl 7)))
						       (vector-set! ,cl 7 (bit-and (not_bit_1) (vector-ref ,cl 7)))))

;; ---------------------------------------------------------------
;; Input clause data structure
;; ---------------------------------------------------------------

(define-macro (new-input) `(make-vector 5 #f))

(define-macro (input-type in)          `(vector-ref ,in 0))      
(define-macro (input-number in)        `(vector-ref ,in 1))      
(define-macro (input-name in)          `(vector-ref ,in 2))      
(define-macro (input-variant-index in) `(vector-ref ,in 3))      
(define-macro (input-in-tableau in)    `(vector-ref ,in 4))      

(define-macro (set-input-type in arg)          `(vector-set! ,in 0 ,arg))
(define-macro (set-input-number in arg)        `(vector-set! ,in 1 ,arg))
(define-macro (set-input-name in arg)          `(vector-set! ,in 2 ,arg))
(define-macro (set-input-variant-index in arg) `(vector-set! ,in 3 ,arg))
(define-macro (set-input-in-tableau in arg)    `(vector-set! ,in 4 ,arg))

;; ---------------------------------------------------------------
;; Subgoal data structure
;; ---------------------------------------------------------------

(define-macro (subgoal-size) `9)
(define-macro (new-subgoal) `(let ((v (make-vector (subgoal-size) #f))) (vector-set! v 8 0) v))
(define-macro (subgoal? sg) `(and (vector? ,sg) (= (vector-length ,sg) (subgoal-size))))
(define-macro (propositional-subgoal? sg) `(and (subgoal? ,sg) (= (predsym-arity (subgoal-pred ,sg)) 0)))
(define-macro (context-lemma? sg) `(eq? (input-type (clause-input (subgoal-clause ,sg))) 'clemma))

(define-macro (subgoal-misc sg)                 `(vector-ref ,sg 0))
(define-macro (subgoal-pred sg)                 `(vector-ref ,sg 1))
(define-macro (subgoal-terms sg)                `(vector-ref ,sg 2))
(define-macro (subgoal-links sg)                `(vector-ref ,sg 3))
(define-macro (subgoal-clause sg)               `(vector-ref ,sg 4))
(define-macro (subgoal-weight sg)               `(vector-ref ,sg 5))
(define-macro (subgoal-next sg)                 `(vector-ref ,sg 6))
(define-macro (subgoal-local-instances sg)      `(vector-ref ,sg 7))
;; bit vector flag access macros
(define-macro (subgoal-sign sg)                 `(if (= (bit-and (bit_0) (vector-ref ,sg 8)) 0) '- '+))
(define-macro (subgoal-selected sg)             `(not (= (bit-and (bit_1) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-unlinkable sg)           `(not (= (bit-and (bit_2) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-ordered-eq sg)           `(not (= (bit-and (bit_3) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-simplified sg)           `(not (= (bit-and (bit_4) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-ground sg)               `(not (= (bit-and (bit_5) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-local sg)                `(not (= (bit-and (bit_6) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-semi-local sg)           `(not (= (bit-and (bit_7) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-split-lit sg)            `(not (= (bit-and (bit_8) (vector-ref ,sg 8)) 0)))
(define-macro (subgoal-expansion-lit sg)        `(not (= (bit-and (bit_9) (vector-ref ,sg 8)) 0)))

(define-macro (subgoal-depth sg)                `(clause-depth (subgoal-clause ,sg)))
(define-macro (subgoal-predecessor sg)          `(clause-path (subgoal-clause ,sg)))
(define-macro (subgoal-clause-number sg)        `(input-number (clause-input (subgoal-clause ,sg))))

(define-macro (lemma-rel-vector sg)             `(clause-path (subgoal-clause ,sg)))

(define-macro (set-subgoal-misc sg arg)                `(vector-set! ,sg 0 ,arg))
(define-macro (set-subgoal-pred sg arg)                `(vector-set! ,sg 1 ,arg))
(define-macro (set-subgoal-terms sg arg)               `(vector-set! ,sg 2 ,arg))
(define-macro (set-subgoal-links sg arg)               `(vector-set! ,sg 3 ,arg))
(define-macro (set-subgoal-clause sg arg)              `(vector-set! ,sg 4 ,arg))
(define-macro (set-subgoal-weight sg arg)              `(vector-set! ,sg 5 ,arg))
(define-macro (set-subgoal-next sg arg)                `(vector-set! ,sg 6 ,arg))
(define-macro (set-subgoal-local-instances sg arg)     `(vector-set! ,sg 7 ,arg))
;; bit vector flag access macros
(define-macro (set-subgoal-sign sg arg)                `(if (eq? ,arg '+)
							    (vector-set! ,sg 8 (bit-or (bit_0) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_0) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-selected sg arg)            `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_1) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_1) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-unlinkable sg arg)          `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_2) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_2) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-ordered-eq sg arg)          `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_3) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_3) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-simplified sg arg)          `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_4) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_4) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-ground sg arg)              `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_5) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_5) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-local sg arg)               `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_6) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_6) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-semi-local sg arg)          `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_7) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_7) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-split-lit sg arg)           `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_8) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_8) (vector-ref ,sg 8)))))
(define-macro (set-subgoal-expansion-lit sg arg)       `(if ,arg 
							    (vector-set! ,sg 8 (bit-or (bit_9) (vector-ref ,sg 8)))
							    (vector-set! ,sg 8 (bit-and (not_bit_9) (vector-ref ,sg 8)))))

(define-macro (set-lemma-rel-vector sg arg)             `(set-clause-path (subgoal-clause ,sg) ,arg))

;; ---------------------------------------------------------------
;; Term data structures
;; ---------------------------------------------------------------

(define-macro (new-predsym) `(make-vector 16 #f))

(define-macro (predsym-symbol p)          `(vector-ref ,p 0))
(define-macro (predsym-arity p)           `(vector-ref ,p 1))
(define-macro (predsym-number p)          `(vector-ref ,p 2))
(define-macro (predsym-weight p)          `(vector-ref ,p 3))
(define-macro (predsym-pos-context-index p)  `(vector-ref ,p 4))
(define-macro (predsym-neg-context-index p)  `(vector-ref ,p 5))
(define-macro (predsym-pos-gr-index p)    `(vector-ref ,p 6))
(define-macro (predsym-neg-gr-index p)    `(vector-ref ,p 7))
(define-macro (predsym-pos-unit-index p)  `(vector-ref ,p 8))
(define-macro (predsym-neg-unit-index p)  `(vector-ref ,p 9))
(define-macro (predsym-pos-reg-index p)   `(vector-ref ,p 10))
(define-macro (predsym-neg-reg-index p)   `(vector-ref ,p 11))
(define-macro (predsym-pos-path-index p)  `(vector-ref ,p 12))
(define-macro (predsym-neg-path-index p)  `(vector-ref ,p 13))
(define-macro (predsym-pos-purity-index p)  `(vector-ref ,p 14))
(define-macro (predsym-neg-purity-index p)  `(vector-ref ,p 15))

(define-macro (set-predsym-symbol p arg)          `(vector-set! ,p 0 ,arg))
(define-macro (set-predsym-arity p arg)           `(vector-set! ,p 1 ,arg))
(define-macro (set-predsym-number p arg)          `(vector-set! ,p 2 ,arg))
(define-macro (set-predsym-weight p arg)          `(vector-set! ,p 3 ,arg))
(define-macro (set-predsym-pos-context-index p arg)  `(vector-set! ,p 4 ,arg))
(define-macro (set-predsym-neg-context-index p arg)  `(vector-set! ,p 5 ,arg))
(define-macro (set-predsym-pos-gr-index p arg)    `(vector-set! ,p 6 ,arg))
(define-macro (set-predsym-neg-gr-index p arg)    `(vector-set! ,p 7 ,arg))
(define-macro (set-predsym-pos-unit-index p arg)  `(vector-set! ,p 8 ,arg))
(define-macro (set-predsym-neg-unit-index p arg)  `(vector-set! ,p 9 ,arg))
(define-macro (set-predsym-pos-reg-index p arg)   `(vector-set! ,p 10 ,arg))
(define-macro (set-predsym-neg-reg-index p arg)   `(vector-set! ,p 11 ,arg))
(define-macro (set-predsym-pos-path-index p arg)  `(vector-set! ,p 12 ,arg))
(define-macro (set-predsym-neg-path-index p arg)  `(vector-set! ,p 13 ,arg))
(define-macro (set-predsym-pos-purity-index p arg)  `(vector-set! ,p 14 ,arg))
(define-macro (set-predsym-neg-purity-index p arg)  `(vector-set! ,p 15 ,arg))

(define-macro (new-functor) `(make-vector 4 #f))

(define-macro (functor-symbol sg)     `(vector-ref ,sg 0))
(define-macro (functor-arity sg)      `(vector-ref ,sg 1))
(define-macro (functor-number sg)     `(vector-ref ,sg 2))
(define-macro (functor-weight sg)     `(vector-ref ,sg 3))

(define-macro (set-functor-symbol sg arg)     `(vector-set! ,sg 0 ,arg))
(define-macro (set-functor-arity sg arg)      `(vector-set! ,sg 1 ,arg))
(define-macro (set-functor-number sg arg)     `(vector-set! ,sg 2 ,arg))
(define-macro (set-functor-weight sg arg)     `(vector-set! ,sg 3 ,arg))

;; ---------------------------------------------------------------
;; Signature data structure
;; ---------------------------------------------------------------

(define-macro (new-signature) `(make-vector 4 #f))

(define-macro (signature-symbol sg)     `(vector-ref ,sg 0))
(define-macro (signature-arity sg)      `(vector-ref ,sg 1))
(define-macro (signature-number sg)     `(vector-ref ,sg 2))
(define-macro (signature-weight sg)     `(vector-ref ,sg 3))

(define-macro (set-signature-symbol sg arg)     `(vector-set! ,sg 0 ,arg))
(define-macro (set-signature-arity sg arg)      `(vector-set! ,sg 1 ,arg))
(define-macro (set-signature-number sg arg)     `(vector-set! ,sg 2 ,arg))
(define-macro (set-signature-weight sg arg)     `(vector-set! ,sg 3 ,arg))

;; ---------------------------------------------------------------
;; Variable data structure
;; ---------------------------------------------------------------

(define-macro (new-variable) `(cons '() (cons #f #f)))

(define-macro (variable-name var)             `(cddr ,var))
(define-macro (variable-renaming var)         `(cadr ,var))
(define-macro (variable-binding var)          `(car ,var))

(define-macro (set-variable-name var arg)     `(set-cdr! (cdr ,var) ,arg))
(define-macro (set-variable-renaming var arg) `(set-car! (cdr ,var) ,arg))
(define-macro (set-variable-binding var arg)  `(set-car! ,var ,arg))

;; ---------------------------------------------------------------
;; Link Data Structure
;; ---------------------------------------------------------------

(define-macro (new-link) `(make-vector 5 #f))
(define-macro (new-disagreement-link) `(make-vector 6 #f))
(define-macro (new-eq-link) `(make-vector 8 #f))

(define-macro (disagreement-link? link) `(= (vector-length ,link) 6))
(define-macro (eq-link? link)  `(= (vector-length ,link) 8))

(define-macro (link-lita link)              `(vector-ref ,link 0))
(define-macro (link-litb link)              `(vector-ref ,link 1))
(define-macro (link-active link)            `(vector-ref ,link 2))
(define-macro (link-weight link)            `(vector-ref ,link 3))
(define-macro (link-depth link)             `(vector-ref ,link 4))
(define-macro (link-position link)          `(if (eq-link? ,link) (vector-ref ,link 5) #f))
(define-macro (link-misc link)              `(if (disagreement-link? ,link) (vector-ref ,link 5) #f)) ;; for disagreement linking
(define-macro (link-left-to-right link)     `(if (eq-link? ,link) (vector-ref ,link 6) #f))
(define-macro (link-rewriting link)         `(if (eq-link? ,link) (vector-ref ,link 7) #f))

(define-macro (set-link-lita link arg)      `(vector-set! ,link 0 ,arg))
(define-macro (set-link-litb link arg)      `(vector-set! ,link 1 ,arg))
(define-macro (set-link-active link arg)    `(vector-set! ,link 2 ,arg))
(define-macro (set-link-weight link arg)    `(vector-set! ,link 3 ,arg))
(define-macro (set-link-depth link arg)     `(vector-set! ,link 4 ,arg))
(define-macro (set-link-position link arg)  `(if (eq-link? ,link) (vector-set! ,link 5 ,arg)))
(define-macro (set-link-misc link arg)      `(if (disagreement-link? ,link) (vector-set! ,link 5 ,arg))) ;; for disagreement linking
(define-macro (set-link-left-to-right link arg)     `(if (eq-link? ,link) (vector-set! ,link 6 ,arg)))
(define-macro (set-link-rewriting link arg)         `(if (eq-link? ,link) (vector-set! ,link 7 ,arg)))


;;;*************************************************************************
;; ---------------
;; UNIT TESTS
;; ---------------

(define-macro (unit-subgoal sg) `(= 1 (clause-unsimplified (subgoal-clause ,sg))))
(define-macro (unit-clause cl)  `(= 1 (clause-unsimplified ,cl)))
(define-macro (non-unit-subgoal sg) `(< 1 (clause-unsimplified (subgoal-clause ,sg))))
(define-macro (non-unit-clause cl)  `(< 1 (clause-unsimplified ,cl)))

;;;*************************************************************************
;; ---------------
;; TERM TYPE TESTS
;; ---------------

(define-macro (constant term) `(and (pair? ,term) (null? (cdr ,term))))

(define-macro (complex term) 
`(and (pair? ,term) (not (variable ,term)) (pair? (cdr ,term))))

(define-macro (variable term) `(and (pair? ,term) (not (vector? (car ,term))) (pair? (cdr ,term))))

(define-macro (local-variable term) `(and (pair? ,term) (not (vector? (car ,term))) (pair? (cdr ,term)) (pair? (cddr ,term))))

(define-macro (shared-variable term) `(and (pair? ,term) (not (vector? (car ,term))) (pair? (cdr ,term)) (not (pair? (cddr ,term)))))

(define-macro (domain-element term) `(and (pair? ,term) (not (cdr ,term))))

(define-macro (non-variable term) `(and (pair? ,term) (vector? (car ,term))))

(define-macro (free-variable term) `(and (pair? ,term) (eq? (car ,term) '())))

(define-macro (bound-variable term) `(and (pair? ,term) (or (integer? ,term) (pair? (car ,term)))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; VARIABLE BINDING, EQUATION EXTRACTION AND TRAIL MODIFICATION
;; ------------------------------------------------------------

(define-macro (push-trail data)
`(begin (trail-next) (set-car! (car trail) ,data)))

(define-macro (trail-next)
`(if (null? (cdr trail))
     (begin (set-cdr! trail (list (cons #f trail)))
            (set! trail (cdr trail)))
     (set! trail (cdr trail))))

(define-macro (trail-stop)
`(begin (trail-next) (set-car! (car trail) #f)))

(define-macro (trail-previous) `(set! trail (cdar trail)))

(define-macro (trail-content) `(caar trail))

(define-macro (bind var term)
;; binds var to term destructively, next trail position is used
`(begin (set-car! ,var ,term)
        (trail-next)
        (set-car! (car trail) ,var)
	#t))

(define-macro (bind-var var term) `(set-car! ,var ,term))

(define-macro (unbind-var var) `(set-car! ,var '()))

(define-macro (unify-lists-stop k l)
`(begin (trail-stop) (unify-lists ,k ,l)))

(define-macro (unify-terms-stop s t)
`(begin (trail-stop) (unify-terms ,s ,t)))

(define-macro (extension-unify-lists-stop k l)
`(begin (trail-stop) 
        (set! optimized-unification 0) 
        (extension-unify-lists ,k ,l)))

(define-macro (match-lists-stop k l)
`(begin (trail-stop) (match-termlists ,k ,l)))

(define-macro (match-terms-stop s t)
`(begin (trail-stop) (match-terms ,s ,t)))

(define-macro (trail-subgoal sg)
;; binds var to term destructively, next trail position is used
`(begin
   (trail-next)
   (set-car! (car trail) ,sg)
   #t))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index macros
;; ------------------------------------------------------------

;; Index macros need to be partially adapted to support a count of valid sub-and-cells

(define-macro (new-or-and-cell)             `(cons (cons #f '()) '()))

(define-macro (and-content i)               `(caar ,i))      
(define-macro (and-next i)                  `(cdar ,i))      
(define-macro (or-current i)                `(car ,i))      
(define-macro (or-next i)                   `(cdr ,i))      

(define-macro (set-and-content i arg)       `(set-car! (car ,i) ,arg))      
(define-macro (set-and-next i arg)          `(set-cdr! (car ,i) ,arg))      
(define-macro (set-or-current i arg)        `(set-car! ,i ,arg))      
(define-macro (set-or-next i arg)           `(set-cdr! ,i ,arg))      

(define-macro (index-variable t) `(integer? ,t))
(define-macro (index-functor t)  `(vector? ,t))
(define-macro (index-constant t) `(and (vector? ,t) (= (functor-arity ,t) 0)))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Term trail modification
;; ------------------------------------------------------------

;; New macros for the term trail are needed

(define-macro (push-term-trail data)
`(begin (term-trail-next) (set-car! (car term-trail) ,data)))

(define-macro (pop-term-trail)
`(if (term-trail-content)
     (begin
       (set! term-trail (cdar term-trail))
       (caadr term-trail))
     #f))

(define-macro (term-trail-next)
`(if (null? (cdr term-trail))
     (begin (set-cdr! term-trail (list (cons #f term-trail)))
            (set! term-trail (cdr term-trail)))
     (set! term-trail (cdr term-trail))))

(define-macro (term-trail-stop)
`(begin (term-trail-next) (set-car! (car term-trail) #f)))

(define-macro (term-trail-previous) `(set! term-trail (cdar term-trail)))

(define-macro (term-trail-content) `(caar term-trail))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Equality macros
;; ------------------------------------------------------------

(define-macro (eq-subgoal? subgoal) `(eq? (predsym-symbol (subgoal-pred ,subgoal)) 'equal))

(define-macro (equation? subgoal)
  `(and (eq? (subgoal-sign ,subgoal) '+)
       (eq? (predsym-symbol (subgoal-pred ,subgoal)) 'equal)))

(define-macro (disequation? subgoal)
  `(and (eq? (subgoal-sign ,subgoal) '-)
       (eq? (predsym-symbol (subgoal-pred ,subgoal)) 'equal)))
