(module path
	(include "dctp.sch")
	(export path-vector-size)
	(export (initialize-path-literal-vectors) (add-path-literal subgoal) (add-unit-lemma subgoal) 
		(delete-path-literal subgoal) (potential-linked-lits subgoal) (potential-linked-lemmas subgoal) 
		(positive-eq-lits) (positive-eq-lemmas) (unit-eq-lits) (enlarge-path-literal-vectors))
	(import (globals "globals.scm") (preprocess "preprocess.scm") (display "display.scm")
		(nr-initial-predicates parser "parser.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define pos-path-literals #f)
(define neg-path-literals #f)
(define pos-path-lemmas #f)
(define neg-path-lemmas #f)
(define path-vector-size 0)

(define unit-equations '())

;;;*************************************************************************
;; ------------------------------------------------------------
;; Initialization
;; ------------------------------------------------------------

(define (initialize-path-literal-vectors)
  (set! pos-path-literals (make-vector (+ 1 nr-initial-predicates) '()))
  (set! neg-path-literals (make-vector (+ 1 nr-initial-predicates) '()))
  (set! pos-path-lemmas (make-vector (+ 1 nr-initial-predicates) '()))
  (set! neg-path-lemmas (make-vector (+ 1 nr-initial-predicates) '()))
  (set! path-vector-size (+ 1 nr-initial-predicates)))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Initialization
;; ------------------------------------------------------------

(define (enlarge-path-literal-vectors)
  (let ((old-length (vector-length pos-path-literals)))
    (set! pos-path-literals (copy-vector pos-path-literals (* 2 old-length)))
    (set! neg-path-literals (copy-vector neg-path-literals (* 2 old-length)))
    (set! pos-path-lemmas (copy-vector pos-path-lemmas (* 2 old-length)))
    (set! neg-path-lemmas (copy-vector neg-path-lemmas (* 2 old-length)))
    (set! path-vector-size (* 2 old-length))
    (do ((i old-length (+ i 1)))
	((= i (* 2 old-length)) #t)
      (vector-set! pos-path-literals i '())
      (vector-set! neg-path-literals i '())
      (vector-set! pos-path-lemmas i '())
      (vector-set! neg-path-lemmas i '()))))
    

;;;*************************************************************************
;; ------------------------------------------------------------
;; Adding literals
;; ------------------------------------------------------------

(define (add-path-literal subgoal)
  (let (lists pred-num)
    (if (and (unit-subgoal subgoal) (equation? subgoal))
	(set! unit-equations (cons subgoal unit-equations)))
    (set! pred-num (predsym-number (subgoal-pred subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! lists pos-path-literals)
	(set! lists neg-path-literals))
    (if (null? (vector-ref lists pred-num))
	(let ((new-sg (cons subgoal '())))
	  (vector-set! lists pred-num new-sg))
	(vector-set! lists pred-num (cons subgoal (vector-ref lists pred-num))))))

(define (add-unit-lemma subgoal)
  (let (lists pred-num)
    (if (equation? subgoal)
	(set! unit-equations (cons subgoal unit-equations)))
    (set! pred-num (predsym-number (subgoal-pred subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! lists pos-path-lemmas)
	(set! lists neg-path-lemmas))
    (if (null? (vector-ref lists pred-num))
	(let ((new-sg (cons subgoal '())))
	  (vector-set! lists pred-num new-sg))
	(vector-set! lists pred-num (cons subgoal (vector-ref lists pred-num))))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Removing literals
;; ------------------------------------------------------------

(define (delete-path-literal subgoal)  
  (let (lists list-ends pred-num)
    (if (and (unit-subgoal subgoal) (equation? subgoal))
	(set! unit-equations (remq! subgoal unit-equations)))
    (set! pred-num (predsym-number (subgoal-pred subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! lists pos-path-literals)
	(set! lists neg-path-literals))
    (if (or (null? (vector-ref lists pred-num)) (not (eq? subgoal (car (vector-ref lists pred-num)))))
	(begin
	  (display "path literal deletion error for subgoal ")
	  (display-subgoal subgoal)
	  (display " on depth ")
	  (display (subgoal-depth  subgoal))
	  (newline)
	  (if (not (null? (vector-ref lists pred-num)))
	      (begin
		(display "found subgoal ")
		(display-subgoal (car (vector-ref lists pred-num)))
		(display " of depth ")
		(display (subgoal-depth (car (vector-ref lists pred-num))))
		(newline))))
	(begin
	  ;;(set-car! (vector-ref lists pred-num) #f)
	  (vector-set! lists pred-num (cdr (vector-ref lists pred-num)))))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Obtaining the list of possible linked path literals
;; ------------------------------------------------------------

(define (potential-linked-lits subgoal)
  (if (eq? '+ (subgoal-sign subgoal))
      (vector-ref neg-path-literals (predsym-number (subgoal-pred subgoal)))
      (vector-ref pos-path-literals (predsym-number (subgoal-pred subgoal)))))

(define (potential-linked-lemmas subgoal)
  (if (eq? '+ (subgoal-sign subgoal))
      (vector-ref neg-path-lemmas (predsym-number (subgoal-pred subgoal)))
      (vector-ref pos-path-lemmas (predsym-number (subgoal-pred subgoal)))))

(define (positive-eq-lits)
      (vector-ref pos-path-literals (predsym-number equality-pred)))

(define (positive-eq-lemmas)
      (vector-ref pos-path-lemmas (predsym-number equality-pred)))

(define (unit-eq-lits)
  unit-equations)
