(* $Id: context_unifier_search.mli,v 1.3 2004/12/28 14:06:47 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** search for context unifiers

    All found context unifiers are checked with {!Context_unifier_check},
    and if they pass, added to the candidate set
    by calling the appropriate callback functions of
    {!Context_unifier.context_unifier_space}.
 *)


(** {6 Types} *)


type config = Config.config
type bound = Bound.bound
type subst =  Subst.subst
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type context_partner = Context_unifier.context_partner



(** {2 Search} *)

(** search for context unifiers. *)
module type Search = 
  sig
    (** [search_context_unifiers config bound
	space partners fixed_index active_partner subst non_empty_remainder]
	searchs for all context unifiers of the clause [space]
	with the current context and the new context literal [context_partner].

	The search must have already been initialized by unifying
	[context_partner] with the clause literal at position [fixed_index]
	in the current ordering	of [space],
	yielding the initial partial context unifier [subst].
	[partners] must be set to [context_partner] at position [fixed_index],
	If [subst] already has a non-empty remainder,
	then [non_empty_remainder] must be set to true.
    *)
    val search_context_unifiers:
      config ->
      bound ->
      context_unifier_space ->
      context_partners ->
      int ->
      context_partner ->
      subst ->
      bool ->
      bool -> (* closing already found *)
      bool (* found closing context unifier candidate? *)
  end


module SearchClose: Search

(** search for {e Assert} and {e Close} context unifiers only. *)
module SearchAssert: Search

(** search for {e Split} context unifiers only. *)
module SearchSplit: Search

module SearchCloseAssert: Search

(** search for {e Assert}, {e Close}, and {e Split} context unifiers. *)
module SearchAll: Search
