(* $Id: context_unifier_space.mli,v 1.16 2004/12/28 14:06:47 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** a problem clause

    Stores a clause with some additional information
    like a unique id, shared variables, and so on,
    as specified in {!Context_unifier.context_unifier_space}.

    Registers all clause literals at {!Problem_literals},
    and initiates the search for new context unifiers
    when informed by {!Problem_literals} that a new literal
    has been added to the context.
*)

(** {6 Types} *)

type config = Config.config
type statistic = Statistic.statistic
type bound = Bound.bound
type literal = Term.literal
type clause = Term.clause
type state = State.state
type context_unifier_space = Context_unifier.context_unifier_space
type raw_context_unifier = Context_unifier.raw_context_unifier
type problem_literals = Problem_literals.problem_literals


(** {6 Functions} *)


(** {2 Creation} *)

(** [create: config state statistic bound problem_literals clause
    process_close_candidate process_assert_candidate process_split_candidate is_lemma]
  - creates a {!Context_unifier_space.context_unifier_space} for a {!Term.clause}
    using [process_close_candidate] resp. [process_assert_candidate]
    resp. [process_split_candidate]
    as callback functions to be called
    when assert resp. split candidates are computed.
  - and registers the literals at {!Problem_literals}.
*)
val create:
  config ->
  state ->
  statistic ->
  bound ->
  problem_literals ->
  clause ->
  (raw_context_unifier -> unit) ->
  (raw_context_unifier -> bool) ->
  (raw_context_unifier -> unit) ->
  bool ->
  context_unifier_space



(** {2 Backtracking} *)

(** for means of efficiency dependent on {!Problem_literals.backtrack},
    as the resolve information per input clause literal is used. *)
val backtrack: config -> context_unifier_space -> unit
