(* $Id: discrimination_tree.ml,v 1.8 2004/12/28 14:06:48 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

(* a basic implementation of a discrimination tree term indexing tree.
   Reference: Peter Graf, Term Indexing, Springer-Verlag, New York, NY, 1995

   Copyright (C) 2004 John Wheeler 
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*)

(* This implementation is based on code found in the file
   dtIndex.ml contained in the source code for KRHyper, copyright
   2003, by Universitaet Koblenz-Landau.  The file dtIndex.ml
   included the following copyright notice:

   * Copyright (C) 2003 Universitaet Koblenz-Landau
   * 
   * This file is part of KRHyper.
   * 
   * KRHyper is free software; you can redistribute it and/or modify it under the
   * terms of the GNU General Public License as published by the Free Software
   * Foundation; either version 2 of the License, or (at your option) any later
   * version.
   * 
   * KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
   * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
   * details.
   * 
   * You should have received a copy of the GNU General Public License along with
   * KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
   * Place, Suite 330, Boston, MA 02111-1307 USA
   * 
   * Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)

(* The current author (John Wheer) has adapted the basic implementation of discrimination tree
   term indexing found in the dtIndex.ml file for use in the darwin program and 
   extended the implementation to handle both universal and paremetric variables.  
   This adaptation and extension retained the basic outlines of implementing functions, 
   but rewrote both the call interfaces to match the darwin term indexing 
   interface (see term_index.ml* for details) and the code that handles the term and
   data items that are stored in the darwin term index.  Approximately 70% 
   of the code in the file dtIndex.ml has been modified or rewritten and approximately 
   300 lines of code of additional code were created to implement extensions to 
   address universal and parametric variables.
*)

   

(* NOTE: after changing this from a pure module to a module containing a class
   the source is kind of messy *)




(*** types ***)

type term = Term.term
type literal = Term.literal
type subst = Subst.subst
type 'a data = 'a Term_indexing.data
type 'a iterator = 'a Term_indexing.iterator

(*
let interest =
  (*Read_darwin.to_term "is_a_theorem(equivalent(equivalent(X, Y), equivalent(X, Y)))"*)
(*  Read_darwin.to_term (*"is_a_theorem(equivalent(equivalent(equivalent(_0, equivalent(equivalent(equivalent(_1, _2), equivalent(equivalent(_2, _3), _1)), _3)), _4), equivalent(_4, _0)))";;*)*)

(*    Read_darwin.to_term "product(_0, _1, _0)";;*)


(*    Read_darwin.to_term "niemalss";;*)

      (*      print_endline ("interest: " ^ Term.term_to_string interest);*)

(*  Read_darwin.to_term "a_truth(knight(me), =0)";;*)

(*Read_darwin.to_term "is_a_theorem(equivalent(equivalent(equivalent(_0, _1), equivalent(_1, equivalent(_2, _0))), equivalent(_3, equivalent(_2, _3))))";;*)

(*Read_darwin.to_term "is_a_theorem(equivalent(equivalent(equivalent(_0, _1), _2), equivalent(_2, equivalent(_1, _0))))";;*)

(*Read_darwin.to_term "less_than_or_equal(=0, q(lower_bound, =0))"*)
Read_darwin.to_term "less_than_or_equal(=0, q(lower_bound, l))"


let debug = ref false;;
*)

(* hkey *)

type hkey =
    Variable 
  | Parameter 
  | Symbol of Term.symbol 

let hkey_equal (k1: hkey) (k2: hkey) : bool =
  match k1, k2 with
      Variable, Variable ->
	true
    | Parameter, Parameter ->
	true
    | Symbol s1, Symbol s2 ->
	Symbol.equal s1 s2
    | _ ->
	false
	  
let hkey_hash (k1: hkey) : int =
  match k1 with
    | Variable ->
	0
    | Parameter ->
	1
    | Symbol s ->
	2 + Symbol.arity s

let term_hkey (term: Term.term ): hkey =
  match term with
      Term.Var var ->
	begin
	  if Var.is_universal var then
	    Variable
	  else
	    Parameter
	end
    | Term.Const const ->
	Symbol const
    | Term.Func (symbol, _, _) ->
	Symbol symbol

let hkey_arity (hkey: hkey) : int =
  match hkey with
      Symbol s ->
	Symbol.arity s
    | _ -> 0

module HKeyHashtbl =
  Hashtbl.Make(
    struct
      type t = hkey
      let equal = hkey_equal
      let hash = (*Hashtbl.hash*) hkey_hash
    end)

module type Table =
sig
  type key = hkey
  type 'a t
  val make_table: unit -> 'a t
  val add: 'a t -> key -> 'a -> unit
  val remove: 'a t -> key -> unit
  val replace: 'a t -> key -> 'a -> unit
  val find: 'a t -> key -> 'a
  val iter: (key -> 'a -> unit) -> 'a t -> unit
  val iter_and_throw_away: (key -> 'a -> unit) -> 'a t -> unit
  val fold: (key -> 'a -> 'c -> 'c) -> 'a t -> 'c -> 'c
  val length: 'a t -> int
end

module Table =
struct
  type key = hkey
  type 'a t = 'a HKeyHashtbl.t
  let make_table () = HKeyHashtbl.create 8
  let add = HKeyHashtbl.add
  let remove = HKeyHashtbl.remove
  let replace = HKeyHashtbl.replace
  let find = HKeyHashtbl.find
  let iter = HKeyHashtbl.iter
  let iter_and_throw_away f t = 
    (* Hashtables from the standard library do not to allow removal
       of entries while iterating. *)
    HKeyHashtbl.iter f t;
    HKeyHashtbl.clear t
  let fold = HKeyHashtbl.fold 
  let length = HKeyHashtbl.length
end


exception FoundVar
    
    
type fterm =
    hkey list

type 'data entry = {
  term: term;
  data: 'data;
}
      
type 'data entries = {
  mutable list: 'data entry list;
  tail: fterm;
}
    
type 'data node =
    DtInnerNode of 'data table
  | DtLeafNode of 'data entries
      
and 'data table =
    'data node Table.t
      
(*
type 'data remove =
  | RemoveDone
  | RemoveContinue
  | RemoveShrink of 'data entries
*)

(*** functions ***)

let make_entry (t: term ) (d: 'data) : 'data entry =
  {term=t ; data=d }
    
let get_term (e: 'data entry) : term =
  e.term
    
let get_data (e: 'data entry) : 'data =
  e.data


(* flatten term *)

let flatten_term term : fterm =
  let rec flatten_term' term' acc =
    match term' with
      | Term.Var var ->
	  if Var.is_universal var then
	    Variable :: acc
	  else
	    Parameter :: acc
	      
	| Term.Const symbol ->
	    Symbol symbol :: acc

	| Term.Func (symbol, terms, _) ->
	    (* flatten subterms from right to left *)
	    let new_acc =
	      Array.fold_right
		(fun term acc ->
		   flatten_term' term acc
		)
		terms
		acc
	    in
(*
	    let rec flatten_term'' acc i =
	      if i < 0 then
		acc

	      else begin
		(* flatten this subterm *)
		let acc' =
		  flatten_term' terms.(i) acc
		in
		  (* continue with next subterm *)
		  flatten_term'' acc' (i - 1)
	      end
	    in
	    let new_acc =
	      flatten_term'' acc (Array.length terms - 1)
	    in*)
	      Symbol symbol :: new_acc
  in
    flatten_term' term []
(*      
let rec flatten_term term =
  flatten_term_1 term []
    
and flatten_term_1 term rest =
  let hkey = term_hkey term in
    match term with
	Term.Func ( k, termarry, _ ) ->
	  hkey :: flatten_term_1_args
	    rest
	    termarry
	    (hkey_arity hkey)
	    0
      | _ -> hkey :: rest
	    
and flatten_term_1_args rest args arity i =
  if i >= arity then
    rest
  else
    flatten_term_1
      args.(i)
      (flatten_term_1_args
	 rest
	 args
	 arity
	 (i + 1))
*)
    
(* skip some terms in fterm. *)
(* count is the number of terms to be skipped. *)


let hkey_to_string hkey : string =
  match hkey with
      Symbol sym ->
	(Symbol.name sym) ^ "/" ^ (string_of_int (Symbol.arity sym))
    | Variable ->
	"*U/0"
    | Parameter ->
	"*P/0"

let rec fterm_to_string x : string =
    match x with
      | [] ->
	  ""

      | h :: t ->
	  hkey_to_string h ^ " " ^  fterm_to_string t
	

let rec fterm_equal x y =
  match x, y with
    | [], [] ->
	true

    | xh :: xt, yh :: yt when
	hkey_equal xh yh ->
	fterm_equal xt yt

    | _ ->
	false



let rec skip_terms_fterm fterm count =
  if count = 0 then
    fterm
  else
    match fterm with 
	hkey::fterm1 ->
	  skip_terms_fterm fterm1 (count + (hkey_arity hkey) - 1)
      | _ ->
	  raise (Failure "discrimination_tree:skip_terms_fterm: Bad flattened term structure.")



(* x can match y *)
let rec fterm_equal_generalization ~(p_preserving:bool) x y =
  match x, y with
    | [], [] ->
	true

    | xh :: xt, yh :: yt ->
	if hkey_equal xh yh then begin
	  fterm_equal_generalization ~p_preserving:p_preserving xt yt
	end

	else if
	  xh == Variable
	then begin
	  fterm_equal_generalization ~p_preserving:p_preserving xt (skip_terms_fterm y 1)
	end

	else if
	  not p_preserving
	  &&
	  xh == Parameter
	then begin
	  fterm_equal_generalization ~p_preserving:p_preserving xt (skip_terms_fterm y 1)
	end

	else begin
	  false
	end

    | _ ->
	print_endline (fterm_to_string x);
	print_endline (fterm_to_string y);
	failwith "fterm_equal_generalization"

let fterm_equal_instance ~(p_preserving:bool) x y =
  fterm_equal_generalization ~p_preserving:p_preserving y x

let rec fterm_equal_unifiable ~(p_preserving:bool) x y =
(*  print_endline ("\nfterm_equal_unifiable");
  print_endline (fterm_to_string x);
  print_endline (fterm_to_string y);*)
  match x, y with
    | [], [] ->
	true

    | xh :: xt, yh :: yt ->
	if hkey_equal xh yh then begin
	  fterm_equal_unifiable ~p_preserving:p_preserving xt yt
	end

	else if
	  xh == Variable
	then begin
	  fterm_equal_unifiable ~p_preserving:p_preserving xt (skip_terms_fterm y 1)
	end

	else if
	  yh == Variable
	then begin
	  fterm_equal_unifiable ~p_preserving:p_preserving (skip_terms_fterm x 1) yt
	end

	else if
	  not p_preserving
	  &&
	  xh == Parameter
	then begin
	  fterm_equal_unifiable ~p_preserving:p_preserving xt (skip_terms_fterm y 1)
	end

	else if
	  not p_preserving
	  &&
	  yh == Parameter
	then begin
	  fterm_equal_unifiable ~p_preserving:p_preserving (skip_terms_fterm x 1) yt
	end

	else begin
	  false
	end

    | _ ->
	print_endline (fterm_to_string x);
	print_endline (fterm_to_string y);
	failwith "fterm_equal_unifiable"


(*** instance, genralization, and unification predicates ***)
(* thanks to Alex Fuchs for pointing out how to code these *)
(* in terms of Unificaition.* funcitons. *)
    
let is_instance ?(p_preserving: bool = false) query_term candidate_term =
  (* returns true if candidate_term is an instance of query_term *)
  try
    ignore (Unification.match_terms ~p_preserving:p_preserving 
              query_term 0 candidate_term 1);
    true
  with
    | Unification.UNIFICATION_FAIL ->
        false
	  
let is_generalization ?(p_preserving: bool = false) query_term candidate_term = 
  (* returns true if candidate_term is a generalization of query_term *)
  try
    ignore(Unification.match_terms ~p_preserving:p_preserving
             candidate_term 0 query_term 1);
    true
  with
    | Unification.UNIFICATION_FAIL ->
        false
	  
let is_unifiable  ?(p_preserving: bool = false) query_term candidate_term  = 
  (* returns true if candidate_term is unifiable with query_term *)
  try
    ignore (Unification.unify_terms ~p_preserving:p_preserving
				    query_term 0 candidate_term 1);
    true
  with
    | Unification.UNIFICATION_FAIL ->
	false 
	  
	  
exception ITERATOR_EMPTY = Term_indexing.ITERATOR_EMPTY


(* an element matching the request has been found in the index.
   unfortunately, just returning it with (FOUND data)
   is not possible due to the polymorphic exception.
*)
exception FOUND



(*  class index (database: Term.database) =*)
class ['data] predicate_index (__data: 'data data) =

object (self)
    
  (**** values ***)

  (* object providing functions on the data type *)
  val _data = __data

  val mutable _root =  DtInnerNode (Table.make_table ())
  val mutable _size = 0;
  
  
  (* ugly: this simulates raise (FOUND 'data) *)
  val mutable _found : 'data option = None
					
  method private set_found data =
    _found <- Some data
      
  method private get_found =
    match _found with
      | None ->
	  failwith "Discrimination_tree.get_found"
	    
      | Some data ->
	  let data =
	    _found
	  in
	    _found <- None;
	    data
	      
	      
	      
      
      
	
  (*** creation ***)
      
  method clear =
    _root <- DtInnerNode (Table.make_table ());
    _size <- 0
      
  method size =
    _size 
      


  (*** add ***)


  method add ?(unique: bool = false) (term: term) (data: 'data) : unit =
(*    print_endline ("add: " ^ Term.term_to_string term);*)
(*    print_endline (self#to_string);*)
    match _root with
      | DtInnerNode table ->
	  let fterm =
	    flatten_term term
	  in
(*	    print_endline ("add0");*)
	    self#insert_term ~unique:unique (make_entry term data) fterm table;
	    _size <- _size + 1;
(*	    print_endline ("add done");*)


      | _ ->
	  raise (Failure "discrimination_tree:add: Bad index toplevel type.")

      

  method private insert_expand_tail (table: 'data table) (entries: 'data entries)
    (entry1: 'data entry) (hkey: hkey) (tail1: fterm) : unit =
(*    print_endline ("insert_expand");*)
    
    let tail2 : fterm =
      entries.tail
    in
    let entry2 =
      match entries.list with
	| entry2 :: _ ->
	    entry2

      | _ ->
	  failwith "discrimination_tree:insert_term:insert_expand_tail"

    in
(*	  print_endline ("insert_expand 0");*)
	  let rec expand (tail1': fterm) (tail2': fterm) (table': 'data table) : unit =
	    match tail1', tail2' with
	      | h1 :: t1, h2 :: t2 ->
(*		  print_endline ("insert_expand 1 ");*)
		  if hkey_equal h1 h2 then begin
(*		    print_endline ("insert_expand 2");*)
		    (* create new node and continue *)
		    let table'' =
		      Table.make_table ()
		    in
		    let node =
		      DtInnerNode table''
		    in
		      Table.add table' h1 node;
		      expand t1 t2 table''
		  end

		  else begin
(*		    print_endline ("insert_expand 3");*)
		    let node1 =
		      DtLeafNode {
			list = [ entry1 ];
			tail = t1;
		      }
		    in
(*		    print_endline ("insert_expand 3a");*)
		    let node2 =
		      DtLeafNode {
			list = entries.list;
			tail = t2;
		      }
		    in
(*		    print_endline ("insert_expand 3b");*)
		      Table.add table' h1 node1;
(*		    print_endline ("insert_expand 3c");*)
		      Table.add table' h2 node2;
(*		    print_endline ("insert_expand 3d");*)
		  end

	      | [], [] ->
		  failwith "discrimination_tree:insert_term:insert_expand_tail 3"

	      | _ ->
		  failwith "discrimination_tree:insert_term:insert_expand_tail 2"
	    
	  in
	    expand (hkey :: tail1) (hkey :: tail2) table;
(*	    print_endline ("insert_expand done");*)


  method private insert_term ~(unique: bool) (entry: 'data entry) (fterm: fterm) (table: 'data table) : unit =
(*    print_endline ("insert_term " ^ Term.term_to_string (get_term entry));*)
    match fterm with
      | [] ->
	  failwith "discrimination_tree:insert_term: Bad index structure 1."

      | hkey :: tail ->
	  begin
	    try 
	      begin
		let node =
		  Table.find table hkey
		in
(*		  print_endline ("insert_term found");*)
		  match node with
		    | DtLeafNode entries ->
			begin
			  if fterm_equal entries.tail tail then begin

			    if
			      unique
			      &&
				List.exists
				(fun entry' ->
				   Term.term_equal (get_term entry) (get_term entry')
				)
				entries.list
			    then begin
			      ()
			    end


(*			    print_endline ("insert_term found 4");*)
(*			      print_endline (self#entrylist_to_string entries.list);*)
			    else begin
			      entries.list <- entry :: entries.list;
			    end

(*			    print_endline ("insert_term found 4b");*)
(*			      print_endline (self#entrylist_to_string entries.list);*)
			  end
(*				
			  else if tail = [] then begin
			    
			    failwith "AA";
			  end
*)
			  else begin  
(*			  (* really a leaf, or some compacted tail? *)
			  match tail with
			    | [] ->
				print_endline ("insert_term found 1");
				(* new variant of existing leaf term *)
				entries.list <- entry :: entries.list

			    | _ ->*)
(*				print_endline ("insert_term found 2");*)
				Table.remove table hkey;
				self#insert_expand_tail table entries entry hkey tail
			  end
			end
			  
		    | DtInnerNode table1 ->
			(* shared prefix does already exist *)
(*			print_endline ("insert_term found 3");*)
			self#insert_term ~unique:unique entry tail table1
	      end
	    with
	      | Not_found ->
(*		  print_endline ("insert_term not found");*)
		  let node =
		    DtLeafNode {
		      list = [ entry ];
		      tail = tail;
		    }
		  in
		    Table.add table hkey node
		      (*
		  begin
		    match tail with
		      | [] ->
			  (* term fully processed *)
			  let node =
			    DtLeafNode {list=[ (entry []) ]}
			  in
			    Table.add table hkey node
			      
		      | _ ->
			  (* create new node and continue *)
			  let table1 =
			    Table.make_table ()
			  in
			  let node =
			    DtInnerNode table1
			  in
			    Table.add table hkey node;
			    self#insert_term entry tail table1
		  end
		      *)
	  end

(*
  method private insert_term (entry: 'data entry) (term: term) (table: 'data table) (n: int) : 'data table * int =
    (* n is a counter of the terms to be processed. It is used to
       find out when we are at the last subterm in the preorder. *)
    let hkey = term_hkey term in
    let hkey_arity = (hkey_arity hkey) in
    let n1 = n + hkey_arity - 1 in

      try 
	  begin
	let node =
	  Table.find table hkey
	in
	  match node with
	    | DtLeafNode entries ->
		if Config.debug && n1 != 0 then begin
		  failwith "discrimination_tree:insert_term: Bad index structure 1."
		end;

		entries.list <- entry :: entries.list;
		table, n1 (* Return the table just to stay well typed. *)

	    | DtInnerNode table1 ->
		if Config.debug && n1 == 0 then begin
		  failwith "discrimination_tree:insert_term: Bad index structure 2."
		end;

		begin
		  match term with
		    | Term.Func ( _, terms, _ ) ->
			Array.fold_left
			  (fun (table', n') term ->
			     self#insert_term entry term table' n'
			  )
			  (table1, n1)
			  terms

		    | _ ->
			table1, n1
		end
	  end

      with
	| Not_found ->
	    begin
	      if n1 == 0 then begin (* At the last subterm in the preorder. *)
		let node =
		  DtLeafNode {list=[entry]; tail = []}
		in
		  Table.replace table hkey node; (* We could also use replace here. *)
		  table, n1
		  
	      end

	      else begin
		let table1 =
		  Table.make_table ()
		in
		let node =
		  DtInnerNode table1
		in
		  Table.replace table hkey node; (* We could also use replace here. *)
		  begin
		    match term with
		      | Term.Func ( _, terms, _ ) ->
			  Array.fold_left
			    (fun (table', n') term ->
			       self#insert_term entry term table' n'
			    )
			    (table1, n1)
			    terms

		      | _ ->
			  table1, n1
		  end;
		  
	      end
	    end



  method private insert term entry =
    match _root with
	DtInnerNode table ->
	  ignore (self#insert_1 entry term table 1);
	  _size <- _size + 1

      | _ ->
	  raise (Failure "discrimination_tree:insert: Bad index toplevel type.")
	    
  method private insert_1 entry term table n =
    (* n is a counter of the terms to be processed. It is used to
       find out when we are at the last subterm in the preorder. *)
    let hkey = term_hkey term in
    let hkey_arity = (hkey_arity hkey) in
    let n1 = n + hkey_arity - 1 in
    let node =
      try 
	Table.find table hkey
      with Not_found ->
	let node =
	  if n1 = 0 then (* At the last subterm in the preorder. *)
	    DtLeafNode {list=[]; tail = []}
	  else
	    DtInnerNode (Table.make_table ())
	in
	  Table.add table hkey node; (* We could also use replace here. *)
	  node
    in
      match n1, node with
	  0, DtLeafNode entries ->
	    entries.list <- entry :: entries.list;
	    table, n1 (* Return the table just to stay well typed. *)
	| _, DtInnerNode table1 ->
	    ( match term with
		  Term.Func ( _, termarry, _ ) ->
		    self#insert_1_args
		      entry
		      termarry
		      hkey_arity
		      0
		      table1
			n1
		| _ -> table1, n1 )
	| _ , _ ->
	    raise (Failure "discrimination_tree:insert_1: Bad index structure.")
	      
  method private insert_1_args entry args arity i table n =
    if i >= arity then
      table, n
    else
      let table1, n1 = self#insert_1 entry args.(i) table n in
	self#insert_1_args entry args arity (i+1) table1 n1
	
	  
  method add (term: term) (data: 'data) : unit =
    match _root with
      | DtInnerNode table ->
(*	  let fterm =
	    flatten term
	  in*)
	    self#insert_term (make_entry term data) term table 1;
	    _size <- _size + 1

      | _ ->
	  raise (Failure "discrimination_tree:add: Bad index toplevel type.")


	  *)


  (*** iter ***)
      
  method private iter_all_1 (func: term -> 'data -> unit) node =
    match node with
	DtInnerNode table ->
	  Table.iter (fun k v -> self#iter_all_1 func v) table
      | DtLeafNode entries ->
	  List.iter (fun entry -> func entry.term entry.data ) entries.list
	    
  method private iter_all (func: term -> 'data -> unit) : unit =
    if _size > 0 then
      self#iter_all_1 func _root
	
  method iter (func: term -> 'data -> unit) : unit =
    self#iter_all func
      

  method private fold_all_1:
    'a. ('a -> term -> 'data -> 'a) -> 'a -> 'data node -> 'a =
    fun func acc node ->
      match node with
	  DtInnerNode table ->
	    Table.fold (fun _ v acc -> self#fold_all_1 func acc v) table acc
	| DtLeafNode entries ->
	    List.fold_left (fun acc entry -> func acc entry.term entry.data ) acc entries.list
	    
  method private fold_all:
    'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a =
    fun func acc ->
      if _size > 0 then
	self#fold_all_1 func acc _root
      else
	acc
	
  method fold:
    'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a =
    fun func acc ->
      self#fold_all func acc

  (*** remove ***)
      
  (*** remove a variant ***)
      (*
  method private is_empty_table table =
    try
      Table.iter 
	(fun k v -> raise FoundVar) table;
      true
    with FoundVar ->
      false
      *)












  method private remove_variant (term: term) (data: 'data option) : bool =
    let fterm =
      flatten_term term
    in
      try
	let x : bool =
	  self#rm_variant term fterm _root data
	in
	  _size <- _size - 1;
	  true
      with
	| Not_found ->
	    false


  method private rm_variant (term: term) (fterm: fterm) (node: 'data node) (data: 'data option) : bool =
    match node with
      | DtInnerNode table ->
	  begin
	    match fterm with
	      | [] ->
		  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
		    
	      | hkey :: fterm1 ->
		  begin
(*		    try*)
		      let node1 =
			Table.find table hkey
		      in
		      let empty =
			self#rm_variant term fterm1 node1 data
		      in
			if empty then begin
			  Table.remove table hkey;
			  Table.length table == 0;
			end

			else begin
			  false
			end
(*		  with
		    | Not_found ->
			raise Not_found *)
		  end
		    
	  end
      | DtLeafNode entries ->
	  self#list_remove_variant term fterm entries data;
	  (entries.list = [])
	    

  method private list_remove_variant (term: term) (tail: fterm) (entries: 'data entries) (data: 'data option) : unit =
    if not (fterm_equal entries.tail tail) then begin
      raise Not_found
    end

    else begin
      let rec list_remove_variant' (entries: 'data entry list) : 'data entry list =
	match entries with
	  | [] ->
	      raise Not_found
		
	  | entry :: tail ->
	      if Term.are_terms_variants (get_term entry) term then begin
		match data with
		  | Some termdata when 
		      not (_data#is_equal (get_data entry) termdata) ->
			entry :: list_remove_variant' tail
			  
		  | _ ->
		      tail
	      end
		
	      else
		entry :: list_remove_variant' tail
      in
(*	try*)
	  let pruned =
	    list_remove_variant' entries.list
	  in
	    entries.list <- pruned;
(*	with
	  | Exit ->
	      false	*)
    end

















(* 
  method private remove_variant (term: term) (data: 'data option) : bool =
    let fterm =
      flatten_term term
    in
      try
	let x : 'data remove =
	  self#rm_variant term fterm _root data
	in
(*	  begin
	    match x with
	      | RemoveDone
	      | RemoveContinue ->
		  ()
		    
	      | RemoveShrink entries ->
		  let hkey =
		    match term with
		      | Term.Const symbol
		      | Term.Func (symbol, _, _) ->
			  Symbol symbol

		      | Term.Var _ ->
			  failwith "Discrimination_tree.remove_variant"
		  in
		    begin
		      match _root with
			| DtInnerNode table ->
			    let fterm =
			      match entries.tail with
				| _ :: tail ->
				    tail

				| _ ->
			      
			    Table.add table hkey (DtLeafNode entries);

			| DtLeafNode _ ->
			    failwith "Discrimination_tree.remove_variant 2"
		    end
	  end;
*)
	_size <- _size - 1;
	true
      with
	| Not_found ->
	    false


  method private rm_variant (term: term) (fterm: fterm) (node: 'data node) (data: 'data option) : 'data remove =
    match node with
      | DtInnerNode table ->
	  begin
	    match fterm with
	      | [] ->
		  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
		    
	      | hkey :: fterm1 ->
(*	  if !debug then begin
	    print_endline ("hkey : " ^ hkey_to_string hkey);
	  end;*)
		  begin
(*		    try*)
		      let node1 =
			Table.find table hkey
		      in
		      let prune =
			self#rm_variant term fterm1 node1 data
		      in
			match prune with
			  | RemoveDone ->
(*	  if !debug then begin
	    print_endline ("hkey Remove 0 : " ^ hkey_to_string hkey);
	  end;*)
			      RemoveDone


			  | RemoveContinue ->
(*	  if !debug then begin
	    print_endline ("hkey Remove 1 : " ^ hkey_to_string hkey);
	  end;*)
			      Table.remove table hkey;
			      
			      if Table.length table == 0 then begin
(*	  if !debug then begin
	    print_endline ("hkey Remove 2 : " ^ hkey_to_string hkey);
	  end;*)
				RemoveContinue
			      end
				
			      else begin
(*	  if !debug then begin
	    print_endline ("hkey Remove 3 : " ^ hkey_to_string hkey);
	  end;*)
				if Table.length table == 1 then begin
				  Table.fold
				    (fun hkey' node _ ->
				       match node with
					 | DtInnerNode _ ->
					     RemoveDone
					       
					 | DtLeafNode entries ->
					     RemoveShrink { entries with tail = (*hkey :: *)hkey' :: entries.tail }
				    )
				    table				
				    RemoveDone
				end

				else begin
(*	  if !debug then begin
	    print_endline ("hkey Remove 4 : " ^ hkey_to_string hkey);
	  end;*)
				  RemoveDone
				end

			      end


			  | RemoveShrink entries ->
(*	  if !debug then begin
	    print_endline ("hkey Remove 5 : " ^ hkey_to_string hkey);
	  end;*)
			      Table.remove table hkey;

			      if node == _root then begin
(*				if !debug then begin
				  print_endline ("hkey Remove 8 : " ^ hkey_to_string hkey);
				end;*)
				Table.add table hkey (DtLeafNode entries);
				RemoveDone;
			      end

			      else if Table.length table == 0 then begin
(*	  if !debug then begin
	    print_endline ("hkey Remove 6 : " ^ hkey_to_string hkey);
	  end;*)
				RemoveShrink { entries with tail = hkey :: entries.tail }
			      end
				
			      else begin
(*	  if !debug then begin
	    print_endline ("hkey Remove 7 : " ^ hkey_to_string hkey);
	  end;*)
				Table.add table hkey (DtLeafNode entries);
				RemoveDone;
			      end;
(*			      



			if prune then begin
			  Table.remove table hkey;

			  if Table.length table == 0 then begin
			    RemoveContinue
			  end
			  
			  else begin
			    if Table.length table == 1 then begin
			      Table.fold_left
				(fun hkey node ->
				   match node with
				     | DtInnerNode _ ->
					 RemoveDone

				     | DtLeafNode entries ->
					 RemoveShrink entries + hkey
				)
				RemoveDone
				table				
			    end
			  end;
			end

			else begin
			  RemoveDone
			end
*)
(*		    with
		      | Not_found ->
			  raise Not_found*)
		  end
	  end
	    
      | DtLeafNode entries ->
	  self#list_remove_variant term fterm entries data;
	  if entries.list = [] then
	    RemoveContinue
	  else
	    RemoveDone



(*
  method private rm_variant (term: term) (fterm: fterm) (node: 'data node) (data: 'data option) : bool =
    match node with
      | DtInnerNode table ->
	  begin
	    match fterm with
	      | [] ->
		  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
		    
	      | hkey :: fterm1 ->
		  if !debug then begin
		    print_endline ("hkey : " ^ hkey_to_string hkey);
		  end;
		  begin
(*		    try*)
		    let node1 =
		      Table.find table hkey
		      in
		    let prune =
		      self#rm_variant term fterm1 node1 data
		    in
		      if prune then begin
			if !debug then begin
			  print_endline ("hkey Remove 0 : " ^ hkey_to_string hkey);
			end;
			
			Table.remove table hkey;
		      end;
		      
		      
		      if Table.length table == 0 then begin
			if !debug then begin
			  print_endline ("hkey Remove 2 : " ^ hkey_to_string hkey);
			end;
			true
		      end
			  
		      else if Table.length table == 1 then begin
			if !debug then begin
			  print_endline ("hkey Remove 3 : " ^ hkey_to_string hkey);
			end;
			
			let entries =
			  Table.fold
			    (fun hkey' node _ ->
			       match node with
				 | DtInnerNode _ ->
				     None
				       
				 | DtLeafNode entries ->
				     Some (hkey', entries)
			    )
			    table
			    None
			in
			  begin
			    match entries with
			      | None ->
				  ()
				      
			      | Some (key, entries) ->
				  Table.replace
				    table
				    key
				    (DtLeafNode { entries with tail = key :: entries.tail })
			  end;

			  false
		      end

		      else begin
			false
		      end
			
		  end
	  end
	    
      | DtLeafNode entries ->
	  self#list_remove_variant term fterm entries data;
	  entries.list = []
    *)


  method private list_remove_variant (term: term) (tail: fterm) (entries: 'data entries) (data: 'data option) : unit =
(*    if !debug then begin
      print_endline ("list_remove_variant");
    end;*)
    if not (fterm_equal entries.tail tail) then begin
(*      if !debug then begin
	print_endline ("list_remove_variant NOP");
      end;*)
      raise Not_found
    end

    else begin
(*      if !debug then begin
	print_endline ("list_remove_variant YEP");
      end;*)
      let rec list_remove_variant' (entries: 'data entry list) : 'data entry list =
	match entries with
	  | [] ->
	      raise Not_found
		
	  | entry :: tail ->
	      if Term.are_terms_variants (get_term entry) term then begin
		match data with
		  | Some termdata when 
		      not (_data#is_equal (get_data entry) termdata) ->
		      print_endline ("list_remove_variant 1");
			entry :: list_remove_variant' tail
			
		  | _ ->
		      tail
	      end
		
	      else begin
		entry :: list_remove_variant' tail
	      end
      in
      let pruned =
	list_remove_variant' entries.list
      in
	entries.list <- pruned;
    end
*)

(*

  method private remove_variant (term: term) (data: 'data option) : bool =
    let fterm =
      flatten_term term
    in
    let _, removed =
      self#rm_variant term fterm _root data
    in
      if removed then
	_size <- _size - 1;

      removed


  method private rm_variant (term: term) (fterm: fterm) (node: 'data node) (data: 'data option) : bool * bool =
	    match node with
	      | DtInnerNode table ->
	  begin
    match fterm with
      | [] ->
	  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
	    
      | hkey :: fterm1 ->
(*	  if !debug then begin
	    print_endline ("hkey : " ^ hkey_to_string hkey);
	  end;*)
(*		  if !debug then begin
		    print_endline ("inner");
	  end;*)
		  begin
		    try
		      let node1 =
			Table.find table hkey
		      in
		      let empty, removed =
			self#rm_variant term fterm1 node1 data
		      in
			if empty then begin
			  Table.remove table hkey;
			  ((self#is_empty_table table), removed)
			end

			else begin
			  (false, removed)
			end
		    with
		      | Not_found ->
			  (false, false)
		  end
	  end

	      | DtLeafNode entries ->
(*		  if !debug then begin
		    print_endline ("leaf");
		  end;*)
		  begin
		    let removed =
		      self#list_remove_variant term fterm entries data
		    in
		      if removed then
			(entries.list = []), true

		      else
			false, false
		  end

(*	    
  method private rm_variant (term: term) (fterm: fterm) (node: 'data node) (data: 'data option) : bool * bool =
    match fterm, node with
      | hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      let node1 =
		Table.find table hkey
	      in
	      let empty, removed =
		self#rm_variant term fterm1 node1 data
	      in
		if empty then begin
		  Table.remove table hkey;
		  ((self#is_empty_table table), removed)
		end

		else begin
		  (false, removed)
		end
	    with
	      | Not_found ->
		  (false, false)
	  end

      | [], DtLeafNode entries ->
	  begin
	    let entries', removed =
	      self#list_remove_variant term entries.list data
	    in
	      if removed then begin
		entries.list <- entries';
		(entries.list = [], true)
	      end

	      else
		false, false
	  end

      | _, _ ->
	  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
	  *)		  

  method private list_remove_variant (term: term) (tail: fterm) (entries: 'data entries) (data: 'data option) : bool =
(*    if !debug then begin
      print_endline ("list_remove_variant");
    end;*)
    
    if not (fterm_equal entries.tail tail) then begin
      (*
      print_endline ("NO!");
      print_endline ("Entries: " ^ fterm_to_string entries.tail);
      print_endline ("Remove: " ^ fterm_to_string tail);*)
      false
    end

    else begin
(*      print_endline ("YO!");*)
      let rec list_remove_variant' (entries: 'data entry list) : 'data entry list =
	match entries with
	  | [] ->
	      raise Exit
		
	  | entry :: tail ->
	      if Term.are_terms_variants (get_term entry) term then begin
		match data with
		  | Some termdata when 
		      not (_data#is_equal (get_data entry) termdata) ->
			entry :: list_remove_variant' tail
			
		  | _ ->
		      tail
	      end
		
	      else
		entry :: list_remove_variant' tail
      in
	try
	  let pruned =
	    list_remove_variant' entries.list
	  in
	    entries.list <- pruned;
	    true
	with
	  | Exit ->
	      false	
    end
*)
(*
  method private remove_variant term (data: 'data option) =
    match (self#rm_variant term 
	     (flatten_term term)
	     _root data  ) with
	(empty, foundp) ->
	  ignore (if foundp then 
		    _size <- _size - 1);
	  foundp
	    
  method private rm_variant term fterm node (data: 'data option) =
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      let node1 = Table.find table hkey in
		match (self#rm_variant term fterm1 node1 data) with
		    (empty , foundp) ->
		      if empty then
			begin
			  Table.remove table hkey;
			  ((self#is_empty_table table), foundp)
			end
		      else
			(false, foundp)
	    with Not_found ->
	      (false, false)
	  end
      | [], DtLeafNode entries ->
	  begin
	    match self#list_remove_variant term entries.list data with
		(entlist, foundp) ->
		  begin
		    entries.list <- entlist ;
		    (entries.list = [], foundp)
		  end
	  end
      | _, _ ->
	  raise (Failure "discrimination_tree:rm_variant: Bad index structure.")
		  
  method private list_remove_variant term entries (data: 'data option)=
    match entries with
	[] ->
	  ([] , false)
      | entry :: entries1 ->
	  if Term.are_terms_variants (get_term entry) term  then 
	    match data with
		Some termdata -> 
		  if _data#is_equal (get_data entry) termdata then
(*		  if (get_data entry) = termdata then*)
		    (entries1, true)
		  else (
		    match self#list_remove_variant term entries1 data with 
			(rest, foundp) -> (entry :: rest , foundp) 
		  )
	      | None ->
		    (entries1, true)
	  else
	    match self#list_remove_variant term entries1 data with
		(rest, foundp) ->
		  (entry :: rest , foundp) 

		  *)

  method remove (term: term) (data: 'data option) : bool =
(*    if !debug || Term.term_equal interest term then begin
      print_endline ("\nremove " ^ Term.term_to_string term);
      debug := true;
      print_endline (self#to_string);
    end;*)
    
(*    print_endline ("remove");*)
    (* remove - removes p-preserving u_preserving variants only, per alex 040416/1240-0500 . *)
    self#remove_variant term data

  (*** variants ***)

  (* graf notation for candidate set for variants : *)
  (*   variants(N,t) := var(N,[t]) *)
  (*   var(empty,[X]) := emptyset *)
  (*   var(N, []) := {N} *)
  (* if p_preserving, i.e., U to U and P to P, only*)
  (*     var(N, [x, X]) := var(next(N,*U),[X]) U var(next(N,*P),[X])  *)
  (*     var(N, [p, X]) := var(next(N,*P),[X])  *)
  (* else not p_preserving, i.e., [U|P] to [U|P] *)
  (*     var(N, [[x|p], X]) := var(next(N,*U),[X]) U var(next(N,*P),[X])  *)
  (* end if *)
  (*   var(N, [a, X]) := var(next(N,a),[X]) *)
  (*   var(N, [f(t1, ... tn),X]) := var(next(N,f),[t1,...,tn,X]) *)

  method private map_variants ?(p_preserving: bool = true) f term : 'data list=
    if _size > 0 then
      self#map_var ~p_preserving:p_preserving f (flatten_term term) _root
    else
      []
	
  method private map_var ?(p_preserving: bool = true)  f fterm node : 'data list=
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    try
	      begin 
		if p_preserving then
		  begin 
		    match hkey with
			Variable ->
			  List.append 
			    (try
			       let node1 = Table.find table Variable in
				 self#map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
			    (try
			       let node1 = Table.find table Parameter in
				 self#map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
		      | _ ->
			  let node1 = Table.find table hkey in
			    self#map_var ~p_preserving:p_preserving f fterm1 node1
		  end
		else
		  begin 
		    match hkey with
			Variable
		      | Parameter ->
			  List.append 
			    (try
			       let node1 = Table.find table Variable in
				 self#map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
			    (try
			       let node1 = Table.find table Parameter in
				 self#map_var ~p_preserving:p_preserving f fterm1 node1
			     with Not_found ->
			       []
			    )
		      | _ ->
			  let node1 = Table.find table hkey in
			    self#map_var ~p_preserving:p_preserving f fterm1 node1
		  end
	      end
	    with Not_found ->
	      []
	  end
      | _, DtLeafNode entries ->
(* 	  List.map f entries.list *)
(*	  List.fold_right f tail entries []*)
	  f entries fterm 0 []

      | _, _ ->
          raise (Failure "discrimination_tree:map_var: Bad index structure.")


  method private create_leaf_check_all (check_tail: fterm -> fterm -> bool) (predicate: term -> bool) : 'data entries -> fterm -> int -> 'data list -> 'data list =
    let f entries tail count acc =
      let leaf_tail =
	skip_terms_fterm entries.tail count
      in

       if not (check_tail leaf_tail tail) then
	 acc

       else begin
	   List.fold_left
	     (fun acc entry ->
		if predicate (get_term entry) then
		  get_data entry :: acc
		else
		  acc
	     )
	     acc
	     entries.list
	 end
    in
      f

  method private create_leaf_check_first (check_tail: fterm -> fterm -> bool) (predicate: term -> bool) : 'data entries -> fterm -> int -> 'data list -> 'data list =
    let f entries tail count acc =
(*      if !debug then begin
	print_endline ("create_leaf_check_first:");
	print_endline (self#entrylist_to_string 2 entries.list);
      end;*)

      let leaf_tail =
	skip_terms_fterm entries.tail count
      in
      
      if not (check_tail leaf_tail tail) then begin
	acc
      end

      else begin
	   List.iter
	     (fun entry ->
		if predicate (get_term entry) then begin
		  self#set_found (get_data entry);
		  raise FOUND;
		end
	     )
	     entries.list;
	 acc
       end
    in
      f

  method find_all_variants (term: term) : 'data list =
    self#map_variants
      (*
      (fun entry dlist -> 
	 if Term.are_terms_variants (get_term entry) term  then
	   get_data entry :: dlist
	 else
	   dlist)
      *)
(*
      (fun entries tail acc ->
	 if not (fterm_equal tail entries.tail) then
	   acc

	 else begin
	   List.fold_left
	     (fun acc entry ->
		if Term.are_terms_variants (get_term entry) term  then
		  get_data entry :: acc
		else
		  acc
	     )
	     acc
	     entries.list
	 end)
*)
      (self#create_leaf_check_all fterm_equal (Term.are_terms_variants term))
      term

  method find (term: term) : 'data option =
    try
      ignore(
	self#map_variants ~p_preserving:true
	  (self#create_leaf_check_first fterm_equal (Term.term_equal term))
	  term
      );
      None
    with FOUND ->
      self#get_found



  (*** generalization ***)

  (* graf notation for candidate set for generalization : *)
  (*   general(N,t) := gen(N,[t]) *)
  (*   gen(empty,[X]) := emptyset *)
  (*   gen(N, []) := {N} *)
  (* if p_preserving, P to P and U to U, only *)
  (*     gen(N, [x, X]) := gen(next(N,*U),[X]) *)
  (*     gen(N, [p, X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X])  *)
  (*   only a Universal variable may generalize a constant or a function *)
  (*     gen(N, [a, X]) := gen(next(N,*U),[X]) U gen(next(N,a),[X]) *)
  (*     gen(N, [f(t1, ... tn),X]) := gen(next(N,*U),[X]) U *)
  (*                                 gen(next(N,f),[t1,...,tn,X]) *)
  (* else not p_preserving, i.e., [U|P] to [U|P] *)
  (*     gen(N, [[x|p], X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X]) *)
  (*     gen(N, [a, X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X]) U gen(next(N,a),[X]) *)
  (*     gen(N, [f(t1, ... tn),X]) := gen(next(N,*U),[X]) U gen(next(N,*P),[X]) U *)
  (*                                 gen(next(N,f),[t1,...,tn,X]) *)
  (* end if *)

  method private map_generalization_candidates ?(p_preserving: bool = true) ?(find_all: bool = false) f term : 'data list=
(*    if Term.term_equal interest term then begin
      print_endline ("\ngeneralization " ^ Term.term_to_string term);
	
      print_endline (self#to_string);
    end;*)

    if _size > 0 then
      self#map_gen ~p_preserving:p_preserving ~find_all:find_all [] f (flatten_term term) _root
    else
      []
	
  method private map_gen ?(p_preserving: bool = true) ?(find_all: bool = false) (acc: 'data list) f fterm node : 'data list=
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
(*	  if Term.term_equal interest (get_term entry) then begin
	    print_endline ("hkey: " ^ hkey_to_string hkey);
	  end;*)
	  begin
	    if p_preserving then
	      match hkey with
		  Parameter ->
		    let acc =
(*                     List.append *)
                      (try
			 let node1 = Table.find table Variable in
			   self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
			 acc
		      )
		    in
                      (try
			 let node1 = Table.find table Parameter in
			   self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
			 acc
		      )
		|  Variable ->
		    begin 
		      try
			let node1 = Table.find table Variable in
			  self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		      with Not_found ->
			acc
		    end
                |  _ ->
		    let acc =
			try 
			  let node1 = Table.find table hkey in
			    self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
			with Not_found ->
			  acc
		    in

		     (*		    List.append *)
		      (try
			  let node1 = Table.find table Variable in
			    self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc
			      f
			      (skip_terms_fterm fterm 1)
			      node1
			with Not_found ->
			  acc
		      )
(*
		      (match hkey with
			   Variable ->
			     acc
			 | _ ->
			     try 
			       let node1 = Table.find table hkey in
				 self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
			     with Not_found ->
			       acc
		      )*)

	    else (* not p_preserving *)
	      match hkey with
		| Parameter
		| Variable ->
		    let acc =
(*                     List.append *)
                      (try
			 let node1 = Table.find table Variable in
			   self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
			 acc
		      )
		    in
                      (try
			 let node1 = Table.find table Parameter in
			   self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
			 acc
		      )

		| _ ->
		    let acc =
			try 
			  let node1 = Table.find table hkey in
			    self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
			with Not_found ->
			  acc
		    in
		     (*		    List.append *)
		    let acc =
		      (try
			  let node1 = Table.find table Variable in
			    self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc
			      f
			      (skip_terms_fterm fterm 1)
			      node1
			with Not_found ->
			  acc
		      )
		    in
		      (try
			  let node1 = Table.find table Parameter in
			    self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc
			      f
			      (skip_terms_fterm fterm 1)
			      node1
			with Not_found ->
			  acc
		      )
(*
		    let acc =
(*                     List.append *)
                      (try
			 let node1 = Table.find table Variable in
			   self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
			 acc
		      )
		    in
                      (try
			 let node1 = Table.find table Parameter in
			   self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
			 acc
		      )
*)
(*

	      let acc =
(*	      List.append
		(List.append *)
		   (try
		      let node1 = Table.find table Variable in
			self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc
			  f
			  (skip_terms_fterm fterm 1)
			  node1
		    with Not_found ->
		      acc
		   )
	      in
	      let acc =
		   (try
		      let node1 = Table.find table Parameter in
			self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc
			  f
			  (skip_terms_fterm fterm 1)
			  node1
		    with Not_found ->
		      acc
		   )
in
(*		)*)
		( match hkey with
		      Variable 
		    | Parameter ->
			acc
		    | _ ->
			try 
			  let node1 = Table.find table hkey in
			    self#map_gen ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
			with Not_found ->
			  acc
		)
*)
	  end
      | _, DtLeafNode entries ->
(*	  print_endline ("check");*)
(*	  List.fold_right f entries.list acc*)
	  f entries fterm 0 acc

      | _, _ ->
	  raise (Failure "discrimination_tree:map_gen: Bad index structure.")
	    
  method find_all_generalizations ~(p_preserving: bool) (term: term) : 'data list =
    self#map_generalization_candidates ~p_preserving:p_preserving ~find_all:true 
      (self#create_leaf_check_all 
	     (fterm_equal_generalization ~p_preserving:p_preserving)
	 (fun term' -> 
(*				 print_endline ("find_all_generalizations");
				 print_endline ("tree: " ^ Term.term_to_string term');
				 print_endline ("query: " ^ Term.term_to_string term);*)
				 is_generalization ~p_preserving:p_preserving term term'))
(*      (fun entry dlist -> 
         if is_generalization ~p_preserving:p_preserving term (get_term entry) then
           (get_data entry) :: dlist
         else
           dlist)*)
      term

  method find_generalization ~(p_preserving: bool) (term: term) : 'data option =  
    try
      ignore(
	self#map_generalization_candidates ~p_preserving:p_preserving ~find_all:false 
	  (self#create_leaf_check_first 
	     (fterm_equal_generalization ~p_preserving:p_preserving)
	     (fun term' ->
(*				 print_endline ("find_generalizations");
				 print_endline ("tree: " ^ Term.term_to_string term');
				 print_endline ("query: " ^ Term.term_to_string term);*)
				 let x =
				   is_generalization ~p_preserving:p_preserving term term'

				 in
(*				   if Term.term_equal interest term then begin
				     print_endline ("find_generalizations");
				     print_endline ("tree: " ^ Term.term_to_string term');
				     print_endline ("query: " ^ Term.term_to_string term);
				     print_endline ("FOUND: " ^ string_of_bool x);
				   end;*)
				   x
				  ))
(*
	  (fun entry dlist -> 
		  if is_generalization ~p_preserving:p_preserving term (get_term entry) then begin
		    (*raise (FOUND (get_data entry))*)
		    self#set_found (get_data entry);
		    raise FOUND;
		  end
		  else
		    dlist) *)
	       term );
      None
    with FOUND ->
      self#get_found
      

  (*** unifiable ***)

  (* graf notation for skip set : *)
  (*   skip(N) := U over N' sonsof (N) skip'(N', arity(label(N,N')))  *)
  (*   skip'(N,0) := {N} *)
  (*   skip'(N,l) := U over N' sonsof (N) skip'(N',l - 1 +arity(label(N,N'))) *)

  (* graf notation for candidate set for unifiable terms : *)
  (*   unifiable(N,t):= uni(N,[t])  *)
  (*   uni(empty,[X]) := emptyset  *)
  (*   uni(N, []) := {N}  *)
  (* if p_preserving, i.e., P to P only, but either [U to U | U to term(on both U and P) ] allowed *)
  (*     uni(N, [x, X]) := U over M \in skip(N) uni(M,[X])  *)
  (*     uni(N, [p, X]) := uni(next(N,*V),[X]) U uni(next(N,*P),[X])  *)
  (*   only a Universal variable may generalize a constant or a function *)
  (*     uni(N, [a, X]) := uni(next(N,*U),[X]) U uni(next(N,a),[X]) *)
  (*     uni(N,[f(t1, ... tn),X]) := uni(next(N,*U),[X]) U *)
  (*                                uni(next(N,f),[t1,...,tn,X]) *)
  (* else not p_preserving, i.e., [([U|P] to [U|P]) | ([U|P] to term)] allowed *)
  (*     uni(N, [[x|p] , X]) := U over M \in skip(N) uni(M,[X])  *)
  (*     uni(N, [a, X]) := uni(next(N,*U),[X]) U uni(next(N,*P),[X]) U uni(next(N,a),[X]) *)
  (*     uni(N,[f(t1, ... tn),X]) := uni(next(N,*U),[X]) U uni(next(N,*p),[X]) U *)
  (*                                 uni(next(N,f),[t1,...,tn,X]) *)
  (* end if  *)

  method private map_unification_candidates ?(p_preserving: bool = true) ?(find_all: bool = false) f term : 'data list =
(*    print_endline ("unification");*)
(*    if Term.term_equal interest term then begin
      debug := true;
      print_endline ("\nunification " ^ Term.term_to_string term);
	
      print_endline (self#to_string);
    end;*)
    if _size > 0 then
      self#map_unif ~p_preserving:p_preserving ~find_all:find_all [] f (flatten_term term) _root
    else
      []
	
  method private map_unif ?(p_preserving: bool = true) ?(find_all: bool = false) (acc: 'data list) f fterm node : 'data list =
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
(*	  if !debug then begin
	    print_endline ("hkey : " ^ hkey_to_string hkey);
	  end;*)
	  begin
	    if p_preserving then
	      match hkey with
		  Variable ->
(*		    if !debug then begin
		      print_endline ("PV");
		    end;*)
		    self#map_unif_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node 1
                | Parameter ->
(*	  if !debug then begin
		    print_endline ("PP");
	  end;*)
(*                     List.append *)
		    let acc =
                      (try 
		         let node1 = Table.find table Variable in
			   self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
		         acc
		      )
		    in
                      (try 
		         let node1 = Table.find table Parameter in
			   self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
		         acc
		      )
	        | _ ->
(*	  if !debug then begin
		    print_endline ("PS");
	  end;*)
		    let acc =
(* 		    List.append*)
                      (try 
			 let node1 = Table.find table hkey in
(*			 print_endline "SYM";*)
			   self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		       with Not_found ->
(*			 print_endline "NO SYM";*)
			 acc
		      )
		    in
		      (try 
			 let node1 = Table.find table Variable in
(*			 print_endline "FOUND VAR";*)
			   self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc
			     f
			     (skip_terms_fterm fterm 1)
			     node1
		       with Not_found ->
(*			 print_endline "NO VAR";*)
			 acc
		      )
	    else
	      match hkey with
		  Variable 
		| Parameter ->
(*	  if !debug then begin
(*		    print_endline ("UV");*)
	  end;*)
		    self#map_unif_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node 1
		| _ ->
(*	  if !debug then begin
(*		    print_endline ("US");*)
	  end;*)
		    let acc =
		      (try 
			  let node1 = Table.find table hkey in
			    self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
			with Not_found ->
			  acc
		      )
		    in
		    let acc =
(*		    List.append
		      (List.append*)
			 (try 
			    let node1 = Table.find table Variable in
			      self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc
				f
				(skip_terms_fterm fterm 1)
				node1
			  with Not_found ->
			    acc
			 )
		    in
			 (try 
			    let node1 = Table.find table Parameter in
			      self#map_unif ~p_preserving:p_preserving ~find_all:find_all  acc
				f
				(skip_terms_fterm fterm 1)
				node1
			  with Not_found ->
			    acc
			 )
(*		      )*)

	  end
      | _, DtLeafNode entries ->
(*	  print_endline "CHILD";
	  print_endline (self#entrylist_to_string 2 entries.list);*)
(*	  List.fold_right f entries.list acc*)
	  f entries fterm 0 acc

      | _, _ ->
	  raise (Failure "discrimination_tree:map_unif: Bad index structure.")	
	    
  method private map_unif_skip ?(p_preserving: bool = true) ?(find_all: bool = false) (acc: 'data list) f fterm node count : 'data list =
(*	  if !debug then begin
    print_endline ("US");
	  end;*)
    if count = 0 then
      self#map_unif ~p_preserving:p_preserving ~find_all:find_all acc f fterm node
    else
      match node with
	  DtInnerNode table ->
	    (*  val fold : ('a -> 'b -> 'c -> 'c) -> ('a, 'b) t -> 'c -> 'c *)
	    Table.fold
	    (fun hkey node1 (*dlist*) acc ->
(*	  if !debug then begin
    print_endline ("US: " ^ hkey_to_string hkey);
	  end;*)
(*	       List.append *)
	       (self#map_unif_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm node1 (count + (hkey_arity hkey) - 1))
(*	       dlist*)
	    )
	    table
	    acc

	| DtLeafNode entries ->
	    f entries fterm count acc

(*	| _ -> raise (Failure "discrimination_tree:map_unif_skip: Bad index structure.")*)

	    
  method find_all_unifiable ~(p_preserving: bool) (term: term) : 'data list = 
    self#map_unification_candidates ~p_preserving:p_preserving ~find_all:true 
	  (self#create_leaf_check_all 
	     (fterm_equal_unifiable ~p_preserving:p_preserving)
	     (is_unifiable ~p_preserving:p_preserving term))
(*      (fun entry dlist -> 
         if is_unifiable ~p_preserving:p_preserving term (get_term entry) then
           (get_data entry) :: dlist
         else
           dlist)*)
      term 

  method find_unifiable ~(p_preserving: bool) (term: term) : 'data option = 
    try
      ignore(
	self#map_unification_candidates  ~p_preserving:p_preserving ~find_all:false 
	  (self#create_leaf_check_first
	     (fterm_equal_unifiable ~p_preserving:p_preserving)
	     (fun term' ->
				 let x =
				   (is_unifiable ~p_preserving:p_preserving term term')
					(*   is_generalization ~p_preserving:p_preserving term term'*)
				 in
(*				   if Term.term_equal interest term then begin*)
(*				   if !debug then begin
				     print_endline ("find_generalizations");
				     print_endline ("tree: " ^ Term.term_to_string term');
				     print_endline ("query: " ^ Term.term_to_string term);
				     print_endline ("FOUND: " ^ string_of_bool x);
				   end;*)
				   x
					)
	  )
(*	       (fun entry dlist -> 
		  if is_unifiable ~p_preserving:p_preserving term (get_term entry) then begin
		    (*raise (FOUND (get_data entry))*)
		    self#set_found (get_data entry);
		    raise FOUND;
		  end
		  else
		    dlist)*)
	       term);
      None
    with FOUND ->
      self#get_found


  (*** instance ***)

  (* graf notation for candidate set for instances : *)
  (*   instances(N,t) := inst(N,[t]) *)
  (*   inst(empty,[X]) := emptyset *)
  (*   inst(N, []) := {N} *)
  (* if p_preserving, i.e., P to P only, but either [U to U | U to term(on both U and P) ] allowed *)
  (*     inst(N, [x, X]) := U over M \in skip(N) inst(M,[X])  *)
  (*     inst(N, [p, X]) := inst(next(N,*P),[X])  *)
  (* else not p_preserving, i.e., [([U|P] to [U|P]) | ([U|P] to term)] allowed *)
  (*     inst(N, [[x|p] , X]) := U over M \in skip(N) inst(M,[X])  *)
  (* end if *)
  (*   inst(N, [a, X]) := inst(next(N,a),[X]) *)
  (*   inst(N,[f(t1, ... tn),X]) := inst(next(N,f),[t1,...,tn,X]) *)

  method private map_instance_candidates ?(p_preserving: bool = true) ?(find_all: bool = false) f term : 'data list =
(*    print_endline ("instance");*)
    if _size > 0 then
      self#map_inst ~p_preserving:p_preserving ~find_all:find_all [] f (flatten_term term) _root
    else
      []

  method private map_inst ?(p_preserving: bool = true) ?(find_all: bool = false) (acc: 'data list) f fterm node : 'data list=
    match fterm, node with
	hkey::fterm1, DtInnerNode table ->
	  begin
	    if p_preserving then
	      match hkey with
		  Variable ->
		    self#map_inst_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node 1
		| _ ->
		    try
		      let node1 = Table.find table hkey in
			self#map_inst ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		    with Not_found ->
		      acc
	    else
	      match hkey with
		  Variable 
		| Parameter ->
		    self#map_inst_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node 1
		| _ ->
		    try
		      let node1 = Table.find table hkey in
			self#map_inst ~p_preserving:p_preserving ~find_all:find_all acc f fterm1 node1
		    with Not_found ->
		      acc
	  end
      | _, DtLeafNode entries ->
(* 	    ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b *)
(* 	    List.fold_right f entries.list acc*)
	  f entries fterm 0 acc

      | _, _ ->
	  raise (Failure "discrimination_tree:map_inst: Bad index structure.")
(*
  method private map_inst_skip ?(p_preserving: bool = true) ?(find_all: bool = false) (acc: 'data list) f fterm node count : 'data list =
    if count = 0 then
      self#map_inst ~p_preserving:p_preserving ~find_all:find_all acc f fterm node
    else
      match node with
	  DtInnerNode table ->
	    (*  val fold : ('a -> 'b -> 'c -> 'c) -> ('a, 'b) t -> 'c -> 'c *)
	    Table.fold
	    (fun hkey node1 (*dlist*) acc ->
(*	       List.append*)
	       (self#map_inst_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm node1 (count + (hkey_arity hkey) - 1))
(*	       dlist*)
	    )
	    table
	    acc

	| _ -> raise (Failure "discrimination_tree:map_inst_skip: Bad index structure.")
*)


  method private map_inst_skip ?(p_preserving: bool = true) ?(find_all: bool = false) (acc: 'data list) f fterm node count : 'data list =
    if count = 0 then
      self#map_inst ~p_preserving:p_preserving ~find_all:find_all acc f fterm node
    else
      match node with

	| DtInnerNode table ->
	    Table.fold
	    (fun hkey node1 acc ->
	       (self#map_inst_skip ~p_preserving:p_preserving ~find_all:find_all acc f fterm node1 (count + (hkey_arity hkey) - 1))
	    )
	    table
	    acc

	| DtLeafNode entries ->
	    f entries fterm count acc


  method find_all_instances ~(p_preserving: bool) (term: term) : 'data list =
    self#map_instance_candidates ~p_preserving:p_preserving ~find_all:true 
	  (self#create_leaf_check_all
	     (fterm_equal_instance ~p_preserving:p_preserving)
	     (fun term' -> is_instance ~p_preserving:p_preserving term term'))
(*      (fun entry dlist -> 
         if is_instance ~p_preserving:p_preserving term (get_term entry) then
           (get_data entry) :: dlist
         else
           dlist)*)
      term 

  method find_instance ~(p_preserving: bool) (term: term) : 'data option = 
    try 
      ignore(
	self#map_instance_candidates ~p_preserving:p_preserving ~find_all:false 
	  (self#create_leaf_check_first
	     (fterm_equal_instance ~p_preserving:p_preserving)
	     (fun term' -> is_instance ~p_preserving:p_preserving term term'))
(*	       (fun entry dlist -> 
		  if is_instance ~p_preserving:p_preserving term (get_term entry) then begin
		    (*raise (FOUND (get_data entry))*)
		    self#set_found (get_data entry);
		    raise FOUND;
		  end
		  else
		    dlist) *)
	       term );
      None
    with FOUND ->
      self#get_found

  (*** productivity ***)

  method find_strongly_shielding (producer: term) (produced: term) : 'data option =
    failwith "Discrimination_tree.find_strongly_shielding"

  (* find the first or all elements according to the specified function *)
  method find_all_shielding (producer: term) (produced: term) : 'data list =
    failwith "Discrimination_tree.find_all_shielding"

  method find_shielding (producer: term) (produced: term) : 'data option =
    failwith "Discrimination_tree.find_shielding"

  method get_generalization_iterator ~(p_preserving: bool) (term: term) : 'data iterator =
    (* failwith "Discrimination_tree.get_generalization_iterator" *)
    let generalizations =
      self#find_all_generalizations ~p_preserving:p_preserving term
    in
      object (iterator)
	val mutable _elements = generalizations

	method is_empty =
	  match _elements with
	    | [] -> true
	    | _ -> false

	method next =
	  match _elements with
	    | [] ->
		raise ITERATOR_EMPTY

	    | element :: tail ->
		_elements <- tail;
		element
      end


  method get_shielding_iterator (producer: term) (produced: term) : 'data iterator =
    failwith "Discrimination_tree.get_shielding_iterator"

  (*** to_string ***)


  method private indent d : string =
      if d <= 0 then
	""
      else
	"  " ^ self#indent (d - 1)

  method private index_to_string : string =
    let node = _root in
      self#node_to_string node 0

  method private node_to_string node d = 
    match node with
	DtInnerNode table ->
	  Table.fold
	  (fun k v str ->
	     (self#indent d) ^ 
	     "hkey : " ^ (hkey_to_string k) ^ "\n" ^ 
	     (self#node_to_string v (d + 1)) ^ 
	     str 
	  )
	  table
	  ""
      | DtLeafNode entries ->
(*	  (self#indent d) ^ "[ " ^ self#entrylist_to_string entries.list ^ "]\n"*)
	  (self#indent d) ^ "tail : " ^ fterm_to_string entries.tail ^ "\n"
	  ^ self#entrylist_to_string d entries.list

  method private entrylist_to_string (d: int) (entrylist: 'data entry list) : string =
    String.concat "\n"
      (List.map
	 (fun entry ->
	    self#indent d
	    ^ Term.term_to_string (get_term entry)
	    ^ "\n"
	 )
	 entrylist
      )
(*
    match entrylist with
	entry :: entrylist1 ->
	  let data = get_data entry in
	    _data#to_string data ^ ", " ^ self#entrylist_to_string entrylist1

      | _ ->
	  ""*)

  method to_string : string =
    self#index_to_string
      
end


    
class ['data] index
  (__data: 'data data)
   =
object (self)
  inherit ['data] Term_indexing.index
    __data
    (new predicate_index)
end






(*** public interface ***)

let create_predicate_index (data: 'data data) : 'data Term_indexing.predicate_index =
  new predicate_index data


let create_index (data: 'data data) : 'data Term_indexing.index =
  new index data
