(* $Id: log.mli,v 1.2 2004/12/28 14:06:49 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** derivation log *)



(** {6 Types} *)

type literal = Term.literal
type config = Config.config
type state = State.state
type choice_point = State.choice_point
type context = Context.context
type raw_context_unifier = Context_unifier.raw_context_unifier


(** prints the current derivation. *)
class type log =
object
  (** [assert_ literal context_unifier]
      registers an Assert application in the current choice point. *)
  method assert_: literal -> raw_context_unifier -> unit

  (** [split_unit literal context_unifier]
      registers a Unit Split application in the current choice point. *)
  method split_unit: literal -> raw_context_unifier -> unit

  (** [split_left choice_point literal context_unifier]
      registers a Left Split application creating [choice_point]. *)
  method split_left:
    choice_point -> literal -> raw_context_unifier -> unit

  (** notifiy of a Right Split application. *)
  (** [close closing_context_unifier]
      registers an Assert application. *)

  (** [split_right left_choice_point right_choice_point literal context_unifier]
      registers a Right Split application creating [right_choice_point].
      [left_choice_point] is the choice point of the corresponding Left Split.
  *)
  method split_right:
    (*choice_point -> choice_point -> *)literal -> raw_context_unifier -> unit

  (** [close closing_context_unifier]
      registers a Close application. *)
  method close: choice_point -> raw_context_unifier -> unit

  (** registers a branch as 'closed' because of being incomplete
      (see {!Bound}). *)
  method incomplete: choice_point -> unit

  (** registers a choice point as unneeded in the derivation,
      i.e. it has been retracted without contributing anything. *)
  method prune: choice_point -> unit

  (** does any log show any interest in pruning?
      (just a performance optimization) *)
  method is_prune: bool

  (** finalizes the derivation log.
      That is, depending on the log type,
      nothing might happen or the complete derivation might be printed. *)
  method finalize: unit
end


(** {6 Functions} *)

(**
   provides various ways to print the current derivation.
   Creates a log suite based on the config settings, see:
   - {!Config.print_derivation}
   - {!Config.print_derivation_online}
   - {!Config.print_derivation_file}
   - {!Config.print_derivation_dot}
   - {!Config.print_derivation_context_unifier}
   - {!Config.print_derivation_pruned}
*)
val create: config -> state -> context -> log
