(* $Id: selection.mli,v 1.15 2004/12/28 14:06:50 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** selection of a split or assert literal

  manages the computation of candidates for {e Assert} and {e Split}
  with {!Context_unifier} and {!Context_unifier_space}
  after each change of the context. *)


(** {6 Types} *)

type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type clause = Term.clause
type choice_point = State.choice_point
type state = State.state
type selected = Selection_types.selected
type context = Context.context


(** the {e Selection} data structure *)
type selection



(** no further selection possible - satisfiability detected *)
exception SATISFIABLE



(** {6 Functions} *)


(** [create input_clauses lemmas] creates a new [selection]
  over the input clauses. *)
val create: clause list -> clause list -> config -> bound -> statistic -> state -> context -> Random_jump.saved_branch -> selection


(** to be called when a new term is added to the context.
  Updates the set of assert and split candidates.
  May raise {!Context_unifier.CLOSE}. *)
val add: selection -> Context.element -> unit


(** selects a literal to add to the context.
  Raises {!Selection.SATISFIABLE} if no valid candidate literal exists. *)
val select: selection -> selected



(** [backtrack selection retracted_left_split oldest_retracted rigth_split_literal]
    removes candidates based on invalid states.
    takes the last retracted left split choice point,
    plus the corresponding right split literal,
    and returns the information of the corresponding right split. *)
val backtrack: selection -> choice_point -> choice_point list -> unit

val add_lemma: selection -> clause -> unit
val get_lemmas: selection -> clause list


val in_replay: selection -> bool
