(* $Id: selection_assert.mli,v 1.15 2004/12/28 14:06:50 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** selection of assert candidates

    stores all assert candidates,
    and selects one for application.
*)



(** {6 Types} *)

type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type state = State.state
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected = Selection_types.selected
type context = Context.context

type candidates


(** {6 Functions} *)

(** creates an empty candidate repository. *)
val create: config -> bound -> statistic -> state -> context -> candidates

(** adds an assert candidate to the repository. *)
val add: candidates -> raw_context_unifier -> bool

(** selects (priority based) an assert literal from the repository.
    Ignores previously selected candidates,
    and candidates subsumed by the context.

    Returns None, if no further assert candidate exists. *)
val select: candidates -> selected option

(** registers the stored candidates exceeding the depth bound {!Bound}
    if they are not satisfied by the context.
    To be called based on {!Flags.opt_restart}, when a branch
    has to be checked for being exhausted.
*)
val check_exceeding: candidates -> unit

(** removes the assert candidates based on invalid states.
    readds previously used but now from the context removed candidates. *)
val backtrack: candidates -> unit

val has_closing_candidate: candidates -> bool

val size: candidates -> int
