(* $Id: selection_split.ml,v 1.63 2004/12/30 13:04:10 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type config = Config.config
type bound = Bound.bound
type statistic = Statistic.statistic
type choice_point = State.choice_point
type literal = Term.literal
type pureness = Term_attributes.pureness
type subst = Subst.subst
type state = State.state
type context_unifier_space = Context_unifier.context_unifier_space
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected = Selection_types.selected
(*type selection_lookahead = Selection_lookahead.selection_lookahead*)
type context = Context.context

(* an instance computed by a context unifier for a clause and a context literal.
   a remainder literal, if cl_is_remainder_literal is true *)
type candidate_literal = {
  (* the instance *)
  cl_literal: literal;

  (* the context literal used to compute the remainder literal *)
  cl_context_literal: Context.element;

  (* is this a remainder literal? *)
  cl_is_remainder_literal: bool;

  (* based on the nth clause literal, starting from 0 *)
  cl_index: int;
}

type quality = {
  (* properties of the remainder. *)
  cq_remainder_size: int;

  (* properties of the candidate literal. *)
  cq_pureness: pureness;
  cq_term_quality: int;
  cq_generation: int;
}

(* actually all literals of a context unifier, not only the remainder *)
type candidate = {
  (* the computed context unifier *)
  ca_raw_context_unifier: raw_context_unifier;

  (* comparison properties for the selection heuristic *)
  ca_quality: quality;
}


(*let max_uncompressed = 16000*)


(*** comparison ***)
(*
let compare_candidates (first: candidate) (second: candidate) : int =
  let pureness_universal =
    Term.cmp_pureness first.ca_quality.cq_pureness second.ca_quality.cq_pureness
  in
    if pureness_universal <> 0 then
      pureness_universal
    else
      let size =
        compare first.ca_quality.cq_remainder_size second.ca_quality.cq_remainder_size
      in
        if size <> 0 then
          size
        else
 	  let term_quality =
	    compare first.ca_quality.cq_term_quality second.ca_quality.cq_term_quality
	  in
	    if term_quality <> 0 then
	      term_quality
	    else
	      let pureness_generalization =
		Term.cmp_universality first.ca_quality.cq_pureness second.ca_quality.cq_pureness
	      in
		if pureness_generalization <> 0 then
		  pureness_generalization
		else
(*      let size =
	compare
	  (Array.length first.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
	  (Array.length second.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
      in
	if size <> 0 then
	  size
	else*)
		  let generation =
		    compare
		      (Context_unifier.generation_of_candidate first.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
		      (Context_unifier.generation_of_candidate second.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
		  in
		    if generation <> 0 then
		      generation
		    else
		      Context_unifier.compare_context_unifiers first.ca_raw_context_unifier second.ca_raw_context_unifier
*)			


let compare_candidates (first: candidate) (second: candidate) : int =
  let pureness_universal =
    Term_attributes.cmp_pureness first.ca_quality.cq_pureness second.ca_quality.cq_pureness
  in
    if pureness_universal <> 0 then
      pureness_universal
    else begin
      if first.ca_quality.cq_remainder_size < second.ca_quality.cq_remainder_size then
	-1
      else if first.ca_quality.cq_remainder_size > second.ca_quality.cq_remainder_size then
	1
      else begin
 	  if first.ca_quality.cq_term_quality < second.ca_quality.cq_term_quality then
	    -1
	  else if first.ca_quality.cq_term_quality > second.ca_quality.cq_term_quality then
	    1
	  else begin
	    let pureness_generalization =
	      Term_attributes.cmp_universality first.ca_quality.cq_pureness second.ca_quality.cq_pureness
	    in
	      if pureness_generalization <> 0 then
		pureness_generalization
	      else begin
		if first.ca_quality.cq_generation < second.ca_quality.cq_generation then
		  -1
		else if first.ca_quality.cq_generation > second.ca_quality.cq_generation then
		  1
		else
		  Context_unifier.compare_context_unifiers first.ca_raw_context_unifier second.ca_raw_context_unifier
	      end
	  end
      end
    end







exception Empty



(* A totally ordered type and its comparison functions *)
module type ORDERED = sig
  type t

  val eq : t -> t -> bool
  val lt : t -> t -> bool
  val leq : t -> t -> bool
end


module type HEAP = sig
  module Elem : ORDERED

  type heap

  val empty : heap
  val is_empty : heap -> bool
  val size : heap -> int

  val add : Elem.t -> heap -> heap
  val merge : heap -> heap -> heap

  val find_min : heap -> Elem.t  (* raises Empty if heap is empty *)
  val remove_min : heap -> heap  (* raises Empty if heap is empty *)

  val iter : (Elem.t -> unit) -> heap -> unit
    
  val check : heap -> unit
end

(*
module SplayTree (Element : ORDERED) : (HEAP with module Elem = Element) =
struct
  module Elem = Element

  type heap =
    | Empty
    | Node of heap * Elem.t * heap

  let empty =
    Empty

  let is_empty heap =
    heap == Empty

  let splay (heap: heap) (element: Elem.t) : heap * heap =
    match heap with
      | Empty ->
	  failwith "splay empty"

      | Node (left, x, right) ->
	  if Element.leq element x then begin
	    match left with
	      | Empty ->
		  (* no change *)
		  left, right

	      | Node (left', y, right') ->
		  (* zig left *)
		  if Element.eq x' element then begin
		    Node (left', x, Node (right', x', right))
		  end

		  (* zig-zig left *)
		  else if Element.lt x' element then begin
		    let splayed =
		      splay left' element
		    in
		      Node (
		  end

		  else if begin
		  end
	  end

	  else begin
	  end

end


	    *)


module SplayHeap (Element : ORDERED) : (HEAP with module Elem = Element) =
struct
  module Elem = Element

  type heap = E | T of heap * Elem.t * heap

  let empty = E
  let is_empty h = h = E


  let size _heap =

    let rec size' size heap =
      match heap with
	| E ->
	    size

	| T (a, _, b) ->
	    size' (size' (size + 1) a) b
    in
      size' 0 _heap


  let rec partition pivot = function
    | E -> E, E
    | T (a, x, b) as t ->
        if Elem.leq x pivot then
          match b with
          | E -> t, E
          | T (b1, y, b2) ->
              if Elem.leq y pivot then
                let small, big = partition pivot b2 in
                T (T (a, x, b1), y, small), big
              else
                let small, big = partition pivot b1 in
                T (a, x, small), T (big, y, b2)
        else
          match a with
          | E -> E, t
          | T (a1, y, a2) ->
              if Elem.leq y pivot then
                let small, big = partition pivot a2 in
                T (a1, y, small), T (big, x, b)
              else
                let small, big = partition pivot a1 in
                small, T (big, y, T (a2, x, b))






  let rec merge s t = match s, t with
    | E, _ -> t
    | T (a, x, b), _ ->
        let ta, tb = partition x t in
        T (merge ta a, x, merge tb b)

  let rec find_min = function
    | E -> raise Empty
    | T (E, x, _) -> x
    | T (a, x, _) -> find_min a

  let rec remove_min = function
    | E -> raise Empty
    | T (E, _, b) -> b
    | T (T (E, _, b), y, c) -> T (b, y, c)
    | T (T (a, x, b), y, c) -> T (remove_min a, x, T (b, y, c))



  let rec check heap : unit =

    let rec check' heap acc =
      if is_empty heap then
	acc

      else begin
	let min =
	  find_min heap
	and heap' =
	  remove_min heap
	in
	  check' heap' (min :: acc)
      end
    in

    let sorted : Elem.t list =
      check' heap []
    in
    let x : Elem.t option =
      List.fold_left
	(fun acc next ->
	   begin
	     match acc with
	       | None ->
		   ()

	       | Some prev ->
		   if compare prev next < 0 then begin
		     failwith "SplayHeap failure";
		   end
	   end;

	   Some next
	)
	None
	sorted
    in
      ()

  let rec iter (func: Elem.t -> unit) heap : unit =
    match heap with
      | E ->
	  ()

      | T (l, elem, r) ->
	  iter func l;
	  func elem;
	  iter func r


(*  let add x t = let a, b = partition x t in T (a, x, b)*)

  let rec add x t =
    let x =
    match t with
      | E ->
	  T (E, x , E)

      | T (left, x', right) ->
	  if Elem.lt x x' then begin
	    T (add x left, x', right)
	  end

	  else begin
	    T (left, x', add x right)
	  end
    in
      check x;
      x


	    
end



module PairingHeap (Element : ORDERED) : (HEAP with module Elem = Element) =
struct
  module Elem = Element

  type heap = E | T of Elem.t * heap list

  let empty = E
  let is_empty h = h = E

  let size _heap =

    let rec size' size heap =
      match heap with
	| E ->
	    size

	| T (_, x) ->
	    List.fold_left
	      (fun acc heap ->
		 size' acc heap
	      )
	      (size + 1)
	      x
    in
      size' 0 _heap

  let merge h1 h2 = match h1, h2 with
    | _, E -> h1
    | E, _ -> h2
    | T (x, hs1), T (y, hs2) ->
        if Elem.leq x y then T (x, h2 :: hs1)
        else T (y, h1 :: hs2)


  let rec merge_pairs = function
    | [] -> E
    | [h] -> h
    | h1 :: h2 :: hs -> merge (merge h1 h2) (merge_pairs hs)

  let find_min = function
    | E -> raise Empty
    | T (x, hs) -> x

  let remove_min = function
    | E -> raise Empty
    | T (x, hs) -> merge_pairs hs

  let rec iter (func: Elem.t -> unit) heap : unit =
    match heap with
      | E ->
	  ()

      | T (elem, x) ->
	  func elem;
	  List.iter
	    (fun heap' ->
	       iter func heap'
	    )
	    x

  let rec check heap : unit =

    let rec check' heap acc =
      if is_empty heap then
	acc

      else begin
	let min =
	  find_min heap
	and heap' =
	  remove_min heap
	in
	  check' heap' (min :: acc)
      end
    in

    let sorted : Elem.t list =
      check' heap []
    in
    let x : Elem.t option =
      List.fold_left
	(fun acc next ->
	   begin
	     match acc with
	       | None ->
		   ()

	       | Some prev ->
		   if compare prev next > 0 then begin
		     failwith "PairingHeap failure";
		   end
	   end;

	   Some next
	)
	None
	sorted
    in
      ()

  let add x h =
    let x =
      merge (T (x, [])) h
    in
      check x;
      x
end



(*
module Valid =
  SplayHeap (
(*    PairingHeap ( *)
    struct
      type t = candidate

      let compare (first: candidate) (second: candidate) : int =
	compare_candidates first second

      let eq (first: candidate) (second: candidate) : bool =
	first == second

      let lt (first: candidate) (second: candidate) : bool =
	compare first second < 0

      let leq (first: candidate) (second: candidate) : bool =
	compare first second <= 0

    end
  )
  *)


module Valid =
  Heap.Make (
    struct
      type t = candidate

      let compare (first: candidate) (second: candidate) : int =
	compare_candidates first second
    end
  )

(*
type compressed_candidates = {
  mutable front: (candidate * literal * int) option;
  candidates: raw_context_unifier option array;
  mutable current_index: int;
}
*)


(* remainders whose best split literal are of this depth *)
type candidates = {
  cd_config: config;
  cd_bound: bound;
  cd_statistic: statistic;
  cd_state: state;
  cd_context: context;

  (* candidates not considered yet *)
  mutable cd_unprocessed: raw_context_unifier list;

  (* the remainders currently valid for adding to the context
     (at least valid when last checked). *)
  (*mutable*) cd_valid: Valid.t(*heap*);
(*  mutable cd_history: (choice_point * Valid.heap) list;*)

  (* to save memory only a limited amount of candidates is stored
     with full informatio nin cd_valid,
     the others are kept here *)
  mutable cd_invalid: (raw_context_unifier * choice_point) list;

  (* candidates added to the context (at step) *)
  mutable cd_added: (candidate * literal * int * choice_point) list;

(*  (* lookahead for contradictions within unit split candidates *)
  cd_lookahead: selection_lookahead;
  cd_lookahead_exceeding: selection_lookahead;*)

  mutable cd_exceeding: (raw_context_unifier * literal * Bound.complexity) list;

  mutable cd_last_retracted: (candidate * literal * int * choice_point list) list;
  mutable cd_last_retracted_done: (candidate * literal * int * choice_point list) list;

(*  mutable cd_compressed: compressed_candidates list;*)
}












    
(*** creation ***)




let create (config: config) (bound: bound) (statistic: statistic) (state: state)
  (context: context) : candidates =

  {
    cd_config = config;
    cd_bound = bound;
    cd_statistic = statistic;
    cd_state = state;
    cd_context = context;
    cd_unprocessed = [];
    cd_valid = Valid.create () (*Valid.empty*);
(*    cd_history = [];*)
    cd_invalid = [];
    cd_added = [];
(*    cd_lookahead = Selection_lookahead.create 100000;
    cd_lookahead_exceeding = Selection_lookahead.create 10000;*)
    cd_exceeding = [];
    cd_last_retracted = [];
    cd_last_retracted_done = [];
(*    cd_compressed = [];*)
  }




(*** add ***)




(*** build ***)


(* compute the quality of a term based on the deepening bound criterion. *)
let get_term_quality (candidates: candidates) (literal: literal) : int =
  if Config.iterative_deepening candidates.cd_config = Flags.IT_TermDepth then
    Term_attributes.weight_of_literal literal
  else
    Term_attributes.depth_of_literal literal



let find_best_candidate_literal (candidates: candidates) (candidate_literals: candidate_literal list) :
  (candidate_literal * pureness * int * Bound.complexity) =
  let best =
    List.fold_left
      (fun acc candidate_literal ->
	 (* is this a remainder literal? *)
	 if not candidate_literal.cl_is_remainder_literal then begin
	   acc
	 end

	 else begin
	   let pureness = Term_attributes.get_pureness candidate_literal.cl_literal.Term.term
	   and term_quality = get_term_quality candidates candidate_literal.cl_literal
	   and complexity = candidates.cd_bound#get_complexity candidate_literal.cl_literal
	   in
	   let cmp =
	     match acc with
	       | None ->
		   1
		     
	       | Some (best_candidate_literal, best_pureness, best_term_quality, best_complexity) ->
		   (* if possible ignore candidates exceeding the depth bound *)
		   if
		     candidates.cd_bound#exceeds best_complexity
		     &&
		     candidates.cd_bound#exceeds complexity
		   then
		     0
		       
		   else
		     let cmp_complexity =
		       if candidates.cd_bound#exceeds complexity then
			 -1
		       else if candidates.cd_bound#exceeds best_complexity then
			 1
		       else
			 0
		     in
		       if cmp_complexity <> 0 then
			 cmp_complexity
		       else
			 
			 let cmp_pureness =
			   Term_attributes.cmp_pureness best_pureness pureness
			 in
			   if cmp_pureness <> 0 then
			     cmp_pureness
			   else
			     
			     let cmp_term_quality =
			       compare best_term_quality term_quality
			     in
			       if cmp_term_quality <> 0 then
				 cmp_term_quality
			       else
				 Term_attributes.cmp_universality best_pureness pureness
	   in
	     if cmp <= 0 then
	       acc
	     else
	       Some (candidate_literal, pureness, term_quality, complexity)
	 end
      )
      None
      candidate_literals
  in
    match best with
      | None ->
	  failwith "find_split_literal"
	    
      | Some best ->
	  best



(* denotes that a check of a candidate against the context failed
   because of this context element. *)
exception CONFLICT of Context.element

(* builds the candidate literals generated by a context unifier.

   always returns the list in the same order for a context unifier
   in order to make the derivation more stable to (conceptually) independent changes,
   e.g. when find_best_candidate_literal picks the first of the remainder literals
   equal to the heuristics.

   if ~check_productivity is true and the productivity check is enabled,
   the candidate literals are checked for productivity while being built.
   then (and only then) CONFLICT is raised.
*)
let build_candidate_literals ?(check_productivity: bool = false)
    (config: config) (context: context) (raw_context_unifier: raw_context_unifier)
    : candidate_literal list =

  let subst =
    Context_unifier.recompute_unifier raw_context_unifier
  in
    
  let remainder_states =
    Context_unifier.compute_remainder_states raw_context_unifier subst
  in

  let admissible =
    Admissible.make_admissible config raw_context_unifier subst remainder_states
  in

    if Const.debug then begin
      let remainder_choice_points_admissible =
	Context_unifier.compute_remainder_states raw_context_unifier subst
      in
	if not (remainder_states = remainder_choice_points_admissible) then begin
	  failwith "Selection_split.remainder_choice_points";
	end;
    end;

    let input_partners =
      raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_input_partners
    in

    let remainder_literals =
      Array.make
	(Array.length input_partners)
	{
	  cl_context_literal = Context.null_element;
	  cl_literal = Term.null_literal;
	  cl_is_remainder_literal = false;
	  cl_index = -1;
	}
    in

      Array.iter
	(fun input_partner ->
	   let index =
	     input_partner.Context_unifier.ip_index
	   in
	  
	   let context_literal =
	     raw_context_unifier.Context_unifier.rcu_context_partners.(index).Context_unifier.cp_element;
	   in
	  
	   let split_literal =
	     Subst.apply_to_literal 
	       admissible 
	       input_partner.Context_unifier.ip_literal
	       Subst.input_literal_offset 
	   in

	     (* if this literal is not productive abort the computation. *)
	     if check_productivity && Config.productivity config then begin
	       let check_status =
		 Context.check_productive
		   context
		   context_literal.Context.el_literal
		   split_literal
	       in
		 match check_status with
		   | None ->
		       ()
			 
		   | Some conflict ->
		       raise (CONFLICT conflict)
	     end;
	     
	  
	   let remainder_literal = {
	     cl_context_literal = context_literal;
	     cl_literal = split_literal;
	     cl_is_remainder_literal = remainder_states.(index);
	     cl_index = index;
	   }
	   in
	     remainder_literals.(index) <- remainder_literal
	)
	input_partners;
      Array.to_list remainder_literals




(*
let remove_empty_compressed candidates =
  candidates.cd_compressed <-
    List.find_all
    (fun compressed ->
       match compressed.front with
	 | None -> false
	 | Some _ -> true
    )
    candidates.cd_compressed


let compress_candidates candidates =
  if Valid.size candidates.cd_valid >= max_uncompressed then begin
(*    Statistic.inc_global_debug ();
    print_endline ("COMPRESS: " ^ string_of_int (Statistic.get_global_debug ()) ^ " - " ^ string_of_int (List.length candidates.cd_compressed));*)
    remove_empty_compressed candidates;

    let candidate =
      Valid.find_min candidates.cd_valid
    in
      candidates.cd_valid <- Valid.remove_min candidates.cd_valid;

	  let candidate_literals =
	    build_candidate_literals ~check_productivity:false
	      candidates.cd_config candidates.cd_context
	      candidate.ca_raw_context_unifier
	  in
	  let (best_candidate_literal, _, _, _) =
	    find_best_candidate_literal candidates candidate_literals
	  in
	  let split_literal, index =
	    best_candidate_literal.cl_literal, best_candidate_literal.cl_index
	  in
(*
  let assert_literal, index =
    build_assert_literal candidates first.ca_raw_context_unifier
  in
*)
    let new_compressed =
      Array.create
	(Valid.size candidates.cd_valid)
	None
    in
    let counter =
      ref 0
    in
      while not (Valid.is_empty candidates.cd_valid) do
	new_compressed.(!counter) <- Some (Valid.remove_min candidates.cd_valid).ca_raw_context_unifier;
	counter := !counter + 1;
      done;
      (*
      Valid.iter_increasing
	(fun candidate ->
	   new_compressed.(!counter) <- candidate.ca_raw_context_unifier;
	)
	candidates.cd_valid;
      *)
      candidates.cd_valid <- Valid.create ();
      candidates.cd_compressed <-
	{
	  front = Some (candidate, split_literal, index);
	  candidates = new_compressed;
	  current_index = 0;
	}
	:: candidates.cd_compressed;
      
  end
*)

(*** representation ***)

let remainder_literals_to_string_ (candidate_literals: candidate_literal list) : string =
  let candidate_literals_repr =
    List.fold_left
      (fun acc candidate_literal ->
	 if candidate_literal.cl_is_remainder_literal then
	   Term.literal_to_string candidate_literal.cl_literal :: acc
	 else
	   "___" :: acc
      )
      []
      candidate_literals
  in
    "Remain.: " ^ "[" ^ String.concat ", " (List.rev candidate_literals_repr) ^ "]" ^ "\n"



let remainder_literals_to_string (config: config) (context: context)
  (raw_context_unifier: raw_context_unifier) : string =

  let remainder_literals =
    build_candidate_literals config context raw_context_unifier
  in
    remainder_literals_to_string_ remainder_literals




let remainder_to_string_ (candidates: candidates) (title: string)
  (remainder: candidate) (best_remainder_literal: literal) (remainder_literals: candidate_literal list) : string =
    State.active_choice_point_to_string candidates.cd_state
    ^ title ^ ": " ^ Term.literal_to_string best_remainder_literal ^ "\n"
    ^
    if Config.print_derivation_context_unifier candidates.cd_config then
      Context_unifier.raw_context_unifier_to_string remainder.ca_raw_context_unifier
      ^ remainder_literals_to_string_ remainder_literals
    else
      ""

let remainder_to_string (candidates: candidates) (title: string) (remainder: candidate) : string =
  let remainder_literals =
    build_candidate_literals
      candidates.cd_config candidates.cd_context remainder.ca_raw_context_unifier
  in
  let (best_candidate_literal, _, _, _) =
    find_best_candidate_literal candidates remainder_literals
  in
    remainder_to_string_ candidates title remainder best_candidate_literal.cl_literal remainder_literals






(*** context checks ***)

(* checks all candidate literals for productivity.
   returns on the first non-productive one. *)
let rec check_productivity (context: context) (candidate_literals: candidate_literal list)
  : Context.element option =

  match candidate_literals with
    | [] ->
	None
	
    | candidate_literal :: tail ->
	let check_status =
	  Context.check_productive
	    context
	    candidate_literal.cl_context_literal.Context.el_literal
	    candidate_literal.cl_literal
	in
	  begin
	    match check_status with
	      | None ->
		  check_productivity context tail
	    
	      | Some _ ->
		  check_status
	  end


let check_candidate_literals (candidates: candidates)
  (candidate_literals: candidate_literal list) (best_candidate_literal: literal) :
  Context.element option =

  (* the split literal must not be contradictory with the context *)
  let contradictory_element =
    Context.check_contradictory candidates.cd_context best_candidate_literal
  in
    match contradictory_element with
      | Some _ ->
	  contradictory_element
	  
      | None ->
	  let productivity_preventing_element =
	    if Config.productivity candidates.cd_config then
	      check_productivity candidates.cd_context candidate_literals
	    else
	      None
	  in begin
	      match productivity_preventing_element with
		| None ->
		    (* the skolemized negated split literal
		       must not be contradictory with the context.

		       this is equivalent to checking that 
		       the split literal is not p-subsumed by the context.
		       
		       this check is superfluous after a productivity check.
		    *)
		    if Config.productivity candidates.cd_config then
		      None
		    else
		      Context.check_subsumed candidates.cd_context best_candidate_literal

		| _ ->
		    productivity_preventing_element
	    end





(*** add ***)


let add_unprocessed (candidates: candidates) (candidate: raw_context_unifier) : unit =
  candidates.cd_unprocessed <- candidate :: candidates.cd_unprocessed

let add (candidates: candidates) (candidate: raw_context_unifier) : unit =
  Statistic.inc_computed_split_candidates candidates.cd_statistic;
  add_unprocessed candidates candidate



(* adds a candidate to the valid candidate queue.
   returns the candidate if the close lookahead is successful. *)
let add_valid (candidates: candidates) (candidate: candidate) (split_literal: literal) : unit =
(*  compress_candidates candidates;*)
(*  Valid.iter
    (fun x ->
       if candidate == x then
	 failwith "hier";
    )
    candidates.cd_valid;
*)
  try
    Valid.add candidates.cd_valid candidate
(*    candidates.cd_valid <- Valid.add candidate candidates.cd_valid*)
  with
    | Valid.OVERFLOW ->
	raise (Const.NO_SOLUTION "Selection_split.add_valid: candidate set overflow")




let do_build_candidate (candidates: candidates)
    (raw_context_unifier: raw_context_unifier)
    (remainder_size: int) (pureness: pureness) (term_quality: int)
    : candidate =

  let quality = {
    cq_remainder_size = remainder_size;
    cq_pureness = pureness;
    cq_term_quality = term_quality;
    cq_generation = Context_unifier.generation_of_candidate raw_context_unifier.Context_unifier.rcu_context_partners;
  }
  in
  let candidate = {
    ca_raw_context_unifier = raw_context_unifier;
    ca_quality = quality;
  }
  in
    candidate


let build_candidate
  (candidates: candidates) (raw_context_unifier: raw_context_unifier)
  (old_quality: quality option)
  : (candidate * literal * int) option =
  try
    let candidate_literals =
      build_candidate_literals ~check_productivity:true
	candidates.cd_config candidates.cd_context raw_context_unifier
    in

    let best_candidate_literal, pureness, term_quality, complexity =
      find_best_candidate_literal candidates candidate_literals
    in
      
    let split_literal =
      best_candidate_literal.cl_literal;
    in

    let remainder_size =
      List.fold_left
	(fun acc remainder_literal ->
	   if remainder_literal.cl_is_remainder_literal then
	     acc + 1
	   else
	     acc
	)
	0
	candidate_literals
    in

      (* exceeding *)
      if candidates.cd_bound#exceeds complexity then begin
(*
	(* lookahead check *)
	if
	  Config.lookahead_exceeding candidates.cd_config
 	  && 
	  (remainder_size = 1)
	  &&
	  (
	    Selection_lookahead.check candidates.cd_lookahead split_literal 
	    ||
	    Selection_lookahead.check candidates.cd_lookahead_exceeding split_literal 
	  )
	then begin
	  Some (do_build_candidate candidates raw_context_unifier remainder_size pureness term_quality, split_literal, best_candidate_literal.cl_index)
	end

	else*) begin
(*
	  (* lookahead registration *)
	  if
	    Config.lookahead_exceeding candidates.cd_config
 	    && 
	    (remainder_size = 1)
	  then begin
	    Selection_lookahead.add candidates.cd_lookahead_exceeding split_literal raw_context_unifier;
	  end;
*)
	  begin
	    (* keep candidates? *)
	    match Config.restart candidates.cd_config with
	      | Flags.RS_Eager
	      | Flags.RS_Lazy ->
		  (* no, immediately drop the exceeding candidates *)
		  ignore (
		    candidates.cd_bound#register
		      complexity
		      (Context_unifier.get_creation_choice_point raw_context_unifier.Context_unifier.rcu_context_partners));
		  
	    | Flags.RS_Delayed ->
		(* keep the exceeding candidates till an open branch is found. *)
		candidates.cd_exceeding <-
		  (raw_context_unifier, split_literal, complexity) :: candidates.cd_exceeding;
	  end;
	
	  None
	end
      end

      (* regular candidates within the deepening bound. *)	
      else begin
	(* perform applicability checks on the context. *)
	begin
	  match Context.check_contradictory candidates.cd_context split_literal with
	    | None -> ()
	    | Some conflict -> raise (CONFLICT conflict)
	end;
	
	(* subsumption not needed after productivity check. *)
	if not (Config.productivity candidates.cd_config) then begin
	  match Context.check_subsumed candidates.cd_context split_literal with
	    | None -> ()
	    | Some conflict -> raise (CONFLICT conflict)
	end;

	(* lookahead check *)
	let candidate =
	  do_build_candidate candidates raw_context_unifier remainder_size pureness term_quality
	in
(*
	  if
	    (remainder_size = 1)
	    &&
	    (
	      Selection_lookahead.check candidates.cd_lookahead split_literal 
	      ||
		Selection_lookahead.check candidates.cd_lookahead_exceeding split_literal 
	    )
	  then
	    (* success, close the branch with this candidate. *)
	    Some (candidate, split_literal, best_candidate_literal.cl_index)

	  else*) begin
	    (* an applicable candidate has been found, keep it. *)
	    if Config.print_split_candidates candidates.cd_config then begin
	      print_string (remainder_to_string_
			      candidates
			      "New Split Candidate"
			      candidate
			      split_literal
			      candidate_literals);
	    end;
	    (*
	    if candidate.ca_quality.cq_remainder_size = 1 then
	      Selection_lookahead.add candidates.cd_lookahead split_literal candidate.ca_raw_context_unifier;
	    *)
	    add_valid candidates candidate split_literal;
	    
	    None
	  end
      end 
  with
    | CONFLICT conflicting_element ->
	(* candidate is not applicable *)
	begin
	  if Context.is_element_mandatory_for_choice_point
	    candidates.cd_config
	    conflicting_element
	    (Context_unifier.get_creation_cid raw_context_unifier.Context_unifier.rcu_context_partners)
	  then
	    (* permanently not applicable, so drop it. *)
	    ()

	  else
	    (* move it to the invalid candidats *)
	    candidates.cd_invalid <-
  	      (raw_context_unifier, conflicting_element.Context.el_choice_point) :: candidates.cd_invalid
	end;
	
	None
	    
	



    


let rec build_unprocessed_candidates (candidates: candidates) : (candidate * literal * int) option =
  match candidates.cd_unprocessed with
    | [] ->
	None

    | candidate :: tail ->
	candidates.cd_unprocessed <- tail;
	begin
	  match build_candidate candidates candidate None with
	    | None ->
		build_unprocessed_candidates candidates
		  
	    | Some _ as conflicting ->
		conflicting
	end









(*** select ***)


let rec get_best_candidate_uncompressed (candidates: candidates) : (candidate * literal * int) option =
  if Valid.is_empty candidates.cd_valid then
    None

  else begin
    let candidate =
      Valid.(*find*)remove_min candidates.cd_valid
    in
(*      candidates.cd_valid <- Valid.remove_min candidates.cd_valid;*)

      if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
(*	      let x: candidate =*)
(*		candidates.cd_valid <- Valid.remove_min candidates.cd_valid;*)
(*	      in
		if x != candidate then begin
		  failwith "OGA 1";
		end;*)
	(* candidate does no longer exist (due to backtracking) *)
	get_best_candidate_uncompressed candidates;
      end

      else begin
	try
	  let candidate_literals =
	    build_candidate_literals ~check_productivity:true
	      candidates.cd_config candidates.cd_context
	      candidate.ca_raw_context_unifier
	  in
	  let (best_candidate_literal, _, _, _) =
	    find_best_candidate_literal candidates candidate_literals
	  in
	  let split_literal =
	    best_candidate_literal.cl_literal
	  in
(*
	    if candidate.ca_quality.cq_remainder_size = 1 then begin
	      Selection_lookahead.remove candidates.cd_lookahead split_literal candidate.ca_raw_context_unifier;
	    end;
*)
	    (* perform applicability checks on the context. *)
	    begin
	      match Context.check_contradictory candidates.cd_context split_literal with
		| None -> ()
		| Some conflict -> raise (CONFLICT conflict)
	    end;
	
	    (* subsumption not needed after productivity check. *)
	    if not (Config.productivity candidates.cd_config) then begin
	      match Context.check_subsumed candidates.cd_context split_literal with
		| None -> ()
		| Some conflict -> raise (CONFLICT conflict)
	    end;
	    
	    (* applicable candidate found *)
	    Some (candidate, split_literal, best_candidate_literal.cl_index)

	with
	  | CONFLICT conflicting_element ->
	      (* candidate is not applicable *)
	      begin
		if Context.is_element_mandatory_for_choice_point
		  candidates.cd_config
		  conflicting_element
		  (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
		then
		  ()
		else
		  candidates.cd_invalid <-
		    (candidate.ca_raw_context_unifier, conflicting_element.Context.el_choice_point) :: candidates.cd_invalid
	      end;
	      
(*	      let x: candidate =*)
(*		candidates.cd_valid <- Valid.remove_min candidates.cd_valid;*)
(*	      in
		if x != candidate then begin
		  failwith "OGA 2";
		end;*)

	      get_best_candidate_uncompressed candidates;
(*	      get_best_candidate candidates;*)
      end
  end



(*
let rec get_best_candidate_uncompressed (candidates: candidates) : (candidate * literal * int) option =

  if Valid.is_empty candidates.cd_valid then begin
    None
  end

  else begin
    let candidate =
      Valid.min candidates.cd_valid
    in
      (*
    let assert_literal, index =
      build_assert_literal candidates candidate.ca_raw_context_unifier
    in
      *)
      
(*      Selection_lookahead.remove candidates.cd_lookahead assert_literal candidate.ca_raw_context_unifier;*)
      
      if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	(* candidate does no longer exist (due to backtracking) *)
	      let x: candidate =
		Valid.remove_min candidates.cd_valid
	      in
	get_best_candidate_uncompressed candidates;
      end
	
      else begin
	match Context.check_subsumed candidates.cd_context assert_literal with
	  | Some element ->
	      (* candidate is invalid with the context *)
	      if
		Context.is_element_mandatory_for_choice_point
		  candidates.cd_config
		  element
		  (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
	      then begin
		()
	      end
		
	      else begin
		candidates.cd_invalid <-
		(candidate.ca_raw_context_unifier, element.Context.el_choice_point) :: candidates.cd_invalid;
	      end;
	      
	      let x: candidate =
		Valid.remove_min candidates.cd_valid
	      in
	      get_best_candidate_uncompressed candidates;
	      
	  | None ->
	      (* valid candidate found *)
	      Some (candidate, assert_literal, index)
      end
  end
*)
(*
let remove_compressed candidates (compressed_candidates: compressed_candidates) : unit =
  match compressed_candidates.candidates.(compressed_candidates.current_index) with
    | None ->
	compressed_candidates.front <- None

    | Some raw_context_unifier ->
	  let candidate_literals =
	    build_candidate_literals ~check_productivity:false
	      candidates.cd_config candidates.cd_context
	      raw_context_unifier
	  in
	  let best_candidate_literal, pureness, term_quality, complexity =
	    find_best_candidate_literal candidates candidate_literals
	  in
	  let split_literal, index =
	    best_candidate_literal.cl_literal, best_candidate_literal.cl_index
	  in
    let remainder_size =
      List.fold_left
	(fun acc remainder_literal ->
	   if remainder_literal.cl_is_remainder_literal then
	     acc + 1
	   else
	     acc
	)
	0
	candidate_literals
    in

	let candidate =
	  do_build_candidate candidates raw_context_unifier remainder_size pureness term_quality
	in
	  compressed_candidates.candidates.(compressed_candidates.current_index) <- None;
	  if compressed_candidates.current_index < Array.length compressed_candidates.candidates - 1 then begin
	    compressed_candidates.current_index <- compressed_candidates.current_index + 1
	  end;
	  compressed_candidates.front <- Some (candidate, split_literal, index)

	  
let rec advance_compressed candidates (compressed_candidates: compressed_candidates) : unit =

  match compressed_candidates.front with
    | None ->
	()

    | Some (candidate, split_literal, index) ->
	  if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	    (* candidate does no longer exist (due to backtracking) *)
	    remove_compressed candidates compressed_candidates;
	    advance_compressed candidates compressed_candidates;
	  end
	    
	  else begin
	    try
	      begin
		match Context.check_contradictory candidates.cd_context split_literal with
		  | None -> ()
		  | Some conflict -> raise (CONFLICT conflict)
	      end;
	      
	      (* subsumption not needed after productivity check. *)
	      if not (Config.productivity candidates.cd_config) then begin
		match Context.check_subsumed candidates.cd_context split_literal with
		  | None -> ()
		  | Some conflict -> raise (CONFLICT conflict)
	      end;
	      
	  let candidate_literals =
	    build_candidate_literals ~check_productivity:true
	      candidates.cd_config candidates.cd_context
	      candidate.ca_raw_context_unifier
	  in

	      begin	
		  (* valid candidate found *)
		  ()
	      end
	    with
	      | CONFLICT conflicting_element ->
	(* candidate is not applicable *)
	begin
	  if Context.is_element_mandatory_for_choice_point
	    candidates.cd_config
	    conflicting_element
	    (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
	  then
	    (* permanently not applicable, so drop it. *)
	    ()

	  else
	    (* move it to the invalid candidats *)
	    candidates.cd_invalid <-
  	      (candidate.ca_raw_context_unifier, conflicting_element.Context.el_choice_point) :: candidates.cd_invalid
	end;

		  remove_compressed candidates compressed_candidates;
		  advance_compressed candidates compressed_candidates;
	  end
*)
(*


let rec advance_compressed candidates (compressed_candidates: compressed_candidates) : unit =

  match compressed_candidates.front with
    | None ->
	()

    | Some (candidate, assert_literal, index) ->
	  if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	    (* candidate does no longer exist (due to backtracking) *)
	    remove_compressed candidates compressed_candidates;
	    advance_compressed candidates compressed_candidates;
	  end
	    
	  else begin
	    match Context.check_subsumed candidates.cd_context assert_literal with
	      | Some element ->
		  (* candidate is invalid with the context *)
		  if
		    Context.is_element_mandatory_for_choice_point
		      candidates.cd_config
		      element
		      (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
		  then begin
		    ()
		  end
		    
		  else begin
		    candidates.cd_invalid <-
		      (candidate.ca_raw_context_unifier, element.Context.el_choice_point) :: candidates.cd_invalid;
		  end;
		  
		  remove_compressed candidates compressed_candidates;
		  advance_compressed candidates compressed_candidates;
	      
	      | None ->
		  (* valid candidate found *)
		  ()
	  end
*)


let rec get_best_candidate (candidates: candidates) : (candidate * literal * int) option =
  get_best_candidate_uncompressed candidates
(*
  (* process all candidate queues till valid candidate is in front *)
(*  List.iter
    (advance_compressed candidates)
    candidates.cd_compressed;*)

  (* get best compressed *)
(*  let best_compressed (*: compressed_candidates option*) =
    None
(*
    List.fold_left
      (fun acc compressed ->
	 match acc with
	   | None ->
	       Some compressed

	   | Some old ->
	       begin
		 match old.front, compressed.front with
		   | None, None ->
		       None
			 
		   | None, Some _ ->
		       Some compressed
			 
		   | Some _, None ->
		       acc
			 
		   | Some (x, _, _), Some (y, _, _) ->
		       if compare_candidates x y > 0 then
			 Some compressed
		       else
			 acc
	       end
      )
      None
      candidates.cd_compressed*)
  in*)
    begin
      let best_uncompressed =
	get_best_candidate_uncompressed candidates
      in
(*	match best_uncompressed with
	  | None ->
	      begin
		match best_compressed with
		  | None ->
		      None

		  | Some best ->
		      let x =
		      best.front
		      in
		      remove_compressed candidates best;
		      x
	      end
		
	  | Some (candidate, assert_literal, index) ->*)
	      begin
		match best_compressed with
		  | None ->
(*		      let x: candidate =*)
			candidates.cd_valid <- Valid.remove_min candidates.cd_valid;
(*		      in
			best_uncompressed*)
			
		  | Some best ->
		      begin
			match best.front with
			  | None ->
(*			      failwith "bst_frnt"*)
		      let x: candidate =
			Valid.remove_min candidates.cd_valid
		      in
			best_uncompressed


			  | Some (candidate', assert_literal', index') ->
			      if compare_candidates candidate candidate' <= 0 then begin
				let x: candidate =
				  Valid.remove_min candidates.cd_valid
				in
				  best_uncompressed
			      end
				
			      else begin
		      let x =
		      best.front
		      in
		      remove_compressed candidates best;
		      x
			      end
		      end
	      end
    end
*)    
    (* compare with best uncompressed *)
    (*

let rec get_best_candidate (candidates: candidates) : (candidate * literal * int) option =
  if Valid.is_empty candidates.cd_valid then
    None

  else begin
    let candidate =
      Valid.remove_min candidates.cd_valid
    in
      if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then begin
	(* candidate does no longer exist (due to backtracking) *)
	get_best_candidate candidates;
      end

      else begin
	try
	  let candidate_literals =
	    build_candidate_literals ~check_productivity:true
	      candidates.cd_config candidates.cd_context
	      candidate.ca_raw_context_unifier
	  in
	  let (best_candidate_literal, _, _, _) =
	    find_best_candidate_literal candidates candidate_literals
	  in
	  let split_literal =
	    best_candidate_literal.cl_literal
	  in
(*
	    if candidate.ca_quality.cq_remainder_size = 1 then begin
	      Selection_lookahead.remove candidates.cd_lookahead split_literal candidate.ca_raw_context_unifier;
	    end;
*)
	    (* perform applicability checks on the context. *)
	    begin
	      match Context.check_contradictory candidates.cd_context split_literal with
		| None -> ()
		| Some conflict -> raise (CONFLICT conflict)
	    end;
	
	    (* subsumption not needed after productivity check. *)
	    if not (Config.productivity candidates.cd_config) then begin
	      match Context.check_subsumed candidates.cd_context split_literal with
		| None -> ()
		| Some conflict -> raise (CONFLICT conflict)
	    end;
	    
	    (* applicable candidate found *)
	    Some (candidate, split_literal, best_candidate_literal.cl_index)

	with
	  | CONFLICT conflicting_element ->
	      (* candidate is not applicable *)
	      begin
		if Context.is_element_mandatory_for_choice_point
		  candidates.cd_config
		  conflicting_element
		  (Context_unifier.get_creation_cid candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners)
		then
		  ()
		else
		  candidates.cd_invalid <-
		    (candidate.ca_raw_context_unifier, conflicting_element.Context.el_choice_point) :: candidates.cd_invalid
	      end;
	      
	      get_best_candidate candidates;
      end
  end


*)


let check_exceeding candidates =

  let rec check_exceeding exceeding =
    match exceeding with
      | [] ->
	  ()
	    
      | (raw_context_unifier, split_literal, complexity) :: tail ->
(*	      let a = ref false
	      and b = ref false
	      in*)
	  let choice_point =
	    (Context_unifier.get_creation_choice_point raw_context_unifier.Context_unifier.rcu_context_partners)
	  in
	    (* is this literal worth registering? *)
	    if candidates.cd_bound#exceeds_current complexity choice_point then begin
	      try
		let candidate_literals =
		  build_candidate_literals ~check_productivity:true
		    candidates.cd_config candidates.cd_context raw_context_unifier
		in
		  if
		    List.exists
		      (fun candidate_literal ->
			 candidate_literal.cl_is_remainder_literal
			 &&
			   (
			     (
			       begin
				 match Context.check_contradictory candidates.cd_context candidate_literal.cl_literal with
				   | None -> false
				   | _ -> (*print_endline ("0"); *)true
			       end
			     )
			 ||
			   Context.is_satisfied candidates.cd_context candidate_literal.cl_literal
			   )
		      )
		      candidate_literals
		  then
		    ()
		  else
		    let exceeding: bool =
		      candidates.cd_bound#register
			complexity
			choice_point
		    in
(*		      a := true;*)
		      ()
	      with
		| CONFLICT _ ->
(*		    print_endline ("1"); *)
		    ()
	    end;
(*

	      if
		Context.is_satisfied candidates.cd_context split_literal
	      then
		()
	      else if
		begin
		  let candidate_literals =
		    build_candidate_literals ~check_productivity:false
		      candidates.cd_config candidates.cd_context
		      raw_context_unifier
		  in
		    (* is any of the remainder literals satisfied by the context? *)
		    if
		      List.exists
			(fun remainder_literal ->
			   remainder_literal.cl_is_remainder_literal
			   &&
			     (
			     Context.is_satisfied candidates.cd_context remainder_literal.cl_literal
			)
			candidate_literals
		    end
		then
		  ()
			
		else

		let exceeding: bool =
		  candidates.cd_bound#register
		    complexity
		    choice_point
		in
(*		  print_endline ("A exceeding");
		  a := true;*)
		  ()

	      with
		| CONFLICT conflicting_element ->
		    ()
	    end;*)
(*
		begin
		  match build_candidate candidates raw_context_unifier None with
		    | None ->
			print_endline ("NONE:");
			()
			  
		    | Some (candidate, literal, index) ->
			if
			  Context.is_satisfied candidates.cd_context literal
			then
			  ()
			else
			  let exceeding: bool =
			    candidates.cd_bound#register
			      complexity
			      choice_point
			  in
			    print_endline ("A exceeding");
			    a := true;
			    ()
			      
		end;
*)
(*
		begin
		  let candidate_literals =
		    build_candidate_literals ~check_productivity:false
		      candidates.cd_config candidates.cd_context
		      raw_context_unifier
		  in
		    (* is any of the remainder literals satisfied by the context? *)
		    if
		      List.exists
			(fun remainder_literal ->
			   if remainder_literal.cl_is_remainder_literal then
			     print_endline ("X " ^ Term.literal_to_string remainder_literal.cl_literal);

			   remainder_literal.cl_is_remainder_literal
			   &&
			     Context.is_satisfied candidates.cd_context remainder_literal.cl_literal
			)
			candidate_literals
		    then
		      ()
			
		    else begin
		      (* no, so register it *)
(*		      let exceeding =
			candidates.cd_bound#register
			  complexity
			  choice_point
		      in*)
			print_endline ("B exceeding");
			b := true;
			()
		    end;
(*
		    (* for exhaustive restarting any exceeding candidate is sufficient *)
		    if 
		      exceeding
		      && 
			Config.restart candidates.cd_config = Flags.RS_Exhaustive
		    then
		      raise Exit*)
		end;
		
(*	    end;*)
	    if !a != !b then begin
	      Context.print_context candidates.cd_context stdout;
	      print_endline (Context_unifier.raw_context_unifier_to_string raw_context_unifier);
	      print_endline ("AHA");
	      print_newline ();
	      print_newline ();
(*	      failwith "aha";*)
	    end;
*)

 check_exceeding tail
  in
    try
      check_exceeding candidates.cd_exceeding
    with
      | Exit ->
	  ()
	  


let split (candidates: candidates) (candidate: candidate) (split_literal: literal) (index: int) (replay: Random_jump.replay option) : selected =
  (* unit split? *)
  if candidate.ca_quality.cq_remainder_size = 1 then begin
    let context_literals =
      Context_unifier.get_context_literals
	candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners
    in
      State.apply_assert
	candidates.cd_state
	split_literal
	candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
	candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
	index
	context_literals;

      let assert_choice_point =
	State.active_choice_point candidates.cd_state
      in
	candidates.cd_invalid <-
	  (candidate.ca_raw_context_unifier, assert_choice_point) :: candidates.cd_invalid;

      let creation_choice_point =
	Context_unifier.get_creation_choice_point
	  candidate.ca_raw_context_unifier.Context_unifier.rcu_context_partners
      in
      
      {
	Selection_types.candidate_type = Selection_types.UnitSplit;
	Selection_types.literal = split_literal;
	Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
      }
    end
	      
    else begin
      match replay with
	| None ->

(*
	    print_endline ("HISTORY UPDATE:");
	    Valid.iter
	      (fun candidate ->
		 print_endline (remainder_literals_to_string candidates.cd_config candidates.cd_context candidate.ca_raw_context_unifier);
	      )
	      candidates.cd_valid;*)
(*    candidates.cd_history <- (State.active_choice_point candidates.cd_state, candidates.cd_valid) :: candidates.cd_history;*)
(*    candidates.cd_valid <- Valid.empty;*)


      let split_choice_point: choice_point =
	State.apply_split_left
	  candidates.cd_state
	  split_literal
	  candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
	  index;
      in
	candidates.cd_added <-
	  (candidate, split_literal, index, split_choice_point) :: candidates.cd_added;
	
	{
	  Selection_types.candidate_type = Selection_types.SplitLeft;
	  Selection_types.literal = split_literal;
	  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
	}


	| Some (Random_jump.Split saved_literal) ->
	    if not (Term.are_literals_skolem_variants saved_literal split_literal) then begin
	      
(*	      Valid.iter
		(fun candidate ->
		   print_endline (remainder_to_string candidates "Valid" candidate)
		)
		candidates.cd_valid;
	      let next =
		Valid.find_min candidates.cd_valid
	      in*)
	      print_endline ("SAVED: " ^ Term.literal_to_string saved_literal);
	      print_endline ("FOUND: " ^ Term.literal_to_string split_literal);
	      print_endline (Context_unifier.raw_context_unifier_to_string candidate.ca_raw_context_unifier);
(*	      print_endline (remainder_to_string candidates "NEXT" candidate);
	      let cmp =
		compare_candidates candidate next
	      in
		print_endline (string_of_int cmp);*)
	      failwith "Selection_split.split.replay 1";
	      (*
		    let right_split_literal =
		      Term.request_negated_literal
			(Term.request_skolemized_literal split_literal)
		    in
	      State.apply_split_right
		candidates.cd_state
		right_split_literal
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
		index
		[];
		      
      let assert_choice_point =
	State.active_choice_point candidates.cd_state
      in
	candidates.cd_invalid <-
	  (candidate.ca_raw_context_unifier, assert_choice_point) :: candidates.cd_invalid;
		{
		  Selection_types.candidate_type = Selection_types.SplitRight;
		  Selection_types.literal = right_split_literal;
		  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
		}
	      *)
	    end

	    else begin
(*	      candidates.cd_history <- (State.active_choice_point candidates.cd_state, candidates.cd_valid) :: candidates.cd_history;*)


	      let split_choice_point: choice_point =
		State.apply_split_left
		  candidates.cd_state
		  split_literal
		  candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
		  index;
	      in
		candidates.cd_added <-
		  (candidate, split_literal, index, split_choice_point) :: candidates.cd_added;
		
		{
		  Selection_types.candidate_type = Selection_types.SplitLeft;
		  Selection_types.literal = split_literal;
		  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
		}
	    end

	| Some (Random_jump.Left saved_literal) ->
	    if not (Term.are_literals_skolem_variants saved_literal split_literal) then begin
	      
	      print_endline ("SAVED: " ^ Term.literal_to_string saved_literal);
	      print_endline ("FOUND: " ^ Term.literal_to_string split_literal);
	      print_endline (Context_unifier.raw_context_unifier_to_string candidate.ca_raw_context_unifier);
	      failwith "Selection_split.split.replay 2";
	      (*
		    let right_split_literal =
		      Term.request_negated_literal
			(Term.request_skolemized_literal split_literal)
		    in
	      State.apply_split_right
		candidates.cd_state
		right_split_literal
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
		index
		[];
		      
      let assert_choice_point =
	State.active_choice_point candidates.cd_state
      in
	candidates.cd_invalid <-
	  (candidate.ca_raw_context_unifier, assert_choice_point) :: candidates.cd_invalid;
		{
		  Selection_types.candidate_type = Selection_types.SplitRight;
		  Selection_types.literal = right_split_literal;
		  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
		}
	      *)
	    end

	    else begin
	      State.apply_split_right
		candidates.cd_state
		split_literal
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
		index
		[];
		      
      let assert_choice_point =
	State.active_choice_point candidates.cd_state
      in
	candidates.cd_invalid <-
	  (candidate.ca_raw_context_unifier, assert_choice_point) :: candidates.cd_invalid;
		{
		  Selection_types.candidate_type = Selection_types.SplitRight;
		  Selection_types.literal = split_literal;
		  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
		}
	    end


	| Some (Random_jump.Right saved_literal) ->
	    let right_split_literal =
	      Term.request_negated_literal
		(Term.request_skolemized_literal split_literal)
	    in
	    if not (Term.are_literals_skolem_variants saved_literal right_split_literal) then begin
	      
	      Valid.iter
		(fun candidate ->
		   print_endline (remainder_to_string candidates "Valid" candidate)
		)
		candidates.cd_valid;

	      print_endline ("SAVED: " ^ Term.literal_to_string saved_literal);
	      print_endline ("FOUND: " ^ Term.literal_to_string right_split_literal);
	      print_endline (Context_unifier.raw_context_unifier_to_string candidate.ca_raw_context_unifier);
	      failwith "Selection_split.split.replay 3";
	    end

	    else begin
	      State.apply_split_right
		candidates.cd_state
		right_split_literal
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
		candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
		index
		[];
	      
	      let assert_choice_point =
		State.active_choice_point candidates.cd_state
	      in
		candidates.cd_invalid <-
		  (candidate.ca_raw_context_unifier, assert_choice_point) :: candidates.cd_invalid;
		{
		  Selection_types.candidate_type = Selection_types.SplitRight;
		  Selection_types.literal = right_split_literal;
		  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
		}
	    end
    end


let select_right_split (candidates: candidates) : selected option =

  let rec select_right_split' () =
    match candidates.cd_last_retracted with
      | [] ->
	  None
	    
      | ((candidate, split_literal, index, backjump_level) as x) :: tail ->
	  candidates.cd_last_retracted <- tail;
	  
	  if
	    List.exists
	      State.is_choice_point_invalid
	      backjump_level
	  then begin
	    select_right_split' ()
	  end
	    
	  else begin
	    begin
	      match backjump_level with
		| [] ->
		    (* incomplete branch - arbitrary backtracking,
		       no actual right split learned.
		       apply this right split once and then forget about it. *)
		    ()
		      
		| _ ->
		    candidates.cd_last_retracted_done <- x :: candidates.cd_last_retracted_done;
	    end;
	    
	    (* right split already done by lemma unit propagation? *)
	    let contradictory_element =
	      Context.check_contradictory candidates.cd_context split_literal
	    in
	      match contradictory_element with
		| Some element ->
		    (* yes, so close *)
		    select_right_split' ()
		      
		| None ->
		    let right_split =
		      Term.request_negated_literal
			(Term.request_skolemized_literal split_literal)
		    in
		      State.apply_split_right
			candidates.cd_state
			right_split
			candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
			candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
			index
			backjump_level;
		      
		      (*		      failwith ("RIGHT SPLIT: " ^ Term.literal_to_string right_split);*)
		      (*	      print_endline ("RIGHT SPLIT: " ^ Term.literal_to_string right_split);*)
		      Some
			{
			  Selection_types.candidate_type = Selection_types.SplitRight;
			  Selection_types.literal = right_split;
			  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
			}
	  end
  in
    select_right_split' ()


let select (candidates: candidates) (replay: Random_jump.replay option) : selected option =
  (*
  begin
    match replay with
      | None ->
	  ()
	    
      | Some _ ->
	  failwith "replay";
  end;
  *)

  (*
  let right_split =
    match candidates.cd_last_retracted with
      | None ->
	  None

      | Some (candidate, split_literal, index, backjump_level) ->
	  candidates.cd_last_retracted <- None;

	  if
	    List.exists
	      State.is_choice_point_invalid
	      backjump_level
	  then
	    failwith "AHA";

	  (*
	  if (Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier) then begin
	    None
	  end

	  else*) begin
	    (* right split already done by lemma unit propagation? *)
	    let contradictory_element =
	      Context.check_contradictory candidates.cd_context split_literal
	    in
	      match contradictory_element with
		| Some _ ->
		    (* yes, so ignore the right split *)
		    None
		    
		| None ->
		    let right_split =
		      Term.request_negated_literal
			(Term.request_skolemized_literal split_literal)
		    in
		      State.apply_split_right
			candidates.cd_state
			right_split
			candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_clause
			candidate.ca_raw_context_unifier.Context_unifier.rcu_space.Context_unifier.cus_value
			index
			backjump_level;

(*		      failwith ("RIGHT SPLIT: " ^ Term.literal_to_string right_split);*)
		      (*	      print_endline ("RIGHT SPLIT: " ^ Term.literal_to_string right_split);*)
		      Some
			{
			  Selection_types.candidate_type = Selection_types.SplitRight;
			  Selection_types.literal = right_split;
			  Selection_types.raw_context_unifier = candidate.ca_raw_context_unifier;
			}
	  end
  in
  *)


  let right_split =
(*      print_endline (string_of_int (List.length candidates.cd_last_retracted));
      print_newline ();
      print_newline ();*)
      select_right_split candidates
  in

    match right_split with
      | Some _ ->
	  right_split
	  
      | None ->
	  begin

  let candidate =
    match build_unprocessed_candidates candidates with
      | Some _ as conflicting ->
(*	  failwith "conflicting";*)
	  conflicting

      | None ->
	  get_best_candidate candidates
  in

    match candidate with
      | None ->
	  None

      | Some (candidate, split_literal, index) ->
	  Some (split candidates candidate split_literal index replay)
	  end




(*** backtrack ***)

(* when using Selection_lookahead a candidate may be chosen
   before all candidates are built,
   thus some still valid candidates might be left *)
let backtrack_unprocessed (candidates: candidates) : unit =
  candidates.cd_unprocessed <-
  List.find_all
    (fun raw_context_unifier ->
       not (Context_unifier.is_raw_context_unifier_invalid raw_context_unifier)
    )
    candidates.cd_unprocessed

(*
let rec backtrack_invalid (candidates: candidates) : unit =
  match candidates.cd_history with
    | [] ->
	candidates.cd_valid <- Valid.empty

    | (cp, valid) :: tail ->
	candidates.cd_history <- tail;

	if State.is_choice_point_valid cp then begin
	  if not (State.choice_point_equal (State.active_choice_point candidates.cd_state) cp) then begin
	    failwith "IM BT";
	  end;
(*
	    print_endline ("HISTORY BACKTRACK:");
	    Valid.iter
	      (fun candidate ->
		 print_endline (remainder_literals_to_string candidates.cd_config candidates.cd_context candidate.ca_raw_context_unifier);
	      )
	      candidates.cd_valid;*)

	  candidates.cd_valid <- valid;
	end

	else begin
	  backtrack_invalid candidates;
	end
*)

let rec backtrack_invalid (candidates: candidates) : unit =
  candidates.cd_invalid <-
    List.find_all
      (fun (raw_context_unifier, conflicting_choice_point) ->
	 if Context_unifier.is_raw_context_unifier_invalid raw_context_unifier then
	   false
	     
	 else if State.is_choice_point_invalid conflicting_choice_point then begin
	   add_unprocessed candidates raw_context_unifier;
	   false
	 end
	   
	 else
	   true
      )
      candidates.cd_invalid



let backtrack_added (candidates: candidates) (retracted: choice_point) (explanation: choice_point list) : unit =
(*  print_endline ("backtrack_added: " ^ State.choice_point_to_string retracted);*)
  candidates.cd_added <-
    List.find_all
    (fun (candidate, split_literal, index, choice_point) ->
       if State.choice_point_equal choice_point retracted then begin
	 (* reenter it to being processed -
	    its invalid right now because the right split has preference.
	    but if the right split is not done due to unit propagation,
	    and another close occurs before the right split is even checked for,
	    this candidate becomes valid again.
	 *)
(*	 print_endline ("FOUND");*)
	 add_unprocessed candidates candidate.ca_raw_context_unifier;

(*	 begin
	   match explanation with
	     | [] ->
		 (* incomplete branch - arbitrary backtracking,
		    no actual right split learned *)
		 ()

	     | _ ->*)
		 candidates.cd_last_retracted <- (candidate, split_literal, index, explanation) :: candidates.cd_last_retracted;
(*	 end;*)

	 false;
       end
       
       else if Context_unifier.is_raw_context_unifier_invalid candidate.ca_raw_context_unifier then
	 false
	   
       else if (State.is_choice_point_invalid choice_point) then begin
	 ignore (add_valid candidates candidate split_literal);
	 false
       end
	 
       else
	 true
    )
    candidates.cd_added
  

(*
let rec backtrack_added (candidates: candidates) (retracted: choice_point) (explanation: choice_point list) : unit =
(*  print_endline ("backtrack_added: " ^ State.choice_point_to_string retracted);*)
  match candidates.cd_added with
    | [] ->
	()

    | (candidate, split_literal, index, choice_point) :: tail ->
	candidates.cd_added <- tail;

       if State.choice_point_equal choice_point retracted then begin
	 (* reenter it to being processed -
	    its invalid right now because the right split has preference.
	    but if the right split is not done due to unit propagation,
	    and another close occurs before the right split is even checked for,
	    this candidate becomes valid again.
	 *)
	 candidates.cd_last_retracted <- (candidate, split_literal, index, explanation) :: candidates.cd_last_retracted;
       end
       
       else
	 backtrack_added candidates retracted explanation
*)


let backtrack (candidates: candidates) (retracted: choice_point) (explanation: choice_point list) : unit =
  (* right splits can't be kept for random jumping *)
(*  if Config.random_jump candidates.cd_config then begin*)
    candidates.cd_last_retracted <- [];
(*  end
  else begin
  candidates.cd_last_retracted <-
    List.fold_left
    (fun acc ((_, _, _, backjump_level) as x) ->
       if List.for_all State.is_choice_point_valid backjump_level then
	  x :: acc
       else
	 acc
    )
    candidates.cd_last_retracted
    candidates.cd_last_retracted_done;
  end;*)
    candidates.cd_last_retracted_done <- [];

(*
    List.find_all
    (fun (_, _, _, backjump_level) ->
       List.for_all State.is_choice_point_valid backjump_level
    )
    candidates.cd_last_retracted;
    *)
(*
  Selection_lookahead.backtrack candidates.cd_lookahead;
  Selection_lookahead.backtrack candidates.cd_lookahead_exceeding;
*)
  candidates.cd_exceeding <-
  List.find_all
    (fun (raw_context_unifier, _, _) ->
       not (Context_unifier.is_raw_context_unifier_invalid raw_context_unifier)
    )
    candidates.cd_exceeding;


  backtrack_unprocessed candidates;

  backtrack_invalid candidates;

  backtrack_added candidates retracted explanation;

  match candidates.cd_last_retracted with
    | [] ->
	failwith "backtrack retract last"

    | _ ->
	()


let size candidates =
  Valid.size candidates.cd_valid
