(* $Id: selection_split.mli,v 1.12 2004/12/28 14:06:51 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** selection of split candidates

    stores all split candidates (remainders),
    and selects one for application.
*)



(** {6 Types} *)

type config = Config.config
type choice_point = State.choice_point
type bound = Bound.bound
type statistic = Statistic.statistic
type literal = Term.literal
type state = State.state
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected = Selection_types.selected
type context = Context.context

type candidates



(** {6 Functions} *)

(** creates an empty candidate repository. *)
val create: config -> bound -> statistic -> state -> context -> candidates

(** adds a split candidate to the repository. *)
val add: candidates -> raw_context_unifier -> unit

(** selects (priority based) a split literal from the repository.
    Ignores previously selected candidates,
    and irrelevant candidates,
    i.e candidates subsumed by, contradictory with, or not productive
    with the context.

    Returns None, if no further split candidate exists. *)
val select: candidates -> Random_jump.replay option -> selected option

(** registers the stored candidates exceeding the depth bound {!Bound}
    if they are not satisfied by the context.
    To be called based on {!Flags.opt_restart}, when a branch
    has to be checked for being exhausted.
*)
val check_exceeding: candidates -> unit

(** [backtrack candidates toggle_choice_point right_split_literal]
    removes all split candidates based on invalid/retracted choice points,
    readds previously added or invalid candidates to the candidate set,
    and transform the candidate associated with the
    left split [toggle_choice_point] from a left to a right split,
    thereby using [the right_split_literal].
*)
val backtrack: candidates -> choice_point -> choice_point list -> unit


(** textual representation of the remainder of the context unifier as a clause. *)
val remainder_literals_to_string: config -> context ->
  raw_context_unifier -> string

val size: candidates -> int

val select_right_split: candidates -> selected option
