(* $Id: symbol.ml,v 1.8 2004/12/28 14:06:52 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type counter = Counter.counter


(* a symbol is represented by its id *)
type symbol = {
  name: string;
  arity: int;
  id: int;
}


type symbol_table = {
  (* incrementing id counter for all symbols. *)
  id_counter: counter;

  (* mapping from a symbol's name and arity to the symbol. *)
  symbols: (string * int, symbol) Hashtbl.t;
}



(*** global symbol table *)

let global =
  {
    id_counter = Counter.create ();
    symbols = Hashtbl.create 128;
  }



(*** decomposition ***)


let name (symbol: symbol) : string =
  symbol.name

let arity (symbol: symbol) : int =
  symbol.arity

let id (symbol: symbol) : int =
  symbol.id


(*** comparison ***)

let equal (symbol1: symbol) (symbol2: symbol) : bool = 
  symbol1 == symbol2





(*** representation ***)

let to_string (symbol: symbol) : string =
(*  name_of_symbol symbol ^ "/" ^ string_of_int (arity_of_symbol symbol) *)
  name symbol






(*** creation ***)

let create_symbol (name: string) (arity: int) : symbol =
  try
    (* does the symbol already exist? *)
    let symbol =
      Hashtbl.find global.symbols (name, arity)
    in
      symbol
  with
    | Not_found ->
	(* no, so create and register it *)
	let symbol = {
	  name = name;
	  arity = arity;
	  id = Counter.next global.id_counter;
	}
	in
	  Hashtbl.add global.symbols (name, arity) symbol;
	  symbol



let create_skolem_name id =
  "__" ^ string_of_int id ^ "__"
  

(* create a skolem constant. not even stored anywhere *)
let create_skolem () : symbol =
  let id =
    Counter.next global.id_counter
  in
    {
      name = create_skolem_name id;
      arity = 0;
      id = id;
    }


let is_skolem (symbol: symbol) : bool =
  symbol.arity = 0
  &&
  symbol.name = (create_skolem_name symbol.id)
  




(*** constants ***)


(* create and register *)
let equality =
  create_symbol "=" 2
