(* $Id: symbol.mli,v 1.9 2004/12/28 14:06:52 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** constant and function symbols

  A symbol consists of
  - its name (eg {e p}, {e f} or {e a})
  - its arity (0 for a constant)
  - its unique id
    
  All symbols are irrevocably registered in a global symbol table.
  Different symbols with the same name but a different arity might exist.
*)


(** {6 Types} *)


(** a constant or function symbol *)
type symbol




(** {6 Functions} *)



(** {2 Special Symbols} *)

(** these special symbols are automatically added to the symbol table. *)

val equality: symbol


(** {2 Creation} *)


(** [Symbol.create_symbol name arity] registers a symbol and returns it. *)
val create_symbol: string -> int -> symbol


(** creates a unique fresh skolem constant. *)
val create_skolem: unit -> symbol





(** {2 Decomposition} *)

(** the symbol name. *)
val name: symbol -> string

(** the symbol arity. *)
val arity: symbol -> int

(** the unique symbol id. Useful for hashing on symbols or terms. *)
val id: symbol -> int



(** {2 Comparison} *)

(** the same symbol? *)
val equal: symbol -> symbol -> bool

(** checks if this symbol is a skolem constant. *)
val is_skolem: symbol -> bool


(** {2 String Representation} *)

(** currently just the name of the symbol. *)
val to_string: symbol -> string

