(* $Id: term_indexing.ml,v 1.8 2004/12/28 14:06:53 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type term = Term.term
type literal = Term.literal
type subst = Subst.subst

exception ITERATOR_EMPTY

class virtual ['data] data =
object
  method virtual is_equal: 'data -> 'data -> bool
  method virtual to_string: 'data -> string
end
  
class virtual ['data] iterator =
object
  method virtual is_empty: bool
  method virtual next: 'data
end

class type ['data] predicate_index =
object
  method add: ?unique:bool -> term -> 'data -> unit
  method remove: term -> 'data option -> bool 
  method clear: unit
  method size: int
    
  method iter: (term -> 'data -> unit) -> unit
  method fold : 'a. ('a -> term -> 'data -> 'a) -> 'a -> 'a
  method get_generalization_iterator: p_preserving:bool -> term -> 'data iterator
  method get_shielding_iterator: term -> term -> 'data iterator
    
  method find_all_variants: term -> 'data list
  method find_all_generalizations: p_preserving:bool -> term -> 'data list
  method find_all_unifiable: p_preserving:bool -> term -> 'data list
  method find_all_instances: p_preserving:bool -> term -> 'data list
  method find_all_shielding: term -> term -> 'data list

  method find: term -> 'data option    
  method find_generalization: p_preserving:bool -> term -> 'data option
  method find_unifiable: p_preserving:bool -> term -> 'data option
  method find_instance: p_preserving:bool -> term -> 'data option
  method find_strongly_shielding: term -> term -> 'data option
  method find_shielding: term -> term -> 'data option
    
  method to_string: string
end


module Table = Term.LiteralTypeTable

class ['data] index
  (__data: 'data data)
  (__create_index: 'data data -> 'data predicate_index)
   =
object (self)
  val _data = __data
  val _create_index = __create_index

  val _indexes: ('data predicate_index) Table.t = Table.create 16

  method private create_index (literal: literal) : 'data predicate_index =
    let new_index =
      __create_index _data
    in
      Table.add _indexes literal new_index;
      new_index
  

  method find (literal: literal) : 'data predicate_index =
    try
      Table.find _indexes literal
    with
      | Not_found ->
	  self#create_index literal

  method iter (func: 'data predicate_index -> unit) : unit =
    Table.iter
      (fun _ index ->
	 func index
      )
      _indexes

  method fold:
    'a. ('a -> literal -> 'data predicate_index -> 'a) -> 'a -> 'a =
    fun func acc ->
    Table.fold
      (fun literal index acc ->
	 func acc literal index
      )
      _indexes
      acc

end
