#include "demodulate.h"

#include "../ladr/ioutil.h"

/* Private definitions and types */

/* Maintain the indexes for forward and backward demodulation.
 */

static Mindex Demod_idx;
static Mindex Back_demod_idx;

/*************
 *
 *   init_demodulator_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_demodulator_index(Mindextype mtype, Uniftype utype, int fpa_depth)
{
  Demod_idx = mindex_init(mtype, utype, fpa_depth);
}  /* init_demodulator_index */

/*************
 *
 *   init_back_demod_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_back_demod_index(Mindextype mtype, Uniftype utype, int fpa_depth)
{
  Back_demod_idx = mindex_init(mtype, utype, fpa_depth);
}  /* init_back_demod_index */

/*************
 *
 *   index_demodulator()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void index_demodulator(Clause c, int type, Indexop operation, Clock clock)
{
  clock_start(clock);
  idx_demodulator(c, type, operation, Demod_idx);
  clock_stop(clock);
}  /* index_demodulator */

/*************
 *
 *   index_back_demod()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void index_back_demod(Clause c, Indexop operation, Clock clock, BOOL enabled)
{
  if (enabled) {
    clock_start(clock);
    index_clause_back_demod(c, Back_demod_idx, operation);
    clock_stop(clock);
  }
}  /* index_back_demod */

/*************
 *
 *   destroy_demodulation_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void destroy_demodulation_index(void)
{
  mindex_destroy(Demod_idx);
  Demod_idx = NULL;
}  /* destroy_demodulation_index */

/*************
 *
 *   destroy_back_demod_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void destroy_back_demod_index(void)
{
  mindex_destroy(Back_demod_idx);
  Back_demod_idx = NULL;
}  /* destroy_back_demod_index */

/*************
 *
 *   demodulate_clause()
 *
 *************/

/* DOCUMENTATION
Demodulate Clause c, using demodulators alreadly known to this package.
If any rewriting occurs, the justification is appended to
the clause's existing justification.
*/

/* PUBLIC */
void demodulate_clause(Clause c, int limit, BOOL print, BOOL lex_order_vars)
{
#if 1
  static BOOL limit_message = FALSE;
  int lim = limit == -1 ? INT_MAX : limit;
  fdemod_clause(c, Demod_idx, &lim, lex_order_vars);
  if (lim == 0 && print) {
    if (limit_message == FALSE) {
      limit_message = TRUE;
      fprintf(stderr, "Demod_limit (see stdout)\n");
      printf("\nDemod_limit: "); f_clause(c);
      printf("\nDemod_limit (%d, size=%d) termiated the demodulation of a clause.\n"
	     "The most recent kept clause is %d.\n"
	     "This message will not be repeated.\n\n",
	     limit, clause_symbol_count(c), clause_ids_assigned());
      fflush(stdout);
    }
    else {
      printf("Demod_limit\n");
      fflush(stdout);
    }
  }
#else
  demod_clause(c, Demod_idx);
#endif
}  /* demodulate_clause */

/*************
 *
 *   back_demodulatable()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist back_demodulatable(Clause demod, int type, BOOL lex_order_vars)
{
  return back_demod_indexed(demod, type, Back_demod_idx, lex_order_vars);
}  /* back_demodulatable */

/*************
 *
 *   back_demod_idx_report()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void back_demod_idx_report(void)
{
  printf("Back demod index: ");
  p_fpa_density(Back_demod_idx->fpa);
}  /* back_demod_idx_report */
