// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.util.TSTPCreator;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/07/15 11:35:21 $ $Revision: 1.28 $
*/
public class Spass extends ATP {
  
  public String SPASS_HOME = getSystemProperty("spass.home", "SPASS_HOME");
  public String[] spass_envs = new String[] {"SPASS_HOME="+SPASS_HOME};
  public String spassExecutable = findSystemExecutable(SPASS_HOME, "spass", "Spass.exe");
  private static String startProof = "Here is a proof";
  private static String endProof = "Formulae used";
  
  private static String sep = File.separator;
  private static String spassOptions = " -PGiven=0 -PProblem=0 -DocProof  -TimeLimit=";
    
  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Spass");
  private static String logFileName = getSystemProperty("spass.log.file", "SPASS_LOG_FILE"); 
  private static FileAppender appender = newAppender(logFileName);
  
  public Spass()
  {
    super();
    name = "Spass_2_1";
    acceptedConcept = "TstpProblem";
    resultConcept = "FoAtpResult";
    proofConcept = "TstpCnfRefutation";
    calculusConcept = "spass_res_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  public String[] callProver(String problem, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      inputFile = File.createTempFile("spass", ".in");
      outputFile = File.createTempFile("spass", ".out");

      /*      -PGiven=0 -PProblem=0 -DocProof -TimeLimit=%d %s */
      //      System.out.println(inputFile + " " + outputFile);
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        String newProb = tstp2X.tstp2X(problem, 
                                       "dfg",
                                       "rm_equality:rstfp");
        writeFile(inputFile, newProb);
        logger.debug("Spass format: " + newProb);
        //        System.out.println("path " + inputFile.getPath());
        
        String command = spassExecutable + " " +  spassOptions + time + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[SPASS] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.debug("[SPASS] starting prover...");
        this.start(command, spass_envs, defaultNice, time);
        //      this.exec(commands, spass_envs, false, true);
        logger.debug("reading result..." + outputFile.getAbsolutePath());
        String result = this.readFile(outputFile);
        if ((result != null) && (result.length() > 100000))
          return  new String[] {"Unknown", ""};
        
        logger.debug("[SPASS] transforming to TSTP...");
        //        String tstp = x2tstp.x2tstp(buf.toString(), "SPASS---2.1");
        String tstp = x2tstp.x2tstp(result, "SPASS---2.1");
        logger.info("[SPASS] TSTP..." + tstp);

//         File tstpFile = File.createTempFile("spass", ".tstp");      
//         writeFile(tstpFile, tstp);
        String status = extractStatusFromTSTP(tstp);
        return new String[] {status, tstp+"\n"};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Spass] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Spass] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Spass e = new Spass();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(20,20);
      {
        String servF1;
        //       String problem = e.readFile(servF1);
        //servF1 = msHome + "/examples/owl/fof.owl";
        servF1 = msHome + "/examples/owl/fof3.owl";
        //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
        // servF1 = msHome + "/src/b";
//        logger.info("result1: " + result);

        String ind = e.readFile(servF1);
        //        System.out.println("problem: " + ind);
        
        //       String language = "tptp";
        //       TimeResource tr = new TimeResource();
        //       tr.setCpuTime(300);
        //       tr.setWallClockTime(300);
        //      logger.info("prob: " + ind);
        //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/NLP/NLP043+1.p");
        
        //String result = e.callProver(problem, language, tr, "");
        String result = e.proveProblem(ind, timeInd);
        //String result = e.proveProblemXml(ind, timeInd);
          //          e.writeFile(msHome + "/examples/owl/spass-proof2.owl", result);
          //        e.writeFile(msHome + "/examples/owl/proof3.owl", result);
        
          //        logger.info("result1: " + result);
        
          //         result = e.proveProblem(ind);
        
        System.out.println("result2: " + result);
        
      
      }
  }
  
  

} // end Spass.java

//         String[] split = result.split("\n");
//         StringBuffer buf = new StringBuffer();
//         boolean started = false;
//         for(int i=0; i<split.length; i++){
//           logger.info("[line] " + split[i]);
//           if (!started) {
//             started = split[i].startsWith(startProof);
//             if (started) {
//               buf.append(split[i]);
//               buf.append("\n");
//             }
//           } else {
//             buf.append(split[i]);
//             buf.append("\n");
//             if (split[i].startsWith(endProof)) 
//               i = split.length;
//           }
//         }
//         logger.debug("[SPASS] result: "+ buf.toString());


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
