// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.atp;

import java.io.*;
import java.util.*;
import java.rmi.Remote;
import java.rmi.RemoteException;
import javax.xml.namespace.QName;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;

import org.mindswap.query.ValueMap;

import org.mathweb.owls.OWLSParser;
import org.mathweb.util.Java2WSDL;
import org.mathweb.onto.OWLJDom;
import org.mathweb.serv.ReasoningService;

public class SpassClient
{

  public String readFile(String fileName){
    try {
      FileInputStream isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
	sbRead.append((char) nChar);
	nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    }
  }

  public void waitMillis(long millis)
  {
    long t1 = System.currentTimeMillis();
    while ((System.currentTimeMillis() - t1) < millis);
  }
  
  public static void main(String [] args) {
    String msHome = System.getProperty("mathserv.home");
    OWLJDom owlJDom = OWLJDom.getInstance();
    String timeInd = owlJDom.createTimeIndi(8, 8);
    try {
      String hostname;
      if ((args.length > 0) && (!args[0].trim().equals(""))) 
	hostname = args[0];
      else 
	hostname = "localhost";
      
      OWLSParser owlsParser = new OWLSParser();
      
      SpassClient c = new SpassClient();
      //c.waitMillis(3000);
      
      String fofProb = c.readFile(msHome + "/examples/owl/fof2.owl");

      String wsdl = null;

      Vector ms = new Vector();
      ms.add("proveProblem");
      ms.add("proveProblemXml");
            
      wsdl = (new Java2WSDL()).java2WSDL(hostname, "SpassService", "org.mathweb.atp.Spass" , ms );
      System.out.println(" wsdl:  " + wsdl);

      ReasoningService service = owlsParser.parseServiceFile(msHome + "/etc/services/owls/1.1/SpassService.owl", wsdl);
      System.out.println("Service " + service);

      Hashtable map = new Hashtable();
      map.put("tstp_problem", fofProb);
      map.put("time_res", timeInd);
      ValueMap resTab = service.invokeOWLS(map);
      System.out.println("Sent problem and got '" + resTab + "'");
    } catch (Exception e) {
      e.printStackTrace();
      System.err.println(e.toString());
    }
  }
}
