// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

//  For testing:
//

package org.mathweb.atp;

import java.io.*;
import java.net.URI;
//import java.util.StringTokenizer;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Enumeration;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import dk.capgemini.tc.JProf;

import org.mathweb.util.TSTPCreator;

import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;

//import monet.openmath.lang.*;
//import monet.openmath.lang.*;

import org.mathweb.core.BasicService;

/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2005/08/09 16:54:38 $ $Revision: 1.15 $
*/
public class TPTPClassifier extends BasicService {
  
  public String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String CLASSIFIER_HOME = getSystemProperty("tptp.classifier.home", "TPTP_CLASSIFIER_HOME");
  public String[] classifier_envs = new String[] {"TPTP_HOME="+TPTP_HOME, "TPTP="+TPTP_HOME+sep+"TPTP"};
  public String classifierExecutable = findSystemExecutable(CLASSIFIER_HOME, "SPCForProblem", "SPCForProblem.exe");
  public String acceptedConcept = "TstpProblem";

  private OWLJDom owlJDom;

  private HashMap classifier_map = new HashMap();
  
  private static Logger logger = Logger.getLogger("Classifier");
  private static String logFileName = getSystemProperty("tptp.classifier.log.file", "TPTP_CLASSIFIER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
  
  public TPTPClassifier()
  { //THM_RFO_EQU_FOF
    //mkTempDir("tptp_classifier");
    //    System.out.println("Temporary dir is "+ tempDir);
    
    classifier_map.put("EPR","essentiallyProp");
    classifier_map.put("RFO","realFirstOrder");
    classifier_map.put("FOF","fofForm");
    classifier_map.put("CNF","cnfForm");
    classifier_map.put("HRN","Horn");
    classifier_map.put("NHN","notHorn");
    classifier_map.put("EQU","equality");
    classifier_map.put("NEQ","noEquality");
    classifier_map.put("SEQ","someEquality");
    classifier_map.put("PEQ","pureEquality");
    classifier_map.put("UEQ","unitPureEquality");
    classifier_map.put("NUE","nonUnitPureEquality");
    classifier_map.put("THM","theorem");
    classifier_map.put("SAT","satisfiable");
    classifier_map.put("NKS","notKnownSatisfiable");
    classifier_map.put("NKC","notKnownCounterSatisfiable");
    
    owlJDom = OWLJDom.getInstance();
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    Enumeration enum = logger.getAllAppenders();
//     for( ; enum.hasMoreElements() ;) {
//       System.out.println("appender: " +enum.nextElement());
//     }
    if (!enum.hasMoreElements()) logger.addAppender(appender);
//     FileAppender appender = null;
//     try {
//       appender = new FileAppender(new SimpleLayout(), logFileName, false);
//     } catch(Exception e) {}
//     logger.addAppender(appender);
  }
  

  
  public String analyseProblem(String provingProblem)
  {
    long startTime = System.currentTimeMillis();
    long cpuStart = JProf.getCurrentThreadCpuMillis();
    String probID = null;
    boolean valid = false; 
    //    logger.info("[Classify]: " + provingProblem);
    
    try{
      Element root = owlJDom.getRootElement(provingProblem);
      String xmlBase = root.getAttributeValue("base", Namespace.XML_NAMESPACE);

      logger.info("... parsing, took me " + (System.currentTimeMillis() - startTime) + " msecs");
      logger.debug("[Classify] : " + root + " " + owlJDom);
      //        List cs = root.getChildren();
      //        for(int i=0; i<cs.size();i++) logger.debug(cs.get(i));
      List children = root.getChildren();
      Element rightChild = null;
      for(int i=0; i<children.size(); i++){
          Element c = (Element) children.get(i);
          if ((c.getName()).indexOf("Problem") > -1) {
              String potentialID = owlJDom.valueOf(owlJDom.RDF_ID, c);
              if (potentialID != null) {
                  rightChild = c;
                  probID = xmlBase + potentialID;
                  break;
              }
          }
      }
      if (probID == null) 
        return owlJDom.createFailure(xmlBase, startTime, cpuStart, "Problem has to be of class " + acceptedConcept + "!");
      logger.info("[Classify] prob ID: " + probID);
 //      valid = reasoner.tellIndividual(provingProblem);
//       if (!valid)
//         return owlJDom.createFailure(probID, startTime, cpuStart, "Problem has to be valid RDF description of an OWL individual of class TstpProblem!");
      Element formal = (Element) OWLJDom.PROVING_PROBLEM_FORMAL_DESCRIPTIONS.selectSingleNode(rightChild);
      if (formal == null) 
          return owlJDom.createFailure(probID, startTime, cpuStart, "Problem has to be valid RDF description of an OWL individual of class TstpProblem!");
      String problem = formal.getTextTrim();
      
      URI probURI = URI.create(probID);
      //       String problem = reasoner.valueOf("formalDescription", probURI);
      //      logger.info("[Classify] prob: " + problem);
      String pClass = classifyProblem(problem);
      logger.info("[Classify] SPC: " + pClass);
      if (pClass == null){
        logger.info("[Classify] prob: " + provingProblem);
        return owlJDom.createFailure(probID, startTime, cpuStart, "Could not classify problem " + probID);
      }
      
      return createResult(probID, startTime, cpuStart, pClass);
    } catch(Exception e){
      return owlJDom.createFailure(probID, startTime, cpuStart, e.getMessage());
    }
  }
  
  public String createResult(String probID, long startTime, long cpuStart, String pClass) {
    try{
      Element rdf = new Element("RDF", OWLJDom.RDF_NS);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      rdf.setAttribute("base", OWLJDom.MW_NS.getURI(), Namespace.XML_NAMESPACE);
      Element classE = new Element("FOProblemClass",OWLJDom.MW_NS);
      classE.setAttribute("about", OWLJDom.MW_NS_STR+pClass, OWLJDom.RDF_NS);

      Element timeres = OWLJDom.createTimeResource(startTime, cpuStart);

      rdf.addContent(classE); 
      rdf.addContent(timeres); 
      return owlJDom.newXMLDocument(rdf);
    }
    catch(Exception e){
      return owlJDom.createFailure(probID, startTime, cpuStart, e.getMessage());
    }
  }
  
  public String classifyProblem(String problem){
    
    File inputFile = null;
    File outputFile = null;
    try {
      inputFile = File.createTempFile("classifier", ".in");
      outputFile = File.createTempFile("classifier", ".out");
      String inputFileName = inputFile.getAbsolutePath();
      String outputFileName = outputFile.getAbsolutePath();
      String command = classifierExecutable + " " + inputFileName + "> " +outputFileName;
      //      System.out.println("prob: " + command);
    
      // write the TPTP problem into the file
      this.writeFile(inputFile, problem);
      
      //        System.out.println("command: " + command);
      // call the SPCForProblem classifier 
      String[] commands = {"sh", "-c", command};
      this.exec(commands, classifier_envs, false, true);
      String result = readFile(outputFile);
      if (result == null) 
        return null;
      //      System.out.print(result + "\n");
      if (result.indexOf("ERROR") > -1)
        return null;
      else 
        return result.trim();
      //        System.out.println("[TPTPClassifier] problem classified as: " + result);
      //         StringTokenizer st = new StringTokenizer(result, "_");
      //         HashSet feature_set = new HashSet();
      //         String token;
      //         while (st.hasMoreTokens()){
      //           token = st.nextToken().trim();
      //           //TODO handle EQU properly
      //           if (!(token.equals("THM") || token.equals("SAT") || token.equals("EQU")))
      //             feature_set.add(classifier_map.get(token));
      //         }
      //         Iterator iter = feature_set.iterator();
      //         //        System.out.println("features: ");
      //         //         while (iter.hasNext()) {
      //         //           System.out.println(iter.next());
      //         //         }
      //         if (inputFile != null) inputFile.delete();
      //         if (outputFile != null) outputFile.delete();
      //         return feature_set;
      //       }
      //       return null;      
      
    } catch(Exception e){
      e.printStackTrace();
      return null;
    } finally{
      if (inputFile != null) inputFile.delete();
      if (outputFile != null) outputFile.delete();
    }
  }

  public String classifyFile(String inputFileName)
  {
    //    System.out.print(inputFileName + " " );
    String problem = this.readFile(inputFileName);
    return classifyProblem(problem);
  }


  public void classifyFileList(String listFileName, String prefix)
  {
    FileOutputStream outputFile = null; 
    
    String spcFileName = listFileName;

    String[] spcLines = readFile(spcFileName).split("\n");
    int problems = spcLines.length;
    StringBuffer output = new StringBuffer();
    
    for(int i = 0; i< problems; i++){
      String line = spcLines[i].trim();
      String fileName = prefix + "/" + line;
      String spc = null;
      try{
        spc = classifyFile(fileName);
      }catch(Exception e) {
        e.printStackTrace();
      } 
      output.append(line);
      output.append("  ");
      if (spc != null)
        output.append(spc);
      output.append("\n");
      if (spc.equals("CNF_NKS_EPR"))
        System.out.println(line);
    }
  }
  
  public void countSPCs(String listFileName)
  {
    String[] spcLines = readFile(listFileName).split("\n");
    int problems = spcLines.length;
    Hashtable spcs = new Hashtable();
    
    for(int i = 0; i< problems; i++){
      String line = spcLines[i].trim();
      String[] parts = line.split("   ");
      String spc = parts[1].trim();
      System.out.println(spc);
      if (spcs.containsKey(spc)){
        Integer pre = (Integer) spcs.get(spc);
        spcs.put(spc, new Integer(pre.intValue() + 1));
      } else {
        spcs.put(spc, new Integer(1));
      }
    }
    System.out.println(spcs);
  }

  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
    TPTPClassifier c = new TPTPClassifier();
    TSTPCreator creator = new TSTPCreator();
    if (argv.length>1) {
      String fileList = argv[0];
      String prefix = argv[1];
      c.classifyFileList(fileList, prefix);
    } 
//     if (argv.length>0) {
//       c.countSPCs(argv[0]);
//     } 
    //else {
//     String servF1;
//     servF1 = msHome + "/examples/owl/fof3.owl";
//     String ind = c.readFile(servF1);
//     //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/NLP/NLP061+1.p");
//     ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN880-1.p");
 
//     String res = c.analyseProblem(ind);
//     System.out.println(res);
    //    }
  }
  
 
} // end TPTPClassifier.java
    
      
