// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.util.TSTPCreator;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer Universitaet des Saarlandes
@version $Date: 2005/07/15 11:35:21 $ $Revision: 1.16 $
*/
public class Vampire extends ATP {
  
  public String VAMPIRE_HOME = getSystemProperty("vampire.home", "VAMPIRE_HOME");
  public String[] vampire_envs = new String[] {"VAMPIRE_HOME="+VAMPIRE_HOME};
  public String vampireExecutable = findSystemExecutable(VAMPIRE_HOME, "vampire", "Vampire.exe");
  
  private static String sep = File.separator;
  private static String vampireOptions = " -PGiven=0 -PProblem=0 -DocProof  -TimeLimit=";
    
  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Vampire");
  private static String logFileName = getSystemProperty("vampire.log.file", "VAMPIRE_LOG_FILE"); 
  private static FileAppender appender = newAppender(logFileName);
  
  public Vampire()
  {
    super();
    name = "Vampire_7_0";
    acceptedConcept = "TstpProblem";
    resultConcept = "FoAtpResult";
    proofConcept = "TstpCnfRefutation";
    calculusConcept = "vampire_res_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  public String[] callProver(String problem, String language, String format, int time, String options) throws Exception
  {
    File origFile = null;
    File inputFile = null;
    File outputFile = null;

    try{
      //      origFile = File.createTempFile("vampire", ".orig");
      inputFile = File.createTempFile("vampire", ".in");
      outputFile = File.createTempFile("vampire", ".out");

      //      writeFile(origFile, problem);

      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        String newProb = tstp2X.tstp2X(problem, 
                                       "oldtptp",
                                       "none");
        writeFile(inputFile, newProb);
        logger.debug("Vampire format: " + newProb);
        //        System.out.println("path " + inputFile.getPath());
        
        /*         vampire -t %d %s*/
        
        String command = vampireExecutable + " -t " + time + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[VAMPIRE] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.debug("[VAMPIRE] starting prover...");
        this.start(command, vampire_envs, defaultNice, time);
        //      this.exec(commands, vampire_envs, false, true);
        logger.debug("reading result..." + outputFile.getAbsolutePath());
        String result = this.readFile(outputFile);

        if ((result != null) && (result.length() > 100000))
          return  new String[] {"Unknown", ""};
        
        //        logger.debug("[VAMPIRE] result..." + result);
        logger.debug("[VAMPIRE] transforming to TSTP...");
        String tstp = x2tstp.x2tstp(result, "Vampire---7.0");
        logger.info("[VAMPIRE] TSTP..." + tstp);

//         File tstpFile = File.createTempFile("vampire", ".tstp");      
//         writeFile(tstpFile, tstp);
        String status = extractStatusFromTSTP(tstp);
        return new String[] {status, tstp+"\n"};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Vampire] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Vampire] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Vampire e = new Vampire();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(20,20);
      {
        String servF1;
        //       String problem = e.readFile(servF1);
        //servF1 = msHome + "/examples/owl/fof.owl";
        servF1 = msHome + "/examples/owl/fof3.owl";
        //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
        // servF1 = msHome + "/src/b";
//        logger.info("result1: " + result);

        String ind = e.readFile(servF1);
        //        System.out.println("problem: " + ind);
        
        //       String language = "tptp";
        //       TimeResource tr = new TimeResource();
        //       tr.setCpuTime(300);
        //       tr.setWallClockTime(300);
        //      logger.info("prob: " + ind);
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN880-1.p");
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN455+1.p");
        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG111+1.p");

        //        String result = e.callProver(problem, language, tr, "");
        String result = e.proveProblem(ind, timeInd);
        //String result = e.proveProblemXml(ind, timeInd);
        
        System.out.println("result2: " + result);
      }
  }
  

} // end Vampire.java

//         String[] split = result.split("\n");
//         StringBuffer buf = new StringBuffer();
//         boolean started = false;
//         for(int i=0; i<split.length; i++){
//           logger.info("[line] " + split[i]);
//           if (!started) {
//             started = split[i].startsWith(startProof);
//             if (started) {
//               buf.append(split[i]);
//               buf.append("\n");
//             }
//           } else {
//             buf.append(split[i]);
//             buf.append("\n");
//             if (split[i].startsWith(endProof)) 
//               i = split.length;
//           }
//         }
//         logger.debug("[VAMPIRE] result: "+ buf.toString());


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
