// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

/** For testing:
*/
package org.mathweb.broker;

import java.io.*;
import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;
import java.util.HashSet;
import java.util.Set;

import java.util.Enumeration;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;

import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.input.*;
import org.jdom.output.*;

import salvo.jesus.graph.*;

//import monet.openmath.lang.*;
//import monet.openmath.lang.*;


import org.mathweb.core.BasicService;
import org.mathweb.core.OMUtils;

import org.mathweb.onto.*;

// import org.mathweb.broker.ServiceRegistry;
// import org.mathweb.broker.ServicePlanner;
// import org.mathweb.broker.BrokerGUI;
// import org.mathweb.broker.GuiEvent;

import org.mathweb.atp.TPTPClassifier;

import org.mathweb.serv.ReasoningService;
import org.mathweb.serv.ReasoningProblem;
import org.mathweb.serv.Query;
import org.mathweb.serv.Input;
import org.mathweb.serv.PreCondition;

import org.mathweb.owls.OWLSParser;

/**
   @author Juergen Zimmer, Saarland University, Germany
   @version $Date: 2005/06/24 19:11:55 $ $Revision: 1.42 $
*/
public class Broker extends BasicService {

  private static Logger logger = Logger.getLogger("Broker");
  private String logFileName = getSystemProperty("broker.log.file", "BROKER_LOG_FILE");
  //  private String brokerEmptyImage = getSystemProperty("broker.empty.image", "BROKER_EMPTY_IMAGE");
  
  private ServiceRegistry serviceRegistry;
  //  private ServiceMatcher serviceMatcher = new ServiceMatcher();
  private ServicePlanner servicePlanner = new ServicePlanner();
  private Query currentQuery = null;

  private ServiceExecutor executor = new ServiceExecutor();
  private ArrayList executionGraphs = new ArrayList();
  private ArrayList executionImages = new ArrayList();
  private DirectedGraph currentExecutionGraph = null; 
  private int executionGraphIndex = -1;
  
  /**  protected Ontology brokerOntology = BrokerOntology.getInstance(); */
  protected OWLJDom owlJDom;

  private OWLSParser owlsParser = new OWLSParser();
  private TPTPClassifier tptpClassifier = new TPTPClassifier();
  private BrokerGUI gui = null;
  //  private OMUtils omUtils = new OMUtils();
 
  /** GuiEvent to start the agent activity **/
  final static int SEND_QUERY = 2;
  final static int PREV_PLAN = 3;
  final static int NEXT_PLAN = 4;
  final static int EXEC_PLAN = 5;

  //  protected Ontology brokerOntology = BrokerOntology.getInstance();
  
  public Broker(boolean withGui) {
    owlJDom = OWLJDom.getInstance();
    serviceRegistry = new ServiceRegistry();

    if (withGui){
      gui = new BrokerGUI(this, "A first Broker GUI");
      gui.setVisible(true);
      serviceRegistry.loadDefaultOwlsServices(owlsParser);
      showServices(serviceRegistry.allNames());
    }
    
    logger.setLevel(Level.OFF);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    logger.addAppender(appender);

  }

  public Broker()
  {
    this(false);
  }
  
  public String registerService(String description)  {
    try { 
      logger.info(" descr: "+description);
      ReasoningService service = owlsParser.parseService(description);
      logger.info(" parsed!!!");
      return this.registerService(service);
    } catch(Exception e){
      e.printStackTrace();
      long cpuTime = 0;
      return owlJDom.createFailure(null, System.currentTimeMillis(), cpuTime, "Broker could not interpret service!");
    }
  }

  public String deregisterService(String serviceName)  {
    try { 
      serviceRegistry.remove(serviceName);
      return owlJDom.createConfirmation("De-registration succeeded!");
    } catch(Exception e){
      e.printStackTrace();
      long cpuTime = 0;
      return owlJDom.createFailure(serviceName, System.currentTimeMillis(), cpuTime, "Broker could not interpret service!");
    }
  }

  protected String registerService(ReasoningService service)  {
    if (service.getName() == null){
      logger.info("The service has no name!!!");
    }
    if (serviceRegistry.put(service)){
      showServices(serviceRegistry.allNames());
      return owlJDom.createConfirmation("Registration succeeded!");
    } else {
      long cpuTime = 0;
      return owlJDom.createFailure(service.getName(), System.currentTimeMillis(), cpuTime, "Registration did NOT succeed!");
    }
  }
  
  
  public String answerQuery(String id, String description)  {
    try{    
      if (gui != null) gui.setQuery(description);
      Query query = owlsParser.parseQuery(description);
      logger.info("The query: " + query);
      return answerQuery(id, query);
    } catch(Exception e){
      e.printStackTrace();
      long cpuTime = 0;
      return owlJDom.createFailure(id, System.currentTimeMillis(), cpuTime, "Broker could not interpret query description!");
    }
  }

  protected String answerQuery(String id, Query query)  {
    try{    
      currentQuery = query;
      executionImages = new ArrayList();
      executionGraphIndex = -1;
      
      executionGraphs = servicePlanner.findProdigyPlans(query, serviceRegistry);
      
      //      executionGraphs = servicePlanner.findPopPlans(query, serviceRegistry);
      
      /* throw away subsumed plans: 
       ============================ */
      List subsumed = new ArrayList();
      for(int i=0; i< executionGraphs.size(); i++){
        DirectedGraph agraph = (DirectedGraph) executionGraphs.get(i);
        for(int j=0; j< executionGraphs.size(); j++){
          if (i != j){
            DirectedGraph bgraph = (DirectedGraph) executionGraphs.get(j);
            boolean subsumes = servicePlanner.subsumes(agraph, bgraph);
//             System.out.println("graph " + i + " subsumes graph " +j+ "?: " + 
//                                subsumes);
            if (subsumes)
              subsumed.add(bgraph);
          }}
      }
      executionGraphs.removeAll(subsumed);

      Vertex startNode = servicePlanner.getStartNode();
      
      if (startNode != null){
        ServiceMerger merger = new ServiceMerger();
        Set merged = merger.compose(new HashSet(executionGraphs), startNode);
        if (merged != null) executionGraphs.addAll(merged);
      }
      
      if (gui != null) {
        if (executionGraphs.size() >0) {
          currentExecutionGraph = (DirectedGraph) executionGraphs.get(0);
          executionGraphIndex = 0; 
          BufferedImage image = graph2Image(currentExecutionGraph, 0);
          executionImages.add(image);
          gui.showImage((BufferedImage) executionImages.get(executionGraphIndex));
        } else{
          long cpuTime = 0;
          return owlJDom.createFailure(id, System.currentTimeMillis(), cpuTime, "Broker could not find any execution plans!");
        }
        
        for(int l=1; l< executionGraphs.size(); l++){
          DirectedGraph graph = (DirectedGraph) executionGraphs.get(l);
          BufferedImage image = graph2Image(graph, l);
          executionImages.add(image);
        }
      }
      
      //return owlJDom.createFailure("Invokation of services has to be implemented!");
      logger.info("#plans = " +executionGraphs.size());
      
      return executionGraphs.size() + " execution plans found!";
      //       return executor.executeGraph(graph, query, serviceRegistry);
      //     } catch(java.net.ConnectException e){
      //       e.printStackTrace();
      //       return owlJDom.createFailure("Broker could not execute service:\n" + e);
    } catch(Exception e){
      e.printStackTrace();
      long cpuTime = 0;
      return owlJDom.createFailure(id, System.currentTimeMillis(), cpuTime, "Broker encountered a problem!\n" + e);
    }
  }
  
  
  private ReasoningProblem analyseProblem(ReasoningProblem problem){
    //     ArrayList inputs = problem.getInputs();
    //     Iterator it = inputs.iterator();
    //     Input in;
    //     ArrayList preConds = problem.getPreconds();
    
    //     while (it.hasNext()){
    //       in = (Input)it.next();
    //       logger.info("next input: " + in.toString());
      
    //       if (omUtils.equal(in.getSignature(), tptpOMStr)){
    //         OMString content = (OMString) in.getContent();
    //         HashSet features = tptpClassifier.classifyProblem(content.getString());
    //         Iterator featit = features.iterator();
    //         String precondStr = "";
    //         String inputName = in.getName();
    //         logger.info("problem features: " + features.toString());
    //         while (featit.hasNext()){
    //           precondStr = precondStr + featit.next() + "(" + inputName + ").\n";
    //         }
    //         OMString newOMStr = new OMString(precondStr);
    //         PreCondition newPrecond = new PreCondition(features.toString(),new OMString(precondStr));
    //         preConds.add(newPrecond);
    //         logger.info("new precond: " + newPrecond.toString());
    //       }}
    //     problem.setPreconds(preConds);
    return problem;
  }


  //   private String callService(ServiceEntry serviceEntry, ReasoningProblem problem)
  //   {
  //     try{
  //       ProvingProblem p = new ProvingProblem();
  //       TimeResource tr = new TimeResource();
  //       Logic l = new Logic();
  //       tr.setWallClockTime(60);
  //       p.setTimeResource(tr);
    
  //       String language = "tptp";
  //       Input probInput = problem.getInput("tptp");
  //       logger.info("probInput: " + probInput);
  //       logger.info("entry: " + serviceEntry.toString());
      
  //       OMString prob = (OMString) probInput.getContent();
    
  //       p.setLanguage(language);
  //       p.setFormalDescription(prob.getString());
  //       p.setNlDescription(problem.getName());
  //       p.setLogic(l);
    
  //       ACLMessage msg = new ACLMessage(ACLMessage.QUERY_IF);
  //       msg.addReceiver(serviceEntry.getAgent());
  //       msg.setLanguage(codec.getName());
  //       msg.setOntology(atpOntology.getName());
  //       getContentManager().fillContent(msg, p);
  //       send(msg);
  //       return "Query sent to service "+serviceEntry.getService().getName() + "...";
  //     } catch (CodecException e) {
  //       e.printStackTrace();
  //       return e.getMessage();
  //     }
  //     catch (OntologyException e) {
  //       e.printStackTrace();
  //       return e.getMessage();
  //     } catch (Exception e) {
  //       e.printStackTrace();
  //       return e.getMessage();
  //     }
  //   }
  
    
  public void onGuiEvent(GuiEvent ev) {
    String result;
    switch (ev.getType()) {
    case SEND_QUERY: 
      String queryOWLS = (String)ev.getParameter(0);
      Query query = owlsParser.parseQuery(queryOWLS);
      logger.info("query: " + query.toString());
      result = answerQuery("gui",query);
      logger.info("[Broker] result: "+ result);
      showResult(result);
      break;
    case PREV_PLAN: 
      previousGraph();
      break;
    case NEXT_PLAN: 
      nextGraph();
      break;
    case EXEC_PLAN: 
      showResult("Trying to execute current plan!");
      result = executeCurrentGraph();
      showResult(result);
      break;
    default:
      System.err.println("Received unexpexcted GuiEvent of type:"+ev.getType());
    }
  }
  
  public void showResult(String result)
  {
    if (gui != null) gui.setResult(result);
  }

  public void showServices(String serviceNames)
  {
    if (gui != null) gui.setServices(serviceNames);
  }
  
  public String executeCurrentGraph()
  {
    long cpuTime = 0;
    try{    
      if ((currentExecutionGraph != null) && (currentQuery != null))
        return executor.executeGraph(currentExecutionGraph, currentQuery, false);
      else 
        return owlJDom.createFailure(null, System.currentTimeMillis(), cpuTime, "No query or execution plan available!");
      //     } catch(java.net.ConnectException e){
      //       e.printStackTrace();
      //       return owlJDom.createFailure("Broker could not execute service:\n" + e);
    } catch(Exception e){
      e.printStackTrace();
      return owlJDom.createFailure(null, System.currentTimeMillis(), cpuTime, "Broker encountered a problem!\n" + e);
    }
  }
  
  public void nextGraph()
  {
    if ((!executionGraphs.isEmpty()) && 
        (executionGraphIndex >= 0) && (executionGraphIndex < executionGraphs.size() -1)){
      executionGraphIndex = executionGraphIndex + 1;
      currentExecutionGraph = (DirectedGraph) executionGraphs.get(executionGraphIndex);
      if (gui != null) gui.showImage((BufferedImage) executionImages.get(executionGraphIndex));
    }
  }
  public void previousGraph()
  {
    if ((!executionGraphs.isEmpty()) && (executionGraphIndex > 0)){
      executionGraphIndex = executionGraphIndex - 1;
      currentExecutionGraph = (DirectedGraph) executionGraphs.get(executionGraphIndex);
      if (gui != null) gui.showImage((BufferedImage) executionImages.get(executionGraphIndex));
    }
  }
  
  
  private BufferedImage graph2Image(DirectedGraph graph, int nr)
  {
    File dotFile = null;
    File jpgFile = null;
    try{
      String dot = servicePlanner.graph2Dot(graph);
      logger.info(dot);
      writeFile(msHome + "/etc/data/graph/graph" + nr + ".dot", dot);
      
      dotFile = File.createTempFile("broker", ".dot");
      jpgFile = File.createTempFile("broker", ".jpg");
      writeFile(dotFile, dot);
      String command = "/usr/bin/dot -v -Tjpg "+dotFile.getAbsolutePath() + 
        " -o" + jpgFile.getAbsolutePath();
      logger.info("[Broker] starting dot...\n" + command);
      String[] commands = {"sh", "-c", command};
      this.exec(commands, new String[0], false, true);
      if (jpgFile.exists()) {
        BufferedImage image = ImageIO.read(jpgFile);
        dotFile.delete();
        //        jpgFile.delete();
        return image;
      }
    } catch(Exception e){
      logger.debug("[Broker] Exception " + e);
    } finally{
      try{ 
        if (dotFile != null) dotFile.delete();
        //        if (jpgFile != null) jpgFile.delete();
      } catch(Exception e){ logger.debug("[Otter] Exception " + e); }
    }
    return null;
  }
  
  public void shutdown()
  {
    //    serviceRegistry.shutdownLocal();
    return;
  }
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
    Broker b;
    boolean gui = ((argv.length > 0) && (argv[0].equals("gui")));
    b = new Broker(gui);
    //OWLSParser c = new OWLSParser();
    
 //    String[] s  = new String[10];
//     System.out.println("service cnf");
//     s[0] = b.readFile(msHome + "/etc/services/tramp-cnf-service.xml");
//     System.out.println("service nd");
//     s[1] = b.readFile(msHome + "/etc/services/tramp-nd-service.xml");
//     System.out.println("service e");
//     s[2] = b.readFile(msHome + "/etc/services/e-service.xml");
//     System.out.println("service otter");
//     s[3] = b.readFile(msHome + "/etc/services/otterfier-service.xml");
//     s[4] = b.readFile(msHome + "/etc/services/otter-service.xml");
//     s[5] = b.readFile(msHome + "/etc/services/spass-service.xml");
//     s[6] = b.readFile(msHome + "/etc/services/apros-service.xml");
//     for (int i =0; i<7; i++){
//       //      System.out.println("service " + s[i]);
//       b.registerService(s[i]);
//       //      System.out.println("action: " + acts.ge.toLongString());
//     }

//     if (!gui){
//       System.out.println("query");
//       String query = b.readFile(msHome + "/examples/queries/query1.xml");
//       String answer = b.answerQuery("bar",query);
//       System.out.println("broker answer: " + answer);
//     }
    
    ReasoningService[] s  = new org.mathweb.serv.ReasoningService[10];
//     System.out.println("service cnf");
//     int counter =1;

    
//     s[0] = c.parseServiceFile(msHome + "/etc/services/owls/OtterService.owl");
//     s[1] = c.parseServiceFile(msHome + "/etc/services/owls/EService.owl");
//     s[2] = c.parseServiceFile(msHome + "/etc/services/owls/TrampCNFService.owl");
//     s[3] = c.parseServiceFile(msHome + "/etc/services/owls/TrampNDService.owl");
//     s[4] = c.parseServiceFile(msHome + "/etc/services/owls/OtterfierService.owl");
//    s[4] = c.parseServiceFile(msHome + "/etc/services/owls/1.1/FlotterCNFService.owl");
    
//     for (int i =0; i<5; i++){
//       b.registerService(s[i]);
//     }
    if (!gui){
      Query  query;
      String answer;
      
      // //     query = c.parseQueryFile(msHome + "/examples/queries/owls/result-query1.owl");
      // //     answer = b.answerQuery("foo", query);
      String q;
      //      q = b.readFile(msHome + "/examples/queries/owls/1.1/nd-query1.owl");
      q = b.readFile(msHome + "/examples/queries/owls/1.1/cnf-result-query1.owl");
      answer = b.answerQuery("foo", q);
    }
    
      //     String res = b.executeCurrentGraph();
    
  }
  
} // end Broker.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

