// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.broker;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2004/08/16 10:02:07 $ $Revision: 1.1 $
*/

public class GuiEvent 
{
    //#APIDOC_EXCLUDE_BEGIN
    protected Object source;
    protected int type;
    //#APIDOC_EXCLUDE_END

  private List parameters;

    /**
       Create a GUI event.
       @param eventSource The logical source of this event.
       @param eventType An integer value, identifying the kind of this
       event.
    */ 
  public GuiEvent(Object eventSource, int eventType)
  {
    source = eventSource;
    type = eventType;	
    parameters = new ArrayList();
  }

    /**
       Retrieve the kind of this GUI event.
       @return The kind of this event.
    */
    public int getType()
    {
	return(type);
    }

    /**
       Retrieve the logical source of this GUI event.
       @return The event source, or <code>null</code> if no source was
       set.
    */
    public Object getSource()
    {
	return(source);
    }

  /** Add a new parameter to this event.
  * @param param is the parameter 
  **/
  public void addParameter(Object param) {
    parameters.add(param);
  }     

  /**
   * Get the parameter in the given position.
   * @return the Object with the parameter.
   **/
  public Object getParameter(int number) {
    return parameters.get(number);
  }

  /**
   * Get an Iterator over all the parameters.
   * @return An iterator, scanning the event parameter list.
   **/
  public Iterator getAllParameter() {
    return parameters.iterator();
  }
}
