// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

//  For testing:

package org.mathweb.broker;

import java.io.*;
import java.util.*;

import salvo.jesus.graph.*;
import salvo.jesus.graph.visual.*;
import salvo.jesus.graph.visual.layout.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.onto.OntoReasoner;
import org.mathweb.onto.PellReasoner;
import org.mathweb.onto.OWLJDom;
import org.mathweb.serv.*;

import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.parser.*;
import org.mathweb.reasoning.planning.*;
import org.mathweb.reasoning.basicplanner.*;
import org.mathweb.reasoning.partialorderplanner.*;
//import org.mathweb.reasoning.regressionplanner.*;

import org.mathweb.graph.*;
import org.mathweb.broker.ServiceEntry;

import org.jdom.*;

// import java.awt.*;
// import java.awt.event.*;
// import java.applet.*;
// import javax.swing.*;
// import javax.swing.tree.*;

class ServicePlanner extends BasicService 
{
  private PartialOrderPlanner planner;
  private PartialOrderPlan plan;

  private Hashtable serviceHash;
  private Hashtable nodeHash;
  private Hashtable edgeHash;

  private OntoReasoner reasoner;
  private String withGUI = getSystemProperty("service.planner.gui.flag", "SERVICE_PLANNER_GUI_FLAG");
  private static Logger logger = Logger.getLogger("ServicePlanner");
  private String logFileName = getSystemProperty("planner.log.file", "PLANNER_LOG_FILE");
  private String prodigyExec = getSystemProperty("prodigy.exec", "PRODIGY_EXEC");
  private String prodigyDump = getSystemProperty("prodigy.dump", "PRODIGY_DUMP");
  private String prodigyTypes = getSystemProperty("prodigy.types", "PRODIGY_TYPES");
  private String prodigyTimeout = getSystemProperty("prodigy.timeout", "PRODIGY_TIMEOUT");
  private String ACL_HOME = getSystemProperty("acl.home", "ACL_HOME");
  private String[] acl_envs = new String[] {"ACL_HOME="+ACL_HOME};
  private String aclExecutable = findSystemExecutable(ACL_HOME, "lisp", "alisp");
  private String lispLogFileName = getSystemProperty("acl.log.file", "ACL_LOG_FILE");
  private static final int defaultNice = 15;

  private OWLJDom owlJDom = OWLJDom.getInstance();

  private class MyComp implements Comparator
  {
    public int compare(Object o1, Object o2)
    {
      if ((o1 instanceof GraphImpl) && (o2 instanceof GraphImpl)){
        Graph g1 = (GraphImpl) o1;
        Graph g2 = (GraphImpl) o2;
        int v1 = g1.getVerticesCount();
        int v2 = g2.getVerticesCount();
        //        System.out.println("comp: " + v1 + " " + v2);
        
        if (v1 < v2) return -1;
        else if (v1 > v2) return 1;
      }
      return 0;
    }
  }
  
  public ServicePlanner()
  {
    super();
    reasoner =  PellReasoner.getInstance();

    logger.setLevel(Level.DEBUG);
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    logger.addAppender(appender);
  }

  public Vertex getStartNode()
  {
    return (Vertex) nodeHash.get("Source");
  }
  
  public Vertex getEndNode()
  {
    return (Vertex) nodeHash.get("Sink");
  }
    

  public ArrayList findPopPlans(Query query, ServiceRegistry registry)
  {
    ArrayList graphs = new ArrayList();
    
    DirectedGraph graph = null;
    graph = findPlanPOP(query, registry);
    while(graph != null) {
      graphs.add(graph);
      graph = generateNextPOP();
      
    }
    return graphs;

  }
  
  public DirectedGraph findPlanPOP(Query query, ServiceRegistry registry)
  {
    //    reasoner =  PellReasoner.getNewReasoner();
    //    logger.debug("query to match: " + query.toString());
    
    List inputs = query.getProblem().getInputs();
    for(int i=0; i<inputs.size(); i++){
      Input in = (Input) inputs.get(i);
      reasoner.tellIndividual(in.getContents());
    }
    if (!reasoner.isValid()) return null;

    StringBuffer planningProb = new StringBuffer();
    ArrayList services = registry.getAllEntries(); // list of reasoning services
    //    logger.debug("all services: " + services.toString());
    
    Vector actions = new Vector(); // vector of planning operators for LogicSystem
    
    serviceHash = new Hashtable();
    for (int i =0; i< services.size(); i++){
      //logger.debug("service " + services.get(i).toString());
      ServiceEntry e = (ServiceEntry) services.get(i);
      ReasoningService s = (ReasoningService) e.getService();
      serviceHash.put(s.getName(), s);
      Action a = s.toPlanningOperator();
      if (i<10) 
        System.out.println("operator: " + a.toLongString());
      //        logger.debug("operator: " + a.toLongString());
      actions.add(a);
    }
    
//     actions.addAll(reasoner.rulesToActions());
    
    for (int i =0; i< actions.size(); i++){
      planningProb.append(((Action) actions.get(i)).toProbString());
    }
    
    LogicSystem sys = new LogicSystem(actions, reasoner);
    State initial = query.getInitialState();
    State goal = query.getGoalState();
    logger.debug("init state: " + initial.toString());
    logger.debug("goal state: " + goal.toString());
    planningProb.append("initial {\n");
    planningProb.append(initial.toString());
    planningProb.append("}\n");
    planningProb.append("goal {\n");
    planningProb.append(goal.toString());
    planningProb.append("}\n");
    writeFile(msHome +"/etc/data/broker/prob", planningProb.toString());
    
    planner = new PartialOrderPlanner(sys, initial, goal);
    planner.setMaxPlans(5);
    plan = null;
    if (withGUI.equals("yes")) {
      POPGUI popgui = new POPGUI(planner);
      plan = popgui.generatePlan();
    } else 
      plan = planner.generatePlan();
    
    logger.debug("plan " + plan);
    HashSet acts = plan.getActions();
    Iterator it = acts.iterator();
    while (it.hasNext()) logger.debug("act: " + it.next());
    
    if (!(plan == null)){
      DirectedGraph g = plan.toGraph(serviceHash);
      logger.debug("graph " + g);
      return g;
    }
    else
      return null;
    
    //     RegressionPlanner planner = new RegressionPlanner(sys);
    //     LinearPlan plan = planner.generatePlan(initial, goal);
    
    //     BasicPlanner planner = new BasicPlanner(sys);
    //     Plan plan = planner.generatePlan(initial, goal);
    
    //     logger.debug(" plan " + plan);
  }
  public DirectedGraph generateNextPOP()
  {
    plan = planner.generateNextPlan();
    logger.debug("plan " + plan);
    if (!(plan == null)){
      DirectedGraph g1 = plan.toGraph(serviceHash);
      logger.debug("graph " + g1);
      return g1;
    }else return null;
  }
  

  public ArrayList findProdigyPlans(Query query, ServiceRegistry registry)
  {
    //    reasoner =  PellReasoner.getNewReasoner();
    StringBuffer description = new StringBuffer();
    int counter = 1;
    serviceHash = new Hashtable();
    nodeHash = new Hashtable();
    edgeHash = new Hashtable();
    
    List queryInputs = query.getProblem().getInputs();
    List queryOutputs = query.getProblem().getOutputs();

    for(int i=0; i<queryInputs.size(); i++){
      Input in = (Input) queryInputs.get(i);
      reasoner.tellIndividual(in.getContents());
    }
    if (!reasoner.isValid()) return null;
   
    description.append("(create-problem-space 'mathserv :current t)\n\n");
//     description.append(reasoner.getProdigyClassTree());
    String typeDecls = readFile(prodigyTypes);
    description.append(typeDecls);
    description.append("\n");
    
    ArrayList services = registry.getAllEntries(); // list of reasoning services
    for (int i =0; i< services.size(); i++){
      ServiceEntry e = (ServiceEntry) services.get(i);
      ReasoningService s = (ReasoningService) e.getService();
      serviceHash.put(s.getURI().substring(7).toUpperCase(), s);
      System.out.println("s: " + s.getName().toUpperCase());
     
      description.append(s.toProdigyOperator(counter));
      counter += ReasoningService.maxObjs;
      description.append("\n");
    }
    description.append(query.toProdigyProblem());
    //  System.out.println("descr: " + description);

    File inputFile = null;
    File outputFile = null;
    
    try{
      inputFile = File.createTempFile("prodigy", ".in");
      outputFile = File.createTempFile("prodigy", ".out");
      writeFile(inputFile, description.toString());
      writeFile(msHome + "/etc/data/broker/test.lisp", description.toString());
      int timeout = Integer.parseInt(prodigyTimeout);
      
      String command = prodigyExec + " " + aclExecutable + " " + prodigyDump + 
        " " + inputFile.getAbsolutePath() + " " + timeout + " " + outputFile.getAbsolutePath()  +
        " >" + lispLogFileName + "-prodigy.log";
      System.out.println("command: " + command);
    
      String[] commands = {"sh", "-c", command};
      logger.debug("[ServicePlanner] starting prodigy...");
      this.start(command, acl_envs, defaultNice, timeout + 2);
      String plansXML = readFile(outputFile);
      System.out.println("plans: " + plansXML);

      writeFile(msHome + "/etc/data/broker/test.xml", plansXML);
      
      
      Element root = owlJDom.getRootElement(plansXML);
      //      System.out.println("root: " + root);
      List plans = root.getChildren("plan");
      //      System.out.println("plans: " + plans);
      ArrayList graphs = new ArrayList();
      
      VertexImpl startVertex = new VertexImpl("Source");
      VertexImpl endVertex = new VertexImpl("Sink");
      nodeHash.put("Source", startVertex);
      nodeHash.put("Sink", endVertex);
      
      for(int i=0; i<plans.size(); i++){
        //      for(int i=0; i<1; i++){
        Hashtable inputsMap = new Hashtable();
        Hashtable outputsMap = new Hashtable();
        Hashtable outputArgsMap = new Hashtable();
        
        DirectedGraph graph = new DirectedGraphImpl();
        graph.add(startVertex);
        graph.add(endVertex);
        
        for(int m=0; m<queryInputs.size(); m++){
          Input qin = (Input) queryInputs.get(m);
//           String qinName = qin.getName().toUpperCase();
//           String paramName = query.getName().toUpperCase() + "#" + qinName;
          String qinName = qin.getName();
          String paramName = query.getName().toUpperCase() + "#" + qinName.toUpperCase();
          outputsMap.put(paramName, startVertex);
          outputArgsMap.put(paramName, qinName);
        }
        
        //        System.out.println("args: " + outputArgsMap);
        Element plan = (Element) plans.get(i);
        //        System.out.println("planE: " + plan);
        List ops = plan.getChildren("operator");
        
        // for each instantiated operator in the plan look at its arguments (args)
        for(int k=0; k<ops.size(); k++){
          Element op = (Element) ops.get(k);
          String opName = op.getAttributeValue("name").trim();
          System.out.println("op: " + opName);
          ArrayList args = new ArrayList(op.getChildren("arg"));
          //          System.out.println("args: " + args);

          // get the object names of the arguments 
          for(int l=0; l<args.size(); l++){
            Element arg = (Element) args.get(l);
            //            System.out.println("arg: " + arg.getAttributeValue("name"));
            args.set(l, arg.getAttributeValue("name"));
          }
          
          // retrieve the reasoning service corresponding to the operator
          ReasoningService serv = (ReasoningService) serviceHash.get(opName);
          //          System.out.println("serv: " + serv.getName());
          String servName = serv.getName();
//           String[] split = servName.split("#");
//           if (split.length > 1){ servName = split[1];}
          VertexImpl v = null;
          
          if (nodeHash.containsKey(servName))
            v = (VertexImpl) nodeHash.get(servName);
          else {
            // create a new node for this service
            v = new VertexImpl(servName);
            v.setObject(serv);
            nodeHash.put(servName, v);
          }
          
          // TODO: Is this dangerous? If a service is called more than once we can't differentiate the occurences anymore!!!
          graph.add(v);
          
          // now do something with the args...
          List inputs = serv.getProblem().getInputs();
          for(int m=0; m<inputs.size(); m++){
            Input in = (Input) inputs.get(m);
            String iName = in.getName();
            String argObj = (String) args.get(m);
            //            inputsMap.put(servName + "#" + iName, argObj);
	    
            // lookup the source node and the output parameter name feeding
            // this input parameter via the common object.
            VertexImpl source = (VertexImpl) outputsMap.get(argObj);
            String oName = (String) outputArgsMap.get(argObj);

            //            System.out.println("source output param: " + oName);
            
            if (source != null){
//               Edge e = new DirectedEdgeImpl(source, v);
              Annotation ann = new Annotation(oName, iName);
              AnnotatedEdge e = new AnnotatedEdge(source, v, ann);          
              String label;
              e.setFollowVertexLabel(false);
              //               label = ann.toString();
              label = oName + "::" + iName;
              e.setLabel(label.toLowerCase());
              graph.addEdge(e);	      
            }
          }

          // put the output of this node in the outputsMap table.
          // with the object in the planning operator as a key and the node as a value
          int offset = inputs.size();
          List outputs = serv.getProblem().getOutputs();
          for(int m=0; m<outputs.size(); m++){
            Output out = (Output) outputs.get(m);
            String oName = out.getName();
            String outArg = (String) args.get(offset + m);
            outputsMap.put(outArg, v);
            outputArgsMap.put(outArg, oName);
          }
        }

        // last node: connect the last nodes in the graph with the outputs of the End node.
        for(int n=0; n<queryOutputs.size(); n++){
          Output out = (Output) queryOutputs.get(n);
          String oName = out.getName();
          String qoutName = query.getName().toUpperCase() + "#" + oName.toUpperCase();
          VertexImpl source = (VertexImpl) outputsMap.get(qoutName);
          String sourceArgName = (String) outputArgsMap.get(qoutName);
          if (source != null){
            //            Edge e = new DirectedEdgeImpl(source, endVertex);
            //            System.out.println("edge: " + sourceArgName + " " + oName);
            Annotation ann = new Annotation(sourceArgName, oName);
            AnnotatedEdge e = new AnnotatedEdge(source, endVertex, ann);
            e.setFollowVertexLabel(false);
            String label = ann.toString();
            ////e.setLabel(oName + "::" + out.getSignature());
            //   e.setLabel(sourceArgName + "::" + oName);
            e.setLabel(label);
            graph.addEdge(e);	      
          }
        }
        //    if (!reasoner.isValid()) return null;
        
        graphs.add(graph);
        //        
      }
      MyComp comparator = new MyComp();
      
      // sort the graphs according to size (number of nodes)
       Collections.sort(graphs, comparator);
       return graphs;
       
    } catch(Exception e){
      e.printStackTrace();
      logger.debug("[ServicePlanner] Exception " + e);
      return null;
    } finally{
      try{ 
        //        if (inputFile != null) inputFile.delete();
        //        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[ServicePlanner] Exception " + e); }
    }
  }
  /* prodigy-mw ${lisp} prodigy-i386_linux.dxl /home/jzimmer/import/mathserve/thirdParty/prodigy/working/domains/mathserv/domain5.lisp 3.0 /home/jzimmer/test/prodigy5.out
   */
  
  private VertexImpl findStartVertex(DirectedGraph graph)  {
    VertexImpl source = null;
    for(Iterator it = graph.getVerticesIterator(); it.hasNext();){
      VertexImpl v = (VertexImpl) it.next();
      if (v.getLabel().substring(0,5).equals("Source")) {
        source = v; 
      }}
    return source;
  }
  
  private int distance(VertexImpl from, VertexImpl to, DirectedGraph graph)  {
    return 1;
  }

  private VertexImpl containsModObj(VertexImpl v, List vcs)
  {
    Object obj = v.getObject();
    for(int i = 0; i< vcs.size(); i++){
      VertexImpl ver = (VertexImpl) vcs.get(i);
      Object o = ver.getObject();
      if (o == obj) return ver;
    }
    return null;
  }

  /**
   * alist - blist
   */
  private List listDifferenceMod(List alist, List blist){
    List result = new ArrayList();
    for(int i=0; i< alist.size(); i++){
      VertexImpl v = (VertexImpl) alist.get(i);
      if (containsModObj(v, blist) == null)
        result.add(v);
    }
    return result;
  }


  private List listDifference(List alist, List blist){
    List result = new ArrayList();
    for(int i=0; i< alist.size(); i++){
      VertexImpl v = (VertexImpl) alist.get(i);
      if (!blist.contains(v))
        result.add(v);
    }
    return result;
  }
  
    
//   public Strin plans2Golog(List graphs)
//   {
//     StringBuffer program = new StringBuffer();
//     for(int i=0; i< graphs.size(); i++){
//       DirectedGraph graph = (DirectedGraph) graphs.get(i);
      
//   }
  
  /**
   * Create an ND proof for a TSTP FOF problem given a FO-ATP result.
   * 
   * @param agraph A directed graph with a source vertex 'Start' and a sink 'Finish'
   * @param bgraph Another directed graph with a source vertex 'Start' and a sink 'Finish'
   * @return true, if agraph subsumes bgraph, i.e. bgraph looks exactly like agraph except
   that it has one or more verteces in between two adjacent verteces of agraph. <br>
   Example: agraph contains A->C and bgraph contains A->B->D->C instead. <br>
   B and D are called redundant steps.
   NOTE: THIS ALGO IS FAST BUT _WRONG_!!! 
  */
  public boolean subsumes(DirectedGraph agraph, DirectedGraph bgraph)
  {
    //    System.out.println("subsumes: " + agraph + ":" + bgraph);
    ArrayList avcs = new ArrayList(agraph.getVertexSet());
    ArrayList bvcs = new ArrayList(bgraph.getVertexSet());
    //    List diff = listDifferenceMod(bvcs, avcs);
    List diff = listDifference(bvcs, avcs);
    List reversediff = listDifference(avcs, bvcs);
    if (diff.isEmpty()) return false;
    if (diff.size() > 1) return false;
    if (!reversediff.isEmpty()) return false;
    
    List ins = null;
    List outs = null;
    List seen = new ArrayList();
    
    for(int i = 0; i<diff.size(); i++){
      VertexImpl v = (VertexImpl) diff.get(i);
      //      System.out.println("extra node: " + v);
      if (!seen.contains(v)) {
        ins = bgraph.getIncomingAdjacentVertices(v);
        outs = bgraph.getOutgoingAdjacentVertices(v);
        
        boolean apath = false; 
        boolean end = false;
        
        do{
          if ((ins.size() == 1 && (outs.size() == 1))){
            VertexImpl bin = (VertexImpl) ins.get(0);
            VertexImpl bout = (VertexImpl) outs.get(0);
            VertexImpl ain;            
            VertexImpl aout;
            
//             VertexImpl ain = containsModObj(bin, avcs);
            if (avcs.contains(bin)) ain = bin;
            else ain = null;
//             VertexImpl aout = containsModObj(bout, avcs);
            if (avcs.contains(bout)) aout = bout;
            else aout = null;
            
            //            System.out.println(" nodes: " + bin + " " + bout + " " +ain + " " +aout);
            if(ain != null) 
              if (aout != null){
                // found corresponding vertices in agraph
                end = true;
                // now check whether they are directly adjacent in agraph
                Edge e = agraph.getEdge(ain, aout);
                if (e != null)
                  apath = true; 
                else 
                apath = false;
              } else {
                seen.add(bout);
                outs = bgraph.getOutgoingAdjacentVertices(bout);
              }
            else {
              seen.add(bin);
              ins = bgraph.getIncomingAdjacentVertices(bin);
            }
          } else {end = true; apath = false; }
        } while (!end);
        if (!apath) 
          return false;
      }
    }
    return true;
  }
  
    

  
  /**
     Merging Plans:
     ==============
start: the full set of all plans. 
1) build subsets of plans for all types of root nodes (services) reachable directly from Start node
 (take node with distance 1, where distance max lenghts of paths to that node)
2) in each subset with cardinality > 1, try do merge(subset) plans as far as possible

merge(subset):
--------------
1. if all nodes of all plans are the same, simply collapse them to one node in the new graph.
2. go one level deeper (look into graph algo's that start in source node and explore graph in levels 
   a level $n$  is the set of all nodes of certain distance $n$ from current node.
3. if all nodes (services) on certain level are of the same classification (eg. FO-ATP), then  merge them together to a Choice node. 
4 the paths in the choice node are just the remainding paths of the single graphs.

Problems:
---------
- how do we map the I/O parameters properly?
- for the remainding paths in the graph we also have to get the dataflow right...


    Execution of merged Plans:
    ==========================
- At Choice nodes, the broker has to use analysis services (or just built-in) and the specialisations of the services to put a rating on the nodes. Then it chooses the service with the highest rating 
- For this, ratings have to be numerical values of expected success on a problem class.
- for the analysis we need to see which parameters correspond (looking at
  the types (OWL classes) of the parameters).

Actually!:
----------
- We only need to see whether all edges have the same type, and they must,
  because the preceeding service is all the same for all plans, so we just
  find the input of a certain type and look for the specialisation value on
  this input in the Profile.

  */    

  public static String graph2Dot(DirectedGraph g) 
  {
    StringBuffer dot = new StringBuffer();
    dot.append("digraph g {\n");
    //fontnames == Times Helvetica Courier Symbol Times-Italic Times-Bold   don't work!
    String fontname = "courier"; 
    VertexImpl v1, v2;
    try{
      Iterator it = g.getVerticesIterator();
      while(it.hasNext()){
        Object obj = it.next();
        System.out.println("obj: " + obj.getClass().getName());
        VertexImpl v = (VertexImpl) obj;        
        String nodeLabel = v.getLabel();
        //         if (!(nodeLabel.equals("Sink") || nodeLabel.equals("Source"))){
        //           ReasoningService serv = (ReasoningService) v.getObject();
        //            nodeLabel = serv.getName();
          
        
        //         String label = v.getLabel();
        //         String[] parts = label.split("/");
        //         int last = parts.length -1;
        //        String nodeLabel = v.getLabel();
        
        //         if (last > 0)
          //           nodeLabel = parts[last];
        System.out.println("node: " + nodeLabel);
        
        dot.append("node" + nodeLabel + "[shape=ellipse label=\"" + nodeLabel + "\" fontname=\""+fontname+"\" fontsize=8];\n");
      }
      Set edges = g.getEdgeSet();
      it = edges.iterator();
      while(it.hasNext()){
        Object obj = it.next();
        //        System.out.println(obj + " type: " + obj.getClass().getName());
        if (obj instanceof AnnotatedEdge){
          AnnotatedEdge e = (AnnotatedEdge) obj;
          
          //          String label = e.getLabel();
          Object obj1 = e.getVertexA();          
          Object obj2 = e.getVertexB();          
          System.out.println("obj1: " + obj1.getClass().getName());
          System.out.println("obj2: " + obj2.getClass().getName());
          v1 = (VertexImpl) obj1 ;
          v2 = (VertexImpl) obj2;
          
          Annotation ann = e.getAnnotation();
          String edgeLabel = "(" + ann.getSource() + ", " +  ann.getTarget() + ")";
          
          //         String[] split = edgeLabel.split("::");
          //         if (split.length > 1){ edgeLabel = split[1];}
          
          //        dot.append("\"node" +  v1.getLabel() + "\" -> \"node" + v2.getLabel() + "\" [label=\""+e.getLabel()+"\" fontname=\"courier\" fontsize=8];\n");
//           String[] parts = v1.getLabel().split("/");
//           int last = parts.length -1;
          String nodeLabel1 = v1.getLabel();
//           if (last > 0)
//             nodeLabel1 = parts[last];
//           parts = v2.getLabel().split("/");
//           last = parts.length -1;
          String nodeLabel2 = v2.getLabel();
//           if (last > 0)
//             nodeLabel2 = parts[last];
          
          System.out.println("edge: " + nodeLabel1 + " --> " + nodeLabel2);
          
          dot.append("\"node" +  nodeLabel1 + "\" -> \"node" + nodeLabel2 + "\" [label=\""+edgeLabel+"\" fontname=\"courier\" fontsize=8];\n");
          
        }
      }
      
      dot.append("}\n");
      logger.debug("dot graph " + dot);
    } catch(Exception e){
      e.printStackTrace();
      logger.debug(e);
    }
    return dot.toString();
    
  }
        /**
graph2Dot() creates Strings like the following which can be used 
as input for the graphviz dot tool:

digraph g {
node [shape = record,height=.1];
node0[label = "<f0> |<f1> G|<f2> "];
node1[label = "<f0> |<f1> E|<f2> "];
node2[label = "<f0> |<f1> B|<f2> "];

"node0":f2 -> "node4":f1;
"node0":f0 -> "node1":f1;
"node1":f0 -> "node2":f1;
"node1":f2 -> "node3":f1;
"node2":f2 -> "node8":f1;
"node2":f0 -> "node7":f1;
"node4":f2 -> "node6":f1;
"node4":f0 -> "node5":f1;
}
      */
  

  public static void main(String argv[]) throws Exception {
    System.out.println("null == null" + (null == null));
    
  }
    

}

  
  

    
    
    //     out1 = graph1.getOutgoingAdjacentVertices(source1);
//     out2 = graph1.getOutgoingAdjacentVertices(source2);
    
//     try{
//       while (!end){
//         if (out1.size() != out2.size())
//           return false;
//         List objs1 = new ArrayList();
//         List newOut1 = new ArrayList();
//         List newOut2 = new ArrayList();
        
//         for(int i = 0; i<out1.size(); i++){
//           VertexImpl v = (VertexImpl) out1.get(i);
//           newOut1.addAll(graph1.getOutgoingAdjacentVertices(v));
//           objs1.add(v.getObject());
//         }
        
//         for(int j = 0; j<out2.size(); j++){
//           VertexImpl v = (VertexImpl) out2.get(j);
//           Object obj = v.getObject();
//           // if one of the vertices is not in graph1 
//           if (!objs1.contains(obj)){
//             // get the next nodes in this branch of graph2
//             List next  = graph2.getOutgoingAdjacentVertices(v);
//             if (next.size() != 1){
//               return false;
//             }
//             else {
//               v = (VertexImpl) next.get(0);
//               obj = v.getObject();
//               if (!objs1.contains(obj)){
//                 newOut2.addAll(graph2.getOutgoingAdjacentVertices(v));
//                 subsumes = true;
//                 else
//                   return false;
//               }
//             }
//           } else {
//             newOut2.addAll(graph2.getOutgoingAdjacentVertices(v));
//           }
//         }
//         if ((newOut1.size() == 1) && (newOut2.size()==1))
    

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */


//         ReasoningService start = new ReasoningService();
//         start.setName("Source");
//         ReasoningProblem startProb = new ReasoningProblem();
//         startProb.outputs = queryInputs;
//         startProb.postconds = query.getProblem().getPreconds();
//         start.setProblem(startProb);

//         ReasoningService end = new ReasoningService();
//         start.setName("Finish");
//         ReasoningProblem endProb = new ReasoningProblem();
//         endProb.inputs = queryOutputs;
//         endProb.preconds = query.getProblem().getPostconds();
//         end.setProblem(endProb);
        
// hack to avoid Otterfier loops
//     description.append("(Control-Rule reject-wrong-binding-for-www.mathweb.org/owl/trans/OtterfierService.owl2");
//     description.append("  (if (and");
//     description.append("        (applicable-operator (www.mathweb.org/owl/trans/OtterfierService.owl <old_result>  <new_result>))");
//     description.append("        (or");
//     description.append("         (exact-type-of-object <old_result> OWL-FOBrFPResult))");
//     description.append("         ))");
//     description.append("  (then sub-goal))");
