// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

//  For testing:

package org.mathweb.broker;

import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.mathweb.owls.OWLSParser;
import org.mathweb.serv.ReasoningService;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;
import org.mathweb.util.Java2WSDL;
import org.mathweb.broker.ServiceEntry;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2005/06/21 09:20:05 $ $Revision: 1.21 $
*/
public class ServiceRegistry extends BasicService {

  private Hashtable table;
  private ArrayList serviceList;
  private ArrayList serviceNames;
  
  private String serviceConfigFile = getSystemProperty("mathserv.services.config", "MATHSERV_SERVICES_CONFIG");
  private String serviceDir = getSystemProperty("mathserv.services.dir", "MATHSERV_SERVICES_DIR");

  private static OWLJDom owlJDom = OWLJDom.getInstance();
  private Logger logger = Logger.getLogger("ServiceRegistry");
  
  public static final Namespace WSDD_NS = Namespace.getNamespace("wsdd","http://xml.apache.org/axis/wsdd/");
  public static final XPath servicex = newXPath("./wsdd:service");
  public static final XPath classNamex = newXPath("./wsdd:parameter[@name='className']/@value");
  public static final XPath methodsx = newXPath("./wsdd:parameter[@name='allowedMethods']/@value");
  public static final XPath msdlx = newXPath("./wsdd:parameter[@name='msdlDescription']/@value");
  public static final XPath owlsx = newXPath("./wsdd:parameter[@name='owlsDescription']/@value");

  //***********************************************
  public ServiceRegistry()
  {
    table = new Hashtable();
    serviceList = new ArrayList();
    serviceNames = new ArrayList();
    
    logger.setLevel(Level.INFO);
  }
  
  public synchronized boolean remove(String aName)
  {
    boolean existing = serviceNames.contains(aName);
    if (existing) {
      serviceNames.remove(aName);
      
    }
    return true;
  }
  

  public boolean put(ReasoningService service){
    String aName = service.getURI();
    //    System.out.println("serv Name: " + aName);
    return put (aName, service);
  }
  
  public synchronized boolean put(String aName, ReasoningService service){
    boolean existing = serviceNames.contains(aName);
    if (existing) {
      aName = generateNewName(aName);
    }
    ServiceEntry newServiceEntry = new ServiceEntry(aName, service);
    serviceList.add(newServiceEntry);
    serviceNames.add(service.getName());
    return true;
  }
  
  public ReasoningService get(String serviceName){
    for(int i=0; i<serviceList.size(); i++){
      ServiceEntry entry = (ServiceEntry) serviceList.get(i);
      if(entry.getID().equals(serviceName))
        return entry.getService();
    }
    return null;
  }
  
  public ArrayList getAllEntries()
  {
    return serviceList;
  }
  
  private String generateNewName(String oldname)
  {
    return new String(oldname + System.currentTimeMillis());
  }
  
  public String allNames()
  {
    StringBuffer result = new StringBuffer();
    int size = serviceNames.size();
    for (int i= 0; i< size; i++) {
      result.append((String) serviceNames.get(i));
      if (i< size-1) 
        result.append(", ");
    }
    return result.toString();
  } 
 
//   public void remove(String aName)   {
//     synchronized(this) {
//       table.remove(aName);
//     }
//   }

  public void loadDefaultOwlsServices(OWLSParser owlsParser)
  {
    Java2WSDL java2WSDL = new Java2WSDL();
    
    String servXML = null;
    Element root = null;
    List servs = null;
    servXML = readFile(serviceConfigFile);
    if (servXML != null) 
      try{
        root = owlJDom.getRootElement(servXML);    
        servs = servicex.selectNodes(root);
      } catch(Exception e) {
        e.printStackTrace();
        return;
      }
    if (root == null) return;
    
    for(int i=0 ; i<servs.size(); i++){ 
      try { 
        Element s = (Element) servs.get(i);
        String servName = s.getAttributeValue("name");
        String className = classNamex.valueOf(s);
        String methods = methodsx.valueOf(s);
        List fileNameList  = owlsx.selectNodes(s);
        //        String fileName = owlsx.valueOf(s);
//         System.out.println(" fileName:  " +serviceDir + this.sep + "1.1" + this.sep + fileName );
//         System.out.println(" methods:  " + methods);
//         System.out.println(" owls files:  " + list);
        for(int fileI=0; fileI<fileNameList.size(); fileI++){
          Attribute nameAttr = (Attribute) fileNameList.get(fileI);
          String fileName = nameAttr.getValue();
          if ((fileName != null) && (!fileName.equals(""))){
            String description = readFile(serviceDir + this.sep + "1.1" + this.sep + fileName);
            Vector ms = new Vector();
            String[] mNames = methods.split(",");
            for(int k = 0; k< mNames.length; k++)
              ms.add(mNames[k]);
            //System.out.println(" methods:  " + ms);
            String wsdl = null;
            
            if ((className != null) && (!className.trim().equals("")))
              wsdl = java2WSDL.java2WSDL(servName, className, ms);
            //System.out.println(" wsdl:  " + wsdl);
            
            ReasoningService service = owlsParser.parseService(description, wsdl);
            //ReasoningService service = owlsParser.parseService(description, null);
            //          logger.info(" parsed!!!");
            if (service != null) put(service);
          }
        }
      } catch(Exception e){
        e.printStackTrace();
      }
    }
  }
  
  public void shutdownLocal()
  {
    ArrayList services = getAllEntries(); // list of reasoning services
    
    for (int i =0; i< services.size(); i++){
      logger.debug("shutting down service " + services.get(i).toString());
      ServiceEntry e = (ServiceEntry) services.get(i);
      ReasoningService s = (ReasoningService) e.getService();
      s.shutdown();
    }
  }
  
  
  public static XPath newXPath(String x)
  {
    try{
      XPath xpath =  XPath.newInstance(x);
      xpath.addNamespace(WSDD_NS);
      return xpath;
    } catch(Exception e){
      return null;
    }
  }

  public String toString()  {
    StringBuffer result = new StringBuffer();
//     for (Enumeration e = table.elements(); e.hasMoreElements() ;) {
//       ArrayList alist = (ArrayList) e.nextElement();
//       for (int i= 0; i< alist.size(); i++) {
//         ServiceEntry entry = (ServiceEntry) alist.get(i);
//         result = result + "\n agent " + entry.getID() + ":\n  ";
//         Object obj = entry.getService();
//         if (obj != null) { result = result + obj.toString();
//         }
//       }
//     }
    return result.toString();
  }
  
} // end ServiceRegistry.java
    
      
