/*
 * tuProlog - Copyright (C) 2001-2002  aliCE team at deis.unibo.it
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.mathweb.broker;

import java.io.*;
import java.util.ArrayList;

import alice.util.Tools;
import alice.tuprolog.*;
import alice.tuprolog.lib.JavaLibrary;

import org.mathweb.core.*;


// run org.mathweb.broker.TuProlog /home/jzimmer/golog/IndiGolog/elevatorWithGUI/swi/interpreterN.swi "indiprolog(control1)."


/**
 * Provides a prolog virtual machine embedded in a separate thread.
 * It needs a theory and optionally a goal.
 * It parses the theory, solves the goal and stops.
 *
 * @see alice.tuprolog.Prolog
 *
 *
 *
 */
public class TuProlog extends BasicService {
  
  private Prolog      core;
  private String      theoryText;
  private InputStream theoryInputStream;
  private String      goalText;
    private static boolean spy = false;
  //  private static boolean spy = true;
  private int counter = 0;
    
  private OutputListener defaultOutputListener = new OutputListener() {
      public void onOutput(OutputEvent ev) {
	System.out.println(ev.getMsg());
      }};
  
  private WarningListener defaultWarningListener = new WarningListener() {
      public void onWarning(WarningEvent ev) {
	System.out.println(ev.getMsg());
      }};

  private SpyListener defaultSpyListener = new SpyListener() {
      public void onSpy(SpyEvent ev) {
	System.out.println(ev.getMsg());
      }};
  
  
  /**
   * Builds a prolog agent providing it a theory
   *
   * @param theory the text representing the theory
   */
  public TuProlog(){
    System.out.println("foo");
    
    theoryText=null;
    goalText = null;
    
    core=new Prolog();
    core.setWarning(true);
    core.setSpy(spy);
    try{
      Library lib = core.loadLibrary("alice.tuprolog.lib.JavaLibrary");
      ((alice.tuprolog.lib.JavaLibrary)lib).register(new Struct("engine"), this);
    } catch(Exception e){}
    
    core.addOutputListener(defaultOutputListener);
    core.addWarningListener(defaultWarningListener);
    core.addSpyListener(defaultSpyListener);
  }

  public String holds(ArrayList list)
  {
    String[] resultTerm = new String[list.size()];
    System.out.println("[java] holds : " + list);
    
    for(int k=0; k< list.size(); k++){
      resultTerm[k] = list.get(k).toString();
    }
//     if (resultTerm[0].equals("status"))
//       return "false";
//     else 
      return "true";
  }
  

  public String[] execute(ArrayList list)
  {
    String[] resultTerm = new String[list.size()];
        
    if (list.size() > 0) {
      String serviceName = (String) list.get(0);
      if (serviceName.equals("stop"))
	System.exit(0);
      ArrayList result = new ArrayList();
      result.add(serviceName);
      
      System.out.println("[java]: executing " + list);
      for(int i= 1 ; i< list.size(); i++){
	Object a = list.get(i);
	if (a != null){
	  result.add(a);
	} 
	else {
	  counter++;
	  result.add(new Struct("http://www.foo.org/service#result"+counter));
	}
      }
      //      Term[] termArray = new Term[result.size()];
      for(int k=0; k< result.size(); k++){
	Object res = result.get(k);
// 	Term resTerm = null;
// 	if (res instanceof String)
// 	  resTerm = new Struct((String) res);
// 	else if (res instanceof java.lang.Double) {
// 	  java.lang.Double d = (java.lang.Double) res;
// 	  resTerm = new alice.tuprolog.Double(d.doubleValue());
// 	}else 
// 	  resTerm = new Struct(res.toString());
	
	resultTerm[k] = res.toString();
	//	termArray[k] = resTerm;
      }
      //      resultTerm = new Struct(termArray);
    }
    System.out.print("[java] result: ");
    for(int l=0; l<resultTerm.length; l++){
      System.out.print(resultTerm[l]);
      System.out.print(" ");
    }
    System.out.println("");
    
    return resultTerm;
  }
  
  public static class TuPrologThread extends Thread {
    TuProlog agent;
    TuPrologThread(TuProlog agent){
      this.agent=agent;
    }
    final public void run(){
      agent.body();
    }
  }
  
  /**
   * Starts agent execution
   */
  final public void spawn(){
    new TuPrologThread(this).start();
  }
  
  public void body(){
    try {
      if (theoryText==null){
	core.setTheory(new Theory(theoryInputStream));
      }  else {
	System.out.println("setTheory");
	core.setTheory(new Theory(theoryText));
      }
      if (goalText!=null){
	SolveInfo info = core.solve(goalText);
	Term sol = info.getSolution();
	
	System.out.println("info: " + info.isSuccess());
	System.out.println("sol: " + sol);
      }
    } catch (Exception ex){
      System.err.println("invalid theory or goal." + ex.getMessage());
      ex.printStackTrace();
    }
  }

  public void test()
  {
    String msHome = System.getProperty("mathserv.home");
    System.out.println("bar");
    theoryText = readFile(msHome + "/thirdParty/golog/dtgolog/interpreter.tupl");
    goalText = "run(comp('http://www.foo.org/owl/queries/ResultQuery1.owl#fof_problem12345', N))."; 
    //    theoryText = readFile(msHome + "/thirdParty/golog/dtgolog/mathserv2.tupl");
    //    goalText = "bp(comp(fof1, N),5,Pol,U,Prob,outputFile).";
    //     System.out.println("golog prog: " + prog);

    
    counter = 0;
  }
  

  public static void main(String args[]){
    String theo = null;
    String goal = null;
    
    try{ 
      TuProlog engine = new TuProlog();
      engine.test();
      engine.spawn();
    } catch (Exception ex){
      System.err.println("invalid theory." + ex.getMessage());
      ex.printStackTrace();
    }
  }
}
