// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

/** For testing:

runt org.mathweb.eval.XmlRpcBroker atpChoice Test1 3

*/
package org.mathweb.broker;

import java.io.*;
import java.util.*;

import java.util.Enumeration;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.xmlrpc.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 
import org.apache.log4j.PatternLayout; 

import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDFS;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.graph.*;
import com.hp.hpl.jena.reasoner.*;
import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.reasoner.rulesys.*;
import com.hp.hpl.jena.util.*;
import com.hp.hpl.jena.vocabulary.ReasonerVocabulary;

import org.jdom.*;

import org.mathweb.onto.*;
import org.mathweb.serv.*;
import org.mathweb.broker.*;
import org.mathweb.util.*;
import org.mathweb.owls.*;
import org.mathweb.core.BasicService;


/**
   @author Juergen Zimmer, Saarland University, Germany
   @version $Date: 2005/07/04 16:10:59 $ $Revision: 1.15 $
*/
public class XmlRpcBroker extends BasicService {
  
  private static Logger logger = Logger.getLogger("XmlRpcBroker");
  private static String logFileName = getSystemProperty("xmlrpcbroker.log.file", "XMLRPCBROKER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);

  private Random rand = new Random();
  private List concepts;
  
  private static OWLSParser owlsParser = new OWLSParser();
  private ServiceRegistry serviceRegistry = new ServiceRegistry();
  private TSTPCreator creator = new TSTPCreator();

  protected static final String sep = File.separator;
  private String msHome = System.getProperty("mathserv.home");
  
  private String choiceProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpChoice");
  private String splitProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpSplit");
  private String splitJoinProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpSplitJoin");
  private String optProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpOpt-300");
  private String zenonProgram = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "zenonOpt");
  
  private WebServer xmlRpcServer;
  private int serverPort = 0;
  
  public XmlRpcBroker()  {
    super();
    
    serviceRegistry.loadDefaultOwlsServices(owlsParser);
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
    logger.info("new broker object!");
  }
  
  
  private void startXmlRpcServer(int givenPort) throws ClassNotFoundException {
    //    XmlRpc.setDriver(Class.forName(".SAXDriver"));
    boolean success = false;
    serverPort = givenPort;
    int defaultPort = givenPort;
    do{
      try{
        xmlRpcServer = new WebServer(serverPort);
        xmlRpcServer.start();
        System.out.println("Started Java XMLRPC server on port "+serverPort);
        xmlRpcServer.addHandler ("$default", this);
        success = true;
      }catch(IOException e){
        serverPort++;
      }
      if ((serverPort - defaultPort) > 40) {
        System.out.println("More than 40 ports over "+ defaultPort + " are bound. Giving up!");
        serverPort = 0;
        break;
      }
    } while (!success);
  }
  
  
  public String proveProblemXml(String tstpProblem, int time)
  {
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    //    System.out.println(owlsQuery);
    Query query1 = owlsParser.parseQuery(owlsQuery);
    String msHome = System.getProperty("mathserv.home");
    String program = readFile(msHome + sep + "thirdParty" + sep + "golog" + sep + "dtgolog" + sep + "atpOptXml");
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(program, query1, serviceRegistry, time, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemChoice(String tstpProblem, int time)
  {
      //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query2 = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(choiceProgram, query2, serviceRegistry, time, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemSplit(String tstpProblem, int time)
  {
    //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query3 = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(splitProgram, query3, serviceRegistry, time, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemSplitJoin(String tstpProblem, int time)
  {
    //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query3 = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(splitJoinProgram, query3, serviceRegistry, time, false);
    //    System.out.println("result: "+ result);
    return result;
  }

  public String proveProblemOpt(String tstpProblem, int time)
  {
    //    System.out.println(owlsQuery);
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    GologExecutor executor = new GologExecutor();
    Query query4 = owlsParser.parseQuery(owlsQuery);
    String result = executor.executeProgram(optProgram, query4, serviceRegistry, time, false);
    //    System.out.println("result: "+ result);
    return result;
  }


  public String proveProblemFocal(String tstpProblem, int time)
  {
    String owlsQuery = creator.tstp2OWLSQuery(tstpProblem);
    Query query5 = owlsParser.parseQuery(owlsQuery);
    GologExecutor executor = new GologExecutor();
    String result = executor.executeProgram(zenonProgram, query5, serviceRegistry, time, false);
    System.out.println("result: "+ result);
    return result;
  }

  /*  the name of a golog program  and of a TSTP SPC file and time in seconds */
  public static void main(String argv[]) throws Exception {
    int portNumber = 0;
    if (argv.length > 0) {
      portNumber = Integer.parseInt(argv[0]);
    }
    if (portNumber > 0) {
      XmlRpcBroker broker = new XmlRpcBroker();
      broker.startXmlRpcServer(portNumber);
    }
  }
  
} // end XmlRpcBroker.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

