 // -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.core;

import java.io.*;

public class CommandThread extends Thread
{
  private String command;
  private String[] envs;
  private Process myProcess;
  
  static boolean verbose=true;
  
  public CommandThread(String cmd, String[] es)
  {
    command = cmd;
    envs = es;
  }
  
  public void run()
  {
    String[] commands = {"/bin/sh", "-c", command}; 
    //System.out.println("executing threaded: " + commands);
    this.exec(commands, envs, false, false);
  }
    
  public boolean exec(String[] commands, String[] envs, boolean printResults, boolean wait) 
  {
    try {
      // Start running command, returning immediately.
      myProcess = java.lang.Runtime.getRuntime().exec(commands,envs);
      
      // Print the output. Since we read until
      // there is no more input, this causes us
      // to wait until the process is completed

      System.out.println("printing output?: " + printResults);
      if(printResults) {
	BufferedInputStream buffer = new BufferedInputStream(myProcess.getInputStream());
	BufferedReader  commandResult = new BufferedReader(new InputStreamReader(buffer));
	String s = null;
	try {
	  while ((s = commandResult.readLine()) != null)
	    System.out.println("Output: " + s);
	  commandResult.close();
	  if (myProcess.exitValue() != 0) {
	    if (verbose)
	      printError(commands + " -- myProcess.exitValue() != 0");
	    return(false);
	  }
	  // Ignore read errors; they mean process is done
	} catch (Exception e) {}
                
	// If you don't print the results, then you
	// need to call waitFor to stop until the process
	// is completed
      } else if (wait) {
	try {
	  System.out.println(" reading return value!");
	  int returnVal = myProcess.waitFor();
	  if (returnVal != 0) {
	    if (verbose)
	      printError(commands);
	    return(false);
	  }
	} catch (Exception e) {
	  if (verbose)
	    printError(commands, e);
	  return(false);
	}
      }
    } catch (Exception e) {
      if (verbose)
	printError(commands, e);
      return(false);
    } finally{
//       if (myProcess != null) {
//         try{ InputStream is = myProcess.getErrorStream();
//         System.out.println(" exec() error stream " + is);
//         is.close();
//         } catch (Exception e){      e.printStackTrace();
//         }
//         try{ InputStream is = myProcess.getInputStream();
//         System.out.println(" exec() input stream " + is);
//         is.close();
//         } catch (Exception e){      e.printStackTrace();
//         }
//         try{ OutputStream os = myProcess.getOutputStream();
//         System.out.println(" exec() output stream " + os);
//         os.close();
//         } catch (Exception e){      e.printStackTrace();
//         }
//         myProcess.destroy();
//       }
    }
    
    return(true);
    
  }

  protected static void printError(String command,
                                   Exception e) {
    System.out.println("Error doing exec(" +
                       command + "): " + e.getMessage());
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
  
  protected static void printError(String[] commands,
                                   Exception e) {
    String mess = "";
    for(int i=0; i< commands.length; i++) 
      mess = mess + " " + commands[i];
    System.out.println("Error doing exec(" +
                       mess + "): " + e.getMessage());
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
  
  protected static void printError(String[] commands) {
    String mess = "";
    for(int i=0; i< commands.length; i++) 
      mess = mess + " " + commands[i];
    System.out.println("Error doing exec(" +
                       mess + ").");
    System.out.println("Return value not 0 ");
  }

  protected static void printError(String command) {
    System.out.println("Error doing exec(" +
                       command + ").");
    System.out.println("Did you specify the full " +
                       "pathname?");
  }
    

  protected static void printError(String command, int value) {
    System.out.println("Error executing '" +
                       command + " " + value + "'.");
  }

  public void shutdown()
  {
    myProcess.destroy();
  }
  
}
