// package <your-package>

package org.mathweb.core;

import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Calendar;
import java.io.*;
import java.net.*;

import org.apache.xmlrpc.*;

/** 
	A facility class to centralize and call all methods offered by the mathweb object broker.
	*/

public class LispClient 
{
  public final String url;
  public static String defaultURL = "http://gentzen.ags.uni-sb.de:12345/RPC2/Broker";
  public org.apache.xmlrpc.XmlRpcClient broker_connection;
  
  public LispClient(String url) throws MalformedURLException {
    // --------------------------------------
    if (url == null)
      this.url = defaultURL;
    else 
      this.url = url;
    
    initXmlRpcClient();
  }
  
  public void initXmlRpcClient() throws MalformedURLException {
    // -----------------------------
    this.broker_connection = new org.apache.xmlrpc.XmlRpcClient(url);
  }
  
  public Object call(String url, String methodName, Vector params)  throws IOException, XmlRpcException {
    try {
      org.apache.xmlrpc.XmlRpcClient service_connection = new org.apache.xmlrpc.XmlRpcClient(url);
      Object result = service_connection.execute(methodName, params);
      return result;
    }
    catch( IOException e) {
      System.out.println("Trouble with RPC call " + e);
      throw e;
    }
    catch (XmlRpcException e) {
      System.out.println("Trouble with RPC call " + e);
      throw e;
    }
  }
	
  public String createService(String serviceName) throws IOException, XmlRpcException {
    try {
      Vector params = new Vector();
      params.addElement(serviceName);
      System.out.println("Sending xml-rpc request: getService with one parameter: \"" + serviceName + '"');
      String result = "";
      result = (String) broker_connection.execute("getService", params);
      return result;
    }
    catch( IOException e) {
      System.out.println("Trouble on  invoking createService " + e);
      initXmlRpcClient();
      throw e;
    }
    catch (XmlRpcException e) {
      System.out.println("Trouble on  invoking createService " + e);
      initXmlRpcClient();
      throw e;
    }
  }
  
  public static void main (String args[]) {
    try {
      String broker_url = "http://monster.ags.uni-sb.de:30000/RPC2/Broker";
      LispClient client = new LispClient(broker_url);
      String otter_problem_str =  "set(auto). list(usable). a | b. -a | b. -a | -b. a | -b. end_of_list.";
      
      String dfg_problem_str = "begin_problem(Pelletier54).\n"+
	"list_of_descriptions.\n"+
        "name({*Pelletiers Problem No. 54*}).\n"+
        "author({*Christoph Weidenbach*}).\n"+
	"status(unsatisfiable).\n"+
	"description({*Problem taken in revised form from\n"+
        "              the \"Pelletier Collection\",\n"+
	"              Journal of Automated Reasoning, Vol. 2, No. 2, pages 191-216*}).\n"+
	"end_of_list.\n"+
	"list_of_symbols.\n"+
	"  predicates[(F,2)].\n"+
	"end_of_list.\n"+
	"list_of_formulae(axioms).\n"+
	"  formula(forall([U],exists([V],forall([W],equiv(F(W,V),equal(W,U)))))).\n"+
	"end_of_list.\n"+
	
	"list_of_formulae(conjectures).\n"+
	"  formula(not(exists([U],forall([V],equiv(F(V,U),forall([W],implies(F(V,W),exists([X],and(F(X,W),not(exists([Y],and(F(Y,W),F(Y,X))))))))))))).\n"+
	"end_of_list.\n"+
	"end_problem.";
      
      String msc006_1 = "% BEGIN_AXIOMS\n"+
	"% NAME: transitive(p)\n"+
	"p(X, Z) :- p(X, Y), p(Y, Z)."+
	"% NAME: transitive(q)"+
	"q(X, Z) :- q(X, Y), q(Y, Z)."+
	"% NAME: symmetric(q)"+
	"q(Y, X) :- q(X, Y)."+
	"% NAME: all_related(p,q)"+
	"p(X, Y); q(X, Y)."+
	"% NAME: not_total(p)"+
	"- p(a, b)."+
	"% NAME: goal "+
	"?- q(c, d)."+
	"% END_AXIOMS";

      // ----------------------------------------------
      //    First, we call Otter on the otter_problem
      // ----------------------------------------------
      String otter_url = client.createService("OTTER");
      System.out.println("Otter URL: " + otter_url);
      
      Hashtable argument = new Hashtable();
      Hashtable replyWith = new Hashtable();
      replyWith.put("state", " ");
      replyWith.put("output", "all");
      replyWith.put("time", "all");
      
      argument.put("1", otter_problem_str);
      argument.put("syntax", "otter");
      argument.put("timeout", new Integer(20));
      argument.put("replyWith", replyWith);
      
      Vector params = new Vector();
      params.addElement(argument);
      
      Object result = client.call(otter_url, "prove", params);
      System.out.println("Result: " + result);
      //  never forget to kill the service object! :-)
      Object res = client.call(otter_url, "kill", params);
 
      // ----------------------------------------------
      //   , we call Spass on the dfg_problem
      // ----------------------------------------------
      String spass_url = client.createService("SPASS");
      System.out.println("\nSpass URL: " + spass_url);
      
      argument.clear(); 
      argument.put("1", dfg_problem_str);
      argument.put("syntax", "dfg");
      params = new Vector();
      params.addElement(argument);
      
      result = client.call(spass_url, "prove", params);
      System.out.println("Result: " + result);
      //  never forget to kill the service object! :-)
       res = client.call(spass_url, "kill", params);
      
//        // ----------------------------------------------
//        //    now we call the concurrent ATP service
//        //    with dfg_problem and msc006_1 
//        //  still a bit complicated I agree... :-(
//        // ----------------------------------------------
//        String atp_url = client.createService("ATP");
//        System.out.println("Atp URL: " + atp_url);
      
//        Hashtable atp_problem_map = new Hashtable();
//        atp_problem_map.put("spass", dfg_problem_str); // spass only understands dfg :-(
//        atp_problem_map.put("protein", msc006_1);      // protein gets protein syntax

//        Vector provers = new Vector();
//        provers.addElement("spass");
//        provers.addElement("bliksem");
//        provers.addElement("protein");
      
//        Hashtable problem = new Hashtable();
//        problem.put("id", new Integer(123444));       // some integer ...
//        problem.put("problem", atp_problem_map);
//        problem.put("provers", provers);
//        problem.put("mode", "all");
//        problem.put("timeout", new Integer(30));

//        params = new Vector();
//        params.addElement(problem);
//        result = client.call(atp_url, "proveProblem", params);
//        System.out.println("Result: " + result);
//        res = client.call(atp_url, "kill", params);

//        // --------------------------------------------
//        //     Try a simple Maple computation 
//        // --------------------------------------------
//        String maple_url = client.createService("MAPLE");
//        System.out.println("Maple URL: " + maple_url);
//        params = new Vector();
//        params.addElement("2+3");
//        result = client.call(maple_url, "eval", params);
//        System.out.println("Result: " + result);
//        res = client.call(maple_url, "kill", params);
//        //---------------------------------------------
    }
    catch( IOException e) {
      System.out.println("Trouble with maple invokation... " + e);
    }
    catch (XmlRpcException e) {
      System.out.println("Trouble with maple invokation... " + e);
    }
    System.exit(0);
  }

} // class MathWebBroker
