// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

/** sample calls:

runt org.mathweb.eval.TPTPEval lambda.ags.uni-sb.de 8080 proveProblemChoice /home/jzimmer/import/mathserve/thirdParty/TPTPWorld/TPTP/Problems RandomCollection-2005-6-1-13:2-500 10

*/

package org.mathweb.eval;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Calendar;
import javax.xml.rpc.ParameterMode;

import sun.misc.Signal;
import sun.misc.SignalHandler;

import org.apache.axis.client.Call;
import org.apache.axis.client.Service;
import org.apache.axis.encoding.XMLType;

import org.jdom.*;

import org.mathweb.onto.OWLJDom;
import org.mathweb.serv.*;
import org.mathweb.broker.*;
import org.mathweb.util.*;
import org.mathweb.owls.*;

/**
   @author Juergen Zimmer, Saarland University, Germany
   @version $Date: 2005/07/15 11:22:19 $ $Revision: 1.21 $
*/
public class TPTPEval {
  
  static private int delay = 1;

  protected static String msHome = System.getProperty("mathserv.home");
  
  public String TPTP_HOME = msHome + "/thirdParty/TPTPWorld";
  
  private static OWLJDom owlJDom = OWLJDom.getInstance();    
  
  private boolean busy = false;
  private boolean shutdown = false;

  //  private OWLSParser owlsParser = new OWLSParser();
  
  private TSTPCreator creator = new TSTPCreator();

  static class MyHandler implements SignalHandler {
    
    TPTPEval myObj;
    
    public MyHandler(TPTPEval evalObj){
      super();
      myObj = evalObj;
    }
    
    public void handle(Signal sig) {
      myObj.shutdown();
    }
  }
  
  /** 
   * Reads the contens of a file in a java.lang.String.
   * @param fileName the name of the file to read.
   * @return The contents of the file.
   */
  public String readFile(String fileName){
    FileInputStream isFile  = null;
    try {
      isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    } finally {
      if (isFile != null)
        try{ isFile.close(); } catch (Exception e) {}
    }
  }
  

  
  /** 
   * Writes a string into a file
   * @param fileName the name of the file to write into.
   * @param content the string to be written.
   * @return True if the file as written successfully, false otherwise.
   */
  protected boolean writeFile(String fileName, String content){
    FileOutputStream outFile = null;
    try {
      outFile = new FileOutputStream(fileName) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    } finally {
      if (outFile != null)
        try{ outFile.close(); } catch (Exception e) {}
    }
  }
  
  public void eval(String programName, String spcName, int time)
  {
    Calendar rightNow = Calendar.getInstance();
    String outputFileName = msHome + "/etc/eval/"+spcName+ "-" + programName + "-"+ time + "sec-" +
      rightNow.get(Calendar.DAY_OF_MONTH) + "-" + rightNow.get(Calendar.MONTH) + "-" + rightNow.get(Calendar.YEAR) + ".dat";
    FileOutputStream outputFile = null; 
    
    String spcFileName = msHome + "/thirdParty/TPTPWorld/NewSPCLists/"+spcName;
    String programFileName = msHome + "/thirdParty/golog/dtgolog/"+programName;
    
    try{
      outputFile = new FileOutputStream(outputFileName) ;
      String program = readFile(programFileName);
      String[] spcLines = readFile(spcFileName).split("\n");
      int problems = spcLines.length;
      String[] probNames = new String[problems];
      String[] tstpFileNames = new String[problems];
      for(int i = 0; i< problems; i++){
        String line = spcLines[i];
        String[] parts = line.split("   ");
        if(parts.length>1){
          String prob = parts[0].trim();
          String group = prob.substring(0,3);
          probNames[i]= prob;
          tstpFileNames[i] = TPTP_HOME+ "/TPTP/Problems/"+ group + "/" + prob + ".p";
           String owlsQuery = creator.tstpFile2OWLSQuery(tstpFileNames[i]);
           //          writeFile(msHome + "/thirdParty/TPTPWorld/TPTP/OWLProblems/"+group+ "/"+prob+".owl", owlsQuery);
          
          //          Query query = owlsParser.parseQuery(owlsQuery);
//           String result = executor.executeProgram(program, query, serviceRegistry, time, false);
//           Element resultRoot = owlJDom.getRootElement(result);
//           Double wcDouble = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(resultRoot);
//           Double cpuDouble = (Double) OWLJDom.TIME_RESOURCE_CPU.numberValueOf(resultRoot);
//           String status = OWLJDom.STATUS.valueOf(resultRoot);
//           if ((status == null) || (status.equals(""))) status = "Unknown";
//           System.out.println("status: " + status);
          
//           String content =  cpuDouble.intValue() + " \t" + wcDouble.intValue() + " \t" + probNames[i] + " \t" + status + "\n";
//           outputFile.write(content.getBytes());
//           outputFile.flush();
        }
      }
      
    }catch(Exception e) {
      e.printStackTrace();
    }finally{
      try{
        outputFile.close();
      }catch(Exception e) {
        e.printStackTrace();
      }
    }
  }

  public void evalRemote(String endpoint, String fileListName, String prefix, String method, int time)
  {
    //    String method = "proveProblemOpt";
    //    String endpoint = "http://lambda.ags.uni-sb.de:8080/axis/services/Broker";
    Calendar rightNow = Calendar.getInstance();
    
    String outputFileName = msHome + "/etc/eval/"+fileListName+ "-remote-" + method + "-"+ time + "sec-" +
      rightNow.get(Calendar.DAY_OF_MONTH) + "-" + (rightNow.get(Calendar.MONTH) + 1) + "-" + rightNow.get(Calendar.YEAR) + ".dat";
    FileOutputStream outputFile = null; 
    
    String spcFileName = msHome + "/thirdParty/TPTPWorld/NewSPCLists/"+fileListName;
    
    try{
      if (! ((method.equals("proveProblemOpt")) || (method.equals("proveProblemSplit")) ||
             (method.equals("proveProblemSplitJoin")) || (method.equals("proveProblemChoice"))))
        throw new Exception("The method " + method + " is not valid!");
    
      outputFile = new FileOutputStream(outputFileName) ;
      String[] spcLines = readFile(spcFileName).split("\n");
      int problems = spcLines.length;
      String[] probNames = new String[problems];
      String[] tstpFileNames = new String[problems];
      Integer timeInt = new Integer(time - delay);

      Service service = new Service();
      Call call = (Call) service.createCall();

      call.setTargetEndpointAddress(new java.net.URL(endpoint));
      call.setOperationName( method );
      call.addParameter("in0", XMLType.XSD_STRING, ParameterMode.IN);
      call.addParameter("in1", XMLType.XSD_INT, ParameterMode.IN);
      call.setReturnType(XMLType.XSD_STRING);
      call.setTimeout(new Integer((time + 40)*1000));
      
      for(int i = 0; i< problems; i++){
        
        if (this.shutdown) i = problems;
        else {
          String line = spcLines[i];
          String prob = line;
          String group = "UKN";
          String[] parts = line.split("   ");
          if (parts.length > 1){
            prob = parts[0].trim();
          }
          if (line.length() > 2) 
            group = prob.substring(0,3);
          probNames[i]= prob;
          
          //        tstpFileNames[i] = TPTP_HOME+ "/TPTP/Problems/"+ group + "/" + prob + ".p";
          tstpFileNames[i] = prefix + "/" + group + "/" + prob + ".p";
        //          System.out.println("prob: " + prob + "   group: " + group);
          
          String content =  probNames[i] + " \t 0 \t 0 \t Unknown \n";
          try{ 
            //          String owlsQuery = creator.tstpFile2OWLSQuery(tstpFileNames[i]);
            //          writeFile(msHome + "/thirdParty/TPTPWorld/TPTP/OWLProblems/"+group+ "/"+prob+".owl", owlsQuery);
            //String result = (String) call.invoke( new Object [] {owlsQuery, timeInt });
            
            String problem = readFile(tstpFileNames[i]);
            String result = (String) call.invoke( new Object [] {problem, timeInt });
            
            //          System.out.println(result);
            Element resultRoot = owlJDom.getRootElement(result);
            Double wcDouble = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(resultRoot);
            Double cpuDouble = (Double) OWLJDom.TIME_RESOURCE_CPU.numberValueOf(resultRoot);
            String status = OWLJDom.STATUS.valueOf(resultRoot);
            String rdfID = OWLJDom.POTENTIAL_ID.valueOf(resultRoot);
            
            if ((status == null) || (status.equals(""))) status = "Unknown";
            else {
              String[] split = status.split("#");
              if (split.length > 1) status = split[1];
            }
            
            content =  probNames[i] + " \t" + cpuDouble.intValue() + " \t" + wcDouble.intValue() + " \t" + rdfID + "    \t" +  status + "\n";
            if ((status.equals("Unsatisfiable")) || (status.equals("Theorem")))
              try{
                Element proof = (Element) OWLJDom.PROOF_FORMAL_PROOFS.selectSingleNode(resultRoot);
                if (proof != null){
                  String tstpFileName = msHome + "/etc/data/tstp/"+ rdfID + ".tstp";
                  writeFile(tstpFileName, proof.getTextTrim());
                }  
              } catch(Exception e){
                e.printStackTrace();
                System.exit(1);
              }
            else if ((status.equals("CounterSatisfiable")) || (status.equals("Satisfiable")))
              try{
                Element proof = (Element) OWLJDom.PROOF_FORMAL_PROOFS.selectSingleNode(resultRoot);
                if (proof != null){
                  String tstpFileName = msHome + "/etc/data/tstp/"+ rdfID + ".tstp";
                  writeFile(tstpFileName, proof.getTextTrim());
                }  
              } catch(Exception e){
                e.printStackTrace();
                System.exit(1);
              }   
          } catch(Exception e){
            e.printStackTrace();
            //            System.exit(1);
          }
          System.out.println(content);
          outputFile.write(content.getBytes());
          outputFile.flush();
        }
      }
    }catch(Exception e) {
      e.printStackTrace();
    }finally{
      try{
        outputFile.close();
      }catch(Exception e) {
        e.printStackTrace();
      }
    }
  }


  public synchronized void shutdown(){
    this.shutdown = true;
  }
    
  /*  the name of a golog program  and of a TSTP SPC file and time in seconds */
  public static void main(String[] args) {
    TPTPEval c = new TPTPEval();
    Signal.handle(new Signal("INT") , new MyHandler(c));
    Signal.handle(new Signal("TERM") , new MyHandler(c));

    try{
      if (args.length < 6) {
        System.out.println("Usage: TPTPEval <hostname> <port> <method> <dir prefix for TSTP files> <TSTP problem list file> <time limit> [&proxy host &proxy port]");
        System.exit(1);
      } else {
        if (args.length > 7){
          String proxyHost = args[6];
          String proxyPort = args[7];
          if ((!proxyHost.equals("")) && (!proxyPort.equals(""))){
            System.out.println("using proxy: " + proxyHost + ":" + proxyPort);
            System.setProperty("http.proxyHost", proxyHost); 
            System.setProperty("http.proxyPort", proxyPort); 
          }}
        String endpoint = "http://" + args[0] + ":" + args[1] + "/axis/services/Broker";
        String method = args[2];
        String prefix = args[3];
        String probListFileName = args[4];
        Integer timeInt = new Integer(args[5]);
        int time = timeInt.intValue();
        
        c.evalRemote(endpoint, probListFileName, prefix, method, time);
      }
    } catch(Exception e){
      e.printStackTrace();
      System.out.println("FAILURE");
        System.out.println("Usage: TPTPEval <hostname> <port> <method> <dir prefix for TSTP files> <TSTP problem list file> <time limit> [&proxy host &proxy port]");
      System.exit(1);
    }
  }

} // end TPTPEval.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

