package org.mathweb.graph;

import salvo.jesus.graph.*;
import org.mathweb.graph.Annotation;

/**
 * Represents a directed annotated edge in a graph.
 * <p>
 * To do:<br>
 * Expose the <tt>LabeledEdge</tt> implementation so that
 * <tt>setFollowVertexLabel()</tt> can be called to turn it on or off.
 *
 * @author		Juergen Zimmer
 * @version             $Id: AnnotatedEdge.java,v 1.3 2005/01/10 18:47:21 jzimmer Exp $
 */

public class AnnotatedEdge extends DirectedEdgeImpl implements DirectedEdge {
   
  private Annotation annotation; 
  
  /**
   * Creates an AnnotatedEdge object.
    *
    * @see		Vertex
    */
  public AnnotatedEdge(Vertex a, Vertex b, Annotation annotation ) {
    super(a, b);
    this.annotation = annotation;
  }

  public Annotation getAnnotation() {
    return this.annotation;
  }

  public void setAnnotation(Annotation ann) {
    this.annotation = ann;
  }

  /**
    * Returns a String representation of the Edge.
    * By default, the format is:
    * fromVertex.toString() + "->" + toVertex.toString()
    *
    * @return	The String representation of the Edge
    * @see		Vertex
    */
//   public String toString(){
//     return getLabel() + annotation;
//   }

  /**
    * Creates a clone of this Edge. This calls the AnnotatedEdge constructor,
    * thereby creating a new instance of AnnotatedEdge. However, the vertices
    * in both endpoints of the Edge are not cloned.
    *
    * @return  A clone of an instance of Edge.
    */
  protected Object clone(){
    return new AnnotatedEdge(vertexA, vertexB, annotation);
  }

}

