// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
/*
 * Created on Dec 22, 2004
 */
package org.mathweb.onto;

import java.io.*;
import java.util.*;
import java.net.*;

import org.jdom.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 

import com.hp.hpl.jena.util.LocationMapper;

import org.mathweb.core.BasicService;

public class OntPolicyReader extends BasicService 
{
  public static final String RDF_NS_STR = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
  public static final String RDFS_NS_STR = "http://www.w3.org/2001/XMLSchema#";
  public static final String JENA_NS_STR = "http://jena.hpl.hp.com/schemas/2003/03/ont-manager#";
  
  public static final Namespace RDF_NS = Namespace.getNamespace("rdf", RDF_NS_STR );
  public static final Namespace JENA_NS = Namespace.getNamespace("jena", JENA_NS_STR);

  private String policyConfigFile = getSystemProperty("mathserv.ontology.policy", "MATHSERV_ONTOLOGY_POLICY");
  public static OntPolicyReader theInstance = (new OntPolicyReader()).readConfig();
  private LocationMapper locMap;


  public OntPolicyReader readConfig()
  {
    locMap = new LocationMapper();
    OWLJDom owlJDom = OWLJDom.getInstance();
    String rdf = readFile(policyConfigFile);
    try {
      Element root = owlJDom.getRootElement(rdf);
      List mappings = root.getChildren("OntologySpec", JENA_NS);
      for(int i = 0; i< mappings.size(); i++){
	Element map = (Element) mappings.get(i);
	Element uri = map.getChild("publicURI", JENA_NS);
	Element alt = map.getChild("altURL", JENA_NS);
	if ((uri != null) && (alt != null)){
	  String uriStr = uri.getAttributeValue("resource", RDF_NS);
	  String altStr = alt.getAttributeValue("resource", RDF_NS);
	  //	  System.out.println("map: " + uriStr + " --> " + altStr);
	  locMap.addAltEntry(uriStr, altStr);
	}
      }
    } catch(Exception ex){
      ex.printStackTrace();
    }
    return this;
  }
  

  public LocationMapper getMapper()
  {
    return locMap;
  }
  
}
