// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.onto;

import java.net.*;

public interface OntoReasoner
{
  
  /** 
   * @param instance the name of an OWL instance
   * @param className the name of an OWL class.
   * @return true if the individual is of the given class
   */
  public boolean hasClass(String instance, String className);
  
  
  /** 
   * @param class1 the name of an OWL class
   * @param class2 the name of another OWL class.
   * @return true if class1 is an asserted or inferred subclass of class2
   */
  public boolean subClassOf(String class1, String class2);
  
  /** 
   * Tells an individual in the OWL KB.
   * @param indStr the RDF XML string of the individual
   */
  public boolean tellIndividual(String indStr);
  
  
  /** 
   * Untells an individual in the OWL KB.
   * @param indStr the RDF XML string of the individual
   */
  public boolean untellIndividual(String indStr);
  
  
  public void assertLiteral(String property, URI indiURI, URI valueURI);

  /** 
   * @param className the name of an OWL class.
   * @return the rdf:ID of the last individual told of this class
   */
  public String nameOfLastIndividual(String owlClass);
  

  public StringBuffer getProdigyClassTree();
  

  /** 
   * @param property the name of a MathServ OWL property
   * @param individual the name of a MathServ OWL individual in the current model.
   * @return The string value of 'property' of the 'individual.
   */
  public String valueOf(String property, URI individual);

  /** 
   * @param property the name of a MathServ OWL property
   * @param individual the name of a MathServ OWL individual in the current model.
   * @return The string value of 'property' of the 'individual.
   */
  public String valueOf(String property, String individual);

  /** 
   * @return true if the ontology with the individuals told so far is consistent.
   */
  public boolean isValid();
  
  
  
}
  
