package org.mathweb.reasoning.basicplanner;
import java.util.*;
import org.mathweb.reasoning.search.*;
import org.mathweb.reasoning.planning.*;
/**
 * A basic planner that does uninformed search to generate plans.
 *
 * @author Harlan Crystal, Daniel Terry
 * @date October 28, 2003
 */
public class BasicPlanner implements Planner {

    /** The logic system we exist in */
    private LogicSystem system;

    public BasicPlanner(LogicSystem system) {
        this.system = system;
    }

    public Plan generatePlan(State initial, State goal) {
        Plan plan = new Plan();
        BasicPlannerNode start = new BasicPlannerNode(0, initial, plan);

        //Fringe fringe = new Queue();
        Fringe fringe = new IterativeDeepeningStack();
        fringe.addToFringe(start);

        while (fringe.hasNext()) {
            BasicPlannerNode curNode = (BasicPlannerNode) fringe.getNext();
            State curState = curNode.getState();
            Plan curPlan = curNode.getPlan();
            int curDepth = curNode.getDepth();
	    //System.out.println("\n\nCUR STATE: \n" + curState);

            if (goal.isSatisfiedBy(curState))
                return curPlan;

	    for (Iterator it = system.getApplicableActions(curState).iterator(); it.hasNext(); ) {
		BoundAction a = (BoundAction) it.next();
		//System.out.println("Generating Successor of:\n" + a);
		Plan nextPlan = curPlan.addActionToEnd(a);
		BasicPlannerNode nextNode = new BasicPlannerNode(curDepth + 1, a.applyAction(curState), nextPlan);
		//System.out.println("Successor:\n" + nextNode.getState());
		fringe.addToFringe(nextNode);
	    }
        }
	System.out.println("BasicPlanner::generatePlan: No Solution Found");

        return null;
    }
}
