package org.mathweb.reasoning.basicplanner;
import org.mathweb.reasoning.planning.State;
import org.mathweb.reasoning.planning.Plan;
import org.mathweb.reasoning.search.Node;
/**
 * Represents a single node in the search space.
 *
 * Just a tuple of the current state, and the current plan.
 *
 * Needed for the book keeping of doing search.
 *
 * @author Harlan Crystal, Daniel Terry
 * @date October 28, 2003
 */
public class BasicPlannerNode extends Node {
    final private State state;
    final private Plan plan;
    public static int nodeCount = 0;

    /**
     * Creates a new node.
     * 
     * @param depth The depth of this node in the search tree
     * @param state The state associated with this node.
     * @param plan The plan associated with this node.
     */
    public BasicPlannerNode(int depth, State state, Plan plan) {
        super(depth);
        this.plan = plan;
        this.state = state;
        nodeCount++;
    }

    /**
     * @return The state associated with this node.
     */
    public State getState() {
        return state;
    }

    /**
     * @return The plan associated with this node.
     */
    public Plan getPlan() {
        return plan;
    }

    public int hashCode() {
        return super.hashCode() ^ plan.hashCode() ^ state.hashCode();
    }

    public boolean equals(BasicPlannerNode n) {
        return super.equals(n) && plan.equals(n.plan) && state.equals(n.state);
    }

    public boolean equals(Object o) {
        try {
            return equals((BasicPlannerNode) o);
        } catch (ClassCastException e) {
            return false;
        }
    }

    public String toString() {
        return super.toString() + ", plan: " + plan;
    }
}
