package org.mathweb.reasoning.basicplanner;

import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
import org.mathweb.reasoning.parser.Parser;

import java.util.Date;

/**
 * Used for testing the basic planner.
 *
 * @author Harlan Crystal, Daniel Terry
 */
public class PlannerTest {
    BasicPlanner planner;
    LogicSystem system;
    Parser pp;


    public PlannerTest(String filename) throws Exception {
        pp = Parser.loadFromFile(filename);
        system = new LogicSystem(pp.getActions());
        planner = new BasicPlanner(system);
    }

    public void test(State initial, State goal) {
        System.out.println("INITIAL:");
        System.out.println(initial);
        System.out.println("GOAL:");
        System.out.println(goal);
	Date startDate = new Date();
        Plan plan = planner.generatePlan(initial, goal);
	Date finishDate = new Date();
	System.out.println(startDate + "\n" + finishDate);
        System.out.println("PLAN:");
        System.out.println(plan);
    }
    
    /**
     * Takes in an argument which is a problem file.
     */
    public static void main(String [] args) {
        try{ 
            System.out.println("Planner test 1:");
            System.out.println("---------------");
            PlannerTest test = new PlannerTest(args[0]);
            test.test(test.pp.getInitialState(), test.pp.getGoalState());
            System.out.println("Num nodes: " + BasicPlannerNode.nodeCount);
        } catch (Exception e) {
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }
}
