package org.mathweb.reasoning.logic;

/**
 * A base class for any object which wishes to cache it's hashcode upon
 * first call to hashcode.
 *
 * @author Harlan Crystal, Daniel Terry
 */
public abstract class CachedHashCodeObject {
    /** The precomputed hashcode.  if it is zero, it is assumed to be uninitialized */
    protected int myhashcode;

    public CachedHashCodeObject() {
        myhashcode = 0;
    }

    /** 
     * This method computes the cached hashcode.
     *
     * It normally is called by the hashCode() function upon first use,
     * but could be called to force hashcode recomputation.
     */
    protected abstract void computeHashCode();

    public int hashCode() {
        /* there is a ridiculously remote chance that this object's real 
         * hashcode is 0, in which case it is recomputing hashcode every time.
         * this barely ever happens, and isn't much of a loss when it is */
        if (myhashcode == 0)
            computeHashCode();
        
        return myhashcode;
    }
}
