package org.mathweb.reasoning.logic;

import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.ArrayList;

import org.mathweb.reasoning.partialorderplanner.NameGen;

/**
 * A Conjunct is predicate and a list of arguments
 *
 * An immutable object.
 *
 * @author Daniel Terry, Harlan Crystal
 * @date October 22, 2003
 */
public class Conjunct extends CachedHashCodeObject {
    final private ArrayList literals;

    /**
     * Construct a new literal.
     *
     * @param predicate The predicate
     * @param argument The list of arguments of the literal.
     * @param truth True if this literal is true, false if this literal is false.
     */
  public Conjunct(ArrayList literals)
  {
    this.literals = literals;
  }
  

  protected void computeHashCode() {
    int value = 2;
    for (int i = 0; i < literals.size(); i++)
      value = value ^ ((Literal) literals.get(i)).hashCode();
    myhashcode = value;
  }
  
  /**
   * @return The number of arguments.
   */
  public ArrayList getLiterals() {
    return literals;
  }
  
  
  /**
   * @return The set of unbound Variables in this literal
   */
  public Set getVariables() {
    HashSet set = new HashSet();
    Set lset;
    for (int i = 0; i < literals.size(); i++){
      Literal lit = (Literal) literals.get(i);
      lset = lit.getVariables();
      set.addAll(lset);
    }
    return set;
  }
  
  public String toString()
  {
    StringBuffer str = new StringBuffer();
    str.append("(");
    for (int i = 0; i < literals.size(); i++){
      Literal lit = (Literal) literals.get(i);
      str.append(lit.toString());
      if (i < (literals.size()-1)) str.append(" /\\ ");
    }
    str.append(")");
    return str.toString();
  }
  
}
