package org.mathweb.reasoning.logic;

import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.ArrayList;

import org.mathweb.reasoning.partialorderplanner.NameGen;

/**
 * A Implication is predicate and a list of arguments
 *
 * An immutable object.
 *
 * @author Daniel Terry, Harlan Crystal
 * @date October 22, 2003
 */
public class Implication extends CachedHashCodeObject {
    final private Literal head;
    final private ArrayList body;

    /**
     * Construct a new literal.
     *
     * @param predicate The predicate
     * @param argument The list of arguments of the literal.
     * @param truth True if this literal is true, false if this literal is false.
     */
  public Implication(Literal head, ArrayList body)
  {
    this.head = head;
    this.body = body;
  }
  

  protected void computeHashCode() {
    int value = 2;
    for (int i = 0; i < body.size(); i++)
      value = value ^ ((Literal) body.get(i)).hashCode();
    value = value ^ head.hashCode();
    myhashcode = value;
  }
  
  /**
   * @return The number of arguments.
   */
  public ArrayList getBody() {
    return body;
  }


  public Literal getHead() {
    return head;
  }
  
  
  /**
   * @return The set of unbound Variables in this literal
   */
  public Set getVariables() {
    HashSet set = new HashSet();
    Set lset;
    for (int i = 0; i < body.size(); i++){
      Literal lit = (Literal) body.get(i);
      lset = lit.getVariables();
      set.addAll(lset);
    }
    set.addAll(head.getVariables());
    return set;
  }
  
  public String toString()
  {
    StringBuffer str = new StringBuffer();
    str.append("(");
    for (int i = 0; i < body.size(); i++){
      Literal lit = (Literal) body.get(i);
      str.append(lit.toString());
      if (i < (body.size()-1)) str.append(" /\\ ");
    }
    str.append(" => ");
    str.append(head.toString());
    str.append(")");
    return str.toString();
  }
  
}
