package org.mathweb.reasoning.logic;

import java.util.*;
import org.mathweb.reasoning.planning.*;

/**
 * A bunch of tests to check out logic engine.
 *
 * @author Harlan Crystal, Daniel Terry.
 */
public class LogicTests {

    public static void main(String[] args) {
	//findBindingsTest(); //The old way
	//findBindingsTestFile(); //The way of the future
	//bindPreconditionsToStateTest();

// 	LogicLoader ll = new LogicLoader("test.txt");
// 	System.out.println(ll);
// 	Vector bindings = Actor.findBindings(ll.getLiteralRange(0,2), ll.getLiteralRange(2,3), new BindingTable());
// 	System.out.println("Experimental Solution:");
// 	for(int i = 0; i < bindings.size(); i++){
// 	    System.out.println(bindings.get(i));
// 	}
	
	
//	applyActionTest();

    }

    public static void applyActionTest(){
	LogicLoader ll = new LogicLoader("logic/applyActionTest.txt");
	System.out.println(ll);
	State s =  (State)ll.states.get(0);
        LogicSystem system = new LogicSystem(ll.actions);
	System.out.println("Experimental Solution:");
        for (Iterator it=system.generateSuccessors(s).iterator(); it.hasNext(); ) {
            State nextState = (State) it.next();
            System.out.println("New State:\n" + nextState + "\n\n");
        }

        /*
	Vector bindings = Actor.bindPreconditionsToState(a,s);

	for(int i = 0; i < bindings.size(); i++){
	    System.out.println(bindings.get(i));
	    System.out.println("New State:\n" + 
			       a.applyAction(s,(BindingTable)bindings.get(i))+"\n\n");
	}
        */



	/*Actual Solutions:
	  VARIBLE::v1-->CONSTANT::a
	  VARIBLE::v3-->CONSTANT::a
	  VARIBLE::v2-->CONSTANT::b
	  VARIBLE::v0-->CONSTANT::b
	  
	  VARIBLE::v1-->CONSTANT::d
	  VARIBLE::v3-->CONSTANT::a
	  VARIBLE::v2-->CONSTANT::b
	  VARIBLE::v0-->CONSTANT::e

	  VARIBLE::v1-->CONSTANT::a
	  VARIBLE::v3-->CONSTANT::d
	  VARIBLE::v2-->CONSTANT::e
	  VARIBLE::v0-->CONSTANT::b
	  
	  VARIBLE::v1-->CONSTANT::d
	  VARIBLE::v3-->CONSTANT::d
	  VARIBLE::v2-->CONSTANT::e
	  VARIBLE::v0-->CONSTANT::e */
    }

//     public static void bindPreconditionsToStateTest(){
// 	LogicLoader ll = new LogicLoader("logic/bindPreconditionsToStateTest.txt");
// 	System.out.println(ll);
// 	Vector bindings = Actor.bindPreconditionsToState((Action)ll.actions.get(0), (State)ll.states.get(0));
// 	System.out.println("Experimental Solution:");
// 	for(int i = 0; i < bindings.size(); i++){
// 	    System.out.println(bindings.get(i));
// 	}
//     }

    //Same as below function findBindingsTest, but uses the LogicLoader
    public static void findBindingsTestFile() {
	//This test finds all the possible bindings
	//from
	// On(v1,v2)
	// On(v2,v3)
	//to
	// On(a,b)
	// On(b,c)
	// On(c,d)
	// On(b,e)
	//
	//Solutions: (v1=a,v2=b,v3=c), (v1=a,v2=b,v3=e), (v1=b,v2=c,v3=d)

	LogicLoader ll = new LogicLoader("logic/findBindingsTest1.txt");
	Vector bindings = Actor.findBindings(ll.getLiteralRange(4,6), ll.getLiteralRange(0,4), new BindingTable(), new BindingConstraints());
	System.out.println("Experimental Solution:");
	for(int i = 0; i < bindings.size(); i++){
	    System.out.println(bindings.get(i));
	}

	System.out.println("Theoretical Solution:");
	System.out.println("(v1=a,v2=b,v3=c), \n(v1=a,v2=b,v3=e), \n(v1=b,v2=c,v3=d)");
    }

    
    public static void findBindingsTest() {
 	Predicate on = new Predicate("On");

	//This test finds all the possible bindings
	//from
	// On(v1,v2)
	// On(v2,v3)
	//to
	// On(a,b)
	// On(b,c)
	// On(c,d)
	// On(b,e)
	//
	//Solutions: (v1=a,v2=b,v3=c), (v1=a,v2=b,v3=e), (v1=b,v2=c,v3=d)

  	Argument[] onArguments1 = new Argument[2];
  	onArguments1[0] = new Constant("a");
  	onArguments1[1] = new Constant("b");

  	Argument[] onArguments2 = new Argument[2];
  	onArguments2[0] = new Constant("b");
  	onArguments2[1] = new Constant("c");

  	Argument[] onArguments3 = new Argument[2];
  	onArguments3[0] = new Constant("c");
  	onArguments3[1] = new Constant("d");

  	Argument[] onArguments4 = new Argument[2];
  	onArguments4[0] = new Constant("b");
  	onArguments4[1] = new Constant("e");

	Vector constLits = new Vector(4);
	constLits.add(new Literal(on, onArguments1, true));
	constLits.add(new Literal(on, onArguments2, true));
	constLits.add(new Literal(on, onArguments3, true));
	constLits.add(new Literal(on, onArguments4, true));


  	Argument[] onArguments5 = new Argument[2];
  	onArguments5[0] = new Variable("v1");
  	onArguments5[1] = new Variable("v2");

  	Argument[] onArguments6 = new Argument[2];
  	onArguments6[0] = new Variable("v2");
  	onArguments6[1] = new Variable("v3");

	Vector varLits = new Vector(2);
	varLits.add(new Literal(on, onArguments5, true));
	varLits.add(new Literal(on, onArguments6, true));

	Vector bindings = Actor.findBindings(varLits, constLits, new BindingTable(), new BindingConstraints());
	System.out.println("Experimental Solution:");
	for(int i = 0; i < bindings.size(); i++){
	    System.out.println(bindings.get(i));
	}

	System.out.println("Theoretical Solution:");
	System.out.println("(v1=a,v2=b,v3=c), \n(v1=a,v2=b,v3=e), \n(v1=b,v2=c,v3=d)");
    }




}
