package org.mathweb.reasoning.partialorderplanner;

/**
 * A singleton class that generates new names.
 *
 * @author Daniel Terry, Harlan Crystal
 */
public class NameGen {
    /** The next number from this name generator */
    private int next;
    /** The initial number used for this name generator */
    private int initial;
    /** The single instance of this class */
    private static NameGen theInstance = null;

    /**
     * Private constructor, because this is a singleton design pattern.
     *
     * @param next The next number to use for this name generator.
     */
    private NameGen(int next){
 	this.next = next;
	this.initial = next;
    }
    
    /**
     * @return The single instance of this class.
     */
    public static NameGen getNameGen(){
 	if(theInstance == null)
 	    theInstance = new NameGen(1000);
 	return theInstance;
    }

    /**
     * @return The next number to use for unique naming.
     */
    public int getNext(){
	return next++;
    }

    /**
     * Resets this name generator to the inital number again.
     */
    public void reset() {
        next = initial;
    }

     public Object clone(){
 	NameGen x = new NameGen(next);
        x.initial = initial;
        return x;
     }
}
