package org.mathweb.reasoning.partialorderplanner;
import javax.swing.*;
import javax.swing.tree.*;
import java.awt.*;
import java.awt.event.*;

import org.mathweb.reasoning.parser.*;
import org.mathweb.reasoning.planning.*;

public class POPGUI extends JFrame {
    POPPanel popPanel;
    ActionPanel actionPanel;
    SearchTreePanel searchTreePanel;
    PartialOrderPlanner planner;
    ActionTree actionTree;
    POPTree popTree;
    SearchTree searchTree;

    public POPGUI(String filename) {
	loadPlanner(filename);

	getContentPane().setLayout(new BorderLayout());

	popPanel = new POPPanel(planner.getRoot().getPlan());
	getContentPane().add(popPanel, BorderLayout.WEST);

	actionPanel = new ActionPanel((org.mathweb.reasoning.planning.Action)planner.getRoot().getPlan().getActions().iterator().next());
	getContentPane().add(actionPanel, BorderLayout.EAST);

	searchTreePanel = new SearchTreePanel(planner);
	getContentPane().add(searchTreePanel, BorderLayout.CENTER);

	popTree = popPanel.getTree();
	popTree.addMouseListener(new POPTreeMouseListener());

	actionTree = actionPanel.getTree();
	//actionTree.addMouseListener(new SearchTreeMouseListener());

	searchTree = searchTreePanel.getTree();
	searchTree.addMouseListener(new SearchTreeMouseListener());

	setSize(1000, 400);
	setVisible(true);
    }

    public POPGUI(PartialOrderPlanner pop) {
      planner = pop;
      
      getContentPane().setLayout(new BorderLayout());
      
      popPanel = new POPPanel(planner.getRoot().getPlan());
      getContentPane().add(popPanel, BorderLayout.WEST);
      
      actionPanel = new ActionPanel((org.mathweb.reasoning.planning.Action)planner.getRoot().getPlan().getActions().iterator().next());
      getContentPane().add(actionPanel, BorderLayout.EAST);
      
      searchTreePanel = new SearchTreePanel(planner);
      getContentPane().add(searchTreePanel, BorderLayout.CENTER);
      
      popTree = popPanel.getTree();
      popTree.addMouseListener(new POPTreeMouseListener());
      
      actionTree = actionPanel.getTree();
      //actionTree.addMouseListener(new SearchTreeMouseListener());
      
      searchTree = searchTreePanel.getTree();
      searchTree.addMouseListener(new SearchTreeMouseListener());
      
      setTitle("MathServ POP GUI");
      setSize(1000, 400);
      setBackground( Color.gray );
      Rectangle bounds = this.getBounds();
      setLocation(15 + bounds.x, 15 + bounds.y);
      setVisible(true);
    }
  
    private void loadPlanner(String filename) {
	Parser pp;
	try{
	  System.out.println("loading planning problem...");
	  pp = Parser.loadFromFile(filename);
	  System.out.println("...done");
	  planner = new PartialOrderPlanner(new LogicSystem(pp.getActions()), pp.getInitialState(), pp.getGoalState());
	} catch(Exception e){
	  System.out.println(e);
	}
    }
    
    public PartialOrderPlan generatePlan(){
	return planner.generatePlan();
    }

    public void resetPopTree(PartialOrderPlan plan){
	resetPopTree(plan, null);
    }

    public void resetPopTree(PartialOrderPlan plan1, PartialOrderPlan plan2){
 	popPanel.resetTree(plan1, plan2);
 	popTree = popPanel.getTree();
 	popTree.addMouseListener(new POPTreeMouseListener());
    }

//     public void resetActionTree(org.mathweb.reasoning.planning.Action action){

//     }


    public static void main(String[] args){
	POPGUI popgui = new POPGUI(args[0]);
	popgui.generatePlan();
    }

    private class SearchTreeMouseListener extends MouseAdapter {
	public void mousePressed(MouseEvent e) {
	    TreePath[] selPaths = searchTree.getSelectionPaths();
	    if(selPaths != null) {
		Object[] selects = getTerminalUserObjects(selPaths);
		if(selects.length == 1){
		    try{
			if(e.getClickCount() == 1)
			    resetPopTree((PartialOrderPlan)selects[0]);
			else if(e.getClickCount() == 2){
			    System.out.println("HI");
			    planner.expandPlan((PartialOrderPlan)selects[0]);
			}
		    } catch(ClassCastException cce){ }
		}
		else if(selects.length == 2){
		    try{
			resetPopTree((PartialOrderPlan)selects[0], (PartialOrderPlan)selects[1]);
			//resetPopTree((PartialOrderPlan)userObj1, (PartialOrderPlan)userObj2);
		    } catch(ClassCastException cce){ }
		}
	    }
	}
	public Object[] getTerminalUserObjects(TreePath[] paths){
	    Object[] objs = new Object[paths.length];
	    for(int i = 0; i < paths.length; i++){
		objs[i] = ((DefaultMutableTreeNode)paths[i].getPathComponent(paths[i].getPathCount()-1)).
		    getUserObject();
	    }
	    return objs;
	}
    }


    private class POPTreeMouseListener extends MouseAdapter {
	public void mousePressed(MouseEvent e) {
	    TreePath[] selPaths = popTree.getSelectionPaths();
	    if(selPaths != null){
		Object userObj1 = ((DefaultMutableTreeNode)selPaths[0].getPathComponent(selPaths[0].getPathCount()-1)).
		    getUserObject();
		if(userObj1.getClass() == org.mathweb.reasoning.planning.Action.class){
		    actionPanel.resetTree((org.mathweb.reasoning.planning.Action)userObj1);
		}
	    }
	}
    }
}

