package org.mathweb.reasoning.partialorderplanner;
import org.mathweb.reasoning.search.Node;
/**
 * A search node specialized to hold POP data.
 *
 * @author Harlan Crystal, Daniel Terry
 * @date December 6, 2003
 */
public class POPNode extends Node {
    /** The partial order plan associated with this node */
    private final PartialOrderPlan plan;

    /**
     * Construct a new popnode.
     *
     * @param depth The depth of this node in the search tree.
     * @param plan The partial order plan associated with this search node.
     */
    public POPNode(int depth, PartialOrderPlan plan) {
        super(depth);
        this.plan = plan;
    }

    /**
     * @return The partial order plan associated with this node.
     */
    public PartialOrderPlan getPlan() {
        return plan;
    }

    public int hashCode() {
        return super.hashCode() ^ plan.hashCode();
    }

    public boolean equals(POPNode p) {
        return super.equals(p) && plan.equals(p.plan);
    }

    public boolean equals(Object o) {
        try {
            return equals((POPNode) o);
        } catch (ClassCastException e) {
            return false;
        }
    }

    public String toString() {
        return super.toString() + ", plan: " + plan;
    }
}
