package org.mathweb.reasoning.partialorderplanner;

import java.util.ArrayList;
import java.util.Hashtable;

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.tree.*;

import salvo.jesus.graph.*;
import salvo.jesus.graph.visual.*;
import salvo.jesus.graph.visual.layout.*;

import org.mathweb.reasoning.parser.*;
import org.mathweb.reasoning.planning.*;

public class POPTest extends JFrame {
  PartialOrderPlanner planner;
  Vertex        v1, v2, v3, v4, v5, v6, v7;
  Vertex        v8, v9, v10, v11, v12, v13, v14;
  Vertex        v15, v16, v17, v18, v19;
  Graph         graph;
  VisualGraph	  vGraph;
  GraphEditor	  gEdit;
  ForceDirectedLayout  layout;
  
  public POPTest(){
    try{
      graph = new GraphImpl();
      vGraph = new VisualGraph();
      gEdit = new GraphEditor();

      vGraph.setGraph( graph );
      gEdit.setVisualGraph( vGraph );
      
      this.getContentPane().setLayout( new GridLayout(1,2));
      this.getContentPane().add( gEdit );
      
      layout = new ForceDirectedLayout( vGraph );
      vGraph.setGraphLayoutManager( layout );
      
      this.addWindowListener(new WindowAdapter() {
	  public void windowClosing(WindowEvent e) { System.exit(0); }
      });
	
      Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
      Dimension frameSize = new Dimension( screenSize.width - 80, screenSize.height - 80 );
      
      this.setSize( frameSize );
      this.setLocation((int)(screenSize.getWidth() - frameSize.getWidth()) / 2, (int)(screenSize.getHeight() - frameSize.getHeight()) / 2);
      
      //    getContentPane().setLayout(new BorderLayout());
    } catch(Exception e){
      System.out.println(e);
    }
  }
  
  public POPTest(String filename) {
    loadPlanner(filename);
    try{
      graph = new GraphImpl();
      vGraph = new VisualGraph();
      gEdit = new GraphEditor();

      vGraph.setGraph( graph );
      gEdit.setVisualGraph( vGraph );
      
      this.getContentPane().setLayout( new GridLayout(1,2));
      this.getContentPane().add( gEdit );
      
      layout = new ForceDirectedLayout( vGraph );
      vGraph.setGraphLayoutManager( layout );
      
      this.addWindowListener(new WindowAdapter() {
	  public void windowClosing(WindowEvent e) { System.exit(0); }
      });
	
      Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
      Dimension frameSize = new Dimension( screenSize.width - 80, screenSize.height - 80 );
      
      this.setSize( frameSize );
      this.setLocation((int)(screenSize.getWidth() - frameSize.getWidth()) / 2, (int)(screenSize.getHeight() - frameSize.getHeight()) / 2);
      
      //    getContentPane().setLayout(new BorderLayout());
    } catch(Exception e){
      System.out.println(e);
    }
  }
  
  
  private void loadPlanner(String filename) {
    Parser pp;
    try{
      System.out.println("loading planning problem...");
      pp = Parser.loadFromFile(filename);
      System.out.println("...done");
      planner = new PartialOrderPlanner(new LogicSystem(pp.getActions()), pp.getInitialState(), pp.getGoalState());
      System.out.println("planner: " + planner);
    } catch(Exception e){
      System.out.println(e);
    }
  }
  
  public PartialOrderPlan generatePlan(){
    return planner.generatePlan();
  }
  
  //     public void resetActionTree(planning.Action action){

  //     }
  public void draw(Graph g) throws Exception
  {
    this.setTitle( "Graph output" );
    this.setVisible(true);
    vGraph.setGraph(g);
    gEdit.setVisualGraph( vGraph );
    System.out.println("graph:" + g);
    
    this.getContentPane().setLayout( new GridLayout(1,2));
    this.getContentPane().add( gEdit );
    
    OrthogonalLineLayout layout = new OrthogonalLineLayout( vGraph );
    //StraightLineLayout layout = new StraightLineLayout( vGraph );  
    //LayeredTreeLayout layout = new LayeredTreeLayout( vGraph );
    //ForceDirectedLayout  layout = new ForceDirectedLayout(vGraph);
    vGraph.setGraphLayoutManager( layout );
    
    this.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) { System.exit(0); }
      });
    
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    Dimension frameSize = new Dimension( screenSize.width - 80, screenSize.height - 80 );
    
    this.setSize( frameSize );
    this.setLocation((int)(screenSize.getWidth() - frameSize.getWidth()) / 2, (int)(screenSize.getHeight() - frameSize.getHeight()) / 2);
    layout.layout();
  }
  
  
  public static void main(String[] args){
    try {
      POPTest pop = new POPTest(args[0]);
      pop.setTitle( "Graph output" );
      pop.setVisible(true);
      PartialOrderPlan plan = pop.generatePlan();
      ArrayList alist = plan.linearize();
      Graph g = plan.toGraph(new Hashtable());
      pop.draw(g);
    }
    catch(Exception e){
      e.printStackTrace();
    }
    
  }
  
}

      
//     v1 = new VertexImpl( "1" );
//     v2 = new VertexImpl( "2" );
//     v3 = new VertexImpl( "3" );
//     v4 = new VertexImpl( "4" );
//     v5 = new VertexImpl( "5" );
//     v6 = new VertexImpl( "6" );
//     v7 = new VertexImpl( "7" );
    
//     v8 = new VertexImpl( "8" );
//     v9 = new VertexImpl( "9" );
//     v10 = new VertexImpl( "10" );
//     v11 = new VertexImpl( "11" );
//     v12 = new VertexImpl( "12" );
//     v13 = new VertexImpl( "13" );
//     v14 = new VertexImpl( "14" );

//     v15 = new VertexImpl( "15" );
//     v16 = new VertexImpl( "16" );
//     v17 = new VertexImpl( "17" );
//     v18 = new VertexImpl( "18" );
//     v19 = new VertexImpl( "19" );

//     graph.add( v1 );
//     graph.add( v2 );
//     graph.add( v3 );
//     graph.add( v4 );
//     graph.add( v5 );
//     graph.add( v6 );
//     graph.add( v7 );

//     graph.add( v8 );
//     graph.add( v9 );
//     graph.add( v10 );
//     graph.add( v11 );
//     graph.add( v12 );
//     graph.add( v13 );
//     graph.add( v14 );

//     graph.add( v15 );
//     graph.add( v16 );
//     graph.add( v17 );
//     graph.add( v18 );
//     graph.add( v19 );

//     graph.addEdge( v1, v2 );
//     graph.addEdge( v2, v3 );
//     graph.addEdge( v3, v4 );
//     graph.addEdge( v4, v5 );
//     graph.addEdge( v5, v6 );
//     graph.addEdge( v6, v7 );
//     graph.addEdge( v7, v1 );

//     graph.addEdge( v1, v8 );
//     graph.addEdge( v8, v9 );
//     graph.addEdge( v9, v10 );
//     graph.addEdge( v10, v11 );
//     graph.addEdge( v11, v12 );
//     graph.addEdge( v12, v13 );
//     graph.addEdge( v13, v1 );

//     graph.addEdge( v1, v14 );
//     graph.addEdge( v14, v15 );
//     graph.addEdge( v15, v16 );
//     graph.addEdge( v16, v17 );
//     graph.addEdge( v17, v18 );
//     graph.addEdge( v18, v19 );
//     graph.addEdge( v19, v1 );

