package org.mathweb.reasoning.partialorderplanner;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
import org.mathweb.reasoning.parser.Parser;

/**
 * Runs the partial order planner on a test file.
 *
 * @author Daniel Terry, Harlan Crystal
 */
public class POPTests {
    PartialOrderPlanner planner;
    LogicSystem system;
    Parser pp;

    public POPTests(String filename) {
	try{
	    pp = Parser.loadFromFile(filename);
	    system = new LogicSystem(pp.getActions());
	    planner = new PartialOrderPlanner(system, pp.getInitialState(), pp.getGoalState());
	} catch(Exception e){
	    System.out.println(e);
	}
    }

    public void test() {
        System.out.println("INITIAL:\n" + pp.getInitialState());
        System.out.println("GOAL:\n" + pp.getGoalState());
	System.out.println("\n<<//==-----Begin-Planning-----==//>>");
        PartialOrderPlan plan = planner.generatePlan();
        System.out.println("PLAN:");
        System.out.println(plan);
    }

    /**
     * Runs the partial order planner on given test file.
     *
     * @param args The first command line argument to the program should be the filename of the testfile.
     */
    public static void main(String [] args) {
	try{
	    System.out.println("POP Planner test");
	    System.out.println("---------------");
	    POPTests test = new POPTests(args[0]);
            test.test();
	} catch (Exception e){
            System.err.println(e.getMessage());
            e.printStackTrace();
        }
    }
}
