package org.mathweb.reasoning.partialorderplanner;
import javax.swing.*;
import javax.swing.tree.*;
import java.util.*;
import java.awt.*;
import javax.swing.border.*;

public class POPTree extends DTTree {
    public final PartialOrderPlan plan1;
    private final PartialOrderPlan plan2;
    private boolean showStartFinOrders = false;
    private boolean showStartFinActions = true;

    public POPTree (PartialOrderPlan plan1){
	super();
	this.plan1 = plan1;
	this.plan2 = null;
	setBorder(new LineBorder(Color.BLUE, 3));
	rootNode.setUserObject("Plan");
	createNodes();
    }

    public POPTree (PartialOrderPlan plan1, PartialOrderPlan plan2){
	super();
	this.plan1 = plan1;
	this.plan2 = plan2;
	setBorder(new LineBorder(Color.BLUE, 3));
	rootNode.setUserObject("Plan");
	createNodes();
	initCellRenderer();
    }

    private void initCellRenderer(){
	setCellRenderer(new POPTreeCellRenderer(plan1, plan2));
    }

    private Set removeStartFinOrders(Set orders){
	HashSet ret = new HashSet();
	for(Iterator it = orders.iterator(); it.hasNext(); ){
	    OrderConstraint cur = (OrderConstraint)it.next();
	    if(!cur.getEarly().getTask().getName().equals("Start") && !cur.getLate().getTask().getName().equals("Finish"))
		ret.add(cur);
	}
	return ret;
    }

    private Set removeStartFinActions(Set actions){
	HashSet ret = new HashSet();
	for(Iterator it = actions.iterator(); it.hasNext(); ){
	    org.mathweb.reasoning.planning.Action cur = (org.mathweb.reasoning.planning.Action)it.next();
	    if(!cur.getTask().getName().equals("Start") && !cur.getTask().getName().equals("Finish"))
		ret.add(cur);
	}
	return ret;
    }


    /**
     * @return the union of the two sets where two items are checked for equality using equalsByActionName
     **/
//     private Set unionByActionName(Set set1, Set set2){
// 	for(Iterator it1 = set1.keySet().iterator(); it1.hasNext()){
// 	    for(Iterator it2 = set2.keySet().iterator(); it2.hasNext(); ){
// 		Object
// 	    }
// 	}

//     }

    private void createNodes() {
	DefaultMutableTreeNode actions = new DefaultMutableTreeNode("Actions");
	addNodeAndExpand(rootNode, actions);
	Set allActions = plan1.getActions();
	if(null != plan2)
	    allActions.addAll(plan2.getActions());
	if(!showStartFinActions)
	    allActions = removeStartFinActions(allActions);
	addAllChildren(allActions, actions);

	DefaultMutableTreeNode orderings = new DefaultMutableTreeNode("Ordering Constraints");
	addNodeAndExpand(rootNode, orderings);
	Set allOrders = plan1.getOrderingConstraints();
	if(null != plan2)
	    allOrders.addAll(plan2.getOrderingConstraints());
	if(!showStartFinOrders)
	    allOrders = removeStartFinOrders(allOrders);
	addAllChildren(allOrders, orderings);

	DefaultMutableTreeNode links = new DefaultMutableTreeNode("Causal Links");
	addNodeAndExpand(rootNode, links);
	Set allLinks = plan1.getCausalLinks();
	if(null != plan2)
	    allLinks.addAll(plan2.getCausalLinks());
	addAllChildren(allLinks, links);

	DefaultMutableTreeNode bcs = new DefaultMutableTreeNode("Binding Constraints");
	addNodeAndExpand(rootNode, bcs);
	Set allBcs = plan1.getBindingConstraints().getSet();
	if(null != plan2)
	    allBcs.addAll(plan2.getBindingConstraints().getSet());
	addAllChildren(allBcs, bcs);

	DefaultMutableTreeNode conditions = new DefaultMutableTreeNode("Open Conditions");
	addNodeAndExpand(rootNode, conditions);
	Set allConditions = plan1.getOpenPreconditions();
	if(null != plan2)
	    allConditions.addAll(plan2.getOpenPreconditions());
	addAllChildren(allConditions, conditions);
    }
}
