package org.mathweb.reasoning.partialorderplanner;
import javax.swing.tree.*;
import javax.swing.*;
import java.awt.*;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;

public class POPTreeCellRenderer extends DefaultTreeCellRenderer {
    PartialOrderPlan plan1;
    PartialOrderPlan plan2;

    public POPTreeCellRenderer(PartialOrderPlan plan1) {
	super();
	this.plan1 = plan1;
	this.plan2 = null;
    }

    public POPTreeCellRenderer(PartialOrderPlan plan1, PartialOrderPlan plan2) {
	super();
	this.plan1 = plan1;
	this.plan2 = plan2;
    }

    public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, 
					   boolean leaf, int row, boolean hasFocus) {
	Component c = super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);

	if(null != plan2){
	    Object userObj = ((DefaultMutableTreeNode)value).getUserObject();
	    if(userObj.getClass() == org.mathweb.reasoning.planning.Action.class){
		org.mathweb.reasoning.planning.Action userAction = (org.mathweb.reasoning.planning.Action)userObj;
		if(plan1.hasAction(userAction) && !plan2.hasAction(userAction))
		    c.setForeground(Color.RED);
		else if(!plan1.hasAction(userAction) && plan2.hasAction(userAction))
		    c.setForeground(Color.GREEN);
	    }
	    else if(userObj.getClass() == OrderConstraint.class){
		OrderConstraint userOC = (OrderConstraint)userObj;
		if(plan1.hasOrderConstraint(userOC) && !plan2.hasOrderConstraint(userOC))
		    c.setForeground(Color.RED);
		else if(!plan1.hasOrderConstraint(userOC) && plan2.hasOrderConstraint(userOC))
		    c.setForeground(Color.GREEN);
	    }
	    else if(userObj.getClass() == CausalLink.class){
		CausalLink userCL = (CausalLink)userObj;
		if(plan1.hasCausalLink(userCL) && !plan2.hasCausalLink(userCL))
		    c.setForeground(Color.RED);
		else if(!plan1.hasCausalLink(userCL) && plan2.hasCausalLink(userCL))
		    c.setForeground(Color.GREEN);
	    }
	    else if(userObj.getClass() == BindingConstraint.class){
		System.out.println("HIBC");
		BindingConstraint userBC = (BindingConstraint)userObj;
		System.out.println("p1: " + plan1.hasBindingConstraint(userBC));
		System.out.println("p2: " + plan2.hasBindingConstraint(userBC));

		if(plan1.hasBindingConstraint(userBC) && !plan2.hasBindingConstraint(userBC))
		    c.setForeground(Color.RED);
		else if(!plan1.hasBindingConstraint(userBC) && plan2.hasBindingConstraint(userBC))
		    c.setForeground(Color.GREEN);
	    }
	    else if(userObj.getClass() == OpenCondition.class){
		OpenCondition userOC = (OpenCondition)userObj;
		if(plan1.hasOpenCondition(userOC) && !plan2.hasOpenCondition(userOC))
		    c.setForeground(Color.RED);
		else if(!plan1.hasOpenCondition(userOC) && plan2.hasOpenCondition(userOC))
		    c.setForeground(Color.GREEN);
	    }
	}
	return c;
    }
}
