package org.mathweb.reasoning.planning;
import org.mathweb.reasoning.logic.Literal;
import org.mathweb.reasoning.logic.BindingTable;
import org.mathweb.reasoning.logic.BindingConstraints;
/**
 * Represents a Bound or Instantiated Action.  This is an action
 * in which there are no variables.
 *
 * @author Harlan Crystal, Daniel Terry
 * @date November 5, 2003
 */
public class BoundAction extends Action {
    public BoundAction(Task task, Literal [] preconditions, 
                       Literal [] postconditions) {
        super(task, preconditions, postconditions, new BindingConstraints());
        
        if (!isBound())
            throw new IllegalArgumentException("BoundAction class with variables");
    }

    /**
     * Creates a bound action from an action.
     *
     * @param a The action to copy.
     * @exception IllegalArgumentException thrown if a is not bound.
     */
    public BoundAction(Action a) {
        super(a.getTask(), a.preconditions, a.postconditions, new BindingConstraints());
        if (!a.isBound())
            throw new IllegalArgumentException("Can not craete a BoundAction from an Action with variables in it");
    }

    public boolean isBound() {
        return true;
    }

    /**
     * @throws IllegalArgumentException You can not apply bindings to a bound
     * action.
     */
    public Action applyBindings(BindingTable bindings) {
        throw new IllegalArgumentException("Can not apply bindings to a bound action");
    }

    /**
     * This function only works if this action has no variables in it, which
     * can be achieved by calling applyBindings
     *
     * @return a new State that is the result of applying this action's postconditions to the given state
     */
    public State applyAction(State s){
	State newState = s;

	//Loop through each post condition literal
 	for(int post = 0; post < postconditions.length; post++){
 	    Literal curPost = postconditions[post];

            if(curPost.isTrue())
		newState = newState.addLiteral(curPost);
	    else
		newState = newState.removeLiteral(curPost);
 	}
 	return newState;
    }
    


}
