package org.mathweb.reasoning.planning;
import org.mathweb.reasoning.logic.*;
import java.util.*;
import java.io.Serializable;

/**
 * Represents a plan, which is a sequence of actions.
 *
 * An immutable object.
 *
 * @author Harlan Crystal, Daniel Terry
 * @date October 28, 2003
 */
public class Plan implements Serializable {
    private Vector list;
    
    /**
     * Constructs an empty plan.
     */
    public Plan() {
        list = new Vector();
    }

    /**
     * Copy constructor.
     */
    public Plan(Plan copyme) {
        this.list = (Vector) copyme.list.clone();
    }

    /**
     * Adds an action to the end of the list of items in the plan.
     *
     * @param action The action to add.
     * @return A copy of this plan with the action added to the end.
     */
    public Plan addActionToEnd(Action action) {
        Plan newPlan = new Plan(this);
        newPlan.list.add(action);
        return newPlan;
    }

    /**
     * Adds an action to the beginning of the list of actions in the plan.
     *
     * @param action The action to add.
     * @return A copy of this plan with the action added to the front.
     */
    public Plan addActionToFront(Action action) {
        Plan newPlan = new Plan();
        newPlan.list.add(action);
        newPlan.list.addAll(this.list);
        return newPlan;
    }

    /**
     * @return An iterator of the actions in this plan.
     */
    public Iterator iterator() {
        return list.iterator();
    }

    public String toString() {
        String ret = "";
        for (Iterator it = list.iterator(); it.hasNext(); ) {
            Action a = (Action) it.next();
            //ret += "\t\t" + a.getTask() + "\n";
	    ret += a + "\n";
        }
        
        return ret;
    }
}
