package org.mathweb.reasoning.search;
import java.util.*;
/**
 * A special purpose fringe for doing iterative deepening.
 *
 * @author Harlan Crystal, Daniel Terry
 * @date November 22, 2003
 */
public class IterativeDeepeningStack implements Fringe {
    private Vector stack; /* the current stack of fringe nodes */
    private int currentDepthLimit; /* the current depth limit */
    private boolean hasHitDepthLimit; /* is true if we have already hit the depth limit */
    private Node root; /* The root of the tree we're doing iterative deepening on */
    private int numDropped;
    private int numSearched;

    public IterativeDeepeningStack() {
        stack = new Vector();
        currentDepthLimit = 1;
        hasHitDepthLimit = false;
        root = null;
        numDropped = 0;
        numSearched = 0;
    }

    public void addAllToFringe(Collection nodes) {
        for (Iterator it=nodes.iterator(); it.hasNext(); )
            addToFringe((Node) it.next());
    }

    public void addToFringe(Node nextNode) {
        if (root == null) {
            root = nextNode;
            //System.out.println("Root is: " + root.alphaHashCode());
        }

        if (nextNode.getDepth() > currentDepthLimit) {
            hasHitDepthLimit = true;
            numDropped++;
            //System.out.println("Dropped with depth " + nextNode.getDepth()+ ": " + nextNode.alphaHashCode());
        } else {
            stack.add(nextNode);
            //System.out.println("Add " +stack.size()+" : " + getDebugString());
        }
    }

    public Node getNext() {
        if (stack.isEmpty() && !hasHitDepthLimit) {
            return null;
        }
        
        if (stack.isEmpty() && hasHitDepthLimit) {
            hasHitDepthLimit = false;
            currentDepthLimit++;
            System.out.println("Restarting search at increased depth of " + currentDepthLimit + ". Dropped " + numDropped + ". Searched: " + numSearched);
            numSearched = 0;
            numDropped = 0;
            return root;
        }
        numSearched++;
        Node next = (Node) stack.get(stack.size()-1);
        stack.remove(stack.size()-1);
        //System.out.println("Rm " +stack.size()+": " + getDebugString());
        //Node next = (Node) stack.get(0);
        //stack.remove(0);
        return next;
    }

    public boolean hasNext() {
        return hasHitDepthLimit || !stack.isEmpty();
    }
    
    public int size() {
        return stack.size();
    }

    public String toString() {
        String ret = "{";
        for (Iterator it = stack.iterator(); it.hasNext(); )
            ret += it.next();
        return ret +"}";
    }
}
