// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.serv;

//import nl.tue.win.riaca.openmath.lang.OMObject;

import java.util.ArrayList;
import java.util.List;

import org.mathweb.serv.*;
import org.mathweb.reasoning.logic.*;

/**
 */
public class PreCondition {
  
  public String name;
  public String docu;
  public CachedHashCodeObject condition;

  public PreCondition ()
  {
    name = "";
    docu = "";
    condition = null;
  }
  
  public PreCondition (String doc, CachedHashCodeObject obj){
    docu = doc;
    condition = obj;
  }
  
  public CachedHashCodeObject getCondition()
  {
    return condition;
  }
  
  public void setCondition(CachedHashCodeObject obj)
  {
    condition = obj;
  }
 
  public void setName(String n)
  {
    this.name =n;
  }

public String getName()
  {
    return name;
  }
    
  public String toString()
  {
    return  name + ": " +  condition.toString();
  }

  public List toPlanningLiterals(List inputNames){
    ArrayList ls = new ArrayList();
    if (condition instanceof Literal) {
      Literal lit = (Literal) condition;
      // check whether the both arguments of the literal are variables and whether they 
      // are the inputs of the service. If so, then add it to the list of
      // literals. Otherwise omit it.
      Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
      String an0, an1;
      if ((args[0] instanceof Variable) && (args[1] instanceof Variable)){
        String varName0 = args[0].getName().toLowerCase();
        String varName1 = args[1].getName().toLowerCase();
        //        if ((inputNames.contains(varName0)) && inputNames.contains(varName1)){
          ls.add(lit);
          //        }
      }
    }
    else if (condition instanceof Conjunct){
      Conjunct conj = (Conjunct) condition;
      List lits = conj.getLiterals();
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        // check whether the both arguments of the literal are variables and whether they 
        // are the inputs of the service. If so, then add it to the list of
        // literals. Otherwise omit it.
        Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
        String an0, an1;
        if ((args[0] instanceof Variable) && (args[1] instanceof Variable)){
          String varName0 = args[0].getName().toLowerCase();
          String varName1 = args[1].getName().toLowerCase();
          //          if ((inputNames.contains(varName0)) && inputNames.contains(varName1)){
            ls.add(lit);
            //          }
        }
      }
    }
    return ls;
  }
  
} // end PreCondition

  
    

