// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.serv;

import java.net.URI;
import java.net.URL;
import java.util.List;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Vector;
import java.util.Enumeration;

import org.mathweb.serv.*;
import org.mathweb.reasoning.logic.*;
import org.mathweb.reasoning.planning.*;
//import org.apache.axis.client.Service;
import javax.xml.namespace.QName;

import org.mindswap.owls.OWLSFactory;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.process.execution.ProcessExecutionListener;
import org.mindswap.owls.process.Process;
import org.mindswap.owls.process.AtomicProcess;
//import org.mindswap.owls.process.Input;
import org.mindswap.owls.process.InputList;
import org.mindswap.owls.process.OutputList;
import org.mindswap.owls.process.ConditionList;
import org.mindswap.owls.process.Condition;
import org.mindswap.query.ValueMap;
import org.mindswap.owls.process.Parameter;
import org.mindswap.owls.profile.Profile;
import org.mindswap.owls.profile.ServiceParameter;
import org.mindswap.owls.profile.ServiceCategory;
import org.mindswap.owls.service.Service;
import org.mindswap.owls.grounding.Grounding;
import org.mindswap.owls.grounding.AtomicGrounding;
import org.mindswap.owls.grounding.WSDLAtomicGrounding;
import org.mindswap.owls.process.execution.ProcessExecutionEngine;
import org.mindswap.wsdl.WSDLOperation;
import org.mindswap.wsdl.WSDLParameter;
import org.mindswap.wsdl.WSDLService;

/**

 */
public class ReasoningService {
  
  public String name;
  public String uri;
  
  public ReasoningProblem problem;
  public TaxonomyEntry taxonomy ;
  public String directive;
  private WSDLService wsdlService;
  private String operation;
  private Service owlsService = null;

  private ArrayList paramNames;
  private ArrayList newVars;
  public static int maxObjs = 1;
  
  private static URI serviceParameter = URI.create("http://www.daml.org/services/owl-s/1.0DL/Profile.owl#serviceParameter");

  private static URI onEffect = URI.create("http://www.mathweb.org/owl/mathweb.owl#onEffect");
  private static URI whenCondURI = URI.create("http://www.mathweb.org/owl/mathweb.owl#whenCondition");
  private static URI propPred = URI.create("http://www.w3.org/2003/11/swrl#propertyPredicate");
  private static URI arg1 = URI.create("http://www.w3.org/2003/11/swrl#argument1");
  private static URI arg2 = URI.create("http://www.w3.org/2003/11/swrl#argument2");

  //  private static boolean fake = true;
  private static boolean fake = false;

  //  public Service wsdlOperation;
    
  public String getName()  { return name; }

  public String getURI(){
    return uri;}
  public void setURI(String newuri){
    uri = newuri;}


  public ReasoningProblem getProblem()
  {
    return problem;
  }
  public String getDirective()
  {
    return directive;
  }

  public WSDLService getWsdlService()
  {
    return wsdlService;
  }
  
  public void trim()
  {
    name = name.trim();
    directive = directive.trim();
    problem.trim();
    taxonomy.trim();
  }
    

  public void setName(String name)
  {
    this.name = name;
  }

  public void setOwlsService(Service serv)
  {
    this.owlsService = serv;
  }

  public Service getOwlsService()
  {
    return this.owlsService;
  }

  public void setWsdlGrounding(String wsdlFileName, String operation)
  {
    try{
      wsdlService = WSDLService.createService("file:" + wsdlFileName);
      this.operation = operation;
    } catch(Exception e){
      System.out.println(e);
      e.printStackTrace();
      return;
    }
    
  }
  
  public ValueMap invokeOWLS(Hashtable map) throws Exception
  {
    ProcessExecutionEngine exec = OWLSFactory.createExecutionEngine();
//     exec.addExecutionListener(new ProcessExecutionListener() {
//         String TAB = "   ";
//         String indent = "   ";
         
//         public void setCurrentExecuteService(Process p) {
//           System.out.println(indent + "Start executing process " + p);
//           indent += TAB;
//         }
        
//         public void printMessage(String message) {
//           if(message.equals("[DONE]"))
//             indent = indent.substring(0, indent.length()-TAB.length());
//           System.out.println(indent + message);
//         }

//         public void finishExecution(int retCode) {
//           System.out.println("Finished execution " + 
//                              ((retCode == ProcessExecutionListener.EXEC_ERROR)
//                               ? "with errors" : "successfully"));				
//         }			
//       });		

    AtomicProcess process = (AtomicProcess) owlsService.getProcess();
    //    System.out.println("process: " + process);
    WSDLAtomicGrounding atomic = (WSDLAtomicGrounding) process.getGrounding();
//     System.out.println("grounding: " + atomic);
//     System.out.println("grounding: " +atomic.getClass().getName());
//     System.out.println("op: " + atomic.getOperation());
    Hashtable results = new Hashtable();
//     List groundings = ground.getAtomicGroundings();
//     if (groundings.isEmpty())
//       return results;
//     AtomicGrounding atomic = (AtomicGrounding) groundings.get(0);
//     Process process = owlsService.getProcess();
// new for 1.1.0 
    ValueMap vMap = new ValueMap();
    //    ValueMap vMap = OWLSFactory.createValueMap();
    
    for(Enumeration e= map.keys(); e.hasMoreElements(); ){
      String iName = (String) e.nextElement();
      
      //      System.out.println("input: " +process.getInputs().getParameter(iName) +" --> " + map.get(iName));
      
      //    URI uri = owlsService.getURI();
//       URI paramUri = new URI(uri.getScheme(),
//                              uri.getSchemeSpecificPart(),
//                              iName);
      //      System.out.println("param: "+  paramUri);
      //       Parameter in = process.getParameter(paramUri);
      //       System.out.println("mapping: " + in + " -> " + map.get(iName));
      //       vMap.setValue(in, map.get(iName));
      
// new for 1.1.0 
      vMap.setDataValue(process.getInputs().getParameter(iName), ((String) map.get(iName)));		
//      vMap.setValue(process.getInputs().getParameter(iName), map.get(iName));		
    }
    ValueMap resMap= null;
    try {
       resMap = exec.execute(process, vMap);	
    } catch(Exception e){
      System.out.println(e);
      e.printStackTrace();
    }
    //    System.out.println("resultMap: " + resMap);
    return resMap;
  }
  
  public ValueMap invoke(Hashtable map) throws Exception
  {
    if (owlsService != null)
      return invokeOWLS(map);
// new for 1.1.0 
    ValueMap vMap = new ValueMap();
//    ValueMap vMap = OWLSFactory.createValueMap();
    
    List ops = wsdlService.getOperations();
    //    System.out.println("ops: " + ops);
    
    Hashtable results = new Hashtable();
    if (fake){
      List outputs = problem.getOutputs();
      for(int i = 0; i < outputs.size(); i++) {		
        Output servOut = (Output) outputs.get(i);
        results.put(servOut.getSignature(), "fake");		
      }
    }
    else{
      WSDLOperation op = wsdlService.getOperation(operation);
      if(op == null) op = wsdlService.getOperation(getLocalName(operation));
      
      for(int i = 0; i < op.getInputs().size(); i++) {		
        WSDLParameter opIn = op.getInput(i);
        Input servIn = problem.getInput(i);
        Object value = map.get(servIn.getName());
        opIn.setValue(value);
      }
      op.invoke();
      
      for(int i = 0; i < op.getOutputs().size(); i++) {		
        WSDLParameter opOut = op.getOutput(i);
        String outputValue = opOut.getValue().toString();
        Output servOut = problem.getOutput(i);
        results.put(servOut.getSignature(), outputValue);		
      }
    }
    return vMap ;
  }

  public static String getLocalName(URI uri) {
    return getLocalName(uri.toString());
  }
  public static String getLocalName(String uri) {
    int index = splitPos(uri);
    
    if(index == -1) return null;
    
    return uri.substring(index + 1);
  }
  private static int splitPos(String uri) {
    int pos = uri.indexOf("#");
    
    if(pos == -1) pos = uri.lastIndexOf("/");
    
    return pos; 
  }
  
  public void setProblem(ReasoningProblem prob)
  {
    problem = prob;
  }

  public void shutdown()
  {
    try{
      List ops = wsdlService.getOperations();
      //      System.out.println("ops: " + ops);
      WSDLOperation op = wsdlService.getOperation("shutdown");
      op.invoke();
    }catch(Exception e){
      System.out.println("shutdown error: " + e);
      e.printStackTrace();
    }
  }
  
  

  /** transforms services into planning operators, such as:  
action {
  tramp
  pre {
    type(prob1, FOFPROBLEM),
    type(cnf, CNFPROBLEM),
    type(proof1, FORESOLUTIONPROOF),
    constraint(EQUIVALENTTO, cnf, prob1),
    constraint(FOR, proof1, cnf)
  }
  post {
    type(proof2, NDPROOF),
    constraint(FOR, proof2, prob1) 
  }
  constraints {
    prob1 != cnf
  }
}
  */

  public Action toPlanningOperator()
  {
    List inputs = problem.getInputs();
    List outputs = problem.getOutputs();
    List preconds= problem.getPreconds();
    List postconds= problem.getPostconds(); 

    List inputNames = new ArrayList();
    
    ArrayList ls = new ArrayList(); // list of literals
    for(int i=0; i<inputs.size(); i++){
      //      System.out.println("in: " + inputs.get(i).getClass().getName());
      
      Input e = (Input) inputs.get(i);
      Predicate p = new Predicate("type");
      Argument[] args = { new org.mathweb.reasoning.logic.Variable(e.getName()),
                          new org.mathweb.reasoning.logic.Constant(e.getSignature())};
      ls.add(new org.mathweb.reasoning.logic.Literal(p, args, true));
      inputNames.add(e.getName().toLowerCase());
    }

    for(int i=0; i<preconds.size(); i++){
      //      System.out.println("pre: " + preconds.get(i).getClass().getName());
      PreCondition pre = (PreCondition) preconds.get(i);
      List l = pre.toPlanningLiterals(inputNames);
      ls.addAll(l);
    }
    org.mathweb.reasoning.logic.Literal[] pre= new org.mathweb.reasoning.logic.Literal[ls.size()];
    Object[] ps = ls.toArray();
    for(int i=0; i<ps.length; i++){
      //      System.out.println("lit: " + ps[i].getClass().getName());
      pre[i] = (org.mathweb.reasoning.logic.Literal) ps[i];
    }
    

    ls = new ArrayList();
    for(int i=0; i<outputs.size(); i++){
      //      System.out.println("out: " + outputs.get(i).getClass().getName());
      
      Output e = (Output) outputs.get(i);
      Predicate p = new Predicate("type");
      Argument[] args = { new org.mathweb.reasoning.logic.Variable(e.getName()),
                          new org.mathweb.reasoning.logic.Constant(e.getSignature())};
      ls.add(new org.mathweb.reasoning.logic.Literal(p, args, true));
    }
    for(int i=0; i<postconds.size(); i++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      PostCondition po = (PostCondition) postconds.get(i);
      List l = po.toPlanningLiterals(inputNames);
      ls.addAll(l);
    }
    org.mathweb.reasoning.logic.Literal[] post = new org.mathweb.reasoning.logic.Literal[ls.size()];
    ps = ls.toArray();
    for(int i=0; i<ps.length; i++){
      //      System.out.println("lit: " + ps[i].getClass().getName());
      post[i] = (org.mathweb.reasoning.logic.Literal) ps[i];
    }
    return new Action(new Task(name), pre, post, new BindingConstraints());
  }
  
  private String getLitString(Literal lit, List paramNames){
    // check whether the both arguments of the literal are variables and whether they 
    // are the inputs of the service. If so, then add it to the list of
    // literals. Otherwise omit it.
    Argument[] args = { lit.getArgument(0), lit.getArgument(1)};
    String an0, an1;
    if ((args[0] instanceof Variable) && (args[1] instanceof Variable)){
      String varName0 = args[0].getName().toLowerCase();
      String varName1 = args[1].getName().toLowerCase();
      String type = null; 
      System.out.println("params: " + varName0 + " " + varName1 + " " + paramNames);
     if (!paramNames.contains(varName0)) {
        Variable v0 = (Variable) args[0];
        type = v0.getType();
        if (!newVars.contains(varName0))
          newVars.add(varName0);
      }
      if (!paramNames.contains(varName1)){
        Variable v1 = (Variable) args[1];
        type = v1.getType();
        if (!newVars.contains(varName1))
          newVars.add(varName1);
      }
      return "(" + lit.getPredicate().getName() + " <"+ varName0 + "> <"+varName1+">)";
    }
    return "";
  }

  private ArrayList getCondStrings(CachedHashCodeObject condition, List paramNames){
    ArrayList condStrs = new ArrayList();
    
    if (condition instanceof Literal) {
      Literal lit = (Literal) condition;
      condStrs.add(getLitString(lit, paramNames));
    }
    else if (condition instanceof Conjunct){
      Conjunct conj = (Conjunct) condition;
      List lits = conj.getLiterals(); 
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        // check whether the both arguments of the literal are variables and whether they 
        // are the inputs of the service. If so, then add it to the list of
        // literals. Otherwise omit it.
        condStrs.add(getLitString(lit, paramNames));
      }
    }
    else if (condition instanceof Implication){
      Implication imp = (Implication) condition;
      List lits = imp.getBody();
      Literal head = imp.getHead();
      String impStr = "(if (and\n";
      for(int i=0; i<lits.size(); i++){
        Literal lit = (Literal) lits.get(i);
        String litStr = getLitString(lit, paramNames);
        impStr += "        " + litStr + "\n";
      }
      impStr += "       )\n       ((add " + getLitString(head, paramNames) + ")))";
      condStrs.add(impStr);
    }
    return condStrs;
  }
  
  public String toProdigyOperator()
  {
    return toProdigyOperator(1);
  }
  

  public String toProdigyOperator(int objectCounter)
  {
    List inputs = problem.getInputs();
    List outputs = problem.getOutputs();
    HashMap locals = problem.getLocals();
    List preconds = problem.getPreconds();
    List postconds= problem.getPostconds(); 
    String params = "";
    String precondsDecl = "";
    String precondsStr = "";
    String effectsDecl = "";
    String effects = "";
    String rejectRule = "";
    String rejectRuleAppend = "";
    String objects = "";
    String opName = getURI().substring(7);
    
    this.paramNames = new ArrayList();
    rejectRule += "(Control-Rule reject-wrong-binding-for-";
    rejectRule += name;
    rejectRule += "\n  (if (and\n        (applicable-operator (";
    rejectRule += opName;
    rejectRuleAppend += "        (or\n ";
    
    /*
      all inputs have to be bound.
     */
    for(int i=0; i<inputs.size(); i++){
      System.out.println("in: " + inputs.get(i).getClass().getName());
      
      Input e = (Input) inputs.get(i);
      params = params + " <"+e.getName() + ">";
      precondsDecl = precondsDecl + "  (<" +e.getName() + "> OWL-" +e.getSignature() + ")\n";
      precondsStr += "  (bound <" +e.getName() + ">)\n";
      rejectRule += " <"+e.getName() + "> ";
      paramNames.add(e.getName().toLowerCase());
    }

    /*
      all output parameters have to be NOT bound before the operator application.
    */
    for(int i=0; i<outputs.size(); i++){
      //      System.out.println("out: " + outputs.get(i).getClass().getName());
      
      Output e = (Output) outputs.get(i);
      String owlType = " OWL-" +e.getSignature();
      params = params + " <"+e.getName() + ">";
      precondsDecl = precondsDecl + "  (<" +e.getName() + ">" + owlType + ")\n";
      precondsStr += "  (~ (bound <" +e.getName() + ">))\n";
      rejectRule += " <"+e.getName() + ">";
      rejectRuleAppend += "        (~ (exact-type-of-object <" + e.getName() + ">" + owlType + "))\n";

      //      effects += "(del (unbound <" +e.getName() + ">))\n";
      for(int x=0; x< maxObjs; x++)
        objects += "(pinstance-of object_"+ e.getName() + "_" + objectCounter++ + owlType + ")\n";
      effects += "   (add (bound <" +e.getName() + ">))\n";
      paramNames.add(e.getName().toLowerCase());
    }
    rejectRule += "))\n";
    rejectRuleAppend += "         )))\n  (then sub-goal))\n";
    
    newVars= new ArrayList();
    
    for(int k=0; k<preconds.size(); k++){
      //      System.out.println("pre: " + preconds.get(i).getClass().getName());
      PreCondition pre = (PreCondition) preconds.get(k);
      CachedHashCodeObject condition = pre.getCondition();
      ArrayList strs = getCondStrings(condition, paramNames);
      for (int s=0; s<strs.size(); s++) {
        String str = (String) strs.get(s);
        precondsStr += "  " + str + "\n";
      }
    }
    /*
      free variables (not bound in precondDecls) have to be introduced. Since the type is not known, we choose OWL-Thing
    */
    for(int l=0; l<newVars.size(); l++){
      String type = "OWL-Thing";
      String name = (String) newVars.get(l);
      if (locals.containsKey(name)){
        type = "OWL-" + ((String) locals.get(name));
      }
      precondsDecl += "  (<" +name + "> " + type + ")\n";
    }
    ArrayList oldVars = new ArrayList(newVars);
    
    newVars= new ArrayList();
    for(int k=0; k<postconds.size(); k++){
      //      System.out.println("post: " + postconds.get(i).getClass().getName());
      PostCondition po = (PostCondition) postconds.get(k);
      CachedHashCodeObject condition = po.getCondition();
      ArrayList strs = getCondStrings(condition, paramNames);
      
      for (int s=0; s<strs.size(); s++) { 
        String str = (String) strs.get(s);
        
        if (condition instanceof Implication){
          effects += "   "+ str + "\n";
        } else{
          effects += "   (add "+ str + ")\n";          
        }
      }
    }

    for(int l=0; l<newVars.size(); l++){
      String name = (String) newVars.get(l);
      if (!oldVars.contains(name))
        effectsDecl += "(<" +name + "> OWL-Thing)\n";
    }
    
    StringBuffer buf = new StringBuffer();
    buf.append("(OPERATOR ");
    buf.append(opName + "\n");
    buf.append("(params " + params + ")\n");
    buf.append("(preconds \n (\n"+precondsDecl+ " )\n (and\n" + precondsStr + " ))\n");
    buf.append("(effects \n ("+effectsDecl+ " )\n (\n" + effects + ")))\n\n");

    buf.append(objects);
    buf.append("\n\n");
    
    buf.append(rejectRule);
    buf.append(rejectRuleAppend);
    
    
    
// (Control-Rule reject-wrong-binding-otter
// 	      (if (and
// 		   (applicable-operator (OtterATP <x> <result>))
// 		   (~ (exact-type-of-object <result> OWL-FOBrFPResult))
// 		   ))
//               (then sub-goal))


    return buf.toString();
  }

  
  
  public String toString()
  {
    String result = "service `" + name + "' with directive "+ directive;
    if (problem != null) { result = result + " solves: \n" + problem.toString(); }
    return result;
  }
  
}


//   public String getProdigyPreferenceRule()
//   {
//     if (owlsService == null) return "";
//     Profile prof = owlsService.getProfile();
//     ServiceParameter param = prof.getServiceParameter(serviceParameter);
//     OWLResource value = param.getParameter();
    
//     OWLResource whenCond = value.getProperty(whenCondURI);
//     System.out.println(whenCond);
//     String oneff = value.getProperty(onEffect).toString();
//     System.out.println(oneff);
//     String whenCondPred = whenCond.getProperty(propPred).toString();
//     String argName1 = whenCond.getProperty(arg1).toString();
//     String argName2 = whenCond.getProperty(arg2).toString();
//     Predicate pred = new Predicate(whenCondPred); 
//     Argument[] args = new Argument[2];
//     args[0] = new Variable(argName1);
//     args[1] = new Variable(argName2);
//     Literal whenCondLit =  new Literal(pred, args, true);
//     String whenCondLitStr = getCondString(whenCondLit, new ArrayList());
    
//     System.out.println("whenCond: " + whenCondLitStr);
    
//     CachedHashCodeObject condition = null;
//     List postconds= problem.getPostconds(); 
//     for(int k=0; k<postconds.size(); k++){
//       //      System.out.println("post: " + postconds.get(i).getClass().getName());
//       PostCondition po = (PostCondition) postconds.get(k);
//       if (po.getName().equals(oneff))
//         condition = po.getCondition();
//     }
//     String condStr = getCondString(condition, new ArrayList());
//     System.out.println("cond: " + condStr);
    
//     StringBuffer rule = new StringBuffer();
//     rule.append("(Control-Rule prefer-");
//     rule.append(this.name);
//     rule.append("\n  (if (and\n    (current-goal ");
//     rule.append(condStr);
//     rule.append(")\n    (true-in-state ");
//     rule.append(whenCondLitStr);
//     rule.append(")\n    ))\n  (then prefer operator ");
//     rule.append(this.name);
//     rule.append(" <anything>))\n");
    
    
//     /*
//       (Control-Rule prefer-ep-1
// 	      (if (and
// 		   (current-goal (resultFor <x> <y>))
// 		   (true-in-state (problem-class <y> sat_epr_fof))
// 		   ))
// 	      (then prefer operator EpATP <anything>))
// */
//       return rule.toString();
//   }
