package org.mathweb.texmacs;

import java.util.*;
import java.text.ParseException;

import edu.cmu.apros.translation.firstorder.FirstOrderTranslator;
import edu.cmu.apros.language.*;
import edu.cmu.apros.theory.Assertion;

public class Apros2Tptp
{

    FirstOrderTranslator translator = new FirstOrderTranslator();

    public String apros2Tptp(Assertion ass)
    {
        StringBuffer tptp = new StringBuffer();

        tptp.append("% =============================================\n");
        tptp.append("% Apos problem:\n");
        tptp.append("% " + ass.toString() + ".\n");
        tptp.append("% =============================================\n");
        for(int i = 0; i< ass.getNumPremises(); i++)            {
            Formula prem = ass.getPremise(i);
            tptp.append("input_formula(premise"+i+", axiom, (\n");
            tptp.append(formula2Tptp(prem));
            tptp.append(")).\n\n");
        }
        Formula goal = ass.getGoal();
        tptp.append("input_formula(conj, conjecture, (\n");
        tptp.append(formula2Tptp(goal));
        tptp.append(")).\n\n");
        return tptp.toString();
    }
    
    private String formula2Tptp(Formula form)
    {
        FormulaType type = form.getType();
        if (form instanceof AtomicFormula){
            AtomicFormula f = (AtomicFormula) form;
            StringBuffer res = new StringBuffer();

            // relation and function symbols must be in lower case 
            String rel = f.getRelationSymbol().toLowerCase();
            int numTerms = f.getNumTerms();

            res.append(rel);
            if (numTerms > 0) res.append("(");
            
            for(int i = 0; i< numTerms; i++)            {
                Term term = f.getTerm(i);
                res.append(term2Tptp(term));
                if (i< numTerms-1) res.append(",");
            }
            if (numTerms > 0) res.append(")");
            return res.toString();
        }
        
        if (form instanceof Disjunction){
            return "(" +formula2Tptp(form.getSubformula(0)) + " | " + formula2Tptp(form.getSubformula(1)) + ")";
        }
        if (form instanceof Conjunction){
            return "(" +formula2Tptp(form.getSubformula(0)) + " & " + formula2Tptp(form.getSubformula(1)) + ")";
        }
        if (form instanceof Implication){
            return "(" +formula2Tptp(form.getSubformula(0)) + " => " + formula2Tptp(form.getSubformula(1)) + ")";
        }
        if (form instanceof Existential){
            return "(? [" + ((Existential)form).getVariable().toString().toUpperCase() + "] : \n " + formula2Tptp(form.getSubformula(0)) + ")";
        }
        if (form instanceof Universal){
            return "(! [" + ((Existential)form).getVariable().toString().toUpperCase() +"] : \n " + formula2Tptp(form.getSubformula(0)) + ")";
        }
        if (form instanceof Negation){
            return "~(" + formula2Tptp(form.getSubformula(0)) + ")";
        }
        return "formula type " + type + " not known. ";
    }
    
    private String term2Tptp(Term term) {
        if (term instanceof Variable){
            // variables must be in upper case 
            return term.getSymbol().toUpperCase();
        }
        if (term instanceof Constant){
            // constants must be in lower case 
            return term.getSymbol().toLowerCase();
        }
        if (term instanceof ComplexTerm){

            ComplexTerm t = (ComplexTerm) term;
            StringBuffer res = new StringBuffer();
            
            // relation and function symbols must be in lower case 
            String fun = t.getSymbol().toLowerCase();
            int numTerms = t.getNumSubterms();

            res.append(fun);
            if (numTerms > 0) res.append("(");
            for(int i = 0; i< numTerms; i++)            {
                Term st = t.getSubterm(i);
                res.append(term2Tptp(st));
                if (i< numTerms-1) res.append(",");
            }
            if (numTerms > 0) res.append(")");
            return res.toString();
        }
        return "Can't translate terms of type " + term.getType() + " right now!";
    }

    private Assertion parseAssertion(String s) {
        Assertion assertion = null;
        try {
            assertion = translator.parseAssertion(s);
        } catch (ParseException e) {
            //unparseable= true;
            e.printStackTrace(); //To change body of catch statement use Options | File Templates.
            //return null;
        }
        return assertion;
}

    public void testtranslate()
    {
//         System.out.println(apros2Tptp(parseAssertion("{}|=P->P")));
        
//         System.out.println(apros2Tptp(parseAssertion("{P}|=P->P")));
//         System.out.println(apros2Tptp(parseAssertion("{P, S}|=P&S")));
//         System.out.println(apros2Tptp(parseAssertion("{}|=(P & Q) -> (Q & P)")));
//         System.out.println(apros2Tptp(parseAssertion("{P}|=Pv.S")));
//         System.out.println(apros2Tptp(parseAssertion("{(E.x)(P -> F(x)), (E.x)(F(x) -> P)} |= (E.x)((P -> F(x))&(F(x)->P))")));
// 	System.out.println(apros2Tptp(parseAssertion("{}|= (A.x)(A.y)Subset(x,t) <-> ((A.z)Member(z,x) -> Member(z,y))")));

	System.out.println(apros2Tptp(parseAssertion("{}|= (A.x)Group(x) -> Member(identity(x),x) & (A.y)(Member(y,x) -> (Equal(multiply(x,y,identity(x)), y) & Equal(multiply(x,identity(x),y),y))))")));
	
// {(A.u)(A.f)(Subset(u,f) <-> (A.x)(Member(x,u)->Member(x,f))), 
//  (A.g)(Group(g) -> (Member(identity(g),g) & (A.x)(Member(x,g) -> (Equal(multiply(g,x,identity(g)), x) & Equal(multiply(g,identity(g),x),x))))),
//  (A.g)(A.x)((Group(g) & Member(x,g))->(Member(inverse(g,x),g) & Equal(multiply(g,x,inverse(g,x)),indentity(g)) & Equal(multiply(g,inverse(g,x),x),identity(g)))), 
// Group(f), 
// Subset(u,f), 
// (A.x)(A.y)(Member(x,u) & Member(y,u)) -> Member(multiply(f,inverse(f,y)),u)}
//  |= (A.v)(Member(v,u) -> Member(inverse(f,v),u))
    }
    

    public static void main(String[] args) {
        System.out.println("Bonjour!!!");
        try {
            Apros2Tptp tts = new Apros2Tptp();
            tts.testtranslate();
        }
        catch (Exception e){
            System.out.println("error on main");
        }
    }

}// end class Apros2Tptp

/*  a valid TPTP problem description:
input_formula(ax46,axiom,(
    ! [U,V,W,X] : 
      ( ( nonreflexive(U,V)
        & agent(U,V,W)
        & patient(U,V,X) )
     => ~ equal(W,X) )   )).

input_formula(co1,conjecture,(
    ~ ( ? [U] : 
          ( actual_world(U)
          & ? [V,W,X,Y,Z,X1] : 
              ( male(U,X)
              & male(U,V)
              & man(U,V)
              & of(U,W,V)
              & cannon(U,W)
              & ! [X2] : 
                  ( member(U,X2,X)
                 => ? [X3] : 
                      ( event(U,X3)
                      & agent(U,X3,V)
                      & patient(U,X3,X2)
                      & present(U,X3)
                      & nonreflexive(U,X3)
                      & fire(U,X3)
                      & from_loc(U,X3,W) ) )
              & six(U,X)
              & group(U,X)
              & ! [X4] : 
                  ( member(U,X4,X)
                 => shot(U,X4) )
              & cry(U,Y)
              & revenge(U,Z)
              & event(U,X1)
              & agent(U,X1,X)
              & patient(U,X1,Y)
              & present(U,X1)
              & nonreflexive(U,X1)
              & scream(U,X1)
              & of(U,X1,Z) ) ) )   )).
%--------------------------------------------------------------------------
*/
