package org.mathweb.texmacs;

//import junit.framework.TestCase;
/*
 java org.mathweb.test.TestTexmacsServer 
*/

import java.util.*;
import java.net.*;
import java.io.*;
import java.text.ParseException;

import org.apache.xmlrpc.*;
import edu.cmu.apros.translation.firstorder.FirstOrderTranslator;
import edu.cmu.apros.language.*;
import edu.cmu.apros.theory.Assertion;

import org.mathweb.texmacs.Apros2Tptp;


/**
 * Created by mathweb.org.
 * User: jzimmer
 * Date: June 22, 2004
 * Time: 12:04:59 PM
 */
public class TestTexmacsServer {
    
    private String tString = new String();
    private String id = "";
    public static String defaultURL = "http://king.ags.uni-sb.de:30000/RPC2/Broker";
    public XmlRpcClient broker_connection;

    FirstOrderTranslator translator = new FirstOrderTranslator();
    Apros2Tptp tptpTranslator = new Apros2Tptp();
    
    /**
     * Standard constructor (please keep).
     */
    public TestTexmacsServer(String name) 
    {
        id = name;
        initXmlRpcClient();
    }
    
    public void testServer() {
        try{
            while(true){
                ServerSocket serverSocket = null;
                try {
                    serverSocket = new ServerSocket(4445);
                } catch (IOException e) {
                    System.err.println("Could not listen on port: 4445.");
                    System.exit(1);
                }
                
                Socket clientSocket = null;
                try {
                    clientSocket = serverSocket.accept();
                } catch (IOException e) {
                    System.err.println("Accept failed.");
                    System.exit(1);
                }
                
                PrintWriter out = new PrintWriter(clientSocket.getOutputStream(), true);
                BufferedReader in = new BufferedReader(
                                                       new InputStreamReader(clientSocket.getInputStream()));
                String inputLine, outputLine;
                //KnockKnockProtocol kkp = new KnockKnockProtocol();
                //outputLine = kkp.processInput(null);
                // out.println(outputLine);
                
                while ((inputLine = in.readLine()) != null) {
                    try{
                        // outputLine = kkp.processInput(inputLine);
                        tString = "";
                        outputLine = null;
                        //unparseable = false;
                        //unprovable = false;
                        System.out.println("inputline is: " + inputLine);
                        
                        outputLine = processInput(inputLine);
                        
                        System.out.println(outputLine);
                        out.println(outputLine);
                        out.flush();
                        // outputLine = null;
                        System.out.println("after outprinting");
                        //if (outputLine.equals("Bye."))
                        //  break;
                    } catch(Exception e){
                        System.out.println("error " +e);
                    }
                }
                System.out.println("OUT");
                out.close();
                in.close();
                clientSocket.close();
                serverSocket.close();
            }
        } catch(IOException e) {
            System.out.println("Trouble in main server loop... " + e);
        }
    }
    

    private String processInput(String theInput) {
            // parse the Apros syntax
            String[] split = theInput.split("#");
            String command = "";
            String data = "";
            System.out.println("split: " + split.length + " " +split[0]  + " " +split[1] );
            if (split.length>1) {  command = split[0]; data = split[1];}
            else { command = "prove"; data = theInput;
            }
            
            if (command.equals("prove")) return callProver(data);
            if (command.equals("compute")) return callCAS(data);
            return "Command " + theInput + " not known!";
            
    }

    private String callCAS(String theInput)
    {
        try{
            // request an Maple object from the MathWeb broker
            String e_url = createService("MAPLE");
            System.out.println("Maple URL: " + e_url);
            
            // compose XML-RPC call
            Vector params = new Vector();
            params.addElement(theInput);
            Object result = call(e_url, "eval", params);
            
            Object res = call(e_url, "kill", params);
            String value = (String) result;
            
            // simply return the state the prover has determined.
            return "Maple returned: " +value;
        }
        catch(IOException e) {
            System.out.println("Trouble with MathWeb service invocation... " + e);
            return "IO exception while calling MathWeb: " + e;
        }
        catch (XmlRpcException e) {
            System.out.println("Trouble with MathWeb service invocation... " + e);
            return "XML-RPC exception while calling MathWeb: " + e;
        }
    }

    
    private String callProver(String theInput)
    {
        try{
            Assertion assertion = parseAssertion(theInput);
            if (assertion == null){
                return "SYNTAX ERROR";
            }
            // get the TPTP version of the conjecture
            String tptpAss = tptpTranslator.apros2Tptp(assertion);
            System.out.println("TPTP problem is: \n" + tptpAss);
            
            // request an E prover object from the MathWeb broker
            String e_url = createService("EPROVER");
            System.out.println("E URL: " + e_url);
            
            // compose XML-RPC call
            Hashtable argument = new Hashtable();
            Hashtable replyWith = new Hashtable();
            replyWith.put("state", " ");
            replyWith.put("output", "all");
            replyWith.put("time", "all");
            
            argument.put("1", tptpAss);
            argument.put("syntax", "tptp");
            argument.put("timeout", new Integer(20));
            argument.put("replyWith", replyWith);
            
            Vector params = new Vector();
            params.addElement(argument);
            
            // call the 'prove' method of the E service
            Object result = call(e_url, "prove", params);
            System.out.println("Result: " + result);
            
            Hashtable t = (Hashtable) result;
            
            //  never forget to kill the service object! :-)
            Object res = call(e_url, "kill", params);
            String state = (String) t.get("state");
            
            // simply return the state the prover has determined.
            return "Prover E says: " +state;
        }
        catch(IOException e) {
            System.out.println("Trouble with MathWeb service invocation... " + e);
            return "IO exception while calling MathWeb: " + e;
        }
        catch (XmlRpcException e) {
            System.out.println("Trouble with MathWeb service invocation... " + e);
            return "XML-RPC exception while calling MathWeb: " + e;
        }
    }
    
    private Assertion parseAssertion(String s) {
        Assertion assertion = null;
        try {
            assertion = translator.parseAssertion(s);
        } catch (ParseException e) {
            //unparseable= true;
            e.printStackTrace(); //To change body of catch statement use Options | File Templates.
            //return null;
        }
        return assertion;
}
    
    
    /* ****************************************************************
                                 The XML RPC stuff  
    ***************************************************************** */
    public void initXmlRpcClient() {
        try{
            this.broker_connection = new XmlRpcClient(defaultURL);
        }catch (MalformedURLException e) {System.exit(1);
        }
    }
    
    public Object call(String url, String methodName, Vector params)  throws IOException, XmlRpcException {
        try {
            XmlRpcClient service_connection = new XmlRpcClient(url);
            Object result = service_connection.execute(methodName, params);
            return result;
        }
        catch( IOException e) {
            System.out.println("Trouble with RPC call " + e);
            throw e;
        }
        catch (XmlRpcException e) {
            System.out.println("Trouble with RPC call " + e);
            throw e;
        }
    }
    
    public String createService(String serviceName) throws IOException, XmlRpcException {
        try {
            Vector params = new Vector();
            params.addElement(serviceName);
            System.out.println("Sending xml-rpc request: getService with one parameter: \"" + serviceName + '"');
            String result = "";
            result = (String) broker_connection.execute("getService", params);
            return result;
        }
        catch( IOException e) {
        System.out.println("Trouble on  invoking createService " + e);
        initXmlRpcClient();
        throw e;
        }
        catch (XmlRpcException e) {
            System.out.println("Trouble on  invoking createService " + e);
            initXmlRpcClient();
            throw e;
        }
    }
    
    public static void main(String[] args) {
        System.out.println("Texmacs server running on port 4445!");
        TestTexmacsServer tts = new TestTexmacsServer("tts");
        tts.testServer(); 
    }
} //end class
