// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/


package org.mathweb.trans;

import java.io.*;
import java.util.Vector;
import java.net.MalformedURLException;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;

import org.mathweb.core.BasicService;
import org.mathweb.core.CommandThread;

import org.mathweb.onto.OWLJDom;
import org.mathweb.trans.OmegaXmlRpcHandler;

/**
@author Juergen Zimmer Universitaet des Saarlandes
@version $Date: 2005/06/29 14:07:14 $ $Revision: 1.4 $
*/
public class Omega extends BasicService {

  /** Lisp executables and environments */
  private String ACL_HOME = getSystemProperty("acl.home", "ACL_HOME");
  private String aclExecutable = findSystemExecutable(ACL_HOME, "lisp", "alisp");
  
  private String omegaDump = getSystemProperty("omega.dump", "OMEGA_DUMP");
  private String omegaExecutable = getSystemProperty("omega.exec", "OMEGA_EXEC");
  private String omegaXtermExecutable = getSystemProperty("omega.exec.xterm", "OMEGA_EXEC_XTERM");
  private String omegaEmacs = getSystemProperty("omega.emacs.exec", "OMEGA_EMACS_EXEC");
  private String omegaIlisp = getSystemProperty("omega.ilisp", "OMEGA_ILISP");
  private String DISPLAY = getSystemProperty("omega.display", "OMEGA_DISPLAY");
  private String emacsGeometry = getSystemProperty("omega.emacs.geometry", "OMEGA_EMACS_GEOMETRY");
  private String[] acl_envs = new String[] {"ACL_HOME="+ACL_HOME, "DISPLAY="+DISPLAY};


  private static Logger logger = Logger.getLogger("Omega");
  private static String logFileName = getSystemProperty("omega.log.file", "OMEGA_LOG_FILE");

  private static String lispLogFileName = getSystemProperty("acl.log.file", "ACL_LOG_FILE");

  /** For communication with Lisp process */
  private static String sep = File.separator;
  private int serverPort = 0;
  private int defaultPort = 12345;
  private CommandThread lispThread;
  
  private WebServer xmlRpcServer;
  private String lispServerURL = "";
  private org.apache.xmlrpc.XmlRpcClient lispServerConnection;
  
  private static final int defaultNice = 15;
  private Throwable exception;
  private boolean started = false;
  private boolean ready = false;
  private int problemCounter = 1;
  private int ndCounter = 1;
  private static int logfileCounter = 1;
  private static int lisplogfileCounter = 1;
  private String user = "foo";
  
  public Omega(String usr)
  {
    mkTempDir("omega");
    //    owlJDom = OWLJDom.getInstance();

    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    FileAppender appender = newAppender(logFileName);
    logger.removeAllAppenders();
    logger.addAppender(appender);

    xmlRpcServer = null;
    lispServerConnection = null;
    started = false;
    ready = false;
    user = usr;
    //    startOmega();
  }

  /**
     load-problems typed-set
     prove test-bool-prop-set87
     expand-default-subproblem  
      CONCL (NDLINE) conclusion line: [CONC]
      SUPPORTS (NDLINE-LIST) A list of support-lines.: [()]
      PROHIB-DEFS (THY-ASS-LIST) A list of definitions prohibited for expansion.: [(= DEFINED EQUIV)]
      OMEGA: ;;;CSM Arbitrary [2]: 0 provers have to be killed
  */

  public String callHoSetDefsElim(String postProb) throws Exception {
    String noDefsProb = null;
    File probFile = null;
    try{
      System.out.println("POST prob: "+ postProb);
      String tempDirPath = tempDir.getAbsolutePath();
      String probFileName = tempDirPath + sep + "problem" + problemCounter + ".post";
      System.out.println("prob file: "+ probFileName);
      problemCounter ++;
      probFile = new File(probFileName);
      writeFile(probFile, postProb);
      Vector params = new Vector();
      params.addElement(probFileName);
      
      Object result = callLispCommand("read-problem", params);
      System.out.println("result of read-problem " + result);
      
      params = new Vector();
      result = callLispCommand("expand-default-subproblem", params);
      System.out.println("result: " + result + result.getClass().getName());
      noDefsProb = (String) result;
      return noDefsProb;
    } catch (XmlRpcException e) {
      logger.debug("Omega error " + e.code + ":" + e.getMessage());
      throw new Exception("Omega error (code " + e.code + "): " + e.getMessage());
    } finally{
      try{ 
        if (probFile != null) probFile.delete();
      } catch(Exception e){ logger.debug("[Ho2Fo] Exception " + e); }
    }
  }
  
  
  private void startXmlRpcServer() throws ClassNotFoundException {
    //    XmlRpc.setDriver(Class.forName(".SAXDriver"));
    boolean success = false;
    serverPort = defaultPort;
    do{
      try{
        xmlRpcServer = new WebServer(serverPort);
        xmlRpcServer.start();
        logger.debug("Started Java XMLRPC server on port "+serverPort);
        xmlRpcServer.addHandler ( "$default", new OmegaXmlRpcHandler(this));
        success = true;
      }catch(IOException e){
        serverPort++;
      }
      if ((serverPort - defaultPort) > 40) {
        logger.debug("More than 40 ports over "+ defaultPort + " are bound. Giving up!");
        serverPort = 0;
        break;
      }
    } while (!success);
  }
  
  public Omega startOmegaEmacs(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String ilisp = readFile(omegaIlisp);
        String tempDirPath = tempDir.getAbsolutePath();
        String inputFileName = tempDirPath + sep + "omega.el";
        String emacsStartup = ilisp + "\n" +
          "(setq loui-program " + 
          "\"" + omegaExecutable + " " + aclExecutable + " " + omegaDump + " " + serverURL + "\") ";
        writeFile(inputFileName, emacsStartup);
        
        if (serverPort != 0) {
          String command2 = omegaEmacs + " " +
            "-bg white -fg black -debug-init -display " + DISPLAY + " -name \"Omega Service (Emacs)\" " +
            "-geometry " + emacsGeometry + " -font 8x13" +
            " -l " + inputFileName + " -f loui";
          logger.debug("omega command: " + command2);
          lispThread = runThreaded(command2, acl_envs);
          logger.debug("after execution");
          started= true;
          return this;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Omega:" +e);
        logger.fatal("Something went wrong when starting Omega:" +e);
      }
    }
    return this;
  }

  public Omega startOmegaXterm(){
    if (!started){
      try {
        startXmlRpcServer();
        String serverURL = "http://localhost:" + serverPort;
        String tempDirPath = tempDir.getAbsolutePath();
        
        if (serverPort != 0) {
          String command2 =  omegaXtermExecutable + " " + aclExecutable + " " + omegaDump + " " + serverURL + " localhost:0.0 ";
          logger.debug("omega command: " + command2);
          lispThread = runThreaded(command2, acl_envs);
          started = true;
          return this;
        }
      }
      catch(ClassNotFoundException e) {
        System.err.println("Something went wrong when starting Omega:" +e);
        logger.fatal("Something went wrong when starting Omega:" +e);
      }
    }
    return this;
  }

  public Omega startOmega() {
      if (!started){
        try {
          startXmlRpcServer();
          String serverURL = "http://localhost:" + serverPort;
          String tempDirPath = tempDir.getAbsolutePath();
          String outputFileName = tempDirPath + sep + "omega" + ".out";
          
          if (serverPort != 0) {
            String command2 = aclExecutable + " -I " + omegaDump + 
              " -qq -e \"(progn (ags::rpc~start-server '(\\\"" + serverURL + 
              "\\\" \\\"initLispServer\\\" \\\""+ user + "\\\")) " +
              "(setf excl::*restart-app-function* (lambda () (omega::omega~top))))\" > " + lispLogFileName + user + (lisplogfileCounter++) + ".log 2>&1";
            logger.debug("omega command: " + command2);
            lispThread = runThreaded(command2, acl_envs);
          }
          started = true;
          return this;
        }
        catch(Exception e) {
          System.err.println("Something went wrong when starting Omega:" +e);
          logger.fatal("Something went wrong when starting Omega:" +e);
        }
      }
      return this;
  }
 
  public boolean isStarted() { return this.started; }
  public boolean isReady() { return this.ready; }
  public void setReady() { this.ready = true; }
  
  public void setLispServerConnection(String url)
  {
    try{
      lispServerURL = url;
      lispServerConnection = new org.apache.xmlrpc.XmlRpcClient(url);
    } catch (MalformedURLException e){
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    }
  }
  

  public Object callLispCommand(String command, Vector params) throws XmlRpcException
  {
    try{
      Object result = lispServerConnection.execute(command, params);
      logger.debug("Result: " + result + " " + result.getClass());
      return result;
    } catch (MalformedURLException e){
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    } catch(IOException e) {
      logger.debug("Trouble with RPC call " + e.getStackTrace());
      e.printStackTrace();
    }
    return null;
  }
  
  public void shutdown()
  {
    try { 
      Vector params = new Vector();
      if (lispServerConnection != null) {
        logger.debug("shutting down lisp process!");
        callLispCommand("shutdown", params);
        logger.debug("waiting 4 secs!");
        waitMillis(4000);
        logger.debug("waited 4 secs for lisp process!");
        xmlRpcServer.removeHandler("$default");
        xmlRpcServer.shutdown();
        lispThread.shutdown();
        logger.debug("lisp process terminated!");
      }
    } catch (XmlRpcException e) {
      logger.debug("Trouble with RPC call " + e);
      e.printStackTrace();
    } finally {
      removeTempDir();
    }
  }
  
  
  public void waitMillis(long millis)
  {
    long t1 = System.currentTimeMillis();
    while ((System.currentTimeMillis() - t1) < millis);
  }
  
  public void test () {
    try {
      String setProb = readFile(msHome + "/thirdParty/omega/xmp/test-bool-prop-set88.post");
      String result = null;
      result = callHoSetDefsElim(setProb);
      System.out.println(result);
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  } //end test()
  
  
  public static void main(String argv[]) throws Exception {
    Omega e = new Omega("test");
    //e.startOmegaEmacs();
    //e.startOmegaXterm();
    e.startOmega();
    while (! e.isReady()) {
      System.out.println("not ready");
      e.waitMillis(1000);
    }
    e.test();
    e.shutdown();
    System.exit(0);
        
  }
  
} // end Omega.java

      
