package org.mathweb.trans;

import java.util.Vector;
import java.net.MalformedURLException;

import org.apache.xmlrpc.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

/** The XML-RPC server for the lisp process */
public class OmegaXmlRpcHandler implements XmlRpcHandler {
  
  private static Logger logger = Logger.getLogger("OmegaXmlRpcHandler");
  private static String msHome = System.getProperty("mathserv.home");
  private static String logFileName = msHome + "/logs/omegaxmlrpchandler.log";
  private static FileAppender appender = newAppender(logFileName);
  private Omega omega;
  
  public OmegaXmlRpcHandler(Omega om)
  {
    super();
    this.omega = om;
    logger.setLevel(Level.ALL);
    logger.addAppender(appender);
    logger.debug("Omega XmlRpc handler started.");
  }
  
  public static FileAppender newAppender(String logFileName){
    FileAppender appender = null;
    try {
      appender = new FileAppender(new SimpleLayout(), logFileName, false);
    } catch(Exception e) {}
    return appender;
  }

  public Object execute (String methodName, Vector params ) 
  {
    try{ 
      logger.debug("RPC call method: " + methodName);
      logger.debug("args: " + params);
      for (int i=0; i<params.size(); i++){
	logger.debug("args " + params.get(i) + " " + params.get(i).getClass());
      }
      if (methodName.equals("initLispServer")) {
	System.out.println("params for initLispServer" + params);
        String lispServerURL = (String) params.get(1);
	logger.debug("lisp server: " + lispServerURL);
	omega.setLispServerConnection(lispServerURL);
	return "T";
      }
      else if (methodName.equals("ready")) {
	logger.debug("lisp says that it is ready!");
	omega.setReady();
	return "T";
      }
    } catch(Exception e){
      e.printStackTrace();
    }
    return null;
    }
}
