// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.util;

import java.io.*;
import java.util.*;
import java.net.URI;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;
 
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2005/06/17 17:37:12 $ $Revision: 1.1 $
*/
public class OmdocCreator {
  private Quoter quoter = new Quoter();
  private static String xmlBase = "http://www.mathweb.org/omdoc/Problems#";
  private static OWLJDom owlJDom = OWLJDom.getInstance();
  private static String sep = File.separator;
  
  public String omdoc2OWLSQuery(String omdocProb) { 
    String msHome = System.getProperty("mathserv.home");
    try{
      String name = "generated_omdoc_problem_"+System.currentTimeMillis();
      System.out.println(name);
      String tail = readFile(msHome + sep + "etc" + sep + "omdoc" + sep + "omdoc-tail.xml");
      String head = readFile(msHome + sep + "etc" + sep + "omdoc" + sep + "omdoc-head.xml"); 
      return head + name + "\">\n"+
        "  <mw:formalDescription rdf:datatype=\"http://www.w3.org/2001/XMLSchema#string\">\n" +
        quoter.quote(omdocProb) +tail;
    } catch(Exception e) {
      return null;
    }
  }
  
  public String omdocFile2OWLSQuery(String omdocFileName)
  {
    String omdocProb = readFile(omdocFileName);
    return omdoc2OWLSQuery(omdocProb);
  }

  public String omdocFile2OWLProblem(String omdocFileName)
  {
    try{
      File file = new File(omdocFileName);
      String name = file.getName().replace('-', '_').trim();
      name = name.replace('+', '_');
      int nameIndex = name.indexOf(".omdoc");
      if (nameIndex > -1) 
        name = name.substring(0, nameIndex);
      
      System.out.println("new name: " + name);
      
      String omdocProb = readFile(omdocFileName);
      
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      rdf.setAttribute("base", xmlBase, Namespace.XML_NAMESPACE);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      Element prob;
      prob = new Element("OmdocProvingProblem",OWLJDom.MW_NS);
      
      prob.setAttribute("ID", name, OWLJDom.RDF_NS);
      Element formalProb = new Element("formalDescription",OWLJDom.MW_NS);
      formalProb.addContent(omdocProb);
      Element probLogic = new Element("inLogic",OWLJDom.MW_NS); 
      probLogic.setAttribute("resource", "typed_lambda_calc", OWLJDom.RDF_NS);
      
      prob.addContent(formalProb);
      prob.addContent(probLogic);
      rdf.addContent(prob);
      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      return null;
    }
  }
  
  
  public String readFile(String fileName){
    try {
      FileInputStream isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    }
  }
    
  
  public static void main(String[] args) {
    OmdocCreator creat  = new OmdocCreator();
    String result= creat.omdocFile2OWLProblem(args[0]);
    System.out.println(result);
    System.exit(0);
  }
}

  
  
