// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.util;

import java.io.*;
import java.util.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2005/06/29 14:07:14 $ $Revision: 1.3 $
*/
public class TPTPCrop extends BasicService {
  
  public String readFile(String fileName){
    try {
      FileInputStream isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    }
  }
    
    public boolean writeFile(String fileName, String content){
    try {
      FileOutputStream outFile = new FileOutputStream(fileName) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    }
  }
  
  public void filter(String inputFileName)  {
    try{
      String string = readFile(inputFileName); 
      StringBuffer buff = new StringBuffer();
      String[] lines = string.split("\n");
      for(int i=0; i< lines.length; i++){
        String line = lines[i];
        String[] parts = line.split("   ");
        if(parts.length>1){
          String file = parts[0];
          String group = file.substring(0,3);
          String filename = "/amd/nfs/roc/disk/ptn075/dream-src/TPTPWorld/TPTP/Problems/"+ group + "/" + file + ".p";
          File testfile = new File(filename);
          if (!testfile.exists()){
            System.out.println("   " + file);
          }else {
            buff.append(line);
            buff.append("\n");
          }
        }
      }
      writeFile("/amd/nfs/roc/disk/ptn075/dream-src/TPTPWorld/SPCLists/new/"+inputFileName, buff.toString());
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  }
  
  public void extractRunTimes(String inputFileName)  {
    try{
      String input = readFile(msHome + "/thirdParty/TPTPWorld/rawdata/"+inputFileName); 
      StringBuffer buff = new StringBuffer();
      String[] lines = input.split("\n");
      boolean afterHeader = false;
      for(int i=0; i< lines.length; i++){
        String line = lines[i];
        if ((!afterHeader) && (line.indexOf("being prepared") == -1)){
          afterHeader = true;
        }
        if (afterHeader){
          if (line.substring(0,6).equals("RESULT"))
            if (line.indexOf("Timeout") == -1){
              int cpuStart = line.indexOf("CPU = ", 20);
              String cpuTimeStr = line.substring(cpuStart + 5, cpuStart + 14);
              System.out.println("cpustr: " + cpuTimeStr);
              String wcTimeStr = line.substring(cpuStart + 20, cpuStart + 28);
              System.out.println("wcstr: " + wcTimeStr);
              Float cpu = new Float(Float.parseFloat(cpuTimeStr) * 1000);
              Float wc = new Float(Float.parseFloat(wcTimeStr) * 1000);
              buff.append(cpu.intValue() + "\t" + wc.intValue() + "\n");
            }
        }
      }
      System.out.println(buff);
      writeFile(msHome + "/etc/data/eval/"+inputFileName, buff.toString());
    }    catch(Exception ex){
      ex.printStackTrace();
    }
  }

  
  public static void main(String[] args) {
    TPTPCrop crop  = new TPTPCrop();
    //    crop.filter(args[0]);
    crop.extractRunTimes(args[0]);
    System.exit(0);
  }
}

  
  
