// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.util;

import java.io.*;
import java.util.*;

import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2005/07/11 15:03:03 $ $Revision: 1.6 $
*/
public class TPTPShuffle {
  private Random random =  new Random(System.currentTimeMillis());

  
  public String readFile(String fileName){
    try {
      FileInputStream isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    }
  }
    
    public boolean writeFile(String fileName, String content){
    try {
      FileOutputStream outFile = new FileOutputStream(fileName) ;
      outFile.write(content.getBytes());
      outFile.close();
      return true;
    } catch (IOException ioe) {
      return false;
    }
  }
  
  
  public void shuffleSPCs(int finalNumber)  { 
    String msHome = System.getProperty("mathserv.home");
    try{
      String input = readFile(msHome + "/thirdParty/TPTPWorld/NewSPCLists/probNumbers");
      StringBuffer buff = new StringBuffer();
      String[] lines = input.split("\n");
      String[] spcs = new String[lines.length];
      Double[] rates = new Double[lines.length];
      
      double problems = 0.0;
      
      for(int i=0; i< lines.length; i++){
        String line = lines[i];
        // skip the second entry for that problem:
        String[] parts = line.split(" ");
        if (parts.length > 1){
          spcs[i] = parts[0].trim();
          String intStr = parts[1].trim();
          int num = (new Integer(intStr)).intValue();
          problems += num;
          rates[i] = new Double(num);
        }
      }
      
      StringBuffer output = new StringBuffer();
      int picked = 0;
      for(int i=0; i< lines.length; i++){
        //         if (spcs[i].indexOf("FOF") > -1){
        if (true){
          String spc = readFile(msHome + "/thirdParty/TPTPWorld/NewSPCLists/"+spcs[i]);
          String[] spcProbs = spc.split("\n");
          int maxProbs = spcProbs.length;
          // the relative number of problems to be chosen out of this SPC
          System.out.println("rates " + rates[i]);
          
          int num =  (new Double(rates[i].doubleValue() / problems * finalNumber)).intValue();
          picked += num;
          System.out.println("choosing " + num + " problems out of " + spcs[i]);
          
          // choose num  problems out of this SPC
          for(int k=0; k< num; k++){
            int index  = random.nextInt(maxProbs);
            output.append(spcProbs[index]);
            output.append("\n");
          }
        }
      }
      
      System.out.println("picked " + picked);
      Calendar rightNow = Calendar.getInstance();
      String outputFileName = msHome + "/thirdParty/TPTPWorld/NewSPCLists/RandomCollection-" +
        rightNow.get(Calendar.YEAR) + "-" + (rightNow.get(Calendar.MONTH) + 1) + "-" + 
        rightNow.get(Calendar.DAY_OF_MONTH) + "-" +
        rightNow.get(Calendar.HOUR_OF_DAY) + ":" + rightNow.get(Calendar.MINUTE)  + "-" + finalNumber;
      writeFile(outputFileName, output.toString());
    } catch(Exception ex){
      ex.printStackTrace();
    }
  }
  
  
  public static void main(String[] args) {
    TPTPShuffle shuffle  = new TPTPShuffle();
    shuffle.shuffleSPCs(500);
    System.exit(0);
  }
}

/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

