// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathWeb agent system

Authors: Juergen Zimmer

email: jzimmer@mathweb.org
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB nformatik                                            
Universitaet des Saarlandes                                          
Bau 36, 4. Stock                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathWeb, see 
http://www.mathweb.org/mathweb/agents

See the file "LICENSE" or
http://www.mathweb.org/mathweb/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.util;

import java.io.*;
import java.util.*;
import java.net.URI;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.jdom.Namespace;
import org.jdom.JDOMException;
import org.jdom.xpath.XPath;
 
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer, Saarland University, Germany
@version $Date: 2005/06/30 17:28:32 $ $Revision: 1.8 $
*/
public class TSTPCreator extends BasicService {
  private Quoter quoter = new Quoter();
  private static String xmlBase = "http://www.tptp.org/Problems#";
  private static OWLJDom owlJDom = OWLJDom.getInstance();
  private static String sep = File.separator;
  
  public String tstpFOF2OWLSQuery(String tstpFOFProb) { 
    return tstp2OWLSQuery(tstpFOFProb, "fof");
  }
  
  public String tstpCNF2OWLSQuery(String tstpCNFProb) {
    return tstp2OWLSQuery(tstpCNFProb, "cnf");
  }
  
  public String tstp2OWLSQuery(String tstpProb) { 
    if (tstpProb.indexOf("cnf(") > -1) 
      return tstp2OWLSQuery(tstpProb, "cnf");
    else
      return tstp2OWLSQuery(tstpProb, "fof");
  }
  
  public String tstp2OWLSQuery(String tstpProb, String format) { 
    try{
      String name = "generated_tstp_" + format + "_problem_"+System.currentTimeMillis();
      //      System.out.println(name);
      String tail = readFile(msHome + sep + "etc" + sep + "tstp" + sep + "tptp-"+format+"-tail.xml");
      String head = readFile(msHome + sep + "etc" + sep + "tstp" + sep + "tptp-"+format+"-head.xml"); 
      return head + name + "\">\n"+
        "  <mw:formalDescription rdf:datatype=\"http://www.w3.org/2001/XMLSchema#string\">\n" +
        quoter.quote(tstpProb) +tail;
    } catch(Exception e) {
      return null;
    }
  }
  
  public String tstpFile2OWLSQuery(String tstpFileName)
  {
    try{
      File file = new File(tstpFileName);
      String name = file.getName().replace('-', '_');
      //      System.out.println(name);
      String tstpProb = readFile(tstpFileName);
      String format = null;
      if (tstpFileName.indexOf("+") > -1) {
        format = "fof";
        name = name.replace('+', '_') + "_fof";
      }
      else format = "cnf";
      name = name.replace('.', '_');
      name = name + "_" + System.currentTimeMillis();
      
      String tail = readFile(msHome + sep + "etc" + sep + "tstp" + sep + "tptp-"+format+"-tail.xml");
      String head = readFile(msHome + sep + "etc" + sep + "tstp" + sep + "tptp-"+format+"-head.xml"); 
      return head + name + "\">\n"+
        "  <mw:formalDescription rdf:datatype=\"http://www.w3.org/2001/XMLSchema#string\">\n" +
        quoter.quote(tstpProb) +tail;
    } catch(Exception e) {
      return null;
    }
  }

  public String tstp2OWLProblem(String tstpFileName)
  {
    try{
      File file = new File(tstpFileName);
      String name = file.getName().replace('-', '_').trim();
      //      System.out.println(name);
      String tstpProb = readFile(tstpFileName);
      
      Element rdf = new Element("RDF",  OWLJDom.RDF_NS);
      rdf.setAttribute("base", xmlBase, Namespace.XML_NAMESPACE);
      rdf.addNamespaceDeclaration(OWLJDom.MW_NS);
      Element prob;
      if (tstpFileName.indexOf("+") > -1) {
        prob = new Element("TstpFOFProblem",OWLJDom.MW_NS);
        name = name.replace('+', '_') + "_fof";
      } else{
        prob = new Element("TstpCNFProblem",OWLJDom.MW_NS);
      }
      prob.setAttribute("ID", name, OWLJDom.RDF_NS);
      Element format = new Element("format",OWLJDom.MW_NS);
      format.setAttribute("resource",  OWLJDom.MW_NS_STR+"TSTP", OWLJDom.RDF_NS);
      Element formalProb = new Element("formalDescription",OWLJDom.MW_NS);
      formalProb.addContent(tstpProb);
      Element probLogic = new Element("inLogic",OWLJDom.MW_NS); 
      probLogic.setAttribute("resource", "fol_plus_equ", OWLJDom.RDF_NS);

      prob.addContent(formalProb);
      prob.addContent(probLogic);
      rdf.addContent(prob);
      return owlJDom.newXMLDocument(rdf);
      
    } catch(Exception e) {
      return null;
    }
  }
  
  
  public String readFile(String fileName){
    try {
      FileInputStream isFile = new FileInputStream(fileName);
      int nChar;
      StringBuffer sbRead = new StringBuffer();
      nChar = isFile.read();
      while (nChar > -1) {
        sbRead.append((char) nChar);
        nChar = isFile.read();
      }
      isFile.close();
      return sbRead.toString();
    } catch (IOException ioe) {
      return null;
    }
  }
    
  
  public static void main(String[] args) {
    TSTPCreator creat  = new TSTPCreator();
    String result= creat.tstpFile2OWLSQuery(args[0]);
    result= creat.tstp2OWLProblem(args[0]);
    System.out.println(result);
    System.exit(0);
  }
}

  
  
