%------------------------------------------------------------------------------
:-op(100,fx,++).
:-op(100,fx,--).
:-op(401,fy,~).
%------------------------------------------------------------------------------
%----Check every member of the first list is a member of the second
all_members([],_).

all_members([H|T],L):-
    member(H,L),
    all_members(T,L).
%------------------------------------------------------------------------------
%----Check if a list of literals subsumed another
literals_subsume(SubsumingLiterals,SubsumedLiterals):-
%----Verify to avoid instiating variables
    \+ ( \+ (
        numbervars(SubsumedLiterals,0,_),
        all_members(SubsumingLiterals,SubsumedLiterals))).
%------------------------------------------------------------------------------
%----Set if two clauses have the same literals, modulo renaming
%----Need to avoid instantiation here
same_literals([],[]).
    
same_literals([Literal1|RestOfLiterals1],Literals2):-
    select(Literal2,Literals2,RestOfLiterals2),
    numbervars(Literal1,0,NumberOfVariables),
    numbervars(Literal2,0,NumberOfVariables),
    Literal1 == Literal2,
    same_literals(RestOfLiterals1,RestOfLiterals2).
%------------------------------------------------------------------------------
%----TSTP conversion procedures
convert_tstp_to_literals((LHS + RHS),Literals):-
    !,
    convert_tstp_to_literals(LHS,LHSLiterals),
    convert_tstp_to_literals(RHS,RHSLiterals),
    append(LHSLiterals,RHSLiterals,Literals).

convert_tstp_to_literals(~ Atom,[--Atom]):-
    !.

convert_tstp_to_literals(Atom,[++Atom]).
%------------------------------------------------------------------------------
extract_name_and_tptp_literals(Clause,Name,Literals):-
    Clause =.. [cnf,Name,_Status,TSTPClause|_],
    !,
    convert_tstp_to_literals(TSTPClause,Literals).

extract_name_and_tptp_literals(input_clause(Name,_Status,Literals),
Name,Literals).
%------------------------------------------------------------------------------
do_comparison(Name1,Literals1,Name2,Literals2,same(Name1,Name2)):-
    same_literals(Literals1,Literals2),
    !.

do_comparison(Name1,Literals1,Name2,Literals2,subsumes(Name1,Name2)):-
    literals_subsume(Literals1,Literals2),
    !.

do_comparison(Name1,Literals1,Name2,Literals2,subsumed(Name1,Name2)):-
    literals_subsume(Literals2,Literals1),
    !.

do_comparison(Name1,_,Name2,_,unrelated(Name1,Name2)).
%------------------------------------------------------------------------------
extract_and_do_comparison(Clause1,Clause2,Result):-
    extract_name_and_tptp_literals(Clause1,Name1,Literals1),
    extract_name_and_tptp_literals(Clause2,Name2,Literals2),
    !,
    do_comparison(Name1,Literals1,Name2,Literals2,Result).

extract_and_do_comparison(_,_,failed).
%------------------------------------------------------------------------------
compare_clauses:-
    set_prolog_flag(char_conversion,on),
    char_conversion('|','+'),
    argument_list([_,Atom1,Atom2]),
    !,
    read_from_atom(Atom1,Clause1),
    read_from_atom(Atom2,Clause2),
write(Clause1),nl,write(Clause2),nl,
    extract_and_do_comparison(Clause1,Clause2,Result),
    write(Result),
    nl,
    halt.

compare_clauses:-
    write(failed),
    nl,
    halt.
%------------------------------------------------------------------------------
:-initialization(compare_clauses).
%------------------------------------------------------------------------------
