//-----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>

#include "DataTypes.h"
#include "Utilities.h"
#include "Tokenizer.h"
#include "Parsing.h"
#include "Signature.h"
#include "Examine.h"
#include "List.h"
#include "Tree.h"
#include "Compare.h"
#include "Modify.h"
#include "Interpret.h"
#include "PrintTSTP.h"

#include "GDV.h"
//-----------------------------------------------------------------------------
void GlobalInterruptHandler(int TheSignal) {

    extern int GlobalInterrupted;

    printf("GDV has been interrupted\n");
    GlobalInterrupted = 1;
}
//-----------------------------------------------------------------------------
//----Extract the contents of a term (wot's in the ()s)
int ExtractTermArguments(String Term) {

    char * Open;
    char * Close;

    if ((Open = strchr(Term,'(')) != NULL && (Close = strrchr(Term,')')) !=
NULL) {
        *Close = '\0';
        strcpy(Term,Open+1);
        return(1);
    } else {
        return(0);
    }
}
//-----------------------------------------------------------------------------
char * MakePrintableList(StringParts Names,int NumberOfNames,String ListNames) {

    int NameNumber;

    strcpy(ListNames,"");
    for (NameNumber = 0;NameNumber < NumberOfNames;NameNumber++) {
        strcat(ListNames,Names[NameNumber]);
        strcat(ListNames," ");
    }

    return(ListNames);
}
//-----------------------------------------------------------------------------
int ProcessCommandLine(int argc,char * argv[],OptionsType * OptionValues) {

    int OptionChar;

    OptionValues->Quietness = 1;
    OptionValues->ForceContinue = 0;
    OptionValues->VerifyLeaves = 0;
    strcpy(OptionValues->ProblemFileName,"");
    OptionValues->CheckParentRelevance = 0;
    OptionValues->CheckRefutation = 0;
    strcpy(OptionValues->TheoremProver,DEFAULT_THEOREM_PROVER);
    strcpy(OptionValues->ModelFinder,DEFAULT_MODEL_FINDER);
    strcpy(OptionValues->Saturator,DEFAULT_SATURATOR);
    strcpy(OptionValues->UnsatisfiabilityChecker,
DEFAULT_UNSATISFIABILITY_CHECKER);
    OptionValues->TimeLimit = DEFAULT_TIME_LIMIT;
    OptionValues->KeepFiles = 0;
    strcpy(OptionValues->KeepFilesDirectory,DEFAULT_KEEP_FILES_DIRECTORY);
    strcpy(OptionValues->DerivationFileName,"--");
    
    while ((OptionChar = getopt(argc,argv,"q:fli:k:rcp:u:m:s:t:h")) != -1) {
        switch (OptionChar) {
            case 'q':
                OptionValues->Quietness = atoi(optarg);
                break;
            case 'f':
                OptionValues->ForceContinue = 1;
                break;
            case 'l':
                OptionValues->VerifyLeaves = 1;
                break;
            case 'i':
                strcpy(OptionValues->ProblemFileName,optarg);
                break;
            case 'k':
                OptionValues->KeepFiles = 1;
                strcpy(OptionValues->KeepFilesDirectory,optarg);
                break;
            case 'r':
                OptionValues->CheckParentRelevance = 1;
                break;
            case 'c':
                OptionValues->CheckRefutation = 1;
                break;
            case 'p':
                strcpy(OptionValues->TheoremProver,optarg);
                break;
            case 'u':
                strcpy(OptionValues->UnsatisfiabilityChecker,optarg);
                break;
            case 'm':
                strcpy(OptionValues->ModelFinder,optarg);
                break;
            case 's':
                strcpy(OptionValues->Saturator,optarg);
                break;
            case 't':
                OptionValues->TimeLimit = atoi(optarg);
                break;
            case 'h':
            case '?':
                printf("Usage: %s <options> <derivation file>\n",argv[0]);
                printf("-q<quietness>  - control amount of output\n");
                printf("-f             - force continue on failure\n");
                printf("-l             - verify leaves\n");
                printf("-i<input file> - specify problem file\n");
                printf("-k<directory>  - keep discharge files in director\n");
                printf("-r             - check relevance of logical consequences\n");
                printf("-c             - derivation should be a refutation\n");
                printf("-p             - FOF theorem prover System---Version\n");
                printf("-u             - unsatifiability checker System---Version\n");
                printf("-m             - model finder System---Version\n");
                printf("-s             - saturator System---Version\n");
                printf("-t<time limit> - CPU limit for discharge attempts\n");
                printf("-h             - print this help\n");
                exit(EXIT_FAILURE);
                break;
            default:
                printf("ERROR: Something wrong in getopt\n");
                exit(EXIT_FAILURE);
                break;
        }
    }

//----Check for problem file name
    if (argv[optind] != NULL) {
        strcpy(OptionValues->DerivationFileName,argv[optind]);
    }

//----Echo if in verbose mode
    if (OptionValues->Quietness == 0) {
        printf("Command line options:\n");
        printf("    Quietness %d\n",OptionValues->Quietness);
        printf("    ForceContinue %d\n",OptionValues->ForceContinue);
        printf("    VerifyLeaves %s\n",OptionValues->VerifyLeaves?
(!strcmp(OptionValues->ProblemFileName,"")?"from derivation":
OptionValues->ProblemFileName):"0");
        printf("    CheckParentRelevance %d\n",
OptionValues->CheckParentRelevance);
        printf("    CheckRefutation %d\n",OptionValues->CheckRefutation);
        printf("    TheoremProver %s\n",OptionValues->TheoremProver);
        printf("    ModelFinder %s\n",OptionValues->ModelFinder);
        printf("    Saturator %s\n",OptionValues->Saturator);
        printf("    UnsatisfiabilityChecker %s\n",
OptionValues->UnsatisfiabilityChecker);
        printf("    TimeLimit %d\n",OptionValues->TimeLimit);
        printf("    KeepFiles %s\n",OptionValues->KeepFiles?
OptionValues->KeepFilesDirectory:"0");
        printf("    DerivationFileName %s\n",OptionValues->DerivationFileName);
    }

    return(1);
}
//-----------------------------------------------------------------------------
void ExtractBasename(String Path,String Basename) {

    char * Slash;

    if ((Slash = strrchr(Path,'/')) == NULL) {
        strcpy(Basename,Path);
    } else {
        strcpy(Basename,Slash+1);
    }

}
//-----------------------------------------------------------------------------
void EmptyAndDeleteDirectory(String Directory) {

    String UNIXCommand;

    sprintf(UNIXCommand,"rm -rf %s",Directory);
    system(UNIXCommand);
}
//-----------------------------------------------------------------------------
int CreateDirectory(String Directory,String DerivationFileName) {

    String DerivationFileBasename;

    if (!strcmp(DerivationFileName,"--")) {
        sprintf(DerivationFileBasename,"GDV%06d",getpid());
    } else {
        ExtractBasename(DerivationFileName,DerivationFileBasename);
    }

    strcat(Directory,"/");
    strcat(Directory,DerivationFileBasename);
    strcat(Directory,".gdv");

//----Delete any previous version
    EmptyAndDeleteDirectory(Directory);
//----And make the new one
    if (mkdir(Directory,0755) != 0) {
        perror("Creating working directory");
        return(0);
    } else {
        return(1);
    }
}
//-----------------------------------------------------------------------------
FILE * OpenFileInMode(OptionsType OptionValues,String Directory,
String BaseName,char * Extension,char * Mode,String FileName) {

    FILE * Stream;

    strcpy(FileName,Directory);
    strcat(FileName,"/");
    strcat(FileName,BaseName);
    strcat(FileName,".");
    strcat(FileName,Extension);
    if ((Stream = fopen(FileName,Mode)) == NULL) {
        QPRINTF(OptionValues,1)("ERROR: Could not open file %s in %s mode\n",FileName,Mode);
        return(NULL);
    } else {
        return(Stream);
    }
}
//-----------------------------------------------------------------------------
void RemoveFile(OptionsType OptionValues,String FileName) {

    if (unlink(FileName) == -1) {
        perror("Deleting file");
        QPRINTF(OptionValues,1)("ERROR: Could not delete file %s\n",FileName);
    }
}
//-----------------------------------------------------------------------------
int MakeListOfParents(OptionsType OptionValues,LISTNODE Head,
String FormulaName,StringParts ParentNames,int NumberOfParents,
LISTNODE * ParentList) {

    int ParentNumber;
    ANNOTATEDFORMULA ParentAnnotatedFormula;

    *ParentList = NULL;
    for (ParentNumber = 0;ParentNumber < NumberOfParents;ParentNumber++) {
        if (strstr(ParentNames[ParentNumber],"theory(") !=
ParentNames[ParentNumber]) {
            if ((ParentAnnotatedFormula = GetAnnotatedFormulaFromListByName(
Head,ParentNames[ParentNumber])) == NULL) {
                QPRINTF(OptionValues,2)(
"FAILURE: %s has a non-existent parent %s\n",FormulaName,
ParentNames[ParentNumber]);
                FreeListOfAnnotatedFormulae(ParentList);
                return(0);
            } else {
                AddListNode(ParentList,*ParentList,ParentAnnotatedFormula);
            }
        }
    }

    return(1);
}
//-----------------------------------------------------------------------------
void PrintArrayOfAnnotatedFormulaeToFile(FILE * Stream,ANNOTATEDFORMULAArray
AnnotatedFormulae,int NumberOfFormulae,StatusType Status) {

    int FormulaNumber;
    StatusType OldStatus;
    StatusType OldSubStatus;

    for (FormulaNumber = 0;FormulaNumber < NumberOfFormulae;FormulaNumber++) {
        if (AnnotatedFormulae[FormulaNumber] != NULL) {
            OldStatus = GetStatus(AnnotatedFormulae[FormulaNumber],
&OldSubStatus);
            SetStatus(AnnotatedFormulae[FormulaNumber],Status,nonstatus);
            PrintAnnotatedTSTPNode(Stream,AnnotatedFormulae[FormulaNumber],
tptp,1);
            SetStatus(AnnotatedFormulae[FormulaNumber],OldStatus,OldSubStatus);
        }
    }
}
//-----------------------------------------------------------------------------
void PrintAnnotatedFormulaeToFile(FILE * Stream,ANNOTATEDFORMULA
AnnotatedFormula,StatusType Status) {

    StatusType OldStatus;
    StatusType OldSubStatus;

    OldStatus = GetStatus(AnnotatedFormula,&OldSubStatus);
    SetStatus(AnnotatedFormula,Status,nonstatus);
    PrintAnnotatedTSTPNode(Stream,AnnotatedFormula,tptp,1);
    SetStatus(AnnotatedFormula,OldStatus,OldSubStatus);

}
//-----------------------------------------------------------------------------
void PrintListOfAnnotatedTSTPNodesWithStatus(FILE * Stream,LISTNODE Head,
StatusType Status) {

    while (Head != NULL) {
        PrintAnnotatedFormulaeToFile(Stream,Head->AnnotatedFormula,Status);
        Head = Head->Next;
        if (Head != NULL) {
            fprintf(Stream,"\n");
        }
    }
}
//-----------------------------------------------------------------------------
void AddUsefulInfoTag(ANNOTATEDFORMULA AnnotatedFormula,SIGNATURE Signature,
char * UsefulInfo) {

    if (!AddUsefulInformationToAnnotatedFormula(AnnotatedFormula,Signature,
UsefulInfo)) {
        CodingError("Could not add a verified tag");
    } 
}
//-----------------------------------------------------------------------------
void AddVerifiedTag(ANNOTATEDFORMULA AnnotatedFormula,SIGNATURE Signature,
char * TagValue) {

    String VerifiedTag;

    sprintf(VerifiedTag,"verified(%s)",TagValue);
    AddUsefulInfoTag(AnnotatedFormula,Signature,VerifiedTag);
}
//-----------------------------------------------------------------------------
int GetSystemResult(OptionsType OptionValues,String ObligationFileName,
String ATPSystem,String PutResultHere,String PutOutputFileNameHere) {

    String UNIXCommand;
    FILE * SystemOutputFile;
    String ObligationFileBasename;
    FILE * SystemPipe;
    int GotResult;
    String SystemOutputLine;
    char * SaysPart;
    char * CPUPart;

    sprintf(UNIXCommand,"%s/%s -q0 %s %d %s",TPTP_HOME,SYSTEM_ON_TPTP,ATPSystem,
OptionValues.TimeLimit,ObligationFileName);

    ExtractBasename(ObligationFileName,ObligationFileBasename);
    if ((SystemOutputFile = OpenFileInMode(OptionValues,
OptionValues.KeepFilesDirectory,ObligationFileBasename,"dis","w",
PutOutputFileNameHere)) == NULL) {
        return(0);
    }

    if ((SystemPipe = popen(UNIXCommand,"r")) == NULL) {
        perror("Running SystemOnTPTP");
        QPRINTF(OptionValues,1)("ERROR: Could not start %s\n",UNIXCommand);
        RemoveFile(OptionValues,PutOutputFileNameHere);
        return(0);
    }

//----Read SystemOnTPTP output echoing to file and looking for RESULT
    GotResult = 0;
    while (fgets(SystemOutputLine,STRINGLENGTH,SystemPipe) != NULL) {
        fputs(SystemOutputLine,SystemOutputFile);
//DEBUG printf("Line is %s",SystemOutputLine);
        if (!GotResult && 
strstr(SystemOutputLine,"RESULT: ") == SystemOutputLine &&
(SaysPart = strstr(SystemOutputLine," says ")) != NULL &&
(CPUPart = strstr(SystemOutputLine," - CPU =")) != NULL) {
            *CPUPart = '\0';
            strcpy(PutResultHere,SaysPart+6);
            *CPUPart = ' ';
            GotResult = 1;
//DEBUG printf("Got the result %s\n",PutResultHere);
        }
    }

    fclose(SystemOutputFile);
    pclose(SystemPipe);

    return(GotResult);
        
}
//-----------------------------------------------------------------------------
int CheckTheorem(OptionsType OptionValues,String ObligationFileName) {

    String SystemResult;
    String SystemOutputFileName;
    int Correct;

    if (GetSystemResult(OptionValues,ObligationFileName,
OptionValues.TheoremProver,SystemResult,SystemOutputFileName)) {
        Correct = !strcmp(SystemResult,"Theorem");
    } else {
        Correct = 0;
    }

    RemoveFile(OptionValues,ObligationFileName);
    if (!OptionValues.KeepFiles) {
        RemoveFile(OptionValues,SystemOutputFileName);
    }

    return(Correct);
}
//-----------------------------------------------------------------------------
int CheckSatisfiable(OptionsType OptionValues,String ObligationFileName) {

    String SystemResult;
    String SystemOutputFileName;
    int Correct;

//-----1 means no result yet
    Correct = -1;

//----Try finite model finder
    if (Correct == -1 && GetSystemResult(OptionValues,ObligationFileName,
OptionValues.ModelFinder,SystemResult,SystemOutputFileName)) {
        if (!strcmp(SystemResult,"Satisfiable")) {
            Correct = 1;
        } else if (!strcmp(SystemResult,"Unsatisfiable")) {
            Correct = 0;
        }
    }

//----Try saturator
    if (Correct == -1 && GetSystemResult(OptionValues,ObligationFileName,
OptionValues.Saturator,SystemResult,SystemOutputFileName)) {
        if (!strcmp(SystemResult,"Satisfiable")) {
            Correct = 1;
        } else if (!strcmp(SystemResult,"Unsatisfiable")) {
            Correct = 0;
        }
    }

//----Try prover in reverse
    if (Correct == -1 && GetSystemResult(OptionValues,ObligationFileName,
OptionValues.UnsatisfiabilityChecker,SystemResult,SystemOutputFileName)) {
        if (!strcmp(SystemResult,"Satisfiable")) {
            Correct = 1;
        } else if (!strcmp(SystemResult,"Unsatisfiable")) {
            Correct = 0;
        }
    }

    RemoveFile(OptionValues,ObligationFileName);
    if (!OptionValues.KeepFiles) {
        RemoveFile(OptionValues,SystemOutputFileName);
    }

    if (Correct == -1) {
        QPRINTF(OptionValues,1)(
"WARNING: Satisfiablity FAILURE with no Unsatisfiability found\n");
        Correct = 0;
    }

    return(Correct);
}
//-----------------------------------------------------------------------------
int CorrectlyInferred(OptionsType OptionValues,ANNOTATEDFORMULA Target,
String FormulaName,LISTNODE ParentAnnotatedFormulae,char * ParentNames,
String SZSStatus,char * Comment) {

    FILE * ObligationFile;
    String ObligationFileName;
    int Correct;

    if (!strcmp(SZSStatus,"thm")) {
        Correct = 1;
        if (OptionValues.CheckParentRelevance) {
            if (OptionValues.CheckRefutation && FalseAnnotatedFormula(Target)) {
                QPRINTF(OptionValues,1)(
"WARNING: %s is $false - parent satisfiability is unverified\n",FormulaName);
            } else {
                if ((ObligationFile = OpenFileInMode(OptionValues,
OptionValues.KeepFilesDirectory,FormulaName,"psat.obl","w",
ObligationFileName)) == NULL) {
                    Correct = 0;
                }
                PrintListOfAnnotatedTSTPNodesWithStatus(ObligationFile,
ParentAnnotatedFormulae,axiom);
                fclose(ObligationFile);
                if (CheckSatisfiable(OptionValues,ObligationFileName)) {
                    QPRINTF(OptionValues,2)(
"SUCCESS: %s has sat parents %s %s\n",FormulaName,ParentNames,
Comment != NULL?Comment:"");
                } else {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s does not have sat parents %s %s\n",FormulaName,ParentNames,
Comment != NULL?Comment:"");
                    Correct = 0;
                }
            }
        }


        if (!GlobalInterrupted && (Correct || OptionValues.ForceContinue)) {
            if ((ObligationFile = OpenFileInMode(OptionValues,
OptionValues.KeepFilesDirectory,FormulaName,"thm.obl","w",
ObligationFileName)) == NULL) {
                return(0);
            }
            PrintListOfAnnotatedTSTPNodesWithStatus(ObligationFile,
ParentAnnotatedFormulae,axiom);
            PrintAnnotatedFormulaeToFile(ObligationFile,Target,conjecture);
            fclose(ObligationFile);
            if (CheckTheorem(OptionValues,ObligationFileName)) {
                if (ParentNames != NULL) {
                    QPRINTF(OptionValues,2)(
"SUCCESS: %s is a %s of %s %s\n", FormulaName,SZSStatus,ParentNames,
Comment != NULL?Comment:"");
                }
            } else {
                if (ParentNames != NULL) {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s is not a %s of %s %s\n",FormulaName,SZSStatus,ParentNames,
Comment != NULL?Comment:"");
                }
                Correct = 0;
            }
        }
        return(Correct);
    } else if (!strcmp(SZSStatus,"cth")) {
        Negate(Target,0);
        Correct = CorrectlyInferred(OptionValues,Target,FormulaName,
ParentAnnotatedFormulae,ParentNames,"thm","(Negated cth)");
        Negate(Target,1);
        return(Correct);
    } else {
        QPRINTF(OptionValues,1)(
"WARNING: Cannot cope with SZS status %s\n",SZSStatus);
        return(0);
    }

}
//-----------------------------------------------------------------------------
int AllParentsExist(OptionsType OptionValues,LISTNODE Head) {

    LISTNODE Target;
    int OKSoFar;
    StatusType SubStatus;
    String AllParentNames;
    StringParts ParentNames;
    int NumberOfParents;
    String FormulaName;
    LISTNODE ParentList;
    String InferenceRule;

    OKSoFar = 1;

    Target = Head;
    while (OKSoFar && Target != NULL) {
        if (GetStatus(Target->AnnotatedFormula,&SubStatus) == derived ||
SubStatus == derived) {
            GetName(Target->AnnotatedFormula,FormulaName);
            if (GetInferenceRule(Target->AnnotatedFormula,InferenceRule) == 
NULL) {
                QPRINTF(OptionValues,2)(
"FAILURE: %s has no inference rule\n",FormulaName);
                OKSoFar = 0;
            } else {
                GetParentNames(Target->AnnotatedFormula,AllParentNames);
                if ((NumberOfParents = Tokenize(AllParentNames,ParentNames,
"\n")) == 0) {
                    QPRINTF(OptionValues,1)(
"WARNING: %s is derived from no parents\n",FormulaName);
                }
                if (!strcmp(InferenceRule,"") && NumberOfParents != 1) {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s is a copy of not one parent\n",FormulaName);
                    OKSoFar = 0;
                } else if (MakeListOfParents(OptionValues,Head,FormulaName,
ParentNames,NumberOfParents,&ParentList)) {
                    FreeListOfAnnotatedFormulae(&ParentList);
                } else {
                    OKSoFar = 0;
                }
            }
        }
        Target = Target->Next;
    }

    return(OKSoFar);
}
//-----------------------------------------------------------------------------
int ACyclic(OptionsType OptionValues,LISTNODE Head) {

    char * AllNames;
    StringParts Names;
    StringParts CopyOfNames;
    int NumberOfNames;
    String AllParentNames;
    StringParts ParentNames;
    int NumberOfParentNames;
    int Child;
    int Parent;
    int Name;
    int ParentFound;
    int RootFound;
    int Cyclic;

    AllNames = (char *)Malloc(sizeof(String));
    GetAllNames(Head,&AllNames);
    NumberOfNames = Tokenize(AllNames,Names,"\n");
    RootFound = 1;

//----Keep looping until no more roots can be found
    while (RootFound) {
        RootFound = 0;
        memcpy(CopyOfNames,Names,sizeof(StringParts));
//----For each name that is still in the game (not already been a root)
        for (Child = 0;Child < NumberOfNames;Child++) {
            if (Names[Child] != NULL) {
                GetParentNames(GetAnnotatedFormulaFromListByName(Head,
Names[Child]),AllParentNames);
//DEBUG printf("%s has parents %s\n",Names[Child],AllParentNames);
//----If it has no parents, it cannot be part of a cycle
                if ((NumberOfParentNames = Tokenize(AllParentNames,
ParentNames,"\n")) == 0) {
                    Names[Child] = NULL;
                }
//----If it has parents, none of the parents are roots
                for (Parent=0;Parent < NumberOfParentNames;Parent++) {
//DEBUG printf("    Looking for parent %s\n",ParentNames[Parent]);
                    ParentFound = 0;
                    for (Name=0;!ParentFound && Name < NumberOfNames;Name++) {
//----If a node is a parent, and not knocked out, then not a root this time
                        if (CopyOfNames[Name] != NULL && Names[Name] != NULL &&
!strcmp(Names[Name],ParentNames[Parent])) {
//DEBUG printf("      Found %s\n",ParentNames[Parent]);
                            CopyOfNames[Name] = NULL;
                            ParentFound = 1;
                        }
                    }
                }
            }
        }

//----Look for nodes that remain - they are not parents and therefore are roots
//----Knock them out of the game
        for (Child = 0;Child < NumberOfNames;Child++) {
            if (CopyOfNames[Child] != NULL) {
//DEBUG printf("%s is a root\n",CopyOfNames[Child]);
                Names[Child] = NULL;
                RootFound = 1;
            }
        }
    }

    Cyclic = 0;
    for (Child = 0;Child < NumberOfNames;Child++) {
        if (Names[Child] != NULL) {
            Cyclic = 1;
//DEBUG printf("%s is part of a cycle\n",Names[Child]);
        }
    }

    Free((void **)&AllNames);

    return(!Cyclic);
}
//-----------------------------------------------------------------------------
void RemoveNonFalse(LISTNODE * Head) {

    LISTNODE ToRemove;

    while (*Head != NULL) {
        if (!FalseAnnotatedFormula((*Head)->AnnotatedFormula)) {
            ToRemove = *Head;
            *Head = (*Head)->Next;
            FreeListNode(&ToRemove);
        } else {
            Head = &((*Head)->Next);
        }
    }

}
//-----------------------------------------------------------------------------
LISTNODE GetFalseNodes(LISTNODE Head) {

    LISTNODE FalseList;

    FalseList = DuplicateListOfNodes(Head,0);
    RemoveNonFalse(&FalseList);
    return(FalseList);

}
//-----------------------------------------------------------------------------
LISTNODE GetFalseRoots(LISTNODE Head) {

    LISTNODE FalseRootList;

    FalseRootList = GetRootList(Head);
    RemoveNonFalse(&FalseRootList);
    return(FalseRootList);
}
//-----------------------------------------------------------------------------
ANNOTATEDFORMULA HasFalseRoot(OptionsType OptionValues,LISTNODE Head) {

    LISTNODE FalseRootList;
    LISTNODE LastOne;
    ANNOTATEDFORMULA LastAnnotatedNode;

    FalseRootList = GetFalseRoots(Head);
    if (FalseRootList == NULL) {
        QPRINTF(OptionValues,2)("FAILURE: Derivation has no roots\n");
        return(NULL);
    } else {
        LastOne = FalseRootList;
        while (LastOne->Next != NULL) {
            LastOne = LastOne->Next;
        }
        LastAnnotatedNode = LastOne->AnnotatedFormula;
        FreeListOfAnnotatedFormulae(&FalseRootList);
        return(LastAnnotatedNode);
    }
}
//-----------------------------------------------------------------------------
int IsAncestor(LISTNODE Head,ANNOTATEDFORMULA Ancestor,ANNOTATEDFORMULA 
Descendant,int THMNodesOnly) {

    String AncestorName;
    String AllParentNames;
    int NumberOfParents;
    StringParts ParentNames;
    int ParentNumber;
    ANNOTATEDFORMULA Parent;
    String DescendantName;
    String ParentStatus;

    GetName(Ancestor,AncestorName);
    GetName(Descendant,DescendantName);
    GetParentNames(Descendant,AllParentNames);
    NumberOfParents = Tokenize(AllParentNames,ParentNames,"\n");
    for (ParentNumber=0;ParentNumber < NumberOfParents;ParentNumber++) {
        if (strstr(ParentNames[ParentNumber],"theory(") !=
ParentNames[ParentNumber]) {
            if (!strcmp(AncestorName,ParentNames[ParentNumber])) {
                return(1);
            } else {
                Parent = GetAnnotatedFormulaFromListByName(Head,
ParentNames[ParentNumber]);
                if ((!THMNodesOnly || 
(GetInferenceInfoTerm(Parent,"status",ParentStatus) != NULL && 
!strcmp(ParentStatus,"status(thm)"))) &&
IsAncestor(Head,Ancestor,Parent,THMNodesOnly)) {
                    return(1);
                }
            }
        }
    }
    return(0);
}
//-----------------------------------------------------------------------------
int IsAncestorOfOneInList(LISTNODE Head,ANNOTATEDFORMULA Ancestor,LISTNODE
ListOfDescendants,int THMNodesOnly) {

    while (ListOfDescendants != NULL) {
        if (IsAncestor(Head,Ancestor,ListOfDescendants->AnnotatedFormula,
THMNodesOnly)) {
            return(1);
        } else {
            ListOfDescendants = ListOfDescendants->Next;
        }
    }
    return(0);
}
//-----------------------------------------------------------------------------
int WellFormedProofsByContradiction(OptionsType OptionValues,LISTNODE Head,
SIGNATURE Signature,int * NumberOfProofsByContradiction) {

    LISTNODE Target;
    int OKSoFar;
    String FormulaName;
    String AllParentNames;
    int NumberOfParents;
    StringParts ParentNames;
    ANNOTATEDFORMULA FalseParent;
    ANNOTATEDFORMULA AncestorParent;
    ANNOTATEDFORMULA SwapParent;
    char * AncestorName;
    String ProcessedTag;

    OKSoFar = 1;
    *NumberOfProofsByContradiction = 0;
    Target = Head;
    while (OKSoFar && Target != NULL) {
        GetName(Target->AnnotatedFormula,FormulaName);
        GetParentNames(Target->AnnotatedFormula,AllParentNames);
        NumberOfParents = Tokenize(AllParentNames,ParentNames,"\n");
        if (NumberOfParents == 2) {
            FalseParent = GetAnnotatedFormulaFromListByName(Head,
ParentNames[0]);
            AncestorParent = GetAnnotatedFormulaFromListByName(Head,
ParentNames[1]);
            if (FalseAnnotatedFormula(FalseParent)) {
                AncestorName = ParentNames[1];
            } else if (FalseAnnotatedFormula(AncestorParent)) {
                SwapParent = FalseParent;
                FalseParent = AncestorParent;
                AncestorParent = SwapParent;
                AncestorName = ParentNames[0];
            } else {
                FalseParent = NULL;
            }
            if (FalseParent != NULL) {
                if (!IsAncestor(Head,AncestorParent,FalseParent,0)) {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s has a badly formed proof by contradiction\n",FormulaName);
                    OKSoFar = 0;
                } else {
                    (*NumberOfProofsByContradiction)++;
                    sprintf(ProcessedTag,"proved_by_contradiction(%s)",
AncestorName);
                    AddUsefulInfoTag(Target->AnnotatedFormula,Signature,
ProcessedTag);
                }
            }
        }
        Target = Target->Next;
    }

    return(OKSoFar);
}
//-----------------------------------------------------------------------------
ANNOTATEDFORMULA MakeSplitDefinitionFor(ANNOTATEDFORMULA SplitChild,
SIGNATURE Signature) {

    String FormulaName;
    ANNOTATEDFORMULA SplitDefinition;
    FORMULA Formula;
    String CreatorString;

//----Check it is a disjunction
    Formula = SplitChild->AnnotatedFormulaUnion.AnnotatedTSTPFormula.
FormulaWithVariables->Formula;
    if (Formula->Type != binary || 
Formula->FormulaUnion.BinaryFormula.Connective != disjunction || 
Formula->FormulaUnion.BinaryFormula.LHS->Type != atom ||
GetArity(Formula->FormulaUnion.BinaryFormula.LHS->FormulaUnion.Atom) != 0) {
        return(NULL);
    }

    strcpy(FormulaName,GetSymbol(Formula->FormulaUnion.BinaryFormula.LHS->
FormulaUnion.Atom));
    strcat(FormulaName,"_esplit_defn");
    SplitDefinition = DuplicateAnnotatedFormula(SplitChild,Signature);
    Formula = SplitDefinition->AnnotatedFormulaUnion.AnnotatedTSTPFormula.
FormulaWithVariables->Formula;
    SetSyntax(SplitDefinition,tptp_fof);
    SetName(SplitDefinition,FormulaName);
    SetStatus(SplitDefinition,definition,nonstatus);
    Formula->FormulaUnion.BinaryFormula.Connective = equivalence;
    UniversalQuantifyFormula(&(Formula->FormulaUnion.BinaryFormula.RHS),
SplitDefinition->AnnotatedFormulaUnion.AnnotatedTSTPFormula.
FormulaWithVariables->Variables);
    NegateFormula(&(Formula->FormulaUnion.BinaryFormula.RHS));
    EnsureShortForm(SplitDefinition);
    sprintf(CreatorString,"creator(gdv,[esplit_defn(%s)])",GetName(SplitChild,
FormulaName));
    SetSourceFromString(SplitDefinition,Signature,CreatorString);

    return(SplitDefinition);
}
//-----------------------------------------------------------------------------
int MakeESplitDefinitions(OptionsType OptionValues,LISTNODE Head,SIGNATURE
Signature,LISTNODE * SplitDefinitions) {

    LISTNODE Target;
    LISTNODE Sibling;
    LISTNODE SplitSiblings;
    int OKSoFar;
    String InferenceRule;
    String InferenceInfo;
    String ProcessedTag;
    String AllParentNames;
    String ParentName;
    String SiblingName;
    String AllSiblingParentNames;
    int SiblingNumber;
    ANNOTATEDFORMULA SplitDefinition;
    String SplitDefinitionName;
    String SplitSiblingsNames;

    *SplitDefinitions = NULL;
    OKSoFar = 1;
    Target = Head;
    while (OKSoFar && Target != NULL) {
        GetName(Target->AnnotatedFormula,SiblingName);
//----Check if derived by an esplit and not already processed
        GetInferenceRule(Target->AnnotatedFormula,InferenceRule);
        if (GetInferenceInfoTerm(Target->AnnotatedFormula,InferenceRule,
InferenceInfo) != NULL && ExtractTermArguments(InferenceInfo) &&
strstr(InferenceInfo,"esplit,") == InferenceInfo && GetUsefulInfoTerm(
Target->AnnotatedFormula,"psuedo_split_from",1,ProcessedTag) == NULL) {
//----Get the split parent
            GetParentNames(Target->AnnotatedFormula,AllParentNames);
//----Crude check that there's just one parent
            if (strchr(AllParentNames,'\n') != strrchr(AllParentNames,'\n')) {
                QPRINTF(OptionValues,2)(
"FAILURE: %s has been esplit from more than one parent\n",SiblingName);
                OKSoFar = 0;
            } else {
                SplitSiblings = NULL;
                AddListNode(&SplitSiblings,SplitSiblings,Target->
AnnotatedFormula);

                strcpy(ParentName,AllParentNames);
                *strchr(ParentName,'\n') = '\0';
                sprintf(ProcessedTag,"psuedo_split_from(%s)",ParentName);
                AddUsefulInfoTag(Target->AnnotatedFormula,Signature,
ProcessedTag);

                strcpy(SplitSiblingsNames,"psuedo_split_to(");
                strcat(SplitSiblingsNames,SiblingName);

//----Go through and look for split siblings (with same parent)
                Sibling = Target->Next;
                while (OKSoFar && Sibling != NULL) {
                    if (GetInferenceInfoTerm(Sibling->AnnotatedFormula,
InferenceRule,InferenceInfo) != NULL && ExtractTermArguments(InferenceInfo) &&
strstr(InferenceInfo,"esplit,") == InferenceInfo && GetUsefulInfoTerm(
Sibling->AnnotatedFormula,"psuedo_split_from",1,ProcessedTag) == NULL) {
                        GetName(Sibling->AnnotatedFormula,SiblingName);
                        GetParentNames(Sibling->AnnotatedFormula,
AllSiblingParentNames);
//----Another sibling
                        if (!strcmp(AllParentNames,AllSiblingParentNames)) {
                            AddListNode(&SplitSiblings,SplitSiblings,
Sibling->AnnotatedFormula);
                            AddUsefulInfoTag(Sibling->AnnotatedFormula,
Signature,ProcessedTag);
                            strcat(SplitSiblingsNames,",");
                            strcat(SplitSiblingsNames,SiblingName);
                        }
                    }
                    Sibling = Sibling->Next;
                }
                strcat(SplitSiblingsNames,")");
                AddUsefulInfoTag(GetAnnotatedFormulaFromListByName(Head,
ParentName),Signature,SplitSiblingsNames);

//----Make the definitions for all except the first in the list (which was
//----the last found and I assume to be the link clause)
                Sibling = SplitSiblings->Next;
                SiblingNumber = 1;
                while (OKSoFar && Sibling != NULL) {
                    GetName(Sibling->AnnotatedFormula,SiblingName);
                    if ((SplitDefinition = MakeSplitDefinitionFor(
Sibling->AnnotatedFormula,Signature)) == NULL) {
                        QPRINTF(OptionValues,2)(
"FAILURE: %s cannot be converted to a split definition\n",SiblingName);
                    } else {
                        AddListNode(SplitDefinitions,*SplitDefinitions,
SplitDefinition);
                        sprintf(ProcessedTag,"esplit_defn(%s)",
GetName(SplitDefinition,SplitDefinitionName));
                        AddUsefulInfoTag(Sibling->AnnotatedFormula,Signature,
ProcessedTag);
                    }
                    Sibling = Sibling->Next;
                }
                FreeListOfAnnotatedFormulae(&SplitSiblings);
            }
        }
    Target = Target->Next;
    }

//----Make sure that the definitions are unique by looking at their names
    if (!UniquelyNamed(*SplitDefinitions)) {
        QPRINTF(OptionValues,2)(
"FAILURE: Multiple definition of an esplit name\n");
        OKSoFar = 0;
    }

    if (!OKSoFar) {
        FreeListOfAnnotatedFormulae(SplitDefinitions);
    }

//DEBUG PrintListOfAnnotatedTSTPNodes(stdout,*SplitDefinitions,tptp);

    return(OKSoFar);
}
//-----------------------------------------------------------------------------
int GetSPASSSplitPosition(char * InferenceInfo,char * PutSplitPositionHere) {

    char * Start;

    strcpy(PutSplitPositionHere,InferenceInfo);
    if ((Start = strstr(PutSplitPositionHere,"position(")) != NULL &&
Start[strlen(Start)-1] == ']') {
        Start[strlen(Start)-1] = '\0';
        strcpy(PutSplitPositionHere,Start);
        return(ExtractTermArguments(PutSplitPositionHere));
    } else {
        return(0);
    }
}
//-----------------------------------------------------------------------------
int InsertSPASSSplitInfo(OptionsType OptionValues,LISTNODE Head,
SIGNATURE Signature,int * NumberOfExplicitSplits) {

    int OKSoFar;
    LISTNODE Target;
    LISTNODE Sibling;
    String InferenceRule;
    String InferenceInfo;
    String ProcessedTag;
    String AllParentNames;
    String ParentName;
    String SiblingName;
    String AllSiblingParentNames;
    String SplitPosition;
    String NextSplitPosition;
    String SplitSiblingsNames;

    OKSoFar = 1;
    *NumberOfExplicitSplits = 0;
    Target = Head;
    while (OKSoFar && Target != NULL) {
        GetName(Target->AnnotatedFormula,SiblingName);
//----Check if derived by an sr_split and not already processed
        GetInferenceRule(Target->AnnotatedFormula,InferenceRule);
        if (GetInferenceInfoTerm(Target->AnnotatedFormula,InferenceRule,
InferenceInfo) != NULL && ExtractTermArguments(InferenceInfo) &&
strstr(InferenceInfo,"sr_split,") == InferenceInfo && GetUsefulInfoTerm(
Target->AnnotatedFormula,"explicit_split_from",1,ProcessedTag) == NULL &&
//----Get the split position - must be a left
GetSPASSSplitPosition(InferenceInfo,SplitPosition) &&
SplitPosition[strlen(SplitPosition)-1] == 'l') {
//----Get the split parent
            GetParentNames(Target->AnnotatedFormula,AllParentNames);
//----Crude check that there's just one parent
            if (strchr(AllParentNames,'\n') != strrchr(AllParentNames,'\n')) {
                QPRINTF(OptionValues,2)(
"FAILURE: %s has been sr_split from more than one parent\n",SiblingName);
                OKSoFar = 0;
            }
            if (OKSoFar) {
                (*NumberOfExplicitSplits)++;

                strcpy(ParentName,AllParentNames);
                *strchr(ParentName,'\n') = '\0';
                sprintf(ProcessedTag,"explicit_split_from(%s)",ParentName);
                AddUsefulInfoTag(Target->AnnotatedFormula,Signature,
ProcessedTag);
                strcpy(SplitSiblingsNames,"explicit_split_to(");
                strcat(SplitSiblingsNames,SiblingName);

//----Look for the right child next
                strcpy(NextSplitPosition,SplitPosition);
                NextSplitPosition[strlen(NextSplitPosition)-1] = 'r';
//----Go through and look for split siblings (with same parent)
                Sibling = Target->Next;
                while (OKSoFar && Sibling != NULL) {
                    if (GetInferenceInfoTerm(Sibling->AnnotatedFormula,
InferenceRule,InferenceInfo) != NULL && ExtractTermArguments(InferenceInfo) &&
strstr(InferenceInfo,"sr_split,") == InferenceInfo && GetUsefulInfoTerm(
Sibling->AnnotatedFormula,"psuedo_split_from",1,ProcessedTag) == NULL) {
                        GetName(Sibling->AnnotatedFormula,SiblingName);
                        GetParentNames(Sibling->AnnotatedFormula,
AllSiblingParentNames);
//----The sibling or anti-kid
                        if (!strcmp(AllParentNames,AllSiblingParentNames) &&
GetSPASSSplitPosition(InferenceInfo,SplitPosition) &&
!strcmp(SplitPosition,NextSplitPosition)) {
                            AddUsefulInfoTag(Sibling->AnnotatedFormula,
Signature,ProcessedTag);
                            strcat(SplitSiblingsNames,",");
                            strcat(SplitSiblingsNames,SiblingName);
                            strcat(SplitSiblingsNames,")");
//----Stop the search and clear the required split position
                            Sibling = NULL;
                            strcpy(NextSplitPosition,"");
                        } else {
                            Sibling = Sibling->Next;
                        }
                    } else {
                        Sibling = Sibling->Next;
                    }
                }
//----By now the next required split position should be empty
                if (strcmp(NextSplitPosition,"")) {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s has no explicit split right child\n",ParentName);
                    OKSoFar = 0;
                } else {
                    AddUsefulInfoTag(GetAnnotatedFormulaFromListByName(Head,
ParentName),Signature,SplitSiblingsNames);
                }
            }
        }
        Target = Target->Next;
    }
    return(OKSoFar);
}
//-----------------------------------------------------------------------------
int ExplicitSplitsIndependent(OptionsType OptionValues,LISTNODE Head,
int * NumberOfExplicitSplits) {

    int OKSoFar;
    LISTNODE Target;
    LISTNODE InnerTarget;
    String FormulaName;
    int OccurenceNumber;
    String UsefulInfo;
    StringParts SplitChildrenNames;
    ANNOTATEDFORMULA SplitChild1;
    ANNOTATEDFORMULA SplitChild2;
    String InnerName;
    ANNOTATEDFORMULA RefutationRoot;
    LISTNODE FalseList;

//----Already checked that false root exists
    if (OptionValues.CheckRefutation) {
        RefutationRoot = HasFalseRoot(OptionValues,Head);
        FalseList = GetFalseNodes(Head);
    }
    OKSoFar = 1;
    *NumberOfExplicitSplits = 0;
    Target = Head;
    while (OKSoFar && Target != NULL) {
        OccurenceNumber = 1;
//----Look for explicitly split formulae
        while (GetUsefulInfoTerm(Target->AnnotatedFormula,"explicit_split_to",
OccurenceNumber,UsefulInfo) != NULL) {
            OccurenceNumber++;
            GetName(Target->AnnotatedFormula,FormulaName);
            (*NumberOfExplicitSplits)++;
            ExtractTermArguments(UsefulInfo);
            if (Tokenize(UsefulInfo,SplitChildrenNames,",") != 2) {
                CodingError("Wrong number of split children");
            }
            SplitChild1 = GetAnnotatedFormulaFromListByName(Head,
SplitChildrenNames[0]);
            SplitChild2 = GetAnnotatedFormulaFromListByName(Head,
SplitChildrenNames[1]);
//----No formula may have both split children as an ancestor
            InnerTarget = Head;
            while (OKSoFar && InnerTarget != NULL) {
                if (IsAncestor(Head,SplitChild1,InnerTarget->AnnotatedFormula,0) 
&& IsAncestor(Head,SplitChild2,InnerTarget->AnnotatedFormula,0)) {
                    OKSoFar = 0;
                    QPRINTF(OptionValues,2)(
"FAILURE: %s has both split siblings %s and %s as ancestors\n",
GetName(InnerTarget->AnnotatedFormula,InnerName),SplitChildrenNames[0],
SplitChildrenNames[1]);
                }
                InnerTarget = InnerTarget->Next;
            }
//----If child is an ancestor of the refutation node, not going through non-thm
//----Both sides must have a false descendant
            if (OptionValues.CheckRefutation &&
((IsAncestor(Head,SplitChild1,RefutationRoot,1) && 
!IsAncestorOfOneInList(Head,SplitChild2,FalseList,1)) ||
(IsAncestor(Head,SplitChild2,RefutationRoot,1) && 
!IsAncestorOfOneInList(Head,SplitChild1,FalseList,1)))) {
                OKSoFar = 0;
                QPRINTF(OptionValues,2)(
"FAILURE: One of the split siblings %s and %s is not refuted\n",
SplitChildrenNames[0],SplitChildrenNames[1]);
            }
        }
        Target = Target->Next;
    }
    if (OptionValues.CheckRefutation) {
        FreeListOfAnnotatedFormulae(&FalseList);
    }
    return(OKSoFar);
}
//-----------------------------------------------------------------------------
int StructuralVerification(OptionsType OptionValues,LISTNODE * Head,
SIGNATURE Signature) {

    int OKSoFar;
    LISTNODE SplitDefinitions;
    int NumberOfInstances;

    OKSoFar = 1;

    if (!GlobalInterrupted && OKSoFar) {
        if (UniquelyNamed(*Head)) {
            QPRINTF(OptionValues,2)(
"SUCCESS: Derivation has unique formula names\n");
        } else {
            QPRINTF(OptionValues,2)(
"FAILURE: Derivation has duplicate formula names\n");
            OKSoFar = 0;
        }
    }

    if (!GlobalInterrupted && OKSoFar) {
        if (AllParentsExist(OptionValues,*Head)) {
            QPRINTF(OptionValues,2)(
"SUCCESS: All derived formulae have parents and inference information\n");
        } else {
            OKSoFar = 0;
        }
    }

    if (!GlobalInterrupted && OKSoFar) {
        if (ACyclic(OptionValues,*Head)) {
            QPRINTF(OptionValues,2)("SUCCESS: Derivation is acyclic\n");
        } else {
            QPRINTF(OptionValues,2)("FAILURE: Derivation is cyclic\n");
            OKSoFar = 0;
        }
    }

//----Check a refutation has a false root (at least one)
    if (!GlobalInterrupted && OKSoFar && OptionValues.CheckRefutation) {
        if (HasFalseRoot(OptionValues,*Head) != NULL) {
            QPRINTF(OptionValues,2)(
"SUCCESS: Derivation could be a refutation\n");
        } else {
            OKSoFar = 0;
        }
    }

//----Check that all nodes that have a false parent have two parents, and
//----the second is an ancestor of the false
    if (!GlobalInterrupted && OKSoFar && OptionValues.CheckRefutation) {
        if (WellFormedProofsByContradiction(OptionValues,*Head,Signature,
&NumberOfInstances)) {
//----Report only if there are some
            if (NumberOfInstances > 0) {
                QPRINTF(OptionValues,2)(
"SUCCESS: Derivation has well formed proofs by contradiction\n");
            }
        } else {
            OKSoFar = 0;
        }
    }

//----Add definitions for E's psuedo splitting
    if (!GlobalInterrupted && OKSoFar) {
        if (MakeESplitDefinitions(OptionValues,*Head,Signature,
&SplitDefinitions)) {
//----Report only if there are some
            if (SplitDefinitions != NULL) {
                QPRINTF(OptionValues,2)(
"SUCCESS: All esplits have had definitions added\n");
                while (SplitDefinitions != NULL) {
                    AddListNode(Head,*Head,SplitDefinitions->AnnotatedFormula);
                    SplitDefinitions = SplitDefinitions->Next;
                }
                FreeListOfAnnotatedFormulae(&SplitDefinitions);
            }
        } else {
            OKSoFar = 0;
        }
    }

    if (!GlobalInterrupted && OKSoFar) {
        if (InsertSPASSSplitInfo(OptionValues,*Head,Signature,
&NumberOfInstances)) {
//----Report only if there are some
            if (NumberOfInstances > 0) {
                QPRINTF(OptionValues,2)(
"SUCCESS: All sr_splits have both children in place\n");
            }
        } else {
            OKSoFar = 0;
        }
    }

    if (!GlobalInterrupted && OKSoFar) {
        if (ExplicitSplitsIndependent(OptionValues,*Head,&NumberOfInstances)) {
//----Report only if there are some
            if (NumberOfInstances > 0) {
                QPRINTF(OptionValues,2)(
"SUCCESS: All explicit splits are independent and refuted\n");
            }
        } else {
            OKSoFar = 0;
        }
    }

    return(OKSoFar);
}
//-----------------------------------------------------------------------------
void GetSplitDefinitionNames(OptionsType OptionValues,LISTNODE Head,
ANNOTATEDFORMULA AnnotatedFormula,String SplitDefinitionNames) {

    String UsefulInfo;
    int NumberOfSiblings;
    StringParts SiblingNames;
    int SiblingNumber;
    String SiblingInfo;

    strcpy(SplitDefinitionNames,"");
    if (GetUsefulInfoTerm(AnnotatedFormula,"psuedo_split_to",1,UsefulInfo) == 
NULL) {
        CodingError("Missing psuedo_split_to term");
    }
    ExtractTermArguments(UsefulInfo);
    NumberOfSiblings = Tokenize(UsefulInfo,SiblingNames,",");
    for (SiblingNumber = 0;SiblingNumber < NumberOfSiblings;SiblingNumber++) {
        if (GetUsefulInfoTerm(GetAnnotatedFormulaFromListByName(Head,
SiblingNames[SiblingNumber]),"esplit_defn",1,SiblingInfo) != NULL) {
            ExtractTermArguments(SiblingInfo);
            strcat(SplitDefinitionNames,SiblingInfo);
            strcat(SplitDefinitionNames,"\n");
        }
    }
}
//-----------------------------------------------------------------------------
int RuleSpecificVerification(OptionsType OptionValues,LISTNODE Head,
SIGNATURE Signature) {

    LISTNODE Target;
    int AllOKSoFar;
    String UsefulInfo;
    int UserCheckParentRelevance;
    String FormulaName;
    int NumberOfParents;
    StringParts ParentNames;
    String ListParentNames;
    LISTNODE ParentAnnotatedFormulae;
    LISTNODE Parent;
    String SplitDefinitionNames;
    ANNOTATEDFORMULA SplitParent;
    int OccurenceNumber;

    Target = Head;
    AllOKSoFar = 1;
    while (!GlobalInterrupted && (AllOKSoFar || OptionValues.ForceContinue) &&
Target != NULL) {
        GetName(Target->AnnotatedFormula,FormulaName);
//DEBUG PrintAnnotatedTSTPNode(stdout,Target->AnnotatedFormula,tptp,1);

//----esplit parent to be inferred from its children, with relevance check
        if (GetUsefulInfoTerm(Target->AnnotatedFormula,"psuedo_split_to",1,
UsefulInfo) != NULL) {
            UserCheckParentRelevance = OptionValues.CheckParentRelevance;
            OptionValues.CheckParentRelevance = 1;
//----Get the split children's names
            ExtractTermArguments(UsefulInfo);
            NumberOfParents = Tokenize(UsefulInfo,ParentNames,",");
            MakePrintableList(ParentNames,NumberOfParents,ListParentNames);
//----Make list of split children
            MakeListOfParents(OptionValues,Head,FormulaName,
ParentNames,NumberOfParents,&ParentAnnotatedFormulae);
            if (!CorrectlyInferred(OptionValues,Target->AnnotatedFormula,
FormulaName,ParentAnnotatedFormulae,ListParentNames,"thm","")) {
                AllOKSoFar = 0;
            } 
//----Free the parents list
            FreeListOfAnnotatedFormulae(&ParentAnnotatedFormulae);
            OptionValues.CheckParentRelevance = UserCheckParentRelevance;
        }

//----esplit child, to be inferred from the parent and the definitions
        if (GetUsefulInfoTerm(Target->AnnotatedFormula,"psuedo_split_from",1,
UsefulInfo) != NULL) {
//----Get the split parent's name
            ExtractTermArguments(UsefulInfo);
//----Get the split definitions
            if ((SplitParent = GetAnnotatedFormulaFromListByName(Head,
UsefulInfo)) == NULL) {
                CodingError("Missing split parent");
            } 
            GetSplitDefinitionNames(OptionValues,Head,SplitParent,
SplitDefinitionNames);
            strcat(UsefulInfo,"\n");
            strcat(UsefulInfo,SplitDefinitionNames);
            NumberOfParents = Tokenize(UsefulInfo,ParentNames,"\n");
            MakePrintableList(ParentNames,NumberOfParents,ListParentNames);
//----Make list of split children
            MakeListOfParents(OptionValues,Head,FormulaName,ParentNames,
NumberOfParents,&ParentAnnotatedFormulae);
            if (!CorrectlyInferred(OptionValues,Target->AnnotatedFormula,
FormulaName,ParentAnnotatedFormulae,ListParentNames,"thm","")) {
                AllOKSoFar = 0;
            } else {
                AddVerifiedTag(Target->AnnotatedFormula,Signature,
"psuedo_split");
            }
//----Free the parents list
            FreeListOfAnnotatedFormulae(&ParentAnnotatedFormulae);
        }

//----sr_split parent negation to be inferred from its negated children, 
//----with relevance check
        OccurenceNumber = 1;
        while (GetUsefulInfoTerm(Target->AnnotatedFormula,"explicit_split_to",
OccurenceNumber,UsefulInfo) != NULL) {
            OccurenceNumber++;
            UserCheckParentRelevance = OptionValues.CheckParentRelevance;
            OptionValues.CheckParentRelevance = 1;
//----Get the split children's names
            ExtractTermArguments(UsefulInfo);
            NumberOfParents = Tokenize(UsefulInfo,ParentNames,",");
            MakePrintableList(ParentNames,NumberOfParents,ListParentNames);
//----Make list of split children
            MakeListOfParents(OptionValues,Head,FormulaName,
ParentNames,NumberOfParents,&ParentAnnotatedFormulae);
            NegateListOfAnnotatedTSTPNodes(ParentAnnotatedFormulae,0);
            Negate(Target->AnnotatedFormula,0);
            if (!CorrectlyInferred(OptionValues,Target->AnnotatedFormula,
FormulaName,ParentAnnotatedFormulae,ListParentNames,"thm",
"(Negated formulae for sr_split)")) {
                AllOKSoFar = 0;
            } else {
                Parent = ParentAnnotatedFormulae;
                while (Parent != NULL) {
                    AddVerifiedTag(Parent->AnnotatedFormula,Signature,
"explicit_split");
                    Parent = Parent->Next;
                }
            }
            Negate(Target->AnnotatedFormula,1);
            NegateListOfAnnotatedTSTPNodes(ParentAnnotatedFormulae,1);
//----Free the parents list
            FreeListOfAnnotatedFormulae(&ParentAnnotatedFormulae);
            OptionValues.CheckParentRelevance = UserCheckParentRelevance;
        }

//----Proofs by contradiction
        if (GetUsefulInfoTerm(Target->AnnotatedFormula,
"proved_by_contradiction",1,UsefulInfo) != NULL) {
//----Get the parent's name
            ExtractTermArguments(UsefulInfo);
            NumberOfParents = 1;
            ParentNames[0] = UsefulInfo;
            MakeListOfParents(OptionValues,Head,FormulaName,ParentNames,
NumberOfParents,&ParentAnnotatedFormulae);
            NegateListOfAnnotatedTSTPNodes(ParentAnnotatedFormulae,0);
            if (!CorrectlyInferred(OptionValues,Target->AnnotatedFormula,
FormulaName,ParentAnnotatedFormulae,UsefulInfo,"thm",
"(Negated parent for PbC)")) {
                AllOKSoFar = 0;
            } else {
                AddVerifiedTag(Target->AnnotatedFormula,Signature,
"proved_by_contradiction");
            }
            NegateListOfAnnotatedTSTPNodes(ParentAnnotatedFormulae,1);
//----Free the parents list
            FreeListOfAnnotatedFormulae(&ParentAnnotatedFormulae);
        }

        Target = Target->Next;
    }

    return(AllOKSoFar);
}
//-----------------------------------------------------------------------------
void SplitIntoSatisfiableLists(OptionsType OptionValues, LISTNODE Head,
LISTNODE * PositiveHead,LISTNODE * NegativeHead,LISTNODE * NeitherHead) {

    int PositiveOrNegative;
    FILE * ObligationFile;
    String ObligationFileName;

    *PositiveHead = NULL;
    *NegativeHead = NULL;
    *NeitherHead = NULL;
    while (Head != NULL) {
        PositiveOrNegative = 0;
        if (AnnotatedFormulaTrueInInterpretation(Head->AnnotatedFormula,
positive)) {
            AddListNode(PositiveHead,*PositiveHead,Head->AnnotatedFormula);
            PositiveOrNegative = 1;
        }
        if (AnnotatedFormulaTrueInInterpretation(Head->AnnotatedFormula,
negative)) {
            AddListNode(NegativeHead,*NegativeHead,Head->AnnotatedFormula);
            PositiveOrNegative = 1;
        }
        if (!PositiveOrNegative) {
            AddListNode(NeitherHead,*NeitherHead,Head->AnnotatedFormula);
        }
        Head = Head->Next;
    }

//----Check if Neither list is satisfiable
    if ((ObligationFile = OpenFileInMode(OptionValues,
OptionValues.KeepFilesDirectory,"problem","neither.obl","w",
ObligationFileName)) == NULL) {
        FreeListOfAnnotatedFormulae(NeitherHead);
    } else {
        PrintListOfAnnotatedTSTPNodesWithStatus(ObligationFile,*NeitherHead,
axiom);
        fclose(ObligationFile);
        OptionValues.KeepFiles = 0;
        if (CheckSatisfiable(OptionValues,ObligationFileName)) {
            QPRINTF(OptionValues,0)(
"WOOHOOO: All problem formulae are in a satisfiable set\n");
        } else {
            QPRINTF(OptionValues,2)(
"WARNING: Some problem formulae are not in a satisfiable set\n");
//----Free the neither list (making the head NULL) if no use
            FreeListOfAnnotatedFormulae(NeitherHead);
        }
    }
}
//-----------------------------------------------------------------------------
int LeafVerification(OptionsType OptionValues,LISTNODE Head,SIGNATURE
Signature) {

    StatusType SubStatus;
    LISTNODE ProblemHead;
    LISTNODE PositiveHead;
    LISTNODE NegativeHead;
    LISTNODE NeitherHead;
    LISTNODE Target;
    String FormulaName;
    String UsefulInfo;
    LISTNODE ProblemTarget;
    String ProblemTargetName;
    int AllOKSoFar;
    int ThisOneOK;

//----Get the problem file into memory
    if ((ProblemHead = ParseFileOfFormulae(OptionValues.ProblemFileName,NULL,
Signature,1,NULL)) == NULL) {
        QPRINTF(OptionValues,1)(
"ERROR: Could not parse problem file %s\n",OptionValues.ProblemFileName);
        QPRINTF(OptionValues,2)("FAILURE: Leaf nodes unverified\n");
        return(0);
    }
    if (OptionValues.Quietness == 0) {
        printf("Problem file contents:\n");
        PrintListOfAnnotatedTSTPNodes(stdout,ProblemHead,tptp);
    }

//----Split problem into satisfiable lists
    SplitIntoSatisfiableLists(OptionValues,ProblemHead,&PositiveHead,
&NegativeHead,&NeitherHead);

//----For each derivation node, check if the same as a problem node, or
//----can be inferred from one of the satisfiable lists
    Target = Head;
    AllOKSoFar = 1;
    while (!GlobalInterrupted && (AllOKSoFar || OptionValues.ForceContinue) &&
Target != NULL) {
        if (!(GetStatus(Target->AnnotatedFormula,&SubStatus) == derived ||
SubStatus == derived) && GetUsefulInfoTerm(Target->AnnotatedFormula,"verified",
1,UsefulInfo) == NULL) {
            GetName(Target->AnnotatedFormula,FormulaName);
//----Don't verify definitions inserted by GDV
            if (GetSourceTerm(Target->AnnotatedFormula,UsefulInfo) != NULL && 
strstr(UsefulInfo,"creator(gdv") == UsefulInfo) {
                QPRINTF(OptionValues,1)(
"WARNING: %s is a leaf created by GDV, and not verified\n",FormulaName);
                ThisOneOK = 1;
            } else {
                ThisOneOK = 0;

//----Look for an original that has been copied
                ProblemTarget = ProblemHead;
                while (!ThisOneOK && ProblemTarget != NULL) {
                    if (SameFormulaInAnnotatedFormulae(Target->AnnotatedFormula,
ProblemTarget->AnnotatedFormula,1)) {
                        QPRINTF(OptionValues,2)(
"SUCCESS: %s is a copy of %s\n",FormulaName,
GetName(ProblemTarget->AnnotatedFormula,ProblemTargetName));
                        ThisOneOK = 1;
                    }
                    ProblemTarget = ProblemTarget->Next;
                }

//----If not a copy, try some inferencing
                OptionValues.CheckParentRelevance = 0;
                if (!ThisOneOK && PositiveHead != NULL && 
CorrectlyInferred(OptionValues,Target->AnnotatedFormula,FormulaName,
PositiveHead,NULL,"thm","")) {
                    QPRINTF(OptionValues,2)(
"SUCCESS: %s is a thm of the positive input formulae\n",FormulaName);
                    ThisOneOK = 1;
                }
                if (!ThisOneOK && NegativeHead != NULL && 
CorrectlyInferred(OptionValues,Target->AnnotatedFormula,FormulaName,
NegativeHead,NULL,"thm","")) {
                    QPRINTF(OptionValues,2)(
"SUCCESS: %s is a thm of the negative input formulae\n",FormulaName);
                    ThisOneOK = 1;
                }
                if (!ThisOneOK && NeitherHead != NULL && 
CorrectlyInferred(OptionValues,Target->AnnotatedFormula,FormulaName,
NeitherHead,NULL,"thm","")) {
                    QPRINTF(OptionValues,2)(
"SUCCESS: %s is a thm of the neither input formulae\n",FormulaName);
                    ThisOneOK = 1;
                }

                if (ThisOneOK) {
                    AddVerifiedTag(Target->AnnotatedFormula,Signature,"leaf");
                } else {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s cannot be shown to be a thm of the input formulae\n",FormulaName);
                }

                AllOKSoFar *= ThisOneOK;
            }
        }
        Target = Target->Next;
    }

//----Free the satisfiable lists and the problem list
    FreeListOfAnnotatedFormulae(&PositiveHead);
    FreeListOfAnnotatedFormulae(&NegativeHead);
    FreeListOfAnnotatedFormulae(&NeitherHead);
    FreeListOfAnnotatedFormulae(&ProblemHead);

    return(AllOKSoFar);
}
//-----------------------------------------------------------------------------
int DerivedVerification(OptionsType OptionValues,LISTNODE Head,
SIGNATURE Signature) {

    extern int GlobalInterrupted;
    LISTNODE Target;
    String VerifiedTag;
    int AllOKSoFar;
    StatusType SubStatus;
    String FormulaName;
    String InferenceRule;
    String AllParentNames;
    String ListParentNames;
    StringParts ParentNames;
    int NumberOfParents;
    LISTNODE ParentAnnotatedFormulae;
    String SZSStatus;

    Target = Head;
    AllOKSoFar = 1;
//----Work through list looking for derived nodes
    while (!GlobalInterrupted && (AllOKSoFar || OptionValues.ForceContinue) &&
Target != NULL) {
        if ((GetStatus(Target->AnnotatedFormula,&SubStatus) == derived ||
SubStatus == derived) && GetUsefulInfoTerm(Target->AnnotatedFormula,
"verified",1,VerifiedTag) == NULL) {
            GetName(Target->AnnotatedFormula,FormulaName);

            GetInferenceRule(Target->AnnotatedFormula,InferenceRule);
//----Get the parents' in various ways
            GetParentNames(Target->AnnotatedFormula,AllParentNames);
            NumberOfParents = Tokenize(AllParentNames,ParentNames,"\n");
            MakePrintableList(ParentNames,NumberOfParents,ListParentNames);
            MakeListOfParents(OptionValues,Head,FormulaName,
ParentNames,NumberOfParents,&ParentAnnotatedFormulae);

//----Copied formula
            if (!strcmp(InferenceRule,"")) {
                if (SameFormulaInAnnotatedFormulae(
Target->AnnotatedFormula,ParentAnnotatedFormulae->AnnotatedFormula,1)) {
                    QPRINTF(OptionValues,2)(
"SUCCESS: %s is a copy of %s\n",FormulaName,ParentNames[0]);
                } else {
                    QPRINTF(OptionValues,2)(
"FAILURE: %s is not a copy of %s\n",FormulaName,ParentNames[0]);
                    AllOKSoFar = 0;
                }

//----Inferred formula
            } else {
//----Get SZS status
                if (GetInferenceInfoTerm(Target->AnnotatedFormula,
"status",SZSStatus) == NULL || ! ExtractTermArguments(SZSStatus)) {
                    QPRINTF(OptionValues,1)(
"WARNING: Cannot get SZS status for %s, assuming thm\n",FormulaName);
                    strcpy(SZSStatus,"thm");
                }
                if (!CorrectlyInferred(OptionValues,
Target->AnnotatedFormula,FormulaName,ParentAnnotatedFormulae,ListParentNames,
SZSStatus,"")) {
                    AllOKSoFar = 0;
                } else {
                    AddVerifiedTag(Target->AnnotatedFormula,Signature,
SZSStatus);
                }
            }
//----Free the parents list
            FreeListOfAnnotatedFormulae(&ParentAnnotatedFormulae);
        }
        Target = Target->Next;
    }

    return(AllOKSoFar);
}
//-----------------------------------------------------------------------------
void GetProblemFileName(ANNOTATEDFORMULA AnnotatedFormula,
char * ProblemFileName) {

    String PossibleFileName;

//----If no name given look in first formula
    if (strcmp(ProblemFileName,"")) {
        strcpy(PossibleFileName,ProblemFileName);
    } else if (GetFileSourceNameAndNode(AnnotatedFormula,PossibleFileName) != 
NULL) {
        *strchr(PossibleFileName,'\n') = '\0';
        if (PossibleFileName[0] == '\'') {
            PossibleFileName[strlen(PossibleFileName)-1] = '\0';
            strcpy(PossibleFileName,PossibleFileName+1);
        }
    } else {
        strcpy(PossibleFileName,"");
    }

//----Check that it's a readable file
    if (strcmp(PossibleFileName,"") && access(PossibleFileName,R_OK) == 0) {
        strcpy(ProblemFileName,PossibleFileName);
    }
} 
//-----------------------------------------------------------------------------
void ConvertToFOF(OptionsType OptionValues,LISTNODE Head) {

    LISTNODE Target;

    Target = Head;
    while (Target != NULL) {
        if (!UniversalFOFify(Target->AnnotatedFormula)) {
            QPRINTF(OptionValues,4)("ERROR: Could not FOFify\n");
            exit(EXIT_FAILURE);
        }
        Target = Target->Next;
    }
    if (OptionValues.Quietness == 0) {
        printf("Derivation file contents in FOF:\n");
        PrintListOfAnnotatedTSTPNodes(stdout,Head,tptp);
    }

}
//-----------------------------------------------------------------------------
int main(int argc,char * argv[]) {

    extern int GlobalInterrupted;
    OptionsType OptionValues;
    LISTNODE Head;
    SIGNATURE Signature;
    int OKSoFar;
    String UNIXCommand;

    GlobalInterrupted = 0;
    if (signal(SIGQUIT,GlobalInterruptHandler) == SIG_ERR ||
signal(SIGQUIT,GlobalInterruptHandler) == SIG_ERR) {
        perror("Setting interrupt handler");
        exit(EXIT_FAILURE);
    }

    sprintf(UNIXCommand,"%s/%s",TPTP_HOME,SYSTEM_ON_TPTP);
    if (access(UNIXCommand,X_OK) == -1) {
        perror(UNIXCommand);
        exit(EXIT_FAILURE);
    }

    if (!ProcessCommandLine(argc,argv,&OptionValues)) {
        QPRINTF(OptionValues,4)("ERROR: Invalid command line arguments\n");
        exit(EXIT_FAILURE);
    }

//----Read the derivation file
    Signature = NewSignature();
    SetNeedForNonLogicTokens(0);
    if ((Head = ParseFileOfFormulae(OptionValues.DerivationFileName,NULL,
Signature,1,NULL)) == NULL) {
        QPRINTF(OptionValues,4)(
"ERROR: Could not parse %s\n",OptionValues.DerivationFileName);
        exit(EXIT_FAILURE);
    }
    if (OptionValues.Quietness == 0) {
        printf("Derivation file contents:\n");
        PrintListOfAnnotatedTSTPNodes(stdout,Head,tptp);
    }
//----Get problem file name sorted out
    GetProblemFileName(Head->AnnotatedFormula,OptionValues.ProblemFileName);

//----Create working directory
    if (!GlobalInterrupted) {
        if (!CreateDirectory(OptionValues.KeepFilesDirectory,
OptionValues.DerivationFileName)) {
            QPRINTF(OptionValues,4)(
"ERROR: Could not create working directory %s\n",
OptionValues.KeepFilesDirectory);
            exit(EXIT_FAILURE);
        }
    }

//----Start verification
    OKSoFar = 1;

//----Structural verification - failure cannot be forced past
    if (!GlobalInterrupted && (OKSoFar || OptionValues.ForceContinue)) {
        QPRINTF(OptionValues,0)("Start structural verification\n");
        if (!StructuralVerification(OptionValues,&Head,Signature)) {
            OKSoFar = 0;
            if (OptionValues.ForceContinue) {
                OptionValues.ForceContinue = 0;
                QPRINTF(OptionValues,2)(
"FAILURE: Structural failure, cannot be forced to continue\n");
            }
        }
    }
//DEBUG PrintListOfAnnotatedTSTPNodes(stdout,Head,0);

//----Convert to FOF for semantic parts
    if (!GlobalInterrupted && (OKSoFar || OptionValues.ForceContinue)) {
        QPRINTF(OptionValues,0)("Convert to FOF\n");
        ConvertToFOF(OptionValues,Head);
    }

//----Inference rule specific parts
    if (!GlobalInterrupted && (OKSoFar || OptionValues.ForceContinue)) {
        QPRINTF(OptionValues,0)("Start rule specific verification\n");
        OKSoFar *= RuleSpecificVerification(OptionValues,Head,Signature);
    }

//----Initials verification
    if (!GlobalInterrupted && (OKSoFar || OptionValues.ForceContinue) &&
OptionValues.VerifyLeaves) {
        QPRINTF(OptionValues,0)("Start leaf verification\n");
        OKSoFar *= LeafVerification(OptionValues,Head,Signature);
    }

//----Derived verification
    if (!GlobalInterrupted && (OKSoFar || OptionValues.ForceContinue)) {
        QPRINTF(OptionValues,0)("Start derived verification\n");
        OKSoFar *= DerivedVerification(OptionValues,Head,Signature);
    }

//----Delete working directory if not longer needed
    if (!OptionValues.KeepFiles) {
        QPRINTF(OptionValues,0)("Clean up files\n");
        EmptyAndDeleteDirectory(OptionValues.KeepFilesDirectory);
    }

//----Free memory
    FreeListOfAnnotatedFormulae(&Head);
    FreeSignature(&Signature);
    if (GlobalInterrupted) {
        OKSoFar = 0;
    }

    QPRINTF(OptionValues,3)("%s\n",OKSoFar?"SUCCESS":"FAILURE");

    return(EXIT_SUCCESS);
}
//-----------------------------------------------------------------------------
