#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "Examine.h"
#include "Parsing.h"
#include "PrintTSTP.h"
#include "List.h"
//-----------------------------------------------------------------------------
//----Skip comments, blank lines, etc
LISTNODE GetLogicNode(LISTNODE Head) {

    while (Head != NULL && Head->AnnotatedFormula->Syntax != tptp_fof &&
Head->AnnotatedFormula->Syntax != tptp_cnf) {
        Head = Head->Next;
    }

    return(Head);
}
//-----------------------------------------------------------------------------
LISTNODE NewListNode(ANNOTATEDFORMULA AnnotatedFormula) {

    LISTNODE NewNode;

    NewNode = (LISTNODE)Malloc(sizeof(ListNodeType));
    NewNode->Next = NULL;
    NewNode->AnnotatedFormula = AnnotatedFormula;
    NewNode->AnnotatedFormula->NumberOfUses++;

    return(NewNode);
}
//-----------------------------------------------------------------------------
void AddListNode(LISTNODE * From,LISTNODE Next,ANNOTATEDFORMULA
AnnotatedFormulae) {

    LISTNODE NewNode;

    NewNode = NewListNode(AnnotatedFormulae);
    NewNode->Next = Next;
    *From = NewNode;
}
//-----------------------------------------------------------------------------
void FreeListNode(LISTNODE * FreeThis) {

    FreeAnnotatedFormula(&((*FreeThis)->AnnotatedFormula));
    Free((void **)FreeThis);
}
//-----------------------------------------------------------------------------
LISTNODE DuplicateListOfNodes(LISTNODE Head,int KeepNonLogicNodes) {

    LISTNODE RootList;
    LISTNODE * Current;
    LISTNODE Target;

    RootList = NULL;
    Current = &RootList;
    Target = Head;
    if (!KeepNonLogicNodes) {
        Target = GetLogicNode(Target);
    }
    while (Target != NULL) {
        AddListNode(Current,NULL,Target->AnnotatedFormula);
        Current = &((*Current)->Next);
        Target = Target->Next;
        if (!KeepNonLogicNodes) {
            Target = GetLogicNode(Target);
        }
    }

    return(RootList);
}
//-----------------------------------------------------------------------------
void FreeListOfAnnotatedFormulae(LISTNODE * Head) {

    if (*Head != NULL) {
        if ((*Head)->Next != NULL) {
            FreeListOfAnnotatedFormulae(&((*Head)->Next));
        }
        FreeListNode(Head);
    }
}
//-----------------------------------------------------------------------------
//----Second must be address of malloced String so it can be realloced 
//----if necessary
char * GetAllNames(LISTNODE Head,char ** Names) {

    String Name;
    int NamesMax = STRINGLENGTH;

    strcpy(*Names,"");
    while ((Head = GetLogicNode(Head)) != NULL) {
        GetName(Head->AnnotatedFormula,Name);
        strcat(Name,"\n");
        ExtendString(Names,Name,&NamesMax);
        Head = Head->Next;
    }

    return(*Names);
}
//-----------------------------------------------------------------------------
int UniquelyNamed(LISTNODE Head) {

    char * AllNames;
    StringParts Names;
    int NumberOfNames;
    int OKSoFar;
    int Original;
    int Copy;

    AllNames = (char *)Malloc(sizeof(String));
    GetAllNames(Head,&AllNames);
    NumberOfNames = Tokenize(AllNames,Names,"\n");
    OKSoFar = 1;

    for (Original = 0;OKSoFar && Original < NumberOfNames;Original++) {
        for (Copy = Original+1;OKSoFar && Copy < NumberOfNames;Copy++) {
            if (!strcmp(Names[Original],Names[Copy])) {
                OKSoFar = 0;
            }
        }
    }
    Free((void **)&AllNames);

    return(OKSoFar);
}
//-----------------------------------------------------------------------------
LISTNODE * GetNodeFromListByNumber(LISTNODE * Head,int Number) {

    do {
//----Skip non-formula records
        while ((*Head) != NULL && 
(*Head)->AnnotatedFormula->Syntax != tptp_fof &&
(*Head)->AnnotatedFormula->Syntax != tptp_cnf) {
            Head = &((*Head)->Next);
        }
//----Is it the one we want?
        if ((*Head) != NULL) {
            if(Number == 1) {
                return(Head);
            } else {
//----If not, move on
                Head = &((*Head)->Next);
                Number--;
            }
        }
    } while ((*Head) != NULL);

    return(NULL);
}
//-----------------------------------------------------------------------------
ANNOTATEDFORMULA GetAnnotatedFormulaFromListByNumber(LISTNODE Head,int Number) {

    LISTNODE * NodePointer;

    if ((NodePointer = GetNodeFromListByNumber(&Head,Number)) !=
NULL) {
        assert((*NodePointer) != NULL);
        return((*NodePointer)->AnnotatedFormula);
    } else {
        return(NULL);
    }
}
//-----------------------------------------------------------------------------
//----The first parameter is a pointer to the pointer to the node in the real 
//----data structure, so that the function can return a pointer to the real
//----pointer in the data structure. This would not be possible if a pointer 
//----to the node were received, as then it would be in a local variable and
//----not be anything in the data structure.
LISTNODE * GetNodeFromListByAnnotatedFormulaName(LISTNODE * Head,
char * Name) {

    String FormulaName;

    do {
//----Skip non-formula records
        while ((*Head) != NULL && 
(*Head)->AnnotatedFormula->Syntax != tptp_fof &&
(*Head)->AnnotatedFormula->Syntax != tptp_cnf) {
            Head = &((*Head)->Next);
        }
//----Is it the one we want?
        if ((*Head) != NULL) {
            if (!strcmp(GetName((*Head)->AnnotatedFormula,FormulaName),Name)) {
                return(Head);
            } else {
//----If not, move on
                Head = &((*Head)->Next);
            }
        }
    } while ((*Head) != NULL);

    return(NULL);
}
//-----------------------------------------------------------------------------
ANNOTATEDFORMULA GetAnnotatedFormulaFromListByName(LISTNODE Head, char * Name) {

    LISTNODE * NodePointer;

    if ((NodePointer = GetNodeFromListByAnnotatedFormulaName(&Head,Name)) !=
NULL) {
        assert((*NodePointer) != NULL);
        return((*NodePointer)->AnnotatedFormula);
    } else {
        return(NULL);
    }
}
//-----------------------------------------------------------------------------
