#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "DataTypes.h"
#include "Utilities.h"
#include "Tokenizer.h"
#include "Parsing.h"
#include "Signature.h"
#include "Examine.h"
#include "Modify.h"
#include "List.h"
#include "ListStatistics.h"
#include "PrintTSTP.h"
#include "ParseTSTP.h"
//-----------------------------------------------------------------------------
LISTNODE * FindPointerToFirstFormula(LISTNODE * Head) {

    while (*Head != NULL && 
!LogicalAnnotatedFormula((*Head)->AnnotatedFormula)) {
        Head = &((*Head)->Next);
    }
    return(Head);
}
//-----------------------------------------------------------------------------
LISTNODE MakeBlankLine(void) {

    static ANNOTATEDFORMULA BlankLine = NULL;

//----Make it once only and reuse
    if (BlankLine == NULL) {
        BlankLine = NewAnnotatedFormula(blank_line);
    }

    return(NewListNode(BlankLine));
    
}
//-----------------------------------------------------------------------------
LISTNODE MakeReflexivity(SyntaxType Syntax,SIGNATURE Signature) {

    String Axiom;

    if (Syntax == tptp_fof) {
        sprintf(Axiom,"fof(reflexivity,axiom,\n    (! [X] : equal(X,X) )).");
    }
    if (Syntax == tptp_cnf) {
        sprintf(Axiom,"cnf(reflexivity,axiom,\n    ( equal(X,X) )).");
    }
//DEBUG printf("The string reflexivity is ==%s==\n",Axiom);
    return(ParseStringOfFormulae(Axiom,Signature,0,NULL));
}
//-----------------------------------------------------------------------------
LISTNODE MakeSymmetry(SyntaxType Syntax,SIGNATURE Signature) {

    String Axiom;

    if (Syntax == tptp_fof) {
        sprintf(Axiom,"fof(symmetry,axiom,\n    (! [X,Y] : (equal(X,Y) => equal(Y,X)))).");
    }
    if (Syntax == tptp_cnf) {
        sprintf(Axiom,"cnf(symmetry,axiom,\n    (~equal(X,Y) | equal(Y,X))).");
    }
//DEBUG printf("The string symmetry is ==%s==\n",Axiom);
    return(ParseStringOfFormulae(Axiom,Signature,0,NULL));
}
//-----------------------------------------------------------------------------
LISTNODE MakeTransitivity(SyntaxType Syntax,SIGNATURE Signature) {

    String Axiom;

    if (Syntax == tptp_fof) {
        sprintf(Axiom,"fof(transitivity,axiom,\n    (! [X,Y,Z] : ( (equal(X,Y) & equal(Y,Z) ) => equal(X,Z)))).");
    }
    if (Syntax == tptp_cnf) {
        sprintf(Axiom,"cnf(transitivity,axiom,\n    (~equal(X,Y) | ~equal(Y,Z) | equal(X,Z))).");
    }
//DEBUG printf("The string transitivity is ==%s==\n",Axiom);
    return(ParseStringOfFormulae(Axiom,Signature,0,NULL));
}
//-----------------------------------------------------------------------------
void MakeTermSubstitutionPair(char * Symbol,int Arity,int Position,
String TermX,String TermY,String Variables) {

    String IndexString;
    int Index;

    sprintf(TermX,"%s(",Symbol);
    sprintf(TermY,"%s(",Symbol);
    strcpy(Variables,"");
    for (Index=1;Index <= Arity;Index++) {
        if (Index > 1) {
            strcat(TermX,",");
            strcat(TermY,",");
            strcat(Variables,",");
        }
        if (Index == Position) {
            strcat(TermX,"X");
            strcat(TermY,"Y");
            strcat(Variables,"X,Y");
        } else {
            strcat(TermX,"Z");
            strcat(TermY,"Z");
            strcat(Variables,"Z");
            sprintf(IndexString,"%d",Index);
            strcat(TermX,IndexString);
            strcat(TermY,IndexString);
            strcat(Variables,IndexString);
        }
    }
    strcat(TermX,")");
    strcat(TermY,")");

//DEBUG printf("pair %d of %d for %s are ==%s==%s==\n",Position,Arity,Symbol,TermX,TermY);
}
//-----------------------------------------------------------------------------
LISTNODE MakeFunctionSubstitution(char * Symbol,int Arity,int Position,
SyntaxType Syntax,SIGNATURE Signature) {

    String Axiom;
    String TermX,TermY,Variables;

    MakeTermSubstitutionPair(Symbol,Arity,Position,TermX,TermY,Variables);
    if (Syntax == tptp_fof) {
        sprintf(Axiom,"fof(%s_substitution_%d,axiom,\n    (! [%s] : ( equal(X,Y) => equal(",Symbol,Position,Variables);
        strcat(Axiom,TermX);
        strcat(Axiom,",");
        strcat(Axiom,TermY);
        strcat(Axiom,")))).");
    }
    if (Syntax == tptp_cnf) {
        sprintf(Axiom,"cnf(%s_substitution_%d,axiom,\n    (~equal(X,Y) | equal(%s,%s))).",Symbol,Position,TermX,TermY);
    }
//DEBUG printf("The string axiom is ==%s==\n",Axiom);

    return(ParseStringOfFormulae(Axiom,Signature,0,NULL));
}
//-----------------------------------------------------------------------------
LISTNODE MakePredicateSubstitution(char * Symbol,int Arity,int Position,
SyntaxType Syntax,SIGNATURE Signature) {

    String Axiom;
    String TermX,TermY,Variables;

    MakeTermSubstitutionPair(Symbol,Arity,Position,TermX,TermY,Variables);
    if (Syntax == tptp_fof) {
        sprintf(Axiom,"fof(%s_substitution_%d,axiom,\n    (![%s] : ( (equal(X,Y) & %s ) => %s))).",Symbol,Position,Variables,TermX,TermY);
    }
    if (Syntax == tptp_cnf) {
        sprintf(Axiom,"cnf(%s_substitution_%d,axiom,\n    (~equal(X,Y) | ~%s | %s)).",Symbol,Position,TermX,TermY);
    }
//DEBUG printf("The string axiom is ==%s==\n",Axiom);

    return(ParseStringOfFormulae(Axiom,Signature,0,NULL));

}
//-----------------------------------------------------------------------------
LISTNODE * LinkInBefore(LISTNODE * BeforeNode,LISTNODE NewNode) {

    NewNode->Next = *BeforeNode;
    *BeforeNode = NewNode;
    return(&(NewNode->Next));

}
//-----------------------------------------------------------------------------
void AddEqualityFormulae(char * SymbolUsage,char * Functors,SyntaxType Syntax,
LISTNODE * BeforeNode,SIGNATURE Signature) {

    char * Symbol;
    char * NextSymbol;
    char * Slash;
    char * NextSlash;
    int Arity;
    int Index;

    BeforeNode = LinkInBefore(BeforeNode,MakeReflexivity(Syntax,Signature));
    BeforeNode = LinkInBefore(BeforeNode,MakeBlankLine());
    BeforeNode = LinkInBefore(BeforeNode,MakeSymmetry(Syntax,Signature));
    BeforeNode = LinkInBefore(BeforeNode,MakeBlankLine());
    BeforeNode = LinkInBefore(BeforeNode,MakeTransitivity(Syntax,Signature));
    BeforeNode = LinkInBefore(BeforeNode,MakeBlankLine());

    Symbol = strtok_r(Functors,"\n",&NextSymbol);
    while (Symbol != NULL) {
        Slash = strtok_r(Symbol,"/",&NextSlash);
        Slash = strtok_r(NULL,"/",&NextSlash);
        Arity = atoi(Slash);
        for (Index=1;Index <= Arity;Index++) {
//DEBUG printf("Making function substitution for %s position %d\n",Symbol,Index);
            BeforeNode = LinkInBefore(BeforeNode,MakeFunctionSubstitution(
Symbol,Arity,Index,Syntax,Signature));
            BeforeNode = LinkInBefore(BeforeNode,MakeBlankLine());
        }
        Symbol = strtok_r(NULL,"\n",&NextSymbol);
    }
    *Functors = '\0';
    Symbol = strtok_r(SymbolUsage,"\n",&NextSymbol);
    while (Symbol != NULL) {
        Slash = strtok_r(Symbol,"/",&NextSlash);
        Slash = strtok_r(NULL,"/",&NextSlash);
        Arity = atoi(Slash);
        for (Index=1;Index <= Arity;Index++) {
            if (strcmp(Symbol,"equal")) {
//DEBUG printf("Making predicate substitution for %s position %d\n",Symbol,Arity);
                BeforeNode = LinkInBefore(BeforeNode,MakePredicateSubstitution(
Symbol,Arity,Index,Syntax,Signature));
                BeforeNode = LinkInBefore(BeforeNode,MakeBlankLine());
            }
        }
        Symbol = strtok_r(NULL,"\n",&NextSymbol);
    }
}
//-----------------------------------------------------------------------------
int main(int argc, char *argv[]) {

    SIGNATURE Signature;
    LISTNODE Head;
    PrintFormatType Format;
    int ArgOffset;
    int KeepNonLogicals;
    char * SymbolUsage;
    char * Functors;
    LISTNODE * BeforeNode;
    int NumberOfFOF;
    int NumberOfCNF;

    ArgOffset = 0;

//----Change output length if specified
    Format = tptp;
    if (argc > ArgOffset+3 && !strcmp(argv[ArgOffset+1],"-f")) {
        Format = StringToPrintFormat(argv[ArgOffset+2]);
        ArgOffset += 2;
    }

//----Decide whether or not to keep non-logicals
    if (argc > ArgOffset+2 && !strcmp(argv[ArgOffset+1],"-k")) {
        KeepNonLogicals = 1;
        ArgOffset++;
    } else {
        KeepNonLogicals = 0;
    }

//----One signature for all testing
    Signature = NewSignature();

//----Reading formulae one-by-one from stdin. 
    SetNeedForNonLogicTokens(KeepNonLogicals);
    if ((Head = ParseFileOfFormulae(argv[ArgOffset+1],NULL,Signature,1,
NULL)) != NULL) {
        NumberOfFOF = (int)ListCount(Head,fof_nodes);
        NumberOfCNF = (int)ListCount(Head,cnf_nodes);
        if ((BeforeNode = FindPointerToFirstFormula(&Head)) == NULL) {
            printf("ERROR: No formulae\n");
            exit(EXIT_FAILURE);
        }
//DEBUG printf("First formula\n");
//DEBUG PrintAnnotatedTSTPNode(stdout,(*BeforeNode)->AnnotatedFormula,tptp,1);
        SymbolUsage = (char *)Malloc(sizeof(String));
        GetListOfAnnotatedFormulaSymbolUsage(Head,&SymbolUsage,&Functors);
        if (NumberOfFOF > 0 && NumberOfCNF == 0) {
            AddEqualityFormulae(SymbolUsage,Functors,tptp_fof,BeforeNode,
Signature);
        } else if (NumberOfFOF == 0 && NumberOfCNF > 0) {
            AddEqualityFormulae(SymbolUsage,Functors,tptp_cnf,BeforeNode,
Signature);
        } else {
            printf("ERROR: Cannot do statistics for empty or mixed types\n");
            exit(EXIT_FAILURE);
        }
        Free((void **)&SymbolUsage);
        PrintListOfAnnotatedTSTPNodes(stdout,Head,Format);
        FreeListOfAnnotatedFormulae(&Head);
    } else {
        printf("ERROR: Failed to open %s\n",argv[ArgOffset+1]);
    }

    return(EXIT_SUCCESS);
}
//-----------------------------------------------------------------------------
