//-----------------------------------------------------------------------------
//----Program to limit the CPU time available to a job
//----Usage: CPULimitedRun [-qN] <TimeLimit> <Job>
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <limits.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
//-----------------------------------------------------------------------------
#define STRING_LENGTH 80
#define MAX_ARGUMENTS 30

#define READ_INDEX 0
#define WRITE_INDEX 1
#define STDIN 0
#define STDOUT 1
//-----------------------------------------------------------------------------
typedef char String[STRING_LENGTH];
typedef char *StringName;
//-----------------------------------------------------------------------------
void XCPUHandler(int TheSignal) {

    if (TheSignal != SIGXCPU) {
        printf("Error, %d is not SIGXCPU\n",TheSignal);
    } else {
        printf("SGIXCPU caught\n");
        exit(-1);
    }
}
//-----------------------------------------------------------------------------
void SetCPUTimeLimit(int CPUTimeLimit) {

    struct rlimit ResourceLimits;

//----Set the signal handler
//----No point - signals are reset on exec
//----For some reason this is still needed, or the signal doesn't work
    if (signal(SIGXCPU,XCPUHandler) == SIG_ERR) {
        perror("CPULimitedRun - Setting signal handler:");
        exit(-1);
    }

//----Limit the CPU time. Need to get old one for hard limit field
    if (getrlimit(RLIMIT_CPU,&ResourceLimits) == -1) {
        perror("CPULimitedRun - Getting resource limit:");
        exit(-1);
    }
//----Set new CPU limit in ms (sent in secs)
    ResourceLimits.rlim_cur = CPUTimeLimit-1; //---- -1 because its weird
    if (setrlimit(RLIMIT_CPU,&ResourceLimits) == -1) {
        perror("CPULimitedRun - Setting resource limit:");
        exit(-1);
    }
}
//-----------------------------------------------------------------------------
//----Prevent core dumps that occur on timeout
void SetNoCoreDump(void) {

    struct rlimit ResourceLimits;

//----Get old resource limits
    if (getrlimit(RLIMIT_CORE,&ResourceLimits) == -1) {
        perror("CPULimitedRun - Getting resource limit:");
        exit(-1);
    }
//----Set new core limit to 0
    ResourceLimits.rlim_cur = 0;
    if (setrlimit(RLIMIT_CORE,&ResourceLimits) == -1) {
        perror("CPULimitedRun - Setting resource limit:");
        exit(-1);
    }
}
//-----------------------------------------------------------------------------
int main(int NumberOfArguments,char *Arguments[]) {

    int TimeLimit;
    int ArgumentNumber;
    int QuietnessLevel;
    int ArgumentOffset;

//----Check the time limit is one
    if (NumberOfArguments >= 2 && strstr(Arguments[1],"-q") == Arguments[1]) {
        QuietnessLevel = atoi(&Arguments[1][2]);
        ArgumentOffset = 1;
    } else {
        QuietnessLevel = 1;
        ArgumentOffset = 0;
    }

    if (NumberOfArguments - ArgumentOffset >= 3 && 
(TimeLimit = atoi(Arguments[ArgumentOffset+1])) > 0) {
        if (QuietnessLevel == 0) {
            printf(
"CPULimitedRun: ----------------------------------------------------------\n");
            printf("CPULimitedRun: %s ",Arguments[ArgumentOffset+2]);
            for (ArgumentNumber=ArgumentOffset+3;
ArgumentNumber<NumberOfArguments; ArgumentNumber++) {
                printf("%s ",Arguments[ArgumentNumber]);
            }
            printf("\n");
            printf("CPULimitedRun: Time limit is %ds\n",TimeLimit);
//----Output the PID for possible later use
            printf("CPULimitedRun: PID is %d\n",getpid());
            printf(
"CPULimitedRun: ----------------------------------------------------------\n");
            fflush(stdout);
        }
        SetCPUTimeLimit(TimeLimit);
        SetNoCoreDump();

//----Execute the ATP system
        execvp(Arguments[ArgumentOffset+2],Arguments+ArgumentOffset+2);
        perror("CPULimitedRun - Cannot exec program:");
    } else {
        printf("Usage: CPULimitedRun [-qN] <TimeLimit> <Job>\n");
    }

    return(-1);
}
//-----------------------------------------------------------------------------
