/*---------------------------------------------------------------------------*/
/*----Program to limit the CPU and WC time available to a job                */
/*----Usage: TimeLimitedRun [-qN] <CPUTimeLimit> <WCTimeLimit> <Job>         */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <limits.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
/*---------------------------------------------------------------------------*/
#define STRING_LENGTH 80
#define MAX_ARGUMENTS 30

#define READ_INDEX 0
#define WRITE_INDEX 1
#define STDIN 0
#define STDOUT 1
/*---------------------------------------------------------------------------*/
typedef char String[STRING_LENGTH];
typedef char *StringName;
/*---------------------------------------------------------------------------*/
void XCPUHandler(int TheSignal) {

    if (TheSignal != SIGXCPU) {
        printf("Error, %d is not SIGXCPU\n",TheSignal);
    } else {
        printf("SGIXCPU caught\n");
        exit(-1);
    }
}
/*---------------------------------------------------------------------------*/
void SetCPUTimeLimit(int CPUTimeLimit) {

    struct rlimit ResourceLimits;

/*----Set the signal handler                                                 */
/*----No point - signals are reset on exec                                   */
/*----For some reason this is still needed, or the signal doesn't work       */
    if (signal(SIGXCPU,XCPUHandler) == SIG_ERR) {
        perror("TimeLimitedRun - Setting signal handler:");
        exit(-1);
    }

/*----Limit the CPU time. Need to get old one for hard limit field           */
    if (getrlimit(RLIMIT_CPU,&ResourceLimits) == -1) {
        perror("TimeLimitedRun - Getting resource limit:");
        exit(-1);
    }
/*----Set new CPU limit in ms (sent in secs)                                 */
    ResourceLimits.rlim_cur = CPUTimeLimit-1; /*---- -1 because its weird */
    if (setrlimit(RLIMIT_CPU,&ResourceLimits) == -1) {
        perror("TimeLimitedRun - Setting resource limit:");
        exit(-1);
    }
}
/*---------------------------------------------------------------------------*/
void SetMemoryLimit(int MemoryLimit) {

    struct rlimit ResourceLimits;

/*----Limit the memory. Need to get old one for hard limit field           */
    if (getrlimit(RLIMIT_AS,&ResourceLimits) == -1) {
        perror("TimeLimitedRun - Getting resource limit:");
        exit(-1);
    }
/*----Set new CPU limit in ms (sent in secs)                                 */
    ResourceLimits.rlim_cur = MemoryLimit; /*---- -1 because its weird */
    if (setrlimit(RLIMIT_AS,&ResourceLimits) == -1) {
        perror("TimeLimitedRun - Setting resource limit:");
        exit(-1);
    }
}
/*---------------------------------------------------------------------------*/
/*----Prevent core dumps that occur on timeout                               */
void SetNoCoreDump(void) {

    struct rlimit ResourceLimits;

/*----Get old resource limits                                                */
    if (getrlimit(RLIMIT_CORE,&ResourceLimits) == -1) {
        perror("TimeLimitedRun - Getting resource limit:");
        exit(-1);
    }
/*----Set new core limit to 0                                                */
    ResourceLimits.rlim_cur = 0;
    if (setrlimit(RLIMIT_CORE,&ResourceLimits) == -1) {
        perror("TimeLimitedRun - Setting resource limit:");
        exit(-1);
    }
}
/*---------------------------------------------------------------------------*/
int main(int argc,char *argv[]) {

    int CPUTimeLimit;
    unsigned int WCTimeLimit;
    unsigned int WCTimeLimit2;
    int MemoryLimit;
    int ArgNumber;
    int QuietnessLevel;
    int ArgOffset;

/*----Check the time limit is one                                            */
    if (argc >= 2 && strstr(argv[1],"-q") == argv[1]) {
        QuietnessLevel = atoi(&argv[1][2]);
        ArgOffset = 1;
    } else {
        QuietnessLevel = 1;
        ArgOffset = 0;
    }

    if (argc - ArgOffset >= 4) {
        CPUTimeLimit = atoi(argv[ArgOffset+1]); 
        WCTimeLimit = (unsigned int) atoi(argv[ArgOffset+2]); 
        if (isdigit(argv[ArgOffset+3][0])) {
            MemoryLimit = atoi(argv[ArgOffset+3]);
            ArgOffset++;
        } else {
            MemoryLimit = 0;
        }

        if (QuietnessLevel == 0) {
            printf(
"TimeLimitedRun: ----------------------------------------------------------\n");
            printf("TimeLimitedRun: %s ",argv[ArgOffset+3]);
            for (ArgNumber=ArgOffset+4;ArgNumber<argc;ArgNumber++)
                printf("%s ",argv[ArgNumber]);
            printf("\n");
            printf("TimeLimitedRun: CPU time limit is %ds\n",CPUTimeLimit);
            printf("TimeLimitedRun: WC  time limit is %ds\n",WCTimeLimit);
            if (MemoryLimit > 0) {
                printf("TimeLimitedRun: Memory   limit is %dbytes\n",
MemoryLimit);
            }
/*----Output the PID for possible later use                                  */
            printf("TimeLimitedRun: PID is %d\n",getpid());
            printf(
"TimeLimitedRun: ----------------------------------------------------------\n");
            fflush(stdout);
        }
        if (CPUTimeLimit > 0) {
            SetCPUTimeLimit(CPUTimeLimit);
        }
        if (WCTimeLimit > 0) {
            alarm(WCTimeLimit);
        }
        if (MemoryLimit > 0) {
            SetMemoryLimit(MemoryLimit);
        }
        SetNoCoreDump();

/*----Execute the ATP system                                                 */
        execv(argv[ArgOffset+3],argv+ArgOffset+3);
        fprintf(stderr,"Could not exec %s\n",argv[ArgOffset+3]);
        perror("TimeLimitedRun - Cannot exec program:");
    } else {
        printf(
"Usage: TimeLimitedRun [-qN] <CPUTimeLimit> <WCTimeLimit> <Job>\n");
    }

return(-1);
}
/*---------------------------------------------------------------------------*/
