#!/usr/bin/perl -w

use strict "vars";
#------------------------------------------------------------------------------
sub ep {
    my($InputLine) = @_;
    my $Language;
    my $Name;
    my @AlternativeNames;
    my $Source;
    my $SourceRecord;
    my @SourceParts;
    my @Temp;
    my $Formula;
    my @Source;
    my $Type;
    my @UsefulInfos;
    my @TPTPLiterals;


#DEBUG    print("\n====$InputLine====\n");

#----Clause normal form, cnf, is assumed to be the language of the file.
    $Language = "cnf";
#----Extract main components
    ($Name,$Formula,$Source) = ($InputLine =~ /\s*(\d*)\s*\:\s*\[(.*)\]\s*\:\s*(.*)$/);
#DEBUG     print("Parts are ==$Name==\n==$Source==\n==$Formula==\n");

#----If some error, just return the language
    if (!defined($Name) || !defined($Source) || !defined($Formula)) {
        return($Language);
    }

#----Check if initial or inferred
    if ($Source eq "initial") {
        $Type = "initial";
        @UsefulInfos = ();
        @Source = ("file");
    } elsif ($Source =~ /\d+/) {
#----Inferred by copy
        $Type = "derived";
        @UsefulInfos = ("iquote('$Source')");


#----Get the source parts
     @SourceParts = split(/[\(\,\)]/,$Source);

    } else {
        $Type = "derived";
        @UsefulInfos = ("iquote('$Source')");
	$SourceRecord = $Source;
#DEBUG  print("SourceRecord is == $SourceRecord\n");
	$SourceRecord =~ s/([^a-zA-Z0-9])/ $1 /g;
	@SourceParts = split(/\s+/, $SourceRecord);
#DEBUG  print("Sources  are == @SourceParts\n");
        @Source = MakeParents(@SourceParts);	
    }
#----Fix up the clause.
    @TPTPLiterals = ();

#----Replace ",--" with ";~" for splitting (to recognize this particular ",")
    $Formula =~ s/\,\s*-\s*-\s*/\;\~/g;
#----Replace ",++" with ";" for splitting (to recognize this particular ",")
    $Formula =~ s/\,\s*\+\s*\+\s*/\;/g;
#----Remove the left "++"
    $Formula =~ s/\s*\+\s*\+\s*//g;
#----Replace the left "--" with "~"
    $Formula =~ s/\s*-\s*-\s*/\~/g;    
#----Remove the "$" before "true"
    $Formula =~ s/\s*\$\s*//g;
#DEBUG    print("Revised Formular: $Formula");
    @TPTPLiterals = split(/;/,$Formula);
    return($Language,$Name,$Type,[@TPTPLiterals],[@Source],[@UsefulInfos]);
}
#----------------------------------------------------------------------------
sub MakeParents{
    my @Parts = @_;
    my @ParentsInfo;
    my @InferenceInfos;
    my $InferenceRule;
    my @Parents;
    my @Parent;
    my $Symbol;
    my @Symbols;
    my $Count;

    my @EqualityInferenceRules = (
        "er", #----Equality resolution
        "pm", #----Paramodulation
        "ef", #----Equality factoring
        "rw", #----Rewriting
        "sr", #----Simplify reflect
        #"ar", #----AC-resolution
        #"cn", #----Clause normalize
    );

    my @NonLogicalConsequenceRules = (
        "split",  #----Splitting
    );
	
	$InferenceRule = shift(@Parts);
    $InferenceRule =~ tr/A-Z/a-z/;
#----Get rid of the outmost brackets	
	shift(@Parts);
 	pop(@Parts);
#DEBUG  print("Parts are: @Parts\n");
	$Count = 0;
	@Symbols = ();
	push(@Parts,',');
#DEBUG	print("Parts are: @Parts\n");
	foreach $Symbol (@Parts) {
		push(@Symbols,($Symbol));
#DEBUG	print("Symbols are: @Symbols\n");

		if ($Symbol eq '(') {
			$Count++;
		}
		if ($Symbol eq ')') { 
			$Count--;
		}
		if ($Count eq 0 && $Symbol eq ','){
			pop(@Symbols); #Get rid of the last ','
			if (scalar(@Symbols) == 1) {
				push(@Parents,$Symbols[0]);
			} else { 
#DEBUG		                print("Symbols are: @Symbols\n");
			        @Parent = MakeParents(@Symbols);
				push(@Parents, [MakeArray(@Parent)]);
			}	
			@Symbols = ();
		}
	}
#----Check if equality parent is needed
        if (grep(/$InferenceRule/,@EqualityInferenceRules)) {
            push(@Parents,("theory(equality)"));
        }
#----Check if logical consequence or not
        if (grep(/$InferenceRule/,@NonLogicalConsequenceRules)) {
            @InferenceInfos = ("status(sat)");
        } else {
            @InferenceInfos = ("status(thm)");
        }
        @ParentsInfo = ("inference",$InferenceRule,[@InferenceInfos],[@Parents]);
#DEBUG		print("ParentsInfo is: @ParentsInfo");
	return(@ParentsInfo);
}
#------------------------------------------------------------------------------
#sub MakeArray {
#    my @TheArray = @_;
#    return(@TheArray);
#}
#------------------------------------------------------------------------------
return 1;
#------------------------------------------------------------------------------

