#!/usr/local/bin/perl -w

use strict "vars";
#------------------------------------------------------------------------------
sub spass {
    my ($InputLine,$SolutionType,%GlobalMemory) = @_;

    my $MoreInputLine;
    my $Language;
    my $Name;
    my $Source;
    my $Formula;
    my @AlternativeNames;
    my $RecordedSource;
    my $InferenceRule;
    my $SplitLevel;
    my $SplitPosition;
    my @Parents;
    my @FinalParents;
    my $NumberOfParents;
    my $Parent;
    my $Rule;
    my $Type;
    my @Source;
    my @UsefulInfos;
    my @InferenceInfos;
    my @TPTPLiterals;
    my $UPPredicate;
    my @SPASSLiterals;
    my $Sign;
    my $Literal;

    my @EqualityInferenceRules = (
        "eqr",  #----Equality resolution
        "eqf",  #----Equality factoring
        "mpm",  #----Merging paramodulation
        "spr",  #----Superposition right
        "spl",  #----Superposition left
        "spm",  #----Paramodulation
        "opm",  #----Ordered paramodulation
        "rew",  #----Rewriting
        "crw",  #----Conditional rewriting
        "aed",  #----Assignment equation deletion
        "res",  #----Binary resolution, which resolves equal modulo symmetry
        "obv",  #----Resolution with reflexivity
    );
    my %SpecialStatusRules = (
        "spt" => "split",  #----Splitting
    );

#DEBUG print(stderr "====$InputLine====\n");
#----Discard not real input lines
    if ($InputLine eq ".") {
        return(undef);
    }

#----Join broken lines
    while ($InputLine !~ /\.$/ && defined($MoreInputLine = <STDIN>)) {
        $InputLine .= $MoreInputLine;
        chomp($InputLine);
    }

#----Clause normal form, cnf, is assumed to be the language of the Otter file.
    $Language = "cnf";
#----Extract main components
    ($Name,$Source,$Formula) = ($InputLine =~ /^([\d]*)\[(.*)\] *(.*)\. *$/);
#DEBUG print(stderr "Parts are ==$Name==$Source==$Formula==\n");

#----If some error, just return the language
    if (!defined($Name) || !defined($Source) || !defined($Formula)) {
        return($Language);
    }

#----Get the name
    @AlternativeNames = split(/,/,$Name);
    $Name = $AlternativeNames[0];

#----Record Source for output, for inference_details item
    $RecordedSource = $Source;

#----Extract the inference rule
    ($InferenceRule) = ($Source =~ /^\d*:(\w+)/);
    $InferenceRule =~ tr/A-Z/a-z/;
    $Source =~ s/^\d*:\w+:*//;

    @Parents = ();
#----Extract Source list
    while ($Source =~ /\d[\.\d]*/) {
#DEBUG print(stderr "Extract parent from ==$Source==\n");
        ($Parent) = ($Source =~ /(\d[\.\d]*)/);
#DEBUG print(stderr "Parent is ==$Parent==\n");
        $Source =~ s/^\Q$Parent\E//;
#----Clean parent after removing the raw form from source
        $Parent =~ s/\..*//;
        $Source =~ s/^,//;
        push(@Parents,($Parent));
    }
    $NumberOfParents = scalar(@Parents);

#----Check if equality parent is needed
    if (grep(/$InferenceRule/,@EqualityInferenceRules)) {
        push(@Parents,("theory(equality)"));
    }

#----Fix up the clause.
    @TPTPLiterals = ();
#----Remove * and + tags
    $Formula =~ s/\*//g;
    $Formula =~ s/\+//g;
#----Remove sort ||. They are antecedent anyway
    $Formula =~ s/\|\| *//;
#----Remove those dumb _f and _p symbols that format.dfg inserts (omit if big_p
#----for SYN problems - aaargh)
    $Formula =~ s/_f(\W)/$1/g;
    $Formula =~ s/big_p(\W)/temp${$}temp$1/g;
    $Formula =~ s/_p(\W)/$1/g;
    $Formula =~ s/temp${$}temp(\W)/big_p$1/g;
#----Convert Skolem predicates to lower case
    $Formula =~ s/SkP/skP/g;
    $Formula =~ s/SkC/skC/g;
#DEBUG print("Clean clause is $Formula\n");
#----Check if the empty clause
    if ($Formula ne "-> \.") {
#----Split on spaces. The -> will be a literal, and can be used to
#----change the sign from initial -- (antecedent literals)
        $Sign = "~";
        @SPASSLiterals = split(/ +/,$Formula);
        foreach $Literal (@SPASSLiterals) {
#DEBUG print("translating $Literal\n");
#----Check if implication
            if ($Literal eq "->") {
                $Sign = "";
            } else {
#DEBUG print("and got $Sign$Literal\n");
                push(@TPTPLiterals,("$Sign$Literal"));
            }
        }
    }

#----Check if logical consequence or not
    @InferenceInfos = ();
    if (exists($SpecialStatusRules{$InferenceRule})) {
        if ($SpecialStatusRules{$InferenceRule} eq "split" && 
$SolutionType eq "Refutation") {
            ($SplitLevel) = ($RecordedSource =~ /(\d*):Spt/);
            if (exists($GlobalMemory{"splitposition"})) {
#DEBUG print(stderr "old level is $GlobalMemory{'splitlevel'} new level is $SplitLevel old splitposition is $GlobalMemory{'splitposition'}\n");
#----Fall down to earlier split level
                while ($SplitLevel < $GlobalMemory{"splitlevel"}) {
                    if ($GlobalMemory{"splitposition"} =~ /r$/) {
                        $GlobalMemory{"splitposition"} =~ s/r$//;
                        $GlobalMemory{"splitlevel"}--;
                    } else {
                        print("ERROR: Split confused at $RecordedSource\n");
                    }
                }
#DEBUG print(stderr "after dropping rs old level is $GlobalMemory{'splitlevel'} new level is $SplitLevel old splitposition is $GlobalMemory{'splitposition'}\n");
#----If at the same level, then L - A - R
                if ($SplitLevel == $GlobalMemory{"splitlevel"}) {
                    if ($GlobalMemory{"splitposition"} =~ /[la]$/) {
#DEBUG printf(stderr "@TPTPLiterals ==%d==$NumberOfParents==\n",scalar(@TPTPLiterals));
#----Anti-kid for unit splits only
                        if (scalar(@TPTPLiterals) == 1 && 
$NumberOfParents == 3) {
                            $GlobalMemory{"splitposition"} =~ s/.$/a/;
#----For anti-kids, remove the second parent which is the split clause
                            $Parent = shift(@Parents);
                            shift(@Parents);
                            unshift(@Parents,($Parent));
                            $NumberOfParents--;
                        } else {
                            $GlobalMemory{"splitposition"} =~ s/.$/r/;
#----For right side of split, remove the left $false parent
                            shift(@Parents);
                            $NumberOfParents--;
                        }
                    } else {
                        print(
"ERROR: Split incomplete at $RecordedSource\n");
                    }
                } elsif ($SplitLevel == $GlobalMemory{"splitlevel"}+1) {
                    $GlobalMemory{"splitposition"} .= "l";
                } else {
                    print("ERROR: Split weird at $RecordedSource\n");
                }
            } else {
                $GlobalMemory{"splitposition"} = "l";
            }
#DEBUG print(stderr "new split $RecordedSource gives level $SplitLevel at $GlobalMemory{'splitposition'}\n");
            $GlobalMemory{"splitlevel"} = $SplitLevel;
            $SplitPosition = $GlobalMemory{"splitposition"};
#----sr_split is a spass refutation split
            push(@InferenceInfos,
("$InferenceRule(sr_split,[position($SplitPosition)])"));
        } elsif ($SpecialStatusRules{$InferenceRule} eq "split" && 
$SolutionType eq "Assurance") {
#AAARGH - HAVE TO TAKE Assurance UNTIL SPASS indicates ends of saturations
#AAAARGH $SolutionType eq "Saturation") 
#----ss_split is a spass saturation split
            push(@InferenceInfos,
("$InferenceRule(ss_split,[])"));
        }
    } else {
        push(@InferenceInfos,("status(thm)"));
    }

#----No duplicate parents
    @FinalParents = ();
    foreach $Parent (@Parents) {
        if (!grep(/$Parent/,@FinalParents)) {
            push(@FinalParents,($Parent));
        }
    }
    if (@FinalParents) {
        $Type = "derived";
        @Source = ("inference",$InferenceRule,[@InferenceInfos],
[@FinalParents]);
        @UsefulInfos = ("iquote('$RecordedSource')");
    } else {
        $Type = "initial";
        @Source = ("file");
        @UsefulInfos = ();
    }


    return($Language,$Name,$Type,[@TPTPLiterals],[@Source],[@UsefulInfos],
%GlobalMemory);
}
#------------------------------------------------------------------------------
return 1;
#------------------------------------------------------------------------------
