#!/usr/bin/perl -w

use strict "vars";
#------------------------------------------------------------------------------
sub vampire {
    my ($InputLine) = @_;

    my $Language;
    my $Source;
    my $Parents;
    my $Name;
    my $InferenceRule;
    my $Formula;
    my $Type;
    my @Source;
    my @UsefulInfos;
    my $Prefix;
    my $Quantifier;
    my $Variables;
    my $Suffix;

#DEBUG print("\n====$InputLine====\n");

#----Extract main components
    ($Source) = ($InputLine =~ /(\[[^]]*])/);
#DEBUG print("Source is $Source\n");

    if ($InputLine =~ /, input]/ || $InputLine =~ /, name introduction]/) {
        $Type = "initial";
        ($Name,$Formula) = ($InputLine =~ /\[(\d+), [^]]*]\s*(.*)/);
        $Parents = "";
        $InferenceRule = "";
    } else {
        $Type = "derived";
        ($Parents,$Name,$InferenceRule,$Formula) = ($InputLine =~ 
/\[([\d,]*)->(\d+), ([^]]*)]\s*(.*)/);
    }
#DEBUG print("Type is $Type\n");
#DEBUG print("Parts are ==$Source==$Parents==$Name==\n==$InferenceRule==\n==$Formula==\n");
#----Fix inference rule
    $InferenceRule =~ s/,/_/g;
    $InferenceRule =~ s/ /_/g;
    $InferenceRule =~ tr/A-Z/a-z/;
#----If some error, just return the language
    if (!defined($Parents) || !defined($Name) || !defined($InferenceRule) || 
!defined($Formula)) {
        return("fof");
    }

#----Check for non-CNF operators
    if ($Formula =~ /! / || $Formula =~ /=>/ || $Formula =~ /\? / ||
$Formula =~ /&/) {
        $Language = "fof";
    } else {
        $Language = "cnf";
    }

#----Check if initial or inferred
    if ($Type eq "initial") {
        @Source = ("file");
        @UsefulInfos = ();
    } else {
        @Source = MakeSource($Type,$InferenceRule,$Parents);
        @UsefulInfos = ("iquote('$Source')");
    }

#----Fix up the formula
    $Formula =~ s/#/\$false/;
    $Formula =~ s/\\\//|/g;
    while ($Formula =~ /\([\?!] [A-Z][^)]*\)/) {
        ($Prefix,$Quantifier,$Variables,$Suffix) = ($Formula =~
/(.*)\(([\?!]) ([A-Z][^)]*)\)(.*)/);
#DEBUG print("==$Prefix==$Quantifier==$Variables==$Suffix==\n");
        $Variables =~ s/ /,/g;
        $Formula = "$Prefix$Quantifier [$Variables] : $Suffix";
#DEBUG print("DeQed ==$Formula==\n");
    }
#DEBUG print("Revised Formula: $Formula\n");

    return($Language,$Name,$Type,[($Formula)],[@Source],[@UsefulInfos]);
}
#----------------------------------------------------------------------------
sub MakeSource {
    my ($Type,$InferenceRule,$Parents) = @_;

    my @Source;
    my @InferenceInfos;
    my $Rule;
    my $Status;

    my @EqualityInferenceRules = (
        "superposition",
        "demodulation",
    );
    my @NonLogicalConsequenceRules = (
        "skolemization",
    );

#----Check if equality parent is needed
    foreach $Rule (@EqualityInferenceRules) {
        if ($InferenceRule =~ /$Rule/) {
            $Parents .= ",theory(equality)";
            last;
        }
    }
#----Check if logical consequence or not
    $Status = "thm";
    foreach $Rule (@NonLogicalConsequenceRules) {
        if ($InferenceRule =~ /$Rule/) {
            $Status = "sat";
            last;
        }
    }
    @InferenceInfos = ("status($Status)");

    @Source = ("inference",$InferenceRule,[@InferenceInfos],[($Parents)]);
#DEBUG print("Source is: @Source");
	return(@Source);
}
#------------------------------------------------------------------------------
return 1;
#------------------------------------------------------------------------------
