/*------------------------------------------------------------------*/
/*      affc.c
/*      copyrighted by Monty Newborn 2005
/*
/*      The functions in this file transform the clauses in a file
/*      "infile" into an encoded form which is used as input to the
/*      theorem-proving program in compiled_code_buffer.
/*
/*      functions:
/*         yyerror()......................#1
/*         tree_add().....................#2
/*         tree_find()....................#3
/*         pf_init()......................#4
/*         pf_alloc().....................#5
/*         pf_initsym()...................#6
/*         pf_getnode()...................#7
/*         pf_add().......................#8
/*         pf_find()......................#9
/*         c_init().......................#10
/*         c_alloc()......................#11
/*         c_link().......................#12
/*         emit_to_code_space()...........#13
/*         psave()........................#14
/*         vinit()........................#15
/*         vfind()........................#16
/*         vsave()........................#17
/*         idcopy().......................#18
/*         yyback().......................#19
/*         yylook().......................#20
/*         yylex()........................#21
/*         emit_word()....................#22
/*         emit_bytes()...................#23
/*         predicate_out()................#24
/*         pick_next_lit()................#25
/*         clause_out()...................#26
/*         yyparse()......................#27
/*         compile_clauses()>>>>>>>>>>>...#28
/*
/*------------------------------------------------------------------
/*
/*         FUNCTION LINKAGE OF AFFC.C (all functions are in affc.c)
/*
/*  compile_clauses()
/*      pf_initsym()
/*          pf_init()
/*      c_init()
/*      yyparse()
/*          yylex()
/*              yylook()
/*                  yyback()
/*              yyerror()
/*          vinit() 
/*          c_link() 
/*          c_alloc()
/*          emit_to_code_space()
/*              yyerror()
/*          psave()
/*              pf_find()         
/*                  tree_find() 
/*              pf_getnode()
/*                  pf_alloc()
/*                  yyerror()
/*              pf_add()
/*                  tree_add()
/*          vsave()             
/*              vfind()
/*              yyerror()
/*          idcopy() 
/*          clause_out()
/*              emit_bytes() 
/*              pick_next_lit() 
/*                  predicate_out()
/*                      emit_word()
/*              emit_word()
/*          yyerror()
/*
/*------------------------------------------------------------------*/
 
#include <stdio.h> 
#include <strings.h>
 
#include "clauseS.h"
#include "const.h"
 
# define T_NOT      257
# define T_OR       258
# define T_LPAR     259
# define T_RPAR     260
# define T_ID       261
# define T_COMMA    262
# define T_THEOREM  263
 
# define input()    if(yysptr > yysbuf) yytchar = U(*--yysptr);\
		    else {\
		       if(o_tptp == 0) {/*yytchar = getc(infile);*/}\
		       else {yytchar = *theobufptr++;}}\
		    if(yytchar == 4 || yytchar == -1) yytchar = 0;
 
# define U(x)       x
 
# define unput(c)   {yytchar = (c);\
		    *yysptr++ = yytchar;}
 
# define MAXVCHARS        200
# define YYERRCODE        256
# define YYFLAG           -1000
# define YYLAST           49
# define YYLERR           yysvec
# define YY_LINE_LENGTH   160
# define YYMAXDEPTH       100
# define YYNEWLINE        10
# define YYU(x)           x

/* put cr in output buffer */
 
#define EMITreturn *comp_code_buffer_index++='\n'
 
extern void fatal();
extern void get_clauses();
 
extern int  nc;                 /* # clauses in negated conclusion */
 
extern pfnode pfspace[];             /* pf symbol records go here  */
extern int kode[];
extern char name_buff[];
extern cnode pred_ptr_list[];
 
extern char theoCodeBuffer[200000];
extern char compiled_code_buffer[];  /* buffer: compiled clauses   */
extern char *comp_code_buffer_index;
extern int compiled_code_size;
extern int o_tptp;
 
extern cnode *ctbl[];                          /* clause table     */
extern char *pidtbl[];
extern char *fidtbl[];
extern int skolem[MAX_FXNS_IDS];
 
//extern FILE *infile;

/*-----------------------------------------------------------------*/
/* internally defined functions                                    */
/*-----------------------------------------------------------------*/

void yyerror();
pfnode *tree_add();
pfnode *tree_find();
void pf_init();
pfnode *pf_alloc();
void pf_initsym();
pfnode *pf_getnode();
void pf_add();
pfnode *pf_find();
void c_init();
cnode *c_alloc();
void c_link();
void emit_code_to_space();
int psave();
void vinit();
int vfind();
int vsave();
void idcopy();
int yyback();
int yylook();
int yylex();
void emit_word();
void emit_bytes();
void predicate_out();
void pick_next_lit();
void clause_out();
void yyparse();
void compile_clauses(); 

/*-----------------------------------------------------------------*/
/* internally defined constants                                    */
/*-----------------------------------------------------------------*/

int yytchar, yyleng, yymorfg;
int yyprevious;
int yylval, yyval;
int yychar;                       /* current input token number    */
int *yyfnd; 
int vn, vid;                         /* next var slot available    */
int yyerrflag;                       /* error recovery flag        */
 
pfnode *pfree;                             /* available pfnodes    */
int pfcount;                             /* # pfnodes free         */
int pid, fid;                              /* unique ID generator  */
 
char *nbp;                              /* pfnode names go here    */
pfnode *pftree;                            /* symbol table         */
 
int nextcl;
char *theobufptr;
 
cnode *c_free;
 
int cp;                                       /* code goes here    */
 
struct yysvf {
       struct yywork *yystoff;
       struct yysvf *yyother;
       int *yystops;};
	
struct yysvf *yyestate;
struct yysvf *yylstate[YY_LINE_LENGTH], **yylsp, **yyolsp;
 
char yytext[YY_LINE_LENGTH];       /* storage area for input line   */
char ytext[YY_LINE_LENGTH];
char yysbuf[YY_LINE_LENGTH];
char *yysptr = yysbuf;
 
char vname[MAXVCHARS],*vp;                         /* names go here */
 
struct {char *v_name;                           /* variable's name  */
	int v_id;}                             /* variable's number */
	variabl[MAX_VARS];
 
/*------------------------------------------------------------------*/
/*      yyerror(msg)                .......................#1
/*
/*      called by:      c_alloc()               affc.c
/*                      emit_to_code_space()    affc.c
/*                      vsave()                 affc.c
/*                      yylex()                 affc.c
/*                      yyparse()               affc.c
/*
/*      calls:          nobody
/*
/*      returns:        Nil
/*------------------------------------------------------------------*/
 
void yyerror(msg)
char *msg;
{
printf("\n\nAt least one error is in your clauses.\nEdit them.");
printf(" and then try to prove the theorem again.\n\n");
printf("Check clause %d or %d for errors.\n",nextcl+1,nextcl+2);
printf("The error may be: %s.\n",msg);
printf("\nPlease type 'Enter'.\n\n");
fatal(" ");
}
 
/*------------------------------------------------------------------*/
/*
/*      these routines will handle threaded binary tree manipulations.
/*      the node types defined here are simply the required header
/*      for any nodes to be linked into a tree. As long as the first
/*      four fields in the node are the same these routines will work.
/*
/*      Note: the bits field is open for user nodes excepty that the
/*      first two bits (least significant) are used for thread
/*      indicators and must not be fiddled with
/*
/*      functions:
/*
/*------------------------------------------------------------------*/
 
/*------------------------------------------------------------------*/
/*      tree_add(t,n)                .........................#2
/*
/*      an empty tree consists of a 0 (NULL) pointer. When a record is
/*      added the root pointer (tree pointer) will point to the
/*      root node of the tree
/*
/*      tree_add will add a node to the threaded tree. Nodes with
/*      keys less than the parent node go left and greater than
/*      or equal go right.
/*
/*      parameters:   t : tree pointer (root node)
/*                    n : node to add
/*
/*      called by:     pf_add()        affc.c
/*
/*      calls:         nobody
/*
/*      returns:       t : updated root node pointer.
/*------------------------------------------------------------------*/ 
 
pfnode *tree_add(t,n)
pfnode *t,*n;
{
pfnode *p;                                     /* parent pointer    */
 
if(t == 0) {
   n->p_llink = 0;                         /* first node in tree    */
   n->p_rlink = 0;
   n->p_bits &= ~(P_LTHREAD | P_RTHREAD);    /* clear thread bits   */
   return(n);}
 
/*--------------------------------------------------------------*/
/* tree is not empty, we walk down the tree and insert the node */
/*--------------------------------------------------------------*/
 
for(p = t;;) {
   if(strcmp(n->p_name,p->p_name) < 0) {
      if(p->p_llink == 0 || p->p_bits & P_LTHREAD) {
	     n->p_rlink = p;
	     n->p_llink = p->p_llink;
	     n->p_bits |= P_RTHREAD | (p->p_bits & P_LTHREAD);
	     p->p_llink = n;
	     p->p_bits &= ~P_LTHREAD;
	     return(t);}
		
      p = p->p_llink;} 
   else {
      if(p->p_rlink == 0 || p->p_bits & P_RTHREAD) {
	     n->p_llink = p;
	     n->p_rlink = p->p_rlink;
	     n->p_bits |= P_LTHREAD | (p->p_bits & P_RTHREAD);
	     p->p_rlink = n;
	     p->p_bits &= ~P_RTHREAD;
	     return(t);}
	    
      p = p->p_rlink;}}
}
 
/*------------------------------------------------------------------*/
/*      *tree_find(t,k)               .........................#3
/*
/*      tree_find will locate a record in a threaded binary tree when
/*      given a pointer to the key string. If found the pointer to
/*      the record is returned else 0 (NULL)
/*
/*      called by:  pf_find()      affc.c
/*
/*      calls:      nobody
/*
/*      returns:    t  or  0 (NULL)
/*------------------------------------------------------------------*/ 
 
pfnode *tree_find(t,k)
pfnode *t;
char *k;
{
int c; 
 
while(t != 0) {
   if((c = strcmp(k,t->p_name)) == 0) {
      return(t);}                                      /* found it! */
	
   if(c < 0) {
      if(t->p_llink == 0 || t->p_bits & P_LTHREAD) {
         break;}
      t = t->p_llink;}
   else {
      if(t->p_rlink == 0 || t->p_bits & P_RTHREAD) {
         break;}
      t = t->p_rlink;}}
 
return(0);
}
 
/*------------------------------------------------------------------*/
/*      pf_init()                      .........................#4
/*
/*      pf_init will initialize the pfspace so that it consists of a
/*      linked list of pfnodes (llink).
/*
/*      called by:  pf_initsym()          affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void pf_init()
{
int i;
    
pfree = 0;
 
for(i = 0; i < MAX_PF; i++) {
   pfspace[i].p_llink = pfree;            /* create chain of llink  */
   pfree = &pfspace[i];}              /* link all pfnodes together  */
 
for(i = 0; i < MAX_FXNS_IDS; i++) {
   pidtbl[i] = 0;
   fidtbl[i] = 0;
   skolem[i] = 0;}
	
pfcount = MAX_PF;                          /* # pfnodes left        */
pid = fid = 1;                      /* PIDs/FIDs start at 1         */
}
 
/*------------------------------------------------------------------*/
/*      pf_alloc()                   ...........................#5
/*
/*      pf_alloc will allocate and return a pfnode. if none are left
/*      we return a 0 (null) pointer
/*
/*      called by:  pf_getnode()              affc.c
/*
/*      calls:      nobody
/*
/*      returns:    pf
/*------------------------------------------------------------------*/ 
 
pfnode *pf_alloc()
{
pfnode *pf;
    
pf = pfree;
if(pf != 0) {                                /* return top element */
   pfree = pf->p_llink;
   pfcount--;}
return(pf);
}
 
/*------------------------------------------------------------------*/ 
/*      pf_initsym()                 .........................#6
/*
/*      pf_initsym will initialize the pfnode symbol table
/*
/*      called by:  compile_clauses()    affc.c
/*
/*      calls:      pf_init()
/*
/*      returns:    nil
/*------------------------------------------------------------------*/ 
 
 
void pf_initsym()
{
nbp = name_buff;                             /* name buffer         */
pf_init();                                   /* create pfnodes      */
pftree = 0;
}
 
/*------------------------------------------------------------------*/ 
/*      pf_getnode(name)              .........................#7
/*
/*      pf_getnode will allocate and rturn an initialized pfnode.the
/*      default pfnode is a predicate with 0 arguments.
/*
/*      parameters:     name : ^ asciz string that is name
/*
/*      note: the name is copied into a local char array since it
/*      probably comes from yytext and is likely to disappear.
/*
/*      called by:  psave()          affc.c
/*
/*      calls:      pf_alloc()       affc.c
/*                  yyerror()
/*
/*      returns:    p
/*------------------------------------------------------------------*/ 
 
pfnode *pf_getnode(name)
char *name;
{
pfnode *p;
 
if((p = pf_alloc()) == 0) {
   yyerror("out of memory (pftree)");}
	
p->p_bits  = 0;
p->p_nargs = -1;                  /* undefined number of arguments  */
p->p_name  = nbp;
do *nbp++  = *name; while(*name++ != '\0');
if(nbp - &name_buff[MAX_CHAR] >= 0) {
   yyerror("out of memory (name_buff)");}
return(p);
}
 
/*------------------------------------------------------------------*/ 
/*      pf_add(p)                    .........................#8
/*
/*      pf_add will add an identifier to the symbol table (assuming
/*      that it is not already in it). A unique ID is also given. We
/*      stuff the pointer to the pfnode into the fid/pid tables to be
/*      indexed by ID.
/*
/*      called by:  psave()              affc.c
/*
/*      calls:      tree_add()           affc.c
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void pf_add(p)
pfnode *p;
{
if((p->p_bits & P_PRED) == P_PRED) {
   p->p_id = pid;
   pidtbl[pid++] = (char *) p;}
else {
   p->p_id = fid;
   fidtbl[fid++] = (char *) p;}
 
pftree = tree_add(pftree,p);
}
 
/*------------------------------------------------------------------*/ 
/*      pf_find(name)                   .........................#9
/*
/*      pf_find will return a pointer to the pfnode if there exists
/*      one with a passed name. 0 (null) if none found
/*
/*      called by:  psave()              affc.c
/*
/*      calls:      tree_find()          affc.c
/*
/*      returns:    tree_find(pf,name)
/*------------------------------------------------------------------*/ 
 
pfnode *pf_find(name)
char *name;
{
return(tree_find(pftree,name));
}
 
/*------------------------------------------------------------------*/
/*      c_init()              .........................#10
/*
/*      c_init will initialize pred_ptr_list so that the free list
/*      contains all nodes
/*
/*      called by:  compile_clauses()           affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void c_init()
{
int i;
	
c_free = 0;
	
for(i = 0; i < MAX_PRED; i++) {
   pred_ptr_list[i].c_next = c_free;
   c_free = &pred_ptr_list[i];}
 
for(i = 0; i < MAX_BASE_CLAUSES; i++) {
   ctbl[i] = 0;}
 
nextcl = cp = 0;
}
 
/*------------------------------------------------------------------*/
/*      c_alloc()                  .........................#11
/*
/*      called by:  yyparse()           affc.c
/*
/*      calls:      nobody
/*
/*      returns:    ptr to clause
/*------------------------------------------------------------------*/
 
cnode *c_alloc()
{
cnode *c;
	
if(c_free == 0) {
   yyerror("out of memory (c_alloc)");} 
c = c_free;
c_free = c->c_next;
return(c);
}
 
/*------------------------------------------------------------------*/ 
/*      c_link(c,l)              .........................#12
/*
/*      the predicate data is stored as a list of items in the
/*      code array. Clauses are rings of cnodes where each node
/*      points to its predicate.
/*
/*      called by:  yyparse()             affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void c_link(c,l)
int c;                                         /* clause number     */
cnode *l;                                        /* literal to add  */
{
if(l == 0) return;
 
if(ctbl[c] == 0) {                      /* first literal in clause  */
   l->c_next = l;
   l->c_prev = l;
   ctbl[c] = l;}
		 
else {
   l->c_next = ctbl[c];
   l->c_prev = ctbl[c]->c_prev;
   ctbl[c]->c_prev->c_next = l;
   ctbl[c]->c_prev = l;}
}
 
/*------------------------------------------------------------------*/
/*      emit_to_code_space(c,code)  .........................#13
/*
/*      emit a word of code to the code space
/*
/*      called by:  yyparse()               affc.c
/*
/*      calls:      yyerror()               affc.c
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void emit_to_code_space(c,code)
int c, *code;
{
if(cp >= MAX_COMPILED_THM_SIZE)  {
   yyerror("out of memory (code)");}
 
code[cp++] = c;
}
 
/*------------------------------------------------------------------*/
/*      psave(name,type)           .......................#14
/*
/*      psave will save a lexeme of a predicate and
/*      assign an identifier to the lexeme
/*
/*      called by:  yyparse()               theo.c
/*
/*      calls:      pf_add()                affc.c
/*                  pf_getnode()            affc.c
/*                  pf_find()               affc.c
/*
/*      returns:    pf_getnode(name)->p_id
/*------------------------------------------------------------------*/
 
int psave(name,type)
char *name;
int type;
{
pfnode *p;
 
if((p = pf_find(name)) != 0) {              /* is it already there? */
   return(p->p_id);}                                        /* yes! */
    
p = pf_getnode(name);                       /* no, add it to table  */
 
p->p_bits = type;
pf_add(p);
return(p->p_id);
}
 
/*------------------------------------------------------------------*/ 
/*     vinit()            .......................#15
/*
/*      vinit is called before each clause to init the variable
/*      name stuff
/*
/*      called by:  yyparse()         affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void vinit()
{
vn  = 0;
vid = 0;
vp  = vname;
}
 
/*------------------------------------------------------------------*/
/*      vfind()                    .......................#16
/*
/*      vfind will return the index into the var array of the variable
/*      (if it is there) else -1
/*
/*      called by:  vsave()         affc.c
/*
/*      calls:      nobody
/*
/*      returns:    v or -1
/*------------------------------------------------------------------*/ 
 
int vfind(name)
char *name;
{
int i;
	
for(i = 0; i < vn; i++)
   if(strcmp(variabl[i].v_name,name) == 0) {
      return(i);}
return(-1);
}
 
/*------------------------------------------------------------------*/
/*      vsave()                       .......................#17
/*
/*      vsave will stuff an identifier into the
/*      variable table and return its id
/*
/*      called by:  yyparse()           affc.c
/*
/*      calls:      vfind()             affc.c
/*
/*      returns:    variabl[v].v_id  or  vid++
/*------------------------------------------------------------------*/
 
int vsave(name)
char *name;
{
int v;
	
if((v = vfind(name)) != -1) {
   return(variabl[v].v_id);}
 
if(vn >= MAX_VARS) {
   yyerror("out of memory (increase MAX_VARS)");}
 
variabl[vn].v_name = vp;
variabl[vn++].v_id = vid;
	
do {                                        /* copy variable name  */
   *vp++ = *name;
   } while(*name++ != '\0');
 
if(vp - &vname[MAXVCHARS] >= 0) {
   yyerror("out of memory (increase MAXVCHARS)");}
	
return(vid++);
}
 
/*------------------------------------------------------------------*/ 
/*      idcopy()                      .......................#18
/*
/*      idcopy saves a copy of the yytext (needed in for arguments)
/*
/*      called by:  yyparse()           affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void idcopy(from,to)
char *from,*to;
{
int i;
 
for(i = 1; i < YY_LINE_LENGTH; i++) {
   *to++ = *from;
   if(*from++ == '\0') {
      return;}}
 
*from = '\0';
}
 
/*------------------------------------------------------------------*/
 
 
short yyexca[] = {-1, 1, 0, -1, -2, 0, -1, 2, 0, 1, -2, 4, };
 
short yyact[] = {
 
  31,  22,  25,  27,   5,  29,  23,  24,  22,  25,
  12,  10,  18,  23,  22,  13,  12,  19,  16,  23,
  14,  13,  20,   8,   9,   3,  26,  21,   6,  11,
   7,   4,   2,   1,   0,  15,   0,   0,  17,   0,
   0,   0,   0,   0,  30,   0,   0,   0,  28 };
 
short yypact[] = {
 
-259,-1000,-259,-1000,-246,-1000,-1000,-238,-1000,-1000,
-240,-241,-1000,-1000,-246,-1000,-248,-1000,-1000,-253,
-1000,-256,-1000,-1000,-1000,-242,-255,-1000,-1000,-1000,
-260,-1000 };
 
short yypgo[] = {
   0,  33,  32,  25,  31,  30,  23,  24,  29,  17,
  22,  27,  26 };
 
short yyr1[] = {
   0,   1,   2,   2,   4,   3,   3,   5,   5,   6,
   6,   7,   7,   7,   8,   8,   9,   9,  10,  10,
  10,  12,  11,  11 };
 
short yyr2[] = {
   0,   1,   1,   2,   0,   2,   1,   1,   3,   1,
   2,   1,   3,   4,   1,   1,   1,   3,   1,   3,
   4,   1,   1,   1 };
 
short yychk[] = {
-1000,  -1,  -2,  -3,  -4, 263,  -3,  -5,  -6,  -7,
 257,  -8, 256, 261, 258,  -7, 259,  -6, 260,  -9,
 -10, -11, 256, 261, 260, 262, -12, 259, -10, 260,
  -9, 260 };
 
short yydef[] = {
   4,  -2,  -2,   2,   0,   6,   3,   5,   7,   9,
   0,  11,  14,  15,   0,  10,   0,   8,  12,   0,
  16,  18,  22,  23,  13,   0,   0,  21,  17,  19,
   0,  20 };
 
 
 
/*------------------------------------------------------*/ 
 
int yyvstop[] ={
0,  10,  0,
 
8,  10,  0,
 
8,  0,
 
7,  10,  0,
 
3,  10,  0,
 
4,  10,  0,
 
2,  10,  0,
 
9,  10,  0,
 
7,  10,  0,
 
1,  10,  0,
 
5,  10,  0,
 
7,  0,
 
9,  0,
 
7,  0,
 
7,  0,
 
7,  0,
 
7,  0,
 
7,  0,
 
6,  7,  0,  0};
 
struct yywork { char  verify, advance; } yycrank[] = {
0,0,    0,0,    1,3,    0,0,
0,0,    0,0,    0,0,    10,15,
0,0,    0,0,    1,4,    1,5,
15,0,   4,5,    4,5,    10,15,
10,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    1,6,
4,5,    0,0,    0,0,    0,0,
10,15,  1,7,    1,8,    0,0,
0,0,    1,9,    2,7,    2,8,
0,0,    0,0,    2,9,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
1,10,   0,0,    0,0,    0,0,
0,0,    2,10,   0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    1,11,   0,0,    0,0,
0,0,    0,0,    2,11,   0,0,
0,0,    1,12,   0,0,    1,13,
0,0,    0,0,    2,12,   0,0,
2,13,   6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,0,    6,0,
6,0,    6,0,    6,14,   6,0,
0,0,    0,0,    6,0,    0,0,
6,0,    6,0,    0,0,    0,0,
6,0,    0,0,    6,0,    6,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    6,0,    6,0,
6,0,    0,0,    6,0,    0,0,
6,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    6,0,
6,0,    6,0,    0,0,    0,0,
6,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    6,0,
6,0,    6,0,    6,0,    6,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,0,   11,0,   11,0,
11,0,   11,14,  11,0,   0,0,
0,0,    11,0,   0,0,    11,0,
11,0,   0,0,    0,0,    11,0,
0,0,    11,0,   11,0,   0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    11,0,   11,0,   11,0,
0,0,    11,0,   0,0,    11,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    11,0,   11,0,
11,0,   0,0,    0,0,    11,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    11,16,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    11,0,   11,0,
11,0,   11,0,   11,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
14,0,   14,0,   14,0,   14,0,
0,0,    14,0,   0,0,    0,0,
14,0,   0,0,    14,0,   14,0,
0,0,    0,0,    14,0,   0,0,
14,0,   14,0,   0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
14,0,   14,0,   14,0,   0,0,
14,0,   0,0,    14,0,   0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    14,0,   14,0,   14,0,
0,0,    0,0,    14,0,   0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    14,0,   14,0,   14,0,
14,0,   14,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,0,
16,0,   16,0,   16,0,   16,14,
16,0,   0,0,    0,0,    16,0,
0,0,    16,0,   16,0,   0,0,
0,0,    16,0,   0,0,    16,0,
16,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    16,0,
16,0,   16,0,   0,0,    16,0,
0,0,    16,0,   0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
16,0,   16,0,   16,0,   0,0,
0,0,    16,0,   0,0,    0,0,
0,0,    0,0,    16,17,  0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
16,0,   16,0,   16,0,   16,0,
16,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,0,   17,0,
17,0,   17,0,   17,14,  17,0,
0,0,    0,0,    17,0,   0,0,
17,0,   17,0,   0,0,    0,0,
17,0,   0,0,    17,0,   17,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    17,0,   17,0,
17,0,   0,0,    17,0,   0,0,
17,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    17,0,
17,0,   17,0,   0,0,    0,0,
17,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    17,18,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    17,0,
17,0,   17,0,   17,0,   17,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,0,   18,0,   18,0,
18,0,   18,14,  18,0,   0,0,
0,0,    18,0,   0,0,    18,0,
18,0,   0,0,    0,0,    18,0,
0,0,    18,0,   18,0,   0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    18,0,   18,0,   18,0,
0,0,    18,0,   0,0,    18,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    18,0,   18,0,
18,0,   0,0,    0,0,    18,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    18,19,  0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    18,0,   18,0,
18,0,   18,0,   18,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
19,0,   19,0,   19,0,   19,0,
0,0,    19,0,   0,0,    0,0,
19,0,   0,0,    19,0,   19,0,
0,0,    0,0,    19,0,   0,0,
19,0,   19,0,   0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
19,0,   19,0,   19,0,   0,0,
19,0,   0,0,    19,0,   0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    19,0,   19,0,   19,0,
0,0,    0,0,    19,0,   0,0,
0,0,    0,0,    0,0,    19,20,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    19,0,   19,0,   19,0,
19,0,   19,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,0,
20,0,   20,0,   20,0,   20,14,
20,0,   0,0,    0,0,    20,0,
0,0,    20,0,   20,0,   0,0,
0,0,    20,0,   0,0,    20,0,
20,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    20,0,
20,0,   20,0,   0,0,    20,0,
0,0,    20,0,   0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
20,0,   20,0,   20,0,   0,0,
0,0,    20,0,   0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    20,21,  0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
20,0,   20,0,   20,0,   20,0,
20,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   21,0,   21,0,
21,0,   21,0,   0,0,    21,0,
0,0,    0,0,    21,0,   0,0,
21,0,   21,0,   0,0,    0,0,
21,0,   0,0,    21,0,   21,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    21,0,   21,0,
21,0,   0,0,    21,0,   0,0,
21,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    21,0,
21,0,   21,0,   0,0,    0,0,
21,0,   0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    0,0,
0,0,    0,0,    0,0,    21,0,
21,0,   21,0,   21,0,   21,0,
0,0};
 
struct yysvf yysvec[] = {
0,              0,              0,
yycrank+-1,     0,              0,
yycrank+-6,     yysvec+1,       0,
yycrank+0,      0,              yyvstop+1,
yycrank+4,      0,              yyvstop+3,
yycrank+0,      yysvec+4,       yyvstop+6,
yycrank+-132,   0,              yyvstop+8,
yycrank+0,      0,              yyvstop+11,
yycrank+0,      0,              yyvstop+14,
yycrank+0,      0,              yyvstop+17,
yycrank+-6,     0,              yyvstop+20,
yycrank+-259,   0,              yyvstop+23,
yycrank+0,      0,              yyvstop+26,
yycrank+0,      0,              yyvstop+29,
yycrank+-386,   yysvec+6,       yyvstop+32,
yycrank+-2,     yysvec+10,      yyvstop+34,
yycrank+-513,   0,              yyvstop+36,
yycrank+-640,   0,              yyvstop+38,
yycrank+-767,   0,              yyvstop+40,
yycrank+-894,   yysvec+16,      yyvstop+42,
yycrank+-1021,  0,              yyvstop+44,
yycrank+-1148,  yysvec+6,       yyvstop+46,
0,              0,              0};
struct yywork *yytop = yycrank + 1275;
struct yysvf *yybgin = yysvec + 1;
char yymatch[] = {
00  ,01  ,01  ,01  ,01  ,01  ,01  ,01  ,
01  ,011 ,012 ,01  ,01  ,01  ,01  ,01  ,
01  ,01  ,01  ,01  ,01  ,01  ,01  ,01  ,
01  ,01  ,01  ,01  ,01  ,01  ,01  ,01  ,
011 ,01  ,'"' ,01  ,'"' ,'"' ,01  ,'"' ,
01  ,01  ,'"' ,'"' ,01  ,'"' ,01  ,01  ,
'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,01  ,01  ,01  ,'"' ,01  ,'"' ,
01  ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,'"' ,01  ,01  ,01  ,'"' ,'"' ,
01  ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,'"' ,
'"' ,'"' ,'"' ,01  ,01  ,01  ,01  ,01  ,
0};
char yyextra[] = {
0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,
0};
 
/*-----------------------------------------------------------------*/
/*      yyback(p,m)            .......................#19
/*
/*      called by:    yylook()       affc.c          
/*
/*      calls:        nobody
/*
/*      returns:      0,1
/*-----------------------------------------------------------------*/
 
int yyback(p,m)
int *p,m;
{
if(p == 0) {
   return(0);}
while(*p != 0) {
   if(*p++ == m) return(1);}
return(0);
} 
 
/*-----------------------------------------------------------------*/
/*      yylook()                   .......................#20
/*
/*      called by:  yylex()            affc.c
/*
/*      calls:      yyback()           affc.c
/*
/*      returns:    *yyfnd++   or  0
/*-----------------------------------------------------------------*/
 
int yylook()
{
struct yywork *yyt;
struct yysvf *yystate, **lsp;
struct yysvf *yyz;
int yych;
struct yywork *yyr;
char *yylastch;
	
/* start off machines */
 
if(yymorfg == 0) {yylastch = yytext;}
else {
   yymorfg = 0;
   yylastch = yytext + yyleng;}
	  
for(;;) {
   lsp = yylstate;
   yyestate = yystate = yybgin;
   if(yyprevious == YYNEWLINE) {
      yystate++;}
   for(;;) {
      yyt = yystate->yystoff;
      if((int) yyt == (int) yycrank) {  /* may not be any trans  */
	     yyz = yystate->yyother;
	     if(yyz == 0) {
	        break;}
	     if(yyz->yystoff == yycrank) {
	        break;}}
      input()
      *yylastch++ = yych = yytchar;

tryagain:
      yyr = yyt;
      if((int) yyt > (int) yycrank) {
	     yyt = yyr + yych;
	     if(yyt <= yytop && yyt->verify + yysvec == yystate) {
	        if(yyt->advance + yysvec == YYLERR) {
	           unput(*--yylastch);
	           break;}
	        *lsp++ = yystate = yyt->advance + yysvec;
	        goto contin;}}
			 
       else if((int) yyt < (int) yycrank) {
	      yyt = yyr = yycrank + (yycrank - yyt);
	      yyt = yyt + yych;
	      if(yyt <= yytop && yyt->verify + yysvec == yystate) {
	         if(yyt->advance + yysvec == YYLERR) {
	            unput(*--yylastch); 
	            break;}
	         *lsp++ = yystate = yyt->advance + yysvec;
	         goto contin;}
 
	      yyt = yyr + YYU(yymatch[yych]);
	      if(yyt <= yytop && yyt->verify + yysvec == yystate) {
	         if(yyt->advance + yysvec == YYLERR) {
	            unput(*--yylastch); 
	            break;}
	         *lsp++ = yystate = yyt->advance + yysvec;
	         goto contin;}}
			 
       if((yystate = yystate->yyother) && 
		         (int) (yyt = yystate->yystoff) != (int) yycrank) {
	      goto tryagain;}
       else  {
	      unput(*--yylastch);
	      break;}
		
contin:
         ;}
   while(lsp-- > yylstate) {
      *yylastch-- = 0;
      if(*lsp != 0 && (yyfnd = (*lsp)->yystops) && *yyfnd > 0) {
	     yyolsp = lsp;
	     if(yyextra[*yyfnd]) {
	        while(yyback((*lsp)->yystops,-*yyfnd) !=
		                       1 && lsp > yylstate) {
		       lsp--;
		       unput(*yylastch--);}}
						
	     yyprevious = YYU(*yylastch);
	     yylsp = lsp;
	     yyleng = yylastch - yytext + 1;
	     yytext[yyleng] = 0;
	     return(*yyfnd++);}
      unput(*yylastch);}
 
   if(yytext[0] == 0) {                   /*  && feof(yyin)   */
      yysptr = yysbuf;
      return(0);}
   input()
   yyprevious = yytext[0] = yytchar;                    
   yylastch = yytext;}
}
 
/*------------------------------------------------------------------*/
/*      yylex()                     .......................#21
/*
/*      called by:  yyparse()           affc.c
/*
/*      calls:      yylook()            affc.c
/*                  yyerror()           affc.c
/*
/*      returns:    T_OR  or  T_COMMA  or ....  or 0
/*------------------------------------------------------------------*/
 
int yylex()
{
int intstr;
 
while((intstr = (int) yylook()) >= 0)  {
   switch(intstr) {
   case 0:
      return(0); 
   case 1:
      return(T_OR);
   case 2:
      return(T_COMMA);
   case 3:
      return(T_LPAR);
   case 4:
      return(T_RPAR);
   case 5:
      return(T_NOT);
   case 6:
      return(T_THEOREM);
   case 7:
      if(strcmp(yytext,"negated_conclusion") == 0) return(T_THEOREM);
      return(T_ID);
   case 8:
      ;
      break;
   case 9:
      ;
      break;
   case 10:
      yyerror("illegal character");
      break;
   case -1:
      break;
   default:
      yyerror("Not clear");}}
return(0); 
}
 
 
/*------------------------------------------------------------------*/
/*      emit_word()                  .........................#22
/*
/*      called by:  clause_out()               affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void emit_word(word)                              /* emit word   */
int word;
{
int i; 
char *word_pointer;
	
word_pointer = (char *) &word;
for(i = 0; i < 4; i++) {
   *comp_code_buffer_index++ = *word_pointer++;}
}
 
/*------------------------------------------------------------------*/
/*      emit_bytes()                   ......................#23
/*
/*      called by:  clause_out()                  affc.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void emit_bytes(string,bytes)                     /* emit bytes  */
char *string;
int bytes;
{
int i;
	
for(i = 0; i < bytes; i++) {
   *comp_code_buffer_index++ = *string++;} 
}
 
/*------------------------------------------------------------------*/ 
/*      predicate_out()            .........................#24
/*
/*      called by:  pick_next_lit()          affc.c
/*
/*      calls:      emit_word()              affc.c
/*
/*      returns:    nil
/*------------------------------------------------------------------*/ 
 
void predicate_out(p)
int *p;
{
int word_count;
emit_word(*p);
for(word_count = *p++ & ARGMASK; word_count > 0; word_count--) {
   emit_word(*p);
   word_count += *p++ & ARGMASK;}
}
 
/*------------------------------------------------------------------*/ 
/*      pick_next_lit()           .........................#25
/*
/*      emits the literal with the lowest pid and remove it
/*
/*      called by:  clause_out()            affc.c
/*
/*      calls:      predicate_out()         affc.c
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void pick_next_lit(i,code)
int i;
int *code;
{
cnode *c,*s;
int p1,p2;
 
s = ctbl[i];                                /* next literal    */
    
predicate_out(&code[s->c_pred]);          /* send the literal  */
 
if(s == s->c_next) {ctbl[i] = 0;}
else {
   s->c_prev->c_next = s->c_next;
   s->c_next->c_prev = s->c_prev;
   if(ctbl[i] == s) {ctbl[i] = s->c_next;}}
}
 
/*------------------------------------------------------------------*/
/*      clause_out()            ........................#26
/*
/*      clause_out will send out the predicate/function symbol tables
/*      followed by the clauses
/*
/*      called by:    yyparse()           affc.c
/*
/*      calls:        pick_next_lit()     affc.c
/*                    emit_word()         affc.c
/*                    emit_bytes()        affc.c
/*
/*      returns:      nil
/*------------------------------------------------------------------*/ 
 
void clause_out(code)
int *code;
{
int i, longlen;
pfnode *p;

comp_code_buffer_index = compiled_code_buffer; 
if(nc == 0) {nc = nextcl;}
 
/*--------------------------------------------------------------*/
/* first, we send the symbol table out. we send the             */
/* predicate IDs (sorted by ID) then the function IDs (sorted)  */
/*--------------------------------------------------------------*/
 
for(i = 1; (p = (pfnode *) pidtbl[i]) != 0; i++) {
   longlen = (int) strlen(p->p_name);
   emit_bytes(p->p_name,longlen);       
   EMITreturn;}                                /* carr. return  */
		
EMITreturn;     
	
for(i = 1; (p = (pfnode *) fidtbl[i]) != 0; i++) {
   longlen = (int) strlen(p->p_name);
   emit_bytes(p->p_name,longlen); 
   EMITreturn;} 
 
EMITreturn;
 
/*--------------------------------------------------------------*/
/* next, the # base clauses, the number of clauses in the       */
/* negated conclusion, and the total number of clauses are sent */
/* followed by the clauses themselves                           */
/*--------------------------------------------------------------*/
 
emit_word(nextcl);                  /* number of base clauses   */
emit_word(nc);                      /* negated concl. clauses   */
 
/*--------------------------------------------------------------*/
/* we write the clauses out with a -1 to terminate each clause  */
/*--------------------------------------------------------------*/
 
for(i = 0; i < nextcl; i++) {
   emit_word(0); emit_word(0);                      /* parents  */
   while(ctbl[i] != 0) {
      pick_next_lit(i,code);}
   emit_word(-1);}
 
/*--------------------------------------------------------------*/
/* for compatibility with affx, we emit some zero's             */
/*--------------------------------------------------------------*/
 
for(i = 0; i < 15; i++) {
   emit_word(0);}
 
/*------------------------------------------------------------*/
/*  determine the # of chars in compiled_code_buffer          */
/*------------------------------------------------------------*/

compiled_code_size = comp_code_buffer_index - compiled_code_buffer;   
}
 
/*------------------------------------------------------------------*/
/*      yyparse()                     .......................#27
/*
/*      parser for yacc output
/*
/*      called by:         compile_clauses()       affc.c
/*
/*      calls:             yylex()                 affc.c             
/*                         clause_out()            affc.c
/*                         vinit()                 affc.c
/*                         c_link()                affc.c
/*                         c_alloc()               affc.c
/*                         emit_to_code_space()    affc.c
/*                         yyerror()               affc.c
/*
/*      returns:           nil
/*------------------------------------------------------------------*/
 
void yyparse()
{
short yys[YYMAXDEPTH];
int yyv[YYMAXDEPTH];
short yyj, yym, yyw;
int *yypvt;
short yystate, *yyps, yyn;
int *yypv;
short *yyxi;
 
yystate   = 0;
yychar    = -1;
yyerrflag = 0;
yyps      = &yys[-1];
yypv      = &yyv[-1];
 
yystack:                 /* put a state and value onto the stack */
 
if(++yyps > &yys[YYMAXDEPTH]) {yyerror("yacc s.o");}
		
*yyps = yystate;
++yypv;
*yypv = yyval;
 
yynewstate:
yyn = yypact[yystate];
 
if(yyn <= YYFLAG) {goto yydefault;}            /* simple state   */
 
if(yychar < 0) {
   if((yychar = yylex()) < 0)  {yychar = 0;}}
 
if((yyn += yychar) < 0 || yyn >= YYLAST) {goto yydefault;}
 
if(yychk[yyn = yyact[yyn]] == yychar) {          /* valid shift  */
   yychar = -1;
   yyval = yylval;
   yystate = yyn;
   if(yyerrflag > 0) {--yyerrflag;}
   goto yystack;}
 
yydefault:
 
/* default state action */
	
if((yyn = yydef[yystate]) == -2) {
   if(yychar < 0) if((yychar = yylex()) < 0) {yychar = 0;}
		
/* look through exception table */
 
   for(yyxi = yyexca;(*yyxi != (-1)) || (yyxi[1] != yystate); yyxi+=2);
   while(*(yyxi += 2) >= 0) {
      if(*yyxi == (short) yychar) {break;}}
   if((yyn = yyxi[1]) < 0) {return;}}
 
if(yyn == 0) {
		
/*----------------------------------------------*/
/* error ... attempt to resume parsing          */
/*----------------------------------------------*/
		
   switch((int) yyerrflag) {
 
      case 0:                                  /* brand new error */
	  yyerror("syntax error");
 
yyerrlab:
      case 1:
      case 2:     /* incompletely recovered error ... try again   */
 
	  yyerrflag = 3;
 
	    /* find a state where "error" is a legal shift action */
 
	  while(yyps >= yys) {
	     yyn = yypact[*yyps] + YYERRCODE;
	     yyw = yychk[yyact[yyn]]; 
	     if(yyn >= 0 && yyn < YYLAST && yyw == YYERRCODE) {
	        yystate = yyact[yyn];      /* sim shift of "error" */
	        goto yystack;}
	     yyn = yypact[*yyps];
 
/* the current yyps has no shift on "error", pop st     */
 
	     --yyps;
	     --yypv;}
 
/* there is no state on stack with an error shift...abort   */
 
yyabort:                       
	  return;
  
      case 3:            /* no shift yet;clobber input char */
	  if(yychar == 0) {goto yyabort;}
	  yychar = -1;
	  goto yynewstate;}}   /* try again in the same state   */
 
/* reduction by production yyn */
		
yyps -= yyr2[yyn];
yypvt = yypv;
yypv -= yyr2[yyn];
yyval = yypv[1];
yym   = yyn;
		       
		       /* consult goto table to find next state  */
   
yyn = yyr1[yyn];
yyj = yypgo[yyn] + *yyps + 1;
if(yyj >= YYLAST || yychk[ yystate = yyact[yyj]] != -yyn) {
   yystate = yyact[yypgo[yyn]];}
switch(yym) {
 
   case 1:
   clause_out(kode);
   break;                             /* print out clause  */
   case 4:
   vinit(); 
   break;
   case 5:
   nextcl++;
   nc++;
   break;
   case 6:
   nc = 0; 
   break;
   case 9:
   c_link(nextcl, (cnode *) yypvt[-0]); 
   break;
   case 10:
   {cnode *c;
   c = (cnode *) yypvt[-0];
   kode[c->c_pred] |= NEGMASK;
   c_link(nextcl,c);} 
   break;
   case 11:
   yyval = yypvt[-0]; 
   break;
   case 12:
   yyval = yypvt[-2]; 
   break;
   case 13:
   {cnode *c;
   c = (cnode *) yypvt[-3];
   kode[c->c_pred] |= yypvt[-1];
   yyval = yypvt[-3];} 
   break;
   case 14:
   yyerror("predicate expected"); 
   break;
   case 15:
   {cnode *c;
   c = c_alloc();
   c->c_pred = cp;
   yyval = (int) c;
   emit_to_code_space(psave(yytext,P_TYPE) << ARGSHIFT,kode);} 
   break;
   case 16:
   yyval = 1; 
   break;
   case 17:
   yyval = yypvt[-2] + 1; 
   break;
   case 18:
   emit_to_code_space(vsave(ytext) << ARGSHIFT,kode); 
   break;
   case 20:
   kode[yypvt[-2]] |= yypvt[-1]; 
   break;
   case 21:
   yyval = cp;
   emit_to_code_space(FCNMASK+(psave(ytext,0)<<ARGSHIFT),kode); 
   break;
   case 22:
   yyerror("identifier expected"); 
   break;
   case 23:
   idcopy(yytext,ytext); 
   break;}
goto yystack;                        /* stack new state and value  */
 
}
 
/*------------------------------------------------------------------*/ 
/*      compile_clauses()         .......................#28
/*
/*      called by:  main()                  theo.c
/*
/*      calls:      pf_initsym()            affc.c
/*                  c_init()                affc.c
/*                  yyparse()               affc.c
/*
/*      returns:    nil
/*------------------------------------------------------------------*/
 
void compile_clauses()
{
int i;

get_clauses();

yyerrflag = 0;                  /* error recovery flag        */
yymorfg = 0;
yyprevious = YYNEWLINE;

theobufptr = theoCodeBuffer;

nc = 0;

pf_initsym();
c_init();
yyparse();
}

