/*-------------------------------------------------------*/
/* atom.c
/*
/* Copyright Newborn Software 1995, 1996, 1997, 1998, 1999
/*
/* atom.c contains 
/*
/*   1. atom_sel_heuristics().... controls the order of
/*                                selecting the atoms
/*
/*   2.  ash0()
/*   3.	 ash1()
/*   4.	 ash2()
/*   5.	 ash3()
/*   6.	 ash4()
/*   7.  ash5a()
/*   8.  ash5b()
/*   9.  ash5c()
/*
/* The following function is called from herby.c
/* and adds clauses to the original set of base clauses
/* 
/*   10.  resolution_stage()
/*   11.  resolution_stage2()
/*
/*-----------------------------------------------
/*   function calling order:
/*
/*    atom_sel_heuristic()
/*       ash1()
/*          binary_resolvents3()
/*          install_ash_clause()
/*          print_clause()
/*       ash2()
/*          binary_resolvents2()
/*          binary_resolvents3() 
/*          install_ash_clause()
/*          print_clause()
/*       ash3()
/*          binary_resolvents2()
/*          binary_resolvents3()
/*          install_ash_clause()
/*          print_clause()
/*       ash4()
/*          binary_resolvents2()
/*          binary_resolvents3()
/*          install_ash_clause()
/*          install_unit_and_test()
/*          print_clause()
/*       ash5a()
/*           copy_atom_from_clause()
/*           copy_gnd_atom_from_cls()
/*           install_hb_clause()
/*           calc_numb_var_in_lit()
/*       ash5b()
/*           copy_atom_from_clause()
/*           copy_gnd_atom_from_cls()
/*           install_hb_clause()
/*       ash5c()
/*          copy_gnd_atom_from_cls()
/*          copy_atom_from_clause()
/*          calc_numb_var_in_lit()
/*          install_hb_clause()
/*
/*    resolution_stage()
/*      binary_resolvents()
/*      print_clause()
/*
/*--------------------------------------------------------*/
 
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "constHS.h"
#include "clauseHS.h"

/*------------------------------------------------------------*/
/* externally defined functions                               */
/*------------------------------------------------------------*/
 
extern void fatal();                      /* herby.c          */
extern int time_now();

extern void print_clause();               /* hdump.c          */
extern void print0();

extern int unify();                       /* unify.c          */

extern int gen_lit();                     /* hgen.c           */
extern clause *generate_clause2();
extern clause *copy_atom_from_clause();
extern clause *copy_gnd_atom_from_cls();
extern clause *gen_unified_lit();

extern int install_hb_clause();           /* build.c          */
extern int install_ash_clause();
extern int install_unit_and_test();

extern void calc_numb_var_in_lit();       /* hhash.c          */
extern int calc_lit_hashcode();
extern void calc_clause_hashcode();
extern int add_clause_to_ht();

extern int binary_resolvents();           /* infer.c          */
extern int binary_resolvents2();
extern int binary_resolvents3();
extern void clear_var_substitution_tble();

/*------------------------------------------------------------*/
/* externally defined constants                               */
/*------------------------------------------------------------*/

extern char *OutPtr;
 
extern int nextc, nexthb, ply, base, base0, base1, nc;
extern int iter_depth;
extern int build_start_time;
extern int o_interactive;
extern clause *clist[];
extern clause *unit_list[];
extern clause *hb_list[];
extern int fail[];
extern int heuristic_applied[];
extern int branch[];
extern dnode depth[];

extern int good_atoms;
extern int s_nod;
extern int x_debug;
extern int number_constants, var_in_lit;
extern int const_alt, last_installed, nextunit;
extern int last_atom_sign;

extern int *var_substitution_tbl[]; /* herby.c..var substitut'n tbl */
 
extern int max_var;
 
extern int *eptr;                     /* where to emit clause code  */

extern char input_buffer[];

/*---------------------------------------*/
/* internally defined functions          */
/*---------------------------------------*/

int atom_sel_heuristics();
int ash1();
int ash2();
int ash3();
int ash4();
int ash5a();
int ash5b();
int ash5c();
int resolution_stage();
int resolution_stage2();

/*---------------------------------------*/
/* internally defined constants          */
/*---------------------------------------*/

int c_pick;
int ash5_pick[200];
int apick;

/*---------------------------------------------------------*/
/* atom_sel_heuristics()               ..............#1
/*
/* called by: build_tree()   search.c
/*
/* calls:     ash1()          atom.c
/*            ash2()          atom.c
/*            ash3()          atom.c
/*            ash4()          atom.c
/*            ash5a()         atom.c
/*            ash5b()         atom.c
/*            ash5c()         atom.c
/*            print_clause()  hdump.c
/*            time_now()      herby.c
/*
/* returns:   1 if atom found; 0 otherwise
/*----------------------------------------------------------*/
 
int atom_sel_heuristics()
{
clause *c1, *c2, *c_hb;
literal *lit;
int i, j, k, ash_type, ash_clause_number,ash_literal_index;
int ash3_result, ash4_result;
char zz;

if(nexthb == 0) {
  for(i = 0; i < 200; i++) {ash5_pick[i] = 0;}
  apick = 0;}

/*----------------------------------------*/
/* decide on the next constant to         */
/* substitute for a variable in an atom.  */
/* rotate from one constant to the next.  */
/*----------------------------------------*/

if(const_alt++ >= number_constants) {
   const_alt = 1;}

/*----------------------------------------*/
/* ash1: see if two clauses on clist      */
/*       to yield the NULL clause....     */
/*       If so, select most general       */
/*       substitution instance of the     */
/*       two clauses as next atom.        */
/*----------------------------------------*/

for(i = depth[ply - 1].d_ec - 1; i >= depth[ply - 1].d_bc; i--) { 
   c1 = clist[i];
   if(c1->c_nlit == 1) {
      for(j = i - 1; j >= 0; j--) { 
         c2 = clist[j];
         if(c2->c_nlit == 1) {
            if(ash1(c1,c2)) {
               heuristic_applied[ply - 1] = 1;
//             print_clause(hb_list[ply - 1]);
//             print0("   H1 ");
               return(1);}}}}}

/*-----------------------------------------*/
/* ash2: search for a 2-1-1 resolvable     */
/*       trio. If found, add one or two    */
/*       atoms.                            */
/*-----------------------------------------*/

for(i = 0; i < nextc ; i++) { 
   c1 = clist[i];
   if(c1->c_nlit == 1) {
      for(j = nextc - 1; j >= 0; j--) { 
            c2 = clist[j];
            if(c2->c_nlit == 2) {
               if(ash2(c1,c2)) { 
                  heuristic_applied[ply - 1] = 2;
//                print_clause(hb_list[ply - 1]);
//                print0("   H2 ");
	          if(nexthb > ply) {
//                   print0("\n");
//                   print_clause(hb_list[ply]);
                     heuristic_applied[ply] = -2;
//                   print0("   -H2 ");
		     }
                  return(1);}}}}}

/*---------------------------------------------*/
/* ash3:
/*---------------------------------------------*/

for(i = 0; i < nextc; i++) { 
   c1 = clist[i];
   if(c1->c_nlit != 1) {continue;}
   for(j = nextc - 1; j >= 0; j--) { 
      c2 = clist[j];
      if(c2->c_nlit == 3) {
         if(ash3_result = ash3(c1,c2)) {   /* ash3 returns #atoms */
            heuristic_applied[ply - 1] = 3;
//          print_clause(hb_list[ply - 1]);
//          print0("   H3 ");
	    if(ash3_result) {
//             print0("\n");
	       }
	    k = 0;
	    while(ash3_result > 0) {
               heuristic_applied[ply + k] = -3;
//             print_clause(hb_list[ply + k]);
               k++;
//             print0("   -H3 ");
	       if(ash3_result != 1) {
//                print0("\n");
		  }
	       ash3_result--;} 
            return(1);}}}}

/*------------------------------------------------*/
/* ash4:
/*------------------------------------------------*/

if(ash4_result = ash4()) {
   heuristic_applied[ply - 1] = 4;
// print_clause(hb_list[ply - 1]);
// print0("   H4");
   if(ash4_result % 20 == 2) {  /* ash4_results = 10*installed atoms */
//    print0("b ");
         }            /*                 + 1 =>a or +2 =>b */
   else {                       /*-----------------------------------*/
//    print0("a ");
      }
   if(ash4_result > 12) {/*print0("\n");*/}
   k = 0;
   ash4_result = ash4_result - 10;
   while(ash4_result > 10) {
      heuristic_applied[ply + k] = -4;
//    print_clause(hb_list[ply + k]);
      k++;
//    print0("   -H4 ");
      if(ash4_result > 12) {/*print0("\n");*/}
      ash4_result = ash4_result - 10;}
   return(1);}

/*-------------------------------------------------*/
/* ash5a and ash5b .... alternate each time
/*-----------------------------------------*/
/* ash5c:
/*-----------------------------------------*/


if(ply % 3 == 0) {
if(ash_type = ash5c()) {
   heuristic_applied[ply - 1] = 5;
/*
   print_clause(hb_list[ply - 1]);
   print0("   H5c");
   if(ash_type/10000 == 0) {print0("i");}
   if(ash_type/10000 == 1) {print0("ii");}
   if(ash_type/10000 == 2) {print0("iii");}
   if(ash_type/10000 == 3) {print0("iv");}
*/
   c_pick = ash_type % 10000;
// printf(".%d ",c_pick);
// OutPtr += sprintf(OutPtr,".%d ",c_pick); 
   return(1);}}


if(ply % 2 == 1) {
   if(ash_type = ash5a()) {
      heuristic_applied[ply - 1] = 5;
/*
      print_clause(hb_list[ply - 1]);
      print0("   H5a");
      if(ash_type/10000 == 0) {print0("i");}
      if(ash_type/10000 == 1) {print0("ii");}
      if(ash_type/10000 == 2) {print0("iii");}
      if(ash_type/10000 == 3) {print0("iv");}
*/
      c_pick = ash_type % 10000;
//    printf(".%d ",c_pick);
//    OutPtr += sprintf(OutPtr,".%d ",c_pick); 
      return(1);}

   if(ash_type = ash5b()) {
//    printf("Hb5 install at ply %d\n\n",ply);
//    print_clause(hb_list[ply - 1]);
      heuristic_applied[ply - 1] = 5;
/*
      print0("   H5b");
      if(ash_type/10000 == 0) {print0("i");}
      if(ash_type/10000 == 1) {print0("ii");}
      if(ash_type/10000 == 2) {print0("iii");}
      if(ash_type/10000 == 3) {print0("iv");}
      printf(".%d ",ash_type % 10000);
      OutPtr += sprintf(OutPtr,".%d ",ash_type & 10000);*/ 
      return(1);}}

if(ply % 2 == 0) {
   if(ash_type = ash5b()) {
      heuristic_applied[ply - 1] = 5;
//    print_clause(hb_list[ply - 1]);
//    print0("   H5b");
//    if(ash_type/10000 == 0) {print0("i");}
//    if(ash_type/10000 == 1) {print0("ii");}
//    if(ash_type/10000 == 2) {print0("iii");}
//    if(ash_type/10000 == 3) {print0("iv");}
//    printf(".%d ",ash_type % 10000);
//    OutPtr += sprintf(OutPtr,".%d ",ash_type % 10000); 
      return(1);}

   if(ash_type = ash5a()) {
      heuristic_applied[ply - 1] = 5;
//    print_clause(hb_list[ply - 1]);
//    print0("   H5a");
//    if(ash_type/10000 == 0) {print0("i");}
//    if(ash_type/10000 == 1) {print0("ii");}
//    if(ash_type/10000 == 2) {print0("iii");}
//    if(ash_type/10000 == 3) {print0("iv");}
      c_pick = ash_type % 10000;
//    printf(".%d ",c_pick);
//    OutPtr += sprintf(OutPtr,".%d ",c_pick); 
      return(1);}}
  
/*-----------------------------------------*/
/* ash5c:
/*-----------------------------------------*/

if(ash_type = ash5c()) {
   heuristic_applied[ply - 1] = 5;
/*
   print_clause(hb_list[ply - 1]);
   print0("   H5c");
   if(ash_type/10000 == 0) {print0("i");}
   if(ash_type/10000 == 1) {print0("ii");}
   if(ash_type/10000 == 2) {print0("iii");}
   if(ash_type/10000 == 3) {print0("iv");}
*/
   c_pick = ash_type % 10000;
// printf(".%d ",c_pick);
// OutPtr += sprintf(OutPtr,".%d ",c_pick); 
   return(1);}

return(0);
}
/*----------------------------------------------------------*/
/* ash0()                  ........................ #2
/*
/* ash0 permits the user to decide on the atom.
/*
/* called by: atom_sel_heuristic()         atom.c
/*
/* calls:     print_clause()               hdump.c
/*            grab_number()                herby.c
/*            calc_numb_var_in_lit()       hhash.c
/*            copy_atom_from_clause()      hgen.c
/*            copy_gnd_atom_from_cls()     hgen.c
/*            install_hb_clause()          build.c
/*
/* returns:   1 if atom is selected, 0 otherwise
/*---------------------------------------------------------*/

ash0()
{
int i,j;
int ash_clause_number, ash_literal_index, ash_constant_index;
char zz;
clause *c1, *c_hb;
literal *lit;

reselect:

   print0("\nEnter clause number, literal index, and constant index:");
   j = 0;
   while((zz = getchar()) != '\n') {
      if(zz == '\b') {
         if(j > 0) {j--;printf("\b \b");}}
      else {
         input_buffer[j] = zz; printf("%c",input_buffer[j++]);}}
   input_buffer[j] = '\n';
   printf("\n");

/*---------------------------------------------------*/
/* If the user entered a clause number and index     */
/* (and constant index, (this is optional)           */
/* then determine clause and index and use for atom  */
/* after grounding if necessary.  Otherwise....      */
/* if the user entered:                              */
/*            (1) CR => herby pick this atom         */
/*            (2) 'q' => herby pick future atoms     */
/*            (3) 'Q' => herby pick future atoms and */
/*                       turn off debug option       */
/*            (4) 'l' => list atoms selected to far  */
/*---------------------------------------------------*/

   if(input_buffer[0] == '\n') {
      return(0);}
   if(input_buffer[0] == 'q') {
      o_interactive = 0;
      return(0);}
   if(input_buffer[0] == 'Q') {
      o_interactive = 0;x_debug = 0;
      return(0);}
   if(input_buffer[0] == 'l') {
      for(i = 0; i < nexthb; i++) {
         printf("\n");print_clause(hb_list[i]);}
      printf("\n");goto reselect;}
   for(i = 1; i < 128; i++) {
      if(input_buffer[i] == ' ') {break;}
      if(input_buffer[i] == '\n') {goto reselect;}}

/*--------------------------------*/
/* select a constant..........    */
/*--------------------------------*/

   ash_constant_index = 1;
   for(j = i + 1; j < 128 ; j++) {
      if(input_buffer[j] == ' ') {
	 ash_constant_index = grab_number(++j);
	 if(ash_constant_index > number_constants) {
	    goto reselect;}
	 break;}
      if(input_buffer[j] == '\n') {break;}}

   ash_clause_number = grab_number(0);
   ash_literal_index = grab_number(++i);
   if(ash_clause_number > nextc) {goto reselect;}
   c1 = clist[ash_clause_number - 1];
   if(c1->c_nlit < ash_literal_index) goto reselect;

   j= 1;
   for(lit = c1->c_lit ; ; lit = lit->l_next) {
      if((lit->l_id & ELIMLIT) != 0) {continue;}
      if(j == ash_literal_index) {break;}
      j++;}
   calc_numb_var_in_lit(lit,c1->c_nv);
   if(var_in_lit == 0) { 
      c_hb = copy_atom_from_clause(c1,lit);
      if(install_hb_clause(c_hb)) {goto det_lit_sign;}}
   else { 
      c_hb = copy_gnd_atom_from_cls(c1,lit,ash_constant_index);
      if(install_hb_clause(c_hb)) {goto det_lit_sign;}}
   goto reselect;

det_lit_sign:

   /*-------------------------------------*/
   /* determine whether lit or comp??     */
   /*-------------------------------------*/

   while(input_buffer[i] != ' ') {i++;}
   if(input_buffer[i + 1] == '-') {
      c_hb->c_lit->l_id &= ~NEGMASK;}
   else {
      c_hb->c_lit->l_id |= NEGMASK;}
   return(1);}

 
/*-------------------------------------------------------*/
/* ash1()               ...........................#3
/*
/* ash1 looks for a 1-1 resolvable pair
/*
/* called by:    atom_sel_heuristic()        atom.c
/*
/* calls:        binary_resolvents3()        atom.c
/*               install_ash_clause()        atom.c
/*               print_clause()              hdump.c
/*
/* returns:     1 if hb atom added; otherwise returns a 0
/*--------------------------------------------------------*/
 
int ash1(c1,c2)
clause *c1, *c2;
{
clause *c3;
int nextc0;
literal *l1;

nextc0 = nextc;
 
if(binary_resolvents3(c1,c2,1,1,0)) {
   c3 = clist[nextc - 1];
 
/*-----------------------------------------------*/
/* clause c3 is the most common instance of      */
/* clauses c1 and c2.                            */
/* clause c3 may contain eliminated literals.    */
/* unwanted literals must be removed and the     */
/* resulting literal grounded if necessary       */
/*-----------------------------------------------*/

   if(install_ash_clause(c3)) {
      if(x_debug > 1) {
         print0("USING ASH1: \n");
	 print0("C1 and C2 resolve to the NULL clause\n");
         print0("\nC1:  ");
         print_clause(c1);
         print0("\nC2:  ");
         print_clause(c2);
	 print0("\nThe atom generated is:");
         print_clause(c3);
         print0("\n\n");}

      nextc = nextc0;
      return(1);}}

nextc = nextc0;
return(0);
}
 
/*---------------------------------------------------------*/
/* ash2()                ..........................#4
/*
/* ash2 looks for a 1-2-1 hyper-resolvable trio
/*
/* called by:    atom_sel_heuristic()              atom.c
/*
/* calls:        binary_resolvents2()              infer.c
/*               binary_resolvents3()              infer.c
/*               install_ash_clause()              build.c
/*               print_clause()                    hdump.c
/*
/* returns:     1 if hb atom added; otherwise returns a 0
/*-----------------------------------------------------------*/
 
int ash2(c1,c2)
clause *c1, *c2;
{
clause *c3, *c4, *c5;
int j, k, nextc0, nexthb0;

nextc0 = nextc;
nexthb0 = nexthb;

for(j = 1; j <= 2 ; j++) { 
   nextc = nextc0;
   if(binary_resolvents2(c1,c2,1,j,0)) {

/*------------------------------------------------*/
/* c1, a unit clause resolves with c2, with       */
/* two literals to produce c4 with one literal    */
/*------------------------------------------------*/

      c4 = clist[nextc - 1]; 
      for(k = 0; k < nextc - 1; k++) {
         c3 = clist[k];
         if(c3->c_nlit != 1) {continue;}
         nextc = nextc0 + 1; 
         if(binary_resolvents2(c4,c3,1,1,1)) {
            c5 = clist[nextc - 1];
	    if(install_ash_clause(c5)) {

/*-------------------------------------------------*/
/* c4, a unit produced from c1 and c2 has been     */
/* successfully resolved with c3 to produce c5     */
/* which has been installed as the next atom.      */
/* Now the unified literal from the resolution of  */
/* c1 and c1 will be added as the next atom if     */
/* it meets certain conditions.                    */
/*-------------------------------------------------*/

	       nextc = nextc0 + 1;
	       binary_resolvents3(c1,c2,1,j,0);
	       c5 = clist[nextc - 1];
	       install_ash_clause(c5);
if(x_debug > 1) {

print0("USING ASH2: \n");
print0("C1,C2 resolve to C4, which resolves with C3\n");
print0("\nC1: ");print_clause(c1);
print0("\nC2: ");print_clause(c2);
print0("\nC3: ");print_clause(c3);
print0("\nC4: ");print_clause(c4);
print0("\n");
print0("The atom generated is:");
print_clause(c5);
print0("\n\n");}

               nextc = nextc0;
               return(1);}}}}}
	  
nextc = nextc0;
nexthb = nexthb0;
return(0);
}
 
/*--------------------------------------------------------*/
/* ash3() 	        ...........................#5
/*
/* ash3 looks for a 1-3-1-1 hyper-resolvable quad
/*
/* called by:    atom_sel_heuristic()   atom.c
/*
/* calls:        binary_resolvents2()   infer.c
/*               install_ash_clause()   build.c
/*               print_clause()         hdump.c
/*
/* returns:      1 if hb atom added; otherwise returns a 0
/*-----------------------------------------------------------*/
 
int ash3(c1,c2)
clause *c1, *c2;
{
clause *c3, *c4, *c5, *c6, *c7, *c8, *c_hb;
int i, j, k, m, n, nextc0, installed_atoms;
literal *lit;

nextc0 = nextc;
installed_atoms = 0;
c6 = 0;
 
   nextc = nextc0;
   if(binary_resolvents2(c1,c2,1,1,0)) {
      c5 = clist[nextc - 1];		

/*-----------------------------------*/
/* c5 is the resolvent of c1 and c2  */
/*-----------------------------------*/

      for(k = 0 ; k < nextc - 1; k++) {
         c3 = clist[k];
         if(c3->c_nlit != 1) {continue;}

/*------------------------------------------------*/
/* resolve each of the two literals of c5 with c3 */
/*------------------------------------------------*/

            nextc = nextc0 + 1;
            if(binary_resolvents2(c5,c3,1,1,0)) {
               c6 = clist[nextc - 1];	

/*------------------------------------------------*/
/* c6 is a unit and the resolvent of c3 & c5      */ 
/*------------------------------------------------*/

               for(m = 0; m < nextc - 2; m++) {
                 nextc = nextc0 + 2;
                 c4 = clist[m]; 
                 if(c4->c_nlit != 1) {continue;}
                 if(binary_resolvents3(c6,c4,1,1,0)) {
                    c7 = clist[nextc - 1];       

/*-----------------------------------*/
/* c7 is the resolvent of c4 and c6  */
/*-----------------------------------*/

                    if(install_ash_clause(c7)) {
                       nextc = nextc0 + 3;
                       binary_resolvents3(c1,c2,1,1,0);
                       c8 = clist[nextc - 1];
                       if(install_ash_clause(c8)) {
                          nextc = nextc0 +4;
                          installed_atoms++;}
                       else {
                          nextc = nextc0 + 3;}
                       binary_resolvents3(c3,c5,1,1,0);
                       c8 = clist[nextc - 1];
                       if(install_ash_clause(c8)) {
                          installed_atoms++;}
                       nextc = nextc0 + 1;
if(x_debug > 1) {
print0("USING ASH3: \n");
print0("C1,C2 resolve to C5, which resolves with C3 to C6\n");
print0("C4 and C6 resolve\n");
print0("\nC1: ");print_clause(c1);
print0("\nC2: ");print_clause(c2);
print0("\nC3: ");print_clause(c3);
print0("\nC4: ");print_clause(c4);
print0("\nC5: ");print_clause(c5);
print0("\nC6: ");print_clause(c6);
print0("\nThe atom generated is:");
print_clause(c7);
print0("\n\n");}

                        nextc = nextc0;
                        return(installed_atoms);}}}}}}
nextc = nextc0;
return(0);
}
 
/*-----------------------------------------------------------*/
/* ash4()                 ...........................#6
/*
/* ash4 looks for a 1-2 and/or a 1-3-1 units and sees if
/* two are complementary
/*
/* called by:    atom_sel_heuristic()        atom.c
/*
/* calls:        binary_resolvents2()        atom.c 
/*               binary_resolvents3()        atom.c 
/*               install_unit_and_test()     atom.c
/*               print_clause()              hdump.c
/*               time_now()                  herby.c
/*
/* returns:      1 if hb atom added; otherwise returns a 0
/*------------------------------------------------------------*/
 
int ash4()
{
clause *c1, *c2, *c3, *c4, *c5, *c6;          
int i, j, k, ia, ib, nextc0, installed_atoms; 
int ash4cont;

installed_atoms = 0;
nextc0 = nextc;
ash4cont = 0;

for(i = 0; i < nextc0; i++) {

   c1 = clist[i];    
   if(c1->c_nlit != 1) {continue;}

/*------------------------------------*/
/*------- c1 is a unit clause---------*/
/*------------------------------------*/
 
   for(j = nextc0 - 1 ; j >= 0; j--) {
      c2 = clist[j];     
      if(c2->c_nlit < 2 || c2->c_nlit > 3) continue;
      if(c2->c_nlit == 2) {
 
/*------------------------------------------*/
/*----c2 has 2 literals ( 1-2..pattern?)----*/
/*------------------------------------------*/
 
         for(ia = 1;  ia <=2;  ia++) {
            nextc = nextc0; 
            if(binary_resolvents2(c1,c2,1,ia,0)) {
               c3 = clist[nextc - 1];

/*-----------------------------------------------------*/
/* c3, the resolvent of c1 and c2 has one literal      */
/* and now save it on ulist and see if it resolves     */
/* with another unit on ulist.  If so, use as atom.    */
/*-----------------------------------------------------*/

               if(install_unit_and_test(c3)) { 
                  nextc = nextc0 + 2;
		  binary_resolvents3(c1,c2,1,ia,0);
                  c6 = clist[nextc - 1];
                  if(install_ash_clause(c6)) {

/*-----------------------------------------------------*/
/* install second atom, the resolved away literal      */
/* in the resolution of c1 and c2.                     */
/*-----------------------------------------------------*/

                     installed_atoms++;}

                  if(x_debug > 1) {
                     print0("USING ASH4a: \n");
                     print0("C1,C2 resolve to C3\n");
                     print0("\nC1: ");print_clause(c1);
                     print0("\nC2: ");print_clause(c2);
                     print0("\nC3: ");print_clause(c3);
                     print0("\n\n");}

                  nextc = nextc0;
                  return(10*installed_atoms +11);}}}}

/*----------------------------------------------------*/
/*-see if c2 has 3 literals ( 1-3..... pattern?)------*/
/*----------------------------------------------------*/

      if(c2->c_nlit == 3) {
         for(ia = 1; ia <= 3; ia++) {
            nextc = nextc0;
            if(binary_resolvents2(c1,c2,1,ia,0)) {
               c4 = clist[nextc - 1]; 
 
/*-------------------------------------------------------------*/
/* ----c5, a 1-3 clause, with 2 lits,                          */
/* has been generated; see if it resolves with a unit        --*/
/*-------------------------------------------------------------*/
 
               for(k = 0; k < nextc0; k++) {
                  c3 = clist[k];
                  if(c3->c_nlit != 1) {continue;}
                  for(ib = 1; ib <= 2; ib++) {
                     nextc = nextc0 + 1; 

                     if(binary_resolvents2(c3,c4,1,ib,0)) {
                        c5 = clist[nextc - 1];
 
/*------------------------------------------------------*/
/*----a 1-3-1 unit has been generated, place on ulist---*/
/*------------------------------------------------------*/
 
                        if(install_unit_and_test(c5)) { 			
                           nextc = nextc0 + 2;
		 	   binary_resolvents3(c1,c2,1,ia,0);
                           c6 = clist[nextc - 1];
                           if(install_ash_clause(c6)) {
                              installed_atoms++;}
                           nextc = nextc0 + 3;
			   binary_resolvents3(c3,c4,1,ib,0);
                           c6 = clist[nextc - 1];
                           if(install_ash_clause(c6)) {
                              installed_atoms++;}

                           if(x_debug > 1) {
                              print0("USING ASH4b: \n");
                              print0("C1,C2 resolve to C4, which resolves with C3\n");
                              print0("\nC1: ");print_clause(c1);
                              print0("\nC2: ");print_clause(c2);
                              print0("\nC3: ");print_clause(c3);
                              print0("\nC4: ");print_clause(c4);
                              print0("\n");
                              print0("The atom generated is:");
                              print_clause(c5);
                              print0("\n\n");}

                           nextc = nextc0;
                           return(10 * installed_atoms + 12);}
                       else {break;}}
		   nextc = nextc0;}}}}}}}
nextc = nextc0;
return(0);
}
 
/*---------------------------------------------------------------*/
/* ash5a()				.................#7
/*
/* adds atom to hb list if it appears as a ground instance
/* of a literal in T(N) or has at most one variable.
/* a constant is substituted for the variable.
/*
/* called by:    atom_sel_heuristic()      atom.c
/*
/* calls:        copy_atom_from_clause()   hgen.c
/*     		 install_hb_clause()       build.c
/*               calc_numb_var_in_lit()    hhash.c
/*               copy_gnd_atom_from_cls()  hgen.c
/*
/* returns:      1 if atom installed, 0 otherwise
/*--------------------------------------------------------------*/
 
int ash5a()
{
int i, j, jj, c_total, c_ptr, c_best, c_best_total;
clause *c, *c_hb;
literal *lit, *lit2, *c_best_lit;


for(i = nextc - 1; i > base0 ; i--) {
   c = clist[i]; 
   if(c->c_nlit == 1) {
      for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
         if((lit->l_id & ELIMLIT) == 0 &&
                (lit->l_id & NEGMASK) == last_atom_sign) {
            calc_numb_var_in_lit(lit,c->c_nv);

/*------------------------------------------------------------*/
/* if the number of variables in literal is zero, copy atom   */
/* from clause and addd to hb_list                            */
/*------------------------------------------------------------*/

            if(var_in_lit == 0) { 
               c_hb = copy_atom_from_clause(c,lit);
               if(install_hb_clause(c_hb)) {
                  last_atom_sign = last_atom_sign ^ NEGMASK;
                  return(i + 1);}}}}}}

for(i = nextc - 1; i > base0 ; i--) {
   c = clist[i]; 
   if(c->c_nlit == 1) {
      for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
         if((lit->l_id & ELIMLIT) == 0 &&
		 (lit->l_id & NEGMASK) == ~last_atom_sign) {
            calc_numb_var_in_lit(lit,c->c_nv);

/*------------------------------------------------------------*/
/* if the number of variables in literal is zero, copy atom   */
/* from clause and addd to hb_list                            */
/*------------------------------------------------------------*/

            if(var_in_lit == 0) { 
               c_hb = copy_atom_from_clause(c,lit);
               if(install_hb_clause(c_hb)) {
		  last_atom_sign = last_atom_sign ^ NEGMASK;
                  return(i + 1 + 10000);}}}}}}

for(i = nextc - 1; i > base0 ; i--) {
   c = clist[i]; 
   if(c->c_nlit == 1) {
      for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
         if((lit->l_id & ELIMLIT) == 0) {
            calc_numb_var_in_lit(lit,c->c_nv);

/*------------------------------------------------------------*/
/* if there is one variable in literal, copy atom and ground  */
/* it and then add to hb_list.                                */
/*------------------------------------------------------------*/

            if(var_in_lit == 1) {
               c_hb = copy_gnd_atom_from_cls(c,lit,const_alt);
               if(install_hb_clause(c_hb)) {
                  return(i + 1 + 20000);}}}}}}

for(i = 0; i < nextc ; i++) {
   c = clist[i]; 
   if(c->c_nlit == 2) { 
      for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
         if((lit->l_id & ELIMLIT) == 0) {
            calc_numb_var_in_lit(lit,c->c_nv);

/*------------------------------------------------------------*/
/* if the number of variables in literal is zero, copy atom   */
/* from clause and addd to hb_list                            */
/*------------------------------------------------------------*/
            if(var_in_lit == 0) { 
               c_hb = copy_atom_from_clause(c,lit);
               if(install_hb_clause(c_hb)) {
                  return(i + 1 + 30000);}}

/*------------------------------------------------------------*/
/* if there is one variable in literal, copy atom and ground  */
/* it and then add to hb_list.                                */
/*------------------------------------------------------------*/

            if(var_in_lit == 1) {
               c_hb = copy_gnd_atom_from_cls(c,lit,const_alt);
               if(install_hb_clause(c_hb)) {
                  return(i + 1 + 30000);}}}}}}  

return(0);
}
 
/*-------------------------------------------------------------*/
/* ash5b()                      ......................#8
/*
/* picks a ground atom on the unit clause list for
/* the next atom to explore
/*
/* called by:  atom_sel_heuristic()         atom.c
/*
/* calls:      copy_atom_from_clause()      hgen.c
/*             copy_gnd_atom_from_cls()     hgen.c
/*             install_hb_clause()          build.c
/*
/* returns:    unit number if clause installed
/*             otherwise returns 0
/*--------------------------------------------------------------*/

int ash5b()
{
int i,j,match;
clause *c_unit, *c_hb, *c;

/*---------------------------------------------------*/
/* look for unit clause of correct polarity with no  */
/* variables                                         */
/*---------------------------------------------------*/

for(i = 0; i < nextunit; i++) {
   c_unit = unit_list[i];
   if(c_unit->c_nv == 0 && 
		(c_unit->c_lit->l_id & FCNMASK) == last_atom_sign) {
      for(j = 0; j < nextc; j++) {
         match = 0;
	 c = clist[j];
	 if(c->c_nlit != 1) {continue;}
         if(c->c_X == c_unit->c_X && c->c_Y == c_unit->c_Y) {
            match = 1;
            break;}}
      if(match == 0) {
         c_hb = copy_atom_from_clause(c_unit,c_unit->c_lit);
         if(install_hb_clause(c_hb)) {
            last_atom_sign = last_atom_sign ^ NEGMASK;
            return(i + 1);}}}}
	   
/*----------------------------------------------------*/
/* if correct polarity cannot be found, search for    */
/* unit clause with incorrect polarity                */
/*----------------------------------------------------*/

for(i = 0; i < nextunit; i++) {
   c_unit = unit_list[i];
   if(c_unit->c_nv == 0 && 
		(c_unit->c_lit->l_id & FCNMASK) == ~last_atom_sign) {
      for(j = 0; j < nextc; j++) {
         match = 0;
	 c = clist[j];
	 if(c->c_nlit != 1) {continue;}
         if(c->c_X == c_unit->c_X && c->c_Y == c_unit->c_Y) {
            match = 1;
            break;}}
      if(match == 0) {
         c_hb = copy_atom_from_clause(c_unit,c_unit->c_lit);
         if(install_hb_clause(c_hb)) {
            last_atom_sign = last_atom_sign ^ NEGMASK;
            return(i + 1 + 10000);}}}}

/*---------------------------------------------------------*/
/* if there is a unit clause with one variable, ground     */
/* variable.                                               */
/*---------------------------------------------------------*/

for(i = 0; i < nextunit; i++) {
   c_unit = unit_list[i];
   if(c_unit->c_nv == 1) { 
      c_hb = copy_gnd_atom_from_cls(c_unit,c_unit->c_lit,const_alt);
      if(install_hb_clause(c_hb)) {
	 last_atom_sign = last_atom_sign ^ NEGMASK;
         return(i + 1 + 20000);}}}

/*---------------------------------------------------------*/
/* if there is a unit clause with more than one var, grnd  */
/* variable.                                               */
/*---------------------------------------------------------*/

for(i = 0; i < nextunit; i++) {
   c_unit = unit_list[i];
   if(c_unit->c_nv > 1) { 
      c_hb = copy_gnd_atom_from_cls(c_unit,c_unit->c_lit,const_alt);
      if(install_hb_clause(c_hb)) {
	 last_atom_sign = last_atom_sign ^ NEGMASK;
         return(i + 1 + 30000);}}}

return(0);
}


/*----------------------------------------------------------------*/
/* ash5c()			           ..............#9
/*
/*
/* adds atom to hb list if it appears as a ground instance
/* of a literal from a clause in T(N) with at least two 
/* literals. 
/*
/* called by:   atom_sel_heuristics()      atom.c
/*
/* calls:       copy_atom_from_clause()    hgen.c
/*              copy_gnd_atom_from_cls()   hgen.c
/*              install_hb_clause()        build.c
/*              calc_numb_var_in_lit()     herby.c
/*
/* returns:     1 if atom installed, 0 otherwise
/*----------------------------------------------------------------*/
 
int ash5c()
{
int i;
clause *c, *c_hb;
literal *lit;

for(i = base0; i >= base0 - nc + 1; i--) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next ) {
      if((lit->l_id & ELIMLIT) == 0) {
         calc_numb_var_in_lit(lit,c->c_nv);
         if(var_in_lit == 0) {
            c_hb = copy_atom_from_clause(c, lit);
            if(install_hb_clause(c_hb)) {
               return(i + 1);}}
         if(var_in_lit == 1) {
            c_hb = copy_gnd_atom_from_cls(c,lit,1);
            if(install_hb_clause(c_hb)) {
               return(i + 1);}}}}}

for(i = base0; i >= base0 - nc + 1 ; i--) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next ) {
      if((lit->l_id & ELIMLIT) == 0) {
         calc_numb_var_in_lit(lit,c->c_nv);
         c_hb = copy_gnd_atom_from_cls(c,lit,1);
         if(install_hb_clause(c_hb)) {
            return(i + 1 + 20000);}}}}

for(i = base0 - nc; i >= 0 ; i--) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next ) {
      if((lit->l_id & ELIMLIT) == 0) {
         calc_numb_var_in_lit(lit,c->c_nv);
         if(var_in_lit == 0) {
            c_hb = copy_atom_from_clause(c, lit);
            if(install_hb_clause(c_hb)) {
               return(i + 1);}}
         if(var_in_lit == 1) {
            c_hb = copy_gnd_atom_from_cls(c,lit,1);
            if(install_hb_clause(c_hb)) {
               return(i + 1);}}}}}

for(i = nextc - 1; i > base0 ; i--) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next ) {
      if((lit->l_id & ELIMLIT) == 0) {
         calc_numb_var_in_lit(lit,c->c_nv);
         if(var_in_lit == 0) {
            c_hb = copy_atom_from_clause(c, lit);
            if(install_hb_clause(c_hb)) {
               return(i + 1 + 10000);}}
         if(var_in_lit == 1) {
            c_hb = copy_gnd_atom_from_cls(c,lit,1);
            if(install_hb_clause(c_hb)) {
               return(i + 1 + 10000);}}}}}

for(i = base0 - nc; i >= 0 ; i--) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next ) {
      if((lit->l_id & ELIMLIT) == 0) {
         calc_numb_var_in_lit(lit,c->c_nv);
         c_hb = copy_gnd_atom_from_cls(c,lit,1);
         if(install_hb_clause(c_hb)) {
            return(i + 1 + 20000);}}}}

for(i = nextc - 1; i > base0 ; i--) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next ) {
      if((lit->l_id & ELIMLIT) == 0) {
         calc_numb_var_in_lit(lit,c->c_nv);
         c_hb = copy_gnd_atom_from_cls(c,lit,1);
         if(install_hb_clause(c_hb)) {
            return(i + 1 + 30000);}}}}

return(0);
}

/*------------------------------------------------------------*/
/* resolution_stage()     .............. #10
/*  
/* initially resolve base clauses together
/*
/* called by:  main()                 herby.c
/*
/* calls:      binary_resolvents()    infer.c
/*             print_clause()         hdump.c
/*
/* returns:    nil
/*-------------------------------------------------------------*/

int resolution_stage()
{
int i,j, k, l, number_inf, number_literals, bin_infs;
clause *c1, *c2;

number_inf = 0;
bin_infs = -1;

/*-----------------------------------------------*/
/* now resolve unit base clauses with the other  */
/* clauses to produce new base clauses.  begin   */
/* with unit base clauses with no variables,     */
/* the one variable, then two, and so on until   */
/* either there are no more clauses to add or    */
/* the number of clauses added equals the number */
/* of base clauses or exceeds 600                */
/*-----------------------------------------------*/

for(k = base0; k > 0; k--) {
   for(i = base0; i >= k; i--) {
      c1 = clist[i];
      for(j = base0 + number_inf; j >= 0; j--) {
	 if(i == j) {continue;}
         c2 = clist[j];
         bin_infs = binary_resolvents(c1,c2);

/*------------------------------------------------------------*/
/* if NULL clause found, return proof                         */
/*------------------------------------------------------------*/

	 if(clist[nextc - 1]->c_nlit == 0) {
	    printf("\nCONTRADICTION FOUND IN BCR HEURISTIC\n");
	    fflush(stdout);
	    OutPtr += sprintf(OutPtr,
	           "\nCONTRADICTION FOUND IN BCR HEURISTIC\n");
            for(l = nextc - 1; l > base; l--) {
	       c1 = clist[l];
	       if(c1->c_nlit == 0 || (c1->c_bits & LIST)) {
	          c1->c_bits |= LIST;
                     c1->c_lpar->c_bits |= LIST;
		     c1->c_rpar->c_bits |= LIST;}}
               for(l = base + 1; l < nextc; l++) {
                  c1 = clist[l];
                  if(c1->c_bits & LIST) {
                     OutPtr += sprintf(OutPtr,"\n");
                     print_clause(c1);}}
               OutPtr += sprintf(OutPtr,"\n");
               return(2);}

/*----------------------------------------------------------*/
/* if NULL clause not found, decide whether to loop again   */
/*----------------------------------------------------------*/

            number_inf += bin_infs;
	    if(number_inf > 600 || number_inf > base0 + 25) {
	       goto prepare_return;}
            if(number_inf > 25) {goto prepare_return;}}}}


/*---------------------------------------------------*/
/* if debug level > 1, print clauses generated       */
/* by the base clause resolution heuristic           */
/*---------------------------------------------------*/

prepare_return:
if(x_debug > 3) {
   OutPtr += sprintf(OutPtr,
	  "\n%d clauses are added by BCR Heuristic:\n",number_inf);
   for(i = base + 1; i <= base + number_inf; i++) {
      c1 = clist[i];
      print0("\nRes_stage: ");
      print_clause(c1);}
   print0("\n\n");}

base = base0 + number_inf; 
base1 = base;
return(1);
}

/*------------------------------------------------------------*/
/* resolution_stage2()     .............. #11
/*  
/* initially resolve base clauses together
/*
/* called by:  main()                 herby.c
/*
/* calls:      binary_resolvents()    infer.c
/*             print_clause()         hdump.c
/*
/* returns:    nil
/*-------------------------------------------------------------*/

int resolution_stage2()
{
int ii, i, j, k, l, number_inf, number_literals, bin_infs;
clause *c1, *c2, *c3;
int base00;
int infs;

number_inf = 0;
bin_infs = -1;
base00 = base;

/*-----------------------------------------------*/
/* now resolve unit base clauses with the other  */
/* clauses to produce new base clauses.  begin   */
/* with unit base clauses with no variables,     */
/* the one variable, then two, and so on until   */
/* either there are no more clauses to add or    */
/* the number of clauses added equals the number */
/* of base clauses or exceeds 600                */
/*-----------------------------------------------*/

for(i = base; i <= base; i++) {
   for(j = 0; j <= base; j++) {
      if(i == j) {continue;}
      c1 = clist[i];
      c2 = clist[j];
      if(simple_subsume(c1,c2)) {c2->c_bits |= DEAD;}
      else if(simple_subsume(c2,c1)) {c1->c_bits |= DEAD;}
      if(c2->c_nlit > 4) {continue;}
      if(c1->c_nlit < 3) {
         for(ii = 1; ii <= c2->c_nlit; ii++) {
	    bin_infs = 0;
            infs = binary_resolvents2(c1,c2,1,ii,0);

/*------------------------------------------------------------*/
/* if NULL clause found, return proof                         */
/*------------------------------------------------------------*/

            if(infs == 0) continue;
	    else bin_infs += infs;
	    if(clist[nextc - 1]->c_nlit == 0) {
	       clist[i + 1] = clist[nextc - 1];
	       clist[i + 1]->c_id = i + 2;
	       nextc = i + 2;
	       base = i + 1;
	       base1 = base; 
	       OutPtr += sprintf(OutPtr,
	                "\nCONTRADICTION FOUND IN BCR2 HEURISTIC\n");
	       printf("\nCONTRADICTION FOUND IN BCR2 HEURISTIC\n");
               fflush(stdout);
               for(l = nextc - 1; l > base; l--) {
	          c1 = clist[l];
	          if(c1->c_nlit == 0 || (c1->c_bits & LIST)) {
	             c1->c_bits |= LIST;
                     c1->c_lpar->c_bits |= LIST;
		     c1->c_rpar->c_bits |= LIST;}}
               return(2);}
	    c3 = clist[nextc - 1];
            if(c3->c_nlit == 1) {
	       add_clause_to_ht(c3->c_X,c3->c_Y,c3->c_id,c3->c_nlit);}
            if(bin_infs != 0 && clist[nextc - 1]->c_nlit > 0) {
	       nextc--;
	       continue;}
	    
            if(simple_subsume(clist[nextc - 1],c2)) {
	       c2->c_bits |= DEAD;}

/*----------------------------------------------------------*/
/* if NULL clause not found, decide whether to loop again   */
/*----------------------------------------------------------*/

            number_inf += bin_infs;}}}}


/*---------------------------------------------------*/
/* if debug level > 1, print clauses generated       */
/* by the base clause resolution heuristic           */
/*---------------------------------------------------*/

if(x_debug >= 1) {
   OutPtr += sprintf(OutPtr,
	  "\n%d clauses are added by BCR2 Heuristic:\n",number_inf);
   for(i = 0; i <= base + number_inf; i++) {
      c1 = clist[i];
      print0("\nRes_stage2: ");
      print_clause(c1);}
   print0("\n\n");}
base = base + number_inf; 
base1 = base;
return(1);
}
