/*----------------------------------------------------------------*/
/* build.c           HERBY
/*
/* Copyright Newborn Software 1995 - 1999, 2000
/*
/* functions:
/* init_search_space()     ......................#1
/* install_clause()        ......................#2
/* install_hb_clause()     ......................#3
/* install_ash_clause()..........................#4
/* install_unit_and_test().......................#5
/* int calc_inferences()   ......................#6
/* build_tree()            ......................#7
/* save_failure_pair()     ......................#8
/* simple_subsume()        ......................#9
/*----------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#include "constHS.h"
#include "clauseHS.h"

/*----------------------------------------------------------------*/
/* externally defined functions                                   */
/*----------------------------------------------------------------*/

extern int time_now();           /* herby.c:   get time of day    */

extern int binary_resolvents();  /* infer.c:   inference rules    */
extern void clear_var_substitution_tble();

extern clause *copy_gnd_atom_from_cls();  /* hgen.c       */ 
extern clause *copy_atom_from_clause();
extern clause *copy_clause();
extern clause *gen_unified_lit();

extern void print_clause();      /* hdump.c: print clause fxn     */
extern void print0();

extern int atom_sel_heuristics(); /* atom.c: picks hb atoms       */

/*----------------------------------------------------------------*/
/* externally defined constants                                   */
/*----------------------------------------------------------------*/

extern char *OutPtr;
 
extern int base, ply, iter_depth;
extern int nextc, nexthb, nextunit;
extern int phase;

extern int x_debug;	
extern int o_maxtime, build_start_time, build_time;
extern int o_fulltree;
extern int o_charmax;
extern int x_complete_proof;

extern int s_nod;                       /* # nodes in tree        */
extern int open_problems;

extern int var_in_lit;
extern int const_alt;

extern int clause_space[];            /* where clauses are stored  */
extern clause *clist[];               /* pointers to clauses       */
extern int hb_space[];
extern clause *hb_list[];
extern int unit_space[];
extern clause *unit_list[];

extern *rand_array;
extern int numb_rows, numb_cols, numb_entries;
extern int number_preds, number_fxns;
extern int rows_less_nfxns;

extern hnode *clause_hash_table;
extern int o_arraysize;

extern dnode depth[];            /* pointers to pointers of clauses  */
extern int fail[];
extern int fail_count[];
extern int dead_left_branches[];
extern int dead_right_branches[];
extern int heuristic_applied[];
extern int branch[];

extern int fail_clause1, fail_clause1x;
extern int fail_clause1_ply, fail_clause1_node;

extern int fail_clause_left[200][18];

extern int fail_lits1[MAX_LIT_IN_CLAUSE];

extern int fail_clause2, fail_clause2x;
extern int fail_clause2_ply, fail_clause2_node;
extern int fail_lits2[MAX_LIT_IN_CLAUSE];

extern int fail_clause3, fail_clause3x;
extern int fail_clause3_ply, fail_clause3_node;
extern int fail_lits3[MAX_LIT_IN_CLAUSE];

extern int fail_clause4, fail_clause4x;
extern int fail_clause4_ply, fail_clause4_node;
extern int fail_lits4[MAX_LIT_IN_CLAUSE];
extern int last_fail;

extern int orig_branch_sign[HSPACE_CLAUSES];

/*------------------------------------------------------------------*/
/* internally defined functions                                     */
/*------------------------------------------------------------------*/

void init_search_space();
int install_clause();
int install_hb_clause();
int install_ash_clause();
int install_unit_and_test();
int calc_inferences();
int build_tree();
void save_failure_pair();
int simple_subsume();
void init_random_hash_array();

/*-------------------------------------------------------------------*/
/* init_search_space()            ......................#1
/*
/* sets pointers to where first clause on
/* clist, hb_list and unit_list goes.
/*
/* called from:   main()       herby.c
/*
/* calls:         nobody
/*
/* returns:       nil
/*-------------------------------------------------------------------*/

void init_search_space()
{	
int cspace_ptr;

for(cspace_ptr = 0; cspace_ptr < CSPACE_SIZE; cspace_ptr++) {
   clause_space[cspace_ptr] = 0;}

nextc = 0;
clist[nextc] = (clause *) clause_space;

nexthb = 0;
hb_list[nexthb] = (clause *) hb_space;

nextunit = 0;
unit_list[nextunit] = (clause *) unit_space;

clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);
}

/*------------------------------------------------------------------*/
/* install_clause(c)             ......................#2
/*
/* places clause c at the end of clist.
/*
/* install_clause() is called after the clause has been
/* generated by generate_clause().  It is pointed to by c.
/* Install_clause() has the job of deciding whether to
/* actually list the clause with the other clauses at that
/* level in the tree.
/*
/* called by:  binary_resolvents()         infer.c
/*
/* calls:      simple_subsume()            build.c
/*
/* returns     1 if the clause was indeed installed,
/*             0 if the clause was not installed
/*------------------------------------------------------------------*/
 
int install_clause(c)
clause *c;
{
int cl_ptr;
dnode *d;
clause *c1;
literal *lit;

int i;

/*--------------------------------------------*/
/* if clist is full, don't add clause         */
/*--------------------------------------------*/

if(nextc > CSPACE_CLAUSES - 2) {
   printf("\nToo many clauses(%ld) \n",nextc);fflush(stdout);
   return(-1);}

/*--------------------------------------------*/
/* if clause is on clist, don't add it again  */
/*--------------------------------------------*/

for(cl_ptr = 0; cl_ptr < nextc; cl_ptr++) {
   c1 = clist[cl_ptr]; 
   if(c1->c_X == c->c_X && c1->c_Y == c->c_Y) {return(0);}}

/*---------------------------------------------*/
/* if clause is s-subsumed, don't add          */
/*---------------------------------------------*/

if(c->c_nlit != 0) {
   for(cl_ptr = base + 1; cl_ptr < nextc; cl_ptr++) {
      c1 = clist[cl_ptr];
      if(c1->c_bits & DEAD) {continue;}
      if(simple_subsume(c1,c)) {return(0);}}}

/*---------------------------------------------*/
/* don't add clause if subsumed by atom on path*/
/*---------------------------------------------*/

for(i = 0; i < ply; i++) {
   c1 = hb_list[i];
   if(simple_subsume(c1,c)) {return(0);}}

/*---------------------------------------------*/
/* if no lits, mark as a failure node          */
/*---------------------------------------------*/

if(c->c_nlit == 0) {
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {

/*---------------------------------------------*/
/* give atom credit for failure.  this is      */
/* used to see if an atom actually causes      */
/* a failuree in this part of the tree.        */
/* If it doesn't, the atom is totally          */
/* ineffective in causing a failure node.      */
/*---------------------------------------------*/

      if(c->c_lit->l_next == 0) {
	 if(branch[ply] == 0) {
            dead_left_branches[ply - 1] = c->c_rpar->c_id;}
	 else {
            dead_right_branches[ply - 1] = c->c_rpar->c_id;}} 

      fail_count[lit->l_X - 1] += 1;}
   fail[ply + 1] = 2;}

d = &depth[ply];
d->d_tc = nextc;
clist[nextc++] = c;
clist[nextc] = c->c_end;
return(1);
}
  
/*------------------------------------------------------------------*/
/* install_hb_clause(c)             ......................#3
/*
/* installs a clause on the atom list
/*
/* called by:    ash5(), ash6a(), ash6b(),        atom.c
/*               ash6c()
/*
/* calls:        Nobody
/*
/* returns:      1 if clause installed, 0 if NULL clause installed
/*------------------------------------------------------------------*/
 
int install_hb_clause(c)
clause *c;
{
clause *c1;
literal *lit1;
int i;

/*-----------------------------------------*/
/* if already on list, do not add again    */
/*-----------------------------------------*/

for(i = 0; i < nexthb; i++) {
   c1 = hb_list[i];
   if(c1->c_X == c->c_X && c1->c_Y == c->c_Y) {return(0);}
   if(c1->c_X == -c->c_X && c1->c_Y == -c->c_Y) {return(0);}}

/*------------------------------------------*/
/* since not found, add it to list          */
/*------------------------------------------*/

hb_list[nexthb++] = c;
hb_list[nexthb] = c->c_end;
return(1);
}

 
/*---------------------------------------------------------*/
/* install_ash_clause()        .....................#4
/*
/* places a clause generated during ash on clist
/*
/* called by:  ash1(), ash2(), ash3(),    atom.c
/*             install_unit_and_test()    atom.c
/*
/* calls:      copy_gnd_atom_from_cls()   hgen.c
/*             copy_atom_from_clause()    hgen.c
/*             calc_numb_var_in_lit()     hhash.c
/*
/* returns:    1 => installed; 0 => not installed
/*----------------------------------------------------------*/
 
int install_ash_clause(c)
clause *c;
{
literal  *lit, *lit1;
clause *c1, *c_hb;
int i;

for(lit = c->c_lit ; lit != 0; lit = lit->l_next) {
   if((lit->l_id & ELIMLIT) == 0) {goto calc_lit_vars;}}

for(lit = c->c_lit ; lit != 0; lit = lit->l_next) {
   if(lit->l_X == ply || 
       (((lit->l_id & ELIMLIT) != 0) && phase == 1)) {
      break;}}

calc_lit_vars:
if(c->c_nv != 0) {calc_numb_var_in_lit(lit,c->c_nv);} 
else {var_in_lit = 0;}

if(var_in_lit == 1) {
   lit->l_id = lit->l_id & ~ELIMLIT;
   c_hb = copy_gnd_atom_from_cls(c,lit,const_alt);}
else if(var_in_lit == 0) {
   lit->l_id = lit->l_id & ~ELIMLIT;
   c_hb = copy_atom_from_clause(c,lit);}
else {return(0);} 

c_hb->c_lit->l_id = lit->l_id;

/*-------------------------------------------*/
/* if c1 or its complement is on hblist,     */
/* don't add again.                          */
/*-------------------------------------------*/

for(i = 0; i < nexthb; i++) {
   c1 = hb_list[i];
   if(c1->c_X == c_hb->c_X && c1->c_Y == c_hb->c_Y) {return(0);}
   if(c1->c_X == -c_hb->c_X && c1->c_Y == -c_hb->c_Y) {return(0);}}

/*-------------------------------------------*/
/* add clause to hb_list, increment nextc,   */
/*  set ptr to next clause                   */
/*-------------------------------------------*/

hb_list[nexthb++] = c_hb;   
hb_list[nexthb] = c_hb->c_end; 
return(1);
}
 
/*----------------------------------------------------------------*/
/* install_unit_and_test()          ......................#5
/*
/* installs a unit clause on units_list and checks t see if the
/* one added is the complement of an other
/*
/* called by:  ash4()                         atom.c
/*
/* calls:      binary_resolvents2()           atom.c
/*             install_ash_clause()           atom.c
/*             copy_clause()                  hgen.c
/*             clear_var_substitution_tble()  infer.c
/*             unify()                        unify.c
/*             gen_unified_lit()              hgen.c
/*
/* returns:    1 if complementary pair found, otherwise 0
/*----------------------------------------------------------------*/
 
int install_unit_and_test(c3)
clause *c3;
{
clause *c4, *c5, *c6, *c7, *c8;
int i, nextc0, j, k;
int c3_length;
int dont_sort;

/*-------------------------------------------*/
/*  first test to see if c3 is on unit list  */
/*-------------------------------------------*/

if(nextunit > 1000) return(0);

nextc0 = nextc; 
dont_sort = 0;

for(i = 0; i < nextunit; i++) {
   c7 = unit_list[i]; 
   if(c3->c_X == c7->c_X && c3->c_Y == c7->c_Y) {return(0);}}

/*-------------------------------------------*/
/* c3 is not on unit list, see if it can     */
/* be installed, i.e., not too many clauses  */
/*-------------------------------------------*/
 
if(nextunit > USPACE_CLAUSES - 3) {
   dont_sort = 1;
   c6 = c3;}
 
else { 
      c6 = copy_clause(c3,2);
      nextunit++;
      unit_list[nextunit] = c6->c_end;}

for(i = nextunit - 2; i >= 0; i--) {
   c4 = unit_list[i]; 
   nextc = nextc0;
   if(binary_resolvents2(c6,c4,1,1,0))  {
      if(install_ash_clause(c6)) {
	 nexthb--; nextc = nextc0;
         clear_var_substitution_tble(c4->c_bits,c4->c_nv);
         clear_var_substitution_tble(c6->c_bits,c6->c_nv);
         unify(&c6->c_lit->l_id,&c4->c_lit->l_id);
         c8 = gen_unified_lit(c6);
         if(!install_ash_clause(c8)) {continue;};
         nextc = nextc0;
         return(1);}}}

/*------------------------------------------*/
/* sort unit clause list by length          */
/*------------------------------------------*/

if(dont_sort == 0) {
   c3_length = (int) c6->c_end - (int) c6->c_lit;

   for(i = 0; i < nextunit - 1; i++) {
      c8 = unit_list[i];
      if(c3_length < (int) c8->c_end - (int) c8->c_lit) {
         for(j = nextunit - 1; j > i; j--) {
	    unit_list[j] = unit_list[j - 1];}
         unit_list[i] = c6; 
         break;}}}

nextc = nextc0;
return(0);
}

/*-----------------------------------------------------------------*/
/* calc_inferences()    ......................#6
/*
/* called by:  build_tree()            build.c
/*
/* calls:      binary_resolvents()     infer.c
/*             print_clause()          hdump.c
/*             simple_subsume()        build.c
/*
/* returns:    0, or the number of inferences generated
/*-----------------------------------------------------------------*/
 
int calc_inferences(plyM1)
int plyM1;
{
clause *c,*c1;
clause *hb_unit;
int cl_ptr, cl_ptr1, cl_pt_last;
int number_of_inferences;
int new_inferences;
int next1;
int i;

number_of_inferences = 0;
next1 = nextc;

/*-------------------------------------------------*/
/* resolve each clause on clist with atom hb_unit  */
/* if null clause is found, exit with a 0          */
/*-------------------------------------------------*/

   hb_unit = hb_list[plyM1];

/*-------------------------------------------------*/
/* first if unit clause is known to cause a failure*/
/* at this level, use it to generate a failure     */
/* clause.                                         */
/*-------------------------------------------------*/

   if(branch[ply] == 0 && dead_left_branches[ply - 1] != 0) {
      c = clist[dead_left_branches[ply - 1] - 1];
      new_inferences = binary_resolvents(hb_unit,c);
      return(0);}
   if(branch[ply] == 1 && dead_right_branches[ply - 1] != 0) {
      c = clist[dead_right_branches[ply - 1] - 1];
      new_inferences = binary_resolvents(hb_unit,c);
      return(0);}

   for(cl_ptr = 0; cl_ptr < nextc ; cl_ptr++) {
      c = clist[cl_ptr];
      if(c->c_bits & DEAD) {continue;} 
      new_inferences = binary_resolvents(hb_unit,c);
      if(new_inferences == -1) {return(-1);}
      number_of_inferences += new_inferences;
      if(fail[ply + 1] >= 2) {return(0);}}

/*-------------------------------------------------*/
/* eliminate clauses that are s-subsumed           */
/*-------------------------------------------------*/

cl_pt_last = nextc - number_of_inferences -1;
for(cl_ptr = nextc - 1; cl_ptr > cl_pt_last; cl_ptr--){
    c = clist[cl_ptr];
    if(c->c_bits & DEAD) {continue;}
    for(cl_ptr1  = cl_ptr - 1; cl_ptr1 >= next1; cl_ptr1--) {
       c1 = clist[cl_ptr1];
       if(c1->c_bits & DEAD) {continue;}
       if(simple_subsume(c,c1)) {c1->c_bits |= DEAD;}}}

/*-------------------------*/
/* debugging message       */
/* print clauses generated */
/*-------------------------*/

if(x_debug > 1) {
   printf("\n\nPRINT INFERENCES GENERATED\n\n");
   for(cl_ptr = nextc-number_of_inferences;cl_ptr < nextc;cl_ptr++) {
      print_clause(clist[cl_ptr]);
      print0("\n");}
   printf("\nEND OF INFERENCES GENERATED\n\n");}

return(number_of_inferences);
}							
 
/*----------------------------------------------------------------*/
/* build_tree()                  ......................#7
/*
/* build_tree will return TRUE if we have generated the
/* empty clause during our explorations of the tree.
/* Note: we assume that there are at least two base clauses
/*
/* called by:  prove_by_iterative_deepening()    herby.c
/*
/* calls:      calc_inferences()                 build.c
/*             atom_sel_heuristics()             atom.c
/*             print_clause()                    hdump.c
/*             time_now()                        herby.c
/*             save_failure_pair()               build.c
/*
/* returns:    1  => proof found on current iteration
/*             -1 => no proof -- and at root
/*             0  => no proof -- out of time and no proof
/*             2  => no proof -- no atom generated by heuristics
/*             4  => no proof -- too many clauses on clist 
/*----------------------------------------------------------------*/
 
int build_tree()
{
dnode *d;
int number_of_clauses;
int advance;
clause *c, *c1;
int i, j, jjj;
literal  *lit;
char gcxx;
int ash;

advance = 1;
for(i = 0; i < HSPACE_CLAUSES ; i++) {
   dead_left_branches[i] = dead_right_branches[i] = 0;
   heuristic_applied[i] = 0;
   branch[i] = fail[i] = fail_count[i] = 0;}

for(i = 0; i < 200; i++) {
   for(j = 0; j <18; j++) {fail_clause_left[i][j] = 0;}}

depth[0].d_tc = base;               /* initial generating clause  */
depth[0].d_ec = base + 1;
depth[0].d_bc = 0;

nextc = base + 1; 

/*
printf("NEXTC=%ld TIME=%d XARS=%d\n",nextc,o_maxtime,o_charmax);
fflush(stdout);
OutPtr += sprintf(OutPtr,
       "NEXTC=%ld TIME=%d XARS=%d\n",nextc,o_maxtime,o_charmax);
*/

open_problems = 0;

/*--------------------------------------------------------------*/
/*      determine where to store first clause                   */
/*--------------------------------------------------------------*/

clist[nextc] = clist[base]->c_end;

/*---------------------------------------------------------------*/
/*       begin construction                                      */
/*---------------------------------------------------------------*/

for(ply = 1;;) {
evaluate: 
   d = &depth[ply];             
   if(ply == 0) { 

/*--------------------------------------------------------------*/
/* if at root of the tree and there are no open_problems,       */
/* return with "proof found".  Otherwise, no proof found.       */
/*--------------------------------------------------------------*/

      if(open_problems == 0) {return(1);}
      else {return(-1);}}

/*--------------------------------------------------------------*/
/* if node is a failure node or if it is at max search depth:   */
/*--------------------------------------------------------------*/

   if(fail[ply] == 2 || ply == iter_depth + 1) {
      if(x_debug > 0) {

/*---------------------------------------------------------------*/
/* print path to first and last open problems                    */
/*---------------------------------------------------------------*/

         print0("\nPATH: ");
         for(i = 1; i < ply; i++) {
            printf("%ld",branch[i]);
	    OutPtr += sprintf(OutPtr,"%ld",branch[i]);
            if(i % 5 == 0) {print0(" ");}}
         print0("\n");
	    }

/*---------------------------------------------------------------*/
/* if at maximum depth but node is NOT a failure, then,          */
/* depending on whether full tree is to be constructed before    */
/* incrementing iter_depth:                                      */
/*---------------------------------------------------------------*/
      
      if(ply == iter_depth + 1 && fail[ply] < 2) {
         if(x_debug > 0) {print0("\n");}

/*---------------------------------------------------------------*/
/* if not generating full tree (o_fulltree = 0) increment        */
/* search depth and use new value of iter_depth as maximum       */
/* construction depth.                                           */
/*---------------------------------------------------------------*/

         if(o_fulltree == 0) {
            iter_depth++;
            goto select_atom;}

/*---------------------------------------------------------------*/
/* in this case (o_fulltree = 1) the fulltree will be built on   */
/* each iteration before going on to the next depth              */
/*---------------------------------------------------------------*/

         if(x_debug > 0) {print0("X");}
         open_problems++;}

/*----------------------------------------------*/
/* print out tree as construction progresses    */
/*----------------------------------------------*/

      else{
         if(ply == iter_depth + 1) {

/*---------------------------------------------------------*/
/* at maximum depth and a failure node is arrived at       */
/* get clause, c,  that caused the failure                 */
/*---------------------------------------------------------*/

            c = clist[nextc - 1];
	    if(advance > 0) {save_failure_pair(c);}
	    if(fail_clause2 != 0) {
               if(x_debug > 1) {
	          print0(" FAILS: ");
	          print_clause(c);}
	       ply = 0;
	       return(1);}}

         else {

/*------------------------------------------------------------*/
/* a failure, but not at maximum depth.                       */
/*------------------------------------------------------------*/

            c = clist[nextc - 1];
	    if(advance > 0) {save_failure_pair(c);}
	    if(fail_clause2 != 0) {
	       ply = 0;
	       return(1);}
            if(x_debug > 1) {
               print0(" FAILS: ");
               if(advance == 1) {print_clause(c);}}}}

/*----------------------------------------------------------*/
/* back up the tree and go to evaluate                      */
/*----------------------------------------------------------*/

back_up_tree:
      fail[ply] = 0;
      ply--;
      advance = -1;
      fail[ply]++;
      if(fail_count[ply - 1] == 0 && ply > 0) {
         if(x_debug > 0) {
            print0("\nPATH: ");
            for(i = 1; i < ply; i++) {
               printf("%ld",branch[i]);
	       OutPtr += sprintf(OutPtr,"%ld",branch[i]);
	       if(i % 5 == 0) print0(" ");}
            print0(" FAILS: \n");}

/*-----------------------------------------------------*/
/* complement the atom so that pruning won't throw off */ 
/* alternating order of atoms at this level            */
/*-----------------------------------------------------*/

         c = hb_list[ply - 1];
	 if(fail[ply] == 1) {
         c->c_lit->l_id = c->c_lit->l_id ^ NEGMASK;
         c->c_X = -c->c_X;
         c->c_Y = -c->c_Y;
         c->c_lit->l_X = -c->c_lit->l_X;
         c->c_lit->l_Y = -c->c_lit->l_Y;
	 goto back_up_tree;}}
      goto evaluate;}

select_atom:
   fail[ply + 1] = 0;
   s_nod++; 

/*------------------------------------------------------*/
/* check time every 20 nodes, return if time expired    */
/*------------------------------------------------------*/

   if((s_nod % 20) == 0) {
      if(time_now() - build_start_time > o_maxtime) {
         return(0);}}

/*------------------------------------------------------*/
/* select herbrand base element                         */
/* if no atom selected, return a 2                      */
/*------------------------------------------------------*/

   if(ply > nexthb) {  
      if(time_now() - build_start_time > o_maxtime) {
         return(0);}
      if(x_debug > 0) {
         print0("\nGENERATE ATOM ");}
      if((ash = atom_sel_heuristics()) == 0) {return(2);}

/*-------------------------------------------------------*/
/* print the time taken by the search, the number of     */
/* nodes in the search tree, the number of clauses       */
/* on clist, and the number of unit clauses on unit_list */
/*-------------------------------------------------------*/

/*
      printf(" T%ld  N%ld  C%ld  U%ld\n",
          time_now() - build_start_time,s_nod,nextc - base - 1,
          nextunit);fflush(stdout);
      OutPtr += sprintf(OutPtr," T%ld  N%ld  C%ld  U%ld\n",
          time_now() - build_start_time,s_nod,nextc - base - 1,
          nextunit);*/}


/*--------------------------------------------------------*/
/* pick next hb element for construction                  */
/*--------------------------------------------------------*/

pick_next_hb:
   if(nexthb == 595) {
//    printf("\nNo room for more atoms!\n");
//    fflush(stdout);
      return(2);}

   if(nexthb < ply) { 

/*--------------------------------------------------------*/
/* if there is no atom for this level in the tree, return */
/* this will appen if the atom selection heuristics fail  */
/* to select an atom.                                     */
/*--------------------------------------------------------*/

      printf("\nNo HB atom! nexthb = %d ply = %d\n",nexthb,ply);
      fflush(stdout);
      OutPtr += sprintf(OutPtr,
             "\nNo HB atom! nexthb = %d ply = %d\n",nexthb,ply);
      return(2);}

/*--------------------------------------------------------*/
/* get herbrand base atom for this level                  */
/* complement it from the last time it was used           */
/*--------------------------------------------------------*/

   c = hb_list[ply - 1];
   c->c_lit->l_id = c->c_lit->l_id ^ NEGMASK;
   c->c_X = -c->c_X;
   c->c_Y = -c->c_Y;
   c->c_lit->l_X = -c->c_lit->l_X;
   c->c_lit->l_Y = -c->c_lit->l_Y;
   if(orig_branch_sign[ply - 1] == 0) {
      if(c->c_lit->l_id & NEGMASK) orig_branch_sign[ply - 1] = -1;
      else orig_branch_sign[ply - 1] = 1;}

/*--------------------------------------------------------*/
/* resolve hb element with set of base cls, add to clist  */
/*--------------------------------------------------------*/   

   if(x_debug > 1) {
      print0("\nBranch on atom: ");
      print_clause(c); 
      printf(" to node %ld ",s_nod);
      OutPtr += sprintf(OutPtr," to node %ld ",s_nod);
      gcxx = getchar();
      if(gcxx == '1') x_debug = 1;
      else if(gcxx == '2') x_debug = 2;
      else if(gcxx != '\n') {while((gcxx = getchar()) != '\n') {};}
      print0("\n");}

/*----------------------------------------------------------*/
/* determine where to store clauses generated at this node  */
/*----------------------------------------------------------*/

   nextc  = depth[ply - 1].d_ec;
   c = clist[nextc - 1];
   clist[nextc] = c->c_end; 
   d->d_tc = d->d_bc = nextc;

/*----------------------------------------------------------*/
/* generate clauses at this node.  place them on clist.     */
/* if too many clauses, return a 4.                         */
/*----------------------------------------------------------*/

   if(x_debug > 0) { 
      printf("\nGEN CLAUSES AT NODE %d: ",s_nod);
      OutPtr += sprintf(OutPtr,"\nGEN CLAUSES AT NODE %d: ",s_nod);}
   fail_count[ply - 1] = 0;
   number_of_clauses = calc_inferences(ply - 1);
   if(number_of_clauses < 0) {return(4);}
   if(number_of_clauses == 0) {
      if(x_debug > 0) printf("0 CLAUSES GENERATED!\n"); 
      d->d_tc = d->d_bc;}
   d->d_ec = depth[ply - 1].d_ec + number_of_clauses;
   if(nextc > CSPACE_CLAUSES - 3) {return(4);}

   if(x_debug > 0  && fail[ply + 1] < 2) {
      printf("GENERATED %d CLAUSES\n",number_of_clauses);
      OutPtr += sprintf(OutPtr,"%d CLAUSES GENERATED.\n",number_of_clauses);}

/*--------------------------------------------------------*/
/* prepare to advance deeper in the tree                  */
/*--------------------------------------------------------*/

   branch[ply]++;
   ply++;
   advance = 1;
   if(fail[ply] < 2) {branch[ply] = 0;}
   goto evaluate;}
}

/*------------------------------------------------------------*/
/* save_failure_pair()                                    #8
/*
/* keeps track of the two shallowest failure nodes and info
/* about them for use in constructing proof in add_clause()
/*
/* failure_clause1 and failure_clause2 record info on the
/* two clauses that most recently failed to close part of
/* the semantic tree.
/*
/* failure_clause3 and failure_clause4 keep track of info on
/* the two brother clauses that failed at the shallowed level
/* in the tree.  When the semantic tree is finally closed,
/*
/* it is clauses in failure_clause3 and failure_clause4 that
/* are used to generate the next clause in the resolution-
/* refutation proof.
/*-----------------------------------------------------------
/* called by: build_tree()  build.c
/*
/* calls:     nobody
/*
/* returns:   nil
/*-------------------------------------------------------------*/

void save_failure_pair(c)
clause *c;
{
int i, j;
literal *lit;

if(fail_clause2 != 0) return;

last_fail = c->c_id;

if(s_nod == fail_clause1_node + 1 
       && branch[ply - 1] == 2
       && s_nod > 1
       && fail_clause1_ply == ply - 1) {

/*---------------------------------------*/
/* save info on the right clause that    */
/* caused a node to fail                 */
/*---------------------------------------*/

// printf("Right failure at ply %d and node %d saved...\n",ply - 1,s_nod);
// fflush(stdout);
// for(i = 0; i < ply - 1; i++) {printf("%d",branch[i + 1]);}printf("\n");
   fail_clause2 = c->c_rpar->c_id;
   fail_clause2x = c->c_indices & RIGHT_INDEX_MASK;
   fail_clause2_ply = ply - 1;
   fail_clause2_node = s_nod;

   i = 0;
   fail_clause_left[ply - 1][0] = c->c_rpar->c_id;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      fail_clause_left[ply - 1][i++ + 1] = lit->l_X;}

   i = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      fail_lits2[i++] = lit->l_X;}
   fail_lits2[i] = 0;

   if(fail_clause2_ply < fail_clause3_ply || fail_clause3_ply == 0) {
      fail_clause3 = fail_clause1;
      fail_clause3x = fail_clause1x;
      fail_clause3_ply = fail_clause1_ply;
      fail_clause3_node = s_nod;

      for(i = 0; ;i++) {
         fail_lits3[i] = fail_lits1[i];
         if(fail_lits1[i] == 0) {break;}}

      fail_clause4 = fail_clause2;
      fail_clause4x = fail_clause2x;
      fail_clause4_ply = fail_clause3_ply;
      fail_clause4_node = s_nod;

      for(i = 0; ;i++) {
         fail_lits4[i] = fail_lits2[i];
         if(fail_lits2[i] == 0) {break;}}}}

/*---------------------------------------*/
/* save info about right parent failure  */
/*---------------------------------------*/

else {
   if(branch[ply - 1] == 1) {
//    printf("\nLeft failure at ply %d and node %d saved...\n",ply-1,s_nod);
//    for(i = 0; i < ply - 1; i++) {printf("%d",branch[i + 1]);}printf("\n");
      fail_clause1 = c->c_rpar->c_id;
      fail_clause1x = c->c_indices & RIGHT_INDEX_MASK;
      fail_clause1_ply = ply - 1;
      fail_clause1_node = s_nod;
      i = 0;
      fail_clause_left[ply - 2][0] = c->c_rpar->c_id;
      for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
         fail_clause_left[ply - 2][i + 1] = lit->l_X;
         fail_lits1[i++] = lit->l_X;}
      fail_lits1[i] = 0;
/*    printf("FAILS: ");
      print_clause(c);
      printf("\n");*/
        }}

return;
}

/*----------------------------------------------------------*/
/* simple_subsume()          .................... #9
/*
/* test to see if clause c1 subsumes clause c2
/*
/* called by:   install_clause()                  build.c
/*              calc_inferences()                 build.c
/*
/* calls:       nobody
/*
/* returns:     1 if c1 subsumes c2, otherwise returns 0
/*-----------------------------------------------------------*/

int simple_subsume(c1,c2)
clause *c1,*c2;
{
literal *l1, *l2;

if(c1->c_nlit > c2->c_nlit) return(0);

for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   if(l1->l_id & ELIMLIT) {continue;}
   for(l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
      if(l2->l_id & ELIMLIT) {continue;}
      if(l1->l_X == l2->l_X && l1->l_Y == l2->l_Y) {break;}}
   if(l2 == 0) {return(0);}}
return(1);
}
/*-----------------------------------------------------------------*/
/*      init_random_hash_array()    ........#9
/*
/*      initialize the rand_array to random numbers 
/*
/*      called from:   main()                      theo.c 
/*
/*
/*      returns:       nil
/*-----------------------------------------------------------------*/
 
void init_random_hash_array()
{
rnode *r1;
int i;
unsigned int num_ent;
	
/*--------------------------------------------------------------*/
/* the number of rows in rand_array = MAX_VARS + numb_fxns      */
/* the number of columns in rand_array = numb_chars             */
/*--------------------------------------------------------------*/
 
numb_rows = MAX_VARS + number_fxns + number_preds;
numb_cols = 200;
rows_less_nfxns = numb_rows - number_fxns - 1;
 
/*--------------------------------------------------------------*/
/* the number of entries in the array rand_array is the         */
/* product of the numb_rows * numb_cols                         */ 
/*--------------------------------------------------------------*/
 
numb_entries = numb_rows * numb_cols;
num_ent = (unsigned int) (numb_entries * 2 + 1); 
if(num_ent < 50) {
   numb_entries = 25;
   num_ent = 51;}

/*--------------------------------------------------------------*/
/* allocate space for rand_array                                */
/*--------------------------------------------------------------*/

rand_array = (rnode *) calloc(num_ent * 2,4);

/*--------------------------------------------------------------*/
/* make 32-bit random number for each element in rand_array     */
/*--------------------------------------------------------------*/
 
srand(2);
 
for(i = 0 ; i < (int) numb_entries; i++) {
   r1 = &rand_array[i];
   r1->r_hash1 = rand()*16384 + rand();
   r1->r_hash2 = rand()*16384 + rand();}
}
