/*-----------------------------------------------------------------*/
/* hdump.c                     HERBY......clause printing
/*
/* Copyright Newborn Software 1995 - 1999, 2000
/*
/* functions:
/* the following two functions actually print the clause
/*
/* print_arguments()       .......................#1
/* print_clause()          .......................#2
/* print_clause2()         .......................#2a
/* print0()                .......................#3
/*-----------------------------------------------------------------*/
 
#include <stdio.h>
#include "constHS.h"
#include "clauseHS.h"

/*-----------------------------------------------------------------*/
/* externally defined constants                                    */
/*-----------------------------------------------------------------*/
 
extern char *pidtbl[], *fidtbl[];
 
extern char *OutPtr;
 
/*-----------------------------------------------------------------*/
/* internally defined functions                                    */
/*-----------------------------------------------------------------*/

int *print_arguments();
void print_clause();
void print_clause2();
void print0();

/*-----------------------------------------------------------------*/
/* print_arguments(p)          .......................#1
/*
/* print the arguments of the predicate
/*
/* called by:    print_clause()           hdump.c
/*
/* calls:        nobody
/*
/* returns:      pointer to next argument
/*-----------------------------------------------------------------*/
 
int *print_arguments(p)
int *p;
{
int a, p_variable;
static char var_[] =
   "xyzuvwtsrqpklmnNMLKTSRQPWVUZYXxyzuvwtsrqpklmnNMLKPQRSTWVUZYX";
 
if((*p & FCNMASK) == 0) {
	
/*-----------------------------------------------------------------*/
/* print a variable                                                */
/*-----------------------------------------------------------------*/
	
   p_variable = *p;
   if(p_variable >= MAX_VARS) {
      p_variable = p_variable - MAX_VARS;}     /* standardize */
// printf("%c",var_[p_variable]); 
   OutPtr += sprintf(OutPtr,"%c",var_[p_variable]); 
   return(p + 1);}
    
/*-----------------------------------------------------------------*/
/* print a function                                                */
/*-----------------------------------------------------------------*/
 
//printf("%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);
OutPtr += sprintf(OutPtr,"%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);
	
/*-----------------------------------------------------------------*/
/* recursively print the remaining arguments of p                  */
/*-----------------------------------------------------------------*/
 
for(a = *p++ & ARGMASK; a > 0; a--)  p = print_arguments(p);
return(p);
}
 
/*-----------------------------------------------------------------*/
/* print_clause(c)               .......................#2
/*
/* called by:      many functions
/*
/* calls:          print_arguments()      hdump.c
/*
/* returns:        nil
/*-----------------------------------------------------------------*/
 
void print_clause(c)
clause *c;
{
int i, *a;
literal *l;
static char AB[] = ".abcdefghijklmnopqrstuvwxyz";
 
if((c->c_bits & DEAD) == 0) {
// printf("%3d",c->c_id);
   OutPtr += sprintf(OutPtr,"%3d",c->c_id);
   if(c->c_bits & LIST) {
//    printf(">");
      OutPtr += sprintf(OutPtr,">");}
//  else printf(":");
//  printf(" ");   
   }
else {
// printf("%3d# ",c->c_id);
   OutPtr += sprintf(OutPtr,"%3d# ",c->c_id);}

if(c->c_lpar != 0 && c->c_indices != 0) {
   if(c->c_rpar == 0) {
 
/*-----------------------------------------------------------------*/
/* clause is a binary factor; print merged indices                 */
/*-----------------------------------------------------------------*/ 
 
/*    printf("(%d%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
*/
      OutPtr += sprintf(OutPtr,"(%d%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
      if(c->c_lpar->c_id<10) {
         print0(" ");}}
   else {
 
/*-----------------------------------------------------------------*/
/* clause is a binary resolvent;     print parents and indices     */
/*-----------------------------------------------------------------*/ 
	
/*
       printf("(%d%c,%d%c",c->c_lpar->c_id, 
	        AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	        AB[c->c_indices & RIGHT_INDEX_MASK]);       
*/
       OutPtr += sprintf(OutPtr,"(%d%c,%d%c",c->c_lpar->c_id, 
	        AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	        AB[c->c_indices & RIGHT_INDEX_MASK]);
 
       print0(") ");
	 		
       if(c->c_lpar->c_id < 10) print0(" ");
       if(c->c_rpar->c_id < 10) print0(" ");}}
 
/*-----------------------------------------------------------------*/
/* if null clause, print []                                        */
/*-----------------------------------------------------------------*/
 
if(c->c_lit == 0) {print0("[]\n");}
		 
/*-----------------------------------------------------------------*/
/* else print each literal of the clause                           */
/*-----------------------------------------------------------------*/
 
else {
   for(l = c->c_lit; l != 0; l = l->l_next) {

/*-----------------------------------------------------------------*/
/* print negation symbol if negative literal                       */
/*-----------------------------------------------------------------*/
	
      if(l->l_id & NEGMASK)  {print0("~");}

/*-----------------------------------------------------------------*/
/* print predicate name                                            */
/*-----------------------------------------------------------------*/
				 
//    printf("%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]); 
      OutPtr += sprintf(OutPtr,"%s",pidtbl[(l->l_id&IDMASK) >> ARGSHIFT]);

/*-----------------------------------------------------------------*/
/* now print the arguments                                         */
/*-----------------------------------------------------------------*/
 
      a = &l->l_id;
      a++;
      for(i = l->l_id & ARGMASK; i > 0; i--) {a = print_arguments(a);}
 
/*-----------------------------------------------------------------*/
/* print a # if literal was eliminated and atom that eliminated lit*/
/*-----------------------------------------------------------------*/
 
      if(l->l_id & ELIMLIT) {
         print0("#");
//       printf("%d ",l->l_X);
         OutPtr += sprintf(OutPtr,"%d ",l->l_X);}

/*-----------------------------------------------------------------*/
/* print a space after each literal                                */
/*-----------------------------------------------------------------*/

      if(l->l_next) {print0(" ");}}}
 
return;
}

/*---------------------------------------------------------------*/
/* print_clause2()
/*---------------------------------------------------------------*/

void print_clause2(c)
clause *c;
{
int *cq;

cq = &c->c_id;

printf("\n%8x ID%d ",cq,*cq++);
printf("\n%8x END %x ",cq,*cq++);
printf("\n%8x BITS %x ",cq,*cq++);
printf("\n%8x HASH %x %x ",cq,*cq++,*cq++);
printf("\n%8x L %x R %x ",cq,*cq++,*cq++);
printf("%8x IX %x ",cq,*cq++);
printf("\n%8x CLIT %x ",cq,*cq++);
printf("%8x NLIT %d ",cq,*cq++);
printf("%8x CCGE %x ",cq,*cq++);
printf("%8x Vars %d ",cq,*cq++);
printf("..\n");
while((int) cq < (int) c->c_end) { 
   printf("\n%8x %8x ",cq,*cq);
   cq++;}
   printf("\n");
}


/*---------------------------------------------------------------*/
/*      print0()         .......................#3
/*
/*      prints a string to the screen AND to the disk
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
	
void print0(msg)
char *msg;
{
OutPtr += sprintf(OutPtr,"%s",msg);	
}
