/*------------------------------------------------------------------*/
/* hgen.c                 HERBY......clause generation
/*
/* Copyright Newborn Software 1995 - 1999, 2000
/*
/* functions:
/*    generate_clause()       .......................#1
/*    generate_clause2() ............................#2
/*    emit_lit()           ..........................#3
/*    gen_lit()       ...............................#4
/*    copy_atom_from_clause()   .....................#5
/*    emit_ground_lit()    ..........................#6
/*    gen_ground_lit()     ..........................#7
/*    copy_gnd_atom_from_cls()  .....................#8
/*    copy_clause()..................................#9
/*    gen_unified_lit()  ...........................#10
/*    hbgen_clause()  ..............................#11
/*------------------------------------------------------------------*/
 
#include <stdio.h>
 
#include "constHS.h"
#include "clauseHS.h"

/*------------------------------------------------------------------*/
/* externally defined functions                                     */
/*------------------------------------------------------------------*/
 
extern int calc_lit_hc();       /* hhash.c..calc literal hash code  */
extern int is_comp_in_ht();
extern void calc_clause_hashcode(); /* hhash.c.calc cls hash code   */
extern void clear_var_substitution_tble();  /* infer.c              */
extern void print_clause();            /* hdump                     */
extern void print_clause2();           /* hdump                     */

/*------------------------------------------------------------------*/
/* externally defined constants                                     */
/*------------------------------------------------------------------*/

extern int  *var_substitution_tbl[];/* herby.c..var substitut'n tbl */
extern clause *clist[];         /*  build.c...pointers to clauses   */
extern clause *hb_list[];
extern clause *unit_list[];
 
extern int nextc, nexthb, nextunit, base, ply;
extern int o_charmax, o_litmax, o_resref;
extern int x_debug;
extern int max_var;
extern int phase;
extern int long_lit;
extern int s_nod;

/*------------------------------------------------------------------*/
/* internally defined functions                                     */
/*------------------------------------------------------------------*/
 
clause *generate_clause();
clause *generate_clause2();
int *emit_lit();
int gen_lit();
clause *copy_atom_from_clause();
int *emit_ground_lit();
int gen_ground_lit();
clause *copy_gnd_atom_from_cls();
clause *copy_clause();
clause *gen_unified_lit();
clause *hbgen_clause();

/*------------------------------------------------------------------*/
/* internally defined constants                                     */
/*------------------------------------------------------------------*/
 
int *eptr;                            /* where to emit clause code  */
int testt;

int new_lit_chars;

/*------------------------------------------------------------------*/
/* generate_clause(c1,c2,lit1,lit2,ix1,ix2)    ........#1
/*
/* will generate a clause from one or two parent cls
/*
/* parameters:     c1,c2: left and right parents
/*                 lit1,lit2: literals to leave out
/*                 ix1,ix2 are the literal indices
/*
/* A clause will not be gen. for various reasons (too many lits,
/* or the clause is a tautology)
/*
/* called by:    binary_resolvents()       infer.c
/*               binary_factors()          infer.c
/*               add_clause()              herby.c
/*
/* calls:        gen_lit()                 hgen.c
/*               calc_clause_hashcode()    hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *generate_clause(c1,c2,lit1,lit2,ix1,ix2)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2;
{
clause *c, *k;                                /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2, *lit;
literal *pl;                             /* current predicate group */
int oldnv;
int iz2;
int lit_found, kk;
	
c = clist[nextc];                      /* where to put the clauses  */

if(ix2 < 0) {iz2 = -ix2;}
else {iz2 = ix2;}
 
c->c_bits = 0;

/*-----------------------------------------------*/
/* determine left and right parents of clause    */
/*-----------------------------------------------*/

c->c_lpar = c1;
 
k = c2;
if(k != 0 && ix2 > 0 && phase != 0) {
   while(k->c_id > base + 1) {k = k->c_rpar;}}
c->c_rpar = k;
   
c->c_nv   = 0;                                /* filled in later   */
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0;                                /* literal chain     */
 
prevl = 0;                               /* prev literal in scan   */
eptr = (int  *) (c + 1);                /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */
l2 = 0;                         /* assume non-null right parent    */
if(c2 != 0) l2 = c2->c_lit;

while(l1 != 0 || l2 != 0) {          /* while there are literals  */
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/

   curl = (literal *) eptr;           /* literal that we're making   */
   oldnv = c->c_nv;
 
   if(l1 != 0 && l1 == lit1) {l1 = l1->l_next; continue;}

   if(phase == 0 || o_resref == 0) {
      if(l2 != 0 && l2 == lit2) {l2 = l2->l_next; continue;}}

   if(ix2 < 0) {
      if(l2 != 0 && l2 == lit2) {l2 = l2->l_next; continue;}}

/*------------------------------------------------*/
/* generate one more literal from input clause c1 */
/*------------------------------------------------*/
 
   if(l2 == 0 || l1 != 0) {
      if(gen_lit(c,l1)) {return(0);}
      if((l1->l_id & ELIMLIT) != 0) {
         curl->l_id = l1->l_id;
         curl->l_X = l1->l_X;}
      l1 = l1->l_next;}
   else {

/*-------------------------------------------------*/
/* generate one more literal from input clause c2  */
/*-------------------------------------------------*/

      if(gen_lit(c,l2)) {return(0);}
      if(ix2 > 0) {
         if((l2->l_id & ELIMLIT) != 0) {
            curl->l_id = l2->l_id;
            curl->l_X = l2->l_X;}
         else {
            if(l2 != 0 && l2 == lit2) { 
               curl->l_id = curl->l_id | ELIMLIT;
               curl->l_X = ply;}}}      
      l2 = l2->l_next;}

/*----------------------------------------------------------*/
/* we scan back to see if there are any matching hashcodes. */
/* If so we check for a tautology and if so, return a 0.    */
/*----------------------------------------------------------*/
	 	
   for(pl = c->c_lit; pl != 0; pl = pl->l_next) {
      if(pl->l_X == -curl->l_X && pl->l_Y == -curl->l_Y) { 
         if((curl->l_id ^ pl->l_id) != 0) {
	    return(0);}}

      if(pl->l_X == curl->l_X && pl->l_Y == curl->l_Y && phase != 2) {
         c->c_nv   = oldnv;
         eptr = (int *) curl;
         goto idn;}}

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if((curl->l_id & ELIMLIT) == 0) {c->c_nlit++;}
   if(prevl == 0)  c->c_lit = curl;
   else  prevl->l_next = curl;
   prevl = curl;
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/
 		
   if(c->c_nv > 4*max_var && ix2 > 0) {return(0);}
   if(c->c_nlit > o_litmax && phase < 2) {return(0);}
 
idn:
;}
if(c->c_nlit == 1) c->c_bits |= UNIT;             /* a unit clause  */
c->c_end = (clause *) eptr;                  /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + iz2; /* parent lit indices */
c->c_id = nextc + 1;                       /* give clause an ID    */

/*---------------------------------------------------------------*/
/* see if a literal can be resolved away by a hashtable hit..    */
/*---------------------------------------------------------------*/

if(lit_found = is_comp_in_ht(c)) {

   kk = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      kk++;
      if(kk == lit_found) {
         lit->l_X = 400;
	 lit->l_X |= ELIMLIT;}}}

calc_clause_hashcode(c);

return(c);
}

/*------------------------------------------------------------------*/ 
/* generate_clause2()    .....................#2
/*
/* similar to generate_clause in gen.c, but used by
/* the ashs.
/*
/* parameters:     c1,c2: left and right parents
/*                 lit1,lit2: literals to leave out
/*
/* A clause will not be gen. for various reasons (too many lits,
/* or the clause is a tautology)
/*
/* called from:  binary_resolvents2()      atom.c
/*
/* calls:        gen_lit()                 hgen.c
/*               calc_clause_hashcode()    hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *generate_clause2(c1,c2,lit1,lit2,ix1,ix2,add_lit)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2, add_lit;
{
clause *c, *k;                                /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2;
literal *pl, *cg;                        /* current predicate group */
int i;
int oldnv;
int iz2;

c = clist[nextc];                      /* where to put the clauses  */
 
iz2 = ix2;
 
c->c_bits = 0;
c->c_lpar = c1;                         /* parents of this clause   */
 
k = c2;
if(k != 0 && ix2 >0) {
   while(k->c_id > base + 1) {k = k->c_rpar;}}
c->c_rpar = k;
   
c->c_nv   = 0;                                /* filled in later   */
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0;                                /* literal chain     */
 
i = -1;                                    /* when to change cg    */
cg = 0;                                    /* xsum comparisons     */
prevl = 0;                               /* prev literal in scan   */
eptr = (int *) (c + 1);                 /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */
l2 = c2->c_lit;
 
while(l1 != 0 || l2 != 0) {          /* while there are literals  */

/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
   curl = (literal *) eptr;          /* literal that we're making   */
   oldnv = c->c_nv;
 
   if((l1 != 0 && l1 == lit1) || 
       ((l1 != 0) && (l1->l_id & ELIMLIT) != 0)) {
      l1 = l1->l_next; continue;}            

   if(add_lit == 0) {
      if((l2 != 0 && l2 == lit2) || 
          ((l2 != 0) && (l2->l_id & ELIMLIT) != 0)) {
         l2 = l2->l_next; 
         continue;}}           
   else {
      if((l2 != 0) && (l2->l_id & ELIMLIT) != 0) {
         l2 = l2->l_next; 
         continue;}}           

   if(l1 != 0) {
      if(gen_lit(c,l1)) {return(0);}
      l1 = l1->l_next;}
   else {
      if(gen_lit(c,l2)) {return(0);}
      l2 = l2->l_next;}
                
/*----------------------------------------------------------*/
/* we scan back in the current group to see if there        */
/* are any matching literal checksums. If so we perform     */
/* a literal comparison and a) remove the literal if        */
/* it is identical to another; or b) cancel the gen'tion    */
/* of the clause if the literals are identical but of       */
/* opposite signs (tautology).                              */
/*----------------------------------------------------------*/
	
   if((curl->l_id & IAMASK) == i) {
      for(pl = cg; pl != 0; pl = pl->l_next) {
         if((pl->l_X == -curl->l_X && pl->l_Y ==-curl->l_Y)) { 
            if((curl->l_id ^ pl->l_id) != 0) {return(0);}
            c->c_nv   = oldnv;
            eptr = (int *) curl;
            goto idn;}}}
   else {
 
/*----------------------------------------------------------*/
/* if we change IDs (i.e. change groups) we set up another  */
/* cg for later xsum scanning                               */
/*----------------------------------------------------------*/
 
      i = curl->l_id & IAMASK;
      cg = curl;}
 
/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if((curl->l_id & ELIMLIT) == 0) {c->c_nlit++;}
   if(prevl == 0)  {c->c_lit = curl;}
   else {prevl->l_next = curl;}
   prevl = curl;
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/
 		
   if(c->c_nv > 2 * max_var  && ix2 > 0) {return(0);}
 
idn:    ;
	}
if(c->c_nlit == 1) c->c_bits |= UNIT;             /* a unit clause  */
c->c_end = (clause *) eptr;                  /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + iz2; /* parent lit indices */
c->c_id = nextc + 1;                        /* give clause an ID    */
calc_clause_hashcode(c);

return(c);
}
 
/*--------------------------------------------------------------*/ 
/* emit_lit(c,l,i)           ......................#3
/*
/* emit_lit is a recursive fxn that will apply the subs
/* of the var_substitution_tbl to a given agmnt and write the 
/* expanded version.
/* NOTE: we use the var_substitution_tbl as a renumber array 
/* (ie. holds the new value of variables) and so no real pointer
/* can be below MAXVARS*2, and NULL must be 0.
/*
/* parameters:     c : ^ clause header
/*                 l : ^ arguments
/*                 i : number of arguments
/*                 d : current depth (pass 0 to start).
/*
/* Note: we write out the expanded version at the current
/* "eptr" location.
/*
/* called from:    gen_lit()               hgen.c
/*
/* calls:          emit_lit()              hgen.c
/*
/* returns:        pointer past arguments
/*                 NULL if fcn nesting too deep.
/*--------------------------------------------------------------*/
 
int  *emit_lit(c,l,ii)
clause *c;
int *l,ii;
{
int sl, i;

/*---------------------------------------------------------------*/
/* for each argument of the predicate (and functions)            */
/*---------------------------------------------------------------*/

for(i = ii; i > 0; i--) {
   sl = *l;

/*---------------------------------------------------------------*/
/* if argument is a function, do the following                   */
/*---------------------------------------------------------------*/

   if((sl & FCNMASK) != 0) {           /* FCN:just copy it over */
      *eptr++ = sl;
      l++;
      if((sl & ARGMASK) != 0) {
         if((l = emit_lit(c,l,sl&ARGMASK)) == 0) {return(0);}}}

/*---------------------------------------------------------------*/
/* otherwise, if argument is a variable, do the following        */
/*---------------------------------------------------------------*/

   else { 
      if((unsigned int) var_substitution_tbl[sl] < MAX_VARS_X2) {
         if(var_substitution_tbl[sl] == 0) {
            var_substitution_tbl[sl] = (int *) ++c->c_nv;}

/*---------------------------------------------------------------*/
/* this is the simple case, where a new variable has been        */
/* found, or one previously found has had another variable       */
/* substituted for it previously.                                */
/*---------------------------------------------------------------*/
 
         *eptr++ = (int) var_substitution_tbl[sl] - 1;}
      else if(emit_lit(c,var_substitution_tbl[sl],1) == 0) {return(0);}
      l++;}}

return(l);
}
 
/*----------------------------------------------------------------*/
/* gen_lit(c,lit)     .....................#4
/*
/* called from:   generate_clause()          hgen.c
/*                generate_clause2()         hgen.c
/*                copy_atom_from_clause()    hgen.c
/*                copy_clause()              hgen.c
/*                gen_unified_lit()          hgen.c
/*                hbgen_clause()             hgen.c
/*
/* calls:         emit_lit()                 hgen.c
/*                calc_lit_hc()              hhash.c
/*
/* returns:       0 => literal is ok 
/*                1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
int gen_lit(c,lit)
clause *c;
literal *lit;
{
literal *new_lit;
	
new_lit = (literal *) eptr;                /* new literal  */
new_lit->l_next = 0;
new_lit->l_id =  lit->l_id;
 
eptr = (int  *) (new_lit + 1);           /* args go here */
if(emit_lit(c,(int *) (lit+1),lit->l_id & ARGMASK) == 0) {return(1);}

/*------------------------------------------*/
/* determine a hash code for the literal    */
/* and receive the number of char in lit    */
/*------------------------------------------*/

new_lit_chars = calc_lit_hc(new_lit);
 
/*------------------------------------------*/
/* if too many chars in lit, kill clause    */
/*------------------------------------------*/
 
if(new_lit_chars > o_charmax) {return(1);}
return(0);
}
 
/*------------------------------------------------------------------*/ 
/* copy_atom_from_clause(c1,lit1)    ..................#5
/*
/* parameters:   c:  clause from which atom is copied
/*               lit: literal to copy
/*
/* called by:    ash6a(),ash6b(),ash6c()      atom.c
/*               install_ash_clause()         build.c
/*
/* calls:        gen_lit()                    hgen.c
/*               calc_clause_hashcode()       hhash.c
/*               clear_var_substitution_tble  infer.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *copy_atom_from_clause(c,lit)
clause *c;
literal *lit;
{
clause *new_clause;                           /* new clause         */

clear_var_substitution_tble(c->c_bits,c->c_nv);

new_clause = hb_list[nexthb];          /* where to put the clauses  */
 
new_clause->c_lpar = c;                 /* parents of this clause   */
new_clause->c_rpar = 0;
new_clause->c_nv   = 0;                        /* filled in later   */
new_clause->c_nlit = 1;                  /* count the # of literals */
 
eptr = (int *) (new_clause + 1);        /* lit's go after cl header */
 
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
new_clause->c_lit = (literal *) eptr;       /* literal being made   */
 
/*------------------------------------------------------*/
/* if l1 or l2 is a hashcode literal, generate          */
/* a hashcode literal                                   */
/*------------------------------------------------------*/
          
gen_lit(new_clause,lit);
              
new_clause->c_bits = UNIT;                        /* a unit clause  */
new_clause->c_end = (clause *) eptr;         /* pointer past clause */
new_clause->c_indices = 0;                    /* parent lit indices */
new_clause->c_id = nexthb + 1;              /* give clause an ID    */
 
calc_clause_hashcode(new_clause);
 
return(new_clause);
}
 
/*--------------------------------------------------------------*/ 
/* emit_ground_lit(l,i,gc)    .....................#6
/*
/* emit_ground_lit is a recursive fxn that will apply the subs
/* of the var_substitution_tbl to a given agmnt and write the 
/* expanded version.
/* NOTE: we use the var_substitution_tbl as a renumber array 
/* (ie. holds the new value of variables) and so no real pointer
/* can be below MAXVARS*2, and NULL must be 0.
/*
/* parameters:     
/*                 l : pointer to literal or term
/*                 i : number of arguments
/*                 gc: index of constant to substitute for vars
/*
/* Note: we write out the expanded version at the current
/* "eptr" location.
/*
/* called from:    gen_ground_lit()        hgen.c
/*
/* calls:          emit_ground_lit()       hgen.c
/*
/* returns:        pointer past arguments
/*                 NULL if fcn nesting too deep.
/*--------------------------------------------------------------*/
 
int  *emit_ground_lit(l,ii,gc)
int *l,ii,gc;
{
int i, sl;
 
for(i = ii; i > 0; i--) {                 /* for each argument  */
   sl = *l;
   if((sl & FCNMASK) != 0) {           /* FCN:just copy it over */
      *eptr++ = sl;
      l++;
      if((sl & ARGMASK) != 0) {l = emit_ground_lit(l,sl&ARGMASK,gc);}}
   else {
      *eptr++ = ((gc << ARGSHIFT) | FCNMASK);
      l++;}}
      
return(l);
}
 
/*----------------------------------------------------------------*/
/* gen_ground_lit(lit,gc)     ..................#7
/*
/* called from:   copy_gnd_atom_from_cls()    hgen.c
/*
/* calls:         emit_ground_lit()           hgen.c
/*                calc_lit_hc()               hhash.c
/*
/* returns:       0 => literal is ok 
/*                1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
int gen_ground_lit(lit,new_lit,gc)
literal *new_lit, *lit;
int gc;
{
int *p, *q, a;
	
p = &lit->l_id;
q = &new_lit->l_id;
*q++ = *p;

for(a = *p++ &ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) {
      *q = *p;q++;
      a += *p++ & ARGMASK;}
   else {
      *q = gc;
      *q = NEGMASK | (*q << 5);
      q++;
      p++;}}

/*------------------------------------------*/
/* determine a hash code for the literal    */
/* and receive the number of char in lit    */
/*------------------------------------------*/
 
new_lit_chars = calc_lit_hc(new_lit);
return((int) q);
}
 
/*------------------------------------------------------------------*/ 
/* copy_gnd_atom_from_cls(c1,lit1,gc)  ...............#8
/*
/* copy literal lit1 from clause c1 assigning gc to all
/* variables in the literal
/*
/* parameters:     c1 = clause to get atom from
/*                 lit = literal in clause to ground
/*                 gc  = index of constant to replace vars
/*
/* called from:  install_ash_clause()           atom.c
/*
/* calls:        clear_var_substitution_tble()  infer.c
/*               gen_ground_lit()               hgen.c
/*               calc_clause_hashcode()         hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *copy_gnd_atom_from_cls(c1,lit1,gc)
clause *c1;
literal *lit1;
int  gc;                                 /* ground constant   */
{
clause *c;                               /* new clause        */
literal *curl;                           /* current literal   */
int *eq;

clear_var_substitution_tble(c1->c_bits,c1->c_nv);
c = hb_list[nexthb];             /* where to put the clauses  */
c->c_lpar = c1;                   /* parents of this clause   */
c->c_rpar = 0;
c->c_nv   = 0;                           /* filled in later   */
c->c_nlit = 1;
 
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
eq = (int *) (c + 1);
c->c_lit = (literal *) eq;    /* literal that we're making   */
c->c_lit->l_id = lit1->l_id;
c->c_lit->l_next = 0;

c->c_end = (clause *) gen_ground_lit(lit1,c->c_lit,gc);
                 
c->c_bits = UNIT;                          /* a unit clause  */
c->c_indices = 0;                     /* parent lit indices */
c->c_id = nexthb + 1;               /* give clause an ID    */
calc_clause_hashcode(c);
 
return(c);
}
 
/*-------------------------------------------------------------*/ 
/* copy_clause(c1,liste)    ..............................#9
/*
/* c1 = clause to copy
/* liste = list on which to copy clause
/*
/* called by:    add_clause()                   herby.c
/*
/* calls:        clear_var_substitution_tble()  infer.c
/*               gen_lit()                      hgen.c
/*               calc_clause_hashcode()         hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*-------------------------------------------------------------*/
 
clause *copy_clause(c1,liste)
clause *c1;
int liste;
{
clause *c;                                  /* new clause       */
literal *curl, *prevl;    /* current literal; previous literal  */
literal *l1;

clear_var_substitution_tble(c1->c_bits,c1->c_nv);

if(liste == 1) {c = clist[nextc];} /* where to put the clauses  */ 
else if(liste == 3) {c = hb_list[nexthb];}
else c = unit_list[nextunit];

c->c_bits = c1->c_bits & ~VID;
c->c_lpar = c1->c_lpar;             /* parents of this clause   */
c->c_rpar = c1->c_rpar;
c->c_nv   = 0;                            /* filled in later   */
c->c_change = c1->c_change;

prevl = 0; 
eptr= (int  *) (c + 1);             /* lit's go after cl header */
 
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {    

/*-----------------------*/
/* while there are lits  */	
/*-----------------------*/
   
   curl = (literal *) eptr;     /* literal that we're making   */
 
/*------------------------------------------------------*/
/* if l1 or l2 is a hashcode literal, generate          */
/* a hashcode literal                                   */
/*------------------------------------------------------*/
            
   gen_lit(c,l1);

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if(prevl == 0)  {c->c_lit = curl;}
   else  {prevl->l_next = curl;}        
   prevl = curl;}
 
c->c_nlit = c1->c_nlit;
c->c_end = (clause *) eptr;             /* pointer past clause */
c->c_indices = 0;                       /* parent lit indices */
c->c_id = c1->c_id ;

calc_clause_hashcode(c);

return(c);
}

/*------------------------------------------------------------*/
/* gen_unified_lit()        .........................#10
/*
/* forms the unit clause that is the common instance of
/* c1 and c2.  If htere is an ungrounded variable, it is
/* grounded.
/*
/* called by:        install_unit_and_test()     build.c
/*                   binary_resolvents3()        infer.c
/* calls:
/* returns:  a pointer to the clause
/*------------------------------------------------------------*/

clause *gen_unified_lit(c1)
clause *c1;
{
clause *c;                                    /* new clause         */
literal *l1;
	
c = clist[nextc];                      /* where to put the clauses  */

/*-----------------------------------------------*/
/* determine left and right parents of clause    */
/*-----------------------------------------------*/

c->c_lpar = 0;
c->c_rpar = 0;

c->c_nv   = 0;                                /* filled in later   */
c->c_lit  = 0;                                /* literal chain     */
 
eptr = (int  *) (c + 1);                /* lit's go after cl header */
 
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   if((l1->l_id & ELIMLIT) == 0) {break;}}

c->c_lit = (literal *) eptr;

/*------------------------------------------------*/
/* generate one more literal from input clause c1 */
/*------------------------------------------------*/

if(gen_lit(c,l1)) {
   o_charmax = new_lit_chars;
   printf("\nIncrease term size to %d\n",o_charmax);
   }

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
c->c_nlit = 1;
c->c_bits = UNIT;                                 /* a unit clause  */
c->c_end = (clause *) eptr;                  /* pointer past clause */
c->c_id = nextc + 1;                       /* give clause an ID    */
c->c_indices = 0;
calc_clause_hashcode(c);
return(c);
}
 
/*------------------------------------------------------------------*/
/* hbgen_clause(h,c)    ..........................#11
/*
/* will gen a cls on clist by resolving hb-atom h with 
/* literal lit of clause c.
/*
/* A clause will not be gen. for various reasons (too many lits,
/* or the clause is a tautology)
/*
/* called by:    binary_resolvents()       infer.c
/*
/* calls:        gen_lit()                 hgen.c
/*               calc_clause_hashcode()    hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *hbgen_clause(h,c,lit)
clause *h, *c;
literal *lit;
{
clause *c_new, *k;                            /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2;
literal *pl;                             /* current predicate group */
int iz2;
int oldnv;

c_new = clist[nextc];                  /* where to put the clauses  */
 
c_new->c_bits = 0;

/*-----------------------------------------------*/
/* determine left and right parents of clause    */
/*-----------------------------------------------*/

c_new->c_lpar = h;
 
k = c;

while(k->c_id > base + 1) {k = k->c_rpar;}

c_new->c_rpar = k;

c_new->c_nv   = 0;                            /* filled in later   */
c_new->c_nlit = 0;                       /* count the # of literals */
c_new->c_lit  = 0;                            /* literal chain     */
 
prevl = 0;                               /* prev literal in scan   */
eptr = (int  *) (c_new + 1);            /* lit's go after cl header */
 
l1 = h->c_lit;                                    /* ptr to hb-atom */

l2 = c->c_lit;
iz2 = 1;
while(l2 != lit) {l2 = l2->l_next;iz2++;}
l2 = c->c_lit;

while(l2 != 0) {                      /* while there are literals  */
 
   curl = (literal *) eptr;          /* literal that we're making   */
   oldnv = c_new->c_nv;
 
/*------------------------------------------------*/
/* generate one more literal from input clause c1 */
/*------------------------------------------------*/
 
   if(gen_lit(c_new,l2)) {return(0);}
   if((l2->l_id & ELIMLIT) != 0) {
      curl->l_id = l2->l_id;
      curl->l_X = l2->l_X;}
   else {
      if(l2 == lit) { 
         curl->l_id = curl->l_id | ELIMLIT;
         curl->l_X = ply;}}
   l2 = l2->l_next;

/*----------------------------------------------------------*/
/* we scan back to see if there are any matching hashcodes. */
/* If so we check for a tautology and if so, return a 0.    */
/*----------------------------------------------------------*/
	 	
   for(pl = c_new->c_lit; pl != 0; pl = pl->l_next) {
      if(pl->l_X == -curl->l_X && pl->l_Y == -curl->l_Y) { 
         if((curl->l_id ^ pl->l_id) != 0) {return(0);}}}

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if((curl->l_id & ELIMLIT) == 0) {c_new->c_nlit++;}
 
   if(prevl == 0)  c_new->c_lit = curl;
   else  prevl->l_next = curl;
   prevl = curl;
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/
 		
   if(c_new->c_nv > 4 * max_var) {return(0);}
 
idn:
;}

if(c_new->c_nlit == 1) c_new->c_bits |= UNIT;     /* a unit clause  */
c_new->c_end = (clause *) eptr;              /* pointer past clause */
c_new->c_indices = (1<<LEFT_INDEX_SHIFT) + iz2; /* parent lit ind */
c_new->c_id = nextc + 1;                   /* give clause an ID    */
calc_clause_hashcode(c_new);

return(c_new);
}
