/*------------------------------------------------------------------*/
/*      gen.c
/*      copyrighted by Monty Newborn 2005
/*
/*      functions:
/*         generate_clause()        0141......................#1
/*         generate_iclause()       1509......................#21
/*         emit_literal()           0297......................#2
/*         generate_literal()       0343......................#3
/*         emit_literal2()          1037......................#12
/*         generate_literal2()      0997......................#11
/*         repair_literal()         0399......................#4
/*         eval_fxns()              0548......................#7
/*         prepare_subarray()       0717......................#8
/*         add_string()             0824......................#10
/*         print_sub_array()        1067......................#13
/*         copy_clause()            0751......................#9
/*         copy_clause3()           1176......................#15
/*         copy_clause_all()        1317......................#18
/*         copy_clause_elim()       1382......................#19
/*         generate_literal3()      1452......................#20
/*         test_equality()          0443......................#5
/*         test_simple_equality()   0485......................#6
/*         test_const_equal()       1612......................#22
/*         hgenerate_clause()
/*         hgenerate_literal()
/*         hgen_normalize()
/*         generate_clause_p2()
/*-----------------------------------------------------------------*/
 
#include <stdio.h>

#include "const.h"
#include "clauseS.h"

/*------------------------------------------------------------------*/
/* externally defined functions                                     */
/*------------------------------------------------------------------*/
 
extern int calc_lit_hc();       /* hash.c..calc literal hash code   */
extern void calc_clause_hashcode(); /* hash.c..calc cls hash code   */
extern void decide_to_add1();
extern void print_clause();
extern void print_clause2();
extern int add_clause_to_ht();
 
/*------------------------------------------------------------------*/
/* externally defined constants                                     */
/*------------------------------------------------------------------*/

extern int *var_substitution_tbl[]; /* theo.c..var substitution tbl */
extern clause *clist[];           /* theo.c...pointers to clauses   */
extern clause *ulist[];
extern char *fidtbl[];

extern int ply, nextc, nextu, base;
extern int s_nod, s_nod0;
extern int o_maxlit;        /* max # lits per clause (0 = no limit) */
extern int o_maxterm;
extern int in_elim;
extern int o_maxvar;
extern int equal_pred;
extern int phase;
extern int s_subfxn;
extern int o_step, o_print;
extern int this_computer;
extern int reinstate[1000], reinstate_parent[1000];
extern int o_debug;

/*------------------------------------------------------------------*/
/* internally defined functions                                     */
/*------------------------------------------------------------------*/

clause *generate_clause();
clause *generate_iclause();
int *emit_literal();
int generate_literal();
int *emit_literal2();
void generate_literal2();
void repair_literal();
int eval_fxns();
void prepare_subarray();
void add_string();
void print_sub_array();
clause *copy_clause();
void copy_clause3();
void copy_clause_all();
int copy_clause_elim();
void generate_literal3();
int test_equality();
int test_simple_equality();
int test_const_equal();
int replace_constant();
clause *hgenerate_clause();
int hgenerate_literal();
void hgen_normalize();
clause *generate_clause_p2();
 
/*------------------------------------------------------------------*/
/* internally defined constants                                     */
/*------------------------------------------------------------------*/
 
static int *ep, *eq;                  /* where to emit clause code  */
int constants_in_clause;
int var_in_lit;
int sub_array[65536][6];             /* function substitution array */
int sub_array_last;
int const_found_equal;
int replace_array[256];
int *replace_ptr;
int vars_sub[64];
int vars_sub_ptr;

int low_vars;
int low_array[64];
int high_vars;
int high_array[64];

/*------------------------------------------------------------------*/ 
/*      *generate_clause(c1,c2,lit1,lit2,ix1,ix2)    ........#1
/*
/*      generate a clause from one or two parent cls
/*      Apply all changes currently in the 
/*      var_substitution_tbl while we gen the new lits. 
/*      The variables are st'dized apart. Note that we destroy
/*      the var_substitution_tbl when we generate the clause
/*
/*      parameters:     c1,c2: left and right parents
/*                      lit1,lit2: literals to leave out
/*                      ix1, ix2 are indices of resolved literals
/*
/*      A clause will not be gen. for various reasons (too many lits,
/*      or the clause is a tautology)
/*
/*      called from:  binary_resolutions()      infer.c
/*                    binary_factors()          infer.c
/*                    bres_ply_1()              elim.c
/*                    bres_link()               elim.c
/*                    bres_replace()            elim.c
/*                    bres_proof_line()         pass2.c
/*                    factor_proof_line()       pass2.c
/*
/*      calls:        generate_literal()        gen.c
/*                    calc_clause_hashcode()    hash.c
/*                    test_equality()           gen.c
/*
/*      returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *generate_clause(c1,c2,lit1,lit2,ix1,ix2)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2;
{
clause *c;                                    /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2 , *pl;
int oldnv;
int lnumber;

c = clist[nextc];                      /* where to put the clauses  */

/*-----------------------------*/
/* initialize clause header    */
/*-----------------------------*/              
 
c->c_bits = ACTIVE;                       /* assume active clause   */
c->c_lpar = c1;                         /* parents of this clause   */
c->c_rpar = c2;
c->c_nv   = 0;                                 /* filled in later   */
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0;
c->c_change = 0;

/*--------------------------*/
/* prepare to scan literals */
/*--------------------------*/
 
constants_in_clause = 0;
prevl = 0;                                /* prev literal in scan   */
ep = (int *) (c + 1);                   /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */

if(c2) {l2 = c2->c_lit;}
else {l2 = 0;}
while(l1 || l2) {                      /* while there are literals  */ 
 
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
   if(l1 && ((l1 == lit1) || (l1->l_id & ELIMLIT))) { 
      l1 = l1->l_next; continue;}            
   if(l2 && (l2 == lit2 || (l2->l_id & ELIMLIT))) {
      l2 = l2->l_next; continue;}
 
/*----------------------------------------------------------*/
/* save this information in case literal is deleted---      */
/* in the case that it is identical to another and must     */
/* be deleted                                               */
/*----------------------------------------------------------*/
 
   curl = (literal *) ep;    /* literal that we're making   */
   oldnv = c->c_nv;
	    
   if(!l2 || l1) {
      if(generate_literal(c,l1)) {return(0);}
      l1 = l1->l_next;}
   else {
      if(generate_literal(c,l2)) {return(0);}
      l2 = l2->l_next;}
	       
/*----------------------------------------------------------*/
/* eliminate predicates with Equal(x,x) as tautologies      */
/*----------------------------------------------------------*/
 
   if((curl->l_id & 0x1000FFFF) == equal_pred && phase) { 
      if(test_equality(curl) > 0) {
         if(phase < 3) {
	    return(0);}}}    /* equal tautology */
 
/*----------------------------------------------------------*/
/* we scan back in the current group to see if there        */
/* are any matching literal checksums. If so we perform     */
/* a literal comparison and a) remove the literal if        */
/* it is identical to another; or b) cancel generation      */
/* of the clause if the literals are identical but of       */
/* opposite signs (tautology).                              */
/*----------------------------------------------------------*/
		
tautology_check:
   for(pl = c->c_lit; pl != 0; pl = pl->l_next) {
      if(pl->l_X == -curl->l_X && pl->l_Y == -curl->l_Y) {

/*-----------------------------------------------------------*/
/* clause is a tautology                                     */
/*-----------------------------------------------------------*/

            if(in_elim) {
	       return(0);}}                     /* tautology */

       if(pl->l_X == curl->l_X && pl->l_Y == curl->l_Y){ 
	       
/*-----------------------------------------------------------*/
/* merge clause, delete literal (by not adding it)           */
/*-----------------------------------------------------------*/
 
          c->c_nv    = oldnv;
          c->c_bits |= MERGE;                 /* remove lit  */
          ep = (int *) curl;
          goto idn;}}
 
/*-----------------------------------------------------------*/
/* the literal has survived, we patch the previous lit       */
/* so that it points here                                    */
/*-----------------------------------------------------------*/
 
   if(!prevl) {c->c_lit = curl;}
   else {prevl->l_next = curl;}
	
   prevl = curl;
   c->c_nlit++; 
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/

   if(c->c_nlit > o_maxlit) {
      if(o_maxlit)  {return(0);}}
   if(c->c_nv > o_maxvar) {return(0);}

idn: o_maxlit = o_maxlit;}
if(c->c_nlit == 1) {c->c_bits |= UNIT;}           /* a unit clause  */
c->c_end = (clause *) ep;                    /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + ix2; /* parent lit indices */
c->c_id = base + 1 + ply;                   /* give clause an ID    */

calc_clause_hashcode(c);
return(c);
}

/*------------------------------------------------------------------*/ 
/*      *generate_iclause(c1,c2,lit1,lit2,ix1,ix2)    ........#2
/*
/*      generate a clause from one or two parent cls
/*      Apply all changes currently in the 
/*      var_substitution_tbl while we gen the new lits. 
/*      The variables are st'dized apart. Note that we destroy
/*      the var_substitution_tbl when we generate the clause
/*
/*      parameters:     c1,c2: left and right parents
/*                      lit1,lit2: literals to leave out
/*                      ix1, ix2 are indices of resolved literals
/*
/*      A clause will not be gen. for various reasons (too many lits,
/*      or the clause is a tautology)
/*
/*      called from:  binary_resolutions()      infer.c
/*                    binary_factors()          infer.c
/*                    bres_ply_1()              elim.c
/*                    bres_link()               elim.c
/*                    bres_replace()            elim.c
/*                    bres_proof_line()         pass2.c
/*                    factor_proof_line()       pass2.c
/*
/*      calls:        generate_literal()        gen.c
/*                    calc_clause_hashcode()    hash.c
/*
/*      returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *generate_iclause(c1,c2,lit1,lit2,ix1,ix2)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2;
{
clause *c;                                    /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2 , *pl;
int oldnv;
int lnumber;

c = ulist[nextu];                      /* where to put the clauses  */

/*-----------------------------*/
/* initialize clause header    */
/*-----------------------------*/              
 
c->c_bits = ACTIVE| BASE;                /* assume active clause   */
c->c_lpar = c1;                        /* parents of this clause   */
c->c_rpar = c2;

c->c_nv   = 0;
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0;
c->c_change = 0;

/*--------------------------*/
/* prepare to scan literals */
/*--------------------------*/
 
constants_in_clause = 0;
prevl = 0;                                /* prev literal in scan   */
ep = (int *) (c + 1);                   /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */
l2 = 0;
if(c2 != 0) l2 = c2->c_lit;

while(l1 || l2) {                      /* while there are literals  */ 

   if(l1 && ((l1 == lit1) || (l1->l_id & ELIMLIT))) {
      l1 = l1->l_next; continue;}
   if(l2 && (l2 == lit2 || (l2->l_id & ELIMLIT))) {
      l2 = l2->l_next; continue;}

// if(l1 == lit1) {l1 = l1->l_next; continue;} 
// if(l2 == lit2) {l2 = l2->l_next; continue;}

   curl = (literal *) ep;    /* literal that we're making   */
   oldnv = c->c_nv;
	    
   if(!l2 || l1) {
      if(generate_literal(c,l1)) {return(0);}
      l1 = l1->l_next;}
   else {
      if(generate_literal(c,l2)) {return(0);}
      l2 = l2->l_next;}
		
tautology_check:
   for(pl = c->c_lit; pl != 0; pl = pl->l_next) {
       if(pl->l_X == curl->l_X && pl->l_Y == curl->l_Y) { 
	
/*-----------------------------------------------------------*/
/* merge clause, delete literal (by not adding it)           */
/*-----------------------------------------------------------*/
 
          c->c_nv    = oldnv;
          c->c_bits |= MERGE;                 /* remove lit  */
          ep = (int *) curl;
          goto idn;}}
 
/*-----------------------------------------------------------*/
/* the literal has survived, we patch the previous lit       */
/* so that it points here                                    */
/*-----------------------------------------------------------*/
 
   if(!prevl) {c->c_lit = curl;}
   else {prevl->l_next = curl;}
	
   prevl = curl;
   c->c_nlit++; 
idn: o_maxlit = o_maxlit;}

if(c->c_nlit == 1) {c->c_bits |= UNIT;}           /* a unit clause  */
c->c_end = (clause *) ep;                    /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + ix2; /* parent lit indices */
c->c_id = base + 1 + ply;                   /* give clause an ID    */

calc_clause_hashcode(c);
return(c);
}
 
/*----------------------------------------------------------------*/ 
/*      emit_literal(c,l,ii)           ......................#3
/*
/*      emit_literal is a recursive fxn that will apply the subs
/*      of the var_substitution_tbl to a given agmnt and write the 
/*      expanded version.
/*
/*      parameters:     c : ^ clause header
/*                      l : ^ arguments
/*                      i : number of arguments
/*
/*      Note: we write out the expanded version at the current
/*      "ep" location.
/*
/*      called from:    generate_literal()      gen.c
/*
/*      calls:          emit_literal()          gen.c
/*
/*      returns:        pointer past arguments
/*                      NULL if fcn nesting too deep.
/*--------------------------------------------------------------*/
 
int *emit_literal(c,l,ii)
clause *c;
int *l,ii;
{
int sl, i;
 
for(i = ii; i > 0; i--) {                 /* for each argument  */
   sl = *l;
   if(sl & FCNMASK) {                   /* FCN:just copy it over */
      *ep++ = sl;
      l++;
      if(sl & ARGMASK) {
	 if((l = emit_literal(c,(int) l,sl & ARGMASK)) == 0) {
	    return(0);}}
      else constants_in_clause++;}
   else {
      if((unsigned int) var_substitution_tbl[sl] < MAX_VARS_X2) {
	 if(!var_substitution_tbl[sl]) {
	    var_substitution_tbl[sl] = (int *) ++c->c_nv;}
 
/*----------------*/
/* write it out   */
/*----------------*/
 
	  *ep++ = (int) var_substitution_tbl[sl] - 1;}
 
      else if(!emit_literal(c,var_substitution_tbl[sl],1)) {
	 return(0);} 
      l++;}}
return(l);
}
 
/*----------------------------------------------------------------*/
/*      generate_literal(c,lit)     .....................#4
/*
/*      add literal lit from clause c to new clause
/*
/*      called from:   generate_clause()          gen.c
/*
/*      calls:         emit_literal()             gen.c
/*                     calc_lit_hc()              hash.c
/*
/*      returns:       0 => literal is ok 
/*                     1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
int generate_literal(c,lit)
clause *c;
literal *lit;
{
literal *new_literal;
int new_lit_terms;
	
new_literal = (literal *) ep;                    /* new literal  */
new_literal->l_next = 0;
new_literal->l_id = lit->l_id & ~VARMASK;
var_in_lit = 0;

ep = (int *) (new_literal + 1);                 /* args go here */
 
if(!emit_literal(c,(int) (lit + 1),lit->l_id & ARGMASK)) {
   return(1);}
 
/*------------------------------------------*/
/* determine a hash code for the literal    */
/* and receive the number of term in lit    */
/*------------------------------------------*/
 
new_lit_terms = calc_lit_hc(new_literal,c);

/*------------------------------------------*/
/* if number of variables in literal is not */
/* equal to zero, save in l_id              */
/*------------------------------------------*/

if(var_in_lit) {new_literal->l_id |= VARMASK;}

/*------------------------------------------*/
/* if too many terms in lit, kill clause    */
/*------------------------------------------*/

if(new_lit_terms > o_maxterm) {
   if(in_elim) {return(1);}}
 
return(0);
}
 
/*----------------------------------------------------------------*/ 
/*      emit_literal2(c,l,ii)           ......................#5
/*
/*      emit_literal is a recursive fxn that will apply the subs
/*      of the var_substitution_tbl to a given agmnt and write the 
/*      expanded version.
/*
/*      parameters:     c : ^ clause header
/*                      l : ^ arguments
/*                      i : number of arguments
/*
/*      Note: we write out the expanded version at the current
/*      "ep" location.
/*
/*      called from:    generate_literal()      gen.c
/*
/*      calls:          emit_literal()          gen.c
/*
/*      returns:        pointer past arguments
/*                      NULL if fcn nesting too deep.
/*--------------------------------------------------------------*/
 
int *emit_literal2(c,l,ii)
clause *c;
int *l,ii;
{
int sl, i, jj;
int loop;

for(i = ii; i > 0; i--) { 
   sl = (int) *l;
   if(sl & FCNMASK) {        
      *ep++ = sl;
      l++;
      if(sl & ARGMASK) {
	  if((l = emit_literal2(c,(int) l,sl & ARGMASK)) == 0) {
	     return(0);}}
      else constants_in_clause++;}

   else {
      if((unsigned int) var_substitution_tbl[sl] < MAX_VARS_X2) {
         if(var_substitution_tbl[sl] == 0) {
	    var_substitution_tbl[sl] = (unsigned int) sl + 1;
	       if(c->c_nv < (((unsigned int) sl + 1) & 0x0000003F)) 
	          c->c_nv = ((unsigned int) sl + 1) & 0x0000003F;}

/*----------------*/
/* write it out   */
/*----------------*/

         *ep++ = (int) var_substitution_tbl[sl] - 1;}

      else {

         if(!emit_literal2(c,var_substitution_tbl[sl],1)) {
            return(0);}}
      l++;}}
return(l);
}

/*----------------------------------------------------------------*/
/*      generate_literal2(c,lit)     .....................#6
/*
/*      add literal lit from clause c to new clause
/*
/*      called from:   generate_clause()          gen.c
/*
/*      calls:         emit_literal()             gen.c
/*                     calc_lit_hc()              hash.c
/*
/*      returns:       0 => literal is ok 
/*                     1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
void generate_literal2(c,lit)
clause *c;
literal *lit;
{
literal *new_literal;
	
new_literal = (literal *) ep;                    /* new literal  */
new_literal->l_next = 0;
new_literal->l_id = lit->l_id;
var_in_lit = 0;
new_literal->l_X = lit->l_X;
new_literal->l_Y = lit->l_Y;

ep = (int *) (new_literal + 1);                 /* args go here */
 
emit_literal2(c,(int) ( lit + 1),lit->l_id & ARGMASK);
}

/*--------------------------------------------------------------*/
/*      repair_literal(c,index)       ......................#7
/*
/*      replaces a literal with its equivalent hash code when, for
/*      example, the literal is of the form ~Equal(x,x) since this
/*      literal always resolves with Equal(x,x).
/*
/*      called from:    install_clause()            search.c
/*
/*      calls:          calc_clause_hashcode()      hash.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void repair_literal(c,index)
clause *c;
int index;
{
int lit_index;
literal *l;
 
l = c->c_lit;
for(lit_index = index; lit_index > 1; lit_index--) {l = l->l_next;}
 
/*-------------------------------*/
/* remove literal by setting     */
/* the bit l->l_id & ELIMLIT     */ 
/* to record literal             */
/* is resolved away              */
/*-------------------------------*/
 
l->l_id |= ELIMLIT;
 
/*-------------------------------*/
/* if only one literal remains,  */
/* set c_bits field for unit cls */
/*-------------------------------*/
 
if(--(c->c_nlit) == 1) {c->c_bits |= UNIT;}
calc_clause_hashcode(c);
return;
}
 
/*---------------------------------------------------------------*/
/*     eval_fxns()                   ................... #8
/*
/* scans the literal to see if fxns can be evaluated and 
/* substituted for.  For example, if the theorem has an
/* axiom that says times(0,x) is 0, then make this evaluation.
/*
/*     called by:
/* 
/*     calls:
/*
/*     returns: NIL
/*---------------------------------------------------------------*/

int eval_fxns(c,lit)
clause *c;
literal *lit;
{
int *p, *pp, tt, a, i, ii, t, k, lit_end;
int kount, var_count, position;

p = &lit->l_id;
position = 1;

for(k = 0; k < c->c_nv; k++) {vars_sub[k] = 0;}
vars_sub_ptr = 0;

for(a =*p++ & ARGMASK; a > 0; a--) {
   if((*p & FCNMASK) && ((*p & ARGMASK) != 0)) { 

/*-------------------------------------------------*/
/* function found:                                 */
/* compare string with those stored; clear vst     */
/* pp is pointer to secondary string being scanned */
/* i is the row of the sub_array under study       */
/*-------------------------------------------------*/

      pp = p; 
      i = 1;
      var_count = 0;

/*-------------------------------------------------*/
/* now loop through substring for match            */
/*-------------------------------------------------*/

loophere: 
      if(((sub_array[i][0] & 0xFFFF0000) == 0xFFFF0000) 
	    && ((*pp & NEGMASK) == 0 || ((*pp & ARGMASK) == 0))) {

/*-------------------------------------------------*/
/* symbol in sub_array is a var, in string a const */
/* or a var                                        */
/*-------------------------------------------------*/

         if((vars_sub[sub_array[i][0] & 0x0000FFFF]) == 0) {

/*-------------------------------------------------*/
/* first time variable or constant has been seen   */
/*-------------------------------------------------*/

            if((*pp & 0x80000000) != 0) {
               vars_sub[sub_array[i][0] & 0x0000003F] = *pp;}
            else {
               vars_sub[sub_array[i][0] & 0x0000003F] = 1 + *pp;}}

/*-------------------------------------------------*/
/* var seen before                                 */
/*-------------------------------------------------*/

         else if((vars_sub[sub_array[i][0]&0x0000FFFF]&0x80000000)!=0) {
            if(vars_sub[sub_array[i][0] & 0x0000FFFF] != *pp) {
               goto get_brother;}}
         else if((vars_sub[sub_array[i][0]&0x0000FFFF]&0x80000000)==0) {
	    if(vars_sub[sub_array[i][0] & 0x0000FFFF] != *pp + 1) {    

/*-------------------------------------------------*/
/* different from previous val assigned to var     */
/*-------------------------------------------------*/

               goto get_brother;}} 

         if(sub_array[i][1]  & 0x40000000) {goto terminal;}
         else {
            i = sub_array[i][1];     /* go deeper on child */
	    var_count++;
            pp++;goto loophere;}}
      else {
         if(sub_array[i][0] == *pp) {
            if(sub_array[i][1]  & 0x40000000) {goto terminal;}
	    else {
               i = sub_array[i][1];  /* go deeper on child */
               pp++;goto loophere;}}
         if((sub_array[i][0] & 0xFFFF0000) == 0xFFFF0000 
			        && (sub_array[i][1] & 0x40000000)
			        && (*pp & 0x80000000)
		  		&& (*pp & 0x0000003F)
				&& var_count == 0
				&& sub_array[i][1] == 0xE0000000) {
	      s_subfxn++;goto terminal2;}

/*-------------------------------------------------*/
/* get brother                                     */
/*-------------------------------------------------*/

         else {
get_brother:
            i = sub_array[i][2]; 
            if(i == -1) {goto adjust;} 
	    goto loophere;}}
adjust:
      for(k = 0; k < c->c_nv; k++) {vars_sub[k] = 0;}
      vars_sub_ptr = 0;
      position++;
      a += *p++ & ARGMASK;}
   else {
      for(k = 0; k < c->c_nv; k++) {vars_sub[k] = 0;}
      vars_sub_ptr = 0;
      position++; p++;}}

return(0);
terminal:
if(lit->l_next == 0) {lit_end = (int) c->c_end;}
else {lit_end = (int) lit->l_next;}

/*--------------------------------------*/
/* replace string with a variable       */
/* or a constant as a result of a subst */
/*--------------------------------------*/

/*
if((sub_array[i][1] & 0x0FFF000) != 0) {
   if((sub_array[i][1] & 0xE0000000) == 0xE0000000) {
      *p = (sub_array[i][1] & 0x00070000) >> 16;
      *p = *p | ((sub_array[i][1] & 0x0FF80000) >> 13) | 0x80000000;
      p++;
      *p = vars_sub[(sub_array[i][1] & 0x0000FFFF)];
      if((*p & 0x80000000) == 0) {
         *p = *p - 1;
         tt = *p << 16;}
      else {tt = (*p << 11) | 0x00800000;}
      p++; pp++;
      while((unsigned int) pp < (unsigned int) lit_end) {
         *p = *pp; p++; pp++;}
      goto patch_clause;}
   else if((sub_array[i][1] & 0xC0000000) == 0xC0000000) {
      *p = (sub_array[i][1] & 0x00070000) >> 16;
      *p = *p | ((sub_array[i][1] & 0x0FF80000) >> 13) | 0x80000000;
      p++;
      *p = sub_array[i][1] & 0x8000FFFF;
      tt = ((*p & 0x0000FFFF) << 11) | 0x00800000;
      p++; pp++;
      while((unsigned int) pp < (unsigned int) lit_end) {
         *p = *pp; p++; pp++;}
      goto patch_clause;}}
*/

if((sub_array[i][1] & 0xE0000000) == 0xE0000000) {
   *p = vars_sub[(sub_array[i][1] & 0x0000FFFF)];
   if((*p & 0x80000000) == 0) {
      *p = *p - 1;
      tt = *p << 16;}
   else {tt = (*p << 11) | 0x00800000;}
   p++; pp++;
   while((unsigned int) pp < (unsigned int) lit_end) {
      *p = *pp; p++; pp++;}
   goto patch_clause;}

/*----------------------------------------*/
/* replace string with a constant         */
/*----------------------------------------*/

else if((sub_array[i][1] & 0xC0000000) == 0xC0000000) {
   *p = sub_array[i][1] & 0x8000FFFF;
   tt = ((*p & 0x0000FFFF) << 11) | 0x00800000;
   p++; pp++;
   while((unsigned int) pp < (unsigned int) lit_end) {
      *p = *pp; p++; pp++;}
   goto patch_clause;}

terminal2:
if(lit->l_next == 0) {lit_end = (int) c->c_end;}
else {lit_end = (int) lit->l_next;}

tt = 0x00FF0000;
while((unsigned int) pp < (unsigned int) lit_end) {
   *p = *pp; p++; pp++;}

patch_clause:
*p = (position << 24) + i;
*p = *p | tt;
p++;
while((unsigned int) p < (unsigned int) lit_end) {*p++ = 0;}

/*------------------------------------------------*/
/* save pointer into sub_array for reconstruction */
/* bits 24-32: poistion where deletion starts     */
/* bits 16-23: var/const to substitute... if var  */
/*             actually v1 = 0, etc..             */
/* bits 00-15: sub table location of sub.         */
/*------------------------------------------------*/

return(1);
}
 
/*---------------------------------------------------------------*/
/*      prepare_subarray()                 .............. #9
/*
/*      prepares the function evaluation array
/*
/*      called by:
/*
/*      calls:
/*
/*      returns: NIL
/*---------------------------------------------------------------*/

void prepare_subarray()
{
clause *c;
int i;

sub_array[0][0] = 0;
sub_array[0][1] = 1;
sub_array[0][2] = -1;
sub_array[0][2] = -1;

sub_array_last = 0;

for(i = 0; i < base; i++) {
   c = clist[i];
   if((c->c_bits & ACTIVE) == 0) continue;
   if(c->c_nlit == 1 && sub_array_last <  65000
        && phase > 0) {
       decide_to_add1(c);}}

const_found_equal = 0;
}
 
/*------------------------------------------------------------*/
/* add_string()                         ............. #10
/*
/* called by: install_clause()        search.c
/*
/* calls:     nobody
/*
/* returns:   nill
/*------------------------------------------------------------*/

void add_string(c,ptr1,ptr2,string_length,numb_char)
clause *c;
int *ptr1, *ptr2, string_length, numb_char;
{
int i, p, q, k, length, var_id, *ptr3;
int parent, final_parent, brother;
int vars_in_string;
int temp_fxn;

length = string_length;
i = 1;
var_id = 0;
parent = -1;
final_parent = 0;
vars_in_string = 0;

/*---------------------------*/
/* clear var sub array       */
/*---------------------------*/

for(k = 0; k < c->c_nv; k++) {vars_sub[k] = 0;}
vars_sub_ptr = 0;

top: 
if(length == 0) return;

if((*ptr2 & FCNMASK) == 0) { 
   if(!vars_sub[*ptr2 & MAX_VARS_M1]) {
      vars_sub[*ptr2 & MAX_VARS_M1] = vars_sub_ptr++ + 1;}

   if(sub_array[i][0] == 
      ((vars_sub[*ptr2 & MAX_VARS_M1] - 1) | 0xFFFF0000)) {

/*---------------------------*/
/* variable found in string  */
/*---------------------------*/

      vars_in_string++;
      length--;
      ptr2++; 
      parent = i;
      i = sub_array[i][1];   /* set i to first child */
      goto top;}
   else if(sub_array[i][2] != -1) {
      i = sub_array[i][2];
      goto top;}}

if(sub_array[i][0] == *ptr2 || (((*ptr2 & FCNMASK) == 0) 
     && (sub_array[i][0] == 
         ((vars_sub[*ptr2 & MAX_VARS_M1] - 1) | 0xFFFF0000)))) { 

/*---------------------------*/
/* function found in string  */
/*---------------------------*/

   if((sub_array[i][1] & 0xF0000000) != 0) {return;}

   length--;
   if(length == 0) {return;}

   ptr2++; 
   parent = i;
   i = sub_array[i][1];   /* set i to first child */
   goto top;}

if(sub_array[i][2] != -1) {
   i = sub_array[i][2];       /* set i to next brother */
   goto top;}

/*-------------------------------------------*/
/* now add remaining part of the string to   */
/* the table.                                */
/*-------------------------------------------*/

else {
   sub_array_last++;

   if(parent == -1) {
      p = 1;

set_ptr:
      q = sub_array[p][2];
      if(q == -1) {sub_array[p][2] = sub_array_last;}
      else {
         p = q;
	 goto set_ptr;}}
   else { 
      brother = sub_array[parent][1];
      if((*ptr2 & 0xF0000000) != 0) {sub_array[parent][1] = sub_array_last;}
      else {
         sub_array[sub_array_last][2] = -1;
	 parent = -1;
         sub_array[i][2] = sub_array_last;}}

   while(length-- > 0) {
      if((*ptr2 & FCNMASK) == 0) {
         vars_in_string++;
         if(!vars_sub[*ptr2 & MAX_VARS_M1]) {
            vars_sub[*ptr2 & MAX_VARS_M1] = vars_sub_ptr++ + 1;}
         sub_array[sub_array_last][0] = (0xFFFF0000 | 
            vars_sub[*ptr2 & MAX_VARS_M1]) - 1;
	 *ptr2++;}
      else {sub_array[sub_array_last][0] = *ptr2++;}

      if(length > 0) {
         sub_array[sub_array_last][1] = sub_array_last + 1;
	 if(parent != -1) {
             sub_array[sub_array_last++][2] = brother;
	     parent = -1;}
	 else {sub_array[sub_array_last++][2] = -1;}}

/*-------------------------------------------*/
/* now add replacement character...          */
/* first, save hashcode of clause            */
/*-------------------------------------------*/

      else {
         sub_array[sub_array_last][3] = c->c_X;
         sub_array[sub_array_last][4] = c->c_Y;
         sub_array[sub_array_last][5] = s_nod - s_nod0;

/*-------------------------------------------*/
/* test to see if one or two chars saved for */
/* replacement.                              */
/*-------------------------------------------*/

         if(numb_char == 1) {

            if((*ptr1 & FCNMASK) == 0) { 

/*-------------------------------------------*/
/* replace string with a variable            */
/*-------------------------------------------*/

               if(!vars_sub[*ptr1 & MAX_VARS_M1]) {
                  vars_sub[*ptr1 & MAX_VARS_M1] = vars_sub_ptr++;}
               
               sub_array[sub_array_last][1] = (0xE0000000 |
                            (vars_sub[*ptr1 & MAX_VARS_M1])) - 1;

               if(parent != -1) {
	          final_parent = parent;
                  sub_array[sub_array_last++][2] = brother;
	          parent = -1;}
	       else {sub_array[sub_array_last++][2] = -1;}}
	     
	    else {

/*---------------------------------------------*/
/* replace string with a constant              */
/*---------------------------------------------*/

               sub_array[sub_array_last][1] = *ptr1 | 0xC0000000;
	       if(parent != -1) {
                  final_parent = parent;
                  sub_array[sub_array_last++][2] = brother;
	          parent = -1;}
	       else {sub_array[sub_array_last++][2] = -1;}}}
          
         else {

/*---------------------------------------------*/
/* replace string of length 4+ by fxn c or v   */
/*---------------------------------------------*/

	    printf("-a-");
	    ptr3 = ptr1;
	    ptr3++;

            if((*ptr3 & FCNMASK) == 0) { 

/*-------------------------------------------*/
/* replace string with a variable            */
/* and then add in the function. make sure   */
/* function fits.                            */
/*-------------------------------------------*/

               if(!vars_sub[*ptr3 & MAX_VARS_M1]) {
                  vars_sub[*ptr3 & MAX_VARS_M1] = vars_sub_ptr++;}
               
               sub_array[sub_array_last][1] = (0xE0000000 |
                            (vars_sub[*ptr3 & MAX_VARS_M1])) - 1;

	       temp_fxn = *ptr1 & 0x00000007;
	       temp_fxn = (((*ptr1 >> 3) | temp_fxn) & 0x0000FFFF) << 16;

	       sub_array[sub_array_last][1] += temp_fxn;
               if(parent != -1) {
	          final_parent = parent;
                  sub_array[sub_array_last++][2] = brother;
	          parent = -1;}
	       else {sub_array[sub_array_last++][2] = -1;}}
	     
            else {

/*---------------------------------------------*/
/* replace string with a constant              */
/*---------------------------------------------*/

               sub_array[sub_array_last][1] = *ptr3 | 0xC0000000;
	       temp_fxn = *ptr1 & 0x00000007;
	       temp_fxn = (((*ptr1 >> 3) | temp_fxn) & 0x0000FFFF) << 16;
	       sub_array[sub_array_last][1] += temp_fxn;
	       if(parent != -1) {
                  final_parent = parent;
                  sub_array[sub_array_last++][2] = brother;
	          parent = -1;}
               else {sub_array[sub_array_last++][2] = -1;}}}}}}

sub_array_last--;

/*
print_clause(c);
print_sub_array(c);
*/

return;
}

/*--------------------------------------------------------------*/
/* print_sub_array  ...............................#11
/*--------------------------------------------------------------*/

void print_sub_array(c)
clause *c;
{
int i;


for(i = 0; i <= sub_array_last; i++) {
   printf("\n%x %9x %9x %9x %9x %9x",
		   i,sub_array[i][0],sub_array[i][1],
		   sub_array[i][2],sub_array[i][3],sub_array[i][4]); 
   if((sub_array[i][0] & 0xFFFF0000) != 0xFFFF0000) {
      printf(" %s",fidtbl[(sub_array[i][0] >> 5) & 0x000000FF]);}}


print_clause(c);
printf("\n\n");
return;
}

/*-------------------------------------------------------------*/ 
/* copy_clause(c)    ..............................#12
/*
/* c1 = clause to copy
/* liste = list on which to copy clause
/*
/* called by:    add_clause()                   herby.c
/*
/* calls:        clear_var_substitution_tble()  infer.c
/*               gen_lit()                      hgen.c
/*               calc_clause_hashcode()         hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*-------------------------------------------------------------*/
 
clause *copy_clause(c1,lit_index,elim_index)
clause *c1;
int lit_index, elim_index;
{
clause *c;                                  /* new clause       */
literal *curl, *prevl;    /* current literal; previous literal  */
literal *l1;
int e_index;

clear_var_substitution_tble(c1->c_bits,c1->c_nv);

c = clist[nextc];

c->c_bits = c1->c_bits;
c->c_lpar = c1;                    /* parents of this clause   */
c->c_rpar = c1;
c->c_nv   = 0;                            /* filled in later   */
c->c_bits |= ACTIVE;
c->c_change = 0;

prevl = 0; 
ep = (int  *) (c + 1);             /* lit's go after cl header */
e_index = 0;
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   e_index++;
   if((l1->l_id & ELIMLIT) == 0) {
      if(e_index != elim_index) {

/*-----------------------*/
/* while there are lits  */
/*-----------------------*/
   
         curl = (literal *) ep;       /* literal that we're making   */
 
/*------------------------------------------------------*/
/* if l1 or l2 is a hashcode literal, generate          */
/* a hashcode literal                                   */
/*------------------------------------------------------*/
            
         generate_literal2(c,l1);

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
         if(prevl == 0)  {c->c_lit = curl;}
         else  {prevl->l_next = curl;}        
         prevl = curl;}}}

if(c->c_lpar->c_bits & NC) {c->c_bits |= NC;}
c->c_nlit = c1->c_nlit;
c->c_end = (clause *) ep;               /* pointer past clause */
c->c_indices = (lit_index << LEFT_INDEX_SHIFT) + lit_index;
c->c_id = c1->c_id + 1;
c->c_nv = c1->c_nv; 
calc_clause_hashcode(c);

return(c);
}

/*-------------------------------------------------------------*/ 
/* copy_clause3(c)    ..............................#13
/*
/* c1 = clause to copy
/* liste = list on which to copy clause
/*
/* called by:    add_clause()                   herby.c
/*
/* calls:        clear_var_substitution_tble()  infer.c
/*               gen_lit()                      hgen.c
/*               calc_clause_hashcode()         hhash.c
/*
/* returns:      ptr to new clause or NULL if none generated
/*-------------------------------------------------------------*/
 
void copy_clause3(c1)
clause *c1;
{
clause *c, *u;                                  /* new clause       */
literal *l1;
int i;

c = ulist[nextu];

c->c_id      = c1->c_id;
c->c_lpar    = c1->c_lpar;
c->c_rpar    = c1->c_rpar;
c->c_indices = c1->c_indices;
c->c_nlit    = c1->c_nlit;
c->c_nv      = c1->c_nv + 1;                 /* filled in later   */
c->c_bits    = ACTIVE | BASE | NC;
if(c1->c_bits & VID) {c->c_bits |= VID;}
if(c->c_nlit == 1) c->c_bits = c->c_bits | UNIT;
c->c_change  = c1->c_change;
c->c_X       = c1->c_X;
c->c_Y       = c1->c_Y;

ep = (int *) (c + 1);              /* lit's go after cl header */
c->c_lit = (literal *) ep;
generate_literal3(c1,c);
c->c_end = (clause *) ep;             /* pointer past clause */
nextu++;
ulist[nextu] = c->c_end;
return;
}

/*----------------------------------------------------------------*/
/* copy_clause_all()                ....................#14
/*
/* called by: bsub_instantiated_proof
/*
/* calls:     clear_var_substitution_tble()
/*            generate_literal2()
/*            calc_clause_hashcode()
/*
/* copies a clause in its entirety to ulist
/*----------------------------------------------------------------*/

void copy_clause_all(c)
clause *c;
{
clause *c1;                                 /* new clause       */
literal *curl, *prevl;    /* current literal; previous literal  */
literal *l1;

clear_var_substitution_tble(c->c_bits,c->c_nv);

c1 = ulist[nextu];

c1->c_id = c->c_id;
c1->c_bits = c->c_bits;
c1->c_lpar = c->c_lpar;             /* parents of this clause   */
c1->c_rpar = c->c_rpar;
c1->c_indices = c->c_indices;
c1->c_nv   = 0;                            /* filled in later   */
c1->c_bits |= ACTIVE;
c1->c_change = c->c_change;
c1->c_nlit = c->c_nlit;

prevl = 0; 
ep = (int  *) (c1 + 1);             /* lit's go after cl header */

for(l1 = c->c_lit; l1 != 0; l1 = l1->l_next) {

/*-----------------------*/
/* while there are lits  */
/*-----------------------*/
   
   curl = (literal *) ep;       /* literal that we're making   */
 
/*------------------------------------------------------*/
/* if l1 or l2 is a hashcode literal, generate          */
/* a hashcode literal                                   */
/*------------------------------------------------------*/
            
   generate_literal2(c,l1);

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if(prevl == 0)  {c1->c_lit = curl;}
   else  {prevl->l_next = curl;}        
   prevl = curl;}

c1->c_end = (clause *) ep;               /* pointer past clause */
calc_clause_hashcode(c1);
return;
}

/*----------------------------------------------------------------*/
/* copy_clause_elim()                ....................#15
/*
/* called by: bsub_instantiated_proof
/*
/* calls:     clear_var_substitution_tble()
/*            generate_literal2()
/*
/* copies a clause in its entirety to ulist
/*----------------------------------------------------------------*/

int copy_clause_elim(c)
clause *c;
{
clause *c1;                                 /* new clause       */
literal *curl, *prevl;    /* current literal; previous literal  */
literal *l1;

clear_var_substitution_tble(c->c_bits,c->c_nv);

c1 = clist[nextc];

c1->c_bits = c->c_bits;
c1->c_lpar = 0;                     /* parents of this clause   */
c1->c_rpar = 0;
c1->c_bits |= ACTIVE;
c1->c_change = c->c_change;
c1->c_nlit = 0;

prevl = 0; 
ep = (int  *) (c1 + 1);             /* lit's go after cl header */

for(l1 = c->c_lit; l1 != 0; l1 = l1->l_next) {

/*-----------------------*/
/* while there are lits  */
/*-----------------------*/
   
   if(l1->l_id & ELIMLIT) continue;
   curl = (literal *) ep;       /* literal that we're making   */
 
/*------------------------------------------------------*/
/* if l1 or l2 is a hashcode literal, generate          */
/* a hashcode literal                                   */
/*------------------------------------------------------*/
            
   generate_literal2(c,l1);
   c1->c_nlit++;

/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if(prevl == 0)  {c1->c_lit = curl;}
   else  {prevl->l_next = curl;}        
   prevl = curl;}

c1->c_end = (clause *) ep;               /* pointer past clause */
c1->c_indices = 0;
c1->c_nv = c->c_nv;
nextc++; 
clist[nextc] = c1->c_end;
c1->c_id = nextc;
return(1);
}

/*----------------------------------------------------------------*/
/*      generate_literal3(lit)     .....................#16
/*
/*      add literal lit from clause c to new clause
/*
/*      called from:   copy_clause3()             gen.c
/*
/*      calls:         emit_literal()             gen.c
/*                     calc_lit_hc()              hash.c
/*
/*      returns:       0 => literal is ok 
/*                     1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
void generate_literal3(c1,c)
clause *c1, *c;
{
int delta;
int *eq1, *ep1, eq_next_lit;

eq = &c1->c_lit->l_next;
ep = &c->c_lit->l_next;
eq_next_lit = *eq;

while((unsigned int) eq < (unsigned int) c1->c_end) {
      if((unsigned int) eq == eq_next_lit || 
              (unsigned int) eq == (unsigned int) &c1->c_lit->l_next) {
         eq_next_lit = *eq;
	 if(eq_next_lit == 0) {
	    *ep = *eq;
	    ep++;eq++;}
	 else {
	    delta = (eq_next_lit - (int) eq)/4;
	    *ep = ep + delta;
	    ep++;eq++;}}
      else {
	 *ep = *eq;
	 ep++;eq++;}}
}
 
/*--------------------------------------------------------------*/
/*      test_equality(l)       ......................#17
/*
/*      tests to see if equal(string1,string1) is the case
/*
/*      called from:   generate_clause()               gen.c
/*                     install_clause()                search.c
/*
/*      calls:         nobody
/*
/*      returns:    0 => if equal(string1,string1) is not the case
/*                  1 => if equal(string1,string1) is the case
/*--------------------------------------------------------------*/
 
int test_equality(l_equal)
literal *l_equal;
{
int *p, *p1, a;

p1 = p = &l_equal->l_id;

for(a = (*p++ & ARGMASK) ; a > 1 ; a--) {
   if(*p & FCNMASK) {
      a += *p++ & ARGMASK;}
   else {p++;}}
 
/*--------------------------------------*/
/* p now points to beginning of 2nd     */
/* half of the equality predicate.      */
/* p1 now points to the beginning.      */
/* now a scan is made comparing each    */
/* symbol in the two halfs              */
/*--------------------------------------*/

for(a = (*p1++ & ARGMASK) ; a > 1 ; a--) {
   if(*p != *p1) {return(0);} 
   if(*p1 & FCNMASK) {
      a +=*p1++ & ARGMASK;
      p++;}
   else {p++;p1++;}}

return(1);
}

/*--------------------------------------------------------------*/
/*      test_simple_equality(l)       ......................#18
/*
/*      tests to see if equal(string1,string1) is the case
/*
/*      called from:   generate_clause()               gen.c
/*                     install_clause()                search.c
/*
/*      calls:         nobody
/*
/*      returns:    0 => if equal(string1,string1) is not the case
/*                  1 => if equal(string1,string1) is the case
/*--------------------------------------------------------------*/
 
int test_simple_equality(c_equal)
clause *c_equal;
{
literal *l_equal;
int *p, *p1, *pL1, *pL2, *pt, p2, a, once;

l_equal = c_equal->c_lit;
once = 0;

if((l_equal->l_id & 0x0000FFFF) != equal_pred) {
   return(0);}

p1 = p = &l_equal->l_id;

for(a = (*p++ & ARGMASK) ; a > 0 ; a--) {
   if(a == 1 && !once) {pL1 = p - 1;}
   if(*p & FCNMASK) {
      a += *p++ & ARGMASK;}
   else {p++;}}
 
/*--------------------------------------*/
/* p now points to beginning of 2nd     */
/* half of the equality predicate.      */
/* p1 now points to the beginning.      */
/* now a scan is made comparing each    */
/* symbol in the two halfs              */
/*--------------------------------------*/

pL2 = p - 1;
p1++;
return(0);

if((int) p == (int) p1 + 4 && *p != *p1 && (*p & ARGMASK) == 0) {
   p2 = *p1;
   *p1 = *p;
   *p = p2;
   calc_lit_hc(l_equal,c_equal);
   calc_clause_hashcode(c_equal);
   add_clause_to_ht(c_equal->c_X,c_equal->c_Y,c_equal->c_nlit);
   p2 = *p1;
   *p1 = *p;
   *p = p2;
   calc_lit_hc(l_equal,c_equal);
   calc_clause_hashcode(c_equal);
   return(1);}

return(0);
}

/*--------------------------------------------------------------*/
/*      test_const_equal(c_equal,l_equal)   ..............#19
/*
/*      tests to see if equal(const1,const2) is the case
/*
/*      called from:   generate_clause()               gen.c
/*                     install_clause()                search.c
/*
/*      calls:         copy_clause_elim()              gen.c
/*                     calc_lit_hc()                   hash.c
/*                     calc_clause_hashcode()          hash.c
/*
/*      returns:    0 => if equal(string1,string1) is not the case
/*                  1 => if equal(string1,string1) is the case
/*--------------------------------------------------------------*/
 
int test_const_equal(c_equal,l_equal)
clause *c_equal;
literal *l_equal;
{
int *p, *pp, const1, const2;

p = &l_equal->l_id;

if((*p++ & 0x8000FFFF) != equal_pred) return(0);
const1 = *p++;
const2 = *p;

if(((const1 | const2) & ARGMASK) == 0 && const1 != const2) {
   if((const1 & FCNMASK) != 0 && (const2 & FCNMASK) != 0) {
      const1 = (const1 & IDMASK) >> 6;
      const2 = (const2 & IDMASK) >> 6;
      return(const1 + (const2 << 8));}}
return(0);
}

/*----------------------------------------------------------------*/
/*     replace_constant(c_equal,l_equal) ................ #20
/*
/*     replaces const2 with const 1 throughout theorem
/*
/*     called by: 
/*
/*     calls:
/*
/*     returns:
/*
/*-----------------------------------------------------------------*/

int replace_constant(c_equal,const_replace)
clause *c_equal;
int const_replace;
{
int *p, *pp, i, a, aa, const1, const2, sub_const, mod_cls;
clause *c, *c2;
literal *lit, *lit2;

/*----------------------------------------------*/
/* c_equal is of the form Equal(cosnt1,const2)  */
/* - - - - - - - - - - - - - - - - - - - - - -  */ 
/* scan base clauses to replace clauses with    */
/* constant const2 with const1                  */
/*----------------------------------------------*/

const1 = (const_replace & 0x000000FF) << 6;
const1 |= FCNMASK;

const2 = (const_replace >> 8) << 6;
const2 |= FCNMASK;

mod_cls = 0;
for(i = 0; i <= base; i++) {
   sub_const = 0;
   c = clist[i];
   if((c->c_bits & ACTIVE) == 0) continue;
   if(c == c_equal) {
      reinstate[i] = 2;
      c->c_bits &= ~ACTIVE;}
   else {
      for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
         if(sub_const != 0) break;
         p = &lit->l_id;
         for(a = *p++ & ARGMASK; a > 0; a--) {
            if(*p & FCNMASK) {
	       if(*p == const2) {
	          copy_clause_elim(c);
	          c->c_bits &= ~ACTIVE;
                  reinstate[i] = 2;
                  reinstate_parent[base + 1] = c_equal->c_id; 
		  base++;
                  mod_cls = 1;

/*-------------------------------------------------------------------*/
/* elim const2 by replacing with const1
/*-------------------------------------------------------------------*/

                  c2 = clist[nextc - 1];
		  for(lit2 = c2->c_lit; lit2 != 0; lit2 = lit2->l_next) {
                     pp = &lit2->l_id;
                     for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
                        if(*pp & FCNMASK) {
                           if(*pp == const2) {*pp = const1;}
                           aa += *pp++ & ARGMASK;} 
                        else {pp++;}}
                     calc_lit_hc(lit2,c2);}
                  calc_clause_hashcode(c2);
                  c2->c_change = (const1 &  0x000FFFc0) << 2; 
                  c2->c_change += (const2 & 0x000FFFc0) >> 6;
                  c2->c_change |= 0x80000000;
                  c2->c_lpar = c2->c_rpar = c;
                  sub_const = 1;
                  break;}
               a += *p++ & ARGMASK;}
	    else {p++;}}
         if(sub_const == 1) break;}}}
return(mod_cls);
}

/*------------------------------------------------------------------*/ 
/*      *hgenerate_clause(c1,c2,lit1,lit2,ix1,ix2)    ........#21
/*
/*      generate a clause from one or two parent cls
/*      Apply all changes currently in the 
/*      var_substitution_tbl while we gen the new lits. 
/*      The variables are st'dized apart. Note that we destroy
/*      the var_substitution_tbl when we generate the clause
/*
/*      parameters:     c1,c2: left and right parents
/*                      lit1,lit2: literals to leave out
/*                      ix1, ix2 are indices of resolved literals
/*
/*      A clause will not be gen. for various reasons (too many lits,
/*      or the clause is a tautology)
/*
/*      called from:  binary_resolutions()      infer.c
/*                    binary_factors()          infer.c
/*                    bres_ply_1()              elim.c
/*                    bres_link()               elim.c
/*                    bres_replace()            elim.c
/*                    bres_proof_line()         pass2.c
/*                    factor_proof_line()       pass2.c
/*
/*      calls:        generate_literal()        gen.c
/*                    calc_clause_hashcode()    hash.c
/*                    test_equality()           gen.c
/*
/*      returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *hgenerate_clause(c1,c2,lit1,lit2,ix1,ix2)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2;
{
clause *c;                                    /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2 , *pl;
int oldnv;
int lnumber;

c = clist[nextc];                      /* where to put the clauses  */

/*-----------------------------*/
/* initialize clause header    */
/*-----------------------------*/              
 
c->c_bits = ACTIVE;                       /* assume active clause   */
c->c_lpar = c1;                         /* parents of this clause   */
c->c_rpar = c2;
c->c_nv   = 0;                                 /* filled in later   */
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0;
c->c_change = 0;

/*--------------------------*/
/* prepare to scan literals */
/*--------------------------*/
 
constants_in_clause = 0;
prevl = 0;                                /* prev literal in scan   */
ep = (int *) (c + 1);                   /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */

if(c2) {l2 = c2->c_lit;}
else {l2 = 0;}
while(l1 || l2) {                      /* while there are literals  */ 
 
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
   if(l1 && ((l1 == lit1) || (l1->l_id & ELIMLIT))) { 
      l1 = l1->l_next; continue;}            
   if(l2 && (l2 == lit2 || (l2->l_id & ELIMLIT))) {
      l2 = l2->l_next; continue;}
 
/*----------------------------------------------------------*/
/* save this information in case literal is deleted---      */
/* in the case that it is identical to another and must     */
/* be deleted                                               */
/*----------------------------------------------------------*/
 
   curl = (literal *) ep;    /* literal that we're making   */
   oldnv = c->c_nv;
	    
   if(!l2 || l1) {
      if(hgenerate_literal(c,l1)) {return(0);}
      l1 = l1->l_next;}
   else {
      if(hgenerate_literal(c,l2)) {return(0);}
      l2 = l2->l_next;}
	       
/*----------------------------------------------------------*/
/* eliminate predicates with Equal(x,x) as tautologies      */
/*----------------------------------------------------------*/
 
   if((curl->l_id & 0x1000FFFF) == equal_pred && phase) { 
      if(test_equality(curl) > 0) {
         if(phase < 3) {
	    return(0);}}}    /* equal tautology */
 
/*----------------------------------------------------------*/
/* we scan back in the current group to see if there        */
/* are any matching literal checksums. If so we perform     */
/* a literal comparison and a) remove the literal if        */
/* it is identical to another; or b) cancel generation      */
/* of the clause if the literals are identical but of       */
/* opposite signs (tautology).                              */
/*----------------------------------------------------------*/
		
tautology_check:
   for(pl = c->c_lit; pl != 0; pl = pl->l_next) {
      if(pl->l_X == -curl->l_X && pl->l_Y == -curl->l_Y) {

/*-----------------------------------------------------------*/
/* clause is a tautology                                     */
/*-----------------------------------------------------------*/

            if(in_elim && phase != 3) {
	       return(0);}}}                    /* tautology */

/*-----------------------------------------------------------*/
/* the literal has survived, we patch the previous lit       */
/* so that it points here                                    */
/*-----------------------------------------------------------*/
 
   if(!prevl) {c->c_lit = curl;}
   else {prevl->l_next = curl;}
	
   prevl = curl;
   c->c_nlit++; 
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/

   if(c->c_nlit > o_maxlit) {
      if(o_maxlit)  {return(0);}}
   if(c->c_nv > o_maxvar) {return(0);}

idn: o_maxlit = o_maxlit;}
if(c->c_nlit == 1) {c->c_bits |= UNIT;}           /* a unit clause  */
c->c_end = (clause *) ep;                    /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + ix2; /* parent lit indices */
c->c_id = base + 1 + ply;                   /* give clause an ID    */

calc_clause_hashcode(c);
return(c);
}

/*----------------------------------------------------------------*/
/*      hgenerate_literal(c,lit)     .....................#22
/*
/*      add literal lit from clause c to new clause
/*
/*      called from:   hgenerate_clause()          gen.c
/*
/*      calls:         hemit_literal()             gen.c
/*                     calc_lit_hc()              hash.c
/*
/*      returns:       0 => literal is ok 
/*                     1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
int hgenerate_literal(c,lit)
clause *c;
literal *lit;
{
literal *new_literal;
int new_lit_terms;
	
new_literal = (literal *) ep;                    /* new literal  */
new_literal->l_next = 0;
new_literal->l_id = lit->l_id & ~VARMASK;
var_in_lit = 0;

ep = (int *) (new_literal + 1);                 /* args go here */
 
if(!emit_literal2(c,(int) (lit + 1),lit->l_id & ARGMASK)) {
   return(1);}
 
/*------------------------------------------*/
/* determine a hash code for the literal    */
/* and receive the number of term in lit    */
/*------------------------------------------*/
 
new_lit_terms = calc_lit_hc(new_literal,c);

/*------------------------------------------*/
/* if number of variables in literal is not */
/* equal to zero, save in l_id              */
/*------------------------------------------*/

if(var_in_lit) {new_literal->l_id |= VARMASK;}

/*------------------------------------------*/
/* if too many terms in lit, kill clause    */
/*------------------------------------------*/

if(new_lit_terms > o_maxterm) {
   if(in_elim) {return(1);}}
 
return(0);
}

/*------------------------------------------------------------------*/
/*      hgen_normalize()               ............. #23
/*------------------------------------------------------------------*/

void hgen_normalize(c)
clause *c;
{
literal *lit;
int a, *p;
int next_avail;
int j, high_low_sum;

for(a = 0; a < 64; a++) low_array[a] = 0;
for(a = 0; a < 64; a++) high_array[a] = 0;

//printf("HGEN: ");print_clause(c);
low_vars = 0;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   p = &lit->l_id;
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {a += *p++ & ARGMASK;}
      else {
         if(*p < 64) {
	    if(low_array[*p + 1] == 0) {
	       if(*p + 1 > low_vars) low_vars = *p + 1;
	       low_array[*p + 1] = *p + 1;}}
         p++;}}}

high_vars = 0;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   p = &lit->l_id;
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {a += *p++ & ARGMASK;}
      else {
         if(*p >= 64) {
	    if(high_array[*p + 1 - 64] == 0) {
	       high_array[*p + 1 - 64] = *p + 1;
	       if(*p + 1 > high_vars) high_vars = *p + 1;}}
         p++;}}}

if(high_vars == 0) return;

high_low_sum = low_vars + high_vars - 64;
c->c_nv = high_low_sum;

for(a = 1; a <= high_low_sum; a++) {
   if(high_array[a] != 0 && low_array[a] == 0) {
      high_array[a] = a;
      low_array[a] = -1;}}

for(a = 1; a <= high_low_sum; a++) {
   j = 0;
   if(high_array[a] != 0 && low_array[a] == -1) continue;
   if(high_array[a] != 0) {
      while(low_array[a + j] != 0) j++;      
      high_array[a] = a + j;
      low_array[a + j] = -1;}}

for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   p = &lit->l_id;
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {a += *p++ & ARGMASK;}
      else {
         if(*p >= 64) {
	    *p = high_array[*p + 1 - 64] - 1;}
         p++;}}}
//printf("HGEN: ");print_clause(c);

}








/*------------------------------------------------------------------*/ 
/*      *generate_clause_p2(c1,c2,lit1,lit2,ix1,ix2)    ........#24
/*
/*      generate a clause from one or two parent cls
/*      Apply all changes currently in the 
/*      var_substitution_tbl while we gen the new lits. 
/*      The variables are st'dized apart. Note that we destroy
/*      the var_substitution_tbl when we generate the clause
/*
/*      parameters:     c1,c2: left and right parents
/*                      lit1,lit2: literals to leave out
/*                      ix1, ix2 are indices of resolved literals
/*
/*      A clause will not be gen. for various reasons (too many lits,
/*      or the clause is a tautology)
/*
/*      called from:  binary_resolutions()      infer.c
/*                    binary_factors()          infer.c
/*                    bres_ply_1()              elim.c
/*                    bres_link()               elim.c
/*                    bres_replace()            elim.c
/*                    bres_proof_line()         pass2.c
/*                    factor_proof_line()       pass2.c
/*
/*      calls:        generate_literal()        gen.c
/*                    calc_clause_hashcode()    hash.c
/*                    test_equality()           gen.c
/*
/*      returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *generate_clause_p2(c1,c2,lit1,lit2,ix1,ix2)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2;
{
clause *c;                                    /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
literal *l1, *l2 , *pl;
int oldnv;
int lnumber;

c = clist[nextc];                      /* where to put the clauses  */

/*-----------------------------*/
/* initialize clause header    */
/*-----------------------------*/              
 
c->c_bits = ACTIVE;                       /* assume active clause   */
c->c_lpar = c1;                         /* parents of this clause   */
c->c_rpar = c2;
c->c_nv   = 0;                                 /* filled in later   */
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0;
c->c_change = 0;

/*--------------------------*/
/* prepare to scan literals */
/*--------------------------*/
 
constants_in_clause = 0;
prevl = 0;                                /* prev literal in scan   */
ep = (int *) (c + 1);                   /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */

if(c2) {l2 = c2->c_lit;}
else {l2 = 0;}
while(l1 || l2) {                      /* while there are literals  */ 
 
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
   if(l1 && ((l1 == lit1) || (l1->l_id & ELIMLIT))) { 
      l1 = l1->l_next; continue;}            
   if(l2 && (l2 == lit2 || (l2->l_id & ELIMLIT))) {
      l2 = l2->l_next; continue;}
 
/*----------------------------------------------------------*/
/* save this information in case literal is deleted---      */
/* in the case that it is identical to another and must     */
/* be deleted                                               */
/*----------------------------------------------------------*/
 
   curl = (literal *) ep;    /* literal that we're making   */
   oldnv = c->c_nv;
	    
   if(!l2 || l1) {
      if(generate_literal(c,l1)) {return(0);}
      l1 = l1->l_next;}
   else {
      if(generate_literal(c,l2)) {return(0);}
      l2 = l2->l_next;}
	       
/*----------------------------------------------------------*/
/* eliminate predicates with Equal(x,x) as tautologies      */
/*----------------------------------------------------------*/
 
   if((curl->l_id & 0x1000FFFF) == equal_pred && phase) { 
      if(test_equality(curl) > 0) {
         if(phase < 3) {
	    return(0);}}}    /* equal tautology */
 
/*----------------------------------------------------------*/
/* we scan back in the current group to see if there        */
/* are any matching literal checksums. If so we perform     */
/* a literal comparison and a) remove the literal if        */
/* it is identical to another; or b) cancel generation      */
/* of the clause if the literals are identical but of       */
/* opposite signs (tautology).                              */
/*----------------------------------------------------------*/
		
tautology_check:
   for(pl = c->c_lit; pl != 0; pl = pl->l_next) {
      if(pl->l_X == -curl->l_X && pl->l_Y == -curl->l_Y) {

/*-----------------------------------------------------------*/
/* clause is a tautology                                     */
/*-----------------------------------------------------------*/

            if(in_elim) {
	       return(0);}}}                    /* tautology */

/*-----------------------------------------------------------*/
/* the literal has survived, we patch the previous lit       */
/* so that it points here                                    */
/*-----------------------------------------------------------*/
 
   if(!prevl) {c->c_lit = curl;}
   else {prevl->l_next = curl;}
	
   prevl = curl;
   c->c_nlit++; 
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/

   if(c->c_nlit > o_maxlit) {
      if(o_maxlit)  {return(0);}}
   if(c->c_nv > o_maxvar) {return(0);}

idn: o_maxlit = o_maxlit;}
if(c->c_nlit == 1) {c->c_bits |= UNIT;}           /* a unit clause  */
c->c_end = (clause *) ep;                    /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + ix2; /* parent lit indices */
c->c_id = base + 1 + ply;                   /* give clause an ID    */

calc_clause_hashcode(c);
return(c);
}

