/*------------------------------------------------------------*/
/* herby.c        Herbrand Base Theorem Prover.... HERBY
/*
/* Copyright Newborn Software 1995 - 1999, 2000
/*
/* functions:
/* time_now()          ...................#1
/* clear_fail_pair_info()   ............. #2
/* init_options()      .................. #3
/* grab_number()       .................. #4
/* fatal()             .................. #5
/* main(argc,argv) ---------------->>>>>>>#6
/* enter_theorem_name() ................. #7
/* read_buffer_word().................... #8
/* sort_const_fxns()   .................. #9
/* is_fxn_a_const()    .................. #10
/* interchange_fxn_const()................#11
/* prepare_clauses()   .................. #12
/* get_pred_fxn_symbols().................#13
/* format_clause()         ...............#14
/* calculate_hashcodes()   ...............#15
/* calc_max_char_in_any_lit().............#16
/* count_lit_characters()  ...............#17
/* prove_by_iterative_deepening().........#18
/* write_out_results() .................. #19
/* add_clause().......................... #20
/*------------------------------------------------------------------*/
			 
#include <stdlib.h>
#include <strings.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <pvm3.h>

#include "constHS.h"
#include "clauseHS.h"

/*----------------------------------------------------------------*/
/* externally defined functions                                   */
/*----------------------------------------------------------------*/

extern void compile_clauses();          /* function in affc.c     */

extern void init_search_space();        /* functions in build.c   */
extern int build_tree();

extern void init_random_hash_array();   /* functions in hhash.c   */
extern int calc_lit_hc();
extern void calc_clause_hashcode();
extern int calc_alt_hashcodes();

extern clause *generate_clause();       /* functions in hgen.c    */
extern clause *copy_clause();

extern int binary_factors();            /* function in infer.c    */

extern int resolution_stage();          /* function in atom.c     */
extern int resolution_stage2();         /* function in atom.c     */

extern void print_clause();               /* function in hdump.c  */
extern void print0();

extern void equal_predicate();          /* functions in equal.c   */
extern void esa_for_fxns();
extern void esa_for_lits();

extern void rename_functions();        /* function in names.c     */

/*----------------------------------------------------------------*/
/* externally defined constants                                   */
/*----------------------------------------------------------------*/

extern int c_pick;

/*----------------------------------------------------------------*/
/* internally defined functions                                   */
/*----------------------------------------------------------------*/

int time_now();
void clear_fail_pair_info();
int init_options();
int grab_number();
void fatal();
int main();
void enter_theorem_name();
void read_buffer_word();
void sort_constants_fxns();
int is_fxn_a_const();
void interchange_fxn_const();
void prepare_clauses();
int get_pred_fxn_symbols();
void format_clause();
void calculate_hashcodes();
void calc_max_char_in_any_lit();
int count_lit_characters();
int prove_by_iterative_deepening();
void write_out_results();
void add_clause();

/*----------------------------------------------------------------*/
/* internally defined constants                                   */
/*----------------------------------------------------------------*/

pfnode pfspace[PF_SPACE_SIZE];          /* arrays used in affc.c  */ 
int kode[MAX_COMPILED_THM_SIZE];
char name_buff[MAX_CHAR];
cnode pred_ptr_list[PRED_PTR_SIZE];

int    clause_space[CSPACE_SIZE]; /* space for clist clauses      */
clause *clist[CSPACE_CLAUSES];    /* first few are base clauses   */

int    hb_space[HSPACE_SIZE];     /* space for atoms              */
clause *hb_list[HSPACE_CLAUSES];  /* pointers to atoms            */
clause *hb_list2[HSPACE_CLAUSES];  /* pointers to atoms            */

int  unit_space[USPACE_SIZE];      /* space for units             */
clause *unit_list[USPACE_CLAUSES];    /* pointers to units        */

rnode *rand_array;                /* arrary for random numbers    */
dnode  depth[HSPACE_CLAUSES];     /* depth information            */
int numb_rows;
int numb_cols;
int numb_entries;
int number_fxns, number_preds;
int rows_less_nfxns;

hnode *clause_hash_table;
unsigned int o_arraysize = 1048576;
int array_sizem1 = 1048575;

int fail[HSPACE_CLAUSES];
int fail_count[HSPACE_CLAUSES];
int last_fail;
int orig_branch_sign[HSPACE_CLAUSES];
int orig_branch_sign2[HSPACE_CLAUSES];

int dead_left_branches[HSPACE_CLAUSES];
int dead_right_branches[HSPACE_CLAUSES];
int heuristic_applied[HSPACE_CLAUSES];

int branch[HSPACE_CLAUSES];       /* current search path array    */
int branch2[HSPACE_CLAUSES];      /* current search path array    */
int fail_frequency[HSPACE_CLAUSES][2];
int hb_length[HSPACE_CLAUSES][2];

/*----------------------------------------------------------------*/
/* the following data structures save information on about the    */
/* two clauses that cause failure nodes and will be used to       */
/* construct a closed tree later.                                 */
/* fail_clause1,fail_clause1x are the clause number and literal   */
/* index of the last clause to fail at maximum construction depth */
/* fail_clause2,fail_clause2x are the clause number and literal   */
/* index of the second last clause to fail at maximum             */
/* construction depth.  these two clauses are brothers and        */
/* are resolved together when seeking a res-ref proof             */
/*                                                                */
/* fail_lits1 and fail_lits2 record the index of the atom         */
/* that caused each of the literals in the two clauses to fail    */
/*----------------------------------------------------------------*/

int fail_clause1,fail_clause1x,fail_clause1_ply,fail_clause1_node;
int fail_clause2,fail_clause2x,fail_clause2_ply,fail_clause2_node;
int fail_clause3,fail_clause3x,fail_clause3_ply,fail_clause3_node;
int fail_clause4,fail_clause4x,fail_clause4_ply,fail_clause4_node;

int fail_lits1[MAX_LIT_IN_CLAUSE];
int fail_lits2[MAX_LIT_IN_CLAUSE];
int fail_lits3[MAX_LIT_IN_CLAUSE];
int fail_lits4[MAX_LIT_IN_CLAUSE];

int fail_clause_left[200][18];

int build_start_time;
int build_time;

int base;                    /* number of base clauses            */
int base0;                   /* original number of base clauses   */
int base1;                   /* base clauses after prelim. stage  */
int nc;                      /* # clauses in negated conclusion   */

int iter_depth;                    /* current iteration depth     */
int ply;                           /* current construction depth  */

int nextc;                     /* location where next clause goes */
int nexthb;
int nextunit;

int last_atom_sign;                   /* = 0 if +lit, = 1 if -lit */
int open_problems;

int slash_char;

FILE *infile, *outfile;


char results[200000];
char *OutPtr;
   
char input_buffer[256]; /* buffer for input file name and options */ 

char *theorem_ptr, *theorem_out_ptr, *theorem_out_ptr2;
char theorem_name[256]; /* name of theorem read in                 */
char theorem_out_name[256];/* name of file where output is placed  */
char theorem_out_name2[256];/* new thm name if solved  */
int  theorem_out_name2end; /* ptr to last char in output thm*/
int tough, unknown, satisfiable;

char compiled_code_buffer[MAX_INPUT_THM_SIZE];
char *comp_code_buffer_index;  /* index on compiled_code_buf      */
int  compiled_code_size;
int  ww;  /* temp loc for reading chars from compiled_code_buf    */

int  *var_substitution_tbl[MAX_VARS*2]; /* var substitution table */
int  pstack[1000],*psp;            /* used in unify.c             */

int equal_pred;       /* =1 if "equal" or "Equal" is in theorem   */
int equal_identity;         /* the clause number of "Equal(x,x)"  */
int rename_fxns, rename_preds;

int const_alt; 
int max_var;
int max_lit;
int phase;        /* phase: elim = 0, 1st pass = 1, 2nd pass = 2  */
int number_preds;  /* number of different preds in input clauses  */
int number_fxns;   /* number of different fxns in input clauses   */
int number_constants; /* # of different constants in input cls    */
int number_const_nc;                /* number of constants in nc  */
int max_c_in_lit;              /* maximum characters in a literal */
int var_in_lit;
int first_attempt;

char *pidtbl[MAX_PRED_IDS];   /* ptrs to strings for predicates   */
char *fidtbl[MAX_FXNS_IDS];   /* ptrs to strings for functions    */
cnode *ctbl[MAX_BASE_CLAUSES];

/*----------------------------------------------------------------*/ 
/* The following are vars that record info on the construction.   */
/*----------------------------------------------------------------*/
 
int s_nod;                        /* # nodes on all iterations    */
int s_htentry, s_hthit, s_htfail;
int s_ltentry, s_lthit, s_ltfail;

/*----------------------------------------------------------------*/
/* The following variables record how many theorems have been
/* tried, how many have been proved, and how long it took.
/*----------------------------------------------------------------*/

int batch_tried;
int batch_solved;
int batch_start_time, batch_total_time;

/*----------------------------------------------------------*/
/* the following are variables that (in most cases)         */
/* the user can set as desired.  They control the search.   */
/*----------------------------------------------------------*/

int o_litmax;     /* max literals in a resolvent                  */
int o_charmax;    /* max chars in literal                         */
int o_scharmax;   /* max char in literal in construction phase    */
int o_maxtime;    /* max construction time in seconds             */
int x_charmax;    /* x_charmax = 0 => calc o_charmax              */
int x_debug;      /* user is supposed to be unaware of this optn  */
int o_resref;     /* tells HERBY to find res-ref proof            */
int o_fulltree;   /* constructs the full tree at each iteration   */
int o_interactive;/* tells HERBY to enter atoms interactively     */
int o_batch;
int o_tptp;       /* tells HERBY theorem is from TPTP Library     */
int o_phase0;     /* tells HERBY to perform prelim resol. phase   */
int x_complete_proof;

int mytid;
char buf[20000];
int myparent;
int this_computer;
int spawned;
int theorem_number;
int theorem_rating;

/*----------------------------------------------------------------*/
/* time_now()                    .............. #1
/*
/* called by:  main()                         herby.c
/*             prove_by_iterative_deepening() herby.c
/*             build_tree()                   build.c
/*             atom_selection_heuristics()    atom.c
/*             ash4()                         atom.c
/*
/* calls:      Nobody
/*
/* returns:    time of day as a int integer
/*---------------------------------------------------------------*/

int time_now()
{
return((int) time(0));
}

/*--------------------------------------------------------------*/
/* clear_fail_pair_info()                        .............. #2
/*
/* init statistical parameters gathered during the construction
/*
/* called from:    main()       herby.c
/*
/* calls:          nobody
/*
/* returns:        nil
/*---------------------------------------------------------------*/

void clear_fail_pair_info()
{
fail_clause1 = fail_clause1x = fail_clause2 = fail_clause2x = 0;
fail_clause3 = fail_clause3x = 0;
fail_clause4 = fail_clause4x = 0;
fail_clause1_ply = fail_clause1_node = 0;
fail_clause2_ply = fail_clause2_node = 0;
fail_clause3_ply = fail_clause3_node = 0;
fail_clause4_ply = fail_clause4_node = 0;
}
 
/*----------------------------------------------------------------*/
/* init_options()                          .............. #3
/*
/* reads the options that the user requests in the
/* command line.
/*
/* called from:  enter_theorem_name()    herby.c
/*
/* calls:        grab_number()           herby.c
/*
/* returns:      the number of chars in the name of the theorem
/*               or a -1 if help was requested
/*----------------------------------------------------------------*/
 
int init_options(argc,argv)
int argc;
char *argv[];
{
int i,j;

/*------------------------------------------*/
/* initialize all options to default values */
/*------------------------------------------*/

o_phase0    = 1;
o_fulltree  = 0;      /* normally do not construct full tree   */
o_resref    = 1;

o_litmax    = 0;
o_charmax   = 16;
x_charmax   = 0;
o_scharmax  = 0;

o_maxtime   = 3600;
o_interactive = 0;
x_debug     = 0;      /* 1 => turns debug info from pass2 on   */
o_tptp      = 0;      /* theorem assume to be in tgtp format   */

s_ltentry = s_lthit = s_ltfail = s_htentry = s_hthit = s_htfail = 0;

/*------------------------------------------*/
/* scan the arguments passed to the program */
/* if a "?" appears anywhere on the line    */
/* print out a "help" message               */
/*------------------------------------------*/

for(i = 76; i > 1 ; i--) {
   if(input_buffer[i - 2] == '?') {
        
/*--------------------------------------------------------------*/
/* a help message, triggered by a "?" anywhere on input line    */
/*--------------------------------------------------------------*/
      printf(
      "\n------------------------------------------------------");
      printf(
      "\n-----------------------------------------------------");
      printf("\n");
		
      for(i = 0; i < 256 ; i++) 
         input_buffer[i] = ' ';     /* clear input buffer  */
      return(-1);}                  /* -1 => re-enter      */
			
   if(input_buffer[i] == ' ') {
			
/*--------------------------------------------------------------*/
/* if rx is found, set o_resref = x                             */
/*--------------------------------------------------------------*/
						
      if(input_buffer[i + 1] == 'r') {
         o_resref = grab_number(i + 2);
         if(o_resref < 0 || o_resref > 1) {
            printf("\nbad argument\n");return(-1);}}

/*--------------------------------------------------------------*/
/* if fx is found, set o_fulltree = x                           */
/*--------------------------------------------------------------*/
						
      if(input_buffer[i + 1] == 'f') {
         o_fulltree = grab_number(i + 2);
         if(o_fulltree < 0 || o_fulltree > 1) {
            printf("\nbad argument\n");return(-1);}}}}

/*---------------------------------------------------*/
/* end of option list:                               */	
/* blank out input_buffer except for name of theorem */
/*---------------------------------------------------*/
	
for(i = 0; i < 255; i++) {
   if(input_buffer[i] == ' ' || input_buffer[i] == '\n') {
      j = i;
      for(j = i; j < 256 ; j++) {input_buffer[j] = ' ';}
      break;}}

return(i);         /* return the number of characters in the name  */
}

/*-----------------------------------------------------------------*/
/* grab_number(s)                       .............. #4
/*
/* reads a string of characters and converts them to a
/* integer equivalent
/*
/* called from:  init_options()        herby.c
/*
/* calls:        nobody
/*
/* returns:      an integer correspond to the string s
/*-----------------------------------------------------------------*/

int grab_number(s)
int s;
{
int n;
 	
for(n = 0; ; s++) {
   if(input_buffer[s] >= '0' && input_buffer[s] <= '9')
      n = n * 10 + ( (int) (input_buffer[s] - '0') );
   else {return(n);}}
}							

/*-----------------------------------------------------------------*/ 
/* fatal(msg)                     .............. #5
/*
/* fatal will print an error message on the screen and quit
/*
/* called by:    main()                  herby.c
/*               enter_theorem_name()    herby.c
/*               prepare_clause()        herby.c
/*               format_clause()         herby.c
/*               yyerror()               affc.c
/*               install_clist_clause()  atom.c
/*
/* calls:        nobody
/*
/* returns:      to the operating system after reporting
/*               error
/*-----------------------------------------------------------------*/
 
void fatal(msg)
char *msg;
{
   printf("....%s\n",msg);fflush(stdout);
   if(*msg == 'D') {
      printf("\nTheorems Tried %d\n Theorems Solved %d",
	 batch_tried,batch_solved);fflush(stdout);
      batch_total_time = time_now() - batch_start_time;
      printf("\nTotal Time (in seconds) %d\n\n",
	   batch_total_time);fflush(stdout);}
   exit(0);
}

/*-----------------------------------------------------------------*/
/* main()                             .............. #6
/*
/* The theorem prover enters here. We set the options specified
/* by the user, read in the clause system, prove the theorem,
/* and write the proof out.
/*
/* calls:   fatal()                        herby.c
/*          time_now()                     herby.c
/*          enter_theorem_name()           herby.c
/*          compile_clauses()              affc.c
/*          init_search_space()            build.c
/*          clear_fail_pair_info()         herby.c
/*          prepare_clauses()              herby.c
/*          init_random_hash_array()       hhash.c
/*          calculate_hashcodes()          herby.c
/*          resolution_stage()             herby.c
/*          prove_by_iterative_deepening() herby.c
/*          add_clause()                   herby.c
/*          write_out_results()            herby.c
/*-----------------------------------------------------------------*/

int main(argc,argv)
int argc;
char *argv[];
{
int zz;
char decision_char;   /* char to decide whether to do another thm  */
int proved, i, j, jj, k, m, mc, nc;
int resol_results;
clause *c;
literal *lit;
int next_clause;
int removed;

o_batch = 0;
batch_start_time = time_now();
batch_tried = 0;
batch_solved = 0;

if(removed = remove("/u0/prof/newborn/octoslave.core") == 0) {
   printf("Killed /u0/prof/newborn/octoslave.core\n");
   fflush(stdout);}
if(removed = remove("~/octoslave.core") == 0) {
   printf("Killed ~/octoslave.core\n");
   fflush(stdout);}
if(removed = remove("./octoslave.core") == 0) {
   printf("Killed ./octoslave.core\n");
   fflush(stdout);}

theorem_number = 0;

mytid = pvm_mytid();
myparent = pvm_parent();

gethostname(buf,64);
pvm_initsend(PvmDataDefault);
pvm_pkstr(buf);
pvm_send(myparent, CNAME);

this_computer = 0;
OutPtr = &results[0];

begin_proof:
o_phase0    = 1;
o_fulltree  = 0;      /* normally do not construct full tree   */
o_litmax    = 0;
o_charmax   = 16;
x_charmax   = 0;
o_scharmax  = 0;
o_interactive = 0;
x_debug     = 0;      /* 1 => turns debug info from pass2 on   */
s_ltentry = s_lthit = s_ltfail = s_htentry = s_hthit = s_htfail = 0;
o_resref = 1;
o_maxtime   = 3600;
phase       = 0;
iter_depth  = 0;
const_alt   = 0;
equal_pred = 0;
equal_identity = 0;
last_atom_sign = 0;
first_attempt = 0;
	
compile_clauses();

/*--------------------------------------------------*/
/* init arrays for phase 0, then fix up clauses     */
/*--------------------------------------------------*/

init_search_space();          /* init search space  */
prepare_clauses();

/*---------------------------------------------------*/
/* initialize the random_hash_array and then hash    */
/* all base clauses                                  */
/*---------------------------------------------------*/

init_random_hash_array(25,number_fxns);
calculate_hashcodes();
	
for(i = 0; i < HSPACE_CLAUSES; i++) {orig_branch_sign[i] = 0;}

base0 = base;
base1 = -1;
build_start_time = time_now(); /* start time         */ 
x_complete_proof = 0;

/*-----------------------------------------------------*/
/* resolution stage....form resolvents of base clauses */
/*-----------------------------------------------------*/

if(o_phase0 == 1) {
   if((proved = resolution_stage()) == 2) {
      goto phase3;}}

loop_here:
phase = 1;
ply = 0;
c_pick = 0;
clear_fail_pair_info();     /* zero our statistics variables  */

proved = prove_by_iterative_deepening(); /* prove the theorem */

if(x_debug > 2 ) {
   print0("\nUNIT LIST\n");
   for(i = 0 ; i < nextunit ; i++) {
      print0("\n");
      print_clause(unit_list[i]);}
   print0("\nEND OF UNIT LIST\n");}

if(proved != 1 || o_resref == 0) {
   ply = 0;
   goto phase3;}

phase = 2;
x_complete_proof = 1;

/*-------------------------------*/
/* when forming resolution-ref   */
/* proof, add next base clause   */
/*-------------------------------*/

add_clause();

if(clist[base]->c_nlit != 0) {
   if(resolution_stage2() == 2) goto phase3;
   nextunit = 0;

/*-------------------------------*/
/* eliminate dead atoms and      */
/* reset atom signs to orig val  */
/*-------------------------------*/

/*
   printf("FAIL ARRAY\n\n");
   for(i = 0; i <= nexthb; i++) {
      for(j = 0; j < 18; j++) {
         if(j == 0) {printf("\n");}
	 if(j == 1) {
	    printf("H%3d ",heuristic_applied[i]);}
         printf("%3d ",fail_clause_left[i][j]);}}
    printf("\n\n END OF FAIL ARRAY\n\n");
    */

/*
   printf("HBLIST BEFORE DELETE\n\n");
   for(i = 0; i < nexthb; i++) {
      print_clause(hb_list[i]);
   printf("\n");}
*/

/*
   printf("INITIALIZE BRANCH SIGNS\n\n");
   for(i = 1; i <= nexthb; i++) {
      printf("%d",branch[i]);}
   printf("\n");
*/

   for(i = 0; i < nexthb; i++) {
      c = hb_list[i]; 
      if(branch[i + 1] == 1) {
         c->c_lit->l_id = c->c_lit->l_id ^ NEGMASK;}
      calc_lit_hc(c->c_lit);
      calc_clause_hashcode(c);}

   for(m = 0; m <= nexthb; m++) {
      hb_length[m][0] = 0;
      fail_frequency[m][0] = m + 1;
      fail_frequency[m][1] = 0;}

// printf("DETERMINE HB FAIL FREQUENCIES\n\n");
   for(m = nexthb - 1; m >= 0; m--) {
      mc = 0;
      for(j = 1; j < 18; j++) {
         if(fail_clause_left[nexthb][j] == m + 1) {
            fail_frequency[m][1] += 1;
            mc = 1;}}
      for(i = m - 1; i < nexthb; i++) {
         for(j = 1; j < 18; j++) {
            if(fail_clause_left[i][j] == m + 1) {
               fail_frequency[m][1] += 1;
               mc = 1;}}}}

// printf("SORTED LENGTHS\n\n");
   for(i = 0; i <= nexthb; i++) {
      hb_length[i][0] = i + 1;
      hb_length[i][1] = (((int) hb_list[i]->c_end - (int) hb_list[i]) - 64)/4;}

   if(x_debug > 0) {
      for(i = 0; i < nexthb; i++) {
         printf("H %d L %d\n",hb_length[i][0],hb_length[i][1]);}}

   if(x_debug > 0) {
      printf("\nORIGINAL UNSORTED FAIL FREQUENCIES\n"); 
      for(i = 0; i < nexthb; i++) {
         printf("H %d: %d\n",
	      fail_frequency[i][0],fail_frequency[i][1]);}}


   for(i = 0; i < nexthb - 1; i++) {
      mc = fail_frequency[i][1];
      jj = 0;
      for(j = i + 1; j < nexthb; j++) {
         nc = fail_frequency[j][1];
         if(mc < nc) {
            mc = nc;
            jj = j;}}
      if(jj != 0) {
	 mc = fail_frequency[jj][1];
	 nc = fail_frequency[jj][0];

         for(m = jj; m > i; m--) {
         fail_frequency[m][1] = fail_frequency[m - 1][1];
         fail_frequency[m][0] = fail_frequency[m - 1][0];}
	 fail_frequency[i][1] = mc;
	 fail_frequency[i][0] = nc;}}
	    
  for(i = 0; i < nexthb; i++) {
     hb_list2[i] = hb_list[i];
//   printf("%d: %d\n",fail_frequency[i][0],fail_frequency[i][1]);
     }
// printf("\nFINAL aHB LIST....\n\n");

   for(i = 0; i < nexthb; i++) {
      for(j = i + 1; j < nexthb; j++) {
         if(hb_length[i][1] > hb_length[j][1]) {
	    mc = hb_length[i][1];
	    nc = hb_length[i][0];
	    hb_length[i][1] = hb_length[j][1];
	    hb_length[j][1] = mc;
	    hb_length[i][0] = hb_length[j][0];
	    hb_length[j][0] = nc;}}}

   j = 0;
   jj = mc = nexthb;

   for(i = 0; i < nexthb; i++) {
      if(fail_frequency[i][1] < 1) {mc--;continue;}
      hb_list[j] = hb_list2[fail_frequency[i][0] - 1];
      hb_list[j]->c_id = j + 1;
      j++;}

   for(k = j; k < nexthb; k++) {
      hb_list[k] = 0;}

   nexthb = j;

   next_clause = hb_list[0]->c_end;

   for(i = 1; i < nexthb; i++) {
     if(hb_list[i] == 0) continue;
     if(hb_list[i]->c_end > next_clause) {next_clause = hb_list[i]->c_end;}}
   hb_list[nexthb] = next_clause;

   if(x_debug > 0) {
      printf("\nFINAL HB LIST....\n\n");
      for(i = 0; i < nexthb; i++) {
         print_clause(hb_list[i]);printf("\n");}
      printf("\nEND OF HB LIST\n");} 

   for(i = 0; i <= jj; i++) {
      for(j = 0; j < 18; j++) {
         fail_clause_left[i][j] = 0;}}
   goto loop_here;}

phase3: 
phase = 3;
write_out_results(proved);
	
/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/

free(rand_array);
fclose(infile);
fclose(outfile);

/*----------------------------------------------*/
/* prove another theorem?                       */
/*----------------------------------------------*/

another_theorem:

i = 1;
while(i < 2) {i = i;}
}

/*-------------------------------------------------------------*/
/* enter_theorem_name()           ................#7
/*
/* read in the name of the theore to be proved
/*
/* called by:   main()                 herby.c
/*
/* calls:       init_options()         herby.c
/*              fatal()                herby.c
/*
/* returns:     nil
/*--------------------------------------------------------------*/

void enter_theorem_name(argc,argv)
int argc;
char *argv[];
{
int number_chars;      /* number of characters in name of theorem  */ 
int bad_name;        /* bad_name = 0 until theorem found in file   */
int j, jj, jjj, k, m;
int ask_for_help;     /* normally 0, set to 1 if "?" on input line */
char zz;

bad_name    = 0;

while(bad_name == 0) {
   for(j = 0 ; j < 256 ; j++) {
      theorem_name[j] = 0;
      theorem_out_name[j] = theorem_out_name2[j] = 0;
      input_buffer[j] = ' ';}
   ask_for_help = 0;
   while(ask_for_help == 0) {
      j = 0;
      printf("\nEnter name of theorem (type '?' for help): ");
      while((zz = getchar()) != '\n') {
         if(zz == '\b') {
            if(j > 0) {
               j--; printf("\b \b");}}
         else if(zz == EOF) {
	    fatal("DONE!");}
         else {
            input_buffer[j++] = zz;}}
      input_buffer[j++] = ' ';

      for(jj = 1; jj < argc; jj++) {
         ++argv;
         jjj = 0;
         while((input_buffer[j] = (*argv)[jjj++]) != '\0') {
	    j++;}
         input_buffer[j++] = ' ';}

      number_chars = init_options(argc,argv); /* set/clear options */

      printf("\n");
      for(jj = 0; jj < number_chars; jj++) {
         printf("%c",input_buffer[jj]);}
      printf("\n");

      if(number_chars != -1) {ask_for_help++;printf("\n");}}         

   m = 0;
   for(j = 0 ; j < number_chars ; j++)  { 
      if(input_buffer[j] == '.') {m = j;}
      if(input_buffer[j] == '/') {m = 0;}
      theorem_name[j] = input_buffer[j];}
   for(j = number_chars ; j < 256 ; j++ ) {
      theorem_name[j] = 0;}
    
/*--------------------------------------------------------------*/
/* give output file the same as input file with a suffix ".h"   */    
/*--------------------------------------------------------------*/
    
   if( m != 0) {
      for(j = 0 ; j < number_chars ; j++)  {
         theorem_out_name[j] = input_buffer[j];
	 theorem_out_name2[j] = input_buffer[j];
         if(m == j) {
            theorem_out_name[j+1] = 'h';
            theorem_out_name2[j+1] = 'H';
            break;}}}
   else {
      for(j = 0 ; j < number_chars + 1; j++) {
         theorem_out_name[j] = input_buffer[j];
         theorem_out_name2[j] = input_buffer[j];
         if(input_buffer[j] == ' ') {
            theorem_out_name[j] = '.';
            theorem_out_name2[j] = '.';
            theorem_out_name[j+1] = 'h';
            theorem_out_name2[j+1] = 'H';
            break;}}}
			
   theorem_ptr = &theorem_name[0];
   theorem_out_ptr = &theorem_out_name[0];
   theorem_out_ptr2 = &theorem_out_name2[0];

   if((infile = fopen(theorem_ptr,"r")) == 0) {
      printf(
      "Cannot open theorem. You may have typed the wrong name.\n\n");
      if(argc > 1) fatal(" ");}
   else bad_name = 1;}

if((outfile = fopen(theorem_out_ptr,"w")) == 0) {
   fclose(outfile);
   fatal("\nCannot open output file:\nType cr and rerun");}

OutPtr = &results[0];

theorem_out_name2end = j + 2;
		
for(m = number_chars - 1; m >= 0; m--) {
   if(input_buffer[m] == '/') {break;}
   if(!m) {m = -1;}}
slash_char = m;

OutPtr += sprintf(OutPtr,"Theorem:");
for(j = 0 ; j < number_chars; j++) {
   OutPtr += sprintf(OutPtr,"%c",theorem_name[j]);}
OutPtr += sprintf(OutPtr,"\n\n");
}
 
/*--------------------------------------------------------------*/
/* read_buffer_word()               .............. #8
/*
/* read_buffer_word will read a word (integer) from the input file
/*
/* called by:    prepare_clauses()     herby.c
/*               format_clause()       herby.c
/*
/* calls:        nobody
/*
/* returns:      nil
/*----------------------------------------------------------------*/

void read_buffer_word()
{
union u_tag {
   int w; 
   char ttext[4];} 
   temp;
			
int j;
	
for(j = 0; j < 4; j++) {
   temp.ttext[j] = *comp_code_buffer_index++;}
ww = temp.w;
}

/*--------------------------------------------------------------*/
/*      is_fxn_a_constant()                .............. #13
/*
/*      called by:      sort_constants_fxns()       tgtp.c
/*
/*      calls:          nobody
/*
/*      returns:        1 if fxn is a constant, 0 otherwise
/*-----------------------------------------------------------*/
 
int is_fxn_a_constant(i,top_clause)
int i,top_clause;
{
int k, a, *p;
clause *c;
literal *l;
	
for(k = top_clause; k <= base ; k++) { 
   c = clist[k];
   for(l = c->c_lit ; l != 0 ; l = l->l_next) {  /* scan lits  */
      p = &l->l_id;
      for(a = (*p++ & ARGMASK) ; a > 0 ; a--) {
	     if((*p & FCNMASK) != 0) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           if((*p & ARGMASK) == 0) return(1);
	           else return(0);}
	        a += *p++ & ARGMASK;}
	     else {p++;}}}}
return(0);
}


/*--------------------------------------------------------------*/
/*      sort_constants_fxns()           .............. #9
/*
/*      called by:   prepare_clauses()          herby.c
/*
/*      calls:       is_fxn_a_constant()        herby.c
/*                   interchange_fxn_const()    herby.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/

void sort_constants_fxns()
{
int i, k, a, *p;
char *fide, *ii;
clause *c;
literal *l;

number_constants = 0;

for(i = 1; i <= number_fxns ; i++) {
   if(is_fxn_a_constant(i,0)) {
      number_constants++;	
      if(i >= number_fxns) break;}
   else {
      for(k = number_fxns; k > i; k--) {
         if(is_fxn_a_constant(k,0)) {
            interchange_fxn_const(i,k);
            number_constants++;
            break;}}}}  

if(number_constants == 0) {
   number_constants++;
   number_fxns++;
   ii = input_buffer;
   *ii = 'a'; ii++;*ii = 0;
   fidtbl[number_fxns] = malloc(strlen(input_buffer) + 1);
   strcpy(fidtbl[number_fxns],input_buffer);
   fide = fidtbl[number_fxns];
   fidtbl[number_fxns] = fidtbl[1];
   fidtbl[1] = fide;
   for(k = 0; k <= base ; k++) {
      c = clist[k];
      for(l = c->c_lit; l != 0; l = l->l_next) {
         p = &l->l_id;
         for(a = (*p++ & ARGMASK) ; a > 0; a--) {
            if((*p & FCNMASK) != 0) {
               *p = *p + 32;
               a += *p++ & ARGMASK;}
            else {p++;}}}}}
 
/*-----------------------------------------------*/
/*   now resort the constants in the nc to the   */
/*   top of the list.                            */
/*-----------------------------------------------*/

number_const_nc = 0;
for(i = 1; i <= number_constants; i++) {
   if(is_fxn_a_constant(i,base - nc + 1)) { 
      number_const_nc++; 
      continue;}
   else {
      for(k =  i + 1; k <= number_constants; k++) {
         if(is_fxn_a_constant(k,base - nc + 1)) {
            number_const_nc++;
            interchange_fxn_const(i,k);
            break;}}}} 

for(i = 1; i <= number_fxns; i++) {
   printf(" %s",fidtbl[i]);
   OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
   if(i == number_constants) {print0(" :");}
   if(i == number_const_nc) {print0(" .");}}
print0("\n");
}

/*--------------------------------------------------------------*/
/* interchange_fxn_const()          ............. #11
/*
/* called by:      sort_constants_fxns()      herby.c
/*
/* calls:          nobody
/*
/* returns:        nil
/*--------------------------------------------------------------*/

void interchange_fxn_const(i,j)
int i,j;
{
int k, *p, a;
clause *c;
literal *l;
char *fide;
	
for(k = 0; k <= base ; k++) {
   c = clist[k];
   for(l = c->c_lit ; l != 0 ; l = l->l_next) {
      p = &l->l_id;
      for(a = (*p++ & ARGMASK) ; a > 0 ; a--) {
         if((*p & FCNMASK) != 0) {
            if(((*p & IDMASK) >> ARGSHIFT) == i) {
               *p = (*p & ~IDMASK) + (j << ARGSHIFT);}
            else if(((*p & IDMASK) >> ARGSHIFT) == j) {
               *p = (*p & ~IDMASK) + (i << ARGSHIFT);}
            a += *p++ &ARGMASK;}
         else {p++;}}}}
			
fide = fidtbl[i];
fidtbl[i] = fidtbl[j];
fidtbl[j] = fide;
}

/*--------------------------------------------------------------*/
/* prepare_clauses()                .............. #12
/*
/* prepare_clauses will read the clauses currently stored
/* in the input_buffer and format them, placing the resulting
/* clauses in cspace with pointers to each in the clist
/*
/* NOTES:
/* - we assume that all variables are numbered from 0..n so
/*   that n+1 is the number of distinct variables
/*
/* - we pass the predicate and function symbol tables on to
/*   the output directly since we have no need for them
/*
/* called from:    main()                          herby.c
/*
/* calls:          get_pred_fxn_symbols()          herby.c
/*                 read_buffer_word()              herby.c
/*                 format_clause()                 herby.c
/*                 calc_max_char_in_any_lit()      herby.c
/*                 sort_constants_fxns()           herby.c
/*                 equal_predicate()               equal.c
/*                 esa_for_fxns()                  equal.c
/*                 esa_for_lits()                  equal.c
/*                 fatal()                         herby.c
/* 
/* returns:        nil
/*-------------------------------------------------------------*/

void prepare_clauses()
{
static int characters[] =
         {0,10,12,15,20,25,35,35,35,35,35,35};
int n, i;

/*--------------------------------------------------*/
/* set pointer to beginning of compiled_code_buf    */
/*--------------------------------------------------*/

comp_code_buffer_index = compiled_code_buffer;

/*--------------------------------------------------------------*/
/* We first pass the symbol table on to the output file.        */
/* This symbol table is two sets of non-empty lines followed    */
/* by an empty line (the predicate and function symbol tables   */
/*--------------------------------------------------------------*/
    
OutPtr += sprintf(OutPtr,"Predicates: "); 
number_preds = get_pred_fxn_symbols(pidtbl); /* pred sym table   */
sprintf(OutPtr++,"\n");

OutPtr += sprintf(OutPtr,"\nFunctions: ");
number_fxns = get_pred_fxn_symbols(fidtbl);/* function sym table */

read_buffer_word(); 

n = ww; 
base = ww - 1;              /* clist[base] is last base clause  */

read_buffer_word();nc = ww;   /* # clauses in negated conclusion */
	
if(n < 2) fatal("not enough clauses");	

/*--------------------------------------------*/
/* read the clauses and put them in the clist */
/*--------------------------------------------*/
  
while(n-- > 0) {format_clause();}

sort_constants_fxns();
calc_max_char_in_any_lit();

if(x_charmax == 0) {
   if(max_c_in_lit <= 11) o_scharmax = characters[max_c_in_lit];
   else o_scharmax = 40;}
	
o_charmax = o_scharmax;	

if(x_charmax != 0 && o_scharmax < max_c_in_lit) {
   o_scharmax = max_c_in_lit;}

/*--------------------------------------------------*/
/* determine the max number of literals in any      */
/* retained clause and the maximum number of        */
/* different variable names in any retained literal */
/*--------------------------------------------------*/

max_lit = 0;
max_var = 0;
for(n = 0 ; n <= base ; n++) {
   if(clist[n]->c_nlit > max_lit) max_lit = clist[n]->c_nlit;
   if(clist[n]->c_nv > max_var)  max_var = clist[n]->c_nv;}

if(o_litmax == 0) {o_litmax = max_lit - 1;}
max_lit = 2 * max_lit;

rename_functions();

if(rename_preds) {

/*----------------------------------------*/
/* if predicates have been renamed, list  */
/*----------------------------------------*/

   OutPtr += sprintf(OutPtr,"Renamed predicates: ");
   for(i = 1; i <= number_preds; i++) {
      OutPtr += sprintf(OutPtr," %s",pidtbl[i]);}
   OutPtr += sprintf(OutPtr,"\n");}
  
if(rename_fxns) {

/*----------------------------------------*/
/* if functions have been renamed, list   */
/*----------------------------------------*/

   OutPtr += sprintf(OutPtr,"Renamed functions: ");
   for(i = 1; i <= number_fxns; i++) {
      OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
      if(i == number_constants) {OutPtr += sprintf(OutPtr," :");}
      if(i == number_const_nc) {OutPtr += sprintf(OutPtr," .");}}
   OutPtr += sprintf(OutPtr,"\n");}

equal_predicate();
esa_for_fxns();
esa_for_lits();

}
 

/*--------------------------------------------------------------*/ 
/*      get_pred_fxn_symbols()        .............. #13
/*
/*      get_pred_fxn_symbols will read in a symbol table and write
/*      it to the output file (and to the log file if there is one
/*
/*      called from:   prepare_clauses()          tgtp.c
/*
/*      calls:         nobody
/*
/*      returns:       the number of symbols (pred/fxns)
/*---------------------------------------------------------------*/
 
int get_pred_fxn_symbols(qtbl)
char *qtbl[];
{
int a, k;
char *ii, *jj;
static char Eq[] = "Equal";
static char eq[] = "equal";
		
for(a = 1; ;a++) {  

/*----------------------------------------*/
/* check first if at least one pred, fxn  */           
/*----------------------------------------*/

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++; 
      return(0);}

/*----------------------------------------*/
/* set qtbl[a] to point to beginning of   */
/* the a-th predicate or function         */
/*----------------------------------------*/

   qtbl[a] = ii = comp_code_buffer_index;

   while(*comp_code_buffer_index != 0xa) { 
      comp_code_buffer_index++;}
 
/*----------------------------------------*/
/* replace terminating '0xa' with a 0     */
/*----------------------------------------*/
  
   *comp_code_buffer_index++ = 0;
 
/*--------------------------------------------------*/
/* print string of characters representing pred     */    
/*--------------------------------------------------*/
    
if(qtbl == pidtbl) {
   OutPtr += sprintf(OutPtr,"%s ",qtbl[a]);}
 
/*--------------------------------------------------*/
/* test for equality predicate                      */
/*--------------------------------------------------*/

   jj = ii;
	
   for(k = 0; k < 6 ; k++) {if(*jj++ != Eq[k]) {k = 0; break;}}
   if(k != 0) {equal_pred =  (a << ARGSHIFT) + 2;}

   jj = ii;
   for(k = 0; k < 6 ; k++) {if(*jj++ != eq[k]) {k = 0; break;}}
   if(k != 0) {equal_pred =  (a << ARGSHIFT) + 2;}
    
   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++;
      return(a);}} 
}							
 
/*--------------------------------------------------------------*/
/* format_clause()                   .............. #14
/*
/* format_clause will read clause from the input and stuff
/* it into the clist
/*
/* called from:    prepare_clauses()   herby.c
/*
/* calls:          read_buffer_word()  herby.c
/*                 fatal()             herby.c
/* 
/* returns:        nil
/*--------------------------------------------------------------*/

void format_clause()
{
clause *c;
literal *previous_literal;
int *p,a,i;
int var_detected[32];
	
ply = 0;	
for(i = 0; i < 32 ; i++) {var_detected[i] = 0;}

read_buffer_word();read_buffer_word();/* these are leading 0's */
 
/*------------------------------------------------------*/
/* initialize clause header (this is not part of the    */
/* input file, we add it as we read in the clause)      */
/*------------------------------------------------------*/
 
c = clist[nextc++];                  /* the new clause         */
c->c_id = nextc;
c->c_bits =  BASE;                 /* denotes base clause      */
if(nextc > base - nc +1) {c->c_bits |= NC;} 
c->c_rpar = c->c_lpar = 0;      /* no parents (base clause)    */
c->c_nlit = 0;                             /* # literals       */
c->c_nv   = 0;                      /* # distinct variables    */
	
p = (int *) (c + 1);            /* where predicate code goes   */
previous_literal = 0;                /* literal chain          */
 
/*----------------------------------------*/
/* read in each literal, end clause at -1 */
/*----------------------------------------*/	

read_buffer_word(); 
while(ww != -1) {
   if(previous_literal == 0) {c->c_lit = (literal *) p;}
   else {previous_literal->l_next = (literal *) p;}
      
   previous_literal = (literal *) p;
   previous_literal->l_id = ww;
   previous_literal->l_next = 0;
 
   p = (int *) (previous_literal + 1);
 
   for(a = ww & ARGMASK; a > 0; a--) {
      read_buffer_word();
      if(ww & FCNMASK) {                          /* function? */
         *p++ = ww;                            /* yes       */
         a += ww & ARGMASK;}                   /* end if    */ 
      else {
         ww >>= ARGSHIFT;                      /* var ID    */
         *p++ = ww;
         if(var_detected[ww] == 0) {
            var_detected[ww]++;
            c->c_nv++;}}}
   c->c_nlit++;
   read_buffer_word();}                       /* end while  */
 
c->c_end = (clause *) p;
clist[nextc] = c->c_end;
	
/*--------------------------------------------------*/
/* if clause has only one literal, set UNIT flag on */
/*--------------------------------------------------*/

if(c->c_nlit == 1) {c->c_bits |= UNIT;}
if(c->c_nv > MAX_VARS) {fatal("Too many variables in clause");}
if(c->c_nlit == 0) fatal("Empty clause in base");
}

/*---------------------------------------------------------------*/
/* calculate_hashcodes()            .............. #15
/*
/* called by:   main()                      herby.c
/*
/* calls:       calc_lit_hc()               hhash.c
/*              calc_clause_hashcode()      hhash.c
/*
/* returns:     nil
/*--------------------------------------------------------------*/
 
void calculate_hashcodes()
{
clause *c;
literal *l;
int i, l_cnt, o_charmax_tmp;
	
o_charmax_tmp = o_charmax;
o_charmax = 200;

for(i = 0 ; i <= base ; i++) {
   c = clist[i];
   for(l = c->c_lit ; l != 0 ; l = l->l_next) {l_cnt = calc_lit_hc(l);}
   calc_clause_hashcode(c);
   if(c->c_nlit == 1) {
      add_clause_to_ht(c->c_X,c->c_Y,c->c_id,1);
      if(c->c_nv > 0) {
         calc_alt_hashcodes(c->c_lit,c);}}}

o_charmax = o_charmax_tmp;
return;
}	

/*--------------------------------------------------------------*/
/* calc_max_char_in_any_lit()           .............. #16
/*
/* calculates the maximum number of characters in any literal
/* o_schartmax records this maximum number
/*
/* called by:      prepare_clauses()         herby.c
/*
/* calls:          count_lit_characters()    herby.c
/*
/* returns:        nil
/*--------------------------------------------------------------*/

void calc_max_char_in_any_lit()
{
clause *c;
literal *lit;
int i, lit_char_cnt;

max_c_in_lit = 0;

for(i = 0; i <= base; i++) {
   c = clist[i];
   for(lit = c->c_lit ; lit != 0 ; lit = lit->l_next) {
      lit_char_cnt = count_lit_characters(lit);
      if(lit_char_cnt > max_c_in_lit) {max_c_in_lit = lit_char_cnt;}}}
return;
}
			
/*---------------------------------------------------------------*/
/* count_lit_characters()          .............. #17
/*
/* count_lit_characters counts chars in lit l
/*
/* called by:    calc_max_char_in_any_lit()        herby.c
/*
/* calls:        nobody
/*
/* returns:      number of char's in literal l
/*----------------------------------------------------------------*/
 
int count_lit_characters(l)
literal *l;                    /* l is a literal in clause c      */
{
int a,*p;
int literal_chars;         /* count of the number of chars in lit */
	
p = &l->l_id;        
literal_chars = 0;     
	
for(a = *p++ & ARGMASK; a > 0; a--) {
   if((*p & FCNMASK) != 0) { 
      literal_chars++;  /* function; inc count; det. # arguments   */
      a += *p++ & ARGMASK;}
   else {		
      literal_chars++;    /* variable; inc count                   */
      p++;}}

return(literal_chars);
}

/*----------------------------------------------------------------*/
/* prove_by_iterative_deepening()      .............. #18
/*
/* prove will actually perform the iteratively deepening
/* depth-first construction.  We return the depth at which we stop
/*
/* called by:      main()                     herby.c
/*
/* calls:          build_tree()               build.c
/*                 time_now()                 herby.c
/*
/* returns:        integer indicating the result of search
/*---------------------------------------------------------------*/

int prove_by_iterative_deepening()
{
int s;              /* normally s=0; if proof found, s = 1       */

//OutPtr += sprintf(OutPtr,"\n");

for(iter_depth = 1; iter_depth < MAX_PROOF_LENGTH ;iter_depth++) {

/*--------------------------*/
/* set node count to zero   */
/*--------------------------*/

   s_nod = 0;

/*--------------------------*/
/* Begin construction!!!!   */
/*--------------------------*/

   s = build_tree();	

/*----------------------------------------*/
/* save the current time                  */
/*----------------------------------------*/

   build_time = time_now();
		
/*----------------------------------------*/	
/* print time, nodes, open pbs, atoms     */
/*----------------------------------------*/

/*
   OutPtr += sprintf(OutPtr,
                "\nTIME:%d ",build_time - build_start_time);
   OutPtr += sprintf(OutPtr,
                "(NODES: %d, OPEN PROBLEMS: %d, ATOMS: %d)\n",
                s_nod+1,open_problems,iter_depth);
*/

/*------------------------------*/
/* if proof found, print *, ret */
/*------------------------------*/

   if(build_time - build_start_time > o_maxtime && phase == 1) {
      printf("TOO MUCH TIME -- NO PROOF OBTAINED\n");fflush(stdout);
      return(0);}
							
   if(s == 2) {
      printf("NO ATOM PICKED BY ASHs -- NO PROOF OBTAINED\n");
      fflush(stdout);
      return(0);}

   if(s == 4) {
      printf("NO SPACE ON CLAUSE LIST -- NO PROOF OBTAINED\n");
      fflush(stdout);
      return(0);}

   if(s == 0) {
      printf("TOO MANY ATOMS -- NO PROOF OPTAINED\n");
      fflush(stdout);
      return(0);}

   if(s == 1)  {
//    OutPtr += sprintf(OutPtr,"** Proof Found! **\n"); 
//    printf("** Proof Found! **\n");fflush(stdout); 
      return(1);}}
               
/*----------------------*/		
/* return with no proof */
/*----------------------*/

return(0);
}
 
/*------------------------------------------------------------*/
/* write_out_results()         .............. #19
/*
/* write_out_results writes out the cl's that form the proof
/* (if a proof found) and write statistics out. it writes
/* the results on the screen and to a file "in-theorem.u".
/*
/* called by:      main()          herby.c
/*
/* calls:          print_clause()  hdump.c
/*
/* returns:        nil
/*---------------------------------------------------------------*/
 
void write_out_results(proof_found)
int proof_found;
{
int i;		
int tbt ;     /* total build time    */
int t_temp, t1_temp, t_power;
char t1_char;
int leading_zeros;
clause *c, *ci;
char *OutPtr2;

/*--------------------------------------------------*/
/* set screen printing to print to a depth of 1000. */
/*--------------------------------------------------*/

x_debug = 0; 
	
if(o_resref == 1) {

/*----------------------------------------------------------------*/
/* determine which of the inferred clauses were used in the proof.*/
/* fix c->c_bits by adding LIST bit for inferred clauses that     */
/* should be printed. for each clause in the proof found in phases*/
/* 1,2, check LIST bit of parents                                 */
/*----------------------------------------------------------------*/

   for(i = 1; i <= ply; i++) {
      c = clist[depth[i].d_tc];
      if(c->c_lpar) {
         ci = c->c_lpar;
         ci->c_bits |= LIST;} /* check lft par of clist[depth[i].d_tc*/
      if(c->c_rpar)  {            /* check if c is a factor      */
         ci = c->c_rpar;
         ci->c_bits |= LIST;}}/* check rt par of clist[depth[i].d_tc */

/*-------------------------------------------------------------*/
/* now check LIST bit of other inferred cl's to det parents  */
/*-------------------------------------------------------------*/

   for(i = base ; i > base0 ; i--)  {
      c = clist[i];
      if(c->c_nlit == 0) c->c_bits |= LIST; /* null cls is inc */
      if((c->c_bits & LIST) == LIST) {
         if(c->c_lpar) {
	    ci = c->c_lpar;
            ci->c_bits |= LIST;}
         if(c->c_rpar) {
	    ci = c->c_rpar;
            ci->c_bits |= LIST;}}}}

/*----------------------------------*/
/* write out the base clauses       */
/*----------------------------------*/

if(proof_found == 1) {
  OutPtr += sprintf(OutPtr,"THE PROOF\n\n");}
	
OutPtr += sprintf(OutPtr,"\n\nGiven axioms:\n");
	
/*--------------------------------------------------------------*/
/* write out the original clauses and the used inferred clauses */
/*--------------------------------------------------------------*/

for(i = 0; i <= base0; i++) {
   if(i <= base0) {
      print_clause(clist[i]);
      OutPtr += sprintf(OutPtr,"\n");}		
   if(i == (base0 - nc))  {
      OutPtr += sprintf(OutPtr,"\nNegated conclusion:\n");}}

/*---------------------------------------------------------------*/
/* write out atoms used to construct closed semantic tree        */
/*---------------------------------------------------------------*/

/*
if(proof_found) {
   OutPtr += sprintf(OutPtr,"\nAtoms used to construct closed semantic tree:\n");}
else {OutPtr += sprintf(OutPtr,"\nNo proof found.  Atoms used:\n");}

for(i = 0; i < nexthb ; i++) {
   print_clause(hb_list[i]); 
   OutPtr += sprintf(OutPtr,"\n");
   printf("\n");}
*/

/*----------------------------------------------------------------*/
/* write out resolution-refutation proof if one was found         */
/*----------------------------------------------------------------*/

if((base0 != base && x_debug > 2) || o_resref == 1) {
   OutPtr += sprintf(OutPtr,"\nResolution clauses:\n");
   for(i = base0 + 1 ; i <= base ; i++) {
      print_clause(clist[i]);
      OutPtr += sprintf(OutPtr,"\n");}
   OutPtr += sprintf(OutPtr,"\nEnd of proof\n\n");
   fflush(stdout);}
/*
if(o_resref == 1) {
   printf("\nProof:\n");fflush(stdout);
   OutPtr += sprintf(OutPtr,"\nProof:\n");
   for(i = base1 + 1; i <= base ; i++) {
      print_clause(clist[i]);
      if(clist[i]->c_nlit == 0) {break;}
      printf("\n");
      fflush(stdout);
      OutPtr += sprintf(OutPtr,"\n");}
   printf("END OF PROOF\n\n");
   fflush(stdout);
   OutPtr += sprintf(OutPtr,"END OF PROOF\n\n");}
*/

//printf("Now output theorem's results = %d\n",OutPtr - &results[0]);

OutPtr2 = &results[0];

//printf("Done output theorem's results\n");

tbt = build_time - build_start_time;	
  	
leading_zeros = 0;
for(t_power = 100000; ; t_power = t_power/10) {
   if(tbt > t_power) {
      t1_temp = tbt/t_power;
      t1_char = t1_temp + 48;
      theorem_out_name2[theorem_out_name2end++] = t1_char;
      leading_zeros = 1;
      tbt = tbt - t1_temp*t_power;}
   else {
      if(leading_zeros == 1 || t_power == 1) {
	 t1_char = 48;
	 theorem_out_name2[theorem_out_name2end++] = t1_char;
	 leading_zeros = 1;}}
   if(t_power == 1) {break;}}

if(proof_found > 0) {
   theorem_out_name2[theorem_out_name2end++] = 'L';
   theorem_out_name2[theorem_out_name2end++] = 'H';
   i = rename(theorem_out_ptr,theorem_out_ptr2);}

/*----------------------------------*/
/* print statistics on the screen   */
/*----------------------------------*/    
   	
//printf("\nTotal Time: %d sec\n",tbt);
//printf("BASE CLAUSES: %d   REVISED SET OF BASE CLAUSES: %d ",
//   base0 + 1, base1 + 1);
//printf("\nNodes: %d  ATOMS: %d\n",s_nod+1,nexthb);

/*--------------------------------------------------*/        
/* output to theorem.p (same as output to screen)   */
/*--------------------------------------------------*/
        
OutPtr += sprintf(OutPtr,"\nTotal Time: %d sec\n",tbt);

printf("C%d sending proof of theorem\n",this_computer);fflush(stdout);
OutPtr += sprintf(OutPtr,"$$");

results[0] = 'b';
results[1] = '0' + (this_computer/10);
results[2] = '0' + (this_computer % 10);
results[3] = theorem_number;
for(i = 0; i < 25; i++) {
   results[5 + i] = theorem_out_name2[i];}

pvm_initsend(PvmDataDefault);
pvm_pkbyte(results,200000,1);
pvm_send(myparent, MPROOF);

}


/*-----------------------------------------------------------------*/
/* add_clause()        ...........................#20
/*
/* adds a clause to the set of base clauses as dictated
/* by the semantic tree.  in particular, if two clauses,
/* c1 and c2 caused failures at the children of some node,
/* then their resolvent will cause that node to fail if
/* the resolvent is added to the base clauses.
/*
/* it should be noted that while several clauses may be
/* added to the set of base clauses, all but one of them
/* are made DEAD clauses and are not used in the construction
/* of future semantic trees.
/*
/* called by:    main()                       herby.c
/*
/* calls:        generate_clause()            hgen.c
/*               unify()                      unify.c
/*               standardize()                infer.c
/*               copy_clause()                hgen.c
/*               print_clause()               hdump.c
/*               clear_var_substitution_tble  infer.c
/*               binary_factors()             infer.c
/*
/* returns:      nil
/*-----------------------------------------------------------------*/
 
void add_clause()

{
clause *c, *c1, *c2;
literal *l1, *l2, *lit, *lit2;
int i, j, k, kk, m, mc, repeat_add, max3, max4;

nextc = base + 1;

/*----------------------------------------*/
/* get two clauses that caused failures   */
/*----------------------------------------*/

c1 = clist[fail_clause3 - 1];
c2 = clist[fail_clause4 - 1];
 
begin_add:

/*
print0("\nRESOLVE CLAUSES:\n");

print_clause(c1);
print0("\n");
print_clause(c2);
print0("\n");
*/

//printf("nextc = %d RESOLVE ON ATOM %d:\n",nextc,fail_clause3_ply);
//fprintf(outfile,"RESOLVE ON ATOM %d:\n",fail_clause3_ply);

//print_clause(hb_list[fail_clause3_ply - 1]);
//print0("\n");

/*-------------------------------------*/
/* the two clauses may, in fact, be    */
/* the same clause.  in this case a    */
/* copy of the clause is necessary     */
/*-------------------------------------*/

if(c1 == c2) {

/*------------------------------------------------------*/
/* make a copy of c2                                    */
/* place it at the end of ulist                         */
/* it is out of the way there. ulist                    */
/* is cleared before the next start of                  */
/* constructing the next semantic tree.                 */
/*------------------------------------------------------*/
 
   clear_var_substitution_tble(c2->c_bits,c2->c_nv);
   unify(&c2->c_lit->l_id,&c2->c_lit->l_id);
   c = copy_clause(c2,2);
   c->c_rpar = c2->c_rpar;
   c2 = c;}
 
/*-------------------------------------*/
/* now c1 must be factored on the atom */
/* at the last level in the tree -- if */
/* it is possible to do so.  first     */
/* find literal at that was resolved   */
/* away by atom at last level.         */
/*-------------------------------------*/

for(i = 0; fail_lits3[i] != 0; i++) {
   if(c1->c_nlit < 2) {break;}
   if(fail_lits3[i] != fail_clause3_ply) {continue;}

/*-------------------------------------*/
/* attempt to factor on literal i of   */
/* clause c1                           */
/*-------------------------------------*/

   for(k = i + 1 ; fail_lits3[k] != 0; k++) {
      if(fail_lits3[i] == fail_lits3[k]) {

/*-------------------------------------*/
/* c1 can be factored on lit i and k   */
/*-------------------------------------*/

         binary_factors(c1, i + 1, k + 1);

/*-------------------------------------*/
/* now, c1 is not necessary to use     */
/* when trying to form new semantic    */
/* tree so we make it a DEAD clause    */
/* the factor, however, may be used!   */
/*-------------------------------------*/

         if(c1->c_id > base + 1) {c1->c_bits |= DEAD;}
         c1 = clist[nextc];

//       print0("ADD FACTOR OF C1:\n");
//       print_clause(c1);
//       print0("\n");

/*--------------------------------------*/
/* reindex literals with the atom       */
/* index that caused that literal to    */
/* fail.                                */
/*--------------------------------------*/

         for(kk = k; fail_lits3[kk] != 0; kk++) {
            fail_lits3[kk] = fail_lits3[kk + 1];}

         clist[nextc + 1] = clist[nextc]->c_end;
         nextc++;}}}

/*-------------------------------------*/
/* now c2 must be factored on the atom */
/* at the last level in the tree       */
/*-------------------------------------*/

if(c1->c_nlit == 1) {
   calc_clause_hashcode(c1);
   add_clause_to_ht(c1->c_X,c1->c_Y,c1->c_id,1);}


for(j = 0; fail_lits4[j] != 0; j++) {
   if(c2->c_nlit < 2) {break;}
   if(fail_lits4[j] != fail_clause4_ply) {continue;}
   for(k = j + 1; fail_lits4[k] != 0; k++) {
      if(fail_lits4[j] == fail_lits4[k]) {
         binary_factors(c2, (int) j + 1, (int) k + 1);
         if(c2->c_id > base + 1) {c2->c_bits |= DEAD;}
         c2 = clist[nextc];

//       print0("ADD FACTOR OF C2:\n");
//       print_clause(c2);
//       print0("\n");

         for(kk = k; fail_lits4[kk] != 0; kk++) {
            fail_lits4[kk] = fail_lits4[kk + 1];}

         clist[nextc + 1] = clist[nextc]->c_end;
         nextc++;}}}
 
if(c2->c_nlit == 1) {
   calc_clause_hashcode(c2);
   add_clause_to_ht(c2->c_X,c2->c_Y,c2->c_id,1);}

/*------------------------------------------------------*/
/* now both c1 and c2 have been factored as much as     */
/* necessary on the atom at the last level.             */
/*------------------------------------------------------*/

standardize(c1,c2);
	
clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);

l1 = c1->c_lit;
for(i = 0; ; i++){
// printf("fl3=%d ",fail_lits3[i]);
   if(fail_lits3[i] == fail_clause3_ply) {i++;break;}
   l1 = l1->l_next;}

l2 = c2->c_lit;
for(j = 0; ;j++) {
   if(fail_lits4[j] == fail_clause4_ply) {j++;break;}
   l2 = l2->l_next;}

/*------------------------------------------------------*/
/* unify two literals l1 and l2 and then form resolvent */
/*------------------------------------------------------*/

unify(&l1->l_id,&l2->l_id);
c = generate_clause(c1,c2,l1,l2, i, -j);

//print0("BINARY RESOLVENT IS:  \n");
//print_clause(c);
//print0("\n");
if(c1->c_id > base + 1) {c1->c_bits |= DEAD;}
if(c2->c_id > base + 1) {c2->c_bits |= DEAD;}
c->c_id = nextc + 1;
clist[nextc + 1] = c->c_end;
nextc++;

j = 0;
for(i = 0; fail_lits3[i] != 0 ; i++) {
   if(fail_lits3[i] != fail_clause3_ply) {
      fail_lits3[j++] = fail_lits3[i];}}

for(k = 0; fail_lits4[k] != 0 ; k++) {
   if(fail_lits4[k] != fail_clause4_ply) {
      fail_lits3[j++] = fail_lits4[k];}}
fail_lits3[j] = 0;

if(c->c_nlit < 2) {goto prepare_exit;}

repeat_factoring:
for(i = 0; fail_lits3[i] != 0 ; i++) {
   for(k = i + 1; fail_lits3[k] != 0; k++) {
      if(fail_lits3[i] == fail_lits3[k]) {

/*--------------------------------------------------*/
/* the literals li and lk can be factored.  the     */
/* fail_lits must be reindexed to take into account */
/* the removal of factored literals                 */
/*--------------------------------------------------*/

	    fail_lits3[j] = 0;
            for(j = k ; fail_lits3[j + 1] != 0 ; j++) {
               fail_lits3[j] = fail_lits3[j + 1];}
            fail_lits3[j] = 0;
            binary_factors(c, i + 1, k + 1);
            c->c_bits |= DEAD;
            clist[nextc + 1] = clist[nextc]->c_end;
            c = clist[nextc];

//          print0("FACTOR RESOLVENT:  \n");
//          print_clause(c);
//          print0("\n");
            nextc++;
	    goto repeat_factoring;}}}

prepare_exit:

/*--------------------------------------------------*/
/* eliminate identical literals to form factored    */
/* clause.                                          */
/*--------------------------------------------------*/

if(c->c_nlit == 0) goto now_exit; 

final_factoring:
i = 0;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   i++;
   k = i;
   for(lit2 = lit->l_next; lit2 != 0; lit2 = lit2->l_next) {
      k++;
      if(lit->l_X == lit2->l_X && lit->l_Y == lit2->l_Y) {
	 fail_lits3[j] = 0;
         for(j = k; fail_lits3[j + 1] != 0; j++) {
	    fail_lits3[j] = fail_lits3[j + 1];}
         fail_lits3[j] = 0;
         binary_factors(c,i,k);
         c->c_bits |= DEAD;
         clist[nextc + 1] = clist[nextc]->c_end;
         c = clist[nextc];

//       print0("FINAL FACTOR TWO IDENTICAL LITERALS:\n");
//       print_clause(c);
	 nextc++;
//       print0("\n");
	 goto final_factoring;}}}

/*--------------------------------------------*/
/* see if more inferences should be generated */
/*--------------------------------------------*/

if(c->c_nlit == 1) {
   calc_clause_hashcode(c);
   add_clause_to_ht(c->c_X,c->c_Y,c->c_id,1);}

//printf("See if more inferences might be generated\n");

   repeat_add = 0;
   max4 = 0;
   for(i = 0; i < 16; i++) {
      fail_lits4[i] = fail_lits3[i];
      if(fail_lits4[i] > max4) max4 = fail_lits4[i];
      if(fail_lits4[i] == fail_clause1_ply - 1) {repeat_add++;} }
   c2 = c;

see_if_more:
//printf("\nSEE IF MORE..\n");

if(fail_clause_left[fail_clause1_ply - 2][0] != 0) {
/* printf("\nPerhaps another clause at ply %d can be generated from clause %d\n",
       ply, fail_clause_left[fail_clause1_ply - 2][0]);
   printf("Resolve together (?): ");
   print_clause(clist[fail_clause_left[fail_clause1_ply - 2][0] - 1]);
   printf("\n");
   print_clause(c);
   printf("\n");
*/

   c1 = clist[fail_clause_left[fail_clause1_ply - 2][0] - 1];
     
   max3 = 0;
   for(i = 1; i < 17; i++) {
      fail_lits3[i - 1] = fail_clause_left[fail_clause1_ply - 2][i];
      if(fail_lits3[i - 1] > max3) max3 = fail_lits3[i - 1];}

   if(max3 > max4) {
      fail_clause3_ply--;
      fail_clause1_ply--;
      goto see_if_more;}

   fail_clause3_ply--;
   fail_clause4_ply = fail_clause3_ply;
   fail_clause1_ply--;
   if(max3 == max4 && c1->c_nlit  + c2->c_nlit > 2) goto begin_add;}

//printf("\nDONE ADDING...\n");

if(c2->c_nlit == 1) {
   calc_clause_hashcode(c2);
//   printf("ADD HASHCODE OF FACTORED CLAUSE: ");print_clause(c2);
   add_clause_to_ht(c2->c_X,c2->c_Y,c2->c_id,1);}

//printf("Fix base clauses...Eliminate subsumed clauses\n");
for(i = 0; i < nextc; i++) {
   c1 =  clist[i];
   if((c1->c_bits & DEAD) !=  0) {continue;}
   for(j = 0; j < nextc; j++) {
      if(i == j) {continue;}
      c2 = clist[j];
      if((c2->c_bits & DEAD) !=  0) {continue;}
      if(simple_subsume(c1,c2)) {
//       printf("clause %d subsumes clause %d\n",i + 1,j + 1);
//       print_clause(c1); printf("\n");print_clause(c2);printf("\n\n");
         c2->c_bits |= DEAD;}
      else if(simple_subsume(c2,c1)) {
//       printf("Clause %d subsumes clause %d\n",j + 1,i + 1);
//       print_clause(c2); printf("\n");print_clause(c1);printf("\n\n");
         c1->c_bits |= DEAD;}}}

now_exit:
base = nextc - 1;

//printf("Done fixing base clauses...\n");

/*-------------------------------------------------*/
/* scan fail_clause_left array to find atoms that  */
/* do not appear in array.  These atoms can be     */
/* eliminated on the next iteration.               */
/*-------------------------------------------------*/

if(x_debug > 0) {
   printf("PRINT DURING DEBUGGING\n\n");
   for(i = 0; i <= base ; i++) {
      print_clause(clist[i]);
      printf("\n");}
   printf("\nEND OF BASE CLAUSES PRINTED DURING DEBUGGING\n\n");}
return;
}
