/*--------------------------------------------------------------*/
/* infer.c           HERBY.......generate inferences
/*
/* Copyright Newborn Software 1995 - 1999, 2000
/*
/* functions:
/*    binary_resolvents()         ................... #1
/*    binary_resolvents2()        ................... #2
/*    binary_resolvents3()        ................... #3
/*    binary_factors()            ................... #4
/*    standardize()               ................... #5
/*    clear_var_substitution_tble()  ................ #6
/*--------------------------------------------------------------*/
 
#include <stdio.h>

#include "constHS.h"
#include "clauseHS.h"

/*--------------------------------------------------------------*/
/* externally defined functions                                 */
/*--------------------------------------------------------------*/
 
extern int unify();                 /* unify.c: unify lits      */

extern int install_clause();        /* build.c: instal cls      */

extern clause *generate_clause();   /* hgen.c: gen clause       */
extern clause *generate_clause2();
extern clause *gen_unified_lit();

extern clause *hbgen_clause(); 

extern void print_clause();
extern void print_clause2();

/*--------------------------------------------------------------*/
/* externally defined  constants                                */
/*--------------------------------------------------------------*/


extern char *OutPtr;

extern clause *clist[];         /* first few are base clauses   */
extern int *var_substitution_tbl[];    /* variable subs table   */
extern int fail[];
 
extern int nextc, ply; 
extern int x_debug;
extern int phase;
extern int s_nod;

/*--------------------------------------------------------------*/
/* internally defined functions                                 */
/*--------------------------------------------------------------*/

int binary_resolvents();
int binary_resolvents2();
int binary_resolvents3();
int binary_factors();
void standardize();
void clear_var_substitution_tble();

int br01 = 0,br02,br03;

/*------------------------------------------------------------------*/
/* binary_resolvents(c1,c2)         ..................... #1
/*
/* generates all resolvents of two clauses that are produced
/* from binary resolution.  resolvents are listed one after
/* the other on clist[] in the next open location.
/*
/* parameters:     c1,c2 : clauses to resolve
/*
/* called by:  calc_inferences()                build.c
/*             resolution_stage()               atom.c
/*
/* calls:      standardize()                    infer.c
/*             clear_var_substitution_tble()    infer.c
/*             unify()                          unify.c
/*             hbgen_clause()                   hgen.c
/*             generate_clause()                hgen.c
/*             install_clause()                 build.c
/*
/* returns:    # of clauses produced
/*------------------------------------------------------------------*/
 
int binary_resolvents(c1,c2)
clause *c1,*c2;
{
literal *l1, *l2;
int id1, id2;
clause *c;
int ix1, ix2;	
int numb_resolvents;
int installed;
             
standardize(c1,c2);                   /* standardize vars apart   */

numb_resolvents = 0;       /* count of the # of binary resolvents */

ix1 = 1;
for(l1 = c1->c_lit ; l1 != 0 ; l1 = l1->l_next) {
    ix2 = 1;
    for(l2 = c2->c_lit ; l2 != 0 ; l2 = l2->l_next) {
	if((l2->l_id & ELIMLIT) == 1) {continue;}
        id1 = l1->l_id & IDMASK;               /* predicate IDs   */
        id2 = l2->l_id & IDMASK;
 
/*----------------------------------------------------------*/
/* if IDs are the same then we attempt unification.         */
/*----------------------------------------------------------*/
 
        if(id1 == id2) {
           if((l1->l_id ^ l2->l_id) == NEGMASK) {
              clear_var_substitution_tble(c1->c_bits,c1->c_nv);
              clear_var_substitution_tble(c2->c_bits,c2->c_nv);
              if(unify(&l1->l_id,&l2->l_id)) {

                 if(phase == 1) {c = hbgen_clause(c1,c2,l2);}
                 else {c = generate_clause(c1,c2,l1,l2,ix1,ix2);}
                 if(c != 0) {
		    installed = install_clause(c);
                    if(installed == 1) {
                       numb_resolvents++;
                       if(fail[ply + 1] == 2) {return(numb_resolvents);}}
	            else {
		       if(installed < 0) {return(-1);}}}}}}
 
             ix2++;}
      ix1++;}
return(numb_resolvents);
}


/*-----------------------------------------------------------*/
/* binary_resolvents2()	       .................#2
/*
/* forms the binary resolvent of 
/* clauses c1 and c2 on literals x1 and x2
/*
/* called by: ash1()                         atom.c
/*            ash2()                         atom.c
/*            ash3()                         atom.c
/*            ash4()                         atom.c
/*            install_unit_and_test()        atom.c
/*
/* calls:     standardize()                  infer.c
/*            clear_var_substitution_tble()  infer.c
/*            unify()                        unify.c
/*            generate_clause2()             atom.c
/*				
/* returns:   1 if clause is istalled; otherwise 0.
/*-------------------------------------------------------------*/
 
int binary_resolvents2(c1,c2,x1,x2,add_lit)
clause *c1, *c2;
int x1, x2, add_lit;
{
clause *c;
literal *lit1, *lit2;
int xx1, xx2;

standardize(c1,c2);
 
/*----------------------------------------------*/
/* find appropriate literals; skip eliminated   */
/* literals.                                    */
/*----------------------------------------------*/

xx1 = 0;
for(lit1 = c1->c_lit; lit1 != 0; lit1 = lit1->l_next) {
   if((lit1->l_id & ELIMLIT) == 0) {
      xx1++;
      if(xx1 == x1) {break;}}}

br01++; 

xx2 = 0;
for(lit2 = c2->c_lit; lit2 != 0; lit2 = lit2->l_next  ) {
   if((lit2->l_id & ELIMLIT) == 0) { 
      xx2++;
      if(xx2 == x2) {break;}}}

if((lit1->l_id & IDMASK) != (lit2->l_id & IDMASK)) {return(0);}
if((lit1->l_id ^ lit2->l_id) != NEGMASK) {return(0);}

clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);

if(!unify(&lit1->l_id,&lit2->l_id)) {return(0);}
c = generate_clause2(c1,c2,lit1,lit2,x1,x2,add_lit); 
if(c == 0) {return(0);}

if(nextc > CSPACE_CLAUSES -2) {
   OutPtr += sprintf(OutPtr,"\nToo many clauses (%d) \n",nextc);
   fatal("\nQuitting search\n");}
clist[nextc++] = c;
clist[nextc] = c->c_end;

return(1);
}
 
/*-----------------------------------------------------------*/
/* binary_resolvents3()	       .................#3
/*
/* forms the binary resolvent of 
/* clauses c1 and c2 on literals x1 and x2
/*
/* binary_resolvents3 forms the most common instance
/* of the resolved-upon literal in the two clauses,
/* c1 and c2.  for example, if c1 = p(x,a) | q(b,y)
/* c2 = r(a,b) | ~q(x,y), then this function would
/* produce q(b,x).
/*
/* called by: ash1()                         atom.c
/*
/* calls:     standardize()                  infer.c
/*            clear_var_substitution_tble()  infer.c
/*            unify()                        unify.c
/*            gen_unified_lit()              hgen.c
/*				
/* returns:   1 if clause is istalled; otherwise 0.
/*-------------------------------------------------------------*/
 
int binary_resolvents3(c1,c2,x1,x2,add_lit)
clause *c1, *c2;
int x1, x2, add_lit;
{
clause *c;
literal *lit1, *lit2;
int xx1, xx2;

standardize(c1,c2);
 
/*----------------------------------------------*/
/* find appropriate literals; skip eliminated   */
/* literals.                                    */
/*----------------------------------------------*/

xx1 = 0;
for(lit1 = c1->c_lit; lit1 != 0 ; lit1 = lit1->l_next) {
   if((lit1->l_id & ELIMLIT) == 0) {
      xx1++;
      if(xx1 == x1) {break;}}}
 
xx2 = 0;
for(lit2 = c2->c_lit; lit2 != 0; lit2 = lit2->l_next  ) {
   if((lit2->l_id & ELIMLIT) == 0) { 
      xx2++;
      if(xx2 == x2) {break;}}}

if((lit1->l_id & IDMASK) != (lit2->l_id & IDMASK)) {
   return(0);}
if((lit1->l_id ^ lit2->l_id) != NEGMASK) {
   return(0);}

clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);

if(!unify(&lit1->l_id,&lit2->l_id)) {return(0);}
c = gen_unified_lit(c1); 

if(nextc > CSPACE_CLAUSES -2) {
   printf("\nToo many clauses (%d) \n",nextc);
   fatal("\nQuitting search\n");}
clist[nextc++] = c;
clist[nextc] = c->c_end;
return(1);
}


/*-------------------------------------------------------------*/
/* binary_factors()                 ................... #4
/*
/* factoring is an inference rule that produces a new clause
/* from a given non-unit cl. A typical example of factoring
/* is the following:
/*
/*                     p(x) | p(y)
/*
/*      can be factored into p(x). What we do is attempt to find an
/*      mgu such that we collapse two literals of a clause into a
/*      singleton.
/*
/*      binary_factors will generate all factors of a given cause.
/*      Since the new clause will have only one parent, we use the
/*      left parent (and NULL the right parent).
/*
/* parameters: c : non-unit clause to factor
/*             ix1: index of one literal to factor
/*             ix2: index of second literal to factor
/*
/* called by:  add_clause()                    herby.c
/*
/* calls:      clear_var_substitution_tble()   infer.c
/*             unify()                         unify.c
/*             generate_clause()               hgen.c
/*
/* returns:    the # of factors produced
/*-------------------------------------------------------------*/
 
int binary_factors(c1,ix1,ix2)
clause *c1;
int ix1, ix2;
{
literal *l1, *l2;
clause *c;
int j;
 

l1 = c1->c_lit;   for( j = 1 ; j < ix1; j++) l1 = l1->l_next;
l2 = c1->c_lit;   for( j = 1 ; j < ix2; j++) l2 = l2->l_next;

clear_var_substitution_tble(c1->c_bits,c1->c_nv);
unify(&l1->l_id,&l2->l_id); 
c = generate_clause(c1,0,l2,0,ix1,-ix2);
c->c_id = nextc +1;
clist[nextc] = c;
clist[nextc + 1] = c->c_end;
}
 
/*--------------------------------------------------------------*/
/* standardize(c1,c2)            ...................  #5
/*
/* renames variables
/*
/* called by:  binary_resolvents()          infer.c
/*             binary_resolvent2()          infer.c
/*             binary_resolvent3()          infer.c
/*             add_clause()                 herby.c
/*
/* calls:      nobody
/*
/* returns:    nil
/*--------------------------------------------------------------*/
 
void standardize(c1,c2)
clause *c1,*c2;
{
int i, a, *p;
literal *l;
 
/*---------------------------------------------*/
/* if variable in one clauses are numbered     */
/* beginning at 0 in one clauses and 200 in    */
/* the other, var are already standardized,    */
/* and the program should return               */
/*---------------------------------------------*/
 
if((c1->c_bits & VID) != (c2->c_bits & VID)) {return;}
 
/*---------------------------------------------*/
/* fix offset of variables in clause c2        */
/*---------------------------------------------*/
 
if((c2->c_bits & VID) == 0) {
   c2->c_bits |= VID;
   i = MAX_VARS;}        /* if now 0, add 200  */
else {
   c2->c_bits &= ~VID;
   i = -MAX_VARS;}      /* otherwise, sub 200  */
   
/*----------------------------------------------*/
/* now scan clause while renumbering variables  */
/*----------------------------------------------*/
 
for( l = c2->c_lit; l != 0 ; l = l->l_next) {
   p = &l->l_id;
   for(a = *p++ & ARGMASK; a > 0 ; a--) {
      if((*p & FCNMASK) != 0) {
         a += *p++ & ARGMASK;}
      else {
         *p += i;
         p++;}}}
}
 
/*-------------------------------------------------------------*/
/* clear_var_substitution_tble(i,n) .............. #6

/*
/* used to zero a section of the vst table (i.e.
/* set each entry to NULL). called
/* before performing unification
/*
/* calling arguments:
/*             i = c_base (ptr to first variable in cl
/*             n = c_nv (number of distinct var in cl
/*
/* called by:  binary_resolvents()         infer.c
/*             binary_resolvents2()        atom.c
/*             binary_factors()            infer.c
/*             add_clause()                herby.c
/*             copy_atom_from_clause()     hgen.c
/*             copy_gnd_atom_from_cls()    hgen.c
/*             copy_clause()               hgen.c
/*
/* calls:      nobody
/*
/* returns:    nil
/*--------------------------------------------------------------*/
 
void clear_var_substitution_tble(i,n)
int i,n;
{
int **p, j;

if((i & VID) == 0) j = 0;
else j = MAX_VARS;
 
for(p = &var_substitution_tbl[j]; n > 0; n--) *p++ = 0;
}
