/*----------------------------------------------------------*/
/* modify.c (Version for Theo)                              */     
/* copyrighted by Monty Newborn 2005                        */
/*                                                          */
/* Added December 2002                                      */
/*                                                          */
/* Controls the modification of clauses                     */
/*     determine_potential_changes().............#1         */
/*     modify_clause()...........................#2         */
/*     form_weakened_clause()....................#3         */
/*     print_change_array()......................#5         */
/*     reinstate_base_clause()...................#7         */
/*     bres_instantiated_proof().................#8         */
/*     bfac_instantiated_proof().................#9         */
/*     bsub_instantiated_proof().................#10        */
/*     filter_ulist()............................#11        */
/*     skip_equal_all_vars().....................#12        */
/*     skip_equal_first_last()...................#13        */
/*     print_xl_array()..........................#14        */ 
/*     det_dependencies()........................#15        */
/*     analyze_decoposeability().................#16        */
/*----------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "const.h"
#include "clause.h"

/*-----------------------------------------------------------*/
/* externally defined functions                              */
/*-----------------------------------------------------------*/

extern void print_clause();
extern int grab_number();
extern void standardize();
extern void clear_var_substitution_tble();
extern int unify();
extern clause *generate_iclause();
extern void calc_clause_hashcode();
extern int time_now();
extern void clean_var_names();
extern void replace_vars();
extern void copy_clause_all();
extern void normalize_literal();
extern int simple_subsume2();
		   
/*-----------------------------------------------------------*/
/* externally defined constants                              */
/*-----------------------------------------------------------*/

extern char *OutPtr;
extern int ply, nextu;
extern int base, org_base, nc;
extern int batch_tried;
extern int o_maxvar, o_maxlit, o_maxterm;
extern int o_batchtime, o_maxtime;
extern int s_starttime;
extern int s_nod;
extern int s_nod_last;
extern int learn_nod;
extern int o_debug;
extern int o_learn, weakened_ply, weakened_clause,o_weakened_index;
extern int length_of_proof;
extern int clauseToWeaken;
extern int weakened_proof_remember;
extern char *weakProofBeginsHere, *weakProofEndsHere;
extern int good_clauses[MAX_BASE_CLAUSES];
extern int good_inferred_cls;
extern int satisfiable;
extern int equal_pred;
extern clause *clist[];
extern clause *ulist[];
extern int u_space;
extern char input_buffer[];
extern int cl[PROOF_ARRAY_SIZE][14];
extern int xl[PROOF_ARRAY_SIZE][14];
extern int equal_xyyzxz;
extern dnode depth[DEPTH_PTR_SPACE];
extern int cade_format;
extern int spawned;
extern int reinstate_parent[1000];
extern int questionableUsed;
extern int numb_trivial_proofs;
extern int this_computer;

/*-----------------------------------------------------------*/
/* internally defined functions                              */
/*-----------------------------------------------------------*/

void determine_potential_changes();
void modify_clause();
void print_change_array();
void form_weakened_clause();
void reinstate_base_clause();
int bres_instantiated_proof();
int bfac_instantiated_proof();
int bsub_instantiated_proof();
void filter_ulist();
int skip_equal_all_vars();
int skip_equal_first_last();
void print_xl_array();
void det_dependencies();
void analyze_decomposeability();

/*-----------------------------------------------------------*/
/* internally defined constants                              */
/*-----------------------------------------------------------*/

char clause_number[5];
char literal_number[5];
char char_number[5];
char old_char[80];
char new_char[5];
int replaced_char;
int deleted_terms[200];
int questionable[2000];
int total_questionable;
int excellent_weakened_proof;
int used, used_nc;

/*------------------------------------------------------------*/
/* change_array:                                              */
/*         column 0:  batch_tried                             */
/*         column 1:  clause number                           */
/*         column 2:  literal number                          */
/*         column 3:  position in literal                     */
/*         column 4:  substitute (variable)                   */
/*         cols 5-9:  characters substituted for              */
/*------------------------------------------------------------*/

int change_array[50][10];
int change_level;

/*------------------------------------------------------------*/
/* potential_change_array:                                    */
/*          column 0: clause                                  */
/*          column 1: literal                                 */
/*          column 2: position                                */
/*          column 3: char                                    */
/*          column 4: #var in clause                          */
/*          column 5: substituted char.                       */
/*------------------------------------------------------------*/

int p_change_array[2000][7];
int p_change_number;
int randy, randy2, randy_last, randy2_last;
int repeat_weakening;

int decompose, decompose2, most_decomposible, decomposible_parts;
int var_vector[16][100], best_var_vector[16][100], split_temp[16];
int split_pbs, unsolved_split_pbs, solved_split_lits[16];
int unsolved_exists_pbs;
int solved_split_pbs[16];
int exists_split_pbs[16];
int exists_split_lits[16];
int pick_split_pbm;

/*------------------------------------------------------------*/
/* determine_potential_changes() ..................#1         */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:   skip_equal_all_vars()                             */
/*          skip_equal_first_last()                           */
/*          tdump.c/print_clause()                            */
/*                                                            */
/* determines a list of potential changes that can be made to */
/* the base clauses to weaken them.  At most 2000 changes can */
/* be listed on potential_change_array[][].                   */
/*------------------------------------------------------------*/

void determine_potential_changes()
{
int i, j, jj, k, kk, a, *p, v_cnt, position, litl_index;
int jjtemp, var_cnt[64], base_print[2000];
literal *lit;
clause *c;
int second_half;

/*--------------------------------------------------*/
/* Heuristic function to weight weakeneings:        */
/*     1: Constant in set-of-support                */
/*     2: function in set of support                */
/*     3: literal in set-of-support                 */
/*     4: Constant in axioms                        */
/*     5: Function in axioms                        */
/*     6: Literal in axioms                         */
/*--------------------------------------------------*/

/*--------------------------------------------------*/
/* List positions where a constant can be replaced  */
/*--------------------------------------------------*/

for(i = 0; i < 2000; i++) {
   for(j = 0; j < 7; j++) {p_change_array[i][j] = 0;}}

k = -1;
for(i = base; i >= 0; i--) {

/*--------------------------------------*/
/* for base clauses                     */
/*--------------------------------------*/

   for(j = 0; j < 64; j++) {var_cnt[j] = 0;}
   c = clist[i];
   if(c->c_bits & EQCLS) {continue;}
   if(c->c_nlit == 1 && skip_equal_all_vars(c)) {continue;}
   v_cnt = 0;
   kk = k;
   litl_index = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      p = &lit->l_id;
      litl_index++;position = second_half = 0;
      for(a = *p++ & ARGMASK; a > 0; a--) {
	 position++;
	 if(a == 1 && second_half == 0) {second_half = position;}
         if(*p & FCNMASK) {
            if((*p & ARGMASK) == 0) {
               if(!skip_equal_first_last(c,p)) {

/*----------------------------------------*/
/* the replacement of any constant by a   */
/* variable is a potential change.        */
/* for clauses in the set-of-support      */
/*----------------------------------------*/

                  if(k < 1999) {
		     k++;
	             p_change_array[k][0] = c->c_id;
	             p_change_array[k][1] = litl_index;
	             p_change_array[k][2] = position;
                     p_change_array[k][3] = *p;
		     if(c->c_id >= base - nc + 1) {
		        if((c->c_lit->l_id & FCNMASK) != 0) {
			   p_change_array[k][6] = 50000 - position;}
                        else p_change_array[k][6] = 100000 - position;}
                     else p_change_array[k][6] = 400000 - position;}}}
	    else {

/*---------------------------------------*/
/* replace function with variable        */
/* but..... first ELIMINATE weakenings   */
/* for unit clauses                      */
/* where pred is ~equal(f....,g....)     */
/* and weakening replaces f with var     */
/* or g with a var, as in these cases    */
/* weakening won't help much!!.          */
/*---------------------------------------*/

               if(lit->l_id == (equal_pred | FCNMASK) && c->c_nlit==1) {
                  if(position == 1 || position == second_half) { 
		     goto Skip_saving;}}

	       if(k < 1999) {
                  k++;
	          p_change_array[k][0] = c->c_id;
	          p_change_array[k][1] = litl_index;
	          p_change_array[k][2] = position;
	          p_change_array[k][3] = *p;
		  if(c->c_id >= base - nc + 1) {
		     if((c->c_lit->l_id & FCNMASK) != 0) {
                        p_change_array[k][6] = 80000 - position;}
                     else p_change_array[k][6] = 200000 - position;}
                  else p_change_array[k][6] = 500000 - position;}}
Skip_saving:
            a += *p++ & ARGMASK;}
         else {
	    if(var_cnt[*p + 1] == 0) {
	       var_cnt[*p + 1] = *p + 1;
	       v_cnt++;
	       if(v_cnt > o_maxvar) goto Skip_saving;}
            else {
               if(k < 1999) {
                  k++;
                  p_change_array[k][0] = c->c_id;
                  p_change_array[k][1] = litl_index;
                  p_change_array[k][2] = position;
                  if(c->c_id >= base - nc) {
                     p_change_array[k][6] = 600000 - position;}
                  else {
                     if(litl_index == c->c_nlit) {
                        p_change_array[k][6] = 600000 - position;}
                     else {
                        p_change_array[k][6] = 600000 - position;}}}}
	   
            p++;}}}

   for(j = kk + 1; j <= k; j++) {p_change_array[j][4] = v_cnt;}}

/*-------------------------------------------------*/
/* list literals that can be deleted from clauses  */
/* with more than 2 literals.                      */
/*-------------------------------------------------*/

for(i = base; i >= 0; i--) {
   c = clist[i];
   if(c->c_bits & EQCLS) {continue;}
   litl_index = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      litl_index++;
      if((litl_index == c->c_nlit && litl_index > 1) || (c->c_nlit > 1 && (i >= base - nc))) {
         if(k < 1999) {
            k++;
            p_change_array[k][0] = c->c_id;
            p_change_array[k][1] = litl_index;
            p_change_array[k][2] = -1;
            p_change_array[k][3] = -1;
            p_change_array[k][4] = c->c_nv;
            if(c->c_id >= base - nc + 1) {
               p_change_array[k][6] = 300000 - c->c_id * 100 - litl_index;}
            else p_change_array[k][6] = 600000 - c->c_id * 100 - litl_index;}}}}

if(p_change_array[k][0] == 0 && k >= 0) k--;

/*-----------------------------------------*/
/* sort based on heuristic score           */
/*-----------------------------------------*/

if( k != -1) {
   for(i = 0; i < k; i++) {
      for(j = i + 1; j <= k; j++) {
         if(p_change_array[i][6] <= p_change_array[j][6]) {continue;}
         for(jj = 0; jj < 7; jj++) {
            jjtemp = p_change_array[i][jj];
            p_change_array[i][jj] = p_change_array[j][jj];
            p_change_array[j][jj] = jjtemp;}}}}

/*-------------------------------------------------*/
/* print potential changes                         */
/*-------------------------------------------------*/

if(o_debug > 1) {
   if(k == -1) {
      printf("\n\nPOTENTIAL CHANGE ARRAY: no modifications possible\n");
      p_change_number = 0;
      return;}
   else {
      printf("\n\nPOTENTIAL CHANGE ARRAY: %d modifications possible\n",k+1);}
      printf(" # Clause  Literal  Position  Contents  Variables Heuristic\n");}

for(i = 0; i < 2000; i++) base_print[i] = 0;

if(o_debug > 1) {
   for(i = 0;  i <= k; i++) {
      base_print[p_change_array[i][0]]++;
      printf("%2d %5d    %3d       %2d      %8x     %2d    %1d\n",
	 i,
	 p_change_array[i][0],
	 p_change_array[i][1],
	 p_change_array[i][2],
	 p_change_array[i][3],
	 p_change_array[i][4],
	 p_change_array[i][6]);}
   print0("\n\n");}

/*--------------------------------------------------*/
/* print clauses that can be modified               */
/*--------------------------------------------------*/

p_change_number = k + 1;
return;
}

/*------------------------------------------------------------*/
/* modify_clause() .................................#2        */
/*                                                            */
/* modifies a clause as desired                               */
/*                                                            */
/* called from: main()             theo.c                     */
/*                                                            */
/* calls:       determine_potential_changes()                 */
/*              print_clause()                                */
/*              grab_number()                                 */
/*              clean_var_names()                             */
/*                                                            */
/*------------------------------------------------------------*/

void modify_clause()
{
int f, i, j, k, kk, l, m, mm, n, nn, lit_end, nv;
int dt;                            /* index for deleted_terms */
char ichar;
int a, aa, *p, *pp, *ppp, position;
clause *c, *c_nc;
literal *lit;
int timeTook, timeToGo;
int good1, good2, randy_temp;

/*---------------------------------------------*/
/* determine clause to modify                  */
/*---------------------------------------------*/

analyze_decomposeability(0);
}

/*------------------------------------------------------------*/
/* form_weakened_clause() ........................#3          */
/*                                                            */
/* mark clauses that were derived from modified clauses       */
/* and their descendants                                      */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:     Nobody                                          */
/*                                                            */
/*------------------------------------------------------------*/

void form_weakened_clause(randy)
int randy;
{
int j, k, kk, l, f, m, nv, n, dt, aa, lit_end;
int position, a, *p, *pp, *ppp;
clause *c;
literal *lit;

j = p_change_array[randy][0];
k = p_change_array[randy][1];
l = p_change_array[randy][2];
f = p_change_array[randy][3];
m = p_change_array[randy][4];
nv = p_change_array[randy][4];
c = clist[j - 1];

print0("Weaken clause: ");
print_clause(c);

/*----------------------------------------------*/
/* now carry out the weakening of the clause    */
/* first determine the literal to weaken.       */
/*----------------------------------------------*/

kk = k;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   if(--kk == 0) break;}

if(l != 0 && l != -1) {
   p = &lit->l_id;
   position = 1;
   if(nv > o_maxvar - 3) {o_maxvar++;}
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {
         if(position++ == l) {
            if((*p & ARGMASK) == 0) {

/*------------------------------------------*/
/* replace constant with a variable!        */
/*------------------------------------------*/

               replaced_char = *p;*p = nv++;break;}
            else {

/*------------------------------------------*/
/* replace function with a variable!        */
/*------------------------------------------*/

               pp = p;
	       if(randy != randy2) dt = 1;
	       else dt = 101;

/*------------------------------------*/
/* find end of function               */
/*------------------------------------*/

	       for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
                  if(*pp & FCNMASK) {aa += *pp++ & ARGMASK;}
                  else {pp++;}}

/*------------------------------------*/
/* p points to fxn to be replaced     */
/* pp points to first term after fxn  */
/*------------------------------------*/

               if(lit->l_next != 0) lit_end = (int) lit->l_next;
	       else lit_end = (int) c->c_end;

	       for(ppp = p; (int) ppp < lit_end; ppp++) {
		  deleted_terms[dt++] = *ppp;}

               if(randy != randy2) deleted_terms[0] = dt - 1;
	       else deleted_terms[100] = dt - 1;

/*-------------------------------------*/
/* find end of literal                 */
/*-------------------------------------*/

                replaced_char = *p;
                *p = nv++;
                p++;
                while((int) pp < (int) lit_end) {
                   *p = *pp;
                   p++;pp++;}
                break;}}
          a += *p++ & ARGMASK;}
       else {
          if(position++ == l) {
             replaced_char = *p;
             n = *p;*p = nv++;break;}
          p++;}}}

/*-------------------------------------*/
/* delete literal                      */
/*-------------------------------------*/

else if(l == -1) {
   replaced_char = 0;
   c->c_nlit--;
   lit->l_id = lit->l_id | ELIMLIT;}

return;
}

/*-----------------------------------------------------------*/
/* print_change_array() ........................#5           */
/*                                                           */
/* called by:                                                */
/*                                                           */
/* calls:      Nobody                                        */
/*                                                           */
/*-----------------------------------------------------------*/

void print_change_array()
{
int i;

}

/*----------------------------------------------------------------*/
/* skip_equal_all_vars() ..........................#12            */
/*                                                                */
/* called by:                                                     */
/*                                                                */
/* calls:                                                         */
/*                                                                */
/* don't use a clause of form equal(x,y) where x,y the same       */
/* variable for learning.  The clause is of no help               */
/*                                                                */
/* return: 0 if clause is not of useless form                     */
/* return: 1 if clause is useless                                 */
/*----------------------------------------------------------------*/

int skip_equal_all_vars(c)
clause *c;
{
int *p;

if(c->c_nlit > 1) {return(0);}
if((c->c_lit->l_id & 0x8000FFFF) != (equal_pred | FCNMASK)) {return(0);}

/*----------------------------------*/
/* first test for equal(a,b)        */
/*----------------------------------*/

p = &c->c_lit->l_id;
p++;
if((*p & FCNMASK) && ((*p & ARGMASK) == 0)) {
   p++;
   if((*p & FCNMASK) && ((*p & ARGMASK) == 0)) {return(1);}}

/*-----------------------------------*/
/* now test for equal(xxxxx,constant)*/
/* (to be completed??)               */
/*-----------------------------------*/

p = &c->c_lit->l_id;
p++;
if((*p & FCNMASK) == 0) {
   p++;
   if((*p & FCNMASK) == 0) {return(1);}}
return(0);
}

/*----------------------------------------------------------------*/
/* skip_equal_first_last() ...........................#13         */
/*                                                                */
/* called by:                                                     */
/*                                                                */
/* calls:      nobody                                             */
/*                                                                */
/* don't use a clause of form equal(c,xxxxxy) or equal(xxxx,c)    */
/*                                                                */
/* return: 0 if clause is not of useless form                     */
/* return: 1 if clause is useless                                 */
/*----------------------------------------------------------------*/

int skip_equal_first_last(c,pp)
int *pp;
clause *c;
{
int *p, a;
int string_length, string_length2;
int *const_ptr, *const_ptr2;

if(c->c_nlit > 1) {return(0);}
if((c->c_lit->l_id & 0x8000FFFF) != (equal_pred | FCNMASK)) {return(0);}

/*----------------------------------*/
/* first test for equal(a,b)        */
/*----------------------------------*/

string_length = 0; string_length2 = 1;

p = &c->c_lit->l_id;

for(a = *p++ & ARGMASK; a > 1; a--) {
   if(string_length == 0) {const_ptr = p;}
   string_length++;
   if(*p & FCNMASK) {a += *p++ & ARGMASK;}
   else {p++;}}

const_ptr2 = p;

if((*p & FCNMASK) == 0) {p++;}
else {
   for(a = *p++ & ARGMASK; a > 0; a--) {
      string_length2++;
      if(*p & FCNMASK) {a += *p++ & ARGMASK;}
      else {p++;}}}

if(string_length * string_length2 == string_length ||
   string_length * string_length2 == string_length2) {

   if(string_length == 1 && string_length2 > 1 && pp == const_ptr &&
            (*pp & FCNMASK)) {return(1);}

   if(string_length2 == 1 && string_length > 1 && pp == const_ptr2 &&
            (*pp & FCNMASK)) {return(1);}}

/*-----------------------------------*/
/* now test for equal(xxxxx,constant)*/
/* (to be completed??)               */
/*-----------------------------------*/

return(0);
}

/*---------------------------------------------------------------*/
/* print_xl_array()
/*
/*---------------------------------------------------------------*/

void print_xl_array()
{
int i;

printf("XL ARRAY\n");

for(i = 0; xl[i][0] != 0; i++) {
   printf(
      "%d  L %4d,%d  R %4d,%d H %d  I: %d  ST %8x CH %x H %d,%d",
      (xl[i][0] >> 16) & CIDMASK2,                    /* block   */
      (xl[i][0] >> 4) & PARENT_MASK,                  /* l_par   */
      (xl[i][0] & INDEX_MASK),                        /* index1  */
      ((xl[i][1] >> 4) & PARENT_MASK),
      (xl[i][1] & INDEX_MASK),
      xl[i][3],
      xl[i][2],
      xl[i][4],
      xl[i][5],
      xl[i][6],
      xl[i][7]);
   printf(" HR %d H %d,%d",xl[i][8],xl[i][9],xl[i][10]);
   if(xl[i][11] == 1) {printf(" R");}
   if(xl[i][11] == 2) {printf(" E");}
   if(xl[i][12] != 0) {printf(" X%d",xl[i][12]);}
   if(xl[i][13] != 0) {printf(" S%x",xl[i][13]);}
   printf("\n");}
 
print0("\n");
return;         
}
 

/*---------------------------------------------------------------*/
/*  det_dependencies()
/*
/*---------------------------------------------------------------*/

void det_dependencies(c)
clause *c;
{
int i, j;
int c_left,c_right;
int marked; 

i = 0;
while((xl[i][0] >> 16) != c->c_id) {
  i++;}

/*-------------------------------------------*/
/* mark clause                               */
/*-------------------------------------------*/

xl[i][13] = xl[i][0] >> 16; 

mark_more:
c_left = (xl[i][0] & 0x0000FFF0) >> 4;
for(j = 0; xl[j][0] != 0; j++) {
   if(((xl[j][0] & 0xFFFF0000) >> 16) == c_left && xl[j][13] == 0) {
      xl[j][13] = xl[j][0] >> 16;
      break;}}

if(xl[i][1] & 0x0000FFF0) {
   c_right = (xl[i][1] & 0x0000FFF0) >> 4;
   for(j = 0; xl[j][0] != 0; j++) {
      if(((xl[j][0] & 0xFFFF0000) >> 16) == c_right && xl[j][13] == 0) {
         xl[j][13] = xl[j][0] >> 16;
         break;}}}

xl[i][13] = - xl[i][13];
i = 0;
while(xl[i][13] <= 0 && xl[i][0] != 0) i++;
if(xl[i][0] == 0) return;
else goto mark_more;

}

/*--------------------------------------------------------------------*/
/* analyze_decomposeability()       ....................... #16
/*
/* called by:  main()                    theoM.c (theoS.c)
/*
/* calls:      calc_clause_hashcode()    hash.c 
/*--------------------------------------------------------------------*/

void analyze_decomposeability(last_split)
int last_split;
{
clause *c;
literal *l, *lit, prev_lit;
int i, ii, j, k, kk, m, mm, mmm, nn, pp, *p, a, var_rows;
int empty_col, empty_col_count;
int computer_temp;
int lit_count;
int lit_has_var, clause_has_var;
int q, jq, xx, yy;

if(weakened_ply == 1) return;

most_decomposible = -1;
pick_split_pbm = -1;
decomposible_parts = 0;
split_pbs = 0;
unsolved_split_pbs = 0;
decompose2 = -1;

for(i = 0; i < 16; i++) {
   solved_split_pbs[i] = exists_split_pbs[i] = exists_split_lits[i] = 0;
   solved_split_lits[i] = 0;}

for(i = base - nc + 1; i <= base; i++) {
   c = clist[i];
   if(c->c_nlit > decomposible_parts && c->c_nlit > 1) {
      for(k = 0; k < 16; k++) {
         for(j = 0; j < 16; j++) {var_vector[j][k] = 0;}}

      jq = 0;
      lit_count = 0;
      for(lit = c->c_lit; lit != 0;lit = lit->l_next) {
         lit_has_var = 0;
	 p = &lit->l_id;
	 for(a = *p++ & ARGMASK; a > 0; a--) {
	    if(*p & FCNMASK) {a += *p++ & ARGMASK;}
	    else {
	       lit_has_var = 1;
	       p++;}}

/*-------------------------------------------*/
/* if lit has no var, add row                */
/*-------------------------------------------*/

          if(lit_has_var == 0) {
	     var_vector[jq++][lit_count] = 1;}
          lit_count++;}

      for(j = jq; j < c->c_nv + jq; j++) {
         m = 0;
         for(lit = c->c_lit; lit != 0;lit = lit->l_next) {

/*--------------------------------------------*/
/* scan i-th literal for variable j           */
/*--------------------------------------------*/

            p = &lit->l_id;
            for(a = *p++ & ARGMASK; a > 0; a--) {
               if(*p & FCNMASK) {a += *p++ & ARGMASK;}
               else {
                  if((*p & MAX_VARS_M1) == j - jq) {
		     var_vector[j][m] = 1;}
                  p++;}} 
            m++;}}

/*---------------------------------------------*/
/* merge rows with common variables            */
/*---------------------------------------------*/

      var_rows = c->c_nv + jq;
      for(q = 0; q < var_rows - 1; q++) {
pass_again:
         for(m = q + 1; m < var_rows; m++) {
            for(j = 0; j < c->c_nlit; j++) {
               if(var_vector[q][j]==1 && var_vector[m][j]==1) {
                  for(ii = 0; ii < c->c_nlit; ii++) {
                     var_vector[q][ii] |= var_vector[m][ii];}
                  for(nn = m; nn < var_rows - 1; nn++) {      
                     for(pp = 0; pp < c->c_nlit; pp++) {
                        var_vector[nn][pp]=var_vector[nn+1][pp];}}
                  var_rows--;
/*
                  if(this_computer == 0) {
                     printf("VAR_VECTOR....\n");
                     for(mm = 0; mm < var_rows; mm++) {
                        printf("\n");
                        for(kk = 0; kk < c->c_nlit; kk++) {
	                   printf("%d ",var_vector[mm][kk]);}}}
*/
                   if(var_rows == q + 1) goto
	           print_vector;
                   goto pass_again;}}}}
print_vector:

/*----------------------------------------------*/
/* delete empty rows (too many vars).           */
/*----------------------------------------------*/

      for(mm = var_rows - 1; mm >= 0; mm --) {
         clause_has_var = 0;
         for(kk = 0; kk < c->c_nlit; kk++) {
	    if(var_vector[mm][kk] != 0) clause_has_var++;}
         if(clause_has_var == 0) {  /*delete row */
	    for(nn = mm; nn < var_rows - 1; nn++) {
	       for(pp = 0; pp < c->c_nlit; pp++) {
	          var_vector[nn][pp]=var_vector[nn+1][pp];}}
            var_rows--;}}
	    
      
/*----------------------------------------------*/
/* determine number of subclauses               */
/*----------------------------------------------*/

      decompose = var_rows;
      decompose2 = decompose;
      empty_col_count = 0;

      mmm = decompose;

      if(this_computer == 0) 
      for(mm = 0; mm < decompose2; mm++) {
         lit_count = 0;
         for(kk = 0; kk < c->c_nlit; kk++) {
            if(var_vector[mm][kk]) lit_count++;}
         var_vector[mm][kk] = lit_count;}

      if(decomposible_parts < decompose) {
         decomposible_parts = decompose;
         most_decomposible = c->c_id;

         for(mm = 0; mm < decompose2; mm++) {
            for(kk = 0; kk < c->c_nlit; kk++) {
               best_var_vector[mm][kk] = var_vector[mm][kk];}
            best_var_vector[mm][kk] =
	    var_vector[mm][kk];}}}}

/*---------------------------------------*/
/* Print out the final vector array      */
/* Clause called "most_decomposible" is  */
/* the clause to decompose.              */
/*---------------------------------------*/

if(decomposible_parts < 2) {
   if(this_computer == 0) 
       printf("No good clause to decompose!\n");
   most_decomposible = -1;
   return;}

for(mm = 0; mm < decomposible_parts; mm++) {
   for(kk=0; kk < clist[most_decomposible-1]->c_nlit+1; kk++) {}}

/*----------------------------------------*/
/* sort FINAL VECTOR                      */
/*----------------------------------------*/

for(mm = 0; mm < decomposible_parts - 1; mm++) {
   for(nn = mm + 1; nn < decomposible_parts; nn++) {
      if(best_var_vector[nn][kk-1] < best_var_vector[mm][kk-1]) {
         for(pp = 0; pp < clist[most_decomposible - 1]->c_nlit + 1;pp++) {
	    split_temp[pp] = best_var_vector[mm][pp];
            best_var_vector[mm][pp] = best_var_vector[nn][pp];
	    best_var_vector[nn][pp] = split_temp[pp];}}}}

if(this_computer == 0) {
   printf("\nDECOMPOSITION_VECTOR\n");
   for(mm = 0; mm < decomposible_parts; mm++) {
      printf("\n");
      for(kk=0; kk<clist[most_decomposible-1]->c_nlit+1; kk++) {
         printf("%d ",best_var_vector[mm][kk]);}}
   printf("\n");}

printf("DECOMPOSE: ");
print_clause(clist[most_decomposible - 1]);
printf("\n");

/*----------------------------------------*/
/* decide literals to keep for clause     */
/*----------------------------------------*/

unsolved_split_pbs=unsolved_exists_pbs=decomposible_parts;
split_pbs = decomposible_parts;
for(i = 1; i <= split_pbs; i++) {solved_split_lits[i] = 1;}
if(this_computer == 0) return;

if(weakened_ply == 0) 
   pick_split_pbm =
          1+(((this_computer-1)%spawned)%decomposible_parts);
mm = 0;
if(decomposible_parts > 0) c = clist[most_decomposible - 1];

for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   lit->l_id = lit->l_id & ~ELIMLIT;
   if(best_var_vector[pick_split_pbm - 1][mm++] == 0) {
      c->c_nlit--;
      lit->l_id = lit->l_id | ELIMLIT;}}

calc_clause_hashcode(c);

//split_pbs = decompose2;
//for(i = 0; i < split_pbs; i++) {solved_split_lits[i] = 1;}
//printf("C%d splits on problem %d\n",this_computer,pick_split_pbm);fflush(stdout);

computer_temp = this_computer;
this_computer = 0;
print_clause(clist[most_decomposible - 1]);
printf("\n");
this_computer = computer_temp;

return;
} 
