/*----------------------------------------------------------------*/
/*      search.c
/*
/*
/*      functions:
/*      install_clause()        ......................#1
/*      int calc_inferences()   ......................#2
/*      search_tree()           ......................#3
/*      advance_to_successor()  ......................#4
/*      simple_subsume2()       ......................#5
/*      esa2()                  ......................#6
/*      decide_to_add1()        ......................#7
/*----------------------------------------------------------------*/
 
#include <stdio.h>
#include <pvm3.h>
 
#include "const.h"
#include "clause.h"

/*----------------------------------------------------------------*/
/* externally defined functions                                   */
/*----------------------------------------------------------------*/
 
extern void fatal();                /* can be found in theo.c     */
extern int time_now();
 
extern int test_equality();         /* can be found in ....gen.c  */
extern void repair_literal();
extern void add_string();
extern int eval_fxns();

extern int binary_resolvents();  /* infer.c:   inference rules    */
extern int binary_factors();
extern int hash_resolvent();
extern int function_eval();
 
extern void print_clause();      /* tdump.c: print clause fxn     */
extern void print_continuation();
extern int halted_search_options();
extern void print0();

extern int is_clause_in_ht();                      /*   hash.c    */
extern int add_clause_to_ht();
extern int is_unit_in_ht();
extern int is_comp_in_ht();
extern int calc_alt_hashcodes();
extern void calc_alt_hash();
extern void normalize_literal();
extern int calc_lit_hc();
extern void calc_clause_hashcode();
extern int test_alt_hashcodes();
extern int is_test_in_ht();
extern void print_hash_code_list();
 
extern int on_hash_code_list();                      /* pass2.c    */
extern void save_line();
extern int alt_on_hash_code_list();

extern int unify();                                  /* unify.c    */

extern int test_equality_fxn();
extern int fix_proof();

/*-----------------------------------------------------------------*/
/* externally defined constants                                    */
/*-----------------------------------------------------------------*/
 
extern int base, ply, nextc, iter_depth, nc;  /* theo.c     */
extern int org_base;

extern int search_start_time;
extern int search_time;
extern int equal_pred;
extern int equal_clist_index;
extern int units_begin;
extern int phase;
extern int number_constants;
extern int number_const_hash;
 
extern int o_extend, o_substitution, o_hash, in_elim;      
extern int o_print, o_step, x_cr;
extern int o_print_low, o_print_high;

extern int o_merge, o_subsume;
extern int x_stop;      
extern int o_maxtime;
extern int o_maxlit;
extern int o_debug;
extern int s_nod;                          /* # nodes in tree        */
extern int s_nod0;
extern int s_subs_suc;
extern int s_res, s_fac;
extern int s_subsume;                      /* count of s-subsumed cl */
extern int s_htentry;
extern int s_ltentry;
extern int magic_hashx,magic_hashy;
 
extern int hash_code_active_number;                /* pass2.c        */
extern int proof_length;
extern int lit_in_hash;
extern int alt_hx, alt_hy;
extern int v1_count, v2_count;

extern int sub_array_last;                       /* gen.c            */
extern int const_found_equal;

extern int myparent;
extern int this_computer;
extern char *OutPtr;
extern int weakened_ply, weakened_clause;

/*-------------------------------------------------------------------*/
/* clist is a list of ptrs to clauses that lie in clause_space.      */
/* During the search each group of pointers belongs to a particular  */
/* depth. Since the clauses at each depth are ordered by their score */
/* the list is not ordered (i.e. the last clist item does not        */
/* necessarily point to the last clause). However, a group of cls    */
/* at a particular depth are consecutive. Pointers to the clauses    */
/* at each level are in the array depth[].                           */
/*-------------------------------------------------------------------*/
 
extern int *clause_space[];             /* where clauses are stored  */
extern clause *clist[];                 /* pointers to clauses       */
extern dnode depth[];            /* pointers to pointers of clauses  */
extern int hr_hashcodes[3][50];

/*-------------------------------------------------------------------*/
/* internally defined functions                                      */
/*-------------------------------------------------------------------*/
 
int install_clause();
int calc_inferences();
int search_tree();
int advance_to_successor();
int simple_subsume2();
void esa2();
void decide_to_add1();

/*-------------------------------------------------------------------*/
/* internally defined constant                                       */
/*-------------------------------------------------------------------*/
 
int magic_ptr;                  /* ptr to unit clause at end of line */
int alt_HX, alt_HY;
int elim_clauses;
 
/*-------------------------------------------------------------------*/
/*      install_clause(c)             ......................#1
/*
/*      install_clause() is called after the clause has been
/*      generated by generate_clause().  It is pointed to by c.
/*      Install_clause() has the job of deciding whether to
/*      actually list the clause with the other clauses at that
/*      level in the tree.
/*
/*      install_clause adds a clause to the group of clauses
/*      at ply. We assume that this has been properly
/*      initialized. We insert the new clause so that all clauses
/*      at this ply are sorted by the number of literals.
/*      The clause is located at clist[nextc].
/*
/*          clist:      .
/*
/*              [       ] <- d_bc
/*              [       ]
/*              [       ] <- d_tc
/*              [   c   ] <- nextc
/*
/*      called by:  binary_resolvents()         infer.c
/*                  binary_factors()            infer.c
/*                  bres_replace()              elim2.c
/*                  bres_link()                 elim2.c
/*
/*      calls:      simple_subsume2()           search.c
/*                  test_equality()             gen.c
/*                  repair_literal()            gen.c
/*                  add_clause_to_ht()          hash.c
/*                  is_clause_in_ht()           hash.c
/*                  is_comp_in_ht()             hash.c
/*                  calc_clause_hashcode()      hash.c
/*                  calc_alt_hashcodes()        hash.c
/*                  test_alt_hashcodes()        hash.c
/*                  on_hash_code_list()         pass2.c
/*                  alt_on_hash_code_list()     pass2.c
/*                  normalize_literal()         hash.c
/*                  is_test_in_ht()             hash.c
/*                  save_line()                 pass2.c
/*                  print_continuation()        tdump.c
/*
/*      returns     1 if the clause was indeed installed,
/*                  0 if the clause was not installed
/*------------------------------------------------------------------*/
 
int install_clause(c)
clause *c;
{
int i, k;
int comp_lit;
literal *l;
literal *llit;
int aa, *pp;
dnode *d;
literal *l_elim, *l_elim2;
int l_index;
int found, deep_in_tree;
int lits_with_vars;
literal *lit_with_var, *lit_with_var2;
literal *tlit;
int lit_count;
int lit_index, lit_index2, lit_index3;
int c_found;
int h1, h2, ha, hb;
int h1temp, h2temp, h3temp, h4temp, h5temp, h6temp;
int c1_index, c2_index,c3_index;
int repaired;
int constant_val;
int clause_in_table;
int orig_number_lits;
int alt_hx_X, alt_hy_Y;

d = &depth[ply];                      /* insert at current ply     */
comp_lit = 0;
clause_in_table = -1;
orig_number_lits = c->c_nlit;

/*------------------------------------------------------*/
/* if the null clause has been found at this node,      */
/* then lit_in_hash will be 2.  in that case, no        */
/* more clauses should be generated at this node.       */
/*------------------------------------------------------*/

if(lit_in_hash >= 2) return(0);

/*--------------------------------------------------------------*/
/* if the merge option is on and the transposition table option */
/* is on and the clause is not a merge ancestor and the cluase  */
/* is in the extended-search part of the tree, and if the clause*/
/* is found in the clause_hash_table, don't search it           */
/*--------------------------------------------------------------*/

if(o_substitution == 1 && phase) {
   repaired = 0;
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) == 0) {
         while(eval_fxns(c,l)) {
            l->l_id |= EVALMASK;
            repaired++;
	    s_subs_suc++;
            calc_lit_hc(l,c);
            if(c->c_nlit > 1) {
               for(tlit = c->c_lit; tlit != 0; tlit = tlit->l_next) {

/*---------------------------------------------------*/
/* If tautology after evaluation, don't install.     */
/*---------------------------------------------------*/

                  if(tlit == l) {continue;}
                  if(l->l_X == -tlit->l_X &&
                     l->l_Y == -tlit->l_Y) {
                     return(0);}}}}}}

   if(repaired) {
      if(c->c_nlit == 1) {
         if((c->c_lit->l_id & 0x8000FFFF) == equal_pred) {
            if(test_equality(c->c_lit)) {
	       return(0);}}} 
      calc_clause_hashcode(c);}}

if(ply >= iter_depth && o_merge && o_hash) {deep_in_tree = 1;}
else {deep_in_tree = 0;}

/*--------------------------------------------------------------*/
/*  See Section 9.11 in the text
/*--------------------------------------------------------------*/

if(deep_in_tree || (c->c_nlit == 1 && c->c_nv < 2)) {
   if(is_clause_in_ht(c))  {
   clause_in_table = 1;}}

/*---------------------------------------------------------------*/
/* subsumption test:  if the clause is s-subsumed by either a    */
/* base clause or an ancestor clause, delete it from the list    */
/*---------------------------------------------------------------*/

if(in_elim) {
   for(i = base; i >= 0; i--) {             /* try base clauses  */
      if(clist[i]->c_nlit > c->c_nlit) {
         if(i < org_base - nc + 1) {break;}}
      if(clist[i]->c_bits & ACTIVE) {
         if(simple_subsume2(clist[i],c))  {return(0);}}}
   for(i = ply - 1; i > 0; i--) {                 /* try anc. cls */
      if(clist[depth[i].d_tc]->c_nlit <= c->c_nlit) {
         if(simple_subsume2(clist[depth[i].d_tc],c)) {
            return(0);}}}}

/*----------------------------------------------------------------*/
/* examine clause to see if one of its literals is the equality   */
/* predicate. if so, test for Equal(x,x) in clause and if found   */
/* delete the clause. otherwise if not, see if a literal can be   */
/* eliminated because the hash code of its complement is on       */
/* the unit clause hash table.                                    */
/*----------------------------------------------------------------*/

if(c->c_nlit > 1 && phase && lit_in_hash != 2) {
   l_index = 0;
   comp_lit = 0;        
   for(l = c->c_lit; l != 0; l = l->l_next) {        
      l_index++;
      if((l->l_id & 0x8000FFFF) == (equal_pred | FCNMASK)
		      && !(l->l_id & ELIMLIT)) {
         if(test_equality(l) > 0) { 
	    comp_lit = l_index;         /* remove this literal */
	    break;}}
      if((l->l_id & 0x8000FFFF) == equal_pred
		      && !(l->l_id & ELIMLIT)) {
         if(test_equality(l) > 0) { 
	    return(0);}}}                /* delete this clause */

   if(!comp_lit) {
      comp_lit = is_comp_in_ht(c);}
 
/*-----------------------------------------------------*/
/* comp found, its index in cl = comp_lit; delete lit. */       
/*-----------------------------------------------------*/
		
   if(comp_lit) {
      repair_literal(c,comp_lit);
      if(deep_in_tree) {
	 if(is_clause_in_ht(c)) {clause_in_table = 1;}}}}
						
/*----------------------------------------------------------------*/
/* examine clause to see if the complement of one of its literals */
/* is in the clause hash table                                    */
/* if complement of literal is in hash table, eliminate literal,  */
/* replace literal with its hash code for later reevaluation      */
/*----------------------------------------------------------------*/

while(comp_lit && c->c_nlit > 1 
          && phase && lit_in_hash != 2) {
   comp_lit = is_comp_in_ht(c);
   if(comp_lit) {
      repair_literal(c,comp_lit);
      if(deep_in_tree) {        
	 if(is_clause_in_ht(c)) {clause_in_table = 1;}}}}

/*------------------------------------------------------*/
/* see if a 2-lit clause with one literal having        */
/* variables, and the other without any variables       */
/* can be resolved away...                              */
/*                                                      */
/* for example, if c = P(x) | Q(a), then if the hash    */
/* code for P(a) or P(b) etc has been previously found  */
/* then P(x) can be resolved away by a hash table resol */
/*------------------------------------------------------*/

if(c->c_nlit > 1 && lit_in_hash != 2 && phase) {
   lits_with_vars = 0;
   lit_with_var = 0;
   lit_with_var2 = 0;
   lit_count = 0;
   lit_index = 0;
   for(l = c->c_lit; l != 0; l = l->l_next) {
      lit_count++;
      if(!(l->l_id & ELIMLIT)) {
	 if(l->l_id & VARMASK) {
	    lit_with_var2 = lit_with_var;
	    lit_with_var = l;
	    lit_index2 = lit_index;
	    lit_index = lit_count;
	    lits_with_vars++;}}}

   if(lits_with_vars == 1) {
      if(test_alt_hashcodes(lit_with_var,c)) { 

/*---------------------------------------------------------*/
/* Yes, literal can be resolved away, yielding unit clause */
/*---------------------------------------------------------*/

	 lit_with_var->l_X = alt_hx;
	 lit_with_var->l_Y = alt_hy;
	 repair_literal(c,lit_index);}}}

/*---------------------------------------------------------*/
/* Now, if clause has three literals and one variable,     */
/* see if literals can be resolved away by same constant.  */
/*---------------------------------------------------------*/

if(c->c_nlit == 3 && c->c_nv == 1 && lit_in_hash != 2 && phase) {
   c1_index = 0;
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if(c1_index != 0) {break;}
      if(l->l_id & ELIMLIT) {continue;}
      if(!(l->l_id & VARMASK)) {continue;}
      c1_index++;
      for(c_found = 1; c_found <= number_constants; c_found++) {
         calc_alt_hash(l,0,-1,c_found,0);
         if(is_test_in_ht(-alt_hx,-alt_hy)) {
            l_elim = l;h1 = alt_hx;h2=alt_hy;
	    c2_index = c1_index;
next_lit1:  c2_index++;
	    if(l->l_next) {
               l = l->l_next;
               if(l->l_id & ELIMLIT) {goto next_lit1;}
   	       if(l->l_id & VARMASK) {
                  calc_alt_hash(l,0,-1,c_found,0);
		  if(is_test_in_ht(-alt_hx,-alt_hy)) {
                     l_elim2 = l;ha = alt_hx;hb = alt_hy;
		     c3_index = c2_index;
next_lit2:           c3_index++;
		     if(l->l_next) {
		        found = 0;
                        l = l->l_next;
                        if(l->l_id & ELIMLIT) {goto next_lit2;}
   	                if(l->l_id & VARMASK) {
                        calc_alt_hash(l,0,-1,c_found,0);
		        if(is_test_in_ht(-alt_hx,-alt_hy)) {
			   /*print0("&");*/
			   l_elim->l_X = h1; l_elim->l_Y = h2;
			   l_elim->l_id += c_found << 20;
			   l_elim2->l_X = ha; l_elim2->l_Y = hb;
			   l->l_X = alt_hx; l->l_Y = alt_hy;
			   l_elim->l_id |= ELIMLIT;
			   l_elim2->l_id |= ELIMLIT;
			   l->l_id |= ELIMLIT;
			   c->c_nlit = 0;
			   calc_clause_hashcode(c);
			   lit_in_hash = 2;
			   c->c_bits |= MERGE;}
		        else { 
			   if(hash_code_active_number) {
			      found = on_hash_code_list(alt_hx,alt_hy);
			      if(found) {
                                 h1temp = l_elim->l_X;
				 h2temp = l_elim->l_Y;
				 l_elim->l_X = h1;
				 l_elim->l_Y = h2;
				 h3temp = l_elim2->l_X;
				 h4temp = l_elim2->l_Y;
				 l_elim2->l_X = ha;
				 l_elim2->l_Y = hb;
				 h5temp = l->l_X;
				 h6temp = l->l_Y;
				 l->l_X = alt_hx;
				 l->l_Y = alt_hy;
				 l_elim->l_id |= ELIMLIT;
				 l_elim2->l_id |= ELIMLIT;
			         l_elim->l_id += c_found << 20;

/*--------------------------------------------*/
/* replace variable (x) with constant         */
/*--------------------------------------------*/

                                for(llit=c->c_lit;llit!=0;llit=llit->l_next) {
			           pp = &llit->l_id;
			           for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
			              if(*pp & FCNMASK) {
			                 aa += *pp++ & ARGMASK;}
                                      else {
			                 *pp++ = (c_found << 5) | NEGMASK;}}}

				 calc_clause_hashcode(c);
		                 alt_hx_X = alt_hx;alt_hy_Y = alt_hy;
			         save_line(c,1);
		                 alt_hx = alt_hx_X;alt_hy = alt_hy_Y;
				 l_elim->l_id &= ~ELIMLIT;
				 l_elim2->l_id &= ~ELIMLIT;
				 l->l_X = h5temp;
				 l->l_Y = h6temp;
				 l_elim2->l_X = h3temp;
				 l_elim2->l_Y = h4temp;
				 l_elim->l_X = h1temp;
				 l_elim->l_Y = h2temp;
				 calc_clause_hashcode(c);
			         if(!hash_code_active_number) {
				    lit_in_hash = 2;}}}}
		        add_clause_to_ht(alt_hx,alt_hy,1);
		        if(lit_in_hash == 2) {goto finish_install;}
			else goto next_lit2;}}}}}}}}}
		



/*---------------------------------------------------------*/
/* Now, if clause has two literals
/*---------------------------------------------------------*/

if(c->c_nlit == 2 && lit_in_hash != 2 && phase) {
   esa2(c);
   if(c->c_bits & EQCLS) {c->c_bits ^= ACTIVE;}}

/*---------------------------------------------------------*/
/* Now, if clause has two literals and one variable,       */
/*---------------------------------------------------------*/

if(c->c_nlit == 2 && c->c_nv == 1 && lit_in_hash != 2 && phase) {
   c1_index = 0;
   for(l = c->c_lit; l != 0; l = l->l_next) {
      c1_index++;
      if(l->l_id & ELIMLIT) {continue;}
      if(!(l->l_id & VARMASK)) {continue;}
      for(c_found = 1; c_found <= number_constants; c_found++) {
	 calc_alt_hash(l,0,-1,c_found,0);
         if(is_test_in_ht(-alt_hx,-alt_hy)) {
	    l_elim = l;
	    h1 = alt_hx;h2 = alt_hy;
            c2_index = c1_index;
            next_lit:
	    c2_index = c2_index + 1;
	    if(l->l_next) {
	       l = l->l_next;
               if(l->l_id & ELIMLIT) {goto next_lit;}
   	       if(l->l_id & VARMASK) {
                  calc_alt_hash(l,0,-1,c_found,0);
                  if(is_test_in_ht(-alt_hx,-alt_hy)) {
		     /*print0("@");*/
                     l_elim->l_X = h1;l_elim->l_Y = h2;
		     l->l_X = alt_hx;l->l_Y = alt_hy;
		     l->l_id |= ELIMLIT;
                     l_elim->l_id |= ELIMLIT;
		     l_elim->l_id += c_found << 20;
                     c->c_nlit = 0;

                     calc_clause_hashcode(c);
                     lit_in_hash = 2;
                     l_elim->l_X = h1;l_elim->l_Y = h2;
                     c->c_bits |= MERGE;}
		  else { 
		     found = 0;
                     if(hash_code_active_number) {
                        add_clause_to_ht(alt_hx,alt_hy,1);
                        found = on_hash_code_list(alt_hx,alt_hy);
                        if(found) {
                           h1temp = l_elim->l_X;
                           h2temp = l_elim->l_Y;
                           l_elim->l_X = h1;
                           l_elim->l_Y = h2;
                           h3temp = l->l_X;
                           h4temp = l->l_Y;
                           l->l_X = alt_hx;
                           l->l_Y = alt_hy;
                           l_elim->l_id |= ELIMLIT;
			   c->c_nlit = 1;

/*--------------------------------------------*/
/* replace variable (x) with constant         */
/*--------------------------------------------*/


                           for(llit=c->c_lit;llit!=0;llit=llit->l_next) {
	                      pp = &llit->l_id;
			           for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
			              if(*pp & FCNMASK) {
			                 aa += *pp++ & ARGMASK;}
                                      else {
			                 *pp++ = (c_found << 5) | NEGMASK;}}}


		           l_elim->l_id += c_found << 20;
                           calc_clause_hashcode(c);
			   /*print0("%");*/
		           alt_hx_X = alt_hx;alt_hy_Y = alt_hy;
			   save_line(c,1);
		           alt_hx = alt_hx_X;alt_hy = alt_hy_Y;
                           l_elim->l_id &= ~ELIMLIT;
                           l->l_X = h3temp;
                           l->l_Y = h4temp;
                           l_elim->l_X = h1temp;
                           l_elim->l_Y = h2temp;
			   c->c_nlit = 2;
                           calc_clause_hashcode(c);
                           if(!hash_code_active_number) {
                              lit_in_hash = 2;}}}
                    if(found == 0) {
		       add_clause_to_ht(alt_hx,alt_hy,1);
		       goto finish_install;}}}}}}}}

/*---------------------------------------------------------------*/
/* if there is only one lit in the clause execute the following  */
/*---------------------------------------------------------------*/

if(c->c_nlit == 1 && phase && lit_in_hash != 2) {

/*------------------------------------------*/
/* find first real literal                  */
/*------------------------------------------*/
 
   for(l = c->c_lit; l->l_id & ELIMLIT ; l = l->l_next) {;}

   if(c->c_nv) {
      normalize_literal(l,c);} 
   if(c->c_nv && l != c->c_lit) {
      calc_lit_hc(l,c);
      calc_clause_hashcode(c);}
   lit_in_hash = add_clause_to_ht(l->l_X,l->l_Y,1);

   if(lit_in_hash != 2) {
      lit_in_hash = calc_alt_hashcodes(l,c);}


   if(!hash_code_active_number && (lit_in_hash == 2)) {
 
/*---------------------------------------*/
/* in phase 1:if contradiction found.... */
/*---------------------------------------*/
 
      c->c_bits |= MERGE;}        /* make a merge clause */
		
   if(hash_code_active_number) {  
 
/*----------------------------------------------*/
/* during phase 2: when looking for complements */
/*----------------------------------------------*/

      found = on_hash_code_list(l->l_X,l->l_Y);
      
      if(!found) {
         found = alt_on_hash_code_list(l,c);}

      if(found) { 
	 /*print0("#");*/
         save_line(c,0); 
         if(!hash_code_active_number) {
            lit_in_hash = 2;}}}} 

/*----------------------------------------------------------*/
/* determine where the next clause will be placed           */
/*----------------------------------------------------------*/

if(clause_in_table == 1 && c->c_nlit == orig_number_lits 
          && c->c_nlit != 0 && lit_in_hash == 0) {return(0);}

finish_install:
d->d_tc = nextc++; 
clist[nextc] = c->c_end;

/*----------------------------------------------------------*/
/* We insert the new clause so that clauses with fewer      */
/* literals are considered first.                           */
/* However, we skip this part of the code while in elim     */
/*----------------------------------------------------------*/
 
if(in_elim) {
   for(k = d->d_tc; k >= d->d_bc; k--) {
      if(k == d->d_bc) {break;}
      if(clist[k-1]->c_nlit >= c->c_nlit) {break;}
         clist[k] = clist[k-1];}}
else k = d->d_tc;
 
/*-----------------------------------------------------------*/
/* If too many clauses are generated, quit search            */
/* The maximum number is CSPACE_CLAUSES                      */
/* This constant is defined in const.h                       */
/*-----------------------------------------------------------*/

if(k > CSPACE_CLAUSES - 2) {
   printf("\nToo many clauses(%d) \n",k);
   for(i = 0; i <= ply; i++) {
      d = &depth[i];
      printf("\nPly %ld clauses: %d ",i, d->d_tc);}
   print_continuation();
   getchar();
   fatal("\nQuitting search\n");}
 
if((unsigned int) c->c_end > (unsigned int) d->d_rc) {
   d->d_rc = c->c_end;}
 
/*------------------------------------------------------------*/
/* if either the left or right parent is in the NC, set this  */
/* flag in c_bits                                             */
/*------------------------------------------------------------*/

if(phase) { 
if(c->c_lpar->c_bits & NC) {
   c->c_bits |= NC;}
else if(c->c_rpar) { 
   if(c->c_rpar->c_bits & NC) {
      c->c_bits |= NC;}}}
 
/*------------------------------------------------------------*/
/* place clause on clist and add it to hash table             */
/*------------------------------------------------------------*/

clist[k] = c;
if(c->c_nlit > 1) {
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);}

decide_to_add1(c);
if(c->c_nlit == 0) lit_in_hash = 2;
if(lit_in_hash == 2) { 
   magic_ptr = k;}

return(1);
}
  
/*-----------------------------------------------------------------*/
/*      calc_inferences()                ......................#2
/*
/*      calc_inferences will generate all inferences of a clause
/*      with all base clauses and all ancestor clauses (depending
/*      on the options) at the current ply. We return the number
/*      of resolvents generated.
/*
/*      called by:  search_tree()           search.c
/*
/*      calls:      binary_resolvents()     infer.c
/*                  binary_factors()        infer.c
/*                  is_unit_in_ht()         hash.c
/*                  simple_subsume2()       search.c
/*
/*      returns:    0,nc
/*-----------------------------------------------------------------*/
 
int calc_inferences()
{
clause *c;
literal *l;
int i, j;
int opt;
clause *c1;
dnode *d;
int number_of_inferences;
int number_of_factors;
int iter_horizon, merge_depth, nc_depth;
int start_clause;
int c1_index, constant_val;
int new_inferences;
int ind;

number_of_inferences = 0;
number_of_factors = 0;

hr_hashcodes[0][ply] = 0;
hr_hashcodes[1][ply] = 0;
hr_hashcodes[2][ply] = 0;

/*--------------------------------------------------------------*/
/* if the extended search option is off and ply > iter_depth    */
/* don't generate any clauses at this node                      */
/*--------------------------------------------------------------*/

if(!o_extend) {
   if(ply > iter_depth) {return(0);}}
		
/*--------------------------------*/
/* c = clause to expand           */
/*--------------------------------*/
 
c = clist[depth[ply - 1].d_tc];        

/*--------------------------------*/
/* check for literal subsumption  */
/*--------------------------------*/

if(c->c_nlit > 1) {
   if((ind = is_unit_in_ht(c)) > 0) {
      return(0);}}

/*---------------------------------*/
/* adjust search horizon taking    */
/* into account whether the node   */
/* is a merge or a nc descendent   */
/*---------------------------------*/

if(ply > 1) {
   merge_depth = (c->c_bits & MERGE) >> 2;
   nc_depth = (c->c_bits & NC) >> 8;
   nc_depth = nc_depth + merge_depth;
   if (nc_depth > 1) { nc_depth = 1;}
   iter_horizon = iter_depth + nc_depth;}
else {
   iter_horizon = iter_depth;}
 
opt = (ply  <= iter_horizon) ?
	  (o_merge * MERGE) : UNIT;

/*--------------------------------------------------*/
/* next, we generate all factors of this clause     */
/*--------------------------------------------------*/

if(c->c_bits & ACTIVE && c->c_nlit > 1) {
   number_of_factors = binary_factors(c);
   number_of_inferences += number_of_factors;}

/*--------------------------------------------------*/
/* if c is a base clause generate all resolvents    */
/* only with clauses with small clause id's         */
/* otherwise generate resolvents with all base      */
/* clauses                                          */
/*--------------------------------------------------*/

if(ply > iter_horizon) {
   start_clause = units_begin;}
else {
   start_clause = 0;}

if(c->c_bits & BASE) {
   for(i = c->c_id - 2 ; i >= 0 ; i--) {
      if(lit_in_hash == 2) {break;}
      number_of_inferences += binary_resolvents(c,clist[i],opt);}}
else {
   for(i = base; i >= start_clause; i--) {
      if(lit_in_hash == 2) {break;}
      number_of_inferences += binary_resolvents(c,clist[i],opt);}}

/*----------------------------------------------------------*/
/* now that the base resolvents are in, we can generate     */
/* the resolvents of this clause with its ancestors         */
/*----------------------------------------------------------*/

for(i = ply - 2; i > 0; i--) {
   if(lit_in_hash == 2) {break;}
   number_of_inferences += 
	  binary_resolvents(c,clist[depth[i].d_tc],opt);}

/*-----------------------------------------------------------*/
/* if no inferences generated thus far at this node, generate*/
/* instantiated clause, if possible.  See if litteral can    */
/* be instantiated and then resolved away.                   */
/*-----------------------------------------------------------*/

if(ply > iter_depth && number_of_inferences == 0) {
   new_inferences = number_of_inferences;
   if(c->c_nlit == 2 && phase && lit_in_hash != 2) {
      if(c->c_bits & ACTIVE) {
         c1_index = 0;
         for(l = c->c_lit; l != 0; l = l->l_next) {
            if(l->l_id & ELIMLIT) {continue;}
            c1_index++;
            alt_HX = 0;
	    if(constant_val = test_alt_hashcodes(l,c)) {
	       if(v1_count != 0 && v2_count == 0) {
	          constant_val = constant_val & 0x0000FFFF;}
	       else if(v1_count == 0 && v2_count == 0) {
	          v1_count = 1;
	          constant_val = constant_val + 1;}
	       else if(v1_count == 0 && v2_count != 0) {
	          v1_count = v2_count;
	          v2_count = 0;
	          constant_val = constant_val>>16;}
               alt_HX = alt_hx;alt_HY = alt_hy;
               number_of_inferences +=
	          hash_resolvent(c,l,c1_index,constant_val);
               if(number_of_inferences != new_inferences) {
                  alt_HX = 0;}
               break;}}}}}

d = &depth[ply];

/*--------------------------------------------------*/
/* delete s-subsumed clauses by other clauses       */
/* at this node                                     */
/*--------------------------------------------------*/

for(i = d->d_bc ; i < d->d_tc ; i++) {
   if(clist[i]->c_bits & ACTIVE) {
      for(j = i + 1; j <= d->d_tc; j++) {
	     if(clist[j]->c_bits & ACTIVE) {
	        if(simple_subsume2(clist[j],clist[i])) {
	           clist[i]->c_bits = (clist[i]->c_bits&(~ACTIVE));
	           break;}}}}}

/*---------------------------------------------------*/
/* if in extended search and clause c has n literals */
/* where n is greater than 2, and the number of      */
/* inferences generated is only one (so no units     */
/* can be produced, terminate search...........      */
/*---------------------------------------------------*/

if(ply > iter_depth) {
   if(c->c_nlit > 2) {
      if(number_of_inferences  == 1) { 
         c1 = clist[d->d_bc];
            if(c1->c_bits & ACTIVE) {
               if(c1->c_nlit >= c->c_nlit - 1) {
                  c1->c_bits = c1->c_bits & (~ACTIVE);}}}}}
 
return(number_of_inferences);
}                                                       
 
/*----------------------------------------------------------------*/
/*      search_tree()                  ......................#3
/*
/*      search will return TRUE if we have generated the
/*      empty clause during our explorations of the tree.
/*      Note: we assume that there are at least two base clauses
/*
/*      called by:  iterative_deepening()             theo.c
/*
/*      calls:      calc_inferences()                 search.c
/*                  advance_to_successor()            search.c
/*                  halted_search_options()           tdump.c
/*                  print_clause()                    tdump.c
/*
/*      returns:    1 if proof found on current iteration
/*                  0 otherwise
/*----------------------------------------------------------------*/
 
int search_tree()
{
int adv_val, number_of_clauses;
dnode *d;
clause *c; 

depth[0].d_tc = base;               /* initial generating clause  */
depth[0].d_ec = base;
depth[0].d_bc = 0;
x_stop = 0;
nextc = base + 1;
clist[nextc] = depth[0].d_rc = clist[base]->c_end;
if(s_nod == 0) elim_clauses = 0;

for(ply = 1;;) {
   d = &depth[ply];             /* current depth node   */
		
/*------------------------------------------------------*/
/* we generate all possible resolvents and factors      */
/* with the current ply's generating clause             */
/*                                                      */
/* x_stop is normally 0, but can be set to 1 during     */
/* the search by the user when searching the tree       */
/* in an interactive mode                               */
/*------------------------------------------------------*/

   number_of_clauses = 0;
   if(!x_stop) {            
      nextc  = depth[ply - 1].d_ec + 1;
      clist[nextc] = depth[ply - 1].d_rc;
      d->d_rc = 0;
      d->d_tc = d->d_bc = nextc;
      number_of_clauses = calc_inferences();

      if(!number_of_clauses) {
         d->d_tc = d->d_bc;}
      d->d_ec = depth[ply - 1].d_ec + number_of_clauses;}
   x_stop = 0;
	
   if(!number_of_clauses) {
      while((adv_val = advance_to_successor())) {

/*-------------------------------------------------*/
/* if adv_val=2, proof found, quit search          */
/* if adv_val=1 and search is at root => no proof  */
/* if adv_val=0, exit while and search this branch */
/*               deeper                            */
/*-------------------------------------------------*/  

         if(adv_val == 2) {
            lit_in_hash = 0; 
            depth[ply].d_tc = magic_ptr;
            return(1);}
         else if(adv_val == 3) {
	    if(x_stop == 2) {return(0);}
	    return(3);}
         else if(ply-- <= 1) {return(0);}
	 nextc = depth[ply].d_tc;              
	 clist[nextc] = depth[ply - 1].d_rc;}
      continue;}
 
/*--------------------------------------------------*/
/* continue deeper into tree                        */
/*--------------------------------------------------*/

   s_nod++;

   if(!clist[d->d_tc]->c_rpar) {s_fac++;}
   else {s_res++;} 

/*--------------------------------------------------*/
/* every 512 nodes searched, check the time to see  */
/* whether to stop. Don't stop in phase 2.          */
/*--------------------------------------------------*/

   if((s_nod % 512) == 0 && myparent == 0) {
      if(pvm_nrecv(-1,MPROOF) > 0) {
	 if(fix_proof() != 0) {
	    return(5);}}}

   if(((s_nod % 512) ==  0) && phase < 2) {
      if(time_now() - search_start_time >= o_maxtime) {
	 return(2);}} 
   if(ply <= o_print)  {
      if(s_nod >= o_print_low && s_nod <= o_print_high) {
         x_cr = 1;
         print_clause(clist[d->d_tc]);
         x_cr = 0;
         if(ply <= o_step) {
	    if(halted_search_options() == 1) {return(3);}}}
      if(x_stop == 2) {return(0);}}
 
/*--------------------------------------------------*/
/* since the empty clause has the highest score     */
/* we check the top clause of the current ply       */
/* to see if it is IT.                              */
/*--------------------------------------------------*/
	
   if(!clist[d->d_tc]->c_nlit) {return(1);}
		
   if(clist[d->d_tc]->c_nlit == 1) {
      if(lit_in_hash == 2) { 
         depth[ply].d_tc = magic_ptr;
	 lit_in_hash = 0;             /* complement in hash table  */
	 return(1);}}                 /* proof found               */
 
   ply++;}                            /* inc. search depth         */
}                                                       
 
/*-----------------------------------------------------------------*/
/*      advance_to_successor()          .....................#4
/*
/*      advance will modify a depth node so that we can consider
/*      the next sub-AFF tree. We return TRUE if we cannot
/*      advance the pointer and a backup is in order
/*
/*      p  = ply - 1
/*
/*      called by:  search_tree()       search.c
/*
/*      calls:      print_clause()      tdump.c
/*
/*      returns:    0 => more branches to consider at this node
/*                  1 => no more branches to consider
/*                  2 => proof found either as a null clause
/*                       or a complement in hash table
/*------------------------------------------------------------------*/
 
int advance_to_successor()
{
dnode *d;
int k, hso, litl_index, same_index;

d = &depth[ply - 1];

if(lit_in_hash == 2) {
   depth[ply].d_tc = magic_ptr;
   if(ply != 1) {
      if(!clist[d->d_tc]->c_rpar) {s_fac++;} 
      else {s_res++;}
      if(ply - 1 <= o_print) {
	 ply--; 
         x_cr = 1;
	 if(s_nod >= o_print_low && s_nod <= o_print_high) {
            print_clause(clist[d->d_tc]);}
         x_cr = 0;
         ply++;}}
   return(2);}

if(d->d_tc == d->d_bc) {return(1);}

d->d_tc--;                  /* decrement pointer to next clause     */

litl_index = 0;
if(d->d_tc == d->d_ec - 1 && clist[d->d_bc]->c_nlit > 2) {
   litl_index = clist[d->d_ec]->c_indices >> 6;
   for(k = d->d_bc; k < d->d_ec; k++) {
      if(clist[k]->c_indices >> 6 != litl_index) {
         litl_index = -1;}}}
if(litl_index == 2) {
   elim_clauses += d->d_ec - d->d_bc + 1;
   if(ply > 1) return(1);}
 
if(ply != 1) {
   if(!clist[d->d_tc]->c_rpar) {s_fac++;} else {s_res++;}
   if(ply - 1 <= o_print) {
      if(s_nod >= o_print_low && s_nod <= o_print_high) {
         ply--; 
         x_cr = 1;
         print_clause(clist[d->d_tc]);
         x_cr = 0;
	 hso = 0;
         if(ply <= o_step) {
            hso = halted_search_options();}
         ply++;
         if(hso == 1) {return(3);}
         if(x_stop == 2) {return(3);}}}
   if(!clist[d->d_tc]->c_nlit) {return(2);}}

return(0);
}
 
/*------------------------------------------------------------------*/
/*      simple_subsume2(c1,c2)         ......................#5
/*
/*      called by:      install_clause()    search.c
/*                      calc_inferences()   search.c
/*
/*      calls:          nobody
/*
/*      Returns "1" if c1 subsumes in the simplest way c2,
/*      ie., s-subsumes (test to see if each literal in c1
/*      appears in c2. Actually hash codes of literals are checked)
/*------------------------------------------------------------------*/
 
int simple_subsume2(c1,c2)
clause *c1,*c2;
{
literal *l1,*l2;               /* pointers to literals in c1, c2    */
 
for(l1 = c1->c_lit ; l1 != 0; l1 = l1->l_next) {   /* scan c1 lit s */
   if(!(l1->l_id & ELIMLIT)) {
 
/*--------------------------------------*/
/* test to see if hashcode of l1 is     */
/* the same as the hashcode of some     */
/* literal in c2. if so, set lit_found  */
/* equal to 1                           */
/*--------------------------------------*/
 
      for(l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
         if(!(l2->l_id & ELIMLIT)) {
	    if(l1->l_X == l2->l_X && l1->l_Y == l2->l_Y) {
	       break;}}}
 
/*--------------------------------------*/
/* if no literal in l2 has hashcode of  */
/* l1, return with "no s-subsumption    */
/*--------------------------------------*/
 
      if(!l2)  {return(0);}}}

/*---------------------------------------------*/
/* if no simple subsumption strat.,don't elim. */
/*---------------------------------------------*/
 
if(!o_subsume) {return(0);}

s_subsume++; 
return(1);
}

/*---------------------------------------------*/
/* esa2() ...........................#6
/*---------------------------------------------*/

void esa2(c)
clause *c;
{
literal *l1, *l2;
int v1, v2, *p;

l1 = c->c_lit;
l2 = l1->l_next;

if((l1->l_id & 0x0000FFFF) != equal_pred) {return;}
if((l2->l_id & 0x0000FFFF) != equal_pred) {return;}
if(l1->l_id ^ l2->l_id ^ NEGMASK) {return;}

if((l1->l_id & NEGMASK) == 0) {
   l1 = l2;
   l2 = c->c_lit;}
					    
/*-------------------------------------------------------*/
/*   clause with two equal lits. l1 is the "~" one.      */
/*   now see if l1 is of the form equal(x,y)             */
/*-------------------------------------------------------*/

p = &l1->l_id;
p++;
v1 = *p++;
v2 = *p;
if(((v1&FCNMASK) == 0) && ((v2&FCNMASK) == 0 && v1 != v2)) {
   if(test_equality_fxn(l2,v1,v2)) {
      c->c_bits |= EQCLS;}}

return;
}

/*-----------------------------------------------------------*/
/* decide_to_add1()  .................#7
/*-----------------------------------------------------------*/

void decide_to_add1(c)
clause *c;
{
literal *l;
int a, *p;
int string_length, string_length2;
int *cp, *const_ptr, *const_ptr2, *const_ptr3;
int v_found, sl;

if(c->c_nlit == 1 && sub_array_last <  65000
		&& phase > 0 && o_substitution == 1) {
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) == 0) {break;}}
   if((l->l_id & 0x8000FFFF) == equal_pred) {

/*-----------------------------------------------------------*/
/* equalty predicate -- looking at literal l ....            */
/* now determine the number of characters in the left        */
/* and right side of the predicate. store in string_length   */
/* and string_length2.                                       */
/*-----------------------------------------------------------*/

      string_length = 0; string_length2 = 1;

      p = &l->l_id;
      for(a = *p++ & ARGMASK; a > 1; a--) {
         if(string_length == 0) {const_ptr = p;}
	 string_length++;
         if(*p & FCNMASK) {
            a += *p++ & ARGMASK;}
         else {
            p++;}}
      const_ptr2 = p;
      if((*p & FCNMASK) == 0) {
         p++;}
      else {
         for(a = *p++ & ARGMASK; a > 0; a--) {
	    string_length2++;
            if(*p & FCNMASK) {
               a += *p++ & ARGMASK;}
	    else {p++;}}}

      if(string_length * string_length2 == string_length ||
           string_length * string_length2 == string_length2) {

/*--------------------------------------------------*/
/* if const_ptr is a var, make sure var appears     */
/* in right side of string, and similarly, if       */
/* const_ptr2 points to a var....                   */
/*--------------------------------------------------*/

            v_found = 0;
            if((*const_ptr & FCNMASK) == 0) {
               v_found++;
               cp = const_ptr2;
               for(sl = 1; sl <= string_length2; sl++) {
                  if(*cp++ == *const_ptr) {
                     v_found = 0;
		     break;}}}
            else if((*const_ptr2 & FCNMASK) == 0) {
               v_found++;
               cp = const_ptr;
               for(sl = 1; sl <= string_length; sl++) {
                  if(*cp++ == *const_ptr2) {
                     v_found = 0;
		     break;}}}

	 if(string_length*string_length2 != 1) { 
            if(v_found == 0) {
               if(string_length == 1) {
	          add_string(c,const_ptr,const_ptr2,string_length2);}
	       else {

/*----------------------------------------------------------*/
/* otherwise, replace left side with constant on right side */
/*----------------------------------------------------------*/

                 add_string(c,const_ptr2,const_ptr,string_length);}}}}}}
return;
}
