/*-----------------------------------------------------------------*/
/*      tdump.c
/*
/*
/*      functions:
/*
/*      the following three functions print the clause
/*
/*      print_arguments()       .......................#1
/*      print_clause()          .......................#2
/*      print_clause2()         .......................#2a
/*      print_all_clauses()     .......................#2b 
/*
/*      information on the options available are printed out by:
/*
/*      help_halted_search()       ....................#3
/*
/*      user can request information or change options or
/*      guide the search by choosing various options.
/*
/*      halted_search_options() .......................#4
/*      print_continuation()    .......................#5
/*      dump_calloc_info()      .......................#6
/*
/*      prints a string to the screen and to the disk
/*
/*      print0()                .......................#7
/*
/*      print_substitutions()   .......................#8
/*
/*      stat_dump()             .......................#9
/*
/*-----------------------------------------------------------------*/
 
#include "const.h"
#include "clause.h"
 
#include <stdio.h>

/*------------------------------------------------------------------*/
/* externally defined functions                                     */
/*------------------------------------------------------------------*/
 
extern void print_cl_array();                     /* in pass2.c     */
extern void print_hash_code_list(); 
extern int time_now();                            /* in theo.c      */
extern void fatal();
extern int grab_number();

/*------------------------------------------------------------------*/
/* externally defined constants                                     */
/*------------------------------------------------------------------*/
 
extern char input_buffer[];
extern clause *clist[];              /* pointers to clauses         */
extern dnode depth[];            /* pointers to pointers of clauses */
extern int cl[PROOF_ARRAY_SIZE][14];
extern int sub_array[65536][6];
extern int sub_array_last;
extern pass_number;
extern int questionable[2000];
extern int phase, big_bug;
extern int o_step, o_print, x_cr;
extern int base, ply, iter_depth, s_nod, s_res, s_fac;
extern int s_nod0, org_base;
extern int s_ltentry;
extern int s_htentry, s_hthit, s_htfail, s_subs_suc;
extern int s_subfxn, s_too_many_terms;
extern int x_stop, o_debug, o_maxlit;
extern int o_maxterm;
extern char *pidtbl[], *fidtbl[];
extern int search_start_time;
extern int search_time;
extern int proof_length;
extern int tab_amount;       /* number of tabs before printing cls  */
extern int equal_pred;
extern int this_computer;
 
extern char *OutPtr;

int *print_arguments();
void print_clause();
void print_clause2();
void print_all_clauses();
void help_halted_search();
int halted_search_options();
void print_continuation();
void dump_calloc_info();
void print0();
void print_substitutions();
void stat_dump();

/*------------------------------------------------------------------*/
/*      print_arguments(p)          .......................#1
/*
/*      print the arguments of the predicate
/*
/*      called by:    print_clause()           tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      Nil
/*------------------------------------------------------------------*/
 
int *print_arguments(p,terms)
int *p, terms;
{
int a, p_prime, period;
static char var_[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";
 
period = 0;
if((*p & FCNMASK) == 0) {
	
/*------------------*/
/* print a variable */  
/*------------------*/
	
   p_prime = *p;        
   if(p_prime >= MAX_VARS) {
      p_prime = p_prime - MAX_VARS;}            /* standardize */
   if(this_computer == 0) {
      printf("%c",var_[p_prime]);} 
   OutPtr += sprintf(OutPtr,"%c",var_[p_prime]);
   terms--;
   if(terms == 0) {
      if(this_computer == 0) {
         printf("]");}
      OutPtr += sprintf(OutPtr,"]");}
   else {
      if(this_computer == 0) {
         printf(",");}
      OutPtr += sprintf(OutPtr,",");}
   return(++p);}
    
/*------------------*/
/* print a function */  
/*------------------*/
 
if(this_computer == 0) {
   printf("%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);}
OutPtr += sprintf(OutPtr,"%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);
	

if((*p&ARGMASK) != 0) {
   if(this_computer == 0) printf("[");
   OutPtr += sprintf(OutPtr,"[");}

terms--;

/*------------------------------------------------*/
/* recursively print the remaining arguments of p */
/*------------------------------------------------*/
 
for(a = *p++ & ARGMASK; a > 0; a--)  {
   p = print_arguments(p,a);}

if(terms == 0) {
   if(this_computer == 0) printf("]");
   OutPtr += sprintf(OutPtr,"]");}
else {
   if(this_computer == 0) printf(",");
   OutPtr += sprintf(OutPtr,",");}

return(p);
}
 
/*-----------------------------------------------------------------*/
/*      print_clause(c)               .......................#2
/*
/*      called by:      write_out_results()       theo.c
/*                      search_tree()             search.c
/*                      advance_to_successor()    search.c
/*
/*      calls:          print_arguments()         tdump.c
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void print_clause(c)
clause *c;
{
int i, j, *a, litend;
int substitution_info, con1, var1, con2, var2;
literal *l, *litrl;
static char AB[] = ".abcdefghijklmnopqrstuvwxyz";
static char varchars[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";

/*--------------------------------------------------------*/
/* decide how far to tab before beginning to print clause */
/*--------------------------------------------------------*/

if((c->c_bits & BASE) == 0 && c->c_lpar != 0) {
   if(c->c_lpar->c_id != c->c_id - 1 && c->c_id > base + 1) {
      tab_amount = c->c_id - base - 2;}}
else {
   tab_amount = 0;}
 
if((c->c_bits & BASE) == 0) {
   for(i = c->c_id - base - 2 - tab_amount; i > 0; i--) {
      if(this_computer == 0) {
         printf(" ");}
      sprintf(OutPtr++," ");}}

if(this_computer == 0) {
   printf(" ");} 
sprintf(OutPtr++," ");
				
if((c->c_bits & ACTIVE) != 0) {
   if((c->c_bits & NC) != 0) {
				
/*--------------------------------------*/
/* active clause in NC; denote by S     */
/*--------------------------------------*/
 
      if((c->c_bits & BASE) == BASE) {
         if((c->c_bits & LIST) == LIST) {
	    if(this_computer == 0) {
	       printf("%2dS>",c->c_id);}
	    OutPtr += sprintf(OutPtr,"%2dS>",c->c_id);}
         else {
	    if(this_computer == 0) {
	       printf("%2dS ",c->c_id);}
	    OutPtr += sprintf(OutPtr,"%2dS ",c->c_id);}}
      else {
	 if(this_computer == 0) {
            printf("%2dS>",c->c_id);}
	 OutPtr += sprintf(OutPtr,"%2dS>",c->c_id);}}
   else {               
 
/*--------------------------------------*/
/* active clause not in NC; leave " "   */
/*--------------------------------------*/
 
      if((c->c_bits & BASE) == BASE) {
         if((c->c_bits & LIST) == LIST) {
            if(this_computer == 0) {
	       printf("%2d >",c->c_id);}
	    OutPtr += sprintf(OutPtr,"%2d >",c->c_id);}
	 else {
	    if(this_computer == 0) {
	       printf("%2d: ",c->c_id);}
	      OutPtr += sprintf(OutPtr,"%2d: ",c->c_id);}}
      else {
	 if(this_computer == 0) {
            printf("%2d: ",c->c_id);}
	 OutPtr += sprintf(OutPtr,"%2d: ",c->c_id);}}}
 
else {
 
/*--------------------------------------*/
/* clause is not active; denote by #    */
/*--------------------------------------*/
 
   if((c->c_bits & LIST) == LIST) {
      if(this_computer == 0) {
         printf("%2d#>",c->c_id);}
      OutPtr += sprintf(OutPtr,"%2d#>",c->c_id);}
   else {
      if(this_computer == 0) {
         printf("%2d# ",c->c_id);}
      OutPtr += sprintf(OutPtr,"%2d# ",c->c_id);}}

if(c->c_lpar != 0) {
   if(c->c_bits & EVALMASK) {

/*-------------------------------------*/
/* clause is a substitution            */
/*-------------------------------------*/

      if(this_computer == 0) {
	 if(c->c_rpar != 0 ) {
         printf("[%d%c,%d%c] ",c->c_lpar->c_id, 
               AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}}      
      if(c->c_rpar != 0) {
      OutPtr += sprintf(OutPtr,"[%d%c,%d%c] ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}

      if(c->c_lpar->c_id < 10) {
         if(this_computer == 0) {
	    printf(" ");}
	  OutPtr += sprintf(OutPtr," ");}
	  
      if(c->c_rpar != 0) {
         if(c->c_rpar->c_id < 10) {
            if(this_computer == 0) {
	       printf(" ");}
            OutPtr += sprintf(OutPtr," ");}}}

   else if(c->c_rpar == 0 && c->c_indices != 0) {
 
/*--------------------------*/
/* clause is a factor       */
/*--------------------------*/
 
      if(this_computer == 0) {
         printf("(%d%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);}
      OutPtr += sprintf(OutPtr,"(%d%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
      if(c->c_lpar->c_id < 10) {
         if(this_computer == 0) {
	    printf(" ");}
         OutPtr += sprintf(OutPtr," ");}}

   else if(c->c_lpar == c->c_rpar && 
      ((c->c_indices & LEFT_INDEX_MASK)>>6) ==
	 (c->c_indices & RIGHT_INDEX_MASK) ||
	 c->c_change != 0) {

/*----------------------------*/
/* clause is an instantiation */ 
/*----------------------------*/
 
      if(this_computer == 0) {
         printf("%d|",c->c_lpar->c_id);}
      OutPtr += sprintf(OutPtr,"%d|",c->c_lpar->c_id);
      substitution_info = c->c_change;
      if(substitution_info == 0) {substitution_info = cl[ply - 1][5];}
      var1 = substitution_info & 0x000000FF;
      con1 = (substitution_info >> 8) & 0x000000FF;
      var2 = (substitution_info >> 16) & 0x000000FF;
      con2 = (substitution_info >> 24) & 0x000000FF;

      if(var2 != 0) {
         if(this_computer == 0) {
            printf("{%s/%c,%s/%c} ",
              fidtbl[con1],varchars[var1],fidtbl[con2],varchars[var2]);}
	 OutPtr += sprintf(OutPtr,"{%s/%c,%s/%c} ",
              fidtbl[con1],varchars[var1],fidtbl[con2],varchars[var2]);}
      else {
         if(this_computer == 0) {
            printf("{%s/%c} ",
	              fidtbl[con1],varchars[var1]);}
	 OutPtr += sprintf(OutPtr,"{%s/%c} ",
	              fidtbl[con1],varchars[var1]);}
	   
      if(c->c_lpar->c_id < 10) {
	 if(this_computer == 0) {
	    printf(" ");}
         OutPtr += sprintf(OutPtr," ");}}

   else if(phase == 3 && cl[ply - 1][5] == -1) {

	 /*----------------------------------*/
	 /* clause is a hash resolvent       */
	 /*----------------------------------*/ 

      if(c->c_bits & MERGE) {	
	 if(this_computer == 0) {
	    if(c->c_rpar != 0 && c->c_lpar != 0 ) {
            printf("{%d%c*%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}}       
	 if(c->c_lpar != 0 && c->c_rpar != 0) {
	    OutPtr += sprintf(OutPtr,"{%d%c*%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}}
      else { 
	 if(c->c_rpar != 0 && c->c_lpar != 0 ) {
	 if(this_computer == 0) {
            printf("{%d%c,%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}       
	 OutPtr += sprintf(OutPtr,"{%d%c,%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}}
      if(c->c_lpar != 0) {
         if(c->c_lpar->c_id < 10) {
            if(this_computer == 0) {
	       printf(" ");}
            OutPtr += sprintf(OutPtr," ");}}

      if(c->c_rpar != 0) {
         if(c->c_rpar->c_id < 10) {
            if(this_computer == 0) {
	       printf(" ");}
            OutPtr += sprintf(OutPtr," ");}}}

   else {
 
	 /*----------------------------------*/
	 /* clause is a binary resolvent     */
	 /*----------------------------------*/ 

      if((c->c_bits & MERGE) && c->c_lpar != 0 && c->c_rpar != 0) {	
	 if(this_computer == 0) {
            printf("(%d%c*%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}       
	 OutPtr += sprintf(OutPtr,"(%d%c*%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}
      else { 
	 if(c->c_lpar != 0 && c->c_rpar != 0) {
	 if(this_computer == 0) {
            printf("(%d%c,%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}       
	 OutPtr += sprintf(OutPtr,"(%d%c,%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}}

      if(c->c_lpar != 0) {
      if(c->c_lpar->c_id < 10) {
         if(this_computer == 0) {
	    printf(" ");}
         OutPtr += sprintf(OutPtr," ");}}

      if(c->c_rpar != 0) {
         if(c->c_rpar->c_id < 10) {
            if(this_computer == 0) {
	       printf(" ");}
            OutPtr += sprintf(OutPtr," ");}}}}
			 
/*----------------------------------------------*/
/* if null clause, print []                     */
/*----------------------------------------------*/

if(c->c_lit == 0) {
   if(questionable[c->c_id] == 1) {
      if(this_computer == 0) {
	 printf("[] ... ?\n");}
      OutPtr += sprintf(OutPtr,"[] ... ?\n");}
   else {
      if(this_computer == 0) {
	 printf("[]\n");}
      OutPtr += sprintf(OutPtr,"[]\n");}}

/*----------------------------------------------*/
/* else print clause                            */
/*----------------------------------------------*/
 
else {
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {
	 if(this_computer == 0) {
	    printf("<");}
         OutPtr += sprintf(OutPtr,"<");}
      
      if((l->l_id & NEGMASK) != 0)  {
	 if(this_computer == 0) {
	    printf("~");}
         OutPtr += sprintf(OutPtr,"~");}
				 
    if(this_computer == 0) {
       printf("%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]);} 
    OutPtr += sprintf(OutPtr,
	   "%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]);
 
    a = &l->l_id;
    a++;
    for(i = l->l_id & ARGMASK; i > 0; i--) {
       /*if((l->l_id & 0x0000FFFF) == equal_pred && i == 1) {
	  if(this_computer == 0) {
	     printf(":");}
          OutPtr += sprintf(OutPtr,":");}*/
       if(i == (l->l_id & ARGMASK)) {
          if(this_computer == 0) {
	     printf("[");}
          OutPtr += sprintf(OutPtr,"[");}
       a = print_arguments(a,i);}
 
/*----------------------------------------------*/
/* see if literal is represented by hash code   */
/*----------------------------------------------*/              

    if((l->l_id & ELIMLIT) != 0) {
       if(this_computer == 0) {
	  printf(">");}
       OutPtr += sprintf(OutPtr,">");
       if(l->l_id & CONSMASK) {
	  if(this_computer == 0) {
	     printf(" {%s/x}",fidtbl[(l->l_id & CONSMASK) >> 20]);}
	  OutPtr += sprintf(OutPtr,
		" {%s/x}",fidtbl[(l->l_id & CONSMASK) >> 20]);}}

    if((l->l_id & EVALMASK) != 0) {
       if(l->l_next == 0) {
          litend = (int) c->c_end;} 
       else {
          litend = (int) l->l_next;} 

       while((int) a != litend) {
          if(*a != 0) {
             if(this_computer == 0) {
                printf(" +%8x",*a);}
             OutPtr += sprintf(OutPtr," +%8x",*a);}
          a++;}}

/*----------------------------------------------*/
/* print a space after each literal             */
/*----------------------------------------------*/

      if(l->l_next != 0) {
	 if(this_computer == 0) {
	    printf(" ");}
         OutPtr += sprintf(OutPtr," ");}}}

if(o_debug > 0 || big_bug != 0) {
   if(this_computer == 0) {
      printf(" H:%d %d",c->c_X,c->c_Y);}
   OutPtr += sprintf(OutPtr," H:%d %d",c->c_X,c->c_Y);
   if(this_computer == 0) {
      printf(" NODE: %d %d",s_nod,s_nod - s_nod0);}
   OutPtr += sprintf(OutPtr," NODE: %d %d",s_nod, s_nod - s_nod0);}

if(questionable[c->c_id] == 1 && c->c_nlit != 0) {
   if(this_computer == 0) {
      printf(" ... ?");}
   OutPtr += sprintf(OutPtr," ... ?");}

if(o_step <= 0 || (o_step > 0 && ply < o_step)) { 
   if(!x_cr) {
      if(this_computer == 0) {
         printf("\n");}
      OutPtr += sprintf(OutPtr,"\n");}}
if(x_cr && ply > o_step) {
   if(this_computer == 0) {
      printf("\n");}
   OutPtr += sprintf(OutPtr,"\n");}
/*if(this_computer != 0) {fflush(stdout);} */
fflush(stdout);
return; 
}

/*---------------------------------------------------------------*/
/* print_clause2()
/*---------------------------------------------------------------*/

void print_clause2(c)
clause *c;
{
int *cq;

cq = &c->c_id;

printf("\n");
if(this_computer != 0) fflush(stdout);
printf("A%8x ID%d ",cq,*cq++);
printf("A%8x END %x ",cq,*cq++);
printf("A%8x BITS %x \n",cq,*cq++);
if(this_computer != 0) fflush(stdout);

printf("A%x H %x %x ",cq,*cq++,*cq++);
printf("A%x L %x R %x ",cq,*cq++,*cq++);
printf("A%x IX %x \n",cq,*cq++);
if(this_computer != 0) fflush(stdout);

printf("A%x CLIT %x ",cq,*cq++);
printf("A%x NLIT %d ",cq,*cq++);
printf("A%x CCGE %x ",cq,*cq++);
printf("A%x V %d ",cq,*cq++);
printf("..\n");
if(this_computer != 0) fflush(stdout);
while((int) cq < (int) c->c_end) { 
   printf("A%x %8x ",cq,*cq);
   cq++;}
printf("\n");
if(this_computer != 0) fflush(stdout);
}

/*---------------------------------------------------------------*/
/* print_all_clauses()
/*---------------------------------------------------------------*/

void print_all_clauses()
{
int i;

for(i = 0; i <= base; i++) {print_clause(clist[i]);}
printf("\n");
if(this_computer == 0) OutPtr += sprintf(OutPtr,"\n");
}

/*---------------------------------------------------------------*/
/*      help_halted_search()           .....................#3
/*
/*      print information to help human during search
/*
/*      called by:        halted_search_options()     tdump.c
/*
/*      calls:            nobody
/*
/*      returns:          Nil
/*---------------------------------------------------------------*/
 
void help_halted_search()
{
printf("\n\n  - - - - - - - - - - - - - - - - - - - - - - - - - - -");
printf("\n |  The following keys control the search:");
printf("\n |");
printf("\n | 'q'  => quit prove, try another thoerem");
printf("\n | 'k'  => quit iteration, begin the next one");
printf("\n | 'r'  => quit iteration, repeat iteration");
printf("\n | 'x'  => terminate search on this branch");
printf("\n | 't'  => print total search time");
printf("\n | 'n'  => print total search nodes");
printf("\n | 'a'  => print partially formed proof in cl_array");
printf("\n | 'h'  => print hash_code_list");
printf("\n | 'c'  => print current continuation and stats");
printf("\n | 'D'  => print substitution array");
printf("\n | 'y'  => print tree structure");
printf("\n | 'vx' => print clause x on clist");
printf("\n | 'V'  => stop printing clauses from clist");
printf("\n | '+'  => increment step and print depth");
printf("\n | '-'  => decrement step and print depth");
printf("\n | 'p'  => print step and print depth");
printf("\n | 'd'  => turn on debug option");
printf("\n | '?'  => print this list!!");
printf("\n | 'CR' => type a 'return' to continue with the search");
printf("\n |");
printf("\n  - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
}
 
/*---------------------------------------------------------------*/
/*      halted_search_options()         .....................#4
/*
/*      called by:    search_tree()              search.c
/*
/*      calls:        time_now()                 theo.c
/*                    print_continuation()       tdump.c
/*                    print_cl_array()           pass2.c
/*                    print_hash_code_list()     pass2.c
/*                    grab_number()              theo.c
/*                    help_halted_search()       tdump.c
/*                    print_clause()             tdump.c
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
 
int halted_search_options()
{
int zz, i, o_debugt;
int kk, clause_to_print, current_time;
char in_char;
dnode *dd;
clause *c;
 
print0(" ");
 
clause_to_print = -1;
 
search_options:

for(i = 0; i < 80 ; i++) {
   input_buffer[i] = getchar();
   if(input_buffer[i] == '\n') {break;}}

zz = input_buffer[0];
if(this_computer == 0) OutPtr += sprintf(OutPtr,"%c",(char) zz);
 
if(clause_to_print >= 0 && zz == '\n') {
   goto print_this_clause;}
   
clause_to_print = -1;
 
/*-----------------------------------------*/
/* if 'q', quit search                     */
/*-----------------------------------------*/
 
if(zz == 'q') {
   print0("\n\nDo you wish to quit(Y or N)? ");
   zz = getchar();
   printf("%c ",zz); 
   if(this_computer == 0) OutPtr += sprintf(OutPtr,"%c ",zz);
   if(zz == 'y' || zz == 'Y') {
      return(1);}
   else goto search_options;}
 
/*-----------------------------------------*/
/* if 'k', kill iteration, start next      */
/*-----------------------------------------*/
 
if(zz == 'k')  {
   print0("\nQuitting this iteration, beginning the next.\n");
   x_stop = 2;
   return(0);}
 
/*-----------------------------------------*/
/* if 'r', repeat iteration                */
/*-----------------------------------------*/
 
if(zz == 'r')  {
   x_stop = 2; 
   iter_depth--;
   print0("\nQuitting this iteration, but repeating it.\n");
   return(0);}
 
/*-----------------------------------------*/
/* if 'x', terminate search at this node   */
/*-----------------------------------------*/
 
if(zz == 'x')  {
   x_stop = 1;
   return(0);}
 
/*-----------------------------------------*/
/* if 't', print search time in seconds    */
/* if 'n', print nodes searched            */
/*-----------------------------------------*/
 
if(zz == 't') {
   current_time = time_now() - search_start_time;
   printf("TIME: %d sec.",current_time); 
   if(this_computer == 0) OutPtr += sprintf(OutPtr,"TIME: %d sec.",current_time);
   print0("\n");
   goto search_options;}
 
if(zz == 'n') {
   printf("NODES: %d",s_nod);
   if(this_computer == 0) OutPtr += sprintf(OutPtr,"NODES: %d",s_nod);
   print0("\n");
   goto search_options;}
 
/*----------------------------------------*/
/* if 'a', print proof array              */
/*----------------------------------------*/
 
if(zz == 'a') {
   if(!proof_length) {
      print0("NO PROOF FOUND YET!\n");
      goto search_options;}
   else {
      print0("PROOF AS FOUND THUS FAR:\n");
      printf("\nProof length = %d ",proof_length);
      if(this_computer == 0) 
	 OutPtr += sprintf(OutPtr,"\nProof length = %d ",proof_length);
      print_cl_array(PROOF_ARRAY_SIZE - proof_length);
      goto search_options;}}
 
/*----------------------------------------*/
/* if 'h', print hash_code_list           */
/*----------------------------------------*/
 
if(zz == 'h') {
   if(!proof_length) {
      print0("NO PROOF FOUND YET AND THUS NO HASH CODES TO FIND\n");
      goto search_options;}
   o_debugt = o_debug;
   o_debug = 1;
   print_hash_code_list();
   print0("\n");
   o_debug = o_debugt;
   goto search_options;}
		
/*------------------------------------------*/
/* if 'c', print current continuation       */
/*------------------------------------------*/
 
if(zz == 'c') {
   print0("CURRENT CONTINUATION UNDER SEARCH:");
   printf("\n (NODES: %d) <UNITS: %d>",s_nod,s_ltentry);
   if(this_computer == 0)
      OutPtr += sprintf(OutPtr,"\n (NODES: %d) <UNITS: %d>",s_nod,s_ltentry);
   current_time = time_now() - search_start_time;
   printf(" (TIME: %d) (ITER: %d) ",current_time,iter_depth);
   if(this_computer == 0) 
      OutPtr += sprintf(OutPtr," (TIME: %d) (ITER: %d) ",
                                      current_time,iter_depth);
   printf("(l%d) (x%d) ",o_maxlit,o_maxterm);
   if(this_computer == 0) 
      OutPtr += sprintf(OutPtr,"(l%d) (x%d) ",o_maxlit,o_maxterm);
   printf("<LEN: %d> ",proof_length);
   if(this_computer == 0)
      OutPtr += sprintf(OutPtr,"<LEN: %d> ",proof_length);
   print_continuation();
   goto search_options;}
 
/*------------------------------------------*/
/* if 'y', print current tree info          */
/*------------------------------------------*/
 
if(zz == 'y') {
   print0("CURRENT SEARCH INFORMATION:\n");
   dd = &depth[0];
   printf("\nBase clauses:    %3d %3d %3d %8d",
		 dd->d_bc + 1, dd->d_tc + 1, dd->d_ec + 1, dd->d_rc);
   printf("\n--------------------------------");
   if(this_computer == 0) 
      OutPtr += sprintf(OutPtr,"\nBase clauses:    %3d %3d %3d %8d",
	 dd->d_bc + 1, dd->d_tc + 1, dd->d_ec + 1, dd->d_rc);
   if(this_computer == 0) 
      OutPtr += sprintf(OutPtr,"\n--------------------------------");
 
   for(kk = 1; kk <= ply; kk++) {
      dd = &depth[kk];
      printf("\nPly  %2d clauses: %3d %3d %3d %8d",
	   kk, dd->d_bc + 1, dd->d_tc + 1, dd->d_ec + 1, dd->d_rc);}
   print0("\n");
   goto search_options;}
 
/*----------------------------------------*/
/* if 'vx', print clause x on clist       */
/*----------------------------------------*/
 
if(zz == 'v') {
   clause_to_print = (int) grab_number(1) - 1;
 
print_this_clause:
   dd = &depth[ply];
   if(clause_to_print > dd->d_ec || clause_to_print < 0) {
      print0("\nInvalid clause index.\n");
      clause_to_print = -1;
      goto search_options;}
   c = clist[clause_to_print];
   print_clause(c);
   print0("\n");
   clause_to_print++;
   goto search_options;}

if(zz == 'V') {
   clause_to_print = -1;
   goto search_options;}
		
/*------------------------------------------*/
/* if '+', increment step,print depth       */
/* if '-', decrement step,print depth       */
/* if 'p', print out step,print depth       */
/*------------------------------------------*/
 
if(zz == '+') {
   o_print++; 
   o_step++;
   printf(" P%d S%d\n",o_print,o_step);
   if(this_computer == 0) OutPtr += sprintf(OutPtr," P%d S%d\n",o_print,o_step);
   goto search_options;}
 
if(zz == '-' ) {
   if(o_step) {
      o_print--;
      o_step--;         
      printf(" P%d S%d\n",o_print,o_step);
      if(this_computer == 0) 
	 OutPtr += sprintf(OutPtr," P%d S%d\n",o_print,o_step);
      goto search_options;}
   else {
      printf(" P%d S%d\n",o_print,o_step);
      if(this_computer == 0)
	 OutPtr += sprintf(OutPtr," P%d S%d\n",o_print,o_step);
      goto search_options;}}
			
if(zz == 'p') {
   printf(" P%d S%d\n",o_print,o_step);
   if(this_computer == 0) 
      OutPtr += sprintf(OutPtr," P%d S%d\n",o_print,o_step);
   goto search_options;}        
 
/*-------------------------------------------*/
/* if 'd', then if debug option is on, turn  */
/* it off, else if off, turn it on           */
/*-------------------------------------------*/
 
if(zz == 'd')  {
   if(!o_debug) {
      o_debug = 1;
      goto search_options;}
   else {
      o_debug = 0;      
      goto search_options;}}   

/*-------------------------------------------*/
/* if 'D', print substitution table          */
/*-------------------------------------------*/

if(zz == 'D') {
   print_substitutions();
   goto search_options;}
 
/*-------------------------------------------*/
/* if '?', print halted_search_options       */
/*-------------------------------------------*/
 
if(zz == '?' ) {
   help_halted_search();
   goto search_options;}
 
if(zz == 10) {      /* if zz is a carriage return, return */
   return(0);}
 
else {
   print0("\n?? ");
   goto search_options;}
 
}
 
/*---------------------------------------------------------------*/
/*      print_continuation()            .......................#5
/*
/*      prints the current continuation under search
/*
/*      called by:   halted_search_options()         tdump.c
/*                   install_clause()                search.c
/*                   prove_by_iterative_deepening()  theo.c
/*
/*      calls:       print_clause()                  tdump.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void print_continuation()
{
int i;
clause *c;
	
print0("\n");
for(i = 1 ; i <= ply ; i++) {
   c = clist[depth[i].d_tc];
   print_clause(c);}
return;
}
 
/*---------------------------------------------------------------*/
/*      dump_calloc_info()         ....................#6
/*
/*      called by:    init_clause_hash_table()   theo.c
/*                    init_random_hash_array()   theo.c
/*                    init_search_space()        theo.c
/*
/*      calls:        fatal()                    theo.c
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
 
void dump_calloc_info(name,nameptr,index,data_type,size)
char *name;
int nameptr;
int index;
char *data_type;
int size;
{
if(o_debug > 1) {
   if(index == -1) {
      printf("\n%s \t@%d  #%d  %s",name,nameptr,size,data_type);
      if(this_computer == 0) 
	 OutPtr += sprintf(OutPtr,
	     "\n%s \t@%d  #%d  %s",name,nameptr,size,data_type);}
 
   else {
      if(index % 4 == 0) {
	     printf("\n%s[%d] \t@%d  #%d  %s",
			  name,index,nameptr,size,data_type);
	     if(this_computer == 0)
		OutPtr += sprintf(OutPtr,"\n%s[%d] \t@%d  #%d  %s",
			 name,index,nameptr,size,data_type);}
      else {
	     printf("  @%d",nameptr);
	     if(this_computer == 0)
		OutPtr += sprintf(OutPtr,"  @%d",nameptr);}}
   getchar();}
 
if(!nameptr) {
   if(index == -1) {
      printf("Insufficient memory (%d) for %s",size,name);
      if(this_computer == 0)
	 OutPtr += sprintf(OutPtr,"Insufficient memory (%d) for %s",size,name);}
   else {
      printf("Insufficient memory for %s[%d]",name,index);
      if(this_computer == 0)
	 OutPtr += sprintf(OutPtr,"Insufficient memory for %s[%d]",name,index);}
   fatal("\nQuitting PROVE.\n");}      
return;
}       
	
/*---------------------------------------------------------------*/
/*      print0()         .......................#7
/*
/*      prints a string to the screen and to the disk
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
	
void print0(msg)
char *msg;
{
printf("%s",msg);fflush(stdout);
OutPtr += sprintf(OutPtr,"%s",msg);
}

/*-----------------------------------------------------------------*/
/*      print_substitutions()
/*
/*      called by:
/*
/*      calls:
/*
/*      returns:  NIL
/*------------------------------------------------------------------*/

void print_substitutions()

{
int i, j, k, m, n, sub_stack[200], top_stack;
int returned, deep;
static char var_[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";

if(sub_array_last == 0) return;

print0("\n");


i = 1;
top_stack = 0;
returned = 0;
sub_stack[0] = 0;
deep = 0;

printf("\n\n");
if(this_computer == 0) OutPtr += sprintf(OutPtr,"\n\n");
				 
loop_top:
if(i == 1 && top_stack == 1) {print0("\n\n");
return;}


if(returned == 1) {
   returned = 0;
   for(k = 0; k < top_stack ; k++) {deep++;print0("\t");}}

if((sub_array[i][0] & 0xFFFF0000) == 0xFFFF0000) {
      if(this_computer == 0)
	 OutPtr += sprintf(OutPtr,"%d %c",
	 deep,var_[sub_array[i][0] & 0x0000FFFF]);
      printf("%d %c",deep,var_[sub_array[i][0] & 0x0000FFFF]);}
   else if(sub_array[i][0] != 0x00001000) {
/*      OutPtr += sprintf(OutPtr,"%d. F%d", deep,
          ((sub_array[i][0]) & 0x0000FFE0) >> ARGSHIFT);
      printf("%d. F%d", deep,
          ((sub_array[i][0]) & 0x0000FFE0) >> ARGSHIFT);}
*/      if(this_computer == 0) 
	  OutPtr += sprintf(OutPtr,"%d %s ", deep,
          fidtbl[((sub_array[i][0]) & 0x0000FFE0) >> ARGSHIFT]);
      printf("%d %s ", deep,
          fidtbl[((sub_array[i][0]) & 0x0000FFE0) >> ARGSHIFT]);}

   if((sub_array[i][1] & 0xF0000000) == 0) {

/*---------------------------------------------*/
/* go deeper on this path                      */
/*---------------------------------------------*/
      deep++;
      print0("\t");

      sub_stack[top_stack++] = i;
      i = sub_array[i][1];}	
   else { 

/*---------------------------------------------*/
/* terminal node, print contents               */
/*---------------------------------------------*/

      if((sub_array[i][1] & 0xE0000000) == 0xC0000000) {
         if(this_computer == 0) OutPtr += sprintf(OutPtr," (%s: %d)",
	        fidtbl[(sub_array[i][1] & 0x0000FFFF) >> ARGSHIFT],
		i);
         printf(" (%s: %d)",
	        fidtbl[(sub_array[i][1] & 0x0000FFFF) >> ARGSHIFT],
		i);}
      else {
         if(this_computer == 0) OutPtr += sprintf(OutPtr," (%c: %d) ",
	     var_[sub_array[i][1] & 0x0000FFFF],
	     i);
         printf(" (%c: %d)",
	     var_[sub_array[i][1] & 0x0000FFFF],
             i);}
      print0("\n");deep = 0;returned++; 

get_brother: 
    if(sub_array[i][2] != -1) {
	sub_stack[top_stack] = sub_array[i][2];
	i = sub_array[i][2];
	goto loop_top;}
    else { 
        i = sub_stack[--top_stack];
	goto get_brother;}}

   goto loop_top;
}	

/*-----------------------------------------------------------------*/
/* stat_dump()
/*
/*-----------------------------------------------------------------*/

void stat_dump()
{
if(this_computer == 0) {
   printf("%d {T%d N%d R%d F%d C%d H%d h%d U%d E%d s%d t%d L%d} ",
      iter_depth,(search_time -search_start_time),
      s_nod - s_nod0,s_res,s_fac,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);}
OutPtr += sprintf(OutPtr,
     "%d {T%d N%d R%d F%d C%d H%d h%d U%d E%d s%d t%d L%d} ",
      iter_depth,(search_time - search_start_time),
      s_nod - s_nod0,s_res,s_fac,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);
if(this_computer == 0) {
   printf("\n\n");}
OutPtr += sprintf(OutPtr,"\n\n");
}
