/*-----------------------------------------------------------------*/
/*      tdump.c
/*      copyrighted by Monty Newborn 2005
/*
/*      functions:
/*
/*      the following three functions print the clause
/*
/*      print_arguments()       .......................#1
/*      print_clause()          .......................#2
/*      print_clause2()         .......................#2a
/*      print_literal2()        .......................#2c
/*      print_all_clauses()     .......................#2b 
/*
/*      information on the options available are printed out by:
/*
/*      help_halted_search()       ....................#3
/*
/*      user can request information or change options or
/*      guide the search by choosing various options.
/*
/*      halted_search_options() .......................#4
/*      print_continuation()    .......................#5
/*      dump_calloc_info()      .......................#6
/*
/*      prints a string to the screen and to the disk
/*
/*      print0()                .......................#7
/*
/*      print_substitutions()   .......................#8
/*
/*      stat_dump()             .......................#9
/*
/*-----------------------------------------------------------------*/
 
#include "const.h"
#include "clauseS.h"
 
#include <stdio.h>

/*------------------------------------------------------------------*/
/* externally defined functions                                     */
/*------------------------------------------------------------------*/
 
extern void print_cl_array();                     /* in pass2.c     */
extern void print_hash_code_list(); 
extern int time_now();                            /* in theo.c      */
extern void fatal();
extern int grab_number();

/*------------------------------------------------------------------*/
/* externally defined constants                                     */
/*------------------------------------------------------------------*/
 
extern char input_buffer[];
extern clause *clist[];              /* pointers to clauses         */
extern clause *ulist[];              /* pointers to clauses         */
extern dnode depth[];            /* pointers to pointers of clauses */
extern int cl[PROOF_ARRAY_SIZE][14];
extern int sub_array[65536][6];
extern int sub_array_last;
extern pass_number;
extern int questionable[2000];
extern int phase, big_bug;
extern int o_step, o_print, x_cr;
extern int base, org_base, ply, iter_depth, s_nod, s_res, s_fac;
extern int nextu;
extern int s_nod0;
extern int s_ltentry,s_lthit,s_ltfail;
extern int s_htentry, s_hthit, s_htfail, s_subs_suc;
extern int s_subfxn, s_too_many_terms;
extern int x_stop, o_debug, o_maxlit;
extern int o_maxterm;
extern char *pidtbl[], *fidtbl[];
extern int search_start_time;
extern int search_time;
extern int proof_length;
extern int tab_amount;       /* number of tabs before printing cls  */
extern int equal_pred;
extern int reinstate_parent[1000];
 
extern char *OutPtr;
extern int this_computer;

/*------------------------------------------------------------------*/
/* internally defined functions                                     */
/*------------------------------------------------------------------*/
 
int *print_arguments();
void print_clause();
void print_clause2();
void print_literal2();
void print_all_clauses();
void help_halted_search();
int halted_search_options();
void print_continuation();
void dump_calloc_info();
void print0();
void print_substitutions();
void stat_dump();

/*------------------------------------------------------------------*/
/*      print_arguments(p)          .......................#1
/*
/*      print the arguments of the predicate
/*
/*      called by:    print_clause()           tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      Nil
/*------------------------------------------------------------------*/
 
int *print_arguments(p,terms)
int *p, terms;
{
int a, p_prime;
static char var_[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";
int terms_org;

if((*p & FCNMASK) == 0) {
	
/*------------------*/
/* print a variable */  
/*------------------*/
   	
   p_prime = *p;        
   if(p_prime >= MAX_VARS) {
      p_prime = p_prime - MAX_VARS;}            /* standardize */
   if(this_computer == 0) printf("%c",var_[p_prime]); 
   OutPtr += sprintf(OutPtr,"%c",var_[p_prime]); 
   terms--;
   if(terms == 0) {
      if(this_computer == 0) printf("]");
      OutPtr += sprintf(OutPtr,"]");}
   else {
      if(this_computer == 0) printf(",");
      OutPtr += sprintf(OutPtr,",");}
   return(++p);}
    
/*------------------*/
/* print a function */  
/*------------------*/
 
if(this_computer == 0) printf("%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);
OutPtr += sprintf(OutPtr,"%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);

if((*p&ARGMASK) != 0) {
   if(this_computer == 0) printf("[");
   OutPtr += sprintf(OutPtr,"[");}

terms--;

/*------------------------------------------------*/
/* recursively print the remaining arguments of p */
/*------------------------------------------------*/
 
for(a = *p++ & ARGMASK; a > 0; a--)  {
   p = print_arguments(p,a);}

if(terms == 0) {
   if(this_computer == 0) printf("]");
   OutPtr += sprintf(OutPtr,"]");}
else {
   if(this_computer == 0) printf(",");
   OutPtr += sprintf(OutPtr,",");}

return(p);
}
 
/*-----------------------------------------------------------------*/
/*      print_clause(c)               .......................#2
/*
/*      called by:      write_out_results()       theo.c
/*                      search_tree()             search.c
/*                      advance_to_successor()    search.c
/*
/*      calls:          print_arguments()         tdump.c
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void print_clause(c)
clause *c;
{
int i, j, *a, mm, litend;
int substitution_info, con1, var1, con2, var2;
literal *l, *litrl;
static char AB[] = ".abcdefghijklmnopqrstuvwxyz";
static char varchars[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";

/*--------------------------------------------------------*/
/* decide how far to tab before beginning to print clause */
/*--------------------------------------------------------*/

if((c->c_bits & BASE) == 0) {
   if(c->c_lpar->c_id != c->c_id - 1 && c->c_id > base + 1) {
      tab_amount = c->c_id - base - 2;}}
else {
   tab_amount = 0;}
 
if((c->c_bits & BASE) == 0) {
   for(i = c->c_id - base - 2 - tab_amount; i > 0; i--) {
            OutPtr += sprintf(OutPtr," ");
	    if(this_computer == 0) printf(" ");}}       

if((c->c_bits & ACTIVE) != 0) {
   if((c->c_bits & NC) != 0) {
				
/*--------------------------------------*/
/* active clause in NC; denote by S     */
/*--------------------------------------*/
 
      if((c->c_bits & BASE) == BASE) {
         if((c->c_bits & LIST) == LIST) {
	    if(this_computer == 0) printf("%2dS>",c->c_id);
	    OutPtr += sprintf(OutPtr,"%2dS>",c->c_id);}
         else {
	    if(this_computer == 0) printf("%2dS ",c->c_id);
	    OutPtr += sprintf(OutPtr,"%2dS ",c->c_id);}}
      else {
         if(this_computer == 0) printf("%2dS>",c->c_id);
	 OutPtr += sprintf(OutPtr,"%2dS>",c->c_id);}}
   else {               
 
/*--------------------------------------*/
/* active clause not in NC; leave " "   */
/*--------------------------------------*/
 
      if((c->c_bits & BASE) == BASE) {
         if((c->c_bits & LIST) == LIST) {
	    if(this_computer == 0) printf("%2d >",c->c_id);
            OutPtr += sprintf(OutPtr,"%2d >",c->c_id);}
	 else {
	    if(this_computer == 0) printf("%2d: ",c->c_id);
	    OutPtr += sprintf(OutPtr,"%2d: ",c->c_id);}}
      else {
         if(this_computer == 0) printf("%2d: ",c->c_id);
	 OutPtr += sprintf(OutPtr,"%2d: ",c->c_id);}}}
 
else {
 
/*--------------------------------------*/
/* clause is not active; denote by #    */
/*--------------------------------------*/
 
   if((c->c_bits & LIST) == LIST) {
      if(this_computer == 0) printf("%2d#>",c->c_id);
      OutPtr += sprintf(OutPtr,"%2d#>",c->c_id);}
   else {
      if(this_computer == 0) printf("%2d# ",c->c_id);
      OutPtr += sprintf(OutPtr,"%2d# ",c->c_id);}}

if(c->c_lpar != 0) {
   if(c->c_bits & EVALMASK) {

/*-------------------------------------*/
/* clause is a substitution            */
/*-------------------------------------*/

      if(this_computer == 0) printf("[%d%c,%d%c] ",c->c_lpar->c_id, 
               AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
      OutPtr += sprintf(OutPtr,"[%d%c,%d%c] ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);
      if(c->c_lpar->c_id < 10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}
      if(c->c_rpar->c_id < 10) 
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}
      
   else if(c->c_rpar == 0 && c->c_indices != 0) {
 
/*--------------------------*/
/* clause is a factor       */
/*--------------------------*/
 
      if(this_computer == 0) printf("(%d%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
      OutPtr += sprintf(OutPtr,"(%d%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
      if(c->c_lpar->c_id<10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}}

   else if(c->c_lpar == c->c_rpar && 
      ((c->c_indices & LEFT_INDEX_MASK)>>6) ==
	 (c->c_indices & RIGHT_INDEX_MASK) ||
	 c->c_change != 0) {

/*-------------------------------*/
/* clause is an instantiation    */ 
/* or substitution of constants. */
/*-------------------------------*/
 
      if(this_computer == 0) printf("%d | ",c->c_lpar->c_id);
      OutPtr += sprintf(OutPtr,"%d | ",c->c_lpar->c_id);

      substitution_info = c->c_change;
      if(substitution_info == 0) {substitution_info = cl[ply - 1][5];}
      var1 = substitution_info & 0x000000FF;
      con1 = (substitution_info >> 8) & 0x000000FF;
      var2 = (substitution_info >> 16) & 0x000000FF;
      con2 = (substitution_info >> 24) & 0x000000FF;

      if(c->c_id <= base + 1 && reinstate_parent[c->c_id - 1] != 0) {
	 if(this_computer == 0) {
            printf("%d ",reinstate_parent[c->c_id - 1]);}
         OutPtr += sprintf(OutPtr,"%d ",reinstate_parent[c->c_id - 1]);}

      else if(var2 == 0 && con2 != 0) {
//    else if(c->c_rpar != 0 && c->c_rpar != c->c_lpar) {
         if(this_computer == 0) {
	    printf("%d ",c->c_rpar->c_id);}
         OutPtr += sprintf(OutPtr,"%d ",c->c_rpar->c_id);}
      else {
         if(this_computer == 0) printf("... ");
         OutPtr += sprintf(OutPtr,"... ");}

      if(con1 == 0) {
         if(this_computer == 0) printf("{null/null} ");
         OutPtr += sprintf(OutPtr,"{null/null ");}

      else if(var2 == 0 && con2 != 0) {
         if(this_computer == 0) {printf("{%s/%s} ",fidtbl[con1],fidtbl[var1]);}
	 OutPtr += sprintf(OutPtr,"{%s/%s} ",fidtbl[con1],fidtbl[var1]);}

      else if(var2 != 0) {
         OutPtr += sprintf(OutPtr,"{%s/%c,%s/%c} ",
         fidtbl[con1],varchars[var1],fidtbl[con2],varchars[var2]);
         if(this_computer == 0) {
	    printf("{%s/%c,%s/%c} ",
             fidtbl[con1],varchars[var1],fidtbl[con2],varchars[var2]);}}
      else {
         OutPtr += sprintf(OutPtr,"{%s/%c} ",
	 fidtbl[con1],varchars[var1]);
         if(this_computer == 0) {
	    printf("{%s/%c} ",fidtbl[con1],varchars[var1]);}}

      if(c->c_lpar->c_id<10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}}

   else if(phase == 3 && cl[ply - 1][5] == -1) {

	 /*----------------------------------*/
	 /* clause is a hash resolvent       */
	 /*----------------------------------*/ 

      if(c->c_bits & MERGE) {	
         if(this_computer == 0) printf("{%d%c*%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
	 OutPtr += sprintf(OutPtr,"{%d%c*%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}
      else { 
         if(this_computer == 0) printf("{%d%c,%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
	 OutPtr += sprintf(OutPtr,"{%d%c,%d%c} ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}
      if(c->c_lpar->c_id < 10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}
      if(c->c_rpar->c_id < 10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}}

   else {
 
	 /*----------------------------------*/
	 /* clause is a binary resolvent     */
	 /*----------------------------------*/ 

      if(c->c_bits & MERGE) {	
         if(this_computer == 0) printf("(%d%c*%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
	 OutPtr += sprintf(OutPtr,"(%d%c*%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}
      else { 
         if(this_computer == 0) printf("(%d%c,%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
	 OutPtr += sprintf(OutPtr,"(%d%c,%d%c) ",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}
      if(c->c_lpar->c_id < 10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}
      if(c->c_rpar->c_id < 10) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}}}
			 
/*----------------------------------------------*/
/* if null clause, print []                     */
/*----------------------------------------------*/

if(c->c_lit == 0) {
   if(questionable[c->c_id] == 1) {
      if(this_computer == 0) printf("[] ... ?\n");
      OutPtr += sprintf(OutPtr,"[] ... ?\n");}
   else {
      if(this_computer == 0) printf("[]\n");
      OutPtr += sprintf(OutPtr,"[]\n");}}
		 
/*----------------------------------------------*/
/* else print clause                            */
/*----------------------------------------------*/

else {
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {
         if(this_computer == 0) printf("<");
	 OutPtr += sprintf(OutPtr,"<");}
      
      if((l->l_id & NEGMASK) != 0)  {
         if(this_computer == 0) printf("~");
	 OutPtr += sprintf(OutPtr,"~");}     
				 
    if(this_computer == 0) 
       printf("%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]); 
    OutPtr += sprintf(OutPtr,
	   "%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]);
 
    a = &l->l_id;
    a++;
    for(i = l->l_id & ARGMASK; i > 0; i--) {
       if(i == (l->l_id & ARGMASK)) {
          if(this_computer == 0) printf("[");
	  OutPtr += sprintf(OutPtr,"[");}
       a = print_arguments(a,i);}
 
/*----------------------------------------------*/
/* see if literal is represented by hash code   */
/*----------------------------------------------*/              

    if((l->l_id & ELIMLIT) != 0) {
      if(this_computer == 0)  printf(">");
      OutPtr += sprintf(OutPtr,">");

       if(l->l_id & CONSMASK) {
	  if(this_computer == 0) 
	      printf(" {%s/x}",fidtbl[(l->l_id & CONSMASK) >> 20]);
	  OutPtr += sprintf(OutPtr,
		 " {%s/x}",fidtbl[(l->l_id & CONSMASK) >> 20]);}
       if(l->l_id & CONSMASK2) {
	  if(this_computer == 0) 
	    printf(" {%s/y}",fidtbl[(l->l_id & CONSMASK2) >> 16]);
	  OutPtr += sprintf(OutPtr,
		 " {%s/y}",fidtbl[(l->l_id & CONSMASK2) >> 16]);}}

    if((l->l_id & EVALMASK) != 0) {
       if(l->l_next == 0) {
          litend = (int) c->c_end;} 
       else {
          litend = (int) l->l_next;} 

       while((int) a != litend) {
          if(*a != 0) {
             if(this_computer == 0) printf(" +%8x",*a);
             OutPtr += sprintf(OutPtr," +%8x",*a);}
          a++;}}

/*----------------------------------------------*/
/* print a space after each literal             */
/*----------------------------------------------*/

      if(l->l_next != 0) {
         if(this_computer == 0) printf(" ");
	 OutPtr += sprintf(OutPtr," ");}}}

if(o_debug > 1 || big_bug != 0) {
   printf(" H:%d %d",c->c_X,c->c_Y);
   OutPtr += sprintf(OutPtr," H:%d %d",c->c_X,c->c_Y); 
   printf(" NOD: %d:%d ",s_nod,s_nod - s_nod0);
   OutPtr += sprintf(OutPtr," NOD: %d ",s_nod);}

if(o_debug > 0) {
   printf("VARS = %d ",c->c_nv);}
   
if(questionable[c->c_id] == 1 && c->c_nlit != 0) {
   if(this_computer == 0) printf(" ... ?");
   OutPtr += sprintf(OutPtr," ...?");}

for(i = 0; i < nextu; i++) {
if(c == ulist[i]) goto dont_print_w;}

if(c->c_id > org_base + 1 && c->c_lpar == 0) {
   for(mm = 0; mm < nextu; mm++) {
      if(c->c_X == ulist[mm]->c_X && c->c_Y == ulist[mm]->c_Y) {
         if(this_computer == 0) 
	       printf("    (Clause %d in weakened proof)",ulist[mm]->c_id);
         OutPtr += sprintf(OutPtr,
                "    (Clause %d in weakened proof)",ulist[mm]->c_id);
         break;}}}
dont_print_w:

if(o_step <= 0 || (o_step > 0 && ply < o_step)) { 
   if(!x_cr) {
   if(this_computer == 0) printf("\n");
   OutPtr += sprintf(OutPtr,"\n");}}
if(x_cr && ply > o_step) {
   if(this_computer == 0) printf("\n");
   OutPtr += sprintf(OutPtr,"\n");}
return; 
}

/*---------------------------------------------------------------*/
/* print_clause2()
/*---------------------------------------------------------------*/

void print_clause2(c)
clause *c;
{
int *cq;

cq = &c->c_id;

printf("print_clause2\n");fflush(stdout);
printf("\n");fflush(stdout);
printf("%10x ID%x ",cq,*cq++);
printf("\n");fflush(stdout);
printf("  %8x END %x ",cq,*cq++);
printf("\n");fflush(stdout);
printf("  %8x BITS %x ",cq,*cq);
printf("\n");fflush(stdout);

if(*cq & UNIT) printf("UNIT ");
if(*cq & MERGE) printf("MERGE ");
if(*cq & BASE) printf("BASE ");
if(*cq & DEAD) printf("DEAD ");
if(*cq & VID) printf("VID ");
printf("\n");fflush(stdout);
if(*cq & ACTIVE) printf("ACTIVE ");
if(*cq & NC) printf("NC ");
if(*cq & ACTIVE) printf("ACTIVE ");
if(*cq & LIST) printf("LIST ");
if(*cq & EQCLS) printf("EQCLS ");
if(*cq & HRES) printf("HRES ");
printf("\n");fflush(stdout);

cq++;

printf("%8x H %x %x \n",cq,*cq++,*cq++);
fflush(stdout);
printf("%8x L %x R %x ",cq,*cq++,*cq++);
printf("  %8x IX %x \n",cq,*cq++);
fflush(stdout);
printf("%8x CLIT %x ",cq,*cq++);
printf("%8x NLIT %d ",cq,*cq++);
printf("%8x CCGE %x ",cq,*cq++);
printf("%8x Vars %d ",cq,*cq++);
printf("....");
while((unsigned int) cq < (unsigned int) c->c_end) { 
   printf("%8x %8x \n",cq,*cq);
fflush(stdout);
   cq++;}
printf("\n");
fflush(stdout);
}

/*---------------------------------------------------------------*/
/* print_literal2()
/*---------------------------------------------------------------*/

void print_literal2(l)
literal *l;
{
int *lptr, a, kk;

printf("PRINT LITERAL\n");
kk = 0;
lptr = &l->l_id;
printf("LITID: %x %x\n",lptr,*lptr);
printf("HASHC: %d %d\n",l->l_X,l->l_Y);

for(a = *lptr++ & ARGMASK; a > 0; a--) {
   if(kk++ == 100) return;
   printf("\n%x %x",lptr,*lptr);
   if(*lptr & FCNMASK) {
      a += *lptr++ & ARGMASK;}

   else {
      lptr++;}}
printf("\nEND LITERAL\n\n");
}


/*---------------------------------------------------------------*/
/* print_all_clauses()
/*---------------------------------------------------------------*/

void print_all_clauses()
{
int i;

for(i = 0; i <= base; i++) {print_clause(clist[i]);}
if(this_computer == 0) printf("\n");
OutPtr += sprintf(OutPtr,"\n");
}

/*---------------------------------------------------------------*/
/*      print_continuation()            .......................#5
/*
/*      prints the current continuation under search
/*
/*      called by:   halted_search_options()         tdump.c
/*                   install_clause()                search.c
/*                   prove_by_iterative_deepening()  theo.c
/*
/*      calls:       print_clause()                  tdump.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void print_continuation()
{
int i;
clause *c;
	
if(this_computer == 0) printf("\n");
OutPtr += sprintf(OutPtr,"\n");

for(i = 1 ; i <= ply ; i++) {
   c = clist[depth[i].d_tc];
   print_clause(c);}
return;
}
 
/*---------------------------------------------------------------*/
/*      print0()         .......................#7
/*
/*      prints a string to the screen and to the disk
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
	
void print0(msg)
char *msg;
{
printf("%s",msg);
OutPtr += sprintf(OutPtr,"%s",msg);
fflush(stdout);
}

/*-----------------------------------------------------------------*/
/* stat_dump()
/*
/*-----------------------------------------------------------------*/

void stat_dump()
{
   printf("%d {T%d N%d R%d F%d C%d H%d h%d U%d u%d f%d E%d s%d t%d L%d} ",
      iter_depth,(search_time -search_start_time),
      s_nod - s_nod0,s_res,s_fac,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);
   OutPtr += sprintf(OutPtr,
     "%d {T%d N%d R%d F%d C%d H%d h%d U%d u%d f%d E%d s%d t%d L%d} ",
      iter_depth,(search_time - search_start_time),
      s_nod - s_nod0,s_res,s_fac,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);
   printf("\n\n");OutPtr += sprintf(OutPtr,"\n\n");
}
