/*-------------------------------------------------------------*/
/*      theoM.c
/*
/*
/*      functions:
/*      time_now()..................... #1
/*      init_stats()................... #2
/*      init_options()................. #3
/*      grab_number().................. #4
/*      fatal()........................ #5
/*      enter_theorem_name()........... #6
/*      init_search_space()............ #7
/*      init_clause_hash_table()....... #8
/*      init_random_hash_array()....... #9           
/*      free_space_close_files()....... #10
/*      read_buffer_word()............. #11
/*      interchange_fxn_const()........ #12
/*      is_fxn_a_constant()............ #13
/*      sort_constants_fxns().......... #14
/*      format_clause()................ #15
/*      count_lit_characters()......... #16
/*      max_term_in_any_lit().......... #17
/*      get_pred_fxn_symbols()......... #18
/*      prepare_clauses().............. #19
/*      print_search_options()......... #19aa
/*      prepare_subarray............... #19a (mar 01)
/*      hash_all_base_clauses()........ #20
/*      iterative_deepening().......... #21
/*      get_clause_stats()............. #22
/*      write_out_results()............ #23
/*      main()........>>>>>>>>>>>>>>>>> #24
/*-------------------------------------------------------------*/
/*
/*      LINKAGE OF FUNCTIONS
/*
/*main()--------------<<<<<<<<<<<<<<<<<<<<<<<<<.........theo.c
/*
/* begin_proof:
/* 
/*       enter_theorem_name()...........................theo.c
/*       compile_clauses()..............................affc.c
/*       init_cl_array()................................pass2.c
/*       init_stats()...................................theo.c
/*       init_search_space()............................theo.c
/*       prepare_clauses()..............................theo.c
/*             get_pred_fxn_symbols()...................theo.c
/*             read_buffer_word().......................theo.c
/*             format_clause()..........................theo.c
/*                   read_buffer_word().................theo.c
/*             sort_constants_fxns()....................theo.c
/*                   is_fxn_a_constant()................theo.c
/*                   interchange_fxn_const()............theo.c
/*             max_term_in_any_lit()....................theo.c
/*                   count_lit_characters().............theo.c
/*       init_random_hash_array().......................theo.c
/*       hash_all_base_clauses()........................theo.c
/*             calc_lit_hc()............................hash.c
/*             calc_clause_hashcode()...................hash.c
/*             add_clause_to_ht().......................hash.c
/*             calc_alt_hashcodes().....................hash.c
/* 
/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simply input clauses if possible.                  */
/*----------------------------------------------------*/
/* 
/*        eliminate()...................................elim.c
/* 
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/
/* 
/*        iterative_deepening().........................theo.c
/*              init_clause_hash_table()................theo.c
/*              add_clause_to_ht()......................hash.c
/*              calc_alt_hashcodes()....................hash.c
/*              search_tree()...........................search.c
/*              save_line().............................pass2.c
/*              time_now()..............................theo.c
/* 
/*        stack_hash_codes()............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/* put on stack new hash codes that must yet be       */
/* found on another pass through phase 2.             */
/*----------------------------------------------------*/
/* 
/*       iterative_deepening()..........................theo.c
/*       stack_hash_codes().............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* set phase to 3, check proof, then print results    */
/*----------------------------------------------------*/
/* 
/*       rearrange_cl_array()...........................pass2.c
/*       search_proof_line()............................pass2.c
/*       write_out_results()............................theo.c
/*             print_clause()...........................tdump.c
/*             get_clause_stats().......................theo.c
/*
/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/
/* 
/*       free_space_close_files().......................theo.c
/*
/*------------------------------------------------------------------*/
			 
#include "const.h"
#include "clause.h"
 
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <pvm3.h>
 
/*----------------------------------------------------------------*/
/* external functions called from this file                       */
/*----------------------------------------------------------------*/

extern void CompileTptp();               /* function in tptpi.c   */
extern int CompareStrings();

extern void compile_clauses();          /* function in affc.c     */
 
extern int search_tree();               /* functions in search.c  */
 
extern int calc_lit_hc();                /* functions in hash.c   */
extern int calc_alt_hashcodes();
extern void calc_clause_hashcode();
extern int add_clause_to_ht();
extern void normalize_literal();
extern void sort_hash_array();
 
extern void eliminate();                 /* function in elim.c    */
 
extern void init_cl_array();             /* functions in pass2.c  */
extern void save_line(); 
extern void stack_hash_codes();
extern void rearrange_cl_array();
extern void search_proof_line();
extern void print_hash_code_list();

extern void print_clause();              /* function in tdump.c   */
extern void dump_calloc_info();
extern void print0();
extern void print_continuation();
extern void print_substitutions();

extern void equal_predicate();

extern void prepare_subarray();             /* function in gen.c  */
extern void copy_clause3();
extern int test_simple_equality();

extern void rename_functions();
extern void modify_clause();

/*----------------------------------------------------------------*/
/* hash_code_list_active is a bit mapping of the hash codes on    */
/* the hash code list that remain to be found.  bit i is a 1      */
/* if hash code i has yet to be found.                            */
/*----------------------------------------------------------------*/
 
extern int hash_code_active_number;
extern int killer_hashx, killer_hashy;

/*----------------------------------------------------------------*/
/* the hash code of the clause whose complement is found in the   */
/* clause_hash_table implying a contradiction; it is thus magic!  */
/*----------------------------------------------------------------*/

extern int magic_hashx, magic_hashy;
extern int alt_HX;

/*----------------------------------------------------------------*/
/* array for substitution strings                                 */
/*----------------------------------------------------------------*/

extern int sub_array[65536][6];
extern int sub_array_last;
extern int const_found_equal;

/*----------------------------------------------------------------*/
/* array for substitution strings used in proof                   */
/*----------------------------------------------------------------*/

extern int deleted_clauses;

extern int cl[PROOF_ARRAY_SIZE][14];
extern int questionable[2000];

extern int change_array[50][10];
extern int change_level;

extern int p_change_array[2000][7];
extern int p_change_number;
extern int randy, repeat_weakening;
extern int excellent_weakened_proof;

/*----------------------------------------------------------------*/
/* internally defined functions                                   */
/*----------------------------------------------------------------*/

int time_now();
void init_stats();
int init_options();
int grab_number();
void fatal();
void enter_theorem_name();
void init_search_space();
void init_clause_hash_table();
void init_random_hash_array();
void free_space_close_files();
void read_buffer_word();
void interchange_fxn_const();
int is_fxn_a_constant();
void sort_constants_fxns();
void format_clause();
int count_lit_characters();
void max_term_in_any_lit();
int get_pred_fxn_symbols();
void print_search_options();
void prepare_clauses();
void hash_all_base_clauses();
int iterative_deepening();
void get_clause_stats();
void get_good_clauses();
void write_out_results();
int main();
 
/*----------------------------------------------------------------*/
/* arrays used for reading in input file and converting           */
/* to format compatible with the search.                          */
/*----------------------------------------------------------------*/
 
char compiled_code_buffer[MAX_COMPILED_THM_SIZE];
char *comp_code_buffer_index;
int compiled_code_size[14];
 
pfnode pfspace[PF_SPACE_SIZE];
int kode[MAX_INPUT_THM_SIZE];
char name_buff[MAX_CHAR];
cnode pred_ptr_list[PRED_PTR_SIZE];

/*----------------------------------------------------------------*/
/* space for output, stored temporarily until output to disk file */
/*----------------------------------------------------------------*/

char *OutPtr2;
char *OutPtr;
char results[300000];
char results_slave[300000];

/*----------------------------------------------------------------*/
/* space for clause list generated in elim.c for simplifying      */
/*----------------------------------------------------------------*/

int  elim_inferences[ELIM_ARRAY_SIZE]; /* clause list for elim.c  */

/*----------------------------------------------------------------*/
/* clist contains pointers to the clauses (base clauses and       */ 
/* clauses generated during the search.)                          */
/*----------------------------------------------------------------*/

clause *clist[CSPACE_CLAUSES];    /* first few are base clauses   */
int good_clauses[MAX_BASE_CLAUSES];
clause *ulist[500];

int decompose, decompose2, most_decomposible, decomposible_parts;
int var_vector[16][100], best_var_vector[16][100], split_temp[16];
int split_pbs, unsolved_split_pbs, solved_split_lits[16];
int unsolved_exists_pbs;
int solved_split_pbs[16];
int exists_split_pbs[16];
int exists_split_lits[16];
int pick_split_pbm;

/*----------------------------------------------------------------*/
/* each entry in depth is the top, bottom, and current clause     */
/* under search as well as a pointer to the current clause.       */
/* each entry is thus four int  words.                            */
/*----------------------------------------------------------------*/

dnode  depth[DEPTH_PTR_SPACE];    /* depth ptr info               */

/*----------------------------------------------------------------*/
/* hr_hashcodes store the hashcode that was used in an            */
/* instantiation (in calc_inferences())                           */
/*----------------------------------------------------------------*/

int hr_hashcodes[3][50];          /* array to store hr hashcodes  */

/*----------------------------------------------------------------*/
/* the following two arrays occupy a large amount of memory       */
/* the space is dynamically allocated and is currently set to:    */
/* clause space: 256K words     clause_hash_table: 256K entries   */
/*----------------------------------------------------------------*/  

int  *clause_space;
int *u_space;
hnode *clause_hash_table;           /* ptrs to clause hash tables */
unsigned int o_arraysize;
int array_sizem1;
unsigned int cspace_size;
int array_power;

/*----------------------------------------------------------------*/
/* rand_array is dynamically allocated. its size depends on the   */
/* particular theorem.                                            */
/*----------------------------------------------------------------*/
 
rnode *rand_array;
 
/*----------------------------------------------------------------*/
/*  arrays for storing name of input and outpout theorem          */
/*----------------------------------------------------------------*/

char *theorem_ptr, *theorem_out_ptr, *theorem_out_ptr2;
char *orig_thm_ptr;
char orig_thm_name[256]; /* name of theorem read in               */
char theorem_name[256];  /* name of theorem read in               */
char theorem_out_name[256];/* name of file where output is placed */
char theorem_out_name2[256];/* if sol found, new name of theorem  */
int  theorem_out_name2end; /* ptr to last char in output file name*/
int number_chars;     /* number of characters in name of theorem  */ 
int slash_char;            /* location of last slash of input thm */

/*----------------------------------------------------------------*/
/* variables that are concerned with recording time               */
/*----------------------------------------------------------------*/
 
int elim_starting_time; 
int eliminate_time;
int search_start_time;
int s_starttime;
int search_time;
int phase1_time;                      /* time for phase 1         */
 
int base;                           /* ID base last base clause   */
int org_base;                 /* original number of base clauses  */
int clauses_alive;       /* count of the number of ACTIVE clauses */
int nc;                          /* # clauses in negated concl.   */
int iter_depth;                    /* current iteration depth     */
int solved_iteration;            /* iteration number when solved  */
int ply;                           /* current search depth        */
int nextc;                     /* location where next clause goes */
int nextu;
int lit_in_hash;
int sub_array_last1;
int rename_fxns, rename_preds;
int length_of_proof;
int outside_range, satisfiable, unknown, tough, theorem_rating;
int weakened_ply, weakened_clause;
int weakened_time;
int proof_length;
 
/*----------------------------------------------------------------*/
/* number_rows,       numb_cols, numb_entries in the random array */
/*----------------------------------------------------------------*/
 
int numb_rows, numb_cols, numb_entries, rows_less_nfxns;
 
/*----------------------------------------------------------------*/
/* variables concerned with the state of the search               */
/*----------------------------------------------------------------*/
 
int in_elim;                    /* in_elim = 0 when in eliminate  */
int simple_sol;              /* simple_sol = 1 if elim finds sol  */
int extra_pass;     /* denotes whether more than 1 pass was done  */
int phase; /* search phase: elim = 0, 1st pass = 1, 2nd pass = 2  */
int big_bug;      /* set to 1 when a hash error misleads prover   */

FILE *infile, *outfile;
 
char input_buffer[256]; /* buffer for input file name and options */ 
 
int  ww;  /* temp loc for reading chars from compiled_code_buffer */
 
int  *var_substitution_tbl[MAX_VARS*2]; /* var substitution table */
int  pstack[500],*psp;            /* used in unify.c             */
 
int  equal_pred;     /* =1 if "equal" or "Equal" pred is in thm   */
int  equal_clist_index;       /* the clause id of the eq identity */
int  units_begin;     /* the index of the first unit clause       */

int max_var;
int number_preds;  /* number of different preds in input clauses  */
int number_fxns;   /* number of different fxns in input clauses   */
int number_constants; /* # of different constants in input cls    */
int number_const_nc;  /* number of constants in negated conclus   */
int number_const_hash; /* number of constants for hashing         */
int max_lit;
int max_term_in_lit;                /* maximum terms in a literal */
int max_term_in_nc;                 /* fattest literal in nc      */
int max_vars_in_lit;
int tab_amount;      /* used to keep track of the number of tabs  */

int  maxl, maxv, maxx; /* max lits, vars, trms in clause in proof */
 
char *pidtbl[MAX_PRED_IDS];   /* ptrs to strings for predicates   */
char *fidtbl[MAX_FXNS_IDS];   /* ptrs to strings for functions    */
int skolem[MAX_FXNS_IDS];     /* 1 if constant in nc only         */

cnode *ctbl[MAX_BASE_CLAUSES];
 
int batch_start_time, batch_total_time;
int batch_tried, batch_solved, batch_satisfiable;
int batch_solved_array[2000];

/*----------------------------------------------------------------*/ 
/* The following are vars that record statistics on the search.   */
/*----------------------------------------------------------------*/
 
int  s_nod;                       /* # nodes on all iterations    */
int  s_nod0;                      /* # nodes at beginning of pass */
int  s_res;              /* # resolvents from binary resolution   */
int  s_fac;              /* # factors inferred                    */
int  s_subsume;          /* # of s-subsumed clauses (see 6.14)    */
int  s_subfxn;           /* subsumed ending in fxn                */
 
int  s_htentry; /* # of entries in clause hash table (see 6.10.3) */
int  s_hthit;   /* # of clause hash table hits                    */
int  s_htelim;  /* # of clauses elim by ht hits (see 6.10.4)      */
int  s_htfail;  /* # of failures to insert clause in hash table   */
 
int  s_htentry_phase1;    /* # of entries in cl ht on phase 1     */
int  s_hthit_phase1;      /* # of ht hits in phase 1              */
int  s_htelim_phase1;     /* # of clauses elim by ht hits in ph 1 */
int  s_htfail_phase1;     /* # of ht failures in phase 1          */

int  s_ltentry_phase1;    /* # of entries in lit ht on phase 1    */
int  s_lthit_phase1;      /* # of unit ht resolutions in phase 1  */
int  s_ltfail_phase1;     /* # of unit clause failures in phase 1 */
int  s_unit_res_phase1;   /* # of unit ht resolutions in phase 1  */
int  s_unit_sub_phase1;   /* # of unit clause subsump in phase 1  */
int  s_subsume_phase1;    /* # of s-subsumed clauses in phase 1   */

int  s_ltentry;           /* # of single literal clauses          */
int  s_lthit;     /* # unit hash table resolutions (see 6.11.3)   */
int  s_ltfail;    /* # unit clause failures on entering ht        */
int  s_unit_res;  /* # of hash table resolutions (see 6.11.2)     */
int  s_unit_sub;  /* # of unit clause subsumptions (see 6.11.3)   */
int  s_default;   /* equals zero if default parameters are used   */

int  s_subs;      /* number of attempts to find a string subst.   */
int  s_subs_suc;  /* number of successful "   "   "   "    "      */

int  s_too_many_terms;
 
/*----------------------------------------------------------------*/
/* the following are variables that (in most cases)               */
/* the user can set as desired.  They control the search.         */
/*----------------------------------------------------------------*/
 
int o_elim;       /* o_elim = 1 => do elim, else do not do elim   */
int o_merge;      /* MERGE = ancestors must be merge clauses      */
int o_nc;         /* 1 = nc-strategy                              */
int o_hash;       /* 1 = clause hash table used in proof          */
int o_extend;     /* UNIT = allow unit extension past maxd        */
int o_substitution; /* 1 => use sub. table; 0 => don't use        */
int o_subsume;    /* 1 = simple-subsumbtion used (0 => not used)  */
int o_maxlit;     /* n = max size of clauses (0 = no limit)       */
int o_maxvar;     /* o_maxvar = max # vars in a clause            */
int o_maxterm;    /* max terms in literal                         */
int o_maxtermM1;  /* max terms in literal - 1                     */
int o_maxsterm;   /* max term in literal in search phase          */
int o_maxtime;    /* max time to search for a proof               */
int o_batchtime;
int o_deltalit;
int o_deltaterm;
int o_constants;
int o_maxtermT, o_maxlitT, o_maxvarT;
int max_lit0;
 
int o_step;       /* step depth                                   */
int o_print;      /* print depth                                  */
int o_print_low;  /* first node to print                          */
int o_print_high; /* last node to print                           */
int o_debug;      /* user is supposed to be unaware of this optn  */
int o_relevent;   /* = 1 => print only clauses involved in proof  */
int o_elimprint;  /* o_elimprint = 1 => print elim info           */
int o_tptp;       /* o_tptp = 1 => input file in tptp format      */
int o_batch;      /* o_batch = 1 if batch file is input           */
int o_learn;      /* 0 => no learning, 1 => learning              */
int o_learnM;
int o_diff_top;   /* level at which theorems are eliminated       */
int o_diff_bottom;/* level at which theorems are eliminated       */

int x_maxterm;    /* x_maxterm = 0 => calc o_maxterm              */
int x_stop;       /* can be set to 1 by user in interactive mode  */
int x_cr;         /* when printing clause, add cr when o_cr = 1   */

/*----------------------------------------------------------------*/
/* Parallel search parameters                                     */
/*----------------------------------------------------------------*/

int my_tid;
int myparent;
int tid[400];           /* permits 400 cmputers in the network    */
int cc;
char buf[20000];
int spawned, to_spawn;  /* Number of slaves spawned, to spawn     */
int info;               /* pvm running tasks info */
int ntask;
struct pvmtaskinfo *taskp;
int spawned_alive[400];
int theorem_number;
int fix_proof_once;
int add_time;
int SlaveFoundProof;
int this_computer;
int SolvedBy;
int o_print_slave;
int cade_format;
int slaves_killed;

static char alphabet32[] =
 {'0','1','2','3','4','5','6','7','8','9',
  'a','b','c','d','e','f','g','h','i','j',
  'k','l','m','n','o','p','q','r','s','t',
  'u','v','w','x','y','z','A','B','C','D',
  'E','F','G','H','I','J','K','L','M','N'};

/*----------------------------------------------------------------*/
/*----------------------------------------------------------------*/

static int term_column[] =
{0,0,0,0,0,0,1,1,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,7,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8};

static int tms3[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
  10,   12,   14,   17,   20,   23,   28,   35,   50,  /*05*/
   0,   14,   16,   20,   23,   26,   32,   36,   50,  /*07*/
   0,    0,   18,   22,   25,   28,   32,   40,   50,  /*09*/
   0,    0,    0,   24,   27,   30,   33,   42,   50,  /*12*/
   0,    0,    0,    0,   27,   30,   33,   42,   50,  /*15*/
   0,    0,    0,    0,    0,   32,   35,   44,   50,  /*18*/
   0,    0,    0,    0,    0,    0,   36,   46,   50,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   48,   50,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   50,  /*+25*/
};


static int tms4[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
   9,   12,   14,   16,   18,   21,   24,   30,   40,  /*05*/
   0,   14,   16,   18,   18,   21,   24,   30,   40,  /*07*/
   0,    0,   18,   18,   20,   22,   24,   30,   40,  /*09*/
   0,    0,    0,   20,   20,   24,   25,   32,   40,  /*12*/
   0,    0,    0,    0,   20,   24,   25,   32,   40,  /*15*/
   0,    0,    0,    0,    0,   24,   26,   32,   40,  /*18*/
   0,    0,    0,    0,    0,    0,   26,   32,   40,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   32,   40,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   40,  /*+25*/
};

static int tms5[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
   9,   12,   14,   16,   18,   19,   22,   23,   24,  /*05*/
   0,   14,   16,   18,   18,   19,   22,   23,   24,  /*07*/
   0,    0,   18,   18,   20,   19,   22,   23,   24,  /*09*/
   0,    0,    0,   20,   20,   19,   22,   23,   24,  /*12*/
   0,    0,    0,    0,   20,   19,   22,   23,   24,  /*15*/
   0,    0,    0,    0,    0,   19,   22,   23,   24,  /*18*/
   0,    0,    0,    0,    0,    0,   22,   23,   24,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   23,   24,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   24,  /*+25*/
};

static int tms[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
  10,   10,   10,   10,   10,   10,   10,   10,   10,  /*05*/
   0,   10,   10,   11,   11,   11,   11,   11,   10,  /*07*/
   0,    0,   10,   11,   12,   12,   12,   12,   10,  /*09*/
   0,    0,    0,   12,   12,   12,   12,   12,   10,  /*12*/
   0,    0,    0,    0,   15,   15,   15,   12,   10,  /*15*/
   0,    0,    0,    0,    0,   14,   14,   12,   10,  /*18*/
   0,    0,    0,    0,    0,    0,   14,   12,   10,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   12,   10,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   10,  /*+25*/
};

/*----------------------------------------------------------------*/
/*      time_now()                    ...................... #1
/*
/*      added to consolidate all system calls to theo 
/*
/*      called by:  main()                         theo.c
/*                  iterative_deepening            theo.c
/*                  halted_search_options()        tdump.c
/*                  eliminate()                    elim.c
/*
/*      calls:      nobody
/*
/*      returns:    time of day as a an integer
/*---------------------------------------------------------------*/
 
int time_now()
{
return(time(0));
}
 
/*--------------------------------------------------------------*/
/*      init_stats()                   ..................... #2
/*
/*      init statistical parameters gathered during the search
/*
/*      called from:    main()       theo.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void init_stats()
{
s_nod  = 0; 
s_fac  = 0;             
s_res  = 0;
s_nod  = 0;             
s_nod0 = 0;
s_subs = 0;
s_subs_suc = 0;
solved_iteration = 0;
 
/*---------------------------------------------------------*/
/* stats for clause hash table and lit clause hash table   */
/*---------------------------------------------------------*/
	
s_htentry = 0;       
s_hthit   = 0;      
s_htfail  = 0;
s_ltentry = 0;       
s_lthit   = 0;      
s_ltfail  = 0; 
s_htelim  = 0;
s_unit_res = 0;
s_unit_sub = 0;
s_subsume = 0;
s_subfxn = 0;
s_too_many_terms = 0;
 
tab_amount = 0;        
	
}
 
/*----------------------------------------------------------------*/
/*      init_options()                          .............. #3
/*
/*      reads the search options that the user requests in the
/*      command line.
/*
/*      called from:  enter_theorem_name()    theo.c
/*
/*      calls:        grab_number()           theo.c
/*
/*      returns:      the number of chars in the name of the theorem
/*                    or a -1 if help was requested
/*----------------------------------------------------------------*/
 
int init_options()
{
int i, j;
unsigned int k;
 
/*-----------------------------------------------------------------*/
/* initialize options to default values                            */
/*-----------------------------------------------------------------*/
    
o_elim      = 1;  /* Option 1... elim phase: 1/0/2 => do/skip/only */
o_merge     = 1;  /* Option 2... merge strat: 1/0 => off/on        */
o_nc        = 0;  /* Option 3... linear-nc: 0/1 => off/on          */
o_hash      = 1;  /* Option 4... clause ht on: 1/0=> cht off/on    */
o_extend    = 1;  /* Option 5... extended search: 1/0=> on/off     */
o_substitution = 1; /* Option 18.. use substitution table: 1/0     */
o_subsume   = 1;  /* Option 6... s-subsumption: 1/0 => use/don't   */
o_maxlit    = 0;  /* Option 7... maximum # lits in a cls in proof  */
o_maxterm   = 16; /* Option 8... maximum # terms in a literal      */
x_maxterm   = 0;
o_maxsterm  = 0;
 
o_maxvar    = 0;   /* Option 9... maximum # diff vars in a clause  */

                        /* Option 10.. size of clause_hash_table   */

o_arraysize = 4*4194304; 
array_sizem1 = 4*4194304 - 1;  
cspace_size = 2097152;
array_power = 24;

o_tptp      = 3;   /* Option 11.. 0/1 not in/in tptp format        */
/*o_maxtime   = 3600;*/
                   /* Option 12.. maximum time for proof..seconds  */
o_batchtime = 3600;

o_elimprint = 0;   /* Option 13.. 0/1 do/ do not give phase 0 info */
o_step      = 0;   /* Option 14.. step through tree to ply o_step  */ 
o_print     = 0;   /* Option 15.. print tree to ply o_step         */
o_print_low = 0;
o_print_high= 1000000000;

o_debug     = 0;   /* Option 16.. debug option: 0/1 => off/on      */
o_relevent  = 0;   /* Option 17.. 0/1 do/do not prt all proof cls  */

/*-----------------------------------------------------------------*/
/*        END OF OPTIONS                                           */
/*-----------------------------------------------------------------*/

s_default   = 0;
big_bug     = 0;       /* set to 1 if a hash error found inphase 3 */
x_cr        = 0;       /* set to 1 => print cr when printing cls   */

/*------------------------------------------*/
/* scan the arguments passed to the program */
/* if a "?" appears anywhere on the line    */
/* print out a "help" message               */
/*------------------------------------------*/
 
for(i = 250 ; i > 1; i--) {
   if(input_buffer[i - 1] == 'l' && input_buffer[i - 2] == '-') {
        break;}
   if(input_buffer[i - 2] == '?') {
	
/*--------------------------------------------------------------*/
/* a help message, triggered by a "?" anywhere on input line    */
/*--------------------------------------------------------------*/

printf("\n--------------------------------------------------------");
printf("\n| Enter theorem name and then options as in this example:");
      printf(
      "\n| ");
      printf(
      "\n| STARK103.THM k1 m0 z1 h0 n0 b0 l4 x6 v6 t7 e1 p3 s2 d0");
      printf(
      "\n| ");
      printf(
      "\n| where:");
      printf(
      "\n|   k0: skip elim; k1 carry out elim; k2 carry out only elim");
      printf(
      "\n|   m0: turn off linear-merge strategy");
      printf(
      "\n|   z1: turn on linear-nc search strategy");
      printf(
      "\n|   h0: turn off hash table for clauses");
      printf(
      "\n|   n0: turn off extended search strategy");
      printf(
      "\n|   b0: turn off s-subsumption strategy");
      printf(
      "\n|   l4 x6 v6: set max # of lits, terms and vars to 4,6,6");
      printf(
      "\n|   t7: set maximum search time to 7 seconds");
      printf(
      "\n|   e1: print info on elim fxn; e2: include resolutions");
      printf(
      "\n|   p3: print three levels of search tree");
      printf(
      "\n|   s2: step two levels of search tree");
      printf(
      "\n|   d1: turn on debugger to obtain information on search");
      printf(
      "\n|   r1: when proof is found, list only relevant clauses");
      printf(
      "\n|   Qx: Consider only theorems easier than this level");
      printf(
      "\n|   qx: Consider only theorems only more difficult");
      printf(
      "\n|   a524288: set clause_hash_table to 512K entries");
      printf(
      "\n| By default, k=m=1, z=0, h=n=b=1, t=3600, e=p=s=d=r=0.");
      printf(
      "\n| Default values for l,x,v depend on the theorem.");
      printf(
      "\n| To prove a batch of theorems add 'batch' to command line");
      printf(
      "\n| and the redirection symbol '<' and a file where theorems");
      printf(
      "\n| are listed.");
      printf(
      "\n------------------------------------------------------");
		
      for(i = 0; i < 256; i++) {
         input_buffer[i] = ' ';}           /* clr input buffer  */
      return(-1);}                       /* -1 => re-enter      */
			
   if(input_buffer[i] == ' ') {
			
/*------------------------------------------------------*/
/* P0 => min slave printing; P1 => normal slave print   */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'P') {
	 o_print_slave = grab_number(i + 2);
	 if(o_print_slave > 2 || o_print_slave < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------*/
/* L0 => no Learn strategy; L1 => uses Learn startegy   */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'L') {
	 o_learn = grab_number(i + 2);
	 o_learnM = o_learn;
	 if(o_learn > 1 || o_learn < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------*/
/* Qx => eliminate theorems who are more difficult      */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'Q') {
	 o_diff_top = grab_number(i + 2);
	 if(o_diff_top > 100 || o_diff_top < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------*/
/* qx => eliminate theorems easier than x               */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'q') {
	 o_diff_bottom = grab_number(i + 2);
	 if(o_diff_bottom > 100 || o_diff_bottom < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------*/
/* k0 => skip elim phase, k1 => carry out elim phase    */
/* k2 => carry out only elim phase                      */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'k') {
         s_default = 1;
	 o_elim = grab_number(i + 2);
	 if(o_elim > 3 || o_elim < 0) {
	    printf("\nbad argument\n");return(-1);}}
	
/*------------------------------------------------------*/
/* m0 => no merge strategy; m1 => uses merge startegy   */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'm') {
         s_default = 1;
	 o_merge = grab_number(i + 2);
	 if(o_merge > 1 || o_merge < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------*/
/* z0 => no nc strategy; z1 => uses nc strategy         */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'z') {
         s_default = 1;
	 o_nc = grab_number(i + 2);
         if(o_nc > 1 || o_nc < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* h0 => turn off clause_hash_table h1 => turn on clause_hash_table */
/*------------------------------------------------------------------*/
 
      if(input_buffer[i + 1] == 'h') {
         s_default = 1;
	 o_hash = grab_number(i + 2);
	 if(o_hash > 1 || o_hash < 0) {
	    printf("\nbad argument\n");return(-1);}}
 
/*--------------------------------------------------------------*/
/* n0 => no ext'd search str'gy; n1 => uses ext'd search str'gy */
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'n') {
         s_default = 1;
	 o_extend = grab_number(i + 2);
	 if(o_extend > 1 || o_extend < 0) {
	    printf("\nbad argument\n");return(-1);}}

/*--------------------------------------------------------------*/
/* c0 => no substitution array; c1 => use substitution array    */
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'c') {
         s_default = 1;
	 o_substitution = grab_number(i + 2);
	 if(o_substitution > 1 || o_substitution < 0) {
	    printf("\nbad argument\n");return(-1);}}
	
/*--------------------------------------------------------------*/
/* batch => set o_batch to 1, for a batch file
/* b0 => no simple subsumptions; b1 => yes..simple subsumptions 
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'b') {
	 if(input_buffer[i + 2] == 'a' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'c' &&
            input_buffer[i + 5] == 'h') {
	    o_batch = 1;
	    continue;}
         o_subsume = grab_number(i + 2);
         if(o_subsume  > 1 || o_subsume  < 0) {
            printf("\nbad argument\n");return(-1);}}
	    
/*--------------------------------------------------------------*/
/* if lx is found, set max number of literals in a clause to x  */
/* x is restricted to fall between 2 and 20                     */
/*--------------------------------------------------------------*/
						
      if(input_buffer[i + 1] == 'l') {
         s_default = 1;
	 o_maxlit = grab_number(i + 2);
         if(o_maxlit < 1 || o_maxlit > 20) {
	    printf("\nbad argument\n");return(-1);}}
			
/*------------------------------------------------------------------*/
/* if xX is found, set max number of terms in a literal to X        */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'x') {
         s_default = 1;
         o_maxsterm = grab_number(i + 2);
         x_maxterm = 1;
         if(o_maxsterm < 2 || o_maxsterm > 200) {
            printf("\nbad argument\n");return(-1);}}
			
/*------------------------------------------------------------------*/
/* if vX is found, set max number of variables in a clause to X     */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'v') {
         s_default = 1;
         o_maxvar = grab_number(i + 2);
         if(o_maxvar < 2 || o_maxvar > 32) {
            printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* if tptp is found, set o_tptp to 1................................*/
/* if tX is found, set max search time to X                         */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 't') {
	 if(input_buffer[i + 2] == 'p' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'p' &&
	    input_buffer[i + 5] == '3') {
	    o_tptp = 3;
	    continue;}

	 if(input_buffer[i + 2] == 'p' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'p' &&
	    input_buffer[i + 5] == '2') {
	    o_tptp = 2;
	    continue;}

	 if(input_buffer[i + 2] == 'p' &&
	    input_buffer[i + 3] == 't' &&
	    input_buffer[i + 4] == 'p' &&
	    input_buffer[i + 5] == '0') {
	    o_tptp = 0;
	    continue;}


         o_batchtime = grab_number(i + 2);
         if(o_batchtime < 1 || o_batchtime > 200000) {
            printf("\nbad argument\n");return(-1);}}
			
/*-------------------------------------------------------------------*/
/* fist order formula format => o_tptp = 4
/*-------------------------------------------------------------------*/

       if(input_buffer[i + 1] == 'f') {
          if(input_buffer[i + 2] == 'o' &&
            input_buffer[i + 3] == 'f') {
             o_tptp = 4;}}

/*----------------------------------------------*/
/* if e0 => turn elim info printout off,        */
/* if e1 => set eliminate info printout on      */
/* if e2 => set elim info printout on,          */
/*          and list resolutions                */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'e') {
         o_elimprint = grab_number(i + 2);
         if(o_elimprint < 0 || o_elimprint > 20) {
            printf("\nbad argument\n");return(-1);}}

/*----------------------------------------------*/
/* if px is found, set printing depth to x      */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'p') {
         o_print = grab_number(i + 2);
         if(o_print < o_step) {
            o_print = o_step;}
         if(o_print < 0 || o_print > 40) {
            printf("\nbad argument\n");return(-1);}}
			
/*----------------------------------------------*/
/* if Hx is found, print until node x is found  */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'H') {
         o_print_high = grab_number(i + 2);
         if(o_print_high < 0 || o_print_high > 1,000,000,000) {
            printf("\nbad argument\n");return(-1);}}

/*----------------------------------------------*/
/* if Bx is found, begin printing  after node x */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'B') {
         o_print_low = grab_number(i + 2);
         if(o_print_low < 0) {
            printf("\nbad argument\n");return(-1);}}

/*--------------------------------------------------------------*/
/* if sx is found, set depth of stepping during printing to x   */
/*--------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 's') {
         o_step = grab_number(i + 2);
         if(o_print < o_step) {
            o_print = o_step;}
         if(o_step < 0 || o_step > 40) {
            printf("\nbad argument\n");return(-1);}}
			
/*----------------------------------*/
/* if d is found, set debugger on?  */
/*----------------------------------*/
			
      if(input_buffer[i + 1] == 'd') {
         o_debug = grab_number(i + 2);
         if(o_debug  > 3 || o_debug  < 0) {
            printf("\nbad argument\n");return(-1);}}
				
/*------------------------------------------------------------------*/
/* if r1 is found , set o_relevent=1 to print only relevant clauses */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'r') {
         o_relevent = grab_number(i + 2);
         if(o_relevent > 1 || o_relevent < 0) {
            printf("\nbad argument\n");return(-1);}}

/*------------------------------------------------------------------*/
/* axxxxx determines the size of clause_space and the size of       */
/* clause_hash_table.  It must be a power of 2!!!                   */
/*------------------------------------------------------------------*/

      if(input_buffer[i + 1] == 'a') {
         o_arraysize = (unsigned int) grab_number(i + 2);
         array_sizem1 = o_arraysize - 1;
         for(k = 16384; k != o_arraysize ; k = k * 2) {
            if(k > o_arraysize) {
               printf("\nbad argument;");
               if(k == 16384) {printf(" not enough memory space.\n");}
               else {printf(" memory space must be a power of 2.\n");}
               return(-1);}}}}}	
         
/*---------------------------------------------------*/
/* end of option list:                               */ 
/* blank out input_buffer except for name of theorem */
/*---------------------------------------------------*/
	
for(i = 0; i < 255; i++) {
   if(input_buffer[i] == ' ' || input_buffer[i] == '\n') {
      j = i;
      for(j = i; j < 256 ; j++) input_buffer[j] = ' ';
      break;}}

/*--------------------------------------------------*/ 
/* return the number of characters in the name      */
/*--------------------------------------------------*/

return(i);
}
 
/*-----------------------------------------------------------------*/
/*      grab_number(s)                       .............. #4
/*
/*      reads a string of characters and converts them to a
/*      integer equivalent
/*
/*      called from:  init_options()              theo.c
/*                    halted_search_options()     tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      an integer correspond to the string s
/*-----------------------------------------------------------------*/
 
int grab_number(s)
int s;
{
int n;
	
for(n = 0; ; s++) {
   if(input_buffer[s] >= '0' && input_buffer[s] <= '9')
      n = n * 10 + ( (int) (input_buffer[s] - '0') );
   else
      return(n);}
}                                                       
 
/*-----------------------------------------------------------------*/ 
/*      fatal(msg)                          ................. #5
/*
/*      fatal will print an error message on the screen and quit
/*
/*      called by:    main()                theo.c
/*                    prepare_clauses()     theo.c
/*                    format_clause()       theo.c
/*                    eliminate()           elim.c
/*                    enter_theorem_name()  theo.c
/*
/*      calls:        time_now()            theo.c
/*
/*      returns:      to the operating system after reporting
/*                    error
/*-----------------------------------------------------------------*/
 
void fatal(msg)
char *msg;
{
printf("....%s\n",msg);
OutPtr += sprintf(OutPtr,"....%s\n",msg); 

if(*msg == 'D') {
  printf(
        "\n Theorems Tried %d\n Theorems Solved %d\n Satisfiable %d\n",
	 batch_tried,batch_solved,batch_satisfiable);
  OutPtr += sprintf(OutPtr,
	"\n Theorems Tried %d\n Theorems Solved %d\n Satisfiable %d\n",
	 batch_tried,batch_solved,batch_satisfiable);

  batch_total_time = time_now() - batch_start_time;

  printf("\n Total Time (in seconds) %d\n\n",
            batch_total_time);
  OutPtr += sprintf(OutPtr,
	     "\n Total Time (in seconds) %d\n\n",
            batch_total_time);}
   exit(0);
}
 
 
/*-------------------------------------------------------------*/
/*      enter_theorem_name()           ................#6
/*
/*      read in the name of the theorem to be proved
/*
/*      called by:   main()                    theo.c
/*
/*      calls:       CompareStrings()          tptpi.c
/*                   init_options()            theo.c
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void enter_theorem_name(argc,argv)
int argc;
char *argv[];
{
int  bad_name;       /* bad_name = 0 until theorem found in file   */
int i, j, jj, jjj, m;
int  ask_for_help;    /* normally 0, set to 1 if "?" on input line */
int zz;

bad_name = ask_for_help = cade_format = 0;

++argv;
++argv;
if((*argv)[0] == '-' && (*argv)[1] == 'l') {
   (*argv)[0] = ' ';
   (*argv)[1] = ' ';
   cade_format = 1;}
argv--;
argv--;

if(weakened_ply != 0) {
   cade_format = 1;
   for(j = 0 ; j < 256 ; j++) {
      theorem_name[j] = 0;
      theorem_out_name[j] =  0;
      theorem_out_name2[j] = 0;
      input_buffer[j] = ' ';}
 for(j = 0; j < 256; j++) {
    input_buffer[j] = orig_thm_name[j];}
   j = 0;
   goto get_args;}

while(!bad_name) {
   for(j = 0 ; j < 256 ; j++) {
      orig_thm_name[j] = 0;
      theorem_name[j] = 0;
      theorem_out_name[j] =  0;
      theorem_out_name2[j] = 0;
      input_buffer[j] = ' ';}
   ask_for_help = 0;
   while(!ask_for_help) {
      j = 0;
      if(cade_format == 0) {
      printf("\nEnter name of theorem (type '?' for help): ");

/*-------------------------------------*/
/* read from the terminal if no        */
/* srguments on command line           */
/*-------------------------------------*/

      while((zz = getchar()) != '\n') { 
         if(zz == (int) '\b') {
            if(j > 0) {
	       j--; printf("\b \b");}}
         else if(zz == -1) {
            fatal("DONE!");}
         else {
	    input_buffer[j] = orig_thm_name[j] = (char) zz;
	    j++;}} 
      input_buffer[j] = orig_thm_name[j] = ' ';
      j++;}

      for(jj = 1; jj < argc; jj++) {
         ++argv;
         jjj = 0;
         while((orig_thm_name[j] = input_buffer[j] = (*argv)[jjj++]) 
		  != '\0') {
            j++;}
         input_buffer[j] = orig_thm_name[j] = ' ';
	 j++;}
get_args:

      number_chars = init_options();       /* set search options */
      for(jj = 0; jj < number_chars; jj++) {
         printf("%c",input_buffer[jj]);}
      printf("\n");
      if(number_chars != -1) {
         ask_for_help++;
         printf("\n");}}         

   m = 0; 
   for(j = 0 ; j < number_chars ; j++) {  
      if(input_buffer[j] == '.') {m = j;}
      if(input_buffer[j] == '+') {if(!m) m = j;}
      if(input_buffer[j] == '/') {m = 0;}
      theorem_name[j] = input_buffer[j];
      orig_thm_name[j] = theorem_name[j];}

   for(j = number_chars ; j < 256 ; j++ ) {
      theorem_name[j] = 0;}

/*--------------------------------------------------------------*/
/* give output file the same as input file with a suffix ".u"   */    
/*--------------------------------------------------------------*/

   if(m) { 
      for(j = 0; j < m; j++)  {
         theorem_out_name[j] = input_buffer[j];
         theorem_out_name2[j] = input_buffer[j];
         if(j == m) {
            break;}}
      input_buffer[m] = '.';
      theorem_out_name[m] = theorem_out_name2[m] = '.';}
   else {
      for(j = 0 ; j < number_chars + 1 ; j++)  {
         theorem_out_name[j] = input_buffer[j];
         theorem_out_name2[j] = input_buffer[j];
         if(input_buffer[j] == ' ') {
            theorem_out_name[j] = '.';
            theorem_out_name2[j] = '.';
	    break;}}}
			
   theorem_ptr = &theorem_name[0];
   theorem_out_ptr = &theorem_out_name[0];
   theorem_out_ptr2 = &theorem_out_name2[0];
 
   if((infile = fopen(theorem_ptr,"r")) == 0) {
      printf(
      "Cannot open theorem. You may have typed the wrong name.\n\n");
      if(argc > 1) {fatal(" ");}}
   else bad_name = 1;}

if(weakened_ply == 0) {
   if((outfile = fopen(theorem_out_ptr,"w")) == 0) {
      fclose(outfile);
      fatal("\nCannot open output file:\nType cr and rerun");}

OutPtr = &results[0];
OutPtr2 = OutPtr;

*OutPtr++ = 0;     /* put 0s in first 3 locations -- replaced later */
*OutPtr++ = 0;
*OutPtr++ = 0;

o_maxtime = o_batchtime;}

for(m = number_chars - 1; m >= 0; m--) {
   if(input_buffer[m] == '/') {break;}}
if(!m) {m = -1;}

slash_char = m;

for(jj = m + 1; jj < number_chars; jj++) {
   OutPtr += sprintf(OutPtr,"%c",input_buffer[jj]);}

sprintf(OutPtr++,"\n");
if(weakened_ply == 0) {
   theorem_out_name2end = j + 1;}

}
 
/*-------------------------------------------------------------------*/
/*      init_search_space()            ......................#7
/*
/*      allocates space for clauses.
/*      allocates space for clause hash table.
/*      allocates space for pointers at each level in tree
/*
/*      called from:   main()                 theo.c
/*
/*      calls:         dump_calloc_info()     tdump.c
/*
/*      returns:       nil
/*-------------------------------------------------------------------*/
 
void init_search_space()
{       
clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);

clause_space = (int *) calloc(cspace_size,4);   
nextc = 0;
clist[nextc] = (clause *) clause_space;

if(nextu == 0) {
   u_space = (int *) calloc(10000,4);   
   ulist[nextu] = (clause *) u_space;}
}
 
/*-----------------------------------------------------------------*/
/*      init_clause_hash_table()          .................#8
/*
/*      initialize the clause_hash_table to all 0's
/*
/*      called by:      main()                             theo.c
/*                      iterative_deepening()              theo.c
/*
/*      calls:          dump_calloc_info()                 tdump.c
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void init_clause_hash_table()
{
int i, j;

/*----------------------------------------------*/
/* allocate 4 * HT_SIZE int  words for the hash */
/* table.  initially set all words to 0.        */
/*----------------------------------------------*/
 
clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);
dump_calloc_info(
      "clause_hash_table",
      (int) clause_hash_table,-1,"hnodes",o_arraysize);

/*----------------------------------------------*/
/* clear hr_hashcodes array                     */
/*----------------------------------------------*/

for(i = 0; i < 3; i++) {
   for(j = 0; j < 50; j++) {hr_hashcodes[i][j] = 0;}}

return;
}
 
/*-----------------------------------------------------------------*/
/*      init_random_hash_array()    ........#9
/*
/*      initialize the rand_array to random numbers 
/*
/*      called from:   main()                      theo.c 
/*
/*      calls:         dump_calloc_info()           tdump.c
/*
/*      returns:       nil
/*-----------------------------------------------------------------*/
 
void init_random_hash_array()
{
rnode *r1;
int i;
unsigned int num_ent;
	
/*--------------------------------------------------------------*/
/* the number of rows in rand_array = MAX_VARS + numb_fxns      */
/* the number of columns in rand_array = numb_chars             */
/*--------------------------------------------------------------*/
 
numb_rows = MAX_VARS + number_fxns + number_preds;
numb_cols = 200;
rows_less_nfxns = numb_rows - number_fxns - 1;
 
/*--------------------------------------------------------------*/
/* the number of entries in the array rand_array is the         */
/* product of the numb_rows * numb_cols                         */ 
/*--------------------------------------------------------------*/
 
numb_entries = numb_rows * numb_cols;
num_ent = (unsigned int) (numb_entries * 2 + 1); 
if(num_ent < 50) {
   numb_entries = 25;
   num_ent = 51;}

 
/*--------------------------------------------------------------*/
/* allocate space for rand_array                                */
/*--------------------------------------------------------------*/

rand_array = (rnode *) calloc(num_ent * 2,4);
dump_calloc_info(
 "rand_array..........",(int) rand_array,-1,"rnodes",num_ent);
	

/*--------------------------------------------------------------*/
/* make 32-bit random number for each element in rand_array     */
/*--------------------------------------------------------------*/
 
srand(1);
 
for(i = 0 ; i < (int) numb_entries; i++) {
   r1 = &rand_array[i];
   r1->r_hash1 = rand()*16384 + rand();
   r1->r_hash2 = rand()*16384 + rand();}

}
 
/*-------------------------------------------------------------*/
/*      free_space_close_files()  ...................#10
/*
/*      frees space and closes files used in phases 1 and 2
/*
/*      called by:   main()                 theo.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void free_space_close_files()
{
int i, j;

free(clause_hash_table);
free(rand_array);
free(clause_space);  
fclose(infile);
if(weakened_clause == -1 || excellent_weakened_proof || satisfiable) {
// printf("KILL SLAVES\n");
   fclose(outfile); 
   slaves_killed = 0;
   for(i = 0; i < spawned; i++) {
      if(spawned_alive[i] == 0) {
         if((j = pvm_kill(tid[i])) != 0) {
            printf("\nprocess %d not killed -- Error(%d)\n",i,j);}
	 else {
	    slaves_killed++;
	    spawned_alive[i] = 1;}}}}
//printf("FREE_SPACE: %d slaves killed\n",slaves_killed);
}
 
/*--------------------------------------------------------------*/
/*      read_buffer_word()               .............. #11
/*
/*      read a word (integer) from the input file
/*
/*      called by:    prepare_clauses()     theo.c
/*                    format_clause()       theo.c
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void read_buffer_word()
{
union u_tag {
   int w; 
   char ttext[4];} 
   temp;
			
int j;
	
for(j = 0; j < 4; j++) {
   temp.ttext[j] = *comp_code_buffer_index++;}
ww = temp.w;
}
 
 
/*--------------------------------------------------------------*/
/*      interchange_fxn_const()          ............. #12
/*
/*       called by:      sort_constants_fxns()      theo.c
/*
/*       calls:          nobody
/*
/*       returns:         nil
/*--------------------------------------------------------------*/
 
void interchange_fxn_const(i,j)
int i,j;
{
int k, *p, a;
clause *c;
literal *l;
char *fide;
	
for(k = 0; k <= base ; k++) {
   c = clist[k];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      p = &l->l_id;
      for(a = (*p++ & ARGMASK); a > 0; a--) {
	     if(*p & FCNMASK) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           *p = (*p & ~IDMASK) + (j << ARGSHIFT);}
	        else if(((*p & IDMASK) >> ARGSHIFT) == j) {
	           *p = (*p & ~IDMASK) + (i << ARGSHIFT);}
	        a += *p++ &ARGMASK;}
	     else {
	        p++;}}}}
			
fide = fidtbl[i];
fidtbl[i] = fidtbl[j];
fidtbl[j] = fide;
}
 
/*--------------------------------------------------------------*/
/*      is_fxn_a_constant()                .............. #13
/*
/*      called by:      sort_constants_fxns()       theo.c
/*
/*      calls:          nobody
/*
/*      returns:        1 if fxn is a constant, 0 otherwise
/*-----------------------------------------------------------*/
 
int is_fxn_a_constant(i,top_clause,bottom_clause)
int i,top_clause,bottom_clause;
{
int k, a, *p;
clause *c;
literal *l;
	
for(k = top_clause; k <= bottom_clause ; k++) { 
   c = clist[k];
   
   for(l = c->c_lit; l != 0; l = l->l_next) {  /* scan lits  */
      p = &l->l_id;
      for(a = (*p++ & ARGMASK); a > 0; a--) {
	     if(*p & FCNMASK) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           if((*p & ARGMASK) == 0) return(1);
	           else return(0);}
	        a += *p++ & ARGMASK;}
	     else {
	        p++;}}}}
return(0);
}
 
/*--------------------------------------------------------------*/
/*      sort_constants_fxns()           .............. #14
/*
/*      sort the constants and functions so that the constants
/*      are placed at the top of the list of functions and constants
/*      once that is done, place the constants in the nc at the
/*      head of the constants list.
/*
/*      called by:   prepare_clauses()          theo.c
/*
/*      calls:       is_fxn_a_constant()        theo.c
/*                   interchange_fxn_const()    theo.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void sort_constants_fxns()
{
int i, k;
 
/*-----------------------------------------------*/
/*    put the constants at the head of the list  */
/*-----------------------------------------------*/
 
number_constants = 0;
for(i = 1; i <= number_fxns ; i++) {
   if(is_fxn_a_constant(i,0,base)) {
      number_constants++;       
      if(i >= number_fxns) break;}
   else {
      for(k = number_fxns; k > i; k--) {
	     if(is_fxn_a_constant(k,0,base)) {
	        interchange_fxn_const(i,k);
	        number_constants++;
	        break;}}}}  

/*-----------------------------------------------*/
/*   now resort the constants in the nc to the   */
/*   top of the list.                            */
/*-----------------------------------------------*/

number_const_nc = 0;
for(i = 1; i <= number_constants; i++) {

   if(is_fxn_a_constant(i,base - nc + 1,base)) { 
      number_const_nc++; 
      continue;}
   else {
      for(k = i + 1; k <= number_constants; k++) {
	 if(is_fxn_a_constant(k,base - nc + 1,base)) {
            number_const_nc++;
	    interchange_fxn_const(i,k);
	    break;}}}} 

if(number_constants <= 5) {
   number_const_hash = number_constants;}
else {
   number_const_hash = 5;}

for(i = 1; i <= number_constants; i++) {
   skolem[i] = 1;
   if(is_fxn_a_constant(i,0, base - nc)) {
      skolem[i] = 0;}} 


for(i = 1; i <= number_fxns; i++) {
   if(o_debug > 0) {
      printf(" %d",i);
      OutPtr += sprintf(OutPtr," %d",i);}
      printf(" %s ",fidtbl[i]);
      OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
   if(skolem[i] == 1) print0("(S)");
   if(i == number_constants) {
      printf(" :");
      OutPtr += sprintf(OutPtr," :");}
   if(i == number_const_nc) {
      printf(" .");
      OutPtr += sprintf(OutPtr," .");}}
printf("\n");
sprintf(OutPtr++,"\n"); 
}

/*--------------------------------------------------------------*/
/*      format_clause()                   .............. #15
/*
/*      format_clause will read clause from the input and stuff
/*      it into the clist
/*
/*      called from:    prepare_clauses()   theo.c
/*
/*      calls:          read_buffer_word()  theo.c
/*                      fatal()             theo.c
/* 
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void format_clause()
{
clause *c;
literal *previous_literal;
int *p,a;
	
ply = 0;       
read_buffer_word();read_buffer_word();/* these are leading 0's */
 
/*------------------------------------------------------*/
/* initialize clause header (this is not part of the    */
/* input file, we add it as we read in the clause)      */
/*------------------------------------------------------*/
 
c = clist[nextc++];                  /* the new clause         */
c->c_id = nextc;
c->c_bits = ACTIVE | MERGE | BASE; /* base clauses are merges  */
if(nextc > base - nc +1) {
    c->c_bits |= NC;}                      /* clause in NC     */
c->c_rpar = c->c_lpar = 0;       /* no parents (base clause)   */
c->c_lit = 0;                        /* ^ first literal        */
c->c_nlit = 0;                             /* # literals       */
c->c_nv   = 0;                       /* # distinct variables   */
	
p = (int  *) (c + 1);           /* where predicate code goes   */
previous_literal = 0;                 /* literal chain         */
 
/*----------------------------------------*/
/* read in each literal, end clause at -1 */
/*----------------------------------------*/    
 
read_buffer_word(); 
while(ww != -1) {
   if(!previous_literal) {               /* link last literal  */
      c->c_lit = (literal *) p;}
   else {
      previous_literal->l_next = (literal *) p;}
      
   previous_literal = (literal *) p;
   previous_literal->l_id = ww;
   previous_literal->l_next = 0;
 
   p = (int *) (previous_literal + 1);
 
   for(a = ww & ARGMASK; a > 0; a--) {
      read_buffer_word();
      if(ww & FCNMASK) {                          /* function? */
	     *p++ = ww;                            /* yes       */
	     a += ww & ARGMASK;}                   /* end if    */ 
      else {
	     ww >>= ARGSHIFT;                      /* var ID    */
	     *p++ = ww;
	     if(ww > c->c_nv) {
	        c->c_nv = ww;}}}                     /* end for */
   
/*----------------------------------------------------------------*/
/* transform equality predicates to standard form                 */
/*----------------------------------------------------------------*/
 
   c->c_nlit++;
		
   read_buffer_word();}                       /* end while  */
 
/*-----------------------------------------------------------------*/
/* set up a ptr to the next clause, pointed to by clist[nextc]     */
/*-----------------------------------------------------------------*/
 
c->c_end = (clause *) p;
clist[nextc] = c->c_end;
	
/*--------------------------------------------------*/
/* if clause has only one literal, set UNIT flag on */
/*--------------------------------------------------*/
 
if(c->c_nlit == 1) {
   c->c_bits |= UNIT;}
 
if(++c->c_nv > MAX_VARS) {
   printf("Clause %d has too many variables!!\n",nextc);
   OutPtr += sprintf(OutPtr,
	   "Clause %d has too many variables!\n",nextc);
   fatal("Too many variables in clause");}
 
if(!c->c_nlit) {
   fatal("Empty clause in base");}
}
 
/*---------------------------------------------------------------*/
/*      count_lit_characters()               .............. #16
/*
/*      count_lit_characters counts terms in lit l
/*
/*      called from:  max_term_in_any_lit()        theo.c
/*
/*      calls:        nobody
/*
/*      returns:      number of terms in literal l
/*----------------------------------------------------------------*/
 
int count_lit_characters(l)
literal *l;                    /* l is a literal in clause c      */
{
int a,*p;
int literal_terms;         /* count of the number of terms in lit */
int var_count;
	
p = &l->l_id;        
literal_terms = 0;     
var_count = 0;
	
for(a = *p++ & ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) { 
      literal_terms++;  /* function; inc count; det. # arguments   */
      a += *p++ & ARGMASK;}
   else {               
      literal_terms++;
      var_count++;
      p++;}}              /* variable; inc count                   */

if(var_count > max_vars_in_lit) {
   max_vars_in_lit = var_count;}

return(literal_terms);
}
 
/*--------------------------------------------------------------*/
/*      max_term_in_any_lit()           .............. #17
/*
/*      calculates the maximum number of terms in any literal
/*      o_maxsterm records this maximum number
/*
/*      called from:    prepare_clauses()         theo.c
/*
/*      calls:          count_lit_characters()    theo.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void max_term_in_any_lit()
{
clause *c;
literal *lit;
int clause_index, ncc;
int lit_term_cnt;

max_term_in_nc = 0;
max_term_in_lit = 0;
max_vars_in_lit = 0;

if(!nc) ncc = 1; else ncc = nc; 

for(clause_index = 0; clause_index<= base; clause_index++) {
   c = clist[clause_index];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      lit_term_cnt = count_lit_characters(lit);
      if(lit_term_cnt > max_term_in_nc && 
                      clause_index >= base - ncc + 1 &&
                      clause_index <= base) {
         max_term_in_nc = lit_term_cnt;}
      if(lit_term_cnt > max_term_in_lit) {
	     max_term_in_lit = lit_term_cnt;}}}

return;
}
			
/*--------------------------------------------------------------*/ 
/*      get_pred_fxn_symbols()        .............. #18
/*
/*      get_pred_fxn_symbols will read in a symbol table and write
/*      it to the output file (and to the log file if there is one
/*
/*      called from:   prepare_clauses()          theo.c
/*
/*      calls:         nobody
/*
/*      returns:       the number of symbols (pred/fxns)
/*---------------------------------------------------------------*/
 
int get_pred_fxn_symbols(qtbl)
char *qtbl[];
{
int a, k;
char *ii, *jj;
static char Eq[] = "Equal";
static char eq[] = "equal";
static char Eqish[] = "equalish";

for(a = 1; ;a++) {  

/*----------------------------------------*/
/* check first if at least one pred, fxn  */           
/*----------------------------------------*/

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++; 
      return(0);}

/*----------------------------------------*/
/* set qtbl[a] to point to beginning of   */
/* the a-th predicate or function         */
/*----------------------------------------*/

   qtbl[a] = ii = comp_code_buffer_index;

   while(*comp_code_buffer_index != 0xa) { 
      comp_code_buffer_index++;}
 
/*----------------------------------------*/
/* replace terminating '0xa' with a 0     */
/*----------------------------------------*/
  
   *comp_code_buffer_index++ = 0;
 
/*--------------------------------------------------*/
/* print string of characters representing pred     */    
/*--------------------------------------------------*/
    
if(qtbl == pidtbl) {
   printf("%s ",qtbl[a]);
   OutPtr += sprintf(OutPtr,"%s ",qtbl[a]);}
 
/*--------------------------------------------------*/
/* test for equality predicate                      */
/*--------------------------------------------------*/
   jj = ii;
	
   for(k = 0; k < 6 ; k++) {
      if(*jj++ != Eq[k]) {k = 0; break;}}
   if(k) { 
      equal_pred =  (a << ARGSHIFT) + 2;}

   jj = ii;
   for(k = 0; k < 6 ; k++) {
      if(*jj++ != eq[k]) {k = 0; break;}}
   if(k) { 
      equal_pred =  (a << ARGSHIFT) + 2;} 
/* 
   jj = ii;
   for(k = 0; k < 9 ; k++) {
      if(*jj++ != Eqish[k]) {k = 0; break;}}
   if(k) { 
      equal_pred =  (a << ARGSHIFT) + 2;} 
*/
   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++;
      return(a);}}                      

}                                                       

/*--------------------------------------------------------------*/
/* void print_search_options()
/*--------------------------------------------------------------*/

void print_search_options()
{
printf("OPT:");
printf(" kmzhnc %d%d%d%d%d%d",o_elim,o_merge,o_nc,o_hash,
                   o_extend,o_substitution);
printf(" xvl %d %d %d",o_maxterm,o_maxvar,o_maxlit);
printf(" t%d C%d",theorem_number,this_computer);
printf(" Ran%d",randy);
printf(" T%d",o_maxtime);
if(weakened_clause != -1) {
   print_clause(clist[weakened_clause]);}
else printf("\n");
return;  
}

/*--------------------------------------------------------------*/
/*      prepare_clauses()                .............. #19
/*
/*      prepare_clauses will read the clauses currently stored
/*      in the input_buffer and format them, placing the resulting
/*      clauses in cspace with pointers to each in the clist
/*
/*      NOTES:
/*      - we assume that all variables are numbered from 0..n so
/*        that n+1 is the number of distinct variables
/*
/*      - we pass the predicate and function symbol tables on to
/*        the output directly since we have no need for them
/*
/*      called from:    main()                          theo.c
/*
/*      calls:          get_pred_fxn_symbols()          theo.c
/*                      read_buffer_word()              theo.c
/*                      format_clause()                 theo.c
/*                      max_term_in_any_lit()           theo.c
/*                      sort_constants_fxns()           theo.c
/*                      equality_predicate()            equal.c
/*                      fatal()                         theo.c
/* 
/*      returns:        nil
/* -------------------------------------------------------------*/
 
void prepare_clauses()
{
static int characters[] =
	 {0,3,6,9,12,15,18,21,24,26,28,30,32};
int i, n, x1, x2;
int min_lit;
int return_size;
int spd, max_char_difference, char_difference;
 
/*--------------------------------------------------*/
/* set pointer to beginning of compiled_code_buffer */
/*--------------------------------------------------*/

   comp_code_buffer_index = compiled_code_buffer;

if(weakened_ply == 0) {
   compiled_code_size[1] = -compiled_code_size[0];
   compiled_code_size[3] = spawned;
   compiled_code_size[3] = -compiled_code_size[3];
   compiled_code_size[4] = o_maxlit;
   compiled_code_size[5] = o_deltalit;
   compiled_code_size[6] = o_maxsterm;
   compiled_code_size[7] = o_deltaterm;
   compiled_code_size[8] = o_nc;
   compiled_code_size[9] = o_constants;
   compiled_code_size[10] = theorem_number;
   compiled_code_size[11] = o_batchtime;
   compiled_code_size[12] = o_learn + (o_print_slave*2);
   compiled_code_size[13] = theorem_rating;}

/*--------------------------------------------------------------*/
/* We first pass the symbol table on to the output file.        */
/* This symbol table is two sets of non-empty lines followed    */
/* by an empty line (the predicate and function symbol tables   */
/*--------------------------------------------------------------*/

printf("Predicates: ");
OutPtr += sprintf(OutPtr,"Predicates: ");
number_preds = get_pred_fxn_symbols(pidtbl); /* pred sym table   */
printf("\n");
sprintf(OutPtr++,"\n");

printf("Functions:");
OutPtr += sprintf(OutPtr,"Functions:");
number_fxns = get_pred_fxn_symbols(fidtbl);/* function sym table */
read_buffer_word(); 
 
n = ww; 
base = ww - 1;               /* clist[base] is last base clause  */
org_base = base;              /* original number of base clauses */
 
read_buffer_word();nc = ww;   /* # clauses in negated conclusion */
	
if(n < 2) {
   fatal("theorem has one clause; proof impossible, exiting PROVE");}
 
/*--------------------------------------------*/
/* read the clauses and put them in the clist */
/*--------------------------------------------*/
  
while(n-- > 0) {
   format_clause();}
 
sort_constants_fxns();
 
max_term_in_any_lit();

max_lit = 0;
min_lit = 1000;
for(n = 0; n <= base; n++) {
   if(clist[n]->c_nlit > max_lit) {
      max_lit = clist[n]->c_nlit;}
   if(clist[n]->c_nlit < min_lit) {
      min_lit = clist[n]->c_nlit;}}
 
/*-----------------------------------------------*/
/* given the maximum number of terms in          */
/* any literal in any base clause,               */
/* determine o_maxsterm, the maximum number of   */
/* terms in any literal kept during the          */
/* search.  Using the default value:             */
/*                                               */
/*       max_term_in_lit    o_maxsterm           */
/*           0                  0                */
/*           1                  3                */
/*           2                  6                */
/*           3                  9                */
/*           4                 12                */
/*           5                 15                */
/*           6                 18                */
/*           7                 21                */
/*           8                 24                */
/*           9                 26                */
/*          >10        2 * max_term_in_lit + 8   */
/*                                               */
/* the user can set o_maxsterm using the xi      */
/* option.                                       */
/*-----------------------------------------------*/
 
if(x_maxterm == 0) {
   x1 = term_column[max_term_in_nc];
   x2 = term_column[max_term_in_lit];
   if(x1 > 25) {x1 = 25;}
   if(x2 > 25) {x2 = 25;}

   if(max_lit <= 3) {
      o_maxterm = tms3[x1 * 9 + x2];}
   
   if(max_lit == 4) {
      o_maxterm = tms4[x1 * 9 + x2];}
   
   if(max_lit == 5) {
      o_maxterm = tms5[x1 * 9 + x2];}
   
   if(max_lit > 5) {
      o_maxterm = tms[x1 * 9 + x2];}
   o_maxtermM1 = o_maxterm - 1;}

if(x_maxterm) {
   o_maxterm = o_maxsterm;
   o_maxtermM1 = o_maxterm - 1;
   if(o_maxsterm < max_term_in_lit) {
      o_maxsterm = 2*max_term_in_lit+8;}}
 
/*--------------------------------------------------*/
/* determine the max number of literals in any      */
/* base clause and the maximum number of            */
/* different variable names in any base literal     */
/*--------------------------------------------------*/
 
max_var = 0;
	
for(n = 0; n <= base; n++) {
   if(clist[n]->c_nv > max_var)  {
      max_var = clist[n]->c_nv;}}
if(o_maxvar == 0) {
   o_maxvar = 2 * max_var;
   if(o_maxvar < max_vars_in_lit) {
      o_maxvar = max_vars_in_lit + 3;}}
      
if(o_maxlit == 0) {
   o_maxlit = max_lit + min_lit - 2;
   if(o_maxlit < 3) {o_maxlit = 3;}}
if(max_lit < o_maxlit) {max_lit = o_maxlit;} 

/*--------------------------------------*/
/* Parallel theo code                   */
/*--------------------------------------*/

if(o_maxlit == 0) {
   if(max_lit > 4) {
      o_maxlit = ((((this_computer + 2) % 5) + 2) * max_lit)/5;}
   else {
      o_maxlit = ((((this_computer + 2) % 5) + 3) % 5) + 2;}}
      max_lit0 = max_lit;
   if(max_lit < o_maxlit) {max_lit = o_maxlit;} 

/*--------------------------------------*/
/* search for equal predicate           */
/*--------------------------------------*/

rename_functions();

if(rename_preds) {

/*----------------------------------------*/
/* if predicates have been renamed, list  */
/*----------------------------------------*/

   print0("Renamed predicates: ");
   for(i = 1; i <= number_preds; i++) {
      printf(" %s",pidtbl[i]);
      OutPtr += sprintf(OutPtr," %s",pidtbl[i]);}
   print0("\n");} 


if(rename_fxns) {

/*----------------------------------------*/
/* if functions have been renamed, list   */
/*----------------------------------------*/

   print0("Renamed functions: ");
   for(i = 1; i <= number_fxns; i++) {
      printf(" %s",fidtbl[i]);
      OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
      if(i == number_constants) {
          printf(" :");
          OutPtr += sprintf(OutPtr," :");}
      if(i == number_const_nc) {
          printf(" .");
          OutPtr += sprintf(OutPtr," .");}}
   printf("\n");
   sprintf(OutPtr++,"\n");} 

equal_predicate();

if(equal_pred && number_preds == 1) {
   if(o_maxterm < 50) {
	o_maxterm = o_maxterm * 1.5;
	o_maxtermM1 = o_maxterm - 1;
	o_maxsterm = o_maxsterm;}}

return;
}
 
/*---------------------------------------------------------------*/
/*      hash_all_base_clauses()            .............. #20
/*
/*      calculates a hash code for each literal and a hash code
/*      for each base clause.  It adds the clause to the literal
/*      hash table if there is only one literal and to the
/*      the clause hash table in any case
/*
/*      called from: main()                      theo.c
/*
/*      calls:       calc_lit_hc()               hash.c
/*                   calc_clause_hashcode()      hash.c
/*                   add_clause_to_ht()          hash.c
/*                   calc_alt_hashcodes()        hash.c
/*		     normalize_literal()         hash.c
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void hash_all_base_clauses()
{
clause *c;
literal *l;
int i;
int l_cnt, o_maxterm_tmp;
	
o_maxterm_tmp = o_maxterm;
o_maxterm = 200; 
 
for(i = 0; i <= base; i++) {
   c = clist[i];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {continue;}
      l_cnt = calc_lit_hc(l,c);}
   calc_clause_hashcode(c);
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(c->c_nlit == 1) {
      for(l = c->c_lit; l != 0; l = l->l_next) {
         if((l->l_id & ELIMLIT) == 0) {break;}}
      normalize_literal(l,c);
      add_clause_to_ht(l->l_X,l->l_Y,1);
      /*print_clause(c);*/
      if(c->c_nv) {  
         calc_alt_hashcodes(l,c);}
      if(test_simple_equality(c)) {
	 /*print_clause(c);*/}}}

o_maxterm = o_maxterm_tmp;

/*-----------------------------------*/
/* find first unit clause            */
/*-----------------------------------*/

units_begin = 0;
for(i = 0; i <= base; i++) {
   c = clist[i];
   if(c->c_nlit > 1) {continue;}
   else {
      units_begin = i;
      break;}}
return;
}       
 
					
/*----------------------------------------------------------------*/
/*      iterative_deepening()      .............. #21
/*
/*      prove will actually perform the iteratively deepening
/*      depth first search.  We return the depth at which we stop
/*
/*      called by:      main()                     theo.c
/*
/*      calls:          init_clause_hash_table()   theo.c
/*                      add_clause_to_ht()         hash.c
/*                      calc_alt_hashcodes()       hash.c
/*                      normalize_literal()        hash.c
/*                      save_line()                pass2.c
/*                      print_continuation()       tdump.c
/*                      search_tree()              search.c
/*                      time_now()                 theo.c
/*
/*      returns:        integer regarding the result of search
/*---------------------------------------------------------------*/
 
int iterative_deepening()
{
int s;              /* normally s = 0; if proof found, s = 1     */
int i; 
clause *c;
int total_nodes, total_nodesM1, total_nodesM2;

total_nodes = total_nodesM1 = total_nodesM2 = 0;

if(o_print) {
   print0("\n\nSearch dump: \n");}
 
/*---------------------------------------------------------------*/
/*                                                               */
/* Main iterative deepening loop begins here                     */
/*                                                               */
/*---------------------------------------------------------------*/
	
iter_depth = 0;
s_nod0 = s_nod;
 
/*----------------------------------------------------------------*/
/* if the number of base clauses is less than 9, turn off         */
/* merge option                                                   */
/*----------------------------------------------------------------*/

if(org_base < 9) {
   o_merge = 0;}

/*----------------------------------------------------------------*/
/* before beginning search, clear out hash tables                 */
/*----------------------------------------------------------------*/ 
 
free(clause_hash_table);
init_clause_hash_table();       
 
/*------------------------------------------*/
/* add clauses to clause_hash_table         */
/* and unit clauses to literal_hash_table   */
/*------------------------------------------*/

while(iter_depth < MAX_ITERATIONS) {                    

for(i = 0; i <= base; i++) {                         
   c = clist[i];
   if((c->c_bits & ACTIVE) == ACTIVE) { 
      add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);}

   if(c->c_nlit == 1) {
      normalize_literal(c->c_lit,c);
      add_clause_to_ht(c->c_lit->l_X,c->c_lit->l_Y,1);
      if(c->c_nv) {
         calc_alt_hashcodes(c->c_lit,c);}
      test_simple_equality(c);}} 

/*--------------------------*/
/* inc iteration depth      */    
/*--------------------------*/
	
   iter_depth++;

   if(o_print)  {
      printf("\nDepth %d\n\n",iter_depth);
      OutPtr += sprintf(OutPtr,"\nDepth %d\n\n",iter_depth);}
		
/*--------------------------*/
/* Begin search!!!!         */
/*--------------------------*/

   s = search_tree(); 

/*----------------------------------------*/
/* save the current time                  */
/*----------------------------------------*/

   search_time = time_now();
    
/*----------------------------------------------------------*/
/* extended printing routine: on each iteration print:      */
/*          1. time so far                                  */
/*          2. nodes searched on this pass                  */
/*          3. clause hash table entries                    */
/*          4. clause hash table failures                   */
/*          5. unit clause hash table entries               */
/*          6. unit clause hash table failures              */
/*----------------------------------------------------------*/
	
   printf("%d {T%d N%d-%d C%d H%d h%d U%d E%d t%d C%d} ",
      iter_depth,(search_time -search_start_time),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,
      sub_array_last,
      theorem_number,this_computer);
   OutPtr += sprintf(OutPtr,
     "%d {T%d N%d-%d C%d H%d h%d U%d E%d t%d C%d} ",
      iter_depth,(search_time - search_start_time),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,
      sub_array_last,
      theorem_number,this_computer);
 
/*------------------------------*/
/* if proof found, print *, ret */
/*------------------------------*/

   if(s != 1) {
      print0("\n");}

/*----------------------------------------*/
/* if proof found (s != 0), and in        */
/* phase 1, then save line                */
/*----------------------------------------*/
 
   if(s == 1 && phase == 1) {
      save_line(0,0);}
   if(s == 1) {
      stack_hash_codes();
      if(hash_code_active_number == 0) {
         print0("$\n");}
      else if(phase == 1) {
         print0("*\n");
	 print0("\n");
	 print0("Master knows proof exists\n");}
      else print0("\n");
      return(1);} 

/*------------------------------*/
/* too much time,...quit        */
/*------------------------------*/

   if(SlaveFoundProof == 1) return(5);

   if(s == 2)  {
      print0("\nTime has expired; no proof obtained\n");
      ply = 0;
      return(2);}
      

/*------------------------------*/
/* compare size of trees        */
/* searched, and if appropriate */
/* widen search parameters      */
/*------------------------------*/

total_nodesM2 = total_nodesM1;
total_nodesM1 = total_nodes;
total_nodes = s_nod - s_nod0;

if((iter_depth> 3 && (total_nodes - total_nodesM1) == 
         (total_nodesM1 - total_nodesM2)) ||
	 total_nodes == 0) {
   o_maxlit = 2*o_maxlit;
   if(o_maxlit > 15) o_maxlit = 15;
   o_maxterm = 2*o_maxterm;
   printf("X%d l%d ",o_maxterm,o_maxlit);
   OutPtr += sprintf(OutPtr,"X%d l%d ",o_maxterm,o_maxlit);
   if(o_maxterm > 100) o_maxterm = 100;}

/*------------------------------*/
/* giving up search,....quit    */
/*------------------------------*/
 
   if(s == 3)  {
      print0("\n\nGiving up the search; no proof obtained\n");
      ply = 0;
      return(2);}
   
   if(s_nod == 0 && iter_depth > 10) {
      print0("\n\nNo proof obtained; check your clauses!\n");
      ply = 0;
      return(2);}}
	       
/*----------------------*/              
/* return with no proof */
/*----------------------*/
 
return(4);
}

/*-------------------------------------------------------------*/
/*      get_clause_stats()         .............. #22
/*
/*      determine the number of chars, vars, and lits
/*      in clause c
/*
/*      called by:     write_out_results    theo.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*
/*--------------------------------------------------------------*/

void get_clause_stats(c)
clause *c;
{
literal *l;
int lit_length;

if(c->c_nlit > maxl) maxl = c->c_nlit;
if(c->c_nv   > maxv) maxv = c->c_nv;
for(l = c->c_lit; l != 0; l = l->l_next) {
   if(l->l_next) {
      lit_length = ((int) l->l_next - (int) l)/4 - 4;}
   else {
      lit_length = ((int) c->c_end - (int) l)/4 - 4;}
   if(lit_length > maxx) {
      maxx = lit_length;}}

}

/*------------------------------------------------------------*/
/*      get_good_clauses()
/*
/*------------------------------------------------------------*/

void get_good_clauses()
{
clause *c;
int i;

return;
if(o_learn == 0) return;

for(i = 0;i <= org_base + 1; i++) {
   if(good_clauses[i] > 0) { 
      clist[i]->c_bits = clist[i]->c_bits | NC;}} 

}



/*------------------------------------------------------------*/
/*      write_out_results()         .............. #23
/*
/*      write_out_results writes out the cl's that form the proof
/*      (if a proof found) and write statistics out. it writes
/*      the results on the screen and to a file "in-theorem.u".
/*
/*      called by:      main()              theo.c
/*
/*      calls:          print_clause()      tdump.c
/*                      get_clause_stats()  theo.c
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void write_out_results(proof_found)
int proof_found;
{
clause *c, *ci, *c_next;
int i,j,k,ii;         
int total_search_time;     /* total search time                 */
int good_inferred_cls;     /* # inferred clauses used in proof  */
int yes_inferred_cls;
int t_temp, t1_temp, t_power;
int x_temp;
char t1_char;
int leading_zeros;
int cls_id;

/*--------------------------------------------------*/
/* Parallel theo code                               */
/*--------------------------------------------------*/

if(SlaveFoundProof == 1) proof_found = 1;

/*--------------------------------------------------*/
/* set screen printing to print to a depth of 1000. */
/*--------------------------------------------------*/

if(((proof_found == 1 || simple_sol == 1) && weakened_clause == -1)
            || excellent_weakened_proof == 1) {
   batch_tried++; batch_solved++;
   batch_solved_array[batch_tried - 1] = 1;
   if(weakened_clause == -1) { 
      printf("THE PROOF\n\n",theorem_number);
      OutPtr += sprintf(OutPtr,"THE PROOF\n\n",
	    theorem_number);}
   else { 
      printf("Proof obtained to weakened theorem %d!\n",theorem_number);
      OutPtr += sprintf(OutPtr,
	      "Proof obtained to weakened theorem %d!\n",theorem_number);}
   if(o_batch == 1) {
      printf(" (%d Solved, %d Tried, %d Satisfiable) ",
	 batch_solved,batch_tried,batch_satisfiable);
      for(ii = 0; ii < batch_tried; ii++) {
         printf("%d",batch_solved_array[ii]);
	 if(ii % 10 == 9) printf(" ");}
      OutPtr += sprintf(OutPtr," (%d Solved, %d Tried, %d Satisfiable) ",
	 batch_solved,batch_tried,batch_satisfiable);}
   printf("\n");}
else {
   if(proof_found != 1 && weakened_clause == -1) {
      batch_tried++;
      printf("No Proof Obtained.\n");
      OutPtr += sprintf(OutPtr,"No Proof Obtained\n");
      if(o_batch == 1) {
         printf(" (%d Solved, %d Tried, %d Satisfiable) ",
	    batch_solved,batch_tried,batch_satisfiable);
         for(ii = 0; ii < batch_tried; ii++) {
            printf("%d",batch_solved_array[ii]);
	    if(ii % 10 == 9) printf(" ");}

        OutPtr+=sprintf(OutPtr," (%d Solved, %d Tried, %d Satisfiable) ",
	    batch_solved,batch_tried,batch_satisfiable);}
      printf("\n");}}

o_print = 1000; o_step = -1; o_debug = 0; 
maxx = maxl = maxv = 0;
	
/*----------------------------------------------------------------*/
/* determine which of the inferred clauses were used in the proof.*/  
/* fix c->c_bits by adding LIST bit for inferred clauses that     */ 
/* should be printed. for each clause in the proof found in phases*/
/* 1,2, check LIST bit of parents                                 */
/*----------------------------------------------------------------*/

for(i = 1; i <= ply; i++) {
   c = clist[depth[i].d_tc];
   if(c->c_lpar) {
      ci = c->c_lpar;
      ci->c_bits |= LIST;} /* check lft par of clist[depth[i].d_tc*/
   if(c->c_rpar)  {                /* check if c is a factor      */
      ci = c->c_rpar;
      ci->c_bits |= LIST;}}/* check rt par of clist[depth[i].d_tc */
			
/*-------------------------------------------------------------*/
/* now check LIST bit of other inferred cl's to det parents    */
/*-------------------------------------------------------------*/

yes_inferred_cls = 0; 
for(i = base ; i > org_base ; i--)  {
   c = clist[i];
   if(c->c_nlit == 0) c->c_bits |= LIST; /* null cls is inc */
   if((c->c_bits & LIST) == LIST) {
      yes_inferred_cls++;
      if(c->c_lpar != 0) ci = c->c_lpar;
      ci->c_bits |= LIST;
      if(c->c_rpar != 0) ci = c->c_rpar;
      ci->c_bits |= LIST;}}
			 
good_inferred_cls = 0;        /* count # of inferred cls used   */

/*--------------------------------------------------------------*/
/* if there are axioms, print "axioms". otherwise print         */
/* "negated conclusion".                                        */
/*--------------------------------------------------------------*/
 
if(org_base + 1 - nc) {
   print0("\nAxioms:\n");}
else {
   print0("\nNegated conclusion:\n");}

/*--------------------------------------------------------------*/
/* write out the base clauses and the clauses used from phase 0 */
/*--------------------------------------------------------------*/

tab_amount = 0;

for(i = 0; i <= base; i++) {
   c = clist[i];
   if(c->c_id <= org_base + 1) {
      if((c->c_bits & LIST) == LIST) {
         good_clauses[i]++;}}
   if(i <= org_base) {
      if(o_relevent == 0) {
         if((c->c_bits & LIST) == LIST) {
            get_clause_stats(c);}
         print_clause(c);}
      else if((c->c_bits & LIST) == LIST) {
         get_clause_stats(c);
         print_clause(c);}}
   else {
      if(o_debug > 0) {print_clause(c);}
      if(((c->c_bits & LIST) == LIST) || proof_found > 1) {
         get_clause_stats(c);
         print_clause(c); 
	 if(questionable[c->c_id] == 0 && c->c_nlit == 1) {
	    copy_clause3(c);}
	 good_inferred_cls++;}}         

   if(i == org_base) print0("\n---------------\n");
   if((i == org_base) && (org_base != base)) {
      if(yes_inferred_cls) {
	 print0("\nPhase 0 clauses used in proof:\n");}}
   if(i == (org_base - nc))  { 
      print0("\nNegated conclusion:\n");}} 

/*----------------------------------*/
/* write out the clauses from       */
/* phases 1 and 2 used in proof     */
/* if proof not found in phase 0    */
/*----------------------------------*/
	
if(clist[base]->c_nlit != 0 && ply != 0) {
   print0("\nPhases 1 and 2 clauses used in proof:\n");
   for(i = 1; i <= ply; i++)  {
      c = clist[depth[i].d_tc];
      if(c->c_lpar->c_id != c->c_id - 1 && i != 1) {
         print0("\n");}
      get_clause_stats(c);
      if(questionable[c->c_id] == 0 && c->c_nlit == 1) {
	 if(depth[i+1].d_tc) c_next = clist[depth[i+1].d_tc];
	 if(c_next->c_lpar->c_id != c->c_id) {
	    copy_clause3(c);}}
      if(i > 1) {
	 cls_id = c->c_id;
	 if((questionable[cls_id] == 0)
	    && questionable[cls_id + 1] == 1) {
	    if(cl[i-1][1] == 0) {
	       copy_clause3(clist[depth[i-1].d_tc]);}
	    else {
	       copy_clause3(c);}}}
      print_clause(c);}}

if(((proof_found == 1 || simple_sol == 1) && weakened_clause == -1)
            || excellent_weakened_proof == 1) {
   if(weakened_clause == -1) { 
      printf("THE END\n\n",theorem_number);
      OutPtr += sprintf(OutPtr,"THE END\n\n",
	    theorem_number);}}

statistics:
/*
if(weakened_clause != -1) {
   reconsider_proof(proof_found);
   if(nextu > 0) {
      printf("Inferences derived from modified clause ");
      printf("when variable is replaced by constant\n");}

   for(i = 0; i < nextu; i++) {
      ulist[i]->c_bits |= BASE;
      print_clause(ulist[i]);}}
*/

total_search_time = search_time - search_start_time; 

if(max_lit > 16) {
   theorem_out_name2[theorem_out_name2end++] = 'z';
   goto rename_theorem;}

if(satisfiable == 1) {
   theorem_out_name2[theorem_out_name2end++] = 's';
   goto rename_theorem;}

if(outside_range == 1) {
   theorem_out_name2[theorem_out_name2end++] = 'Z';}

if(weakened_clause == -1 || excellent_weakened_proof == 1 || c->c_nlit == 0) {
   if(tough == 1 && unknown == 0) {
      theorem_out_name2[theorem_out_name2end++] = 'x';}

   if(tough == 1 && unknown == 1) {
      theorem_out_name2[theorem_out_name2end++] = 'X';}

   if(tough != 1) {
      theorem_out_name2[theorem_out_name2end++] =  
			       48 + theorem_rating/10;}

   if(proof_found == 1 && o_learnM != 0 && !outside_range) {
      theorem_out_name2[theorem_out_name2end++] = 'L';
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned % 50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[this_computer/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[this_computer % 50];}

   if(proof_found != 1 && o_learnM != 0 && !outside_range) {
      theorem_out_name2[theorem_out_name2end++] = 'k';
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned % 50];}

   if(proof_found == 1 && o_learnM == 0 && !outside_range) {
      theorem_out_name2[theorem_out_name2end++] = 'T';
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned % 50];
      theorem_out_name2[theorem_out_name2end++] = '0';
      theorem_out_name2[theorem_out_name2end++] = '0';}

   if(proof_found != 1 && o_learnM == 0 && !outside_range) {
      theorem_out_name2[theorem_out_name2end++] = 't';
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned % 50];}

   if(excellent_weakened_proof == 1) {
      batch_tried++; batch_solved++;
      printf("Proof obtained to weakened theorem %d!\n",theorem_number);
      OutPtr += sprintf(OutPtr,
	      "Proof obtained to weakened theorem %d!\n",theorem_number);
      if(o_batch == 1) {
         printf("(%d Solved, %d Tried, %d Satisfiable)\n",
	    batch_solved,batch_tried,batch_satisfiable);
OutPtr+=sprintf(OutPtr,"(%d Solved, %d Tried, %d Satisfiable)\n",
	 batch_solved,batch_tried,batch_satisfiable);}
      theorem_out_name2[theorem_out_name2end++] = 'W';}

/*-----------------------------------------------------*/
/* convert time (in t_temp) to string which is then    */
/* suffixed to the file name.                          */
/*-----------------------------------------------------*/

   theorem_out_name2[theorem_out_name2end++] = '.';

   if((weakened_clause == -1 && !outside_range) ||
       excellent_weakened_proof == 1) {
      leading_zeros = 0;
      if(simple_sol == 1) {phase1_time = 0;}
      t_temp = eliminate_time + phase1_time + weakened_time;
      for(t_power = 125000; ; t_power = t_power/50) {
         if(t_temp >= t_power) {
            t1_temp = t_temp/t_power;
            t1_char = alphabet32[t1_temp];
            theorem_out_name2[theorem_out_name2end++] =  t1_char; 
            leading_zeros = 1;
            t_temp = t_temp - t1_temp*t_power;}
         else {
            if(leading_zeros == 1 || t_power == 1)  {
               t1_char = 48; 
               theorem_out_name2[theorem_out_name2end++] = t1_char;
               leading_zeros = 1;}}
         if(t_power == 1) {
	    break;}}

      if(proof_found == 1) {
         theorem_out_name2[theorem_out_name2end++] = '.';
         leading_zeros = 0;
         t_temp = good_inferred_cls + ply;
         for(t_power = 125000; ; t_power = t_power/50) {
            if(t_temp >= t_power) {
               t1_temp = t_temp/t_power;
               t1_char = alphabet32[t1_temp];
               theorem_out_name2[theorem_out_name2end++] =  t1_char; 
               leading_zeros = 1;
               t_temp = t_temp - t1_temp*t_power;}
            else {
               if(leading_zeros == 1 || t_power == 1)  {
                  t1_char = 48; 
                  theorem_out_name2[theorem_out_name2end++] = t1_char;
                  leading_zeros = 1;}}
            if(t_power == 1) {
	       break;}}}

      theorem_out_name2[theorem_out_name2end++] = '.';
      leading_zeros = 0;
      t_temp =  s_nod;
      for(t_power = 312500000; ; t_power = t_power/50) {
         if(t_temp >= t_power) {
            t1_temp = t_temp/t_power;
            t1_char = alphabet32[t1_temp]; 
            theorem_out_name2[theorem_out_name2end++] =  t1_char; 
            leading_zeros = 1;
            t_temp = t_temp - t1_temp*t_power;}
         else {
            if(leading_zeros == 1 || t_power == 1)  {
               t1_char = 48; 
               theorem_out_name2[theorem_out_name2end++] = t1_char;
               leading_zeros = 1;}}
         if(t_power == 1) {
             break;}}}

if(c->c_nlit && weakened_clause == -1) {
   print0("\nNo Proof Obtained !!\n");}


/*------------------------------------------------------*/
/* if proof found, rename theorem to "theorem.Utime     */
/*------------------------------------------------------*/

if(weakened_clause == -1 || outside_range ||
            c->c_nlit == 0 || excellent_weakened_proof == 1) {
rename_theorem:
   rename(theorem_out_ptr,theorem_out_ptr2);
   weakened_ply = 0;}}
	
/*------------------------------------------------------------------*/
/* first print info on time for elim, phase 1, phase 2, sum of 1,2  */
/*------------------------------------------------------------------*/

if(weakened_clause != -1) {   
   if(nextu > 0) {
      printf("Valid inferences derived from given clauses:\n");
      OutPtr += sprintf(OutPtr,
	     "Valid inferences derived from given clauses:\n");
      for(i = 0; i < nextu; i++) {
         ulist[i]->c_bits |= BASE;
        /* print_clause(ulist[i]);*/}}} 

printf("\nPHASE 0: %d s",eliminate_time);
if(simple_sol == 0) printf(" PHASE 1: %d s",phase1_time);
else printf(" PHASE 1: -NA- ");
if(extra_pass == 1)  {
   printf(" PHASE 2: %d s ",total_search_time - phase1_time);}
else printf(" PHASE 2: -NA- ");
if(simple_sol == 0) {
   printf("  Total Time: %d s\n",
             eliminate_time + total_search_time + weakened_time);
   weakened_time = eliminate_time + total_search_time;}
else printf("\n");
 
/*----------------------------------*/
/* print statistics on the screen   */
/*----------------------------------*/    
	
printf("NOD: %-10d\t",s_nod);               /* nodes in tree    */
printf("RES: %-10d\t",s_res);            /* binary resolv'ts    */
printf("FAC: %-10d\t",s_fac);            /* # factors           */
printf("T: %d  ",o_maxterm);              /* max terms in lit    */
printf("V: %d  ",o_maxvar);               /* max vars in clause  */
printf("L: %d\n",o_maxlit);               /* max lits in clause  */

printf("CTE: %-10d\t",s_htentry_phase1); /* clause tble entries */
printf("CTH: %-10d\t",s_hthit_phase1);   /* hash hits           */
printf("CTF: %-10d\t",s_htfail_phase1);  /* hash failures       */
printf("CSZ: %-10d\n",o_arraysize);              /* size of ht  */
 
printf("UTE: %-10d\t",s_ltentry_phase1); /* # unit cl ht phase1 */
printf("UTH: %-10d\t",s_lthit_phase1);   /* unit cls hits       */
printf("UTF: %-10d\t",s_ltfail_phase1); /* unit clause failures */
printf("SBA: %-10d\n",sub_array_last1);

printf("BAS: %-10d\tRED: %-10d\t",org_base + 1,clauses_alive);
 
/*------------------------------------------------------------------*/
/* print proof length: number of inferred clauses that are used     */
/* plus the number of clauses found in the search (phases 1,2)      */
/*------------------------------------------------------------------*/
	
if(proof_found == 1) {
   length_of_proof = good_inferred_cls + ply;
   printf("LEN: %d+%d=%d\t",
           good_inferred_cls,ply,length_of_proof);}
	
/*----------------------------------------------------------*/
/* print options: (1) elim strategy, (2) merge strategy     */
/* (3) linear-nc strat. (4) clause hash table               */
/* (5) literal hash table, and (6) extended search,         */
/* (7) s-subsumption strategy.                              */
/* options are only printed if they are not default values. */
/*----------------------------------------------------------*/
	
printf("OPT:");
printf(" k%d",o_elim);
printf(" m%d",o_merge);
printf(" z%d",o_nc);
printf(" h%d",o_hash);
printf(" n%d",o_extend);
printf(" c%d",o_substitution);
printf(" b%d",o_subsume);
printf(" t%d",o_maxtime);
printf("\n");   

/*--------------------------------------------------*/        
/* output to theorem.p (same as output to screen)   */
/*--------------------------------------------------*/
	
OutPtr += sprintf(OutPtr,"\nPHASE 0: %d s",eliminate_time);

if(simple_sol == 0) {
   OutPtr += sprintf(OutPtr," PHASE 1: %d s ",phase1_time);}
else OutPtr += sprintf(OutPtr," PHASE 1: -NA-  ");
	
if(extra_pass == 1) { 
   OutPtr += sprintf(OutPtr,
     " PHASE 2: %d s ",total_search_time - phase1_time);}
else {
   OutPtr += sprintf(OutPtr," PHASE 2: -NA-  ");}
if(simple_sol == 0) {
   OutPtr += sprintf(OutPtr,
   " Total Time: %d s\n",eliminate_time + total_search_time);}
else OutPtr += sprintf(OutPtr,"\n");

OutPtr += sprintf(OutPtr,"NOD: %-10d\t",s_nod);  
OutPtr += sprintf(OutPtr,"RES: %-10d\t",s_res);
OutPtr += sprintf(OutPtr,"FAC: %-10d\t",s_fac);       
OutPtr += sprintf(OutPtr,"T: %d  ",o_maxterm);
OutPtr += sprintf(OutPtr,"V: %d  ",o_maxvar);
OutPtr += sprintf(OutPtr,"L: %d\n",o_maxlit);        

OutPtr += sprintf(OutPtr,"CTE: %-10d\t",s_htentry_phase1);  
OutPtr += sprintf(OutPtr,"CTH: %-10d\t",s_hthit_phase1);      
OutPtr += sprintf(OutPtr,"CTF: %-10d\t",s_htfail_phase1);      
OutPtr += sprintf(OutPtr,"CSZ: %-10d\n",o_arraysize);     
 
OutPtr += sprintf(OutPtr,"UTE: %-10d\t",s_ltentry_phase1);      
OutPtr += sprintf(OutPtr,"UTH: %-10d\t",s_lthit_phase1);     
OutPtr += sprintf(OutPtr,"UTF: %-10d\t",s_ltfail_phase1);      
OutPtr += sprintf(OutPtr,"SBA: %-10d\n",sub_array_last1);

OutPtr += sprintf(OutPtr,
   "BAS: %-10d\tRED: %-10d\t",org_base + 1,clauses_alive);

if(proof_found == 1) {
   OutPtr += sprintf(OutPtr,
      "LEN: %d+%d=%d\t",
           good_inferred_cls,ply,length_of_proof);}
	
OutPtr += sprintf(OutPtr,"OPT:");
OutPtr += sprintf(OutPtr," k%d",o_elim);
OutPtr += sprintf(OutPtr," m%d",o_merge);
OutPtr += sprintf(OutPtr," z%d",o_nc);
OutPtr += sprintf(OutPtr," h%d",o_hash);
OutPtr += sprintf(OutPtr," n%d",o_extend);
OutPtr += sprintf(OutPtr," c%d",o_substitution);
OutPtr += sprintf(OutPtr," b%d",o_subsume);
OutPtr += sprintf(OutPtr," t%d",o_maxtime);
OutPtr += sprintf(OutPtr,"\n");

if(weakened_clause == -1 || excellent_weakened_proof == 1 
            || satisfiable == 1) {
   if(proof_found == 1) {
      OutPtr2 = &results[0];
      fopen(theorem_out_name2,"w");
      while(OutPtr2 <= OutPtr) {
         fprintf(outfile,"%c",*OutPtr2++);}}}


else if(fix_proof_once == 0) weakened_ply = 1;

printf("\n");
 
}
 
/*-----------------------------------------------------------*/
/* fixproof()
/*-----------------------------------------------------------*/

int fix_proof()
{
int q;
clause *c;
int final_base;
int i, j, k;
char *char_temp;
int proof_from, selected_pbm;

pvm_upkbyte(results_slave,300000,1);
if((int) results_slave[3] != theorem_number) {
   printf("Early_return %d \n",(int) results_slave[3]);
   return(0);}

proof_from = (results_slave[1] - '0') * 10;
proof_from =  proof_from + (results_slave[2] - '0');

if(split_pbs != 0 && proof_from < spawned) {
   goto fix_split_pbs;}

/*------------------------------------*/
/* slave knows that proof exists      */
/*------------------------------------*/

if(results_slave[0] == 'a' && results_slave[4] == 0) {
   if(add_time != 0) {return(0);}
      add_time++;
      o_maxtime = 360000;
      printf(
      "\nSlave %d knows proof exists after %d seconds! Extend allotted time\n",
       proof_from,time_now() - elim_starting_time);

   slaves_killed = 0;
   for(i = 0; i <spawned; i++) { 
      if(i + 1 == proof_from) {
	 continue;}
      if(spawned_alive[i] == 0) {
         if((k = pvm_kill(tid[i])) != 0) {
	    printf("\nProcess %d not killed -- Error(%d)\n",i,k);}
         else {
	    spawned_alive[i] = 1;
	    slaves_killed++;}}}
   return(0);}

/*------------------------------------*/
/* save only one proof                */
/*------------------------------------*/

if(fix_proof_once != 0)  {return(0);}
fix_proof_once = 1;

phase = 3;

printf("Proof Found: slave %d found proof after %d seconds!\n",
          proof_from,time_now() - elim_starting_time);
OutPtr += sprintf(OutPtr,
       "Proof Found: slave %d found proof after %d seconds!\n",
          proof_from,time_now() - elim_starting_time);

SlaveFoundProof = 1;

/*--------------------------------------*/
/* fix file name for theorem's results  */
/*--------------------------------------*/

OutPtr2 = &results_slave[6];

if(tough == 1 && unknown == 0) {
   theorem_out_name2[theorem_out_name2end++] = 'x';}

if(tough == 1 && unknown == 1) {
   theorem_out_name2[theorem_out_name2end++] = 'X';}

if(tough != 1) {
   theorem_out_name2[theorem_out_name2end++] =  
			       48 + theorem_rating/10;}
for(i = 0; i < 24; i++) {
   theorem_out_name2[theorem_out_name2end++] = *OutPtr2++;}

/*---------------------------------------*/
/* transfer proof from results_slave to  */
/* output file.                          */
/*---------------------------------------*/

OutPtr2 = &results_slave[30];
*OutPtr2 = '\n';
OutPtr2 = &results_slave[5];
i= 0;
while(i < 300000) {
   fprintf(outfile,"%c",*OutPtr2);
   char_temp = OutPtr2;  
   i++;
   OutPtr2++;
   if((*char_temp == '$') && (*OutPtr2 == '$')) {break;}}

/*---------------------------------------*/
/* update counts on solved problems      */
/*---------------------------------------*/

batch_tried++; batch_solved++;
if(o_batch == 1) {
   printf(" (%d Solved, %d Tried, %d Satisfiable)",
	 batch_solved,batch_tried,batch_satisfiable);
   OutPtr += sprintf(OutPtr," (%d Solved, %d Tried, %d Satisfiable)",
	 batch_solved,batch_tried,batch_satisfiable);}
printf("\n");

theorem_out_ptr2 = &theorem_out_name2[0];
rename(theorem_out_ptr,theorem_out_ptr2);
weakened_ply = 0;
weakened_clause = -1;
free_space_close_files();
return(1);


fix_split_pbs:
selected_pbm = ((proof_from - 1) % split_pbs) + 1;
//printf("Selected split problem = %d\n",selected_pbm);

if(results_slave[0] == 'a' && results_slave[4] == 0) {
   if(add_time == 0 && unsolved_split_pbs == 1 &&
           solved_split_lits[selected_pbm] == 1) {
      printf( 
      "Split problem %d info from computer %d not saved as not yet",
             selected_pbm, proof_from);
      printf(" a proof. Extend allotted time.\n");
      printf("Master knows split proof exists\n");
      if(add_time != 0 && unsolved_split_pbs == 1) {return(0);}
      add_time++;
      o_maxtime = 360000;
      return(0);}
   else {
      printf(
      "Split problem %d info from computer %d not saved as not yet",
      selected_pbm, proof_from);
      printf(" a proof.\n");
      return(0);}}


if(solved_split_lits[selected_pbm] != 1) {
  printf("Split problem %d sent by computer %d solved already\n",
          selected_pbm,proof_from);
  return(0);}

solved_split_lits[selected_pbm] = -proof_from;
unsolved_split_pbs--;

if(unsolved_split_pbs == 0) {
   printf(
 "Split problem %d solved by computer %d; no unsolved split problems!\n",
           selected_pbm,proof_from);
   printf("Split Proof Found\n");}

else {
   printf("Split problem %d solved by computer %d; unsolved split problems: %d/%d -- ",
           selected_pbm,proof_from,unsolved_split_pbs,split_pbs);

   for(i = 1; i <= split_pbs; i++) {
      if(solved_split_lits[i]  < 0) printf("%d ",solved_split_lits[i]);
      else printf("N ");}
   printf("\n");}

/*-----------------------------------*/
/* This is a solved problem that     */
/* must be saved.  Transfer proof    */
/* from results_slave to output file */
/*-----------------------------------*/

OutPtr2 = &results_slave[30];
*OutPtr2 = '\n';
OutPtr2 = &results_slave[5];
i= 0;
while(i < 300000) {
   fprintf(outfile,"%c",*OutPtr2);
   char_temp = OutPtr2;  
   i++;
   OutPtr2++;
   if((*char_temp == '$') && (*OutPtr2 == '$')) {break;}}

/*
printf("Slave %d found split proof after %d seconds!\n",
          proof_from,time_now() - elim_starting_time);
OutPtr += sprintf(OutPtr,
       "Slave %d found split proof after %d seconds!\n",
          proof_from,time_now() - elim_starting_time);
*/

/*--------------------------------------*/
/* fix file name for theorem's results  */
/*--------------------------------------*/

if(unsolved_split_pbs != 0) return(0);

OutPtr2 = &results_slave[6];

if(tough == 1 && unknown == 0) {
   theorem_out_name2[theorem_out_name2end++] = 'x';}

if(tough == 1 && unknown == 1) {
   theorem_out_name2[theorem_out_name2end++] = 'X';}

if(tough != 1) {
   theorem_out_name2[theorem_out_name2end++] =  
			       48 + theorem_rating/10;}

theorem_out_name2[theorem_out_name2end++] = 'S';

for(i = 0; i < 24; i++) {
   theorem_out_name2[theorem_out_name2end++] = *OutPtr2++;}

SlaveFoundProof = 1;
phase = 3;

batch_tried++; batch_solved++;
if(o_batch == 1) {
   printf(" (%d Solved, %d Tried, %d Satisfiable)",
	 batch_solved,batch_tried,batch_satisfiable);
   OutPtr += sprintf(OutPtr," (%d Solved, %d Tried, %d Satisfiable)",
	 batch_solved,batch_tried,batch_satisfiable);}
printf("\n");

theorem_out_ptr2 = &theorem_out_name2[0];
rename(theorem_out_ptr,theorem_out_ptr2);
weakened_ply = 0;
weakened_clause = -1;
free_space_close_files();
return(5);

}

/*-----------------------------------------------------------------*/
/*      main()                             .............. #24
/*
/*      The theorem prover enters here. We set the options specified
/*      by the  user, read in the clause system, prove the theorem,
/*      and write the proof out.
/*
/*      calls:   fatal()                        theo.c
/*               init_options()                 theo.c
/*               compile_clauses()              affc.c
/*               init_cl_array()                pass2.c
/*               init_clause_space()            theo.c
/*               init_stats()                   theo.c
/*               prepare_clauses()              theo.c
/*               init_random_hash_array()       theo.c
/*               hash_all_base_clauses()        theo.c
/*          (PHASE 0)
/*               eliminate()                    elim.c
/*          (PHASE 1)
/*               init_stats()                   theo.c
/*               iterative_deepening()          theo.c
/*               stack_hash_codes()             pass2.c
/*          (PHASE 2)
/*               iterative_deepening()          theo.c
/*               stack_hash_codes()             pass2.c
/*               rearrange_cl_array()           pass2.c
/*               search_proof_line()            pass2.c
/*          (PHASE 3)
/*               write_out_results()            theo.c
/*               free_space_close_files()       theo.c
/*-----------------------------------------------------------------*/
 
int main(argc,argv)
int argc;
char *argv[];

{
int zz;
int proof_found;
char decision_char;   /* char to decide whether to do another thm  */
int i,j, jj, ck, kk;
int time_delay;
int o_maxtermZ, o_maxvarZ, o_maxlitZ;

o_batch = 0;
o_learn = 0;
o_print_slave = 1;
this_computer = 0;

batch_start_time = time_now();
batch_tried = 0;
batch_solved = 0;
for(i = 0; i <2000; i++) batch_solved_array[i] = 0;
batch_satisfiable = 0;

weakened_ply = 0;
weakened_clause = -1;
fix_proof_once = 0;
repeat_weakening = 0;

o_diff_top = 101;
o_diff_bottom = 0;

to_spawn = 150;

if(remove("/u0/prof/newborn/octoslave.core") == 0) {
   printf("Killed /u0/prof/newborn/octoslave.core\n");}

if(remove("/u0/prof/newborn/herbyslave.core") == 0) {
   printf("Killed /u0/prof/newborn/herbyslave.core\n");}

setlinebuf(stdout);
pvm_catchout(stdout);

theorem_number = 0;
myparent = 0;

my_tid = pvm_mytid();

/*-----------------------------------------*/
/* change_array used in modify.c to keep   */
/* track of changes made to the given thm  */
/*-----------------------------------------*/

for(i = 0; i < 50; i++) {
   for(j = 0; j < 10; j++) {
      change_array[i][j] = 0;}}

/*-----------------------------------------*/
/* potential_change_array is used to store */
/* potential vars and constants  that may  */
/* be changed for a weakened version of thm*/
/*-----------------------------------------*/

for(i = 0; i < 2000; i++) {
   for(j = 0; j < 7; j++) {
      p_change_array[i][j] = 0;}}
excellent_weakened_proof = 0;
p_change_number = 0;
	
begin_proof:
add_time = 0;
if(weakened_ply == 0 || excellent_weakened_proof == 1) {
   outside_range = 0;
   excellent_weakened_proof = 0;


   printf("Begin Proof by killing slaves.\n");
   info = pvm_tasks(0,&ntask,&taskp );
   for(i = 2; i < ntask; i++) {
      pvm_kill(taskp[i].ti_tid);}
   nextu = 0;

   if(remove("/u0/prof/newborn/octoslave.core") == 0) {
      printf("Killed /u0/prof/newborn/octoslave.core\n");}
   if(remove("./octoslave.core") == 0) {
      printf("Killed ./octoslave.core\n");}
   if(remove("~/octoslave.core") == 0) {
      printf("Killed ./octoslave.core\n");}

   if(remove("/u0/prof/newborn/herbyslave.core") == 0) {
      printf("Killed /u0/prof/newborn/herbyslave.core\n");}
   if(remove("./herbyslave.core") == 0) {
      printf("Killed ./herbyslave.core\n");}
   if(remove("~/herbyslave.core") == 0) {
      printf("Killed ./herbyslave.core\n");}

   printf("Done killing slaves.\n");
   weakened_clause = -1;
   weakened_time = 0;
   SolvedBy = 0;
   fix_proof_once = 0;
   for(i = 0; i < MAX_BASE_CLAUSES; i++) {
      good_clauses[i] = 0;}

   for(jj = 0; jj <= to_spawn; jj++) {tid[jj] = 0;}
   spawned = pvm_spawn("octoslave",
         NULL, PvmTaskHost+PvmHostCompl,".",to_spawn, tid);
   spawned += pvm_spawn("herbyslave",
         NULL, PvmTaskHost+PvmHostCompl,".",1, &tid[spawned]);
   printf("\nSlaves spawned = %d\n",spawned);
   
   for(jj = 0; jj < spawned; jj++) {spawned_alive[jj] = 0;}

   theorem_number++;
   change_level = 1;

   for(ck = 0; ck < spawned; ck++) {
      cc = pvm_recv(-1,CNAME);
      pvm_upkstr(buf);
      printf("%d:%x:%s  ",ck + 1,tid[ck],buf);
      if(ck % 3 == 2) {
         printf("\n");}}

   if(spawned != to_spawn + 1) {
// if(spawned != to_spawn) {
      printf("\nCan't start all %d slaves\n",to_spawn + 1);}}
      
else {printf("Reconsider original theorem ");}

proof_found  = 0;
killer_hashx = 0;
killer_hashy = 0;
magic_hashx  = 0;
magic_hashy  = 0;
extra_pass   = 0;
phase        = 0;
iter_depth   = 0;
equal_pred   = 0;
equal_clist_index = -2;
lit_in_hash = 0;
alt_HX = 0;
satisfiable = 0;
unknown = 0;
tough = 0;
theorem_rating = 0;

SlaveFoundProof = 0;

/*------------------------------------------------------*/
/* attempt to determine theorem's name                  */
/*------------------------------------------------------*/
 
enter_theorem_name(argc,argv);

/*--------------------------------------------------*/
/* if input file is in tptp format, call            */
/* CompileTptp()                                    */
/*--------------------------------------------------*/

if(o_tptp != 0) {CompileTptp(o_tptp);}

/*--------------------------------------------------*/
/* allocate space for data structures in affc.c     */
/*                                                  */
/* read in clauses and format for theorem prover    */
/* This is done in the function compile_clauses()   */
/* which is found in affc.c.                        */
/*                                                  */
/* free space used by compile; close input file     */
/*                                                  */
/*--------------------------------------------------*/

compile_clauses();

/*--------------------------------------------------*/
/* init arrays for phase 0, then fix up clauses     */
/*--------------------------------------------------*/

for(i = 0; i < 65536; i++) {
   for(j = 0; j < 6; j++) {sub_array[i][j] = 0;}}

for(i = 0; i < 2000; i++) {questionable[i] = 0;}
init_cl_array();          /* init cl list in pass2.c         */
init_search_space();      /* init search space               */
init_stats();             /* zero our statistics variables   */
prepare_clauses();

if(max_lit > 16) {
   proof_found = 0;
   goto write_results;}

print_search_options();
modify_clause();

/*---------------------------------------------------*/
/* initialize the random_hash_array and then hash    */
/* all base clauses                                  */
/*---------------------------------------------------*/

init_random_hash_array();
hash_all_base_clauses();

if(satisfiable == 1) {
   if(o_batch == 1) batch_satisfiable++;
   proof_found = 0;
   goto write_results;}

if(theorem_rating > o_diff_top || theorem_rating < o_diff_bottom) {
   if(o_batch == 1) {
      outside_range = 1;}
   proof_found = 0;
   goto write_results;}

s_starttime = time_now();

get_good_clauses();

/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simplify input clauses if possible.                */
/* the array elim_inferences stores clauses generated */
/* in elim.                                           */
/*----------------------------------------------------*/

in_elim = 0;
if(o_elim) {
   simple_sol = 0;
   eliminate();}
in_elim = 1;
		
/*----------------------------------------*/
/* if proof found, don't carry out search */
/*----------------------------------------*/
 
if(o_elim == 2) simple_sol = 2;
	
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/

phase = 1;      
ply = 0;
lit_in_hash = 0;

search_start_time = time_now(); /* start time for search  */ 
 
init_stats();           /* zero our statistics variables  */
	
if(simple_sol == 0) {
   prepare_subarray();
   /*print_substitutions();*/
   o_maxlitZ = o_maxlit;o_maxtermZ = o_maxterm;o_maxvarZ = o_maxvar;
   proof_found = iterative_deepening();
   if(proof_found == 5) goto another_theorem;
   lit_in_hash = 0;
   if(proof_found == 4) {
      o_maxlit++;
      o_maxterm = o_maxterm * 2;
      o_maxvar = 30;
      if(o_maxterm > 80) {o_maxterm = 80;}
      printf("X%d v%d l%d ",o_maxterm,o_maxvar,o_maxlit);
      OutPtr += sprintf(OutPtr,
	     "X%d v%d l%d ",o_maxterm,o_maxvar,o_maxlit);
      proof_found = iterative_deepening();}
   /*stack_hash_codes();*/
   if(o_debug > 2) {print_substitutions();}
   phase1_time = search_time - search_start_time;
   sub_array_last1 = sub_array_last;
   solved_iteration = iter_depth;
   s_ltentry_phase1  = s_ltentry;
   s_htentry_phase1  = s_htentry;
   s_hthit_phase1    = s_hthit;
   s_lthit_phase1    = s_lthit;
   s_htfail_phase1   = s_htfail;
   s_ltfail_phase1   = s_ltfail;
   s_htelim_phase1   = s_htelim;
   s_subsume_phase1  = s_subsume;
   s_unit_res_phase1 = s_unit_res;
   s_unit_sub_phase1 = s_unit_sub;
   s_ltentry = s_htentry = s_hthit = s_lthit = 0;
   s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0;
   s_htelim = 0;
	
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/*----------------------------------------------------*/

   phase = 2;
	
   while(hash_code_active_number != 0 && proof_found < 2) {
      ply = 0; 
      if(o_debug > 2) {print_substitutions();}
      for(i = 0; i < 65536; i++) {
         for(j = 0; j < 6; j++) {sub_array[i][j] = 0;}}
      prepare_subarray();
      extra_pass = 1;
      o_maxlit = o_maxlitZ;o_maxterm = o_maxtermZ;o_maxvar = o_maxvarZ;
      proof_found = iterative_deepening();
      if(proof_found == 5) goto another_theorem;
      s_ltentry = s_htentry = s_hthit = s_lthit = 0;
      s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0;
      s_htelim = 0;
      /*stack_hash_codes();*/}}

/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* check proof.  if proof is  ok, then print results  */
/*----------------------------------------------------*/

phase = 3;
if(o_learn == 0 || (o_learn != 0 && weakened_clause == -1)) {
   if(proof_found != 2) {
      printf("Proof Found by Master\n\n");
      OutPtr += sprintf(OutPtr,
          "Proof Found by Master\n\n");}
   slaves_killed = 0;
   for(i = 0; i < 0; i++) {
      if(spawned_alive[i] == 0) {
         if(kk = pvm_kill(tid[i]) != 0) {
            printf("\nprocess %d not killed -- Error(%d)\n",i,kk);}
          else {
	     slaves_killed++;
	     spawned_alive[i] = 1;}}}}

if(simple_sol == 0 && proof_found < 2) {
   rearrange_cl_array();
   search_proof_line();}

write_results:
write_out_results(proof_found);

/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/

free_space_close_files();
 
/*----------------------------------------------*/
/* prove another theorem?                       */
/*----------------------------------------------*/
 
another_theorem:

if(o_batch == 1) {time_delay = time_now();
   while(time_now() - time_delay < 25) {
      if(pvm_nrecv(-1,MPROOF)) {
         pvm_upkbyte(results_slave,300000,1);}}}
printf("\n");
if(o_batch == 1 || 
  (weakened_clause != -1 && excellent_weakened_proof==0)) {
   goto begin_proof;}

if(cade_format == 1) fatal("\nOK, Quitting Now\n");

printf("Another theorem(Y,N)? ");
decision_char = (char) getchar();printf("%c",decision_char);
if(decision_char == '\n') {goto another_theorem;}
zz = getchar();printf("%c",(char) zz);
if(zz != (int) '\n') {printf("%c",'\n');goto another_theorem;}
if(decision_char == 'y' || decision_char == 'Y') goto begin_proof;
else if(decision_char == 'n' || decision_char == 'N') {
   fatal("\nOK, Quitting Now\n");}
else goto another_theorem;

}
