/*-------------------------------------------------------------*/
/*      theo.c
/*      copyrighted by Monty Newborn 2005
/*
/*      functions:
/*      time_now()..................... #1
/*      init_stats()................... #2
/*      init_options()................. #3
/*      grab_number().................. #4
/*      fatal()........................ #5
/*      init_search_space()............ #7
/*      init_clause_hash_table()....... #8
/*      init_random_hash_array()....... #9           
/*      free_space_close_files()....... #10
/*      read_buffer_word()............. #11
/*      interchange_fxn_const()........ #12
/*      is_fxn_a_constant()............ #13
/*      sort_constants_fxns().......... #14
/*      format_clause()................ #15
/*      count_lit_characters()......... #16
/*      max_term_in_any_lit().......... #17
/*      get_pred_fxn_symbols()......... #18
/*      prepare_clauses().............. #19
/*      prepare_subarray............... #19a (mar 01)
/*      hash_all_base_clauses()........ #20
/*      iterative_deepening().......... #21
/*      get_clause_stats()............. #22
/*      write_out_results()............ #23
/*      main()........>>>>>>>>>>>>>>>>> #24
/*-------------------------------------------------------------*/
/*
/*      LINKAGE OF FUNCTIONS
/*
/*main()--------------<<<<<<<<<<<<<<<<<<<<<<<<<.........theo.c
/*
/* begin_proof:
/* 
/*       compile_clauses()..............................affc.c
/*       init_cl_array()................................pass2.c
/*       init_stats()...................................theo.c
/*       init_search_space()............................theo.c
/*       prepare_clauses()..............................theo.c
/*             get_pred_fxn_symbols()...................theo.c
/*             read_buffer_word().......................theo.c
/*             format_clause()..........................theo.c
/*                   read_buffer_word().................theo.c
/*             sort_constants_fxns()....................theo.c
/*                   is_fxn_a_constant()................theo.c
/*                   interchange_fxn_const()............theo.c
/*             max_term_in_any_lit()....................theo.c
/*                   count_lit_characters().............theo.c
/*       init_random_hash_array().......................theo.c
/*       hash_all_base_clauses()........................theo.c
/*             calc_lit_hc()............................hash.c
/*             calc_clause_hashcode()...................hash.c
/*             add_clause_to_ht().......................hash.c
/*             calc_alt_hashcodes().....................hash.c
/* 
/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simply input clauses if possible.                  */
/*----------------------------------------------------*/
/* 
/*        eliminate()...................................elim.c
/* 
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/
/* 
/*        iterative_deepening().........................theo.c
/*              init_clause_hash_table()................theo.c
/*              add_clause_to_ht()......................hash.c
/*              calc_alt_hashcodes()....................hash.c
/*              search_tree()...........................search.c
/*              save_line().............................pass2.c
/*              time_now()..............................theo.c
/* 
/*        stack_hash_codes()............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/* put on stack new hash codes that must yet be       */
/* found on another pass through phase 2.             */
/*----------------------------------------------------*/
/* 
/*       iterative_deepening()..........................theo.c
/*       stack_hash_codes().............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* set phase to 3, check proof, then print results    */
/*----------------------------------------------------*/
/* 
/*       rearrange_cl_array()...........................pass2.c
/*       search_proof_line()............................pass2.c
/*       write_out_results()............................theo.c
/*             print_clause()...........................tdump.c
/*             get_clause_stats().......................theo.c
/*
/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/
/* 
/*       free_space_close_files().......................theo.c
/*
/*------------------------------------------------------------------*/
			 

#include "const.h"
#include "clauseS.h"
#include "compwff.h"
 
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <pvm3.h>

 
/*----------------------------------------------------------------*/
/* external functions called from this file                       */
/*----------------------------------------------------------------*/

extern void CompileTptp();               /* function in tptpi.c   */
extern int CompareStrings();

extern void compile_clauses();          /* function in affc.c     */
 
extern int search_tree();               /* functions in search.c  */
 
extern int calc_lit_hc();                /* functions in hash.c   */
extern int calc_alt_hashcodes();
extern void calc_clause_hashcode();
extern int add_clause_to_ht();
extern void normalize_literal();
extern void eliminate();                 /* function in elim.c    */

extern void init_cl_array();             /* functions in pass2.c  */
extern void save_line(); 
extern void stack_hash_codes();
extern void rearrange_cl_array();
extern void search_proof_line();
extern void print_hash_code_list();

extern void print_clause();              /* function in tdump.c   */
extern void print0();

extern void equal_predicate();

extern void prepare_subarray();             /* function in gen.c  */
extern void copy_clause3();
extern clause *copy_clause();
extern void rename_functions();
extern void modify_clause();
extern void mark_questionable();
extern void reconsider_proof();
extern void det_dependencies();

/*----------------------------------------------------------------*/
/* hash_code_list_active is a bit mapping of the hash codes on    */
/* the hash code list that remain to be found.  bit i is a 1      */
/* if hash code i has yet to be found.                            */
/*----------------------------------------------------------------*/
 
extern int hash_code_active_number;
extern int killer_hashx, killer_hashy;

/*----------------------------------------------------------------*/
/* the hash code of the clause whose complement is found in the   */
/* clause_hash_table implying a contradiction; it is thus magic!  */
/*----------------------------------------------------------------*/

extern int magic_hashx, magic_hashy;
extern int alt_HX;

/*----------------------------------------------------------------*/
/* array for substitution strings                                 */
/*----------------------------------------------------------------*/

extern int sub_array[65536][6];
extern int sub_array_last;
extern int const_found_equal;

/*----------------------------------------------------------------*/
/* array for substitution strings used in proof                   */
/*----------------------------------------------------------------*/

extern int deleted_clauses;

extern int cl[PROOF_ARRAY_SIZE][14];
extern int xl[PROOF_ARRAY_SIZE][14];

extern int questionable[2000];

extern int change_array[50][10];
extern int change_level;

extern int p_change_array[2000][7];
extern int p_change_number;
extern int randy;
extern int excellent_weakened_proof;
extern int proof_length;

extern int reinstate_parent[1000];

extern int o_maxterm_tempy, o_maxvar_tempy,o_maxlit_tempy;

/*----------------------------------------------------------------*/
/* internally defined functions                                   */
/*----------------------------------------------------------------*/

int time_now();
void init_stats();
int init_options();
int grab_number();
void fatal();
void init_search_space();
void init_clause_hash_table();
void init_random_hash_array();
void free_space_close_files();
void read_buffer_word();
void interchange_fxn_const();
int is_fxn_a_constant();
void sort_constants_fxns();
void format_clause();
int count_lit_characters();
void max_term_in_any_lit();
int get_pred_fxn_symbols();
void prepare_clauses();
void hash_all_base_clauses();
int iterative_deepening();
void get_clause_stats();
void write_out_results();
int main();
 
/*----------------------------------------------------------------*/
/* arrays used for reading in input file and converting           */
/* to format compatible with the search.                          */
/*----------------------------------------------------------------*/
 
char compiled_code_buffer[MAX_COMPILED_THM_SIZE];
char compiled_code_saved[MAX_COMPILED_THM_SIZE];
char *comp_code_buffer_index;
char *comp_code_saved_index;
int compiled_code_size[14];
 
pfnode pfspace[PF_SPACE_SIZE];
int kode[MAX_INPUT_THM_SIZE];
char name_buff[MAX_CHAR];
cnode pred_ptr_list[PRED_PTR_SIZE];

/*----------------------------------------------------------------*/
/* space for clause list generated in elim.c for simplifying      */
/*----------------------------------------------------------------*/

int elim_infs[ELIM_ARRAY_SIZE];        /* clause list for elim.c  */

/*----------------------------------------------------------------*/
/* clist contains pointers to the clauses (base clauses and       */ 
/* clauses generated during the search.)                          */
/*----------------------------------------------------------------*/

clause *clist[CSPACE_CLAUSES];    /* first few are base clauses   */
int good_clauses[MAX_BASE_CLAUSES];
int good_inferred_cls;     /* # inferred clauses used in proof  */
clause *ulist[500];

/*----------------------------------------------------------------*/
/* each entry in depth is the top, bottom, and current clause     */
/* under search as well as a pointer to the current clause.       */
/* each entry is thus four int  words.                            */
/*----------------------------------------------------------------*/

dnode  depth[DEPTH_PTR_SPACE];    /* depth ptr info               */

/*----------------------------------------------------------------*/
/* hr_hashcodes store the hashcode that was used in an            */
/* instantiation (in calc_inferences())                           */
/*----------------------------------------------------------------*/

int hr_hashcodes[3][50];          /* array to store hr hashcodes  */

/*----------------------------------------------------------------*/
/* space for output, stored temporarily until output to disk file */
/*----------------------------------------------------------------*/

char *OutPtr;
char results[300000];
char *OutPtrZ;

/*----------------------------------------------------------------*/
/* the following two arrays occupy a large amount of memory       */
/* the space is dynamically allocated and is currently set to:    */
/* clause space: 256K words     clause_hash_table: 256K entries   */
/*----------------------------------------------------------------*/  

int *clause_space;
int *u_space;
hnode *clause_hash_table;           /* ptrs to clause hash tables */
hnode *split_hash_table;
unsigned int o_arraysize;
unsigned int o_splitsize;
int array_sizem1;
unsigned int cspace_size;
int array_power;

/*----------------------------------------------------------------*/
/* rand_array is dynamically allocated. its size depends on the   */
/* particular theorem.                                            */
/*----------------------------------------------------------------*/
 
rnode *rand_array;
 
/*----------------------------------------------------------------*/
/*  arrays for storing name of input and outpout theorem          */
/*----------------------------------------------------------------*/

char *theorem_ptr, *theorem_out_ptr, *theorem_out_ptr2;
char *orig_thm_ptr;
char orig_thm_name[256]; /* name of theorem read in               */
char theorem_name[256];  /* name of theorem read in               */
char theorem_out_name[256];/* name of file where output is placed */
char theorem_out_name2[256];/* if sol found, new name of theorem  */
int  theorem_out_name2end; /* ptr to last char in output file name*/
int number_chars;     /* number of characters in name of theorem  */ 
int slash_char;            /* location of last slash of input thm */

/*----------------------------------------------------------------*/
/* variables that are concerned with recording time               */
/*----------------------------------------------------------------*/
 
int accumulated_time;
int elim_starting_time; 
int eliminate_time;
int search_start_time;
int search_time;
int phase1_time;                      /* time for phase 1         */
 
int base;                           /* ID base last base clause   */
int org_base;                 /* original number of base clauses  */
int clauses_alive;       /* count of the number of ACTIVE clauses */
int nc;                          /* # clauses in negated concl.   */
int iter_depth;                    /* current iteration depth     */
int solved_iteration;            /* iteration number when solved  */
int ply;                           /* current search depth        */
int nextc;                     /* location where next clause goes */
int nextu;
int lit_in_hash;
int sub_array_last1;
int rename_fxns, rename_preds;
int length_of_proof;
int satisfiable, unknown, tough;
int weakened_ply, weakened_clause;
int weakened_clause_count;
int numb_trivial_proofs;
int questionableUsed;
int first_attempt;

int decompose, decompose2, most_decomposible, decomposible_parts;
int var_vector[16][100], best_var_vector[16][100];
int split_pbs, unsolved_split_pbs, solved_split_pbs[16];
int pick_split_pbm;

/*----------------------------------------------------------------*/
/* number_rows,       numb_cols, numb_entries in the random array */
/*----------------------------------------------------------------*/
 
int numb_rows, numb_cols, numb_entries, rows_less_nfxns;
 
/*----------------------------------------------------------------*/
/* variables concerned with the state of the search               */
/*----------------------------------------------------------------*/
 
int in_elim;                    /* in_elim = 0 when in eliminate  */
int simple_sol;              /* simple_sol = 1 if elim finds sol  */
int extra_pass;     /* denotes whether more than 1 pass was done  */
int phase; /* search phase: elim = 0, 1st pass = 1, 2nd pass = 2  */
int big_bug;      /* set to 1 when a hash error misleads prover   */

//FILE *infile;
 
char input_buffer[256]; /* buffer for input file name and options */ 
 
int  ww;  /* temp loc for reading chars from compiled_code_buffer */
 
int  *var_substitution_tbl[MAX_VARS*2]; /* var substitution table */
int  pstack[500],*psp;             /* used in unify.c             */
 
int  equal_pred;     /* =1 if "equal" or "Equal" pred is in thm   */
int  equal_clist_index;       /* the clause id of the eq identity */
int  equal_xyyzxz;
int  units_begin;     /* the index of the first unit clause       */

int max_var;
int number_preds;  /* number of different preds in input clauses  */
int number_fxns;   /* number of different fxns in input clauses   */
int number_constants; /* # of different constants in input cls    */
int number_const_nc;  /* number of constants in negated conclus   */
int number_const_hash; /* number of constants for hashing         */
int max_lit;
int max_term_in_lit;                /* maximum terms in a literal */
int max_term_in_nc;                 /* fattest literal in nc      */
int max_vars_in_lit;
int tab_amount;      /* used to keep track of the number of tabs  */

int  maxl, maxv, maxx; /* max lits, vars, trms in clause in proof */
 
char *pidtbl[MAX_PRED_IDS];   /* ptrs to strings for predicates   */
char *fidtbl[MAX_FXNS_IDS];   /* ptrs to strings for functions    */
int skolem[MAX_FXNS_IDS];     /* 1 if constant in nc only         */

cnode *ctbl[MAX_BASE_CLAUSES];

int begin_main_line, end_main_line, on_main_line;
 
int batch_start_time, batch_total_time;
int batch_satisfiable;
int cade_format;
int proof_found_remember;
int weakened_proof_remember;
char *weakProofBeginsHere;
char *weakProofEndsHere;

/*----------------------------------------------------------------*/ 
/* The following are vars that record statistics on the search.   */
/*----------------------------------------------------------------*/
 
int  s_nod;                       /* # nodes on all iterations    */
int  s_nod_last;
int  learn_nod;
int  s_nod0;                      /* # nodes at beginning of pass */
int  s_res;              /* # resolvents from binary resolution   */
int  s_fac;              /* # factors inferred                    */
int  s_subsume;          /* # of s-subsumed clauses (see 6.14)    */
int  s_subfxn;           /* subsumed ending in fxn                */
 
int  s_htentry; /* # of entries in clause hash table (see 6.10.3) */
int  s_hthit;   /* # of clause hash table hits                    */
int  s_htelim;  /* # of clauses elim by ht hits (see 6.10.4)      */
int  s_htfail;  /* # of failures to insert clause in hash table   */
 
int  s_htentry_phase1;    /* # of entries in cl ht on phase 1     */
int  s_hthit_phase1;      /* # of ht hits in phase 1              */
int  s_htelim_phase1;     /* # of clauses elim by ht hits in ph 1 */
int  s_htfail_phase1;     /* # of ht failures in phase 1          */

int  s_ltentry_phase1;    /* # of entries in lit ht on phase 1    */
int  s_lthit_phase1;      /* # of unit ht resolutions in phase 1  */
int  s_ltfail_phase1;     /* # of unit clause failures in phase 1 */
int  s_unit_res_phase1;   /* # of unit ht resolutions in phase 1  */
int  s_unit_sub_phase1;   /* # of unit clause subsump in phase 1  */
int  s_subsume_phase1;    /* # of s-subsumed clauses in phase 1   */

int  s_ltentry;           /* # of single literal clauses          */
int  s_lthit;     /* # unit hash table resolutions (see 6.11.3)   */
int  s_ltfail;    /* # unit clause failures on entering ht        */
int  s_unit_res;  /* # of hash table resolutions (see 6.11.2)     */
int  s_unit_sub;  /* # of unit clause subsumptions (see 6.11.3)   */
int  s_default;   /* equals zero if default parameters are used   */

int  s_subs;      /* number of attempts to find a string subst.   */
int  s_subs_suc;  /* number of successful "   "   "   "    "      */

int  s_too_many_terms;
 
/*----------------------------------------------------------------*/
/* the following are variables that (in most cases)               */
/* the user can set as desired.  They control the search.         */
/*----------------------------------------------------------------*/
 
int o_elim;       /* o_elim = 1 => do elim, else do not do elim   */
int o_merge;      /* MERGE = ancestors must be merge clauses      */
int o_nc;         /* 1 = nc-strategy                              */
int o_hash;       /* 1 = clause hash table used in proof          */
int o_extend;     /* UNIT = allow unit extension past maxd        */
int o_subsume;    /* 1 = simple-subsumbtion used (0 => not used)  */
int o_maxlit;     /* n = max size of clauses (0 = no limit)       */
int o_maxvar;     /* o_maxvar = max # vars in a clause            */
int o_maxterm;    /* max terms in literal                         */
int o_maxtermM1;  /* max terms in literal - 1                     */
int o_maxsterm;   /* max term in literal in search phase          */
int o_maxtime;    /* max time to search for a proof               */
int o_batchtime;  /* max time to search for a proof               */
int s_starttime;
 
int o_step;       /* step depth                                   */
int o_print;      /* print depth                                  */
int o_print_low;  /* first node to print                          */
int o_print_high; /* last node to print                           */
int o_debug;      /* user is supposed to be unaware of this optn  */
int o_relevent;   /* = 1 => print only clauses involved in proof  */
int o_elimprint;  /* o_elimprint = 1 => print elim info           */
int o_tptp;       /* o_tptp = 1 => input file in tptp format      */
int o_batch;      /* o_batch = 1 if batch file is input           */
int o_learn;      /* 0 => no learning, 1 => learning              */
int o_learn_lits;
int o_learn_char;
int o_print_nodes;
int o_weakened_index;
int clauseToWeaken;

int x_maxterm;    /* x_maxterm = 0 => calc o_maxterm              */
int x_stop;       /* can be set to 1 by user in interactive mode  */
int x_cr;         /* when printing clause, add cr when o_cr = 1   */

int mytid;
int myparent;
int cc;
char buf[20000];
int SlaveFoundProof;
int slave_quit;
int theorem_number;
int theorem_rating;
int rcvd_theorem_number;
int spawned;
int this_computer, this_computer_temp;
 
static int term_column[] =
{0,0,0,0,0,0,1,1,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,7,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8};

static int tms3[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
  10,   12,   14,   17,   20,   23,   28,   35,   50,  /*05*/
   0,   14,   16,   20,   23,   26,   32,   36,   50,  /*07*/
   0,    0,   18,   22,   25,   28,   32,   40,   50,  /*09*/
   0,    0,    0,   24,   27,   30,   33,   42,   50,  /*12*/
   0,    0,    0,    0,   27,   30,   33,   42,   50,  /*15*/
   0,    0,    0,    0,    0,   32,   35,   44,   50,  /*18*/
   0,    0,    0,    0,    0,    0,   36,   46,   50,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   48,   50,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   50,  /*+25*/
};


static int tms4[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
   9,   12,   14,   16,   18,   21,   24,   30,   40,  /*05*/
   0,   14,   16,   18,   18,   21,   24,   30,   40,  /*07*/
   0,    0,   18,   18,   20,   22,   24,   30,   40,  /*09*/
   0,    0,    0,   20,   20,   24,   25,   32,   40,  /*12*/
   0,    0,    0,    0,   20,   24,   25,   32,   40,  /*15*/
   0,    0,    0,    0,    0,   24,   26,   32,   40,  /*18*/
   0,    0,    0,    0,    0,    0,   26,   32,   40,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   32,   40,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   40,  /*+25*/
};

static int tms5[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
   9,   12,   14,   16,   18,   19,   22,   23,   24,  /*05*/
   0,   14,   16,   18,   18,   19,   22,   23,   24,  /*07*/
   0,    0,   18,   18,   20,   19,   22,   23,   24,  /*09*/
   0,    0,    0,   20,   20,   19,   22,   23,   24,  /*12*/
   0,    0,    0,    0,   20,   19,   22,   23,   24,  /*15*/
   0,    0,    0,    0,    0,   19,   22,   23,   24,  /*18*/
   0,    0,    0,    0,    0,    0,   22,   23,   24,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   23,   24,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   24,  /*+25*/
};

static int tms[] = {

/*------------------------------------------------------------*/
/* 5     7     9    12    15    18    21    25   +25 
/*------------------------------------------------------------*/
 
  10,   10,   10,   10,   10,   10,   10,   10,   10,  /*05*/
   0,   10,   10,   11,   11,   11,   11,   11,   10,  /*07*/
   0,    0,   10,   11,   12,   12,   12,   12,   10,  /*09*/
   0,    0,    0,   12,   12,   12,   12,   12,   10,  /*12*/
   0,    0,    0,    0,   15,   15,   15,   12,   10,  /*15*/
   0,    0,    0,    0,    0,   14,   14,   12,   10,  /*18*/
   0,    0,    0,    0,    0,    0,   14,   12,   10,  /*21*/
   0,    0,    0,    0,    0,    0,    0,   12,   10,  /*25*/
   0,    0,    0,    0,    0,    0,    0,    0,   10,  /*+25*/
};

static char alphabet32[] =
 {'0','1','2','3','4','5','6','7','8','9',
  'a','b','c','d','e','f','g','h','i','j',
  'k','l','m','n','o','p','q','r','s','t',
  'u','v','w','x','y','z','A','B','C','D',
  'E','F','G','H','I','J','K','L','M','N'};

/*----------------------------------------------------------------*/
/*      time_now()                    ...................... #1
/*
/*      added to consolidate all system calls to theo 
/*
/*      called by:  main()                         theo.c
/*                  iterative_deepening            theo.c
/*                  halted_search_options()        tdump.c
/*                  eliminate()                    elim.c
/*
/*      calls:      nobody
/*
/*      returns:    time of day as a an integer
/*---------------------------------------------------------------*/
 
int time_now()
{
return(time(0));
}
 
/*--------------------------------------------------------------*/
/*      init_stats()                   ..................... #2
/*
/*      init statistical parameters gathered during the search
/*
/*      called from:    main()       theo.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void init_stats()
{
s_nod  = 0; 
s_fac  = 0;             
s_res  = 0;
s_nod0 = 0;
s_subs = 0;
s_subs_suc = 0;
solved_iteration = 0;
 
/*---------------------------------------------------------*/
/* stats for clause hash table and lit clause hash table   */
/*---------------------------------------------------------*/
	
s_htentry = 0;       
s_hthit   = 0;      
s_htfail  = 0;
s_ltentry = 0;       
s_lthit   = 0;      
s_ltfail  = 0; 
s_htelim  = 0;
s_unit_res = 0;
s_unit_sub = 0;
s_subsume = 0;
s_subfxn = 0;
s_too_many_terms = 0;
 
tab_amount = 0;        
	
}
 
/*----------------------------------------------------------------*/
/*      init_options()                          .............. #3
/*
/*      reads the search options that the user requests in the
/*      command line.
/*
/*      called from:  enter_theorem_name()    theo.c
/*
/*      calls:        grab_number()           theo.c
/*
/*      returns:      the number of chars in the name of the theorem
/*                    or a -1 if help was requested
/*----------------------------------------------------------------*/
 
int init_options()
{
int i, j;
unsigned int k;
 
/*-----------------------------------------------------------------*/
/* initialize options to default values                            */
/*-----------------------------------------------------------------*/
    
o_elim      = 1;  /* Option 1... elim phase: 1/0/2 => do/skip/only */
o_merge     = 1;  /* Option 2... merge strat: 1/0 => off/on        */
o_nc        = 0;  /* Option 3... linear-nc: 0/1 => off/on          */
o_hash      = 1;  /* Option 4... clause ht on: 1/0=> cht off/on    */
o_extend    = 1;  /* Option 5... extended search: 1/0=> on/off     */
o_subsume   = 1;  /* Option 6... s-subsumption: 1/0 => use/don't   */
o_maxlit    = 0;  /* Option 7... maximum # lits in a cls in proof  */
o_maxterm   = 16; /* Option 8... maximum # terms in a literal      */
x_maxterm   = 0;
o_maxsterm  = 0;
 
o_maxvar    = 0;   /* Option 9... maximum # diff vars in a clause  */

                        /* Option 10.. size of clause_hash_table   */

o_arraysize = 4*4194304; 
array_sizem1 = 4*4194304 - 1;  
cspace_size = 2097152;
array_power = 24;

o_tptp      = 3;   /* Option 11.. 0/1 not in/in tptp format        */

o_batchtime = 3600;/* Option 12.. maximum time for proof..seconds  */

o_elimprint = 0;   /* Option 13.. 0/1 do/ do not give phase 0 info */
o_step      = 0;   /* Option 14.. step through tree to ply o_step  */ 
o_print     = 0;   /* Option 15.. print tree to ply o_step         */
o_print_low = 0;
o_print_high= 1000000000;
o_print_nodes = 0;

o_debug     = 0;   /* Option 16.. debug option: 0/1 => off/on      */
o_relevent  = 0;   /* Option 17.. 0/1 do/do not prt all proof cls  */

//spawned     = 0;

/*-----------------------------------------------------------------*/
/*        END OF OPTIONS                                           */
/*-----------------------------------------------------------------*/

s_default   = 0;
big_bug     = 0;       /* set to 1 if a hash error found inphase 3 */
x_cr        = 0;       /* set to 1 => print cr when printing cls   */

/*------------------------------------------*/
/* scan the arguments passed to the program */
/* if a "?" appears anywhere on the line    */
/* print out a "help" message               */
/*------------------------------------------*/
 
         
/*---------------------------------------------------*/
/* end of option list:                               */ 
/* blank out input_buffer except for name of theorem */
/*---------------------------------------------------*/
	
for(i = 0; i < 255; i++) {
   if(input_buffer[i] == ' ' || input_buffer[i] == '\n') {
      j = i;
      for(j = i; j < 256 ; j++) input_buffer[j] = ' ';
      break;}}

/*--------------------------------------------------*/ 
/* return the number of characters in the name      */
/*--------------------------------------------------*/

if(o_learn == 2) {
   o_learn_lits = o_maxlit;
   o_learn_char = o_maxsterm;}
return(i);
}
 
/*-----------------------------------------------------------------*/
/*      grab_number(s)                       .............. #4
/*
/*      reads a string of characters and converts them to a
/*      integer equivalent
/*
/*      called from:  init_options()              theo.c
/*                    halted_search_options()     tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      an integer correspond to the string s
/*-----------------------------------------------------------------*/
 
int grab_number(s)
int s;
{
int n;
	
for(n = 0; ; s++) {
   if(input_buffer[s] >= '0' && input_buffer[s] <= '9')
      n = n * 10 + ( (int) (input_buffer[s] - '0') );
   else
      return(n);}
}                                                       
 
/*-----------------------------------------------------------------*/ 
/*      fatal(msg)                          ................. #5
/*
/*      fatal will print an error message on the screen and quit
/*
/*      called by:    main()                theo.c
/*                    prepare_clauses()     theo.c
/*                    format_clause()       theo.c
/*                    eliminate()           elim.c
/*
/*      calls:        time_now()            theo.c
/*
/*      returns:      to the operating system after reporting
/*                    error
/*-----------------------------------------------------------------*/
 
void fatal(msg)
char *msg;
{
printf("....%s\n",msg);fflush(stdout);
OutPtr += sprintf(OutPtr,"....%s\n",msg); 

if(*msg == 'D') {
   exit(0);}
}
 
/*-------------------------------------------------------------------*/
/*      init_search_space()            ......................6#
/*
/*      allocates space for clauses.
/*      allocates space for clause hash table.
/*      allocates space for pointers at each level in tree
/*
/*      called from:   main()                 theo.c
/*
/*      calls:
/*
/*      returns:       nil
/*-------------------------------------------------------------------*/
 
void init_search_space()
{       

clause_space = (int *) calloc(cspace_size,4);   
nextc = 0;
clist[nextc] = (clause *) clause_space;

if(nextu == 0) {
   u_space = (int *) calloc(10000,4);   
   ulist[nextu] = (clause *) u_space;}
 
clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);
}
 
/*-----------------------------------------------------------------*/
/*      init_clause_hash_table()          .................#7
/*
/*      initialize the clause_hash_table to all 0's
/*
/*      called by:      main()                             theo.c
/*                      iterative_deepening()              theo.c
/*
/*      calls:
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void init_clause_hash_table()
{
int i, j;

/*----------------------------------------------*/
/* allocate 4 * HT_SIZE int  words for the hash */
/* table.  initially set all words to 0.        */
/*----------------------------------------------*/
 
clause_hash_table =  (hnode *) calloc(3 * o_arraysize,4);

/*----------------------------------------------*/
/* clear hr_hashcodes array                     */
/*----------------------------------------------*/

for(i = 0; i < 3; i++) {
   for(j = 0; j < 50; j++) {hr_hashcodes[i][j] = 0;}}

return;
}
 
/*-----------------------------------------------------------------*/
/*      init_random_hash_array()    ........#8
/*
/*      initialize the rand_array to random numbers 
/*
/*      called from:   main()                      theo.c 
/*
/*      calls:
/*
/*      returns:       nil
/*-----------------------------------------------------------------*/
 
void init_random_hash_array()
{
rnode *r1;
int i;
unsigned int num_ent;
	
/*--------------------------------------------------------------*/
/* the number of rows in rand_array = MAX_VARS + numb_fxns      */
/* the number of columns in rand_array = numb_chars             */
/*--------------------------------------------------------------*/
 
numb_rows = MAX_VARS + number_fxns + number_preds;
numb_cols = 200;
rows_less_nfxns = numb_rows - number_fxns - 1;
 
/*--------------------------------------------------------------*/
/* the number of entries in the array rand_array is the         */
/* product of the numb_rows * numb_cols                         */ 
/*--------------------------------------------------------------*/
 
numb_entries = numb_rows * numb_cols;
num_ent = (unsigned int) (numb_entries * 2 + 1); 
if(num_ent < 50) {
   numb_entries = 25;
   num_ent = 51;}

/*--------------------------------------------------------------*/
/* allocate space for rand_array                                */
/*--------------------------------------------------------------*/

rand_array = (rnode *) calloc(num_ent * 2,4);

/*--------------------------------------------------------------*/
/* make 32-bit random number for each element in rand_array     */
/*--------------------------------------------------------------*/
 
srand(2);
 
for(i = 0 ; i < (int) numb_entries; i++) {
   r1 = &rand_array[i];
   r1->r_hash1 = rand()*16384 + rand();
   r1->r_hash2 = rand()*16384 + rand();}
}
 
/*-------------------------------------------------------------*/
/*      free_space_close_files()  ...................#9
/*
/*      frees space and closes files used in phases 1 and 2
/*
/*      called by:   main()                 theo.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void free_space_close_files()
{
free(clause_hash_table);
free(rand_array);
free(clause_space);  
}
 
/*--------------------------------------------------------------*/
/*      read_buffer_word()               .............. #10
/*
/*      read a word (integer) from the input file
/*
/*      called by:    prepare_clauses()     theo.c
/*                    format_clause()       theo.c
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void read_buffer_word()
{
union u_tag {
   int w; 
   char ttext[4];} 
   temp;
			
int j;
	
for(j = 0; j < 4; j++) {
   temp.ttext[j] = *comp_code_buffer_index++;}
ww = temp.w;
}
 
 
/*--------------------------------------------------------------*/
/*      interchange_fxn_const()          ............. #11
/*
/*       called by:      sort_constants_fxns()      theo.c
/*
/*       calls:          nobody
/*
/*       returns:         nil
/*--------------------------------------------------------------*/
 
void interchange_fxn_const(i,j)
int i,j;
{
int k, *p, a;
clause *c;
literal *l;
char *fide;
	
for(k = 0; k <= base ; k++) {
   c = clist[k];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      p = &l->l_id;
      for(a = (*p++ & ARGMASK); a > 0; a--) {
	     if(*p & FCNMASK) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           *p = (*p & ~IDMASK) + (j << ARGSHIFT);}
	        else if(((*p & IDMASK) >> ARGSHIFT) == j) {
	           *p = (*p & ~IDMASK) + (i << ARGSHIFT);}
	        a += *p++ &ARGMASK;}
	     else {
	        p++;}}}}
			
fide = fidtbl[i];
fidtbl[i] = fidtbl[j];
fidtbl[j] = fide;
}
 
/*--------------------------------------------------------------*/
/*      is_fxn_a_constant()                .............. #12
/*
/*      called by:      sort_constants_fxns()       theo.c
/*
/*      calls:          nobody
/*
/*      returns:        1 if fxn is a constant, 0 otherwise
/*-----------------------------------------------------------*/
 
int is_fxn_a_constant(i,top_clause,bottom_clause)
int i,top_clause,bottom_clause;
{
int k, a, *p;
clause *c;
literal *l;
	
for(k = top_clause; k <= bottom_clause ; k++) { 
   c = clist[k];
   
   for(l = c->c_lit; l != 0; l = l->l_next) {  /* scan lits  */
      p = &l->l_id;
      for(a = (*p++ & ARGMASK); a > 0; a--) {
	     if(*p & FCNMASK) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           if((*p & ARGMASK) == 0) return(1);
	           else return(0);}
	        a += *p++ & ARGMASK;}
	     else {
	        p++;}}}}
return(0);
}
 
/*--------------------------------------------------------------*/
/*      sort_constants_fxns()           .............. #13
/*
/*      sort the constants and functions so that the constants
/*      are placed at the top of the list of functions and constants
/*      once that is done, place the constants in the nc at the
/*      head of the constants list.
/*
/*      called by:   prepare_clauses()          theo.c
/*
/*      calls:       is_fxn_a_constant()        theo.c
/*                   interchange_fxn_const()    theo.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void sort_constants_fxns()
{
int i, k;
 
/*-----------------------------------------------*/
/*    put the constants at the head of the list  */
/*-----------------------------------------------*/

number_constants = 0;
for(i = 1; i <= number_fxns ; i++) {
   if(is_fxn_a_constant(i,0,base)) {
      number_constants++;       
      if(i >= number_fxns) break;}
   else {
      for(k = number_fxns; k > i; k--) {
	     if(is_fxn_a_constant(k,0,base)) {
	        interchange_fxn_const(i,k);
	        number_constants++;
	        break;}}}}  

/*-----------------------------------------------*/
/*   now resort the constants in the nc to the   */
/*   top of the list.                            */
/*-----------------------------------------------*/

number_const_nc = 0;
for(i = 1; i <= number_constants; i++) {

   if(is_fxn_a_constant(i,base - nc + 1,base)) { 
      number_const_nc++; 
      continue;}
   else {
      for(k = i + 1; k <= number_constants; k++) {
	 if(is_fxn_a_constant(k,base - nc + 1,base)) {
            number_const_nc++;
	    interchange_fxn_const(i,k);
	    break;}}}} 

if(number_constants <= 5) {number_const_hash = number_constants;}
else {number_const_hash = 5;}

for(i = 1; i <= number_constants; i++) {
   skolem[i] = 1;
   if(is_fxn_a_constant(i,0, base - nc)) {skolem[i] = 0;}}

/*
for(i = 1; i <= number_fxns; i++) {
   if(o_debug > 0) {
      printf(" %d %s ",i, fidtbl[i]);
      OutPtr += sprintf(OutPtr," %d %s",i,fidtbl[i]);}
   if(skolem[i] == 1) print0("(S)");
   if(i == number_constants) {print0(" :");}
   if(i == number_const_nc) {print0(" .");}}
print0("\n");
*/

}

/*--------------------------------------------------------------*/
/*      format_clause()                   .............. #14
/*
/*      format_clause will read clause from the input and stuff
/*      it into the clist
/*
/*      called from:    prepare_clauses()   theo.c
/*
/*      calls:          read_buffer_word()  theo.c
/*                      fatal()             theo.c
/* 
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void format_clause()
{
clause *c;
literal *previous_literal;
int *p,a;
	
ply = 0;       
read_buffer_word();read_buffer_word();/* these are leading 0's */
 
/*------------------------------------------------------*/
/* initialize clause header (this is not part of the    */
/* input file, we add it as we read in the clause)      */
/*------------------------------------------------------*/
 
c = clist[nextc++];                  /* the new clause         */
c->c_id = nextc;
c->c_bits = ACTIVE | MERGE | BASE; /* base clauses are merges  */
if(nextc > base - nc + 1) {c->c_bits |= NC;}
c->c_rpar = c->c_lpar = 0;       /* no parents (base clause)   */
c->c_lit = 0;                        /* ^ first literal        */
c->c_nlit = 0;                             /* # literals       */
c->c_nv   = 0;                       /* # distinct variables   */
	
p = (int  *) (c + 1);           /* where predicate code goes   */
previous_literal = 0;                 /* literal chain         */
 
/*----------------------------------------*/
/* read in each literal, end clause at -1 */
/*----------------------------------------*/    
 
read_buffer_word(); 
while(ww != -1) {
   if(!previous_literal) {               /* link last literal  */
      c->c_lit = (literal *) p;}
   else {
      previous_literal->l_next = (literal *) p;}
      
   previous_literal = (literal *) p;
   previous_literal->l_id = ww;
   previous_literal->l_next = 0;
 
   p = (int *) (previous_literal + 1);
 
   for(a = ww & ARGMASK; a > 0; a--) {
      read_buffer_word();
      if(ww & FCNMASK) {                          /* function? */
	     *p++ = ww;                            /* yes       */
	     a += ww & ARGMASK;}                   /* end if    */ 
      else {
	     ww >>= ARGSHIFT;                      /* var ID    */
	     *p++ = ww;
	     if(ww > c->c_nv) {c->c_nv = ww;}}}      /* end for */
   
/*----------------------------------------------------------------*/
/* transform equality predicates to standard form                 */
/*----------------------------------------------------------------*/
 
   c->c_nlit++;
		
   read_buffer_word();}                       /* end while  */
 
/*-----------------------------------------------------------------*/
/* set up a ptr to the next clause, pointed to by clist[nextc]     */
/*-----------------------------------------------------------------*/
 
c->c_end = (clause *) p;
clist[nextc] = c->c_end;
	
/*--------------------------------------------------*/
/* if clause has only one literal, set UNIT flag on */
/*--------------------------------------------------*/
 
if(c->c_nlit == 1) {c->c_bits |= UNIT;}
 
if(++c->c_nv > MAX_VARS) {
   OutPtr += sprintf(OutPtr,
	   "Clause %d has too many variables!\n",nextc);
   fatal("Too many variables in clause");}
 
if(!c->c_nlit) {fatal("Empty clause in base");}
}
 
/*---------------------------------------------------------------*/
/*      count_lit_characters()               .............. #15
/*
/*      count_lit_characters counts terms in lit l
/*
/*      called from:  max_term_in_any_lit()        theo.c
/*
/*      calls:        nobody
/*
/*      returns:      number of terms in literal l
/*----------------------------------------------------------------*/
 
int count_lit_characters(l)
literal *l;                    /* l is a literal in clause c      */
{
int a,*p;
int literal_terms;         /* count of the number of terms in lit */
int var_count;
	
p = &l->l_id;        
literal_terms = 0;     
var_count = 0;
	
for(a = *p++ & ARGMASK; a > 0; a--) {
   if(*p & FCNMASK) { 
      literal_terms++;  /* function; inc count; det. # arguments   */
      a += *p++ & ARGMASK;}
   else {               
      literal_terms++;
      var_count++;
      p++;}}              /* variable; inc count                   */

if(var_count > max_vars_in_lit) {max_vars_in_lit = var_count;}

return(literal_terms);
}
 
/*--------------------------------------------------------------*/
/*      max_term_in_any_lit()           .............. #16
/*
/*      calculates the maximum number of terms in any literal
/*      o_maxsterm records this maximum number
/*
/*      called from:    prepare_clauses()         theo.c
/*
/*      calls:          count_lit_characters()    theo.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void max_term_in_any_lit()
{
clause *c;
literal *lit;
int clause_index, ncc;
int lit_term_cnt;

max_term_in_nc = 0;
max_term_in_lit = 0;
max_vars_in_lit = 0;

if(!nc) ncc = 1; else ncc = nc; 

for(clause_index = 0; clause_index<= base; clause_index++) {
   c = clist[clause_index];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      lit_term_cnt = count_lit_characters(lit);
      if(lit_term_cnt > max_term_in_nc && 
                      clause_index >= base - ncc + 1 &&
                      clause_index <= base) {
         max_term_in_nc = lit_term_cnt;}
      if(lit_term_cnt > max_term_in_lit) {
	     max_term_in_lit = lit_term_cnt;}}}

return;
}
			
/*--------------------------------------------------------------*/ 
/*      get_pred_fxn_symbols()        .............. #17
/*
/*      get_pred_fxn_symbols will read in a symbol table and write
/*      it to the output file (and to the log file if there is one
/*
/*      called from:   prepare_clauses()          theo.c
/*
/*      calls:         nobody
/*
/*      returns:       the number of symbols (pred/fxns)
/*---------------------------------------------------------------*/
 
int get_pred_fxn_symbols(qtbl)
char *qtbl[];
{
int a, k;
char *ii, *jj;
static char Eq[] = "Equal";
static char eq[] = "equal";
static char Eqish[] = "equalish";

for(a = 1; ;a++) {  

/*----------------------------------------*/
/* check first if at least one pred, fxn  */           
/*----------------------------------------*/

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++; 
      return(0);}

/*----------------------------------------*/
/* set qtbl[a] to point to beginning of   */
/* the a-th predicate or function         */
/*----------------------------------------*/

   qtbl[a] = ii = comp_code_buffer_index;

   while(*comp_code_buffer_index != 0xa) {comp_code_buffer_index++;} 
 
/*----------------------------------------*/
/* replace terminating '0xa' with a 0     */
/*----------------------------------------*/
  
   *comp_code_buffer_index++ = 0;
 
/*--------------------------------------------------*/
/* print string of characters representing pred     */    
/*--------------------------------------------------*/
    
if(qtbl == pidtbl) {
   OutPtr += sprintf(OutPtr,"%s ",qtbl[a]);}

/*--------------------------------------------------*/
/* test for equality predicate                      */
/*--------------------------------------------------*/
   jj = ii;
	
   for(k = 0; k < 6 ; k++) {
      if(*jj++ != Eq[k]) {k = 0; break;}}
   if(k) {equal_pred =  (a << ARGSHIFT) + 2;} 

   jj = ii;
   for(k = 0; k < 6 ; k++) {
      if(*jj++ != eq[k]) {k = 0; break;}}
   if(k) {equal_pred =  (a << ARGSHIFT) + 2;} 

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++;
      return(a);}}                      

}                                                       

/*--------------------------------------------------------------*/
/*      prepare_clauses()                .............. #18
/*
/*      prepare_clauses will read the clauses currently stored
/*      in the input_buffer and format them, placing the resulting
/*      clauses in cspace with pointers to each in the clist
/*
/*      NOTES:
/*      - we assume that all variables are numbered from 0..n so
/*        that n+1 is the number of distinct variables
/*
/*      - we pass the predicate and function symbol tables on to
/*        the output directly since we have no need for them
/*
/*      called from:    main()                          theo.c
/*
/*      calls:          get_pred_fxn_symbols()          theo.c
/*                      read_buffer_word()              theo.c
/*                      format_clause()                 theo.c
/*                      max_term_in_any_lit()           theo.c
/*                      sort_constants_fxns()           theo.c
/*                      equality_predicate()            equal.c
/*                      fatal()                         theo.c
/* 
/*      returns:        nil
/* -------------------------------------------------------------*/
 
void prepare_clauses()
{
static int characters[] = {0,3,6,9,12,15,18,21,24,26,28,30,32};
int i, n, x1, x2;
int min_lit;

/*--------------------------------------------------*/
/* set pointer to beginning of compiled_code_buffer */
/*--------------------------------------------------*/

comp_code_saved_index = compiled_code_saved;
comp_code_buffer_index = compiled_code_buffer;
if(weakened_ply != 0) {
   for(i = 0; i < MAX_COMPILED_THM_SIZE; i++) {
      *comp_code_buffer_index++ = *comp_code_saved_index++;}}
else {
   for(i = 0; i < MAX_COMPILED_THM_SIZE; i++) {
      *comp_code_saved_index++ = *comp_code_buffer_index++;}}

comp_code_buffer_index = compiled_code_buffer;

/*--------------------------------------------------------------*/
/* We first pass the symbol table on to the output file.        */
/* This symbol table is two sets of non-empty lines followed    */
/* by an empty line (the predicate and function symbol tables   */
/*--------------------------------------------------------------*/
    
OutPtr += sprintf(OutPtr,"Predicates: ");
number_preds = get_pred_fxn_symbols(pidtbl); /* pred sym table   */
sprintf(OutPtr++,"\n");

OutPtr += sprintf(OutPtr,"Functions: ");
number_fxns = get_pred_fxn_symbols(fidtbl);/* function sym table */
read_buffer_word(); 
 
n = ww; 

base = ww - 1;               /* clist[base] is last base clause  */
org_base = base;              /* original number of base clauses */
 
read_buffer_word();nc = ww;   /* # clauses in negated conclusion */
	
if(n < 2) {fatal("theorem has one clause; proof impossible. Quitting.");}
 
/*--------------------------------------------*/
/* read the clauses and put them in the clist */
/*--------------------------------------------*/
  
while(n-- > 0) {format_clause();}
 
sort_constants_fxns();
 
max_term_in_any_lit();

max_lit = 0;
min_lit = 1000;
for(n = 0; n <= base; n++) {
   if(clist[n]->c_nlit > max_lit) {max_lit = clist[n]->c_nlit;}
   if(clist[n]->c_nlit < min_lit) {min_lit = clist[n]->c_nlit;}}
 
/*-----------------------------------------------*/
/* given the maximum number of terms in          */
/* any literal in any base clause,               */
/* determine o_maxsterm, the maximum number of   */
/* terms in any literal kept during the          */
/* search.  Using the default value:             */
/*                                               */
/*       max_term_in_lit    o_maxsterm           */
/*           0                  0                */
/*           1                  3                */
/*           2                  6                */
/*           3                  9                */
/*           4                 12                */
/*           5                 15                */
/*           6                 18                */
/*           7                 21                */
/*           8                 24                */
/*           9                 26                */
/*          >10        2 * max_term_in_lit + 8   */
/*                                               */
/* the user can set o_maxsterm using the xi      */
/* option.                                       */
/*-----------------------------------------------*/
 
if(x_maxterm == 0) {
   x1 = term_column[max_term_in_nc];
   x2 = term_column[max_term_in_lit];
   if(x1 > 25) {x1 = 25;}
   if(x2 > 25) {x2 = 25;}

   if(max_lit <= 3) {o_maxterm = tms3[x1 * 9 + x2];}
   if(max_lit == 4) {o_maxterm = tms4[x1 * 9 + x2];}
   if(max_lit == 5) {o_maxterm = tms5[x1 * 9 + x2];}
   if(max_lit > 5) {o_maxterm = tms[x1 * 9 + x2];}
   o_maxtermM1 = o_maxterm - 1;



   o_maxterm =
      o_maxterm*(4+((((this_computer*this_computer+4)/5)*2)%8))/6
      + (this_computer/30);
   o_maxtermM1 = o_maxterm - 1;}

if(x_maxterm) {
   o_maxterm = o_maxsterm;
   o_maxtermM1 = o_maxterm - 1;
   if(o_maxsterm < max_term_in_lit) {o_maxsterm = 2*max_term_in_lit+8;}}
 
/*--------------------------------------------------*/
/* determine the max number of literals in any      */
/* base clause and the maximum number of            */
/* different variable names in any base literal     */
/*--------------------------------------------------*/
 
max_var = 0;
	
for(n = 0; n <= base; n++) {
   if(clist[n]->c_nv > max_var)  {max_var = clist[n]->c_nv;}}
if(o_maxvar == 0) {
   o_maxvar = 2 * max_var;
   if(o_maxvar < max_vars_in_lit) {
      o_maxvar = max_vars_in_lit + 3 + (this_computer/40);}}
      
if(o_maxlit == 0) {
   o_maxlit = max_lit + min_lit - 2;
   if(o_maxlit < 3) {o_maxlit = 3;}}
if(max_lit < o_maxlit) {max_lit = o_maxlit;} 

if(max_lit < 6) {
   o_maxlit =  (this_computer % 5) + 2;
   if(o_maxlit == 2 && base < 40) o_maxlit = 3;}
else {
   if((this_computer % 8) == 0)
      o_maxlit = 2;
   else if((this_computer % 8) == 1)
      o_maxlit = 3;
   else if((this_computer % 8) == 2)
      o_maxlit = 3;
   else if((this_computer % 8) == 3)
      o_maxlit = 4;
   else if((this_computer % 8) == 4)
      o_maxlit = 4;
   else if((this_computer % 8) == 5)
      o_maxlit = 5;
   else if((this_computer % 8) == 6)
      o_maxlit = 6;
   else if((this_computer % 8) == 7)
      o_maxlit = max_lit + 1;}
if(o_maxlit > 15)
   o_maxlit = 15;



/*--------------------------------------*/
/* search for equal predicate           */
/*--------------------------------------*/

rename_functions();

if(rename_preds) {

/*----------------------------------------*/
/* if predicates have been renamed, list  */
/*----------------------------------------*/

   OutPtr += sprintf(OutPtr,"Renamed predicates: ");
   for(i = 1; i <= number_preds; i++) {
      OutPtr += sprintf(OutPtr," %s",pidtbl[i]);}
   OutPtr += sprintf(OutPtr,"\n");}

if(rename_fxns) {

/*----------------------------------------*/
/* if functions have been renamed, list   */
/*----------------------------------------*/

    OutPtr += sprintf(OutPtr,"Renamed functions: ");
   for(i = 1; i <= number_fxns; i++) {
      OutPtr += sprintf(OutPtr," %s",fidtbl[i]);
      if(i == number_constants) {OutPtr += sprintf(OutPtr," :");}
      if(i == number_const_nc) {OutPtr += sprintf(OutPtr," .");}}
   sprintf(OutPtr++,"\n");}

equal_predicate();

if(equal_pred && number_preds == 1) {
   if(o_maxterm < 50) {
      o_maxterm = o_maxterm * 1.5;
      o_maxtermM1 = o_maxterm - 1;
      o_maxsterm = o_maxsterm;}}

return;
}
 
/*---------------------------------------------------------------*/
/*      hash_all_base_clauses()            .............. #19
/*
/*      calculates a hash code for each literal and a hash code
/*      for each base clause.  It adds the clause to the literal
/*      hash table if there is only one literal and to the
/*      the clause hash table in any case
/*
/*      called from: main()                      theo.c
/*
/*      calls:       calc_lit_hc()               hash.c
/*                   calc_clause_hashcode()      hash.c
/*                   add_clause_to_ht()          hash.c
/*                   calc_alt_hashcodes()        hash.c
/*		     normalize_literal()         hash.c
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void hash_all_base_clauses()
{
clause *c;
literal *l;
int i;
int l_cnt, o_maxterm_tmp;
	
o_maxterm_tmp = o_maxterm;
o_maxterm = 200; 
 
for(i = 0; i <= base; i++) {
   c = clist[i];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {continue;}
      l_cnt = calc_lit_hc(l,c);}
   calc_clause_hashcode(c);
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(c->c_nlit == 1) {
      for(l = c->c_lit; l != 0; l = l->l_next) {
         if((l->l_id & ELIMLIT) == 0) {break;}}
      normalize_literal(l,c);
      add_clause_to_ht(l->l_X,l->l_Y,1);
      if(c->c_nv) {calc_alt_hashcodes(l,c);}}}  

o_maxterm = o_maxterm_tmp;

/*-----------------------------------*/
/* find first unit clause            */
/*-----------------------------------*/

units_begin = 0;
for(i = 0; i <= base; i++) {
   c = clist[i];
   if(c->c_nlit > 1) {continue;}
   else {
      units_begin = i;
      break;}}
return;
}       


/*----------------------------------------------------------------*/
/*      iterative_deepening()      .............. #20
/*
/*      prove will actually perform the iteratively deepening
/*      depth first search.  We return the depth at which we stop
/*
/*      called by:      main()                     theo.c
/*
/*      calls:          init_clause_hash_table()   theo.c
/*                      add_clause_to_ht()         hash.c
/*                      calc_alt_hashcodes()       hash.c
/*                      normalize_literal()        hash.c
/*                      save_line()                pass2.c
/*                      search_tree()              search.c
/*                      time_now()                 theo.c
/*
/*      returns:        integer regarding the result of search
/*---------------------------------------------------------------*/
 
int iterative_deepening()
{
int s;              /* normally s = 0; if proof found, s = 1     */
int i; 
clause *c;
literal *l;
int total_nodes, total_nodesM1, total_nodesM2;
int l_cnt;
int units_last_iteration;
int units_this_iteration;

total_nodes = total_nodesM1 = total_nodesM2 = 0;

/*---------------------------------------------------------------*/
/*                                                               */
/* Main iterative deepening loop begins here                     */
/*                                                               */
/*---------------------------------------------------------------*/

	
iter_depth = 0;
units_last_iteration = 0;
units_this_iteration = 0;
s_nod0 = s_nod;
 
/*----------------------------------------------------------------*/
/* if the number of base clauses is less than 9, turn off         */
/* merge option                                                   */
/*----------------------------------------------------------------*/

if(org_base < 9) {o_merge = 0;}

/*----------------------------------------------------------------*/
/* before beginning search, clear out hash tables                 */
/*----------------------------------------------------------------*/ 
 
free(clause_hash_table);
init_clause_hash_table();       
 
/*------------------------------------------*/
/* add clauses to clause_hash_table         */
/* and unit clauses to literal_hash_table   */
/*------------------------------------------*/

while(iter_depth < MAX_ITERATIONS) {                    


for(i = 0; i <= base; i++) {
   c = clist[i];
   for(l = c->c_lit; l != 0; l = l->l_next) {
      if((l->l_id & ELIMLIT) != 0) {continue;}
      l_cnt = calc_lit_hc(l,c);}
   calc_clause_hashcode(c);
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(c->c_nlit == 1) {
      for(l = c->c_lit; l != 0; l = l->l_next) {
         if((l->l_id & ELIMLIT) == 0) {break;}}
      normalize_literal(l,c);
      add_clause_to_ht(l->l_X,l->l_Y,1);
      if(c->c_nv) {calc_alt_hashcodes(l,c);}
      test_simple_equality(c);}}

/*--------------------------*/
/* inc iteration depth      */    
/*--------------------------*/
	
   iter_depth++;

/*--------------------------*/
/* Begin search!!!!         */
/*--------------------------*/

   s = search_tree(); 

/*----------------------------------------*/
/* save the current time                  */
/*----------------------------------------*/

   search_time = time_now();
    
/*----------------------------------------------------------*/
/* extended printing routine: on each iteration print:      */
/*          1. time so far                                  */
/*          2. nodes searched on this pass                  */
/*          3. clause hash table entries                    */
/*          4. clause hash table failures                   */
/*          5. unit clause hash table entries               */
/*          6. unit clause hash table failures              */
/*----------------------------------------------------------*/
	
   if(weakened_clause_count == 0) {
   if(time_now() - s_starttime > 5) {
   printf("c%d: %d {T%d N%d-%d [C%d H%d h%d U%d u%d f%d] E%d s%d t%d L%d}\n",
      this_computer,
      iter_depth,(time_now() - s_starttime),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,proof_length);
      fflush(stdout);}

   OutPtr += sprintf(OutPtr,
     "c%d: %d {T%d N%d-%d [C%d H%d h%d U%d u%d f%d] E%d s%d t%d L%d}\n",
      this_computer,
      iter_depth,(time_now() - s_starttime),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,proof_length);}
 

   else {
  /* printf("C%d: %d {T%d N%d-%d [C%d H%d h%d U%d u%d f%d] E%d s%d t%d L%d}\n",
      this_computer,
      iter_depth,(search_time - search_start_time),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);
   fflush(stdout);
*/
   OutPtr += sprintf(OutPtr,
     "C%d: %d {T%d N%d-%d [C%d H%d h%d U%d u%d f%d] E%d s%d t%d L%d}\n",
      this_computer,
      iter_depth,(time_now() - s_starttime),
      s_nod,s_nod - s_nod0,
      s_htentry,s_hthit,s_htfail,
      s_ltentry,s_lthit,s_ltfail,
      sub_array_last,s_subs_suc,s_subfxn,s_too_many_terms);}
 
/*------------------------------*/
/* if proof found, print *, ret */
/*------------------------------*/

// if(s != 1) {print0("\n");}

/*----------------------------------------*/
/* if proof found (s != 0), and in        */
/* phase 1, then save line                */
/*----------------------------------------*/
 
   if(s == 1 && phase == 1) {
      OutPtr += sprintf(OutPtr,"* ");
      save_line(0,0);}
   if(s == 1) {
      stack_hash_codes();
      if(hash_code_active_number == 0) {
	 if(weakened_clause_count == 0) {
	    if(split_pbs == 0) 
	       OutPtr += sprintf(OutPtr,
	                "\nC%d: THE PROOF!!\n",this_computer);
            else {
	       OutPtr += sprintf(OutPtr,
	         "\n\nC%d: THE PROOF of Split Problem %d of %d.\n\n",
		      this_computer,pick_split_pbm,split_pbs);}
	    fflush(stdout);}
         else {
	    OutPtr += sprintf(OutPtr,
	        "C%d: %d-node %d-step weakened proof obtained!!\n",
			  this_computer,s_nod,proof_length);
	    printf("C%d: %d-node %d-step weakened proof obtained!!\n",
	            this_computer,s_nod,proof_length);
		    fflush(stdout);}}
      else if(phase == 1) {
	 if(weakened_clause_count == 0) {
            if(split_pbs == 0) {
	       printf("c%d: Proof Found; details coming!!\n",this_computer);
	       fflush(stdout);
	       OutPtr += sprintf(OutPtr,
	           "c%d: Proof Found; details coming!!\n",this_computer);}
            else {
	       printf(
               "c%d: Proof Obtained to Split Problem %d; details coming!\n",
	                             this_computer,pick_split_pbm);
	       fflush(stdout);
               OutPtr += sprintf(OutPtr,
               "c%d: Proof Obtained to Split Problem %d; details coming!\n",
		   this_computer,pick_split_pbm);}}

         else {
/*          printf("C%d: %d-node %d-step weakened proof obtained;  details coming!!\n",
	               this_computer,s_nod,proof_length);
            fflush(stdout); */

	    OutPtr += sprintf(OutPtr,
	         "C%d: %d-node %d-step weakened proof obtained;  details coming!!\n",
	                this_computer,s_nod,proof_length);}}

      return(1);} 

/*------------------------------*/
/* too much time,...quit        */
/*------------------------------*/
 
   if(s == 2)  {
      ply = 0;
      return(2);}
      

/*------------------------------*/
/* compare size of trees        */
/* searched, and if appropriate */
/* widen search parameters      */
/*------------------------------*/

total_nodesM2 = total_nodesM1;
total_nodesM1 = total_nodes;
total_nodes = s_nod - s_nod0;
units_this_iteration = s_ltentry;

if(units_this_iteration == units_last_iteration) {
   o_maxlit = 2 + o_maxlit;
   if(o_maxlit > 15) o_maxlit = 15;
   o_maxterm = 2*o_maxterm;
   //printf("X%d l%d ",o_maxterm,o_maxlit);
   OutPtr += sprintf(OutPtr,"X%d l%d ",o_maxterm,o_maxlit);
   if(o_maxterm > 100) o_maxterm = 100;}

else if((iter_depth> 3 && (total_nodes - total_nodesM1) == 
         (total_nodesM1 - total_nodesM2)) ||
	 total_nodes == 0) {
   o_maxlit = 2*o_maxlit;
   if(o_maxlit > 15) o_maxlit = 15;
   o_maxterm = 2*o_maxterm;
   //printf("X%d l%d ",o_maxterm,o_maxlit);
   OutPtr += sprintf(OutPtr,"X%d l%d ",o_maxterm,o_maxlit);
   if(o_maxterm > 100) o_maxterm = 100;}

units_last_iteration = units_this_iteration;

/*------------------------------*/
/* giving up search,....quit    */
/*------------------------------*/
 
   if(s == 3)  {
      //print0("\n\nGiving up the search; no proof found\n");
      ply = 0;
      return(2);}
   
   if(s_nod == 0 && iter_depth > 10) {
      //print0("\n\nNo proof found; check your clauses!\n");
      ply = 0;
      return(2);}}
	       
/*----------------------*/              
/* return with no proof */
/*----------------------*/
 
return(4);
}

/*-------------------------------------------------------------*/
/*      get_clause_stats()         .............. #21
/*
/*      determine the number of chars, vars, and lits
/*      in clause c
/*
/*      called by:     write_out_results    theo.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*
/*--------------------------------------------------------------*/

void get_clause_stats(c)
clause *c;
{
literal *l;
int lit_length;

if(c->c_nlit > maxl) maxl = c->c_nlit;
if(c->c_nv   > maxv) maxv = c->c_nv;
for(l = c->c_lit; l != 0; l = l->l_next) {
   if(l->l_next) {lit_length = ((int) l->l_next - (int) l)/4 - 4;}
   else {lit_length = ((int) c->c_end - (int) l)/4 - 4;}
   if(lit_length > maxx) {maxx = lit_length;}}

}

/*------------------------------------------------------------*/
/*      write_out_results()         .............. #22
/*
/*      write_out_results writes out the cl's that form the proof
/*      (if a proof found) and write statistics out. it writes
/*      the results on the screen and to a file "in-theorem.u".
/*
/*      called by:      main()              theo.c
/*
/*      calls:          print_clause()      tdump.c
/*                      get_clause_stats()  theo.c
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void write_out_results(proof_found)
int proof_found;
{
clause *c, *ci, *c_next;
int i,j,jj,k,m,tc;         
int total_search_time;     /* total search time                 */
int yes_inferred_cls;
int t_temp, t1_temp, t_power;
int x_temp;
char t1_char;
int leading_zeros;
int cls_id;
char *OutPtr2;
int used_nc_clauses, used_axioms, new_clauses, mm;
char *wpbh, *wpehP1, *wp_ptr;
int printAxiomsOnce, printNcOnce;
int weakened_clauses_used;
clause *c_pseudo1, *c_pseudo2;
int c_pseudo_indices;
int pvm_init_results;

/*--------------------------------------------------*/
/* set screen printing to print to a depth of 1000. */
/*--------------------------------------------------*/

//if(proof_found != 1 && weakened_ply == 1) print0("No Proof Obtained.\n\n");

o_print = 1000; o_step = -1; o_debug = 0; 
maxx = maxl = maxv = 0;

/*----------------------------------------------------------------*/
/* determine which of the inferred clauses were used in the proof.*/  
/* fix c->c_bits by adding LIST bit for inferred clauses that     */ 
/* should be printed. for each clause in the proof found in phases*/
/* 1,2, check LIST bit of parents                                 */
/*----------------------------------------------------------------*/

for(i = 1; i <= ply; i++) {
   c = clist[depth[i].d_tc];
   if(c->c_lpar) {
      ci = c->c_lpar;
      ci->c_bits |= LIST;} /* check lft par of clist[depth[i].d_tc*/
   if(c->c_rpar)  {                /* check if c is a factor      */
      ci = c->c_rpar;
      ci->c_bits |= LIST;}}
      
/*-------------------------------------------------------------*/
/* now check LIST bit of other inferred cl's to det parents    */
/*-------------------------------------------------------------*/

yes_inferred_cls = 0; 
for(i = base; i > org_base; i--) {
   c = clist[i];
   if(c->c_nlit == 0) c->c_bits |= LIST; /* null cls is inc */
   if((c->c_bits & LIST) == LIST) {
      for(m = 0; m < nextu; m++) {
         if(ulist[m]->c_X == c->c_X && ulist[m]->c_Y == c->c_Y) {
	   det_dependencies(ulist[m]);
	    break;}}
      yes_inferred_cls++;
      if(reinstate_parent[i + 1] != 0) {
	 clist[reinstate_parent[i + 1] - 1]->c_bits |= LIST;}
      if(c->c_lpar != 0) ci = c->c_lpar;
      ci->c_bits |= LIST;
      if(c->c_rpar != 0) ci = c->c_rpar;
      ci->c_bits |= LIST;}
   if(reinstate_parent[i] != 0) {
      clist[reinstate_parent[i] - 1]->c_bits |= LIST;}}
			 
good_inferred_cls = 0;        /* count # of inferred cls used   */

/*--------------------------------------------------------------*/
/* if there are axioms, print "axioms". otherwise print         */
/* "negated conclusion".                                        */
/*--------------------------------------------------------------*/
/* 
if(org_base + 1 - nc) {
   if(proof_found == 1) {
      if(weakened_ply == 0) {
         weakProofBeginsHere = OutPtr;
         OutPtr += sprint(OutPtr,"\nWEAKENED PROOF\n\nAxioms:\n");}
      else {
         OutPtr += sprint(OutPtr,"\nTHE PROOF\n");
	 if(weakProofBeginsHere != 0) {
	    for(OutPtrZ=weakProofBeginsHere;OutPtrZ<=weakProofEndsHere;OutPtrZ++){ 
	       *OutPtr++ = *OutPtrZ;}
            wpehP1 = weakProofEndsHere;
	    wpehP1++;
	    wpbh = weakProofBeginsHere;
	    for(wp_ptr = wpehP1; wp_ptr < OutPtr; wp_ptr++) {
	       *wpbh++ = *wp_ptr;
	       *wp_ptr = 0;}
            OutPtr = wpbh;}
	 OutPtr += sprintf(OutPtr,"\nAxioms:\n");}}
   else {OutPtr += sprintf(OutPtr,"\nTHE RESULTS (No Proof)\n\nAxioms:\n");}}

else {
   if(proof_found == 1) {
      if(weakened_ply == 0) {
         weakProofBeginsHere = OutPtr;
         OutPtr += sprintf(OutPtr,"\nWEAKENED PROOF\n\nNegated conclusion:\n");}
      else {
         OutPtr += sprintf(OutPtr,"\nTHE PROOF\n");
	 if(weakProofBeginsHere != 0) {
	    for(OutPtrZ=weakProofBeginsHere;OutPtrZ<=weakProofEndsHere;OutPtrZ++){ 
	       *OutPtr++ = *OutPtrZ;}
            wpehP1 = weakProofEndsHere;
	    wpehP1++;
	    wpbh = weakProofBeginsHere;
	    for(wp_ptr = wpehP1; wp_ptr < OutPtr; wp_ptr++) {
	       *wpbh++ = *wp_ptr;
	       *wp_ptr = 0;}
            OutPtr = wpbh;}
	 OutPtr += sprintf(OutPtr,"\nNegated conclusion:\n");}}
   else {OutPtr += sprintf(OutPtr,
        "\nTHE RESULTS (No Proof)\n\nNegated conclusion:\n");}}
*/
/*--------------------------------------------------------------*/
/* write out the base clauses and the clauses used from phase 0 */
/*--------------------------------------------------------------*/

tab_amount = 0;
questionableUsed = 0;

printAxiomsOnce = printNcOnce = 0;
for(i = 0; i <= base; i++) {
   if(org_base + 1 - nc - i) {
      if(printAxiomsOnce++ == 0) OutPtr += sprintf(OutPtr,"\nAxioms:\n");}
   else {if(printNcOnce++ == 0) 
                     OutPtr += sprintf(OutPtr,"\nNegated conclusion:\n");}
   c = clist[i];
   if(c->c_id <= org_base + 1) {
      if((c->c_bits & LIST) == LIST) {good_clauses[i] = 1;}}
   if(i <= org_base) {
      if(o_relevent == 0) {
         if((c->c_bits & LIST) == LIST) {get_clause_stats(c);}
         print_clause(c);}
      else if((c->c_bits & LIST) == LIST) {
         get_clause_stats(c);
         print_clause(c);}}
   else {
      if(o_debug > 0) {print_clause(c);}
      if(((c->c_bits & LIST) == LIST) || proof_found > 1) {
         get_clause_stats(c);
         print_clause(c);
	 if(questionable[c->c_id] == 0) {
	    copy_clause3(c);
	    if(c->c_nlit != 1) {
	       nextu--;
	       ulist[nextu++]->c_id &= ~LIST;}}
	 if(questionable[c->c_id] != 0) questionableUsed++;
	 good_inferred_cls++;}}         

   if(i == org_base) {
      OutPtr += sprintf(OutPtr,"\n---------------\n");}
   if(i == org_base && org_base != base && yes_inferred_cls) {
      weakened_clauses_used = 0;
      for(j = 0; xl[j][0] != 0; j++) {
         if(xl[j][13] != 0) {
	    weakened_clauses_used++;}}

      if(weakened_clauses_used != 0) {
         OutPtr += sprintf(OutPtr,"\nWeakened clauses used in proof:\n");}
      else {
         OutPtr += sprintf(OutPtr,"\nPhase 0 clauses used in proof:\n");}

      for(j = 0; xl[j][0] != 0; j++) {
         if(xl[j][13] < 0) {
            weakened_clauses_used++;
            for(jj = 0; xl[jj][0] != 0; jj++) {
               if(ulist[jj]->c_id == -xl[j][13]) { 
                  ulist[jj]->c_lpar = &c_pseudo1; 
		  ulist[jj]->c_rpar = &c_pseudo2;
		  c_pseudo1 = (xl[j][0] & 0x0000FFF0) >> 4;
		  c_pseudo2 = (xl[j][1] & 0x0000FFF0) >> 4;
	          print_clause(ulist[jj]);
	          break;}}}} 
      if(weakened_clauses_used != 0) {
         OutPtr += sprintf(OutPtr,"\nPhase 0 clauses used in proof:\n");}}}

/*----------------------------------*/
/* write out the clauses from       */
/* phases 1 and 2 used in proof     */
/* if proof not found in phase 0    */
/*----------------------------------*/
	
for(i = ply - 2; i >= 0; i--) {if(cl[i][3] != 0) break;}
begin_main_line = cl[i + 1][0] >> 16;
end_main_line   = cl[ply - 1][0] >> 16;

if(clist[base]->c_nlit != 0 && ply != 0) {
   OutPtr += sprintf(OutPtr,"\nPhases 1 and 2 clauses used in proof:\n");
   for(i = 1; i <= ply; i++)  {
      c = clist[depth[i].d_tc];
      if(c->c_id >= begin_main_line && c->c_id <= end_main_line - 1) {
         on_main_line = 1;}
      else on_main_line = 0;

      if(c->c_lpar->c_id != c->c_id - 1 && i != 1) {
         OutPtr += sprintf(OutPtr,"\n");}
      get_clause_stats(c);

   // tc = this_computer;this_computer = 0;
      print_clause(c);
   // this_computer = tc;

      if(questionable[c->c_id] != 0) questionableUsed++;
      if(questionable[c->c_id] == 0 && c->c_nlit != 0) {
         copy_clause3(c);
         nextu--;
         if(ulist[nextu]->c_nlit != 1) 
	       ulist[nextu]->c_bits = (ulist[nextu]->c_bits & ~LIST);
         nextu++;}

      if(on_main_line == 1 && questionable[c->c_id] == 0) {
         copy_clause3(c);}

      if(i > 1) {
	 cls_id = c->c_id;
	 if((questionable[cls_id] == 0) 
	    && questionable[cls_id + 1] == 1) {
	    copy_clause3(c);}}}}

statistics:
weakened_ply = 0;
for(i = 0; i < 2000; i++) {
   if(questionable[i] != 0) {weakened_ply = 1;}}

if(weakened_ply == 1) {
   reconsider_proof(proof_found);

   for(i = 0; i <2000; i++) {questionable[i] = 0;}

/*---------------------------------------------------------------*/
/* ANALYZE WEAKENED PROOF
/*---------------------------------------------------------------*/

   OutPtr += sprintf(OutPtr,
    "\n----------------------------------------------------------------------*/\n");
   OutPtr += sprintf(OutPtr,"ANALYSIS OF WEAKENED PROOF\n\n");

   OutPtr += sprintf(OutPtr,"Clause %d was weakened:  ",change_array[1][1]);
   print_clause(clist[change_array[1][1] - 1]);

   if(p_change_array[randy][6] ==3 || p_change_array[randy][6] == 6) {
      OutPtr += sprintf(OutPtr,"The clause was weakened by deleting literal %d\n\n",
         change_array[1][2]);}
	    
   else if(p_change_array[randy][6]==1 
                        || p_change_array[randy][6] == 4) {
      OutPtr += sprintf(OutPtr,
          "The clause was weakened by replacing a constant with a variable.\n\n");}

   else if(p_change_array[randy][6] == 2 
                        || p_change_array[randy][6] == 5) {
      OutPtr += sprintf(OutPtr,
          "The clause was weakened by replacing a function with a variable.\n\n");}

   OutPtr += sprintf(OutPtr,"Length of weakened proof is %d inferences.\n",good_inferred_cls + ply);

   OutPtr += sprintf(OutPtr,"Nodes searched for weakened proof is %d.\n",s_nod);

   if(good_inferred_cls + ply > 0 && good_inferred_cls + ply < 4) {
      OutPtr += sprintf(OutPtr,"Weakened theorem too easy!\n");}
  
   OutPtr += sprintf(OutPtr,
       "%d clauses in weakened proof are usable when proving given theorem.\n",
                     nextu);

   new_clauses = 0;
   for(mm = 0; mm < nextu; mm++) {
      if(ulist[mm]->c_bits & ACTIVE) {new_clauses++;}}
   OutPtr += sprintf(OutPtr,"Of these clauses, %d will be added to the base clauses.\n",new_clauses);

   used_axioms = 0;
   for(mm = 0; mm <= org_base - nc; mm++) {
      if(clist[mm]->c_bits & LIST) {used_axioms++;}}
   OutPtr += sprintf(OutPtr,
          "%d of given %d axioms used in weakened proof will be placed in the sos.\n",
      used_axioms,org_base - nc + 1);

   used_nc_clauses = 0;
   for(mm = org_base - nc + 1; mm <= org_base; mm++) {
      if(clist[mm]->c_bits & LIST) {used_nc_clauses++;}}
   OutPtr += sprintf(OutPtr,"%d of given %d clauses in sos were used in weakened proof.\n\n",
	     used_nc_clauses,nc);
   if(good_inferred_cls + ply > 3) {
      OutPtr += sprintf(OutPtr,"Proof of weakened version may be useful.\n\n");}
   else {
      OutPtr += sprintf(OutPtr,"Proof of weakened version not likely to be useful.\n\n");}

   OutPtr += sprintf(OutPtr,
  "\n------------------------------------------------------------------------*/\n");}

/*---------------------------------------------------------------*/
/* END OF ANALYSIS OF WEAKENED PROOF
/*---------------------------------------------------------------*/

done_analysis:
total_search_time = search_time - search_start_time; 

if(max_lit > 16) {
   theorem_out_name2[theorem_out_name2end++] = 'z';
   goto rename_theorem;}

if(satisfiable == 1) {
   theorem_out_name2[theorem_out_name2end++] = 's';
   goto rename_theorem;}

if(weakened_ply == 0 || excellent_weakened_proof == 1) {
   if(tough == 1 && unknown == 0) {
      theorem_out_name2[theorem_out_name2end++] = 'x';}

   if(tough == 1 && unknown == 1) {theorem_out_name2[theorem_out_name2end++] = 'X';}

   if(tough != 1) {theorem_out_name2[theorem_out_name2end++] = 48+theorem_rating/10;}

   if(proof_found == 1) {
      theorem_out_name2[theorem_out_name2end++] = 'L';
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned % 50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[this_computer/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[this_computer % 50];
      theorem_out_name2[theorem_out_name2end++]='.';}

   if(proof_found == 1 && excellent_weakened_proof == 0) {
      OutPtr += sprintf(OutPtr,"C%d: END OF PROOF\n",this_computer);}

   if(proof_found != 1) {
      theorem_out_name2[theorem_out_name2end++] = 't';
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned/50];
      theorem_out_name2[theorem_out_name2end++]=alphabet32[spawned % 50];}

   if(excellent_weakened_proof == 1) {
      theorem_out_name2[theorem_out_name2end++] = 'W';}

/*-----------------------------------------------------*/
/* convert time (in t_temp) to string which is then    */
/* suffixed to the file name.                          */
/*-----------------------------------------------------*/

if(simple_sol == 1) {phase1_time = 0;}

leading_zeros = 0;
t_temp = eliminate_time + phase1_time + eliminate_time + accumulated_time;
for(t_power = 100000; ; t_power = t_power/10) {
   if(t_temp >= t_power) {
      t1_temp = t_temp/t_power;
      t1_char = t1_temp + 48; 
      theorem_out_name2[theorem_out_name2end++] =  t1_char; 
      leading_zeros = 1;
      t_temp = t_temp - t1_temp*t_power;}
   else {
      if(leading_zeros == 1 || t_power == 1)  {
         t1_char = 48; 
         theorem_out_name2[theorem_out_name2end++] = t1_char;
         leading_zeros = 1;}}
   if(t_power == 1) {break;}}

if(proof_found == 1) {
   theorem_out_name2[theorem_out_name2end++] = '.';
   leading_zeros = 0;
   t_temp = good_inferred_cls + ply;
   for(t_power = 100000; ; t_power = t_power/10) {
      if(t_temp >= t_power) {
         t1_temp = t_temp/t_power;
         t1_char = t1_temp + 48; 
         theorem_out_name2[theorem_out_name2end++] =  t1_char; 
         leading_zeros = 1;
         t_temp = t_temp - t1_temp*t_power;}
      else {
         if(leading_zeros == 1 || t_power == 1)  {
            t1_char = 48; 
            theorem_out_name2[theorem_out_name2end++] = t1_char;
            leading_zeros = 1;}}
      if(t_power == 1) {break;}}}

theorem_out_name2[theorem_out_name2end++] = '.';

if(o_learn != 0) {
   if(p_change_array[randy][3] == -1) 
      theorem_out_name2[theorem_out_name2end++] = 'l';
   else if(p_change_array[randy][3] == 0) 
      theorem_out_name2[theorem_out_name2end++] = 'v';
   else if((p_change_array[randy][3] & ARGMASK) == 0) 
      theorem_out_name2[theorem_out_name2end++] = 'c';
   else theorem_out_name2[theorem_out_name2end++] = 'f';

   leading_zeros = 0;
   t_temp = randy;
   for(t_power = 100000; ; t_power = t_power/10) {
      if(t_temp >= t_power) {
         t1_temp = t_temp/t_power;
         t1_char = t1_temp + 48; 
         theorem_out_name2[theorem_out_name2end++] =  t1_char; 
         leading_zeros = 1;
         t_temp = t_temp - t1_temp*t_power;}
      else {
         if(leading_zeros == 1 || t_power == 1)  {
            t1_char = 48; 
            theorem_out_name2[theorem_out_name2end++] = t1_char;
            leading_zeros = 1;}}
      if(t_power == 1) {break;}}

   theorem_out_name2[theorem_out_name2end++] = '.';}

t_temp = s_nod + learn_nod;
leading_zeros = 0;
for(t_power = 10000000; ; t_power = t_power/10) {
   if(t_temp >= t_power) {
      t1_temp = t_temp/t_power;
      t1_char = t1_temp + 48; 
      theorem_out_name2[theorem_out_name2end++] =  t1_char; 
      leading_zeros = 1;
      t_temp = t_temp - t1_temp*t_power;}
   else {
      if(leading_zeros == 1 || t_power == 1)  {
         t1_char = 48; 
         theorem_out_name2[theorem_out_name2end++] = t1_char;
         leading_zeros = 1;}}
   if(t_power == 1) {break;}}

//if(c->c_nlit) {print0("\nNo Proof Obtained !!\n");}

/*------------------------------------------------------*/
/* if proof found, rename theorem to "theorem.Utime     */
/*------------------------------------------------------*/

rename_theorem:
rename(theorem_out_ptr,theorem_out_ptr2);}
	
/*------------------------------------------------------------------*/
/* first print info on time for elim, phase 1, phase 2, sum of 1,2  */
/*------------------------------------------------------------------*/
   
if(weakened_ply == 1) {
   if(nextu > 0 && questionableUsed != 0) {
      OutPtr += sprintf(OutPtr,"Valid inferences derived from given clauses:\n");
      for(i = 0; i < nextu; i++) {ulist[i]->c_bits |= BASE;}}}

/*----------------------------------*/
/* print statistics on the screen   */
/*----------------------------------*/    
	
s_nod_last = s_nod;
 
/*------------------------------------------------------------------*/
/* print proof length: number of inferred clauses that are used     */
/* plus the number of clauses found in the search (phases 1,2)      */
/*------------------------------------------------------------------*/
	
if(proof_found == 1 || excellent_weakened_proof == 1) {
   length_of_proof = good_inferred_cls + ply;}
	
/*--------------------------------------------------*/        
/* output to theorem.p (same as output to screen)   */
/*--------------------------------------------------*/
	
OutPtr += sprintf(OutPtr,"\nPHASE 0: %d s",eliminate_time);

if(simple_sol == 0) {
   OutPtr += sprintf(OutPtr," PHASE 1: %d s ",phase1_time);}
else OutPtr += sprintf(OutPtr," PHASE 1: -NA-  ");
	
if(extra_pass == 1) { 
   OutPtr += sprintf(OutPtr,
     " PHASE 2: %d s ",total_search_time - phase1_time);}
else {
   OutPtr += sprintf(OutPtr," PHASE 2: -NA-  ");}
OutPtr += sprintf(OutPtr,"ACCUM: %d",accumulated_time);
if(simple_sol == 0) {
   OutPtr += sprintf(OutPtr,
   " Total Time: %d s\n",accumulated_time + eliminate_time + total_search_time);}
else OutPtr += sprintf(OutPtr,"\n");

OutPtr += sprintf(OutPtr,"NOD: %-10d\t",s_nod);  
OutPtr += sprintf(OutPtr,"RES: %-10d\t",s_res);
OutPtr += sprintf(OutPtr,"FAC: %-10d\t",s_fac);       
OutPtr += sprintf(OutPtr,"T: %d  ",o_maxterm_tempy);
OutPtr += sprintf(OutPtr,"V: %d  ",o_maxvar_tempy);
OutPtr += sprintf(OutPtr,"L: %d\n",o_maxlit_tempy);        

OutPtr += sprintf(OutPtr,"CTE: %-10d\t",s_htentry_phase1);  
OutPtr += sprintf(OutPtr,"CTH: %-10d\t",s_hthit_phase1);      
OutPtr += sprintf(OutPtr,"CTF: %-10d\t",s_htfail_phase1);      
OutPtr += sprintf(OutPtr,"CSZ: %-10d\n",o_arraysize);     
 
OutPtr += sprintf(OutPtr,"UTE: %-10d\t",s_ltentry_phase1);      
OutPtr += sprintf(OutPtr,"UTH: %-10d\t",s_lthit_phase1);     
OutPtr += sprintf(OutPtr,"UTF: %-10d\t",s_ltfail_phase1);      
OutPtr += sprintf(OutPtr,"SBA: %-10d\n",sub_array_last1);

OutPtr += sprintf(OutPtr,
   "BAS: %-10d\tRED: %-10d\t",org_base + 1,clauses_alive);
if(proof_found == 1) {
   OutPtr += sprintf(OutPtr,
      "LEN: %d+%d=%d\t",
           good_inferred_cls,ply,length_of_proof);}
	
OutPtr += sprintf(OutPtr,"OPT:");
OutPtr += sprintf(OutPtr," k%d",o_elim);
OutPtr += sprintf(OutPtr," m%d",o_merge);
OutPtr += sprintf(OutPtr," z%d",o_nc);
OutPtr += sprintf(OutPtr," h%d",o_hash);
OutPtr += sprintf(OutPtr," n%d",o_extend);
OutPtr += sprintf(OutPtr," b%d",o_subsume);
OutPtr += sprintf(OutPtr," t%d/%d",o_batchtime,o_maxtime);
OutPtr += sprintf(OutPtr,"\n");

if(weakened_ply == 0 || excellent_weakened_proof || satisfiable) {

/*
   printf(
"C%d sending %d-node %d-step proof to theorem %d now\n",
           this_computer, s_nod+learn_nod,
	   proof_length,theorem_number);
   fflush(stdout);
*/

   OutPtr += sprintf(OutPtr,"$$");

   results[0]  = 'b';
   results[1] = '0' + (this_computer/10);
   results[2] = '0' + (this_computer % 10);
   results[3] = theorem_number;
   results[4] = excellent_weakened_proof;

   for(i = 0; i < 25; i++) {
      results[5 + i] = theorem_out_name2[i];}
/*
   for(i = 0; i < 300000; i++) {
      printf("%c",results[i]);
      if(results[i] == '$' && results[i - 1] == '$') break;}
*/
   pvm_init_results = pvm_initsend(PvmDataDefault);
   if(pvm_init_results < 0) {
      fflush(stdout);}

   pvm_pkbyte(results,300000,1);
   pvm_send(myparent, MPROOF);

/*
   printf("C%d sent %d-node proof to theorem %d now\n",
           this_computer, s_nod+learn_nod, theorem_number);
   fflush(stdout);*/
   }


if(proof_found == 1) {weakened_proof_remember = 1;}
}
 
/*-----------------------------------------------------------------*/
/*      main()                             .............. #23
/*
/*      The theorem prover enters here. We set the options specified
/*      by the  user, read in the clause system, prove the theorem,
/*      and write the proof out.
/*
/*      calls:   fatal()                        theo.c
/*               init_options()                 theo.c
/*               compile_clauses()              affc.c
/*               init_cl_array()                pass2.c
/*               init_clause_space()            theo.c
/*               init_stats()                   theo.c
/*               prepare_clauses()              theo.c
/*               init_random_hash_array()       theo.c
/*               hash_all_base_clauses()        theo.c
/*          (PHASE 0)
/*               eliminate()                    elim.c
/*          (PHASE 1)
/*               init_stats()                   theo.c
/*               iterative_deepening()          theo.c
/*               stack_hash_codes()             pass2.c
/*          (PHASE 2)
/*               iterative_deepening()          theo.c
/*               stack_hash_codes()             pass2.c
/*               rearrange_cl_array()           pass2.c
/*               search_proof_line()            pass2.c
/*          (PHASE 3)
/*               write_out_results()            theo.c
/*               free_space_close_files()       theo.c
/*-----------------------------------------------------------------*/
 
int main(argc,argv)
int argc;
char *argv[];

{
int zz;
int proof_found;
char decision_char;   /* char to decide whether to do another thm  */
int i,j,ii;
int send_once;
char *tt;
int o_maxtermZ, o_maxvarZ, o_maxlitZ;
int removed;
int this_computer_tt;

if(removed = remove("/u0/prof/newborn/octoslave.core") == 0) {
   printf("Killed /u0/prof/newborn/octoslave.core\n");
   fflush(stdout);}
if(removed = remove("~/octoslave.core") == 0) {
   printf("Killed ~/octoslave.core\n");
   fflush(stdout);}
if(removed = remove("./octoslave.core") == 0) {
   printf("Killed ./octoslave.core\n");
   fflush(stdout);}

theorem_number = 0;

mytid = pvm_mytid();
myparent = pvm_parent();

gethostname(buf,64);
pvm_initsend(PvmDataDefault);
pvm_pkstr(buf);
pvm_send(myparent, CNAME);


this_computer = 0;
o_batch = 0;
o_weakened_index = -1;
o_learn = 0;
batch_start_time = time_now();
batch_satisfiable = 0;
weakened_ply = 0;
weakened_clause = -1;

/*-----------------------------------------*/
/* change_array used in modify.c to keep   */
/* track of changes made to the given thm  */
/*-----------------------------------------*/

for(i = 0; i < 50; i++) {
   for(j = 0; j < 10; j++) {change_array[i][j] = 0;}}
change_level = 1;

/*-----------------------------------------*/
/* potential_change_array is used to store */
/* potential vars and constants  that may  */
/* be changed for a weakened version of thm*/
/*-----------------------------------------*/

for(i = 0; i < 2000; i++) {
   for(j = 0; j < 7; j++) {p_change_array[i][j] = 0;}}
p_change_number = 0;
excellent_weakened_proof = 0;
	
begin_proof:
proof_found  = 0;
killer_hashx = 0;
killer_hashy = 0;
magic_hashx  = 0;
magic_hashy  = 0;
extra_pass   = 0;
phase        = 0;
iter_depth   = 0;
equal_pred   = 0;
lit_in_hash  = 0;
alt_HX       = 0;
unknown      = 0;
tough        = 0;
equal_xyyzxz = -2;
theorem_rating = 0;
equal_clist_index = -2;
questionableUsed = 0;
slave_quit = 0;
init_options();

if(weakened_ply == 0 || excellent_weakened_proof == 1) {
   for(i = 0; i < MAX_BASE_CLAUSES; i++) {
      nextu = 0;
      good_clauses[i] = 0;}
   excellent_weakened_proof = 0;
   change_level = 1;
   satisfiable = 0;
   unknown = 0;
   tough = 0;
   theorem_rating = 0;
   learn_nod = 0;
   send_once = 0;
   OutPtr = &results[30];

/*------------------------------------------------------*/
/* attempt to determine theorem's name                  */
/*------------------------------------------------------*/


   compile_clauses();
   clauseToWeaken = -1;
   weakened_clause_count = -1;
   s_starttime = time_now();
   if(o_learn == 1) o_maxtime = o_batchtime + 5;
   else o_maxtime = 100000;}
else {
   learn_nod = s_nod;
   o_maxtime = 100000;}

/*--------------------------------------------------*/
/* init arrays for phase 0, then fix up clauses     */
/*--------------------------------------------------*/

restart_proof:
for(i = 0; i < 65536; i++) {
   for(j = 0; j < 6; j++) {sub_array[i][j] = 0;}}

for(i = 0; i < 2000; i++) {questionable[i] = 0;}

init_cl_array();          /* init cl list in pass2.c         */
init_search_space();      /* init search space               */
init_stats();             /* zero our statistics variables   */
prepare_clauses();
proof_found_remember = 0;
modify_clause();

/*---------------------------------------------------*/
/* initialize the random_hash_array and then hash    */
/* all base clauses                                  */
/*---------------------------------------------------*/

init_random_hash_array();
hash_all_base_clauses();

/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simplify input clauses if possible.                */
/* the array elim_inferences stores clauses generated */
/* in elim.                                           */
/*----------------------------------------------------*/

if(o_learn == 2) {
   o_learn_lits = o_maxlit;
   o_learn_char = o_maxterm;}

in_elim = 0;
if(o_elim) {
   simple_sol = 0;
   eliminate();}
in_elim = 1;
		
/*----------------------------------------*/
/* if proof found, don't carry out search */
/*----------------------------------------*/
 
if(o_elim == 2) simple_sol = 2;
	
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/

phase = 1;      
ply = 0;
lit_in_hash = 0;

search_start_time = time_now(); /* start time for search  */ 
 
init_stats();           /* zero our statistics variables  */
	
if(simple_sol == 0) {
   prepare_subarray();
   o_maxlitZ = o_maxlit;o_maxtermZ = o_maxterm;o_maxvarZ = o_maxvar;
   proof_found = iterative_deepening();
   lit_in_hash = 0;
   if(proof_found == 4) {
      o_maxlit++;
      o_maxterm = o_maxterm * 2;
      o_maxvar = 30;
      if(o_maxterm > 80) {o_maxterm = 80;}
//    printf("X%d v%d l%d ",o_maxterm,o_maxvar,o_maxlit);
      OutPtr += sprintf(OutPtr,
	     "X%d v%d l%d ",o_maxterm,o_maxvar,o_maxlit);
      proof_found = iterative_deepening();}
   phase1_time = search_time - search_start_time;
   sub_array_last1 = sub_array_last;
   solved_iteration = iter_depth;
   s_ltentry_phase1  = s_ltentry;
   s_htentry_phase1  = s_htentry;
   s_hthit_phase1    = s_hthit;
   s_lthit_phase1    = s_lthit;
   s_htfail_phase1   = s_htfail;
   s_ltfail_phase1   = s_ltfail;
   s_htelim_phase1   = s_htelim;
   s_subsume_phase1  = s_subsume;
   s_unit_res_phase1 = s_unit_res;
   s_unit_sub_phase1 = s_unit_sub;
   s_ltentry = s_htentry = s_hthit = s_lthit = 0;
   s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0;
   s_htelim = 0;
	
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/*----------------------------------------------------*/

   phase = 2;
   if(weakened_clause_count == 0) {
      printf("K%d: found %d-node %d-step proof exists after %d secs!\n",
                       this_computer,s_nod,ply,time_now() - s_starttime);
      fflush(stdout);
      results[0]  = 'a';
      results[1] = '0' + (this_computer/10);
      results[2] = '0' + (this_computer % 10);
      results[3] = theorem_number;
      results[4] = 0;
      pvm_initsend(PvmDataDefault);
      pvm_pkbyte(results,300000,1);
      pvm_send(myparent,MPROOF);}

   while(hash_code_active_number != 0 && proof_found < 2) {
      ply = 0; 
      for(i = 0; i < 65536; i++) {
         for(j = 0; j < 6; j++) {sub_array[i][j] = 0;}}
      prepare_subarray();
      extra_pass = 1;
      o_maxlit = o_maxlitZ;o_maxterm = o_maxtermZ;o_maxvar = o_maxvarZ;
      proof_found = iterative_deepening();
      lit_in_hash = 0;
      s_ltentry = s_htentry = s_hthit = s_lthit = 0;
      s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0;
      s_htelim = 0;}}

/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* check proof.  if proof is  ok, then print results  */
/*----------------------------------------------------*/

phase = 3;
if(simple_sol == 0 && proof_found < 2) {search_proof_line();}

mark_questionable();

write_results:
write_out_results(proof_found);

/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/

free_space_close_files();
 
/*----------------------------------------------*/
/* prove another theorem?                       */
/*----------------------------------------------*/

another_theorem:
if(weakened_ply == 1 && excellent_weakened_proof == 0) {
   goto begin_proof;}

ii = 1;
while(ii < 2) {ii = ii;}
}
