/*------------------------------------------------------------------*/
/*  tptpi.c
/*  copyrighted by Monty Newborn 2005
/*
/*  A routine that converts tptp-formated theorems to
/*  theo-formated theorems.  Note that all the functions
/*  in this file only call other functions in this file
/*  as shown below.  CompileTptp is called from main().
/*
/*  functions:
/*      ReadTptp()  ................. #1
/*      DetermineRating()............ #1a
/*      ConvertTptp3() .............. #2
/*      InsertChar()................. #3
/*      DeleteChar()................. #4
/*      CheckClauseType()............ #5
/*      CompareStrings()............. #6
/*      SetPtrToNextLine()........... #7
/*      AssignCostToClause()......... #8
/*      DumpClause()................. #9
/*      OutString()..................#10
/*      FinalSort()..................#11
/*      CompileTptp()................#12
/*      add_equality_axioms()........#13
/*      ConvertTptp2() ..............#13
/*      det_number_args()............#15
/*
/*  relation of functions:
/*      CompileTptp()
/*         ReadTptp()
/*         ConvertTptp()
/*            InsertChar()
/*            DeleteChar()
/*            SetPtrToNextLine()
/*            CheckClauseType()
/*               CompareStrings()
/*         AssignCostToClause()
/*         FinalSort()
/*            DumpClause()
/*            OutString()
/*
/*------------------------------------------------------------------*/


#include <stdio.h>
#include <strings.h>
#include <ctype.h>

#include "const.h"

/*-----------------------------------*/
/* externally defined functions      */
/*-----------------------------------*/

extern void print0();
extern void main_fof();
extern void add_skolem_axioms();

/*-----------------------------------*/
/* externally defined constants      */
/*-----------------------------------*/

extern char theorem_name[];
extern int slash_char;
extern FILE *infile;
extern char *OutPtr;
extern int satisfiable;
extern int unknown, tough;
extern int o_batch;
extern int o_tptp;
extern int weakened_proof_remember;
extern char ffresults[];
extern int cade_format;

/*-----------------------------------*/
/* internally defined functions      */
/*-----------------------------------*/

int ReadTptp();
void DetermineRating();
void ConvertTptp3();
void InsertChar();
void DeleteChar();
void CheckClauseType();
int  CompareStrings();
void SetPtrToNextLine();
void AssignCostToClause();
void DumpClause();
void OutString();
void FinalSort();
void CompileTptp();
void add_equality_axioms();
void add_pred_sub_axioms();
void ConvertTptp2();
int det_number_args();

/*----------------------------------*/
/* internally defined constants     */
/*----------------------------------*/

FILE *includeFile;

char tptpCodeBuffer[200000];
char *tptpCodePtr;
char *tptpCodeEndPtr;
char *tptpCodePtrTemp;
char *tptpCodePtrTemp2;
char Rating[40];
int RatingPtr;
char Sat[40];
int SatPtr;
int theorem_rating;
                             
char tptpIncludeCodeBuffer[200000];  /* handling of include directive */
char *currentIncludeCodeBuffer;
char *tptpCodePtrSave;                      /* saved position in the */
char *tptpCodeEndPtrSave;                   /* including file */
int  inInclude;                     /* to prevent recursive includes */
                                    /*-------------------------------*/
char includeName[128];
char includeAltName[128] = {'.','.','/','.','.','/'};

char theoCodeBuffer[200000];
char *theoCodePtr;

int ClauseCount;
int CharCount;

int ClauseType[MAX_BASE_CLAUSES];
int LitCount[MAX_BASE_CLAUSES];
int ClauseLength[MAX_BASE_CLAUSES];
int ClauseHead[MAX_BASE_CLAUSES];

char predicates[4000];
char *predicates_ptr;
char *predList[500];
int predArgs[500];

char functions[4000];
char *functions_ptr;
char *fcnList[500]; 
int fcnArgs[500];

int nextp, nextf;
int add_eq_ax;

/*-----------------------------------------------------------------*/
/*      ReadTptp()                   .......................#1
/*
/*  called by:   CompileTptp()       tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*-----------------------------------------------------------------*/

int ReadTptp()
{
int InChar, i;
char *p1, *p2, *p3;
static char foff[] = "fof(";
static char clausef[] = "input_clause(";

inInclude = 0;
CharCount = 0;
ClauseCount = 0;

for(i = 0; i < MAX_BASE_CLAUSES ; i++) {
   ClauseType[i] = LitCount[i] = ClauseLength[i] = ClauseHead[i] = 0;}

p1 = tptpCodeBuffer;
p2 = tptpIncludeCodeBuffer;
p3 = theoCodeBuffer;

for(i = 0; i < 200000; i++) {
   *p1 = *p2 = *p3 = 0;
   p1++; p2++; p3++;}

tptpCodePtr = tptpCodeBuffer;

/*------------------------------------------*/
/* read theorem                             */
/*------------------------------------------*/

for(;;) {
   *tptpCodePtr = getc(infile);
   CharCount++;
   if(*tptpCodePtr == (char) -1) {break;}
   tptpCodePtr++;}
tptpCodeEndPtr = tptpCodePtr;

currentIncludeCodeBuffer = tptpIncludeCodeBuffer;

/*------------------------------------------*/
/* if fof, return                           */
/*------------------------------------------*/

for(i = 0; i < (int) tptpCodeEndPtr - (int) tptpCodeBuffer; i++) {
   if(strncmp(foff,&tptpCodeBuffer[i],4) == 0) {
      return(1);}
   if(strncmp(clausef,&tptpCodeBuffer[i],13) == 0) {
      return(0);}}

return(0);
}

/*-------------------------------------------------------------------*/
/*  DetermineRating()
/*
/* Called from: CompileClauses()
/*
/*  Searches from rating of theorem and store for later printing
/*
/*-------------------------------------------------------------------*/

void DetermineRating()
{
int i, n, s;

for(i = 0; i < CharCount; i++) {
   if(tptpCodeBuffer[i] == 'R' &&
      tptpCodeBuffer[i+1] == 'a' &&
	 tptpCodeBuffer[i+2] == 't' &&
	 tptpCodeBuffer[i+3] == 'i' &&
	 tptpCodeBuffer[i+4] == 'n' &&
	 tptpCodeBuffer[i+5] == 'g') {
      RatingPtr = 0;
      while(tptpCodeBuffer[i] != '\n' && 
			     tptpCodeBuffer[i] != ',') {
	 Rating[RatingPtr++] = tptpCodeBuffer[i++];}
      if(tptpCodeBuffer[i] != '\n') Rating[RatingPtr] = ' ';
      break;}}


n = 0;
for(s = 0; s < 40 ; s++) {
   if(Rating[s] != '.') {continue;}
   n = n * 10 + (Rating[s - 1] - '0');
   n = n * 10 + (Rating[s + 1] - '0');
   n = n * 10 + (Rating[s + 2] - '0');
   break;}
		 
theorem_rating = n;

for(i = 0; i < CharCount; i++) {
   if(tptpCodeBuffer[i] == 'S' &&
      tptpCodeBuffer[i+1] == 't' &&
	 tptpCodeBuffer[i+2] == 'a' &&
	 tptpCodeBuffer[i+3] == 't' &&
	 tptpCodeBuffer[i+4] == 'u' &&
	 tptpCodeBuffer[i+5] == 's') {
      SatPtr = 0;
      while(tptpCodeBuffer[i] != '\n' && 
			     tptpCodeBuffer[i] != ',') {
	 Sat[SatPtr++] = tptpCodeBuffer[i++];}
      if(tptpCodeBuffer[i] != '\n') Sat[SatPtr] = '\n';
      break;}}

for(i = 0; i <= RatingPtr; i++) {
   OutPtr += sprintf(OutPtr,"%c",Rating[i]);
   printf("%c",Rating[i]);}
print0("   ");
for(i = 0; i <= SatPtr; i++) {
   OutPtr += sprintf(OutPtr,"%c",Sat[i]);
   printf("%c",Sat[i]);}
print0("\n");

for(i = 0; i < 30; i++) {
   if(Rating[i] == ' ' && Rating[i+1] == '1' && 
	    Rating[i + 2] == '.' && Rating[i + 3] == '0'){
      tough = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'o' && Sat[i + 1] == 'p' && Sat[i + 2] == 'e') {
      unknown = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'u' && Sat[i + 1] == 'n' && Sat[i + 2] == 'k') {
      unknown = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'S' && Sat[i + 1] == 'a' && Sat[i + 2] == 't') {
      satisfiable = 1;break;}}
}

/*-------------------------------------------------------------------*/
/*  ConvertTptp3()        ........................... #2
/*
/*  called by:  CompileTptp()      tptpi.c
/*
/*  calls:      InsertChar()       tptpi.c
/*              DeleteChar()       tptpi.c
/*              CheckClauseType()  tptpi.c
/*              SetPtrToNextLine() tptpi.c
/*
/*  returns:    Nil
/*
/*-------------------------------------------------------------------*/

void ConvertTptp3()
{
char *j, *jj, *jjj, *k, InChar;
char char1, char2, char3;
int bracket_count;
int char_count;
int i, ii, kk;
char *CommaPtr, *CRPtr;
int CRfound;
int parenthesis_count;
char *tptpU, *tptpZ, *tptpE, *tptpF, *tptpQ, *ptZ;
char *begin_clause;
int neg_pred;
char EQUAL[] = "equality)";
int to_get_name;
int pred_found;

tptpCodePtr = tptpCodeBuffer;
add_eq_ax = 0;

for(i = 0; i < 4000; i++) {predicates[i] = functions[i] = 0;}
for(i = 0; i < 500; i++) {
   predList[i] = fcnList[i] = 0;
   predArgs[i] = fcnArgs[i] = 0;}

predicates_ptr = predicates;
functions_ptr = functions;
nextp = nextf = 0;


/*----------------------------------*/
/* check for equality symbol        */
/*----------------------------------*/

for(tptpU = tptpCodePtr; *tptpU != (char) -1; tptpU++) {
   if(*tptpU == '=') {add_eq_ax = 1;goto STATE1;}}

STATE1:
if(*tptpCodePtr == 10) {    /* skip over carriage returns */
   SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == ' ') {    /* skip over blank spaces */
   DeleteChar(1);
   goto STATE1;}

if(*tptpCodePtr == '%') {   /* skip over comments; delete line */
   *tptpCodePtr = ';';
/* for(tptpE = tptpCodePtr; *tptpE != 10; tptpE++) {
      if(CompareStrings(tptpE,EQUAL,9)) {
         tptpF = tptpE;
         tptpF--;tptpF--;
	 if(*tptpF == '0') {
	    tptpF--;
	    if(*tptpF == ' ') {
               add_eq_ax = 0;break;}}}}
*/

   SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == 'i') { 
   if(tptpCodePtr[2] == 'c') { 

	      /*-------------------------------*/
              /* handling of include directive */
	      /*-------------------------------*/

      for( ;*tptpCodePtr != '\''; tptpCodePtr++); 
      tptpCodePtr++;

      for(i = 0; *tptpCodePtr!='\''; i++) {
	 includeAltName[i + 6] = includeName[i] = *tptpCodePtr++;}
      includeAltName[i + 6] = includeName[i] = 0;

      for(; *tptpCodePtr != '.'; tptpCodePtr++);
      tptpCodePtr++;

      for(ii = slash_char + 1; ii <= slash_char + 6 + i; ii++) {
	 theorem_name[ii] = includeAltName[ii - slash_char - 1];}

      if(inInclude) { 
         printf("Can't do recursive include...\n"); }
      else { 

/*
	 printf("slash_char=%d\n",slash_char);
         for(kk = 0; kk < 50; kk++) printf("%c",includeName[kk]);
	 printf("\n");
         for(kk = 0; kk < 50; kk++) printf("%c",theorem_name[kk]);
	 printf("\n");
*/

         if((includeFile = fopen(includeName,"r")) == NULL) {
	    if((includeFile = 
		  fopen(theorem_name,"r")) == NULL) {
               printf("Can't include required file\n");}}

         inInclude = 1;
         tptpCodePtrSave = tptpCodePtr;
         tptpCodeEndPtrSave = tptpCodeEndPtr;
         tptpCodePtr = currentIncludeCodeBuffer;

         for(;;) {
             InChar = getc(includeFile);  /* read "include" file */
             *tptpCodePtr = (char) InChar;
             if(InChar == (char) -1) {break;}
             tptpCodePtr++;} 
         fclose(includeFile);
         tptpCodeEndPtr = tptpCodePtr;
         tptpCodePtr = currentIncludeCodeBuffer;

	 for(tptpU = tptpCodePtr; *tptpU != (char) -1; tptpU++) {
	    if(*tptpU == '=') {add_eq_ax = 1;goto STATE1;}} 

         goto STATE1;}}}

if(*tptpCodePtr== 'c' && tptpCodePtr[1]== 'n' && tptpCodePtr[2]=='f') {

      ClauseHead[ClauseCount] = (int) tptpCodePtr; /* NEW CLAUSE */
      LitCount[ClauseCount++] = 1;
      CheckClauseType();
      InsertChar(';');
      SetPtrToNextLine();
      if(ClauseCount < 2) begin_clause = tptpCodePtr;
      goto STATE2;}

if(*tptpCodePtr == (char) -1) {   /*        -1 = EOF               */
   if(inInclude) {               /*-------------------------------*/
      inInclude = 0;
      currentIncludeCodeBuffer = tptpCodeEndPtr + 1;
      tptpCodePtr = tptpCodePtrSave;   /* restore position in the */
      tptpCodeEndPtr = tptpCodeEndPtrSave;      /* including file */
      goto STATE1;}
   return;}                      /*-------------------------------*/

STATE2:

/*--------------------------------------------------------------*/
/* delete spaces or line feeds.                                 */
/*--------------------------------------------------------------*/

if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
   DeleteChar(1);
   goto STATE2;} 

/*--------------------------------------------------------------*/
/* delete lines headed by an "%".                               */
/*--------------------------------------------------------------*/

if(*tptpCodePtr == '%') {
   *tptpCodePtr = ';';
   SetPtrToNextLine();
   goto STATE2;}

/*----------------------------------------------------------------*/
/* end of clause detected.                                        */
/*----------------------------------------------------------------*/

if(*tptpCodePtr == '.') {
   DeleteChar(1);
   tptpCodePtr++;
   ClauseLength[ClauseCount - 1] =
      (int) tptpCodePtr - (int) ClauseHead[ClauseCount - 1];
   goto STATE1;}

/*----------------------------------------------------------------*/
/* process literal.                                               */
/*----------------------------------------------------------------*/

STATE3:
if(*tptpCodePtr == '(' || *tptpCodePtr == ' ') {
   DeleteChar(1);
   goto STATE3;}

if(*tptpCodePtr == '~') {
   tptpCodePtr++;
   goto STATE3;}

if(*tptpCodePtr == '|') {
   tptpCodePtr++;
   goto STATE3;}

/*----------------------------------------------------------------*/
/* scan for equality predicate.                                   */
/*----------------------------------------------------------------*/

STATE3a:
to_get_name = 0;
LitCount[ClauseCount - 1]++;
tptpZ = tptpCodePtr;
while(*tptpCodePtr != '|' && *tptpCodePtr != '.') {
   tptpCodePtr++;
   if(*tptpCodePtr == '=') {
         *tptpCodePtr = ',';
	 tptpCodePtr--;
	 if(*tptpCodePtr == '!') {
	    DeleteChar(1);
	    neg_pred = 1;}
         else neg_pred = 0;
	    
         while(*tptpCodePtr != '.' && *tptpCodePtr != '|') {
	    tptpCodePtr++;}
         InsertChar(')');
	 tptpCodePtr = tptpZ;
         InsertChar('(');
         InsertChar('l');
         InsertChar('a');
         InsertChar('u');
         InsertChar('q');
         InsertChar('e');
	 if(neg_pred == 1) {InsertChar('~');}
	 break;}}

/*------------------------------------------------------------------*/
/* literal fixed up for equality predicate.                         */
/*------------------------------------------------------------------*/

STATE3b:

/*-----------------------------------------------------------------*/
/* fix constants by adding ()                                      */
/*-----------------------------------------------------------------*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '|' && *tptpCodePtr != '.') {
   if(isalnum(*tptpCodePtr) && isupper(*tptpCodePtr)) {
     tptpF = tptpCodePtr; 
      while(isalnum(*tptpCodePtr) || *tptpCodePtr == '_') {
         tptpCodePtr++;}
      continue;}

else if(isalnum(*tptpCodePtr) && (islower(*tptpCodePtr) || isdigit(*tptpCodePtr))) {
      tptpF = tptpZ = tptpCodePtr;
      tptpCodePtr++;
      char_count = 0;
      while(isalnum(*tptpCodePtr) || *tptpCodePtr == '_') {
         char_count++;
         tptpCodePtr++;}
      if(to_get_name == 0) {
         to_get_name++;
         predList[nextp++] = predicates_ptr; 
	 while(tptpF < tptpCodePtr) {
            *predicates_ptr++ = *tptpF++;}
         *predicates_ptr++ = ' ';

         tptpF = tptpCodePtr;
         tptpQ = tptpZ;
	 pred_found = 0;
	 for(i = 0 ; i < nextp - 1; i++) {
	    tptpZ = tptpQ; 
            ptZ = predList[i];
            while(*ptZ ==  *tptpZ && *ptZ != ' ') {
	       ptZ++;tptpZ++;}
	    if(*ptZ == ' ' && (*tptpZ == '(' || *tptpZ == ' ')) {
	       nextp--;
	       pred_found = 1;
	       predicates_ptr = predList[nextp];
	       tptpF++;}}
	 if(pred_found == 0) { 
	       predArgs[nextp - 1] = det_number_args(tptpF,1);}}
      else {
         if(*tptpCodePtr == ',' || *tptpCodePtr == ')' ||
            *tptpCodePtr == ' ' || *tptpCodePtr == (char) -2 || 
	    *tptpCodePtr == 10) {
            InsertChar('^');
	    if(char_count < 2) {
	       InsertChar('C');
	       tptpCodePtr++;}}
         else {
	    fcnList[nextf++] = functions_ptr;
            while(tptpF < tptpCodePtr) {
	       *functions_ptr++ = *tptpF++;}
            *functions_ptr++ = ' ';

            tptpF = tptpCodePtr;
            tptpQ = tptpZ;
	    pred_found = 0;
	    for(i = 0 ; i < nextf - 1; i++) {
	       tptpZ = tptpQ; 
               ptZ = fcnList[i];
               while(*ptZ ==  *tptpZ && *ptZ != ' ') {
	          ptZ++;tptpZ++;}
	       if(*ptZ == ' ' && (*tptpZ == '(' || *tptpZ == ' ')) {
	          nextf--;
		  pred_found = 1;
		  functions_ptr = fcnList[nextf];
	          tptpF++;}}
             if(pred_found == 0) {
	         fcnArgs[nextf - 1] = det_number_args(tptpF,2);}}

         continue;}}

   else tptpCodePtr++;}

/*-------------------------------------------------------*/
/* pass over clause and delete spaces and CRs            */
/*-------------------------------------------------------*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '|' && *tptpCodePtr != '.') {
   if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
      DeleteChar(1);}
   else tptpCodePtr++;}

STATE3c:

/*--------------------------------------------------------*/
/* if at end of clause, replace last two ")" with a       */
/* space, and then replace "." with a space.              */
/*--------------------------------------------------------*/

if(*tptpCodePtr == '.') {
   tptpZ = tptpCodePtr;
   parenthesis_count = 0;
   while(parenthesis_count != 2) {
      tptpCodePtr--;
      if(*tptpCodePtr == ')') {
         parenthesis_count++;
         DeleteChar(1);}}
   while(*tptpCodePtr != '.') tptpCodePtr++;
   *tptpCodePtr = ' ';
   tptpCodePtr++;
   goto STATE1;}

/*---------------------------------------------------------*/
/* replace "|" symbol with space, and move to second       */
/* character past blank before going to to STATE2.         */
/*---------------------------------------------------------*/

else if(*tptpCodePtr == '|') {
   InsertChar(' ');tptpCodePtr++;tptpCodePtr++;tptpCodePtr++;
   goto STATE2;}

}

/*------------------------------------------------------------------*/
/*  InsertChar()           ...................... #3
/*
/*  called by:  ConvertTptp()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*------------------------------------------------------------------*/

void InsertChar(s)
char s;
{
char *tptpCodeEndPtr1;
char *tptpCodeEndPtr2;

tptpCodeEndPtr++;
tptpCodeEndPtr1 = tptpCodeEndPtr;
tptpCodeEndPtr2 = tptpCodeEndPtr1;
tptpCodeEndPtr2--;

while(tptpCodeEndPtr2 >= tptpCodePtr) {
   *tptpCodeEndPtr1 = *tptpCodeEndPtr2;
   tptpCodeEndPtr2--;
   tptpCodeEndPtr1--;}

*tptpCodePtr = s;
return;
}

/*------------------------------------------------------------------*/
/*  DeleteChar()          ..................  #4
/*
/*  called by:   ConvertTptp()     tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*------------------------------------------------------------------*/

void DeleteChar(d)
int d;
{
int i;

for(i = d; i > 0; i--) {
   *tptpCodePtr++ = -2;}
return;
}

/*------------------------------------------------------------------*/
/*  CheckClauseType()        ................  #5
/*
/*  called by:    ConvertTptp()     tptpi.c
/*
/*  calls:        CompareStrings()  tptpi.c
/*
/*  returns:      Nil
/*
/*------------------------------------------------------------------*/

void CheckClauseType()
{
char *ccp;

ccp = tptpCodePtr;
while(*ccp != 10) {
   if(CompareStrings(ccp,"conjecture",10)) {
      ClauseType[ClauseCount - 1] = 2; 
      while(*ccp != '.') {
         if(CompareStrings(ccp,"--",2)) {
	    ClauseType[ClauseCount - 1] = 1;
	    return;}
         ccp++;}
      return;}

   if(CompareStrings(ccp,"hypothesis",10)) {
      ClauseType[ClauseCount - 1] = 3; return;}
   ccp++;}
ClauseType[ClauseCount - 1] = 4;
return;
}

/*------------------------------------------------------------------*/
/*  CompareStrings(s,t,n)       .................  #6
/*
/*  called by:  CheckClauseType()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    1 if strings are same, 0 otherwise
/*
/*------------------------------------------------------------------*/

int CompareStrings(s,t,n)
char s[],t[];
int n;
{
int i;

i = 0;
while(s[i] == t[i]) {
   i++;
   if(i == n) {return(1);}}
return(0);
}

/*------------------------------------------------------------------*/
/*  SetPtrToNextLine()      ....................  #7
/*
/*  called by:   ConvertTptp()      tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*------------------------------------------------------------------*/

void SetPtrToNextLine()
{
while(*tptpCodePtr != 10) {tptpCodePtr++;} 
tptpCodePtr++;
return;
}

/*------------------------------------------------------------------*/
/*  AssignCostToClause()            .......................  #8
/*
/*  called by:  CompileTptp()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*  assigns a cost to clauses so that they are first sorted with
/*  the negated conclusion at the end, then by the number of
/*  literals, then by the length of the clause in characters.
/*------------------------------------------------------------------*/

void AssignCostToClause()
{
int i;

for(i = 0; i < ClauseCount; i++) {
    LitCount[i] =
       100000*ClauseType[i] + 1000*LitCount[i] + ClauseLength[i];}
return;
}

/*------------------------------------------------------------------*/
/*  DumpClause()         .....................  #9
/*
/*  called by:  FinalSort()        tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*------------------------------------------------------------------*/

void DumpClause(Imax)
int Imax;
{
int CRctr;
char *cchar;

CRctr = 0;
tptpCodePtr = (char *) ClauseHead[Imax];

for( ; ; ) {
   if(*tptpCodePtr == (char) -2) {
      tptpCodePtr++;                     /* skip deleted chars */
      continue;}
   *theoCodePtr = *tptpCodePtr;
   if(*theoCodePtr == 10 && CRctr++ == 1) {break;}
   if(*theoCodePtr == '^') {        /* replace "^" with "()" */
      *theoCodePtr++ = '(';
      *theoCodePtr = ')';}
   if(*theoCodePtr == '|') {
      theoCodePtr++;             /* delete OR symbol   */
      *theoCodePtr = ' ';}
   tptpCodePtr++;
   theoCodePtr++;}
theoCodePtr++;
*theoCodePtr++ = 10;

/*
printf("DUMP CLUAE\n");
for(cchar = theoCodeBuffer; cchar < theoCodePtr;cchar++) {
   printf("%c",*cchar);}
printf("\nEND\n");
getchar();
*/

return;
}

/*------------------------------------------------------------------*/
/*  OutString()       ....................  #10
/*
/*  called by:  FinalSort()        tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*  Place a string "s" of length "Length" in the theoCodeBuffer 
/*------------------------------------------------------------------*/

void OutString(s,Length)
char s[];
int Length;
{
int i;

for(i = 0; i < Length ; i++) {*theoCodePtr++ = s[i];}
*theoCodePtr++ = 10;
return;
}

/*------------------------------------------------------------------*/
/*  FinalSort()          ..................... #11
/*
/*  called by:  CompileTptp()         tptpi.c
/*
/*  calls:      AssignCostToClause()  tptpi.c
/*
/*  returns:     Nil
/*        
/*------------------------------------------------------------------*/

void FinalSort()
{
int IndexMax, LitMax, i, NegConclusion;
char *OutputCodePtr;

theoCodePtr = theoCodeBuffer;
NegConclusion = 0;
for( ; ;   ) {
   IndexMax = 0;
   LitMax = 0;
   for(i = 0; i < ClauseCount; i++) {
      if(LitCount[i] > LitMax) {
         LitMax = LitCount[i];
         IndexMax = i;}}
   if(LitMax == 0) {goto finaldump;}
   LitCount[IndexMax] = 0;
   if(NegConclusion == 0 && LitMax < 300000) {
      NegConclusion++;
      OutString("negated_conclusion",18);}

   DumpClause(IndexMax);}

finaldump:
*theoCodePtr = -1;

return;
}
 
/*------------------------------------------------------------------*/ 
/*      compile_tptp()         .......................#12
/*
/*      called by:  main()                  theo.c
/*
/*      calls:      ReadTptp()              tptpi.c
/*                  ConvertTptp()           tptpi.c
/*                  AssignCostToClause()    tptpi.c
/*                  FinalSort()             tptpi.c
/*
/*      returns:    Nil
/*------------------------------------------------------------------*/
 
void CompileTptp(version)
int version;
{
int i;
char *z;
int zz;

if(version == 4) {
fof_thm:
   o_tptp = 4;
   main_fof();
   add_skolem_axioms();
   return;}

else {
   zz = ReadTptp();
      printf("FOF????\n");
   if(zz == 1) {
      printf("FOF!!!!\n");
      cade_format = 1;
      goto fof_thm;}
   DetermineRating();
   if(version == 2) ConvertTptp2();
   else if(version == 3) {
      ConvertTptp3();
      if(add_eq_ax) {
         add_equality_axioms();
         add_pred_sub_axioms();}}
   AssignCostToClause();
   FinalSort();}

/*
printf("\n\nAfter final sort\n\n");
for(z = theoCodeBuffer ; z < theoCodePtr ; z++) {
   printf("%c",*z);}

for(i=0; i < RatingPtr; i++) {
   printf("%c",Rating[i]);
   OutPtr += sprintf(OutPtr,"%c",Rating[i]);}
print0("   ");

for(i=0; i < SatPtr; i++) {
   printf("%c",Sat[i]);
   OutPtr += sprintf(OutPtr,"%c",Sat[i]);}
print0("\n");
getchar();
*/
return;
}


/*-------------------------------------------------------------------*/
/*  ConvertTptp2()        ........................... #2
/*
/*  called by:  CompileTptp()      tptpi.c
/*
/*  calls:      InsertChar()       tptpi.c
/*              DeleteChar()       tptpi.c
/*              CheckClauseType()  tptpi.c
/*              SetPtrToNextLine() tptpi.c
/*
/*  returns:    Nil
/*
/*-------------------------------------------------------------------*/

void ConvertTptp2()
{
char *j, *jj, *jjj, *k, InChar;
char char1, char2, char3;
int bracket_count;
int i, ii;
char *CommaPtr, *CRPtr;
int CRfound;

tptpCodePtr = tptpCodeBuffer;

STATE1:
if(*tptpCodePtr == 10) {    /* skip over carriage returns */
   SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == ' ') {    /* skip over carriage returns */
   DeleteChar(1);
   goto STATE1;}

if(*tptpCodePtr == '%') {   /* skip over comments */
   *tptpCodePtr = ';';
   SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == 'i') { 
   if(tptpCodePtr[2] == 'c') { 

	      /*-------------------------------*/
              /* handling of include directive */
	      /*-------------------------------*/

      for( ;*tptpCodePtr != '\''; tptpCodePtr++); 
      tptpCodePtr++;

      for(i = 0; *tptpCodePtr!='\''; i++) {
	 includeAltName[i + 6] = includeName[i] = *tptpCodePtr++;}
      includeAltName[i + 6] = includeName[i] = 0;

      for(; *tptpCodePtr != '.'; tptpCodePtr++);
      tptpCodePtr++;

      for(ii = slash_char + 1; ii <= slash_char + 6 + i; ii++) {
	 theorem_name[ii] = includeAltName[ii - slash_char - 1];}

      if(inInclude) { 
         printf("Can't do recursive include...\n"); }
      else { 
         if((includeFile = fopen(includeName,"r")) == NULL) {
	    if((includeFile = 
		  fopen(theorem_name,"r")) == NULL) {
               printf("Can't include required file\n");}}

         inInclude = 1;
         tptpCodePtrSave = tptpCodePtr;
         tptpCodeEndPtrSave = tptpCodeEndPtr;
         tptpCodePtr = currentIncludeCodeBuffer;

         for(;;) {
             InChar = getc(includeFile);/* read the included file */
             *tptpCodePtr = InChar;
             if(InChar == (char) -1) {break;}
             tptpCodePtr++;} 
          fclose(includeFile);
          tptpCodeEndPtr = tptpCodePtr;
          tptpCodePtr = currentIncludeCodeBuffer;
          goto STATE1;}}
   else {
      ClauseHead[ClauseCount] = (int) tptpCodePtr;
      LitCount[ClauseCount++] = 1;
      CheckClauseType();
      *tptpCodePtr++ = ';';
      DeleteChar(12);
      SetPtrToNextLine();
      goto STATE2;}}

if(*tptpCodePtr == (char) -1) {  /*        -1 = EOF               */
   if(inInclude) {               /*-------------------------------*/
      inInclude = 0;
      currentIncludeCodeBuffer = tptpCodeEndPtr + 1;
      tptpCodePtr = tptpCodePtrSave;   /* restore position in the */
      tptpCodeEndPtr = tptpCodeEndPtrSave;      /* including file */
      goto STATE1;}
   return;}                      /*-------------------------------*/

STATE2:
if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
   DeleteChar(1);
   goto STATE2;} 
   /*
   if(*tptpCodePtr == '.' || *tptpCodePtr == ',') {
      tptpCodePtr++;
      while(*tptpCodePtr == ' ') {
	 DeleteChar(1);}
      goto STATE1;}
   else {
      tptpCodePtr++;
      goto STATE2;}} */

/*if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
   DeleteChar(1);
   goto STATE1;}*/ 

if(*tptpCodePtr == '%') {
   while(*tptpCodePtr != 10) {
      DeleteChar(1);}
   goto STATE2;}

if(*tptpCodePtr == '.') {
   DeleteChar(1);
   tptpCodePtr++;
   ClauseLength[ClauseCount - 1] =
      (int) tptpCodePtr - (int) ClauseHead[ClauseCount - 1];
   goto STATE1;}

if(*tptpCodePtr == '[') {
   DeleteChar(1);
   goto STATE2;}

if(*tptpCodePtr ==  '+') {  /* 43 = + */
   tptpCodePtr++;
   if(*tptpCodePtr ==  '+' ) {
      tptpCodePtr--;
      DeleteChar(2);
      goto STATE2;}
   tptpCodePtr--;
   goto STATE2;}

if(*tptpCodePtr == '-') {
   tptpCodePtr++;
   if(*tptpCodePtr == '-') {
      tptpCodePtr--;
      DeleteChar(1);
      *tptpCodePtr++ = '~';
      goto STATE2;}
   tptpCodePtr--;
   goto STATE2;}

if(*tptpCodePtr == ',') {
   CRfound = 0;
   CommaPtr = tptpCodePtr;
   tptpCodePtr++;
   while(*tptpCodePtr == ' ' || *tptpCodePtr == 9) {
      DeleteChar(1);}
   if(*tptpCodePtr == '+' || *tptpCodePtr == '-') {
      *CommaPtr = -2;
      goto next_literal;}
   if(*tptpCodePtr != 10) {
      goto STATE2;}
   CRPtr = tptpCodePtr;
   CRfound = 1;
   tptpCodePtr++;
   while(*tptpCodePtr == 32) {
      DeleteChar(1);}
next_literal: 
   tptpCodePtrTemp = tptpCodePtr;
   tptpCodePtrTemp++;

   if(*tptpCodePtr == '+' && *tptpCodePtrTemp == '+') {
      LitCount[ClauseCount - 1]++;
      tptpCodePtr--;
      *tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '|';
      *tptpCodePtr++ = ' ';
      if(CRfound != 0) {*CRPtr = -2;}
      *CommaPtr = -2;
      goto STATE2;}
   if(*tptpCodePtr == '-' && *tptpCodePtrTemp == '-') {
      LitCount[ClauseCount - 1]++;
      tptpCodePtr--;
      *tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '|';
      *tptpCodePtr++ = '~';
      if(CRfound != 0) {*CRPtr = -2;}
      *CommaPtr = -2;
      goto STATE2;}
   *CRPtr = -2;
   goto STATE2;}

if(isalnum(*tptpCodePtr)) {
   j = tptpCodePtr;
   j--;
   k = tptpCodePtr;
   while(*tptpCodePtr != ')' && *tptpCodePtr != ',' &&
         *tptpCodePtr != ' ' && *tptpCodePtr != '(' &&
	 *tptpCodePtr != ']') {
      tptpCodePtr++;
      }
   if(*tptpCodePtr == ' ') {
      *tptpCodePtr = '(';
      bracket_count = 0;
      tptpCodePtr++;
      for(jj = tptpCodePtr ; ; jj++) {
	 if(*jj == '(') {bracket_count++;}
	 if(bracket_count == 0 && (*jj == ',' || *jj == ')')) {
	    jjj = tptpCodePtr;
	    tptpCodePtr = jj;
	    InsertChar(')'); 
	    tptpCodePtr = --jjj;
	    break;}
	 if(*jj == ')') {
	    bracket_count--;}}}

   if(*j == '+' || *j == '-' || isupper(*k) || *tptpCodePtr ==
           '(' || *tptpCodePtr == ' ') { 
      goto STATE2;}

   --tptpCodePtr;
   char1 = *tptpCodePtr;
   --tptpCodePtr;
   char2 = *tptpCodePtr;
   --tptpCodePtr;
   char3 = *tptpCodePtr;
   tptpCodePtr++;tptpCodePtr++;tptpCodePtr++;
   if(char2 == '(' || char2 == ',') {
      if(islower((unsigned char) char1)) {
	 InsertChar('C');
	 tptpCodePtr++;}}
   else {
      if(char3 == '(' || char3 == ',') {
         if(char2 == (char) -2) {
	    if(islower((unsigned char) char1) && char1 > 'o') {
	       InsertChar('C');
	       tptpCodePtr++;}}}}

   InsertChar('^');
   tptpCodePtr++;
   goto STATE2;}

if(*tptpCodePtr == ']') { 
   DeleteChar(2);
   goto STATE2;}
else {
   tptpCodePtr++;
   goto STATE2;}
}

/*---------------------------------------------------*/
/* add_equality_axioms()
/*
/* add equality axioms
/*
/*---------------------------------------------------*/

void add_equality_axioms()
{
char equal_d[] = "equal(x,x)";
char equal_sym[] = "~equal(x,y) | equal(y,x)";
char equal_trans[] = "~equal(x,y) | ~equal(y,z) | equal(x,z)";

char *eq_ptr;
int i;

ClauseHead[ClauseCount] = (int) tptpCodePtr;
LitCount[ClauseCount] = 1;
ClauseType[ClauseCount] = 4;
ClauseLength[ClauseCount++] = 10;

eq_ptr = equal_d;
for(i = 0; i < 10; i++) {
   *tptpCodePtr++ = *eq_ptr++;}
*tptpCodePtr++ = 10;
*tptpCodePtr++ = 10;

ClauseHead[ClauseCount] = (int) tptpCodePtr;
LitCount[ClauseCount] = 2;
ClauseType[ClauseCount] = 4;
ClauseLength[ClauseCount++] = 21;

eq_ptr = equal_sym;
for(i = 0; i < 24; i++) {
   *tptpCodePtr++ = *eq_ptr++;}
*tptpCodePtr++ = 10;
*tptpCodePtr++ = 10;

ClauseHead[ClauseCount] = (int) tptpCodePtr;
LitCount[ClauseCount] = 3;
ClauseType[ClauseCount] = 4;
ClauseLength[ClauseCount++] = 32;

eq_ptr = equal_trans;
for(i = 0; i < 38; i++) {
   *tptpCodePtr++ = *eq_ptr++;}
*tptpCodePtr++ = 10;
*tptpCodePtr++ = 10;

return;
}

/*-----------------------------------------------------*/
/* add_pred_sub_axioms()
/*
/*-----------------------------------------------------*/

void add_pred_sub_axioms()
{
int i, j, m;
char *tCodePtr;
char *pred_chars;
int equ_sub_ax;
char sub_char;

char varb[] = "xyzuvwtsrqponmlkj";
char varb1[] = "x,y,z,u,v,w,t,s,r,q,p,o,n,m,l,k,j";
char eql[] = "equal ";

pred_chars = predicates;
for(i = 0; i < nextp; i++) {
   pred_chars = predList[i];
   if(CompareStrings(eql,pred_chars,6)) continue; 

   equ_sub_ax = predArgs[i];
   sub_char = varb[equ_sub_ax];

   for(j = 0; j < equ_sub_ax; j++) {

      ClauseHead[ClauseCount] = (int) tptpCodePtr;
      LitCount[ClauseCount] = 3;
      ClauseType[ClauseCount] = 4;
      ClauseLength[ClauseCount++] = 20 + 4*equ_sub_ax;

      *tptpCodePtr++ = '~';
      pred_chars = predList[i];
      while(*pred_chars != ' ') *tptpCodePtr++ = *pred_chars++;
      *tptpCodePtr++ = '('; 
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         *tptpCodePtr++ = varb1[m];}
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = ' ';*tptpCodePtr++ = '|';*tptpCodePtr++ = ' ';

      *tptpCodePtr++ = '~';
      *tptpCodePtr++ = 'e';
      *tptpCodePtr++ = 'q';
      *tptpCodePtr++ = 'u';
      *tptpCodePtr++ = 'a';
      *tptpCodePtr++ = 'l';
      *tptpCodePtr++ = '('; 
      *tptpCodePtr++ = varb[j];
      *tptpCodePtr++ = ',';
      *tptpCodePtr++ = sub_char;
      *tptpCodePtr++ = ')'; 


      *tptpCodePtr++ = ' ';*tptpCodePtr++ = '|';*tptpCodePtr++ = ' ';

      pred_chars = predList[i];
      while(*pred_chars != ' ') *tptpCodePtr++ = *pred_chars++;
      *tptpCodePtr++ = '('; 
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         if(m == 2*j) *tptpCodePtr++ = sub_char;
         else *tptpCodePtr++ = varb1[m];}
            
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = 10;
      *tptpCodePtr++ = 10;}}

pred_chars = functions;
for(i = 0; i < nextf; i++) {
   pred_chars = fcnList[i];
   if(CompareStrings(eql,pred_chars,6)) continue; 

   equ_sub_ax = fcnArgs[i];
   sub_char = varb[equ_sub_ax];

   for(j = 0; j < equ_sub_ax; j++) {

      ClauseHead[ClauseCount] = (int) tptpCodePtr;
      LitCount[ClauseCount] = 2;
      ClauseType[ClauseCount] = 4;
      ClauseLength[ClauseCount++] = 20 + 4*equ_sub_ax;

      *tptpCodePtr++ = '~';
      *tptpCodePtr++ = 'e';
      *tptpCodePtr++ = 'q';
      *tptpCodePtr++ = 'u';
      *tptpCodePtr++ = 'a';
      *tptpCodePtr++ = 'l';
      *tptpCodePtr++ = '('; 
      *tptpCodePtr++ = 'x';
      *tptpCodePtr++ = ',';
      *tptpCodePtr++ = 'y';
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = ' ';*tptpCodePtr++ = '|';*tptpCodePtr++ = ' ';

      *tptpCodePtr++ = 'e';
      *tptpCodePtr++ = 'q';
      *tptpCodePtr++ = 'u';
      *tptpCodePtr++ = 'a';
      *tptpCodePtr++ = 'l';
      *tptpCodePtr++ = '('; 

      pred_chars = fcnList[i];
      while(*pred_chars != ' ') { *tptpCodePtr++ = *pred_chars++;}
      *tptpCodePtr++ = '(';
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         if(m != 2*j) *tptpCodePtr++ = varb1[m + 4];
         else *tptpCodePtr++ = 'x';}
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = ',';

      pred_chars = fcnList[i];
      while(*pred_chars != ' ') { *tptpCodePtr++ = *pred_chars++;}
      *tptpCodePtr++ = '(';
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         if(m != 2*j) *tptpCodePtr++ = varb1[m + 4];
         else *tptpCodePtr++ = 'y';}
      *tptpCodePtr++ = ')'; 
     

      *tptpCodePtr++ = ')'; 
      *tptpCodePtr++ = 10;
      *tptpCodePtr++ = 10;}}

}

/*------------------------------------------------------*/
/* det_number_args()
/*
/*------------------------------------------------------*/

int det_number_args(string_check,p_or_f)
char *string_check;
int p_or_f;
{
int arg_count;
char *scan_char;
int brackets;
int bracket_found;

arg_count = 0;
brackets = 0;
bracket_found = 0;
scan_char = string_check;
while(islower(*scan_char) || *scan_char == '_') {
   scan_char++;}

while(*scan_char != '|' || *scan_char != '.') {
   if(*scan_char == ')') brackets--;
   if(*scan_char == '(') { 
      brackets++;
      bracket_found = 1;}
   if(*scan_char == ',' && brackets <= 1) arg_count++;
   if(*scan_char == ')' && brackets <= 0) {
      return(arg_count + bracket_found);} 
   scan_char++;}

}

