/*-------------------------------------------------------*/
/*  tptpj.c
/*
/*
/*  A routine that converts tptp-formated theorems to
/*  theo-formated theorems.  Note that all the functions
/*  in this file only call other functions in this file
/*  as shown below.  CompileFOF is called from main().
/*
/*  functions:
/*      ReadFOF() . ................. #1
/*      fof_DetermineRating()........ #1a
/*      ConvertFOF() ................ #2
/*      fof_InsertChar()............. #3
/*      fof_DeleteChar()............. #4
/*      fof_CheckClauseType()........ #5
/*      fof_CompareStrings()......... #6
/*      fof_SetPtrToNextLine()....... #7
/*      fof_AssignCostToClause()..... #8
/*      fof_DumpClause()............. #9
/*      fof_OutString()..............#10
/*      fof_FinalSort()..............#11
/*      CompileFOF().................#12
/*
/*  relation of functions:
/*      CompileFOF()
/*         ReadFOF()
/*         ConvertFOF()
/*            fof_InsertChar()
/*            fof_DeleteChar()
/*            fof_SetPtrToNextLine()
/*            fof_CheckClauseType()
/*               fof_CompareStrings()
/*         fof_AssignCostToClause()
/*         fof_FinalSort()
/*            fof_DumpClause()
/*            fof_OutString()
/*
/*--------------------------------------------------------*/

#include <stdio.h>
#include <strings.h>
#include <ctype.h>

#include "compwff.h"
#include "clause.h"
#include "const.h"

/*-----------------------------------*/
/* externally defined functions      */
/*-----------------------------------*/

extern void print0();

/*-----------------------------------*/
/* externally defined constants      */
/*-----------------------------------*/

extern char theoCodeBuffer[200000];
extern char *theoCodePtr;
extern int o_tptp;

extern char theorem_name[];
extern int slash_char;
extern FILE *infile;
extern char *OutPtr;
extern int satisfiable, unknown, tough;
extern int o_batch;
extern int ffskolemid;
extern char ffskolem_list[2000];
extern char ffclause_form[200000], *ffclause_form_ptr;

extern char ffresults[200000], *ffOutPtr;

/*-----------------------------------*/
/* internally defined functions      */
/*-----------------------------------*/

void ReadFOF();
void fof_DetermineRating();
void ConvertFOF();
void fof_add_equality_axioms();
void add_equality_sub_axioms();
int fof_det_number_args();
void fof_InsertChar();
void fof_DeleteChar();
void fof_CheckClauseType();
int  fof_CompareStrings();
void fof_SetPtrToNextLine();
void fof_AssignCostToClause();
void fof_DumpClause();
void fof_OutString();
void fof_FinalSort();
void CompileFOF();
void add_skolem_axioms();
void list_preds_fxns();

/*----------------------------------*/
/* internally defined constants     */
/*----------------------------------*/

extern FILE *includeFile;

extern char tptpCodeBuffer[200000];
extern char *tptpCodePtr;
extern char *tptpCodeEndPtr;
extern char *tptpCodePtrTemp;
extern char *tptpCodePtrTemp2;
extern char *tptpCodePtrSave;                      /* saved position in the */
extern char *tptpCodeEndPtrSave;                   /* including file */
extern char Rating[40];
extern int RatingPtr;
extern char Sat[40];
extern int SatPtr;
extern int theorem_rating;
                             
extern char tptpIncludeCodeBuffer[200000]; /* handling of include directive */
extern char *currentIncludeCodeBuffer;
extern int  inInclude;                     /* to prevent recursive includes */
                                    /*-------------------------------*/
char ffincludeName[128];
char ffincludeAltName[128] = {'.','.','/','.','.','/'};

extern int ClauseCount;
extern int CharCount;

int ffClauseType[MAX_BASE_CLAUSES];
int ffLitCount[MAX_BASE_CLAUSES];
int ffClauseLength[MAX_BASE_CLAUSES];
int ffClauseHead[MAX_BASE_CLAUSES];

extern char predicates[4000];
extern char *predicates_ptr;
extern char *predList[500];
extern int predArgs[500];

extern char functions[4000];
extern char *functions_ptr;
extern char *fcnList[500];
extern int fcnArgs[500];

extern int nextp, nextf;
extern int add_eq_ax;

extern int cade_format;

/*-------------------------------------------------*/
/*  ReadFOF()             .......................#1
/*
/*  called by:   CompileFOF()       tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*-------------------------------------------------*/

void ReadFOF()
{
int InChar, i;
char *p1, *p2, *p3;

for(i = 0; i < MAX_BASE_CLAUSES ; i++) {
   ffClauseType[i] = 0;
   ffLitCount[i] = ffClauseLength[i] = ffClauseHead[i] = 0;}

if(cade_format == 1) return;

inInclude = 0;
CharCount = 0;
ClauseCount = 0;

p1 = tptpCodeBuffer;
p2 = tptpIncludeCodeBuffer;
p3 = theoCodeBuffer;

for(i = 0; i < 200000; i++) {
   *p1 = *p2 = *p3 = ' ';
   p1++; p2++; p3++;}

tptpCodePtr = tptpCodeBuffer;

/*------------------------------------------*/
/* read theorem                             */
/*------------------------------------------*/

for(;;) {
   *tptpCodePtr = getc(infile);
   CharCount++;
   if(*tptpCodePtr == (char) -1) {break;}
   tptpCodePtr++;}
tptpCodeEndPtr = tptpCodePtr;

currentIncludeCodeBuffer = tptpIncludeCodeBuffer;

return;
}

/*--------------------------------------------------*/
/* fof_DetermineRating()
/*
/* Called from: CompileClauses()
/*
/* Searches for theorem rating, store for later printing
/*
/*---------------------------------------------------*/

void fof_DetermineRating()
{
int i, n, s;

for(i = 0; i < CharCount; i++) {
   if(tptpCodeBuffer[i] == 'R' &&
      tptpCodeBuffer[i+1] == 'a' &&
	 tptpCodeBuffer[i+2] == 't' &&
	 tptpCodeBuffer[i+3] == 'i' &&
	 tptpCodeBuffer[i+4] == 'n' &&
	 tptpCodeBuffer[i+5] == 'g') {
      RatingPtr = 0;
      while(tptpCodeBuffer[i] != '\n' && 
		     tptpCodeBuffer[i] != ',') {
	 Rating[RatingPtr++] = tptpCodeBuffer[i++];}
      if(tptpCodeBuffer[i] != '\n') {
         Rating[RatingPtr] = ' ';}
      break;}}

n = 0;
for(s = 0; s < 40 ; s++) {
   if(Rating[s] != '.') {continue;}
   n = n * 10 + (Rating[s - 1] - '0');
   n = n * 10 + (Rating[s + 1] - '0');
   n = n * 10 + (Rating[s + 2] - '0');
   break;}
		 
theorem_rating = n;

for(i = 0; i < CharCount; i++) {
   if(tptpCodeBuffer[i] == 'S' &&
      tptpCodeBuffer[i+1] == 't' &&
	 tptpCodeBuffer[i+2] == 'a' &&
	 tptpCodeBuffer[i+3] == 't' &&
	 tptpCodeBuffer[i+4] == 'u' &&
	 tptpCodeBuffer[i+5] == 's') {
      SatPtr = 0;
      while(tptpCodeBuffer[i] != '\n' && 
		     tptpCodeBuffer[i] != ',') {
	 Sat[SatPtr++] = tptpCodeBuffer[i++];}
      if(tptpCodeBuffer[i] != '\n') Sat[SatPtr] = '\n';
      break;}}

OutPtr += sprintf(OutPtr,";");
for(i = 0; i <= RatingPtr; i++) {
   OutPtr += sprintf(OutPtr,"%c",Rating[i]);
   printf("%c",Rating[i]);}

OutPtr += sprintf(OutPtr,"  ");
print0("   ",0);

for(i = 0; i <= SatPtr; i++) {
   OutPtr += sprintf(OutPtr,"%c",Sat[i]);
   printf("%c",Sat[i]);}
print0("\n\n",0);

for(i = 0; i < 30; i++) {
   if(Rating[i] == ' ' && Rating[i+1] == '1' && 
	    Rating[i + 2] == '.' && Rating[i + 3] == '0'){
      tough = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'o' && Sat[i+1] == 'p' && Sat[i+2] == 'e') {
      unknown = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'u' && Sat[i+1] == 'n' && Sat[i+2] == 'k') {
      unknown = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == ' ' && Sat[i+1] == 's' && Sat[i+2] == 'a') {
      satisfiable = 1;break;}}
OutPtr += sprintf(OutPtr,"\n");

}

/*-----------------------------------------------------*/
/*  ConvertFOF()        ........................... #2
/*
/*  called by:  CompileFOF()       tptpi.c
/*
/*  calls:      fof_InsertChar()    tptpi.c
/*              fof_DeleteChar()    tptpi.c
/*              fof_CheckClauseType()   tptpi.c
/*              fof_SetPtrToNextLine()  tptpi.c
/*
/*  returns:    Nil
/*
/*------------------------------------------------------*/

void ConvertFOF()
{
char *j, *jj, *jjj, *k, InChar;
char char1, char2, char3;
int bracket_count;
int i, ii, m;
char *CommaPtr, *CRPtr;
int CRfound;
int in_fxn;
int delete_count;
int parenthesis_count;
char *tptpU, *tptpZ, *tptpE, *tptpF, *tptpM, *tptpN, *tptpQ, *ptZ;
char *begin_clause;
int neg_pred;
char EQUAL[] = "equality)";
int to_get_name;
int pred_found;
int total_lr_brackets, brackets;
int nest_depth, bracket_pairs[64];
int lbracket_count;
char left_side[500],right_side[500];
int length_ls,length_rs;

tptpCodePtr = tptpCodeBuffer;
add_eq_ax = 0;

for(i = 0; i < 4000; i++) {predicates[i] = functions[i] = 0;}
for(i = 0; i < 500; i++) {
   predList[i] = fcnList[i] = 0;
   predArgs[i] = fcnArgs[i] = 0;}

predicates_ptr = predicates;
functions_ptr = functions;
nextp = nextf = 0;

/*----------------------------------*/
/* check for equality symbol        */
/*----------------------------------*/

for(tptpU = tptpCodePtr; *tptpU != (char) -1; tptpU++) {
   if(*tptpU == '=') {
       tptpM = tptpU;
       tptpN = tptpU;
       tptpM--;tptpN++;
       if(*tptpM == ' ' && *tptpN == ' ') {
         add_eq_ax = 1;goto STATE1;}}}

STATE1:

in_fxn = 0;
if(*tptpCodePtr == 10) {    /* skip over carriage returns */
   fof_SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == ' ') {   /* delete blank characters */
   fof_DeleteChar(1);
   goto STATE1;}

if(*tptpCodePtr == '%') {   /* comment: replace with semicolon */
   *tptpCodePtr = ';';
/* for(tptpE = tptpCodePtr; *tptpE != 10; tptpE++) {
       if(fof_CompareStrings(tptpE,EQUAL,9)) {
          tptpF = tptpE;
          tptpF--;tptpF--;
          if(*tptpF == '0') {
             tptpF--;
             if(*tptpF == ' ') {
                add_eq_ax = 0;break;}}}}
*/
   for(tptpE = tptpCodePtr; *tptpE != 10; tptpE++) {
      *tptpE = -2;}

   fof_SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == 'i') { 
   tptpZ = tptpCodePtr;
   if(tptpCodePtr[2] == 'c') { 

/*-------------------------------*/
/* handling of include directive */
/*-------------------------------*/

      for( ;*tptpCodePtr != '\''; tptpCodePtr++); 
      tptpCodePtr++;

      for(i = 0; *tptpCodePtr!='\''; i++) {
	 ffincludeAltName[i + 6] = ffincludeName[i] = *tptpCodePtr++;}
      ffincludeAltName[i + 6] = ffincludeName[i] = 0;

      for(; *tptpCodePtr != '.'; tptpCodePtr++);
      tptpCodePtr++;

      for(ii = slash_char + 1; ii <= slash_char + 6 + i; ii++) {
	 theorem_name[ii] = ffincludeAltName[ii - slash_char - 1];}

      if(inInclude) { 
         printf("Can't do recursive include..."); }
      else { 
         if((includeFile = fopen(ffincludeName,"r")) == NULL) {
	    if((includeFile = 
		  fopen(theorem_name,"r")) == NULL) {
               printf("Can't include required file");}}

         inInclude = 1;
         tptpCodePtrSave = tptpCodePtr;
         tptpCodeEndPtrSave = tptpCodeEndPtr;
         tptpCodePtr = currentIncludeCodeBuffer;

         for(;;) {
             InChar = getc(includeFile);/* read included file */
             *tptpCodePtr = InChar;
             if(InChar == (char) -1) {break;}
             tptpCodePtr++;} 
          fclose(includeFile);
          tptpCodeEndPtr = tptpCodePtr;
          tptpCodePtr = currentIncludeCodeBuffer;
          while(*tptpZ != 10) {
             *tptpZ++ = -2;}

          for(tptpU = tptpCodePtr; *tptpU != (char) -1; tptpU++) {
             if(*tptpU == '=') {
             tptpM = tptpU;
             tptpN = tptpU;
             tptpM--;tptpN++;
             if(*tptpM == ' ' && *tptpN == ' ') {
                add_eq_ax = 1;goto STATE1;}}}

	  goto STATE1;}}}

if(*tptpCodePtr== 'f' && tptpCodePtr[1]== 'o' && tptpCodePtr[2]=='f') {
      ffClauseHead[ClauseCount] = (int) tptpCodePtr; /* NEW CLAUSE */
      ffLitCount[ClauseCount++] = 1;
      fof_CheckClauseType();
      fof_InsertChar(';');
      lbracket_count = 0;
      for(tptpE = tptpCodePtr; *tptpE != 10; tptpE++) {
         if(*tptpE == '(') lbracket_count++; 
         *tptpE = -2;}
      fof_SetPtrToNextLine();
      if(ClauseCount < 2) begin_clause = tptpCodePtr;
      goto STATE2;}

if(*tptpCodePtr == (char) -1) {  /*        -1 = EOF       */
   if(inInclude) {               /*-----------------------*/
      inInclude = 0;
      currentIncludeCodeBuffer = tptpCodeEndPtr + 1;
      tptpCodePtr = tptpCodePtrSave; /* restore position in */
      tptpCodeEndPtr = tptpCodeEndPtrSave; /*including file */
      goto STATE1;}
   return;}                      /*-------------------------*/

STATE2:
printf("State2: %c\n",*tptpCodePtr);

/*--------------------------------------------------------------*/
/* delete spaces or line feeds.                                 */
/*--------------------------------------------------------------*/

if(lbracket_count == 2) {
   lbracket_count++;
      *tptpCodePtr = '(';}

if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
   fof_DeleteChar(1);
   goto STATE2;} 

/*--------------------------------------------------------------*/
/* delete lines headed by an "%".                               */ 
/*--------------------------------------------------------------*/

if(*tptpCodePtr == '%') {
   while(*tptpCodePtr != 10) {
      fof_DeleteChar(1);}
   goto STATE2;}

/*--------------------------------------------------------------*/
/* end of clause detected?                                      */ 
/*--------------------------------------------------------------*/

if(*tptpCodePtr == '.') {
   tptpCodePtr++;
   ffClauseLength[ClauseCount - 1] =
      (int) tptpCodePtr - (int) ffClauseHead[ClauseCount-1];
      tptpCodePtr++;
   goto STATE1;}

/*--------------------------------------------------------------*/
/* fix equality predicate.                                      */
/*--------------------------------------------------------------*/

STATE3a:
tptpZ = tptpCodePtr;
while(*tptpCodePtr != '.') {

   if(*tptpCodePtr == ' ') {
     tptpCodePtr--;
     if(*tptpCodePtr == ' ') {
        *tptpCodePtr = -2;
	*tptpCodePtr++;
	continue;}
      else tptpCodePtr++;}
   tptpCodePtr++;
   if(*tptpCodePtr != '=') continue;
   tptpQ = tptpCodePtr;
   tptpQ--;
   if(*tptpQ == '<') continue;
   tptpQ++;tptpQ++;
   if(*tptpQ == '>') continue;
   *tptpCodePtr = ',';
   tptpCodePtr--;
   tptpQ = tptpCodePtr;
   if(*tptpCodePtr == '!') {
      tptpCodePtr--;
      fof_DeleteChar(2);
      neg_pred = 1;}
   else {
      neg_pred = 0;
      fof_DeleteChar(1);}

   brackets = 0;
   tptpE = tptpCodePtr;
   tptpCodePtr++;
   fof_DeleteChar(1);   
   while(!islower(*tptpCodePtr) && !isupper(*tptpCodePtr)) {
      tptpCodePtr++;}
   while(islower(*tptpCodePtr) || isupper(*tptpCodePtr) || *tptpCodePtr == '_' 
       || *tptpCodePtr == ')' || *tptpCodePtr == '(' || *tptpCodePtr == ',') {
      tptpCodePtr++;}
   while(*tptpCodePtr != ' ' && *tptpCodePtr != 10) {
      tptpCodePtr++;}
   fof_InsertChar(')');
   fof_InsertChar(' ');
   fof_InsertChar(')');

   tptpCodePtr = tptpQ;
   while(*tptpCodePtr == ' ' || *tptpCodePtr == (char) -2) {
      fof_DeleteChar(1);tptpCodePtr--;tptpCodePtr--;}
   while(isalnum(*tptpCodePtr) ||
      *tptpCodePtr == '_' || *tptpCodePtr == '(' ||
      *tptpCodePtr == ')' || *tptpCodePtr == ',')  
      tptpCodePtr--;
   *tptpCodePtr = -2;
   fof_InsertChar('(');
   fof_InsertChar('l');
   fof_InsertChar('a');
   fof_InsertChar('u');
   fof_InsertChar('q');
   fof_InsertChar('e');
   if(neg_pred == 1) {fof_InsertChar('~');}
   fof_InsertChar(' ');
   fof_InsertChar('(');
   fof_InsertChar(' ');}

/*---------------------------------------------------------------*/
/* change "(" and ")" to "{" and "}" as appropriate              */
/*---------------------------------------------------------------*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   if(*tptpCodePtr == '(') {*tptpCodePtr++ = '{';continue;}
   if(*tptpCodePtr == ')') {*tptpCodePtr++ = '}';continue;}
   tptpCodePtr++;}

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   if(isalnum(*tptpCodePtr)  || *tptpCodePtr == '_') {
      tptpCodePtr++;

      brackets = 0;
      while(isalnum(*tptpCodePtr) || *tptpCodePtr == '_') {
         tptpCodePtr++;}
   
      if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
         tptpCodePtr++;
         continue;}
      else {
next_char:
         if(*tptpCodePtr == '{') {
	    *tptpCodePtr = '(';
            brackets++;}
         else if(*tptpCodePtr == '}') {
	    *tptpCodePtr = ')';
            brackets--;}}
         if(brackets == 0) {continue;}
	 else tptpCodePtr++;
	 goto next_char;}
   else tptpCodePtr++;}


/*---------------------------------------------------------------*/
/* change "!" to "@" and "?" to "!"                              */
/*---------------------------------------------------------------*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') { 
   if(*tptpCodePtr == '!') {
      *tptpCodePtr++ = '@';
      while(*tptpCodePtr != ':') {
         if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) *tptpCodePtr = -2;
	 else if(*tptpCodePtr == '[') *tptpCodePtr = -2;
	 else if(*tptpCodePtr == ']') *tptpCodePtr = -2;
	 tptpCodePtr++;}}
      
   else if(*tptpCodePtr == '?') {
      *tptpCodePtr++ = '!';
      while(*tptpCodePtr != ':') {
         if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) *tptpCodePtr = -2;
	 else if(*tptpCodePtr == '[') *tptpCodePtr = -2;
	 else if(*tptpCodePtr == ']') *tptpCodePtr = -2;
	 tptpCodePtr++;}}
   tptpCodePtr++;}

/*---------------------------------------------------------------*/
/* delete : when followed by quant and another :                 */
/*---------------------------------------------------------------*/

/*
printf("0.. Add delimiters for quantifiers\n");
tptpE = ffClauseHead[0];
for(i = 0; *tptpE != '.' ;i++ ) { if(*tptpE != -2) printf("%c",*tptpE);tptpE++;}
printf("\n\n");
getchar();
*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {     /* to end of clause, work back  */
   tptpCodePtr++;}

while((int) tptpCodePtr != (int) tptpZ) {   /* now work back...  */
   if(*tptpCodePtr != ':') {
      tptpCodePtr--;
      continue;}
   else {
      tptpCodePtr--;

quant1:
      while(*tptpCodePtr != '!' && *tptpCodePtr != '@' && 
                     (int)tptpCodePtr != (int)tptpZ) {
         tptpCodePtr--;}
      tptpCodePtr--;
      while((*tptpCodePtr == ' ' || *tptpCodePtr == (char) -2 || 
			      *tptpCodePtr == (char) 10) &&
               (int) tptpCodePtr != (int) tptpZ) {
         tptpCodePtr--;}
      if(*tptpCodePtr == ':') {
         *tptpCodePtr = '$';
	 tptpCodePtr--;
	 goto quant1;}
      else {
         continue;}}}

/*---------------------------------------------------------------*/
/* add brackets to delimit quantifiers..                         */
/*---------------------------------------------------------------*/

/*
printf("1.. Add brackets to quantifiers\n");
tptpE = ffClauseHead[0];
for(i = 0; *tptpE != '.' ;i++ ) { if(*tptpE != -2) printf("%c",*tptpE);tptpE++;}
printf("\n\n");
getchar();
*/


nest_depth = 0;

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   if(*tptpCodePtr == '!' || *tptpCodePtr == '@') {
      if(nest_depth++ != -1) {
         tptpCodePtr--;
	 tptpE = tptpCodePtr;
	 while(*tptpE == ' ' || *tptpE == (char) -2 || *tptpE == (char) 10) {
	    tptpE--;}
         if(*tptpE == '$') {*tptpE = -2;tptpCodePtr++;tptpCodePtr++;continue;}
	 fof_InsertChar('{');
	 tptpCodePtr++;
	 tptpCodePtr++;}
      else {
         tptpCodePtr++;
	 continue;}

      brackets = 1;
      total_lr_brackets = 0;
      tptpE = tptpQ =tptpCodePtr;             /* first char after quant */
      while(*tptpQ != ':') tptpQ++;

      tptpQ++;

      while(*tptpQ == ' ' || *tptpQ == (char) -2 || 
		       *tptpQ == (char) 10 || *tptpQ == '~') {
	 tptpQ++;}
      if(*tptpQ != '{') {

/*-------------------------------------------------------------*/
/* go to end of atom and add bracket.                          */
/*-------------------------------------------------------------*/

         if(*tptpQ == '~') {
	    tptpQ++;}
	 while(*tptpQ == ' ' || *tptpQ == (char) -2 || *tptpQ == 10) tptpQ++;
	 while(isalnum(*tptpQ) || *tptpQ == '_' || 
	               *tptpQ == ',' || *tptpQ == '(' || *tptpQ == ')') {
	    tptpQ++;}
         tptpCodePtr = tptpQ;
	 tptpCodePtr++;
         goto cont_brackets;} 

      tptpCodePtr = tptpQ;
      while(brackets != 1 || total_lr_brackets == 0) {
         if(*tptpCodePtr != '{' && *tptpCodePtr != '}') {tptpCodePtr++;continue;}
         if(*tptpCodePtr == '{') {brackets++;total_lr_brackets++;}
         if(*tptpCodePtr == '}') brackets--;
	 tptpCodePtr++;}
      tptpCodePtr--;
cont_brackets:
      fof_InsertChar('}');
      tptpCodePtr = tptpE;}
   tptpCodePtr++;}


tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   tptpCodePtr++;
   if(*tptpCodePtr == '@' || *tptpCodePtr == '!') {
      tptpE = tptpCodePtr;
      if(tptpCodePtr > tptpZ) {
         tptpCodePtr--;
         while(*tptpCodePtr == (char) 10 || 
	       *tptpCodePtr == ' ' || *tptpCodePtr == (char) -2) {
            if(*tptpCodePtr == ' ') *tptpCodePtr = (char) -2;
            tptpCodePtr--;}
         if(*tptpCodePtr == ':') {*tptpCodePtr = (char) -2;}
	 tptpCodePtr = tptpE;}}}
         
tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   tptpCodePtr++;
   if(*tptpCodePtr == '=') {
      tptpE = tptpCodePtr;
      tptpCodePtr++;
      if(*tptpCodePtr == '>') {
         tptpCodePtr++;
         fof_InsertChar('{');
         brackets = 0;
         while(*tptpCodePtr != '{' && *tptpCodePtr != '}') {
            tptpCodePtr++;}
         if(*tptpCodePtr == '}') { 
            fof_InsertChar('}');
	    tptpCodePtr = tptpE;
	    continue;}
         else if(*tptpCodePtr == '{') {
            tptpCodePtr++;
	    brackets = 1;
	    while(*tptpCodePtr != '.') {
	       if(*tptpCodePtr == '{') brackets++;
	       if(*tptpCodePtr == '}') brackets--;
	       if(brackets == 0) {
	          fof_InsertChar('}');
		  tptpCodePtr = tptpE;
		  break;}
               tptpCodePtr++;}}		  
          tptpCodePtr = tptpE;}}}  


tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   tptpCodePtr++;
   if(*tptpCodePtr == '=') {
      tptpE = tptpCodePtr;
      tptpCodePtr--;
      if(*tptpCodePtr == '<') tptpCodePtr--;
      fof_InsertChar('}');
      tptpCodePtr--;
      brackets = 0;
      while(brackets != -1) {
         if(*tptpCodePtr == '}') brackets++;
	 if(*tptpCodePtr == '{') brackets--;
	 if(brackets == -1) {
	    fof_InsertChar('{');
	    break;}
         tptpCodePtr--;} 
      tptpE++;
      tptpE++;
      tptpCodePtr = tptpE;}}



/*---------------------------------------------------------------*/
/* remove carriage returns.                                      */
/*---------------------------------------------------------------*/

/*
printf(" 3... Add delimiters for quantifiers\n");
tptpE = ffClauseHead[0];
for(i = 0; *tptpE != '.' ;i++ ) { if(*tptpE != -2) printf("%c",*tptpE);tptpE++;}
printf("\n\n");
getchar();
*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   if(*tptpCodePtr == 10) {*tptpCodePtr = -2;}
   tptpCodePtr++;}

/*---------------------------------------------------------------*/
/* scan for "<=".                                                */
/*---------------------------------------------------------------*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   tptpE = tptpCodePtr;
   tptpE++;
   tptpU = tptpE;
   tptpU++;
   if(*tptpCodePtr == '<' && *tptpE == '=' && *tptpU != '>') {
//    printf("\n\nFound backward imp...\n");

/*--------------------------------------------------*/
/* mark left hand side                              */
/*--------------------------------------------------*/

      tptpE = tptpCodePtr;
      tptpE--;
      while(*tptpE != '}' && !isalnum(*tptpE) 
              && *tptpE != '_' && *tptpE != ')') tptpE--; 
//    printf("Right end of left side found =  %x %c\n",tptpE,*tptpE);
      if(isalnum(*tptpE)) {
         tptpU = tptpE;
         while(*tptpU != ' ') tptpU--;         
         goto left_here1;}
      else if(*tptpE == ')') {
         tptpU = tptpE;
         total_lr_brackets = 0;
         brackets = 1;
         while(brackets != 1 || total_lr_brackets == 0) {
            if(*tptpU != '(' && *tptpU != ')') {
	       tptpU--;continue;}
            if(*tptpU == ')') {brackets++;total_lr_brackets++;}
            if(*tptpU == '(') brackets--;
	    tptpU--;}
         while(*tptpU != ' ') tptpU--;
	 goto left_here1;}

      tptpU = tptpE;
      total_lr_brackets = 0;
      brackets = 1;
      while(brackets != 1 || total_lr_brackets == 0) {
         if(*tptpU != '{' && *tptpU != '}') {
	    tptpU--;continue;}
         if(*tptpU == '}') {brackets++;total_lr_brackets++;}
         if(*tptpU == '{') brackets--;
	 tptpU--;}
      tptpU++;
left_here1:
//    printf("Left end found of left side\n = %x",tptpU);
//    printf("Length = %d\n",(int) tptpE - (int) tptpU);
      length_ls = (int) tptpE - (int) tptpU;

      
      i = 0;
      for(tptpF = tptpU ; (int) tptpF <= (int) tptpE ; tptpF++) {
         left_side[i++] = *tptpF;}

/*--------------------------------------------------*/
/* mark right hand side                             */
/*--------------------------------------------------*/

      tptpM = tptpCodePtr;
      while(*tptpM != '{' && !isalnum(*tptpM)) tptpM++;
//    printf("\nLeft end of right side found =  %x\n",tptpM);

      if(isalnum(*tptpM)) {
          tptpN = tptpM;
          while(*tptpN != ' ') tptpN++;
	  goto right_here1;}

      tptpN = tptpM;
      total_lr_brackets = 0;
      brackets = 1;
      while(brackets != 1 || total_lr_brackets == 0) {
         if(*tptpN != '{' && *tptpN != '}') {
	    tptpN++;continue;}
         if(*tptpN == '{') {brackets++;total_lr_brackets++;}
         if(*tptpN == '}') brackets--;
	 tptpN++;}
      tptpN--;
right_here1:
//    printf("Right end found of right side\n = %x",tptpN);
//    printf("Length = %d\n",(int) tptpN - (int) tptpM);
      length_rs = (int) tptpN - (int) tptpM;

      i = 0;
      for(tptpF = tptpM ; (int) tptpF <= (int) tptpN ; tptpF++) {
//       printf("%c",*tptpF);
         right_side[i++] = *tptpF;}

/*---------------------------------------*/
/* interchange sides.                    */
/*---------------------------------------*/

      for(tptpF=tptpU; (unsigned int) tptpF <= (unsigned int) tptpN;
      tptpF++) {
         *tptpF = ' ';}

//printf("Cleared\n");getchar();
      for(i = 0; i <= length_rs; i++) {
         *tptpU = right_side[i];
	 tptpU++;}
//printf("LS fixed\n");getchar();
      *tptpU = '=';
      tptpU++;
       *tptpU = '>';
      tptpU++;

      for(i = 0; i <= length_ls ; i++) {
         *tptpU = left_side[i];
	 tptpU++;}
   tptpCodePtr = tptpZ;
      }

   tptpCodePtr++;} 

/*-----------------------------------------------*/
/* fix <~>                                       */
/*-----------------------------------------------*/

/*
printf("Pre --  repaired clause...\n");
for(tptpN = tptpZ; (unsigned int) tptpN <= tptpCodePtr;tptpN++) {
   if(*tptpN != -2) printf("%c",*tptpN);}
*/

tptpCodePtr = tptpZ;
while(*tptpCodePtr != '.') {
   tptpE = tptpCodePtr;
   tptpE++;
   tptpU = tptpE;
   tptpU++;
   if(*tptpCodePtr == '<' && *tptpE == '~' && *tptpU == '>') {
//    printf("\n\nFound <~> imp...\n");

/*--------------------------------------------------*/
/* mark left hand side                              */
/*--------------------------------------------------*/

      tptpE = tptpCodePtr;
      tptpE--;
      while(*tptpE != '}' && !isalnum(*tptpE) 
              && *tptpE != '_' && *tptpE != ')') tptpE--; 
//    printf("Right end of left side found =  %x %c\n",tptpE,*tptpE);
      if(isalnum(*tptpE)) {
         tptpU = tptpE;
         while(*tptpU != ' ') tptpU--;         
         goto left_here2;}
      else if(*tptpE == ')') {
         tptpU = tptpE;
         total_lr_brackets = 0;
         brackets = 1;
         while(brackets != 1 || total_lr_brackets == 0) {
            if(*tptpU != '(' && *tptpU != ')') {
	       tptpU--;continue;}
            if(*tptpU == ')') {brackets++;total_lr_brackets++;}
            if(*tptpU == '(') brackets--;
	    tptpU--;}
         while(*tptpU != ' ') tptpU--;
	 goto left_here2;}

      tptpU = tptpE;
      total_lr_brackets = 0;
      brackets = 1;
      while(brackets != 1 || total_lr_brackets == 0) {
         if(*tptpU != '{' && *tptpU != '}') {
	    tptpU--;continue;}
         if(*tptpU == '}') {brackets++;total_lr_brackets++;}
         if(*tptpU == '{') brackets--;
	 tptpU--;}
      tptpU++;
left_here2:
//    printf("Left end of left side found = %x\n",tptpU);
//    printf("Length = %d\n",(int) tptpE - (int) tptpU);
      length_ls = (int) tptpE - (int) tptpU;
      i = 0;

      for(tptpF = tptpU ; (int) tptpF <= (int) tptpE ; tptpF++) {
         left_side[i++] = *tptpF;}

/*---------------------------------------*/
/* find start and end of right-hand side */
/*---------------------------------------*/

      tptpM = tptpCodePtr;
      while(*tptpM != '{' && !isalnum(*tptpM)) tptpM++;
//    printf("\nLeft end of right side found =  %x\n",tptpM);

      if(isalnum(*tptpM)) {
          tptpN = tptpM;
          while(*tptpN != ' ') tptpN++;
	  goto right_here2;}

      tptpN = tptpM;
      total_lr_brackets = 0;
      brackets = 1;
      while(brackets != 1 || total_lr_brackets == 0) {
         if(*tptpN != '{' && *tptpN != '}') {
	    tptpN++;continue;}
         if(*tptpN == '{') {brackets++;total_lr_brackets++;}
         if(*tptpN == '}') brackets--;
	 tptpN++;}
      tptpN--;
right_here2:
//    printf("Right end of right side found = %x\n",tptpN);
//    printf("Length = %d\n",(int) tptpN - (int) tptpM);
      length_rs = (int) tptpN - (int) tptpM;

      i = 0;
      for(tptpF = tptpM ; (int) tptpF <= (int) tptpN ; tptpF++) {
         right_side[i++] = *tptpF;}

/*---------------------------------------*/
/* interchange sides.                    */
/*---------------------------------------*/

      tptpCodePtr++;
      *tptpCodePtr = '=';
      tptpCodePtr++;
      tptpCodePtr++;
//    printf("\n\nInsert enough spaces\n");
      for(i = 0 ; i < length_rs + length_ls + 20; i++) {
	 tptpN++;
         fof_InsertChar(' ');}

/*
printf("Done inserting --- Repaired clause...\n\n");
for(tptpM = tptpZ; ;tptpM++) {
   if(*tptpM != -2) printf("%c",*tptpM);
   if(*tptpM == '.') break;}
printf("\n\n");
*/

      for(tptpF=tptpU; (unsigned int) tptpF <= (unsigned int) tptpN;
      tptpF++) {
         *tptpF = ' ';}

/*
printf("Done blanking --- Repaired clause...\n\n");
for(tptpM = tptpZ; ;tptpM++) {
   if(*tptpM != -2) printf("%c",*tptpM);
   if(*tptpM == '.') break;}
printf("\n\n");
*/

/*----------------------------------*/
/* fix (A => ~B) & (~A => B)        */
/*----------------------------------*/

      *tptpU = '{';
      tptpU++;
      *tptpU = '{';
      tptpU++;

      for(i = 0; i <= length_ls; i++) {
         *tptpU = left_side[i];
	 tptpU++;}

      *tptpU = ' ';
      tptpU++;
      *tptpU = '=';
      tptpU++;
      *tptpU = '>';
      tptpU++;
      *tptpU = ' ';
      tptpU++;
      
      *tptpU = '~';
      tptpU++;

      for(i = 0; i <= length_rs ; i++) {
         *tptpU = right_side[i];
	 tptpU++;}
      
      *tptpU = '}';
      tptpU++;
      *tptpU = ' ';
      tptpU++;
      *tptpU = '&';
      tptpU++;
      *tptpU = ' ';
      tptpU++;

/*-----------------------------------*/
/* left half done to here, now right */
/*-----------------------------------*/

      *tptpU = '{';
      tptpU++;
      *tptpU = '~';
      tptpU++;

      for(i = 0; i <= length_ls; i++) {
         *tptpU = left_side[i];
	 tptpU++;}

      *tptpU = ' ';
      *tptpU = '=';
      tptpU++;
      *tptpU = '>';
      tptpU++;
      *tptpU = ' ';
      tptpU++;
      
      for(i = 0; i <= length_rs ; i++) {
         *tptpU = right_side[i];
	 tptpU++;}
      
      *tptpU = '}';
      tptpU++;
      *tptpU = '}';
      tptpU++;

      tptpCodePtr = tptpZ;
      }
      
   tptpCodePtr++;}
/*
printf("\n\nRepaired clause...\n");
for(tptpN = tptpZ; (unsigned int) tptpN <= tptpCodePtr;tptpN++) {
   if(*tptpN != -2) printf("%c",*tptpN);}
*/

goto STATE1;

}

/*--------------------------------------------------*/
/*  fof_InsertChar()      ...................... #3
/*
/*  called by:  ConvertFOF()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*--------------------------------------------------*/

void fof_InsertChar(s)
char s;
{
char *tptpCodeEndPtr1;
char *tptpCodeEndPtr2;

tptpCodeEndPtr++;
tptpCodeEndPtr1 = tptpCodeEndPtr;
tptpCodeEndPtr2 = tptpCodeEndPtr1;
tptpCodeEndPtr2--;

while(tptpCodeEndPtr2 >= tptpCodePtr) {
   *tptpCodeEndPtr1 = *tptpCodeEndPtr2;
   tptpCodeEndPtr2--;
   tptpCodeEndPtr1--;}

*tptpCodePtr = s;
return;
}

/*---------------------------------------------------*/
/*  fof_DeleteChar()          ..................  #4
/*
/*  called by:   ConvertFOF()     tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*---------------------------------------------------*/

void fof_DeleteChar(d)
int d;
{
int i;

for(i = d; i > 0; i--) {*tptpCodePtr++ = -2;}
}

/*---------------------------------------------------*/
/*  fof_CheckClauseType()        ................  #5
/*
/*  called by:    ConvertFOF()          tptpi.c
/*
/*  calls:        fof_CompareStrings()  tptpi.c
/*
/*  returns:      Nil
/*
/*----------------------------------------------------*/

void fof_CheckClauseType()
{
char *ccp;

ccp = tptpCodePtr;
while(*ccp != 10) {
   if(fof_CompareStrings(ccp,"conjecture",10)) {
      ffClauseType[ClauseCount - 1] = 1; return;}
   if(fof_CompareStrings(ccp,"hypothesis",10)) {
      ffClauseType[ClauseCount - 1] = 2; return;}
   ccp++;}
ffClauseType[ClauseCount - 1] = 3;
return;
}

/*----------------------------------------------------*/
/*  fof_CompareStrings(s,t,n)       .................  #6
/*
/*  called by:  fof_CheckClauseType()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    1 if strings are same, 0 otherwise
/*
/*-----------------------------------------------------*/

int fof_CompareStrings(s,t,n)
char s[],t[];
int n;
{
int i;

i = 0;
while(s[i] == t[i]) {
   i++;
   if(i == n) {return(1);}}
return(0);
}

/*-----------------------------------------------------*/
/*  fof_SetPtrToNextLine()      ....................  #7
/*
/*  called by:   ConvertFOF()      tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*-----------------------------------------------------*/

void fof_SetPtrToNextLine()
{
while(*tptpCodePtr != 10) {tptpCodePtr++;} 
tptpCodePtr++;
}

/*-----------------------------------------------------*/
/*  fof_AssignCostToClause()    .......................  #8
/*
/*  called by:  CompileFOF()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*  assigns cost to clauses, sorted with
/*  the negated conclusion at end, then by number of
/*  literals, then by length of clause in characters.
/*------------------------------------------------------*/

void fof_AssignCostToClause()
{
int i;

for(i = 0; i < ClauseCount; i++) {
    ffLitCount[i] =
       100000*ffClauseType[i]+1000*ffLitCount[i]+ffClauseLength[i];}
return;
}

/*-------------------------------------------------*/
/*  fof_DumpClause()         .....................  #9
/*
/*  called by:  fof_FinalSort()        tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*--------------------------------------------------*/

void fof_DumpClause(Imax)
int Imax;
{
int CRctr;
char *j, *k, *m, *c_temp, char_temp, char_temp2;
char *theoCodePtrInit;
int final_brackets;
int init_bracket;

final_brackets = 0;
init_bracket = 0;
tptpCodePtr = (char *) ffClauseHead[Imax];
theoCodePtrInit = theoCodePtr;

for( ; ; ) {
   if(*tptpCodePtr == (char) -2) {
      tptpCodePtr++;           /* skip deleted chars */
      continue;}

   if(*tptpCodePtr == '{' && init_bracket == 0) {
      init_bracket++;
      tptpCodePtr++;
      continue;}

   if(*tptpCodePtr == ';') {
      while(*tptpCodePtr != 10) tptpCodePtr++;
      tptpCodePtr++;
      continue;}

   *theoCodePtr = *tptpCodePtr++;
   if(*theoCodePtr == '|') {
      *theoCodePtr++ = ' ';
      *theoCodePtr++ = '|';
      *theoCodePtr = ' ';}

   if(*theoCodePtr == '&') {
      *theoCodePtr++ = ' ';
      *theoCodePtr++ = '&';
      *theoCodePtr = ' ';}

   if(*theoCodePtr == ':') {
      *theoCodePtr++ = ':';
      *theoCodePtr = ' ';}

   if(*theoCodePtr == '.') {
   *theoCodePtr = ' ';
   theoCodePtr--;
      while(final_brackets < 2) { 
	 if(*theoCodePtr == '}') {
            final_brackets++;	    
	    *theoCodePtr = ' ';
	    theoCodePtr--;}
         else if(*theoCodePtr == ' ') {
            theoCodePtr--;}}
      *theoCodePtr++;
      *theoCodePtr++ = '.';
      break;} 
   theoCodePtr++;}

theoCodePtr++;
*theoCodePtr++ = 10;


/*
printf("Begin print theo Clause\n\n");
for(c_temp = theoCodeBuffer; c_temp < theoCodePtr + 10; c_temp++) {
   printf("%c",*c_temp);}
printf("\n\nEND\n\n");
printf("Dump830\n");
getchar();
*/


return;
}

/*-------------------------------------------------*/
/*  fof_OutString()       ....................  #10
/*
/*  called by:  fof_FinalSort()        tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*  Place string "s" of length "Length" in theoCodeBuffer 
/*--------------------------------------------------*/

void fof_OutString(s,Length)
char s[];
int Length;
{
int i;

for(i = 0; i < Length ; i++) {
   *theoCodePtr++ = s[i];}
*theoCodePtr++ = 10;
return;
}

/*--------------------------------------------------*/
/*  fof_FinalSort()          ..................... #11
/*
/*  called by:  CompileFOF()         tptpi.c
/*
/*  calls:      fof_AssignCostToClause()  tptpi.c
/*
/*  returns:     Nil
/*       
/*---------------------------------------------------*/

void fof_FinalSort()
{
int IndexMax, LitMax, i, NegConclusion;
int ii;
char *OutputCodePtr;

theoCodePtr = theoCodeBuffer;
NegConclusion = 0;
for( ; ;   ) {
   IndexMax = 0;
   LitMax = 0;
   for(i = 0; i < ClauseCount; i++) {
      if(ffLitCount[i] > LitMax) {
         LitMax = ffLitCount[i];
         IndexMax = i;}}
   if(LitMax == 0) {goto finaldump;}
   ffLitCount[IndexMax] = 0;
   if(NegConclusion == 0 && LitMax < 200000) {
      NegConclusion++;
      fof_OutString("conclusion",10);}
   fof_DumpClause(IndexMax);}

finaldump:

*theoCodePtr = -1;


return;
}
 
/*--------------------------------------------------*/ 
/* CompileFOF()         .......................#12
/*
/* called by:  main()                  theo.c
/*
/* calls:      ReadFOF()               tptpi.c
/*             ConvertFOF()            tptpi.c
/*             fof_AssignCostToClause()    tptpi.c
/*             fof_FinalSort()         tptpi.c
/*
/* returns:    Nil
/*---------------------------------------------------*/
 
void CompileFOF()
{
char *tz;

ReadFOF(); 
fof_DetermineRating();
ConvertFOF();
if(add_eq_ax) {
   fof_add_equality_axioms();
   list_preds_fxns();
   add_equality_sub_axioms();
   }

fof_AssignCostToClause();
fof_FinalSort();
theoCodePtr = &theoCodeBuffer[0];
return;
}


/*---------------------------------------------------*/
/* fof_add_equality_axioms()
/*
/* add equality axioms
/*
/*---------------------------------------------------*/

void fof_add_equality_axioms()
{
char equal_d[] = "{@X: equal(X,X) }}.";
char equal_sym[] = "{@X,Y: ~equal(X,Y) | equal(Y,X) }}.";
char equal_trans[] = "{@X,Y,Z: ~equal(X,Y) | ~equal(Y,Z) | equal(X,Z) }}.";

char *eq_ptr, *tptpE;
int i;

ffClauseHead[ClauseCount] = (int) tptpCodePtr;
ffLitCount[ClauseCount] = 1;
ffClauseType[ClauseCount] = 4;
ffClauseLength[ClauseCount++] = 10;

eq_ptr = equal_d;
for(i = 0; i < 20; i++) {
   *tptpCodePtr++ = *eq_ptr++;}
*tptpCodePtr++ = 10;
*tptpCodePtr++ = 10;

ffClauseHead[ClauseCount] = (int) tptpCodePtr;
ffLitCount[ClauseCount] = 2;
ffClauseType[ClauseCount] = 4;
ffClauseLength[ClauseCount++] = 21;

eq_ptr = equal_sym;
for(i = 0; i < 37; i++) {
   *tptpCodePtr++ = *eq_ptr++;}
*tptpCodePtr++ = 10;
*tptpCodePtr++ = 10;

ffClauseHead[ClauseCount] = (int) tptpCodePtr;
ffLitCount[ClauseCount] = 3;
ffClauseType[ClauseCount] = 4;
ffClauseLength[ClauseCount++] = 32;

eq_ptr = equal_trans;
for(i = 0; i < 52; i++) {
   *tptpCodePtr++ = *eq_ptr++;}
*tptpCodePtr++ = 10;
*tptpCodePtr++ = 10;

ffClauseHead[ClauseCount] = (int) tptpCodePtr;
return;
}

/*-----------------------------------------------------*/
/* add_equality_sub_axioms()
/*
/*-----------------------------------------------------*/

void add_equality_sub_axioms()
{
int i, j, m, n;
char *tCodePtr, *tptpZ;
char *pred_chars;
int equ_sub_ax;
char sub_char;

char varb[] = "XYZUVWTSRQPONMLKJ";
char varb1[] = "X,Y,Z,U,V,W,T,S,R,Q,P,O,N,M,L,K,J";
char eql[] = "equal ";



pred_chars = predicates;

for(i = 0; i < nextp; i++) {
   pred_chars = predList[i];

/*----------------------------------------------------*/
/* don't add equal sub axioms for equal pred.         */
/*----------------------------------------------------*/

   if(fof_CompareStrings(eql,pred_chars,6)) continue; 

   equ_sub_ax = predArgs[i];
   sub_char = varb[equ_sub_ax];

/*----------------------------------------------------*/
/* for each argument of the predicate, add a clause.  */
/*----------------------------------------------------*/

   for(j = 0; j < equ_sub_ax; j++) {
      ffClauseHead[ClauseCount] = (int) tptpCodePtr;
      ffLitCount[ClauseCount] = 3;
      ffClauseType[ClauseCount] = 4;
      ffClauseLength[ClauseCount++] = 25 + 6*equ_sub_ax;

      *tptpCodePtr++ = '{';
      *tptpCodePtr++ = '@';

      for(m = 0; m < 2*equ_sub_ax + 1; m++) {
	 *tptpCodePtr++ = varb1[m];}

      *tptpCodePtr++ = ':';

      *tptpCodePtr++ = '~';
      pred_chars = predList[i];
      while(*pred_chars != ' ') *tptpCodePtr++ = *pred_chars++;
      *tptpCodePtr++ = '('; 
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         *tptpCodePtr++ = varb1[m];}
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = ' ';*tptpCodePtr++ = '|';*tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '~';
      *tptpCodePtr++ = 'e';
      *tptpCodePtr++ = 'q';
      *tptpCodePtr++ = 'u';
      *tptpCodePtr++ = 'a';
      *tptpCodePtr++ = 'l';
      *tptpCodePtr++ = '('; 
      *tptpCodePtr++ = varb[j];
      *tptpCodePtr++ = ',';
      *tptpCodePtr++ = sub_char;
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = ' ';*tptpCodePtr++ = '|';*tptpCodePtr++ = ' ';
      pred_chars = predList[i];
      while(*pred_chars != ' ') *tptpCodePtr++ = *pred_chars++;
      *tptpCodePtr++ = '('; 
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         if(m == 2*j) *tptpCodePtr++ = sub_char;
         else *tptpCodePtr++ = varb1[m];}
      *tptpCodePtr++ = ')'; 
      *tptpCodePtr++ = ' ';

      *tptpCodePtr++ = '}';*tptpCodePtr++ = '}';*tptpCodePtr++ = '.';
      *tptpCodePtr++ = 10;}}

pred_chars = functions;
for(i = 0; i < nextf; i++) {
   pred_chars = fcnList[i];
   if(fof_CompareStrings(eql,pred_chars,6)) continue; 

   equ_sub_ax = fcnArgs[i];
   sub_char = varb[equ_sub_ax];

   for(j = 0; j < equ_sub_ax; j++) {

      ffClauseHead[ClauseCount] = (int) tptpCodePtr;
      ffLitCount[ClauseCount] = 2;
      ffClauseType[ClauseCount] = 4;
      ffClauseLength[ClauseCount++] = 20 + 4*equ_sub_ax;

      *tptpCodePtr++ = '{';
      *tptpCodePtr++ = '@';
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
	 *tptpCodePtr++ = varb1[m];}
      *tptpCodePtr++ = ',';
      *tptpCodePtr++ = varb[j];
      *tptpCodePtr++ = '1';
      *tptpCodePtr++ = ':';
      *tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '~';
      *tptpCodePtr++ = 'e';
      *tptpCodePtr++ = 'q';
      *tptpCodePtr++ = 'u';
      *tptpCodePtr++ = 'a';
      *tptpCodePtr++ = 'l';
      *tptpCodePtr++ = '('; 
      *tptpCodePtr++ = varb[j];
      *tptpCodePtr++ = ',';
      *tptpCodePtr++ = varb[j];
      *tptpCodePtr++ = '1';
      *tptpCodePtr++ = ')'; 
      *tptpCodePtr++ = ' ';*tptpCodePtr++ = '|';*tptpCodePtr++ = ' ';
      *tptpCodePtr++ = 'e';
      *tptpCodePtr++ = 'q';
      *tptpCodePtr++ = 'u';
      *tptpCodePtr++ = 'a';
      *tptpCodePtr++ = 'l';
      *tptpCodePtr++ = '('; 

      pred_chars = fcnList[i];
      while(*pred_chars != ' ') { *tptpCodePtr++ = *pred_chars++;}
      *tptpCodePtr++ = '(';
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         *tptpCodePtr++ = varb1[m];}
      *tptpCodePtr++ = ')'; 

      *tptpCodePtr++ = ',';

      pred_chars = fcnList[i];
      while(*pred_chars != ' ') { *tptpCodePtr++ = *pred_chars++;}
      *tptpCodePtr++ = '(';

      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         *tptpCodePtr++ = varb1[m];
         if(varb1[m] == varb[j]) *tptpCodePtr++ = '1';}
      *tptpCodePtr++ = ')'; 
     

      *tptpCodePtr++ = ')'; 
      *tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '}';
      *tptpCodePtr++ = '}';
      *tptpCodePtr++ = '.';
      *tptpCodePtr++ = 10;
      *tptpCodePtr++ = 10;}}
     
ffClauseHead[ClauseCount] = tptpCodePtr;


/*
for(i = 0; i < ClauseCount; i++) {
printf("i=%d ",i);
for(tptpZ = ffClauseHead[i] ; tptpZ < ffClauseHead[i+1]; tptpZ++ ){ 
        printf("%c",*tptpZ);}getchar();}
*/

}

/*-----------------------------------------------------*/
/* add_skolem_axioms()
/*
/*-----------------------------------------------------*/

void add_skolem_axioms()
{
int i, j, m, n, nn, ii, jj;
char *tCodePtr, *tptpZ, *tptpZZ, *tptpQ, *tptpS, *tptpSS, *tptpJ;
char *pred_chars;
int equ_sub_ax;
char sub_char;
int skolem_args, skolem_arg_max;
int skolem_count;

char varb[] = "XYZUVWTSRQPONMLKJ";
char varb1[] = "X,Y,Z,U,V,W,T,S,R,Q,P,O,N,M,L,K,J";
char eql[] = "SK ";

int leading_zeros, t_power, t_temp, t1_temp, t1_char;

skolem_count = 0;
skolem_arg_max = 0;
for(i = 0; i < ffskolemid; i++) {
   if(ffskolem_list[i] > skolem_arg_max) skolem_arg_max = ffskolem_list[i];}

tCodePtr = ffclause_form_ptr;
for(nn = skolem_arg_max; nn > 0; nn--) {
for(i = 0; i < ffskolemid; i++) {
   if(add_eq_ax == 0) break;
   equ_sub_ax = ffskolem_list[i];
   if(equ_sub_ax != nn) continue;

   for(j = 0; j < equ_sub_ax; j++) {
      *tCodePtr++ = ' ';
      *tCodePtr++ = '~';
      *tCodePtr++ = 'e';
      *tCodePtr++ = 'q';
      *tCodePtr++ = 'u';
      *tCodePtr++ = 'a';
      *tCodePtr++ = 'l';
      *tCodePtr++ = '('; 
      *tCodePtr++ = varb[j];
      *tCodePtr++ = ',';
      *tCodePtr++ = varb[j];
      *tCodePtr++ = '1';
      *tCodePtr++ = ')'; 
      *tCodePtr++ = ' ';*tCodePtr++ = '|';*tCodePtr++ = ' ';
      *tCodePtr++ = 'e';
      *tCodePtr++ = 'q';
      *tCodePtr++ = 'u';
      *tCodePtr++ = 'a';
      *tCodePtr++ = 'l';
      *tCodePtr++ = '('; 
      *tCodePtr++ = 'S';
      *tCodePtr++ = 'K';

      t_temp = i;
      leading_zeros = 0;
      for(t_power = 100; ; t_power = t_power/10) {
         if(t_temp >= t_power) {
            t1_temp = t_temp/t_power;
            t1_char = t1_temp + 48;      
            *tCodePtr++  =  t1_char;      
            leading_zeros = 1;
            t_temp = t_temp - t1_temp*t_power;}
         else {
            if(leading_zeros == 1 || t_power == 1)  {
               t1_char = 48;         
               *tCodePtr++ = t1_char;
               leading_zeros = 1;}}
         if(t_power == 1) {break;}}
      
      *tCodePtr++ = '(';
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         *tCodePtr++ = varb1[m];}
      *tCodePtr++ = ')'; 

      *tCodePtr++ = ',';

      *tCodePtr++ = 'S';
      *tCodePtr++ = 'K';

      t_temp = i;
      leading_zeros = 0;
      for(t_power = 100; ; t_power = t_power/10) {
         if(t_temp >= t_power) {
            t1_temp = t_temp/t_power;
            t1_char = t1_temp + 48;      
            *tCodePtr++  =  t1_char;      
            leading_zeros = 1;
            t_temp = t_temp - t1_temp*t_power;}
         else {
            if(leading_zeros == 1 || t_power == 1)  {
               t1_char = 48;         
               *tCodePtr++ = t1_char;
               leading_zeros = 1;}}
         if(t_power == 1) {break;}}

      *tCodePtr++ = '(';
      for(m = 0; m < 2*equ_sub_ax - 1; m++) {
         *tCodePtr++ = varb1[m];
         if(varb1[m] == varb[j]) *tCodePtr++ = '1';}
      *tCodePtr++ = ')'; 
      *tCodePtr++ = ')'; 
     
      *tCodePtr++ = 10;}}}

/*----------------------------------------------------*/
/* work in skolem clauses with others 
/*----------------------------------------------------*/

tptpQ = &theoCodeBuffer[0];
tptpZ = &ffresults[0];
tptpZZ = ffclause_form_ptr;
i = 0;                        /* counter of "|" sym   */
j = 0;                        /* counter of arguments */
ii = 0;                        /* counter of "|" sym   */
jj = 0;                        /* counter of arguments */

while(*tptpZ == 10 || *tptpZ == ';') tptpZ++;
while(*tptpZZ == 10) tptpZZ++;

while(tptpZ < ffOutPtr || tptpZZ < tCodePtr) {
   i = j = ii = jj = 0;
   while(*tptpZ == 10 || *tptpZ == ';') tptpZ++;
   tptpS = tptpZ;
   while(*tptpZ != 10 && tptpZ < ffOutPtr) {
      if(strncmp(tptpZ,"negated_conclusion",18) == 0) {
         i = 500;}

      if((*tptpZ == '(' || *tptpZ == ',') && i != 500) j++;
      if(*tptpZ == '|' && i != 500) i++;
      tptpZ++;
      if(*tptpZ == 10) {
         if(i != 500) i = 1000*(i + 1) + j;
	 *tptpZ++;
	 break;}}

   /* check skolem clauses  */

   while(*tptpZZ == 10) tptpZZ++;
   tptpSS = tptpZZ;
   while(*tptpZZ != 10 && tptpZZ < tCodePtr) {
      if(*tptpZZ == '(' || *tptpZZ == ',') jj++;
      if(*tptpZZ == '|') ii++;
      tptpZZ++;
      if(*tptpZZ == 10) {
         ii = 1000*(ii + 1) + jj;
	 tptpZZ++;
	 break;}}
   
/*-------------------------------------------------------*/
/* compare i and ii, and output max of two 
/*-------------------------------------------------------*/

   if(i >= ii) {
      tptpZZ = tptpSS;
      for(tptpJ = tptpS; tptpJ < tptpZ; tptpJ++) {
         *tptpQ++ = *tptpJ;}}
   else {
      tptpZ = tptpS;
      for(tptpJ = tptpSS; tptpJ < tptpZZ; tptpJ++) {
         *tptpQ++ = *tptpJ;}}}

/*-----------------------------------------------------*/
/* now output clauses
/*-----------------------------------------------------*/

if(tptpQ != theoCodeBuffer) *tptpQ++ = -1;

}



/*------------------------------------------------------*/
/* fof_det_number_args()
/*
/*------------------------------------------------------*/

int fof_det_number_args(string_check,p_or_f)
char *string_check;
int p_or_f;
{
int arg_count;
char *scan_char;
int brackets;
int bracket_found;

arg_count = 0;
brackets = 0;
bracket_found = 0;
scan_char = string_check;
while(islower(*scan_char) || *scan_char == '_') {
   scan_char++;}

while(*scan_char != '|' || *scan_char != '.' || *scan_char != '}' ||
      *scan_char != ' ' || *scan_char != 10 || *scan_char != (char) -2) {
   if(*scan_char == ')') brackets--;
   if(*scan_char == '(') { 
      brackets++;
      bracket_found = 1;}
   if(*scan_char == ',' && brackets <= 1) arg_count++;
   if(*scan_char == ')' && brackets <= 0 || 
                arg_count + bracket_found == 0)  {
      return(arg_count + bracket_found);} 
   scan_char++;}

}


/*-----------------------------------------------------------------*/
/* list_preds_fxns()
/*
/* called by:
/*
/* calls:
/*
/* returns:   NIL
/*-----------------------------------------------------------------*/

void list_preds_fxns()
{

char *tptpE, *tptpF, *ptZ, *tptpZ, *tptpQ;
int char_count;
int pred_found;
int i, j;
char *thm_end;

/*-----------------------------------------------------------------*/
/* fix constants by adding ()                                      */
/*-----------------------------------------------------------------*/

for(j = 0; j < ClauseCount; j++)  {
   tptpCodePtr = ffClauseHead[j];

   if(ffClauseHead[j+1] != 0) thm_end = ffClauseHead[j + 1];
   else thm_end = tptpCodeEndPtr;

// printf("LIST: tptpCodePtr=%x ffClauseHead[j+1]=%x end=%x\n",
//    tptpCodePtr,ffClauseHead[j+1],tptpCodeEndPtr);

   while((int) tptpCodePtr < (int) thm_end) {
      if(isalnum(*tptpCodePtr) && isupper(*tptpCodePtr)) {

/*----------------------------------------------------------------*/
/* pass over variables.                                           */
/*----------------------------------------------------------------*/

         tptpF = tptpCodePtr; 
         while(isalnum(*tptpCodePtr) || *tptpCodePtr == '_') {
	    tptpCodePtr++;}
         continue;}

      else if(isalnum(*tptpCodePtr) && islower(*tptpCodePtr)) {
         tptpE = tptpF = tptpZ = tptpCodePtr;
         tptpCodePtr++;
         char_count = 0;
         while(isalnum(*tptpCodePtr) || *tptpCodePtr == '_') {
            char_count++;
            tptpCodePtr++;}
	 tptpE--;
	 while(*tptpE == (char) -2 || *tptpE == ' ') *tptpE--;
	 if(*tptpE != '(' && *tptpE != ',') {
            predList[nextp++] = predicates_ptr; 
	    while(tptpF < tptpCodePtr) {*predicates_ptr++ = *tptpF++;}
            *predicates_ptr++ = ' ';

            tptpF = tptpCodePtr;
            tptpQ = tptpZ;
	    pred_found = 0;
	    for(i = 0 ; i < nextp - 1; i++) {
	       tptpZ = tptpQ; 
               ptZ = predList[i];
               while(*ptZ ==  *tptpZ && *ptZ != ' ') {
	          ptZ++;tptpZ++;}
	       if(*ptZ == ' ' && (*tptpZ == '(' || *tptpZ == ' ')) {
	          nextp--;
	          pred_found = 1;
	          predicates_ptr = predList[nextp];
	          tptpF++;}}
            if(pred_found == 0) { 
	       predArgs[nextp - 1] = fof_det_number_args(tptpF,1);}}
         else {
            if(*tptpCodePtr == ',' || *tptpCodePtr == ')' ||
                 *tptpCodePtr == ' ' || *tptpCodePtr == (char) -2 || 
	         *tptpCodePtr == 10) {tptpCodePtr++;}
            else {
	       fcnList[nextf++] = functions_ptr;
               while(tptpF < tptpCodePtr) {*functions_ptr++ = *tptpF++;}
               *functions_ptr++ = ' ';

               tptpF = tptpCodePtr;
               tptpQ = tptpZ;
	       pred_found = 0;
	       for(i = 0 ; i < nextf - 1; i++) {
	          tptpZ = tptpQ; 
                  ptZ = fcnList[i];
                  while(*ptZ ==  *tptpZ && *ptZ != ' ') {ptZ++;tptpZ++;}
	          if(*ptZ == ' ' && (*tptpZ == '(' || *tptpZ == ' ')) {
	             nextf--;
		     pred_found = 1;
		     functions_ptr = fcnList[nextf];
	             tptpF++;}}
               if(pred_found == 0) {
                  fcnArgs[nextf - 1] = fof_det_number_args(tptpF,2);}}
            continue;}}
      else tptpCodePtr++;}}


/*
printf("PREDS...\n");
for(i = 0; i < nextp; i++) {
   tptpF = predList[i];
   while(*tptpF != ' ') {
      printf("%c",*tptpF);
      tptpF++;}
   printf(" ARGS=%d ",predArgs[i]);
   printf(" ");}

printf("\n\nFXNS...\n");
for(i = 0; i < nextf; i++) {
   tptpF = fcnList[i];
   while(*tptpF != ' ') {
      printf("%c",*tptpF);
      tptpF++;}
   printf(" ARGS=%d ",fcnArgs[i]);
   printf(" ");}
*/

}

