{-# OPTIONS -fglasgow-exts -cpp #-}
------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Set where

import Data.FiniteMap

import Maybe
  ( isJust
  )
  
-------------------------------------------------------------------------
-- maps and sets

type a :=> b = FiniteMap a b
type Set a   = a :=> ()

-- Using GHC 6.4, this should be commented out.
-- Using GHC 6.2, this should be left as-is.
#if __GLASGOW_HASKELL__ < 604
instance (Show a, Show b) => Show (a :=> b) where
  showsPrec n x = showsPrec n (tabulate x)
#endif

-------------------------------------------------------------------------
-- interface

empty :: (a :=> b)
empty = emptyFM

pair :: a -> b -> (a :=> b)
pair x y = unitFM x y

look :: Ord a => a -> (a :=> b) -> Maybe b
look x tab = lookupFM tab x

union :: Ord a => (a :=> b) -> (a :=> b) -> (a :=> b)
tab1 `union` tab2 = plusFM tab1 tab2

unionWith :: Ord a => (b -> b -> b) -> (a :=> b) -> (a :=> b) -> (a :=> b)
unionWith op tab1 tab2 = plusFM_C op tab1 tab2

delete :: Ord a => a -> (a :=> b) -> (a :=> b)
x `delete` tab = delFromFM tab x

difference :: Ord a => (a :=> b) -> (a :=> b) -> (a :=> b)
tab1 `difference` tab2 = minusFM tab1 tab2

intersect :: Ord a => (a :=> b) -> (a :=> b) -> (a :=> b)
tab1 `intersect` tab2 = intersectFM tab1 tab2

extend :: Ord a => a -> b -> (a :=> b) -> (a :=> b)
extend x y tab = addToFM tab x y

member, notMember :: Ord a => a -> (a :=> b) -> Bool
x `member`    tab = isJust (look x tab)
x `notMember` tab = not (x `member` tab)

subset :: Ord a => (a :=> b) -> (a :=> b) -> Bool
xs `subset` ys = size (xs `difference` ys) == 0

tabulate :: (a :=> b) -> [(a,b)]
tabulate tab = fmToList tab

table :: Ord a => [(a,b)] -> (a :=> b)
table xys = listToFM xys

accumulate :: Ord a => (b -> b -> b) -> [(a,b)] -> (a :=> b)
accumulate op xys = addListToFM_C op empty xys

size :: (a :=> b) -> Int
size = sizeFM

pick :: Ord a => (a :=> b) -> Maybe ((a,b),a :=> b)
pick tab | size tab == 0 = Nothing
         | otherwise     = Just ((a,b), a `delete` tab)
 where
  ((a,b):_) = tabulate tab

mapElt :: Ord a => (b -> c) -> (a :=> b) -> (a :=> c)
mapElt f tab = mapFM (\_ -> f) tab
 
-------------------------------------------------------------------------
-- set specific

set :: Ord a => [a] -> Set a
set xs = listToFM [ (x,()) | x <- xs ]

insert :: Ord a => a -> Set a -> Set a
x `insert` set = extend x () set

unit :: a -> Set a
unit x = unitFM x ()

elements :: (a :=> b) -> [a]
elements set = keysFM set

-------------------------------------------------------------------------
-- the end.
