/* Take a stream of interpretations and remove the isomorphic ones.
 */

#include "../ladr/top_input.h"
#include "../ladr/interp.h"

#define PROGRAM_NAME    "isofilter"
#define PROGRAM_VERSION "2005-A"
#include "../PROGRAM_DATE.h"

static char Help_string[] =
"\nThis program reads a stream of interpretations in portable format\n"
"(from stdin) and takes a command-line argument saying how to print them:\n\n"
"    portable      : one line per operation\n"
"    portable2     : portable, with binary operations in a square\n"
"    tabular       : as nice tables\n"
"    tex           : formatted for LaTeX\n\n";

enum {PORTABLE, PORTABLE2, TABULAR, TEX};

int main(int argc, char **argv)
{
  Term t;
  int type;

  if (string_member("help", argv, argc) ||
      string_member("-help", argv, argc)) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }
  else if (string_member("portable", argv, argc))
    type = PORTABLE;
  else if (string_member("portable2", argv, argc))
    type = PORTABLE2;
  else if (string_member("tabular", argv, argc))
    type = TABULAR;
  else if (string_member("tex", argv, argc))
    type = TEX;
  else {
    printf("%s", Help_string);
    exit(1);
  }

  init_standard_ladr();

  /* Note that we do not read any commands. */
  /* terms(interpretations) and end_of_list are optional. */

  t = read_term(stdin, stderr);
  if (is_term(t, "terms", 1))
    t = read_term(stdin, stderr);

  while (t != NULL && !end_of_list_term(t)) {
    Interp a = compile_interp(t);

    if (type == PORTABLE)
      fprint_interp(stdout, a);
    else if (type == PORTABLE2)
      fprint_interp_2(stdout, a);
    else if (type == TABULAR)
      fprint_interp_tabular(stdout, a);
    else if (type == TEX)
      fprint_interp_tex(stdout, a);

    zap_interp(a);
    zap_term(t);
    t = read_term(stdin, stderr);
  }
  
  exit(0);
}  /* main */

