#ifndef TP_BTM_H
#define TP_BTM_H

#include "termorder.h"
#include "termflag.h"
#include "unify.h"
#include "accanon.h"

/* INTRODUCTION
This package handles "backtrack matching", that is, matching
that allows more than one unifier for a pair of terms, and computes
the unifiers incrementally by backtracking.  As I write this,
we support associative commutative (AC) operations and 
commutative/symmetric (C) operations.  Symbols are declared to be
AC with set_assoc_comm() and C with set_commutative().
The use of Terms and Contexts is similar to ordinary matching,
except that the means for undoing substitutions is different.
*/

/* Public definitions */

typedef struct btm_state * Btm_state;

/* End of public definitions */

/* Public function prototypes from btm.c */

void fprint_btm_mem(FILE *fp, BOOL heading);

void p_btm_mem();

void p_btm_state(Btm_state bt);

Btm_state match_bt_first(Term t1, Context c1, Term t2, int partial);

Btm_state match_bt_next(Btm_state bt1);

void match_bt_cancel(Btm_state bt);

#endif  /* conditional compilation of whole file */
