#ifndef TP_CLAUSE_H
#define TP_CLAUSE_H

#include "tlist.h"
#include "termflag.h"
#include "attrib.h"

/* INTRODUCTION
*/

/* Public definitions */

typedef struct literal * Literal;

struct literal {
  int      sign;
  Term     atom;
  Literal  next;
};

typedef struct clause * Clause;

struct clause {
  int              id;
  int              weight;
  Literal          literals;
  Attribute        attributes;
  struct just      *justification;
  struct clist_pos *containers;
  char             *compressed;

  char             initial;
  char             normal_vars;
  char             used;
};

/* End of public definitions */

/* Public function prototypes from clause.c */

Literal get_literal(void);

void free_literal(Literal p);

Clause get_clause(void);

void fprint_clause_mem(FILE *fp, int heading);

void p_clause_mem();

void zap_literal(Literal l);

void zap_literals(Literal l);

void zap_clause(Clause c);

void fprint_clause(FILE *fp, Clause c);

void p_clause(Clause c);

Literal new_literal(int sign, Term atom);

void append_literal(Clause c, Literal lit);

int clause_length(Clause c);

Literal term_to_literals(Term t, Literal lits);

Clause term_to_clause(Term t);

Term clause_to_term(Clause c);

Term clause_to_term_without_attributes(Clause c);

Term clause_to_term_x(Clause c);

Term lits_to_term(Literal l);

void free_lits_to_term(Term t);

void clause_set_variables(Clause c, int max_vars);

void renumber_variables(Clause c, int max_vars);

void term_renumber_variables(Term t, int max_vars);

int positive_literals(Clause c);

int negative_literals(Clause c);

BOOL positive_clause(Clause c);

BOOL any_clause(Clause c);

BOOL negative_clause(Clause c);

BOOL mixed_clause(Clause c);

int number_of_literals(Clause c);

int number_of_pos_literals(Clause c);

int number_of_neg_literals(Clause c);

BOOL unit_clause(Clause c);

BOOL horn_clause(Clause c);

BOOL definite_clause(Clause c);

int greatest_variable_in_clause(Clause c);

BOOL ground_clause(Clause c);

void upward_clause_links(Clause c);

Clause copy_clause(Clause c);

Clause copy_clause_with_flags(Clause c);

Clause copy_clause_with_flag(Clause c, int flag);

int literal_number(Clause c, Literal l);

int atom_number(Clause c, Term atom);

Literal ith_literal(Clause c, int i);

BOOL true_term(Term t);

BOOL false_term(Term t);

BOOL tautology(Clause c);

void inherit_attributes(Clause parent1, Context subst1,
			Clause parent2, Context subst2,
			Clause child);

Mlist function_symbols_in_clause(Clause c, Mlist g);

Mlist relation_symbols_in_clause(Clause c, Mlist g);

Mlist function_symbols_in_clauses(Plist lst);

Mlist relation_symbols_in_clauses(Plist lst);

Ilist fsym_set_in_clauses(Plist lst);

Ilist rsym_set_in_clauses(Plist lst);

int symbol_occurrences_in_clause(Clause c, int symnum);

Literal remove_null_literals(Literal l);

Literal first_literal_of_sign(Clause c, BOOL sign);

Plist constants_in_clause(Clause c);

BOOL clause_ident(Clause c1, Clause c2);

BOOL horn_clauses(Plist l);

BOOL unit_clauses(Plist l);

int clause_symbol_count(Clause c);

int clause_depth(Clause c);

BOOL min_depth(Literal lit);

#endif  /* conditional compilation of whole file */
