#ifndef TP_FORMULA_H
#define TP_FORMULA_H

#include "listterm.h"
#include "tlist.h"

/* INTRODUCTION
*/

/* Public definitions */

/* formula types */

typedef enum {
  ATOM_FORM=0, AND_FORM, OR_FORM, NOT_FORM, IFF_FORM,
  IMP_FORM, IMPBY_FORM, ALL_FORM, EXISTS_FORM} Ftype;

typedef struct formula * Formula;
struct formula {
  Ftype   type;
  int     arity;
  char    *qvar;     /* quantified variable */
  Formula *kids;     /* for non-atoms */
  Term    atom;      /* for atoms */
  int excess_refs;    /* count of extra references */
};

/* formula preference */

typedef enum { CONJUNCTION,
	       DISJUNCTION
             } Fpref;

/* macros */

#define TRUE_FORMULA(f)  ((f)->type == AND_FORM && (f)->arity == 0)
#define FALSE_FORMULA(f) ((f)->type == OR_FORM  && (f)->arity == 0)

/* End of public definitions */

/* Public function prototypes from formula.c */

void free_formula(Formula p);

void fprint_formula_mem(FILE *fp, BOOL heading);

void p_formula_mem();

unsigned formula_megs(void);

Formula formula_get(int arity, Ftype type);

void zap_formula(Formula f);

Mlist function_symbols_in_formula(Formula f, Mlist g);

Mlist function_symbols_in_formulas(Plist lst);

Mlist relation_symbols_in_formula(Formula f, Mlist g);

Mlist relation_symbols_in_formulas(Plist lst);

Formula term_to_formula(Term t);

Term formula_to_term(Formula f);

void fprint_formula(FILE *fp, Formula f);

void p_formula(Formula c);

unsigned hash_formula(Formula f);

BOOL formula_ident(Formula f, Formula g);

Formula formula_copy(Formula f);

BOOL dual_type(int op);

Formula dual(Formula f);

Formula and(Formula a, Formula b);

Formula or(Formula a, Formula b);

Formula imp(Formula a, Formula b);

Formula impby(Formula a, Formula b);

Formula negate(Formula a);

BOOL quant_form(Formula f);

Formula flatten_top(Formula f);

Formula formula_flatten(Formula f);

Formula nnf2(Formula f, Fpref pref);

Formula nnf(Formula f);

Formula make_conjunction(Formula f);

Formula make_disjunction(Formula f);

void formula_canon_eq(Formula f);

int formula_size(Formula f);

int greatest_qvar(Formula f);

int greatest_symnum_in_formula(Formula f);

void subst_free_var(Formula f, Term target, Term replacement);

Formula eliminate_rebinding(Formula f);

Formula get_quant_form(Ftype type, char *qvar, Formula subformula);

Formula universal_closure(Formula f);

BOOL free_variable(char *svar, Formula f);

Formula formulas_to_conjunction(Plist formulas);

#endif  /* conditional compilation of whole file */
