#ifndef TP_GLIST_H
#define TP_GLIST_H

#include "memory.h"
#include "order.h"

/* INTRODUCTION
This package handles
Plist (singly-linked list of void pointers),
Ilist (singly-linked list of integers), and
Mlist (singly-linked list of <integer,integer> pairs).
*/

/* Public definitions */

typedef struct plist * Plist;

struct plist {
  void *v;
  Plist next;
};

typedef struct ilist * Ilist;

struct ilist {
  int i;
  Ilist next;
};

typedef struct mlist * Mlist;

struct mlist {
  int i;
  int n;
  Mlist next;
};

/* End of public definitions */

/* Public function prototypes from glist.c */

Ilist get_ilist(void);

void free_ilist(Ilist p);

Plist get_plist(void);

void free_plist(Plist p);

Mlist get_mlist(void);

void free_mlist(Mlist p);

void fprint_glist_mem(FILE *fp, BOOL heading);

void p_glist_mem();

Ilist ilist_cat(Ilist p1, Ilist p2);

Ilist ilist_cat2(Ilist p1, Ilist p2);

Ilist ilist_pop(Ilist p);

int ilist_count(Ilist p);

Ilist reverse_ilist(Ilist p);

Ilist copy_ilist(Ilist p);

void zap_ilist(Ilist p);

Ilist ilist_append(Ilist lst, int i);

Ilist ilist_prepend(Ilist lst, int i);

BOOL ilist_member(Ilist lst, int i);

Ilist ilist_subtract(Ilist p1, Ilist p2);

Ilist ilist_removeall(Ilist p, int i);

Ilist ilist_intersect(Ilist p1, Ilist p2);

Ilist ilist_set(Ilist m);

Ilist ilist_rem_dups(Ilist m);

BOOL ilist_is_set(Ilist a);

BOOL ilist_subset(Ilist a, Ilist b);

Plist plist_cat(Plist p1, Plist p2);

Plist plist_cat2(Plist p1, Plist p2);

Plist plist_pop(Plist p);

int plist_count(Plist p);

Plist reverse_plist(Plist p);

Plist copy_plist(Plist p);

void zap_plist(Plist p);

Plist plist_append(Plist lst, void *v);

Plist plist_prepend(Plist lst, void *v);

BOOL plist_member(Plist lst, void *v);

Plist plist_subtract(Plist p1, Plist p2);

BOOL plist_subset(Plist a, Plist b);

void fprint_ilist(FILE *fp, Ilist p);

void p_ilist(Ilist p);

Ilist ilist_copy(Ilist p);

Ilist ilist_remove_last(Ilist p);

Plist plist_remove_last(Plist p);

Ilist alist_insert(Ilist p, int key, int val);

int assoc(Ilist p, int key);

Ilist alist2_insert(Ilist p, int key, int a, int b);

int assoc2a(Ilist p, int key);

int assoc2b(Ilist p, int key);

Ilist alist2_remove(Ilist p, int key);

int ilist_occurrences(Ilist p, int i);

Plist plist_remove(Plist p, void *v);

Ilist ilist_insert_up(Ilist p, int i);

BOOL mlist_multimember(Mlist lst, int i, int n);

BOOL mlist_multisubset(Mlist a, Mlist b);

Mlist multiset_add_n(Mlist p, int i, int n);

Mlist multiset_add(Mlist p, int i);

void zap_mlist(Mlist p);

Mlist mlist_removeall(Mlist p, int i);

Mlist mlist_member(Mlist lst, int i);

Mlist multiset_union(Mlist a, Mlist b);

Ilist multiset_to_set(Mlist m);

int multiset_occurrences(Mlist m, int i);

void p_mlist(Mlist p);

Plist sort_plist(Plist objects,	Ordertype (*comp_proc) (void *, void *));

#endif  /* conditional compilation of whole file */
