#ifndef TP_JUST_H
#define TP_JUST_H

#include "clauseid.h"

/* INTRODUCTION
*/

/* Public definitions */

typedef enum {
  INPUT_JUST,         /* Primary               */
  CLAUSIFY_JUST,      /* Primary               */
  COPY_JUST,          /* Primary    int        */
  BACK_DEMOD_JUST,    /* Primary    int        */
  BACK_UNIT_DEL_JUST, /* Primary    int        */
  NEW_SYMBOL_JUST,    /* Primary    int        */
  BINARY_RES_JUST,    /* Primary    Ilist      */
  HYPER_RES_JUST,     /* Primary    Ilist      */
  UR_RES_JUST,        /* Primary    Ilist      */
  FACTOR_JUST,        /* Primary    Ilist      */
  XXRES_JUST,         /* Primary    Ilist      */
  PARA_JUST,          /* Primary    Parajust   */
  PARA_FX_JUST,       /* Primary    Parajust   */
  PARA_IX_JUST,       /* Primary    Parajust   */
  PARA_FX_IX_JUST,    /* Primary    Parajust   */

  DEMOD_JUST,         /* Secondary  Ilist      */
  UNIT_DEL_JUST,      /* Secondary  Ilist      */
  FLIP_JUST,          /* Secondary  int        */
  XX_JUST,            /* Secondary  int        */
  MERGE_JUST          /* Secondary  int        */

} Just_type;

typedef struct parajust * Parajust;

struct parajust {
  int from_id;
  int into_id;
  Ilist from_pos;
  Ilist into_pos;
};

typedef struct just * Just;

struct just {
  Just_type type;
  Just next;
  union {
    int id;
    Ilist lst;
    Parajust para;
  } u;
};

/* A justification starts with a Primary part and then
   has a sequence (maybe none) of Secondary parts.
   Each type has some data, an integer or Ilist (integer
   list) giving clause IDs, or positions of literals or
   subterms.
*/

/* End of public definitions */

/* Public function prototypes from just.c */

Just get_just(void);

Parajust get_parajust(void);

void fprint_just_mem(FILE *fp, BOOL heading);

void p_just_mem();

Just input_just(void);

Just clausify_just(void);

Just copy_just(Clause c);

Just new_symbol_just(Clause c);

Just back_demod_just(Clause c);

Just back_unit_deletion_just(Clause c);

Just binary_res_just(Clause c1, int n1, Clause c2, int n2);

Just factor_just(Clause c, int lit1, int lit2);

Just xxres_just(Clause c, int lit);

Just resolve_just(Ilist g, Just_type type);

Just para_just(Just_type rule,
	       Clause from, Ilist from_vec,
	       Clause into, Ilist into_vec);

Just para_just_rev_copy(Just_type rule,
			Clause from, Ilist from_vec,
			Clause into, Ilist into_vec);

Just demod_just(Ilist g);

Just unit_del_just(Clause deleter, int literal_num);

Just flip_just(int n);

Just xx_just(int n);

Just merge_just(int n);

Just append_just(Just j1, Just j2);

Just copy_justification(Just j);

void fix_input_just(Clause c);

int jmap1(Ilist map, int i);

char *jmap2(Ilist map, int i, char *a);

void sb_write_just(String_buf sb, Just just, Ilist map);

void p_just(Just j);

void zap_just(Just just);

Ilist get_parents(Clause c);

Plist get_clause_ancestors(Clause c);

int proof_length(Plist proof);

void map_just(Just just, Ilist map);

int just_count(Just j);

void mark_parents_as_used(Clause c);

#endif  /* conditional compilation of whole file */
