#include "string.h"

/* Private definitions and types */

/*************
 *
 *     int str_ident(s, t) --  Identity of strings
 *
 *************/

/* DOCUMENTATION
This function routine checks identity of two strings.
*/

/* PUBLIC */
BOOL str_ident(char *s, char *t)
{
  return strcmp(s, t) == 0;
}  /* str_ident */

/*************
 *
 *   new_str_copy()
 *
 *************/

/* DOCUMENTATION
Return a malloced copy of the given string.  To avoid memory leaks,
call free() on the copy if you finish referring to it.
*/

/* PUBLIC */
char *new_str_copy(char *str)
{
  char *p = malloc(strlen(str)+1);
  strcpy(p, str);
  return p;
}  /* new_str_copy */

/*************
 *
 *   string_member()
 *
 *************/

/* DOCUMENTATION
Is "string" a member of an array of "strings"?
*/

/* PUBLIC */
BOOL string_member(char *string, char **strings, int n)
{
  int i;
  for (i = 1; i < n; i++)
    if (str_ident(strings[i], string))
      return TRUE;
  return FALSE;
}  /* string_member */

/*************
 *
 *   which_string_member()
 *
 *************/

/* DOCUMENTATION
If "string" is a member of an array of "strings", return the index;
else return -1.
*/

/* PUBLIC */
int which_string_member(char *string, char **strings, int n)
{
  int i;
  for (i = 1; i < n; i++)
    if (str_ident(strings[i], string))
      return i;
  return -1;
}  /* which_string_member */
