#include "symbolcheck.h"

/* Private definitions and types */

/*************
 *
 *   function_relation_check()
 *
 *************/

/* DOCUMENTATION
Given a Plist of Clists, check that no symbol is used as both
both a relation and a function.  If the check fails,
argument "fatal" says whether it should be a fatal or just a WARNING.
*/

/* PUBLIC */
void function_relation_check(Ilist fsyms, Ilist rsyms, BOOL fatal)
{
  Ilist intersect_syms = ilist_intersect(fsyms, rsyms);

  if (intersect_syms) {
    Ilist g;
    bell(stderr);
    fprintf(stderr,
	    "\nWARNING: symbols occur as both relation and function"
	    " (see stdout).\n\n ");
    printf("\nWARNING: symbols occur as both relation and function:");
    for (g = intersect_syms; g; g = g->next)
      printf(" %s/%d", sn_to_str(g->i), sn_to_arity(g->i));
    printf(".\n");
    zap_ilist(intersect_syms);
    if (fatal)
      fatal_error("symbol occurs as both relation and function");
  }
}  /* function_relation_check */

/*************
 *
 *   collect_multiples()
 *
 *************/

static
Ilist collect_multiples(Ilist syms)
{
  /* Example: Given  (f/0, f/1, g/0, h/2), 
              return (f/0, f/1).
  */
  Ilist p1;
  Ilist p2;
  Ilist bad_syms = NULL;
  for (p1 = syms; p1; p1 = p1->next) {
    char *s1 = sn_to_str(p1->i);
    for (p2 = p1->next; p2; p2 = p2->next) {
      char *s2 = sn_to_str(p2->i);
      if (str_ident(s1, s2)) {
	if (!ilist_member(bad_syms, p1->i))
	  bad_syms = ilist_prepend(bad_syms, p1->i);
	if (!ilist_member(bad_syms, p2->i))
	  bad_syms = ilist_prepend(bad_syms, p2->i);
      }
    }
  }
  return bad_syms;
}  /* collect_multiples */

/*************
 *
 *   arity_check()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void arity_check(Ilist fsyms, Ilist rsyms, BOOL fatal)
{
  Ilist syms = ilist_cat(ilist_copy(fsyms), ilist_copy(rsyms));
  Ilist bad_syms = collect_multiples(syms);
  
  if (bad_syms) {
    Ilist g;
    bell(stderr);
    fprintf(stderr,
	    "\nWARNING: symbol with different arities"
	    " (see stdout).\n\n ");
    printf("\nWARNING: symbol with different arities:");
    for (g = bad_syms; g; g = g->next)
      printf(" %s/%d", sn_to_str(g->i), sn_to_arity(g->i));
    printf(".\n");
    zap_ilist(bad_syms);
    if (fatal)
      fatal_error("symbol with different arities");
  }
  zap_ilist(syms);
}  /* arity_check */

