#include "msearch.h"

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

static
void print_banner(int argc, char **argv)
{
  int i;
  int pid = 0;
    
  printf("----- %s %s, %s -----\n",
	 PROGRAM_NAME, PROGRAM_VERSION, PROGRAM_DATE);
  printf("Built with library %s %s, %s.\n",
	 LIBRARY_NAME, LIBRARY_VERSION, LIBRARY_DATE);
  printf("The process was started by %s on %s,\n%s",
	 username(), hostname(), get_date());
  pid = my_process_id();

  printf("The command was \"");
  for(i = 0; i < argc; i++)
    printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
  printf("\".");
  if (pid != 0)
    printf("  The process ID is %d.\n\n", pid);
  else
    printf("\n\n");
    
}  /* print_banner */

/*************
 *
 *   init_attrs()
 *
 *************/

void init_attrs(void)
{
  // This will allow these attributes to occur on clauses.
  // Mace4 will ignore these attributes.

  int id;
  id = register_attribute("label",         STRING_ATTRIBUTE);
  id = register_attribute("bsub_hint_add_wt", INT_ATTRIBUTE);
  id = register_attribute("bsub_hint_wt",     INT_ATTRIBUTE);
  id = register_attribute("answer",          TERM_ATTRIBUTE);
  id = register_attribute("action",          TERM_ATTRIBUTE);
  id = register_attribute("action2",         TERM_ATTRIBUTE);
}  /* init_attrs */

/*************
 *
 *   read_mace4_input()
 *
 *************/

static
Plist read_mace4_input(int argc, char **argv, BOOL allow_unknown_things,
		      Mace_options opt)
{
  Plist wild_clauses, wild_formulas;
  Plist wild_terms, hints_clauses, hints_formulas;  /* we won't use these */

  // Tell the top_input package what lists to accept and where to put them.

  // Accept hints, but they will not be used.

  accept_list("hints", CLAUSES,  &hints_clauses);
  accept_list("hints", FORMULAS, &hints_formulas);

  // Accept any other clauses and formulas.  These will be used.

  accept_list("",    CLAUSES,  &wild_clauses);
  accept_list("",    FORMULAS, &wild_formulas);

  // Accept any terms.  These will not be used.

  accept_list("",      TERMS,    &wild_terms);

  // Read commands such as set, clear, op, lex.
  // Read lists, filling in variables given to the accept_list calls.

  read_all_input(argc, argv, stdout, TRUE,
		 allow_unknown_things ? WARN_UNKNOWN : KILL_UNKNOWN);
    
  printf("\n%% Finished reading the input.\n");

  if (wild_terms)
    printf("%%   term list(s) ignored\n");
  if (hints_clauses || hints_formulas)
    printf("%%   hints list(s) ignored\n");

  process_command_line_args(argc, argv, opt);

  // Clausify any formulas and append to the corresponding clause lists.

  wild_clauses = plist_cat(wild_clauses, clausify_formulas(wild_formulas));

  return wild_clauses;
}  /* read_mace4_input */

/*************
 *
 *   main()
 *
 *************/

int main(int argc, char **argv)
{
  struct mace_options opt;
  Plist clauses;
  Mace_results results;

  /* Following says whether to ignore unregognized set/clear/assigns. */
  BOOL prover_compatability_mode = member_args(argc, argv, "-c");

  init_standard_ladr();
  init_mace_options(&opt);  /* We must do this before calling usage_message. */
  init_attrs();  

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help")) {
    usage_message(stderr, &opt);
    exit(1);
  }

  print_banner(argc, argv);

  clauses = read_mace4_input(argc, argv, prover_compatability_mode, &opt);

  results = mace4(clauses, &opt);

  mace4_exit(results->return_code);  /* print messages and exit */

  exit(0);  /* won't happen */

}  /* main */


