#include "msearch.h"
#include "../ladr/order.h"

extern int Number_of_cells;
extern struct cell *Cells;

extern struct cell **Ordered_cells;  /* Permutation of Cells. */

/*************
 *
 *   sum_indexes(t)
 *
 *   Tssume t is an eterm, that is, nonvariable with variable arguments.
 *
 *************/

static
int sum_indexes(Term t)
{
  if (ARITY(t) == 0)
    return -1;
  else {
    int sum = 0;
    int i;
    for (i = 0; i < ARITY(t); i++)
      sum += VARNUM(ARG(t,i));
    return sum;
  }
}  /* sum_indexes */

/*************
 *
 *   compare_cells()
 *
 *   For example, if a < b < g < f, this gives the following order:
 *
 *   a.            g(2).           0 = 0.
 *   b.            f(0,2).         0 = 1.
 *   g(0).         f(2,0).         0 = 2.
 *   f(0,0).       f(1,2).         1 = 0.
 *   g(1).         f(2,1).         1 = 1.
 *   f(0,1).       f(2,2).         1 = 2.
 *   f(1,0).                       2 = 0.
 *   f(1,1).                       2 = 1.
 *                                 2 = 2.
 *
 *   Equality cells are greatest, then sort by
 *
 *   <maximum-index, symbol order, sum-of-indexes>.
 */

static
Ordertype compare_cells(struct cell *a, struct cell *b)
{
  if (a->symbol->attribute == EQUALITY_SYMBOL &&
      b->symbol->attribute == EQUALITY_SYMBOL)  return SAME_AS; /* don't care */
                                                   
  else if (b->symbol->attribute == EQUALITY_SYMBOL)  return LESS_THAN;
                                                       
  else if (a->symbol->attribute == EQUALITY_SYMBOL)  return GREATER_THAN;
                                                       
  else if (a->max_index < b->max_index)              return LESS_THAN;
                                                       
  else if (a->max_index > b->max_index)              return GREATER_THAN;
                                                       
  else if (a->symbol->mace_sn < b->symbol->mace_sn)  return LESS_THAN;
                                                       
  else if (a->symbol->mace_sn > b->symbol->mace_sn)  return GREATER_THAN;

  else if (sum_indexes(a->eterm) <
           sum_indexes(b->eterm))                    return LESS_THAN;
    
  else if (sum_indexes(a->eterm) >
           sum_indexes(b->eterm))                    return GREATER_THAN;
    
  else

    return SAME_AS;  /* For now, let f(0,1) be the same as f(1,0).  */
  
}  /* compare_cells */

/*************
 *
 *   order_cells()
 *
 *************/

void order_cells(void)
{
  int i;

  for (i = 0; i < Number_of_cells; i++)
    Ordered_cells[i] = Cells + i;

  merge_sort((void **) Ordered_cells, Number_of_cells,
             (Ordertype (*) (void*,void*)) compare_cells);

#if 0  /* print the Ordered_cells */
  for (i = 0; i < Number_of_cells; i++)
    fwrite_term_nl(stdout, Ordered_cells[i]->eterm);
#endif

}  /* order_cells */
