/* -------------------------------------------------------------*/
/*      clause.h          data structures for THEO and HERBY
/*      copyrighted by Monty Newborn 2005
/*
/* -------------------------------------------------------------*/
 
/*--------------------------------------------------------------*/ 
/*------    data structure for a literal header (l_id)  --------*/    
/*                                                              */
/*    bit 32: NEGMASK -- 1 if negated predicate, 0 otherwise    */
/*    bit 31: ELIMLIT -- 1 if lit eliminated, 0 otherwise       */
/*    bit 30: VARMASK -- 1 if vers in lit, 0 otherwise          */
/*    bit 29: EVALMASK-- 1 if fxn replacement, 0 otherwise      */
/*    bits28-21......store substituted constant                 */
/*    bits 20-5......name of literal                            */
/*    bits 4-0.......number of arguments in literal             */
/*--------------------------------------------------------------*/ 
 
typedef struct lt {
	struct lt   *l_next;               /*  ptr to next lit  */
	int         l_X,l_Y;               /*   lit's hashcode  */
	int         l_id;}
	literal;
 
/*--------------------------------------------------------------*/ 
/*---------------------    l_id field      ---------------------*/
/*--------------------------------------------------------------*/ 
 
#define IDMASK      0x0000FFC0          /* pid,vid,fid masks    */
#define ARGMASK     0x0000003F            /* 0..31 arguments    */
#define FCNMASK     0x80000000           /* 1=item is function  */
#define NEGMASK     0x80000000  /* in preds, indicates negation */
#define ARGSHIFT    0x00000006   /* number of args in 5 ls bits */
#define ELIMLIT     0x40000000       /* pseudo-resolved lit     */
#define VARMASK     0x20000000       /* if on, vars in lit      */
#define EVALMASK    0x10000000       /* 1 if fxn replacement    */
#define CONSMASK    0x0FF00000  /* store substituted constant   */
#define CONSMASK2   0x000F0000  /* store substituted constant   */

/*--------------------------------------------------------------*/ 
/*-------------   data structure for a clause header   ---------*/    
/*--------------------------------------------------------------*/ 
 
typedef struct cl {
	int         c_id;                      /* clause id     */
	struct cl   *c_end;             /* ptr past end of cls  */
	int         c_bits;                      /* see below   */
	int         c_X,c_Y;              /* clause hash code   */
	struct cl   *c_lpar,*c_rpar;         /* our parents     */
	int         c_indices; /* index of resolvent of parents */
	literal     *c_lit;               /* pointer to 1st lit */
	int         c_nlit;           /* number of  lits in cls */
	int	    c_change;    /* condensed repre. of var ch  */
	int         c_nv;}       /* number distinct vars in cls */
	clause;
 
/*--------------------------------------------------------------*/ 
/*---------------------    c_bits field    ---------------------*/
/*--------------------------------------------------------------*/ 
 
#define UNIT        0x00000001                 /* unit clause   */
#define MERGE       0x00000004                /* merge clause   */
#define BASE        0x00000008                 /* base clause   */
#define DEAD        0x00000020                 /* dead clause   */
#define VID         0x00000040  /* var names start at 0 or 200  */
#define ACTIVE      0x00000080              /*  active clause   */
#define NC          0x00000100                    /* nc clause  */
#define LIST        0x00000400   /*  list in output of proof    */
#define EQCLS       0x00000800         /* equal cls, fxn or lit */
#define HRES	    0x00001000  /* hash resolution		*/
 
/*--------------------------------------------------------------*/ 
/*---------------------  c_indices field   ---------------------*/
/*--------------------------------------------------------------*/ 
 
#define LEFT_INDEX_MASK   0x00000fc0    /* mask for left index  */
#define RIGHT_INDEX_MASK  0x0000003f    /* mask for right index */
#define LEFT_INDEX_SHIFT  6          /* 1st indice in 6 ls bits */

/*--------------------------------------------------------------*/ 
/* ----------- pointers to clause stack ----------------------- */
/* ---------- one structure for each level in tree ------------ */
/*--------------------------------------------------------------*/ 
 
typedef struct {
	int         d_ec;      /* top clause at this level      */
	clause      *d_rc;   /* root clause (first clause ptr)  */
	int         d_tc;  /* clause under search at this level */
	int         d_bc;}      /* bottom clause at this level  */
	dnode;
 
/*--------------------------------------------------------------*/ 
/* ------------ clause hash table data structures --------------*/
/*--------------------------------------------------------------*/ 
 
typedef struct hn {
	int         h_X,h_Y;       /* hashcode of lit (64 bits) */
	short       h_in;           /* iteration number         */
	short	    h_ply;}        /* must be at same ply      */
	hnode;
 
/*--------------------------------------------------------------*/ 
/* -------------- random number table entry---------------------*/
/* -------- each entry is a 64-bit random number    ------------*/
/*--------------------------------------------------------------*/ 
 
typedef struct rn {
	int       r_hash1,r_hash2;}
	rnode;
 
/*--------------------------------------------------------------*/ 
/*---- data structures used by affc.c---------------------------*/
/*--------------------------------------------------------------*/ 
 
typedef struct cn {
	struct      cn *c_next,*c_prev;
	int         c_pred;} 
	cnode;
 
typedef struct pf {
	struct      pf *p_llink,*p_rlink;
	int         p_bits;
	char        *p_name;             /* tree linkage stuff  */
	int         p_id;                   /* unique ID        */
	int         p_nargs;}           /* # arguments of obj   */
	pfnode;
 
#define P_RTHREAD   0x00000001          /* threaded tree stuff  */
#define P_LTHREAD   0x00000002
#define P_PRED      0x00000004    /* pred. bit (1=predicate)    */
#define P_FUNC      0x00000000           /* 0 = function        */
#define P_TYPE      0x00000004           /* type mask           */
 
