/*----------------------------------------------------*/
/*
/* compwff.h
/*
/* Copyright Newborn Software 1993, 1996, 1999
/*
/*----------------------------------------------------*/
 
# define T_IFF 		257L
# define T_IMPLIES 	258L
# define T_AND  	259L
# define T_OR 		260L
# define T_NOT 		261L
# define T_FORALL 	262L
# define T_EXISTS 	263L
# define T_LPAR 	264L
# define T_RPAR 	265L
# define T_LBRACE 	266L
# define T_RBRACE 	267L
# define T_LBRACKET 	268L
# define T_RBRACKET 	269L
# define T_COMMA 	270L
# define T_ID 		271L
# define T_DOT 		272L
# define T_COLON 	273L
# define T_NUM 		274L
# define T_THEOREM 	275L
# define T_CONSTANT 	276L
# define T_PREDICATE 	277L
# define T_FUNCTION 	278L
# define T_STATIC 	279L
# define T_DYNAMIC 	280L
 
#define YYMAXDEPTH      150
#define YYERRCODE       256
#define YYLAST          150
#define YYFLAG        -1000

#define DPMAX     MAXVARS*2
#define VOFFSET        1000       /* standardization offset   */
#define MAXNODES       2000       /* must be >= MAXVARS       */

#define YYLERR      fyysqec
 
#define YYLMAX          200
#define YYNEWLINE        10

#define MAXERR           21    /* max number errors before aborting  */

#define RTHREAD           1     /* right thread bit (1=thread) */
#define LTHREAD           2     /* left thread bit (1=thread)     */

#define MAXID           800
#define MAXCHAR     MAXID*5

#define MAXVARS           500      /* max vars in wff 		*/
#define MAXLEX             50       /* max [] nesting 		*/
#define VNAMESIZE          50            /* 6 chars + null  */

#define CODESIZE      200000L    /* max amount of "code" per wff  */
 

/*------------  data structures for varsym.c   ------------*/
 
typedef struct  {
        char ffv_name[VNAMESIZE];              /* max # chars in var name */
        long ffv_id;                        /* unique clausal var id 	*/
        long ffv_ll;}                      /* lex level of decl	*/
	ffvnode;
 
/*--------------  data structures for pfsym.c --------------*/
 

typedef struct ffpf {     
        struct ffpf *ffp_llink,*ffp_rlink;
	long ffp_bits;
	char *ffp_name;
	long ffp_id;
	long ffp_nargs;}
	ffpfnode;

/*
#define P_RTHREAD       1L   
#define P_LTHREAD       2L
 
#define P_PRED          4L
#define P_FUNC          0L
#define P_TYPE          P_PRED
*/

 
#define MAXARGS 	31L   /* max possible args to a fcn/pred 	*/
#define FARGSHIFT 	5L             /* << ARGSHIFT for code 		*/
 
#define FXNMASK 	0x80000000L      /* "code" word fields 		*/
#define FNEGMASK 	FXNMASK
#define FARGMASK 	0x0000001FL
#define FIDMASK  	0x7FFFFFE0L
 
/*-------------data structures for ops.c -----------------*/
 
typedef struct ffop {long ffo_bits;
	struct ffop *ffo_lhs,*ffo_rhs;} 
	ffopnode;
		  
/*-----------------------------------------------------------------*/ 
/*                                                                 */
/* o_bits bit declarations and meanings. the opnode can represent: */
/*                                                                 */
/*              1) binary expression (e.g. =>)                     */
/*              2) atomic unit (e.g. P(x))                         */
/*              3) wff (with quantifiers)                          */
/*              4) quantifier with variable                        */
/*                                                                 */
/*-----------------------------------------------------------------*/
 
#define O_LNEG  0x0001L          /* negate left hand side 		*/
#define O_RNEG  0x0002L          /* negate right hand side 		*/
#define O_IFF   0x0010L          /* mask for operators 			*/
#define O_IMP   0x0020L
#define O_AND   0x0030L
#define O_OR    0x0040L
#define O_ALL   0x0050L
#define O_EX    0x0060L
#define O_OP    0x00F0L                     /* operator mask  */
#define O_BIN   0x0100L    /* binary operator opnode 	      */
#define O_ATOM  0x0200L          /* pointer to atomic unit    */
#define O_WFF   0x0400L          /* sub-wff indicator         */
#define O_QUAN  0x0800L   /* opnode is a quantified variable  */
#define O_FMT   0x0f00L              /* mask for format type  */
 
/*----------------------------------------------------------*/ 
/*                                                          */
/* if the O_BIN bit is set then the opnode fields are:      */
/*                                                          */
/*      o_lhs : ^ left hand side sub expression             */
/*      o_rhs : ^ right hand side sub expression            */
/*                                                          */
/* the op field is then in effect and the LNEG and RNEG     */
/* bits are used to indicate negation of the left and       */
/* right hand sides                                         */
/*                                                          */
/* sample binary operator:                                  */
/*                                                          */
/*      opnode:         [E1 & ~E2]  O_RNEG, O_AND bits set  */
/*                      /        \                          */
/*                 E1 opnode  E2 opnode                     */
/*                                                          */
/* if the O_ATOM bit is set then :                          */
/*                                                          */
/*      o_lhs : ^ pred data list (actual text for atom)     */
/*      o_rhs : unused                                      */
/*                                                          */
/* the LNEG bit is used to indicate negation of the atom    */
/*                                                          */
/*      opnode:         [~P(x,y,z)]     O_LNEG set          */
/*                                                          */
/*----------------------------------------------------------*/
 
#define ffo_pred ffo_lhs
 
/*----------------------------------------------------------*/ 
/*                                                          */
/* if the O_WFF bit is set then:                            */
/*                                                          */
/*      o_lhs : ^ expression part of the wff                */
/*      o_rhs : ^ opnode list of quantifiers & variables    */
/*                                                          */
/* again the LNEG is used to indicate negation of the wff   */
/*                                                          */
/*      opnode:         [E      Q]                          */
/*                      /        \                          */
/*                 E opnode     [QUAN]->[QUAN]->[QUAN] ...  */
/*                                                          */
/*----------------------------------------------------------*/
 
#define ffo_expr ffo_lhs
#define ffo_qlist ffo_rhs
 
/*----------------------------------------------------------*/ 
/*                                                          */
/* if the O_QUAN bit is set then:                           */
/*                                                          */
/*      o_lhs : next quantified variable                    */
/*      o_rhs : variable ID within wff                      */
/*                                                          */
/* the bits field is otherwise useless                      */
/*                                                          */
/*      opnode: [next   V]                                  */
/*                                                          */
/*----------------------------------------------------------*/
 
#define ffo_next ffo_lhs
#define ffo_var ffo_rhs
 
/*-----------------------------------------------------------*/ 
/*                                                           */
/* For example, the wff:                                     */
/*                                                           */
/*      @x,y!z: P(x) & Q(y) => [P(x) <=> Q(y)].              */
/*                                                           */
/* is represented as:                                        */
/*                                                           */
/*          [WFF E Q]------>[QUAN @x]->[QUAN @y]->[QUAN !z]  */
/*               |                                           */
/*          [BIN E1 => E2]                                   */
/*              /        \                                   */
/*      [BIN E3 & E4]       [WFF E Q]                        */
/*          /       \             \                          */
/*   [ATOM P(x)]  [ATOM Q(y)]   [BIN E5 <=> E6]              */
/*                                  /         \              */
/*                          [ATOM P(x)]    [ATOM Q(y)]       */
/*                                                           */
/*                                                           */
/*-----------------------------------------------------------*/
 
/*---------------------data structures for error.c ----------*/
 
#define E_PFSYMFULL     1L  /* pred/fcd symbol table full     */
#define E_NAMEFULL      2L  /* character buffer full          */
#define E_NOID          3L  /* id missing                     */
#define E_MULTIDEF      4L  /* id defined twice               */
#define E_BADCHAR       5L  /* illegal character              */
#define E_BADINCLUDE    6L  /* bad include filename           */
#define E_NOTFOUND      7L  /* include file not found         */
#define E_UNDEFINED     8L  /* undefined identifier           */
#define E_ARGNUM        9L  /* incorrect number of arguments  */
#define E_NOTYPE        10L /* expected predicate/fxn/const   */
#define E_TOOMANYARGS   11L /* too many args declared         */
#define E_NOPID         12L /* predicate expected             */
#define E_FCNASPRED     13L /* identifier must be a predicate */
#define E_NORBRACKET    14L						
#define E_NORBRACE      15L						
#define E_PREDEXP       16L   /* predicate expected           */
#define E_FUNCEXP       17L   /* function expected            */
#define E_LEXTOODEEP    18L   /* too many nested {}/[]        */
#define E_MULTIVAR      19L   /* var quantified twice         */
#define E_VSYMFULL      20L   /* too many variables           */
#define E_EXPR          21L   /* error in expression          */
#define E_OPSPACE       22L   /* no more opspace              */
#define E_CODESIZE      23L   /* no more predicate space      */
#define E_NODOT         24L   /* period expected              */
#define E_CONSTARG      25L   /* constants cannot have args   */
