/*----------------------------------------------------------*/
/*
/* ops.c           wff conversion to clause form
/*
/* Copyright Newborn Software 1993, 1996, 1999, 2004
/*
/* functions:           
/*  fof_op_alloc().....................#2
/*  fof_op_release()...................#3
/*  fof_atom().........................#4
/*  fof_setop()........................#5
/*  fof_whatop().......................#6
/*  fof_op_rel_expr()..................#7
/*  fof_op_clause()....................#8
/*  fof_op_copy()......................#9
/*  fof_op_iffimpneg().................#10
/*  fof_op_standardize()...............#11
/*  fof_change().......................#12
/*  fof_op_cnform()....................#13
/*  fof_skolinit().....................#14
/*  fof_op_skolemize().................#15
/*  fof_op_emit()......................#16
/*  fof_print_predicate()..............#17
/*  fof_print_argument()...............#18
/*  fof_print_skolem_fxn().............#18a
/*  fof_op_sort()......................#19
/*  fof_print_tree()...................#20
/*
/*------------------------------------------------------------*/
 
 
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
 
#include "compwff.h"
#include "clause.h"
 
/*--------------------------------------------------------------*/
/* externally defined functions 
/*--------------------------------------------------------------*/

extern void fof_emit_code();
extern int ffcode[],ffcodeic;   /* skolem functions go here     */
extern ffpfnode *fof_pf_getnode();/*fcn/pred symboltable stuff  */
extern void fof_pf_add();           /* to add skolem functions  */
extern void fof_print0();

/*--------------------------------------------------------------*/
/* externally defined constants 
/*--------------------------------------------------------------*/

extern int fferrcount;             /* only convert if no errors   */
extern int fftneg;
extern ffpfnode *ffpidtbl[],*fffidtbl[];
extern char ffclause_form[200000], *ffclause_form_ptr;
extern char ffresults[200000], *ffOutPtr;
 
extern char *ffClauseStart[2000],*ffClauseEnd[2000];
extern int ffClauseWeight[2000], ffClauseRank[2000];

extern int first_attempt;
 
/*--------------------------------------------------------------*/
/* internally defined constants 
/*--------------------------------------------------------------*/

int ffskolemid;                     /* skolem functions         */
int ffdpstack[DPMAX],ffdpsp;        /* holds var #s             */
int ffdplex[MAXLEX],ffdlx;          /* marks sub-wff levels     */
int ffsklist[DPMAX];                /* skolemization pointers   */
 
int ffvid;              /* must init on first call to stand..   */
char ffskname[200];                     /* skolem function name */

int fflit_count;
int ffclause_count;
int fffxn_count;

int spaces;

/*--------------------------------------------------------------*/
/* internally defined functions
/*--------------------------------------------------------------*/

ffopnode *fof_op_alloc();
void fof_op_release();
int fof_atom();
void fof_setop();
int fof_whatop();
void fof_op_rel_expr();
void fof_op_clause();
ffopnode *fof_op_copy();
void fof_op_iffimpneg();
void fof_op_standardize();
void fof_change();
void fof_op_cnform();
void fof_skolinit();
void fof_op_skolemize();
void fof_op_emit();
void fof_print_predicate();
int fof_print_argument();
int fof_print_skolem_fxn();
void fof_op_sort();
void fof_print_tree();
 
/*------------------------------------------------------------*/
/* fof_op_alloc()                   ...................#2
/*
/* fof_op_alloc will allocate an opnode,
/* if no nodes are available, exit COMPILE
/*
/* called by:  fof_op_iffimpneg()         ops.c
/*             fof_op_cnform()            ops.c
/*             fof_op_copy()              ops.c
/*             fof_yyparse()              compwff.c
/*             fof_binop()                compwff.c
/*
/* calls:      fof_yerror()               err.c
/*
/* returns:    pointer to ffopnode
/*------------------------------------------------------------*/
 
ffopnode *fof_op_alloc()
{
ffopnode *o;      
    
o = (ffopnode *) malloc(sizeof(ffopnode));
 
if(o == 0) {fof_yerror(E_OPSPACE);exit(0);}
   
o->ffo_bits = 0;                              /* no node */
return(o);
}
 
/*-----------------------------------------------------------*/
/* fof_op_release()            .........................#3
/*
/* op_release will release a passed ffopnode (put it back
/* on the list)
/*
/* called by:  fof_op_skolemize()       ops.c
/*             fof_op_rel_exp()         ops.c
/*      
/* calls:      nobody
/*
/* returns:    nil
/*----------------------------------------------------------*/
 
void fof_op_release(o)
ffopnode *o;
{
if((int) o != 0) free(o);
}
 
/*----------------------------------------------------------*/
/* fof_atom()                .........................#4
/* 
/* called by:  fof_op_skolemize()       ops.c
/*             fof_op_cnform()          ops.c
/* 
/* calls:      nobody
/*
/* returns:    T if opnode is an atom, F if an expression
/*-----------------------------------------------------------*/
 
int fof_atom(w)
ffopnode *w;
{
return((w->ffo_bits & O_FMT) == O_ATOM);
}
 
/*----------------------------------------------------------*/
/* fof_setop()                ........................ #5
/*
/* setop will set the O_OP field of an opnode
/*
/* called by:  fof_op_iffimpneg()      ops.c
/*
/* calls:      nobody
/*
/* returns:    nil
/*---------------------------------------------------------*/
 
void fof_setop(w,op)
ffopnode *w;
int op;
{
w->ffo_bits = (w->ffo_bits & (~O_OP)) | op;
}
 
/*---------------------------------------------------------*/
/* fof_whatop()               ......................... #6
/*
/* called by:  fof_op_iffimpneg()          ops.c
/*             fof_op_skolemize()          ops.c
/*             fof_op_cnform()             ops.c
/*
/* calls:      nobody
/*
/* returns:    the O_OP field of the ffopnode (if not a WFF)
/*----------------------------------------------------------*/
 
int fof_whatop(w)
ffopnode *w;
{
return(w->ffo_bits & O_OP);
}
 
/*--------------------------------------------------------*/
/* fof_op_rel_expr()           ........................ #7
/*
/* fof_op_rel_expr will release an expression back to the free list
/* 
/* called by:  fof_op_clause()          ops.c
/*
/* calls:      fof_op_release()         ops.c
/*
/* returns:    nil
/*---------------------------------------------------------*/
 
void fof_op_rel_expr(o)
ffopnode *o;
{
ffopnode *p,*q;
    
if((int) o == 0) return;           /* just in case        */
 
switch((int) (o->ffo_bits & O_FMT)) {
   case O_BIN:
      fof_op_rel_expr(o->ffo_lhs);   /* release left hand side   */
      fof_op_rel_expr(o->ffo_rhs);   /* release right hand side  */
      fof_op_release(o);           /* and the pointing node    */
      break;
 
   case O_WFF:
      fof_op_rel_expr(o->ffo_expr);/* get rid of expression part */
      for(q = o->ffo_qlist; q != 0;) {
         p = q;                  /* node to release        */
         q = q->ffo_next;        /* next one in chain        */
         fof_op_release(p);}   /* release quantifier       */
 
   case O_QUAN:
   case O_ATOM:
      fof_op_release(o);                  /* single node!  */
      break;}
return;
}
 
 
/*--------------------------------------------------------*/
/* fof_op_clause()               .................... #8
/*
/* called by:  fof_yyparse()                 compwff.c
/*
/* calls:      fof_op_iffimpneg()            ops.c
/*             fof_op_standardize()          ops.c
/*             fof_op_skolemize()            ops.c
/*             fof_op_cnforn()               ops.c
/*             fof_op_emit()                 ops.c
/*             fof_op_rel_expr()             ops.c
/*
/* returns:    nil
/* --------------------------------------------------------*/
 
void fof_op_clause(w,ffthneg)
ffopnode *w;
int ffthneg;
{
int i;

spaces = 0;
if(fferrcount != 0) return;
 
/*----------------------------------------------*/
/*      Step 1: eliminate =>, <=, and <=>.      */
/*      Step 2: Distribute negation symbols     */
/*----------------------------------------------*/

ffvid = VOFFSET; 
fof_skolinit(); 
fof_print_tree(w,1); 

fof_op_iffimpneg(w,ffthneg); /* remove <=>, =>, demote ~ */


fof_print_tree(w,1); 
 
/*----------------------------------------------*/
/* Step 3: Rename variables variables           */
/*----------------------------------------------*/
 
fof_op_standardize(w);     

fof_print_tree(w,0);
 
/*----------------------------------------------*/
/* Step 4: Skolemize existent'ly quant' var's   */
/*----------------------------------------------*/
 
fof_skolinit();fof_op_skolemize(w);   /* remove @, !; skolemize */
 
/*----------------------------------------------*/
/* Step 6: Skolemize existent'ly quant' var's   */
/*----------------------------------------------*/
 

fof_print_tree(w,0);
fof_op_cnform(w);       /* conjunctive normal form  */
 
/*----------------------------------------------*/
/* Step 7: Rewrite as separate clauses          */
/* implicitly connected by an AND function      */
/*----------------------------------------------*/

ffClauseStart[ffclause_count] = ffclause_form_ptr;
fflit_count = 1; 
fffxn_count = 0;
ffclause_count++;

fof_op_emit(w,0); 

ffClauseEnd[ffclause_count - 1]  = ffclause_form_ptr;
if(fftneg) ffClauseWeight[ffclause_count - 1] = 5000 + 100*fflit_count + fffxn_count;
else ffClauseWeight[ffclause_count - 1] = 100000 +100*fflit_count + fffxn_count;

fof_print0("\n\n",0);
//getchar();
fof_op_rel_expr(w);              /* release the expression     */

return;            /* and return to parse() for next wff   */
}
 
/*---------------------------------------------------------*/
/* fof_op_copy()              ....................  #9
/*
/* fof_op_copy will return an exact copy of a passed op structure.
/* NOTE: if we are copying an atom, we copy the code as well.
/*
/* called by:  fof_op_iffimpneg()        ops.c
/*             fof_op_cnform()           ops.c
/*
/* calls:      fof_op_alloc()            ops.c
/*             fof_emit_code()           ops.c
/*
/* returns:        pointer to new node
/*----------------------------------------------------------*/
 
ffopnode *fof_op_copy(w)
ffopnode *w;
{
ffopnode *o,*q,*p;
int c;
int i;
 
if((int) w == 0) return(w);
o = fof_op_alloc();                     /* copy this node   */
o->ffo_bits = w->ffo_bits;   /* copy node exactly (for now) */
o->ffo_lhs = w->ffo_lhs;
o->ffo_rhs = w->ffo_rhs;
 
switch((int) (w->ffo_bits & O_FMT)) {
   case O_WFF:
      p = 0;
      o->ffo_qlist = 0;
      for(q = w->ffo_qlist; q != 0; q = q->ffo_next) {
	       if(p == 0) {
	          p = fof_op_copy(q);
	          o->ffo_qlist = p;} 
	       else {
	          p->ffo_next = fof_op_copy(q);
	          p = p->ffo_next; }}
      o->ffo_expr = fof_op_copy(w->ffo_expr);
      return(o);
 
   case O_QUAN:
      return (o);                        /* already done  */
 
   case O_ATOM:
 
/* we copy the code for the atom as well (the predicate) */
 
      c = (int) o->ffo_pred;
      o->ffo_pred = (ffopnode *) ffcodeic;
      for(i = 0; i >= 0; i--) {
	       fof_emit_code(ffcode[c]);
	       i += ffcode[c] & FARGMASK;
	       c++;}
      return(o);
 
   case O_BIN:
      o->ffo_lhs = fof_op_copy(w->ffo_lhs);
      o->ffo_rhs = fof_op_copy(w->ffo_rhs);
      return(o);}
}
 
/*---------------------------------------------------------*/
/* fof_op_iffimpneg()             ................. #10
/*
/* fof_op_iffimpneg will:
/*
/* a) remove all iff, implies operators (transform to & |);
/* b) reduce scope of negation symbols to atomic units;
/*
/* this is done recursively. since the iff operator
/* expands to two implies operators, the expression can grow.
/*      
/* paramters:   w : ^ ffopnode (wff) to reduce
/*              neg : !=0 if wff to be negated, else 0
/*
/* called by:  fof_op_clause()            ops.c
/*
/* calls:      fof_whatop()               ops.c
/*             fof_op_copy()              ops.c
/*             fof_setop()                ops.c
/*             fof_op_alloc()             ops.c
/*
/* returns:    nil
/*----------------------------------------------------------*/
 
void fof_op_iffimpneg(w,neg)
ffopnode *w;
int neg;
{
ffopnode *e1,*e2,*q;
    
if((int) w == 0) return;
 
switch((int) (w->ffo_bits & O_FMT)) {
 
/* the binary operator format consists of a number of   */
/* expressions. each one is taken separately            */
 
   case O_BIN:
      switch((int) fof_whatop(w)) {
	 case O_IFF:
	    e1 = fof_op_copy(w);      /* will be LHS => RHS */
	    fof_setop(e1,O_IMP);
	    e2 = fof_op_alloc();      /* will be LHS | ~RHS */
	    e2->ffo_bits = w->ffo_bits ^ O_RNEG;
	    e2->ffo_lhs = w->ffo_lhs;
	    e2->ffo_rhs = w->ffo_rhs;
	    fof_setop(e2,O_OR);
	    w->ffo_lhs = e1;           /* becomes e1 & e2 */
	    w->ffo_rhs = e2;
	    w->ffo_bits &= ~(O_LNEG | O_RNEG);
	    fof_setop(w,O_AND);
	    fof_op_iffimpneg(w,neg);
	    return;
 
	 case O_AND:
	    if(neg != 0) {
	       fof_setop(w,O_OR);
	       w->ffo_bits ^= (O_LNEG | O_RNEG);}
	    fof_op_iffimpneg(w->ffo_lhs,w->ffo_bits & O_LNEG);
	    fof_op_iffimpneg(w->ffo_rhs,w->ffo_bits & O_RNEG);
	    w->ffo_bits &= ~(O_LNEG|O_RNEG);
	    return;
 
	 case O_IMP:
	    fof_setop(w,O_OR);
	    w->ffo_bits ^= O_LNEG;
	    /* fall through to OR */
 
	 case O_OR:
	    if(neg != 0) {
	       fof_setop(w,O_AND);
	       w->ffo_bits ^= (O_LNEG | O_RNEG);}
	    fof_op_iffimpneg(w->ffo_lhs,w->ffo_bits & O_LNEG);
	    fof_op_iffimpneg(w->ffo_rhs,w->ffo_bits & O_RNEG);
	    w->ffo_bits &= ~(O_LNEG | O_RNEG);
	    return;
 
	 default:
	    return;}
 
/* atoms are quite easy, we just negate the LNEG bit */
 
   case O_ATOM:
      if(neg != 0) w->ffo_bits ^= O_LNEG;
	 return;
 
/* for wff types, reverse the quantifiers, negate the expression contained */
 
   case O_WFF:
      if(neg != 0) w->ffo_bits ^= O_LNEG;
      if((w->ffo_bits & O_LNEG) != 0) {
	 for(q = w->ffo_qlist; q != 0; q = q->ffo_next) {
	    if(fof_whatop(q) == O_ALL) {fof_setop(q,O_EX);}
	    else {fof_setop(q,O_ALL);}}}
      fof_op_iffimpneg(w->ffo_expr,w->ffo_bits & O_LNEG);
      w->ffo_bits &= ~O_LNEG;
      return;
 
   ffdefault:
      return;}
}
 
/*-------------------------------------------------------*/
/* fof_op_standardize()            ..................... #11
/*
/* fof_op_standardize1 makes sure that all variables in a wff are
/* distinct.  Since all WFF sub-expressions have been st'dized
/* (by the parser) we need only concern ourselves with IFFs.
/* Because the IFF expansion generates a copy
/* (a<=>b: a=>b & b<=a) we must standardize.
/*
/* called by:  fof_op_clause()      ops.c
/*
/* calls:      fof_change()         ops.c
/*
/* returns:    nil
/*-------------------------------------------------------*/
 
void fof_op_standardize(w)
ffopnode *w;
{
ffopnode *q;
    
if((int) w == 0) return;
 
switch((int) (w->ffo_bits & O_FMT)) {
   case O_ATOM:
   case O_QUAN:
      break;
   case O_WFF:
      for(q = w->ffo_qlist; q != 0; q = q->ffo_next) {
               //printf("change var %d with %d\n",q->o_var,vid); 
	       fof_change(w->ffo_expr,q->ffo_var,ffvid);
	       q->ffo_var = (ffopnode *) ffvid;
	       ffvid++;}
      fof_op_standardize(w->ffo_expr);
      break;
   case O_BIN:
      fof_op_standardize(w->ffo_lhs);
      fof_op_standardize(w->ffo_rhs);
      break;
   default:
      break;}
return;
}
 
/*----------------------------------------------------------*/
/* fof_change(w,v,u)               ................... #12
/*
/* called by:  fof_op_standardize()       ops.c
/* 
/* calls:      nobody
/*
/* returns:    nil
/*----------------------------------------------------------*/
 
void fof_change(w,v,u) /* change all v to u in atoms of wff w */
ffopnode *w;
int v,u;
{
int i,p;
 
if((int) w == 0) return;
 
switch((int) (w->ffo_bits & O_FMT)) {
   case O_ATOM:
      p = (int) w->ffo_pred;
      for(i = ffcode[p++] & FARGMASK; i > 0; i--) {
         i += ffcode[p] & FARGMASK;
         if((ffcode[p] & FXNMASK) == 0 && (ffcode[p]>>FARGSHIFT) == v) {
	    ffcode[p] = u << FARGSHIFT;}
	 p++;}
      break;
   case O_QUAN:
      break;
   case O_WFF:
      fof_change(w->ffo_expr,v,u);
      break;
   case O_BIN:
      fof_change(w->ffo_lhs,v,u);
      fof_change (w->ffo_rhs,v,u);
      break;
   default:
      break;}
return;
}
 
/*-----------------------------------------------------------*/
/* fof_op_cnform()                ..............  #13
/*
/* fof_op_cnform will convert an expression that consists of only 
/* & and |      operators (no WFF,QUAN opnodes) into a conjunction
/* of disjunctions.
/*
/* the final form will be a tree where all AND operators appear
/* above the OR operators (in a particular sub-tree).i.e. once
/* an OR operatoris encountered the remaining operators(lower
/* down) are OR operators.
/*
/* called by:  fof_op_clause()             ops.c
/*
/* calls:      fof_atom()                  ops.c
/*             fof_whatop()                ops.c
/*             fof_op_alloc()              ops.c
/*             fof_op_copy()               ops.c
/*
/* returns:    nil
/*----------------------------------------------------------*/
 
void fof_op_cnform(w)
ffopnode *w;
{
ffopnode *o;
 
if((int) w==0 || (fof_atom(w)!=0)) return;  /*stop on at units */
 
/*  first thing to do is guarantee lhs, rhs cnform condition */
 
fof_op_cnform(w->ffo_lhs);                   /* convert LHS        */
fof_op_cnform(w->ffo_rhs);                 /* convert RHS          */
if(fof_whatop(w) == O_AND) {
	 return;}         /* if main c'nective = &, done     */
 
/*-----------------------------------------------------------*/
/* the main connective is an OR, if the LHS and RHS have an
/* AND connective then we must distribute to promote the AND
/*
/*      RHS:            atom            ~atom
/*      LHS
/*      atom            return          return if
/*                                      RHS OP==OR
/*
/*      ~atom           return if       return if
/*                      LHS OP==OR      both OPs == OR
/*----------------------------------------------------------*/
  
if(fof_atom(w->ffo_lhs) != 0) {
   if((fof_atom(w->ffo_rhs) != 0) || fof_whatop(w->ffo_rhs) == O_OR) {
      return;}}
else {
   if(fof_atom(w->ffo_rhs) != 0) {
      if(fof_whatop(w->ffo_lhs) == O_OR) {
	 return;}}
   else
      if(fof_whatop(w->ffo_lhs) == O_OR && fof_whatop(w->ffo_rhs)==O_OR) {
	 return;}}
	
/*-----------------------------------------------------------*/
/* make sure that we have & operator on RHS. Note that we are
/* assured of the fact that there is at least one & operator
/* in the children
/*-----------------------------------------------------------*/
	
if((fof_atom(w->ffo_rhs) != 0) || fof_whatop(w->ffo_rhs) == O_OR) {
   o = w->ffo_lhs;
   w->ffo_lhs = w->ffo_rhs;
   w->ffo_rhs = o;}
	
/*------------------------------------------------------*/
/*
/* suppose we have E1 | [E2 & E3]:
/*
/*                      |
/*              E1              &
/*                           E2   E3
/*
/* we turn this into [E1 | E2] & [E1 & E3]:
/*
/*                      &
/*              |               |
/*           E1   E2         E1   E3
/*
/* -----------------------------------------------------*/
    
o = fof_op_alloc();
o->ffo_bits = O_BIN | O_OR;                    /* LHS OR     */
o->ffo_lhs = w->ffo_lhs;                     /* original LHS   */
o->ffo_rhs = w->ffo_rhs->ffo_lhs;               /* o = E1 | E2   */
 
w->ffo_lhs = o;                             /* LHS of W      */
w->ffo_bits = (w->ffo_bits & ~O_OP) | O_AND;  /* (E1|E2) & ... */
 
o = w->ffo_rhs;                                /* fix RHS    */
o->ffo_bits = (o->ffo_bits & ~O_OP) | O_OR;
o->ffo_lhs = fof_op_copy(w->ffo_lhs->ffo_lhs); /* (E1|E2) & (E1|E3)  */
 
/*---------------------------------------------------------*/
/* we have distributed E1 to E2 and E3, we now must re-convert
/* each side of the main & connective since we have demoted
/* an | operator which now might be sandwiched between some
/* & operators
/*---------------------------------------------------------*/
	
fof_op_cnform(w);
return;
}
 
/*-----------------------------------------------------------*/ 
/* The following routines handle dependencies of variables.
/* This is very much like the lexic level stack for
/* variables and so the structures are the same
/*-----------------------------------------------------------*/ 
 
 
 
/*-----------------------------------------------------------*/
/* fof_skolinit()             ..................... #14
/*
/* called by:  fof_op_clause()         ops.c 
/*
/* calls:      nobody
/*
/* returns:    nil
/*-----------------------------------------------------------*/
 
void fof_skolinit()
{
int i;
    
ffdpsp = -1;
ffdlx = -1;
for(i = 0; i < DPMAX; i++) {ffsklist[i] = -1;}  /* -1 = no skolemize */ 
return;
}
 
/*---------------------------------------------------------*/
/* fof_op_skolemize()            ................... #15
/*
/* fof_op_skolemize will remove all WFF and QUAN opnodes
/* from an expression. existentially quantified variables are
/* converted to functions that depend on the universally
/* quantified variables that preceed it.
/*
/* we must be carefull to preserve all dependencies. For example:
/*
/* @x,y!z: [@i!k : p(i,k)] & [!h : q(x,y,h)].
/*
/* z depends on x and y. k depends on x,y and i.
/* h depends on x and y.
/*
/* skolem functions are of the form $xxxx where xxxx is a
/* number. We create a table of pointers, indexed by variable
/* number, where a non NULL entry points to the skolem function
/* for the identifier.
/*
/* NOTE: since the variables have been standardized apart,
/* all variables start at VOFFSET!!!!
/*
/* called by:  fof_op_clause()        ops.c
/*
/* calls:      fof_atom()             ops.c
/*             fof_whatop()           ops.c
/*             fof_pf_getnode()       pfsym.c
/*             fof_pf_add()           pfsym.c
/*             fof_emit_code()        emit.c
/*             fof_op_release()       ops.c
/*
/* returns:     nil
/*-----------------------------------------------------------*/ 
 
void fof_op_skolemize(w)
ffopnode *w;
{
ffopnode *q, *p;
ffpfnode *ffpf;
int i;
 
if(w == 0) {return;}
 
if(fof_atom(w) != 0) {return;}
 
/*------------------------------------------------------------*/
/* if we hit a binary operator (i.e. not a WFF) we skolemize
/* each side of the expression. All that we are interested in
/* is the WFF nodes so that we can record the skolem functions
/* and get rid of the WFF and QUAN nodes
/*------------------------------------------------------------*/
	
if((w->ffo_bits & O_FMT) == O_BIN) {
   fof_op_skolemize(w->ffo_lhs);
   fof_op_skolemize(w->ffo_rhs);
   return;}
  
/*------------------------------------------------------------*/
/* we have hit a WFF node (nodes are either BIN or WFF at
/* this stage). we scan the qlist adding universally
/* quantified variables to the dependency stack and
/* recording any skolem replacements to be made to
/* the existentially quantified variables. at the same
/* time we remove the qlist and then transform the
/* WFF node into the node of o_expr that it points to
/*------------------------------------------------------------*/
 
ffdplex[++ffdlx] = ffdpsp;                   /* new dependency level */
for(q = w->ffo_qlist; q != 0;) {
   p = q;
   q = q->ffo_next;                     /* advance to next var   */
   if(fof_whatop(p) == O_ALL) {
      ffdpstack[++ffdpsp] = (int) p->ffo_var;}
         /* save dependency */
   else {
      ffsklist[((int) p->ffo_var) - VOFFSET] = ffcodeic;
      sprintf(ffskname,"SK%ld",ffskolemid++);
      ffpf = fof_pf_getnode(ffskname);
      ffpf->ffp_nargs = ffdpsp + 1;         /* args are dependencies */
      ffpf->ffp_bits |= P_FUNC;
      fof_pf_add(ffpf);
      fof_emit_code(FXNMASK + (ffpf->ffp_id << FARGSHIFT) + ffpf->ffp_nargs);
      for(i = 0; i <= ffdpsp; i++) {
	 fof_emit_code(ffdpstack[i] << FARGSHIFT);}}
   fof_op_release(p);}                          /* get rid of VAR */
 
/*-----------------------------------------------------------*/
/* we now skolemize the sub-expression. then copy the main
/* connective of the sub-expression into the WFF node, then
/* release the sub-expression node
/*-----------------------------------------------------------*/
 
q = w->ffo_expr;
fof_op_skolemize(q);
w->ffo_bits = q->ffo_bits;           /* transform WFF node       */
w->ffo_lhs = q->ffo_lhs;
w->ffo_rhs = q->ffo_rhs;
fof_op_release(q);
ffdpsp = ffdplex[ffdlx--];             /* release dependencies     */
return;
}
 
/*-----------------------------------------------------------*/
/* fof_op_compare(o,w)
/*
/* tests literals o and w to see if they are identical
/*
/* called by:
/*
/* calls:
/*
/* returns:  1 if identical, o therwise
/*
/*------------------------------------------------------------*/

int fof_op_compare(o,w)
ffopnode *o, *w;
{
}


/*-----------------------------------------------------------*/
/* fof_op_emit(o)                ................... #16
/*
/* op_emit will dump an entire expression. Skolem functions
/* are substituted for variables if corresponding entry 
/* in the substitution list is not null.
/*
/* called by:  fof_op_clause()          ops.c
/*
/* calls:      fof_print_predicate()    ops.c
/*
/* returns:    nil
/*-----------------------------------------------------------*/
 
void fof_op_emit(o,gprint)
ffopnode *o;
int gprint;
{
ffopnode *q;

switch((int) (o->ffo_bits & O_FMT)) {
   case O_ATOM:
      if((o->ffo_bits & O_LNEG) != 0) {fof_print0("~",gprint);}
      fof_print_predicate(o->ffo_pred,gprint);
      break;
   case O_QUAN:
      if((o->ffo_bits & O_OP) == O_ALL)  {fof_print0("@",gprint);}
      else  {fof_print0("!",gprint);}
      //printf("V%ld",o->o_var);
      if(gprint == 0) ffclause_form_ptr += sprintf(ffclause_form_ptr,"V%ld",o->ffo_var);
      break;
   case O_WFF:
      if((o->ffo_bits & O_LNEG) != 0) {
	       fof_print0("~",gprint);}
      print0("[",gprint);
      for(q = o->ffo_qlist; q != 0; q = q->ffo_next) {
	       fof_op_emit(q,gprint);}
      fof_print0(":",gprint);
      fof_op_emit(o->ffo_expr,gprint);
      fof_print0("]",gprint);
      break;
   case O_BIN:
      fof_op_emit(o->ffo_lhs,gprint);
	    
      switch ((int) (o->ffo_bits & O_OP)) {
	 case O_IFF:     
	    fof_print0(" <=> ",gprint);
	    break;
	 case O_IMP:     
	    fof_print0(" => ",gprint);
	    break;
	 case O_AND:     
	    if(gprint == 0) {
               ffClauseEnd[ffclause_count - 1]  = ffclause_form_ptr;
               if(fftneg) ffClauseWeight[ffclause_count - 1] = 5000 + 100*fflit_count 
	                + fffxn_count;
               else ffClauseWeight[ffclause_count - 1] = 100000 + 
	               100*fflit_count + fffxn_count;
	       fflit_count = 1;
	       fffxn_count = 0;
	       ffclause_count++;}
	    fof_print0("\n",gprint);
	    break;
	 case O_OR:      
	    fof_print0(" | ",gprint);
            if(gprint == 0) fflit_count++;
	    break;
	 default:        
	    fof_print0(" ? ",gprint);
	    break;}
      fof_op_emit(o->ffo_rhs,gprint);
   default:
      break;}
return;
}
 
/*----------------------------------------------------------*/
/* fof_print_predicate()           ................. #17
/*
/* print_predicate will print a predicate in the code space.
/* We use the fidtbl and pitbl to print the function names
/*
/* called by:  fof_op_emit()         ops.c
/*
/* calls:      fof_print_argument()  ops.c
/*
/* returns:     nil
/*----------------------------------------------------------*/
 
void fof_print_predicate(p,gprint)
int p;
int gprint;
{
int w;

if((p & FNEGMASK) != 0) {fof_print0("~",gprint);}
 
//printf(
//   "%s(",ffpidtbl[(ffcode[p] & FIDMASK) >> FARGSHIFT]->ffp_name);
if(gprint == 0) ffclause_form_ptr += sprintf(ffclause_form_ptr,
   "%s(",ffpidtbl[(ffcode[p] & FIDMASK) >> FARGSHIFT]->ffp_name); 
for(w = ffcode[p++] & FARGMASK; w > 0; w--) {
   p = fof_print_argument(p,gprint);
   if(w != 1) {
      if(gprint == 0) fffxn_count++;
      fof_print0(",",gprint);}}

if(gprint == 0) fffxn_count++;
fof_print0(")",gprint);

return;
}
 
/*--------------------------------------------------------*/
/* fof_print_argument(p,gprint)        ................. #18
/*
/* called from:  fof_print_predicate()       ops.c
/*
/* calls:        nobody
/*      
/* returns:      pointer to next argument
/*--------------------------------------------------------*/
 
int fof_print_argument(p,gprint)
int p, gprint;
{
int a;
static char var_print[] = "xyzuvwXYZUVWpqrstPQRSTijklmnIJKLMNABCDEFGHabcdefgh123456789";

if((ffcode[p] & FXNMASK) == 0) {
   a = (ffcode[p] >> FARGSHIFT) - (1 - gprint) * VOFFSET;    /* isolate ID */
   if(ffsklist[a] == -1) {
//    printf("%c",var_print[a]);
      if(gprint == 0) ffclause_form_ptr += sprintf(ffclause_form_ptr,"%c",var_print[a]);}
   else  {
      fof_print_skolem_fxn(ffsklist[a],gprint);}
   return(++p);}
    
//printf(
//   "%s(",fffidtbl[(ffcode[p] & FIDMASK) >> FARGSHIFT]->ffp_name);
if(gprint == 0) ffclause_form_ptr += sprintf(ffclause_form_ptr,
   "%s(",fffidtbl[(ffcode[p] & FIDMASK) >> FARGSHIFT]->ffp_name); 
	
for(a = ffcode[p++] & FARGMASK; a > 0; a--) {
   p = fof_print_argument(p,gprint);
   if(a != 1) {
      if(gprint == 0) fffxn_count++;
      fof_print0(",",gprint);}}
if(gprint == 0) if(*(ffclause_form_ptr - 1) != '(') fffxn_count++;

fof_print0(")",gprint);
	
return(p);
}

/*------------------------------------------------------*/
/* int fof_print_skolem_fxn()
/*------------------------------------------------------*/

int fof_print_skolem_fxn(p,gprint)
int p,gprint;
{
int a;
static char var_print[] = "xyzuvwXYZUVWpqrstPQRSTijklmnIJKLMNABCDEFGHabcdefgh123456789";

if((ffcode[p] & FXNMASK) == 0) {
   a = (ffcode[p] >> FARGSHIFT) - (1 - gprint) * VOFFSET;    /* isolate ID */
   if(ffsklist[a] == -1) {
//    printf("%c",var_print[a]);
      if(gprint == 0) ffclause_form_ptr += sprintf(ffclause_form_ptr,"%c",var_print[a]);}
   else  {
      p = fof_print_argument(ffsklist[a],gprint);
      p--;}
   return(++p);}
    
//printf(
//   "%s(",fffidtbl[(ffcode[p] & FIDMASK) >> FARGSHIFT]->ffp_name);
if(gprint == 0) ffclause_form_ptr += sprintf(ffclause_form_ptr,
   "%s(",fffidtbl[(ffcode[p] & FIDMASK) >> FARGSHIFT]->ffp_name); 
	
for(a = ffcode[p++] & FARGMASK; a > 0; a--) {
   p = fof_print_argument(p,gprint);
   if(a != 1) {
      if(gprint == 0) fffxn_count++;
      fof_print0(",",gprint);}}
if(gprint == 0) if(*(ffclause_form_ptr - 1) != '(') fffxn_count++;

fof_print0(")",gprint);
	
return(p);



}



/*------------------------------------------------------*/
/* fof_void op_sort()
/*
/* sorts clauses by length
/*
/* called by: main_wff()
/*
/* calls:     nobody
/*
/* returns:   nil
/*--------------------------------------------------------*/

void fof_op_sort()
{
int i, j, max_score, min_score_index;
char *cptr;
int print_nc;

/*------------------------------*/
/* first fix clauseStart        */
/*------------------------------*/

print_nc = 0;
for(i = 1; i < ffclause_count; i++) {
   if(ffClauseStart[i] == 0) {
      ffClauseStart[i] = ffClauseEnd[i - 1];
      ffClauseStart[i]++;}}

ffOutPtr += sprintf(ffOutPtr,";\n");

//for(i = 0; i < 80; i++) {printf("%c",ffresults[i]);}

/*
printf("\nClause Stats:...\n\n");
printf("  ##    Start      End        Weight\n\n");
for(i = 0; i < ffclause_count; i++) {
   printf("%4d %10x %10x %10d %c %c\n",
      i,(int) ffClauseStart[i],(int) ffClauseEnd[i], ffClauseWeight[i],
      *ffClauseStart[i], *(ffClauseEnd[i]-1));}
printf("\n\n");

for(i = ffclause_count - 1; i >= 0; i--) {
   if(i + 1 == print_nc) {
      printf("\nnegated_conclusion\n\n");}
   for(j = 0; j < ffclause_count; j++) {
      if(ffClauseRank[j] != i) {continue;}
      for(cptr = ffClauseStart[j]; cptr <= ffClauseEnd[j]; cptr++) {
	 printf("%c",*cptr);}}}
*/

/*----------------------------------*/
/* assign new score based on weight */
/*----------------------------------*/

for(i = 0; i < ffclause_count; i++) {
   max_score = 10000000;
   for(j = 0; j < ffclause_count; j++) {
      if(ffClauseWeight[j] > 0 && ffClauseWeight[j] < max_score) {
         min_score_index = j;
	 max_score = ffClauseWeight[j];}}
   if(max_score < 10000) print_nc++;
   ffClauseWeight[min_score_index] = -ffClauseWeight[min_score_index];
   ffClauseRank[min_score_index] = i;}

//printf("\nClause Stats 2:...\n\n");
//printf("  ##    Start      End        Weight ffclause_count=%d\n\n",ffclause_count);
for(i = 0; i < ffclause_count; i++) {
   ffClauseWeight[i] = -ffClauseWeight[i] + 1;
//   printf("%4d %10x %10x %10d %10d %c\n",
//      i,(int) ffClauseStart[i],(int)ffClauseEnd[i],ffClauseRank[i],
//             ffClauseWeight[i]);
            }


for(i = ffclause_count - 1; i >= 0; i--) {
   if(i + 1 == print_nc) {
      if(first_attempt == 0) printf("\nnegated_conclusion\n\n");
      ffOutPtr += sprintf(ffOutPtr,"\nnegated_conclusion\n\n");}
   for(j = 0; j < ffclause_count; j++) {
      if(ffClauseRank[j] != i) {continue;}
      for(cptr = ffClauseStart[j]; cptr <= ffClauseEnd[j]; cptr++) {
         if(first_attempt == 0) printf("%c",*cptr);
         ffOutPtr += sprintf(ffOutPtr,"%c",*cptr);}}}

first_attempt = 1;

printf("\n\n");

}

/*---------------------------------------------------------------*/
/* fof_print_tree()
/*
/*---------------------------------------------------------------*/

void fof_print_tree(w,gprint)
ffopnode *w;
int gprint;
{
static char var_prt[] = "xyzuvwXYZUVWpqrstPQRSTijklmnIJKLMNABCDEFGHabcdefgh123456789";
ffopnode *o;

return;
printf("\nNode w: %x %lx %x %x",w,w->ffo_bits,w->ffo_lhs,w->ffo_rhs);
if(w->ffo_bits & O_BIN) {
   printf("  BIN ");
   if((w->ffo_bits & O_OP) == O_IFF) printf(" <=> ");
   if((w->ffo_bits & O_OP) == O_IMP) printf(" => ");
   if((w->ffo_bits & O_OP) == O_OR ) printf(" | ");
   if((w->ffo_bits & O_OP) == O_AND) printf(" & ");}
if(w->ffo_bits & O_ATOM) {
   printf("  ATOM ");
   if((w->ffo_bits & O_LNEG) != 0) {printf("~");}
   fof_print_predicate((int)w->ffo_pred,gprint);}

if(w->ffo_bits & O_WFF) printf("  WFF ");
if(w->ffo_bits & O_QUAN) {
   printf("  QUAN ");
   if((w->ffo_bits & O_OP) == O_EX) 
          printf(" !%c ",var_prt[(int)w->ffo_rhs -(1-gprint)*1000]);
   if((w->ffo_bits & O_OP) == O_ALL) 
          printf(" @%c  ",var_prt[(int)w->ffo_rhs -(1-gprint)*1000]);
   o = w;
   while(o->ffo_lhs != 0) {
      o = o->ffo_lhs;
      printf(",%c ",var_prt[(int)o->ffo_rhs -(1-gprint)*1000]);}}
if((w->ffo_bits & O_ATOM) == O_ATOM) {printf(" BU ");spaces--;spaces--;return;}
if((w->ffo_bits & O_QUAN) == O_QUAN) {printf(" BU ");spaces--;spaces--;return;}
printf(" ADV ");spaces++;spaces++;printf("SPACES=%d ",spaces);
if(w->ffo_lhs != 0) fof_print_tree(w->ffo_lhs,gprint);
if(w->ffo_rhs != 0) fof_print_tree(w->ffo_rhs,gprint);
printf(" BU ");spaces--;spaces--;printf("SPACES=%d ",spaces);return;
}
