/*-------------------------------------------------------------*/
/*
/* pfv.c          predicate/function/var/tree symbol table stuff
/*
/* Copyright Newborn Software 1993, 1996, 1999, 2004
/*
/*  fof_pf_init()............................#1
/*  fof_pf_alloc()...........................#2
/*  fof_pf_initsym().........................#3
/*  fof_pf_getnode().........................#4
/*  fof_pf_add().............................#5
/*  fof_pf_find()............................#6
/*
/*                     variable symbol table
/*
/*  fof_v_init()..........................#1
/*  fof_v_lexin().........................#2
/*  fof_v_lexout()........................#3
/*  fof_v_find()..........................#4
/*  fof_v_add()...........................#5
/*
/*              threaded binary tree routines
/*
/*  fof_tree_add().............................#1
/*  fof_tree_find()............................#2
/*
/* emit.c              atomic emit stuff
/*
/* Copyright Newborn Software 1993, 1996, 1999, 2004
/*
/* functions:
/*  fof_emitinit().......................#1
/*  fof_emit_code()......................#2
/*  fof_print0().........................#3
/*  fof_yerror()........................ #1
/*
/*-----------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "compwff.h"
#include "clause.h"
#include "const.h"
 
/*-----------------------------------------------------------*/
/* externally defined constants
/*-----------------------------------------------------------*/

extern char *ffclause_form_ptr;
extern int ffyylineno;

/*-----------------------------------------------------------*/
/* internally defined constants
/*-----------------------------------------------------------*/

typedef struct fftn {struct fftn *ffllink,*ffrlink; /* child links   */
                   int ffbits;                      /* bits field  */
                   char *ffkey;} fftnode;
 
int ffcode[CODESIZE],ffcodeic;

/*-----------------------------------------------------------*/
/* internally defined functions
/*-----------------------------------------------------------*/

void fof_emitinit();
void fof_emit_code();
void fof_print0();
void fof_yerror();
void fof_pf_init();
ffpfnode *fof_pf_alloc();
void fof_pf_initsym();
ffpfnode *fof_pf_getnode();
void fof_pf_add();
ffpfnode *fof_pf_find();
void fof_v_init();
void fof_v_lexin();
void fof_v_lexout();
ffvnode *fof_v_find();
fftnode *fof_tree_add();
fftnode *fof_tree_find();

 
/*------------------------------------------------------------*/
/* internally defined constants
/*------------------------------------------------------------*/

/*----------------------------------------------------------------*/
/* these routines will handle threaded binary tree manipulations.
/* the node types defined here are simply required header for any
/* nodes to be linked into a tree. As long as the first 4 fields in
/* the node are the same these routines will work.
/*
/* Note: the bits field is open for user nodes excepty that the 1st
/* 2 bits (least significant) are used for thread indicators and must
/* not be fiddled with
/*----------------------------------------------------------------*/

static ffpfnode ffpfspace[MAXID];     /* pf symbol records go here*/
static ffpfnode *ffpfree;           /* available ffpfnodes        */
static int ffpfcount;              /* # ffpfnodes free            */
static ffpid, fffid;              /* unique ID generator          */
ffpfnode *ffpidtbl[MAXID+1],*fffidtbl[MAXID+1];
char ffnbuff[MAXCHAR],*ffnbp;           /* ffpfnode names go here */
ffpfnode *ffpftree;                               /* symbol table */

ffvnode ffvstack[MAXVARS];                /* var names go here    */
int ffvsp;                               /* top of vstack index   */
int ffvdisplay[MAXLEX],ffvdx;            /* lex display pointers  */
int ffnextvid;                           /* unique ID generator   */

/*-------------------------------------------------------------*/
/* fof_pf_init()                ......................#1
/*
/* pf_init	will initialize the pfspace so that it
/* consists of a linked list of ffpfnodes (llink).
/*
/* called by:  fof_pf_initsym()          pfsym.c
/*
/* calls:      nobody
/*
/* returns:	   nil
/*-------------------------------------------------------------*/
 
void fof_pf_init()
{
int p;
    
ffpfree = 0;
for(p = 0; p < MAXID; p++) {
   ffpfspace[p].ffp_llink = ffpfree;    /* create chain off llink   */
   ffpfree = &ffpfspace[p];       /* link all ffpfnodes together    */
   ffpidtbl[p + 1] = 0;
   fffidtbl[p + 1] = 0;}
    
ffpfcount = MAXID;                      /* # pfnodes left       */
ffpid = 1;                              /* PIDs start at 1      */
fffid = 1;                              /* FIDs start at 1      */
}
 
/*----------------------------------------------------------*/
/* fof_pf_alloc()                   .....................#2
/*
/* will allocate and return a pfnode. if none are left we return
/* a NULL pointer
/*
/* called by:  fof_pf_getnode()           pfsym.c
/*
/* calls:      nobody
/*
/* returns:    pointer to predicate/function
/*------------------------------------------------------------*/
 
ffpfnode *fof_pf_alloc()
{
ffpfnode *pf;
    
pf = ffpfree;
if((int) pf != 0) {                /* return top element    */
   ffpfree = pf->ffp_llink;
   ffpfcount--;}
return(pf);
}
 
/*-----------------------------------------------------------*/
/* fof_pf_initsym()             ............................#3  
/*
/* will initialize the pfnode symbol table
/*
/* called by:  main_wff()               compwff.c
/*
/* calls:      fof_pf_init()
/*
/* returns:    nil
/*----------------------------------------------------------*/
 
void fof_pf_initsym()
{
ffnbp = ffnbuff;                          /* name buffer    */
fof_pf_init();                          /* create pfnodes   */
ffpftree = 0;
}
 
/*-----------------------------------------------------------*/
/* fof_pf_getnode()                   ..................#4
/*
/* will allocate and return an initialized pfnode.the default
/* pfnode is a predicate with 0 arguments.
/*
/* parameters:     name : ^ asciz string that is name
/* returns:        pfnode
/*
/* note: the name is copied into a local character
/* array since it probably comes from yytext
/* and is likely to disappear.
/*
/* called by:  fof_op_skolemize()         ops.c
/*             fof_yyparse()              compwff.c
/*             fof_pid_gen()              compwff.c
/*				
/* calls:      fof_pf_alloc()             pysym.c
/*             fof_yerror()               err.c
/*
/* returns:    pointer to pf name
/*-----------------------------------------------------------*/
 
ffpfnode *fof_pf_getnode(ffname)
char *ffname;
{
ffpfnode *p;
    
if((int) ((p = fof_pf_alloc())) == 0) {fof_yerror(E_PFSYMFULL); exit(0);}
	
p->ffp_bits = 0;
p->ffp_nargs = -1;          /* undefined number of arguments   */
p->ffp_name = ffnbp;
do *ffnbp++ = *ffname; while (*ffname++ != '\0');
 
if(ffnbp - &ffnbuff[MAXCHAR] >= 0) {fof_yerror(E_NAMEFULL);exit(0);}
 
return(p);
}
 
/*--------------------------------------------------------*/
/* fof_pf_add()            ............................#5
/*
/* fof_pf_add will add an identifier to the symbol table
/* (assuming that it is not already in it). A unique
/* ID is also given. We stuff the pointer to the pfnode
/* into the fid/pid tables to be indexed by ID
/*
/* called by:  fof_op_skolemize()           ops.c
/*             fof_yyparse()                compwff.c
/*             fof_pid_gen()                compwff.c
/*		
/* calls:      fof_tree_add()               tree.c
/*
/* returns:    nil
/*----------------------------------------------------------*/
 
void fof_pf_add(p)
ffpfnode *p;
{
if((p->ffp_bits & P_TYPE) == P_PRED) {
   p->ffp_id = ffpid;
   ffpidtbl[ffpid++] = p;} 
else {
   p->ffp_id = fffid;
   fffidtbl[fffid++] = p;}
 
ffpftree = fof_tree_add(ffpftree,p);
}
 
/*----------------------------------------------------------*/
/* fof_pf_find()               ..........................#6
/*
/* fof_pf_find will return a pointer to the pfnode if there exists
/* one with a passed name. NULL if none found
/*
/* called by:  fof_yyparse()             compwff.c
/*             fof_pid_gen()             compwff.c
/*				
/* calls:      fof_pf_find()             pfsym.c
/*             fof_yerror()              err.c
/*
/* returns:    fof_tree_find(pftree,name)
/*--------------------------------------------------------*/
 
ffpfnode *fof_pf_find(ffname)
char *ffname;
{
return(fof_tree_find(ffpftree,ffname));
}
 
/*----------------------------------------------------------*/
/* fof_v_init()           ..........................#1
/*
/* will initialize the variable name stack structures. This
/* should be called whenever a wff is about to be parsed 
/*
/* called by: fof_yyparse()      compwff.c
/*
/* calls:     nobody
/*
/* returns:   nil
/*----------------------------------------------------------*/
 
void fof_v_init()
{
ffvsp     = -1;                         /* nothing in vstack */
ffvdx     = -1;
ffnextvid = 0;
}
 
/*-----------------------------------------------------------*/
/* fof_v_lexin()         ..........................#2
/*
/* will start a new variable lex level
/*
/* called by:  fof_yyparse()      compwff.c
/*
/* calls:      fof_yerror()       err.c
/*
/* returns:    nil
/*-----------------------------------------------------------*/
 
void fof_v_lexin()
{
if(++ffvdx >= MAXLEX) {fof_yerror(E_LEXTOODEEP);exit(0);}
ffvdisplay[ffvdx] = ffvsp;
}
 
/*-----------------------------------------------------------*/
/* fof_v_lexout()        ..........................#3
/*
/* will delete all variables at the innermost lex level
/*
/* called by:  fof_yyparse()     compwff.c
/*
/* calls:      nobody
/*
/* returns:    nil
/*-----------------------------------------------------------*/
 
void fof_v_lexout()
{
if(ffvdx >= 0) ffvsp = ffvdisplay[ffvdx--];
}
 
/*-----------------------------------------------------------*/
/* fof_v_find()         ..........................#4
/*
/* *fof_v_find will attempt to locate a variable in the
/* variable stack. We search from the top to the bottom
/* so that we find innermost declarations first
/*
/* called by:  fof_v_add()          varsym.c
/*             fof_yyparse()        compwff.c
/*
/* calls:      nobody
/*
/* returns:        pointer to variable in variable stack if there
/*             otherwise returns a 0
/*-----------------------------------------------------------*/
 
ffvnode *fof_v_find(ffname)
char *ffname;
{
int v;
 
for(v = ffvsp; v >= 0; v--) {
   if(strncmp(ffname,ffvstack[v].ffv_name,VNAMESIZE) == 0) {
      return(&ffvstack[v]);}}
  
return(0);
}
 
/*-----------------------------------------------------------*/
/* fof_v_add()           ..........................#5
/*
/* fof_v_add will add a name to the variable symbol table.
/* If the name has already been declared at the current
/* level, we return NULL else we return a pointer the entry
/*
/* called by:  fof_yyparse()        compwff.c
/*
/* calls:      fof_v_find()         varsym.c
/*             fof_yerror()         err.c
/*
/* returns:        pointer to new variable name
/*-----------------------------------------------------------*/
 
ffvnode *fof_v_add(ffname)
char *ffname;
{
ffvnode *v;
 
if((v = fof_v_find(ffname)) != 0) {
   if(v->ffv_ll == ffvdx) {fof_yerror(E_MULTIVAR); return(0);}} 
 
if(++ffvsp >= MAXVARS) {fof_yerror(E_VSYMFULL);exit(0);}
 
v = &ffvstack[ffvsp];
v->ffv_ll = ffvdx;                      /* current lex level   */
v->ffv_id = ffnextvid++;                /* new ID number       */
strncpy(v->ffv_name,ffname,VNAMESIZE);
 
return(v);
}

/*----------------------------------------------------------------*/
/* fof_tree_add()            ..................#1
/*
/* an empty tree consists of a NULL pointer. When a record is
/* added the root pointer (tree pointer) will point to the root
/* node of the tree
/*
/* tree_add will add a node to the threaded tree. Nodes with
/* keys less than the parent node go left and greater than or
/* equal go right.
/*
/* parameters:  t : tree pointer (root node)
/*              n : node to add
/*
/* called by:  fof_pf_add()         pysym.c
/*
/* calls:      nobody
/*
/* returns:    t : updated root node pointer.
/*---------------------------------------------------------------*/
 
fftnode *fof_tree_add(t,n)
fftnode *t,*n;
{
fftnode *p;                                     /* parent pointer */
 
if(t == 0) {
   n->ffllink = 0;                         /* first node in tree  */
   n->ffrlink = 0;
   n->ffbits &= ~(LTHREAD | RTHREAD);      /* clear thread bits   */
   return(n);}
 
/* tree is not empty, we walk down the tree and insert the node */
 
for(p = t;;) {
   if(strcmp(n->ffkey,p->ffkey) < 0) {
      if(p->ffllink == 0 || p->ffbits & LTHREAD) {
         n->ffrlink = p;
         n->ffllink = p->ffllink;
         n->ffbits |= RTHREAD | (p->ffbits & LTHREAD);
         p->ffllink = n;
         p->ffbits &= ~LTHREAD;
         return(t);}
      p = p->ffllink;}
   else {
      if(p->ffrlink == 0 || p->ffbits & RTHREAD) {
         n->ffllink = p;
         n->ffrlink = p->ffrlink;
         n->ffbits |= LTHREAD | (p->ffbits & RTHREAD);
         p->ffrlink = n;
         p->ffbits &= ~RTHREAD;
         return(t);}
      p = p->ffrlink;}}
}
 
/*------------------------------------------------------------*/
/* fof_tree_find()                ....................#2
/*
/* fof_tree_find will locate a record in a threaded binary tree when
/* given a pointer to the key string. If found the pointer to
/* the record is returned else NULL
/*
/* called by:  fof_pf_find()            pysym.c
/*
/* calls:      nobody
/*
/* returns:    t : root node pointer.
/*-----------------------------------------------------------*/
 
fftnode *fof_tree_find(t,k)
fftnode *t;
char *k;
{
int c;
 
while(t != 0) {
   if((c = strcmp(k,t->ffkey)) == 0) return(t);  /* found it! */
   if(c < 0) {
      if(t->ffllink == 0 || t->ffbits & LTHREAD)  break;
      t = t->ffllink;}
   else {
      if(t->ffrlink == 0 || t->ffbits & RTHREAD)  break;
      t = t->ffrlink;}}
    
return(0);
}

/*--------------------------------------------------*/
/* fof_emitinit()        .......................#1
/*
/* initialize stuff for code emission
/*
/* called by:  fof_yyparse()       compwff.c
/*
/* calls:      nobody
/*
/* returns:    nil
/*-------------------------------------------------*/
 
void fof_emitinit()
{
ffcodeic = 0L;
}
 
/*--------------------------------------------------*/
/* fof_emit_code()  	 .......................#2
/*
/* emits a word of code into the codespace
/*
/* called by:  fof_op_skolomize()          ops.c
/*             fof_op_copy()               ops.c
/*             fof_pid_gen()               compwff.c
/*             fof_yyparse()               compwff.c
/*
/* calls:      fof_yerror()                err.c
/*
/* returns:     nil
/*----------------------------------------------------*/
 
void fof_emit_code(c)
int c;
{
 
if(ffcodeic >= CODESIZE) {
   fof_yerror(E_CODESIZE);
   exit(0);}
 
ffcode[ffcodeic++] = c;
return;
}

/*---------------------------------------------------------------*/
/*      fof_print0()         .......................#7
/*
/*      prints a string to the screen and to the disk
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/

void fof_print0(msg)
char *msg;
{
//printf("%s",msg);
ffclause_form_ptr += sprintf(ffclause_form_ptr,"%s",msg);
}

static char *ffmsgtbl[] = {
        "no error",
        "predicate/function symbol table full",
        "predicate/function character array full",
        "identifier expected",
        "multiple definition of identifier",
        "illegal character in source",
        "bad include statement syntax",
        "cannot open include file",
        "identifier undefined",
        "incorrect number of arguments",
        "syntax error in identifier type specification",
        "too many arguments declared",
        "predicate expected",
        "functions must appear as arguments to predicates",
        "missing ]",
        "missing }",
        "predicate expected",
        "function expected",
        "to many nested {}s and []s",
        "variable quantified twice within [] or {}",
        "too many variables defined",
        "syntax error in expression",
        "out of memory (opspace)",
        "out of memory (codespace)",
        "missing .",
        "constants cannot have arguments",
        ""
};
 
int fferrcount = 0;    /* public so others can check err stat  */
static int ffprevmsg = -1;
static int ffprevline = -1;

/*------------------------------------------------------*/
/*     fof_yerror(m)              .................... #1
/*
/*		called by:	..........	
/*
/*		calls:		nobody
/*
/*		returns:	nil
/*-------------------------------------------------------*/
 
void fof_yerror(m)
int m;
{
    
if(ffprevline == ffyylineno && ffprevmsg == m) {
   return;}
ffprevmsg = m;
ffprevline = ffyylineno;

printf("LINE %5ld: ERROR: %s\n",ffyylineno,ffmsgtbl[m]);
ffclause_form_ptr += sprintf(ffclause_form_ptr,
                 "%5ld: %s\n",ffyylineno,ffmsgtbl[m]);

if(++fferrcount < MAXERR) {return;}

printf("%5ld: too many errors\n",ffyylineno);
ffclause_form_ptr += sprintf(ffclause_form_ptr,"%5ld: too many errors\n",ffyylineno);
exit(0);
}
