
/*
 * File Allocator.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Allocator.hpp"
#include "Property.hpp"
#include "Query.hpp"
#include "Unit.hpp"
#include "XML.hpp"
#include "VampireException.hpp"
#include "Tracer.hpp"


#define CHUNK_SIZE 1023

int GlobalAllocator::_remainingPieces = 0;
GlobalAllocator::Dummy* GlobalAllocator::_free = 0;
size_t GlobalAllocator::_totalNumberOfBytesAllocated = 0;
size_t GlobalAllocator::_memoryLimitInBytes = 0;
bool GlobalAllocator::_overflown = false;


#if DEBUG_PREPRO
void* GlobalAllocator::allocate (int words, const string& className)
#else
void* GlobalAllocator::allocate (int words)
#endif
{
  
  ASS(CHUNK_SIZE >= words);

  if (_remainingPieces < words) {
    _free = new Dummy[CHUNK_SIZE];
    if (! _free) {
      ALLOCATION_FAILURE;
    }
    _totalNumberOfBytesAllocated += CHUNK_SIZE * sizeof(Dummy);
    if (! _overflown && _memoryLimitInBytes &&
	_memoryLimitInBytes < _totalNumberOfBytesAllocated) {
      _overflown = true;
      VAMPIRE_TERMINATION("memory limit exceeded",UNKNOWN);
    }
    _remainingPieces = CHUNK_SIZE;
  }

  void* result = _free;
  _free += words;
  _remainingPieces -= words;

  return result;
} 



