
/*
 * File Allocator.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Allocator__
#define __Allocator__

#include <cstdlib>
#include <string>
#include "VampireException.hpp"
#include "Debugger.hpp"


class GlobalAllocator
{
public:
#if DEBUG_PREPRO
  static void* allocate (int words, const string& className);
#else
  static void* allocate (int words);
#endif
  
  static void setMemoryLimitInMegabytes (int lim)
  { 
    _overflown = false;
    _memoryLimitInBytes = size_t(lim) * size_t(1000000); 
  }
  static size_t getUsedMemoryInKilobytes ()
  { return _totalNumberOfBytesAllocated / 1000; }
  static size_t getUsedMemoryInBytes ()
  { return _totalNumberOfBytesAllocated; }

#if DEBUG_PREPRO
#endif
public:
  
  class Dummy 
  {
  private:
    void* _dummy;
  }; 

private:
  
  static int _remainingPieces;
  
  static Dummy* _free;
  
  static size_t _totalNumberOfBytesAllocated;
  
  static size_t _memoryLimitInBytes;
  
  static bool _overflown;
}; 


template <int WORDS>
class Allocator
{
public:
  
#if DEBUG_PREPRO
  static void* allocate (const string& className)
#else
  static void* allocate ()
#endif
  {
    if (_freeList) {
      ASS (Debugger::reallocated(_freeList,WORDS,className));

      void* _result = _freeList;
      _freeList = _freeList->_next;
      return _result;
    }

#if DEBUG_PREPRO
    void* result = GlobalAllocator::allocate(WORDS,className);
#else
    void* result = GlobalAllocator::allocate(WORDS);
#endif
    ASS (Debugger::allocated(result,WORDS,className));
    return result;
   } 

  
#if DEBUG_PREPRO
  static void release (void* obj, const string& className)
#else
  static void release (void* obj)
#endif
  {
    ASS (Debugger::released(obj,WORDS,className));

    Allocator* m = reinterpret_cast <Allocator*>(obj);
    m->_next = _freeList;
    _freeList = m;
  }

private:
  
  static Allocator* _freeList;
  
  Allocator* _next;
  friend class MyTracer;
}; 




#define PIECES(C) (sizeof(C) + sizeof(GlobalAllocator::Dummy) - 1) / \
                   sizeof(GlobalAllocator::Dummy)



#if DEBUG_PREPRO
#define USE_ALLOCATOR(C)                                         \
  void* operator new (size_t)                                    \
  { return Allocator<PIECES(C)>::allocate(className()); }        \
  void operator delete (void* obj)                               \
  { if (obj) Allocator<PIECES(C)>::release(obj,className()); }
#else
#define USE_ALLOCATOR(C)                              \
  void* operator new (size_t)                         \
  { return Allocator<PIECES(C)>::allocate(); }        \
  void operator delete (void* obj)                    \
  { if (obj) Allocator<PIECES(C)>::release(obj); }
#endif


#endif 
