
/*
 * File Atom.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Tracer.hpp"
#include "Atom.hpp"
#include "XML.hpp"
#include "Signature.hpp"



Atom::Atom (const PredicateSymbol& p, const TermList& args)
  :
  _data (new Data (p,args))
{
} 



Atom::Atom (const Term& lhs, const Term& rhs)
  :
  _data (new Data (lhs,rhs))
{
} 



Atom::Atom (const PredicateSymbol& p)
  :
  _data (new Data (p))
{
} 



AtomList::AtomList (const Atom& a)
  : Lst<Atom> ()
{
} 


Atom::Data::Data (const Term& l, const Term& r) 
  :
  _counter (1),
  _functor (Signature::sig.equality()),
  _args (TermList (l, TermList(r)))
{
} 


void Atom::operator = (const Atom& t)
{
  if (t._data) {
    t._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = t._data;
} 

void Atom::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} 





bool Atom::equals (const Atom& a) const
{
  return functor() == a.functor() &&
         args().equals(a.args());
} 






XML::Element* Atom::toXML() const
{
  if (args().isEmpty()) {
    XML::Element* a = new XML::EmptyElement("propositional_constant");
    a->addAttribute(new XML::TextAttribute("name",functor().name()));
    return a;
  }
  
  XML::DeepElement* a = new XML::DeepElement("atom");
  a->addAttribute(new XML::TextAttribute("functor",functor().name()));
  Iterator<Term> ts(args());
  while ( ts.hasNext() ) {
    a->addChild(ts.next().toXML());
  }
  return a;
} 



string Atom::toString () const
{
  if (isEquality()) {
    return args().head().toString() + "=" + args().second().toString();
  }

  
  return functor().name() + args().toString(",");
} 


