
/*
 * File CNF.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "CNF.hpp"
#include "Problem.hpp"
#include "Position.hpp"
#include "Inference.hpp"
#include "InferenceInfo.hpp"


void CNF::clausify (Problem& problem)
{
  if (! problem.hasFormulas()) {
    return;
  }
  
  UnitChain::DelIterator us (problem.giveUnits());
  while (us.hasNext()) {
    Unit u (us.next());
    if (u.unitType() == FORMULA) {
      FormulaList fs (u.formula());
      LiteralList answers;
      if (u.inputType() == CONJECTURE && ! problem.answerLiteral().isNull()) {
	answers.push (problem.answerLiteral());
      }
      clausify(problem, fs, answers, u, Position());
      us.del();
    }
  }
} 



void CNF::clausify (Problem& problem, 
		    const FormulaList& formulas, 
		    const LiteralList& literals,
		    const Unit& parent,
		    const Position& position)
{
  if (formulas.isEmpty()) {
    Inference* inf = new Inference(IR_CLAUSIFY,parent);
    inf->setInfo(new InferenceInfo(IR_CLAUSIFY,position,0));
    Clause c (literals);
    problem.giveUnits().addFirst(Unit(c, inf));
    return;
  }

  
  Formula f(formulas.head());
  FormulaList fs(formulas.tail());

  switch (f.connective()) {
  case ATOM: 
    clausify(problem, fs, 
	     LiteralList(Literal(true, f.atom()),literals), 
	     parent, position);
    return;

  case NOT: 
    clausify(problem, fs, 
	     LiteralList(Literal(false, f.uarg().atom()), literals), 
	     parent, position);
    return;

  case OR: 
    {
      FormulaList newFs(f.args());
      newFs.append(fs);
      clausify (problem, newFs, literals, parent, position);
      return;
    }

  case AND:
    {
      Iterator<Formula> args(f.args());
      int choice = 0;
      while (args.hasNext()) {
	FormulaList newFs(args.next(), fs);
	clausify(problem,newFs,literals,parent,Position(choice++,position));
      }
      return;
    }

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 

