
/*
 * File Clause.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Clause.hpp"
#include "XML.hpp"



Clause::Clause (const LiteralList& literals)
  :
  _data (new Data(literals))
{
} 



void Clause::operator = (const Clause& t)
{
  if (t._data) {
    t._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = t._data;
} 


void Clause::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} 






XML::Element* Clause::toXML() const
{
  if (literals().isEmpty()) {
    return new XML::EmptyElement("empty_clause");
  }
  
  XML::DeepElement* cls = new XML::DeepElement("clause");
  Iterator<Literal> ls(literals());
  while ( ls.hasNext() ) {
    cls->addChild(ls.next().toXML());
  }
  return cls;
} 



string Clause::toString () const
{
  if (isEmpty()) {
    return "#";
  }

  return literals().toString(" \\/ ");
} 



bool Clause::equals (const Clause& c) const
{
  Iterator<Literal> ls1(literals());
  Iterator<Literal> ls2(c.literals());
  while (ls1.hasNext()) {
    if (! ls2.hasNext()) {
      return false;
    }
    if (! ls1.next().equals(ls2.next())) {
      return false;
    }
  }
  return ! ls2.hasNext();
} 
