
/*
 * File Clause.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __Clause__
#define __Clause__


#include "Literal.hpp"


class Clause 
{
 private:
  class Data;

 public:
  
  Clause () : _data (0) {}

  explicit Clause (const LiteralList& lits);
  Clause (const Clause& t);
  ~Clause ();
  void operator= (const Clause& rhs);

  
  const LiteralList& literals () const;

  
  XML::Element* toXML () const;
  
  bool isEmpty () const { return literals().isEmpty(); }
  bool equals (const Clause&) const;

  
  string toString () const;
 private:
  
  Data* _data;
}; 


class Clause::Data
{
 public:
  Data ();
  
  explicit Data (const LiteralList& literals)
    : _counter (1), _literals (literals) {}

  
  const LiteralList& literals () const { return _literals; }
  void ref () { ASS (this); _counter++; }
  void deref ();

  
  CLASS_NAME("Clause::Data");
  USE_ALLOCATOR(Data);
 private:
  
  int _counter;
  LiteralList _literals;
}; 


class ClauseList 
: public Lst<Clause>
{
 public:
  
  ClauseList () : Lst<Clause> () {}
  
  ClauseList (const ClauseList& ts) : Lst<Clause> (ts) {}
  
  explicit ClauseList (const Clause& c) : Lst<Clause> (c) {}
  
  ClauseList (const Clause &hd, const ClauseList& tl)
    : Lst<Clause> (hd,tl) {}
  
  explicit ClauseList (LstData<Clause>* d)
    : Lst<Clause> (d) {}
  
  const LiteralList& literals ();

  
  const ClauseList& tail () const
    { return static_cast<const ClauseList&>(Lst<Clause>::tail()); }
}; 





inline
Clause::~Clause () 
{
  if (_data) {
    _data->deref ();
  }
} 



inline
Clause::Clause (const Clause& c)
  :
  _data (c._data)
{
  if (_data) {
    _data->ref ();
  }
} 


inline
const LiteralList& Clause::literals () const
{
  return _data->literals();
} 


#endif 
