
/*
 * File Compare.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Compare__
#define __Compare__


class PredicateSymbol;
class FunctionSymbol;
class Term;
class Atom;
class TermList;
class Literal;
class Clause;
class Formula;
class FormulaList;
class Unit;
class Problem;
class SymCounter;



enum Comparison
{
  LESS = -1,
  EQUAL = 0,
  GREATER = 1
};



class Compare
{
public:
  
  Compare () {}
  bool lessThan (const Unit&, const Unit&);
  bool lessThan (const Literal&, const Literal&);
  bool lessThan (const Formula& f1, const Formula& f2);
  static Comparison compare (int i1, int i2);
  static Comparison compare (const Term&, const Term&);
  static void createSymbolCounter (const Problem&);
private:
  static bool lessThan (const Clause& c1, const Clause& c2);   
  static Comparison compare (const Atom&, const Atom&);
  static Comparison compare (const Formula&, const Formula&);
  static Comparison compare (const Literal&, const Literal&);
  static Comparison compare (const TermList&, const TermList&);
  static Comparison compare (const FormulaList&, const FormulaList&);
  static Comparison compare (const PredicateSymbol&, const PredicateSymbol&);
  static Comparison compare (const FunctionSymbol&, const FunctionSymbol&);

  
  static SymCounter* _symbolCounter;
}; 



inline 
Comparison Compare::compare (int i1, int i2)
{
  if (i1 > i2)
    return GREATER;

  if (i1 == i2)
    return EQUAL;

  return LESS;
} 


#endif
