
/*
 * File EqualityAxioms.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __EqualityAxioms__
#define __EqualityAxioms__

class Term;
class TermList;
class Atom;
class Clause;
class Formula;
class Unit;

#include "Var.hpp"


class EqualityAxioms
{
public:
  static bool isEqualityAxiom (const Unit&);
private:
  static bool isEqualityAxiom (const Formula&);
  static bool isEqualityAxiom (const Clause&);
  static bool isReflexivityAxiom (const Formula&);
  static bool isReflexivityAxiom (const Clause&);
  static bool isFunctionReflexivityAxiom (const Formula&);
  static bool isFunctionReflexivityAxiom (const Clause&);
  static bool isPredicateReflexivityAxiom (const Formula&);
  static bool isPredicateReflexivityAxiom (const Clause&);
  static bool isSymmetryAxiom (const Formula&);
  static bool isSymmetryAxiom (const Clause&);
  static bool isTransitivityAxiom (const Formula&);
  static bool isTransitivityAxiom (const Clause&);
  static bool isTautology (const Atom&);
  static bool isTransitivity (const Atom& a1, const Atom& a2, const Atom& a3);
  static bool isFunctionMonotonicity (const Atom& a1, const Atom& a2);
  static bool isPredicateMonotonicity (const Atom& a1, const Atom& a2, 
				       const Atom& a3);
  static bool isSwapOf (const Atom&, const Atom&);
  static bool equalsUpTo (const Term& t1,const Term& t2,Var x,Var y);
  static bool equalsUpTo (const TermList& t1,const TermList& t2,Var x,Var y);
};


#endif 

