
/*
 * File FunctionDefinition.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __FunctionDefinition__
#define __FunctionDefinition__


class Term;
class TermList;
class Atom;
class Clause;
class Formula;
class Problem;
class Signature;
class FunctionSymbol;
class UnitList;
class Literal;
class LiteralList;


class FunctionDefinition
{
public:
  FunctionDefinition (Problem&,const Signature &);
  ~FunctionDefinition ();
  int removeAllDefinitions ();
  static bool isFunctionDefinition (const Clause&);
  static bool isFunctionDefinition (const Clause&, Term& lhs, Term& rhs);
  static bool isFunctionDefinition (const Atom&, Term& lhs, Term& rhs);
  static bool isFunctionDefinition (const Formula&);
  static bool isFunctionDefinition (const Formula&, Term& lhs, Term& rhs);

private:
  class Def;

  static bool defines (const Term& t, const Term& lhs);
  static bool occurs (const FunctionSymbol&, const Term&);
  void count (const Clause& c);
  bool unfoldAllDefs ();
  bool unfold (Def& def);
  bool unfold (Term& t,UnitList& parents);
  bool unfold (TermList& t,UnitList& parents);

  void apply (LiteralList& ls,UnitList& parents);
  void apply (Literal& l,UnitList& parents);
  void apply (TermList& ls,UnitList& parents);
  void apply (Term& l,UnitList& parents);

  
  Problem& _problem;
  
  int _numberOfFunctions;
  
  Def* _defs;
  
  int* _counter;
  
  int _removed;
}; 




#endif
