
/*
 * File GapFiller.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __GapFiller__
#define __GapFiller__

#include "Connective.hpp"

class Unit;
class Formula;
class FormulaList;
class VarList;
class TermList;
class InferenceInfoTermList;
class InferenceInfo;
class VarCounter;


class GapFiller
{
public:
  static void fill (Unit& unit);
private:
  
  static void onePremiseFormulaInference (const Unit& conclusion,
					  Unit& premise,
					  InferenceInfo*& info);
  static void markOccurrences (VarCounter&, const Formula&);

  
  static void fillSkolem (Unit& unit);
  static void buildChoiceAxioms (const Unit& premise,
				 Unit& conclusion,
				 const InferenceInfoTermList*);
  static void eliminateAQuantifiers(const Unit& premise,
				    Unit& conclusion,
				    const InferenceInfoTermList*);
  static void termListToVarList (const TermList& from, 
				 VarList& to, 
				 const Unit& premise,
				 const InferenceInfo* info);
  
  static void fillNNF (Unit& unit);
  static void fillNNF (const Unit& premise, 
		       Unit& result, 
		       const InferenceInfo* info);
  static void negateAll (const FormulaList& in, FormulaList& out);
  
  static void fillFlatten (Unit& unit);
  static void fillFlatten (const Unit& premise, 
			   Unit& result, 
			   const InferenceInfo* info);
  static void flatten(FormulaList& fs,Connective con);
  
  static void fillMiniscope (Unit& unit);
  static void fillMiniscope (const Unit& premise, 
			     Unit& result, 
			     const InferenceInfo* info);
  static void fillForallAnd (const Unit& premise,
			     bool isForall,
			     const InferenceInfo* info,
			     Formula& F,
			     const Formula& f);
  static void fillForallOr (const Unit& premise,
			    bool isForall,
			    const InferenceInfoTermList* info,
			    Formula& F,
			    const Formula& f);
  static void fillDummyQuantifierRemoval (const Unit& premise,
					  const InferenceInfoTermList* info,
					  Formula& F,
					  const Formula& f);
  static void fillFlatten (const Unit& premise,
			   const InferenceInfo* info,
			   Formula& F,
			   const Formula& f);
  static void fillQuantifierSwap (const Unit& premise,
				  const InferenceInfo* info,
				  Formula& F,
				  const Formula& f);
};

#endif
