
/*
 * File Inference.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Inference__
#define __Inference__


class InferenceInfo;
#include "Unit.hpp"


class Inference
{
public:
  Inference (InferenceRule r);
  Inference (InferenceRule r, const Unit& premise);
  Inference (InferenceRule r, const UnitList& premises);
  Inference (const UnitList& premises,
	     VampireKernel::Clause::BackgroundFlagIterator& flags);
  Inference (InferenceRule r, const Unit& premise, InferenceInfo* info);
  ~Inference ();

  
  const UnitList& premises () const { return _premises; }
  
  InferenceRule rule () const { return _rule; }
  
  InferenceInfo* info () const { return _info; }
  
  void setInfo (InferenceInfo* info){ _info = info; }

  CLASS_NAME("Inference");
  USE_ALLOCATOR(Inference);

  
  XML::Element* toXML () const;
  static string toString (InferenceRule);
  string toString() const;
protected:
  
  InferenceRule _rule;
  UnitList _premises;
  InferenceInfo* _info;
}; 



#endif 
