
/*
 * File PureLiteral.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Problem.hpp"
#include "Signature.hpp"
#include "PureLiteral.hpp"


class PureLiteral::Counter
{
public:
  Counter();
  
  int pos;
  
  int neg;
  
  const PredicateSymbol* symbol;
}; 


PureLiteral::Counter::Counter()
  :
  pos(0),
  neg(0),
  symbol(0)
{
}


PureLiteral::PureLiteral (Problem& problem,Signature& signature)
  : _problem(problem),
    _signature(signature)
{
} 



int PureLiteral::findAndRemove ()
{
  int removed = 0;

  for (;;) {
    Counter* counter = new Counter[_signature.noOfPreds()];
    if (! counter) {
      ALLOCATION_FAILURE;   
    }

    
    UnitChain::Iterator units (_problem.units());
    while (units.hasNext()) {
      const Unit& u = units.next();
      if (u.unitType() != CLAUSE) {
	continue;
      }
      Iterator<Literal> lits(u.clause().literals());
      while (lits.hasNext()) {
	const Literal& lit = lits.next();
	const PredicateSymbol& pred = lit.atom().functor();
	int p = pred.number();
	Counter& c = counter[p];
	c.symbol = &pred; 
	if (lit.sign()) {
	  c.pos++;
	}
	else {
	  c.neg++;
	}
      }
    }

    bool found = false;
    for (int i = _signature.noOfPreds()-1;i >= 0;i--) {
      const Counter& c = counter[i];
      const PredicateSymbol* p = c.symbol;
      if (! p || p->isEquality()) {
	continue;
      }
      if (c.pos == 0 || c.neg == 0) {
	found = true;
	

	UnitChain::DelIterator del (_problem.giveUnits());
	while (del.hasNext()) {
	  const Unit& u = del.next();
	  if (u.unitType() != CLAUSE) {
	    continue;
	  }
	  Iterator<Literal> lits(u.clause().literals());
	  while (lits.hasNext()) {
	    const Literal& lit = lits.next();
	    const PredicateSymbol& pred = lit.atom().functor();
	    if (p == &pred) {
	      del.del();
	      removed++;
	      break;
	    }
	  }
	}
      }
    }
    delete [] counter;
    if (! found) {
      return removed;
    }
  }
} 


