
/*
 * File Signature.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __signature__
#define __signature__

#define NO_OF_BUCKETS 2039






#include <string>

using namespace std;

#include "Atom.hpp"
#include "List.hpp"


class Symbol;
class FunctionSymbol;
class PredicateSymbol;
class SymbolList;
class Problem;



class Signature 
{ 
 public:

  
  class Iterator {
   public:
    explicit Iterator (SymbolList** buckets);
    bool hasNext ();
    Symbol& next ();
   private:
    SymbolList** _firstBucket;
    SymbolList** _currentBucket;
    SymbolList* _currentList;
  }; 

  
  class FunIterator 
    : public Iterator {
   public:
    explicit FunIterator (Signature& s) :
      Iterator (s._fbuckets)
      {}
  };

  
  class PredIterator 
    : public Iterator {
   public:
    explicit PredIterator (Signature& s) :
      Iterator (s._pbuckets)
      {}
  };

  Signature ();
  ~Signature ();

  
  FunctionSymbol& createFun (const string& name, int arity)
    { return static_cast<FunctionSymbol&>(create(name, arity, 
				        	 _fbuckets, false)); }
  
  PredicateSymbol& createPred (const string& name, int arity)
    { return static_cast<PredicateSymbol&>(create(name, arity, 
						  _pbuckets, true)); }
  void registerKernelNumber (Symbol& symbol, unsigned long number);
  const FunctionSymbol& getFunctionByItsKernelNumber (unsigned long number,
	 					      int arity);
  const PredicateSymbol& getPredicateByItsKernelNumber (unsigned long number,
						        int arity);

  
  PredicateSymbol& equality () { return *_equality; }
  int noOfPreds () const { return _noOfPreds; }
  int noOfFuns () const { return _noOfFuns; }
  void arityCheck () const;

  
  void createAnswerAtom (const TermList& args);
  void addArithmetic ();         
  
  bool isAnswerPredicate (const PredicateSymbol& p) const
    { return &p == _answerPredicate; }
  bool hasAnswerAtom () const
    { return _answerPredicate != 0; }
  const Atom& getAnswerAtom () const;
  string summary ();
  XML::Element* toXML() const;
  FunctionSymbol& newSkolemFunction (int arity);
  void setSkolemPrefix (const string& prefix);
  
  FunctionSymbol* existsFun (const string& name, int arity)
  { return static_cast<FunctionSymbol*>(exists(name,arity,_fbuckets)); }
  
   PredicateSymbol* existsPred (const string& name, int arity)
  { return static_cast<PredicateSymbol*>(exists(name,arity,_pbuckets)); }

  
  void resetSkolemFunctionNumbers () {_lastSkolem = 0;}

  
  static Signature sig;
 private:  
  
  SymbolList * _fbuckets [NO_OF_BUCKETS];
  
  SymbolList * _pbuckets [NO_OF_BUCKETS];
  
  int _noOfPreds;
  
  int _noOfFuns;
  
  int _constants;
  
  int _propositionalSymbols;
  
  int _maxFunArity;
  
  int _maxPredArity;
  
  PredicateSymbol* _answerPredicate;
  
  Atom _answerAtom;
  
  PredicateSymbol* _equality;
  
  Symbol** _kernelSymbols;
  
  int _maxNumberOfKernelSymbols;
  
  int _lastSkolem;
  
  string _skolemPrefix;
 
  static int hash (const string& name);  
  Symbol& create (const string& str, 
		  int arity, 
		  SymbolList** buckets, 
		  bool isPred);

  
  FunctionSymbol* existsFun (const string& name)
  { return static_cast<FunctionSymbol*>(exists(name,_fbuckets)); }
  
  PredicateSymbol* existsPred (const string& name)
  { return static_cast<PredicateSymbol*>(exists(name,_pbuckets)); }
  static Symbol* exists (const string& name, SymbolList** lst);
  static Symbol* exists (const string& name, int arity, SymbolList** lst);
  void expandKernelSymbolsToFit (unsigned long symbolKernelNumber);

 friend class SymCounter;
 friend class FunIterator;
 friend class PredIterator;
}; 


#endif 
