
/*
 * File SignedFormula.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __SignedFormula__
#define __SignedFormula__


#include "Connective.hpp"
#include "Formula.hpp"





class SignedFormula 
{
 private:
  class Data;

 public:
  
  SignedFormula () : _data (0) {}
  SignedFormula (bool sign, const Formula& atom);
  SignedFormula (const SignedFormula& s);
  ~SignedFormula ();
  void operator= (const SignedFormula& rhs);

  
  bool sign () const;
  const Formula& formula () const;
  Connective type () const;

  
  XML::Element* toXML () const;
  bool equals (const SignedFormula&) const;

  
  string toString () const;
 private:
  
  Data* _data;
}; 


class SignedFormula::Data 
{
 public:
  Data ();
  
  Data (bool sign, const Formula& formula)  
    : _counter (1), _sign (sign), _formula (formula) {}
  
  bool sign () const  { return _sign; }
  
  const Formula& formula () const  { return _formula; }

  void ref () { ASS (this); _counter++; }
  void deref ();

  
  CLASS_NAME("SignedFormula::Data");
  USE_ALLOCATOR(Data);
 protected:
  
  int _counter;
  bool _sign;
  Formula _formula;
}; 


class SignedFormulaList 
: public Lst<SignedFormula>
{
 public:
  
  SignedFormulaList () : Lst<SignedFormula> () {}
  
  SignedFormulaList (const SignedFormulaList& ss) : Lst<SignedFormula> (ss) {}
  
  explicit SignedFormulaList (const SignedFormula& s) : Lst<SignedFormula> (s) {}
  SignedFormulaList (const SignedFormula& hd, const SignedFormulaList& tl);
  
  explicit SignedFormulaList (LstData<SignedFormula>* d) : Lst<SignedFormula> (d) {}

  
  const SignedFormulaList& tail () const
    { return static_cast<const SignedFormulaList&>(Lst<SignedFormula>::tail()); }
}; 





inline
SignedFormula::~SignedFormula ()
{
  if (_data) {
    _data->deref ();
  }
} 


inline
SignedFormula::SignedFormula (const SignedFormula& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 


inline
SignedFormula::SignedFormula (bool sign, const Formula& formula)
  :
  _data (new Data(sign,formula))
{
} 


inline
bool SignedFormula::sign () const
{
  return _data->sign();
} 


inline
const Formula& SignedFormula::formula () const
{
  return _data->formula();
} 


#endif 
