
/*
 * File Sort.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __Sort__
#  define __Sort__


#include "VampireException.hpp"






template <typename ToCompare, class Comparator>
class Sort
{
 public:
  
  inline
  Sort (int length,Comparator& comparator) 
    :
    _elems ( new ToCompare[length] ),
    _size (length),
    _length (0),
    _comparator(comparator)
    {
      if (! _elems) {
        ALLOCATION_FAILURE;
      }
    }

  inline ~Sort () 
    {
      delete [] _elems;
    }

  
  inline int length () const 
    { return _length; }

  
  inline ToCompare operator [] (int n) const
    { 
      ASS(n < _length);
      return _elems[n]; 
    } 

   
  inline void add (ToCompare elem)
    {
      ASS(_length < _size);

      _elems[_length++] = elem;
    } 

  
  void removeDuplicates ()
    {
      int cur = 0;

      for (int next = 1; next < _length; next++) {
        if ( _elems[cur] != _elems[next] ) {
          cur++;
        }
      }

      
      _length = cur + 1;
    } 

  
  inline void sort ()
    { sort (0,_length-1); }

  
   inline bool member (const ToCompare elem) const { 
     return member (elem,0,_length-1); 
   }

 protected:  
  ToCompare* _elems;
  int _size;
  int _length;
  Comparator& _comparator;

  
  void sort (int p,int r) 
    {
      ASS(r < _length);

      if (p < r) {
        int q = partition(p,r);
        sort(p,q);
        sort(q+1,r);    
      }
    } 

  
  int partition (int p,int r) 
  {
    ToCompare x = _elems[p];
    int i = p-1;
    int j = r+1;

    for (;;) {
      do --j;
      while (_comparator.lessThan(x,_elems[j]));
      do ++i;
      while (_comparator.lessThan(_elems[i],x));
      if (i < j) {
        
        ToCompare tmp = _elems[i];
        _elems[i] = _elems[j];
        _elems[j] = tmp;
      }
      else
        return j;
    }
  } 

  
  bool member (const ToCompare elem, int fst, int lst) const { 
    for (;;) {
      if (fst > lst) {
	return false;
      }

      int mid = (fst + lst) / 2;
      
      if (_elems[mid] == elem) {
	return true;
      }

      if (_comparator.lessThan(_elems[mid],elem)) {
	lst = mid-1;
      }
      else { 
	fst = mid+1;
      }
    }
  } 
};  


#endif


