
/*
 * File SpiderMode.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Timer.hpp"
#include "Options.hpp"
#include "XML.hpp"
#include "Run.hpp"
#include "Problem.hpp"
#include "Preprocess.hpp"
#include "Refutation.hpp"
#include "SpiderMode.hpp"
#include "Kernel.hpp"
#include "Statistics.hpp"
#include "Property.hpp"


void SpiderMode::run (UnitList& units, Timer& timer)
{

  Run::log->addChild(XML::Element::generalProofAttemptInfo());
  XML::DeepElement* proofAttempt = new XML::DeepElement("proof_attempt");
  Run::log->addChild(proofAttempt);

  Options& options = Options::globalOptions;
  proofAttempt->addChild(options.toXML());

  Problem problem(options.problemName(),units);
  
  Property property(problem);
  property.scanProblem();
  Preprocess::preprocess(problem, options, property);

  
  Kernel kernel;

  
  VampireKernel::Options kernelOptions;
  
  Kernel::passOptions(kernelOptions, options);
  int timeLimit = options.timeLimitInDeciseconds() ? 
                  options.timeLimitInDeciseconds() : 
                  INT_MAX;
  RunResult result = kernel.run(timeLimit, kernelOptions, problem, 
				options, timer, 0);
  switch (result) {
    case REFUTATION: 
      if (options.proof()) {
	Refutation ref (kernel,problem,options);
	proofAttempt->addChild(ref.toXML());
      }
      cout << "+ ";
      break;

  case SATISFIABLE:
    cout << "- ";
    break;

  case UNKNOWN:
    cout << "? ";
    break;
  }

  cout << options.problemName() << ' '
       << timer.elapsedSeconds() << ' ' 
       << options.testId() << '\n';

  
  Statistics statistics (kernel.kernel());
  XML::Element* xmlStat = statistics.toXML();
  if (xmlStat) {
    proofAttempt->addChild(xmlStat);
  }
  
  string resultString = Run::toString(result);
  proofAttempt->addChild(new XML::TextElement("result",resultString));
  Run::log->addChild(new XML::TextElement("result",resultString));

  BK::DestructionMode::makeFast();
  kernel.kernel().closeSession(); 
}  



void SpiderMode::reportAllocationFailure ()
{
  Options& options = Options::globalOptions;

  cout << "? " 
       << options.problemName() << ' '
       << "0.0 "
       << options.testId() << '\n';
} 
