
/*
 * File Substitution.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Substitution__
#define __Substitution__

class Atom;
class Formula;
class FormulaList;

#include "Term.hpp"
#include "List.hpp"



class Substitution
{
public:
  Substitution ();
  ~Substitution ();

  void bind (Var v, const Term& t);
  bool isBound (Var v, Term& t) const;
  bool isBound (Var v) const;
  void undoBinding (Var v);
  void getDomain (VarList& vs) const;
  void getRange (TermList& ts) const;
  bool isInRange (const Term& t) const;
  void apply (Term&) const;
  void apply (TermList&) const;
  void apply (Atom&) const;
  void apply (Formula&);
  void apply (FormulaList&);
  
  Var freshVariable () { return (Var)++_lastNewVar; }
  
  void setLastFreshVariable (int var) { _lastNewVar = var; }
  XML::Element* toXML () const;
protected:
  
  typedef List<Term> Binding;
  
  int _top;
  
  Binding** _bindings;
  
  int _lastNewVar;
  void extend (int minimalNewCapacity);
}; 


#endif 

