
/*
 * File Symbol.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Symbol.hpp"
#include "SymCounter.hpp"
#include "XML.hpp"



XML::Element* Symbol::toXML() const
{
  XML::DeepElement* symbol = new XML::DeepElement(isFunction() ? 
						  "function_symbol" :
						  "predicate_symbol");
  symbol->addAttribute(new XML::TextAttribute("name",_name));
  symbol->addAttribute(new XML::IntegerAttribute("arity",_arity));
  symbol->addAttribute(new XML::IntegerAttribute("number",_number));

  symbol->addChild(new XML::IntegerElement("weight",_weight));
  symbol->addChild(new XML::IntegerElement("precedence",_precedence));
  if (_negativePrecedence) {
    symbol->addChild(new XML::IntegerElement("negative_precedence",
					    _negativePrecedence));
  }

  return symbol;
} 



Symbol::Symbol (const string& name, int arity, int& number) 
  : _name (name),
    _arity (arity),
    _number (number++),
    _kernelNumber (0),
    _weight(1),
    _precedence(0),
    _negativePrecedence(0),
    _properties(0)
{
}


ostream& operator << (ostream& str, const Symbol* p)
{
  str << p->name ();

  return str;
} 



FunctionSymbol::FunctionSymbol (const string& name, int arity, int& count) 
  : Symbol (name,arity,count) 
{
  addProperty(FUNCTION);
}
