
/*
 * File Tracer.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __MyTracer__
#  define __MyTracer__

#if DEBUG_TRACE_ALL
#  define TRACER(Fun) MyTracer tmp (Fun)
#  define CONTROL(description) MyTracer::controlPoint(description)
#  define AFTER(number,command) \
      if (MyTracer::passedControlPoints() >= number) { command };
#  define BETWEEN(number1,number2,command) \
      if (MyTracer::passedControlPoints() >= number1 && \
	  MyTracer::passedControlPoints() <= number2) \
      { command };


#include <iostream>
#include <string>

namespace XML {
  class Element;
};

using namespace std;


enum ControlPointKind {
  
  CP_ENTRY,
  
  CP_EXIT,
  
  CP_MID
}; 


class MyTracer {
 public:
  explicit MyTracer (const string& fun);
  virtual ~MyTracer ();
  static void printStack (ostream&);
  static XML::Element* stackToXML();
  static void controlPoint (const string&);
  static long passedControlPoints () { return _passedControlPoints; }

 private:
  string _fun;
  MyTracer* _previous;
  static MyTracer* _current;

  void printStack (ostream&, int& depth);
  static string _lastControlPoint;
  static long _passedControlPoints;
  static ControlPointKind _lastPointKind;
  static void controlPoint (const string&, ControlPointKind);
  static void outputLastControlPoint (ostream& str);
};


#else
#  define TRACER(Fun) 
#  define CONTROL(description)
#endif

#endif 
