
/*
 * File VampireException.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __VampireException__
#define __VampireException__

#include <string>
#include <iostream>

using namespace std;

#include "RunResult.hpp"

namespace XML {
  class Element;
  class DeepElement;
};



class VampireException 
{
 public:                                
  VampireException (const string& file, int line);

  virtual XML::Element* toXML () const = 0;
  virtual void cry (ostream&) = 0;
  virtual ~VampireException () {}
 protected:
  void outputFileAndLine (ostream&) const;
  void addFileAndLine (XML::DeepElement&) const;
  string _file;
  int _line;
}; 


#define VAMPIRE_EXCEPTION \
      throw VampireException(__FILE__,__LINE__)


class AllocationFailureException 
  : public VampireException
{
 public:                                
  AllocationFailureException (const string& file, int line);
  ~AllocationFailureException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
}; 


#define ALLOCATION_FAILURE \
      throw AllocationFailureException(__FILE__,__LINE__)


class AssertionViolationException 
  : public VampireException
{
public:                                
  AssertionViolationException (const string& file, int line);
  ~AssertionViolationException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
#if DEBUG_TRACE_ALL
private:
  XML::Element* _stack;
#endif
}; 


#define ASSERTION_VIOLATION \
      throw AssertionViolationException(__FILE__,__LINE__)

#if DEBUG_PREPRO


#  define ASS(Cond)                                   \
    if (! (Cond)) {                                   \
      MyTracer::printStack (cout);                    \
      ASSERTION_VIOLATION;                            \
    }                                         

#else 
#  define ASS(Cond)
#endif


class UserErrorException 
  : public VampireException
{
 public:                                
  UserErrorException (const string& file, int line, string _message);
  ~UserErrorException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
 private:                               
  string _message;
}; 


#define USER_ERROR(msg) \
      throw UserErrorException(__FILE__,__LINE__,msg)


class VampireTerminationException 
  : public VampireException
{
 public:                                
  VampireTerminationException (const string& file, 
			       int line, 
			       string reason,
			       RunResult result);
  ~VampireTerminationException () {}
  void cry (ostream&);
  XML::Element* toXML () const;
  RunResult getResult () const { return _result; }
  const string& getReason () const { return _reason; }
 private:                               
  string _reason;
  RunResult _result;
}; 


#define VAMPIRE_TERMINATION(reason,result) \
      throw VampireTerminationException(__FILE__,__LINE__,reason,result)

#if DEBUG_PREPRO
#define CLASS_NAME(name) static string className () { return name; }
#else
#define CLASS_NAME(name)
#endif

#endif 



