//
// File:         AutoSymbolAndHeaderPrecedence.cpp
// Description:  Automatic assignment of precedences to the functors and headers of the signature.
// Created:      Apr 24, 2001, 17:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "AutoSymbolAndHeaderPrecedence.hpp"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_AUTO_FUNCTOR_PRECEDENCE
 #define DEBUG_NAMESPACE "AutoSymbolAndHeaderPrecedence"
#endif 
#include "debugMacros.hpp"
//=================================================  

using namespace VK;


#ifdef DEBUG_ALLOC_OBJ_TYPE
#include "ClassDesc.hpp"
using namespace BK;
ClassDesc 
AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::LeftRightIter::NodeList::_classDesc("AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::LeftRightIter::NodeList",
									      AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::LeftRightIter::NodeList::minSize(),
									      AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::LeftRightIter::NodeList::maxSize());

ClassDesc 
AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::RightLeftIter::NodeList::_classDesc("AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::RightLeftIter::NodeList",
									      AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::RightLeftIter::NodeList::minSize(),
									      AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::RightLeftIter::NodeList::maxSize());

ClassDesc 
AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::Node::_classDesc("AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::Node",
							       AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::Node::minSize(),
							       AutoSymbolAndHeaderPrecedence::ArityFuncSortTree::Node::maxSize());

 
#endif


AutoSymbolAndHeaderPrecedence::AutoSymbolAndHeaderPrecedence() 
  : _dependenceOnArity(0L),
    _inversePolarity(false),
    _kinkyHeaderPrecedence(false),
    _someSymbolPrecedenceSet(false),
    _someHeaderPrecedenceSet(false),
    _minimalFixedSymbolPrecedence(0L),
    _maximalFixedSymbolPrecedence(0L),
    _minimalFixedHeaderPrecedence(0L),
    _maximalFixedHeaderPrecedence(0L),
    _minSymbolNumber(0UL),
    _maxSymbolNumber(0UL),
    _symbolPrecedence(0L,"AutoSymbolAndHeaderPrecedence::_symbolPrecedence"), 
    _symbolPrecedenceIsFixed(false,"AutoSymbolAndHeaderPrecedence::_symbolPrecedenceIsFixed"),
    _positiveHeaderPrecedence(0L,"AutoSymbolAndHeaderPrecedence::_positiveHeaderPrecedence"), 
    _positiveHeaderPrecedenceIsFixed(false,"AutoSymbolAndHeaderPrecedence::_positiveHeaderPrecedenceIsFixed"),
    _negativeHeaderPrecedence(0L,"AutoSymbolAndHeaderPrecedence::_negativeHeaderPrecedence"), 
    _negativeHeaderPrecedenceIsFixed(false,"AutoSymbolAndHeaderPrecedence::_negativeHeaderPrecedenceIsFixed"),
    _arity(0UL,"AutoSymbolAndHeaderPrecedence::_arity"),
    _registeredSymbol(false,"AutoSymbolAndHeaderPrecedence::_registeredSymbol")
{
  CALL("constructor AutoSymbolAndHeaderPrecedence()");   
};



AutoSymbolAndHeaderPrecedence::~AutoSymbolAndHeaderPrecedence()
{
  CALL("destructor ~AutoSymbolAndHeaderPrecedence()");
};


bool AutoSymbolAndHeaderPrecedence::symbolPrecedenceIsReserved(long precedence,ulong& symNum)
{
  CALL("symbolPrecedenceIsReserved(long precedence,ulong& symNum)");
  if (!_someSymbolPrecedenceSet) return false;
  long normPrec = normalisedSymbolPrecedence(precedence);  
  if ((normPrec < _minimalFixedSymbolPrecedence) ||
      (normPrec > _maximalFixedSymbolPrecedence))
    {
      return false;
    };    

  for (ulong sn = 0; sn < _registeredSymbol.size(); sn++)
    if ((_registeredSymbol[sn]) && 
	(_symbolPrecedenceIsFixed.sub(sn)))
      {
	if (_symbolPrecedence[sn] == normPrec)
	  {
	    symNum = sn;
	    return true;
	  };
      };
  return false;
}; // bool AutoSymbolAndHeaderPrecedence::symbolPrecedenceIsReserved(long precedence,ulong& symNum)

bool AutoSymbolAndHeaderPrecedence::headerPrecedenceIsReserved(long precedence,bool& negative,ulong& predNum)
{
  CALL("headerPrecedenceIsReserved(..)");
  if (!_someHeaderPrecedenceSet) return false;
  long normPrec = normalisedHeaderPrecedence(precedence);  
  if ((normPrec < _minimalFixedHeaderPrecedence) ||
      (normPrec > _maximalFixedHeaderPrecedence))
    {
      return false;
    };    
  
  for (ulong pn = 0; pn < _registeredSymbol.size(); pn++)
    if (_registeredSymbol[pn])
      {
	if (_positiveHeaderPrecedenceIsFixed.sub(pn) &&
	    _positiveHeaderPrecedence.sub(pn) == normPrec)
	  {
	    negative = false;
	    predNum = pn;
	    return true;
	  };
	if (_negativeHeaderPrecedenceIsFixed.sub(pn) &&
	    _negativeHeaderPrecedence.sub(pn) == normPrec)
	  {
	    negative = true;
	    predNum = pn;
	    return true;
	  };
      };
  return false;
}; // bool AutoSymbolAndHeaderPrecedence::headerPrecedenceIsReserved(..)

void AutoSymbolAndHeaderPrecedence::assignFixedSymbolPrecedence(ulong symNum,long precedence) 
{ 
  CALL("assignFixedSymbolPrecedence(ulong symNum,long precedence)");

  //DF; cout << "AF FN = " << symNum << "\n";

  _registeredSymbol.sub(symNum) = true;
  long normPrec = normalisedSymbolPrecedence(precedence);  
  if (!_someSymbolPrecedenceSet) 
    { 
      _someSymbolPrecedenceSet = true;
      _minimalFixedSymbolPrecedence = normPrec;
      _maximalFixedSymbolPrecedence = normPrec;
    }
  else 
    if (normPrec > _maximalFixedSymbolPrecedence) { _maximalFixedSymbolPrecedence = normPrec; }
    else 
      if (normPrec < _minimalFixedSymbolPrecedence) { _minimalFixedSymbolPrecedence = normPrec; };
   
  _symbolPrecedence.sub(symNum) = normPrec;    
  _symbolPrecedenceIsFixed.sub(symNum) = true;   
}; // void AutoSymbolAndHeaderPrecedence::assignFixedSymbolPrecedence(ulong symNum,long precedence)


void AutoSymbolAndHeaderPrecedence::assignFixedHeaderPrecedence(bool negative,ulong predNum,long precedence)
{
  CALL("assignFixedHeaderPrecedence(..)");
  _registeredSymbol.sub(predNum) = true;
  long normPrec = normalisedHeaderPrecedence(precedence); 
    
  if (!_someHeaderPrecedenceSet) 
    { 
      _someHeaderPrecedenceSet = true;
      _minimalFixedHeaderPrecedence = normPrec;
      _maximalFixedHeaderPrecedence = normPrec;
    }
  else 
    if (normPrec > _maximalFixedHeaderPrecedence) 
      { 
	_maximalFixedHeaderPrecedence = normPrec; 
      }
    else 
      if (normPrec < _minimalFixedHeaderPrecedence) 
	{ 
	  _minimalFixedHeaderPrecedence = normPrec; 
	};
   
  if (negative)
    {
      _negativeHeaderPrecedence.sub(predNum) = normPrec;    
      _negativeHeaderPrecedenceIsFixed.sub(predNum) = true;         
    }
  else
    {
      _positiveHeaderPrecedence.sub(predNum) = normPrec;    
      _positiveHeaderPrecedenceIsFixed.sub(predNum) = true;  
    };
}; // void AutoSymbolAndHeaderPrecedence::assignFixedHeaderPrecedence(..)


void AutoSymbolAndHeaderPrecedence::assignMaximalSymbolPrecedence(ulong symNum)
{

  //DF; cout << "AMP FN = " << symNum << "\n";

  _registeredSymbol.sub(symNum) = true;
  _someSymbolPrecedenceSet = true;
  _symbolPrecedence.sub(symNum) = maximalSymbolPrecedence();
  _symbolPrecedenceIsFixed.sub(symNum) = true;    
};

void AutoSymbolAndHeaderPrecedence::assignArity(ulong symNum,ulong arity) 
{ 
  CALL("assignArity(ulong symNum,ulong arity)");

  //DF; cout << "AAr FN = " << symNum << "\n";

  _registeredSymbol.sub(symNum) = true; 
  _arity.sub(symNum) = arity;
};
  
void AutoSymbolAndHeaderPrecedence::assignAllPrecedences() 
{ 
  CALL("assignAllPrecedences()");
  
  _maximalFixedSymbolPrecedence = maximalOf(absval(_maximalFixedSymbolPrecedence),absval(_minimalFixedSymbolPrecedence));
  _minimalFixedSymbolPrecedence = -_maximalFixedSymbolPrecedence;

  if (_dependenceOnArity)
    { 
      ArityFuncSortTree sortTree;   
      for (ulong sn = 0; sn < _registeredSymbol.size(); sn++)
	if ((_registeredSymbol[sn]) && 
	    (!_symbolPrecedenceIsFixed.sub(sn))) 
	  {
	    sortTree.insert(ArityFuncPair(_arity.sub(sn),sn));
	  };

      if (_dependenceOnArity > 0L)     
	{
	  // greater arity => greater precedence 
	  ArityFuncSortTree::RightLeftIter sortTreeIter(sortTree);      
	  while (sortTreeIter)
	    { 
	      assignSmallerSymbolPrecedence(sortTreeIter.value().el2); 
	      sortTreeIter.next();     
	    };
	}
      else // _dependenceOnArity < 0L 
	{
	  // greater arity => smaller precedence   
	  ArityFuncSortTree::LeftRightIter sortTreeIter(sortTree);
	  while (sortTreeIter)
	    { 
	      assignSmallerSymbolPrecedence(sortTreeIter.value().el2); 
	      sortTreeIter.next();     
	    };
	};
    }  
  else // !_dependenceOnArity
    {
      for (ulong sn = 0; sn < _registeredSymbol.size(); sn++)
	if ((_registeredSymbol[sn]) && 
	    (!_symbolPrecedenceIsFixed.sub(sn)))
	  assignSmallerSymbolPrecedence(sn);
    };

  // now assign header precedences
   ulong pn;  
  for (pn = 0UL; pn < _registeredSymbol.size(); pn++)
    if (_registeredSymbol[pn])
      {
	if (_positiveHeaderPrecedenceIsFixed.sub(pn))
	  {
	    if (!_negativeHeaderPrecedenceIsFixed.sub(pn))
	      {
		long posHeaderPrecedence = _positiveHeaderPrecedence.sub(pn);
		long negHeaderPrecedence;
		if (_kinkyHeaderPrecedence)
		  {
		    negHeaderPrecedence = 
		      (-1 * posHeaderPrecedence) - 2;
		  }
		else // !_kinkyHeaderPrecedence
		  {
		    if (_inversePolarity)
		      {
			negHeaderPrecedence = posHeaderPrecedence - 1;
		      }
		    else
		      {
			negHeaderPrecedence = posHeaderPrecedence + 1;
		      };
		  };
		_negativeHeaderPrecedenceIsFixed.sub(pn) = true;
		_negativeHeaderPrecedence.sub(pn) = negHeaderPrecedence;
		if (negHeaderPrecedence > _maximalFixedHeaderPrecedence)
		  _maximalFixedHeaderPrecedence = negHeaderPrecedence;
		if (negHeaderPrecedence < _minimalFixedHeaderPrecedence)
		  _minimalFixedHeaderPrecedence = negHeaderPrecedence;
	      };
	  }
	else
	  if (_negativeHeaderPrecedenceIsFixed.sub(pn))
	    {
	      long negHeaderPrecedence = _negativeHeaderPrecedence.sub(pn);
	      long posHeaderPrecedence;
	      if (_kinkyHeaderPrecedence)
		{
		  posHeaderPrecedence = 
		    (-1 * negHeaderPrecedence) - 2;
		}
	      else // !_kinkyHeaderPrecedence
		{
		  if (_inversePolarity)
		    {
		      posHeaderPrecedence = negHeaderPrecedence + 1;
		    }
		  else
		    {
		      posHeaderPrecedence = negHeaderPrecedence - 1;
		    };
		};
	      _positiveHeaderPrecedenceIsFixed.sub(pn) = true;
	      _positiveHeaderPrecedence.sub(pn) = posHeaderPrecedence;
	      if (posHeaderPrecedence > _maximalFixedHeaderPrecedence)
		_maximalFixedHeaderPrecedence = posHeaderPrecedence;
	      if (posHeaderPrecedence < _minimalFixedHeaderPrecedence)
		_minimalFixedHeaderPrecedence = posHeaderPrecedence;
	    };
      };
  
  // now deal with predicates for which neither neither
  // positive nor negative header precedence has been fixed
  _maximalFixedHeaderPrecedence = 
    maximalOf(absval(_maximalFixedHeaderPrecedence),
	      absval(_minimalFixedHeaderPrecedence)) + 4;
  _minimalFixedHeaderPrecedence = -_maximalFixedHeaderPrecedence;
  for (pn = 0UL; pn < _registeredSymbol.size(); pn++)
    if (_registeredSymbol[pn])
      {
	if (!_positiveHeaderPrecedenceIsFixed.sub(pn))
	  {
	    ASSERT(!_negativeHeaderPrecedenceIsFixed.sub(pn));	    
	    if (_kinkyHeaderPrecedence)
	      {
		if (_inversePolarity)
		  {
		    _negativeHeaderPrecedence.sub(pn) = _minimalFixedHeaderPrecedence;
		    _positiveHeaderPrecedence.sub(pn) = _maximalFixedHeaderPrecedence;
		  }
		else
		  {
		    _positiveHeaderPrecedence.sub(pn) = _minimalFixedHeaderPrecedence;
		    _negativeHeaderPrecedence.sub(pn) = _maximalFixedHeaderPrecedence;
		  };
		--_minimalFixedHeaderPrecedence;
		++_maximalFixedHeaderPrecedence;
	      }
	    else // !_kinkyHeaderPrecedence
	      {
		if (_inversePolarity)
		  {
		    _positiveHeaderPrecedence.sub(pn) = _minimalFixedHeaderPrecedence;
		    _negativeHeaderPrecedence.sub(pn) = _minimalFixedHeaderPrecedence - 1;
		  }
		else
		  {
		    _negativeHeaderPrecedence.sub(pn) = _minimalFixedHeaderPrecedence;
		    _positiveHeaderPrecedence.sub(pn) = _minimalFixedHeaderPrecedence - 1;
		  };
		_minimalFixedHeaderPrecedence -= 2;
	      };
	  };
      };

}; // void AutoSymbolAndHeaderPrecedence::assignAllPrecedences() 

void AutoSymbolAndHeaderPrecedence::assignSmallerSymbolPrecedence(ulong symNum) 
{ 
  CALL("assignSmallerSymbolPrecedence(ulong symNum)");
  _minimalFixedSymbolPrecedence = previousNormalisedSymbolPrecedence(_minimalFixedSymbolPrecedence);
  _symbolPrecedence.sub(symNum) = _minimalFixedSymbolPrecedence;
};


//====================================================
