//
// File:         ClauseAssembler.hpp
// Description:  Implements clause sharing.
// Created:      Dec 13, 2001 
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         Was CLAUSE_INDEX in clause_index.h.
// Note:         Integrate and Remove methods need a revision,
//               they became very strange after several modifications.
//============================================================================
#ifndef CLAUSE_ASSEMBLER_H
//=============================================================================
#define CLAUSE_ASSEMBLER_H 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "LiteralSharing.hpp"
#include "variables.hpp"
#include "TemporaryLiteral.hpp"
#include "LiteralProperties.hpp"
namespace VK
{
  
  class SHARING_INDEX;
  class DTREE_NODE;
  class LiteralSelection;
  class BasicTempLitList;
  class Clause;
  class OpenClauseBackground;
  class PagedPassiveClauseStorage;
  class SymbolInfo;
  class TemporaryClause;
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_ASSEMBLER
#define DEBUG_NAMESPACE "ClauseAssembler"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
class ClauseAssembler
{
public:
  typedef TemporaryLiteral::Extras::Handle LitHandle;
 public:
  ClauseAssembler(LiteralSharing* litSharing,const LiteralSelection* litSelection);
  ~ClauseAssembler() { CALL("destructor ~ClauseAssembler()"); };
  void reset();

  bool sharingIsLocked() const 
  {
    return _litSharing->isLocked();
  };

  void setInheritNegativeSelection(bool flag) { _inheritNegativeSelection = flag; };
  void setEqualityIsPossibleFlag(bool flag) 
    {
      _equalityIsPossible = flag;
    };
  void setEliminatedPredicateTable(const SymbolInfo* tbl)
    {
      _eliminatedPredicateTable = tbl;
    };

  void setMaxMaxWeight(long w)
  {
    CALL("setMaxMaxWeight(long w)");
    _maxMaxWeight = w;
  };
  void setMaxWeight(long w)
  {
    CALL("setMaxWeight(long w)");
    _maxWeight = w;
    if (_maxWeight < 1L)
      {
	_maxWeight = 1L;
      }
    else if (_maxWeight > _maxMaxWeight)
      _maxWeight = _maxMaxWeight;
  };

  Clause* assembleShared(TemporaryClause& cl);
  Clause* assembleSharedWithoutSelection(TemporaryClause& cl);
  Clause* assembleUnshared(TemporaryClause& cl);
  bool assembleInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
					   TemporaryClause& cl,
					   long clauseNum,
					   ostream* showNewClause,
					   bool markAsName);
  Clause* assembleSharedCopy(Clause* cl);
  Clause* assembleNameIntro(const TemporaryLiteral& definedLit,
                            const BasicTempLitList& body);

  void removeFromSharing(Clause* cl);
   

private:
  Clause* assembleEmptyClause(OpenClauseBackground& ancestors) const;
  Clause* assembleUnitClause(TemporaryLiteral& lit,
			     OpenClauseBackground& ancestors);

  Clause* assembleClause(BasicTempLitList& lits,
			 OpenClauseBackground& ancestors);
  Clause* assembleClauseWithoutSelection(BasicTempLitList& lits,
					 OpenClauseBackground& ancestors);
 
  Clause* assembleUnsharedClause(BasicTempLitList& lits,OpenClauseBackground& ancestors);
 
  bool assembleClauseInPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
						 BasicTempLitList& lits,
						 OpenClauseBackground& ancestors,
						 long clauseNum,
						 ostream* showNewClause,
						 bool markAsName);
  
   
  DTREE_NODE** Tree(const TERM& header) { return _litSharing->Tree(header); };
 

  TERM* shareTmpLiteral(const TemporaryLiteral& lit,Clause* cl);

  TERM* shareLiteralCopy(TERM* lit);
 
  TERM* assembleUnsharedLit(const TemporaryLiteral& lit);

  void writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
					  BasicTempLitList& lits);  

  void writeIntoPagedPassiveClauseStorage(PagedPassiveClauseStorage& storage,
					  const LitHandle& lit);
 
  TERM* Integrate(const Flatterm* lit,Clause* cl);
  TERM* IntegrateEq(const TERM& header,const Flatterm* arg1,const Flatterm* arg2,Clause* cl);

  void removeFromSharing(TERM* lit);

 private:
  LiteralSharing* _litSharing;
  INT_VAR_RENAMING _intVarRenaming;
  VAR_WEIGHTING _varWeighting;
  TERM::FastOpenUnshared _openUnsharedTerm;
  const LiteralSelection* _litSelection;
  bool _inheritNegativeSelection;
  bool _equalityIsPossible;
  const SymbolInfo* _eliminatedPredicateTable;
  long _maxMaxWeight;
  long _maxWeight;
}; // class ClauseAssembler

}; // namespace VK

//============================================================================
#endif
