//
// File:         ClauseNumIndex.hpp
// Description:  Searching clauses by numbers.
// Created:      Apr 08, 2000, 22:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef CLAUSE_NUM_INDEX_H
//============================================================================
#define CLAUSE_NUM_INDEX_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"    
#include "SkipList.hpp"
#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
namespace VK
{
class Clause;
}; // namespace VK
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_NUM_INDEX
 #define DEBUG_NAMESPACE "ClauseNumIndex"
#endif
#include "debugMacros.hpp"
//============================================================================
namespace VK
{
class ClauseNumIndex
{  
 public:
  // List cannot be made private since it has a static data member
  typedef BK::SkipList<BK::GlobAlloc,Clause*,ulong,2,28,ClauseNumIndex> List;
 public:
  ClauseNumIndex() 
  {
   CALL("constructor ClauseNumIndex()");
   unlock();
  };
  ~ClauseNumIndex() 
    { 
      CALL("destructor ~ClauseNumIndex()");  
    };
  void reset()
  {
   _list.reset();
   unlock();
  };
  bool isLocked() const { return _isLocked; };

  void insert(Clause* cl);
  void remove(Clause* cl);
  Clause* find(ulong num);
  #ifndef NO_DEBUG
   bool clean() const { return !_list; };
  #endif
private:
  void lock() { _isLocked = true; };
  void unlock() { _isLocked = false; };
 private:
  List _list;
  List::Surfer _surfer;
  bool _isLocked;
}; // class ClauseNumIndex

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
