//
// File:         LiteralProperties.inl
// Description:  Various literal labels.
// Created:      Feb 13, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef LITERAL_PROPERTIES_INL
#  define LITERAL_PROPERTIES_INL
//=================================================================
#include "LiteralProperties.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_PROPERTIES
#  define DEBUG_NAMESPACE "LiteralProperties"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{  
  inline
  void LiteralProperties::setSelectionLayer(ulong l)
  {
    CALL("setSelectionLayer(ulong l)");
    if (l > MaximalSelectionLayer) l = MaximalSelectionLayer;
    _data = 
      (_data << (_SizeOfSelectionLayer + _BeforeSelectionLayer)) >>
      (_SizeOfSelectionLayer + _BeforeSelectionLayer);
    _data |= (l << _AfterSelectionLayer);
  }; // void LiteralProperties::setSelectionLayer(ulong l)


  inline
  void LiteralProperties::setDeploymentMode(const DeploymentMode& m)
  {
    CALL("setDeploymentMode(const DeploymentMode& m)");
    ASSERT(m == DefaultDeploymentMode ||
	   m == RuleHead ||
	   m == Subquery);
       
    ulong before = 
      (_data >> (_SizeOfDeploymentMode + _AfterDeploymentMode)) <<
      (_SizeOfDeploymentMode + _AfterDeploymentMode);
    _data = 
      (_data << (_SizeOfDeploymentMode + _BeforeDeploymentMode)) >>
      (_SizeOfDeploymentMode + _BeforeDeploymentMode);
    _data |= before;
    _data |= (m << _AfterDeploymentMode);
  }; // void LiteralProperties::setDeploymentMode(const DeploymentMode& m)


  inline
  void LiteralProperties::setEqualityMode(const EqualityMode& m)
  {
    CALL("setEqualityMode(const EqualityMode& m)");
    ASSERT(m == DefaultEqualityMode ||
	   m == UseEqForResolution ||
	   m == UseEqForResolutionOnly);
       
    ulong before = 
      (_data >> (_SizeOfEqualityMode + _AfterEqualityMode)) <<
      (_SizeOfEqualityMode + _AfterEqualityMode);
    _data = 
      (_data << (_SizeOfEqualityMode + _BeforeEqualityMode)) >>
      (_SizeOfEqualityMode + _BeforeEqualityMode);
    _data |= before;
    _data |= (m << _AfterEqualityMode);
  }; // void LiteralProperties::setEqualityMode(const EqualityMode& m)



  inline
  void LiteralProperties::setSelectionLayerModifier(const SLModifier& m)
  {
    CALL("setSelectionLayerModifier(const SLModifier& m)");
    ASSERT(m == DefaultSLModifier ||
	   m == AccelerateSelection ||
	   m == PostponeSelection);
    
    ulong before = 
      (_data >> (_SizeOfSLModifier + _AfterSLModifier)) <<
      (_SizeOfSLModifier + _AfterSLModifier);
    _data = 
      (_data << (_SizeOfSLModifier + _BeforeSLModifier)) >>
      (_SizeOfSLModifier + _BeforeSLModifier);
    _data |= before;
    _data |= (m << _AfterSLModifier);
    
    ASSERT(selectionLayerModifier() == m);

  }; // void LiteralProperties::setSelectionLayerModifier(const SLModifier& m)


  inline
  void LiteralProperties::setFlagSelected(bool fl)
  {
    CALL("setFlagSelected(bool fl)");
    ulong before = 
      (_data >> (_SizeOfFlagSelected + _AfterFlagSelected)) <<
      (_SizeOfFlagSelected + _AfterFlagSelected);
    _data = 
      (_data << (_SizeOfFlagSelected + _BeforeFlagSelected)) >>
      (_SizeOfFlagSelected + _BeforeFlagSelected);
    _data |= before;
    _data |= ((fl) ? (1UL << _AfterFlagSelected) : 0UL);
  }; // void LiteralProperties::setFlagSelected(bool fl)
  

  inline
  void LiteralProperties::makeSelected()
  {
    CALL("makeSelected()");
    setFlagSelected(true);
  }; //void LiteralProperties::makeSelected()

  inline
  void LiteralProperties::makeUnselected()
  {
    CALL("makeUnselected()");
    setFlagSelected(false);
  }; // void LiteralProperties::makeUnselected()

  inline
  void LiteralProperties::setSelectabilityPrioritySignBit(bool fl)
  {
    CALL("setSelectabilityPrioritySignBit(bool fl)");
    ulong flag = 
      (fl)? (1UL << _AfterSelectabilityPrioritySignBit) : 0UL;
    ulong before = 
      (_data >> (_SizeOfSelectabilityPrioritySignBit + _AfterSelectabilityPrioritySignBit)) <<
      (_SizeOfSelectabilityPrioritySignBit + _AfterSelectabilityPrioritySignBit);
    _data = 
      (_data << (_SizeOfSelectabilityPrioritySignBit + _BeforeSelectabilityPrioritySignBit)) >>
      (_SizeOfSelectabilityPrioritySignBit + _BeforeSelectabilityPrioritySignBit);
    _data |= before;
    _data |= flag;
  }; // void LiteralProperties::setSelectabilityPrioritySignBit(bool fl)


  inline
  void LiteralProperties::setSelectabilityPriority(long p)
  {
    CALL("setSelectabilityPriority(long p)");
    if (p > MaximalSelectabilityPriority) 
      {
	p = MaximalSelectabilityPriority;
      }
    else if (p < MinimalSelectabilityPriority) 
      {
	p = MinimalSelectabilityPriority;
      };
    if (p < 0L)
      {
	setSelectabilityPrioritySignBit(true);
	p = -p;
      }
    else
      setSelectabilityPrioritySignBit(false);
    
    ulong before = 
      (_data >> (_SizeOfSelectabilityPriority + _AfterSelectabilityPriority)) <<
      (_SizeOfSelectabilityPriority + _AfterSelectabilityPriority);
    _data = 
      (_data << (_SizeOfSelectabilityPriority + _BeforeSelectabilityPriority)) >>
      (_SizeOfSelectabilityPriority + _BeforeSelectabilityPriority);
    _data |= before;
    _data |= (static_cast<ulong>(p) << _AfterSelectabilityPriority);
  }; // void LiteralProperties::setSelectabilityPriority(long p)


  inline
  void LiteralProperties::setFiniteSelectabilityInheritanceLimit(ulong l)
  {
    CALL("setFiniteSelectabilityInheritanceLimit(ulong l)");
    if (l > MaximalFiniteSelectabilityInheritanceLimit)
      l = MaximalFiniteSelectabilityInheritanceLimit;
    ++l;
  
    ulong before = 
      (_data >> (_SizeOfSelectabilityInheritanceLimit + _AfterSelectabilityInheritanceLimit)) <<
      (_SizeOfSelectabilityInheritanceLimit + _AfterSelectabilityInheritanceLimit);
    _data = 
      (_data << (_SizeOfSelectabilityInheritanceLimit + _BeforeSelectabilityInheritanceLimit)) >>
      (_SizeOfSelectabilityInheritanceLimit + _BeforeSelectabilityInheritanceLimit);
    _data |= before;
    _data |= (l << _AfterSelectabilityInheritanceLimit);
  }; // void LiteralProperties::setFiniteSelectabilityInheritanceLimit(ulong l)


  inline
  void LiteralProperties::setInfiniteSelectabilityInheritanceLimit()
  {
    CALL("setInfiniteSelectabilityInheritanceLimit()");
    ulong before = 
      (_data >> (_SizeOfSelectabilityInheritanceLimit + _AfterSelectabilityInheritanceLimit)) <<
      (_SizeOfSelectabilityInheritanceLimit + _AfterSelectabilityInheritanceLimit);
    _data = 
      (_data << (_SizeOfSelectabilityInheritanceLimit + _BeforeSelectabilityInheritanceLimit)) >>
      (_SizeOfSelectabilityInheritanceLimit + _BeforeSelectabilityInheritanceLimit);
    _data |= before;
  }; // void LiteralProperties::setInfiniteSelectabilityInheritanceLimit()


  inline
  void LiteralProperties::setSelectabilityCategory(const SelectabilityCategory& c)
  {
    CALL("setSelectabilityCategory(const SelectabilityCategory& c)");
    ASSERT(c == DefaultSelectability ||
	   c == LowSelectability ||
	   c == HighSelectability);
    ((_data >>= _SizeOfSelectabilityCategory) <<= 
     _SizeOfSelectabilityCategory) |= c;
  }; // void LiteralProperties::setSelectabilityCategory(const SelectabilityCategory& c)



  inline
  ulong LiteralProperties::selectionLayer() const
  {
    CALL("selectionLayer() const");
    return (_data << _BeforeSelectionLayer) >> _WithoutSelectionLayer;
  }; // ulong LiteralProperties::selectionLayer() const


  inline
  LiteralProperties::DeploymentMode LiteralProperties::deploymentMode() const
  {
    CALL("deploymentMode() const");
    ulong res = (_data << _BeforeDeploymentMode) >> _WithoutDeploymentMode;
    
    ASSERT(res == DefaultDeploymentMode ||
	   res == RuleHead ||
	   res == Subquery);

    return static_cast<DeploymentMode>(res);
  }; // DeploymentMode LiteralProperties::deploymentMode() const

  
  inline
  LiteralProperties::EqualityMode LiteralProperties::equalityMode() const
  {
    CALL("equalityMode() const");
    ulong res = (_data << _BeforeEqualityMode) >> _WithoutEqualityMode;
    
    ASSERT(res == DefaultEqualityMode ||
	   res == UseEqForResolution ||
	   res == UseEqForResolutionOnly);

    return static_cast<EqualityMode>(res);
  }; // EqualityMode LiteralProperties::equalityMode() const

  inline
  bool LiteralProperties::useEqForResolution() const
  {
    CALL("useEqForResolution() const");
    return equalityMode() != DefaultEqualityMode;
  };
 
  inline
  bool LiteralProperties::useEqForParamodulationFrom() const
  {
    CALL("useEqForParamodulationFrom() const");
    return equalityMode() != UseEqForResolutionOnly;
  };

  inline
  LiteralProperties::SLModifier LiteralProperties::selectionLayerModifier() const
  {
    CALL("selectionLayerModifier() const");
    ulong res = (_data << _BeforeSLModifier) >> _WithoutSLModifier;
    
    ASSERT(res == DefaultSLModifier ||
	   res == AccelerateSelection ||
	   res == PostponeSelection);

    return static_cast<SLModifier>(res);
  }; // SLModifier LiteralProperties::selectionLayerModifier() const


  inline
  bool LiteralProperties::isSelected() const
  {
    CALL("isSelected() const");
    return 
      ((_data << _BeforeFlagSelected) >> _WithoutFlagSelected) != 0UL;
  }; // bool LiteralProperties::isSelected() const


  inline
  bool LiteralProperties::selectabilityPrioritySignBit() const
  {
    CALL("selectabilityPrioritySignBit() const");
    return 
      ((_data << _BeforeSelectabilityPrioritySignBit) >> 
       _WithoutSelectabilityPrioritySignBit) != 0UL;
  }; // bool LiteralProperties::selectabilityPrioritySignBit() const



  inline
  long LiteralProperties::selectabilityPriority() const
  {
    CALL("selectabilityPriority() const");
    ulong absVal =  
      (_data << _BeforeSelectabilityPriority) >> 
      _WithoutSelectabilityPriority;
    if (selectabilityPrioritySignBit())
      {
	return -(static_cast<long>(absVal));
      }
    else
      return static_cast<long>(absVal);
  }; // long LiteralProperties::selectabilityPriority() const 


  inline
  bool LiteralProperties::selectabilityInheritanceLimitIsInfinite() const
  {
    CALL("selectabilityInheritanceLimitIsInfinite() const");
    return 
      ((_data << _BeforeSelectabilityInheritanceLimit) >> 
       _WithoutSelectabilityInheritanceLimit) == 0UL;
  }; // bool LiteralProperties::selectabilityInheritanceLimitIsInfinite() const


  inline
  bool LiteralProperties::selectabilityInheritanceLimitIsFinite() const
  {
    CALL("selectabilityInheritanceLimitIsFinite() const");
    return 
      ((_data << _BeforeSelectabilityInheritanceLimit) >> 
       _WithoutSelectabilityInheritanceLimit) != 0UL;
  }; // bool LiteralProperties::selectabilityInheritanceLimitIsFinite() const


  inline
  bool LiteralProperties::selectabilityInheritanceLimitIsFinite(ulong& limit) const
  {
    CALL("selectabilityInheritanceLimitIsFinite(ulong& limit) const");
    limit = 
      (_data << _BeforeSelectabilityInheritanceLimit) >> 
      _WithoutSelectabilityInheritanceLimit;
    if (limit) 
      {
	--limit;
	return true;
      }
    else
      return false;
  }; // bool LiteralProperties::selectabilityInheritanceLimitIsFinite(ulong& limit) const


  inline
  LiteralProperties::SelectabilityCategory
  LiteralProperties::selectabilityCategory() const
  {
    CALL("selectabilityCategory() const");
    ulong res = 
      (_data << _BeforeSelectabilityCategory) >> _WithoutSelectabilityCategory;
    ASSERT(res == DefaultSelectability ||
	   res == LowSelectability ||
	   res == HighSelectability);
    return static_cast<SelectabilityCategory>(res);
  }; // SelectabilityCategory LiteralProperties::selectabilityCategory() const




  inline
  BK::Comparison
  LiteralProperties::compareSelectabilityForFactoring(const LiteralProperties& prop) const
  {
    CALL("compareSelectabilityForFactoring(const LiteralProperties& prop) const");
    if (selectabilityCategory() > prop.selectabilityCategory()) return BK::Greater;
    if (selectabilityCategory() < prop.selectabilityCategory()) return BK::Less;
 
    // selectabilityCategory() == prop.selectabilityCategory()
    if (selectabilityCategory() == LowSelectability)
      {
	if (selectabilityPriority() < prop.selectabilityPriority()) return BK::Greater;
	if (selectabilityPriority() > prop.selectabilityPriority()) return BK::Less;
      }
    else // selectabilityCategory() != LowSelectability
      {
	if (selectabilityPriority() > prop.selectabilityPriority()) return BK::Greater;
	if (selectabilityPriority() < prop.selectabilityPriority()) return BK::Less;
      };

    // selectabilityPriority() == prop.selectabilityPriority()
#if (defined DEBUG_NAMESPACE) || (defined _SUPPRESS_WARNINGS_)
    ulong inhLim1 = 0UL;
    ulong inhLim2 = 0UL;
#else
    ulong inhLim1;
    ulong inhLim2;
#endif    

    if (selectabilityInheritanceLimitIsFinite(inhLim1))
      {
	if (prop.selectabilityInheritanceLimitIsFinite(inhLim1))
	  {
	    return BK::compare(inhLim1,inhLim2);
	  }
	else
	  return BK::Less;
      }
    else
      if (prop.selectabilityInheritanceLimitIsFinite())
	{
	  return BK::Greater;
	};
    return BK::Equal;
  }; // BK::Comparison LiteralProperties::compareSelectabilityForFactoring(const LiteralProperties& prop) const
}; // namespace VK


//=================================================================

#endif
