//
// File:         Master3.cpp
// Description:  Main loop implementing main_alg 3
// Created:      Apr 29, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      May 20, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Too strong an assertion removed in processCollectedInput3().
//====================================================
#include "GlobalStopFlag.hpp"
#include "Master.hpp"
#include "PagedPassiveClauseStorage.inl"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.hpp"
//====================================================

using namespace BK;
using namespace Gem;
using namespace VK;

void Master::inputClause3(Clause* cl)
{
  CALL("inputClause3(Clause* cl)");
  if (cl->isSupportClause())
    {
      inputSupportClause3(cl);
    }
  else
    inputQueryClause3(cl);
}; // void Master::inputClause3(Clause* cl)

void Master::inputQueryClause3(Clause* cl)
{
  CALL("inputQueryClause3(Clause* cl)");  
  Clause* sharedRetainedAnswer;
  ASSERT(!cl->isSupportClause());
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputQueryClause3_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputQueryClause3(Clause* cl)");
    };


  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;


  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
    {
      _inputClauseCollector.addAncestor(cl);
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{	  
	  ASSERT(!_newInputClause.isNull());
	  _pagedPassive.maximalPriorityOn();
	  _forwardProcessing.load(_newInputClause);

	process_collected_input:
	  while (_forwardProcessing.anotherRetained(_retainedClause))
	    {
	      if (_retainedClause->isAnswer())
		{
		  sharedRetainedAnswer = 
		    _clauseAssembler.assembleShared(*_retainedClause);
		  if (!sharedRetainedAnswer)
		    {
		      ++_statistics.clauses.generated.discardedTooLarge;
		    }
		  else
		    {
		      ++_statistics.clauses.retained.total;
		      ASSERT(sharedRetainedAnswer);
		      ASSERT(sharedRetainedAnswer->isAnswer());
		      sharedRetainedAnswer->setMainSet(Clause::MainSetNew);
		      sharedRetainedAnswer->SetNumber(_clauseNumberGenerator->generate());
		      rememberAnswer3(sharedRetainedAnswer);
		      _proof = sharedRetainedAnswer;
		      terminate(Proof);   
		      disposeObsolete(); 
		      _timer.stop();
		      _timerIsOn = false; 
		      _continuation = Continuation_InputQueryClause3_ProcessCollectedInput;
		      return;
		    };
		}
	      else // !_retainedClause->isAnswer()
		{
		  // transfer to paged passive
		  long clauseNumber = _clauseNumberGenerator->generate();
                  ostream* outputStreamForPagedPassive = 
                    (_parameters.showPassive)? (&cout) : 0;
		  if (!_clauseAssembler.assembleInPagedPassiveClauseStorage(_pagedPassive,
									    *_retainedClause,
									    clauseNumber,
									    outputStreamForPagedPassive,
									    false))
		    {
		      ++_statistics.clauses.generated.discardedTooLarge;
		    }
		  else
		    ++_statistics.clauses.retained.total;

		};
	    }; // while (_forwardProcessing.anotherRetained(_retainedClause))
	  _pagedPassive.maximalPriorityOff(); 
	}; // if (_inputClauseCollector.finaliseClause(_newInputClause))
    }; // if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
      
  _continuation = Continuation_Init;
  _timer.stop();
  _timerIsOn = false;
}; // void Master::inputQueryClause3(Clause* cl)

void Master::inputSupportClause3(Clause* cl)
{
  CALL("inputSupportClause3(Clause* cl)");

  ASSERT(cl->isSupportClause());
  Clause* sharedRetainedClause;
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_InputSupportClause3_ProcessCollectedInput:
      cancelTermination();
      goto process_collected_input;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::inputSupportClause3(Clause* cl)");
    };

  ASSERT(!_terminated);
  ASSERT(!_proof);
  _timer.reset();
  _timer.start();
  _timerIsOn = true;

  if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
    {
      _inputClauseCollector.addAncestor(cl);
      _inputClauseCollector.markAsSupport();
      if (cl->isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
      if (cl->isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
      if (_inputClauseCollector.finaliseClause(_newInputClause))
	{
	  ASSERT(!_newInputClause.isNull());
	  _forwardProcessingForSupport.load(_newInputClause);
	process_collected_input:
	  while (_forwardProcessingForSupport.anotherRetained(_retainedClause))
	    {
	      sharedRetainedClause = 
		_clauseAssembler.assembleShared(*_retainedClause);
	      if (!sharedRetainedClause)
		{
		  ++_statistics.clauses.generated.discardedTooLarge;
		}
	      else
		{
		  ++_statistics.clauses.retained.total;
		  ASSERT(sharedRetainedClause);
		  sharedRetainedClause->setMainSet(Clause::MainSetNew);
		  
		  if (sharedRetainedClause->isAnswer()) 
		    {
		      rememberAnswer0(sharedRetainedClause);
		      _proof = sharedRetainedClause;
		      terminate(Proof); 
		      _timer.stop();
		      _timerIsOn = false;
		      _continuation = Continuation_InputSupportClause3_ProcessCollectedInput;
		      return;
		    };
		  sharedRetainedClause->SetNumber(_clauseNumberGenerator->generate());      
		  INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTRODUCE_CLAUSE_ID(sharedRetainedClause);
		  if (_parameters.useClauseNumIndex) 
		    {
		      sharedRetainedClause->protect();
		      _clauseNumIndex.insert(sharedRetainedClause);
		      sharedRetainedClause->unprotect();
		    };
		  if (_parameters.showPassive) _parameters.showPassive(sharedRetainedClause);
		  integratePassive(sharedRetainedClause);
		  makeActive0(sharedRetainedClause);
		};
	    }; // while (_forwardProcessingForSupport.anotherRetained(_retainedClause))
	  _continuation = Continuation_Init;
	  integrateIntoBackDemIndexFromQueue(); 
	}; // if (_inputClauseCollector.finaliseClause(_newInputClause))

    }; // if (_inputClauseCollector.loadClause(cl,ClauseBackground::Prepro))
}; // void Master::inputSupportClause3(Clause* cl)




void Master::mainLoop3()
{
  CALL("mainLoop3()");
  PagedPassiveClauseStorage::StoredClause selectedClause;
  Clause* recentlyReanimatedClause;

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop3_SelectCurrent:
      cancelTermination();
      _continuation = Continuation_Init;
      goto select_current;
    case Continuation_MainLoop3_BackSimplify:
      cancelTermination();
      goto back_simplify;
    case Continuation_MainLoop3_ProcessNewInference:       
      cancelTermination();
      _continuation = Continuation_Init;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::mainLoop3()");
    };

  ASSERT(!_terminated);
  ASSERT(!_proof);


  _timer.reset();
  _timer.start();
  _timerIsOn = true;
  _startMemory = (float)GlobAlloc::occupiedByObjects(); 

  endOfInput();
  ASSERT(!_terminated);


 select_current: 
  GlobalStopFlag::check(); 
  if (!_pagedPassive.select(selectedClause)) 
    { 
      terminate(EmptyPassive); 
      _timer.stop();
      _timerIsOn = false; 
      return; 
    };

  if (_active.Total() >= _parameters.maxActive) 
    { 
      terminate(MaxActive); 
      _timer.stop();
      _timerIsOn = false; 
      return; 
    };


  if ((_parameters.orphan_murder)
      && (!selectedClause.isVIP())
      && (selectedClause.isOrphan()))
    {      
      ++_statistics.clauses.retained.passiveOrphansMurdered;
      _pagedPassive.popSelected();
      goto select_current;
    };



  _sharedReanimatedClause = reanimate3(selectedClause);
  _pagedPassive.popSelected();

  if (_sharedReanimatedClause)
    {
      ++_statistics.clauses.retained.total;
      _sharedReanimatedClause->SetNumber(_clauseNumberGenerator->generate());
      _sharedReanimatedClause->setMainSet(Clause::MainSetFree); 
      if (_parameters.useClauseNumIndex) 
	{
	  _sharedReanimatedClause->protect();
	  _clauseNumIndex.insert(_sharedReanimatedClause);
	  _sharedReanimatedClause->unprotect();
	};
      if (_parameters.showPassive) _parameters.showPassive(_sharedReanimatedClause);
    }
  else // !_sharedReanimatedClause
    {
      goto select_current; 
    };

  internalTick3();


  if (_sharedReanimatedClause->isAnswer()) 
    { 
      rememberAnswer3(_sharedReanimatedClause);
      _proof = _sharedReanimatedClause;
      _sharedReanimatedClause = 0;
      terminate(Proof);  
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_SelectCurrent;
      return; 
    };


  if (_parameters.backward_subsumption) 
    if (backSubsumeBy(_sharedReanimatedClause)) 
      if (_parameters.orphan_murder) killActiveOrphans3();      
    

  disposeObsolete();
  internalTick3();


 back_simplify:

  backSimplifyBy3(_sharedReanimatedClause);

  if (_proof) 
    {
      terminate(Proof);  
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_BackSimplify;
      return; 
    };    

  disposeObsolete();
  makeActive3(_sharedReanimatedClause);
  recentlyReanimatedClause = _sharedReanimatedClause;
  _sharedReanimatedClause = 0; // not to be destroyed as reanimated
  
  
  if (_parameters.outputStreamForPassiveProfiles &&
      (_active.Total() % _parameters.showPassiveProfileInterval == 0))
    {
      _pagedPassive.outputProfile(*_parameters.outputStreamForPassiveProfiles);
    };

  
  eliminateAllocationBufferDeficit3();  

  _infDispatcher.newQuery(recentlyReanimatedClause);

  

  // some clauses can be waiting in _forwardProcessing,
  // so we call _infDispatcher.nextInference() later

 process_new_inference:
  processCollected3();
  if (_proof) 
    { 
      terminate(Proof); 
      _timer.stop();
      _timerIsOn = false; 
      _continuation = Continuation_MainLoop3_ProcessNewInference;
      return; 
    };
  
  if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
    { 
      terminate(MaxPassive); 
      _infDispatcher.abandonCurrentQuery();
      _timer.stop();
      _timerIsOn = false; 
      return; 
    };
  
  internalTick3();
  
  GlobalStopFlag::check();
  if (_infDispatcher.nextInference()) goto process_new_inference;
  
  _infDispatcher.abandonCurrentQuery();
 

  GlobalStopFlag::check();

  // destroy obsolete clauses
  ASSERT(_queueForBackSimplification.empty());
  disposeObsolete();
  ASSERT(_trash.Empty());
  GlobalStopFlag::check();
  goto select_current; 

};// void Master::mainLoop3()


void Master::backSimplifyBy3(Clause* simplifier)
{
  CALL("backSimplifyBy3(Clause* simplifier)");
  
  Clause* backSimplified; 
  bool somethingSimplified;
  bool onBranch;

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_MainLoop3_BackSimplify:
      somethingSimplified = true;
      goto process_new_inference;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VK::Master::backSimplifyBy3(..)");
    };


  somethingSimplified = false;
  if (backDemodulationQuery(simplifier))
    { 
      while (nextBackDemodulated(onBranch))
	{         
	  if (collectBackSimplified())
	    { 
	      backSimplified = _backDemodulation.simplified();
	      ASSERT(backSimplified->mainSet() == Clause::MainSetActive); 

	      if (onBranch)
		{
		  ++_statistics.inferences.simplifying.onBranchBackwardDemodulation;
		  ++_statistics.clauses.retained.activeSimplifiedByOnBranchDemodulation;
		}
	      else
		{
		  ++_statistics.inferences.simplifying.backwardDemodulation;
		  ++_statistics.clauses.retained.activeSimplifiedByDemodulation;
		};
      
	      enqueueForDisposal(backSimplified);  
	      somethingSimplified = true;

	    process_new_inference:
	      processCollected3(); // not before enqueueForDisposal3(backSimplified)!

	      if (_proof) 
		{ 
		  terminate(Proof);  
		  return; 
		};

	      _continuation = Continuation_Init;

	      if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
		{ 
		  terminate(MaxPassive); 
		  abandonBackDemodulationQuery();
		  return; 
		}; 
	    };
	  GlobalStopFlag::check();  
	  internalTick3();
	}; 
      abandonBackDemodulationQuery();
    }; // if (backDemodulationQuery(simplifier))
  if (_parameters.orphan_murder && somethingSimplified) killActiveOrphans3();
}; // void Master::backSimplifyBy3(Clause* simplifier)

Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl)
{
  CALL("reanimate3(PagedPassiveClauseStorage::StoredClause& cl)");
  ASSERT(_parameters.main_alg == 3); 

  if (!_reanimatedClauseCollector.loadClause(cl.literals(),
					     ClauseBackground::Reanimation))
    return 0;
  _reanimatedClauseCollector.addRules(cl.rules());
  cl.resetAncestors();
  Clause* ancestor;
  while (cl.nextAncestor(ancestor))
    _reanimatedClauseCollector.addAncestor(ancestor);
  _reanimatedClauseCollector.announceNewClause();
  if (!_reanimatedClauseCollector.finaliseClause(_reanimatedClause))
    return 0;

  _forwardProcessing.load(_reanimatedClause);
  if (!_forwardProcessing.anotherRetained(_retainedClause))
    return 0;
  Clause* res = 
    _clauseAssembler.assembleShared(*_retainedClause);
  
  if (!res)
    ++_statistics.clauses.retained.passiveRecycled;

  // some retained may remain in _forwardProcessing
  // and will be transfered to _pagedPassive later

  return res;
}; // Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl) 

void Master::processCollected3()
{
  CALL("processCollected3()");

  Clause* sharedRetainedAnswer;
        
 process_another_retained:
  while (_forwardProcessing.anotherRetained(_retainedClause))
    {
      if (_retainedClause->isAnswer())
	{
	  sharedRetainedAnswer = 
	    _clauseAssembler.assembleShared(*_retainedClause);
	  if (!sharedRetainedAnswer)
	    {
	      ++_statistics.clauses.generated.discardedTooLarge;
	    }
	  else
	    {
	      ++_statistics.clauses.retained.total;
	      ASSERT(sharedRetainedAnswer->isAnswer());
	      sharedRetainedAnswer->setMainSet(Clause::MainSetNew);
	      sharedRetainedAnswer->SetNumber(_clauseNumberGenerator->generate());
	      rememberAnswer3(sharedRetainedAnswer);
	      _proof = sharedRetainedAnswer;
	      return;
	    };
	}
      else // !_retainedClause->isAnswer()
	{
	  // transfer to paged passive
	  long clauseNumber = _clauseNumberGenerator->generate();
	  ostream* outputStreamForPagedPassive = 
	    (_parameters.showPassive)? (&cout) : 0;
          
	  if (!_clauseAssembler.assembleInPagedPassiveClauseStorage(_pagedPassive,
								    *_retainedClause,
								    clauseNumber,
								    outputStreamForPagedPassive,
								    false))
	    {
	      ++_statistics.clauses.generated.discardedTooLarge;
	    }
	  else
	    ++_statistics.clauses.retained.total;

	  eliminateAllocationBufferDeficit3();

	}; 
    }; // while (_forwardProcessing.anotherRetained(_retainedClause))
     
  if (!_newClauseQueue.isEmpty())
    {
      _forwardProcessing.load(_newClauseQueue);
      goto process_another_retained;
    };
}; // void Master::processCollected3()



void Master::makeActive3(Clause* cl)
{
  CALL("makeActive3(Clause* cl)");
  ++(_statistics.clauses.retained.selected);
  cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 
  _active.Insert(cl);
  _infDispatcher.makeActive(cl);
  if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
  if (_parameters.backward_subsumption) 
    { 
      integrateIntoBackwardSubsumptionIndex(cl);
    };
  if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
  if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
  if (_parameters.showActive) _parameters.showActive(cl);
}; // void Master::makeActive3(Clause* cl)


void Master::rememberAnswer3(Clause* cl)
{
  CALL("rememberAnswer3(Clause* cl)");  
  _answers.Insert(cl);
}; // void Master::rememberAnswer3(Clause* cl)

void Master::destroyPassive3()
{
  CALL("destroyPassive3()");
  ASSERT(_passive.isEmpty());
  _pagedPassive.reset();
}; // void Master::destroyPassive3()


bool Master::releaseMemory3(ulong size) // bytes
{
  CALL("releaseMemory3(ulong size)");
  if (size > PagedPassiveClauseStorage::pageSize()) 
    return false;
 try_free_page:
  if (_pagedPassive.recyclePage()) return true;
  if (_pagedPassive.cancelWorstClause()) 
    {
      ++_statistics.clauses.retained.passiveRecycled;
      goto try_free_page; 
    };
  return false;
}; // bool Master::releaseMemory3(ulong size) 

 
bool Master::releaseALotOfMemory3()
{
  CALL("releaseALotOfMemory3()");
  
  bool res = false;

  // first, recycle about 10% of passive clauses
  ulong numOfClausesToRecycle =
    _pagedPassive.numberOfCurrentlyKeptClauses() / 10;
  bool somePassiveClauseRecycled = false;
  while (numOfClausesToRecycle &&
	 _pagedPassive.cancelWorstClause())
    {
      ++_statistics.clauses.retained.passiveRecycled;
      --numOfClausesToRecycle;
      somePassiveClauseRecycled = true;
      BK::GlobalStopFlag::check();
    };
  while (_pagedPassive.recyclePage()) 
    {
      res = true;
    };
 
  
#ifdef CHANGE_WEIGHT_LIMIT_AFTER_MAJOR_RECYCLING
  if (somePassiveClauseRecycled)
    {
      _parameters.maxMaxWeight = _pagedPassive.maxPresentWeight();
      updateWeightLimit(_pagedPassive.maxPresentWeight());
    };
#endif


  
  
#ifdef DONT_RECYCLE_ACTIVE
  // now recycle about 10% of active clauses if possible  
  numOfClausesToRecycle = _active.Size() / 10;

  while (numOfClausesToRecycle)
    {
      Clause* cand = goodActiveCandidateForDisposal();
      if (!cand) return res;
      --numOfClausesToRecycle;
      res = true;
      enqueueForDisposal(cand);
      ++_statistics.clauses.retained.activeRecycled;
      BK::GlobalStopFlag::check();
    };
#endif

  return res;
}; // bool Master::releaseALotOfMemory3()


void Master::killActiveOrphans3()
{
  CALL("killActiveOrphans3()");
  ASSERT(!_active.mainIteratorIsEngaged());
  _active.resetMainIterator();
  for (Clause* cl = _active.Next(); cl; cl = _active.Next())
    if ((cl->Active())
	&& (!cl->IsVIP())
	&& (cl->IsOrphan()))
      {
	enqueueForDisposal(cl);
	++_statistics.clauses.retained.activeOrphansMurdered;
      };
}; // void Master::killActiveOrphans3()



void Master::eliminateAllocationBufferDeficit3()
{
  CALL("eliminateAllocationBufferDeficit3()");
  if (BK::GlobAlloc::activeBufferDeficit())
    {
      long disposalBound = 
	(long)(BK::GlobAlloc::occupiedByObjects() - 
	       (_parameters.allocation_buffer_price * 
		BK::GlobAlloc::activeBufferDeficit()));
      do
	{
	  while (!_pagedPassive.recyclePage())
	    {
	      if (_pagedPassive.cancelWorstClause()) 
		{
		  ++_statistics.clauses.retained.passiveRecycled;
		}
	      else
		{
		  if (BK::GlobAlloc::activeBufferDeficit()) 
		    BK::GlobAlloc::annulBufferDeficit();
		  return;
		};
	    };
	}
      while (BK::GlobAlloc::activeBufferDeficit() && 
	     BK::GlobAlloc::occupiedByObjects() > disposalBound);
      if (BK::GlobAlloc::activeBufferDeficit()) 
	BK::GlobAlloc::annulBufferDeficit();
    };

}; // void Master::eliminateAllocationBufferDeficit3()

//====================================================




