//
// File:         PropositionalSplitting.hpp
// Description:  Core module for propositional splitting without backtracking.
// Created:      Nov 07, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef PROPOSITIONAL_SPLITTING_H
#define PROPOSITIONAL_SPLITTING_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "ExpandingMultiset.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
#include "ComponentName.hpp"
#include "DivIntoComp.hpp"
namespace VK
{
  class TemporaryClause;
  class TemporaryClauseCollector;
  class ComponentName;
}; // namespace VK
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PROPOSITIONAL_SPLITTING
#  define DEBUG_NAMESPACE "PropositionalSplitting"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class PropositionalSplitting
  {
  public:
    enum Style
    {
      BinaryBlocking,
      HyperParallel
    };
    typedef BasicTempLitList::Element LitListElement;
    typedef TemporaryLiteral::Extras::Handle LitHandle; 
  public:
    PropositionalSplitting(ComponentName* componentNaming);
    ~PropositionalSplitting();
    void reset();
    void setStyle(Style s) { _style = s; };
    void setSmallBasicComponentInHyper(bool fl)
    {
      _smallBasicComponentInHyper = fl;
    };
    
    bool split(TemporaryClause& cl,
	       TemporaryClauseCollector* newClauseCollector,
	       ulong& numOfComponents,
	       ulong& numOfNewNames)
    {
      if (!adjustSizes(cl.literals()->length())) return false;
      if (_style == BinaryBlocking)
	return splitWithBinaryBlocking(cl,
				       newClauseCollector,
				       numOfComponents,
				       numOfNewNames);
      return splitWithHyperParallel(cl,
				    newClauseCollector,
				    numOfComponents,
				    numOfNewNames);
    };
    
    
    
  private:
    PropositionalSplitting() {};
    bool splitWithBinaryBlocking(TemporaryClause& cl,
				 TemporaryClauseCollector* newClauseCollector,
				 ulong& numOfComponents,
				 ulong& numOfNewNames);
    bool splitWithHyperParallel(TemporaryClause& cl,
				TemporaryClauseCollector* newClauseCollector,
				ulong& numOfComponents,
				ulong& numOfNewNames);
    bool collectNameIntroduction(TERM::Functor branchNum,
				 TERM::Polarity branchPolarity,
				 BasicTempLitList& component,
				 TemporaryClauseCollector* newClauseCollector,
				 Clause* nameIntroClause);
    bool adjustSizes(ulong numOfLiterals);
  private:
    Style _style;
    bool _smallBasicComponentInHyper;
    ComponentName* _componentNaming;
    DivIntoComp<VampireKernelConst::MaxNumOfVariables> _divisionIntoComponents;
    BK::Array<BK::GlobAlloc,LitListElement*,BK::FastRelocation<LitListElement*>,32UL> _numLitMap; 
    BK::Array<BK::GlobAlloc,BasicTempLitList,BK::CustomRelocation<BasicTempLitList>,32UL> _components;   
    BK::ExpandingMultiset<BK::GlobAlloc,ulong,32UL,PropositionalSplitting> _registeredCompNums;
    BK::Array<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _componentName;  
    BK::Array<BK::GlobAlloc,bool,BK::FastRelocation<bool>,32UL> _newCompName;
    ulong _maximalNumOfLiteralsSoFar;
  }; // class PropositionalSplitting
}; // namespace VK



//===================================================================

#endif
