//
// File:         TemporaryClause.cpp
// Description:  Structure for temporary clauses.
// Created:      Nov 07, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "TemporaryClause.hpp"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//===================================================================

TemporaryClause::Context* TemporaryClause::Context::_current; 
// ^ implicitely initialised by 0

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause"
#endif
#include "debugMacros.hpp"
//===================================================================


TemporaryClause::TemporaryClause() 
{    
};


TemporaryClause::~TemporaryClause()
{
  CALL("destructor ~TemporaryClause()");
}; // TemporaryClause::~TemporaryClause()
  


void TemporaryClause::reset()
{
  CALL("reset()");
  TempLitListExtras::freshList(_literals);
  _background.Reset();
  _persistentClauseNumberIsSet = false;
  _universalId = Context::getCurrent()->generateUniversalId();
}; // void TemporaryClause::reset()



ostream& TemporaryClause::outputAnnounced(ostream& str) const
{
  CALL("output(ostream& str) const");
  str << "ANN " << _universalId << " ..... ";  
  ClauseBackground::outputRules(str << '[',_background.Rules());

  for (Clause* const * anc = _background.end(); 
       anc > _background.begin(); 
       /* nothing here */)
    {
      --anc;
      str << (*anc)->Number() << ' '; 
    };
  str << " ...]";
  return str;
}; // ostream& TemporaryClause::outputAnnounced(ostream& str) const


ostream& TemporaryClause::outputComplete(ostream& str) const
{
  CALL("output(ostream& str) const");
  _literals->output(str << "GEN " << _universalId << " ");
  ClauseBackground::outputRules(str << '[',_background.Rules());
  
  for (Clause* const * anc = _background.end(); 
       anc > _background.begin(); 
       /* nothing here */)
    {
      --anc;
      str << (*anc)->Number() << ' '; 
    };
  str << ']';  
  return str;
}; // ostream& TemporaryClause::outputComplete(ostream& str) const


//=================================================================
