//
// File:         TemporaryClauseQueue.hpp
// Description:  Queues for communicating temporary clauses
//               between different modules.
// Created:      Nov 08, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef TEMPORARY_CLAUSE_QUEUE_H
#define TEMPORARY_CLAUSE_QUEUE_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "ObjectPool.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryClause.hpp"
using namespace std;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_QUEUE
#  define DEBUG_NAMESPACE "TemporaryClauseQueue"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class TemporaryClauseQueue
  {
  public:
    typedef 
    BK::ObjectPool<TemporaryClause>::EntryHandle 
    ClauseHandle;

  public:
    TemporaryClauseQueue();
#ifndef NO_DEBUG
    TemporaryClauseQueue(const char* name);
#endif
    ~TemporaryClauseQueue();
    void reset();
    bool isEmpty() const { return _nextIn == _nextOut; };
    ulong size() const;
    void push(const ClauseHandle& cl);
    void collect(ClauseHandle& cl);
    ClauseHandle pop();
    void popTo(ClauseHandle& cl);
    void popTo(TemporaryClauseQueue& receiver);
    ostream& output(ostream& str) const;
  private:
    TemporaryClauseQueue& operator=(const TemporaryClauseQueue&) 
    { 
      return *this;
    };
    void incNextIn();
    void incNextOut()
    {
      ++_nextOut;
      if (_nextOut == _store.size()) _nextOut = 0UL;
    };

  private:
    BK::Array<BK::GlobAlloc,ClauseHandle,BK::CustomRelocation<ClauseHandle>,32UL> _store;
    ulong _nextIn;
    ulong _nextOut;
#ifndef NO_DEBUG
    const char* _name;
#endif
  }; // class TemporaryClauseQueue
}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE_QUEUE
#  define DEBUG_NAMESPACE "TemporaryClauseQueue"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  inline ulong TemporaryClauseQueue::size() const
  {
    return 
      (_nextIn >= _nextOut) ?
      (_nextIn - _nextOut)
      :
      (_store.size() - (_nextOut - _nextIn));
  }; // ulong TemporaryClauseQueue::size() const
}; // namespace VK

//===================================================================

#endif
