//
// File:         TemporaryLiteral.cpp
// Description:  Temporary literal = 
//                  flatterm representation +
//                  attributes/characteristics
// Created:      Oct 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#include "TemporaryLiteral.inl" 
using namespace BK;
using namespace VK;
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_LITERAL
#  define DEBUG_NAMESPACE "TemporaryLiteral"
#endif
#include "debugMacros.hpp"
//===================================================================

TemporaryLiteral::Context* 
TemporaryLiteral::Context::_current; // implicitely initialised to 0



ulong TemporaryLiteral::minPosEqLHSSize() const
{
  CALL("minPosEqLHSSize() const");
  ASSERT(_debugLiteralComplete);
  ASSERT(isEquality() && isPositive());
  if (isOrderedEquality())
    {
      return arg2()->size();
    }
  else // unordered 
    {
      long w1 = arg1()->size();
      long w2 = arg2()->size();
      return (w1 > w2) ? w2 : w1;
    }; 
}; // ulong TemporaryLiteral::minPosEqLHSSize() const


bool TemporaryLiteral::pushTerm(const Flatterm* term)
{
  CALL("pushTerm(const Flatterm* term)");
  ASSERT(!_debugLiteralComplete);
  Flatterm::ConstIterator iter(term); 
  do
    {
      if (!pushSym(iter.CurrPos()->symbolRef(),
		   iter.CurrPos()->isSubstitutionPosition())) 
	return false;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
  return true;
};  // bool TemporaryLiteral::pushTerm(const Flatterm* term)

 
bool TemporaryLiteral::pushBinaryLit(const TERM& header,
				     const Flatterm* arg1,
				     const Flatterm* arg2,
				     const LiteralProperties& prop)
{
  CALL("pushBinaryLit(const TERM& header,..)");
  ASSERT(header.arity() == 2);
  _openFlatterm.Reset();
  _size = 3UL; // ==  header.arity() + 1 
  _weightCollected = false;
  _variableOccurencesCollected = false;
  _resistanceComputed = false;
  DOP(_debugLiteralComplete = false);
  DOP(_arg1 = 0);
  DOP(_arg2 = 0);
  _openFlatterm.PushNonconstFun(header); 
  _header = header;
  _properties = prop;

  Flatterm::ConstIterator iter(arg1); 
  do
    {
      if (!pushSym(iter.CurrPos()->symbolRef(),false)) return false;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm());
  
  iter.Reset(arg2);
  do
    {
      if (!pushSym(iter.CurrPos()->symbolRef(),false)) return false;
      iter.NextSym();
    }
  while (iter.NotEndOfTerm()); 
  endOfLiteral();
  return true;
}; // bool TemporaryLiteral::pushBinaryLit(const TERM& header,..)


bool TemporaryLiteral::instantiateVar(TemporaryLiteral& new_lit,
				      ulong var,
				      const Flatterm* subst_term,
				      bool unmarkSubstPositionsAbove)
{
  CALL("instantiateVar(..)");
  TERM new_header = header();
  ASSERT(new_header.arity());
  if(new_header.IsEquality()) new_header.MakeUnordered();
  new_lit.pushNonpropHeader(new_header,properties());
  Flatterm::Iterator iter(flatterm());
  Flatterm::ConstIterator subst_iter;
  iter.NextSym(); // skipping the header
  while (iter.NotEndOfTerm())
    {
      TERM sym = iter.CurrPos()->Symbol();  
      if (sym.isVariable() && sym.var() == var) 
	{
	  subst_iter.Reset(subst_term);             
	  do
	    {
	      if (!new_lit.pushSym(subst_iter.CurrPos()->symbolRef(),
				   subst_iter.CurrPos()->isSubstitutionPosition())) 
		return false;             
	      subst_iter.NextSym();
	    }
	  while (subst_iter.NotEndOfTerm());
	}
      else 
	if (!new_lit.pushSym(sym,iter.CurrPos()->isSubstitutionPosition())) 
	  return false;
      iter.NextSym();
    };

  new_lit.endOfLiteral();

  if (unmarkSubstPositionsAbove)
    {
      iter.Reset(flatterm());
      Flatterm::Iterator newLitIter(new_lit.flatterm());
      iter.NextSym();
      newLitIter.NextSym();
      while (iter.NotEndOfTerm())
	{
	  ASSERT(newLitIter.NotEndOfTerm());
	  TERM sym = iter.CurrPos()->Symbol();  
	  if (sym.isVariable() && sym.var() == var) 
	    {
	      new_lit.flatterm()->unmarkSubstitutionPositionsAbove(newLitIter.CurrPos());
	      iter.NextSym();
	      newLitIter.after();
	    }
	  else
	    {
	      iter.NextSym();
              newLitIter.NextSym();
	    };
	};

    };

  return true;
}; // bool TemporaryLiteral::instantiateVar(..)




void TemporaryLiteral::renameVariables(INT_VAR_RENAMING& var_renaming)
{
  CALL("renameVariables(INT_VAR_RENAMING& var_renaming)");
  if (isEquality())
    {
      arg1()->renameVariables(var_renaming);
      arg2()->renameVariables(var_renaming);
    }
  else 
    flatterm()->renameVariables(var_renaming);
};



bool TemporaryLiteral::preferredNegative(TemporaryLiteral& lit)
{
  CALL("preferredNegative(TemporaryLiteral& lit)"); 
  ASSERT(_debugLiteralComplete);
  ASSERT(lit._debugLiteralComplete);
  if (isEquality())
    { 
      ASSERT(isNegative());
      if (lit.isEquality())
	{
	  ASSERT(lit.isNegative());
	  // both are negative equalities
	  return preferredNegativeEq(lit);     
	}
      else return true;
    }
  else
    if (lit.isEquality()) { return false; }
    else // both are not equalities
      {
	return preferredNegativeNoneq(lit);
      };
}; // bool TemporaryLiteral::preferredNegative(TemporaryLiteral& lit)
   



BK::Comparison TemporaryLiteral::compareWithVarWeighting(TemporaryLiteral& lit,const VAR_WEIGHTING& var_weighting)
{
  CALL("compareWithVarWeighting(TemporaryLiteral& lit,const VAR_WEIGHTING& var_weighting)");
  ASSERT(_debugLiteralComplete);
  ASSERT(lit._debugLiteralComplete);

  BK::Comparison hd_cmp = Signature::current()->compareHeaderPrecedences(header().polarity(),
									 header().functor(),
									 lit.header().polarity(),
									 lit.header().functor());

  if (hd_cmp != BK::Equal) return hd_cmp; 
  
  // here the headers are equal
  if (size() > lit.size()) return BK::Greater; 
  if (size() < lit.size()) return BK::Less; 
  

  // compare lexicographically
  BK::Comparison var_cmp;
  const Flatterm* fst = flatterm() + 1;
  const Flatterm* const end_of_fst = flatterm()->after();
  const Flatterm* snd = lit.flatterm();
 check_end:
  if (fst == end_of_fst) return BK::Incomparable;
  snd++;
  if (fst->Symbol() != snd->Symbol())
    {
      if (fst->Symbol().isVariable()) 
	{
	  if (snd->Symbol().isVariable()) 
	    {
	      // different variables
	      var_cmp = var_weighting.Compare(fst->Symbol().var(),snd->Symbol().var());
	      if (var_cmp != BK::Equal) { return var_cmp; };
	    }
	  else 
	    return BK::Less; // a variable is less than a functor
	}
      else 
	{
	  if (snd->Symbol().isVariable()) return BK::Greater;
	  // both are functors
	  return fst->Symbol().comparePrecedence(snd->Symbol()); 
	}; 
    };
  fst++;
  goto check_end;
}; // BK::Comparison TemporaryLiteral::compareWithVarWeighting(TemporaryLiteral& lit,const VAR_WEIGHTING& var_weighting)



BK::Comparison TemporaryLiteral::compare(TemporaryLiteral& lit)
{
  CALL("compare(TemporaryLiteral& lit) const");
  
  if (isEquality())
    {
      if ((lit.isEquality())
	  && (polarity() == lit.polarity()))
	return compareEq(lit); 
      return (isPositive()) ? BK::Less : BK::Greater;
    }
  else
    if (lit.isEquality())
      {
	return (lit.isPositive()) ? BK::Greater : BK::Less;
      };

  // both are not equations
 
  return SimplificationOrdering::current()->compareAtoms(flatterm(),
							 weight(),
							 lit.flatterm(),
							 lit.weight());

}; // BK::Comparison TemporaryLiteral::compare(TemporaryLiteral& lit)

 

    
void TemporaryLiteral::collectVarDepthProfile(VarDepthProfile& profile)
{
  CALL("collectVarDepthProfile(VarDepthProfile& profile)");
  profile.reset();    
  if (_weightCollected)
    {
      if (_weight.monomials().isZero()) return;
    }
  else
    if (_variableOccurencesCollected)
      {
	if (_variableOccurences.isZero()) return; 
      };
         
  Flatterm::Iterator iter(flatterm());
  SmartDepthCounter<VampireKernelConst::MaxTermDepth> depth_counter;
  depth_counter.Reset(1UL); 
  while (iter.NotEndOfTerm())
    {
      depth_counter.NextSym(iter.CurrPos()->symbolRef());  
      if (iter.CurrPos()->isVariable())
	{ 
	  profile.add(depth_counter.Depth());      
	};
   
      iter.NextSym();
    };
}; // void TemporaryLiteral::collectVarDepthProfile(VarDepthProfile& profile)
 
    
ostream& TemporaryLiteral::output(ostream& str) const
{
  if (isEquality())
    { 
      str << '(';
      arg1()->outputTerm(str) << ')' << header() << '(';
      arg2()->outputTerm(str);
      str << ')';
    }
  else 
    flatterm()->outputTerm(str);

  if (Context::getCurrent()->_showLiteralProperties &&
      !_properties.isDefault())
    {
      str << " <@[" << _properties << "]";
    };
  return str; 
}; // ostream& TemporaryLiteral::output(ostream& str) const

float TemporaryLiteral::computeResistance(ulong depthCoeff,float constCoeff,float varCoeff) const
{
  CALL("computeResistance(ulong depthCoeff,float constCoeff,float varCoeff) const");
  if (isEquality())
    {
      if (isOrderedEquality())
	{
	  return arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	}
      else 
	{
	  if (isNegative())
	    {
	      if (arg1()->isVariable())
		{
		  return arg2()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
		};
	      if (arg2()->isVariable())
		{
		  return arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
		};
	      float res1 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      float res2 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      return (res1 < res2) ? res1 : res2;
	    }
	  else
	    {
	      // positive unordered equality
	      float res1 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      float res2 = arg1()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
	      return (res1 < res2) ? res1 : res2;
	    };
	};
    }
  else
    return flatterm()->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
}; // float TemporaryLiteral::computeResistance(ulong depthCoeff,float constCoeff,float varCoeff) const



inline 
BK::Comparison TemporaryLiteral::compareEq(TemporaryLiteral& lit)
{
  CALL("compareEq(TemporaryLiteral& lit)");
  ASSERT(isEquality()); 
  ASSERT(lit.isEquality());
  ASSERT(header().polarity() == lit.header().polarity());

  if (isOrderedEquality())
    {
      if (lit.isOrderedEquality()) return compareEqOO(lit); 
      return compareEqOU(lit);
    }; 
  // this is unordered  

  if (lit.isOrderedEquality()) return BK::inverse(lit.compareEqOU(*this)); 
  // both are unordered
  return compareEqUU(lit);  
}; // BK::Comparison TemporaryLiteral::compareEq(TemporaryLiteral& lit)


inline 
BK::Comparison TemporaryLiteral::compareEqOO(TemporaryLiteral& lit)
{
  CALL("compareEqOO(TemporaryLiteral& lit)");
  ASSERT(isEquality());
  ASSERT(isOrderedEquality());
  ASSERT(header() == lit.header());



  BK::Comparison cmp11 = 
    SimplificationOrdering::current()->compare(arg1(),arg1Weight(),
					       lit.arg1(),lit.arg1Weight()); 
  if (cmp11 != BK::Equal) return cmp11;
  return SimplificationOrdering::current()->compare(arg2(),arg2Weight(),
						    lit.arg2(),lit.arg2Weight()); 
}; // BK::Comparison TemporaryLiteral::compareEqOO(TemporaryLiteral& lit)




inline 
BK::Comparison TemporaryLiteral::compareEqOU(TemporaryLiteral& lit)
{
  CALL("compareEqOU(TemporaryLiteral& lit)");
  // in the comments below * means "incomparable"
  ASSERT(isEquality());
  ASSERT(lit.isEquality());
  ASSERT(isOrderedEquality());
  ASSERT(lit.isUnorderedEquality());
  ASSERT(isPositive() == lit.isPositive());

  // arg1() > arg2()
  switch (SimplificationOrdering::current()->compare(arg1(),arg1Weight(),
						     lit.arg1(),lit.arg1Weight()))
    {
    case BK::Less:
      // arg2() < arg1() < lit.arg1()
      return BK::Less;
	  
    case BK::Equal: 
      // arg1() = lit.arg1() * lit.arg2(), compare arg2() with lit.arg2()
      return SimplificationOrdering::current()->compare(arg2(),arg2Weight(),
							lit.arg2(),lit.arg2Weight());	
    case BK::Greater: 
      // arg1() > lit.arg1(), compare arg1() with lit.arg2()
      return SimplificationOrdering::current()->compare(arg1(),arg1Weight(),
							lit.arg2(),lit.arg2Weight());
    case BK::Incomparable:         
      // compare arg2() with lit.arg1() and lit.arg2()
      switch (SimplificationOrdering::current()->compare(arg2(),arg2Weight(),
							 lit.arg1(),lit.arg1Weight()))
	{
	case BK::Less: 
	  if (SimplificationOrdering::current()->lessOrEqual(arg1(),arg1Weight(),
							     lit.arg2(),lit.arg2Weight()))
	    {
	      return BK::Less;
	    }
	  else
	    return BK::Incomparable;

	case BK::Equal:
	  return SimplificationOrdering::current()->compare(arg1(),arg1Weight(),
							    lit.arg2(),lit.arg2Weight());

	case BK::Greater:
	  // arg1() * lit.arg1(), arg2() > lit.arg1()
	  switch (SimplificationOrdering::current()->compare(arg1(),arg1Weight(),
							     lit.arg2(),lit.arg2Weight()))
	    {
	    case BK::Less: return BK::Less;
	    case BK::Equal: return BK::Greater;
	    case BK::Greater: return BK::Greater;
	    case BK::Incomparable: 
	      if (SimplificationOrdering::current()->greaterOrEqual(arg2(),arg2Weight(),
								    lit.arg2(),lit.arg2Weight()))
		{
		  return BK::Greater;
		}
	      else
		return BK::Incomparable;
	    };

	case BK::Incomparable:
	  // arg1() * lit.arg1(), arg2() * lit.arg1()
	  if (SimplificationOrdering::current()->less(arg1(),arg1Weight(),
						      lit.arg2(),lit.arg2Weight()))
	    {
	      return BK::Less;
	    }
	  else
	    return BK::Incomparable;
	};

    };

  ICP("DE100");
  return BK::Incomparable; 
}; // BK::Comparison TemporaryLiteral::compareEqOU(TemporaryLiteral& lit)





inline 
BK::Comparison TemporaryLiteral::compareEqUU(TemporaryLiteral& lit)
{
  CALL("compareEqUU(TemporaryLiteral& lit)");
  // compare {arg1(), arg2()} with {lit.arg1(), lit.arg2()}
  // "*" in the comments means "incomparable"
  ASSERT(isEquality());
  ASSERT(lit.isEquality());
  ASSERT(isUnorderedEquality());
  ASSERT(lit.isUnorderedEquality());
  ASSERT(isPositive() == lit.isPositive());

  BK::Comparison cmp = SimplificationOrdering::current()->compare(arg1(),arg1Weight(),lit.arg1(),lit.arg1Weight());
  switch (cmp)
    {
    case BK::Greater:
      // arg1() > lit.arg1()
      cmp = SimplificationOrdering::current()->compare(arg1(),arg1Weight(),lit.arg2(),lit.arg2Weight()); 
      switch (cmp)
	{
	case BK::Greater:
	  // arg1() > lit.arg1(), arg1() > lit.arg2()
	  return BK::Greater;
        
	case BK::Less:
	  // arg1() > lit.arg1(), arg1() < lit.arg2(). Not an error since v2.73
	  return SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
       
	case BK::Incomparable:
	  // arg1() > lit.arg1(), arg1() * lit.arg2()
	  cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
	  switch (cmp)
	    {
	    case BK::Greater:
	      // arg1() > lit.arg1(), arg1() * lit.arg2(), arg2() > lit.arg2()
	      return BK::Greater;
            
	    case BK::Less:
	      // arg1() > lit.arg1(), arg1() * lit.arg2(), arg2() < lit.arg2()
	      return BK::Incomparable;
           
	    case BK::Incomparable:
	      // arg1() > lit.arg1(), arg1() * lit.arg2(), arg2() * lit.arg2()
	      return BK::Incomparable;
           
	    case BK::Equal:
	      // arg1() > lit.arg1(), arg1() * lit.arg2(), arg2() = lit.arg2()
	      return BK::Greater;
	    };
        
	case BK::Equal:
	  // arg1() > lit.arg1(), arg1() = lit.arg2() : error because lit.arg1() * lit.arg2()
#ifdef DEBUG_NAMESPACE 
	  ICP("DE20"); 
#endif
	  return BK::Incomparable;
	};

    case BK::Less:
      // arg1() < lit.arg1()
      cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg1(),lit.arg1Weight()); 
      switch (cmp)
	{
	case BK::Greater:
	  // arg1() < lit.arg1(), arg2() > lit.arg1(). Not an error since v2.73
	  return SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
        
	case BK::Less:
	  // arg1() < lit.arg1(), arg2() < lit.arg1()
	  return BK::Less; // was BK::Greater in the CASC-16 version (obvious incompleteness)
        
	case BK::Incomparable:
	  // arg1() < lit.arg1(), arg2() * lit.arg1()
	  cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
	  switch (cmp)
	    {
	    case BK::Greater:
	      // arg1() < lit.arg1(), arg2() * lit.arg1(), arg2() > lit.arg2()
	      return BK::Incomparable;
           
	    case BK::Less:
	      // arg1() < lit.arg1(), arg2() * lit.arg1(), arg2() < lit.arg2()
	      return BK::Less;
           
	    case BK::Incomparable:
	      // arg1() < lit.arg1(), arg2() * lit.arg1(), arg2() * lit.arg2()
	      return BK::Incomparable;
           
	    case BK::Equal:
	      // arg1() < lit.arg1(), arg2() * lit.arg1(), arg2() = lit.arg2()
	      return BK::Less; 
	    }; 
        
	case BK::Equal:
	  // arg1() < lit.arg1(), arg2() = lit.arg1() : error because arg1() * arg2()
#ifdef DEBUG_NAMESPACE 
	  ICP("DE40");
#endif
	  return BK::Incomparable;
	};
      
    case BK::Incomparable:
      // arg1() * lit.arg1()
      cmp = SimplificationOrdering::current()->compare(arg1(),arg1Weight(),lit.arg2(),lit.arg2Weight()); 
      switch (cmp)
	{
	case BK::Greater:
	  // arg1() * lit.arg1(), arg1() > lit.arg2()
	  cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg1(),lit.arg1Weight()); 
	  switch (cmp)
	    {
	    case BK::Greater:
	      // arg1() * lit.arg1(), arg1() > lit.arg2(), arg2() > lit.arg1()
	      return BK::Greater; 
           
	    case BK::Less:
	      // arg1() * lit.arg1(), arg1() > lit.arg2(), arg2() < lit.arg1()
	      return BK::Incomparable;
           
	    case BK::Incomparable:
	      // arg1() * lit.arg1(), arg1() > lit.arg2(), arg2() * lit.arg1()
	      return BK::Incomparable;
           
	    case BK::Equal:
	      // arg1() * lit.arg1(), arg1() > lit.arg2(), arg2() = lit.arg1()
	      return BK::Greater;
	    };
          
	case BK::Less:
	  // arg1() * lit.arg1(), arg1() < lit.arg2()
	  cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg1(),lit.arg1Weight()); 
	  switch (cmp)
	    {
	    case BK::Greater:
	      // arg1() * lit.arg1(), arg1() < lit.arg2(), arg2() > lit.arg1()
	      cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight());
	      ASSERT(cmp != BK::Equal);
	      return cmp;
          
	    case BK::Less:
	      // arg1() * lit.arg1(), arg1() < lit.arg2(), arg2() < lit.arg1()
	      return BK::Less;
            
	    case BK::Incomparable:
	      // arg1() * lit.arg1(), arg1() < lit.arg2(), arg2() * lit.arg1()
	      cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
	      if (cmp == BK::Less) return BK::Less;
	      return BK::Incomparable;
            
	    case BK::Equal:
	      // arg1() * lit.arg1(), arg1() < lit.arg2(), arg2() = lit.arg1()
	      return BK::Less; 
	    }; 
         

	case BK::Incomparable:
	  // arg1() * lit.arg1(), arg1() * lit.arg2()
	  cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg1(),lit.arg1Weight()); 
	  switch (cmp)
	    {
	    case BK::Greater:
	      // arg1() * lit.arg1(), arg1() * lit.arg2(), arg2() > lit.arg1()
	      cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
	      if (cmp == BK::Greater) return BK::Greater;
	      ASSERT(cmp != BK::Equal);
	      return BK::Incomparable;
           
	    case BK::Less:
	      // arg1() * lit.arg1(), arg1() * lit.arg2(), arg2() < lit.arg1()
	      return BK::Incomparable;
           
	    case BK::Incomparable:
	      // arg1() * lit.arg1(), arg1() * lit.arg2(), arg2() * lit.arg1()
	      return BK::Incomparable;
           
	    case BK::Equal:
	      // arg1() * lit.arg1(), arg1() * lit.arg2(), arg2() = lit.arg1()
	      return BK::Incomparable;
	    }; 
        
	case BK::Equal:
	  // arg1() * lit.arg1(), arg1() = lit.arg2()
	  cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg1(),lit.arg1Weight()); 
	  return cmp;
	};
      
    case BK::Equal:
      // arg1() = lit.arg1()
      cmp = SimplificationOrdering::current()->compare(arg2(),arg2Weight(),lit.arg2(),lit.arg2Weight()); 
      return cmp;
    };
  ICP("DE100");
  return BK::Incomparable;
}; // BK::Comparison TemporaryLiteral::compareEqUU(TemporaryLiteral& lit)





inline 
bool TemporaryLiteral::preferredNegativeEq(TemporaryLiteral& lit)
{
  CALL("preferredNegativeEq(TemporaryLiteral& lit)");
  ASSERT(isNegative());
  ASSERT(isEquality());
  ASSERT(lit.isNegative());
  ASSERT(lit.isEquality());
  ulong size1;
  ulong size2;
  if (isOrderedEquality()) { size1 = arg1()->size(); }
  else size1 = size() - 1; 
  if (lit.isOrderedEquality()) { size2 = lit.arg1()->size(); }
  else size2 = lit.size() - 1;   
  if (size1 > size2) return false;
  if (size1 < size2) return true;
  // size1 == size2
  return true; // can be refined  
}; // bool TemporaryLiteral::preferredNegativeEq(TemporaryLiteral& lit)


inline 
bool TemporaryLiteral::preferredNegativeNoneq(TemporaryLiteral& lit)
{
  CALL("preferredNegativeNoneq(TemporaryLiteral& lit)");
  ASSERT(_debugLiteralComplete);
  ASSERT(lit._debugLiteralComplete);
  ASSERT(!isEquality());
  ASSERT(!lit.isEquality());
  ulong size1;
  ulong size2;
  BK::Comparison lexCmp;
  // switch (GlobalHeaderWeighting.Compare(header(),lit.header()))
  switch (Signature::current()->compareHeaderPrecedences(header().polarity(),
							 header().functor(),
							 lit.header().polarity(),
							 lit.header().functor()))
    {
    case BK::Greater: return true;
    case BK::Less: return false;
    case BK::Incomparable: ICP("ICP0"); return false;
    case BK::Equal: 
      size1 = size();
      size2 = lit.size();
      if (size1 > size2) return false;
      if (size1 < size2) return true;
      lexCmp = flatterm()->CompareLex(lit.flatterm());
      if (lexCmp == BK::Less) return true;
      if (lexCmp == BK::Greater) return false;
      return true; // can be refined
    };
  ICP("ICP1");
  return false;
}; // bool TemporaryLiteral::preferredNegativeNoneq(TemporaryLiteral& lit) 



//=================================================================

