#include "binary_resolution.hpp"
#include "Clause.inl"
#include "TemporaryClauseCollector.hpp"
#include "ForwardProcessing.hpp"
#include "Statistics.hpp"

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
#  define DEBUG_NAMESPACE "BIN_ORD_RES"
#endif
#include "debugMacros.hpp"
//=================================================

VK::BIN_ORD_RES::BIN_ORD_RES(OR_INDEX* resIndex) : 
  OPEN_BR(1,"br 1") DCOMMA 
_statistics(0),
  index1(0),
  index2(1),
  ORMachine(resIndex),
  _clauseSizeLimit(VampireKernelConst::MaxClauseSize),
  _clauseCollector(0),
  _forwardProcessing(0),
  _dummyClauseCollector(DOP("BIN_ORD_RES::_dummyClauseCollector")),
  _noInfBetweenRulesFlag(false)
{
  weight_form.SetIndex(index1);
};


bool VK::BIN_ORD_RES::Load(Clause* c)
{
  CALL("Load(Clause* c)");
  if (!eligible(c)) return false;
  // reset everithing here
  ORMachine.SetIndex(index2);
  if (!premise1.Load(index1,c))
    {
      ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
      return false;
    };
  CollectWeightForm(c);
  _sideClauseCanNotBeRule = _noInfBetweenRulesFlag && c->IsRule();
  currentLit = 0;
  _rotationPossible = true;
  _continuation = ContFirstLit;
  return true;
}; // bool BIN_ORD_RES::Load(Clause* c)   

bool VK::BIN_ORD_RES::NextInference()
{
  CALL("NextInference()");
  RES_PAIR_LIST* candidate;
  switch (_continuation)
    {
    case ContFirstLit: goto current_lit;
    case ContFirstFactor: goto first_factor;
    case ContNextFactor: goto next_factor; 
    };
  
  ICP("ICP1");

 current_lit:
  if (currentLit > premise1.lastSelected()) return false; 
  ASSERT(premise1.nthSelectedLit(currentLit));
  if (!premise1.nthSelectedLitEligibleForResolution(currentLit))
    {
      ++currentLit;
      goto current_lit;
    };

  if (!ORMachine.CompileLiteral(premise1.nthSelectedLit(currentLit)))
    {
      ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
      ++currentLit;
      goto current_lit;
    };
  ORMachine.Inversepolarity();
  premise1.Cut(currentLit);
  Unifier::current()->UnbindAll();
  _rotationPossible = true;

 find_first_leaf:
  if (ORMachine.FindFirstLeaf())  
    {
    process_leaf: 
      _residualMainPremiseSize = weight_form.Compute();
      if (_residualMainPremiseSize > _clauseSizeLimit)
	{
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedBySize);
	  goto try_next_leaf;
	};

      ORMachine.ResetCandidatesFromLeaf();
      if (demodulateMainSelectedLits()) 
	{
	  while ((candidate  = ORMachine.NextCandidate()))
	    candidate->TheClause()->protectParamodulationsInto();
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;

    try_next_cand:
      candidate  = ORMachine.NextCandidate();
      if (!candidate) goto try_next_leaf;       
      if (_sideClauseCanNotBeRule && candidate->TheClause()->IsRule()) 
        {
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
        };
      if (_residualMainPremiseSize + 
	  candidate->TheClause()->practicallyMinResidualSize() >
	  _clauseSizeLimit)
        {
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedBySize);
	  // we assume that a whole branch is discarded since 
	  // there may be several factoring inferences afterwards
	  goto try_next_cand;
        };   
            
      if (!premise2.LoadOptimized(index2,candidate))
	{
	  ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
	  goto try_next_cand;
	};

      if (demodulateSideSelectedLits(!_mainSelectedLitDemodulation))
	goto try_next_cand;
      if (subsumeSidePremise()) goto try_next_cand;

      // first resolvent without factoring
       
      if (!CollectResult()) 
        {
	  goto first_factor;
        };
      _continuation = ContFirstFactor;
      return true;
       
    first_factor:
      ResetFactoringState();
       
    next_factor:
      if (NextFactor())
	{
	  if (demodulateMainSelectedLits()) 
	    {
	      premise2.TheClause()->protectParamodulationsInto();
	      goto next_factor;
	    }; 
	  if (demodulateSideSelectedLits(!_mainSelectedLitDemodulation)) 
	    goto next_factor; 
	  if (subsumeMainPremise()) goto next_factor;  
	  if (subsumeSidePremise()) goto next_factor; 
	  if (!CollectResult()) goto next_factor;            
	  _continuation = ContNextFactor;
	  return true;          
	};
  
      // no_more_factors:
      premise2.Unload();
      goto try_next_cand;
     
      // no more candidates in this leaf, try to find another one
    try_next_leaf:
      if (ORMachine.FindNextLeaf()) { goto process_leaf; };
    };
   
  // no more resolvents for this literal
  // try rotation
  if (!_rotationPossible || !premise1.nthSelectedLitRotation(currentLit))
    {
      premise1.UnmarkCut();
      ++currentLit;
      goto current_lit;
    };

  if (!ORMachine.CompileLiteral(premise1.nthSelectedLitRotation(currentLit)))
    {
      ++(_statistics->inferences.deduction.resolution.branchesDiscardedForOtherReasons);
      ++currentLit;
      goto current_lit;
    };
  ORMachine.Inversepolarity();
  _rotationPossible = false;  
  goto find_first_leaf;
}; // bool VK::BIN_ORD_RES::NextInference()


 
bool VK::BIN_ORD_RES::NextFactor()
{
  CALL("NextFactor()");
  if (premise2.WasFactored())
    {
      if (premise2.GetNextFactor())
	{
	  return true;
	}
      else 
	{
	  premise2.ResetFactoring();
	  if (premise1.WasFactored())
	    {
	      return premise1.GetNextFactor();
	    }
	  else
	    {
	      return premise1.GetFirstFactor(premise1.CutLit()+1);
	    };
	};
    }
  else 
    {
      if (premise2.GetFirstFactor(0))
	{
	  return true;
	}
      else
	{
	  if (premise1.WasFactored())
	    {
	      return premise1.GetNextFactor();
	    }
	  else
	    {
	      return premise1.GetFirstFactor(premise1.CutLit()+1);
	    };
	};
    };
}; // bool VK::BIN_ORD_RES::NextFactor()
 

inline bool VK::BIN_ORD_RES::CollectResult() 
{ 
  CALL("CollectResult()");

  if ((!premise1.TheClause()->Active())
      || (!premise2.TheClause()->Active()))
    return false;
  _clauseCollector->openNewClause(ClauseBackground::BinRes,
				  FirstPremise(),
				  SecondPremise());
  _clauseCollector->announceNewClause();
  LiteralProperties::SLModifier slmod1 = LiteralProperties::DefaultSLModifier;
  LiteralProperties::SLModifier slmod2 = LiteralProperties::DefaultSLModifier;


  if (premise1.resolvedLiteralProperties().deploymentMode() != LiteralProperties::DefaultDeploymentMode)
    {
      if (premise1.resolvedLiteralProperties().deploymentMode() == LiteralProperties::RuleHead)
	{
	  slmod1 = LiteralProperties::AccelerateSelection;
	}
      else 
	{
	  ASSERT(premise1.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);	    
	  slmod1 = LiteralProperties::PostponeSelection;
	};	
    };

  if (premise2.resolvedLiteralProperties().deploymentMode() != LiteralProperties::DefaultDeploymentMode)
    {
      if (premise2.resolvedLiteralProperties().deploymentMode() == LiteralProperties::RuleHead)
	{
	  slmod2 = LiteralProperties::AccelerateSelection;
	}
      else
	{
	  ASSERT(premise2.resolvedLiteralProperties().deploymentMode() == LiteralProperties::Subquery);	  
	  slmod2 = LiteralProperties::PostponeSelection;
	};	
    };

  return 
    premise1.collectUnmarkedLiterals(*_clauseCollector,slmod1) && 
    premise2.collectUnmarkedLiterals(*_clauseCollector,slmod2) &&
    _clauseCollector->finaliseClause();
}; // inline bool VK::BIN_ORD_RES::CollectResult() 


inline void VK::BIN_ORD_RES::ResetFactoringState()
{
  premise1.ResetFactoring();
  premise2.ResetFactoring();
}; // void VK::BIN_ORD_RES::ResetFactoringState() 
 
inline void VK::BIN_ORD_RES::CollectWeightForm(const Clause* c)
{
  weight_form.Reset();
  LiteralList* lits = c->LitList();
  for (LiteralList::Iterator iter(lits,lits->nth(lits->numOfSelectedLits()));
       iter.notEnd();
       iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      weight_form.Compile(iter.currentElement()->literal()); 
}; // inline void CollectWeightForm() 


bool VK::BIN_ORD_RES::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (premise1.TheClause()->hasSubsumedPremiseInstance()) return false;
  premise1.TheClause()->blockForSubsumption();

  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  
  if (premise1.collectAllLiterals(_dummyClauseCollector) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != premise1.TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };

 return_false:
  premise1.TheClause()->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.resolution.mainPremisesSubsumed);
 discarded:
  premise1.TheClause()->unblockForSubsumption();
  return true;
}; // bool VK::BIN_ORD_RES::subsumeMainPremise()


bool VK::BIN_ORD_RES::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (premise2.TheClause()->hasSubsumedPremiseInstance()) return false;
  premise2.TheClause()->blockForSubsumption();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);

  if (premise2.collectAllLiterals(_dummyClauseCollector) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      subsumingClause = 
	_forwardProcessing->canBeSubsumed(*_dummyClause);
      ASSERT(subsumingClause != premise2.TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };

 return_false:
  premise2.TheClause()->unblockForSubsumption();
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++(_statistics->inferences.deduction.resolution.sidePremisesSubsumed);
 discarded:
  premise2.TheClause()->unblockForSubsumption();
  return true;
}; // bool VK::BIN_ORD_RES::subsumeSidePremise()


bool VK::BIN_ORD_RES::demodulateMainSelectedLits()
{
  CALL("demodulateMainSelectedLits()");
  Clause* simplifyingClause;
  if (!_mainSelectedLitDemodulation) return false;
  premise1.TheClause()->blockForDemodulationFrom();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);
  if (premise1.collectSubtermsAvailableForParamodulationInto(_dummyClauseCollector,true) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      simplifyingClause = 
	_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
      ASSERT(simplifyingClause != premise1.TheClause());
      if (simplifyingClause) goto demodulated;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
    
 return_false:
  premise1.TheClause()->unblockForDemodulationFrom();
  return false;
  
 demodulated:
  premise1.TheClause()->protectParamodulationsInto();
  ++(_statistics->inferences.deduction.resolution.mainPremisesDemodulated);
 discarded:
  premise1.TheClause()->unblockForDemodulationFrom();
  return true;
}; // bool BIN_ORD_RES::demodulateMainSelectedLits()



bool VK::BIN_ORD_RES::demodulateSideSelectedLits(bool includeMarked)
{
  CALL("demodulateSideSelectedLits(bool includeMarked)");
  Clause* simplifyingClause;
  if (!_sideSelectedLitDemodulation) return false;
  premise2.TheClause()->blockForDemodulationFrom();
  _dummyClauseCollector.openNewClause(ClauseBackground::Unspecified);

  if (premise2.collectSubtermsAvailableForParamodulationInto(_dummyClauseCollector,
							     includeMarked) &&
      _dummyClauseCollector.finaliseClause(_dummyClause))
    {
      simplifyingClause = 
	_forwardProcessing->canBeRewrittenByDemodulation(*_dummyClause);
      ASSERT(simplifyingClause != premise2.TheClause());
      if (simplifyingClause) goto demodulated;
      goto return_false;
    }
  else // analyse the reason
    {
      if ((_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::PropositionalTautology) ||
	  (_dummyClauseCollector.rejectionReason() ==
	   TemporaryClauseCollector::EquationalTautology))
	{
	  ++(_statistics->inferences.deduction.resolution.redundantBranchesDiscarded);
	  goto discarded;
	};
      goto return_false;
    };
    
 return_false:
  premise2.TheClause()->unblockForDemodulationFrom();
  return false;
  
 demodulated:
  premise1.TheClause()->protectParamodulationsInto();
  premise2.TheClause()->protectParamodulationsInto();
  ++(_statistics->inferences.deduction.resolution.sidePremisesDemodulated);
  
 discarded:
  premise2.TheClause()->unblockForDemodulationFrom();
  return true;
}; // bool BIN_ORD_RES::demodulateSideSelectedLits(bool includeMarked)

inline
bool VK::BIN_ORD_RES::eligible(const Clause* cl)
{
  CALL("eligible(const Clause* cl)");
  const LiteralList* lits = cl->LitList();
  ulong numOfSelected = lits->numOfSelectedLits();
  for (const LiteralList::Element* el = cl->LitList()->firstElem();
       numOfSelected;
       (el = el->next(),--numOfSelected))
    {
      ASSERT(el);
      if (el->literal()->IsEquality())
	{
	  if (el->properties().useEqForResolution()) return true;
	}
      else 
	return true;
    };
  return false;
}; // bool BIN_ORD_RES::eligible(const Clause* cl)


//=============================================================
 
 
 
 
 
 
 
 
