// Revised:      Jul 12, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bug fixed in integrateForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)
//               Caused incompleteness. 
#include "or_index.hpp"
#include "Clause.inl"
#include "unification.hpp"
#include "SimplificationOrdering.hpp"
#include "SymbolInfo.hpp"
#include "LiteralProperties.inl"
using namespace BK;
using namespace VK;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
#define DEBUG_NAMESPACE "OR_INDEX"
#endif 
#include "debugMacros.hpp"
//================================================= 


OR_INDEX::OR_INDEX() : 
  _literalIndex(static_cast<OR_INDEX_TREE_NODE*>(0),"OR_INDEX::_literalIndex"),
  forward_para_from_var_tree(0),
  backward_para_tree(0),
  _pos1ChainingTree(static_cast<OR_INDEX_TREE_NODE*>(0),"OR_INDEX::_pos1ChainingTree"),
  _pos2ChainingTree(static_cast<OR_INDEX_TREE_NODE*>(0),"OR_INDEX::_pos2ChainingTree"),
  _neg1ChainingTree(static_cast<OR_INDEX_TREE_NODE*>(0),"OR_INDEX::_neg1ChainingTree"),
  _neg2ChainingTree(static_cast<OR_INDEX_TREE_NODE*>(0),"OR_INDEX::_neg2ChainingTree"),
  _chainingTable(0)
{
  CALL("constructor OR_INDEX()");
  DOP(_literalIndex.freeze());
  pseudo_literal[0] = TERM((TERM::Arity)1,(TERM::Functor)0,Signature::PositivePolarity);
  _specialisedArgComparison = 
    SimplificationOrdering::current()->reserveSpecialisedComparison(); 
}; // OR_INDEX::OR_INDEX()

OR_INDEX::~OR_INDEX() 
{ 
  CALL("destructor ~OR_INDEX()"); 
  ulong i;
  if (DestructionMode::isThorough())
    {        
      for (i = 0; i < _literalIndex.size(); i++)
	{
	  OR_INDEX_TREE_NODE::Destroy(_literalIndex[i]); 
	};
      ASSERT(_paramodulateFromVariables || (!forward_para_from_var_tree));
      OR_INDEX_TREE_NODE::Destroy(forward_para_from_var_tree); 
      OR_INDEX_TREE_NODE::DestroyBPI(backward_para_tree);
      for (i = 0; i < _pos1ChainingTree.size(); i++)
	{
	  OR_INDEX_TREE_NODE::Destroy(_pos1ChainingTree[i]);
	};
      for (i = 0; i < _pos2ChainingTree.size(); i++)
	{
	  OR_INDEX_TREE_NODE::Destroy(_pos2ChainingTree[i]);
	};
      for (i = 0; i < _neg1ChainingTree.size(); i++)
	{
	  OR_INDEX_TREE_NODE::Destroy(_neg1ChainingTree[i]);
	};
      for (i = 0; i < _neg2ChainingTree.size(); i++)
	{
	  OR_INDEX_TREE_NODE::Destroy(_neg2ChainingTree[i]);
	};
      ASSERT(_specialisedArgComparison);
      SimplificationOrdering::current()->releaseSpecialisedComparison(_specialisedArgComparison);
    };
}; // OR_INDEX::~OR_INDEX() 


void OR_INDEX::reset()
{
  CALL("reset()");
  ulong i;  
  for (i = 0; i < _literalIndex.size(); i++)
    {
      OR_INDEX_TREE_NODE::Destroy(_literalIndex[i]);  
      _literalIndex[i] = 0;
    };
 
 
  ASSERT(_paramodulateFromVariables || (!forward_para_from_var_tree));
  OR_INDEX_TREE_NODE::Destroy(forward_para_from_var_tree); 
  forward_para_from_var_tree = 0;


  OR_INDEX_TREE_NODE::DestroyBPI(backward_para_tree);
  backward_para_tree = 0;

  for (i = 0; i < _pos1ChainingTree.size(); i++)
    {
      OR_INDEX_TREE_NODE::Destroy(_pos1ChainingTree[i]);
      _pos1ChainingTree[i] = 0;
    };
  for (i = 0; i < _pos2ChainingTree.size(); i++)
    {
      OR_INDEX_TREE_NODE::Destroy(_pos2ChainingTree[i]);
      _pos2ChainingTree[i] = 0;
    };
  for (i = 0; i < _neg1ChainingTree.size(); i++)
    {
      OR_INDEX_TREE_NODE::Destroy(_neg1ChainingTree[i]);
      _neg1ChainingTree[i] = 0;
    };
  for (i = 0; i < _neg2ChainingTree.size(); i++)
    {
      OR_INDEX_TREE_NODE::Destroy(_neg2ChainingTree[i]);
      _neg2ChainingTree[i] = 0;
    };

}; // void OR_INDEX::reset()



void OR_INDEX::Integrate(Clause* cl)
{
  CALL("Integrate(Clause* cl)");
  LiteralList* lits = cl->LitList();
  LiteralList::Element* firstNonselectedLit = lits->nth(lits->numOfSelectedLits());
  ulong litNum = 0;


  for(LiteralList::Iterator iter(lits);iter.currentElement() != firstNonselectedLit;iter.next()) 
    {
      Integrate(iter.currentElement(),cl,litNum);
      litNum++;
    };
}; // void OR_INDEX::Integrate(Clause* const cl)

void OR_INDEX::Integrate(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)
{
  CALL("Integrate(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)");
  const TERM* litBody = lit->literal();
  if (litBody->arity())
    {
      if (paramodulation) integrateForBackwardSuperposition(litBody,cl,lit_num);
      if (litBody->IsEquality())
	{      
	  if (lit->properties().equalityMode() == LiteralProperties::DefaultEqualityMode)
	    {
	      if (litBody->Positive()) integrateForForwardSuperposition(litBody,cl,lit_num);
	    }
	  else 
	    {
	      ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolution ||
		     lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
	      // this equality literal is available for prop. resolution
	     
	      OR_INDEX_TREE_NODE* leaf = integrateEqForResolution(litBody);
	      ASSERT(leaf->Term().IsReference());
	      ASSERT((!leaf->Term().First()) || (leaf->Term().First() == litBody));
	      leaf->SetTerm(litBody);
	      leaf->AddPair(lit_num,cl); 
	      
	      if (lit->properties().equalityMode() == LiteralProperties::UseEqForResolution)
		{
		  // this equality literal can be used for paramod
		  if (litBody->Positive()) integrateForForwardSuperposition(litBody,cl,lit_num);
		}
	      else
		{
		  ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
		  // nothing to do here
		};
	    };
	}
      else // !litBody->IsEquality()
	{
	  OR_INDEX_TREE_NODE* leaf = IntegrateNonpropLit(litBody);
	  ASSERT(leaf->Term().IsReference());
	  ASSERT((!leaf->Term().First()) || (leaf->Term().First() == litBody));
	  leaf->SetTerm(litBody);
	  leaf->AddPair(lit_num,cl); 
	};
      if (_chainingTable && _chainingTable->transitivityLaw(litBody->functor()))
	{
	  integrateForChaining(litBody,cl,lit_num);
	};
    }
  else
    {
      OR_INDEX_TREE_NODE* leaf = IntegratePropLit(*litBody);    
      ASSERT(leaf->Term().IsReference());
      ASSERT((!leaf->Term().First()) || (leaf->Term().First() == litBody));
      leaf->SetTerm(litBody);
      leaf->AddPair(lit_num,cl); 
    };  
}; // void OR_INDEX::Integrate(const LiteralList::Element* lit,Clause* cl,unsigned lit_num)

void OR_INDEX::integrateBackSupRedex(const TERM* redex,Clause* cl,unsigned lit_num,
					    SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrateBackSupRedex(const TERM* redex,Clause* cl,unsigned lit_num,SimplificationOrdering::StoredConstraint* constr)");
  ASSERT(redex->isComplex());

  //DF; cout << "------- REDEX: " << redex << "\n";

  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_INTEGRATE(redex);

  OR_INDEX_TREE_NODE** tree = &backward_para_tree;
  OR_INDEX_TREE_NODE* leaf;
  pseudo_literal->Args()->SetFirst(const_cast<TERM*>(redex));
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  leaf = IntegrateArgs(tree);
  leaf->SetTerm(redex);
  leaf->addBackSupTriple(lit_num,cl,redex,constr);  
}; // void OR_INDEX::integrateBackSupRedex(const TERM* redex,Clause* cl,unsigned lit_num,SimplificationOrdering::StoredConstraint* constr)

void OR_INDEX::integrateForBackwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)
{
  CALL("integrateForBackwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->arity());
  _previousBackSupRedexes.reset();
  if (lit->IsEquality())
    {
      if (cl->isSupportClause())
	{
	  if ((lit->Positive() && (!_orientPosEqualitiesInSupport)) ||
	      (lit->Negative() && (!_orientNegEqualitiesInSupport)))
	    goto integrate_ordinary; // both arguments and no ord.constr.
	};
      if (lit->IsOrderedEq())
	{
	  // only the first argument can be rewritten
	  ASSERT(lit->Arg1()->IsReference());
	  integrateComplexSubtermForBackwardSuperposition(lit->Arg1()->First(),
							  cl,
							  lit_num,
							  0);
	}
      else // unordered equality
	{
	  const TERM* arg1ref = lit->Arg1();
	  const TERM* arg2ref = arg1ref->Next();
	  if (arg1ref->isVariable() || arg2ref->isVariable()) 
	    goto integrate_ordinary;

	  SimplificationOrdering::StoredConstraint* constraint;
	  
	  _specialisedArgComparison->load(*arg1ref,*arg2ref);

	  if ((_specialisedArgComparison->lhsVarInstMayOrder()) &&
	      ((constraint = SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison))))
	    {
	      integrateComplexSubtermForBackwardSuperposition(arg1ref->First(),
							      cl,
							      lit_num,
							      constraint);

	      SimplificationOrdering::current()->releaseConstraint(constraint);
	    }
	  else // no constraint
	    {
	      integrateComplexSubtermForBackwardSuperposition(arg1ref->First(),
							      cl,
							      lit_num,
							      0);
	    };
      

	  if ((_specialisedArgComparison->rhsVarInstMayOrder()) &&
	      ((constraint = SimplificationOrdering::current()->storedConstraintGreaterOrEqual(_specialisedArgComparison))))
	    {
	      integrateComplexSubtermForBackwardSuperposition(arg2ref->First(),
							      cl,
							      lit_num,
							      constraint);
	      SimplificationOrdering::current()->releaseConstraint(constraint);
	    }
	  else // no constraint
	    {
	      integrateComplexSubtermForBackwardSuperposition(arg2ref->First(),
							      cl,
							      lit_num,
							      0);
	    };
	  
	};
    }
  else // ordinary literal
    {
    integrate_ordinary:
      const bool* nonBasicArgs;
      if (_useInputNonBasicness &&
	  ((nonBasicArgs = lit->nonBasicArguments())))
	{
	  ASSERT(!lit->IsEquality());
	  ulong argNum = 0UL;
	  for (const TERM* arg = lit->Args(); !arg->IsNull(); arg = arg->Next())
	    {
	      if (!arg->isVariable() &&
		  !nonBasicArgs[argNum])
		integrateComplexSubtermForBackwardSuperposition(arg->First(),
								cl,
								lit_num,
								0);
	      ++argNum;
	    };
	}
      else
	{
	  for (const TERM* arg = lit->Args(); !arg->IsNull(); arg = arg->Next())
	    {
	      if (!arg->isVariable())
		integrateComplexSubtermForBackwardSuperposition(arg->First(),
								cl,
								lit_num,
								0);
	    };
	};
    };
}; // void OR_INDEX::integrateForBackwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)


void 
OR_INDEX::integrateComplexSubtermForBackwardSuperposition(const TERM* subterm,
							  Clause* cl,
							  unsigned litNum,
							  SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrateComplexSubtermForBackwardSuperposition(..)");
  ASSERT(subterm->isComplex());
  
  if (_backSupNoDuplicateRedexes)
    {
      for (const TERM** prev = _previousBackSupRedexes.begin();
	   prev != _previousBackSupRedexes.end();
	   ++prev)
	{
	  if (*prev == subterm) return;
	};
    };

  _previousBackSupRedexes.push(subterm);
  integrateBackSupRedex(subterm,cl,litNum,constr);
  
  const bool* nonBasicArgs;
  if (_useInputNonBasicness &&
      ((nonBasicArgs = subterm->nonBasicArguments())))
    {      
      ulong argNum = 0UL;
      for (const TERM* arg = subterm->Args();
	   !arg->IsNull();
	   arg = arg->Next())
	{
	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum]) 
	    integrateComplexSubtermForBackwardSuperposition(arg->First(),
							    cl,
							    litNum,
							    constr);
	  ++argNum;
	};

    }
  else
    {       
      for (const TERM* arg = subterm->Args();
	   !arg->IsNull();
	   arg = arg->Next())
	{
	  if (!arg->isVariable())
	    integrateComplexSubtermForBackwardSuperposition(arg->First(),
							    cl,
							    litNum,
							    constr);
	};
    };
}; // void OR_INDEX::integrateComplexSubtermForBackwardSuperposition(..)


void OR_INDEX::integrateForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)
{
  CALL("integrateForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->IsEquality());
  ASSERT(lit->Positive());
  OR_INDEX_TREE_NODE* leaf;
  const TERM* arg;
  TERM replacing_term;
  bool stop;
  const TERM* arg1ref;
  const TERM* arg2ref;

  if ((!_orientPosEqualitiesInSupport) && cl->isSupportClause())
    goto integrate_unconstrained;


  if (lit->IsOrderedEq()) 
    { 
      integratePreorderedForForwardSuperposition(lit,cl,lit_num); 
      return; 
    };

  arg1ref = lit->Args();
  arg2ref = arg1ref->Next();
  if (arg1ref->IsReference() && arg2ref->IsReference())
    {

      SimplificationOrdering::StoredConstraint* constraint;
      
      _specialisedArgComparison->load(*arg1ref,*arg2ref);

      if (_specialisedArgComparison->lhsVarInstMayOrder())
	{
	  constraint = SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);

	  integrateConstrainedForForwardSuperposition(*arg1ref,*arg2ref,cl,lit_num,constraint); 
	  if (constraint) SimplificationOrdering::current()->releaseConstraint(constraint);
	}
      else
	{
	  integrateConstrainedForForwardSuperposition(*arg1ref,*arg2ref,cl,lit_num,0);
	};
      
      if (_specialisedArgComparison->rhsVarInstMayOrder())
	{
	  constraint = SimplificationOrdering::current()->storedConstraintGreaterOrEqual(_specialisedArgComparison);   
	  integrateConstrainedForForwardSuperposition(*arg2ref,*arg1ref,cl,lit_num,constraint);
	  if (constraint) SimplificationOrdering::current()->releaseConstraint(constraint);
	}
      else
	{
	  integrateConstrainedForForwardSuperposition(*arg2ref,*arg1ref,cl,lit_num,0);
	};
      
      return;
    };

 integrate_unconstrained:
  // integrate the first argument
  arg = lit->Args();
  replacing_term = *(arg->Next());
  stop = false;
 integrate_arg: 
  if (arg->isVariable())
    {
      if (_paramodulateFromVariables)
	{
	  pseudo_literal[1] = *arg;
	  l_iter.Reset(pseudo_literal,1UL);
	  l_iter.Next();
	  leaf = IntegrateArgs(&forward_para_from_var_tree);
	  leaf->SetTerm(*arg);
	  leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);
	}
      else
	{
	  ++(_statistics->inferences.deduction.forwardSuperposition.branchesDiscardedWithParamodulationsFromVariables);
	};
    }
  else // complex term, integrate as ordinary literal
    {
      if (arg->First()->arity())
	{
	  l_iter.Reset(arg->First(),1UL);
	  l_iter.Next();
	  leaf = IntegrateArgs(FParaTree(*(arg->First())));
	}
      else
	{
	  leaf = *(FParaTree(*(arg->First())));
	  if (!leaf) 
	    {
	      leaf = OR_INDEX_TREE_NODE::NewLeaf();
	      *(FParaTree(*(arg->First()))) = leaf;
	    };
	}; 
      leaf->SetTerm(arg->First());
      leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);
    };    
   
  if (stop) return; 
  // integrate the second argument 
  replacing_term = *arg;
  arg = arg->Next();
  stop = true;
  goto integrate_arg;
}; // void OR_INDEX::integrateForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)



inline
void OR_INDEX::integrateConstrainedForForwardSuperposition(const TERM& lhs,const TERM& rhs,Clause* cl,
						  unsigned lit_num,SimplificationOrdering::StoredConstraint* constr)
{
  CALL("integrateConstrainedForForwardSuperposition(..)");
  ASSERT(lhs.IsReference());
  ASSERT(rhs.IsReference());  

  OR_INDEX_TREE_NODE* leaf;
  if (lhs.First()->arity())
    {
      l_iter.Reset(lhs.First(),1UL);
      l_iter.Next();
      leaf = IntegrateArgs(FParaTree(*(lhs.First())));
    }
  else
    {
      leaf = *(FParaTree(*(lhs.First())));
      if (!leaf) 
	{
	  leaf = OR_INDEX_TREE_NODE::NewLeaf();
	  *(FParaTree(*(lhs.First()))) = leaf;
	};
    }; 
  leaf->SetTerm(lhs.First());
  leaf->addForwardSupTriple(lit_num,cl,rhs,constr); 

}; // void integrateConstrainedForForwardSuperposition(const TERM& lhs,const TERM& rhs,Clause* cl,unsigned lit_num,SimplificationOrdering::StoredConstraint* constr)


void OR_INDEX::integratePreorderedForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)
{
  CALL("integratePreorderedForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)");
  ASSERT(lit->IsOrderedEq());
  OR_INDEX_TREE_NODE* leaf;
  // integrate only the first argument
  const TERM* arg = lit->Args();
  ASSERT(arg->IsReference());
  TERM replacing_term = *(arg->Next());   
  if (arg->First()->arity())
    {
      l_iter.Reset(arg->First(),1UL);
      l_iter.Next();
      leaf = IntegrateArgs(FParaTree(*(arg->First())));
    }
  else
    {
      leaf = *(FParaTree(*(arg->First())));
      if (!leaf) 
	{
	  leaf = OR_INDEX_TREE_NODE::NewLeaf();
	  *(FParaTree(*(arg->First()))) = leaf;
	};
    }; 
  leaf->SetTerm(arg->First());
  leaf->addForwardSupTriple(lit_num,cl,replacing_term,0);  
}; // void OR_INDEX::integratePreorderedForForwardSuperposition(const TERM* lit,Clause* cl,unsigned lit_num)



void OR_INDEX::integrateForChaining(const TERM* lit,Clause* cl,unsigned litNum)
{
  CALL("integrateForChaining(const TERM* lit,Clause* cl,unsigned litNum)");
  ASSERT(_chainingTable && _chainingTable->transitivityLaw(lit->functor()));
  if (lit->Positive())
    {
      integratePosForChaining(lit,cl,litNum);
    }
  else
    {
      integrateNegForChaining(lit,cl,litNum);
    };
}; // void OR_INDEX::integrateForChaining(const TERM* lit,Clause* cl,unsigned litNum)


void OR_INDEX::integratePosForChaining(const TERM* lit,Clause* cl,unsigned litNum)
{
  CALL("integratePosForChaining(const TERM* lit,Clause* cl,unsigned litNum)");

  const TERM* arg1ref = lit->Args();
  const TERM* arg2ref = arg1ref->Next();

  bool integrateArg1 = true;
  bool integrateArg2 = true;
  
  bool constrainArg1 = true;
  bool constrainArg2 = true;

  SimplificationOrdering::StoredConstraint* constraint1 = 0;
  SimplificationOrdering::StoredConstraint* constraint2 = 0;

  

  
  _specialisedArgComparison->load(*arg1ref,*arg2ref);
  
  if (!_specialisedArgComparison->lhsVarInstMayOrder())
    constrainArg1 = false;
  if (!_specialisedArgComparison->rhsVarInstMayOrder())
    constrainArg2 = false;


  // check if the first argument needs to be integrated
  // and constrained

  if (arg1ref->isVariable())
    {
      // singleton-variable restriction does not apply
      if (!_chainingThroughVariables)
	{
	  integrateArg1 = false;
	  constrainArg1 = false;
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  
	};
    };
    

  if (integrateArg1 && constrainArg1)
    {
      _specialisedArgComparison->specialiseLessOrEqual();
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;
      
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  constrainArg1 = false;
	  break;
      
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;

	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
	  break; 
	};
    };

  // check if the second argument needs to be integrated
  // and constrained

  if (arg2ref->isVariable())
    {
      // singleton-variable restriction does not apply
      if (!_chainingThroughVariables)
	{
	  integrateArg2 = false;
	  constrainArg2 = false;
	  ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	};
    };
    
  if (integrateArg2 && constrainArg2)
    {
      _specialisedArgComparison->specialiseGreaterOrEqual();
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess:
	  constrainArg2 = false;
	  break;

	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  integrateArg2 = false;
	  constrainArg2 = false;
	  break;

	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  break;
	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	  integrateArg2 = false;
	  constrainArg2 = false;
	  break;
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess:  
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater:  
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile: 
	  // same as below
	  break; 
	};
    };

  ASSERT(!constrainArg1 || integrateArg1);
  ASSERT(!constrainArg2 || integrateArg2);
  
  if (constrainArg1)
    {
      constraint1 = 
	SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);
    };
  
  if (constrainArg2)
    {
      constraint2 = 
	SimplificationOrdering::current()->storedConstraintGreaterOrEqual(_specialisedArgComparison);
    };
  
  if (integrateArg1)
    integratePos1ForChaining(lit->functor(),*arg1ref,*arg2ref,cl,litNum,constraint1);
  
  if (integrateArg2)
    integratePos2ForChaining(lit->functor(),*arg2ref,*arg1ref,cl,litNum,constraint2);


  if (constraint1) 
    SimplificationOrdering::current()->releaseConstraint(constraint1);

  if (constraint2) 
    SimplificationOrdering::current()->releaseConstraint(constraint2);

}; // void OR_INDEX::integratePosForChaining(const TERM* lit,Clause* cl,unsigned litNum)


void 
OR_INDEX::integratePos1ForChaining(ulong predNum,
				   const TERM& head,
				   const TERM& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integratePos1ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  OR_INDEX_TREE_NODE* leaf = 
    IntegrateArgs(&(_pos1ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; // void OR_INDEX::integratePos1ForChaining(..)


void 
OR_INDEX::integratePos2ForChaining(ulong predNum,
				   const TERM& head,
				   const TERM& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integratePos2ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  OR_INDEX_TREE_NODE* leaf = 
    IntegrateArgs(&(_pos2ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; // void OR_INDEX::integratePos2ForChaining(..)



void OR_INDEX::integrateNegForChaining(const TERM* lit,Clause* cl,unsigned litNum)
{
  CALL("integrateNegForChaining(const TERM* lit,Clause* cl,unsigned litNum)");
 
  const TERM* arg1ref = lit->Args();
  const TERM* arg2ref = arg1ref->Next();

  bool integrateArg1 = true;
  bool integrateArg2 = true;
  
  bool constrainArg1 = true;
  bool constrainArg2 = true;

  SimplificationOrdering::StoredConstraint* constraint1 = 0;
  SimplificationOrdering::StoredConstraint* constraint2 = 0;
 
  
  _specialisedArgComparison->load(*arg1ref,*arg2ref);
  
  if (!_specialisedArgComparison->lhsVarInstMayOrder())
    constrainArg1 = false;
  if (!_specialisedArgComparison->rhsVarInstMayOrder())
    constrainArg2 = false;

  // check if the first argument needs to be integrated
  // and constrained

  if (arg1ref->isVariable())
    {
      if (cl->singletonVar(arg1ref->var()))
	{	  
	  integrateArg1 = false;
	  constrainArg1 = false;
	}
      else
	if (!_chainingThroughVariables)
	  {
	    integrateArg1 = false;
	    constrainArg1 = false;
	    ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	    ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  };
    };

  if (integrateArg1 && constrainArg1)
    {
      _specialisedArgComparison->specialiseLessOrEqual();
      
      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;
      
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  constrainArg1 = false;
	  break;
      
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  integrateArg1 = false;
	  constrainArg1 = false;
	  break;

	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater: 
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
	  break; 
	};
    };

  // check if the second argument needs to be integrated
  // and constrained

  if (arg2ref->isVariable())
    {
      if (cl->singletonVar(arg2ref->var()))
	{	  
	  integrateArg2 = false;
	  constrainArg2 = false;
	}
      else
	if (!_chainingThroughVariables)
	  {
	    integrateArg2 = false;
	    constrainArg2 = false;
	    ++(_statistics->inferences.deduction.forwardChaining.branchesDiscardedWithChainingsThroughVariables);
	    ++(_statistics->inferences.deduction.backwardChaining.branchesDiscardedWithChainingsThroughVariables);
	  };
    };

  if (integrateArg2 && constrainArg2)
    {
      _specialisedArgComparison->specialiseGreater();

      switch (_specialisedArgComparison->category())
	{
	case SimplificationOrdering::SpecialisedComparison::AlwaysLess:
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::AlwaysEqual:
	  constrainArg2 = false;
	  break; 
      
	case SimplificationOrdering::SpecialisedComparison::AlwaysGreater:
	  integrateArg2 = false;
	  constrainArg2 = false;
	  break; 
      
	case SimplificationOrdering::SpecialisedComparison::CanBeLessOrEqual:
	  constrainArg2 = false;
	  break; 
      

	case SimplificationOrdering::SpecialisedComparison::CanBeGreaterOrEqual:
	  // same as below
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysLess:
	  // same as below 
	case SimplificationOrdering::SpecialisedComparison::VolatileButNearlyAlwaysGreater: 
	  // same as below 
	case SimplificationOrdering::SpecialisedComparison::AbsolutelyVolatile:
	  break;
	};
    };

  ASSERT(!constrainArg1 || integrateArg1);
  ASSERT(!constrainArg2 || integrateArg2);
  
  if (constrainArg1)
    {
      constraint1 = 
	SimplificationOrdering::current()->storedConstraintLessOrEqual(_specialisedArgComparison);
    };
  
  if (constrainArg2)
    {
      constraint2 = 
	SimplificationOrdering::current()->storedConstraintGreater(_specialisedArgComparison);
    };
  
  if (integrateArg1)
    integrateNeg1ForChaining(lit->functor(),*arg1ref,*arg2ref,cl,litNum,constraint1);
  
  if (integrateArg2)
    integrateNeg2ForChaining(lit->functor(),*arg2ref,*arg1ref,cl,litNum,constraint2);


  if (constraint1) 
    SimplificationOrdering::current()->releaseConstraint(constraint1);

  if (constraint2) 
    SimplificationOrdering::current()->releaseConstraint(constraint2);

}; // void OR_INDEX::integrateNegForChaining(const TERM* lit,Clause* cl,unsigned litNum)



void 
OR_INDEX::integrateNeg1ForChaining(ulong predNum,
				   const TERM& head,
				   const TERM& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integrateNeg1ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  OR_INDEX_TREE_NODE* leaf = 
    IntegrateArgs(&(_neg1ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; // void OR_INDEX::integrateNeg1ForChaining(..)


void 
OR_INDEX::integrateNeg2ForChaining(ulong predNum,
				   const TERM& head,
				   const TERM& tail,
				   Clause* cl,
				   unsigned litNum,
				   SimplificationOrdering::StoredConstraint* constraint)
{
  CALL("integrateNeg2ForChaining(..)");
  pseudo_literal[1] = head;
  l_iter.Reset(pseudo_literal,1UL);
  l_iter.Next();
  OR_INDEX_TREE_NODE* leaf = 
    IntegrateArgs(&(_neg2ChainingTree.sub(predNum)));
  leaf->SetTerm(head);
  leaf->addChainingQuadruple(litNum,cl,tail,constraint);
}; // void OR_INDEX::integrateNeg2ForChaining(..)




OR_INDEX_TREE_NODE* OR_INDEX::IntegratePropLit(TERM lit)
{
  CALL("IntegratePropLit(TERM lit)");
  OR_INDEX_TREE_NODE** tree = Tree(lit);
  if (!(*tree))
    {
      // no tree for this header yet, create it here
      *tree = OR_INDEX_TREE_NODE::NewLeaf();
    };
  return *tree;
}; // OR_INDEX_TREE_NODE* OR_INDEX::IntegratePropLit(TERM lit)

OR_INDEX_TREE_NODE* OR_INDEX::IntegrateNonpropLit(const TERM* lit)
{
  CALL("IntegrateNonpropLit(const TERM* lit)");
  OR_INDEX_TREE_NODE** tree = Tree(*lit);
  l_iter.Reset(lit,1UL);
  l_iter.Next();
  return IntegrateArgs(tree);
}; // OR_INDEX_TREE_NODE* OR_INDEX::IntegrateNonpropLit(const TERM* lit)
 
OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree)
{
  CALL("IntegrateArgs(OR_INDEX_TREE_NODE** tree)");
  var_counter.Reset();
  if (*tree)
    {
      OR_INDEX_TREE_NODE* node = *tree;
      PrefSym sym;
    take_next_sym:
      if (l_iter.CurrentSymIsVar())
	{
	  ulong var_num = l_iter.Symbol().var();
	  sym.MkVar(Unifier::current()->variableBase(var_num));
	  var_counter.Register(var_num);
	}
      else
	{
	  sym.MkFunc(l_iter.Symbol());
	};
    check_node:
      if (node->Symbol() == sym)
	{
	  node = node->Below();
	  if (l_iter.Next())
	    {
	      goto take_next_sym;
	    }
	  else // node must be a leaf
	    {
	      ASSERT(node->IsLeaf());
	      return node;
	    };
	}
      else 
	{
	  if (node->Alternative())
	    {
	      node = node->Alternative();
	      goto check_node;
	    }
	  else // no alternative, new branch must be constructed
	    {
	      OR_INDEX_TREE_NODE* leaf;
	      node->SetAlternative(CreateBranch(leaf));
	      return leaf;
	    };
	};      
    }
  else // empty tree, create branch
    {
      OR_INDEX_TREE_NODE* leaf;
      *tree = CreateBranch(leaf);
      return leaf;
    };
}; // OR_INDEX_TREE_NODE* OR_INDEX::IntegrateArgs(OR_INDEX_TREE_NODE** tree)



OR_INDEX_TREE_NODE* 
OR_INDEX::integrateEqForResolution(const TERM* lit)
{
  CALL("integrateEqForResolution(const TERM* lit)");
  TERM header = *lit;
  ASSERT(header.IsEquality());
  header.MakeUnordered();
  OR_INDEX_TREE_NODE** tree = Tree(header);
  l_iter.Reset(lit,1UL);
  l_iter.Next();
  return IntegrateArgs(tree);
}; // OR_INDEX_TREE_NODE* OR_INDEX::integrateEqForResolution(const TERM* lit)


void OR_INDEX::Remove(Clause* cl)
{   
  LiteralList* allLiterals = cl->LitList();
  LiteralList::Element* firstNonselectedLit = allLiterals->nth(allLiterals->numOfSelectedLits());
  for(LiteralList::Iterator iter(allLiterals); 
      iter.currentElement() != firstNonselectedLit;
      iter.next())
    Remove(iter.currentElement(),cl); 
}; // void OR_INDEX::Remove(Clause* c)

void OR_INDEX::RemovePropLit(const TERM* lit,Clause* cl)
{
  CALL("RemovePropLit(const TERM* lit,Clause* cl)");
  ASSERT(!lit->arity());
  OR_INDEX_TREE_NODE** tree = Tree(*lit);
  ASSERT(*tree);
  ASSERT((*tree)->IsLeaf());
  (*tree)->Remove(cl);
  if (!(*tree)->Clauses()) 
    {
      (*tree)->DeleteLeaf();
      *tree = 0;
    };
}; // void OR_INDEX::RemovePropLit(const TERM* lit,Clause* cl)
 
OR_INDEX_TREE_NODE* OR_INDEX::LeafAndBranch(const TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch)
{
  CALL("LeafAndBranch(const TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch)");
  ASSERT(tree);
  ASSERT(t->arity());
  branch = tree;
  OR_INDEX_TREE_NODE* node = *tree;
  
  if (!node) return 0; 
  
  OR_INDEX_TREE_NODE** node_addr = tree; 
  
  TERM::Iterator iter;  

  iter.Reset(t);

 next_sym:
  if (!iter.Next())
    {
      ASSERT(node->IsLeaf());
      return node;
    };
 check_node:
  if (node->SameSymbol(iter.Symbol()))
    {
      if (node->Alternative()) branch = node_addr;
      node_addr = node->BelowAddr();
      node = node->Below();
      ASSERT(node);
      goto next_sym;     
    }; 
  // try alternative
  if (node->Alternative())
    {
      node_addr = node->AlternativeAddr();
      branch = node_addr;
      node = node->Alternative();
      goto check_node;
    };
  // no appropriate alternative, search fails
  return 0;
}; // OR_INDEX_TREE_NODE* OR_INDEX::LeafAndBranch(const TERM* t,OR_INDEX_TREE_NODE** tree,OR_INDEX_TREE_NODE**& branch)



void OR_INDEX::Remove(const TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)
{  
  CALL("Remove(const TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)");
  
  ASSERT(t->arity());
  OR_INDEX_TREE_NODE** branch;
  OR_INDEX_TREE_NODE* branch_to_delete;
  OR_INDEX_TREE_NODE* node_to_delete;
  
  OR_INDEX_TREE_NODE* leaf = LeafAndBranch(t,tree,branch);

  if (!leaf) return;

  leaf->Remove(cl);

  if (!(leaf->Clauses())) 
    {
      // no more clauses in the leaf, the whole branch must be deleted
      // *branch can't be a leaf here
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
	{
	  node_to_delete = branch_to_delete;
	  branch_to_delete = branch_to_delete->Below();
	  node_to_delete->DeleteNode();
	};
      branch_to_delete->DeleteLeaf(); // leaf
    };  
}; // void OR_INDEX::Remove(const TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)



void 
OR_INDEX::removeEqFromResolutionIndex(const TERM* eqLit,Clause* cl)
{
  CALL("removeEqFromResolutionIndex(const TERM* eqLit,Clause* cl)");
  TERM header = *eqLit;
  ASSERT(header.IsEquality());
  header.MakeUnordered();
  Remove(eqLit,cl,Tree(header));
}; // void OR_INDEX::removeEqFromResolutionIndex(const TERM* eqLit,Clause* cl)


void OR_INDEX::RemoveBPCandidates(const TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)
{  
  CALL("RemoveBPCandidates(const TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)");
  
  ASSERT(t->arity());
  OR_INDEX_TREE_NODE** branch;
  OR_INDEX_TREE_NODE* branch_to_delete;
  OR_INDEX_TREE_NODE* node_to_delete;
  
  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_PREPARE_FOR_REMOVAL(t->Args()->First());


  OR_INDEX_TREE_NODE* leaf = LeafAndBranch(t,tree,branch);

  if (!leaf) 
    {
      INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_REMOVAL_MODE('F');
      return;
    };

  leaf->RemoveBPCandidates(cl);

  if (!(leaf->BPCandidates())) 
    {
      // no more clauses in the leaf, the whole branch must be deleted
      // *branch can't be a leaf here
      INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_REMOVAL_MODE('R');
      branch_to_delete = *branch;
      *branch = (*branch)->Alternative(); 
      while (!(branch_to_delete->IsLeaf()))
	{
	  node_to_delete = branch_to_delete;
	  branch_to_delete = branch_to_delete->Below();
	  node_to_delete->DeleteNode();
	};
      branch_to_delete->DeleteLeaf(); // leaf
    }
  else
    {
      INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_REMOVAL_MODE('C');
    };
}; // void OR_INDEX::RemoveBPCandidates(const TERM* t,Clause* cl,OR_INDEX_TREE_NODE** tree)

void OR_INDEX::RemoveFromBPI(const TERM* lit,Clause* cl)
{
  CALL("RemoveFromBPI(const TERM* lit,Clause* cl)");
  _previousBackSupRedexes.reset();
  
  if (lit->IsEquality() && lit->IsOrderedEq())
    {
      if (cl->isSupportClause())
	{
	  if ((lit->Positive() && (!_orientPosEqualitiesInSupport)) ||
	      (lit->Negative() && (!_orientNegEqualitiesInSupport)))
	    goto remove_ordinary;
	};
      
      // only the first argument was integrated
      ASSERT(lit->Arg1()->IsReference());
      removeComplexSubtermFromBPI(lit->Arg1()->First(),cl);
    }
  else // ordinary literal or unordered equality
    {
    remove_ordinary:
      const bool* nonBasicArgs;
      if (_useInputNonBasicness &&
	  ((nonBasicArgs = lit->nonBasicArguments())))
	{
	  ASSERT(!lit->IsEquality());
	  ulong argNum = 0UL;
	  for (const TERM* arg = lit->Args(); !arg->IsNull(); arg = arg->Next())
	    {
	      if (!arg->isVariable() &&
		  !nonBasicArgs[argNum])
		removeComplexSubtermFromBPI(arg->First(),cl);
	      ++argNum;
	    };
	}
      else
	{
	  for (const TERM* arg = lit->Args(); !arg->IsNull(); arg = arg->Next())
	    {
	      if (!arg->isVariable())
		removeComplexSubtermFromBPI(arg->First(),cl);
	    };
	};
    };

}; // void OR_INDEX::RemoveFromBPI(const TERM* lit,Clause* cl)


void OR_INDEX::removeComplexSubtermFromBPI(const TERM* subterm,Clause* cl)
{
  CALL("removeComplexSubtermFromBPI(const TERM* subterm,Clause* cl)");
  ASSERT(subterm->isComplex());

  if (_backSupNoDuplicateRedexes)
    {
      for (const TERM** prev = _previousBackSupRedexes.begin();
	   prev != _previousBackSupRedexes.end();
	   ++prev)
	{
	  if (*prev == subterm) return;
	};
    };

  _previousBackSupRedexes.push(subterm);
  RemoveBPRedex(subterm,cl);

  const bool* nonBasicArgs;
  if (_useInputNonBasicness &&
      ((nonBasicArgs = subterm->nonBasicArguments())))
    {      
      ulong argNum = 0UL;
      
      for (const TERM* arg = subterm->Args();
	   !arg->IsNull();
	   arg = arg->Next())
	{

	  if (!arg->isVariable() &&
	      !nonBasicArgs[argNum])
	    removeComplexSubtermFromBPI(arg->First(),cl);
	  ++argNum;
	};
    }
  else
    {
      for (const TERM* arg = subterm->Args();
	   !arg->IsNull();
	   arg = arg->Next())
	{
	  if (!arg->isVariable())
	    removeComplexSubtermFromBPI(arg->First(),cl);
	};
    };
}; // void OR_INDEX::removeComplexSubtermFromBPI(const TERM* subterm,Clause* cl)

void OR_INDEX::RemoveBPRedex(const TERM* t,Clause* cl)
{
  CALL("RemoveBPRedex(const TERM* t,Clause* cl)");
  pseudo_literal->Args()->SetFirst(const_cast<TERM*>(t));
  RemoveBPCandidates(pseudo_literal,cl,&backward_para_tree);
}; // void OR_INDEX::RemoveBPRedex(const TERM* t,Clause* cl)

void OR_INDEX::RemoveFromFPI(const TERM* lit,Clause* cl)
{
  CALL("RemoveFromFPI(const TERM* lit,Clause* cl)");
  bool treatAsUnordered = 
    cl->isSupportClause() &&
    (!_orientPosEqualitiesInSupport);
  bool stop = false;
  const TERM* arg = lit->Args();
 remove_arg:
  if (arg->isVariable())
    {
      if (_paramodulateFromVariables)
	{
	  pseudo_literal[1] = *arg;
	  Remove(pseudo_literal,cl,&forward_para_from_var_tree);
	};
    }
  else // complex  
    if (arg->First()->arity())
      {
	Remove(arg->First(),cl,FParaTree(*(arg->First()))); 
      }
    else // constant 
      {
	OR_INDEX_TREE_NODE** tree = FParaTree(*(arg->First()));
       
	if (*tree)
	  {
	    (*tree)->Remove(cl);
	    if (!(*tree)->Clauses()) 
	      {
		(*tree)->DeleteLeaf();
		*tree = 0;
	      };
	  };
      };
  
  if (((!treatAsUnordered) && lit->IsOrderedEq()) || stop) 
    { 
      return;
    }; 
  stop = true;
  arg = arg->Next();
  goto remove_arg;          
}; // void OR_INDEX::RemoveFromFPI(const TERM* lit,Clause* cl)


void OR_INDEX::removeFromChainingIndex(const TERM* lit,Clause* cl)
{
  CALL("removeFromChainingIndex(const TERM* lit,Clause* cl)");

  ASSERT(_chainingTable && _chainingTable->transitivityLaw(lit->functor()));
  if (lit->Positive())
    {
      removePosFromChainingIndex(lit,cl);
    }
  else
    {
      removeNegFromChainingIndex(lit,cl);
    };
}; // void OR_INDEX::removeFromChainingIndex(const TERM* lit,Clause* cl)


void OR_INDEX::removePosFromChainingIndex(const TERM* lit,Clause* cl)
{
  CALL("removePosFromChainingIndex(const TERM* lit,Clause* cl)");
  const TERM* arg1ref = lit->Args();
  const TERM* arg2ref = arg1ref->Next();
  pseudo_literal[1] = *arg1ref;
  Remove(pseudo_literal,cl,&(_pos1ChainingTree.sub(lit->functor())));
  pseudo_literal[1] = *arg2ref;
  Remove(pseudo_literal,cl,&(_pos2ChainingTree.sub(lit->functor())));  
}; // void OR_INDEX::removePosFromChainingIndex(const TERM* lit,Clause* cl)


void OR_INDEX::removeNegFromChainingIndex(const TERM* lit,Clause* cl)
{
  CALL("removeNegFromChainingIndex(const TERM* lit,Clause* cl)");
  const TERM* arg1ref = lit->Args();
  const TERM* arg2ref = arg1ref->Next();
  pseudo_literal[1] = *arg1ref;
  Remove(pseudo_literal,cl,&(_neg1ChainingTree.sub(lit->functor())));
  pseudo_literal[1] = *arg2ref;
  Remove(pseudo_literal,cl,&(_neg2ChainingTree.sub(lit->functor())));  
}; // void OR_INDEX::removeNegFromChainingIndex(const TERM* lit,Clause* cl)


void OR_INDEX::Remove(const LiteralList::Element* lit,Clause* cl)
{
  CALL("Remove(const LiteralList::Element* lit,Clause* cl)");
  const TERM* litBody = lit->literal();
  if (!litBody->arity()) { RemovePropLit(litBody,cl); return; };
  if (paramodulation) RemoveFromBPI(litBody,cl);
  if (litBody->IsEquality())
    {     
      if (lit->properties().equalityMode() == LiteralProperties::DefaultEqualityMode)
	{
	  if (litBody->Positive()) RemoveFromFPI(litBody,cl);
	}
      else
	{
	  ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolution ||
		 lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
	  // this equality literal is available for prop. resolution
	  removeEqFromResolutionIndex(litBody,cl);
	  
	  if (lit->properties().equalityMode() == LiteralProperties::UseEqForResolution)
	    {
	      // this equality literal can be used for paramod
	      if (litBody->Positive()) RemoveFromFPI(litBody,cl);
	    }
	  else
	    {
	      ASSERT(lit->properties().equalityMode() == LiteralProperties::UseEqForResolutionOnly);
	      // nothing to do here
	    };
	};
    }
  else 
    {
      Remove(litBody,cl,Tree(*litBody));
    };
  
  if (_chainingTable && _chainingTable->transitivityLaw(litBody->functor()))
    {
      removeFromChainingIndex(litBody,cl);
    };
}; // void OR_INDEX::Remove(const LiteralList::Element* lit,Clause* cl)




OR_INDEX_TREE_NODE* OR_INDEX::CreateBranch(OR_INDEX_TREE_NODE*& leaf)
{
  PrefSym sym;
  OR_INDEX_TREE_NODE::TAG tag;
  if (l_iter.CurrentSymIsVar())
    {
      if (var_counter.Occurences(l_iter.Symbol().var()))
	{
	  tag = OR_INDEX_TREE_NODE::VAR;
	}
      else // first occurence of the variable
	{
	  var_counter.Register(l_iter.Symbol().var());
	  tag = OR_INDEX_TREE_NODE::VAR_FIRST_OCC;
	};
      sym.MkVar(Unifier::current()->variableBase(l_iter.Symbol().var()));
    }
  else
    {
      if (l_iter.Symbol().arity())
	{
	  tag = OR_INDEX_TREE_NODE::FUNC;
	}
      else
	{
	  tag = OR_INDEX_TREE_NODE::CONST;
	};
      sym.MkFunc(l_iter.Symbol());
    };
  OR_INDEX_TREE_NODE* result = OR_INDEX_TREE_NODE::NewNode(tag,sym,l_iter.Depth());
  OR_INDEX_TREE_NODE* node = result;
  while (l_iter.Next())
    {
      if (l_iter.CurrentSymIsVar())
	{
	  if (var_counter.Occurences(l_iter.Symbol().var()))
	    {
	      tag = OR_INDEX_TREE_NODE::VAR;
	    }
	  else // first occurence of the variable
	    {
	      var_counter.Register(l_iter.Symbol().var());
	      tag = OR_INDEX_TREE_NODE::VAR_FIRST_OCC;
	    };
	  sym.MkVar(Unifier::current()->variableBase(l_iter.Symbol().var()));
	}
      else
	{
	  if (l_iter.Symbol().arity())
	    {
	      tag = OR_INDEX_TREE_NODE::FUNC;
	    }
	  else
	    {
	      tag = OR_INDEX_TREE_NODE::CONST;
	    };
	  sym.MkFunc(l_iter.Symbol());
	};
      node->SetBelow(OR_INDEX_TREE_NODE::NewNode(tag,sym,l_iter.Depth()));
      node = node->Below();
    };
  
  leaf = OR_INDEX_TREE_NODE::NewLeaf();
  node->SetBelow(leaf);
  return result;
}; // OR_INDEX_TREE_NODE* OR_INDEX::CreateBranch(OR_INDEX_TREE_NODE*& leaf)



//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
#define DEBUG_NAMESPACE "OR_INDEX_TREE_NODE"
#endif 
#include "debugMacros.hpp"
//================================================= 

void OR_INDEX_TREE_NODE::Destroy(OR_INDEX_TREE_NODE* tree)
{
  CALL("Destroy(OR_INDEX_TREE_NODE* tree)");
  if (!tree) return;
  if (tree->IsLeaf())
    {
      tree->DeleteLeaf();
    }
  else
    {
      OR_INDEX_TREE_NODE* tree_below = tree->Below();
      OR_INDEX_TREE_NODE* tree_alt = tree->Alternative(); 
      tree->DeleteNode();
      Destroy(tree_below);
      if (tree_alt)
	{
	  Destroy(tree_alt);
	};   
    };
}; // void OR_INDEX_TREE_NODE::Destroy(OR_INDEX_TREE_NODE* tree)

void OR_INDEX_TREE_NODE::DestroyBPI(OR_INDEX_TREE_NODE* tree)
{
  CALL("DestroyBPI(OR_INDEX_TREE_NODE* tree)");
  if (!tree) return;
  if (tree->IsLeaf())
    {
      tree->DeleteBPILeaf();
    }
  else
    {
      OR_INDEX_TREE_NODE* tree_below = tree->Below();
      OR_INDEX_TREE_NODE* tree_alt = tree->Alternative(); 
      tree->DeleteNode();
      DestroyBPI(tree_below);
      if (tree_alt)
	{
	  DestroyBPI(tree_alt);
	};   
    };
}; // void OR_INDEX_TREE_NODE::DestroyBPI(OR_INDEX_TREE_NODE* tree)

#ifndef NO_DEBUG_VIS

 
ostream& OR_INDEX_TREE_NODE::output(ostream& str,unsigned var_offset) const
{
  str << '[' << (ulong)this << "] " << Tag() << ' ';
  switch (Tag())
    {
    case LEAF: str << ' ' << Term(); break;
    case FUNC: 
    case CONST:
    case VAR: 
    case VAR_FIRST_OCC:    
      Symbol().output(str,var_offset) << " [" << (ulong)Below() << "] ["
                                      << (ulong)Alternative() << "] "
                                      << Depth();
      break;
    };
  return str;
}; // ostream& OR_INDEX_TREE_NODE::output(ostream& str,unsigned var_offset) const
#endif


namespace std
{
ostream& operator<<(ostream& str,const OR_INDEX_TREE_NODE::TAG& tag)
{
  switch (tag)
    {
    case OR_INDEX_TREE_NODE::LEAF: return str << "LEAF";
    case OR_INDEX_TREE_NODE::FUNC: return str << "FUNC";
    case OR_INDEX_TREE_NODE::CONST: return str << "CONST"; 
    case OR_INDEX_TREE_NODE::VAR: return str << "VAR";
    case OR_INDEX_TREE_NODE::VAR_FIRST_OCC: return str << "VAR_FIRST_OCC"; 
    default: return str << "UNKNOWN TAG"; 
    };
}; // ostream& operator<<(ostream& str,const OR_INDEX_TREE_NODE::TAG& tag)
}; 

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
#define DEBUG_NAMESPACE "OR_PROCESSOR"
#endif 
#include "debugMacros.hpp"
//================================================= 


inline bool OR_PROCESSOR::CompleteExtraction()
{
  CALL("CompleteExtraction()");
 check_iter:
  
  if (curr_iter->Empty()) 
    {
      return true; 
    };
   
 check_sym:
   
   
  Unifier::Variable* var = curr_iter->CurrVar();
  if (var) // free variable
    {
     
      ASSERT(!(var->Binding()));
      bool same_var;
      
      if (BindWithOccCheck(var,same_var))
	{
       
	  ASSERT(!(same_var && var->Binding()));
	  if (same_var) { *curr_sit = VAR_SAME_VAR; }
	  else { *curr_sit = VAR_ANYTHING; };
	  curr_sit++;
	  curr_iter->SkipSym();
	  goto check_iter;
	}
      else
	{
	  if (BacktrackExtraction()) { goto check_sym; }
	  else { return false; };
	};
    }
  else // functor
    {
     
    check_node:
      switch (OR_INDEX_PATH::CurrNode()->Tag())
        {
	case OR_INDEX_TREE_NODE::FUNC : 
          // currently the same as for OR_INDEX_TREE_NODE::CONST,
          // but can be optimized by jumping to check_sym instead of check_iter
          
	case OR_INDEX_TREE_NODE::CONST : 
         
          if (OR_INDEX_PATH::CurrNode()->Symbol() == curr_iter->CurrSubterm()->Head())
	    { // same functor
            
	      curr_iter->SkipSym();
	      OR_INDEX_PATH::Below();
	      *curr_sit = FUNCTOR_FUNCTOR;
	      curr_sit++;
	      goto check_iter;
	    }
          else // different functors, try alternative
	    {
	      OR_INDEX_PATH::Alternative();
	      if (OR_INDEX_PATH::CurrNode())
		{
		  goto check_node;
		}
	      else // backtrack needed
		{
		  if (BacktrackExtraction()) { goto check_sym; }
		  else { return false; };
		};
	    };
         
	case OR_INDEX_TREE_NODE::VAR :
          {
	    bool free_var;
	    void* bind = OR_INDEX_PATH::Shift(OR_INDEX_PATH::CurrNode()->Symbol().var())->Unref(free_var);
	    
	    Prefix::CorrectBracket(curr_iter->CurrSubterm());
           
	    ASSERT(Prefix::CorrectComplexTerm(curr_iter->CurrSubterm()));
           
	    if (free_var)
	      {
		if (OccurCheck::Occurs(static_cast<Unifier::Variable*>(bind),curr_iter->CurrSubterm()))
		  {
		    // try alternative
		    OR_INDEX_PATH::Alternative();
		    if (OR_INDEX_PATH::CurrNode())
		      {
			goto check_node;
		      }
		    else // backtrack needed
		      {
			if (BacktrackExtraction()) { goto check_sym; }
			else { return false; };
		      };
		  }
		else // the variable doesn't occur in the subterm, make binding
		  {
		    Unifier::current()->BindAlone(static_cast<Unifier::Variable*>(bind),curr_iter->CurrSubterm());
		    curr_iter->SkipTerm();
		    OR_INDEX_PATH::Below();
		    *curr_sit = FUNCTOR_FREE_VAR;
		    curr_sit++;
		    goto check_iter;
		  };
	      }
	    else // variable bound by a complex term, try to unify
	      {
		if (UNIFICATION::UnifyComplex(static_cast<PrefixSym*>(bind),curr_iter->CurrSubterm()))
		  {
		    curr_iter->SkipTerm();
		    OR_INDEX_PATH::Below();
		    *curr_sit = FUNCTOR_BOUND_VAR;
		    curr_sit++;
		    goto check_iter;
		  }
		else // try alternative
		  {
		    OR_INDEX_PATH::Alternative();
		    if (OR_INDEX_PATH::CurrNode())
		      {
			goto check_node;
		      }
		    else // backtrack needed
		      {
			if (BacktrackExtraction()) { goto check_sym; }
			else { return false; };
		      };
		  };
	      };
          };
	case OR_INDEX_TREE_NODE::VAR_FIRST_OCC :
          Prefix::CorrectBracket(curr_iter->CurrSubterm());
          
          ASSERT(Prefix::CorrectComplexTerm(curr_iter->CurrSubterm()));
          
          Unifier::current()->BindAlone(OR_INDEX_PATH::Shift(OR_INDEX_PATH::CurrNode()->Symbol().var()),
				curr_iter->CurrSubterm());
          
          curr_iter->SkipTerm();
          OR_INDEX_PATH::Below();
          *curr_sit = FUNCTOR_FREE_VAR;
          curr_sit++;
          goto check_iter;         
         
#ifdef DEBUG_NAMESPACE
	default : ICP("A2"); return false;
#elif defined _SUPPRESS_WARNINGS_
	default : return false;
#endif 
        }; // switch
      ICP("A3");
      return false;
    };  
}; // bool OR_PROCESSOR::CompleteExtraction()

inline bool OR_PROCESSOR::BacktrackExtraction()
{
  CALL("BacktrackExtraction()");
 rollback:
  ASSERT((!(OR_INDEX_PATH::CurrNode())) || OR_INDEX_PATH::CurrNode()->IsLeaf());
  
  curr_sit--;
   
  switch (*curr_sit)
    {
    case START_OF_EXTRACTION : 
      ForgetIter(); 
      return false;
    case VAR_SAME_VAR :
      {
	bool same_var;
	curr_iter->Backtrack();

	ASSERT(curr_iter->CurrVar());
	ASSERT(!(curr_iter->CurrVar()->Binding()));
       
	if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
	  {
	    if (!(same_var)) { *curr_sit = VAR_ANYTHING; };
	    curr_sit++;
	    curr_iter->SkipSym();
	    return true; 
	  }
	else 
	  {
	    goto rollback; 
	  };
      };
    case VAR_ANYTHING : 
      {
	bool same_var;
	curr_iter->Backtrack();
       
	if (AlterBindingWithOccCheck(curr_iter->CurrVar(),same_var)) 
	  {
	    if (same_var) 
	      {
		*curr_sit = VAR_SAME_VAR; 
	      };
	    curr_sit++; 
	    curr_iter->SkipSym();
	    return true; 
	  }
	else 
	  {
	    goto rollback; 
	  };
      };

    case FUNCTOR_FUNCTOR : 
      curr_iter->Backtrack();
      
      
      ASSERT(curr_iter->CurrSubterm()->Head().IsFunctor());

      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) { return true; }
      else 
	{
	  goto rollback; 
	};

    case FUNCTOR_FREE_VAR :
      Unifier::current()->UnbindLast();
      curr_iter->Backtrack();

      ASSERT(curr_iter->CurrSubterm()->Head().IsFunctor());
      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) { return true; }
      else 
	{
	  goto rollback; 
	};

    case FUNCTOR_BOUND_VAR :
      
      Unifier::current()->UnbindLastSection();
      curr_iter->Backtrack();
      ASSERT(curr_iter->CurrSubterm()->Head().IsFunctor());
      OR_INDEX_PATH::RestoreCurrent();
      OR_INDEX_PATH::Alternative();
      if (OR_INDEX_PATH::CurrNode()) 
	{
	  return true; 
	}
      else 
	{
	  goto rollback; 
	};
#ifdef DEBUG_NAMESPACE
    default : ICP("ICP1"); return false;
#endif
    };  
  ICP("ICP2"); 
  return false; 
}; // bool OR_PROCESSOR::BacktrackExtraction()

bool OR_PROCESSOR::ExtractUnifiable(PrefixSym* word) 
  // word must represent a complex term
{
  NextIter();
  curr_iter->Reset(word);
  *curr_sit = START_OF_EXTRACTION;
  curr_sit++;
  return CompleteExtraction();
}; // bool OR_PROCESSOR::ExtractUnifiable(PrefixSym* word)
 
bool OR_PROCESSOR::AlterUnifiable()
{
  if (BacktrackExtraction()) 
    {
      return CompleteExtraction(); 
    }
  else 
    {
      return false; 
    };
}; // bool OR_PROCESSOR::AlterUnifiable()


#ifndef NO_DEBUG_VIS
namespace std
{
ostream& operator<<(ostream& str,const OR_PROCESSOR::SITUATION& sit) 
{
  switch (sit)
    {
    case OR_PROCESSOR::START_OF_EXTRACTION : str << "START_OF_EXTRACTION"; break;
    case OR_PROCESSOR::VAR_SAME_VAR : str << "VAR_SAME_VAR"; break;
    case OR_PROCESSOR::VAR_ANYTHING : str << "VAR_ANYTHING"; break;
    case OR_PROCESSOR::FUNCTOR_FUNCTOR : str << "FUNCTOR_FUNCTOR"; break;
    case OR_PROCESSOR::FUNCTOR_FREE_VAR : str << "FUNCTOR_FREE_VAR"; break;
    case OR_PROCESSOR::FUNCTOR_BOUND_VAR : str << "FUNCTOR_BOUND_VAR"; break;
    };
  return str;
}; //ostream& operator<<(ostream& str,const OR_PROCESSOR::SITUATION& sit) 
};
#endif


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_INDEX
#define DEBUG_NAMESPACE "OR_INDEX_PATH"
#endif 
#include "debugMacros.hpp"
//================================================= 

#ifndef NO_DEBUG_VIS
ostream& OR_INDEX_PATH::output(ostream& str) const
{
  for (const OR_INDEX_TREE_NODE* const * p = path; p<next; p++)    
    (*p)->output(str,var_offset) << '\n'; 
  str << "CURRENT NODE = ";
  if (curr_node) { curr_node->output(str,var_offset); }
  else str << "[0]";  
  return str << '\n'; 
}; // ostream& OR_INDEX_PATH::output(ostream& str) const
#endif




//=================================================
