#include "or_premise.hpp"

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_PREMISE
 #define DEBUG_NAMESPACE "OR_PREMISE"
#endif
#include "debugMacros.hpp"
//=================================================

using namespace VK;


OR_PREMISE::OR_PREMISE() : 
  _termMemory(DOP("OR_PREMISE::_termMemory")),
  _selectedLiteral(DOP("OR_PREMISE::_selectedLiteral")),
  _selectedLiteralRotation(DOP("OR_PREMISE::_selectedLiteralRotation")),
  _selectedLiteralProperties(DOP("OR_PREMISE::_selectedLiteralProperties")),
  _litIsMarked(DOP("OR_PREMISE::_litIsMarked"))
{
  CALL("constructor OR_PREMISE()"); 
  DOP(_termMemory.freeze());
  DOP(_selectedLiteral.freeze());
  DOP(_selectedLiteralRotation.freeze());
  DOP(_selectedLiteralProperties.freeze());
  DOP(_litIsMarked.freeze());
}; // OR_PREMISE::OR_PREMISE() 


OR_PREMISE::~OR_PREMISE() { CALL("destructor ~OR_PREMISE()"); };

bool OR_PREMISE::load(ulong ind,Clause* c)
{
  CALL("load(ulong ind,Clause* c)");
  ASSERT(!c->isAnswer());
  DOP(_termMemory.unfreeze());
  if (!_termMemory.tryToExpand(c->size() * 2)) return false;
  DOP(_termMemory.freeze());

  _freeTermMemory = _termMemory.memory();
  //free_memory = memory;

  DOP(_selectedLiteral.unfreeze());
  if (!_selectedLiteral.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_selectedLiteral.freeze()); 

  DOP(_selectedLiteralRotation.unfreeze());
  if (!_selectedLiteralRotation.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_selectedLiteralRotation.freeze()); 


  DOP(_selectedLiteralProperties.unfreeze());
  if (!_selectedLiteralProperties.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_selectedLiteralProperties.freeze());

  DOP(_litIsMarked.unfreeze());
  if (!_litIsMarked.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_litIsMarked.freeze());

  clause = c;
  index = ind;

  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  _lastSelected = _allLiterals->numOfSelectedLits() - 1;
  ulong max_lit = 0; 
  for(LiteralList::Iterator iter(_allLiterals);
      iter.currentElement() != _firstNonselectedLit;
      iter.next())
    {
      _selectedLiteralProperties[max_lit] =
	iter.currentElement()->properties();
      _selectedLiteral[max_lit] = _freeTermMemory;
      _litIsMarked[max_lit] = false;
      if (!iter.currentLiteral()->IsEquality())
	{
	  Prefix::StandardTermToPrefix(ind,iter.currentLiteral(),_freeTermMemory);
	  _selectedLiteralRotation[max_lit] = 0;
	}
      else
	{
	  TERM header = *(iter.currentLiteral());
	  header.MakeUnordered();
	  Prefix::binaryStandardTermToPrefix(ind,
					     header,
					     *(iter.currentLiteral()->Arg1()),
					     *(iter.currentLiteral()->Arg2()),
					     _freeTermMemory);
	  if (iter.currentLiteral()->IsUnorderedEq() &&
	      iter.currentElement()->properties().useEqForResolution())
	    {
	      _selectedLiteralRotation[max_lit] = _freeTermMemory;
	      Prefix::binaryStandardTermToPrefix(ind,
						 header,
						 *(iter.currentLiteral()->Arg2()),
						 *(iter.currentLiteral()->Arg1()),
						 _freeTermMemory);
	    }
	  else
	    _selectedLiteralRotation[max_lit] = 0;
	};
      max_lit++;
    };   
  return true;
}; // bool OR_PREMISE::load(ulong ind,Clause* c)

bool OR_PREMISE::loadOptimized(ulong ind,Clause* c)
{
  CALL("loadOptimized(ulong ind,Clause* c)");
  // if there is only one maximal literal we do not transform it
  // into prefix representation
  
  ASSERT(!c->isAnswer());
  DOP(_termMemory.unfreeze());
  if (!_termMemory.tryToExpand(c->size() * 2)) return false;
  DOP(_termMemory.freeze());
   
  _freeTermMemory = _termMemory.memory();
  //free_memory = memory;

  DOP(_selectedLiteral.unfreeze());
  if (!_selectedLiteral.tryToExpand(c->numOfAllLiterals())) return false;
  DOP(_selectedLiteral.freeze()); 


  DOP(_selectedLiteralRotation.unfreeze());
  if (!_selectedLiteralRotation.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_selectedLiteralRotation.freeze()); 


  DOP(_selectedLiteralProperties.unfreeze());
  if (!_selectedLiteralProperties.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_selectedLiteralProperties.freeze());

  DOP(_litIsMarked.unfreeze());
  if (!_litIsMarked.tryToExpand(c->numOfAllLiterals()))
    return false;
  DOP(_litIsMarked.freeze());

  clause = c;
  index = ind;

  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  _lastSelected = _allLiterals->numOfSelectedLits() - 1;

  if (_lastSelected)
    {
      ulong max_lit = 0; 
      for(LiteralList::Iterator iter(_allLiterals);
	  iter.currentElement() != _firstNonselectedLit;
	  iter.next())
	{
	  _selectedLiteralProperties[max_lit] =
	    iter.currentElement()->properties();
	  _selectedLiteral[max_lit] = _freeTermMemory;
	  _litIsMarked[max_lit] = false;

	  
	  if (!iter.currentLiteral()->IsEquality())
	    {
	      Prefix::StandardTermToPrefix(ind,iter.currentLiteral(),_freeTermMemory);
	      _selectedLiteralRotation[max_lit] = 0;
	    }
	  else
	    {
	      TERM header = *(iter.currentLiteral());
	      header.MakeUnordered();
	      Prefix::binaryStandardTermToPrefix(ind,
						 header,
						 *(iter.currentLiteral()->Arg1()),
						 *(iter.currentLiteral()->Arg2()),
						 _freeTermMemory);
	      if (iter.currentLiteral()->IsUnorderedEq() &&
		  iter.currentElement()->properties().useEqForResolution())
		{
		  _selectedLiteralRotation[max_lit] = _freeTermMemory;
		  Prefix::binaryStandardTermToPrefix(ind,
						     header,
						     *(iter.currentLiteral()->Arg2()),
						     *(iter.currentLiteral()->Arg1()),
						     _freeTermMemory);
		}
	      else
		_selectedLiteralRotation[max_lit] = 0;
	    };


	  max_lit++;
	};   
    }
  else
    {
      _selectedLiteralProperties[0] =
	_allLiterals->firstElem()->properties();
      _selectedLiteral[0] = 0; // indicates that optimised loading was used
      _selectedLiteralRotation[0] = 0;
      _litIsMarked[0] = false; 
    };
  return true;
}; // bool OR_PREMISE::loadOptimized(ulong ind,Clause* c) 


#ifndef NO_DEBUG_VIS  
ostream& OR_PREMISE::output(ostream& str) const
{
  str <<  "<<<<<< PREMISE >>>>>>>\n"
      << " Clause : " << clause << "\n Index = " << index
      << "\n Maximal literals: \n";
  for (ulong ml = 0; ml <= _lastSelected; ml++)
    {
      str << "  [" << _litIsMarked[ml] << "] ";
      Prefix::outputTerm(str,_selectedLiteral[ml]);
      if (_selectedLiteralRotation[ml])
	{
	  str << " {rotation: ";
	  Prefix::outputTerm(str,_selectedLiteralRotation[ml]);
	  str << "}";
	};
      str << '\n';
    };
  return str; 
}; // ostream& OR_PREMISE::output(ostream& str) const
#endif

//=================================================
