#ifndef OR_PREMISE_H
//=================================================================
#define OR_PREMISE_H
#ifndef NO_DEBUG_VIS 
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Clause.inl"
#include "prefix.hpp" 
#include "or_index.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
#include "TemporaryClauseCollector.hpp"
#include "LiteralProperties.hpp"
#include "TemporaryLiteral.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_PREMISE
 #define DEBUG_NAMESPACE "OR_PREMISE"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{

class OR_PREMISE
 {
  public: // methods
   OR_PREMISE(); 
   ~OR_PREMISE();



   PrefixSym* nthSelectedLit(ulong n) const
     {
       CALL("nthSelectedLit(ulong n) const");
       ASSERT(n <= _lastSelected);
       return _selectedLiteral[n]; 
     };
   
   PrefixSym* nthSelectedLitRotation(ulong n) const
     {
       CALL("nthSelectedLitRotation(ulong n) const");
       ASSERT(n <= _lastSelected);
       return _selectedLiteralRotation[n]; 
     };
   
   bool nthSelectedLitEligibleForResolution(ulong n) const
   {
     CALL("nthSelectedLitEligibleForResolution(ulong n) const");
     ASSERT(n <= _lastSelected);
     return 
       !_selectedLiteral[n]->Head().Func().IsEquality() ||
       _selectedLiteralProperties[n].useEqForResolution();
   };

   Clause* TheClause() const { return clause; };

   ulong lastSelected() { return _lastSelected; };
   ulong Index() { return index; };
   ulong CutLit() { return cut_lit; };
   const LiteralProperties& resolvedLiteralProperties() const 
   {
     return _resolvedLiteralProperties;
   };

   bool load(ulong ind,Clause* c);
   bool loadOptimized(ulong ind,Clause* c);
   // ^ if there is only one maximal literal we do not transform it
   //   into prefix representation
  
   void Unload() 
   {
   };

   void Cut(ulong lit) 
    {
     CALL("Cut(ulong lit)");
     cut_lit = lit;
     _litIsMarked[cut_lit] = true;
     _resolvedLiteralProperties = 
       _selectedLiteralProperties[cut_lit];
    };
   bool load(ulong ind,RES_PAIR_LIST* res_pair)
    {
     CALL("load(ulong ind,RES_PAIR_LIST* res_pair)");
     if (!load(ind,res_pair->TheClause())) return false;
     Cut((ulong)(res_pair->LitNum()));
     return true;
    };
   bool loadOptimized(ulong ind,RES_PAIR_LIST* res_pair)
    {
     CALL("loadOptimized(ulong ind,RES_PAIR_LIST* res_pair)");
     if (!loadOptimized(ind,res_pair->TheClause())) return false;
     Cut((ulong)(res_pair->LitNum()));
     return true;
    };
   void Cut(TERM* lit) { Cut(clause->NumberOfLiteral(lit)); };
   void UnmarkCut() 
   { 
    CALL("UnmarkCut()");
    _litIsMarked[cut_lit] = false; 
   };
   bool collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,
				LiteralProperties::SLModifier& slmod);
   bool collectAllLiterals(TemporaryClauseCollector& clauseCollector); 
   bool collectSubtermsAvailableForParamodulationInto(TemporaryClauseCollector& clauseCollector,
						      bool collectMarked);
   bool NonmarkedExist() { return nonmarkedExist; };
   
#ifndef NO_DEBUG_VIS  
   ostream& output(ostream& str) const;
#endif
   
  protected: // structure
   Clause* clause;
   ulong index;
   
   BK::Array<BK::GlobAlloc,PrefixSym,BK::FastRelocation<PrefixSym>,256UL> _termMemory;
   PrefixSym* _freeTermMemory;
   
   BK::Array<BK::GlobAlloc,PrefixSym*,BK::FastRelocation<PrefixSym*>,32UL> _selectedLiteral;
   BK::Array<BK::GlobAlloc,PrefixSym*,BK::FastRelocation<PrefixSym*>,32UL> _selectedLiteralRotation;

   BK::Array<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32UL> _selectedLiteralProperties;
   
   LiteralProperties _resolvedLiteralProperties;

   ulong _lastSelected;

   BK::Array<BK::GlobAlloc,bool,BK::FastRelocation<bool>,32UL> _litIsMarked;
   

   LiteralList* _allLiterals;
   LiteralList::Element* _firstNonselectedLit;

   ulong cut_lit;
   bool nonmarkedExist;

 }; // class OR_PREMISE
 
//******************* Definitions: *************************************


inline bool OR_PREMISE::collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,
						LiteralProperties::SLModifier& slmod)
 {
  CALL("collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,LiteralProperties::SLModifier& slmod)");
  ulong currMax = 0;
  LiteralProperties prop;
  nonmarkedExist = (_firstNonselectedLit != 0);   

  while (currMax <= _lastSelected)
   {   
    if (!(_litIsMarked[currMax])) 
     {
      nonmarkedExist = true;
      prop.inheritFrom(_selectedLiteralProperties[currMax],slmod);
      if (!clauseCollector.pushLiteralWithUnifier(_selectedLiteral[currMax],prop))
        return false; 
     };
    currMax++;
   };
  
   
  for(LiteralList::Iterator iter(_allLiterals,_firstNonselectedLit);
      iter.notEnd();
      iter.next())     
    {
      prop.inheritFrom(iter.currentElement()->properties(),slmod);
      if (!clauseCollector.pushLiteralWithUnifier(index,
						  iter.currentLiteral(),
						  prop)) 
	return false;
    };
  return true;
 }; // inline bool OR_PREMISE::collectUnmarkedLiterals(TemporaryClauseCollector& clauseCollector,LiteralProperties::SLModifier& slmod)



inline bool OR_PREMISE::collectAllLiterals(TemporaryClauseCollector& clauseCollector)
 {
  CALL("collectAllLiterals(TemporaryClauseCollector& clauseCollector)");

  // works even if loadOptimised was used

  for(LiteralList::Iterator iter(_allLiterals); iter.notEnd(); iter.next())     
    if (!clauseCollector.pushLiteralWithUnifier(index,
						iter.currentLiteral(),
						iter.currentElement()->properties())) 
     return false;
  
  return true;
 }; // inline bool OR_PREMISE::collectAllLiterals(TemporaryClauseCollector& clauseCollector)



inline 
bool OR_PREMISE::collectSubtermsAvailableForParamodulationInto(TemporaryClauseCollector& clauseCollector,
							       bool collectMarked)
{
  CALL("collectSubtermsAvailableForParamodulationInto(TemporaryClauseCollector& clauseCollector&...)");
  LiteralProperties prop;
  bool atLeastOneLiteralCollected = false;
  ulong litNum = 0UL;
  bool markedCollected = !collectMarked;
  for(LiteralList::Iterator iter(_allLiterals);
      iter.currentElement() != _firstNonselectedLit;
      iter.next())
    {
      if ((!_litIsMarked[litNum]) || (!markedCollected))
	{
	  markedCollected = _litIsMarked[litNum];
	  TERM* lit = iter.currentLiteral();
	  if (lit->IsEquality())
	    {
	      if (lit->IsOrderedEq())
		{
		  // only the first argument
		  if (!clauseCollector.pushPseudoLiteralWithUnifier(index,
								     TERM::dummyPositiveHeader1(),
								     *(lit->Args())))
		    return false;
		}
	      else
		if (!clauseCollector.pushLiteralWithUnifier(index,
							    lit,
							    prop))
		  return false;
	      atLeastOneLiteralCollected = true;
	    }
	  else
	    if (lit->arity())
	      {
		if (!clauseCollector.pushLiteralWithUnifier(index,
							    lit,
							    prop))
		  return false;
		atLeastOneLiteralCollected = true;
	      };
	};
      ++litNum;
    };
  return atLeastOneLiteralCollected;
}; // bool OR_PREMISE::collectSubtermsAvailableForParamodulationInto(emporaryClauseCollector& clauseCollector...)



}; // namespace VK


//=================================================================
#endif
