//
// File:         premise_factoring.cpp
// Description:  OR premises with factoring.
// Created:      Feb 15, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 10, 2001. 
//               The arrays OR_PREMISE::{max_literal,mark} replaced by 
//               extending arrays OR_PREMISE::{_maxLiteral,_litIsMarked}
// Revised:      Dec 11, 2001. 
//               The stack OR_PREMISE_WITH_FACTORING::factored replaced by 
//               extending stack OR_PREMISE_WITH_FACTORING::_factoredLits
//====================================================
#include "premise_factoring.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_FACTORING
 #define DEBUG_NAMESPACE "OR_PREMISE_WITH_FACTORING"
#endif
#include "debugMacros.hpp"
//=================================================

using namespace VK;



OR_PREMISE_WITH_FACTORING::OR_PREMISE_WITH_FACTORING() :
  _factoredLits(DOP("_factoredLits")),
  _rotationRequired(DOP("_rotationRequired"))
{
  CALL("constructor OR_PREMISE_WITH_FACTORING()"); 
  DOP(_factoredLits.freeze());
  DOP(_rotationRequired.freeze());
}; // OR_PREMISE_WITH_FACTORING::OR_PREMISE_WITH_FACTORING() 

OR_PREMISE_WITH_FACTORING::~OR_PREMISE_WITH_FACTORING() 
{ 
  CALL("destructor ~OR_PREMISE_WITH_FACTORING()"); 
};



bool OR_PREMISE_WITH_FACTORING::GetFirstFactor(ulong first_candidate)
{
  CALL("GetFirstFactor(ulong first_candidate)");
  if (FactoringPossible())
    {
      bool rotationRequired;
      _factoredLits.reset();
      _rotationRequired.reset();
      if (FindFirstFactor(first_candidate,rotationRequired))
	{
	  _lastFactoredLitNum = first_candidate;
	  _rotationRequiredForLastFactored = rotationRequired;
	  was_factored = true;
	  factoriseLiteralProperties();
	  ASSERT(_rotationRequired.size() == _factoredLits.size());
	  return true;
	};
    };
  return false;
}; // bool OR_PREMISE_WITH_FACTORING::GetFirstFactor(ulong first_candidate)
 
bool OR_PREMISE_WITH_FACTORING::GetNextFactor()
{
  CALL("GetNextFactor()");

  ulong candidate = _lastFactoredLitNum + 1;
  bool rotationRequired;
 next_factor:
  if (FindFirstFactor(candidate,rotationRequired))
    {
      _factoredLits.push(_lastFactoredLitNum);
      _rotationRequired.push(_rotationRequiredForLastFactored);
      _lastFactoredLitNum = candidate;
      _rotationRequiredForLastFactored = rotationRequired;
      factoriseLiteralProperties();
      ASSERT(_rotationRequired.size() == _factoredLits.size());
      return true;
    }
  else // backtrack needed
    {
      Unifier::current()->UnbindLastSection();

      if (_rotationRequiredForLastFactored)
	{
	  _rotationRequiredForLastFactored = false;
	  if (_mainLiteralRotation)
	    {
	      if (UNIFICATION::UnifyComplex(OR_PREMISE::_selectedLiteral[_lastFactoredLitNum],
					    _mainLiteralRotation))
		{
		  ASSERT(_rotationRequired.size() == _factoredLits.size());
		  return true;
		};
	    }
	  else
	    {
	      ASSERT(OR_PREMISE::_selectedLiteralRotation[_lastFactoredLitNum]);
	      if (UNIFICATION::UnifyComplex(OR_PREMISE::_selectedLiteralRotation[_lastFactoredLitNum],
					    _mainLiteral))
		{
		  ASSERT(_rotationRequired.size() == _factoredLits.size());
		  return true;
		};
	    };
	};
       // cannot be rotated or rotation does not help
	
      OR_PREMISE::_litIsMarked[_lastFactoredLitNum] = false;
      candidate = _lastFactoredLitNum + 1;
      if (_factoredLits.empty())
	{
	  ASSERT(_rotationRequired.empty());
	  if (FindFirstFactor(candidate,rotationRequired))
	    {
	      _lastFactoredLitNum = candidate;
	      _rotationRequiredForLastFactored = rotationRequired;
	      factoriseLiteralProperties();
	      return true;
	    }
	  else // no more factors at all
	    return false;
	}
      else 
	{
	  _lastFactoredLitNum = _factoredLits.pop();
	  _rotationRequiredForLastFactored = _rotationRequired.pop();
	  goto next_factor;        
	};
      
    };
}; // bool OR_PREMISE_WITH_FACTORING::GetNextFactor()
 
bool OR_PREMISE_WITH_FACTORING::FindFirstFactor(ulong& candidate,bool& rotationRequired)
{
  CALL("FindFirstFactor(ulong& candidate,bool& rotationRequired)");
  while (candidate <= _lastSelected)
    {
      if (candidate != cut_lit)
	{
	  if (UNIFICATION::UnifyComplex(OR_PREMISE::_selectedLiteral[candidate],
					_mainLiteral))  
	    {
	      // factor found without rotation
	      OR_PREMISE::_litIsMarked[candidate] = true;
	      rotationRequired = 
		_mainLiteralRotation ||
		OR_PREMISE::_selectedLiteralRotation[candidate];
	      return true;
	    };
	  // try to factorise modulo symmetry
	  if (_mainLiteralRotation)
	    {
	      if (UNIFICATION::UnifyComplex(OR_PREMISE::_selectedLiteral[candidate],
					    _mainLiteralRotation))
		{		  
		  OR_PREMISE::_litIsMarked[candidate] = true;
		  rotationRequired = false;
		  return true;
		};
	    }
	  else if (OR_PREMISE::_selectedLiteralRotation[candidate])
	    {
	      if (UNIFICATION::UnifyComplex(OR_PREMISE::_selectedLiteralRotation[candidate],
					    _mainLiteral))
		{		  
		  OR_PREMISE::_litIsMarked[candidate] = true;
		  rotationRequired = false;
		  return true;
		};
	    };

	  // this literal can not be factored

	};
      // continue search
      ++candidate;
    };
  // no factor found 
  return false;
}; // bool OR_PREMISE_WITH_FACTORING::FindFirstFactor(ulong& candidate,bool& rotationRequired)


inline
void OR_PREMISE_WITH_FACTORING::factoriseLiteralProperties()
{
  CALL("factoriseLiteralProperties()");
  _resolvedLiteralProperties = _selectedLiteralProperties[_lastFactoredLitNum];
  for (ulong* lit = _factoredLits.begin();
       lit != _factoredLits.end();
       ++lit)
    _resolvedLiteralProperties.factoriseWith(_selectedLiteralProperties[*lit]);
}; // void OR_PREMISE_WITH_FACTORING::factoriseLiteralProperties()


//======================================================================
