
/*
 * File Var.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __Var__
#define __Var__

namespace XML {
  class Element;
};

#include <string>
#include "Lst.hpp"



class Var
{
public:
  
  Var () {}
  
  explicit Var (int v) { ASS(v>=0); _var = v; }
  
  Var (const Var& v) { _var = v._var; }
  
  int toInt () const { return _var; }
  
  
  
  static bool lessThan (const Var& v1, const Var& v2) 
    { return v1._var < v2._var; }
  
  bool operator == (const Var& v) const { return _var == v._var; }
  
  bool operator != (const Var& v) const { return _var != v._var; }
  XML::Element* toXML () const;
  string toString () const;
  
  void set (int v) { ASS(v>=0);_var = v; }
private:
  
  int _var;
};


class VarList :
  public Lst<Var>
{
 public:
  
  const VarList& tail () const
    { return static_cast<const VarList&>(Lst<Var>::tail()); }
  VarList (Var head, const VarList& tail);
  VarList (const VarList& lst);
  explicit VarList (Var v); 
  VarList ();

  
  bool equals (const VarList&) const;
  string toString () const;
};


class VarListList :
  public Lst<VarList>
{
 public:
  
  const VarListList& tail () const
    { return static_cast<const VarListList&>(Lst<VarList>::tail()); }
  VarListList (VarList head, const VarListList& tail);
  VarListList (const VarListList& lst);
  explicit VarListList (const VarList& t); 
  VarListList ();
};




inline
VarList::VarList () 
{
}


inline
VarList::VarList (const VarList& lst) 
  :
  Lst<Var> (lst) 
{
}


inline
VarList::VarList (Var v) 
  :
  Lst<Var> (v) 
{
}


inline
VarList::VarList (Var head, const VarList& tail) 
  :
  Lst<Var> (head,tail) 
{
}





inline
VarListList::VarListList () 
{
}


inline
VarListList::VarListList (const VarListList& lst) 
  :
  Lst<VarList> (lst) 
{
}


inline
VarListList::VarListList (VarList head, const VarListList& tail) 
  :
  Lst<VarList> (head,tail) 
{
}


#endif 
