<?xml version="1.0"?>

<!-- ===================================================================

     XFF to TSTP convertor stylesheet
     TODO
     
AUTHORS:

  Petr Pudlak <pudlak@kti.mff.cuni.cz>
  
LICENSE:

    Copyright (C) 2004  Petr Pudlak

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

==================================================================== -->

<xsl:stylesheet
    xmlns="http://www.w3.org/1999/XSL/Transform"
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xff="http://xff.sf.net/xff"
    xmlns:tstp="http://xff.sf.net/xff/ext/tstp">
<!--    xpath-default-namespace="http://www.tptp.org/xtstp" -->

  <xsl:output
    method="text"
    omit-xml-declaration="yes"
    />

  <!-- xsl:import href="copyover.xsl"/ -->

  <!-- commont template definitions -->
  <xsl:template name="replace-string">
    <xsl:param name="text"/>
    <xsl:param name="replace"/>
    <xsl:param name="with"/>
    <xsl:choose>
      <xsl:when test="contains($text,$replace)">
        <xsl:value-of select="substring-before($text,$replace)"/>
        <xsl:value-of select="$with"/>
        <xsl:call-template name="replace-string">
	  <xsl:with-param name="text" select="substring-after($text,$replace)"/>
          <xsl:with-param name="replace" select="$replace"/>
          <xsl:with-param name="with" select="$with"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$text"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>
 

  <!-- list processer -->
  <xsl:template name="clist">
    <xsl:param name="separator"/>
    <xsl:param name="sublist"/>
    <xsl:for-each select="$sublist">
      <xsl:value-of select="$separator" />
      <xsl:apply-templates select="." />
    </xsl:for-each>
  </xsl:template>
  <xsl:template name="list">
    <xsl:param name="separator"/>
    <xsl:param name="sublist"/>

    <xsl:apply-templates select="$sublist[1]" />
    <xsl:call-template name="clist">
      <xsl:with-param name="separator" select="$separator"/>
      <xsl:with-param name="sublist" select="$sublist[position()>1]"/>
    </xsl:call-template>
  </xsl:template>

  <!-- quoted string -->
  <xsl:template name="qstring">
    <xsl:param name="value"/>
    <xsl:text>'</xsl:text>
    <xsl:value-of select="$value"/>
    <xsl:text>'</xsl:text>
  </xsl:template>

  <xsl:template match="xtstp:tstp">
    <xsl:apply-templates/>
  </xsl:template>

  <xsl:template match="xtstp:include">
    <xsl:text>include(</xsl:text>
      <xsl:call-template name="qstring">
        <xsl:with-param name="value" select="@filename"/>
      </xsl:call-template>
      <xsl:call-template name="clist">
        <xsl:with-param name="separator" select="','"/>
        <xsl:with-param name="sublist" select="xtstp:formula-ref"/>
      </xsl:call-template>
    <xsl:text>).&#xa;</xsl:text>
  </xsl:template>
  
  <xsl:template match="xtstp:formula-ref">
    <xsl:value-of select="@name"/>
  </xsl:template>

  <xsl:template match="xtstp:formula">
    <xsl:value-of select="@language"/>
    <xsl:text>(</xsl:text>
      <xsl:value-of select="@name" />
      <xsl:text>,</xsl:text>
      <xsl:value-of select="@type" />
      <xsl:text>,(</xsl:text>
      <xsl:apply-templates/>
      <!-- TODO record extension -->
    <xsl:text>)).&#xa;&#xa;</xsl:text>
  </xsl:template>

  <xsl:template match="xtstp:quantify">
    <xsl:choose>
      <xsl:when test="@type='forall'"> <xsl:text> ! </xsl:text></xsl:when>
      <xsl:when test="@type='exists'"> <xsl:text> ? </xsl:text></xsl:when>
      <xsl:otherwise>
        <xsl:message terminate="yes">
          Unknown quantifier '<xsl:value-of select="@type" />'
        </xsl:message>
      </xsl:otherwise>
    </xsl:choose>
    <xsl:text>[</xsl:text>
      <xsl:call-template name="list">
        <xsl:with-param name="separator" select="','"/>
        <xsl:with-param name="sublist" select="xtstp:qvar/@name"/>
      </xsl:call-template>
    <xsl:text>] : </xsl:text>
    <xsl:apply-templates select="*[node() != 'xtstp:qvar']"/>
  </xsl:template>

  <!-- processess connective subtree
       $symbol  paramter specifies the connective symbol (it is converted
                to a TSTP symbol
       $sublist gives a list of subtrees to process; for each subtree, TSTP
                symbol glyph is emited and then the subtree is processed
		recursively
  -->
  <xsl:template name="connective">
    <xsl:param name="symbol"/>
    <xsl:param name="sublist"/>
    <xsl:for-each select="$sublist">
      <!-- OPTIMIZE: is there a way to store the result for repetive use? -->
      <xsl:choose>
        <xsl:when test="$symbol='or'">  <xsl:text> | </xsl:text></xsl:when>
        <xsl:when test="$symbol='nor'"> <xsl:text> ~| </xsl:text></xsl:when>
        <xsl:when test="$symbol='and'">
	  <xsl:text disable-output-escaping="yes"><![CDATA[ & ]]></xsl:text></xsl:when>
        <xsl:when test="$symbol='impl'">
	  <xsl:text disable-output-escaping="yes"><![CDATA[ => ]]></xsl:text></xsl:when>
        <xsl:when test="$symbol='rimpl'">
	  <xsl:text disable-output-escaping="yes"><![CDATA[ <= ]]></xsl:text></xsl:when>
        <xsl:when test="$symbol='equiv'">
	  <xsl:text disable-output-escaping="yes"><![CDATA[ <=> ]]></xsl:text></xsl:when>
        <xsl:when test="$symbol='xor'">
	  <xsl:text disable-output-escaping="yes"><![CDATA[ <~> ]]></xsl:text></xsl:when>
        <xsl:when test="$symbol='nand'">
	  <xsl:text disable-output-escaping="yes"><![CDATA[ ~& ]]></xsl:text></xsl:when>
        <xsl:otherwise>
          <xsl:message terminate="yes">
            Unknown connective '<xsl:value-of select="$symbol" />'
          </xsl:message>
        </xsl:otherwise>
      </xsl:choose>
      <xsl:apply-templates select="." />
    </xsl:for-each>
  </xsl:template>

 <xsl:template match="xtstp:conn">
   <xsl:text>(</xsl:text>
   <xsl:apply-templates select="*[1]" />
   <xsl:call-template name="connective">
     <xsl:with-param name="symbol" select="@symbol"/>
     <xsl:with-param name="sublist" select="*[position()>1]"/>
   </xsl:call-template>
   <!-- xsl:for-each select="*[position()>1]">
     <xsl:value-of select="../@symbol" />
     <xsl:apply-templates/>
   </xsl:for-each -->
   <xsl:text>)</xsl:text>
 </xsl:template>

  <xsl:template match="xtstp:pred|xtstp:func">
    <xsl:value-of select="@name" />
    <xsl:if test="*">
      <xsl:text>(</xsl:text>
      <xsl:call-template name="list">
        <xsl:with-param name="separator" select="','"/>
        <xsl:with-param name="sublist" select="*"/>
      </xsl:call-template>
      <xsl:text>)</xsl:text>
    </xsl:if>
  </xsl:template>

  <xsl:template match="xtstp:equal">
    <xsl:text>equal(</xsl:text>
    <xsl:apply-templates select="*[1]" />
    <xsl:text>,</xsl:text>
    <xsl:apply-templates select="*[2]" />
    <xsl:text>)</xsl:text>
  </xsl:template>
  
  <xsl:template match="xtstp:not">
    <xsl:text> ~ </xsl:text>
    <xsl:apply-templates/>
  </xsl:template>
  <xsl:template match="xtstp:true">
    <xsl:text>true</xsl:text>
  </xsl:template>
  <xsl:template match="xtstp:false">
    <xsl:text>false</xsl:text>
  </xsl:template>

  <xsl:template match="xtstp:var">
    <xsl:value-of select="@name" />
    <xsl:if test="@type!=''">
      <xsl:text>:</xsl:text>
      <xsl:value-of select="@type"/>
    </xsl:if>
    <!-- TODO: strange number -->
  </xsl:template>
  
  <xsl:template match="xtstp:number">
    <xsl:value-of select="." />
  </xsl:template>
  <xsl:template match="xtstp:qstring">
    <xsl:call-template name="qstring">
      <xsl:with-param name="value" select="."/>
    </xsl:call-template>
  </xsl:template>
  

  <!-- TODO: newlines in headers and comments -->
  <xsl:template match="xtstp:header">
    <xsl:text>%</xsl:text>
    <xsl:if test="@field!=''">
      <xsl:text> </xsl:text>
      <xsl:value-of select="@field"/>
      <xsl:value-of select="substring('            ',string-length(@field))"/>
      <xsl:text> : </xsl:text>
    </xsl:if>
    <xsl:call-template name="replace-string">
     <xsl:with-param name="text" select="text()"/>
     <xsl:with-param name="replace" select="'&#xa;'"/>
     <xsl:with-param name="with" select="'&#xa;%                 '"/>
    </xsl:call-template>
    <xsl:text>&#xa;</xsl:text>
    <xsl:if test="following-sibling::comment[position() = 1]">
      <xsl:text>% Comment         </xsl:text>
    </xsl:if>
  </xsl:template>

  <xsl:template match="xtstp:comment">
    <xsl:text>%</xsl:text>
    <xsl:call-template name="replace-string">
     <xsl:with-param name="text" select="text()"/>
     <xsl:with-param name="replace" select="'&#xa;'"/>
     <xsl:with-param name="with" select="'&#xa;%'"/>
    </xsl:call-template>
    <xsl:text>&#xa;</xsl:text>
  </xsl:template>
  
  
  <!-- TODO: info -->
    <!--
  <xsl:template match="xtstp:info">
    <xsl:text>,</xsl:text>
    <xsl:if test="not(preceding-sibling::source[position()=1])"> 
      <xsl:text>unknown,</xsl:text>
    </xsl:if>
  </xsl:template>
    -->
</xsl:stylesheet>
